package abcm.cloning;

import java.util.HashMap;


/**
 * A Clone Store is a repository of copied objects,
 * indexed by the original objects. It is used when a collection
 * of objects must be copied, but some of the objects within the collection
 * refer to other objects within the collection.
 * 
 * If an object contains only raw data instance variables, making a copy is
 * easy. However, if the class contains instance variables that refer to objects,
 * then a 'deep clone' of the main object implies a new copy of the objects
 * it contains. This is initially simple, however there is a complication
 * because a single object can be contained by more than one main object.
 * (Other reticulations are also possible, so that objects can contain each
 * other, etc.). In these cases, the 'deep clone' of the entire set of objects
 * should contain only one new instance of each object.
 * 
 * The Clone Store contains an indexed list of new copies of each object. In general,
 * when an object tries to make a copy of the instance objects it contains, it should
 * look first in the Clone Store to see if a copy of that object has already been created.
 * If so, it should use the pre-existing copy; if not, it can go ahead and make a new copy.
 * 
 * In fact, this process is already built-in to the Clone Store: requesting a new copy of
 * an object instructs the clone store to check for a copy, and if none is found to create a
 * new one. The appropriate copy, whether newly created or not, is returned.
 * 
 * As one additional point of usage: an object making a deep clone of itself should
 * begin by addint ITSELF to the clone store. Then when related objects make copies of
 * themselves, they will find the original object's copy already available.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 *
 */
public class ABCMCloneStore extends HashMap<ABCMDeepCloneable, ABCMDeepCloneable>{

  /** Serial ID  */
  private static final long serialVersionUID = 1189819950761418727L;

  /**
   * Checks for a pre-existing clone of the specified
   * object; if none exists, one is created. The clone
   * of the object is then returned.
   * @param original original deep cloneable object
   * @return deep clone of that object.
   */
  public ABCMDeepCloneable getClone(ABCMDeepCloneable original){
    if(original == null) return null;
    if(containsKey(original)) return get(original);
    ABCMDeepCloneable clone = original.getDeepClone(this);
    put(original, clone);
    return clone;
  }
  
}
