package abcm.db;

import java.sql.Connection;
import java.sql.Statement;
import java.util.Hashtable;

import abcm.db.components.ABCMdbDatum;
import abcm.db.managers.TruthValueManager;
import abcm.math.BasicRNG;
import abcm.math.SimDate;
import abcm.model.AbstractABCMParser;
import abcm.sim.SimulationRun;
import abcm.utilities.QuotedLineParser;
import abcm.utilities.ScriptReader;

/**
 * ABCMdbDataInitializer is a base class that creates the core of an ABCM database.
 * Extensions of this class can be created to populate a database with specific
 * data or extend the structure of the database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ABCMdbDataInitializer extends AbstractDBDataInitializer {


  /* Hashtables used to make some PK values available to descendant classes */
  /** Hashtable of commands added and their PKs; makes these available to descendant classes */
  protected Hashtable<String, Long> commandPKs            = new Hashtable<String, Long>();
  
  /** Hashtable of code model domain values added and their PKs; makes these available to descendant classes */
  protected Hashtable<String, Long> codeModelDomainValues = new Hashtable<String, Long>();
  
  /** Hashtable of code model option values added and their PKs; makes these available to descendant classes */
  protected Hashtable<String, Long> codeModelOptionValues = new Hashtable<String, Long>();
  
  /** Hashtable of parameter values added and their PKs; makes these available to descendant classes */
  protected Hashtable<String, Long> parameterValues       = new Hashtable<String, Long>();

  
  /** DB Gateway to be used to access the database  */
  protected AbstractABCMDBGateway DB;
  
  /** Connection to be used when inserting data. Note that all data is inserted in one transaction and can be rolled back.  */
  protected Connection con;
  
  
  /** PK assigned to Command Category 'Help'  */
  protected long newPK_Category_Help;
  /** PK assigned to Command Category 'Model Operation'  */
  protected long newPK_Category_ModelOperation;
  /** PK assigned to Command Category 'Space'  */
  protected long newPK_Category_Space;
  /** PK assigned to Command Category 'Time'  */
  protected long newPK_Category_Time; 
  /** PK assigned to Command Category 'Graphs'  */
  protected long newPK_Category_Graphs;  
  /** PK assigned to Command Category 'Interrogative'  */
  protected long newPK_Category_Interrogative;  
  /** PK assigned to Command Category 'Probes and Data'  */
  protected long newPK_Category_ProbesAndData; 
  /** PK assigned to Command Category 'Parameters'  */
  protected long newPK_Category_Parameters;  
  /** PK assigned to Command Category 'File Management'  */
  protected long newPK_Category_FileManagement;  
  /** PK assigned to Command Category 'View'  */
  protected long newPK_Category_View;
  /** PK assigned to Command Category 'Script Processing'  */
  protected long newPK_Category_ScriptProcessing;  
  /** PK assigned to Command Category 'Interface'  */
  protected long newPK_Category_Interface;
  /** PK assigned to Command Category 'DB'  */
  protected long newPK_Category_DB;
  /** PK assigned to Command Category 'Actors' */
  protected long newPK_Category_Actors;
  /** PK assigned to Command Category 'Not Yet Classified'  */
  protected long newPK_Category_NotYetClassified;

  /** PK Assigned to the automatically generated user representing the system  */
  protected long systemUserPK;
  /** PK Assigned to the default Code Model  */
  protected long defaultCodeModelPK;
  
  /** PK Assigned to the System's parameters category  */
  protected long systemParametersCategory;
  
  /** PK Assigned to the reference type 'Book'  */
  protected long newPK_ReferenceType_Book;
  /** PK Assigned to the reference type 'Chapter'  */
  protected long newPK_ReferenceType_Chapter;
  /** PK Assigned to the reference type 'Article'  */
  protected long newPK_ReferenceType_Article;  
  /** PK Assigned to the reference type 'Manuscript'  */
  protected long newPK_ReferenceType_Manuscript;
  /** PK Assigned to the reference type 'Report'  */
  protected long newPK_ReferenceType_Report;
  /** PK Assigned to the reference type 'Other'  */
  protected long newPK_ReferenceType_Other;
  
  /** PK Assigned to 'Create Population' system narrative */ 
  protected long createPopulationNarrative;
  
  /** PK Assigned to 'Stop' narrative */
  protected long stopNarrative;
  
  /**
   * Equals a String[] array with no elements
   */
  protected static final String[] NO_ARGS = new String[] { };
  
  /**
   * Constructor
   * @param db gateway to the database server
   */
  public ABCMdbDataInitializer(AbstractABCMDBGateway db){
    DB = db;
  }
  
  /**
   * Adds the specified datum to the database
   * @param datum datum to be added
   * @return the new PK of the datum entered
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long add(ABCMdbDatum datum) throws Exception{
    return DB.addDatum(con, datum);
  }

  /* General */
  
  /**
   * Adds a term to the ABCM glossary being createds
   * @param term Term being added
   * @param anchor Abbreviation of that term to be used as the HTML anchor
   * @param definition Definition of the term
   * @return the PK of the term added
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addTerm(String term, String anchor, String definition) throws Exception{
    System.out.println("Adding term: " + term);
    return add(DB.getTermsTable().getDatum(term, anchor, definition));
  }
  
  
  /**
   * Adds a comment to the ABCM database
   * @param userPK      the PK of the user to whom the comment will be attributed
   * @param comment     the comment text
   * @return PK of the comment as added
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addComment(long userPK, String comment)throws Exception{
    System.out.println("Adding comment... " + userPK + " " + comment.substring(0, 20) + "...");
    return add(DB.getCommentsTable().getDatum(userPK, DB.currentTimeString(), comment));
  }

  /* Commands */
  
  /**
   * Adds a category to the list of command categories.
   * @param order Order in which category is to be sorted in the list of categories. Care must be taken to avoid duplicates.
   * @param category Name of the category being added
   * @param description Description of the category being added
   * @return PK of the category added
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addCommandCategory(int order, String category, String description) throws Exception{
    System.out.println("Adding command category: " + category);
    return add(DB.getCommandCategoriesTable().getDatum(category, description, order));
  }
  
  /**
   * Adds a command to the list of commands.
   * @param category PK of the category to which this command is to be added. Must be an extant category.
   * @param commandName Name of the command (should include only characters, not spaces or punctuation)
   * @param definition Definition of the command (briefly, what it does)
   * @param commandLineOnly True if the command can only be executed in interactive mode; false otherwise
   * @param probe True if the command can be used by a probe (i.e. returns a value); false if it should not be usable by a probe
   * @param description Description of the action of the command
   * @param argDefs Array of strings defining the names of the arguments this command will take. The length of this array gives th
   * number of arguments.
   * @throws Exception generic exception thrown if any problem occurs
   */
  public void addCommand(long category, String commandName, String definition, boolean commandLineOnly, boolean probe, 
      String description, String[] argDefs) throws Exception{
    System.out.println("Adding Command: " + commandName);
    long comm = add(DB.getNarrativeCommandsTable().getDatum(commandName, argDefs.length, description, 
        commandLineOnly, probe, definition, category));
    commandPKs.put(commandName.toUpperCase(), comm);
    for(int i = 0; i < argDefs.length; i++) add(DB.getNarrativeArgumentsTable().getDatum(comm, i + 1, argDefs[i]));
  }
  
  /**
   * Adds a command to the list of commands; the command being added has no arguments.
   * @param category PK of the category to which this command is to be added. Must be an extant category.
   * @param commandName Name of the command (should include only characters, not spaces or punctuation)
   * @param definition Definition of the command (briefly, what it does)
   * @param commandLineOnly True if the command can only be executed in interactive mode; false otherwise
   * @param probe True if the command can be used by a probe (i.e. returns a value); false if it should not be usable by a probe
   * @param description Description of the action of the command
   * @throws Exception generic exception thrown if any problem occurs
   */
  public void addCommand(long category, String commandName, String definition, boolean commandLineOnly, boolean probe, 
      String description) throws Exception{
    addCommand(category, commandName, definition, commandLineOnly, probe, description, NO_ARGS);
  }
  
  /* Users */
  
  /**
   * Adds a user to the ABCM database being created
   * @param name          user's name in normal format (i.e. John Murphy)
   * @param lastFirst     user's name in Last, First format (i.e. Murphy, John)
   * @param desc          description of the user
   * @return the PK of the user as added to the db
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addUser(String name, String lastFirst, String desc) throws Exception{
    System.out.println("Adding user: " + lastFirst);
    return add(DB.getResearchersTable().getDatum(name, lastFirst, desc));
  }

  
  /**
   * Adds a project user. This adds the user to the basic 'researchers' table, but also to the 'project researchers' table,
   * so that they can use the advanced functions of the model.
   * @param name User's name in normal format (i.e. John Murphy)
   * @param lastFirst User's name in Last, First format (i.e. Murphy, John)
   * @param desc Description of the user
   * @param userName User Name to be used when logging in to an ABCM system
   * @param password Password to be used when logging in to an ABCM system
   * @param admin True if the user is an administrator; false otherwise
   * @return the PK of the user as added to the DB (from the basic 'researchers' table, not the project researcher ID)
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addProjectUser(String name, String lastFirst, String desc, String userName, String password, boolean admin) throws Exception{
    long userPK = addUser(name, lastFirst, desc);
    System.out.println("Adding project user: " + userName);
    add(DB.getProjectResearchersTable().getDatum(userPK, userName, password, admin));
    return userPK;
  }
    

  
  
  /* Simulation */
  
  
  /**
   * Adds a simulation to the database
   * @param name           the name of the simulation to be added
   * @param codeModelFK    PK of the Code Model contained by this simulation
   * @param dataModelFK    PK of the Data Model contained by this simultion
   * @param researcherFK   PK of the Researcher who created this simulation
   * @param description    description of the simulation
   * @return the PK of the Simulation entry in the db
   * @throws Exception generic exception thrown if any problem arises
   */
  public long addSimulation(String name, long codeModelFK, long dataModelFK, long researcherFK, String description) throws Exception{
    System.out.println("Adding simulation " + name);
    return add(DB.getSimulationsTable().getDatum(name, codeModelFK, dataModelFK, DB.currentTimeString(), researcherFK, description));
  }
  
  
  /* Data Model */
  
  /**
   * Adds a Data Model to the database
   * @param name              name of the Data Model to be stored
   * @param description       description of the Data Model to be stored
   * @param historyFK         PK of the History associated with this Data Model
   * @param configurationFK   PK of the Configuration associated with this Data Model
   * @param researcherFK      PK of the Researcher who created this Data Model
   * @return                  the PK of the Data Model as stored in the database
   * @throws Exception generic exception thrown if any problem arises
   */
  public long addDataModel(String name, String description, long historyFK, long configurationFK, long researcherFK) throws Exception{
    System.out.println("Adding data model " + name);
    return add(DB.getDataModelsTable().getDatum(name, description, historyFK, configurationFK, DB.currentTimeString(), researcherFK));
  }
  
  
  /* Configurations */
  /**
   * Adds a Configuration to the database
   * @param configName         name of the Configuration to be stored
   * @param description        description of the Configuration to be stored
   * @param researcherFK       PK of Researcher who created this Configuration 
   * @return the new PK of the Configuration
   * @throws Exception generic exception thrown if any problem arises
   */
  public long addConfiguration(String configName, String description, long researcherFK) throws Exception{
    System.out.println("Adding configuration " + configName);
    return add(DB.getSimulationConfigurationsTable().getDatum(configName, description, DB.currentTimeString(), researcherFK));
  }
  
  
  /* Histories and Narratives */
  
  /**
   * Adds a History to the database
   * @param historyName     name of the History to be stored
   * @param description     description of the History to be stored
   * @param truthValueFK    PK of the Truth Value this History is to be assigned
   * @param researcherFK    PK of the Researcher who created this History
   * @return the PK of the History as stored in the database
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addHistory(String historyName, String description, long truthValueFK, long researcherFK) throws Exception{
    System.out.println("Adding history " + historyName);
    return add(DB.getHistoriesTable().getDatum(historyName, description, truthValueFK, researcherFK));
  }
  
  /**
   * Adds a Narrative to the database. Adds only the entry into the Narratives table, not the instructions.
   * @param name           name of the Narrative to be stored
   * @param description    description of the Narrative to be stored
   * @param researcherFK   PK of the Researcher who created this Narrative
   * @param baseDate       base date from which the dates of the Narrative instructions are reckoned
   * @param isSeasonal     true if the Narrative is a Seasonal Narrative, false otherwise
   * @return the PK of the Narrative as stored in the database
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addNarrative(String name, String description, long researcherFK, String baseDate, boolean isSeasonal) throws Exception{
    System.out.println("Adding narrative " + name);
    return add(DB.getNarrativesTable().getDatum(name, description, researcherFK, SimDate.parseString(baseDate), DB.currentTimeString(), isSeasonal));
  }
  
  /**
   * Adds a Narrative to the database, first as an entry in the Narratives table, but then also all Narrative Instructions.
   * Assumes that the file name given will point to a valid Narrative file (uncompiled).
   * @param name          name of the Narrative to be stored
   * @param fileName      name of the file with the instructions to be stored; name should exclude path and extension
   * @param directory     directory in which narrative file is found
   * @param extension     extension by which narrative file is identified
   * @param description   description of the Narrative to be stroed
   * @param researcherFK  PK of the Researcher who created this Narrative
   * @param baseDate      base date from which the dates of the Narrative instructions are reckoned
   * @param isSeasonal    true if the Narrative is a Seasonal Narrative, false otherwise
   * @return the PK of the Narrative ast stored in the database
   * @throws Exception generic exception thrown if any problem  occurs
   */
  public long addNarrativeWithInstructions(String name, String fileName, String directory, String extension, String description, long researcherFK, 
      String baseDate, boolean isSeasonal) throws Exception{
    long narrID = addNarrative(name, description, researcherFK, baseDate, isSeasonal);
    addNarrativeInstructionsFromFile(fileName, directory, extension, narrID);
    return narrID;
  }
  
  /**
   * Adds the instructions found in the specified file to the Narrative identified by the given Narrative PK
   * @param narrativeFileName name of the file with the instructions to be stored; name should exclude path and extension
   * @param directory         directory in which narrative file is found
   * @param extension         extension by which narrative file is identified
   * @param narrativeFK       PK of the Narrative for which the instructions are to be stored
   * @throws Exception generic exception thrown if any problem occurs
   */
  public void addNarrativeInstructionsFromFile(String narrativeFileName, String directory, String extension, long narrativeFK) throws Exception{
    ScriptReader sp = new ScriptReader(directory + narrativeFileName + extension);
    
    Statement st = con.createStatement();  
    int line = 1;
    int order = 0;
    double dateVal = -1;
    System.err.println("Processing line: " + line);
    while(!sp.EOF()){
      String nextLine = sp.getNextLine();
      
      if(nextLine != null && nextLine.length() > 0){
        nextLine = nextLine.replaceAll("\t"," ");
        String[] inputLine = QuotedLineParser.parseQuoted(nextLine, AbstractABCMParser.MAX_ARGS + 2);
        
        int x = inputLine.length-1;
        while( (inputLine[x].length() == 0) && (x >= 0))  x--;
        int argLength = x - 1;
        String[] newArgs = new String[argLength];
        for(int y = 0; y < argLength; y++){
          newArgs[y] = inputLine[y + 2];
        }
        double commandDate = SimDate.parseString(inputLine[0]);
        if(commandDate != dateVal){
          order = 1;
          dateVal = commandDate;
        }
        else{
          order++;
        }
        if(commandPKs.get(inputLine[1].toUpperCase()) == null){
          System.out.println("NULL COMMAND PK: '" + inputLine[1] + "'");
          System.out.println(commandPKs.containsKey(inputLine[1]));
        }
        
        ABCMdbDatum d = DB.getNarrativeInstructionsTable().getDatum(
            "" + narrativeFK,
            "" + (SimDate.parseString(inputLine[0])),               // Date 
            "" + order,                                             // Order
            "" + commandPKs.get(inputLine[1].toUpperCase()),        // Command ID  
            (newArgs.length > 0 ? newArgs[0] : ""), 
            (newArgs.length > 1 ? newArgs[1] : ""), 
            (newArgs.length > 2 ? newArgs[2] : ""),
            (newArgs.length > 3 ? newArgs[3] : ""),
            (newArgs.length > 4 ? newArgs[4] : ""),
            (newArgs.length > 5 ? newArgs[5] : ""),
            (newArgs.length > 6 ? newArgs[6] : ""),
            (newArgs.length > 7 ? newArgs[7] : ""),
            (newArgs.length > 8 ? newArgs[8] : ""),
            (newArgs.length > 9 ? newArgs[9] : ""),
            "");
        st.execute(d.getInsertSQL(DB.getOwner(), DB.getSQLSyntaxEngine()));
      }
      line++;
      if(line % 50 == 0) System.out.println("   Processing line: " + line);
    }
    sp.closeFile();
    st.close();
  }
  
  /**
   * Adds the specified Narrative as a component of the specified History.
   * @param historyFK       The PK of the History to which the Narrative is to be associated
   * @param narrativeFK     The PK of the Narrative to be added
   * @param baseDate        base date from which the Narrative instructions are to be reckoned in this History
   * @param order           order to be assigned to this Narrative; note that no error checking occurs here, so the calling program
   *    is responsible for ensuring that Narratives assigned to a History fall in proper order with no duplicates (not permitted by db
   *    engine) and no gaps in the sequence (not otherwise enforced).
   * @return the PK of the entry created in the History Narratives table
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addNarrativeToHistory(long historyFK, long narrativeFK, String baseDate, int order) throws Exception{
    System.out.println("Adding narrative " + narrativeFK + " to history " + historyFK);
    return add(DB.getHistoryNarrativesTable().getDatum(historyFK, narrativeFK, SimDate.parseString(baseDate), order));
  }
  
  /**
   * Adds an instruction to a Narrative in the database
   * @param narrativeFK  PK of the Narrative for which the instruction is to be added
   * @param date         simulation date on which the instruction will be executed (relative to the stored base date of the Narrative), as a double
   * @param order        order in which this instruction is to be executed on this date for this Narrative; note that no error
   *    checking is done here, so it is the responsibility of the calling program to ensure that order numbers are non-duplicates 
   *    (enforced by db engine) and without gaps in the sequence (not otherwise enforced).
   * @param commandPK    PK of the command this instruction executes
   * @param arg1         argument 1 to the command
   * @param arg2         argument 2 to the command
   * @param arg3         argument 3 to the command
   * @param arg4         argument 4 to the command
   * @param arg5         argument 5 to the command
   * @param arg6         argument 6 to the command
   * @param arg7         argument 7 to the command
   * @param arg8         argument 8 to the command
   * @param arg9         argument 9 to the command
   * @param arg10        argument 10 to the command
   * @param comment      any comment about this instruction
   * @return PK of the entry in the Narrative Instructions table
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addNarrativeInstruction(long narrativeFK, double date, int order, long commandPK, String arg1,
      String arg2, String arg3, String arg4, String arg5, String arg6, String arg7, String arg8, String arg9, String arg10, String comment) throws Exception{
    return add(DB.getNarrativeInstructionsTable().getDatum(narrativeFK, date, order, 
        commandPK, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, comment));
  }
  
  /**
   * Adds an instruction to a Narrative in the database; uses String format for command and date. If the date cannot be converted to a proper
   * date no error will occur (date stored will be -1), but if the command is not in the commands array an exception (ArrayIndexOutOfBounds) will
   * be thrown.
   * @param narrativeFK  PK of the Narrative for which the instruction is to be added
   * @param date         simulation date on which the instruction will be executed (relative to the stored base date of the Narrative), as a String
   * @param order        order in which this instruction is to be executed on this date for this Narrative; note that no error
   *    checking is done here, so it is the responsibility of the calling program to ensure that order numbers are non-duplicates 
   *    (enforced by db engine) and without gaps in the sequence (not otherwise enforced).
   * @param command      command to be executed, as a String
   * @param arg1         argument 1 to the command
   * @param arg2         argument 2 to the command
   * @param arg3         argument 3 to the command
   * @param arg4         argument 4 to the command
   * @param arg5         argument 5 to the command
   * @param arg6         argument 6 to the command
   * @param arg7         argument 7 to the command
   * @param arg8         argument 8 to the command
   * @param arg9         argument 9 to the command
   * @param arg10        argument 10 to the command
   * @param comment      any comment about this instruction
   * @return PK of the entry in the Narrative Instructions table
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addNarrativeInstruction(long narrativeFK, String date, int order, String command, String arg1,
      String arg2, String arg3, String arg4, String arg5, String arg6, String arg7, String arg8, String arg9, String arg10, String comment) throws Exception{
    if(!this.commandPKs.containsKey(command.toUpperCase())) throw new ArrayIndexOutOfBoundsException();
    return addNarrativeInstruction(narrativeFK, SimDate.parseString(date), order, 
        commandPKs.get(command.toUpperCase()), arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, comment);
  }
  
  
  /* Code Models */
  
  /**
   * Adds a code model domain to the ABCM database
   * @param domainName The name of the domain to be added
   * @param description Description of the domain
   * @param ID ID to be used when specifying this domain
   * @return The PK of the domain added
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addCodeModelDomain(String domainName, String description, String ID) throws Exception{
    System.out.print("Adding code model domain: " + domainName + " ID is " + ID);
    long val = add(DB.getCodeModelOptionDomainsTable().getDatum(domainName, description, ID));
    System.out.println(" val is " + val);
    codeModelDomainValues.put(ID, val);
    return val;
  }
  
  /**
   * Adds a code model domain option value to the ABCM database.
   * @param manualPK Primary Key to be used for this object; these must be managed manually
   * @param optionDomainFK FK linking to the Domain of which this value is an option
   * @param value Value to be associated with this option; input manually rather than generated automatically
   * @param name Name of this domain option value
   * @param code Text of code associated with this option value
   * @param description Description of this option value
   * @return the new PK for this option value
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addCodeModelDomainOptionValue(long manualPK, long optionDomainFK, long value, String name, 
      String code, String description)throws Exception{
    System.out.print("Adding code model domain option values " + name);
    long val = add(DB.getCodeModelOptionValuesTable().getDatum(manualPK, optionDomainFK, value, name, code, description));
    System.out.println(" val is " + val + " but value is " + value);
    this.codeModelOptionValues.put(name, value);
    return val;
  }
  
  /**
   * Adds a new code model to the ABCM database
   * @param title         title for the code model
   * @param researcherFK  PK of researcher who is the owner of this code model
   * @param description   description of this code model
   * @return the new PK for this code model
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addCodeModel(String title, long researcherFK, String description) throws Exception{
    System.out.println("Adding code model " + title);
    return add(DB.getCodeModelsTable().getDatum(title, researcherFK, DB.currentTimeString(), description));
  }
  
  /**
   * Adds a code model/option value link to the database
   * @param codeModelFK Code Model to which this is to be added
   * @param optionDomainFK  Code Model Domain to be used
   * @param optionValueFK   PK of specific Code Model Option Value to be used
   * @return the PK of the newly added entry
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addCodeModelAndOptionValue(long codeModelFK, long optionDomainFK, long optionValueFK) throws Exception{
    System.out.println("Adding code model option value " + codeModelFK + " " + optionDomainFK + " " + optionValueFK);
    return add(DB.getCodeModelsAndOptionValuesTable().getDatum(codeModelFK, optionDomainFK, optionValueFK));
  }
  
  /**
   * Adds a code model/option value link to the database
   * @param codeModelFK Code Model to which this is to be added
   * @param optionDomain  Code Model Domain to be used, as a String. If the String is not found in the
   *   code model domain values array, an ArrayIndexOutOfBounds exception is thrown
   * @param optionValue   Code Model Option Value to be used, as a String. If the String is not found in
   *   the code model option values array, an ArrayIndexOutOfBounds exception is thrown
   * @return the PK of the newly added entry
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addCodeModelAndOptionValue(long codeModelFK, String optionDomain, String optionValue) throws Exception{
    System.out.println("Adding code model and option value: " + optionDomain + " " + codeModelDomainValues.get(optionDomain)+ " " + optionValue  + " " + codeModelOptionValues.get(optionValue));
    if(!this.codeModelDomainValues.containsKey(optionDomain) || !this.codeModelOptionValues.containsKey(optionValue)) throw new ArrayIndexOutOfBoundsException();
    return addCodeModelAndOptionValue(codeModelFK, codeModelDomainValues.get(optionDomain), codeModelOptionValues.get(optionValue));
  }
  
  /* Parameters and Parameter Sets */
  
  
  
  /**
   * Adds a parameter category to the ABCM database
   * @param name Name of the category to be created
   * @param description Description of the category
   * @return the new PK of the newly added parameter category
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addParameterCategory(String name, String description) throws Exception{
    System.out.println("Adding parameter category " + name);
    return add(DB.getParameterCategoriesTable().getDatum(name, description));
  }
  
  /**
   * Adds a parameter to the ABCM database
   * @param name Name of the parameter
   * @param description Description of the parameter
   * @param defaultValue Default value for the parameter
   * @param minimum Minimum value permissible for the parameter
   * @param maximum Maximum value permissible for the parameter
   * @param units Units in which the parameter is to be measured (i.e. Km/hour)
   * @param isEngine True if the parameter is used by the ABCM engine rather than a model component
   * @param parameterCategoryFK PK of the category to which this parameter belongs
   * @param order Order in which this parameter is to be listed
   * @return PK of the newly added parameter
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addParameter(String name, String description, 
      double defaultValue, double minimum, double maximum, String units, 
      boolean isEngine, long parameterCategoryFK, int order )throws Exception{
    System.out.println("Adding parameter " + name);
    long val = add(DB.getParametersTable().getDatum(name, description, DB.currentTimeString(), 
        defaultValue, minimum, maximum, units, 
        isEngine, parameterCategoryFK, order)); 
    parameterValues.put(name, val);
    return val;
  }
  
  /**
   * Adds a parameter set to the ABCM database.
   * @param name Name to be assigned to this parameter set. Must be unique
   * @param description         description of this parameter set
   * @param automatic           true if this is an automatically generated parameter set (generated as part of parameter sweep, perhaps)
   * @param researcherFK        PK of the researcher who 'owns' this parameter set
   * @return the PK of the new parameter set as added to the db
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addParameterSet(String name, String description, boolean automatic, long researcherFK) throws Exception{
    System.out.println("Adding parameter set " + name);
    return add(DB.getParameterSetsTable().getDatum(name, DB.currentTimeString(), automatic, description, researcherFK));
  }
  
  /**
   * Adds a parameter set value to the specified parameter set
   * @param parameterSet the PK of the parameter set to which the value is to be added
   * @param parameter the PK of the parameter to be added
   * @param val the value to be set for this parameter
   * @param truthValue FK to the truth value of this parameter value
   * @return the ID of the newly entered record in the parameter set values table
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addParameterSetValue(long parameterSet, long parameter, double val, int truthValue) throws Exception{
    System.out.println("Adding parameter set value " + parameterSet + " " + parameter + " " + val);
    return add(DB.getParameterSetsAndValuesTable().getDatum(parameterSet, parameter, val, truthValue));
  }
  
  /**
   * Adds a value to a Parameter Set
   * @param parameterSet  PK of the Parameter Set to which the value is to be added
   * @param parameter     name of the parameter to be added. If this is not in the parameter values array, an 
   *    ArrayIndexOutOfBounds exception will be thrown
   * @param val           value to be added
   * @param truthValue    OK of the Truth Value to be assigned to this parameter vale
   * @return the PK of the entry in the Parameter Set Values table
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addParameterSetValue(long parameterSet, String parameter, double val, int truthValue) throws Exception{
    if(!this.parameterValues.containsKey(parameter)) throw new ArrayIndexOutOfBoundsException();
    return addParameterSetValue(parameterSet, parameterValues.get(parameter), val, truthValue);
  }
  
  /* Probes and Probe Sets */
  
  /**
   * Adds a new probe to the database.
   * In theory, an implementation can add probes and probe sets at db creation time;
   * in practice this is usually not used, because probes and probe sets
   * are more specific than the lowest level of abstraction implemented
   * at db creation time.   
   * @param name Name for the probe
   * @param description Description of the probe
   * @param researcherFK PK of the researcher who 'owns' the probe
   * @param dataSeriesName Name for the data series to which this probe sends its data
   * @param dataSeriesDomain Name of the data series domain to which this probe sends its data
   * @param isNumeric True if the probe returns a numeric value, false otherwise
   * @param isMonthly True if the probe is executed monthly, false otherwise
   * @param baseDate Base Date on which the probe is executed
   * @param frequency Frequency at which the probe is executed offset from the base date (i.e. every 365 days)
   * @param retrieve Command string the probe executes to obtain the value it returns
   * @param offset Additional double value indicating the 'offset' from midnight of the date of execution (useful in retrieving
   * stage data by hour, for example)
   * @return the PK of the probe as added
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addProbe(String name, String description, long researcherFK, String dataSeriesName, String dataSeriesDomain, boolean isNumeric,
      boolean isMonthly, SimDate baseDate, int frequency, String retrieve, double offset) throws Exception{
    System.out.println("Adding probe " + name);
    return add(DB.getProbesTable().getDatum(name, description, DB.currentTimeString(), researcherFK, 
        dataSeriesName, dataSeriesDomain, isNumeric, isMonthly, baseDate, frequency, retrieve, offset));
  }
  
  /**
   * Adds a new probe set to this database
   * @param name Name for the new probe set
   * @param researcherFK PK of researcher who is considered to be the owner of this probe
   * @param description Description of this probe
   * @return new PK of the probe as added to this database
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addProbeSet(String name, long researcherFK, String description) throws Exception{
    System.out.println("Adding probe set: " + name);
    return add(DB.getProbeSetsTable().getDatum(name, DB.currentTimeString(), researcherFK, description));
  }
  
  /**
   * Adds a probe to a probe set
   * @param probeSetPK The PK of the probe set to which this probe is being added
   * @param probePK the PK of the probe being added
   * @return the new PK of the ProbeSetProbe entered
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addProbeToProbeSet(long probeSetPK, long probePK) throws Exception{
    System.out.println("Adding probe " + probePK + " to probe set ");
    return add(DB.getProbeSetProbesTable().getDatum(probeSetPK, probePK));
  }
  
  /**
   * Adds an entry to the list of valid types of reference
   * @param type Name for the new type of reference
   * @return the new PK for this type of reference
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addReferenceType(String type) throws Exception{
    System.out.println("Adding reference type " + type);
    return add(DB.getReferenceWorkTypesTable().getDatum(type));
  }

  
  /**
   * Adds a reference to the ABCM database
   * @param title Reference title
   * @param referenceDate Date; can be a String value in any form
   * @param summary Summary of the reference
   * @param abstractText Abstract
   * @param referenceTypeFK Link to the 'Reference Type' table indicating the kind of reference to be added (i.e. book, article)
   * @param researcherFK PK of the researcher who is considered to have added this reference (not the author)
   * @param location Text describing the location of this reference
   * @return the new PK of this reference
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addReference(String title, String referenceDate, String summary, String abstractText, long referenceTypeFK, long researcherFK, String location) throws Exception{
    System.out.println("Adding reference '" + title + "'");
    return add(DB.getReferencesTable().getDatum(title, referenceDate, summary, abstractText, referenceTypeFK, researcherFK, location));
  }
  
  /**
   * Adds an author to a reference in the ABCM db
   * @param referenceFK PK of the reference
   * @param authorOrder numeric order in which this author is to be listed. Note that these must be managed manually in this routine; duplicates
   * and gaps in ordering should be avoided
   * @param researcherFK PK of the researcher who is to be considered the author
   * @return PK of the new ref + author entry
   * @throws Exception generic exception thrown if any problem occurs
   */
  public long addReferenceAuthor(long referenceFK, int authorOrder, long researcherFK)throws Exception{
    System.out.println("Adding author " + researcherFK + " to reference " + referenceFK + " at position " + authorOrder);
    return add(DB.getReferencesAndAuthorsTable().getDatum(referenceFK, authorOrder, researcherFK));
  }  
  
  
  
  
  /* Initialize */
  
  /** 
   * @see abcm.db.AbstractDBDataInitializer#initialize(java.sql.Connection) 
   */
  public void initialize(Connection con) throws Exception{
    this.con = con;
    
    
    //   RunModesTable
    add(DB.getRunModesTable().getDatum(SimulationRun.GUI_MODE,                 "GUI Mode"        ));    
    add(DB.getRunModesTable().getDatum(SimulationRun.BATCH_MODE,               "Batch Mode"      ));
    add(DB.getRunModesTable().getDatum(SimulationRun.COMMAND_LINE_MODE,        "Command Line"    ));
    add(DB.getRunModesTable().getDatum(SimulationRun.ONLINE_INTERACTIVE_MODE,  "Web Interactive" ));
    add(DB.getRunModesTable().getDatum(SimulationRun.SCRIPT_MODE,              "Script Mode"     ));
    add(DB.getRunModesTable().getDatum(SimulationRun.SERVER_MODE,              "Server Mode"     ));
    
    //  TruthValuesTable
    add(DB.getTruthValuesTable().getDatum(TruthValueManager.CERTAIN,       "Absolutely certain", "Certain"       ));
    add(DB.getTruthValuesTable().getDatum(TruthValueManager.RECONSTRUCTED, "Reconstructed",      "Reconstructed" ));
    add(DB.getTruthValuesTable().getDatum(TruthValueManager.INFERRED,      "Inferred",           "Inferred"      ));
    add(DB.getTruthValuesTable().getDatum(TruthValueManager.LIKELY,        "Likely",             "Likely"        ));
    add(DB.getTruthValuesTable().getDatum(TruthValueManager.POSSIBLE,      "Possible",           "Possible"      ));
    add(DB.getTruthValuesTable().getDatum(TruthValueManager.IMPLAUSIBLE,   "Implausible",        "Implausible"   ));
    add(DB.getTruthValuesTable().getDatum(TruthValueManager.IMPOSSIBLE,    "Impossible",         "Impossible"    ));
    add(DB.getTruthValuesTable().getDatum(TruthValueManager.DEBUGGING,     "Debugging",          "Debugging"     ));
    
    //  ReferenceWorkTypesTable
    // Adding basic reference types
    newPK_ReferenceType_Book       = addReferenceType("Book");
    newPK_ReferenceType_Manuscript = addReferenceType("Manuscript");
    newPK_ReferenceType_Article    = addReferenceType("Article");
    newPK_ReferenceType_Chapter    = addReferenceType("Chapter");
    newPK_ReferenceType_Report     = addReferenceType("Report");
    newPK_ReferenceType_Other      = addReferenceType("Other");
    
    
    
    // User (System)
    systemUserPK = addUser("SYSTEM", "SYSTEM", "System");
    
    //  TermsTable
    addTerm("Base Date ",               "BaseDate",   "Narratives are created using absolute dates; the `start date` for a narrative is termed the narrative`s <em>base date</em>. When a narrative is implemented in the simulation, a new `base date` may be specified; this adjusts all of the dates in the narrati");
    addTerm("Configuration",            "config",     "A collection of values, as for a given variety of plant, that will persist throughout the simulation without change.");
    addTerm("Narrative",                "Narrative",  "A Narrative is a sequence of changes to a set of variables that occur at specified times during the simulation run.");
    addTerm("Code Model",               "CodeModel",  "The collection of choices, from the suite of optional algorithms, that makes up the simulation code for a given run. Code Models can be stored for re-use. A Code Model is the dynamic complement to a simulation`s Data Model.");
    addTerm("Parameter",                "Parameter",  "A Parameter is a special kind of value that can be varied from one simulation run to another. Parameters are akin to static data and thus could reasonably be considered part of the data model, but have been grouped separately for convenience and to allow ");
    addTerm("Interrogative",            "Intergtv",   "Most simulation commands returns messages that indicate their success or failure; in contrast, interrogatives return numerical values. These can be used in probes to create data series from which data graphs can be made.");
    addTerm("History",                  "history",    "A history is a combination of Narratives; `Narratives` can be stored individually, so each `streamflow` narrative and each `population` narrative can be separate. The user can choose population narrative and one streamflow narrative and combine them to ma");
    addTerm("Seasonal Narrative",       "SeasNarr",   "A narrative that covers a 365-day period and is repeated every year of the simulation. Used for plant cropping cyclesverting annual flow into daily values, etc.");
    addTerm("Command Line Interface",   "CommLine",   "A place to type instructions directly into the simulation and see how it responds.");
    addTerm("Data Model",               "DataModel",  "A Data Model includes the static data that will be used by the simulation. Typically this includes the Plant Configurations data and the History.");
    
    // Parameter Categories
    systemParametersCategory = addParameterCategory("System", "Parameters required by the ABCM system");
    
    // Parameters
    addParameter("RANDOMNUMBERGENERATORTYPE", "Type of Random Number Generator to be used", 
        BasicRNG.DEFAULT_JAVA_RNG, BasicRNG.DEFAULT_JAVA_RNG, BasicRNG.DEFAULT_JAVA_RNG, "none", 
        true, systemParametersCategory, 1);
    
    // ParameterSet
    long defaultParameterSetPK = addParameterSet("DEFAULT", "Default", true, systemUserPK);
    
    // Add comment to ParameterSet
    long paramSetComment = addComment(systemUserPK, "DEFAULT- Uses all default values; do not add elements");
    add(DB.getParameterSetCommentsTable().getDatum(defaultParameterSetPK, paramSetComment));
    
    //  CommandCategoriesTable, NarrativeCommandsTable and NarrativeArgumentsTable
    // Category: Help
    newPK_Category_Help             = addCommandCategory(      0,   "Help",               "Get help on using the HWM model");
  
      // Help commands
      addCommand(newPK_Category_Help, "Help",   "List all the commands the HWM simulator can use",       true, false, 
          "List the commands available.");
      addCommand(newPK_Category_Help, "Define", "Define a term from the HWM glossary",                   true, false, 
          "Get the definition of a term from the HWM glossary",                                    
          new String[] { "Term to be defined" } );
      addCommand(newPK_Category_Help, "HelpOn", "Get the instructions for a command in the HWM lexicon", true, false, 
          "Returns the instructions for a particular command from the HWM simulation environment", 
          new String[]{ "Command" } );
  
      
    newPK_Category_ModelOperation   = addCommandCategory(  10000,   "Model Operation",    "Creating a model, loading model components, or running the model");
      // Set Random Number Seed--- MISSING!
    
      addCommand(newPK_Category_ModelOperation, "LoadHistory",              "Loads a specific simulation history by name",                 false, false, 
          "Loads a specific history into the simulation. Can only be used on `test` simulations.", 
          new String[]{ "Name of history to be loaded" } );
      addCommand(newPK_Category_ModelOperation, "LoadConfiguration",        "Loads a specific Simulation Configuration by name",           true,  false, 
          "Load a specified configuration into the simulation.",                                   
          new String[]{  "Name of configuration to be loaded" } );
      addCommand(newPK_Category_ModelOperation, "SetCodeModelOption",       "Sets the specified code model option to the specified value", true,  false, 
          "Sets a specified code model option.",                                                  
          new String[]{  "Name of code model option to be set", "New value of code model option"  } );
      addCommand(newPK_Category_ModelOperation, "ViewConfigurationDetails", "View the details of the current configuration",               true,  false, 
          "View the details of the specified configuration.", 
          new String[]{  "Name of configuration to be reviewed"  } );
      addCommand(newPK_Category_ModelOperation, "Step",                     "Step the model forward one day",                              false, false, 
          "Steps the model forward one day.");
      addCommand(newPK_Category_ModelOperation, "StepDays",                 "Step the model forward the specified number of days",         false, false, 
          "Steps the model forward the specified number of days.", 
          new String[]{  "Number of days to step forward"  } );
      addCommand(newPK_Category_ModelOperation, "StepToDate",               "Step the model forward to the specified date",                false, false,
          "Steps the simulation forward to the specified date.", new String[]{ "Date to which the model should be stepped." });
      addCommand(newPK_Category_ModelOperation, "Stop",                     "Marks the end of a narrative.",                               false, false, 
          "Marks the end of a narrative (NO OP)");
      addCommand(newPK_Category_ModelOperation, "Finish",                   "Run the simulation through the end of its history",           true,  false, 
          "Runs the simulation from the current simulation date to one day after the last day of its history.");
      addCommand(newPK_Category_ModelOperation, "CreateModel",              "Create a model",                                              true,  false, 
          "Creates a model with the given simulation, parameter set, probe set, and notes", 
          new String[]{ "Name of the simulation for this model", "Name of the Parameter Set for this model", "Name of the Probe Set for this model", "Notes about this model" } );
      addCommand(newPK_Category_ModelOperation, "CreateTestModel",          "Create a test model",                                         true,  false, 
          "Creates a test model; results from a test model cannot be saved.");
      addCommand(newPK_Category_ModelOperation, "LoadSimulation",           "Load a simulation into a Model object.",                      true,  false, 
          "Loads a simulation into a Model object. Does not create a simulation run nor make the model object saveable.", 
          new String[]{  "The name of the simulation to be loaded" } );
      addCommand(newPK_Category_ModelOperation, "LoadCodeModel",            "Load a code model into a model object",                       true,  false, 
          "Loads a code model into a model object.", 
          new String[]{  "The name of the code model to be loaded" } );
      addCommand(newPK_Category_ModelOperation, "LoadDataModel",            "Load a data model",                                           true,  false, 
          "Loads a data model into the model object", 
          new String[]{  "The name of the data model to be loaded." } );
      addCommand(newPK_Category_ModelOperation, "LoadParameterSet",         "Load a parameter set",                                        true,  false, 
          "Loads a parameter set into an existing model", 
          new String[]{  "The name of the parameter set to load."  } );
      addCommand(newPK_Category_ModelOperation, "ModelStatus",              "Get the status of the current model",                         true,  false, 
          "Gets a report describing the initialization status of the current model object.");
      addCommand(newPK_Category_ModelOperation, "DisposeAndReset",          "Reset the model in the GUI",                                  true,  false, 
          "Destroys the current model session, closing all graphs and disposing of all data. The model object is destroyed- do not use this unless you`re sure!");
      addCommand(newPK_Category_ModelOperation, "ProcessScript",            "Processes a saved HWM script",                                true,  false, 
          "Processes the instructions stored in a saved HWM script", 
          new String[]{  "The name of the script to be processed" } );
      addCommand(newPK_Category_ModelOperation, "ProcessScriptSilent",      "Processes a saved HWM script without displaying results",     true,  false, 
          "Processes the instructions stored in a saved HWM script without displaying theresults", 
          new String[]{  "The name of the script to be processed" } );
      addCommand(newPK_Category_ModelOperation, "Login",                    "Logs a user into the HWM system",                             true,  false, 
          "Logs in a user using the specified username and password", 
          new String[]{  "Username", "Password" } );
      addCommand(newPK_Category_ModelOperation, "ProcessNarrative",         "Processes a saved text file in narrative format",             true,  false, 
          "Processes the instructions in a saved file in narrative format. In narrative format, a leading date precedes the instruction; when this command is used this date is stripped.", 
          new String[]{  "File name of the stored narrative to be processed" } );
      addCommand(newPK_Category_ModelOperation, "StartDemo",                 "Starts a prepared Demo",                                     true,  false, 
          "Starts a pre-created demo", 
          new String[]{  "Name of the demo to be started" } );
      addCommand(newPK_Category_ModelOperation, "ContinueDemo",              "Continues the current Demo",                                 true,  false, 
          "Continues the demo, if one is in progress");
      addCommand(newPK_Category_ModelOperation, "SkipDemoStep",              "Skips the next step in the current demo",                    true,  false, 
          "Skips the next step in the current demo. Does not indicate whether the demo is past the end or not");
      addCommand(newPK_Category_ModelOperation, "SetDemoStep",               "Sets the step number for the current demo",                  true,  false, 
          "Sets the step number for the current demo. Does not indicate whether the demo step is valid or not.", 
          new String[]{ "New step number"  } );
      addCommand(newPK_Category_ModelOperation, "StepDemoBackwards",         "Decrements the demo step counter",                           true,  false,
          "Decrements the demo step counter, allowing a step to be iterated (use `skip demo step` to continue)");
      addCommand(newPK_Category_ModelOperation, "StopDemo",                  "Stops a demo in progress",                                   true,  false, 
          "Stops the specified demo; if no demo is in progress, returns a message to that effect. If a different demo is (also) being run, has no effect.", new String[]{  "Name of demo to be stopped" } );
      addCommand(newPK_Category_ModelOperation, "TurnSeasonalNarrativeOff",  "Turns off the specified seasonal narrative",                 false, false, 
          "Turns off the specified seasonal narrative", 
          new String[]{ "Narrative to be turned off" } );
      addCommand(newPK_Category_ModelOperation, "TurnSeasonalNarrativeON",   "Turns on the specified seasonal narrative",                  false, false, 
          "Turns on the specified seasonal narrative", 
          new String[]{  "Narrative to be turned on"  } );
      addCommand(newPK_Category_ModelOperation, "SwitchSeasonalNarrative",   "Choose 1 seasonal narrative from several",                   false, false, 
          "Given a list of seasonal narratives, turns all of them off but chooses one at random to turn on", 
          new String[]{  "List of seasonal narratives delimited by `:`", "List 1 of seasonal narratives delimited by `:`", "List 2 of seasonal narratives delimited by `:`", "List 3 of seasonal narratives delimited by `:`", "List 4 of seasonal narratives delimited by `:`", "List 5 of seasonal narratives delimited by `:`", "List 6 of seasonal narratives delimited by `:`", "List 7 of seasonal narratives delimited by `:`", "List 8 of seasonal narratives delimited by `:`"   } );
      addCommand(newPK_Category_ModelOperation, "IfUnsuccessful", "Execute a command only if the previous command failed.", false, false, 
          "Executes the specified command only if the previous command returned an error flag", 
          new String[]{ "Command To Be Executed", "Arg1", "Arg2", "Arg3", "Arg4", "Arg5", "Arg6", "Arg7", "Arg8", "Arg9"});
      addCommand(newPK_Category_ModelOperation, "IfSuccessful", "Execute a command only if the previous command succeeded.", false, false, 
          "Executes the specified command only if the previous command succeeded", 
          new String[]{ "Command To Be Executed", "Arg1", "Arg2", "Arg3", "Arg4", "Arg5", "Arg6", "Arg7", "Arg8", "Arg9"});
      addCommand(newPK_Category_ModelOperation, "MarkAndClone", "Mark current state of sim; branch to a clone", false, false, 
          "Marks the current state of the simulation and branches to a clone of it", new String[] { "Name of mark" });
      addCommand(newPK_Category_ModelOperation, "RestoreToMark", "Restores to a previous branch point", false, false, "Restores a simulation to a previous branch point",
          new String[]{ "Name of the branch point."});
      addCommand(newPK_Category_ModelOperation, "AdoptCurrentBranch", "Adopts the current sim branch", false, false, 
          "Adopts the current simulation branch as the main one");
    
    newPK_Category_Space            = addCommandCategory(  20000,   "Space",              "Add or review points in the model`s space");
      addCommand(newPK_Category_Space, "AddPoint",             "Creates a named point that is stored in the model",            false, false, 
          "Add a point to the list of points known to the simulation. Specify the x, y, and z coordinates", 
          new String[]{  "Name of new point to be added", "New point`s X coordinate", "New point`s Y coordinate", "New point`s Z coordinate" } );
      addCommand(newPK_Category_Space, "AddSurfacePoint",      "Creates a named point at the existing surface",            false, false, 
          "Add a point. Specify the x and y coordinates but z is determined from existing surface",
          new String[]{  "Name of new point to be added", "New point`s X coordinate", "New point`s Y coordinate" } );
      addCommand(newPK_Category_Space, "ViewListOfPoints",     "Views the list of points stored by the model",                 true,  false, 
          "Allows the user to review the list of points currently known to the simulation, and their x, y, z coordinates.");
      addCommand(newPK_Category_Space, "GetSVG",               "Gets the SVG map",                                             false, false, 
          "Returns the SVG for a map of the model" );
      addCommand(newPK_Category_Space, "GetX3D",               "Gets the X3D description of the model",                        false, false, 
          "Returns the X3D description of the 3-D model world" );
      addCommand(newPK_Category_Space, "SetNAndEOrientation",  "Sets the north and east orientations",                         false, false, 
          "Sets whether positive numbers are north or south, east or west. Use `Positive` or `+` to set positive, `Negative` or `-` to set negative.", 
          new String[]{  "New North orientation", "New East orientation" } );

    newPK_Category_Time             = addCommandCategory(  30000,   "Time",               "Set or review variables associated with time in the model");
      addCommand(newPK_Category_Time, "SetPeriod",   "Resets the period designation (i.e. `Colonial)",   false, false, 
          "Sets the `period` variable for the simulation. This is just a piece of text that says what `stage` the simulation is in (i.e. Pioneer)", 
          new String[]{  "Period Name" } );
      addCommand(newPK_Category_Time, "ShowDate",    "Shows the current simulation date and period",     true,  false, 
          "Show the simulation date and period.");
      addCommand(newPK_Category_Time, "SetDate",     "Set the simulation date to the specified date",    true,  false, 
          "Set the simulation date.", 
          new String[]{  "New simulation date" } );
    
    newPK_Category_Graphs           = addCommandCategory(  40000,   "Graphs",             "Manage graphs in the HWM Simulation interface");
      addCommand(newPK_Category_Graphs, "CreateDataGraph",     "Create a Data Series Graph",             true,  false, 
          "Create a new data graph.", 
          new String[]{  "Name of new data graph"   } );
      addCommand(newPK_Category_Graphs, "ShowDataGraph",       "Show a Data Graph",                      true,  false, 
          "Show a particular data graph.", 
          new String[]{  "Name of graph to be displayed"  } );
      addCommand(newPK_Category_Graphs, "AddSeriesToGraph",    "Add a series to a graph",                true,  false, 
          "Add the specified data series to the specified graph for analysis.", 
          new String[]{  "Name of the data series to be added", "Name of graph"   } );
      addCommand(newPK_Category_Graphs, "DropSeriesFromGraph", "Drops a data series from a given graph", false, false, 
          "Drops the specified data series from the specified graph; does not remove either the graph or the data series from the current environment (web only)", 
          new String[]{  "Series to be removed", "Graph from which series is to be removed"   } );
      addCommand(newPK_Category_Graphs, "DropGraph",           "Removes a graph from the GUI",           false, false, 
          "Remove the specified graph from the GUI. Does not remove the data series from the model.", 
          new String[]{  "Graph to be removed" } );
    
    
    newPK_Category_Interrogative    = addCommandCategory(  50000,   "Interrogative",      "Get numeric information about the model");
    // Usually interrogatives will be specific to implementations
    
    newPK_Category_ProbesAndData    = addCommandCategory(  60000,   "Probes and Data",    "Manage probes and data series");
      addCommand(newPK_Category_ProbesAndData, "CreateProbe",             "Create a Probe to record data during the model run",                         true,  false,
          "Create a probe.", 
          new String[]{  "Name of new probe", "Base Date for probe", "Frequency at which reporting will occur", "Data Series to which this probe will push results", "The command used by the probe to retrieve data"   } );
      addCommand(newPK_Category_ProbesAndData, "CreateNumericDataSeries", "Create a numeric data series",                                               true,  false, 
          "Create a numeric data series.", 
          new String[]{  "Name for new numeric data series", "Name for variable to be stored" } );
      addCommand(newPK_Category_ProbesAndData, "ViewDataSeriesPrintout",  "View the contents of a data series",                                         true,  false, 
          "View a printout of a given data series`s data.", 
          new String[]{  "Name of data series to be reviewed"  } );
      addCommand(newPK_Category_ProbesAndData, "ListDataSeries",          "List the Data Series that are available in the model",                       true,  false, 
          "List all the data series currently in the simulation.");
      addCommand(newPK_Category_ProbesAndData, "LoadDataSeries",          "Load a data series from the HWM database",                                   true,  false, 
          "Load a specified data series into the GUI for analysis.", 
          new String[]{  "Run number of series to be loaded", "Name of series to be loaded", "Name of graph into which series will be loaded" } );
      addCommand(newPK_Category_ProbesAndData, "SaveDataSeries",          "Save a data series to the HWM database",                                     false, false, 
          "Save a specified data series.", 
          new String[]{  "Name of data series to be saved"   } );
      addCommand(newPK_Category_ProbesAndData, "SaveAllDataSeries",       "Save all the data series for this run",                                      false, false, 
          "Save all the data series that this simulation has.");
      addCommand(newPK_Category_ProbesAndData, "DropProbe",               "Drop a probe from the simulation",                                           true,  false, 
          "Remove a probe from the list of probes in the simulation.", 
          new String[]{  "Name of probe to be dropped"  } );
      addCommand(newPK_Category_ProbesAndData, "TurnProbeOn",             "Turn a probe on",                                                            true,  false, 
          "Turns the specified probe `on` (reporting will take place based on the probe`s base date and frequency or on the 1st of the month, if monthly).", 
          new String[]{  "Name of probe to be turned on"   } );
      addCommand(newPK_Category_ProbesAndData, "TurnProbeOff",            "Turn a probe off",                                                           true,  false, 
          "Turns the specified probe off (suspends reporting)", 
          new String[]{  "Name of probe to be turned off" } );
      addCommand(newPK_Category_ProbesAndData, "CreateMonthlyProbe",      "Create a probe that will record data on the first of each simulation month", false, true,  
          "Creates a probe that collects information on the first day of each simulation month.", 
          new String[]{  "Name of probe", "Name of series to which probe sends results", "Command used by probe to retrieve information"   } );
      addCommand(newPK_Category_ProbesAndData, "ListProbes",              "List all probes",                                                            true,  false, 
          "Lists all the probes in the model object");
      addCommand(newPK_Category_ProbesAndData, "LoadProbeSet",            "Load a specific probe set from the database",                                true,  false, 
          "Loads a specific probe set from the database.", 
          new String[]{ "Name of probe set to be loaded"  } );
      addCommand(newPK_Category_ProbesAndData, "ListGraphs",              "Lists the graphs available in the GUI",                                      true,  false, 
          "Lists all the graphs available in the GUI. Not all graphs are visible.");
      addCommand(newPK_Category_ProbesAndData, "SetProbeOffset",          "Sets the timestamp offset for a probe",                                      false, false, 
          "Set the time that will be added to a datum stored by a probe", 
          new String[]{ "Probe", "Offset" } );
      
      
      
    newPK_Category_Parameters       = addCommandCategory(  70000,  "Parameters",         "Manage parameters and parameter sets");
      addCommand(newPK_Category_Parameters, "SetParameter",   "Set a parameter in the model`s parameter set to a give value", false, false, 
          "Sets the parameter to the specified value", 
          new String[]{  "Name of parameter", "New value" } );
      addCommand(newPK_Category_Parameters, "ShowParameters", "Show the list of parameters and values",                       true,  false, 
          "Displays a complete list of all parameters and their values.");
    
    newPK_Category_FileManagement   = addCommandCategory(  80000,  "File Management",    "Save files");
      addCommand(newPK_Category_FileManagement, "SaveInstructions", "Saves the instructions from an interactive session into a script", true,  false, 
          "Saves only the instructions (and not the HWM responses) from an interactive session into a file specified by the filename argument", 
          new String[]{  "The file to which instructions are saved" } );
    
    newPK_Category_View             = addCommandCategory(  90000,  "View",               "View GUI elememts");
      
    newPK_Category_ScriptProcessing = addCommandCategory( 100000,  "Script Processing",  "Work with pre-stored scripts, including test scripts");
      addCommand(newPK_Category_ScriptProcessing, "Test", "Executes a standard test script", true, false, 
          "Executes a standard test script");

      addCommand(newPK_Category_ScriptProcessing, "ConvertScriptToNarrative", "Converts a script to a narrative", false, false,
          "Converts a script file to a narrative file",
          new String[]{ "Input Script File Name", "Output File Name", "Base Date"});
      addCommand(newPK_Category_ScriptProcessing, "CompileNarrative", "Compiles a narrative file to an uploadable narrative file", false, false,
          "Compiles a narrrative file to an uploadable narrative file that can be stored in the database",
          new String[]{ "Input File Name", "Output File Name"});
      addCommand(newPK_Category_ScriptProcessing, "UploadNarrative", "Uploads a narrative to the database", false, false,
          "Uploads a compiled narrative file for storage in the database",
          new String[]{"Input File Name", "Narrative Name", "Description", "Base Date", "Is Seasonal"});
      
      
    newPK_Category_Interface        = addCommandCategory( 110000,  "Interface",          "Commands that apply to the interface (do not actually affect simulation)");
      addCommand(newPK_Category_Interface, "BookmarkTranscript",   "Saves a bookmark to the current transcript",               false, false, 
          "Places a bookmark with a hyperlink at the current point in the transcript", 
          new String[]{  "Text to use as bookmark; use quotation marks" } );
    
    newPK_Category_DB               = addCommandCategory( 120000, "DB", "Commands that relate to storing data directly into the database");
      addCommand(newPK_Category_DB, "StoreProbe", "Stores a new probe to the database", true, false,
          "Stores a new probe to the database. Will not store probe if one already exists with the specified name.",
          new String[]{"Name", "Description", "Data Series", "Domain", "Numeric", "Base Date", "Frequency", "Retrieval String", "Offset"});      
      addCommand(newPK_Category_DB, "StoreMonthlyProbe", "Stores a new monthly probe to the database", true, false,
          "Stores a new monthly probe to the database. Will not store probe if one already exists with the specified name.",
          new String[]{"Name", "Description", "Data Series", "Domain", "Numeric", "Retrieval String", "Offset"});      
      addCommand(newPK_Category_DB, "StoreProbeSet", "Stores a new probe set to the database", true, false,
          "Stores a new probe set to the database. Will not store probe set if one already exists with the specified name.",
          new String[]{"Name", "Description"});      
      addCommand(newPK_Category_DB, "AddProbeToProbeSet", "Adds a probe to a probe set in the database", true, false,
          "Adds the specified probe to the probe set. Will not add the probe if the probe set cannot be edited.",
          new String[]{"Probe Name", "Probe Set Name"});      
      addCommand(newPK_Category_DB, "StoreHistory", "Stores a history in the database", true, false,
          "Stores the specified history in the database",
          new String[]{"History Name", "Description", "Truth Value"});
      addCommand(newPK_Category_DB, "AddNarrativeToHistory", "Adds a narrative to a history in the database", true, false,
          "Adds the (already stored) narrative to the (already stored) history in the database. Adjusts narrative orders to fit the new narrative in the specified spot.",
          new String[]{"History Name", "Narrative Name", "Base Date", "Order"});
      
    newPK_Category_Actors = addCommandCategory(130000, "Actors", "Sends commands to agents and populations");
      addCommand(newPK_Category_Actors, "CreatePopulation", "Creates a new population of agents", false, false, 
          "Creates a new population of agents, if none exists");
      addCommand(newPK_Category_Actors, "PopulationDo", "Sends a command to the model`s population", true, false, 
          "Sends a command to the model`s population",
          new String[]{"Command to be sent to the population", "Argument 1 to command sent", 
          "Argument 2 to command sent", "Argument 3 to command sent", "Argument 4 to command sent", 
          "Argument 5 to command sent", "Argument 6 to command sent", "Argument 7 to command sent", 
          "Argument 8 to command sent", "Argument 9 to command sent"});
      addCommand(newPK_Category_Actors, "AgentDo", "Sends a command to a specified agent in the model`s population", true, false, 
          "Sends a command to a specified agent in the model`s population",
          new String[]{"ID of Agent to be sent command", "Command to be sent", "Argument 1 to command sent", 
          "Argument 2 to command sent", "Argument 3 to command sent", "Argument 4 to command sent", 
          "Argument 5 to command sent", "Argument 6 to command sent", "Argument 7 to command sent", 
          "Argument 8 to command sent"});
      
      
    newPK_Category_NotYetClassified = addCommandCategory(1000000, "Not yet classified", "Not yet classified");
    
    // Default code model
    defaultCodeModelPK = addCodeModel("Default", systemUserPK, "Default- uses no code model options");
    
    
    // Adding a 'system' comment to the default code model renders it unchangeable (can't be edited)
    long defaultCodeModelCommentPK = add(DB.getCommentsTable().getDatum(systemUserPK, DB.currentTimeString(), "Uses no code model options. DO NOT MODIFY"));
    add(DB.getCodeModelCommentsTable().getDatum(defaultCodeModelPK, defaultCodeModelCommentPK));
    
    
    // There are two narratives that are created by the system
    // One is the 'CreatePopulation' narrative
    createPopulationNarrative = this.addNarrative("CreatePopulation", "Executes the command to create a population", 
        this.systemUserPK, "01-JAN-0001-AD", false);
      addNarrativeInstruction(createPopulationNarrative, "01-JAN-0001-AD", 1, "CreatePopulation", 
          "", "", "", "", "", "", "", "", "", "", "Allows scheduling population creation");
    
    // The other is 'stop' which can be placed anywhere. When a history has instructions up to day X but
    // you need it to run to a later date, you can place the 'stop' command at the later date
    stopNarrative = this.addNarrative("StopNarrative", "Marks the end of a narrative", this.systemUserPK, "01-JAN-0001-AD", false);
      addNarrativeInstruction(stopNarrative, "01-JAN-0001-AD", 1, "Stop", 
          "", "", "", "", "", "", "", "", "", "", "Allows schedule a history to run up to a given date.");
    
  }
  
}
