package abcm.db;



/**
 * Abstract base class for packages that need to initialize data without
 * being part of the lineage of db initializer classes. These classes
 * are usually ones that make use of AbstractModelComponent classes and thus
 * need to add elements to the list of parameters, parameter sets, or to code model
 * domains, options, and code models.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractABCMdbPackageDataInitializer {

  /** DB Initializer to be used to add the required data  */
  protected ABCMdbDataInitializer dbInit;
  /** Count of Parameters  */
  private   int paramCount = 1;
  
  
  /**
   * Constructor
   * @param dbInitializer
   */
  public AbstractABCMdbPackageDataInitializer(ABCMdbDataInitializer dbInitializer){
    dbInit = dbInitializer;
  }
  
  /**
   * This method is called to initialize the data. Confirms that this package
   * has not been initialized before (each instance of this class is called only
   * once per initialization).
   * @throws Exception
   */
  public void init() throws Exception{
    if(!hasBeenInitialized()){
      setHasBeenInitialized();
      initialize();      
    }
  }
  
  /**
   * Confirm that the data has been initialized
   * @return true if the data has been initialized, false otherwise
   */
  protected abstract boolean hasBeenInitialized();
  
  /**
   * Set the 'initialized' flag
   */
  protected abstract void setHasBeenInitialized();
  
  /**
   * Initialize the required data
   * @throws Exception
   */
  protected abstract void initialize() throws Exception;
  
  /**
   * Adds the specified parameter category
   * @param name name of the new parameter category
   * @param description description of the new parameter category
   * @return PK of the new parameter category
   * @throws Exception
   */
  protected long addParameterCategory(String name, String description) throws Exception{
    return dbInit.addParameterCategory(name, description);
  }
  
  /**
   * Adds a new parameter
   * @param name name of the new parameter
   * @param description description of the new parameter
   * @param defaultValue default value for the new parameter
   * @param minimum minimum value of the new parameter
   * @param maximum maximum value of the new parameter
   * @param units units to be used with the new parameter
   * @param parameterCategoryFK category in which the new parameter is to be placed
   * @return new PK of the new parameter
   * @throws Exception
   */
  protected long addParameter(String name, String description, double defaultValue, 
      double minimum, double maximum, String units, long parameterCategoryFK) throws Exception{
    return dbInit.addParameter(name, description, defaultValue, minimum, maximum, units, false, parameterCategoryFK, paramCount++);
  }
  
  // Code Model Domains
  /**
   * Adds a new Code Model Domain
   * @param domainName name of the new domain
   * @param description description of the new domain
   * @param ID ID to be used for this domain 
   * @return new PK for this domain
   * @throws Exception
   */
  protected long addCodeModelDomain(String domainName, String description, String ID) throws Exception{
    return dbInit.addCodeModelDomain(domainName, description, ID);
  }
    
  // Code Model Options
  /**
   * Adds a new code model option domain value
   * @param manualPK manually assigned PK for this element
   * @param optionDomainFK FK identifying the domain to which this element will belong
   * @param value value to be used with this element
   * @param name name to be assigned to this element
   * @param code code associated with this element
   * @param description description of this element
   * @return new PK assigned to this element
   * @throws Exception
   */
  protected long addCodeModelDomainOptionValue(long manualPK, long optionDomainFK, long value, String name, 
      String code, String description)throws Exception{
     return dbInit.addCodeModelDomainOptionValue(manualPK, optionDomainFK, value, name, code, description);
  }
  
  // Code Models
  /**
   * Adds a new Code Model
   * @param title name of the Code Model
   * @param researcherFK FK identifying the researcher who created or owns this Code Model
   * @param description Description of the new Code Model
   * @return new PK assigned to this Code Model
   * @throws Exception
   */
  protected long addCodeModel(String title, long researcherFK, String description) throws Exception{
    return dbInit.addCodeModel(title, researcherFK, description);
  }
  
  // Code Model Domains And Options
  /**
   * Add new linkage of Code Model Domain and Option Value
   * @param codeModelFK link to a specified Code Model 
   * @param optionDomainFK link to a specific Code Model Option Domain
   * @param optionValueFK link to a specific Code Model Option Value
   * @return PK of newly added element
   * @throws Exception
   */
  protected long addCodeModelAndOptionValue(long codeModelFK, long optionDomainFK, long optionValueFK) throws Exception{
    return dbInit.addCodeModelAndOptionValue(codeModelFK, optionDomainFK, optionValueFK);
  }
  
  
}
