package abcm.db;

import java.io.File;
import java.io.FileWriter;
import java.io.StringWriter;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.Statement;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.Vector;

import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.ABCMdbView;
import abcm.db.syntaxengines.AbstractSQLSyntaxEngine;

/**
 * An AbstractDBGateway is an object that facilitates the
 * on-the-fly creation of an SQL server database, and
 * database access thereafter.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractDBGateway {

  /** An SQL Server Constant  */
  public static final long MAX_GUID_VALUE = 2147483147; // This is probably used by the jsp pages... ? Check...
  
  /**
   * A Syntax Engine for creating SQL syntax (in MS-SQL Server or MySQL)
   */
  private static AbstractSQLSyntaxEngine sqlSyntaxEngine;
  
  /** Guest DB ID  */
  protected static String guestDBID;
  /** Guest DB Password  */
  protected static String guestDBPWD;
  /** Administrator DB user ID  */
  protected static String adminDBID;
  /** Administrator DB password  */
  protected static String adminDBPWD;
  
  /** Character used to divide fields in an option list  */
  private static String DEFAULT_OPTION_LIST_FIELD_DIVIDER = " - ";
  
  /** Flag indicating that the order (of an 'ORDER BY' clause) should be ascending  */
  public static final boolean ORDER_BY_ASCENDING  = false;
  /** Flag indicating that the order (of an 'ORDER BY' clause) should be descending  */
  public static final boolean ORDER_BY_DESCENDING = true;

  
  /** Collection of Tables contained in the associated database  */
  protected Vector tableList = new Vector();
  
  /** Collection of Views contained in the associated database  */
  protected Vector viewList  = new Vector();
  
  /** Collection of data to be inserted when the database is first created  */
  protected Vector startData = new Vector(); 
  
  /**
   * Generates a 'WHERE" SQL clause using the specified field name and a numeric value for the comparand.
   * @param field name of the field to be filtered
   * @param val value to be filtered
   * @return the text of the clause
   */
  public static String where(String field, long val){   return sqlSyntaxEngine.where(field, val);}
  
  /**
   * Generates a 'WHERE" SQL clause using the specified field name and a String value for the comparand.
   * @param field name of the field to be filtered
   * @param val value to be filtered
   * @return the text of the clause
   */
  public static String where(String field, String val){ return sqlSyntaxEngine.where(field, val);}
  
  /**
   * Generates an 'ORDER BY' SQL clause in the specified order (ascending or descending).
   * @param field field to be sorted
   * @param descending true if descending order is required, false otherwise
   * @return the text of the clause
   */
  public static String orderBy(String field, boolean descending){ return sqlSyntaxEngine.orderBy(field, descending); }
  
  /**
   * Generates an 'ORDER BY' SQL clause. Assumes ascending order.
   * @param field name of the field to be sorted
   * @return the text of the clause
   */
  public static String orderBy(String field){                     return sqlSyntaxEngine.orderBy(field, ORDER_BY_ASCENDING);                }

  /** Contains a description of any errors this object has encountered  */
  private String errorReport;
  
  
  public AbstractDBGateway(AbstractSQLSyntaxEngine syntaxEngine){
    sqlSyntaxEngine = syntaxEngine;
  }
  

  public AbstractSQLSyntaxEngine getSQLSyntaxEngine(){ return sqlSyntaxEngine; }
  
  /* Table Management */
  /**
   * Adds a table to this gateway
   * @param table the table to be added
   */
  protected final void addTable(ABCMdbTable table){
    removeTable(table);    // Ensure that there are no duplicates
    tableList.add(table);  // Add the table to the collection
  }
  
  /**
   * Removes a table from this gateway
   * @param table the table to be removed
   */
  protected final void removeTable(ABCMdbTable table){    tableList.remove(table);    }
  
  /**
   * Substitutes one table for another in this gateway.
   * @param oldTable the table to be replaced
   * @param newTable the new table
   */
  protected final void substituteTable(ABCMdbTable oldTable, ABCMdbTable newTable){    substituteViewable(tableList, oldTable, newTable);    }
  
  /**
   * Adds a view to this gateway
   * @param view the view to be added
   */
  protected final void addView(ABCMdbView view){
    removeView(view);
    viewList.add(view);
  }
  
  /**
   * Removes a view from this gateway
   * @param view the view to be removed
   */
  protected final void removeView(ABCMdbView view){    viewList.remove(view);    }
  
  /**
   * Substitutes a new view for one already in this gateway
   * @param oldView the view to be replaced
   * @param newView the new view
   */
  protected final void substituteView(ABCMdbView oldView, ABCMdbView newView){    substituteViewable(viewList, oldView, newView);    }
  
  /**
   * Swaps one ABCM Viewable for another from a vector of ABCM Viewables. Used to substitute either a
   * table or a view. Guarantees that the new table occupies the position of the old table, so that
   * order (which sometimes matters) can be preserved. If the 'old' viewable is actually not present, the
   * new one is added at the end of the list
   * @param viewableList list of viewables
   * @param oldViewable the table to be removed
   * @param newViewable the table to be substituted in
   */
  protected final void substituteViewable(Vector viewableList, ABCMdbAbstractViewable oldViewable, ABCMdbAbstractViewable newViewable){
    for(int i = 0; i < viewableList.size(); i++){
      if(viewableList.get(i).equals(oldViewable)){
        viewableList.remove(i);
        viewableList.add(i, newViewable);
        return;
      }
    }
    // If the old table wasn't found
    viewableList.add(newViewable);
  }
  
  /**
   * Moves one table before another in the list of tables. If the target table is not present,
   * nothing is done.
   * @param table the table being inserted
   * @param target the table that serves as the target for the insertion point
   */
  protected final void moveTableBefore(ABCMdbTable table, ABCMdbTable target){
    moveViewableBefore(tableList, table, target);
  }
  
  /**
   * Moves one view before another in the list of views. If the target View is not present,
   * nothing is done.
   * @param view the view being inserted
   * @param target the view that serves as the target for the insertion point
   */  
  protected final void moveViewBefore(ABCMdbView view, ABCMdbView target){
    moveViewableBefore(viewList, view, target);
  }
   
  /**
   * Moves one Viewable in a list to a position immediately before another specified one. If the
   * target Viewable is not present, nothing is done.
   * @param viewableList The list of Viewables into which the new Viewable is to be inserted
   * @param viewable the Viewable to be inserted. 
   * @param target the target point prior to which the new table will be inserted.
   */
  protected final void moveViewableBefore(Vector viewableList, ABCMdbAbstractViewable viewable, ABCMdbAbstractViewable target){
    if(viewableList.contains(target)){
      viewableList.remove(viewable);
      for(int i = 0; i < viewableList.size(); i++){
        if(viewableList.get(i).equals(target)){
          viewableList.add(i, viewable);
          return;
        }
      }
    }
  }
  
  

  /**
   * Get a connection to the database
   * @param id the user ID to be passed to the connection
   * @param pd additional information passed to the connection string
   * @return the Connection established
   */  
  public Connection getConnection(String id, String pd){
    if(id == null) return null;
    if(pd == null) return null;
    Connection con = null;
    try {
      Class.forName(getDriverClassName());
      DriverManager.setLoginTimeout(10);
      con = DriverManager.getConnection(getURL(), id, pd);
      con.setCatalog(getSchema());
      }
    catch (Exception e){
      System.out.println("ERROR in AbstractDBGateway.java: Could not get connection " + getURL() + ": " + e.toString());
      errorReport = "Could not get connection (" + getURL() + ")!";
    }
    return con;    
  }
  
  /**
   * Name of the class of database driver to load
   * @return the class name
   */
  protected abstract String getDriverClassName();
  
  /**
   * Sets the Guest DB ID
   * @param id the new ID
   */
  public void setGuestDBID (String  id){ guestDBID  =  id; }
  
  /**
   * Sets the Guest DB Password
   * @param pwd the new Password
   */
  public void setGuestDBPWD(String pwd){ guestDBPWD = pwd; }
  
  /**
   * Sets the Admin DB ID
   * @param id the new ID
   */
  public void setAdminDBID (String  id){ adminDBID  =  id; }
  
  /**
   * Sets the Admin DB Password
   * @param pwd the new Password
   */
  public void setAdminDBPWD(String pwd){ adminDBPWD = pwd; }
  
  // TODO Replace Guest Connection functionality
//  public Connection getGuestConnection(){    return getConnection(guestDBID, guestDBPWD);    }
  
  /**
   * Gets a Guest Connection to the database
   * @return a Connection to the gatabase with Guest permissions
   */
  public Connection getGuestConnection(){    return getConnection(adminDBID, adminDBPWD);    }
  
  /**
   * Gets an Admin Connection to the database
   * @return a Connection to the database with Admin permissions
   */
  public Connection getAdminConnection(){    return getConnection(adminDBID, adminDBPWD);    }  
    
  /**
   * Gets the URL for the database connection
   * @return a database URL
   */
  public abstract String getURL();
  
  /**
   * Gets the schema for the database Connection
   * @return a String representing a database schema
   */
  public abstract String getSchema();
  
  /**
   * Gets the owner of the database. This is crucial to correctly accessing database elements
   * in some versions of SQL (like SQL Server).
   * @return the owner of the database.
   */
  public abstract String getOwner();
  
  /**
   * Given a String, parse out characters that would make the string
   * not be a valid part of an SQL statement (such as quotation marks).
   * @param originalText The text to be converted
   * @return The converted text.
   */
  public static String makeSQLSafe(String originalText){
    String textToReturn = originalText;
    textToReturn = textToReturn.replace('\'', '`');
    textToReturn = textToReturn.replace('"', '`');
    textToReturn = textToReturn.replace('&', '~');
    return textToReturn;    
  }
  
  /**
   * Given a String, parse characters to make it valid for
   * tex processing as part of normal text. For example,
   * an underscore is used as a command in tex, but and
   * so should be replaced with an escape sequence.
   * @param originalText text to be parsed
   * @return tex-safe text
   */
  public static String makeTexSafe(String originalText){
    String textToReturn = originalText;
    textToReturn = textToReturn.replaceAll("_", "\\\\\\_");
    return textToReturn;
  }
  
  /**
   * Takes an array of Strings and processes
   * all of them using the makeSQLSafe method
   * @param originalText the set of Strings to be converted
   * @return the array of transformed Strings
   */
  public static String[] makeSQLSafe(String[] originalText){
    for(int i = 0; i < originalText.length; i++){
      originalText[i] = makeSQLSafe(originalText[i]);
    }
    return originalText;
  }

  /**
   * Transforms a string so that it can be safely used as a string value
   * in an XML statement. This means that special characters that could
   * be misinterpreted are eliminated.
   * @param originalText the text to be sanitized
   * @return the new version of the text
   */
  public static String makeXMLSafe(String originalText){
    if(originalText == null) return "";
    String textToReturn = originalText;
    textToReturn = textToReturn.replaceAll("&", "&amp;");
    textToReturn = textToReturn.replaceAll("<","&lt;");
    textToReturn = textToReturn.replaceAll(">","&gt;");
    return textToReturn;
  }
  
  /**
   * Generates a string indicating the current time.
   * @return a string indicating the current time
   */
  public final String currentTimeString(){
    return (new SimpleDateFormat(sqlSyntaxEngine.getDateTimeInsertFormat())).format(new Date(System.currentTimeMillis()));
  }
  
  /**
   * Takes an SQL statement that executes an insert action, performs
   * it, and returns a value indicating the primary key of the
   * new record inserted. Returns -1 if the insert failed
   * @param sqlInsert the SQL statement to be executed
   * @return the primary key of the newly inserted record
   * @throws Exception generic Exception
   */  
  public long insertAndReturnPK(String sqlInsert) throws Exception {
    Connection con = this.getAdminConnection();
    Statement st = con.createStatement();
    
    st.executeUpdate(sqlInsert, Statement.RETURN_GENERATED_KEYS);
    long newPK = -1;
    ResultSet rs = st.getGeneratedKeys();
    if(rs.next()) newPK = rs.getLong(1);
    rs.close();
    st.close();
    con.close();
    return newPK;
  }
  
  /**
   * Transforms a string usable as an HTML Option List into
   * the equivalent collection of options, but as radio buttons.
   * @param optionList the option list to be transformed
   * @param name the name of the radio button control created
   * @return a string with the radio button control in HTML format
   */
  public static String convertOptionListToButtonList(String optionList, String name){
    return optionList.replaceAll("<OPTION", "<INPUT TYPE=\"radio\" NAME=\"" + name + "\"").replaceAll("</OPTION>","</INPUT><br>");
  }
    

  /* Getting Option Lists */
  /**
   * Gets an HTML option list from elements in the database.
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleField field that is to be displayed in the option list
   * @return the HTML creating this option list
   */  
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField visibleField){
    return getOptionList(view, valueField, new ABCMdbField[] { visibleField }, null, null, DEFAULT_OPTION_LIST_FIELD_DIVIDER);
  }
  
  /**
   * Gets an HTML option list from elements in the database.
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleField field that is to be displayed in the option list
   * @param fieldDivider character that separates text in the option list
   * @return the HTML creating this option list
   */  
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField visibleField, String fieldDivider){
    String sql = view.getSelectSQL(getOwner(),orderBy(visibleField.getName()));
    return makeOptionList(sql, valueField, new ABCMdbField[] { visibleField }, fieldDivider);
  }
  
  /**
   * Gets an HTML option list from elements in the database.
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleField field that is to be displayed in the option list
   * @param filterField a field that can be used to filter records based on some value
   * @param filterValue value used to filter some records
   * @return the HTML creating this option list
   */  
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField visibleField, 
      ABCMdbField filterField, long filterValue){
    return getOptionList(view, valueField, new ABCMdbField[] { visibleField }, filterField, filterValue, DEFAULT_OPTION_LIST_FIELD_DIVIDER);
  }    
  
  /**
   * Gets an HTML option list from elements in the database.
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleField field that is to be displayed in the option list
   * @param filterField a field that can be used to filter records based on some value
   * @param filterValue value used to filter some records
   * @param fieldDivider character that separates text in the option list
   * @return the HTML creating this option list
   */  
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField visibleField, 
      ABCMdbField filterField, long filterValue, String fieldDivider){
    if(filterField == null) return "";
    String sql = view.getSelectSQL(getOwner(), where(filterField.getName(), filterValue) + " " + orderBy(visibleField.getName()));
    return makeOptionList(sql, valueField, new ABCMdbField[] { visibleField }, fieldDivider);
  }  
  
  /**
   * Gets an HTML option list from elements in the database. Uses default field divider.
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleField field that is to be displayed in the option list
   * @param filterField a field that can be used to filter records based on some value
   * @param filterValue value used to filter some records
   * @return the HTML creating this option list
   */  
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField visibleField, 
      ABCMdbField filterField, String filterValue){
    return getOptionList(view, valueField, new ABCMdbField[] { visibleField }, filterField, filterValue, DEFAULT_OPTION_LIST_FIELD_DIVIDER);
  }  
  
  /**
   * Gets an HTML option list from elements in the database.
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleField field that is to be displayed in the option list
   * @param filterField a field that can be used to filter records based on some value
   * @param filterValue value used to filter some records
   * @param fieldDivider character that separates text in the option list
   * @return the HTML creating this option list
   */  
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField visibleField, 
      ABCMdbField filterField, String filterValue, String fieldDivider){
    if((filterField == null) ^ (filterValue == null)) return ""; // Exclusive OR;  
    String sql = view.getSelectSQL(getOwner(), where(filterField.getName(), filterValue) + " " + orderBy(visibleField.getName()));
    return makeOptionList(sql, valueField, new ABCMdbField[] { visibleField }, fieldDivider);
  }  
  
  /**
   * Gets an HTML option list from elements in the database. Uses default field divider
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleFields fields that are to be displayed in the option list
   * @return the HTML creating this option list
   */
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField[] visibleFields){
    return getOptionList(view, valueField, visibleFields, null, null, DEFAULT_OPTION_LIST_FIELD_DIVIDER);
  }
  
  /**
   * Gets an HTML option list from elements in the database
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleFields fields that are to be displayed in the option list
   * @param fieldDivider character that separates text in the option list
   * @return the HTML creating this option list
   */
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField[] visibleFields, String fieldDivider){
    String sql = view.getSelectSQL(getOwner(),orderBy(visibleFields[0].getName()));
    return makeOptionList(sql, valueField, visibleFields, fieldDivider);
  }
  
  /**
   * Gets an HTML option list from elements in the database. Uses default character divider.
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleFields fields that are to be displayed in the option list
   * @param filterField a field that can be used to filter records based on some value
   * @param filterValue value used to filter some records
   * @return the HTML creating this option list
   */
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField[] visibleFields, 
      ABCMdbField filterField, long filterValue){
    return getOptionList(view, valueField, visibleFields, filterField, filterValue, DEFAULT_OPTION_LIST_FIELD_DIVIDER);
  }    
  
  /**
   * Gets an HTML option list from elements in the database
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleFields fields that are to be displayed in the option list
   * @param filterField a field that can be used to filter records based on some value
   * @param filterValue value used to filter some records
   * @param fieldDivider character that separates text in the option list
   * @return the HTML creating this option list
   */
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField[] visibleFields, 
      ABCMdbField filterField, long filterValue, String fieldDivider){
    if(filterField == null) return "";
    String sql = view.getSelectSQL(getOwner(), where(filterField.getName(), filterValue) + " " + orderBy(visibleFields[0].getName()));
    return makeOptionList(sql, valueField, visibleFields, fieldDivider);
  }  
  
  /**
   * Gets an HTML option list from elements in the database. Uses default character divider.
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleFields fields that are to be displayed in the option list
   * @param filterField a field that can be used to filter records based on some value
   * @param filterValue value used to filter some records
   * @return the HTML creating this option list
   */
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField[] visibleFields, 
      ABCMdbField filterField, String filterValue){
    return getOptionList(view, valueField, visibleFields, filterField, filterValue, DEFAULT_OPTION_LIST_FIELD_DIVIDER);
  }  
  
  /**
   * Gets an HTML option list from elements in the database
   * @param view the view from which the elements are to be selected
   * @param valueField the field from which the 'values' portion of the option list will be taken
   * @param visibleFields fields that are to be displayed in the option list
   * @param filterField a field that can be used to filter records based on some value
   * @param filterValue value used to filter some records
   * @param fieldDivider character that separates text in the option list
   * @return the HTML creating this option list
   */
  public String getOptionList(ABCMdbAbstractViewable view, ABCMdbField valueField, ABCMdbField[] visibleFields, 
      ABCMdbField filterField, String filterValue, String fieldDivider){
    if((filterField == null) ^ (filterValue == null)) return ""; // Exclusive OR;
    String whereClause = "";
    if(filterField != null) whereClause = where(filterField.getName(), filterValue) + " ";
    String sql = view.getSelectSQL(getOwner(), whereClause + orderBy(visibleFields[0].getName()));
    return makeOptionList(sql, valueField, visibleFields, fieldDivider);
  }
  
  /**
   * Makes an HTML option list from the specified sql statement
   * @param sql SQL to be used to create a recordset from which the option list will be drawn
   * @param valueField the name of the field to be used as the option list's 'value' 
   * @param visibleFields fields that should be displayed in the option list
   * @param fieldDivider character that separates multiple fields in the list
   * @return the HTML that generates the option list
   */
  public String makeOptionList(String sql, ABCMdbField valueField, ABCMdbField[] visibleFields, String fieldDivider){
    clearErrorReport();
    String optionList = "";
    try{
      Connection con = getAdminConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      while(rs.next()){
        String id = rs.getString(valueField.getName());
        String val = "";
        for(int i = 0; i < visibleFields.length; i++){
          if(i > 0) val += fieldDivider;
          val += rs.getString(visibleFields[i].getName());        
        }
        optionList+= "<OPTION VALUE=\"" + id + "\">" + val + "</OPTION>";
      }
    }
    catch(Exception e){
      addToErrorReport("Could not complete option list: " + e.getMessage());
    }
    return optionList;
  }
  
  /**
   * Gets a simple HTML Option List from the specified table, using the
   * specified fields for value and key.
   * @param tableName the name of the table from which data should be retrieved
   * @param ValueFieldName the name of the field from which the values in the
   * key column should be taken
   * @param nameFieldName the name of the field from which the values in the
   * identifier column should be taken
   * @return the Option List created
   * @throws Exception generic Exception
   */
  public String getAsStringIndexedOptionList(String tableName, String ValueFieldName, String nameFieldName) throws Exception{
    String optionList = "";
    String sql = "SELECT " + ValueFieldName + ", " + nameFieldName;
    sql += " FROM " + tableName + ";";
    Connection con = getAdminConnection();
    Statement st = con.createStatement();
    
    ResultSet rs = st.executeQuery(sql);
    
    while(rs.next()){
      String val = rs.getString(nameFieldName);
      String id = "" + rs.getString(ValueFieldName);
      optionList+= "<OPTION VALUE=\"" + id + "\">" + val + "</OPTION>";
    }
    rs.close();
    st.close();
    con.close();
   
    return optionList;
  }

  /* Utilities */
  
  /**
   * Sets the 'IDENTITY INSERT' property on the specified table to 'ON'
   * @param tableName name of the table to be set
   * @param con database connection
   * @return true if successful
   */
  public boolean setTableIdentityInsertOn(String tableName, Connection con){
    return setTableIdentityInsert(tableName, true, con);  
  }
  
  /**
   * Sets the 'IDENTITY INSERT' property on the specified table to 'OFF'
   * @param tableName name of the table to be set
   * @param con database connection
   * @return true if successful
   */
  public boolean setTableIdentityInsertOff(String tableName, Connection con){
    return setTableIdentityInsert(tableName, false, con);
  }
  
  /**
   * Sets the 'IDENTITY INSERT' property on the specified table to the specified value
   * @param tableName name of the table
   * @param newStateIsOn flag indicating the desired state
   * @param con database connection
   * @return true if successful, false otherwise
   */
  private boolean setTableIdentityInsert(String tableName, boolean newStateIsOn, Connection con){
    System.out.println("SETTING IDENTITY INSERT for " + tableName);
    boolean successFlag = false;
    try{
      Statement st = con.createStatement();
      st.execute("SET IDENTITY_INSERT " + tableName + " " + (newStateIsOn ? "ON" : "OFF"));
      successFlag = true;
      st.close();      
    }
    catch(Exception E){
      addToErrorReport("Could not set identity insert on table '" + tableName + "' to 'ON': " + E.getMessage());
    }
    return successFlag;
  }
  
  /* Database Creation and Testing */
  
  /**
   * Implementing classes must provide this if the intend to initialize a database
   * @return a data initialization object
   */
  public abstract AbstractDBDataInitializer getDataInitializer();
  
  /**
   * Implementing classes must specify this. If true, then an error
   * during database creation should roll back all transactions so database
   * is returned to original state; if false, an error results in an incomplete
   * database. If true, all transactions are committed if the database creation
   * completes without error.
   * @return true if transaction processing and rollback are used, false if not
   */
  protected abstract boolean useTransactionCommitAndRollback();
  
  /**
   * Implementing classes may override this. If this returns true, 
   * when the database creation terminates successfully all transactions
   * are rolled back and the database is returned to its original state. This
   * is sometimes useful for development. The default is false.
   * Note that classes wishing to roll back for development must
   * also implement the useTransactionCommitAndRollback method to return true. 
   * @return true if rollback is executed, false if not
   */
  protected boolean getRollbackForDevelopment(){ return false; }
  
  /**
   * Create an entire, new database from the collection of views, tables, and starting data.
   * @return true if successful, false otherwise.
   */
  public final boolean createDB(){
    
    System.out.println("Creating database...");
    Connection con = getAdminConnection();
    if(con == null) return false;
    try{
      // If there is to be transaction processing...
      if(useTransactionCommitAndRollback()) con.setAutoCommit(false);
      try{
        createTables(con);
        createViews(con);
        getDataInitializer().initialize(con);        
      }
      catch(Exception E){
        System.out.println("Encountered exception at pt 0: (Rolling back:" + useTransactionCommitAndRollback() + "): " + E.getMessage() + "\n");
        E.printStackTrace(System.out);
        // If there is transaction processing, roll everything back
        if(useTransactionCommitAndRollback()) con.rollback();
        // Return a value of false to indicate failure
        return false;
      }
      System.out.println("DB CREATION WAS SUCCESSFUL.");
      if(getRollbackForDevelopment()){
        if(useTransactionCommitAndRollback()){
          System.out.println("ROLLING BACK FOR CONTINUED DEVELOPMENT...");
          con.rollback();
        }
        else{
          System.out.println("PROBLEM: Implementing class is attempting to roll back for development, but transaction processing is not permitted.");
          System.out.println("Database successfully created and is in place; to remove, delete manually.");
        }
      }
      else{
        if(useTransactionCommitAndRollback()){
          System.out.println("COMMITTING!!!!");
          con.commit();
        }
      }
    }
    catch(Exception E){
      addToErrorReport("Encountered exception at pt 1:" + E.getMessage());
      return false;
    }
    finally{
      try{
        con.close();
      }
      catch(Exception E){
        addToErrorReport("Encountered Exception at pt 2:" + E.getMessage());
      }
    }
    return true;
  }
  
  /**
   * Create in a new databasethe tables listed in this object's collection of tables
   * @param con database connection
   * @throws Exception
   */
  private final void createTables(Connection con) throws Exception{
    System.out.println("Creating tables...");
    for(int i = 0; i < tableList.size(); i++){
      Statement st = con.createStatement();
      ABCMdbTable t = (ABCMdbTable)tableList.get(i);
      String[] sqls = t.getCreateSQL(getOwner(), sqlSyntaxEngine);
      System.out.println("EXECUTE SQL HERE:\n\n");
      for(String s: sqls){
        System.out.println("SECTION: " + s + "\n\n");
        st.execute(s);
      }
      System.out.println("Executed!");
    }
  }
 
  /**
   * Add a given datum to the database.
   * @param con database connection
   * @param datum the datum to be added
   * @return the PK created by the add operation
   * @throws Exception
   */
  protected final long addDatum(Connection con, ABCMdbDatum datum) throws Exception{
    if(datum.getIncludesIdentityPK())  setTableIdentityInsertOn(datum.getTable().getFullName(getOwner()), con);
    Statement st = con.createStatement();
    System.out.println("ADDING: " + datum.getInsertSQL(getOwner(), sqlSyntaxEngine));
    st.executeUpdate(datum.getInsertSQL(getOwner(), sqlSyntaxEngine), Statement.RETURN_GENERATED_KEYS);
    long newPK = -1;
    ResultSet rs = st.getGeneratedKeys();
    if(rs.next()) newPK = rs.getLong(1);    
    if(datum.getIncludesIdentityPK())  setTableIdentityInsertOff(datum.getTable().getFullName(getOwner()), con);
    st.close();
    return newPK;
  }
  
  /**
   * Create in a new database the collection of views represented by the list of views
   * in this object. 
   * @param con database connection
   * @throws Exception
   */
  private final void createViews(Connection con) throws Exception{
    System.out.println("Creating views....");
    for(int i = 0; i < viewList.size(); i++){
      Statement st = con.createStatement();
      ABCMdbView v = (ABCMdbView)viewList.get(i);
      String[] sqls = v.getCreateSQL(getOwner(), sqlSyntaxEngine);
      System.out.println("EXECUTE SQL HERE:\n\n");
      for(String s: sqls){
        System.out.println("SECTION: " + s + "\n\n");
        st.execute(s);
      }
      System.out.println("Executed!");
    }    
  }
  
  
  /**
   * Perform a set of simple 'select' statements that check to see if
   * all tables and fields in the tables listed in this object are in the
   * database. Writes a message indicating success or failure to standard
   * output
   */
  public void testAllTablesAndFields(){
    clearErrorReport();
    for(int i = 0; i < tableList.size(); i++){
      ABCMdbTable T = (ABCMdbTable)tableList.get(i);
      System.out.print("Testing: " + (T.getFullName(getOwner()) + "                                                                                  ").substring(0,60) + "...  ");
      String sql = "SELECT TOP 1 " + T.getSeparatedFieldList() + " FROM " + T.getFullName(getOwner());
      Connection con;
      Statement st;
      ResultSet rst;
      try{
        con = getAdminConnection();
        st = con.createStatement();
        rst = st.executeQuery(sql);
        System.out.println("Passed!");
        rst.close();
        st.close();
        con.close();
      }
      catch(Exception E){
        System.out.println("FAILED!\n\nFailure on table: " + T.getName() + "\n" +
            "SQL: \n\n" + sql + "\n\n" + E.getMessage()); 
      }
    }
  }
  
  /**
   * Perform a simple 'select' statement that checks to see if
   * the specified 'Viewable' is in the database. 
   * @param viewable the object to be found
   * @return true if the table or view is found, false otherwise (or if an error occurred)
   */
  public boolean testViewable(ABCMdbAbstractViewable viewable){
    clearErrorReport();
    System.out.print("Testing: " + (viewable.getName() + "                                                                                  ").substring(0,60) + "...  ");
    boolean retVal = false;
    String sql = "SELECT TOP 1 " + viewable.getSeparatedFieldList() + " FROM " + viewable.getName();
    Connection con;
    Statement st;
    ResultSet rst;
    try{
      con = getAdminConnection();
      st = con.createStatement();
      rst = st.executeQuery(sql);
      
      retVal = true;
      
      rst.close();
      st.close();
      con.close();
      
    }
    catch(Exception E){
      addToErrorReport("FAILED!\n\nFailure on: " + viewable.getName() + "\n" +
                       "SQL: \n\n" + sql + "\n\n" + E.getMessage());
    }
    return retVal;
  }  
  
  /**
   * Determines if a record exists in the specified Viewable such that
   * the specified field contains the specified PK value
   * @param viewable Viewable to be tested
   * @param field field in which value is to be found
   * @param PK value sought
   * @return true if the entry exists, false otherwise
   */
  public boolean entryExists(ABCMdbAbstractViewable viewable, ABCMdbField field, long PK){
    clearErrorReport();
    String SQL = viewable.getSelectSQL(getOwner(), where(field.getName(), PK));
    return entryExists(SQL);
  }
  
  /**
   * Determines if an entry exists in the database. 
   * @param SQL SQL to be execited
   * @return true if the SQL statement returns at least one record
   */
  public boolean entryExists(String SQL){
    boolean retVal = false;
    try{
      Connection con = getGuestConnection(); 
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(SQL);
      
      retVal = rs.next();
      
      rs.close();
      st.close();
      con.close();
    }
    catch(Exception e){
      addToErrorReport("Could not confirm that record exists: " + e.getMessage());
    }
    return retVal;
  }
  
  /**
   * Deletes a record from the database.
   * @param table table from which the record is to be deleted
   * @param PKField field in which the PK of this record will be found
   * @param PK primary key of the record to be deleted
   * @return true if successful (if no error occurred), false if not
   */
  public boolean deleteItem(ABCMdbTable table, ABCMdbField PKField, long PK){
    clearErrorReport();
    boolean retVal = false;
    try{
      Connection con = getGuestConnection();
      Statement st = con.createStatement();
      
      String sql = table.getDeleteItemSQL(getOwner(), where(PKField.getName(), PK));
      st.executeUpdate(sql);
      
      retVal = true;
      
      st.close();
      con.close();      
    }
    catch(Exception e){
      addToErrorReport("Could not delete item from " + table.getName() + ": " + e.getMessage());
    }
    return retVal;
  }

  
  // Some simple lookups
  /**
   * Gets a true-or-false item from the database
   * @param viewable Viewable in which the item is found
   * @param numericField numeric field that can be used to identify the requested record
   * @param numericValue value used for identification
   * @param lookupField field in which boolean value will be found
   * @return the true or false value
   */
  public boolean getBooleanItem(ABCMdbAbstractViewable viewable, ABCMdbField numericField, long numericValue, ABCMdbField lookupField){
    return (getStringItem(viewable, numericField, numericValue, lookupField).compareToIgnoreCase(sqlSyntaxEngine.getBooleanTrueReturnString()) == 0);
  }
  
  /**
   * Gets a true-or-false item from the database
   * @param viewable Viewable in which the item is found
   * @param textField numeric field that can be used to identify the requested record
   * @param comparand text sought
   * @param lookupField field in which boolean value will be found
   * @return the value obtained (if no error occurs)
   */
  public boolean getBooleanItem(ABCMdbAbstractViewable viewable, ABCMdbField textField, String comparand, ABCMdbField lookupField){
    return (getStringItem(viewable, textField, comparand, lookupField).compareToIgnoreCase(sqlSyntaxEngine.getBooleanTrueReturnString()) == 0);
  }
  
  /**
   * Retrieves a boolean value from the database
   * @param sql SQL used to retrieve value
   * @param lookupFieldName field in which value will be found
   * @return the value obtained (if no error occurs)
   */
  public boolean getBooleanItem(String sql, String lookupFieldName){
    return (getStringItem(sql, lookupFieldName).compareToIgnoreCase(sqlSyntaxEngine.getBooleanTrueReturnString()) == 0);    
  }
  
  /**
   * Gets a specific numeric item from the database. The item will be identified by a unique
   * numeric key (possibly the PK).
   * @param viewable the Table or View in which the item will be sought
   * @param numericField field containing the key
   * @param numericValue key to be matched
   * @param lookupField field in which value sought will be found
   * @return the requested value
   */
  public long getNumericItem(ABCMdbAbstractViewable viewable, ABCMdbField numericField, long numericValue, ABCMdbField lookupField){
    clearErrorReport();
    String result = "nothing";
    try{
      result = getStringItem(viewable, numericField, numericValue, lookupField);
      return Long.parseLong(result);
    }
    catch(Exception e){
      addToErrorReport("Could not retrieve numeric item (retrieved: " + result + "); possibly a non-numeric field was specified.");
    }
    return -1;
  }
  
  /**
   * Gets a specific numeric item from the database. Item will be identified by a unique text key.
   * @param viewable Viewable in which item will be found
   * @param stringField field in which key appears
   * @param comparand value for key
   * @param lookupField field in which requested value will be found
   * @return the requested value
   */
  public long getNumericItem(ABCMdbAbstractViewable viewable, ABCMdbField stringField, String comparand, ABCMdbField lookupField){
    clearErrorReport();
    String result = "nothing";
    try{
      result = getStringItem(viewable, stringField, comparand, lookupField);
      return Long.parseLong(result);
    }
    catch(Exception e){
      addToErrorReport("Could not retrieve numeric item (retrieved: " + result + "); possibly a non-numeric field was specified.");
    }
    return -1;
  }
  
  /**
   * Gets a specific numeric item from the database
   * @param SQL SQL to be used to retrieve item
   * @param lookupFieldName field in which item will be found
   * @return the value requested
   */
  public long getNumericItem(String SQL, String lookupFieldName){
    clearErrorReport();
    String result = "nothing";
    try{
      result = getStringItem(SQL, lookupFieldName);
      return Long.parseLong(result);
    }
    catch(Exception e){
      addToErrorReport("Could not retrieve numeric item (retrieved: " + result + "); possibly a non-numeric field was specified.");
    }
    return -1;
  }
  
  /**
   * Gets a text item from the database; item will be identified by a unique numeric value
   * @param viewable Viewable in which item is found
   * @param numericField name of field in which text ID is to be found 
   * @param numericValue value against which IDs are compared
   * @param lookupField field in which desired return value is contained
   * @return the value requested
   */
  public String getStringItem(ABCMdbAbstractViewable viewable, ABCMdbField numericField, long numericValue, ABCMdbField lookupField){
    clearErrorReport();
    String numericFieldName = numericField.getName();
    String lookupFieldName  = lookupField.getName();
    String SQL = viewable.getSelectSQL(getOwner(), where(numericFieldName, numericValue));
    return getStringItem(SQL, lookupFieldName);    
  }
  
  /**
   * Gets a text item from the database; item will be identified by a unique text value
   * @param viewable Viewable in which item is found
   * @param stringField name of field in which text ID is to be found 
   * @param comparand text against which IDs are compared
   * @param lookupField field in which desired return value is contained
   * @return the value requested
   */
  public String getStringItem(ABCMdbAbstractViewable viewable, ABCMdbField stringField, String comparand, ABCMdbField lookupField){
    clearErrorReport();
    String stringFieldName = stringField.getName();
    String lookupFieldName = lookupField.getName();
    String SQL = viewable.getSelectSQL(getOwner(), where(stringFieldName, comparand));
    return getStringItem(SQL, lookupFieldName);    
  }  

  /**
   * Retrieves a text item from the database.
   * @param SQL SQL to be used to retrieve item
   * @param lookupFieldName field in which item will be found
   * @return the value requested
   */
  public String getStringItem(String SQL, String lookupFieldName){
    String retVal = null; 
    try{
      Connection con = getGuestConnection(); 
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(SQL);
      
      if(rs.next()) retVal = rs.getString(lookupFieldName);
      
      rs.close();
      st.close();
      con.close();
      
    }
    catch(Exception e){
      addToErrorReport("Could not retrieve String item: " + e.getMessage());
    }
    return retVal;
  }
  
  /**
   * Updates an item in the database to a new numeric value.
   * @param table table in which the value is found
   * @param PKField field that contains the primary key for this table
   * @param PK primary key of the record to be updated
   * @param valueField field that contains the value to be updated
   * @param newValue new value to be assigned to that field
   * @return true if successful
   */  
  public boolean updateItem(ABCMdbTable table, ABCMdbField PKField, long PK, ABCMdbField valueField, double newValue){
    boolean successFlag = false;
    try{
      String sql = "UPDATE " + table.getFullName(getOwner()) + " " +
                   "SET " + valueField.getName() + " = " + newValue + " " +
                   "WHERE " + PKField.getName() + " = " + PK +
                   ";";
      Connection con = getGuestConnection(); 
      Statement st = con.createStatement();
      st.execute(sql);
      successFlag = true;
      st.close();
      con.close();
    }
    catch(Exception e){
      addToErrorReport("An error occurred while updating value: " + e.getMessage());
    }
    return successFlag;
  }
  
  /**
   * Updates an item in the database to a new text value.
   * @param table table in which the value is found
   * @param PKField field that contains the primary key for this table
   * @param PK primary key of the record to be updated
   * @param valueField field that contains the value to be updated
   * @param newValue new value to be assigned to that field
   * @return true if successful
   */
  public boolean updateItem(ABCMdbTable table, ABCMdbField PKField, long PK, ABCMdbField valueField, String newValue){
    boolean successFlag = false;
    try{
      String sql = "UPDATE " + table.getFullName(getOwner()) + " " +
                   "SET " + valueField.getName() + " = '" + AbstractDBGateway.makeSQLSafe(newValue) + "' " +
                   "WHERE " + PKField.getName() + " = " + PK +
                   ";";
      Connection con = getGuestConnection(); 
      Statement st = con.createStatement();
      st.execute(sql);
      successFlag = true;
      st.close();
      con.close();
    }
    catch(Exception e){
      addToErrorReport("An error occurred while updating value: " + e.getMessage());
    }
    return successFlag;
  }
  
  
  
  /**
   * Gets the current error report
   * @return the current error report.
   */
  public String getErrorReport(){ return errorReport; }
 
  /**
   * Clears the error report.
   */
  public void clearErrorReport(){ errorReport = ""; }
  
  /**
   * Add to the error report. Automatically appends a line separator.
   * @param addendum text to be added
   */
  protected void addToErrorReport(String addendum){
    errorReport += " " + addendum + System.getProperty("line.separator");
  }
  
  
  public Collection<String> getListOfAllTablesInDB(){
    System.out.println("RUNNING LIST TABLES");
    clearErrorReport();
    String sql = sqlSyntaxEngine.getRetrieveNamesOfAllTablesSQL();
    Connection con;
    Statement st;
    ResultSet rst;
    Vector<String> result = new Vector<String>();
    try{
      con = getAdminConnection();
      st = con.createStatement();
      rst = st.executeQuery(sql);
      while(rst.next()){
        result.add(rst.getString(1));
      }
      rst.close();
      st.close();
      con.close();
    }
   catch(Exception e){
      addToErrorReport("Could not retrieve list of tables: " + e.getMessage());
    }
   return result;
  }
  
  public boolean backupAllTablesInDB(String outputFolder){
    Collection<String> tables = getListOfAllTablesInDB();
    Connection con;
    Statement st;
    StringWriter out = new StringWriter();
    try{
      con = getAdminConnection();
      for(String table: tables){
        // Create the output file
        System.out.print("Backing up: " + table + "...");
        out.write("\n\nTABLE: " + table + "\n");
        File F = new File(outputFolder + "\\" + table + ".csv");
        FileWriter FW = new FileWriter(F);
        
        // Retrieve all records
        String sql = "SELECT * FROM " + table + ";";
        ResultSet rst;
        st = con.createStatement();
        rst = st.executeQuery(sql);
        ResultSetMetaData meta = rst.getMetaData();
        int cols = meta.getColumnCount();
        
        // Headers
        String row = "";
        for(int i = 1; i <= cols; i++){
          row += meta.getColumnName(i) + (i < cols ? "," : "\n");   
        }
        out.write(row);
        FW.write(row);
        while(rst.next()){
          row = "";
          for(int i = 1; i <= cols; i++){
            switch(meta.getColumnType(i)){
              case java.sql.Types.BIT:
              case java.sql.Types.BOOLEAN:
                row += (rst.getBoolean(i) ? "true" : "false");
                break;
              case java.sql.Types.DATE:
                row += (DateFormat.getDateTimeInstance().format(rst.getDate(i)));
                break;
              case java.sql.Types.DECIMAL:
              case java.sql.Types.NUMERIC:
                row += ("" + rst.getLong(i));
                break;
              case java.sql.Types.CHAR:
              case java.sql.Types.NCHAR:
              case java.sql.Types.NVARCHAR:
              default:
                row += (rst.getString(i));
                break;
            }
            row +=(i < cols ? "," : "\n");
          }
          out.write(row);
          FW.write(row);
        }
        System.out.println(" done");
        FW.close();
        rst.close();
        st.close();
        
      }
      con.close();
    }
   catch(Exception e){
      addToErrorReport("Could not retrieve list of tables: " + e.getMessage());
      return false;
    }
   System.out.println(out.toString());
   return true;
  }
  
}
