package abcm.db;


/**
 * Interface for all objects that manage data in the HWM
 * database. Provides standard routines for saving and
 * retrieving database objects.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public interface DBManager {
  
  /**
   * Gets an XML version of the complete list of all objects specified.
   * 
   * @return an XML formatted data package for the list of
   * all objects of the required type
   */
  public abstract String getListXML();
  
  /**
   * Gets an XML version of the instance specified. This is the default
   * method; it includes the XML header but does not include comments
   * about the object.
   * 
   * @param PK Primary Key of the instance specified in the HWM database
   * @return XML formatted data package for this instance
   */
  public abstract String getInstanceXML(long PK);

  /**
   * Gets an XML version of the instance specified. This version
   * of the method allows the calling method to specify whether XML headers
   * should be included or not. Comments about the object are not included.
   * 
   * In general, this version of the method would be used in instances
   * where the XML is going to be included in another XML document.
   * 
   * @param PK Primary Key of the instance specified in the HWM database
   * @param includeHeader true if XML headers should be included, false if not
   * @return XML formatted data package for this instance
   */
  public abstract String getInstanceXML(long PK, boolean includeHeader);
  
  /**
   * Gets an XML version of the instance specified. This version
   * of the method allows the user to specify whether XML headers
   * should be included or not. This version also allows the calling method
   * to specify whether comments about this object should be included.
   * 
   * @param PK Primary Key of the instance specified in the HWM database
   * @param includeHeader true if XML headers should be included, false if not
   * @param includeComments true if comments should be included, false if not
   * @return XML formatted data package for this instance
   */
  public abstract String getInstanceXML(long PK, boolean includeHeader, boolean includeComments);
  
  /**
   * Gets an HTML version of the complete list of all objects
   * specified.
   * 
   * @param includeHeaders true if the HTML version should include all the HTML
   * headers for a normal web page.
   * @param xslRootDir root of directory in which xsl sheets are to be found
   * @return HTML version of this list of objects
   */
  public abstract String getListHTML(boolean includeHeaders, String xslRootDir);

  /**
   * Gets an HTML version of a specific instance of the specified
   * kind of object.
   * 
   * @param PK Primary Key in the database for the object requested
   * @param includeHeaders true if the HTML version should include all the HTML
   * headers for a normal web page.
   * @param xslRootDir root of directory in which xsl sheets are to be found
   * @return HTML version of the specified instance of the required object
   */  
  public abstract String getInstanceHTML(long PK, boolean includeHeaders, String xslRootDir);
  
  /**
   * Saves an instance, previously set, of the required type of object
   * to the HWM Database.
   * @return true if successful, false if not
   */
  public abstract boolean saveInstanceToDB();
  
  /**
   * Returns a String containing a message describing the most recent
   * error to have been encountered by this manager object.
   * 
   * @return a message containing the error description
   */
  public abstract String getErrorReport();
  
}
