package abcm.db.components;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import abcm.db.syntaxengines.AbstractSQLSyntaxEngine;
import abcm.db.syntaxengines.SQLDataTypes;

/**
 * An ABCM Viewable is either a Table or a View in the associated database.
 * Both are represented by a collection of fields and can in many ways be
 * treated the same.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class ABCMdbAbstractViewable {
  
    /** Name for this Viewable  */
    protected String    name;
    /** Collection of fields in this Viewable  */
    protected Hashtable fieldList        = new Hashtable();
    /**  Ordered collection of fields in this Viewable */
    protected Vector    orderedFieldList = new Vector(); 
    
    // Some useful data type instances
    protected static SQLDataTypes.SQLDataType    INT       = new SQLDataTypes.SQLINTDataType();
    protected static SQLDataTypes.SQLDataType    VCHAR10   = new SQLDataTypes.SQLVCHARDataType(10);
    protected static SQLDataTypes.SQLDataType    VCHAR16   = new SQLDataTypes.SQLVCHARDataType(16);
    protected static SQLDataTypes.SQLDataType    VCHAR20   = new SQLDataTypes.SQLVCHARDataType(20);
    protected static SQLDataTypes.SQLDataType    VCHAR25   = new SQLDataTypes.SQLVCHARDataType(25);    
    protected static SQLDataTypes.SQLDataType    VCHAR30   = new SQLDataTypes.SQLVCHARDataType(30);
    protected static SQLDataTypes.SQLDataType    VCHAR40   = new SQLDataTypes.SQLVCHARDataType(40);
    protected static SQLDataTypes.SQLDataType    VCHAR50   = new SQLDataTypes.SQLVCHARDataType(50);
    protected static SQLDataTypes.SQLDataType    VCHAR100  = new SQLDataTypes.SQLVCHARDataType(100);
    protected static SQLDataTypes.SQLDataType    VCHAR200  = new SQLDataTypes.SQLVCHARDataType(200);
    protected static SQLDataTypes.SQLDataType    VCHAR255  = new SQLDataTypes.SQLVCHARDataType(255);
    protected static SQLDataTypes.SQLDataType    TEXT      = new SQLDataTypes.SQLTEXTDataType();
    protected static SQLDataTypes.SQLDataType    DATETIME  = new SQLDataTypes.SQLDateTimeDataType();
    protected static SQLDataTypes.SQLDataType    FLOAT8    = new SQLDataTypes.SQLFloatDataType(8);
    protected static SQLDataTypes.SQLDataType    BIT       = new SQLDataTypes.SQLBooleanDataType();
    protected static SQLDataTypes.SQLDataType    CHAR16    = new SQLDataTypes.SQLCHARDataType(16);
    
    /**
     * Constructor
     * @param viewableName name for this Viewable object
     */
    public ABCMdbAbstractViewable(String viewableName){
      name = viewableName;
    }    
    
    /**
     * Adds the specified field to this Viewable
     * @param field field to be added
     * @return the field added
     */
    public ABCMdbField addField(ABCMdbField field){
      fieldList.put(field.getName(), field);
      orderedFieldList.add(field);
      field.setContainer(this);
      return field;
    }
    
    /**
     * Adds a field to this Viewable
     * @param fieldName name of this field
     * @param allowsNulls flag indicating whether field will accept null values
     * @param dataType data type for the field
     * @return the Field object created
     */
    public ABCMdbField createAndAddField(String fieldName, boolean allowsNulls, SQLDataTypes.SQLDataType dataType){
      return addField(new ABCMdbField(fieldName, allowsNulls, dataType));
    }
    
    
    /**
     * Adds a field to this Viewable
     * @param fieldName name of this field
     * @param dataType data type for the field
     * @param isPrimaryKey flag indicating whether this field is a primary key
     * @return the Field object created
     */
    public ABCMdbField createAndAddField(String fieldName, SQLDataTypes.SQLDataType dataType, boolean isPrimaryKey){
      return addField(new ABCMdbField(fieldName, dataType, isPrimaryKey));
    }
    
    /**
     * Adds a field to this Viewable
     * @param fieldName name of the field
     * @param allowsNulls flag indicating whether field will accept null values
     * @param dataType data type for the field
     * @param foreignKeyTable table to which this field is linked as a foreign key
     * @param foreignKeyField field to which this field is linked as a foreign key
     * @return the Field object created
     */
    public ABCMdbField createAndAddField(String fieldName, boolean allowsNulls, SQLDataTypes.SQLDataType dataType, String foreignKeyTable, String foreignKeyField){
      return addField(new ABCMdbField(fieldName, dataType, foreignKeyTable, foreignKeyField, allowsNulls));
    }
        
    /**
     * Adds a field to this Viewable
     * @param fieldName name of the field
     * @param allowsNulls flag indicating whether field will accept null values
     * @param dataType data type for the field
     * @param isIdentityOnly flag indicating whether the field is an identity field (auto-generated ID)
     * @return the Field object created
     */
    public ABCMdbField createAndAddField(String fieldName, boolean allowsNulls, SQLDataTypes.SQLDataType dataType, boolean isIdentityOnly){
      return addField(new ABCMdbField(fieldName, allowsNulls, dataType, isIdentityOnly));
    }
    
    /**
     * Gets the name of this object
     * @return this object's name
     */
    public String getName(){ return name; }
    
    /**
     * Gets the full name of this object; useful for constructing SQL
     * @param dbOwner the database owner
     * @return the full name of the object
     */
    public String getFullName(String dbOwner){ 
      return (dbOwner.length() > 0 ? dbOwner + "." + name : name); 
    }
    
    /**
     * Child classes must provide a means to generate a 'CREATE' statement in SQL that creates the
     * specified object.
     * These may include multiple SQL statements that must (depending on the engine) be
     * executed separately and in order.
     * @param dbOwner the name of the database owner 
     * @return the SQL statement that will create the object
     */
    public abstract String[] getCreateSQL(String dbOwner, AbstractSQLSyntaxEngine syntaxEngine);
    
    /**
     * Gets a SELECT statement that returns all records in this Viewable
     * @param dbOwner the name of the database owner
     * @return the SQL SELECT statement
     */
    public String getSelectAllSQL(String dbOwner){  return "SELECT * FROM " + getFullName(dbOwner);      }    
    
    /**
     * Gets a SELECT statement that returns all records in this Viewable
     * @param dbOwner the name of the database owner
     * @param clause any clause to be appended to the SQL (WHERE, ORDER BY)
     * @return the SQL SELECT statement
     */
    public String getSelectSQL(String dbOwner, String clause){ return getSelectAllSQL(dbOwner) + " " + clause; }
    
    /**
     * Gets the field specified by the field name.
     * @param fieldName the name of the field to be retrieved
     * @return the requested field
     */
    public ABCMdbField getField(String fieldName){   return (ABCMdbField)fieldList.get(fieldName);                   }

    /**
     * Gets a comma-separated list of the fields in this Viewable
     * @return the list of fields as a nicely formatted String
     */
    public String getSeparatedFieldList(){
      String fl = "";
      Enumeration e = fieldList.elements();
      while(e.hasMoreElements()){
        ABCMdbField F = (ABCMdbField)e.nextElement();
        fl += F.getName();
        if(e.hasMoreElements()) fl += ", ";
      }
      return fl;
    }
    
}
