package abcm.db.components;

import abcm.db.syntaxengines.AbstractSQLSyntaxEngine;
import abcm.db.syntaxengines.SQLDataTypes;

/**
 * An ABCMdbField represents a field in either a Table or a View (actually, in a Viewable). It can provide conveniences
 * for accessing fields and their values in the database and in result sets. 
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ABCMdbField {
  
  /** The name of this field  */
  private String  fieldName;
  /** The data type for this field  */
  private SQLDataTypes.SQLDataType  dataType;
  /** If this field is linked to another table, the name of the foreign table  */
  private String  FK_Table;
  /**  If this field is linked to another table, the name of the foreign key field in that table */
  private String  FK_Field;
  /** Flag indicating if this field is a primary key field */
  private boolean primaryKey = false;
  /** Flag if this field is an 'Identity' field (automatically generated ID numbers in SQL Server) */
  private boolean isIdentityOnly = false;
  /** Flag if this field allows null values  */
  private boolean allowNulls = false;
  /** Viewable in which this field occurs  */
  private ABCMdbAbstractViewable inViewable;
  /** True if the field is numeric, false otherwise  */
  private boolean numeric = false;
    
  /**
   * Constructor
   * @param name Name of this field
   */
  public ABCMdbField(String name){
    fieldName = name;
  }
    
  /**
   * Constructor
   * @param name Name of this field
   * @param allowsNulls flag if the field should allow nulls
   * @param type String identifying the type for this field
   */
  public ABCMdbField(String name, boolean allowsNulls, SQLDataTypes.SQLDataType type){
    fieldName  = name;
    dataType   = type;
    allowNulls = allowsNulls;
  }
  
  /**
   * Constructor
   * @param name Name of this field
   * @param allowsNulls flag if the field should allow nulls
   * @param type String identifying the type for this field
   * @param isIdentity true if this is an identity field (auto-generated IDs) 
   */
  public ABCMdbField(String name, boolean allowsNulls, SQLDataTypes.SQLDataType type, boolean isIdentity){
    System.out.println("RUNNING CORRECT CONSTRUCTOR..." + name + " " + isIdentity);
    fieldName      = name;
    dataType       = type;
    allowNulls     = allowsNulls;
    isIdentityOnly = isIdentity;
  }
  
  /**
   * Constructor
   * @param name Name of this field
   * @param type String identifying the type for this field
   * @param isPrimaryKey true if this field will serve as the primary key 
   */
  public ABCMdbField(String name, SQLDataTypes.SQLDataType type, boolean isPrimaryKey){
    fieldName  = name;
    dataType   = type;
    primaryKey = isPrimaryKey;
  }
  
  /**
   * Constructor
   * @param name Name of this field
   * @param type String identifying the type for this field
   * @param foreignKeyTable table to which this field is linked
   * @param foreignKeyField field to which this field is linked
   * @param allowsNulls 
   */
  public ABCMdbField(String name, SQLDataTypes.SQLDataType type, String foreignKeyTable, String foreignKeyField, boolean allowsNulls){
    fieldName  = name;
    dataType   = type;
    FK_Table   = foreignKeyTable;
    FK_Field   = foreignKeyField;
    allowNulls = allowsNulls;
  }    
  
  /**
   * Sets the containing view for this field
   * @param container Viewable within which this field exists
   */
  public void setContainer(ABCMdbAbstractViewable container){ inViewable = container; }
  
  /**
   * Returns the full name of this field, including the database owner.
   * @param dbOwner name of the db owner
   * @return full name of the field
   */
  public String getFullFieldName(String dbOwner){
    if(inViewable == null){
      System.out.println("Full Field Name inViewable is null!");
    }
    return inViewable.getFullName(dbOwner) + "." + fieldName; 
  } 
    
  /**
   * Returns the (simple) name of this field
   * @return the name of this field
   */
  public String getName(){ return fieldName; }

  /**
   * Returns the name of this field with brackets and dots; useful in SQL (i.e. [MyTable].[MyField])
   * @return the 'dotted name' for this field
   */
  public String getDottedName(){
    return( (inViewable != null ? "[" + inViewable.getName() + "]." : "") + "[" + fieldName + "]");
  }
  
  /**
   * Gets a String that generates the SQL required to create this field
   * @return the SQL to create this field
   */
  public String getCreateRepresentation(AbstractSQLSyntaxEngine syntaxEngine){
    String r = "";
    System.out.println("Name:" + fieldName + " Identity only? " + isIdentityOnly);
    if(!isIdentityOnly ) r += fieldName + " " + syntaxEngine.getDataTypeCreateRepesentation(dataType) + (primaryKey     ? " " + syntaxEngine.getPrimaryKeyIdentityFieldCreationKeywords() : "");
    else                 r += fieldName + " " + syntaxEngine.getDataTypeCreateRepesentation(dataType) + (isIdentityOnly ? " " + syntaxEngine.getIdentityOnlyFieldCreationKeywords() : "");
    // Previously creating a foreign key was done as part of the field syntax; this is applicable in Transact-SQL (MS-SQL Server)
    // but not in MySQL, and it's bad design, so it's being removed here. Creating foreign keys will be done as ALTER TABLE statements
//    if((FK_Table != null) && (FK_Field != null) ){
//      r += " " + syntaxEngine.getBasicForeignKeyConstraint(fieldName, FK_Table, FK_Field) ;
//    }
    if(allowNulls) r += " " + (allowNulls ? "NULL" : "NOT NULL");
    return r; 
  }
  
  /**
   * Causes the Field object to reflect that allow nulls is true 
   */
  public void allowNulls(){ allowNulls = true; }
  
  /**
   * Returns true if this field is numeric
   * @return true if this field is numeric
   */
  public boolean isNumeric(){ return numeric; }
 
  public boolean isFK(){ return (FK_Table != null); }
  
  public String getFKTable(){ return FK_Table; }
  public String getFKField(){ return FK_Field; }
  public SQLDataTypes.SQLDataType getDataType(){ return dataType; }
}
