package abcm.db.components;

import java.util.Enumeration;
import java.util.Hashtable;

import abcm.db.syntaxengines.AbstractSQLSyntaxEngine;

/**
 * General class for representing a table 
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class ABCMdbTable extends ABCMdbAbstractViewable {

  /* Static Methods */
  
  /**
   * Gets the given list of fields as a formatted string (comma separated)
   * @param fields fields to list
   * @return the formatted list
   */
  private static String getFieldList(String[] fields){
    String fl = "";
    for(int i = 0; i < fields.length; i++){
      fl += fields[i] + (i < (fields.length - 1) ? ", " : "");
    }
    return fl;
  }
  
  /* Inner Classes */
  /**
   * An ABCMdbIndex can be created and used when the database
   * is initialized to set constraints and relationships on tables.
   * Indexes are always 'unique nonclustered' 
   *
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class ABCMdbIndex{
    
      /** Index Name  */
      private String   indexName;
      /** Collection of fields embraced by this index  */
      private String[] fields;
      
      /**
       * Constructor
       * @param name index name
       * @param fieldList field name(s)
       */
      public ABCMdbIndex(String name, String ... fieldList){
        indexName = name;
        fields    = fieldList;
      }
      
      /**
       * Gets the SQL required to create this index.
       * @param dbOwner database owner name
       * @return text of 'CREATE INDEX' statement
       */
      protected String getCreateSQL(String dbOwner, AbstractSQLSyntaxEngine syntaxEngine){
        return syntaxEngine.getCreateIndexStatement(indexName, getFullName(dbOwner), fields); 
      }      
      
      /**
       * Gets the index name
       * @return the index name
       */
      protected String getName(){ return indexName; }
    }
    
    /**
     * Represents a foreign key that may involve more than one field
     * 
     *  @author John T. Murphy<br/>
     *  Global Institute of Sustainability,<br/>
     *  Arizona State University<br/>
     *  For license and disclaimers, see the Overview.
     */
    protected class ABCMdbForeignKey{
      
      /** Name for this  */
      private String FKName;
      /** Table to which relationship is made  */
      private String FKTable;
      /** Fields in the foreign table  */
      private String[] FKFields;
      /** Fields in this table  */
      private String[] MyFields;
      
      /**
       * Constructor
       * @param fkName name
       * @param fkTable foreign table
       * @param fieldList fields in foreign table
       * @param myFieldList fields in this table
       */
      public ABCMdbForeignKey(String fkName, String fkTable, String[] fieldList, String[] myFieldList){
        FKName = fkName;
        FKTable = fkTable;
        FKFields = fieldList;
        MyFields = myFieldList;
      }
      
      /**
       * Gets the SQL required to create this relationship
       * @return the SQL to create this relationship
       */
      public String getCreateSQL(AbstractSQLSyntaxEngine syntaxEngine){
        return syntaxEngine.getForeignKeyCreateConstraintStatement(getName(), FKName, MyFields, FKTable, FKFields);
      }      
      
      /**
       * Gets the name of this foreign key
       * @return the name of this foreign key
       */
      public String getMultiFieldFKName(){ return FKName; }
    }
       
    /* Instance Variables */
    
    /** Collection of indexes in this table  */
    protected Hashtable indexList          = new Hashtable();
    /** Collection of foreign keys in this table  */
    protected Hashtable fkList             = new Hashtable();
    /** List of multiple-fields in the primary key  */
    protected String[] multiFieldPKFieldList = null;
    
    /**
     * Constructor
     * @param tableName name of the table
     */
    public ABCMdbTable(String tableName){
      super(tableName);      
    }
    
    
        
    /**
     * Adds an index. Name of index must not duplicate existing index; if it does,
     * no action will be taken.
     * @param index index to be added
     */
    public void addIndex(ABCMdbIndex index){
      if(!indexList.containsKey(index.getName())) indexList.put(index.getName(), index);
    }
    
    public ABCMdbField addField(ABCMdbField field){
      super.addField(field);
      if(field.isFK()) this.addFK(new ABCMdbForeignKey(getName() + "FK" + fkList.size(), 
          field.getFKTable(), new String[]{field.getFKField()}, new String[]{field.getName()}));
      return field;
    }
    
    /**
     * Adds a foreign key
     * @param fk foreign key to be added
     */
    public void addFK(ABCMdbForeignKey fk){
      if(!fkList.containsKey(fk.getMultiFieldFKName())){
        fkList.put(fk.getMultiFieldFKName(), fk);
      }
    }
    
    /**
     * Sets the multi-field primary key list for this table.
     * @param fieldList list to be used
     */
    public void setMultiFieldPKFieldList(String ... fieldList){
      multiFieldPKFieldList = fieldList;
    }
    
    /** 
     * @see abcm.db.components.ABCMdbAbstractViewable#getCreateSQL(java.lang.String) 
     */
    public String[] getCreateSQL(String dbOwner, AbstractSQLSyntaxEngine syntaxEngine){
      String[] fieldCreationRepresentations = new String[orderedFieldList.size()];
      for(int i = 0; i < fieldCreationRepresentations.length; i++)
          fieldCreationRepresentations[i] = ((ABCMdbField)orderedFieldList.get(i)).getCreateRepresentation(syntaxEngine);
      
      String[] indexCreationRepresentations = new String[indexList.size()];
      {
        int i = 0;
        Enumeration e = indexList.elements();
        while(e.hasMoreElements()){
          indexCreationRepresentations[i] = (((ABCMdbIndex)e.nextElement()).getCreateSQL(dbOwner, syntaxEngine) + ";");
          i++;
        }          
      }

      // Some syntax engines require single-field FKs to be added separately,
      // others don't.
      String[] foreignKeyCreationRepresentations = new String[fkList.size()];
      {
        int i = 0;
        Enumeration e = fkList.elements();
        while(e.hasMoreElements()){
          foreignKeyCreationRepresentations[i] = (((ABCMdbForeignKey)e.nextElement()).getCreateSQL(syntaxEngine));
          i++;
        }
      }
      
      
      return syntaxEngine.getCreateTableSQL(dbOwner, name, fieldCreationRepresentations, 
          multiFieldPKFieldList, indexCreationRepresentations, foreignKeyCreationRepresentations);
          
//      
//      
//      Vector<String> sqls = new Vector<String>();
//      String sql = "";
//      sql += "CREATE TABLE " + name + " (\n"; 
//      for(int i = 0; i < orderedFieldList.size(); i++){
//        ABCMdbField F = (ABCMdbField)orderedFieldList.get(i);
//        sql += "       " + F.getCreateRepresentation(syntaxEngine) + ((i < (orderedFieldList.size() - 1)) ? ", " : "") + "\n";
//      }
//      
//      
//      if(multiFieldPKFieldList != null){
//        sql += "\nCONSTRAINT " + getName() + "PK PRIMARY KEY NONCLUSTERED (" + multiFieldPKFieldList + ")\n";
//      }
//      sql += ");";
//      sqls.add(sql);
//      // And collect the indexes; these will require separate SQL statements
//      Enumeration e2 = indexList.elements();
//      while(e2.hasMoreElements()){
//        sqls.add(((ABCMdbIndex)e2.nextElement()).getCreateSQL(dbOwner, syntaxEngine) + ";");
//      }
//      // Multiple-Field Foreign Keys also (apparently) require separate SQL statements
//      Enumeration e1 = fkList.elements();
//      
//      while(e1.hasMoreElements()){
//        abcmMultiFieldFK FK = (abcmMultiFieldFK)e1.nextElement();
//        sqls.add(FK.getCreateSQL());
//      }
//      String[] ret = new String[sqls.size()];
//      for(int i = 0; i < ret.length; i++) ret[i]= sqls.get(i);
//      return ret; 
    }
    
    /**
     * Returns SQL that can insert values into this table.
     * @param fieldsAndValues map of fields and the values to be inserted into each field
     * @param dbOwner name of the database owner
     * @return the SQL that will insert the values
     */
    public String getInsertSQL(Hashtable fieldsAndValues, String dbOwner, AbstractSQLSyntaxEngine syntaxEngine){
      String columnList = "( ";
      String valueList  = "( ";
      
      Enumeration e = fieldsAndValues.keys();
      while(e.hasMoreElements()){
        ABCMdbField key = (ABCMdbField)e.nextElement();
        String col = key.getName();
        String val = (String)fieldsAndValues.get(key);
        columnList += col + (e.hasMoreElements() ? ", " : ") ");
        valueList  += syntaxEngine.getInsertFormat(key.getDataType(), val); // "'" + val + "'";
        valueList  += (e.hasMoreElements() ? ", " : ") ");
      }
      
      String sql = "";
      sql += "INSERT INTO " + getFullName(dbOwner) + " " + columnList + " VALUES " + valueList;
      return sql;
    }

    /**
     * Gets SQL that can delete an item from this table
     * @param dbOwner name of the database owner
     * @param whereClause where clause limiting delete to specific records
     * @return the SQL that executes the delete
     */
    public String getDeleteItemSQL(String dbOwner, String whereClause){
      return "DELETE FROM " + this.getFullName(dbOwner) + " " + whereClause;
    }

}
