package abcm.db.managers;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Vector;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.math.SimDate;
import abcm.sim.Analysis;
import abcm.sim.SimulationDataSeries;
import abcm.utilities.XMLAssist;

/**
 * Manages the Analyses stored in the HWM Database.
 * 
 * Provides XML transformation and other routines, including generating images
 * of Analysis graphs.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class AnalysisManager extends AbstractAuditableDBManager {

  /** Name of analysis to be saved              */  
  private String analysisNameToSave;
  /** Description of analysis to be saved       */  
  private String descriptionToSave;
  /** ID of owner of this analysis to be saved  */  
  private long   researcherIDToSave;

  /** Internal manager to manage child table 'Analysis Data Series'  */
  private AnalysisDataSeriesManager ADSM;
  
  /* Inner class */
  
  /**
   * Manages data series associated with analyses
   */
  private class AnalysisDataSeriesManager extends AbstractDBLinkTableInnerManager{

    /** FK of analysis to save */
    private long    analysisFKToSave;
    
    /** FK of data series to save */
    private long    dataSeriesFKToSave; 
    
    /**
     * @param db
     * @param containerClass
     */
    public AnalysisDataSeriesManager(AbstractABCMDBGateway db, AbstractDBManager containerClass) {      super(db, containerClass);    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
     */
    protected ABCMdbTable baseTable() {                     return DB.getAnalysisDataSeriesTable();                                   }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
     */
    protected ABCMdbField baseTablePKField() {              return DB.getAnalysisDataSeriesTable().getField_PK();                     }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
     */
    protected ABCMdbField baseTablePrimaryFKField() {       return DB.getAnalysisDataSeriesTable().getField_AnalysisFK();             }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
     */
    protected ABCMdbField baseTableSecondaryFKField() {     return DB.getAnalysisDataSeriesTable().getField_DataSeriesFK();           }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
     */
    protected ABCMdbAbstractViewable recordsView() {        return DB.getAnalysisDataSeriesView();                                    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
     */
    protected ABCMdbField recordsViewPKField() {            return DB.getAnalysisDataSeriesView().getField_AnalysisDataSeries_PK();   }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
     */
    protected ABCMdbField recordsViewPrimaryFKField() {     return DB.getAnalysisDataSeriesView().getField_Analysis_PK();             }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
     */
    protected ABCMdbField recordsViewPrimaryNameField() {   return DB.getAnalysisDataSeriesView().getField_Analysis_Title();          }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
     */
    protected ABCMdbField recordsViewSecondaryNameField() { return DB.getAnalysisDataSeriesView().getField_DataSeries_Name();         }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
      
    }

    
    /* Saving */
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
     */
    protected void clearInstanceToSave() { 
      analysisFKToSave   = -1;
      dataSeriesFKToSave = -1;
    }
    
    /**
     * @param analysis
     * @param dataSeries
     */
    public void setInstanceToSave(long analysis, long dataSeries){
      analysisFKToSave   = analysis;
      dataSeriesFKToSave = dataSeries;
    }
    
    /**
     * Determines if the instance to be saved is valid. If the instance to be saved
     * has not been set, it will not be valid.
     * @return true if the instance to be saved is valid, false otherwise.
     */
    public boolean instanceToSaveIsValid(){
      return (( analysisFKToSave   != -1) && 
              ( dataSeriesFKToSave != -1));
    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
     */
    protected ABCMdbDatum datumToSave() {      
      return (instanceToSaveIsValid() ?
            DB.getAnalysisDataSeriesTable().getDatum("" + analysisFKToSave, "" + dataSeriesFKToSave)
            : null);
    }

    
    /* Options List */
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#getOptionList(long) 
     */
    public String getOptionList(long PK){
      String retVal = "";
      // This code has to be special because no other creation of option list looks quite like this one. 
      String sql = DB.getAnalysisDataSeriesView().getSelectSQL(DB.getOwner(), AbstractDBGateway.where(DB.getAnalysisDataSeriesView().getField_Analysis_PK().getName(),PK));
      try {
        Connection con = DB.getGuestConnection();
        Statement st = con.createStatement();
    
        ResultSet rs = st.executeQuery(sql);
        
//        while (rs.next()) {
//          retVal += "<OPTION VALUE=\"" + rs.getString(DB.getAnalysisDataSeriesView().getField_DataSeries_PK().getName()) + "\">"
//                              + "Run " + rs.getString(DB.getAnalysisDataSeriesView().getField_SimulationRun_PK().getName()) + ": "
//                                       + rs.getString(DB.getAnalysisDataSeriesView().getField_DataSeries_Name().getName()) + "</OPTION>";
//        }
        
        retVal = (new SimulationDataSeriesManager(DB)).createOptionList(rs, DB.getAnalysisDataSeriesView().getField_DataSeries_PK(), 
            DB.getAnalysisDataSeriesView().getField_DataSeries_Name(), DB.getAnalysisDataSeriesView().getField_SimulationRun_PK());
        
        rs.close();
        st.close();
        con.close();
    
      }
      catch (Exception e) {
        addToErrorReport("Could not get data series as option list: " + e.getMessage());    }
      return retVal;
    }
    
    
    // This adds the errors to the containing
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager#addToErrorReport(java.lang.String) 
     */
    public void addToErrorReport(String addendum){ addError(addendum); }
       
  }
  
  
  /* Constructor */
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public AnalysisManager(AbstractABCMDBGateway db){
    super(db);
    ADSM = new AnalysisDataSeriesManager(DB, this);
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){                  return DB.getAnalysesTable();               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){           return DB.getAnalysesTable().getField_PK(); }
  
  /**
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField()
   */
  protected ABCMdbField baseTableDescriptionField(){  return DB.getAnalysesTable().getField_Description(); }
  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){        return DB.getAnalysesAndResearchersView();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){            return DB.getAnalysesAndResearchersView().getField_Analysis_PK();        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){            return DB.getAnalysesAndResearchersView().getField_Analysis_Title();     }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){                return "analyses"; }  
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
    
    while(rs.next()){
      xml.open("analysis");
        xml.addLine("ID", rs.getLong(DB.getAnalysesAndResearchersView().getField_Analysis_PK().getName()));
        xml.addLine("title", rs.getString(DB.getAnalysesTable().getField_Title().getName()));
        xml.addLine("desc", rs.getString(DB.getAnalysesTable().getField_Description().getName()));
        xml.addLine("date", rs.getString(DB.getAnalysesTable().getField_DateCreated().getName()));
        xml.addLine("createdBy", rs.getString(DB.getAnalysesAndResearchersView().getField_Researcher_LastNameFirstName().getName()));
      
        if(singleInstance){  // Only include if this is a single instance; this is different from most DBManagers, but it is
                             // a practical concession, because a collection of data may be 100s of elements long.
          xml.open("analysisData");
          
            Analysis A = this.getAnalysis(instancePK);  // Get the actual analysis because it's easier to use its own methods to sort data
            
            Vector<String> headers = A.getTableHeaders();
            
            // Headers
            xml.open("seriesList");
            
              for(String header: headers){
                SimulationDataSeries S = (SimulationDataSeries)A.getSeries().get(header);
                String seriesName = S.getName(); 
                String field      = S.getDomain();
                xml.open("series");
                  xml.addLine("seriesName", seriesName);
                xml.closeTag();
              }
              
            xml.closeTag();
    
            String noValue = "-";
            
            // Now get the actual data
            TreeMap rows = A.getRowsInSeriesTable();
   
            Iterator I = rows.keySet().iterator();
            while(I.hasNext()){
              SimDate rowDate = (SimDate)I.next();
              Double row[] = (Double[]) rows.get(rowDate);
              xml.open("recordedOn");
                xml.addLine("date", rowDate.getString());
                for(int i = 0; i < row.length; i++){
                  xml.open("datum");
                    xml.addLine("index", i); 
                    xml.addLine("value", (row[i] != null ? ("" + row[i]) : noValue));
                  xml.closeTag();
                }
              xml.closeTag();
              
            }
          xml.closeTag(); // Analysis Data
          if (includeComments) {
            CommentManager cm = new CommentManager(DB);
            xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.ANALYSIS, instancePK), XMLAssist.INDENT_BLOCK);
          }
        }
      xml.closeTag();
    }
  }
  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlAnalysisListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlAnalysisToHTML.xsl"; }
  
  
  /* Saving */
  
  /**
   * Sets the values to be saved as a new Analysis instance
   * @param name name of the analysis. Must not duplicate any previously saved analysis name
   * @param desc description of the analysis
   * @param researcher primary key identifying researcher who owns this analysis
   */
  public void setInstanceToSave(String name, String desc, long researcher) {
    analysisNameToSave = AbstractABCMDBGateway.makeSQLSafe(name);
    descriptionToSave  = AbstractABCMDBGateway.makeSQLSafe(desc);
    researcherIDToSave = researcher;
  }

  /**
   * Gets the status of the instance to save
   * @return true if the instance to save has been set and is valid, false otherwise
   */
  public boolean instanceToSaveIsValid(){    
    return (analysisNameToSave != null);   
  }
    
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){   
    return DB.getAnalysesTable().getDatum(analysisNameToSave, descriptionToSave, "" + researcherIDToSave, DB.currentTimeString());
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    analysisNameToSave = null;
    descriptionToSave = null;
    researcherIDToSave = -1;
  }
  

  /* Auditable */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView(){         return DB.getCanBeEdited_AnalysesView();                          }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField            canBeEditedViewPKField(){  return DB.getCanBeEdited_AnalysesView().getField_Analysis_PK();   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField            baseTableOwnerField(){     return DB.getAnalysesTable().getField_ResearcherFK();             }
  
  /**
   * Deletes an Analysis from the HWM Database. The Data Series that are part of
   * this Analysis remain stored in the HWM Database, but are no longer grouped
   * together
   * 
   * @param PK primary key of the analysis to be deleted
   * @param user userID, to confirm that user has appropriate permissions
   * @return true if successful, false if not
   */
  public boolean deleteItem(long PK, long user) {
    if(!permitModification(PK, user)) return false;
    else                              return ADSM.deleteElementsByPrimaryFK(PK) &&
                                             super.deleteItem(PK, user);
  } 
  
  /* Error Reporting */
  /**
   * Adds an element to the error report for this object
   * @param addendum item to be added
   */
  public void addError(String addendum){ addToErrorReport(addendum); }
  
  /**
   * Updates the description of this Analysis in the HWM database. Will not
   * update an Analysis that cannot be edited
   * 
   * @param PK the primary key of the Analysis to be updated
   * @param desc the new description to be attached to this Analysis
   * @return true if successful, false if not
   */
  public boolean updateDescription(long PK, String desc) {
    if (!canBeEdited(PK)) {
      errorReport = "Could not update description; Analysis cannot be updated";
      return false;
    }
    String sql = "UPDATE " + DB.getAnalysesTable().getFullName(DB.getOwner())+ " " + 
                 "SET " + DB.getAnalysesTable().getField_Description().getName() + 
                         " = '"  + AbstractABCMDBGateway.makeSQLSafe(desc) + "' " + AbstractDBGateway.where(DB.getAnalysesTable().getField_PK().getName(),PK);
    try {
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();

      st.executeUpdate(sql);

    }
    catch (Exception e) {
      addToErrorReport("Could not update description: " + e.getMessage());
      return false;
    }
    return true;
  }

  /**
   * Gets the table of data for a specific Analysis stored in the HWM Database,
   * as an ASCII table. This first retrieves the XML instance of the Analysis,
   * then performs a special XML transform to return pure ASCII; an additional
   * transformation makes it tab delimited with proper cr/lf.
   * 
   * @param PK primary key of the Analysis as stored in the HWM Database
   * @return the ASCII table requested
   */
  public String getInstanceAsASCIITable(long PK) {
    
    Analysis A = this.getAnalysis(instancePK);  // Get the actual analysis because it's easier to use its own methods to sort data
    
    Vector<String> headers = A.getTableHeaders();

    StringWriter out = new StringWriter();
    
    
    
    // Headers
    for(String header: headers){
      SimulationDataSeries S = (SimulationDataSeries)A.getSeries().get(header);
      String seriesName = S.getName(); 
      String field      = S.getDomain();
      out.write(seriesName + " ");
    }
    out.write("\n");
    
    // Now get the actual data
    TreeMap rows = A.getRowsInSeriesTable();

    Iterator I = rows.keySet().iterator();
    while(I.hasNext()){
      SimDate rowDate = (SimDate)I.next();
      Double row[] = (Double[]) rows.get(rowDate);

      out.write(rowDate.getString() + " ");
      
      
      for(int i = 0; i < row.length; i++){
        out.write((row[i] != null ? ("" + row[i]) : "-") + " ");
      }
      out.write("\n");
    }

    return out.toString();
    
  }

  /**
   * Gets a specified analysis from the HWM Database.
   * @param PK Primary Key of the analysis to be retrieved
   * @return the requested HWMAnalysis
   */
  public Analysis getAnalysis(long PK){
    Analysis analysis = new Analysis();
    // Load each data series and its data
    
    try{
      String sql = DB.getAnalysesTable().getSelectSQL(DB.getOwner(), AbstractDBGateway.where(DB.getAnalysesTable().getField_PK().getName(),PK));
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();

      ResultSet rs = st.executeQuery(sql);
      String name;
      if(rs.next()){
        name = rs.getString(DB.getAnalysesTable().getField_Title().getName());
        sql = DB.getAnalysisDataSeriesView().getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(DB.getAnalysisDataSeriesView().getField_Analysis_PK().getName(),PK));
        
        Statement st2 = con.createStatement();
        ResultSet rs2 = st2.executeQuery(sql);
        
        SimulationDataSeriesManager sdsm = new SimulationDataSeriesManager(DB);
        while(rs2.next()){
          long seriesPK = rs2.getLong(DB.getAnalysisDataSeriesView().getField_DataSeries_PK().getName());
          String seriesName = rs2.getLong(DB.getAnalysisDataSeriesView().getField_SimulationRun_PK().getName()) + "-" + 
              rs2.getString(DB.getAnalysisDataSeriesView().getField_DataSeries_Name().getName());
          SimulationDataSeries sds = sdsm.getDataSeriesFromDB(seriesPK);
          sds.setName(seriesName);
          analysis.addSeries(sds, seriesName);
      
        }
        rs2.close();
        st2.close();
      }
      else{
        analysis = null;
      }
      
      con.close();
      rs.close();
      st.close();
    }
    catch(Exception e){
      addToErrorReport(e.getMessage());
    }
    return analysis;
  }  
  
  /**
   * Gets the list of data series contained in this Analysis as an option list
   * that can be used in an HTML drop-down control
   * 
   * @param PK the primary key of the Analysis to be retrieved
   * @return the option list requested
   */
  public String getListOfSeriesAsOptionList(long PK) {
    return ADSM.getOptionList(PK);    
  }

  /**
   * Remove a data series from this analysis. Note that this can only be done if
   * the analysis can be edited.
   * 
   * @param PK The primary key of the analysis for which this data series is to 
   * be removed
   * @param dsPK  The primary key of the data series to be removed.
   * @param user userID
   * @return true if successful, false if not
   */
  public boolean removeDataSeries(long PK, long dsPK, long user) {
    if (!permitModification(PK, user)) return false;
    return ADSM.deleteElementsByBothPrimaryKeys(PK, dsPK);  
  }

  /**
   * Adds a data series to an Analysis
   * 
   * @param PK the primary key of the Analysis stored in the HWM Database to
   * which the data series is to be added
   * @param dsPK the primary key of data series, stored in the HWM Database, that
   * is to be added to the data series
   * @param user userID
   * @return true if successful, false if not
   */
  public boolean addDataSeries(long PK, long dsPK, long user) {
    if(!permitModification(PK, user)) return false;
    ADSM.setInstanceToSave(PK, dsPK);
    return ADSM.saveInstanceToDB();    
  }

  
  /**
   * Get a string representing a specific Analysis stored in the HWM Database as a
   * graph in SVG format, at a specified width and height.
   * @param PK Primary Key of the Analysis stored in the HWM Database.
   * @param imageWidth width of the graph, in pixels
   * @param imageHeight height of the graph, in pixels
   * @return a String containing the SVG representation of this Analysis graph
   */
  public String getAsSVG(long PK, long imageWidth, long imageHeight){
    Analysis analysis = getAnalysis(PK);
    return analysis.getAsSVG(getName(PK), imageWidth, imageHeight);
  }
  
  /**
   * Get a string representing a specific Analysis stored in the HWM Database as a
   * graph in SVG format, using a default width and height.
   * @param PK Primary Key of the Analysis stored in the HWM Database.
   * @return a String containing the SVG representation of this Analysis graph
   */
  public String getAsSVG(long PK){
    Analysis analysis = getAnalysis(PK);
    return analysis.getAsSVG(getName(PK), 500, 400);
  }

  /**
   * Gets the Analysis graph stored in the HWM Database as an image,
   * output to a file, using a default width and
   * height for the image, and including all date ranges for which data
   * are stored.
   * @param PK Primary Key of the Analysis to be retrieved
   * @param imageOutputLocation path to folder where image should be saved
   * @return file name of the file to which the graph image is output
   */
  public String getImage(long PK, String imageOutputLocation) {
    Analysis analysis = getAnalysis(PK);
    if(analysis == null) return "uninitialized";
    else {
      String name = getName(PK);
      return getImage(analysis, name, imageOutputLocation);
    }    
  }

  /**
   * Gets a graph of the Analysis specified, using a default width and
   * height for the image, and including all date ranges for which data
   * are stored.
   * @param analysis the Analysis to be graphed
   * @param analysisName the name to be used as a label on the graph
   * @param imageOutputLocation path to folder where image should be saved
   * @return file name of the file to which the graph image is output
   */
  public String getImage(Analysis analysis, String analysisName, String imageOutputLocation){
    return getImage(analysis, analysisName, 500, 400, imageOutputLocation);
  }

  /**
   * Gets a graph of the Analysis specified, using a specified
   * width and height, and including all date ranges for which data are stored.
   * @param analysis the Analysis to be graphed
   * @param analysisName the name to be used as a label on the graph
   * @param ImageHeight the height of the graph to be created, in pixels
   * @param ImageWidth the width of the graph to be created, in pixels
   * @param imageOutputLocation path to folder where image should be saved
   * @return file name of the file to which the graph image is output
   */
  public String getImage(Analysis analysis, String analysisName, long ImageHeight, long ImageWidth, 
      String imageOutputLocation){
    return getImage(analysis, analysisName, ImageHeight, ImageWidth, 
        SimDate.convertToString(analysis.earliestTime()),
        SimDate.convertToString(analysis.latestTime()), imageOutputLocation);    
  }

  /**
   * Gets the graph of the Analysis specified, using a specified width and height,
   * and including data only from specified date ranges.
   * @param analysis the Analysis to be graphed
   * @param analysisName the name to be used as a label on the graph
   * @param ImageHeight the height of the graph to be created, in pixels
   * @param ImageWidth the width of the graph to be created, in pixels
   * @param startDate the date of the earliest data to be plotted, in HWM Simulation Date format
   * @param endDate the date of the latest data to be plotted, in HWM Simulation Date format
   * @param imageOutputLocation path to folder where image should be saved
   * @return file name of the file to which the graph image is output
   */
  public String getImage(Analysis analysis, String analysisName, 
      long ImageHeight, long ImageWidth, String startDate, String endDate, String imageOutputLocation){
    String analysisFileName = analysisName + ".svg";
    try{
      File F = new File(imageOutputLocation + "/" + analysisFileName);
      FileWriter FW = new FileWriter(F);
      FW.write(analysis.getAsSVG(analysisName, ImageHeight, ImageWidth, startDate, endDate));
      FW.close();      
    }
    catch(IOException ioe){
      addToErrorReport(ioe.getMessage());
      return "ERROR:" + ioe.getMessage();
    }
    return analysisFileName;
  }   
    
}