package abcm.db.managers;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Hashtable;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMCodeModelsAndResearchersView;
import abcm.db.components.views.ABCMCodeModelsDomainsAndValuesAllView;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.sim.CodeModel;
import abcm.utilities.XMLAssist;

/**
 * Database manager for Code Models
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
*/
public class CodeModelManager extends AbstractAuditableDBManager {

  /** Title of Code Model to save  */
  private String     titleToSave;
  /** Researcher who will own Code Model being saved  */
  private long       researcherToSave;
  /** Description of Code Model to save  */
  private String     descriptionToSave;
  
  
  /** Internal Code Model and Option Values Manager  */
  private CodeModelsAndOptionValuesManager CMAOVM;

  /**
   * Inner class to manage code models and option values
   *
   * @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class CodeModelsAndOptionValuesManager extends AbstractDBLinkTableInnerManager {
  
    /**  Foreign Key Code Model FK to save */
    long codeModelFKToSave;
    /**  Code Model Option Domain FK to save */
    long codeModelOptionDomainFKToSave;
    /**  Code Model Option Value to save */
    long codeModelOptionValueFKToSave;
    
    /**
     * Constructor
     * @param db Instance of an ABCM Gateway to provide database access.
     * @param containerClass Containing Class
     */
    public CodeModelsAndOptionValuesManager(AbstractABCMDBGateway db, AbstractDBManager containerClass){ super(db, containerClass); }
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
     */
    protected ABCMdbTable baseTable(){                  return DB.getCodeModelsAndOptionValuesTable();                          }
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
     */
    protected ABCMdbField baseTablePKField(){           return DB.getCodeModelsAndOptionValuesTable().getField_PK();            }         
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
     */
    protected ABCMdbField baseTablePrimaryFKField(){    return DB.getCodeModelsAndOptionValuesTable().getField_CodeModelFK();   }
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
     */
    protected ABCMdbField baseTableSecondaryFKField(){  return DB.getCodeModelsAndOptionValuesTable().getField_OptionValueFK(); }
  
  
    // From the records View
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
     */
    protected ABCMdbAbstractViewable recordsView(){       return DB.getCodeModelsDomainsAndValuesAllView();                                        }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
     */
    protected ABCMdbField recordsViewPKField(){           return DB.getCodeModelsDomainsAndValuesAllView().getField_CodeModelAndOptionValue_PK();  }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
     */
    protected ABCMdbField recordsViewPrimaryFKField(){    return DB.getCodeModelsDomainsAndValuesAllView().getField_CodeModel_PK();                }
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
     */
    protected ABCMdbField recordsViewPrimaryNameField(){  return DB.getCodeModelsDomainsAndValuesAllView().getField_CodeModel_Title();             }
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
     */
    protected ABCMdbField recordsViewSecondaryNameField(){return DB.getCodeModelsDomainsAndValuesAllView().getField_CodeModelOptionValue_Name();   }
      
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
      ABCMCodeModelsDomainsAndValuesAllView view = DB.getCodeModelsDomainsAndValuesAllView();
      
      int codeModelOptionDomainDescFieldCol  = rs.findColumn(view.getField_CodeModelOptionDomain_Desc().getName());
      int codeModelOptionDomainIDFieldCol    = rs.findColumn(view.getField_CodeModelOptionDomain_ID().getName());
      int codeModelOptionDomainNameFieldCol  = rs.findColumn(view.getField_CodeModelOptionDomain_Name().getName());
      int codeModelOptionValueCodeFieldCol   = rs.findColumn(view.getField_CodeModelOptionValue_Code().getName());
      int codeModelOptionValueDescFieldCol   = rs.findColumn(view.getField_CodeModelOptionValue_Desc().getName());
      int codeModelOptionValueNameFieldCol   = rs.findColumn(view.getField_CodeModelOptionValue_Name().getName());
      int codeModelOptionValueValueFieldCol  = rs.findColumn(view.getField_CodeModelOptionValue_Value().getName());
      
      while(rs.next()){
        xml.open("domain");
          xml.addLine("ID",              rs.getString(codeModelOptionDomainIDFieldCol));
          xml.addLine("domainName",      rs.getString(codeModelOptionDomainNameFieldCol));
          xml.addLine("desc",            rs.getString(codeModelOptionDomainDescFieldCol));
          if(rs.getString(codeModelOptionValueValueFieldCol) != null){
             xml.addLine("optionValue",  rs.getString(codeModelOptionValueValueFieldCol));
             xml.addLine("optionDesc",   rs.getString(codeModelOptionValueDescFieldCol));
             xml.addLine("optionName",   rs.getString(codeModelOptionValueNameFieldCol));
             xml.addLine("code", AbstractABCMDBGateway.makeXMLSafe(rs.getString(codeModelOptionValueCodeFieldCol)));
          } 
          else {
            xml.addLine("optionValue", 0);
            xml.addLine("optionDesc",  "DEFAULT");
            xml.addLine("optionName",  "DEFAULT");
            xml.addLine("code",        "DEFAULT CODE");
            
            if(includeComments){
              CommentManager cm = new CommentManager(DB);
              xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.CODE_MODEL, instancePK), XMLAssist.INDENT_BLOCK);
            }
            
          }          
        xml.closeTag();
      }
    }
  
    /**
     * Adds sets of these records to an XML representation
     * @param xml the XMLAssist object constructing the XML document
     * @param includeComments true if comments are to be included
     * @param codeModelPK the Primary Key of the Code Model to be included
     * @throws Exception
     */
    public void addSetsToXML(XMLAssist xml, boolean includeComments, long codeModelPK) throws Exception{
      String sql = DB.getCodeModelsDomainsAndValuesAllView().getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(DB.getCodeModelsDomainsAndValuesAllView().getField_CodeModel_PK().getName(), codeModelPK));
      Connection con = DB.getAdminConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      addInstancesToXML(xml, rs, includeComments);
      
      rs.close();
      st.close();
      con.close();
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
     */
    protected void clearInstanceToSave(){
      codeModelFKToSave               = -1;
      codeModelOptionDomainFKToSave   = -1;
      codeModelOptionValueFKToSave    = -1;
    }
  
    /**
     * Sets the instance to be saved
     * @param codeModelFK FK to the Code Model
     * @param codeModelOptionDomain Code Model Option Domain
     * @param codeModelOptionValue Code Model Option Value
     */
    public void setInstanceToSave(long codeModelFK, long codeModelOptionDomain, long codeModelOptionValue){
      codeModelFKToSave               = codeModelFK;
      codeModelOptionDomainFKToSave   = codeModelOptionDomain;
      codeModelOptionValueFKToSave    = codeModelOptionValue;  
    }
    
    /**
     * Determines the status of the instance to be saved
     * @return true if the instance to be saved has been set, false otherwise
     */
    public boolean instanceToSaveIsValid(){
      return ((codeModelFKToSave               != -1) &&
              (codeModelOptionDomainFKToSave   != -1) &&
              (codeModelOptionValueFKToSave    != -1));
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
     */
    protected ABCMdbDatum datumToSave(){         
      return (instanceToSaveIsValid() ?
                DB.getCodeModelsAndOptionValuesTable().getDatum("" + codeModelFKToSave, "" + codeModelOptionDomainFKToSave, "" + codeModelOptionValueFKToSave) :
                null);
    }
  
    
  }
 
  
  /* Constructor */

  /**
   * @param db
   */
  public CodeModelManager(AbstractABCMDBGateway db){    
    super(db);  
    CMAOVM = new CodeModelsAndOptionValuesManager(DB, this);
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){             return DB.getCodeModelsTable();                                            }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){      return DB.getCodeModelsTable().getField_PK();                              }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){       return DB.getCodeModelsAndResearchersView().getField_CodeModel_Title();    }
  
  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){   return DB.getCodeModelsAndResearchersView();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){       return DB.getCodeModelsAndResearchersView().getField_CodeModel_PK();       }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){              return "codeModels";                                                       }
    
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMCodeModelsAndResearchersView view = DB.getCodeModelsAndResearchersView();
    
    int codeModelIDFieldCol    = rs.findColumn(view.getField_CodeModel_PK().getName());
    int codeModelTitleFieldCol = rs.findColumn(view.getField_CodeModel_Title().getName());
    int codeModelDescFieldCol  = rs.findColumn(view.getField_CodeModel_Desc().getName());
    int dateCreatedFieldCol    = rs.findColumn(view.getField_CodeModel_DateCreated().getName());
    int createdByFieldCol      = rs.findColumn(view.getField_Researcher_Name().getName());
    int canBeEditedFieldCol    = rs.findColumn(view.getField_CanBeEdited().getName());
    
    while(rs.next()){
      xml.open("codeModel");
        xml.addLine("ID",          rs.getLong(codeModelIDFieldCol));
        xml.addLine("title",       rs.getString(codeModelTitleFieldCol));
        xml.addLine("desc",        rs.getString(codeModelDescFieldCol));
          
        xml.addLine("dateCreated", rs.getString(dateCreatedFieldCol));
        xml.addLine("createdBy",   rs.getString(createdByFieldCol));
        xml.addLine("canBeEdited", ((rs.getString(canBeEditedFieldCol)) != null) ? "Yes"  : "No");
        
        if(singleInstance){
          xml.open("domainValues");
            CMAOVM.addSetsToXML(xml, includeComments, instancePK);        
          xml.closeTag();
          if(includeComments){
            CommentManager cm = new CommentManager(DB);
            xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.CODE_MODEL, instancePK), XMLAssist.INDENT_BLOCK);
          }
        }
      xml.closeTag();
    }    
  }

  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName(){    return "xmlCodeModelListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlCodeModelToHTML.xsl"; }
  
  
  /* Saving */
  
  /**
   * Set the instance to save by specifying the required information
   * @param title the title of the Code Model to save
   * @param researcherID the ID of the owner of the Code Model
   * @param description a description of the Code Model
   */
  /**
   * @param title
   * @param researcherID
   * @param description
   */
  public void setInstanceToSave(String title, long researcherID, String description){
    titleToSave = AbstractABCMDBGateway.makeSQLSafe(title);
    researcherToSave = researcherID;
    descriptionToSave = AbstractABCMDBGateway.makeSQLSafe(description);
  }
  
  /**
   * Sets the values of an instance to be saved to the database.
   * 
   * @param title          name of the Code Model to be saved
   * @param researcherID   PK of the Researcher who created this Code Model
   * @param description    description of this Code Model
   * @param options        Set of options to be saved
   * @param Values
   */
  /**
   * @param title
   * @param researcherID
   * @param description
   * @param options
   * @param Values
   */
  public void setInstanceToSave(String title, long researcherID, String description, String[] options, int[] Values){
    setInstanceToSave(title, researcherID, description);
  }
  
  /**
   * Sets an instance to save; as a shortcut, the user may pass the primary key
   * of a previously entered code model. If -1 is passed, the routine ignores this,
   * but if another value is passed, the routine pings the database to recover the
   * set of options and values already entered for this existing code model (if
   * the PK is valid), and sets the to be saved. The result is that the instance to
   * be saved will be a copy of the previously saved code model.
   * @param title Title of the new code model to be saved.
   * @param researcherID ID of the researcher saving this code model
   * @param description Description of the new code model
   * @param copyOfPK Primary key of the extant code model of which this code 
   * model should be a copy.
   */
  /**
   * @param title
   * @param researcherID
   * @param description
   * @param copyOfPK
   */
  public void setInstanceToSave(String title, long researcherID, String description, long copyOfPK){
    setInstanceToSave(title, researcherID, description);
    if(copyOfPK != -1){
      CodeModel cm = getCodeModelFromDB(copyOfPK);
      if(cm!=null){
        String[] options = new String[cm.countOfValues()];
        int[]    values  = new int[cm.countOfValues()];
        Enumeration e = cm.getKeys();
        int counter = 0;
        while(e.hasMoreElements()){
          options[counter] = (String)e.nextElement();
          values[counter]  = cm.getCodeModelOption(options[counter]);
          counter++;
        }
      }
    }
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){   return (titleToSave != null);      }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    titleToSave       = null;
    researcherToSave  = -1;
    descriptionToSave = null;
  }  
     
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){   
    return DB.getCodeModelsTable().getDatum(titleToSave, "" + researcherToSave, DB.currentTimeString(), descriptionToSave);
  } 
  

  /* Auditing */
  
    // Auditable tables have an owner field in the base table
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField             baseTableOwnerField(){     return DB.getCodeModelsTable().getField_ResearcherFK();            }  
  
  // Auditable tables have a view that will have an entry for each PK that can be edited
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable  canBeEditedView(){         return DB.getCanBeEdited_CodeModelsView();                         }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField             canBeEditedViewPKField(){  return DB.getCanBeEdited_CodeModelsView().getField_CodeModel_PK(); }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#deleteItem(long, long) 
   */
  public boolean deleteItem(long PK, long user) {
    return permitModification(PK, user)         &&
           CMAOVM.deleteElementsByPrimaryFK(PK) &&
           super.deleteItem(PK, user);
  }  

  
  /* Special */
  
  /**
   * Retrieve a specific, already stored code model from
   * the database; create a code model based on this
   * and return it.
   * @param CMPK primary key of the code model to be retrieved
   * @return a new instance of the requested Code Model
   */
  public CodeModel getCodeModelFromDB(long CMPK){
    CodeModel cm = new CodeModel();
    
    try {
      // This sets the options for this model
      String sql = DB.getCodeModelsDomainsAndValuesAllView().getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(DB.getCodeModelsDomainsAndValuesAllView().getField_CodeModel_PK().getName(), CMPK) + " " + 
          AbstractDBGateway.orderBy(DB.getCodeModelsDomainsAndValuesAllView().getField_CodeModelOptionDomain_ID().getName()));
      
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      
      ResultSet rs = st.executeQuery(sql);
      
      while(rs.next()){
        cm.setCodeModelOption(rs.getString(DB.getCodeModelsDomainsAndValuesAllView().getField_CodeModelOptionDomain_ID().getName()), 
                              rs.getInt(DB.getCodeModelsDomainsAndValuesAllView().getField_CodeModelOptionValue_Value().getName()));
      }
      
      rs.close();
      st.close();
      
      con.close();
    }    
    catch (Exception e){
      e.printStackTrace();
      errorReport = "Error in Code Model Manager getCodeModelFromDB(int): " + e.getMessage();
      return new CodeModel();
    } 
    
    return cm;
  }
  
  
  /**
   * Gets a code model from the database by providing its name,
   * as a String.
   * @param codeModelName
   * @return The code model object requested.
   */
  public CodeModel getCodeModelFromDB(String codeModelName){
    return getCodeModelFromDB(getPK(codeModelName));
  }
   
  
  /**
   * Update a code model's options. This can only be done if the code model has not been run
   * @param codeModelID ID number of the code model
   * @param optionCombinationID ID number of the option/value combination to be used
   * @return true if successful, false otherwise.
   */
  public boolean updateCodeModelOptions(long codeModelID, long optionCombinationID){
    if(!canBeEdited(codeModelID)) {
      errorReport = "Could not update code model options because the code model is not editable.";
      return false;
    }
  
    CodeModelOptionValuesManager cmovm = new CodeModelOptionValuesManager(DB);
    long domainID = cmovm.getDomainID(optionCombinationID);
    long optionID = cmovm.getOptionValue(optionCombinationID);
    
    CMAOVM.setInstanceToSave(codeModelID, domainID, optionID);
    return CMAOVM.deleteElementsByBothPrimaryKeys(codeModelID, domainID) && 
           CMAOVM.saveInstanceToDB();    
  }

  
  /* Reporting */
  
  // TODO Review efficiency of getCodeModelReport
  /**
   * Gets a report of a Code Model
   * @param CM Code Model to be described
   * @return a String describing the details of the Code Model
   */
  public String getCodeModelReport(CodeModel CM){
    CodeModelOptionValuesManager cmovm = new CodeModelOptionValuesManager(DB);
    Hashtable optionList = CM.getOptionList();
    String retVal = "Code Model with " + optionList.size() + " keys\n\n";
    Enumeration e = optionList.keys();
    while(e.hasMoreElements()){
      String opt = (String)e.nextElement();
      int val = CM.getCodeModelOption(opt);
      retVal += opt + " value: " + val + " = " + cmovm.getCodeModelOptionValueName(opt, val) + "\n";
    }
    return retVal + "\n";
  } 

  
}
