package abcm.db.managers;

import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Hashtable;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMCodeModelOptionDomainsTable;
import abcm.db.components.views.ABCMCodeModelsForWriteupView;
import abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager;
import abcm.utilities.TextFormatter;
import abcm.utilities.XMLAssist;

/**
 * Database Manager for Code Model Option Domains 
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class CodeModelOptionDomainManager extends AbstractDBStandardizedElementManager {

  /** name of the domain to be saved */ 
  private String domainToSave;
  
  /** Description of the option to be saved */
  private String descriptionToSave;
  
  /** ID to be saved */
  private String IDToSave;
  
  /** Domains and their IDs */
  private Hashtable domainsAndIDs = new Hashtable();
  
  /**
   * Constructor
   * @param db db gateway to be used
   */
  public CodeModelOptionDomainManager(AbstractABCMDBGateway db){ super(db); }
  
  /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  public ABCMdbTable baseTable(){          return DB.getCodeModelOptionDomainsTable();                         }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  public ABCMdbField baseTablePKField(){   return DB.getCodeModelOptionDomainsTable().getField_PK();           }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){ return DB.getCodeModelOptionDomainsTable().getField_DomainName();   }
  
  
  /* XML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){    return baseTable();            }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){        return baseTablePKField();     }
    
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs,
    boolean includeComments) throws Exception {
  
    ABCMCodeModelOptionDomainsTable table = DB.getCodeModelOptionDomainsTable();
  
    int domainIDFieldCol   = rs.findColumn(table.getField_PK().getName());
    int domainNameFieldCol = rs.findColumn(table.getField_DomainName().getName());
    int inCodeFieldCol     = rs.findColumn(table.getField_InCodeID().getName());
    int descFieldCol       = rs.findColumn(table.getField_Description().getName());

    while(rs.next()){
      xml.open("codeModelOptionDomain");
        xml.addLine("optionPK",   rs.getLong(domainIDFieldCol));
        xml.addLine("domainName", rs.getString(domainNameFieldCol));
        xml.addLine("desc",       rs.getString(descFieldCol));
        xml.addLine("inCode",     rs.getString(inCodeFieldCol));        
        
        CodeModelOptionValuesManager cmovm = new CodeModelOptionValuesManager(DB);
        xml.open("options");
          cmovm.addSetToXML(xml, includeComments, rs.getLong(domainIDFieldCol));          
        xml.closeTag();
      xml.closeTag();   
    }
    
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName() {             return "codeModelOptionDomains";    }

  
  /* HTML Creation */
  
  /**
   * Overridden here to do nothing because no instances are required. 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName() {    return null;    }  // This is not used; see CodeModelOptionValuesManager

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {        return "xmlCodeModelDomainListToHTML.xsl";    }

  
  /* Saving */
  
  /**
   * Sets values for a new instance to be saved to the associated db
   * @param domainToSave name of the domain to be saved
   * @param descriptionToSave description to be saved
   * @param IDToSave ID to be saved
   */
  protected void setInstanceToSave(String domainToSave, String descriptionToSave, String IDToSave){
    this.domainToSave      = domainToSave;
    this.descriptionToSave = descriptionToSave;
    this.IDToSave          = IDToSave;
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  protected void clearInstanceToSave() {
    domainToSave      = null;
    descriptionToSave = null;
    IDToSave          = null;
  }
  
  /**
   * Gets the status of the instance to be saved
   * @return true if the instance to be saved has been set and is valid, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return ((domainToSave      != null) &&  
            (descriptionToSave != null) &&
            (IDToSave          != null));
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  protected ABCMdbDatum datumToSave() {
    return (instanceToSaveIsValid() ? 
       DB.getCodeModelOptionDomainsTable().getDatum(domainToSave, descriptionToSave, IDToSave) :
       null);
  }

  
  /* Special */
  
  /**
   * Gets the list of code model options associated with a given domain in a TeX format
   * @param domain requested domain
   * @return text of TeX formatted list
   */
  public String getListOfOptionsAsTex(String domain){
    StringWriter retVal = new StringWriter();
    
    long domainPK = ((domain == null) ? -1 : this.getDomainIDFromDomainName(domain));
    if(domain == null) System.out.println("Domain passed is null; retrieving all (" + domainPK + ")");
    else               System.out.println("Getting list of options: " + domainPK);
    
    try{
      
      ABCMCodeModelOptionDomainsTable codeModelOptionDomains = DB.getCodeModelOptionDomainsTable();
            
      String sql = codeModelOptionDomains.getSelectSQL(DB.getOwner(), AbstractDBGateway.orderBy(codeModelOptionDomains.getField_PK().getName()));
      
      if(domainPK > 0){
        sql = codeModelOptionDomains.getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(codeModelOptionDomains.getField_PK().getName(), domainPK));
      }
      
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      
      ResultSet rs = st.executeQuery(sql);
      
      int PKFieldCol         = rs.findColumn(codeModelOptionDomains.getField_PK().getName());
      int domainNameFieldCol = rs.findColumn(codeModelOptionDomains.getField_DomainName().getName());
      int descFieldCol       = rs.findColumn(codeModelOptionDomains.getField_Description().getName());
      int inCodeFieldCol     = rs.findColumn(codeModelOptionDomains.getField_InCodeID().getName());
      
      
      while(rs.next()){
        retVal.write("\\section{" + TextFormatter.makeTEXSafe(rs.getString(domainNameFieldCol)) + "}" + "\\label{" + rs.getString(inCodeFieldCol) + "}\n");
        retVal.write("In code: \\begin{verbatim}" + rs.getString(inCodeFieldCol) + "\\end{verbatim}\n\n");
        retVal.write("\\noindent\n" + rs.getString(descFieldCol) + "\n\n");
      
        ABCMCodeModelsForWriteupView codeModelsForWriteup = DB.getCodeModelsForWriteupView();
        
        // Left in String (rather than 'int') version... 
        
        String optionDomainPK   = codeModelsForWriteup.getField_CodeModelOptionDomain_PK().getName();
        String optionValueCode  = codeModelsForWriteup.getField_CodeModelOptionValue_Code().getName();
        String optionValueDesc  = codeModelsForWriteup.getField_CodeModelOptionValue_Desc().getName();
        String optionValueName  = codeModelsForWriteup.getField_CodeModelOptionValue_Name().getName();
        String optionValueValue = codeModelsForWriteup.getField_CodeModelOptionValue_Value().getName();
        
        String sql2 = codeModelsForWriteup.getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(optionDomainPK, rs.getLong(PKFieldCol)) + " " + AbstractDBGateway.orderBy(optionValueValue));
        
        Statement st2 = con.createStatement();
        ResultSet rs2 = st2.executeQuery(sql2);
        
        while(rs2.next()){
          retVal.write("\\subsection{" + TextFormatter.makeTEXSafe(rs2.getString(optionValueName)) + "}\n\n");
          retVal.write(AbstractABCMDBGateway.makeTexSafe(rs2.getString(optionValueDesc)) + "\n\n");
          retVal.write("\\begin{lstlisting}[language=Java, breaklines=true]\n");
          retVal.write(rs2.getString(optionValueCode) + "\n");
          retVal.write("\\end{lstlisting}\n\n");
        }
        rs2.close();
        st2.close();
      }
      rs.close();
      st.close();
      con.close();
    }
    catch(Exception e){
      addToErrorReport("Error getting Code Models as Option List: " + e.getMessage());
    }
    
    
    return retVal.toString();
  }
  
  /**
   * Gets the domain ID from the domain name
   * @param option name of the requested domain
   * @return the requested domain ID
   */
  public int getDomainIDFromDomainName(String option){
    if(!(domainsAndIDs.containsKey(option)))  loadDomainIDs();    // Try once; if not there, do a load
    if(!(domainsAndIDs.containsKey(option)))  return -1;          // Try again; if not there, fail gracefully
    else return ((Integer)domainsAndIDs.get(option)).intValue();  // Return result
  }
  
  /**
   * Loads a list of all domain IDs into this class for reference
   */
  private void loadDomainIDs(){
    try {
      ABCMCodeModelOptionDomainsTable table = DB.getCodeModelOptionDomainsTable(); 
      String sql = table.getSelectAllSQL(DB.getOwner());
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      
      ResultSet rs = st.executeQuery(sql);
    
      // If there are any records (there should be)
      while(rs.next()){
        domainsAndIDs.put(rs.getString(table.getField_InCodeID().getName()),
          new Integer(rs.getInt(table.getField_PK().getName())));
      }
    }
    catch (Exception e){
      addToErrorReport("Error in CodeModelOptionDomainManager loadDomainIDs" + e.getMessage());
    }
  }
  
}
