package abcm.db.managers;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Hashtable;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMCodeModelOptionValuesTable;
import abcm.db.managers.abstractparents.AbstractDBLinkTableManager;
import abcm.utilities.XMLAssist;

/**
 * The CodeModelOptionValues table provides the list of valid options for
 * each CodeModelDomain. Each domain + value combination must be unique,
 * and this forms the primary key of the field.
 *  
 * @author John T. Murphy
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class CodeModelOptionValuesManager extends AbstractDBLinkTableManager {
  
  /**
   * Local copy of values list, used for the parser
   */
  private Hashtable<String, Long> valuesList;
  
  
  /**
   * Constructor
   * @param db DBGateway that provides DB access.
   */
  public CodeModelOptionValuesManager(AbstractABCMDBGateway db){ super(db); }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
   */
  protected ABCMdbTable baseTable() {                      return DB.getCodeModelOptionValuesTable();                                     }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField() {               return DB.getCodeModelOptionValuesTable().getField_PRIMARY_KEY();              }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
   */
  protected ABCMdbField baseTablePrimaryFKField() {        return DB.getCodeModelOptionValuesTable().getField_CodeModelOptionDomainFK();  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
   */
  protected ABCMdbField baseTableSecondaryFKField() {      return DB.getCodeModelOptionValuesTable().getField_OptionValue();              }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
   */
  protected ABCMdbField recordsViewPrimaryFKField() {      return DB.getCodeModelOptionValuesTable().getField_CodeModelOptionDomainFK();  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
   */
  protected ABCMdbField recordsViewPrimaryNameField() {    return DB.getCodeModelOptionValuesTable().getField_OptionValue();              }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
   */
  protected ABCMdbField recordsViewSecondaryNameField() {  return DB.getCodeModelOptionValuesTable().getField_Name();                     }  
  
  
  /* XML Creation */ 
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView(){   return DB.getCodeModelOptionValuesTable();                        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField(){       return DB.getCodeModelOptionValuesTable().getField_PRIMARY_KEY(); }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
    
    ABCMCodeModelOptionValuesTable table = DB.getCodeModelOptionValuesTable();
    
    int optionValueNameFieldCol           = rs.findColumn(table.getField_Name().getName());
    int optionValueFieldCol               = rs.findColumn(table.getField_OptionValue().getName());
    int intCodeModelOptionValueIDFieldCol = rs.findColumn(table.getField_PRIMARY_KEY().getName());
    int optionValueDescFieldCol           = rs.findColumn(table.getField_Description().getName());
    // Not used in current implementation
//    int txtCodeFieldCol                   = rs.findColumn(table.getField_Code().getName());
    
    while(rs.next()){
      xml.open("option");
        xml.addLine("valueName", rs.getString(optionValueNameFieldCol));
        // Not used in current implementation
//        if(singleInstance) xml.addLine("codeVerbatim", DB.makeXMLSafe(rs.getString(txtCodeFieldCol)));
        xml.addLine("codeVerbatim", "TBD");
        xml.addLine("optionPK", rs.getString(intCodeModelOptionValueIDFieldCol));
        xml.addLine("optionValue", rs.getLong(optionValueFieldCol));

        xml.addLine("description", rs.getString(optionValueDescFieldCol));
      xml.closeTag();
    }   
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#addSetToXML(abcm.utilities.XMLAssist, boolean, long) 
   */
  public void addSetToXML(XMLAssist xml, boolean includeComments, long domainPK){
    String sql = DB.getCodeModelOptionValuesTable().getSelectSQL(DB.getOwner(), 
              AbstractDBGateway.where(DB.getCodeModelOptionValuesTable().getField_CodeModelOptionDomainFK().getName(), domainPK) + " " + 
              AbstractDBGateway.orderBy(DB.getCodeModelOptionValuesTable().getField_OptionValue().getName()));
    addSetToXML(xml, includeComments, sql);
        
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName() {   return "xmlCodeModelDomainValueToHTML.xsl";    }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {       return null;    } // Not used

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName() {            return null;    } // Not used

  /* Saving */
  // Not implemented
  
  /**
   * Not implemented for this element
   */
  protected void clearInstanceToSave() {   }

  /** 
   * Not implemented for this element.
   * @return null
   */
  protected ABCMdbDatum datumToSave() {            return null;    }

  /* Option List */
  
  /**
   * Note that this object is not a standardized database element, and hence
   * provides its own optionlist methods instead of those built into the
   * AbstractDBStandardizedElementManager family
   * @return an option list of Code Model Options
   */
  public String getOptionList(){
    return DB.getOptionList(DB.getCodeModeOptionsListedView(),
        DB.getCodeModeOptionsListedView().getField_CodeModelOptionValue_PK(),
        new ABCMdbField[] {DB.getCodeModeOptionsListedView().getField_CodeModelOptionDomain_Name(), 
                           DB.getCodeModeOptionsListedView().getField_CodeModelOptionValue_Name()});
  }
  
  
  /**
   * Gets the list of Code Model Options with their domains as an option list.
   * Note that this is a very special option list: the 'Value' component includes
   * both the Domain ID and the codeModelOptionValue. This is because it may be used
   * to create a command in the form:
   * 
   * setCodeModelOption Domain Value
   * 
   * by an interactive web component...
   * @return a string listing the Code Model Options with their domains
   */
  public String getCodeModelOptionsAsExtendedOptionList(){
    String optionList = "";
    try{
      String inCodeDomainIDString = DB.getCodeModeOptionsListedView().getField_CodeModelOptionDomain_ID().getName();
      String optionValue          = DB.getCodeModeOptionsListedView().getField_CodeModelOptionValue_Value().getName();
      String domainName           = DB.getCodeModeOptionsListedView().getField_CodeModelOptionDomain_Name().getName();
      String optionValueName      = DB.getCodeModeOptionsListedView().getField_CodeModelOptionValue_Name().getName();
     
      String sql = DB.getCodeModeOptionsListedView().getSelectAllSQL(DB.getOwner());
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      
      ResultSet rs = st.executeQuery(sql);
      
      while(rs.next()){
        optionList+= "<OPTION VALUE=\"" + rs.getString(inCodeDomainIDString) + ' ' + rs.getLong(optionValue) +"\">" + 
          rs.getString(domainName) + ": " + rs.getString(optionValueName) +
          "</OPTION>";
      }
      
      rs.close();
      st.close();
      con.close();
      
    }
    catch(Exception e){
      addToErrorReport("Exception while getting Code Model Options as extended List: " + e.getMessage());
    }
    return optionList;
  }
  
  
  /* Basic Lookup */
  // Note that this object is not a standardized database element, and hence
  // provides its own methods here instead of those built into the
  // AbstractDBStandardizedElementManager family
  
  /**
   * Given a primary key in the Code Model Option Values Table, returns the domain ID
   * @param PK primary key
   * @return ID of code model option domain
   */
  public long getDomainID(long PK){
    return DB.getNumericItem(DB.getCodeModelOptionValuesTable(), 
                             DB.getCodeModelOptionValuesTable().getField_PRIMARY_KEY(), PK, 
                             DB.getCodeModelOptionValuesTable().getField_CodeModelOptionDomainFK());
  }
  
  /**
   * Gets the option value associated with the Code Model Option Value PK 
   * @param PK Code Model Option Value PK
   * @return value of associated Code Model Option
   */
  public long getOptionValue(long PK){
    return DB.getNumericItem(DB.getCodeModelOptionValuesTable(), 
                             DB.getCodeModelOptionValuesTable().getField_PRIMARY_KEY(), PK, 
                             DB.getCodeModelOptionValuesTable().getField_OptionValue());
  }
  
  /**
   * Gets the name of the Code Model Option Value specified by the IDs
   * @param domainID ID of domain
   * @param valueID ID of value
   * @return name of the Code Model Option Value
   */
  public String getCodeModelOptionValueName(long domainID, long valueID){
    ABCMCodeModelOptionValuesTable table = DB.getCodeModelOptionValuesTable();
      
    String sql = table.getSelectSQL(DB.getOwner(),  
          "WHERE " + table.getField_CodeModelOptionDomainFK().getName() + " = " + domainID + " " +
          "AND " + table.getField_OptionValue() + " = " + valueID);
    
    String retVal =  DB.getStringItem(sql, DB.getCodeModelOptionValuesTable().getField_Name().getName());
    return (retVal == null ? "Unknown option value." : retVal);
  }
  
  /**
   * Gets the name of the Code Model Option Value specified by the name of the domain and the value.
   * @param opt name of domain
   * @param val value
   * @return name of the Code Model Option Value
   */
  public String getCodeModelOptionValueName(String opt, int val){
    CodeModelOptionDomainManager cmodm = new CodeModelOptionDomainManager(DB);
    return getCodeModelOptionValueName(cmodm.getDomainIDFromDomainName(opt), val); 
  }

  /**
   * Gets the integer option value given the String that identifies the option.
   * @param optionID
   * @return the FK of the specified option value
   */
  public long getOptionValue(String optionID){
    
    if(valuesList == null){
      valuesList = new Hashtable<String, Long>();
      try{
        Connection con = DB.getAdminConnection();
        Statement st = con.createStatement();
        String sql = this.baseTable().getSelectAllSQL(DB.getOwner());
        
        ResultSet rs = st.executeQuery(sql);
        
        ABCMCodeModelOptionValuesTable table = DB.getCodeModelOptionValuesTable();
        
        while(rs.next()){
          valuesList.put(rs.getString(table.getField_Name().getName()), new Long(rs.getLong(table.getField_OptionValue().getName())));
        }
      }
      catch(Exception E){
        
      }      
    }
    return(valuesList.containsKey(optionID) ? valuesList.get(optionID) : -1);
  }
 
}
