package abcm.db.managers;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMNarrativeArgumentsTable;
import abcm.db.components.views.ABCMNarrativeArgumentsView;
import abcm.db.managers.abstractparents.AbstractDBLinkTableManager;
import abcm.utilities.XMLAssist;


// This is stretching the use of the 'Link' table; there are
// effectively two keys (because Command No. + Arg no. must be
// unique)...
public class CommandArgumentManager extends AbstractDBLinkTableManager {

  /** FK of command to save  */
  private long    commandFKToSave;
  /** number of argument to save  */
  private int     argNoToSave;
  /** name of argument to save  */
  private String  argNameToSave;
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public CommandArgumentManager(AbstractABCMDBGateway db){ super(db); }
  
  // Base Table
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){                      return DB.getNarrativeArgumentsTable();                                 }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){               return DB.getNarrativeArgumentsTable().getField_PK();                   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
   */
  protected ABCMdbField baseTablePrimaryFKField() {       return DB.getNarrativeArgumentsTable().getField_NarrativeCommandFK();   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
   */
  protected ABCMdbField baseTableSecondaryFKField() {     return DB.getNarrativeArgumentsTable().getField_ArgumentNumber();       }
  
  
  // Records View
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView(){     return DB.getNarrativeArgumentsView();                                  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField(){         return DB.getNarrativeArgumentsView().getField_PK();                    }  
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
   */
  protected ABCMdbField recordsViewPrimaryFKField(){  return DB.getNarrativeArgumentsView().getField_CommandFK();             }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
   */
  public ABCMdbField recordsViewPrimaryNameField(){   return DB.getNarrativeArgumentsView().getField_Command();               }
   
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
   */
  public ABCMdbField recordsViewSecondaryNameField(){ return DB.getNarrativeArgumentsView().getField_ArgName();               }

  /* Saving */
  
  /**
   * Sets an instance of command argument to save
   * @param commandFK FK to command
   * @param argNo number of argument; should not duplicate an existing argument number for this command
   * @param argName name of argument
   */
  public void setInstanceToSave(long commandFK, int argNo, String argName){
    commandFKToSave = commandFK;
    argNoToSave     = argNo;
    argNameToSave   = argName;
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  protected void clearInstanceToSave() {
    commandFKToSave  = -1;
    argNoToSave      = -1;
    argNameToSave    = null;
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return((commandFKToSave != -1  ) &&
           (argNoToSave     != -1  ) &&
           (argNameToSave   != null));
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  protected ABCMdbDatum datumToSave() {
    return (instanceToSaveIsValid() ? DB.getNarrativeArgumentsTable().getDatum("" + commandFKToSave, "" + argNoToSave, argNameToSave) : null);
  }
  
  /* XML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
    
    ABCMNarrativeArgumentsView view = DB.getNarrativeArgumentsView();
    
    int argNoFieldCol    = rs.findColumn(view.getField_ArgNo().getName());
    int argDescFieldCol  = rs.findColumn(view.getField_ArgName().getName());
    
    while(rs.next()){
      xml.open("argument");
        xml.addLine("argNo", rs.getLong(argNoFieldCol));
        xml.addLine("argDesc", rs.getString(argDescFieldCol));
      xml.closeTag();
    }
  }
    
  /**
   * Overridden to return null; no XML is supported
   * @return null
   */
  protected String getInstanceStylesheetName() {    return null;                           }

  /**
   * Overridden to return null; no XML is supported
   * @return null
   */
  protected String getListStylesheetName() {        return null;                           }

  /**
   * Overridden to return null; no XML is supported
   * @return null
   */
  protected String xmlOuterNodeName() {             return null;                           }

  
  /**
   * Gets the Help string associated with the specific command
   * @param commandPK PK of command
   * @return Help message
   */
  public String getHelpString(long commandPK){
    String retVal = "";
    
    ABCMNarrativeArgumentsTable table = DB.getNarrativeArgumentsTable();
    
    String sql = table.getSelectSQL(DB.getOwner(),
        AbstractDBGateway.where(table.getField_NarrativeCommandFK().getName(), commandPK) + " " + 
        AbstractDBGateway.orderBy(table.getField_ArgumentNumber().getName()));
    try{
      Connection con = DB.getAdminConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      int argNoFieldCol   = rs.findColumn(table.getField_ArgumentNumber().getName());
      int argNameFieldCol = rs.findColumn(table.getField_ArgumentName().getName());
      
      while(rs.next()){
        retVal += "   " + rs.getLong(argNoFieldCol) + ": " + rs.getString(argNameFieldCol) + "\n";
      }
    }
    catch(Exception e){
      addToErrorReport("Could not complete retrieval of help string in command arguments: " + e.getMessage());
    }
    return retVal;
  }
  
}
