package abcm.db.managers;

import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMCommandCategoriesInOrderView;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.utilities.XMLAssist;

/**
 * Database Manager for Command Categories 
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class CommandCategoriesManager extends AbstractNonAuditableDBManager {
  
  /** Category name to be saved  */
  private String  categoryNameToSave;
  /** Category description to be saved  */
  private String  categoryDescriptionToSave;
  /** Category order to be saved  */
  private int     categoryOrderToSave;
  /**   */
  private boolean listCommands = false; 
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public CommandCategoriesManager(AbstractABCMDBGateway db){ 
    super(db); 
  }

  /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){                  return DB.getCommandCategoriesTable();                                        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){           return DB.getCommandCategoriesTable().getField_PK();                          }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField() {        return DB.getCommandCategoriesTable().getField_Name();                        }     
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField() 
   */
  protected ABCMdbField baseTableDescriptionField(){  return DB.getCommandCategoriesTable().getField_Description();                 }
  
  
  /* XML Creation */ 
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView() {    return DB.getCommandCategoriesInOrderView();                                  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField() {        return DB.getCommandCategoriesInOrderView().getField_CommandCategory_PK();    }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName() {               return "commandCategories";                                                   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
    
    CommandsManager CM = new CommandsManager(DB);
    
    ABCMCommandCategoriesInOrderView view = DB.getCommandCategoriesInOrderView();
    
    int categoryIDFieldCol   = rs.findColumn(view.getField_CommandCategory_PK().getName());
    int categoryNameFieldCol = rs.findColumn(view.getField_CommandCategory_Name().getName());
    int categoryDescFieldCol = rs.findColumn(view.getField_CommandCategory_Desc().getName());
    
    while(rs.next()){
      xml.open("category");
        xml.addLine("ID",           rs.getLong(categoryIDFieldCol));
        xml.addLine("categoryName", rs.getString(categoryNameFieldCol));
        xml.addLine("desc",         rs.getString(categoryDescFieldCol));
        
        xml.open("commands");
          if(listCommands) CM.addCommandCategorySetToXML(xml, includeComments, rs.getLong(categoryIDFieldCol));
        xml.closeTag();
      xml.closeTag();
    }
  }

  
  /* Saving */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  protected void clearInstanceToSave() {
    categoryNameToSave        = null;
    categoryDescriptionToSave = null;
    categoryOrderToSave       = -1;
  }
  
  /**
   * Sets an instance of a command category to be saved
   * @param name the category name
   * @param description description
   * @param order order in which category is to be listed; should not
   * duplicate any existing order
   */
  public void setInstanceToSave(String name, String description, int order){
    categoryNameToSave        = AbstractDBGateway.makeSQLSafe(name);
    categoryDescriptionToSave = AbstractDBGateway.makeSQLSafe(description);
    categoryOrderToSave       = order;
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return((categoryNameToSave        != null) && 
           (categoryDescriptionToSave != null) &&
           (categoryOrderToSave       != -1  ));
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  protected ABCMdbDatum datumToSave() {
    return ((instanceToSaveIsValid() ? DB.getCommandCategoriesTable().getDatum(categoryNameToSave, 
        categoryDescriptionToSave, "" + categoryOrderToSave): null));
  }
      
  // In an ideal world there would be a method here that updates all existing orders to
  // higher than X; when 'save' is invoked, all numbers higher than X would be incremented
  // to make room for X... someday...  
  
  /**
   * Gets a String that lists all commands in the HWM Command
   * Set (as stored in the HWM Database) and generic help.
   * @return the HWM Command Set listed
   */
  public String getListOfCommandsAsString(){
    StringWriter out = new StringWriter();
    
    try{
      String sql = DB.getNarrativeCommandsByCategoryView().getSelectSQL(DB.getOwner(), 
          " ORDER BY " + DB.getNarrativeCommandsByCategoryView().getField_CommandCategory_Order().getName() + ", " + 
                         DB.getNarrativeCommandsByCategoryView().getField_Command().getName());
      
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
    
      int categoryNameFieldCol = rs.findColumn(DB.getNarrativeCommandsByCategoryView().getField_CommandCategory_Name().getName());
      int commandFieldCol      = rs.findColumn(DB.getNarrativeCommandsByCategoryView().getField_Command().getName());
      int numOfArgsFieldCol    = rs.findColumn(DB.getNarrativeCommandsByCategoryView().getField_NumberOfArguments().getName());
      int commandDescFieldCol  = rs.findColumn(DB.getNarrativeCommandsByCategoryView().getField_Command_Desc().getName());
      
      String category = "";
      while(rs.next()){
        if(rs.getString(categoryNameFieldCol).compareTo(category) != 0){
          category = rs.getString(categoryNameFieldCol);
          out.write(category + ":\n");
        }
        out.write("  " +rs.getString(commandFieldCol) + " (No. of arguments: " + rs.getInt(numOfArgsFieldCol)+")\n");
        out.write("     " + rs.getString(commandDescFieldCol) + "\n");
      }
      
      rs.close();
      st.close();
      con.close();
      
    }
    catch(Exception e){
      addToErrorReport("An error has occurred in CommandCategoriesManager getListOfCommandsAsString(): " + e.getMessage() + e.getLocalizedMessage());
    }  
    
    return out.toString();  
  }

  /* HTML Creation */
  
  /** 
   * Overridden to take no action
   * @return null 
   */
  protected String getInstanceStylesheetName() {  return null;                                   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {      
    return ((!listCommands) ? "xmlCommandCategoryListToHTML.xsl" : "xmlCommandListToHTML.xsl");     
  }
    
  /* Special */
  /**
   * Sets the flag that indicates whether this should return a list of command categories
   * or list all commands
   * @param listCommandsInHTMLList
   */
  public void setListCommands(boolean listCommandsInHTMLList){ listCommands = listCommandsInHTMLList; }
  
}
