package abcm.db.managers;

import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Hashtable;
import java.util.Vector;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMNarrativeCommandsTable;
import abcm.db.components.views.ABCMCommandCategoriesInOrderView;
import abcm.db.components.views.ABCMNarrativeArgumentsView;
import abcm.db.components.views.ABCMNarrativeCommandsView;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.utilities.XMLAssist;

/**
 * Database Manager for Commands 
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class CommandsManager extends AbstractNonAuditableDBManager {

  /** Command to be saved  */
  private String  commandToSave;
  /** Number of arguments being saved  */
  private int     numberOfArgsToSave;
  /** Description to save  */
  private String  descriptionToSave;
  /** Flag indicating if command to be saved is command-line-only  */
  private boolean commandLineOnlyToSave;
  /** Flag indicating if the command to be saved is probe-only  */
  private boolean probeToSave;
  /** Instructions for the command to be saved  */
  private String  instructionsToSave;
  /** FK to the category to which the command to be saved belongs  */
  private long    categoryFKToSave;
  
  /** Internal list of available commands  */
  private Hashtable commandList;
  /** Internal list of numeric versions of available commands  */
  private Hashtable numericCommandList;

  /** Internal Command Categories Manager  */
  private CommandCategoriesManager CCM;
    
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public CommandsManager(AbstractABCMDBGateway db){   
    super(db);  
    CCM = new CommandCategoriesManager(DB);  
  }
  
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){                  return DB.getNarrativeCommandsTable();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){           return DB.getNarrativeCommandsTable().getField_PK();                 }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField() {        return DB.getNarrativeCommandsTable().getField_Command();            }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField() 
   */
  protected ABCMdbField baseTableDescriptionField(){  return DB.getNarrativeCommandsTable().getField_Description();        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView() {    return DB.getNarrativeCommandsView();                                }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField() {        return DB.getNarrativeCommandsView().getField_NarrativeCommand_PK(); }
  

  /**
   * Sets an instance of a command to be saved
   * @param command command to be saved
   * @param numArgs number of arguments to command to be saved
   * @param desc description to be saved
   * @param commandLineOnly flag indicating if command is command-line-only
   * @param probe flag indicating if command is probe-only
   * @param instructions instructions for command
   * @param categoryFK FK to category to which command belongs
   */
  public void setInstanceToSave(String command, int numArgs, String desc, boolean commandLineOnly, boolean probe, String instructions, long categoryFK){
    commandToSave          = AbstractDBGateway.makeSQLSafe(command);
    numberOfArgsToSave     = numArgs;
    descriptionToSave      = AbstractDBGateway.makeSQLSafe(desc);
    commandLineOnlyToSave  = commandLineOnly;
    probeToSave            = probe;
    instructionsToSave     = AbstractDBGateway.makeSQLSafe(instructions);
    categoryFKToSave       = categoryFK;      
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return ((commandToSave          != null ) &&
            (numberOfArgsToSave     != -1   ) &&
            (descriptionToSave      != null ) &&
            (commandLineOnlyToSave  != false) &&
            (probeToSave            != false) &&
            (instructionsToSave     != null ) &&
            (categoryFKToSave       != -1   ));
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  protected void clearInstanceToSave() {
    commandToSave          = null;
    numberOfArgsToSave     = -1;
    descriptionToSave      = null;
    commandLineOnlyToSave  = false;
    probeToSave            = false;
    instructionsToSave     = null;
    categoryFKToSave       = -1;  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  protected ABCMdbDatum datumToSave() {
    return ((instanceToSaveIsValid()) ? 
                     DB.getNarrativeCommandsTable().getDatum(commandToSave, 
                                                             "" + numberOfArgsToSave, 
                                                             descriptionToSave, 
                                                             (commandLineOnlyToSave ? "1" : "0"), 
                                                             (probeToSave           ? "1" : "0"), 
                                                             instructionsToSave, 
                                                             "" + categoryFKToSave) : 
                     null);
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
    
    ABCMNarrativeCommandsView view = DB.getNarrativeCommandsView();
    
    int commandIDFieldCol        = rs.findColumn(view.getField_NarrativeCommand_PK().getName());
    int commandNameFieldCol      = rs.findColumn(view.getField_NarrativeCommand().getName());
    int argumentCountFieldCol    = rs.findColumn(view.getField_NumberOfArguments().getName());
    int descriptionFieldCol      = rs.findColumn(view.getField_Description().getName());
    int commandLineOnlyFieldCol  = rs.findColumn(view.getField_CommandLineOnly().getName());
    int probeFieldCol            = rs.findColumn(view.getField_Probe().getName());
    int instructionsFieldCol     = rs.findColumn(view.getField_NarrativeCommand_Instructions().getName());
    
    while(rs.next()){
      xml.open("command");
        xml.addLine("name",             rs.getString(commandNameFieldCol));
        xml.addLine("argument_count",   rs.getInt(argumentCountFieldCol));
        xml.addLine("commandDesc",      rs.getString(descriptionFieldCol));
        xml.addLine("commandLineOnly", (rs.getBoolean(commandLineOnlyFieldCol) ? "Yes" : "No"));
        xml.addLine("probe",           (rs.getBoolean(probeFieldCol) ? "Yes" : "No"));
        xml.addLine("instructions",     rs.getString(instructionsFieldCol));
        if(rs.getLong(argumentCountFieldCol) > 0){          
          xml.open("arguments");  // TODO Improve performance: this requires too many db hits!
            CommandArgumentManager cam = new CommandArgumentManager(DB);
            cam.addSetToXML(xml, includeComments, rs.getLong(commandIDFieldCol));
          xml.closeTag();  
        }
      xml.closeTag();
    }
  }

  /**
   * Adds sets of commands in a given category to an XMl document
   * @param xml XMLAssist object creating the XML document
   * @param includeComments true if comments are to be included
   * @param categoryFK FK to the category of commands to be included
   * @throws Exception
   */
  protected void addCommandCategorySetToXML(XMLAssist xml, boolean includeComments, long categoryFK) throws Exception {
    String sql = DB.getNarrativeCommandsView().getSelectSQL(DB.getOwner(), 
        AbstractDBGateway.where(DB.getNarrativeCommandsView().getField_CommandCategory_PK().getName(), categoryFK));
    
    Connection con = DB.getAdminConnection();
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);
    
    addInstancesToXML(xml, rs, includeComments);
    
    rs.close();
    st.close();
    con.close();
    
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {        return "xmlCommandListToHTML.xsl";    }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName() {    return null;    }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName() {             return null;    }
    
  
  
  /* Special */
  /**
   * Generates the list of commands and primary keys, identical in
   * format to the code in the interface 'CommandList'. This can
   * be used as a simple way to re-generate that code and ensure that
   * it is complete and up to date.
   */
  public void generateCommandsAndPKsCode(){
    try{
      ABCMNarrativeCommandsTable table = DB.getNarrativeCommandsTable();
      
      String sql = table.getSelectSQL(DB.getOwner(), AbstractDBGateway.orderBy(table.getField_PK().getName()));
                 
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      int IDFieldCol      = rs.findColumn(table.getField_PK().getName());
      int commandFieldCol = rs.findColumn(table.getField_Command().getName());
      
      String out = "";
      while(rs.next()){
        out = "public static final int " + rs.getString(commandFieldCol);
        while(out.length() < 60) out += " ";
        out += " = " + rs.getLong(IDFieldCol) + ";";
        System.out.println(out);
      }
      
      rs.close();
      st.close();
      con.close();
        
    }
    catch(Exception e){
      addToErrorReport("Some occurred: " + e.getMessage());
    }
  }
  
  /**
   * Gets the list of commands in a TeX format
   * @return the TeX formatted list of commands
   */
  public String getTEXListOfCommands(String[] commandsToInclude){
    // Construct Main Header
    StringWriter out = new StringWriter();
    String preamble = "";
    
    Vector<String> include = new Vector<String>();
    Vector<String> categoriesList = new Vector<String>();
    if(commandsToInclude != null) for(String inc: commandsToInclude) include.add(inc);
    
    try{
      Connection con = DB.getGuestConnection();
      
      // Get list of categories      
      Statement st1 = con.createStatement();
      String sql = DB.getCommandCategoriesInOrderView().getSelectAllSQL(DB.getOwner());
      ResultSet categories = st1.executeQuery(sql);
      
      ABCMCommandCategoriesInOrderView commandCategoriesView = DB.getCommandCategoriesInOrderView();
      
      String categoryPKFieldName   = commandCategoriesView.getField_CommandCategory_PK().getName();
      String categoryNameFieldName = commandCategoriesView.getField_CommandCategory_Name().getName();
      String categoryDescFieldName = commandCategoriesView.getField_CommandCategory_Desc().getName();
      
      preamble = "Commands are arranged in the following categories:\n\n";
      preamble += "\\begin{itemize}\n";

      
      // For each category
      while(categories.next()){
        long catPK = categories.getLong(categoryPKFieldName);
        String catName = categories.getString(categoryNameFieldName);
        String catDesc = categories.getString(categoryDescFieldName);
        
        
        // Get list of commands in category
        Statement st2 = con.createStatement();
        sql = DB.getNarrativeCommandsView().getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(DB.getNarrativeCommandsView().getField_CommandCategory_PK().getName(), catPK) + " " +
            AbstractDBGateway.orderBy(DB.getNarrativeCommandsView().getField_NarrativeCommand().getName()));
        ResultSet commands = st2.executeQuery(sql);
        
        ABCMNarrativeCommandsView narrCommandsView = DB.getNarrativeCommandsView();
        
        String commandPKFieldName               = narrCommandsView.getField_NarrativeCommand_PK().getName();
        String narrCommandFieldName             = narrCommandsView.getField_NarrativeCommand().getName();
        String narrCommandNumArgFieldName       = narrCommandsView.getField_NumberOfArguments().getName();
        String narrCommandDescFieldName         = narrCommandsView.getField_Description().getName();
        String narrCommandCommLineOnlyFieldName = narrCommandsView.getField_CommandLineOnly().getName();
        String narrCommandProbeFieldName        = narrCommandsView.getField_Probe().getName();
        String narrCommandInstrFieldName        = narrCommandsView.getField_NarrativeCommand_Instructions().getName();
      
        boolean hasACommand = false;
        StringWriter commandsInThisCategory = new StringWriter();
        
        // For each command
        while(commands.next()){
          
          long commandPK = commands.getLong(commandPKFieldName);
          
          String narrCommand              = commands.getString(narrCommandFieldName);
          long   narrCommandNumArg        = commands.getLong(narrCommandNumArgFieldName);
          String narrCommandDesc          = commands.getString(narrCommandDescFieldName);
          boolean narrCommandCommLineOnly = commands.getBoolean(narrCommandCommLineOnlyFieldName);
          boolean narrCommandProbe        = commands.getBoolean(narrCommandProbeFieldName);
          String narrCommandInstr         = commands.getString(narrCommandInstrFieldName);
    
          // if this command should be included
          if(include.size() == 0 || include.contains(narrCommand)){
            
            hasACommand = true; // Flag- the category has at least one command
            
            // Print command header
            commandsInThisCategory.write("\\subsection*{Command: {\\em " + narrCommand + "}}\n");
            commandsInThisCategory.write(narrCommandInstr + "\n\n");
            
            if(narrCommandNumArg > 0){
            
              Statement st3 = con.createStatement();
              sql = DB.getNarrativeArgumentsView().getSelectSQL(DB.getOwner(), 
                  DB.where(DB.getNarrativeArgumentsView().getField_CommandFK().getName(), commandPK) + " "  + DB.orderBy(DB.getNarrativeArgumentsView().getField_ArgNo().getName()));
              ResultSet arguments = st3.executeQuery(sql);
              
              ABCMNarrativeArgumentsView narrArgsView = DB.getNarrativeArgumentsView();
              
              String argNoFieldName   = narrArgsView.getField_ArgNo().getName();
              String argNameFieldName = narrArgsView.getField_ArgName().getName();
              
              commandsInThisCategory.write("\\begin{tabular}{c|p{6.5cm}}\n");
              commandsInThisCategory.write("Arg No. & Value \\\\\n");
              
              while(arguments.next()){
                // Print each argument
                long   argNo   = arguments.getLong(argNoFieldName);
                String argName = arguments.getString(argNameFieldName);
                
                commandsInThisCategory.write(argNo + " & " + argName + "\\\\\n");
                
              }
              arguments.close();
              st3.close();
              commandsInThisCategory.write("\\end{tabular}\n\n");
              
            }
            else{
              commandsInThisCategory.write("\\noindent\n{\\em This command takes no arguments.}\n\n");
            }
            
            commandsInThisCategory.write("\\noindent\n" + narrCommandDesc + "\n\n");
            commandsInThisCategory.write("\\noindent\nProbe: " + narrCommandProbe + "\n\n");
            commandsInThisCategory.write("\\noindent\nCommand Line Only: " + narrCommandCommLineOnly + "\n\n");
            commandsInThisCategory.write("\n");
          }
        }
        commands.close();
        st2.close();
        // If this category has a command
        if(hasACommand){
          preamble += "  \\item " + catName + "\n";
          
         // Print category header
          out.write("\\section{Command Category: " + catName + "}\n");
          out.write(catDesc + "\n\n");
  
          out.write(commandsInThisCategory.toString());
          
        }
      }
      categories.close();
      st1.close();
      
      preamble += "\\end{itemize}\n\n"; 
      
    }
    catch(Exception E){
      return "";
    }
    return preamble + out.toString();
  }
  
  /**
   * Gets the Help message associated with a particular command 
   * @param commandPK primary key of the command for which help is requested
   * @return the Help message for the specified command
   */
  public String getHelpString(long commandPK){
    String retVal = "";
    retVal += getName(commandPK) + "\n";
    retVal += getDescription(commandPK) + "\n";
    retVal += DB.getStringItem(DB.getNarrativeCommandsTable(), 
                              DB.getNarrativeCommandsTable().getField_PK(), 
                              commandPK, 
                              DB.getNarrativeCommandsTable().getField_Description());
    CommandArgumentManager CAM = new CommandArgumentManager(DB);
    retVal += CAM.getHelpString(commandPK);    
    return retVal;
  }
  
  /**
   * Gets the Help message associated with a particular command
   * @param command the command
   * @return the Help message for the specified command
   */
  public String getHelpString(String command){
    return getHelpString(getPK(command));
  }
  
  /**
   * Returns the list of valid commands as an option list, in HTML
   * format, for insertion into an HTML document as a drop-down
   * selection box. Does not include commands marked with
   * 'blnCommandLineOnly = true'
   * @return a String representing an HTML options list, listing
   * the valid commands for narrative instructions.
   */
  public String getInterfaceCommandsAsOptionList(){
    String retVal = "";
    
    try{
      String sql = DB.getNarrativeCommandsTable().getSelectSQL(DB.getOwner(), 
              AbstractDBGateway.where(DB.getNarrativeCommandsTable().getField_CommandLineOnly().getName(), 0)) + " " +
              AbstractDBGateway.orderBy(DB.getNarrativeCommandsTable().getField_Command().getName());
      
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      
      ResultSet rs = st.executeQuery(sql);
      retVal = "";
      
      int commandIDFieldCol = rs.findColumn(DB.getNarrativeCommandsTable().getField_PK().getName());
      int commandFieldCol   = rs.findColumn(DB.getNarrativeCommandsTable().getField_Command().getName());
      
      while(rs.next()){
        retVal += "<OPTION VALUE=" + rs.getInt(commandIDFieldCol) + ">";
        retVal += rs.getString(commandFieldCol);
        retVal += "</OPTION>";
      }
    }
    catch (Exception e){
      addToErrorReport("CommandsManager getInterfaceCommandsAsOptionsList()." + e.getMessage());
    }  
    
    return retVal;
  }

  /**
   * Get the PK value of the command represented by the given string
   * @return A long representing the command given by the string
   */
  public long getCommandPK(String com){
    long comPK = 0;
    // First, if the string is numeric, just return it as a long
    try{ return Long.parseLong(com); } catch (Exception e){}
    
    // Next get the value from the database
    comPK = getCommandPKFromCommandList(com);
    if(comPK > 0) return comPK;
    
    // Last, try the offline version; this can be overloaded if it has to be
    return offlineParse(com);
    
  }
  
  /**
   * Gets the PK for a specific command from the command list
   * @param com command sought
   * @return PK of that command. If not found, returns zero.
   */
  private long getCommandPKFromCommandList(String com){
    if(commandList == null) initCommandList();
    String command = com.toUpperCase();
    if(!commandList.containsKey(command)) return 0;    
    else return ((Integer)commandList.get(command)).longValue(); 
  }
  
  
  /**
   * Gets the String version of the command given the PK
   * @param commandPK PK sought
   * @return String version of command. If no match is found, returns "UNKNOWN"
   */
  public String getCommand(int commandPK){
    if(numericCommandList == null) initCommandList();
    if(numericCommandList.containsKey(new Integer(commandPK))){
      return (String)numericCommandList.get(new Integer(commandPK));
    }      
    return "UNKNOWN";
  }

  /**
   * Initialize the command list.
   * The command list is the local copy, as a Hashtable, of the
   * database's list of commands in String form and ID numbers
   */  
  private void initCommandList(){
    commandList        = new Hashtable();
    numericCommandList = new Hashtable();

    try{
      String sql = DB.getNarrativeCommandsTable().getSelectAllSQL(DB.getOwner());
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();

      ResultSet rs = st.executeQuery(sql);
      
      int commandIDFieldCol  = rs.findColumn(DB.getNarrativeCommandsTable().getField_PK().getName());
      int commandFieldCol    = rs.findColumn(DB.getNarrativeCommandsTable().getField_Command().getName());
      while (rs.next()){
        commandList.put((rs.getString(commandFieldCol)).toUpperCase(), new Integer(rs.getInt(commandIDFieldCol)));
        numericCommandList.put(new Integer(rs.getInt(commandIDFieldCol)), rs.getString(commandFieldCol));
      }
    }
    catch (Exception e){
      addToErrorReport("Error in CommandsManager initCommandList(): " + e.getMessage());      
    }    
  }
  
  // Intuitively this looks like it should be in this manager, so we'll place a convenience
  // method here; in fact the list is created by the larger outline level of 'Command
  // Category'
  /**
   * Gets the list of commands as a simple String. 
   * @return the list of commands as a String
   */
  public String getListOfCommandsAsString(){                
    return CCM.getListOfCommandsAsString(); 
  }
  
  /**
   * Gets the PK of the specified Command Category
   * @param category category sought
   * @return PK of category. If not found, returns -1.
   */
  public long getCommandCategoryPK(String category){
    return CCM.getPK(category);    
  }
  
  
  /**
   * Overload this to allow the simulation to work when it is
   * not online. The idea is that this method translates a given
   * command (such as 'Login') into a numeric code; it should
   * duplicate the PK codes for the commands in the database.
   * @param parseString
   * @return zero
   */
  protected long offlineParse(String parseString) {  return 0;  }

  /**
   * Gets the list of command categories as an option list
   * @return the HTML option list
   */
  public String getListOfCommandCategoriesAsOptionList(){
    return CCM.getOptionList(true, false);
  }
  
}
