package abcm.db.managers;


import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.Enumeration;
import java.util.Hashtable;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMObjectCommentTable;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMCommentsAndResearchersView;
import abcm.db.components.views.ABCMReferenceAuthorsView;
import abcm.db.components.views.ABCMReferencesAndCommentsView;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.db.managers.abstractparents.AbstractObjectCommentManager;
import abcm.utilities.XMLAssist;

/**
 * Manages comments entered into the HWM database. This includes 
 * adding comments and retrieving the list of comments in 
 * HTML format.<br/><br/>
 * 
 * Comment Manager inner classes are not provided JavaDoc; see their parent class
 * for details.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class CommentManager extends AbstractNonAuditableDBManager {
  
  /** Internal table of comment management objects  */
  private Hashtable objectCommentManagers = new Hashtable();
  
  /** Researcher ID to save  */
  private long researcherIDToSave;
  /** Text of comment to save  */
  private String entryTextToSave;
  
  /** Comment reference manager  */
  private CommentReferencesManager CRM;

  private class CommentReferencesManager extends AbstractDBLinkTableInnerManager{

    private long                     referenceFKToSave;
    private String                   pagesToSave;
    private long                     commentFKToSave;
        
    /* Constructor */
    public CommentReferencesManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }
    
    // Base table
    public ABCMdbTable baseTable() {                       return DB.getReferenceCommentsTable();                                  }

    protected ABCMdbField baseTablePKField(){              return DB.getReferenceCommentsTable().getField_PK();                    }

    protected ABCMdbField baseTablePrimaryFKField(){       return DB.getReferenceCommentsTable().getField_CommentFK();             }

    protected ABCMdbField baseTableSecondaryFKField(){     return DB.getReferenceCommentsTable().getField_ReferenceFK();           }

    protected ABCMdbAbstractViewable recordsView(){        return DB.getReferencesAndCommentsView();                               }
    
    protected ABCMdbField recordsViewPKField(){            return DB.getReferencesAndCommentsView().getField_PK();                 }

    protected ABCMdbField recordsViewPrimaryFKField(){     return DB.getReferencesAndCommentsView().getField_Comment_PK();         }

    protected ABCMdbField recordsViewPrimaryNameField(){   return DB.getReferencesAndCommentsView().getField_Reference_PK();       }

    protected ABCMdbField recordsViewSecondaryNameField(){ return DB.getReferencesAndCommentsView().getField_Reference_Title();    }    
    
    /* XML */

    protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
      ABCMReferencesAndCommentsView view = DB.getReferencesAndCommentsView();
      
      int commentIDFieldCol        = rs.findColumn(view.getField_Comment_PK().getName());
      int commentEntryDateFieldCol = rs.findColumn(view.getField_Comment_Entry_Date().getName());
      int referencePKFieldCol      = rs.findColumn(view.getField_Reference_PK().getName());
      int referenceDateFieldCol    = rs.findColumn(view.getField_Reference_Date().getName());
      int referenceTitleFieldCol   = rs.findColumn(view.getField_Reference_Title().getName());
      int pagesFieldCol            = rs.findColumn(view.getField_Pages().getName());
      
      boolean executedFlag = false;
      if(rs.next()){
        xml.open("CommentRefs");
        executedFlag = true;
          do{
            xml.open("CommentRef", "CommentRefID=\"" + rs.getLong(referencePKFieldCol) + "\"");
              xml.addLine("CommentRefID",    rs.getLong(referencePKFieldCol));
              xml.addLine("CommentRefTitle", rs.getString(referenceTitleFieldCol));
              xml.addLine("CommentRefDate",  rs.getString(referenceDateFieldCol));
              xml.addLine("CommentRefPages", rs.getString(pagesFieldCol));
              
              // Get the authors for this ref
              ABCMReferenceAuthorsView refAuthView = DB.getReferenceAuthorsView();
              
              String sqlRefAuthors = refAuthView.getSelectSQL(DB.getOwner(), 
                  AbstractDBGateway.where(refAuthView.getField_Reference_PK().getName(), rs.getLong(referencePKFieldCol)) + " " + 
                  AbstractDBGateway.orderBy(refAuthView.getField_AuthorOrder().getName()));
              
              Connection con = DB.getAdminConnection();
              Statement st = con.createStatement();
              ResultSet rs2 = st.executeQuery(sqlRefAuthors);
            
              int refAuthPKFieldCol             = rs2.findColumn(refAuthView.getField_PK().getName());
              int refAuthRefPKFieldCol          = rs2.findColumn(refAuthView.getField_Reference_PK().getName());
              int refAuthOrderFieldCol          = rs2.findColumn(refAuthView.getField_AuthorOrder().getName());
              int refAuthResearcherPKFieldCol   = rs2.findColumn(refAuthView.getField_Researcher_PK().getName());
              int refAuthResearcherNameFieldCol = rs2.findColumn(refAuthView.getField_Researcher_LastNameFirstName().getName());
              
              xml.open("RefAuthors");
                while(rs2.next()){            
                  xml.open("author");
                    xml.addLine("order",      rs2.getInt(refAuthOrderFieldCol));
                    xml.addLine("authorName", rs2.getString(refAuthResearcherNameFieldCol));
                  xml.closeTag();
                } 
              xml.closeTag(); // RefAuthors
              rs2.close();
              st.close();
            xml.closeTag(); // CommentRef            
          }while(rs.next());
        if(executedFlag) xml.closeTag(); // Closes CommentRefs tag
        
      }
      
      
    }
    
    public void addSetToXML(XMLAssist xml, long PK, boolean includeHeaders) throws Exception{
      String sql = DB.getReferencesAndCommentsView().getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(DB.getReferencesAndCommentsView().getField_Comment_PK().getName(), PK));
      Connection con = DB.getAdminConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      addInstancesToXML(xml, rs, includeHeaders);
      
      rs.close();
      st.close();
      con.close();
    }

    /* Saving */
    
    public void setInstanceToSave(long referenceFK, String pages, long comment){
      referenceFKToSave = referenceFK;
      pagesToSave       = pages;
      commentFKToSave   = comment;
    }
    
    public boolean instanceToSaveIsValid(){
      return ((referenceFKToSave != -1  ) && 
              (pagesToSave       != null) &&
              (commentFKToSave   != -1  ));
    }
    
    protected void clearInstanceToSave(){
      referenceFKToSave = -1;
      pagesToSave       = null;
      commentFKToSave   = -1;      
    }

    protected ABCMdbDatum datumToSave(){
      return (instanceToSaveIsValid() ? DB.getReferenceCommentsTable().getDatum("" + referenceFKToSave, pagesToSave, "" + commentFKToSave): null);
    }

  } 
    
  /* Comment Managers */
  
  private class SimulationCommentManager     extends AbstractObjectCommentManager {

    public SimulationCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {   return DB.getSimulationCommentsTable();   }
    
    public ABCMdbField getCommentIDField() {         return DB.getSimulationCommentsTable().getField_CommentFK();      }

    public ABCMdbField getObjectIDField() {          return DB.getSimulationCommentsTable().getField_SimulationFK();   }
    
  }

  private class DataModelCommentManager      extends AbstractObjectCommentManager {

    public DataModelCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {  return DB.getDataModelCommentsTable();      }
    
    public ABCMdbField getCommentIDField() {        return DB.getDataModelCommentsTable().getField_CommentFK();         }

    public ABCMdbField getObjectIDField() {         return DB.getDataModelCommentsTable().getField_DataModelFK();       }

  }
  
  private class ConfigurationCommentManager  extends AbstractObjectCommentManager {

    public ConfigurationCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {  return DB.getSimulationConfigurationCommentsTable();                }
    
    public ABCMdbField getCommentIDField() {        return DB.getSimulationConfigurationCommentsTable().getField_CommentFK();                   }

    public ABCMdbField getObjectIDField() {         return DB.getSimulationConfigurationCommentsTable().getField_SimulationConfigurationFK();   }

  }
  
  private class HistoryCommentManager        extends AbstractObjectCommentManager {

    public HistoryCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {  return DB.getHistoryCommentsTable();   }
    
    public ABCMdbField getCommentIDField() {        return DB.getHistoryCommentsTable().getField_CommentFK();      }

    public ABCMdbField getObjectIDField() {         return DB.getHistoryCommentsTable().getField_HistoryFK();      }

  }
  
  private class NarrativeCommentManager      extends AbstractObjectCommentManager {
    
    public NarrativeCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {   return DB.getNarrativeCommentsTable();   }
    
    public ABCMdbField getCommentIDField() {         return DB.getNarrativeCommentsTable().getField_CommentFK();      }

    public ABCMdbField getObjectIDField() {          return DB.getNarrativeCommentsTable().getField_NarrativeFK();    }

  }
  
  private class CodeModelCommentManager      extends AbstractObjectCommentManager {
    
    public CodeModelCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {   return DB.getCodeModelCommentsTable();   }
    
    public ABCMdbField getCommentIDField() {         return DB.getCodeModelCommentsTable().getField_CommentFK();      }

    public ABCMdbField getObjectIDField() {          return DB.getCodeModelCommentsTable().getField_CodeModelFK();    }

  }
  
  private class ParameterSetCommentManager   extends AbstractObjectCommentManager {
     
    public ParameterSetCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {  return DB.getParameterSetCommentsTable();     }
    
    public ABCMdbField getCommentIDField() {        return DB.getParameterSetCommentsTable().getField_CommentFK();        }

    public ABCMdbField getObjectIDField() {         return DB.getParameterSetCommentsTable().getField_ParameterSetFK();   }

  }
  
  private class ParameterCommentManager      extends AbstractObjectCommentManager {
    
    public ParameterCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {  return DB.getParameterCommentsTable();   }
   
    public ABCMdbField getCommentIDField() {        return DB.getParameterCommentsTable().getField_CommentFK();      }

    public ABCMdbField getObjectIDField() {         return DB.getParameterCommentsTable().getField_ParameterFK();    }
   
  }
  
  private class ProbeCommentManager          extends AbstractObjectCommentManager {
    
    public ProbeCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {  return DB.getProbeCommentsTable();   }
    
    public ABCMdbField getCommentIDField() {        return DB.getProbeCommentsTable().getField_CommentFK();      }

    public ABCMdbField getObjectIDField() {         return DB.getProbeCommentsTable().getField_ProbeFK();        }

  }
  
  private class ProbeSetCommentManager       extends AbstractObjectCommentManager {
    
    public ProbeSetCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {   return DB.getProbeSetCommentsTable();   }
    
    public ABCMdbField getCommentIDField() {         return DB.getProbeSetCommentsTable().getField_CommentFK();      }

    public ABCMdbField getObjectIDField() {          return DB.getProbeSetCommentsTable().getField_ProbeSetFK();     }
    
  }
  
  private class AnalysisCommentManager       extends AbstractObjectCommentManager {
    
    public AnalysisCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {   return DB.getAnalysisCommentsTable();  }
    
    public ABCMdbField getCommentIDField() {         return DB.getAnalysisCommentsTable().getField_CommentFK();     }

    public ABCMdbField getObjectIDField() {          return DB.getAnalysisCommentsTable().getField_AnalysisFK();    }
    
  }
  
  private class SummaryCommentManager        extends AbstractObjectCommentManager {
    
    public SummaryCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    public ABCMObjectCommentTable getBaseTable() {   return DB.getSummaryCommentsTable();   }
    
    public ABCMdbField getCommentIDField() {         return DB.getSummaryCommentsTable().getField_CommentFK();      }

    public ABCMdbField getObjectIDField() {          return DB.getSummaryCommentsTable().getField_SummaryFK();      }   
    
  }  
  
  /* Static methods */
  
  /**
   * Retrieves a standard HTML form that allows comments to be entered on any calling
   * web page.
   * @param action the text to be used as the HTML Form's 'action' attribute
   * @param callingObjectValue a value that represents the kind of object calling
   * this web page. These values are generally retrieved from the DB object, which
   * provides static values corresponding to each of the kinds of objects that can
   * receive comments.
   * @param PKValue the primary key of whatever object is currently in the web page
   * that is calling this, and to which the comment will be added.
   * @param redirect a String that is used to redirect the web site after extended
   * comments are added; this is generally the name of the calling web page, along with
   * the PK of the object being commented on, so that the user is returned back to the
   * original object after the extended comment is entered
   * @return a String containing the HTML for a form to enter comments
   */
  public static String getCommentAddForm(String action, long callingObjectValue, 
      long PKValue, String redirect){
    String retVal = 
      "<!-- Add a comment -->" + "\n" +
      "<h2 class=\"standard\">Add a comment</h2>" + "\n" +
      "<table class=\"formInput\">" + "\n" + 
      "  <form name=\"addComment\" action=\"" + action + "\" method=\"post\">" + "\n" +
      "    <input type=\"hidden\" name=\"callingObject\" value=\"" + callingObjectValue + "\"></input>" + "\n" +            
      "    <input type=\"hidden\" name=\"PK\" value=\"" + PKValue + "\"></input>" + "\n" +
      "    <input type=\"hidden\" name=\"redirect\" value=\"" + redirect + "\"></input>" + "\n" +
      "    <tr><td class=\"formInput\" colspan=\"2\"><center><textarea name=\"comment\" rows=\"5\" cols=\"80\"></textarea></center></td></tr>" + "\n" +
      "    <tr><td class=\"formInputHeader\"><center><input type=\"checkbox\" name=\"LINK\">Link comment to additional objects</input></center></td></tr>" + "\n" +
      "    <tr><td class=\"formSubmitLabel\"><center><input type=\"submit\" value=\"Add this comment\"/></center></td></tr>" + "\n" +
      "  </form>" + "\n" +
      "</table>" + "\n";
      
    return retVal;
  }
  
  
  /* Constructor */
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public CommentManager(AbstractABCMDBGateway db){    
    super(db);
    CRM = new CommentReferencesManager(DB, this);
    
    addObjectCommentManager(AbstractABCMDBGateway.SIMULATION,     new SimulationCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.DATA_MODEL,     new DataModelCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.CONFIGURATION,  new ConfigurationCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.HISTORY,        new HistoryCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.NARRATIVE,      new NarrativeCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.CODE_MODEL,     new CodeModelCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.PARAM_SET,      new ParameterSetCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.PARAMETER,      new ParameterCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.PROBE,          new ProbeCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.PROBE_SET,      new ProbeSetCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.ANALYSIS,       new AnalysisCommentManager(DB, this));
    addObjectCommentManager(AbstractABCMDBGateway.SUMMARY,        new SummaryCommentManager(DB, this));    
  }

  /**
   * Adds an object comment manager to the collection of object comment managers
   * herein
   * @param key unique integer key for referring to this manager; usually
   * corresponds to a specific db object type
   * @param manager the manager to be added
   * @return true if successful. Will not be successful if key is a duplicate of one already entered.
   */
  public boolean addObjectCommentManager(int key, AbstractObjectCommentManager manager){
    Integer K = new Integer(key);
    if(objectCommentManagers.containsKey(K)) {
      addToErrorReport("Cannot add manager with key '" + key + "'; a manager with that key already exists.");
      return false;
    }
    objectCommentManagers.put(K, manager);
    return true;
  }
  
  /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){                  return DB.getCommentsTable();                                                }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){           return DB.getCommentsTable().getField_PK();                                  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView() {    return DB.getCommentsAndResearchersView();                                   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField() {        return DB.getCommentsAndResearchersView().getField_Comment_PK();             }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField() {        return DB.getCommentsAndResearchersView().getField_Comment_DateEntryMade();  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField() 
   */
  protected ABCMdbField baseTableDescriptionField(){  return DB.getCommentsAndResearchersView().getField_Comment_Text();           }

  /**
   * Gets the date field from the records view.
   * @return the records view's Date field
   */
  protected ABCMdbField recordsViewDateField(){       return DB.getCommentsAndResearchersView().getField_Comment_DateEntryMade();  }
    
  /* XML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName() {               return "Comments";                                                           }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
    ABCMCommentsAndResearchersView view = DB.getCommentsAndResearchersView();
    
    int commentPKFieldCol      = rs.findColumn(view.getField_Comment_PK().getName());
    int commentDateFieldCol    = rs.findColumn(view.getField_Comment_DateEntryMade().getName());
    int commentTextFieldCol    = rs.findColumn(view.getField_Comment_Text().getName());
    int researcherPKFieldCol   = rs.findColumn(view.getField_Researcher_PK().getName());
    int researcherNameFieldCol = rs.findColumn(view.getField_Researcher_Name().getName());
    
    while(rs.next()){
      xml.open("Comment");
        xml.addLine("ID",          rs.getInt(commentPKFieldCol));
        xml.addLine("Researcher",  rs.getString(researcherNameFieldCol));
        xml.addLine("DateOfEntry", (new SimpleDateFormat()).format(rs.getTimestamp(commentDateFieldCol)));
        xml.addLine("Entry",       rs.getString(commentTextFieldCol));
        
//      if(singleInstance){
        CRM.addSetToXML(xml, rs.getInt(commentPKFieldCol), includeComments);
                  
      xml.closeTag();
    }  
  }

  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlCommentListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlCommentToHTML.xsl"; }
    
  /* Saving */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  protected void clearInstanceToSave(){           
    researcherIDToSave = -1;
    entryTextToSave    = null;
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return ((researcherIDToSave != -1  ) && 
            (entryTextToSave    != null));
  }
  
  /**
   * Sets a comment to be saved
   * @param ID the user ID of the commentor
   * @param entry text of the comment entry
   */
  public void setInstanceToSave(long ID, String entry){
    researcherIDToSave = ID;
    entryTextToSave    = AbstractABCMDBGateway.makeSQLSafe(entry);
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave() {
    return (instanceToSaveIsValid() ? 
        DB.getCommentsTable().getDatum("" + researcherIDToSave, DB.currentTimeString(), entryTextToSave): null);   
  }
  
  
  /**
   * Associates a specific comment to a collection of references in the database. Note:
   * behavior is not atomic and may fail with some references associated and others not.
   * @param commentID PK of comment being referenced
   * @param refList list of references (as a list of Strings but each String must represent
   * a PK of the reference being attached; will fail if a String that cannot be parsed to a long
   * integer is passed)
   * @return true if successful, false otherwise
   */
  public boolean associateCommentToRefs(long commentID, String[] refList){
    boolean retVal = true;
    if(refList == null) return retVal;
    for(int i = 0; i < refList.length; i++){
      long nextLong;
      try{  nextLong = Long.parseLong(refList[i]);      }
      catch(NumberFormatException nfe){       nextLong = -1;        }
      if(nextLong > -1) retVal = associateCommentToRef(commentID, nextLong, "General" ) && retVal;  // One false will poison all, but all execute
      else retVal = false;
    }
    return retVal;
  }

  
  /**
   * Gets the appropriate comment manager for the object type indicated by the key passed.
   * @param objectType integer specifying the object type
   * @return the appropriate comment manager; null if none matching type is found
   */
  protected AbstractObjectCommentManager getObjectCommentManager(int objectType){
    Integer K = new Integer(objectType);
    if(objectCommentManagers.containsKey(K)) return (AbstractObjectCommentManager)objectCommentManagers.get(K);
    else                                     return null;
  }

  /**
   * Associates a comment to a single reference
   * @param commentID comment PK
   * @param refID reference PK
   * @param pages pages specification; comments may refer to specific pages in references
   * @return true if successful
   */
  public boolean associateCommentToRef(long commentID, long refID, String pages){
    CRM.setInstanceToSave(refID, pages, commentID);
    return CRM.saveInstanceToDB();
  }
  
  /**
   * Associates a comment to a single database object.
   * @param objectType integer identifying the kind of object
   * @param commentID PK of the comment to be associated
   * @param PK PK of the object to be associated
   * @return TISFO
   */
  public boolean associateCommentToObject(int objectType, long commentID, long PK){
    boolean successFlag = false;
    if(objectType == AbstractABCMDBGateway.REFERENCE){
      return associateCommentToRef(commentID, PK, "General");      
    }
    AbstractObjectCommentManager ocm = getObjectCommentManager(objectType);
    if(ocm == null){
      addToErrorReport("Could not retrieve object comment manager for object type: " + objectType);
      return false;
    }
    else{
      ocm.setInstanceToSave(PK, commentID);
      return ocm.saveInstanceToDB();
    }
  }
 
  /**
   * Associates a list of objects (of the same kind) to a comment
   * @param objectType key indicating the kind of object
   * @param commentID PK of comment to be associated
   * @param valuesString array of String values representing the PKs of
   * the objects to be associated. Inclusion of a non-numeric value will
   * not cause the method to fail; the value will be ignored without
   * comment.
   */
  public void associateListOfObjectsToComment(int objectType, long commentID, String[] valuesString){
    if(valuesString == null) return;
    for(int i = 0; i < valuesString.length; i++){
      try{ associateCommentToObject(objectType, commentID, Long.parseLong(valuesString[i]));       }
      catch(NumberFormatException nfe){ } // Do nothing... bad form!      
    }
  }
  
  /**
   * Gets the list of comments for a particular object as XML
   * @param objectType integer key indicating the kind of object
   * @param PK PK of the object
   * @return XML representing all comments for that object
   */
  public String getListOfCommentsForObjectAsXML(int objectType, long PK){
    String sql = "";
    if(objectType == AbstractABCMDBGateway.REFERENCE){
      sql = "SELECT * FROM " + recordsView().getFullName(DB.getOwner()) + " " +
                   "INNER JOIN " + DB.getReferenceCommentsTable().getFullName(DB.getOwner()) + " ON " + DB.getReferenceCommentsTable().getField_CommentFK().getFullFieldName(DB.getOwner()) + " = " + recordsViewPKField().getFullFieldName(DB.getOwner()) + " " +
                   "WHERE " + DB.getReferenceCommentsTable().getField_ReferenceFK().getFullFieldName(DB.getOwner()) + " = " + PK + " " +
                   "ORDER BY " + recordsViewDateField().getFullFieldName(DB.getOwner());
    }
    else{
      AbstractObjectCommentManager ocm = (AbstractObjectCommentManager)objectCommentManagers.get(new Integer(objectType));
      if(ocm == null) return "";
      sql = "SELECT * FROM " + recordsView().getFullName(DB.getOwner()) + " " +
                   "INNER JOIN " + ocm.getBaseTable().getFullName(DB.getOwner()) + " ON " + ocm.getCommentIDField().getFullFieldName(DB.getOwner()) + " = " + recordsViewPKField().getFullFieldName(DB.getOwner()) + " " +
                   "WHERE " + ocm.getObjectIDField().getFullFieldName(DB.getOwner()) + " = " + PK + " " +
                   "ORDER BY " + recordsViewDateField().getFullFieldName(DB.getOwner());
    }
    return getXML(sql, false, false);
  }
  
  /**
   * Gets the list of comments for an object as HTML
   * @param objectType integer key indicating the kind of object
   * @param PK PK of the object
   * @param xslRootDir directory in which XSL stylesheet is found
   * @return the HTML list of comments
   */
  public String getListOfCommentsForObjectAsHTML(int objectType, long PK, String xslRootDir){
    return getHTML(getListOfCommentsForObjectAsXML(objectType, PK), false, xslRootDir, getListStylesheetName());
  }
  
  
  
  /**
   * Removes a comment from the Database. Deletes not only the
   * comment but also all associations of that comment to various objects
   * @param commentID the primary key of the comment being deleted
   * @return TISFO
   */
  public boolean deleteItem(long commentID){
    CRM.deleteElementsByPrimaryFK(commentID);    
    Enumeration e = objectCommentManagers.elements();
    while(e.hasMoreElements()){
      AbstractObjectCommentManager ocm = (AbstractObjectCommentManager)e.nextElement();
      ocm.deleteElementsByPrimaryFK(commentID);
    }
    return super.deleteItem(commentID);    
  }

  
  
  /**
   * Gets a comment in a TeX format
   * @param commentID PK of the comment
   * @return TeX version of comment text
   */
  public String getCommentTEX(long commentID){
    StringWriter out = new StringWriter();
    
    try{
      Connection con = DB.getAdminConnection();
      Statement st1 = con.createStatement();
      
      ResultSet entry = st1.executeQuery(DB.getCommentsAndResearchersView().getSelectSQL(DB.getOwner(), 
          AbstractABCMDBGateway.where(DB.getCommentsAndResearchersView().getField_Comment_PK().getFullFieldName(DB.getOwner()), commentID)));
      
      if(entry.next()){
        String commentText   = entry.getString(DB.getCommentsAndResearchersView().getField_Comment_Text().getName());
        String commentAuthor = entry.getString(DB.getCommentsAndResearchersView().getField_Researcher_Name().getName());
        String commentDate   = entry.getString(DB.getCommentsAndResearchersView().getField_Comment_DateEntryMade().getName());
        
        out.write("\\subsubsection*{Comment by " + commentAuthor + ", " + commentDate + "} " + commentText + "\n");
        
        // Get references
        Statement st2 = con.createStatement();
        
        ResultSet refs = st2.executeQuery(DB.getReferenceCommentsTable().getSelectSQL(DB.getOwner(), 
            AbstractABCMDBGateway.where(DB.getReferenceCommentsTable().getField_CommentFK().getName(), commentID)));
        
        String refString = "";
        ReferenceManager RM = new ReferenceManager(DB);
          
        while(refs.next()){
          refString += RM.getReferenceAsTEX(refs.getLong(DB.getReferenceCommentsTable().getField_ReferenceFK().getName()));
        }
        
        if(refString.length() > 0){
          out.write("\n\\noindent\n{\\em References:}\n\n" + refString);
        }
        out.write("\n");
      }      
    }
    catch(Exception E){
      return "";
    }
    
    return out.toString();
  }
  
}
