package abcm.db.managers;


import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.text.SimpleDateFormat;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMDataModelsWithNamesOfElementsView;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.sim.DataModel;
import abcm.utilities.XMLAssist;

/**
 * Provides an interface to manage Data Models as stored in the HWM Database,
 * including routines to retrieve the list of Data Models in XML and HTML formats.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class DataModelManager extends AbstractAuditableDBManager {

  /** Name to be saved  */
  private String            nameToSave;
  /** Description to be saved  */
  private String            descToSave;
  /** History ID to be saved  */
  private long              historyIDToSave;
  /** Configuration ID to be saved  */
  private long              configurationIDToSave;
  /** User ID to be saved  */
  private long              userIDToSave;

  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public DataModelManager(AbstractABCMDBGateway db){
    super(db);
  }

  /* General */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable() {            return DB.getDataModelsTable();                                                 }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField() {     return DB.getDataModelsTable().getField_PK();                                   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){       return DB.getDataModelsTable().getField_Name();     }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){   return DB.getDataModelsWithNamesOfElementsView();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){       return DB.getDataModelsWithNamesOfElementsView().getField_DataModel_PK();       }
  
  
  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){                return "dataModels"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMDataModelsWithNamesOfElementsView view = DB.getDataModelsWithNamesOfElementsView();
    
    int PKFieldCol                   = rs.findColumn(view.getField_DataModel_PK().getName());
    int nameFieldCol                 = rs.findColumn(view.getField_DataModel_Name().getName());
    int descFieldCol                 = rs.findColumn(view.getField_DataModel_Desc().getName());
    int dateCreatedFieldCol          = rs.findColumn(view.getField_DataModel_DateCreated().getName());
    int researcherFieldCol           = rs.findColumn(view.getField_Researcher_Name().getName());
    int canBeEditedFieldCol          = rs.findColumn(view.getField_CanBeEdited().getName());
    int historyTitleFieldCol         = rs.findColumn(view.getField_History_Title().getName());
    int historyPKFieldCol            = rs.findColumn(view.getField_History_PK().getName());
    int configurationNameFieldCol    = rs.findColumn(view.getField_SimulationConfiguration_Name().getName());
    int configurationPKFieldCol      = rs.findColumn(view.getField_SimulationConfiguration_PK().getName());
    int truthValueFieldCol           = rs.findColumn(view.getField_TruthValue().getName());
    
    // Not used:
//    int truthValuePKFieldCol         = rs.findColumn(view.getField_TruthValue_PK().getName());
//    int combinedTruthValueFieldCol   = rs.findColumn(view.getField_CombinedTruthValue().getName());
    
    while(rs.next()){
      xml.open("dataModel");
        xml.addLine("ID",            rs.getString(PKFieldCol));
        xml.addLine("name",          rs.getString(nameFieldCol));
        xml.addLine("desc",          rs.getString(descFieldCol));
        xml.addLine("dateCreated",   (new SimpleDateFormat()).format(rs.getDate(dateCreatedFieldCol)));
        xml.addLine("createdBy",     rs.getString(researcherFieldCol));
        xml.addLine("canBeEdited",   ((rs.getString(canBeEditedFieldCol) != null) ? "Yes" : "No"));
        xml.addLine("history",       rs.getString(historyTitleFieldCol));
        xml.addLine("historyID",     rs.getLong(historyPKFieldCol));
        xml.addLine("configuration", rs.getString(configurationNameFieldCol));
        xml.addLine("configID",      rs.getLong(configurationPKFieldCol));
        xml.addLine("truthValue",    rs.getString(truthValueFieldCol));
        
        if(includeComments){
          CommentManager cm = new CommentManager(DB);
          xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.DATA_MODEL, instancePK), XMLAssist.INDENT_BLOCK);
        }
        
      xml.closeTag();
    }      
  }

  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlDataModelListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlDataModelToHTML.xsl"; }
    
  /**
   * Sets the information required to save a new Data Model to
   * the HWM Database.
   * @param name name of the Data Model
   * @param desc description of the Data Model
   * @param histID primary key of the History included in this Data Model
   * @param configID primary key of the Configuration included in this Data Model
   * @param userID ID of the researcher who is saving this Data Model
   */
  public void setInstanceToSave(String name, String desc, long histID, long configID, long userID){
    nameToSave = name;
    descToSave = desc;
    historyIDToSave = histID;
    configurationIDToSave = configID;
    userIDToSave = userID;
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    nameToSave            = null;
    descToSave            = null;
    historyIDToSave       = -1;
    configurationIDToSave = -1;
    userIDToSave          = -1;
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){
    return DB.getDataModelsTable().getDatum(nameToSave, descToSave, "" + historyIDToSave, 
        "" + configurationIDToSave, DB.currentTimeString(), "" + userIDToSave);
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return (nameToSave != null);
  }
  
  /**
   * Gets the Data Model identified by the specified primary key.
   * @param dataModelPK primary key of the Data Model to be retrieved
   * @return the requested Data Model
   */
  public DataModel getDataModel(long dataModelPK){
    DataModel dm = null;
    try{   
      String sql = DB.getDataModelsTable().getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(DB.getDataModelsTable().getField_PK().getName(), dataModelPK));
      
      Connection con = DB.getGuestConnection(); 
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      
      
      if(rs.next()){
        dm = new DataModel(rs.getString(DB.getDataModelsTable().getField_Name().getName()),
                           rs.getString(DB.getDataModelsTable().getField_Description().getName()),
                           rs.getLong(DB.getDataModelsTable().getField_HistoryFK().getName()),
                           rs.getLong(DB.getDataModelsTable().getField_ConfigurationFK().getName()));
      }
      rs.close();
      st.close();
      con.close();
      
    }
    catch(Exception e) {
      addToErrorReport("Error getting DataModel. " + e.getMessage());
    }
    return dm;
  }

  
  /* Auditable */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField             baseTableOwnerField(){    return DB.getDataModelsTable().getField_ResearcherFK();            }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable  canBeEditedView(){        return DB.getCanBeEdited_DataModelsView();                         }        
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField             canBeEditedViewPKField(){ return DB.getCanBeEdited_DataModelsView().getField_DataModel_PK(); }




}
