package abcm.db.managers;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMTermsTable;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.utilities.XMLAssist;


/**
 * Manages the database entries for the glossary of terms
 * in the HWM database.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class GlossaryManager extends AbstractNonAuditableDBManager {

/** Term to be saved  */
private String termToSave;
/** Description to be saved  */
private String descriptionToSave;
/** Abbreviated form of term (used as HTML anchor) to be saved  */
private String anchorToSave;

  /* Constructor */
 
  /**
   * Constructor
   * @param db db gateway to be used
   */
  public GlossaryManager(AbstractABCMDBGateway db){
    super(db);
  }
  
  /* General */
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  public ABCMdbTable baseTable(){                   return DB.getTermsTable();                        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  public ABCMdbField baseTablePKField(){            return DB.getTermsTable().getField_PK();          }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){          return DB.getTermsTable().getField_Term();        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField() 
   */
  public ABCMdbField baseTableDescriptionField(){   return DB.getTermsTable().getField_Description(); }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){      return DB.getTermsTable();                        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){          return DB.getTermsTable().getField_PK();          }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){                 return "terms";                                    }
  
  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMTermsTable table = DB.getTermsTable();
    
    int PKFieldCol          = rs.findColumn(table.getField_PK().getName());
    int termFieldCol        = rs.findColumn(table.getField_Term().getName());
    int anchorFieldCol      = rs.findColumn(table.getField_Anchor().getName());
    int descriptionFieldCol = rs.findColumn(table.getField_Description().getName());
    
    while(rs.next()){
      xml.open("term", "NUMBER=\"" + rs.getString(DB.getTermsTable().getField_PK().getName()) + "\"");
      xml.addLine("term",            rs.getString(termFieldCol));
      xml.addLine("anchor",          rs.getString(anchorFieldCol));
      xml.addLine("description",     rs.getString(descriptionFieldCol));
      xml.closeTag();
    }
  }  

  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlTermListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlTermToHTML.xsl"; }
  
  
  /* Setting instances to save */
  
  /**
   * Sets the values to be saved as a Glossary entry.
   * @param term Term to be defined
   * @param anchor Name to be given to the anchor tag that locates this
   * term on an HTML page- usually an abbreviated version of the term
   * @param description definition of the term
   */
  public void setTermToSave(String term, String anchor, String description){
    termToSave        = AbstractABCMDBGateway.makeSQLSafe(term);
    anchorToSave      = AbstractABCMDBGateway.makeSQLSafe(anchor);
    descriptionToSave = AbstractABCMDBGateway.makeSQLSafe(description);
  }  
  
  
  /* Saving */
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return ((termToSave        != null) && 
            (anchorToSave      != null) &&
            (descriptionToSave != null));
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){
    return (instanceToSaveIsValid() ? DB.getTermsTable().getDatum(termToSave, anchorToSave, descriptionToSave) : null);
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    termToSave        = null;
    anchorToSave      = null;
    descriptionToSave = null;
  } 

  
  /* Retrieving specific elements */
  
  /**
   * Gets the definition of a term in the glossary given its primary key
   * in the HWM Database.
   * @param PK primary key of a term being retrieved
   * @return the definition of the term
   */
  public String getDefinition(long PK){        return getDescription(PK);     }
  
  /**
   * Gets the definition of a term given the term in String form.
   * @param term the term to be 'looked up'
   * @return the definition of the term
   */
  public String getDefinition(String term){    return getDescription(term);   }

  
  /* TeX output */
  
  /**
   * Gets the glossary of all terms defined in the HWM Database
   * formatted as a TEX document, for inclusion in the manual
   * @return the TEX formatted glossary
   */
  public String getTEXGlossary(){
    String  termField          = DB.getTermsTable().getField_Term().getName();
    String  descriptionField   = DB.getTermsTable().getField_Description().getName();

    String retval = "";
    try{
      
      // Build SQL statement
      String sql = DB.getTermsTable().getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.orderBy(termField)); 

      // Get result set
      Connection con = DB.getGuestConnection(); 
      Statement st = con.createStatement();

      ResultSet rs = st.executeQuery(sql); 
      retval += "\\begin{description}\n";
      while(rs.next()){
        retval += "\\item[" + rs.getString(termField) + "] " + rs.getString(descriptionField) + "\n";
      }
      retval += "\\end{description}\n";
      
      rs.close();
      st.close();
      con.close();
    } 
    catch(Exception e){
      addToErrorReport("Error in GlossaryManager getTEXGlossary(): " + e.getMessage());
    }    
    return retval;
  }
  
}