package abcm.db.managers;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import abcm.db.ABCMXMLTransformer;
import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMHistoriesAndResearchersView;
import abcm.db.components.views.ABCMHistoriesNarrativesAndInstructionsView;
import abcm.db.components.views.ABCMHistoriesNarrativesAndResearchersView;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.math.SimDate;
import abcm.sim.History;
import abcm.utilities.XMLAssist;

/**
 * Manages the entry and retrieval of Histories in the
 * HWM database.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class HistoryManager extends AbstractAuditableDBManager{
  
  /** Title of history to be saved  */
  private String titleToSave;
  /** Description to be saved  */
  private String descToSave;
  /** User ID of owner of the history to be saved  */
  private long userToSave;
  /** Truth value to be saved  */
  private long truthValueToSave;
  /** History Narrative Manager */
  protected InnerHistoryNarrativeManager HNM;
  
  
  /**
   * Inner class to manage the History/Narrative link
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class InnerHistoryNarrativeManager extends AbstractDBLinkTableInnerManager {

    /** FK of History to be saved  */
    private long   historyFKToSave;
    /** FK of Narrative to be saved  */
    private long   narrativeFKToSave;
    /** Base Date to be saved  */
    private double baseDateToSave;
    /** Narrative order to be saved  */
    private int    orderToSave;
    
    /**
     * Constructor
     * @param db         Instance of an ABCM Gateway to provide database access.
     * @param container  pointer to the class within which this one is nested
     */
    public InnerHistoryNarrativeManager(AbstractABCMDBGateway db, HistoryManager container){ super(db, container); }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
     */
    protected ABCMdbTable baseTable(){                     return DB.getHistoryNarrativesTable();                                          }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
     */
    protected ABCMdbField baseTablePKField(){              return DB.getHistoryNarrativesTable().getField_PK();                            }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
     */
    protected ABCMdbField baseTablePrimaryFKField(){       return DB.getHistoryNarrativesTable().getField_HistoryFK();                     }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
     */
    protected ABCMdbField baseTableSecondaryFKField(){     return DB.getHistoryNarrativesTable().getField_NarrativeFK();                   }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
     */
    protected ABCMdbAbstractViewable recordsView(){        return DB.getHistoriesNarrativesAndResearchersView();                           }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
     */
    protected ABCMdbField recordsViewPKField(){            return null;  }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
     */
    protected ABCMdbField recordsViewPrimaryFKField(){     return DB.getHistoriesNarrativesAndResearchersView().getField_History_PK();     }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
     */
    protected ABCMdbField recordsViewPrimaryNameField(){   return DB.getHistoriesNarrativesAndResearchersView().getField_History_Name();   }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
     */
    protected ABCMdbField recordsViewSecondaryNameField(){ return DB.getHistoriesNarrativesAndResearchersView().getField_Narrative_Name(); }
   
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs,
        boolean includeComments) throws Exception {
       ABCMHistoriesNarrativesAndResearchersView view = DB.getHistoriesNarrativesAndResearchersView();
       
       int narrativeIDFieldCol        = rs.findColumn(view.getField_Narrative_PK().getName());
       int narrativeNameFieldCol      = rs.findColumn(view.getField_Narrative_Name().getName());
       int narrativeOrderFieldCol     = rs.findColumn(view.getField_Narrative_Order().getName());
       int narrativeBaseDateFieldCol  = rs.findColumn(view.getField_Narrative_BaseDate().getName());
       int narrativeCreatedByFieldCol = rs.findColumn(view.getField_Researcher_Name().getName());
       int startBaseDateFieldCol      = rs.findColumn(view.getField_Narrative_StartBaseDate().getName());
       
       while(rs.next()){
        xml.open("narrative");
          xml.addLine("ID",            rs.getLong(narrativeIDFieldCol));
          xml.addLine("narrativeName", rs.getString(narrativeNameFieldCol));
          xml.addLine("order",         rs.getLong(narrativeOrderFieldCol));
          xml.addLine("baseDate",      SimDate.convertToString(rs.getDouble(narrativeBaseDateFieldCol)));
          xml.addLine("createdBy",     rs.getString(narrativeCreatedByFieldCol));
          xml.addLine("startBaseDate", SimDate.convertToString(rs.getDouble(startBaseDateFieldCol)));
          
        xml.closeTag();
      }
    }

    /** 
     * Adds a set of records to an XML document. Overrides the superclass because the
     * order is determined by more than one field.
     * @param xml XMLAssist object creating the XML document
     * @param includeComments true if comments are to be included
     * @param orderByField field by which these records are to be ordered (ignored)  
     * @param FK1 foreign key 1
     */
    protected void addSetToXML(XMLAssist xml, boolean includeComments, ABCMdbField orderByField, long FK1){
      String sql = recordsView().getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(recordsViewPrimaryFKField().getName(), FK1) + 
            " ORDER BY " + DB.getHistoriesNarrativesAndResearchersView().getField_Narrative_StartBaseDate().getName() + ", " + 
                           DB.getHistoriesNarrativesAndResearchersView().getField_Narrative_Order().getName());
      System.out.println(sql);
      addSetToXML(xml, includeComments, sql);
    }

    /**
     * Sets the values that are to be saved.
     * @param historyFK
     * @param narrativeFK
     * @param baseDate
     * @param order
     */
    public void setInstanceToSave(long historyFK, long narrativeFK, double baseDate, int order){
      historyFKToSave   = historyFK;
      narrativeFKToSave = narrativeFK;
      baseDateToSave    = baseDate;
      orderToSave       = order;
    }
    
    /**
     * Determines if the instance to be saved is valid
     * @return true if the instance to be saved has been set, false otherwise
     */
    public boolean instanceToSaveIsValid(){
      return((historyFKToSave   != -1) && 
             (narrativeFKToSave != -1) &&
             (baseDateToSave    != -1) &&
             (orderToSave       >  0));
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
     */
    protected void clearInstanceToSave() {
      historyFKToSave   = -1;
      narrativeFKToSave = -1;
      baseDateToSave    = -1;
      orderToSave       = 0;
    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
     */
    protected ABCMdbDatum datumToSave() {
      return (instanceToSaveIsValid() ? 
            DB.getHistoryNarrativesTable().getDatum(historyFKToSave, narrativeFKToSave, baseDateToSave, orderToSave) : 
            null);
    }

    // This is an ordered manager:
    
    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#orderField()
     */
    protected ABCMdbField orderField(){                    return DB.getHistoryNarrativesTable().getField_Order();                          }
    
    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewOrderField()
     */
    protected ABCMdbField recordsViewOrderField(){         return DB.getHistoriesNarrativesAndResearchersView().getField_Narrative_Order(); }
    
    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#getOrderedDominantPKToSave()
     */
    protected long getOrderedDominantPKToSave(){           return historyFKToSave;                                                          }
    
    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#getOrderedNewOrderToSave()
     */
    protected long getOrderedNewOrderToSave(){             return orderToSave;                                                              }

    
  }

  
  
  
  
  
  
  
  
  
  
  
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public HistoryManager(AbstractABCMDBGateway db){
    super(db);
    HNM = new InnerHistoryNarrativeManager(db, this);
  }

  
  
  
  

   /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){               return DB.getHistoriesTable();                                              }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){        return DB.getHistoriesTable().getField_PK();                                }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView() { return DB.getHistoriesAndResearchersView();                                 }  
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField() {     return DB.getHistoriesAndResearchersView().getField_History_PK();           }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField() {     return DB.getHistoriesAndResearchersView().getField_History_Title();        }

  
  /* XML Generation */
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName() { return "histories"; }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
    ABCMHistoriesAndResearchersView view = DB.getHistoriesAndResearchersView();
    int IDFieldCol              = rs.findColumn(view.getField_History_PK().getName()); 
    int titleFieldCol           = rs.findColumn(view.getField_History_Title().getName());
    int descFieldCol            = rs.findColumn(view.getField_History_Description().getName());
    int createdByFieldCol       = rs.findColumn(view.getField_Researcher_Name().getName());
    int canBeEditedFieldCol     = rs.findColumn(view.getField_CanBeEdited().getName());
    int truthValueFieldCol      = rs.findColumn(view.getField_TruthValue_Value().getName());
        
    
    while(rs.next()){
      xml.open("history");
        xml.addLine("ID",          rs.getLong(IDFieldCol));
        xml.addLine("title",       rs.getString(titleFieldCol));
        xml.addLine("desc",        rs.getString(descFieldCol));
        xml.addLine("createdBy",   rs.getString(createdByFieldCol));
        xml.addLine("canBeEdited", ((rs.getString(canBeEditedFieldCol) != null) ? "Yes" : "No"));
        xml.addLine("truthValue",  rs.getString(truthValueFieldCol));
        
        if(singleInstance){
          xml.open("narratives");
            HNM.addSetToXML(xml, includeComments, instancePK);
          xml.closeTag();
          
          if(includeComments){
            CommentManager cm = new CommentManager(DB);
            xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.HISTORY, instancePK), XMLAssist.INDENT_BLOCK);
          }   
        }
      xml.closeTag();
    }
  }
  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlHistoryListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlHistoryToHTML.xsl"; }
    
  
  
  /**
   * Sets the values that will be used to save a new history to the ABCM database
   * @param historyName name to be saved
   * @param historyDesc description to be saved
   * @param userID user ID to be saved
   * @param truthValue truth value to be saved
   * @return true if successful
   */
  public boolean setInstanceToSave(String historyName, String historyDesc, long userID, long truthValue){
    titleToSave      = AbstractABCMDBGateway.makeSQLSafe(historyName);
    descToSave       = AbstractABCMDBGateway.makeSQLSafe(historyDesc);
    userToSave       = userID;
    truthValueToSave = truthValue;
    return false;
  }
  
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave() {
    return DB.getHistoriesTable().getDatum(titleToSave, descToSave, truthValueToSave, userToSave);
  }  
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    titleToSave      = null;
    descToSave       = null;
    userToSave       = -1;
    truthValueToSave = -1;
  }
  

  /* Auditable */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField             baseTableOwnerField(){    return DB.getHistoriesTable().getField_ResearcherFK();          }
  

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable  canBeEditedView(){        return DB.getCanBeEdited_HistoriesView();                       }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField             canBeEditedViewPKField(){ return DB.getCanBeEdited_HistoriesView().getField_History_PK(); }
  
  
  /* Retrieve object from Database */
  /**
   * Gets a History object by retrieving the records from an ABCM db
   * @param PK primary key of History record to be retrieved
   * @return a new ABCM History object created from the appropriate record.
   * Returns null if no matching record is found.
   */
  public History getHistoryFromDB(long PK){
    String historyName = getName(PK);
    if(historyName == null) return null;
    History H = new History(historyName);
    H.setInstructionsFromXML(getInstructionsListXML(PK));
    H.setDate(new SimDate(H.getFirstDate()));
    return H;
  }
  

  
  /* Special */
  
  /**
   * Inserts a Narrative into this History. Adjusts the order numbers so that
   * there are no missing numbers nor duplicates (will move equal and higher numbers
   * upward; will move this number downward if too high).
   * Confirms that the History can be edited before making this change.
   * @param historyPK primary key of the History to which this Narrative is to be
   * added
   * @param narrativePK primary key of the Narrative being added
   * @param baseDate Base Date of this Narrative as implemented in this History
   * @param order order of this Narrative within the Narratives in this History
   * @return true if successful, false otherwise.
   */
  public boolean insertNarrative(long historyPK, long narrativePK, double baseDate, int order){
    if(!canBeEdited(historyPK)){
      addToErrorReport("Could not insert Narrative; History cannot be edited.");
      return false;
    }
    // Pass through to the inner manager, which handles all the issues related to narrative order
    HNM.setInstanceToSave(historyPK, narrativePK, baseDate, order);
    return HNM.saveInstanceToDB();
    
//    if(!incrementNarrativeOrders(historyPK, order)) return false;
//    try{
//      String dbOwner = DB.getOwner();
//      ABCMHistoryNarrativesTable table = DB.getHistoryNarrativesTable();
//      Hashtable toInsert = new Hashtable();
//      toInsert.put(table.getField_HistoryFK(), historyPK + "");
//      toInsert.put(table.getField_NarrativeFK(), narrativePK + "");
//      toInsert.put(table.getField_BaseDate(), baseDate + "");
//      toInsert.put(table.getField_Order(), order + "");
////        String sqlInsert = "INSERT INTO jtmurphy.tblHistoryNarratives (intHistoryID, intNarrativeID, dblStartBaseDate, intOrder) " +
////          "VALUES (" + historyPK + "," + narrativePK + "," + baseDate + "," + order + ");";
//      String sqlInsert = table.getInsertSQL(toInsert, dbOwner);
//      DB.insertAndReturnPK(sqlInsert);
//      collapseNarrativeOrders(historyPK);      
//    }
//    catch(Exception e){
//      addToErrorReport("Could not perform HistoryManager.insertNarrative(): " + e.getMessage() + " " + e.getLocalizedMessage() + " ");
//      return false;
//    }
//    return true;
  }


  
  
  /**
   * Removes a Narrative from a specified History. The Narrative is specified
   * by its order number; remaining Narrative order numbers are adjusted
   * downward, if necessary.
   * @param historyPK primary key of the History being adjusted
   * @param narrativeOrder order number of the Narrative being removed
   * @return true if successful (even if no Narrative with the specified
   * order number was removed.
   */
  public boolean removeNarrative(long historyPK, long narrativeOrder){
    if(!canBeEdited(historyPK)){
      addToErrorReport("Could not remove Narrative; History cannot be edited.");
      return false;
    }
    
    long recordPK = HNM.getPKFromDominantPKAndOrder(historyPK, narrativeOrder);
    if(recordPK == -1){
      addToErrorReport("Could not find matching record");
      return true; // Yes, true- see comments; if no matching record is found, it is as if it were removed
    }
    return HNM.deleteItem(recordPK);
    
//    try{
//      String dbOwner = DB.getOwner();
//      ABCMHistoryNarrativesTable table = DB.getHistoryNarrativesTable();
//      String historyNarrativeTableName = table.getFullName(dbOwner);
//      String historyIDFieldName = table.getField_HistoryFK().getFullFieldName(dbOwner);
//      String orderFieldName = table.getField_Order().getFullFieldName(dbOwner);
//      
//      String sql = "DELETE " +
//                   "FROM " + historyNarrativeTableName + " " +
//                   "WHERE " + historyIDFieldName + " = " + historyPK + " " +
//                   "AND " + orderFieldName + " = " + narrativeOrder + ";";
//      
//      Connection con = DB.getAdminConnection();
//      Statement st = con.createStatement();
//      
//      st.executeUpdate(sql);
//      
//      st.close();
//      con.close();
//      
//      collapseNarrativeOrders(historyPK);
//    }
//    catch(Exception e){
//      addToErrorReport("Could not remove narrative: " + e.getMessage());
//      return false;
//    }
//    return true;
  }

  
  /**
   * Gets an XML formatted list of the instructions for a particular
   * History from the HWM database.
   * The list of instructions is the list of all instructions for the
   * Narratives in this history, adjusted for Base Date and sorted
   * for order.
   * @param PK Primary Key of the History for which the instructions are to be
   * returned.
   * @return String containing XML format of that list of instructions.
   */
  public String getInstructionsListXML(long PK){
    XMLAssist xml = new XMLAssist();
    
    try{
      
      ABCMHistoriesNarrativesAndInstructionsView view = DB.getHistoriesNarrativesAndInstructionsView();
      
      String sql =
          view.getSelectSQL(DB.getOwner(), 
              AbstractDBGateway.where(view.getField_History_PK().getName(), PK) + " " + 
              "ORDER BY " + view.getField_CalculatedInstructionDate().getName() + ", " + 
                            view.getField_NarrativeOrder().getName() + ", " + 
                            view.getField_InstructionOrder().getName());
      
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();

      ResultSet rs = st.executeQuery(sql);
      
      int commandPKCol          = rs.findColumn(view.getField_Command_PK().getName());
      int narrativeNameCol      = rs.findColumn(view.getField_Narrative_Name().getName());
      int narrativeOrderCol     = rs.findColumn(view.getField_NarrativeOrder().getName());
      int instructionOrderCol   = rs.findColumn(view.getField_InstructionOrder().getName());
      int dateCol               = rs.findColumn(view.getField_CalculatedInstructionDate().getName());
      int commandCol            = rs.findColumn(view.getField_Command().getName());
      int[] argCol = new int[11];
      argCol[1]                 = rs.findColumn(view.getField_Arg1().getName());
      argCol[2]                 = rs.findColumn(view.getField_Arg2().getName());
      argCol[3]                 = rs.findColumn(view.getField_Arg3().getName());
      argCol[4]                 = rs.findColumn(view.getField_Arg4().getName());
      argCol[5]                 = rs.findColumn(view.getField_Arg5().getName());
      argCol[6]                 = rs.findColumn(view.getField_Arg6().getName());
      argCol[7]                 = rs.findColumn(view.getField_Arg7().getName());
      argCol[8]                 = rs.findColumn(view.getField_Arg8().getName());
      argCol[9]                 = rs.findColumn(view.getField_Arg9().getName());
      argCol[10]                = rs.findColumn(view.getField_Arg10().getName());
            
      xml.open("historyInstructionSet");
      
      long overallOrder = 0;
      while(rs.next()){
        xml.open("instructionset","overallOrder=\"" + (overallOrder++) + "\"");
        xml.addLine("commandPK", rs.getString(commandPKCol));
        xml.addLine("narrativeName", rs.getString(narrativeNameCol));
        xml.addLine("narrativeOrder", rs.getLong(narrativeOrderCol));
        xml.addLine("instructionOrder", rs.getLong(instructionOrderCol));
        xml.addLine("date", "dateval=\"" + rs.getDouble(dateCol) + "\"", SimDate.convertToString(rs.getDouble(dateCol)));
        xml.addLine("command", rs.getString(commandCol));
        int numOfArgs = rs.getInt("intNumberOfArgs"); 
        if (numOfArgs > 0) {
          xml.open("arguments");
          for(int i = 1; i <= numOfArgs; i++){
            if((rs.getString(argCol[i]) != null) && (rs.getString(argCol[i]).length() > 0)) 
              xml.addLine("arg", "number=\"" + i + "\"", rs.getString(argCol[i]));
          }
          xml.closeTag();
        }
        xml.closeTag();
      }
      xml.closeTag();
      rs.close();
      st.close();
      con.close();
    }
    catch (Exception e){
      errorReport = "Error in History Manager getInstructionsHTML(): " + e.getMessage();
    }
    return xml.getXML();
  }

  /**
   * Gets an HTML version of the list of instructions for a specific
   * History in the HWM database.
   * The list of instructions for a History is the list of instructions
   * for all Narratives in this history, sorted for Base Date and order.
   * @param PK the primary key of the History for which the instructions
   * are requested
   * @param includeHeaders True if the HTML headers should be returned,
   * False if not.
   * @param xslRootDir directory in which xsl stylesheets are found
   * @return A String containing the HTML formatted list of instructions.
   */
  public String getInstructionsHTML(long PK, boolean includeHeaders, String xslRootDir) {
    return ABCMXMLTransformer.getHTML(getInstructionsListXML(PK), 
        xslRootDir + "xmlHistoryInstructionsToHTML.xsl", includeHeaders);
  } 



  /**
   * Deleting a deletable history should remove all the entries in the
   * history narrative manager.
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#deleteItem(long, long)
   */
  public boolean deleteItem(long pk, long user){
    return HNM.deleteElementsByPrimaryFK(pk) && super.deleteItem(pk, user);
  }
  

///**
//* Increments the order numbers of all Narratives of a given History
//* that have order numbers equal to or higher than the given number. 
//* @param historyPK primary key of the History for which the Narratives
//* are being adjusted
//* @param newNarrativeOrder number that marks the new order being inserted
//* and above which (or equal to) other order numbers must be incremented
//* @return true if successful (even if no changes were made), false otherwise.
//*/
//public boolean incrementNarrativeOrders(long historyPK, long newNarrativeOrder){
// try{
//   String dbOwner = DB.getOwner();
//   ABCMHistoryNarrativesTable table = DB.getHistoryNarrativesTable();
//   String historyNarrativeTableName = table.getFullName(dbOwner);
//   String orderFieldName = table.getField_Order().getFullFieldName(dbOwner);
//   String historyIDFieldName = table.getField_HistoryFK().getFullFieldName(dbOwner);
//   String sqlIncrement = 
//     "UPDATE " + historyNarrativeTableName + " " +
//     "SET " + orderFieldName + " = " + orderFieldName + " + 1 " + 
//     "WHERE " + historyIDFieldName + " = " + historyPK + " " + 
//     "AND " + orderFieldName + " >= " + newNarrativeOrder + ";";
//   
//   
////   String sqlIncrement =
////     "UPDATE jtmurphy.tblHistoryNarratives " +
////     "SET intOrder = intOrder + 1 " +
////     "WHERE intHistoryID = " + historyPK + " " +
////     "  AND intOrder >= " + newNarrativeOrder +
////     ";"; 
// 
//   Connection con = DB.getAdminConnection();
//   Statement st = con.createStatement();
//   
//   st.executeUpdate(sqlIncrement);
//   
//   st.close();
//   con.close();
// }
// catch(Exception e){
//   addToErrorReport("Could not increment narrative orders: + " + e.getMessage());
//   return false;
// }
// return true;
//}

///**
//* Collapses the order numbers of all Narratives to eliminate
//* any missing order numbers; numbers should be sequential
//* without gaps.
//* @param historyPK primary key of the History being adjusted
//* @return true if successful (even if no changes were required),
//* false otherwise
//*/
//public boolean collapseNarrativeOrders(long historyPK){
// boolean successFlag = false;
// try {
//   String dbOwner = DB.getOwner();
//   ABCMHistoryNarrativesTable table = DB.getHistoryNarrativesTable();
//   
//   String historyNarrativeTableName = table.getFullName(dbOwner);
//   
//   String historyIDFieldName          = table.getField_HistoryFK().getFullFieldName(dbOwner);
//   String historyNarrativeIDFieldName = table.getField_PK().getFullFieldName(dbOwner);
//   String narrativeIDFieldName        = table.getField_NarrativeFK().getFullFieldName(dbOwner);
//   String orderFieldName              = table.getField_Order().getFullFieldName(dbOwner);
//   String shortOrderFieldName         = table.getField_Order().getName();
//   
//   String sqlGetList =
//       "SELECT " + historyNarrativeIDFieldName + ", " + narrativeIDFieldName + ", " + orderFieldName + " " +
//       "FROM " + historyNarrativeTableName + " " +
//       "WHERE " + historyIDFieldName + " = " + historyPK + " " +
//       "ORDER BY " + orderFieldName +
//       ";"; 
//   
//   Connection con = DB.getGuestConnection();
//   Statement stGetList = con.createStatement();
//   Statement stUpdate = con.createStatement();
//   
//   ResultSet rs = stGetList.executeQuery(sqlGetList);
//   
//   long instrCount = 1;
//   while(rs.next()){
//     if(rs.getInt(shortOrderFieldName)!= instrCount){
//       String sqlUpdate = 
//           "UPDATE " + historyNarrativeTableName + " " +
//           "SET " + shortOrderFieldName + " =" + instrCount + " " +
//           "WHERE " + historyNarrativeIDFieldName + " = " + rs.getInt(historyNarrativeIDFieldName) +
//           ";";
//       stUpdate.executeUpdate(sqlUpdate);
//     }
//     instrCount++;
//   }
//   
//   successFlag = true;
// }
// catch(Exception e){
//   addToErrorReport("Could not collapse Narrative Orders: " + e.getMessage());
// }
// return successFlag;
//}
  
}
