package abcm.db.managers;


import java.sql.ResultSet;
import java.text.SimpleDateFormat;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.views.ABCMJournalEntriesAndResearchersView;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.utilities.XMLAssist;

/**
 * Manages the entry and retrieval of Journal Entries in the
 * HWM database.
 * 
 * Note that Journal Entries are never deleted
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class JournalEntryManager extends AbstractDBManager {

  /** Researcher ID to be saved  */
  private long researcherID = -1;
  /** Text of entry to be saved  */
  private String entryText ="";

  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public JournalEntryManager(AbstractABCMDBGateway db){
    super(db);
  }

  /* XML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView(){   return DB.getJournalEntriesAndResearchersView();                                  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField(){       return DB.getJournalEntriesAndResearchersView().getField_JournalEntry_PK();       }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){ return "JournalEntries"; }
    
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMJournalEntriesAndResearchersView listView = (ABCMJournalEntriesAndResearchersView)recordsView();
    
    int IDFieldCol            = rs.findColumn(listView.getField_JournalEntry_PK().getName());
    int researcherFieldCol    = rs.findColumn(listView.getField_Researcher_Name().getName());
    int dateFieldCol          = rs.findColumn(listView.getField_JournalEntry_DateMade().getName());
    int entryFieldCol         = rs.findColumn(listView.getField_JournalEntry_Entry().getName());
    
    while(rs.next()){
      xml.open("JournalEntry");
        xml.addLine("ID", rs.getInt(listView.getField_JournalEntry_PK().getName()));
        xml.addLine("Researcher", rs.getString(listView.getField_Researcher_Name().getName()));
        xml.addLine("DateOfEntry", (new SimpleDateFormat()).format(rs.getTimestamp(listView.getField_JournalEntry_DateMade().getName())));
        xml.addLine("Entry", rs.getString(listView.getField_JournalEntry_Entry().getName()));
      xml.closeTag();
    }
  }

  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {       return "xmlJournalEntryListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){    return "xmlJournalEntryToHTML.xsl";     }

  
  /* Saving */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    researcherID = -1;
    entryText = null;
  }
  
  /**
   * Set the values to be saved as a new Journal Entry
   * @param ID User ID of the author of this entry
   * @param entry text of the entry
   */ 
  public void setInstanceToSave(long ID, String entry){
    researcherID = ID;
    entryText = AbstractABCMDBGateway.makeSQLSafe(entry);
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return ((researcherID != -1  ) && 
            (entryText    != null));
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){
    return (instanceToSaveIsValid() ? DB.getJournalEntriesTable().getDatum("" + researcherID, DB.currentTimeString(), entryText) : null);
  }  
  
}
