package abcm.db.managers;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.Statement;
import java.util.Vector;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMNarrativeInstructionsTable;
import abcm.db.components.tables.ABCMNarrativesTable;
import abcm.db.components.views.ABCMNarrativesAndInstructionsView;
import abcm.db.components.views.ABCMNarrativesAndResearchersView;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.math.SimDate;
import abcm.utilities.XMLAssist;

/**
 * Manages the entry and retrieval of Narratives in the
 * HWM database. Also includes supporting functions
 * that allow modifications and deletions of Narrative
 * instructions, as well as retrieval of valid instruction
 * commands.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class NarrativeManager extends AbstractAuditableDBManager {
  
  /**
   * Database manager for Narrative Instructions.
   * 
   * @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class NarrativeInstructionManager extends AbstractDBLinkTableInnerManager {

    /** PK of Narrative to save  */
    private long     narrPKToSave;
    /** Date of Narrative Instruction to save  */
    private double   narrInstrDateToSave;
    /** Instruction order to save  */
    private int      instructionOrderToSave;
    /** Instruction command to save */
    private long     commandToSave;
    /** Command arguments to save  */
    private String[] argsToSave;
    /** Comment to save  */
    private String   commentToSave;
    
    /* Constructor */

    /**
     * Constructor
     * @param db Instance of an ABCM Gateway to provide database access.
     * @param containedBy the enclosing class
     */
    public NarrativeInstructionManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }   
  
    /* Base Tables */
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
     */
    protected ABCMdbTable baseTable(){                     return DB.getNarrativeInstructionsTable();                                          }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
     */
    protected ABCMdbField baseTablePKField(){              return DB.getNarrativeInstructionsTable().getField_PK();                            }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
     */
    protected ABCMdbField baseTablePrimaryFKField(){       return DB.getNarrativeInstructionsTable().getField_NarrativeFK();                   }
     
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
     */
    protected ABCMdbField baseTableSecondaryFKField(){     return DB.getNarrativeInstructionsTable().getField_NarrativeCommandFK();            }
    
    
    /* XML Creation*/
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
     */
    protected ABCMdbAbstractViewable recordsView() {       return DB.getNarrativesAndInstructionsView();                                       }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
     */
    protected ABCMdbField recordsViewPKField() {           return DB.getNarrativesAndInstructionsView().getField_NarrativeInstruction_PK();    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
     */
    protected ABCMdbField recordsViewPrimaryFKField(){     return DB.getNarrativesAndInstructionsView().getField_Narrative_PK();               }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
     */
    protected ABCMdbField recordsViewPrimaryNameField(){   return DB.getNarrativesAndInstructionsView().getField_Narrative_Name();             }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
     */
    protected ABCMdbField recordsViewSecondaryNameField(){ return DB.getNarrativesAndInstructionsView().getField_Command();                    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs,
        boolean includeComments) throws Exception {
      ABCMNarrativesAndInstructionsView view = DB.getNarrativesAndInstructionsView();
      
      int relDateCol = rs.findColumn(view.getField_Date().getName());
      int orderCol   = rs.findColumn(view.getField_Order().getName());
      int instructionCol = rs.findColumn(view.getField_Command().getName());
      
      int[] ArgsCols = new int[11];
      ArgsCols[1]  = rs.findColumn(view.getField_Arg1().getName());
      ArgsCols[2]  = rs.findColumn(view.getField_Arg2().getName());
      ArgsCols[3]  = rs.findColumn(view.getField_Arg3().getName());
      ArgsCols[4]  = rs.findColumn(view.getField_Arg4().getName());
      ArgsCols[5]  = rs.findColumn(view.getField_Arg5().getName());
      ArgsCols[6]  = rs.findColumn(view.getField_Arg6().getName());
      ArgsCols[7]  = rs.findColumn(view.getField_Arg7().getName());
      ArgsCols[8]  = rs.findColumn(view.getField_Arg8().getName());
      ArgsCols[9]  = rs.findColumn(view.getField_Arg9().getName());
      ArgsCols[10] = rs.findColumn(view.getField_Arg10().getName());
      int commentsCol  = rs.findColumn(view.getField_Comment().getName());
      int numOfArgsCol = rs.findColumn(view.getField_NumberOfArguments().getName());
  
      while(rs.next()){
        xml.open("instruction");
        xml.addLine("instructiondblDate", rs.getDouble(relDateCol));
        xml.addLine("instructionDate",    SimDate.convertToString(rs.getDouble(relDateCol)));
        xml.addLine("instructionOrder",   rs.getInt(orderCol));
        xml.addLine("instructionCommand", rs.getString(instructionCol));
        int numOfArgs = rs.getInt(numOfArgsCol); 
        if (numOfArgs > 0) {
          xml.open("arguments");
          for(int i = 1; i <= numOfArgs; i++){
            xml.addLine("<arg number=\"" + i + "\">" + rs.getString(ArgsCols[i]) + "</arg>\n");
          }
          xml.closeTag();
        }
        if(rs.getString(commentsCol) != null){
          xml.addLine("comment", rs.getString(commentsCol));
        }
        xml.closeTag();         
      }
    }
    
    // This is overloaded because the order is dependent on two fields
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#addSetToXML(abcm.utilities.XMLAssist, boolean, long) 
     */
    public void addSetToXML(XMLAssist xml, boolean includeComments, long FK1){
      String sql = recordsView().getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(DB.getNarrativesAndInstructionsView().getField_Narrative_PK().getName(), FK1) +
            " ORDER BY " + DB.getNarrativesAndInstructionsView().getField_Date().getName() + ", " + 
                           DB.getNarrativesAndInstructionsView().getField_Order().getName());
      try{
        Connection con = DB.getAdminConnection();
        Statement st = con.createStatement();
        ResultSet rs = st.executeQuery(sql);
        addInstancesToXML(xml, rs, includeComments);
        st.close();
        con.close();
      }
      catch(Exception e){
        addToErrorReport("Could not add narrative instructions to XML: " + e.getMessage());
      }   
    }

      
    /* Basic Lookup */
    
    /**
     * Gets the PK of a narrative instruction
     * @param narrativeFK PK of the narrative
     * @param date date of the instruction
     * @param order order of the instruction
     * @return the requested primary key
     */
    public long getPK(long narrativeFK, double date, int order){
      ABCMNarrativeInstructionsTable table = DB.getNarrativeInstructionsTable();
      String sql = "SELECT " + table.getField_PK().getName() + " " +
                   "FROM " + table.getFullName(DB.getOwner()) + " " + 
                   "WHERE " + table.getField_NarrativeFK().getName() + " = " + narrativeFK + " AND " + 
                              table.getField_InstructionDate().getName() + " = " + date + " AND " + 
                              table.getField_InstructionOrder().getName() + " = " + order;
      long retVal = DB.getNumericItem(sql, table.getField_PK().getName());
      if(retVal == -1)  addToErrorReport(DB.getErrorReport());
      return retVal;
    }    
    
    
    /* Saving */
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
     */
    protected void clearInstanceToSave() {
      narrPKToSave           = -1;
      narrInstrDateToSave    = -1;
      instructionOrderToSave = -1;
      commandToSave          = -1;
    }
  
    /**
     * Sets the instance of a narrative instruction to save
     * @param narrPK PK of the narrative
     * @param narrInstrDate date of the instruction
     * @param instrOrder order of the instruction
     * @param command command to be executed
     * @param Args arguments to the command
     * @param comment comment
     */
    public void setInstanceToSave(long narrPK, double narrInstrDate, int instrOrder, 
        long command, String[] Args, String comment){
      narrPKToSave = narrPK;
      narrInstrDateToSave = narrInstrDate;
      instructionOrderToSave = instrOrder;
      commandToSave = command;
      argsToSave = Args;
      commentToSave = comment;
    }
        
    /**
     * Determines the status of the instance to save
     * @return true if the instance to save has been set
     */
    public boolean instanceToSaveIsValid(){
      return ((narrPKToSave           != -1) &&
              (narrInstrDateToSave    != -1) &&
              (instructionOrderToSave != -1) &&
              (commandToSave          != -1));             
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
     */
    protected ABCMdbDatum datumToSave() {
      return (instanceToSaveIsValid() ? 
          DB.getNarrativeInstructionsTable().getDatum("" + narrPKToSave, "" + narrInstrDateToSave, "" + instructionOrderToSave, "" + commandToSave, 
              ((argsToSave.length > 0) ? argsToSave[0] : ""), 
              ((argsToSave.length > 1) ? argsToSave[1] : ""), 
              ((argsToSave.length > 2) ? argsToSave[2] : ""),
              ((argsToSave.length > 3) ? argsToSave[3] : ""),
              ((argsToSave.length > 4) ? argsToSave[4] : ""), 
              ((argsToSave.length > 5) ? argsToSave[5] : ""),
              ((argsToSave.length > 6) ? argsToSave[6] : ""),
              ((argsToSave.length > 7) ? argsToSave[7] : ""),
              ((argsToSave.length > 8) ? argsToSave[8] : ""),
              ((argsToSave.length > 9) ? argsToSave[9] : ""), 
              ((commentToSave != null) ? commentToSave : "")) : 
          null);
    }    
  
    /* Deleting */
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#deleteItem(long) 
     */
    public boolean deleteItem(long PK){      return false;     }  // This should not be used; may leave order askew
    
    /**
     * Deletes a narrative instruction. Collapses other instruction orders
     * on the same date to eliminate gaps.
     * @param narrPK the PK of the narrative
     * @param narrInstrDate the date of the instruction to be deleted
     * @param instrOrder the order of the instruction to be deleted
     * @return true if successful, false otherwise.
     */
    public boolean deleteItem(long narrPK, double narrInstrDate, int instrOrder){
      long PK = getPK(narrPK, narrInstrDate, instrOrder);
      if(!super.deleteItem(PK)) return false;
      collapseInstructionOrdersOnDate(narrPK, narrInstrDate);
      return true;
    }
    
    /* Special */
    
    /**
     * Inserts a narrative instruction. Shuffles instruction orders
     * and collapses as appropriate, if more than one instruction is
     * to be executed in a single simulation day.
     * @param narrPK primary key of the narrative
     * @param narrInstrDate date at which to insert the instruction
     * @param instrOrder order at which to insert the instruction
     * @param command command to be executed
     * @param Args arguments to the command
     * @param comment comment
     * @return true if successful, false otherwise
     */
    public boolean insertInstruction(long narrPK, double narrInstrDate, int instrOrder, 
        long command, String[] Args, String comment){
      return insertInstruction(narrPK, narrInstrDate, instrOrder, command, Args, comment, true);
    }
    
    /**
     * Inserts a narrative instruction. Optionally shuffles and collapses instruction
     * orders
     * @param narrPK primary key of the narrative
     * @param narrInstrDate date at which to insert the instruction
     * @param instrOrder order at which to insert the instruction
     * @param command command to be executed
     * @param Args arguments to the command
     * @param comment comment
     * @param performCollapse true if instruction orders should be collapsed 
     * @return true if successful, false otherwise
     */
    public boolean insertInstruction(long narrPK, double narrInstrDate, int instrOrder, 
        long command, String[] Args, String comment, boolean performCollapse){
      boolean successFlag = true;
      // set the instance to save
      setInstanceToSave(narrPK, narrInstrDate, instrOrder, command, Args, comment);
      if(performCollapse){
        // Make room in the list of instructions
        successFlag = successFlag && incrementInstructionOrdersOnDate(narrPK, narrInstrDate, instrOrder);
        successFlag = successFlag && saveInstanceToDB();
        successFlag = collapseInstructionOrdersOnDate(narrPK, narrInstrDate) && successFlag; // Run this even if others unsuccessful
        return successFlag;
      }
      else{
        return saveInstanceToDB();
      }
    }
    
    /**
     * Copies the instructions from one narrative to another
     * @param narrativeToCopy PK of narrative to be copied
     * @param newPK PK of new narrative
     * @return true if successful, false otherwise
     */
    public boolean copyInstructions(long narrativeToCopy, long newPK){
      boolean successFlag = false;
      ABCMNarrativeInstructionsTable table = DB.getNarrativeInstructionsTable();
      
      String narrativeInstructionsTableName = table.getFullName(DB.getOwner());
      
      String narrativeIDFieldName         = table.getField_NarrativeFK().getName();
      String dateFieldName                = table.getField_InstructionDate().getName();
      String instructionOrderFieldName    = table.getField_InstructionOrder().getName();
      String commandIDFieldName           = table.getField_NarrativeCommandFK().getName();
      String Arg1FieldName                = table.getField_Arg1().getName();
      String Arg2FieldName                = table.getField_Arg2().getName();
      String Arg3FieldName                = table.getField_Arg3().getName();
      String Arg4FieldName                = table.getField_Arg4().getName();
      String Arg5FieldName                = table.getField_Arg5().getName();
      String Arg6FieldName                = table.getField_Arg6().getName();
      String Arg7FieldName                = table.getField_Arg7().getName();
      String Arg8FieldName                = table.getField_Arg8().getName();
      String Arg9FieldName                = table.getField_Arg9().getName();
      String Arg10FieldName               = table.getField_Arg10().getName();
      String commentFieldName             = table.getField_Comment().getName();
          
      String sqlCopyInstructions = 
            "INSERT INTO " + narrativeInstructionsTableName + " " +
                "( " +
                narrativeIDFieldName + ", " +
                dateFieldName + ", " +
                instructionOrderFieldName + ", " +
                commandIDFieldName + ", " +
                Arg1FieldName + ", " +  Arg2FieldName + ", " +  Arg3FieldName + ", " +  Arg4FieldName + ", " +
                Arg5FieldName + ", " +  Arg6FieldName + ", " +  Arg7FieldName + ", " +  Arg8FieldName + ", " +
                Arg9FieldName + ", " +  Arg10FieldName + ", " +
                commentFieldName +
                " ) " +
            "SELECT " +
                newPK + ", " +
                dateFieldName + ", " +
                instructionOrderFieldName + ", " +
                commandIDFieldName + ", " +
                Arg1FieldName + ", " +  Arg2FieldName + ", " +  Arg3FieldName + ", " +  Arg4FieldName + ", " +
                Arg5FieldName + ", " +  Arg6FieldName + ", " +  Arg7FieldName + ", " +  Arg8FieldName + ", " +
                Arg9FieldName + ", " +  Arg10FieldName + ", " +
                commentFieldName + " " +
            "FROM " + narrativeInstructionsTableName +" " +
            "WHERE " + narrativeIDFieldName + " = " + narrativeToCopy +
            ";";
      try{
        Connection con = DB.getAdminConnection();
        Statement st = con.createStatement();
        st.execute(sqlCopyInstructions);
        successFlag = true;
        st.close();
        con.close();
      }
      catch(Exception e){
        addToErrorReport("Error copying message: " + e.getMessage());
      }    
      return successFlag;
    }
    
    /**
     * Increments the instruction orders on a given date, beginning with
     * a specified lowest number. For example, if the current
     * instructions are 1, 2, 3, 4, 5, and this method is executed
     * with a lowest number of 3, the result will be 1, 2, 4, 5, 6. 
     * @param narrPK primary key of narrative to be updated
     * @param narrInstrDate date on which instructions are to be changed
     * @param lowestOrderValueToIncrement lowest number to increment
     * @return true if successful, false otherwise
     */
    public boolean incrementInstructionOrdersOnDate(long narrPK, double narrInstrDate, int lowestOrderValueToIncrement){
      boolean successFlag = false;
      ABCMNarrativeInstructionsTable table = DB.getNarrativeInstructionsTable();
      
      String narrativeInstructionsTableName = table.getFullName(DB.getOwner());
      
      String narrativeIDFieldName         = table.getField_NarrativeFK().getName();
      String dateFieldName                = table.getField_InstructionDate().getName();
      String instructionOrderFieldName    = table.getField_InstructionOrder().getName();
          
      try {
        String sqlIncrement =
            "UPDATE " + narrativeInstructionsTableName + " " +
            "SET " + instructionOrderFieldName + " = " + instructionOrderFieldName + " + 1 " +
            "WHERE " + narrativeIDFieldName + " = " + narrPK + " " +
            "  AND " + dateFieldName + " = " + narrInstrDate + " " +
            "  AND " + instructionOrderFieldName + " >= " + lowestOrderValueToIncrement +
            ";"; 
        
        Connection con = DB.getAdminConnection();
        Statement st = con.createStatement();
        
        st.executeUpdate(sqlIncrement);
        successFlag = true;
        
        st.close();
        con.close();
      }
      catch (Exception e){
        addToErrorReport("Error in NarrativeManager incrementInstructionOrdersOnDate(): " + e.getMessage());
      }    
      return successFlag;
    }
    
    /**
     * Updates a specific instruction order on a specific date
     * @param narrPK primary key of the narrative to be modified
     * @param instrDate date of the instruction to be adjusted
     * @param oldOrder original instruction order value 
     * @param newOrder new instruction order value
     * @return true if successful, false otherwise
     */
    public boolean updateInstructionOrder(long narrPK, double instrDate, int oldOrder, int newOrder){
      long PK = getPK(narrPK, instrDate, oldOrder);
      if(PK == -1) return false;
      return updateInstructionOrder(PK, newOrder);      
    }
    
    /**
     * Updates a specific instruction order
     * @param instructionPK primary key of the narrative instruction to be changed
     * @param newOrder new order value
     * @return TISFO
     */
    public boolean updateInstructionOrder(long instructionPK, int newOrder){
      boolean successFlag = DB.updateItem(DB.getNarrativeInstructionsTable(), DB.getNarrativeInstructionsTable().getField_PK(), 
          instructionPK, DB.getNarrativeInstructionsTable().getField_InstructionOrder(), newOrder);
      if(!successFlag) addToErrorReport(DB.getErrorReport());
      return successFlag;    
    }  
    
    /**
     * Collapses the instruction orders so that any gaps (i.e. 1, 2, 4, 5) will
     * be closed 
     * @param narrPK Narrative to correct
     * @param instrDate date of instructions to collapse
     * @return TIFSO
     */
    public boolean collapseInstructionOrdersOnDate(long narrPK, double instrDate){
      boolean successFlag = false;
      ABCMNarrativeInstructionsTable table = DB.getNarrativeInstructionsTable();
      
      String narrativeInstructionsTableName = table.getFullName(DB.getOwner());
      
      String narrativeInstructionIDFieldName = table.getField_PK().getName();
      String narrativeIDFieldName            = table.getField_NarrativeFK().getName();
      String dateFieldName                   = table.getField_InstructionDate().getName();
      String instructionOrderFieldName       = table.getField_InstructionOrder().getName();
      
      try {
        String sqlGetList =
            "SELECT " + narrativeInstructionIDFieldName + ", " + instructionOrderFieldName + " " +
            "FROM " + narrativeInstructionsTableName + " " +
            "WHERE " + narrativeIDFieldName + " = " + narrPK + " " +
            "AND " + dateFieldName + " = " + instrDate + " " +
            "ORDER BY " + instructionOrderFieldName + " " +
            ";"; 
        
        Connection con = DB.getGuestConnection();
        Statement stGetList = con.createStatement();
        Statement stUpdate = con.createStatement();
        
        ResultSet rs = stGetList.executeQuery(sqlGetList);
        int instrCount = 1;
        while(rs.next()){
          if(!updateInstructionOrder(rs.getLong(narrativeInstructionIDFieldName), instrCount)){
            successFlag = false;
            break;
          }        
          instrCount++;
        }      
        successFlag = true;
      }
      catch (Exception e){
        addToErrorReport("Error in NarrativeInstructionManager" + e.getMessage());
      }    
      return successFlag;
    }    

    
    public Vector<Vector<String>> getNarrativeSubset(long narrPK){
      return getNarrativeSubset(narrPK, null);
    }
    
    public Vector<Vector<String>> getNarrativeSubset(long narrPK, String command){
      
      ABCMNarrativesAndInstructionsView view = DB.getNarrativesAndInstructionsView();
      
      
      Vector<Vector<String>> ret = new Vector<Vector<String>>();
      String sql = 
        recordsView().getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(view.getField_Narrative_PK().getFullFieldName(DB.getOwner()), 
              narrPK)
            + (!(command == null) ?
                AbstractDBGateway.where(view.getField_Command().getFullFieldName(DB.getOwner()), command)
                : ""));
      Connection con;
      Statement st;
      ResultSet rst;
      
      try{
        con = DB.getAdminConnection();
        st = con.createStatement();
        rst = st.executeQuery(sql);
        
        int arg1Col  = rst.findColumn(view.getField_Arg1().getName());
        int arg2Col  = rst.findColumn(view.getField_Arg1().getName());
        int arg3Col  = rst.findColumn(view.getField_Arg1().getName());
        int arg4Col  = rst.findColumn(view.getField_Arg1().getName());
        int arg5Col  = rst.findColumn(view.getField_Arg1().getName());
        int arg6Col  = rst.findColumn(view.getField_Arg1().getName());
        int arg7Col  = rst.findColumn(view.getField_Arg1().getName());
        int arg8Col  = rst.findColumn(view.getField_Arg1().getName());
        int arg9Col  = rst.findColumn(view.getField_Arg1().getName());
        int arg10Col = rst.findColumn(view.getField_Arg1().getName());
        
        while(rst.next()){
          Vector<String> row = new Vector<String>();
          row.add(rst.getString(view.getField_Date().getName()));
          row.add(rst.getString(view.getField_Command().getName()));
          if(rst.getString(arg1Col)   != null) row.add(rst.getString(arg1Col));
          if(rst.getString(arg2Col)   != null) row.add(rst.getString(arg2Col));
          if(rst.getString(arg3Col)   != null) row.add(rst.getString(arg3Col));
          if(rst.getString(arg4Col)   != null) row.add(rst.getString(arg4Col));
          if(rst.getString(arg5Col)   != null) row.add(rst.getString(arg5Col));
          if(rst.getString(arg6Col)   != null) row.add(rst.getString(arg6Col));
          if(rst.getString(arg7Col)   != null) row.add(rst.getString(arg7Col));
          if(rst.getString(arg8Col)   != null) row.add(rst.getString(arg8Col));
          if(rst.getString(arg9Col)   != null) row.add(rst.getString(arg9Col));
          if(rst.getString(arg10Col)  != null) row.add(rst.getString(arg10Col));
          ret.add(row);
        }       
      }
      catch(Exception E){
        this.addToErrorReport(E.getMessage());
      }
      return ret;
    }
  
  } // End inner class
    
  /** Name of narrative to be saved  */
  private String  narrativeNameToSave;
  /** PK of researcher who is to be owner of narrative to be saved  */
  private long    researcherToSave;
  /** Base Date for narrative to be saved  */
  private double  baseDateToSave;
  /** Flag indicating if narrative to be saved is seasonal  */
  private boolean isSeasonalToSave;
  /** Description of narrative to be saved  */
  private String  descToSave;
  
  /** Internal Narrative Instruction Manager  */
  private NarrativeInstructionManager NIM;
  /** Internal Command Argument Manager  */
  private CommandArgumentManager CAM;
  
  /* Constructor */
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public NarrativeManager(AbstractABCMDBGateway db){   
    super(db);
    NIM = new NarrativeInstructionManager(DB, this);
    CAM = new CommandArgumentManager(DB);
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable() { return DB.getNarrativesTable(); }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField() { return DB.getNarrativesTable().getField_PK(); }

 
  /* XML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView() {   return DB.getNarrativesAndResearchersView();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField(){        return DB.getNarrativesAndResearchersView().getField_Narrative_PK();       }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName() {              return "narratives";                                                       }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField(){           return DB.getNarrativesTable().getField_Name();     }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField() 
   */
  protected ABCMdbField baseTableDescriptionField(){    return DB.getNarrativesTable().getField_BaseDate(); }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    
    ABCMNarrativesAndResearchersView view = DB.getNarrativesAndResearchersView();
      
    int IDCol             = rs.findColumn(view.getField_Narrative_PK().getName());
    int nameCol           = rs.findColumn(view.getField_Narrative_Name().getName());
    int researcherNameCol = rs.findColumn(view.getField_Researcher_Name().getName());
    int descCol           = rs.findColumn(view.getField_Narrative_Desc().getName());
    int baseDateCol       = rs.findColumn(view.getField_Narrative_BaseDate().getName());
    int canBeEditedCol    = rs.findColumn(view.getField_CanBeEdited().getName());
    int dateCreatedCol    = rs.findColumn(view.getField_Narrative_DateCreated().getName());
    int isSeasonalCol     = rs.findColumn(view.getField_Narrative_IsSeasonal().getName());
            
    while(rs.next()){
      xml.open("narrative");
        xml.addLine("ID",              rs.getString(IDCol));
        xml.addLine("name",            rs.getString(nameCol));
        xml.addLine("desc",            rs.getString(descCol));
        xml.addLine("canBeEdited",   ((rs.getString(canBeEditedCol) != null) ? "Yes" : "No"));
        xml.addLine("baseDate",        SimDate.convertToString(rs.getDouble(baseDateCol)));
        xml.addLine("researcher",      rs.getString(researcherNameCol));
        xml.addLine("dateCreated",     rs.getString(dateCreatedCol));
        xml.addLine("isSeasonal",     (rs.getBoolean(isSeasonalCol) ? "Yes" : "No"));
  
        if(singleInstance){ // If this is a single instance...
          xml.open("instructions");
            NIM.addSetToXML(xml, includeComments, instancePK);
          xml.closeTag();
        }       
      xml.closeTag(); // Close 'narrative
    }
  }
  
  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlNarrativeListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlNarrativeToHTML.xsl"; }
  
    
  /**
   * Set the values to be saved as a new Narrative in the HWM database.
   * @param narrName Name for the new narrative.
   * @param narrDesc Description of the new narrative.
   * @param researcher ID Number of the researcher contributing this narrative.
   * @param baseDateString String indicating the Base Date for this narrative.
   * @param isSeasonal boolean if the narrative to be saved is seasonal, false if not 
   * @return true if successful, false if not.
   */
  public boolean setInstanceToSave(String narrName, String narrDesc, 
      long researcher, String baseDateString, boolean isSeasonal){
    
    clearInstanceToSave();
    
    double baseDateValue = SimDate.parseString(baseDateString);
    if(baseDateValue < 0){
      errorReport = "Improper date specified for base date: " + baseDateString;
      return false;
    }
    
    narrativeNameToSave = AbstractABCMDBGateway.makeSQLSafe(narrName);
    descToSave          = AbstractABCMDBGateway.makeSQLSafe(narrDesc);
    researcherToSave    = researcher;
    baseDateToSave      = baseDateValue;
    isSeasonalToSave    = isSeasonal;
    return true;
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){
    return DB.getNarrativesTable().getDatum(narrativeNameToSave, descToSave, "" + researcherToSave, "" + baseDateToSave, 
        DB.currentTimeString(), isSeasonalToSave);
  } 
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    narrativeNameToSave = null;
    descToSave          = null;
    researcherToSave    = -1;
    baseDateToSave      = -1;
    isSeasonalToSave    = false;    
  }
  
  /**
   * Make a copy of a given narrative; will change
   * the 'created by' user. 
   * @param narrToCopy PK of the narrative to be copied
   * @param newNarrativeTitle new title to be assigned to the copy
   * @param newUserID PK of the researcher who creates the new copy
   * @return the new primary key of the copy of the narrative.
   */
  public long copyNarrative(long narrToCopy, String newNarrativeTitle, long newUserID){
    long newPK = 0;
                
    String dateString = DB.currentTimeString();
    
    try {
      Connection con = DB.getAdminConnection();
      Statement stCopyNarrative = con.createStatement();  
      
      // Copy the narrative itself
      ABCMNarrativesTable table = DB.getNarrativesTable();
            
      String sqlCopyNarrative = 
          "INSERT INTO " +
                       table.getFullName(DB.getOwner()) +
                       "(" +
                       table.getField_Name().getName()        + ", " + 
                       table.getField_Description().getName() + ", " + 
                       table.getField_ResearcherFK().getName() + ", " + 
                       table.getField_BaseDate().getName() + ", " + 
                       table.getField_DateCreated().getName() + ", " + 
                       table.getField_Seasonal().getName() +
                       ") " +
          " SELECT '" + newNarrativeTitle + "', " +
                    table.getField_Description().getName() + ", " + 
                    newUserID + ", " +
                    table.getField_BaseDate().getName() + ", " +
                  "'" + dateString + "', " +
                    table.getField_Seasonal().getName() + "\n" +
          " FROM " + table.getFullName(DB.getOwner()) +
          " WHERE " + table.getField_PK().getName() + " = " + narrToCopy + ";";
      
      stCopyNarrative.executeUpdate(sqlCopyNarrative);
      
      // Get the PK of the new narrative
      newPK = getPK(newNarrativeTitle);
     
      // Copy all the instructions to the new narrative
      if(!NIM.copyInstructions(narrToCopy, newPK)) addToErrorReport(NIM.getErrorReport());      
    }
    catch (Exception e) {
      addToErrorReport("Error in NarrativeManager copyNarrative(): " + e.getMessage());
    }
    return newPK;
  }  

  /* Auditable */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField baseTableOwnerField() {  return DB.getNarrativesTable().getField_ResearcherFK();  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView() { return DB.getCanBeEdited_NarrativesView(); }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField canBeEditedViewPKField() { return DB.getCanBeEdited_NarrativesView().getField_Narrative_PK(); }

  
  /* Deleting */
  
  /**
   * Delete a specific narrative from the database. Deletes all associated
   * narrative instructions. Enforces audit rules.
   * @param PK primary key of the Narrative to be deleted
   * @return True if successful, false if not.
   */
  public boolean deleteItem(long PK, long user){
    if(!canBeEdited(PK)){
      addToErrorReport("Cannot delete narrative; narrative cannot be edited");
      return false; // Confirm can-be-edited status
    }
    return  NIM.deleteElementsByPrimaryFK(PK) && 
            super.deleteItem(PK, user);    
  }
  
  /**
   * For a specific Narrative, on a specific date, delete an instruction, specified
   * by instruction order number.
   * @param narrPK Primary key of the Narrative to be modified
   * @param narrInstrDate Date of instruction to be deleted
   * @param instrOrder Order number of instruction to be deleted
   * @return True if successful, False if not.
   */
  public boolean deleteInstructionOnDate(long narrPK, double narrInstrDate, int instrOrder){
    return NIM.deleteItem(narrPK, narrInstrDate, instrOrder);    
  }
  
  /**
   * Inserts a particular instruction on a given date into a Narrative,
   * using the order number specified. Automatically handles adjustments
   * to existing instruction order numbers. Enforces audit rules
   * @param narrPK Primary key of narrative to be modified
   * @param narrInstrDate Date of instruction to be inserted.
   * @param instrOrder Instruction order of the instruction to be inserted
   * @param command primary key of the command to be executed
   * @param Args arguments to that command; may be up to 10.
   * @param performCollapseInstructionOrders setting this to 'false'
   * allows instructions to be added more quickly by bypassing the error
   * check that ensures that all instructions are in proper order (i.e.
   * on a given date there is a 1, 2, 3, but no gaps like 1, 2, 6, etc.)
   * This should only be done when instructions are being added from
   * an automated routine and there is no danger of a problem resulting.
   * @return True if successful, False if not
   */
  public boolean insertInstructionOnDate(long narrPK, double narrInstrDate, int instrOrder, 
      long command, String[] Args, boolean performCollapseInstructionOrders){    
    return insertInstructionOnDate(narrPK, narrInstrDate, instrOrder, command, Args, "", performCollapseInstructionOrders);
  }
  
  /**
   * Inserts a particular instruction on a given date into a Narrative,
   * using the order number specified. Automatically handles adjustments
   * to existing instruction order numbers. Enforces audit rules. 
   * @param narrPK Primary key of narrative to be modified
   * @param narrInstrDate Date of instruction to be inserted.
   * @param instrOrder Instruction order of the instruction to be inserted
   * @param command primary key of the command to be executed
   * @param Args arguments to that command; may be up to 10.
   * @param comment a comment to be associated with this instruction
   * @param performCollapseInstructionOrders setting this to 'false'
   * allows instructions to be added more quickly by bypassing the error
   * check that ensures that all instructions are in proper order (i.e.
   * on a given date there is a 1, 2, 3, but no gaps like 1, 2, 6, etc.)
   * This should only be done when instructions are being added from
   * an automated routine and there is no danger of a problem resulting.
   * @return True if successful, False if not
   */
  public boolean insertInstructionOnDate(long narrPK, double narrInstrDate, int instrOrder, 
      long command, String[] Args, String comment, boolean performCollapseInstructionOrders){
    if(!canBeEdited(narrPK)){
      addToErrorReport("Cannot insert instruction; NarrativeCannotBeEdited.");
      return false;
    }
    return NIM.insertInstruction(narrPK, narrInstrDate, instrOrder, command, Args, comment, performCollapseInstructionOrders);
  }
  
  /**
   * Inserts a particular instruction on a given date into a Narrative,
   * using the order number specified. Automatically handles adjustments
   * to existing instruction order numbers. Enforces audit rules. Omits comment.
   * Automatically updates instruction order numbers. 
   * @param narrPK Primary key of narrative to be modified
   * @param narrInstrDate Date of instruction to be inserted.
   * @param instrOrder Instruction order of the instruction to be inserted
   * @param command primary key of the command to be executed
   * @param Args arguments to that command; may be up to 10.
   * @return True if successful, False if not
   */
  public boolean insertInstructionOnDate(long narrPK, double narrInstrDate, int instrOrder, 
      long command, String[] Args){
    return insertInstructionOnDate(narrPK, narrInstrDate, instrOrder, command, Args, "");
  }  
  
  /**
   * Inserts a particular instruction on a given date into a Narrative,
   * using the order number specified. Automatically handles adjustments
   * to existing instruction order numbers. Enforces audit rules. 
   * Automatically updates instruction order numbers. 
   * @param narrPK Primary key of narrative to be modified
   * @param narrInstrDate Date of instruction to be inserted.
   * @param instrOrder Instruction order of the instruction to be inserted
   * @param command primary key of the command to be executed
   * @param Args arguments to that command; may be up to 10.
   * @param comment comment to be associated with this instruction
   * @return True if successful, False if not
   */
  public boolean insertInstructionOnDate(long narrPK, double narrInstrDate, int instrOrder, 
      long command, String[] Args, String comment){
    return insertInstructionOnDate(narrPK, narrInstrDate, instrOrder, command, Args, comment, true);
  }
  
  /**
   * Changes the base date of a Narrative.
   * Checks first to see if Narrative has been run; will not
   * change base date of a narrative that has already been run.
   * @param narrPK primary key of the narrative to be modified
   * @param newBaseDate new base date to be used for this narrative.
   * @return true if successful, false if not.
   */
  public boolean changeBaseDate(long narrPK, double newBaseDate){
    if(canBeEdited(narrPK))
      if(DB.updateItem(DB.getNarrativesTable(), DB.getNarrativesTable().getField_PK(), narrPK, DB.getNarrativesTable().getField_BaseDate(), newBaseDate))
        return true;
      else{
        addToErrorReport("Could not change base date: " + DB.getErrorReport());
        return false;
      }
    else{
      addToErrorReport("Cannot change base date of narrative; narrative cannot be edited.");
      return false;
    }
  }

  
  
}
