package abcm.db.managers;


import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMParametersTable;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.utilities.XMLAssist;

/**
 *  Manages the storage and retrieval of parameters in the HWM Database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ParameterManager extends AbstractNonAuditableDBManager{
  
  /** Hashtable of descriptions for quick reference */ 
  private Hashtable descriptions;
  /** Hashtable of default values for quick reference */
  private Hashtable defaultValues;
  /** Hashtable of units for quick reference */
  private Hashtable units;
  /** Hashtable of minimum values for quick reference */
  private Hashtable mins;
  /** Hashtable of maximum values for quick reference */
  private Hashtable maxs;
  /** Hashtable of IDs for quick reference */
  private Hashtable IDs;
    
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public ParameterManager(AbstractABCMDBGateway db){
    super(db);
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  public ABCMdbTable baseTable(){                return DB.getParametersTable();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  public ABCMdbField baseTablePKField(){         return DB.getParametersTable().getField_PK();                 }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){   return DB.getParametersTable();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){       return DB.getParametersTable().getField_PK();                 }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){       return DB.getParametersTable().getField_Name();               }
  
  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){ return "parameters"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMParametersTable table = DB.getParametersTable();
    
    int IDFieldCol             = rs.findColumn(table.getField_PK().getName());
    int nameFieldCol           = rs.findColumn(table.getField_Name().getName());
    int descriptionFieldCol    = rs.findColumn(table.getField_Description().getName());
    int dateAddedFieldCol      = rs.findColumn(table.getField_DateCreated().getName());
    int defaultValueFieldCol   = rs.findColumn(table.getField_DefaultValue().getName());
    int minValueFieldCol       = rs.findColumn(table.getField_Minimum().getName());
    int maxValueFieldCol       = rs.findColumn(table.getField_Maximum().getName());
    int unitsFieldCol          = rs.findColumn(table.getField_Units().getName());
    
    
    while(rs.next()){
      xml.open("parameter");
        xml.addLine("ID",             rs.getLong(IDFieldCol));
        xml.addLine("name",           rs.getString(nameFieldCol));
        xml.addLine("desc",           rs.getString(descriptionFieldCol));
        xml.addLine("dateAdded",      rs.getString(dateAddedFieldCol));
        xml.addLine("defaultValue",   rs.getDouble(defaultValueFieldCol));
        xml.addLine("min",            rs.getDouble(minValueFieldCol));
        xml.addLine("max",            rs.getDouble(maxValueFieldCol));
        xml.addLine("unit",           rs.getString(unitsFieldCol));
        
        if(singleInstance){
          CommentManager cm = new CommentManager(DB);
          xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.PARAMETER, instancePK), XMLAssist.INDENT_BLOCK);
        }
      xml.closeTag();
    }
  }
  
  /**
   * Overridden; parametes cannot be saved programmatically.
   * @see abcm.db.managers.abstractparents.AbstractDBManager#saveInstanceToDB()
   * @return false
   */
  public boolean saveInstanceToDB() {    return false;    }

  /** 
   * Takes no action; parameters cannot be saved programmatically 
   */
  public void clearInstanceToSave(){ }  
  


  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlParameterListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlParameterToHTML.xsl"; }
  

  /**
   * Initializes the internal collection of Hash tables and populates them.
   */
  private void initHashTables(){
    IDs           = new Hashtable();
    descriptions  = new Hashtable();
    defaultValues = new Hashtable();
    units         = new Hashtable();
    mins          = new Hashtable();
    maxs          = new Hashtable();
    
    try{
      String sql = DB.getParametersTable().getSelectAllSQL(DB.getOwner());
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      
      ResultSet rs = st.executeQuery(sql);
    
      ABCMParametersTable table = DB.getParametersTable();
      
      int IDFieldCol             = rs.findColumn(table.getField_PK().getName());
      int nameFieldCol           = rs.findColumn(table.getField_Name().getName());
      int descriptionFieldCol    = rs.findColumn(table.getField_Description().getName());
      int defaultValueFieldCol   = rs.findColumn(table.getField_DefaultValue().getName());
      int minValueFieldCol       = rs.findColumn(table.getField_Minimum().getName());
      int maxValueFieldCol       = rs.findColumn(table.getField_Maximum().getName());
      int unitsFieldCol          = rs.findColumn(table.getField_Units().getName());
//      int dateAddedFieldCol      = rs.findColumn(table.getField_DateCreated().getName());
      
      
      while(rs.next()){
        String pName = rs.getString(nameFieldCol);
        
        IDs.put(pName, rs.getString(IDFieldCol));
        descriptions.put(pName, rs.getString(descriptionFieldCol));
        defaultValues.put(pName, rs.getString(defaultValueFieldCol));
        units.put(pName, rs.getString(unitsFieldCol));
        mins.put(pName, rs.getString(minValueFieldCol));
        maxs.put(pName, rs.getString(maxValueFieldCol));
        
      }
      
      rs.close();
      st.close();
      con.close();
      
    }
    catch(Exception e){
      addToErrorReport("Could not get list of parameters as option list: " + e.getMessage());
    }
  }  
  
  /**
   * Gets the description of a specified parameter
   * @param paramName the name of the parameter
   * @return the parameter's description
   */
  public String getDescription(String paramName){
    if(descriptions == null) initHashTables();
    if(!descriptions.containsKey(paramName)) return "Not Found.";
    return (String)descriptions.get(paramName);
  }  

  /**
   * Gets the units in which a parameter is specified
   * @param paramName the name of the parameter
   * @return the units in which that parameter is specified, i.e. mm, m, km, etc.
   */
  public String getUnits(String paramName){
    if(units == null) initHashTables();
    if(!units.containsKey(paramName)) return "Not Found.";
    return (String)units.get(paramName);
  }
  
  /**
   * Gets the default value for the specified parameter
   * @param paramName the name of the parameter
   * @return the default value, as a String
   */
  public String getDefaultValue(String paramName){
    if(defaultValues == null) initHashTables();
    if(!defaultValues.containsKey(paramName)) return "Not Found.";
    return (String)defaultValues.get(paramName);
  }
  

  /**
   * Gets the minimum recommended value for the specified parameter.
   * Returns "Not Found" if the parameter does not exist.
   * @param paramName the name of the parameter 
   * @return the minimum recommended value, as a String
   */
  public String getMin(String paramName){
    if(mins == null) initHashTables();
    if(!mins.containsKey(paramName)) return "Not Found.";
    return (String)mins.get(paramName);
  }
  
  /**
   * Gets the maximum recommended value for the specified parameter.
   * Returns "Not Found" if the parameter does not exist.
   * @param paramName the name of the parameter 
   * @return the maximum recommended value, as a String
   */
  public String getMax(String paramName){
    if(maxs == null) initHashTables();
    if(!maxs.containsKey(paramName)) return "Not Found.";
    return (String)maxs.get(paramName);
  }
  
  /**
   * Gets the ID (primary key in the HWM database for a specified parameter.
   * Returns "Not Found" if the parameter does not exist.
   * @param paramName the name of the parameter
   * @return the parameter's ID, as a String
   */
  public String getID(String paramName){
    if(IDs == null) initHashTables();
    if(!IDs.containsKey(paramName)) return "Not Found.";
    return (String)IDs.get(paramName);
  }
  
  /**
   * Gets a programmatically accessible list (Enumeration)
   * of all the parameters in the HWM Database.
   * @return the list of all HWM parameters
   */
  public Enumeration getParameters(){
    if(descriptions == null) initHashTables();
    return descriptions.keys();
  }
  
  
  /**
   * For a given category, returns a Vector listing the parameters that
   * fall within that category.
   * @param CategoryName the name of the category
   * @return Vector listing parameters in the specified category
   */
  public Vector getParametersInCategory(String CategoryName){
    Vector params = new Vector();
    try{
      String sql = DB.getParametersAndCategoriesView().getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(DB.getParametersAndCategoriesView().getField_ParameterCategory_Name().getName(), CategoryName));
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      
      ResultSet rs = st.executeQuery(sql);      
      
      while(rs.next()){
        params.add(rs.getString(DB.getParametersAndCategoriesView().getField_Parameter_Name().getName()));        
      }
      
      rs.close();
      st.close();
      con.close();
      
    }
    catch(Exception e){
      addToErrorReport("Could not retrieve parameters by category: an error occurred- " + e.getMessage());      
    }
    return params;
  }
  
  /**
   * Returns true if the parameter applies to the HWM engine and not
   * to only one or a set of code models. Parameters for which this is
   * true will always be involved in the simulation's operation;
   * other parameters will only be involved if the code model specifies
   * subroutines that include them. 
   * @param paramName the name of the parameter
   * @return true if it applies to the HWM engine, false otherwise
   */
  public boolean isEngineApplicable(String paramName){
    return DB.getBooleanItem(baseTable(), baseTableNameField(), paramName, DB.getParametersTable().getField_IsEngine());
  }

  
  // Unused
  /**
   * Overridden to take no action; parameters are not saved programmatically
   * @return null
   */
  public ABCMdbDatum datumToSave() {       return null;    }

  /**
   * Overridden to take no action; parameters are not saved programmatically
   * @return null
   */
  public boolean deleteItem(long PK) {
    addToErrorReport("Parameters may not be deleted programmatically.");
    return false;   
  }
    
}
