package abcm.db.managers;


import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMParameterSetsAndResearchersView;
import abcm.db.components.views.ABCMParameterSetsAndValuesView;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.sim.ParameterSet;
import abcm.utilities.XMLAssist;

/**
 * Manages the storage and retrieval of Parameter Sets in the HWM
 * Database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ParameterSetManager extends AbstractAuditableDBManager{
  
  /**   */
  private String                       nameToSave;
  /**   */
  private boolean                      automaticToSave;
  /**   */
  private String                       descToSave;
  /**   */
  private long                         userIDToSave;
  /**   */
  private ParameterSetParameterManager PSPM;

  /**
   * Database Manager for Parameter set Parameters 
   *
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class ParameterSetParameterManager extends AbstractDBLinkTableInnerManager{

    /** FK of the Parameter Set to save  */
    private long   parameterSetFKToSave;
    /** FK of the Parameter to save  */
    private long   parameterFKToSave;
    /** Value to be saved  */
    private double valueToSave;
    /** Truth value to save  */
    private long   truthValueFKToSave;
    
    /**
     * Constructor
     * @param db Instance of an ABCM Gateway to provide database access.
     * @param containedBy enclosing class, also a manager 
     */
    public ParameterSetParameterManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
     */
    protected ABCMdbTable baseTable() {                     return DB.getParameterSetsAndValuesTable();                                }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
     */
    protected ABCMdbField baseTablePKField() {              return DB.getParameterSetsAndValuesTable().getField_PK();                  }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
     */
    protected ABCMdbField baseTablePrimaryFKField() {       return DB.getParameterSetsAndValuesTable().getField_ParameterSet_FK();     }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
     */
    protected ABCMdbField baseTableSecondaryFKField() {     return DB.getParameterSetsAndValuesTable().getField_Parameter_FK();        }
 
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
     */
    protected ABCMdbAbstractViewable recordsView() {        return DB.getParameterSetsAndValuesView();                                 }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
     */
    protected ABCMdbField recordsViewPKField() {            return null; }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
     */
    protected ABCMdbField recordsViewPrimaryFKField() {     return DB.getParameterSetsAndValuesView().getField_ParameterSet_PK();      }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
     */
    protected ABCMdbField recordsViewPrimaryNameField() {   return DB.getParameterSetsAndValuesView().getField_ParameterSet_Name();    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
     */
    protected ABCMdbField recordsViewSecondaryNameField() { return DB.getParameterSetsAndValuesView().getField_Parameter_Name();       }    
    
    
    
     
    
    
    /**
     * Gets the field that represents the FK to the primary table
     * @return primary table FK field
     */
    public ABCMdbField getPrimaryFKField() {       return DB.getParameterSetsAndValuesTable().getField_ParameterSet_FK();  }

    /**
     * Gets the field that represents the name in the primary table, drawn from the view
     * @return primary table name field (view)
     */
    public ABCMdbField getPrimaryNameField() {     return DB.getParameterSetsAndValuesView().getField_ParameterSet_Name(); }
    
    /**
     * Gets the field that represents the FK to the secondary table
     * @return secondary table FK field
     */
    public ABCMdbField getSecondaryFKField() {     return DB.getParameterSetsAndValuesTable().getField_Parameter_FK();     }

    /**
     * Gets the field that represents the name in the secondary table, drawn from the view
     * @return secondary table name field (view)
     */
    public ABCMdbField getSecondaryNameField() {   return DB.getParameterSetsAndValuesView().getField_Parameter_Name();    }
        

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
      ABCMParameterSetsAndValuesView view = DB.getParameterSetsAndValuesView();
      
      int paramIDFieldCol         = rs.findColumn(view.getField_Parameter_PK().getName());
      int paramNameFieldCol       = rs.findColumn(view.getField_Parameter_Name().getName());
      int paramDescFieldCol       = rs.findColumn(view.getField_Parameter_Description().getName());
      int paramUnitsFieldCol      = rs.findColumn(view.getField_Parameter_Units().getName());
      int defaultValueFieldCol    = rs.findColumn(view.getField_Parameter_DefaultValue().getName());
      int valueFieldCol           = rs.findColumn(view.getField_Value().getName());
      
      
      while(rs.next()){
        xml.open("parameter");
          xml.addLine("paramID",   rs.getLong(paramIDFieldCol));
          xml.addLine("paramName", rs.getString(paramNameFieldCol));
          xml.addLine("paramDesc", rs.getString(paramDescFieldCol) + " (" + rs.getString(paramUnitsFieldCol) + ")");
          xml.addLine("default",   rs.getDouble(defaultValueFieldCol));
          xml.addLine("value",    (rs.getString(valueFieldCol) != null ? rs.getString(valueFieldCol) : (rs.getString(defaultValueFieldCol) + " (Default)")));          
        xml.closeTag();
      }      
    }

    /**
     * Sets an instance of a Parameter Set Parameter to save
     * @param parameterSetFK FK to the parameter set
     * @param parameterFK FK to the parameter
     * @param value value to be saved
     * @param truthValue truth value (FK to truth values table)
     */
    public void setInstanceToSave(long parameterSetFK, long parameterFK, double value, long truthValue){
      parameterSetFKToSave = parameterSetFK;
      parameterFKToSave    = parameterFK;
      valueToSave          = value;
      truthValueFKToSave   = truthValue;
    }
    
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
     */
    protected void clearInstanceToSave() {
      parameterSetFKToSave = -1;
      parameterFKToSave    = -1;
      valueToSave          = 0;
      truthValueFKToSave   = -1;  
    }

    /**
     * Determines whether the instance to be saved is valid
     * @return true if the instance to be saved has been set, false otherwise
     */
    public boolean instanceToSaveIsValid(){
      return ((parameterSetFKToSave != -1) && 
              (parameterFKToSave    != -1) &&
              (truthValueFKToSave   != -1));
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
     */
    protected ABCMdbDatum datumToSave() {
      return (instanceToSaveIsValid()  ? 
          DB.getParameterSetsAndValuesTable().getDatum("" + parameterSetFKToSave, "" + parameterFKToSave, "" + valueToSave, "" + truthValueFKToSave): 
          null);           
    }

    
   
  } // End inner class 
  
  
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public ParameterSetManager(AbstractABCMDBGateway db){    
    super(db);
    PSPM = new ParameterSetParameterManager(db, this);
  }

  
  /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){              return DB.getParameterSetsTable();                                            }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){       return DB.getParameterSetsTable().getField_PK();                              }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField(){      return DB.getParameterSetsTable().getField_Name();                            }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView(){ return DB.getParameterSetsAndResearchersView();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField(){     return DB.getParameterSetsAndResearchersView().getField_ParameterSet_PK();    }


  
  /* XML Creation */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName() {  return "parameterSets";  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
    
    ABCMParameterSetsAndResearchersView view = DB.getParameterSetsAndResearchersView();
    
    int paramSetIDFieldCol   = rs.findColumn(view.getField_ParameterSet_PK().getName());
    int paramSetNameFieldCol = rs.findColumn(view.getField_ParameterSet_Name().getName());
    int dateCreatedFieldCol  = rs.findColumn(view.getField_ParameterSet_DateCreated().getName());
    int automaticFieldCol    = rs.findColumn(view.getField_Automatic().getName());
    int descriptionFieldCol  = rs.findColumn(view.getField_ParameterSet_Description().getName());
    int researcherFieldCol   = rs.findColumn(view.getField_Researcher_Name().getName());
    int canBeEditedFieldCol  = rs.findColumn(view.getField_CanBeEdited().getName());
    int truthValueFieldCol   = rs.findColumn(view.getField_ParameterSet_TruthValue().getName());
    
    while(rs.next()){
      xml.open("parameterSet");
        xml.addLine("ID",              rs.getLong(paramSetIDFieldCol));
        xml.addLine("name",            rs.getString(paramSetNameFieldCol));
        xml.addLine("dateCreated",     rs.getString(dateCreatedFieldCol));
        xml.addLine("automatic",      (rs.getBoolean(automaticFieldCol) ? "Yes" : "No"));
        xml.addLine("desc",            rs.getString(descriptionFieldCol));
        xml.addLine("createdBy",       rs.getString(researcherFieldCol));
        xml.addLine("canBeEdited",   ((rs.getString(canBeEditedFieldCol) != null) ? "Yes" : "No"));
        xml.addLine("truthValue",      rs.getString(truthValueFieldCol));
        
        if(singleInstance){
          xml.open("parameters");
            PSPM.addSetToXML(xml, includeComments, rs.getLong(paramSetIDFieldCol));
          xml.closeTag();
        }
        if(includeComments){
          CommentManager cm = new CommentManager(DB);
          xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.PARAM_SET, instancePK), XMLAssist.INDENT_BLOCK);
        }
      xml.closeTag();
    }     
       
  }
  
  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlParameterSetListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlParameterSetToHTML.xsl"; }

  
  /* Saving */
  
  /**
   * Sets the information required to save a new Parameter Set
   * @param name name of the Parameter Set to be Saved
   * @param automatic true if this Parameter Set is generated 
   * automatically, false otherwise (currently automated generation of
   * Parameter Sets is not implemented)
   * @param desc description of this Parameter Set
   * @param userID ID of the user who is creating and saving this Parameter Set
   */
  public void setInstanceToSave(String name, boolean automatic, String desc, long userID){
    nameToSave       = AbstractABCMDBGateway.makeSQLSafe(name);
    automaticToSave  = automatic;
    descToSave       = AbstractABCMDBGateway.makeSQLSafe(desc);
    userIDToSave     = userID;    
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    nameToSave      = null;
    automaticToSave = false;
    descToSave      = null;
    userIDToSave    = -1;
  }
    
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return((nameToSave   != null) && 
           (descToSave   != null) &&
           (userIDToSave != -1));  
  }
    
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave() {
    return DB.getParameterSetsTable().getDatum(nameToSave, DB.currentTimeString(), 
        automaticToSave, descToSave, userIDToSave);
  }
  
  /* Special */
  
  /**
   * Gets the Parameter Set specified by the given primary key
   * @param PK primary key of the requested Parameter Set
   * @return the requested Parameter Set
   */
  public ParameterSet getParameterSet(long PK){
    ParameterSet p = new ParameterSet();
    try{

      String sql = DB.getParameterSetsAndValuesView().getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(DB.getParameterSetsAndValuesView().getField_ParameterSet_PK().getName(), PK));
      
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      while(rs.next()){
        double val = rs.getDouble(DB.getParameterSetsAndValuesView().getField_Parameter_DefaultValue().getName());
        if (rs.getString(DB.getParameterSetsAndValuesView().getField_Value().getName()) != null)
          val = rs.getDouble(DB.getParameterSetsAndValuesView().getField_Value().getName());
        String paramName = rs.getString(DB.getParameterSetsAndValuesView().getField_Parameter_Name().getName());
        p.setParameter(paramName, val);
      }
    }
    catch(Exception e){
      addToErrorReport("Error creating parameter set: " + e.getMessage());
    }
    return p;
  }
  
  
  
  /**
   * Deletes the specified Parameter Set from the HWM Database.
   * Confirms that the Parameter Set can be edited before deleting.
   * @param PK primary key of the Parameter Set to be deleted
   * @param user the ID of the user deleting this item (used to establish permissions)
   * @return true if successful (even if no Parameter Set with
   * the specified primary key exists), false otherwise
   */
  public boolean deleteItem(long PK, long user){
    return permitModification(PK, user) && 
           PSPM.deleteElementsByPrimaryFK(PK) &&
           super.deleteItem(PK, user);
  }

  /**
   * Removes a Parameter and value from a given Parameter Set.
   * Confirms that the Parameter Set can be edited before making
   * this change.
   * @param PK primary key of the Parameter Set to be modified
   * @param paramPK primary key of the Parameter to be removed.
   * @param user the ID of the user deleting this item (used to establish permissions)
   * @return true if successful (even if no match existed)
   */
  public boolean removeParameter(long PK, long paramPK, long user){
    return permitModification(PK, user) &&
           PSPM.deleteElementsByBothPrimaryKeys(PK, paramPK);        
  }

  /**
   * Modifies the value assigned to a given Parameter in the specified
   * Parameter Set.
   * Note that this initially deletes any existing value for this
   * Parameter; if the operation is not completed, the old
   * value may be deleted without being replaced. Note also
   * that there is no obligation for the Parameter to
   * exist; the routine will not provide an error if no
   * Parameter is found with the matching primary key. 
   * @param PK primary key of the Parameter Set to be modified
   * @param param primary key of the Parameter to be modified
   * @param newVal new value to be assigned to this Parameter
   * @param newTruthValue the new truth value to be assigned to this parameter
   * @param user the user saving the new parameter
   * @return true if successful, false otherwise
   */
  public boolean modifyParameter(long PK, long param, double newVal, long newTruthValue, long user){
    if(!permitModification(PK, user)){
      addToErrorReport("Could not remove or modify parameter; parameter set cannot be edited");
      return false;
    }
    PSPM.setInstanceToSave(PK, param, newVal, newTruthValue);
    return removeParameter(PK, param, user) &&  PSPM.saveInstanceToDB();   
  }

  /* Auditable */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField baseTableOwnerField(){        return DB.getParameterSetsTable().getField_ResearcherFK();               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView(){ return DB.getCanBeEdited_ParameterSetsView();                            }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField canBeEditedViewPKField(){     return DB.getCanBeEdited_ParameterSetsView().getField_ParameterSet_PK(); }

  
}
