package abcm.db.managers;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.text.SimpleDateFormat;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMProbesAndResearchersView;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.math.SimDate;
import abcm.sim.Probe;
import abcm.utilities.XMLAssist;

/**
 * Provides means for storing and retrieving Probes from the HWM Database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ProbeManager extends AbstractAuditableDBManager {
  
  /** User ID to be saved  */
  private long     userIDToSave;
  /** Name of Probe to be saved  */
  private String   nameToSave;
  /** Description of Probe to be saved  */
  private String   descToSave;
  /** Name of series to be saved  */
  private String   seriesToSave;
  /** Name of domain to be saved  */
  private String   domainToSave;
  /** Flag indicating if the Probe to be saved is numeric  */
  private boolean  isNumericToSave;
  /** Flag indicating if the Probe to be saved is monthly  */
  private boolean  isMonthlyToSave;
  /** Base Date of Probe to be saved  */
  private double   baseDateToSave;
  /** Frequency of Probe to be saved  */
  private int      frequencyToSave;
  /** Retrieval command of Probe to be saved  */
  private String   retrieveToSave;
  /** Offset of Probe to be saved  */
  private double   offsetToSave;
  
  /* Constructor */
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public ProbeManager(AbstractABCMDBGateway db){
    super(db);
  }
  

  /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable() { return DB.getProbesTable(); }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField() {    return DB.getProbesTable().getField_PK();   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){      return DB.getProbesTable().getField_Name();  }

  
  /* XML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){  return DB.getProbesAndResearchersView();                        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){      return DB.getProbesAndResearchersView().getField_Probe_PK();    }
  
  /**
   * Gets the name field of the records view.
   * @return the records view name field
   */
  public ABCMdbField recordsViewNameField(){    return DB.getProbesAndResearchersView().getField_Probe_Name();  }
  
   
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){             return "probes";                                                }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    
    ABCMProbesAndResearchersView view = DB.getProbesAndResearchersView();
   
    int probeIDFieldCol          = rs.findColumn(view.getField_Probe_PK().getName());
    int probeNameFieldCol        = rs.findColumn(view.getField_Probe_Name().getName());
    int probeDescriptionFieldCol = rs.findColumn(view.getField_Probe_Desc().getName());
    int dateCreatedFieldCol      = rs.findColumn(view.getField_Probe_DateCreated().getName());
    int createdByFieldCol        = rs.findColumn(view.getField_Researcher_Name().getName());
    int dataSeriesNameFieldCol   = rs.findColumn(view.getField_DataSeries_Name().getName());
    int dataSeriesDomainFieldCol = rs.findColumn(view.getField_DataSeries_Domain().getName());
    int isNumericFieldCol        = rs.findColumn(view.getField_Probe_IsNumeric().getName());
    int isMonthlyFieldCol        = rs.findColumn(view.getField_Probe_IsMonthly().getName());
    int baseDateFieldCol         = rs.findColumn(view.getField_Probe_BaseDate().getName());
    int frequencyFieldCol        = rs.findColumn(view.getField_Probe_Frequency().getName());
    int retrieveFieldCol         = rs.findColumn(view.getField_Probe_Retrieve().getName());
    int canBeEditedFieldCol      = rs.findColumn(view.getField_CanBeEdited().getName());
    int offsetFieldCol           = rs.findColumn(view.getField_Probe_Offset().getName());
  
    while(rs.next()){
      xml.open("probe");
        xml.addLine("ID",           rs.getLong(probeIDFieldCol));
        xml.addLine("name",         rs.getString(probeNameFieldCol));
        xml.addLine("desc",         rs.getString(probeDescriptionFieldCol));
        xml.addLine("dateCreated",  (new SimpleDateFormat()).format(rs.getDate(dateCreatedFieldCol)));
        xml.addLine("createdBy",    rs.getString(createdByFieldCol));
        xml.addLine("series",       rs.getString(dataSeriesNameFieldCol));
        xml.addLine("domain",       rs.getString(dataSeriesDomainFieldCol));
        xml.addLine("isNumeric",    (rs.getBoolean(isNumericFieldCol) ? "Yes" : "No"));
        xml.addLine("isMonthly",    (rs.getBoolean(isMonthlyFieldCol) ? "Yes" : "No"));
        xml.addLine("baseDate",     SimDate.convertToString(rs.getDouble(baseDateFieldCol)));
        xml.addLine("frequency",    rs.getDouble(frequencyFieldCol));
        xml.addLine("retrieve",     rs.getString(retrieveFieldCol));
        xml.addLine("canBeEdited",  ((rs.getString(canBeEditedFieldCol) != null) ? "Yes" : "No"));
        xml.addLine("offset",       rs.getDouble(offsetFieldCol));
      xml.closeTag();
    }    
  }
  
  
  /* Setting instance to save */

  /**
   * Set the values of a new Probe Manager to be saved in the HWM Database
   * @param userID User ID of the person who will be the owner of this Probe
   * @param name a name for the probe
   * @param description a description for the probe
   * @param series the series to which the probe will dump data
   * @param domain the domain to which the probe will dump data
   * @param isNumeric true if the probe is numeric; false otherwise
   * @param isMonthly true if the probe is a monthly probe; false otherwise
   * @param baseDate base date for the probe
   * @param frequency frequency at which the probe collects data, in days
   * @param toRetrieve retrieval String for the probe
   * @param offset offset value for within-day probing
   */
  public void setInstanceToSave(long userID, 
                                String name, 
                                String description, 
                                String series, 
                                String domain, 
                                boolean isNumeric, 
                                boolean isMonthly,
                                double baseDate, 
                                int frequency, 
                                String toRetrieve, double offset){
    userIDToSave    = userID;
    nameToSave      = name;
    descToSave      = description;
    seriesToSave    = series;
    domainToSave    = domain;
    isNumericToSave = isNumeric;
    isMonthlyToSave = isMonthly;
    baseDateToSave  = baseDate;
    frequencyToSave = frequency;
    retrieveToSave  = toRetrieve;
    offsetToSave    = Math.abs(offset) % 1d;
    
  }  
  
  
  /* Saving */
  
  /**
   * Determines if the instance to be saved is valid.
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){ return nameToSave != null; } // Only check one- all set at once
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    userIDToSave    = -1;
    nameToSave      = null;
    descToSave      = null;
    seriesToSave    = null;
    domainToSave    = null;
    isNumericToSave = false;
    isMonthlyToSave = false;
    baseDateToSave  = -1;
    frequencyToSave = -1;
    retrieveToSave  = null;
    offsetToSave    = -1;
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){
    if(instanceToSaveIsValid())  
        return DB.getProbesTable().getDatum(nameToSave, 
                                            descToSave, 
                                            DB.currentTimeString(), 
                                            "" + userIDToSave, 
                                            seriesToSave, 
                                            domainToSave, 
                                            (isNumericToSave ? "1" : "0"), 
                                            (isMonthlyToSave ? "1" : "0"), 
                                            "" + baseDateToSave, 
                                            "" + frequencyToSave, 
                                            retrieveToSave, 
                                            "" + offsetToSave);  
     else return null;  
  }

  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlProbeListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlProbeToHTML.xsl"; }

  
  /* Auditable */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField baseTableOwnerField(){    return DB.getProbesTable().getField_ResearcherFK();          }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView(){ return DB.getCanBeEdited_ProbesView();                   }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField canBeEditedViewPKField(){ return DB.getCanBeEdited_ProbesView().getField_Probe_PK();   }
    
  
  /* Retrieving a simulation object */
  
  /**
   * Retrieves a given Probe's information from the ABCM database and creates
   * a programmatic Probe object from it.
   * @param PK primary key of the record to be retrieved
   * @return Probe object created from that record
   */
  public Probe getProbe(long PK){
    Probe p = null;
    try{
      Connection con = DB.getGuestConnection(); 
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(DB.getProbesTable().getSelectSQL(DB.getOwner(), AbstractDBGateway.where(DB.getProbesTable().getField_PK().getName(),PK)));

      if(rs.next()){
        String  name           = rs.getString(DB.getProbesTable().getField_Name().getName());
        String  dataSeriesName = rs.getString(DB.getProbesTable().getField_DataSeriesName().getName());
        String  domainName     = rs.getString(DB.getProbesTable().getField_DataSeriesDomain().getName());
        String  retrieve       = rs.getString(DB.getProbesTable().getField_Retrieve().getName());
        boolean isNumeric      = rs.getBoolean(DB.getProbesTable().getField_IsNumeric().getName());
        boolean isMonthly      = rs.getBoolean(DB.getProbesTable().getField_IsMonthly().getName());
        
        if(isMonthly){
          p = new Probe(name, dataSeriesName, domainName, isNumeric, retrieve);  // This constructor creates a new monthly probe
        }
        else{
          SimDate baseDate    = new SimDate(rs.getDouble(DB.getProbesTable().getField_BaseDate().getName()));
          long    frequency   = (long)rs.getDouble(DB.getProbesTable().getField_Frequency().getName());
          double  offSet      = rs.getDouble(DB.getProbesTable().getField_Offset().getName());
          
          p = new Probe(name, baseDate, frequency, dataSeriesName, domainName, isNumeric, retrieve); // This one creates a new standard probe
          p.setOffset(offSet);
        }
      }
      rs.close();
      st.close();
      con.close();
    }
    catch(Exception e){
      addToErrorReport("An error has occurred in ProbeManager.getProbe() " + e.getMessage());
    }
    return p;    
  }
  
}
