package abcm.db.managers;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.Vector;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMProbeSetProbesView;
import abcm.db.components.views.ABCMProbeSetsAndResearchersView;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.math.SimDate;
import abcm.utilities.XMLAssist;

/**
 * Manages Probe Sets in the HWM Database; provides means to save, retrieve,
 * and display information about them.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ProbeSetManager extends AbstractAuditableDBManager {

  /** Name to be saved  */
  private String                 nameToSave;
  /** Description to be saved  */
  private String                 descToSave;
  /** PK of resarcher to be saved  */
  private long                   userIDToSave;
  
  /** Inner Manager for Probe Sets  */
  private ProbeSetProbesManager  PSPM;
  
  /**
   * Database Manager for Probe Set Probes
   *
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class ProbeSetProbesManager extends AbstractDBLinkTableInnerManager {
  
    /** Probe set FK to be saved  */
    long probeSetFKToSave;
    /** FK to probe to be saved  */
    long probeFKToSave;
    
    /**
     * Constructor
     * @param db Instance of an ABCM Gateway to provide database access.
     * @param containedBy enclosing class
     */
    public ProbeSetProbesManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
     */
    protected ABCMdbTable baseTable(){                     return DB.getProbeSetProbesTable();                           }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
     */
    protected ABCMdbField baseTablePKField(){              return DB.getProbeSetProbesTable().getField_PK();             }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
     */
    protected ABCMdbField baseTablePrimaryFKField(){       return DB.getProbeSetProbesTable().getField_ProbeSetFK();     }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
     */
    protected ABCMdbField baseTableSecondaryFKField(){     return DB.getProbeSetProbesTable().getField_ProbeFK();        }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
     */
    protected ABCMdbAbstractViewable recordsView(){        return DB.getProbeSetProbesView();                            }
        
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
     */
    protected ABCMdbField recordsViewPKField(){            return DB.getProbeSetProbesView().getField_Probe_PK();        }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
     */
    protected ABCMdbField recordsViewPrimaryFKField(){     return DB.getProbeSetProbesView().getField_ProbeSet_PK();     }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
     */
    protected ABCMdbField recordsViewPrimaryNameField(){   return DB.getProbeSetProbesView().getField_ProbeSet_Name();   }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
     */
    protected ABCMdbField recordsViewSecondaryNameField(){ return DB.getProbeSetProbesView().getField_Probe_Name();      } 
    
   
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs,
        boolean includeComments) throws Exception {
      
      ABCMProbeSetProbesView view = DB.getProbeSetProbesView();
      
      int IDFieldCol                = rs.findColumn(view.getField_Probe_PK().getName());
      int nameFieldCol              = rs.findColumn(view.getField_Probe_Name().getName());
      int descFieldCol              = rs.findColumn(view.getField_ProbeSet_Desc().getName());
      int dateCreatedFieldCol       = rs.findColumn(view.getField_ProbeSet_DateCreated().getName());
      int researcherFieldCol        = rs.findColumn(view.getField_Researcher_Name().getName());
      int seriesFieldCol            = rs.findColumn(view.getField_DataSeries_Name().getName());
      int domainFieldCol            = rs.findColumn(view.getField_DataSeries_Domain().getName());
      int isNumericFieldCol         = rs.findColumn(view.getField_Probe_IsNumeric().getName());
      int isMonthlyFieldCol         = rs.findColumn(view.getField_Probe_IsMonthly().getName());
      int baseDateFieldCol          = rs.findColumn(view.getField_Probe_BaseDate().getName());
      int frequencyFieldCol         = rs.findColumn(view.getField_Probe_Frequency().getName());
      int retrieveFieldCol          = rs.findColumn(view.getField_Probe_Retrieve().getName());
      
      while(rs.next()){
        xml.open("probe");
          xml.addLine("ID",           rs.getString(IDFieldCol));
          xml.addLine("name",         rs.getString(nameFieldCol));
          xml.addLine("desc",         rs.getString(descFieldCol));
          xml.addLine("dateCreated",  (new SimpleDateFormat()).format(rs.getDate(dateCreatedFieldCol)));
          xml.addLine("createdBy",    rs.getString(researcherFieldCol));
          xml.addLine("series",       rs.getString(seriesFieldCol));
          xml.addLine("domain",       rs.getString(domainFieldCol));
          xml.addLine("isNumeric",    (rs.getBoolean(isNumericFieldCol) ? "Yes" : "No"));
          xml.addLine("isMonthly",    (rs.getBoolean(isMonthlyFieldCol) ? "Yes" : "No"));
          xml.addLine("baseDate",     SimDate.convertToString(rs.getDouble(baseDateFieldCol)));
          xml.addLine("frequency",    rs.getDouble(frequencyFieldCol));
          xml.addLine("retrieve",     rs.getString(retrieveFieldCol));
        xml.closeTag();
      }    
    }
  
    /**
     * Sets the instance of this Probe Set Probe to be saved
     * @param probeSetFK FK of probe set to be saved
     * @param probeFK FK of probe to be saved
     */
    public void setInstanceToSave(long probeSetFK, long probeFK){
      probeSetFKToSave = probeSetFK;
      probeFKToSave    = probeFK;
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
     */
    protected void clearInstanceToSave(){
      probeSetFKToSave = -1;
      probeFKToSave    = -1;    
    }
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
     */
    protected ABCMdbDatum datumToSave(){
      return DB.getProbeSetProbesTable().getDatum("" + probeSetFKToSave, "" + probeFKToSave);
    }
      
    /* Special */
    /**
     * Gets a Vector of the primary keys of the Probes in a particular Probe Set
     * @param probeSetPK primary key of the Probe Set
     * @return Vector containing the primary keys of the Probes in the specified Probe Set
     */
    public Vector getProbePKs(long probeSetPK){
      Vector v = new Vector();
      try {
        Connection con = DB.getGuestConnection();
        Statement st = con.createStatement();
        String sqlList = DB.getProbeSetProbesTable().getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(DB.getProbeSetProbesTable().getField_ProbeSetFK().getName(), probeSetPK));
        
        ResultSet rs = st.executeQuery(sqlList);
        
        while(rs.next()){
          v.add(new Long(rs.getLong(DB.getProbeSetProbesTable().getField_ProbeFK().getName())));
        }
        
      }
      catch (Exception e) {
        addToErrorReport("Error getting list of probes. " + e.getMessage());
      }
      return v;
    }

    
  }  // end inner class
    
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public ProbeSetManager(AbstractABCMDBGateway db){
    super(db);
    PSPM = new ProbeSetProbesManager(db, this);
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){           return DB.getProbeSetsTable();                                            }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){    return DB.getProbeSetsTable().getField_PK();                              }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){     return DB.getProbeSetsTable().getField_Name();                            }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){ return DB.getProbeSetsAndResearchersView();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){     return DB.getProbeSetsAndResearchersView().getField_ProbeSet_PK();        }
  
  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){                return "probesets"; }
    
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    
    ABCMProbeSetsAndResearchersView view = DB.getProbeSetsAndResearchersView();
    
    int IDFieldCol            = rs.findColumn(view.getField_ProbeSet_PK().getName()); 
    int nameFieldCol          = rs.findColumn(view.getField_ProbeSet_Name().getName());
    int descFieldCol          = rs.findColumn(view.getField_ProbeSet_Desc().getName());
    int dateCreatedFieldCol   = rs.findColumn(view.getField_ProbeSet_DateCreated().getName());
    int createdByFieldCol     = rs.findColumn(view.getField_Researcher_Name().getName());
    int canBeEditedFieldCol   = rs.findColumn(view.getField_CanBeEdited().getName());    
          
    while(rs.next()){
      xml.open("probeset");
        xml.addLine("ID",           rs.getString(IDFieldCol));
        xml.addLine("name",         rs.getString(nameFieldCol));
        xml.addLine("desc",         rs.getString(descFieldCol));
        xml.addLine("dateCreated",  (new SimpleDateFormat()).format(rs.getDate(dateCreatedFieldCol)));
        xml.addLine("createdBy",    rs.getString(createdByFieldCol));
        xml.addLine("canBeEdited",  ((rs.getString(canBeEditedFieldCol) != null) ? "Yes" : "No"));    
    
        if(singleInstance){
          // Add individual probes
          xml.open("probes");
            PSPM.addSetToXML(xml, includeComments, instancePK);
          xml.closeTag();
          
          // Add comments
          if(includeComments){
            CommentManager cm = new CommentManager(DB);
            xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.PARAM_SET, instancePK), XMLAssist.INDENT_BLOCK);
          }
        }
      xml.closeTag();
    }    
  }

  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName(){    return "xmlProbeSetListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlProbeSetToHTML.xsl"; }
  
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){
    return DB.getProbeSetsTable().getDatum(nameToSave, DB.currentTimeString(), "" + userIDToSave, descToSave);
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    nameToSave = null;
    descToSave = null;
    userIDToSave = -1;
  }
  

  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return (nameToSave != null);
  }
  
  
  /**
   * Sets an instance to be saved by specifying the information that
   * must be provided to be saved in the database.
   * @param name the name of this Probe Set
   * @param description a description of this Probe Set
   * @param userID the ID of the user who will be considered the 
   * owner of this Probe Set
   */
  public void setInstanceToSave(String name, String description, long userID){
    nameToSave = name;
    descToSave = description;
    userIDToSave = userID;
  }
  
  
  /**
   * Adds a Probe to a Probe Set in the HWM Database
   * @param probeSetPK the primary key of the Probe Set to which the Probe will be added
   * @param probePK the primary key of the Probe to be added
   * @return true if the Probe is added successfully, false otherwise
   */
  public boolean addAProbe(long probeSetPK, long probePK){
    PSPM.setInstanceToSave(probeSetPK, probePK);
    return PSPM.saveInstanceToDB();    
  }
  

  /** 
   * Deletes a Probe Set from the database. Also deletes all Probe Set's probes.
   * Enforces auditing rules.
   * @param PK primary key of the Probe Set to be Deleted
   * @param user primary key of the user attempting to perform the delete
   * @return true if successful, false otherwise.
   */
  public boolean deleteItem(long PK, long user){
    return permitModification(PK, user)       && 
           PSPM.deleteElementsByPrimaryFK(PK) && 
           super.deleteItem(PK, user);
  }

  
  /**
   * Removes a specific Probe from the given Probe Set
   * @param PK PK of Probe Set from which Probe is to be removed
   * @param probeID PK of Probe to be removed
   * @param user user attempting this deletion. Auditing is enforced
   * @return true if successful, false otherwise
   */
  public boolean removeProbe(long PK, long probeID, long user){
    return permitModification(PK, user) &&
           PSPM.deleteElementsByBothPrimaryKeys(PK, probeID);    
  }
  
  
  /**
   * Gets the list of Probes in a given Probe Set
   * @param PK PK of the Probe Set to be returned
   * @return vector of Probes in that probe set
   */
  public Vector getProbesInProbeSet(long PK){
    Vector probes = new Vector();
    ProbeManager PM = new ProbeManager(DB);
    Vector probeIDs = PSPM.getProbePKs(PK);
    for(int i = 0; i < probeIDs.size(); i++){
      probes.add(PM.getProbe(((Long)probeIDs.get(i)).longValue()));
    }
    return probes;
  }

  
  /* Auditable */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField baseTableOwnerField() {        return DB.getProbeSetsTable().getField_ResearcherFK();           }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView() { return DB.getCanBeEdited_ProbeSetsView();                        }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField canBeEditedViewPKField() {     return DB.getCanBeEdited_ProbeSetsView().getField_ProbeSet_PK(); }

}
