package abcm.db.managers;

import java.sql.ResultSet;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.utilities.XMLAssist;

/**
 * This class extends the AbstractNonAuditableDBManager even though it truncates most of its
 * functionality; the main function kept is option list creation.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ProjectResearcherManager extends AbstractNonAuditableDBManager {

  /* Constructor */
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public ProjectResearcherManager(AbstractABCMDBGateway db){     super(db);    }

  
  /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){                return DB.getProjectResearchersTable();                           }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){         return DB.getProjectResearchersTable().getField_PK();             }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField() {      return DB.getProjectResearchersTable().getField_Username();       }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView() {  return DB.getProjectResearchersTable();                           }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField() {      return DB.getProjectResearchersTable().getField_ResearcherFK();   }
  
  
  
  /* Unimplemented */
  
  /** 
   * Overridden to take no action; all arguments are ignored
   * @param xml XMLAssist that is constructing an XML object
   * @param rs result set containing records to be added
   * @param includeComments true if comments are to be included
   * @throws Exception generic exception 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {         }

  /**
   * Does nothing; project researcher records are not saved programmatically
   * @return null
   */
  public ABCMdbDatum datumToSave() {     return null;     }

  /** 
   * Does nothing; project researcher records are not saved programmatically
   */
  public void clearInstanceToSave(){ }
  
  /** 
   * Overridden to do nothing; project researcher records are not deleted programmatically
   * @return false 
   */
  public boolean deleteItem(long PK) {
    addToErrorReport("Project researcher records cannot be deleted programmatically.");
    return false;    
  }

  /** 
   * Returns null; no XML is created 
   * @return null
   */
  protected String xmlOuterNodeName() {  return null;     }

  
  /* HTML Creation */
  
  /** 
   * Returns null; no stylesheet is applicable
   * @return null 
   */
  protected String getListStylesheetName() {    return null; }
  
  /** 
   * Returns null; no stylesheet is applicable
   * @return null 
   */
  protected String getInstanceStylesheetName(){ return null; }
  
  
  /* Special */
  
  /**
   * Confirms that a user name and password exist as specified
   * @param username the username to be checked
   * @param pwd password to be checked
   * @return true if an entry exists, false otherwise
   */
  public boolean confirmLogin(String username, String pwd){
    String userField = DB.getProjectResearchersTable().getField_Username().getName();
    String pwdField  = DB.getProjectResearchersTable().getField_Password().getName();
    String sql = DB.getProjectResearchersTable().getSelectSQL(DB.getOwner(), "WHERE " + userField + " = '" + username + "' AND " +
                                                                                        pwdField      + " = '" + pwd + "'");
    if(DB.entryExists(sql)) return true;
    
    addToErrorReport(DB.getErrorReport());
    return false;
  }
  
  /**
   * Determines if the specified user is an administrator
   * @param username the user to be checked
   * @return true if the user is an administrator, false if not
   */
  public boolean isAdmin(String username){
    String userField = DB.getProjectResearchersTable().getField_Username().getName();
    String sql = DB.getProjectResearchersTable().getSelectSQL(DB.getOwner(), AbstractDBGateway.where(userField, username));
    return DB.getBooleanItem(sql, DB.getProjectResearchersTable().getField_IsAdministrator().getName());
  }

  /**
   * Gets the researcher ID for the specified project researcher
   * @param ID PK in the project researchers table
   * @return PK for corresponding entry in the researchers table
   */
  public long getResearcherID(long ID){
    return DB.getNumericItem(DB.getProjectResearchersTable(), 
                             DB.getProjectResearchersTable().getField_PK(), ID, 
                             DB.getProjectResearchersTable().getField_ResearcherFK());
  }

  /**
   * Gets the list of project researchers as an option list,
   * but using their Last, First names and the ID from
   * the Researchers table
   * @return the option list
   */
  public String getProjectResearchersAsOptionList(){
    return DB.getOptionList(DB.getProjectResearchersView(), 
                     DB.getProjectResearchersView().getField_Researcher_FK(), 
                     DB.getProjectResearchersView().getField_Researcher_LastNameFirstName());
  }
  
  /**
   * Retrieves a username given a researcher ID number (from the Researchers table).
   * @param ID the FK to the researchers table
   * @return the username, if a matching one is found.
   */
  public String getUsername(long ID){
    return DB.getStringItem(DB.getProjectResearchersTable(), 
                            DB.getProjectResearchersTable().getField_ResearcherFK(),
                            ID, 
                            DB.getProjectResearchersTable().getField_Username());
  }

}
