package abcm.db.managers;

import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMReferenceAuthorsView;
import abcm.db.components.views.ABCMReferencesTypesAndAddedByView;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.utilities.XMLAssist;

/**
 * Manages the entry and retrieval of References in the
 * HWM database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ReferenceManager extends AbstractNonAuditableDBManager {

  /** Title to be saved  */
  private String titleToSave;
  /** Date to be saved (as a String)  */
  private String dateToSave;
  /** Summary to be saved  */
  private String summaryToSave;
  /** Abstract to be saved  */
  private String abstractToSave;
  /** Work Type to be saved  */
  private long   typeOfWorkToSave;
  /** Researcher adding this work  */
  private long   researcherAdding;
  /** Location of work to be saved  */
  private String locationToSave;
  /** List of FKs to Researcher as authors to be saved (in author order)  */
  private long[] authors;

  /** Internal Reference Author Manager  */
  private ReferenceAuthorManager RAM; 
  
  /**
   * Database Manager for 
   *
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class ReferenceAuthorManager extends AbstractDBLinkTableInnerManager {
  
    /** FK to the reference to be saved  */
    public long referenceFKToSave;
    /** order of author to be saved  */
    public int  authorOrderToSave;
    /** PK in researchers table of author to save  */
    public long researcherFKToSave;
    
    /**
     * Constructor
     * @param db Instance of an ABCM Gateway to provide database access.
     * @param containerClass enclosing class
     */
    public ReferenceAuthorManager(AbstractABCMDBGateway db, AbstractDBManager containerClass){ super(db, containerClass); }
        
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
      ABCMReferenceAuthorsView view = DB.getReferenceAuthorsView();
      
      int orderCol  = rs.findColumn(view.getField_AuthorOrder().getName());
      int authorCol = rs.findColumn(view.getField_Researcher_LastNameFirstName().getName());
            
      while(rs.next()){
        xml.addLine("<author ORDER=\"" + rs.getString(orderCol) + "\">" + rs.getString(authorCol) + "</author>");
      }
    }
    
    /**
     * Sets the instance of this reference+author to be saved
     * @param referenceFK FK to the reference
     * @param authorOrder order of the author
     * @param researcherFK FK of the author (in researchers table)
     */
    public void setInstanceToSave(long referenceFK, int authorOrder, long researcherFK ){
      referenceFKToSave = referenceFK;
      authorOrderToSave = authorOrder;
      researcherFKToSave = researcherFK;    
    }
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
     */
    protected void clearInstanceToSave(){
      referenceFKToSave  = -1;
      authorOrderToSave  = -1;
      researcherFKToSave = -1;
    }
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
     */
    protected ABCMdbDatum datumToSave(){
      return ((instanceToSaveIsValid()) ?   
           DB.getReferencesAndAuthorsTable().getDatum("" + referenceFKToSave, "" + authorOrderToSave, "" + researcherFKToSave):  
           null);
    }
  
    /**
     * Determines if the instance to be saved has been set and if valid
     * @return the status of the instance to be saved
     */
    public boolean instanceToSaveIsValid(){
      return ((referenceFKToSave  != -1) &&
              (authorOrderToSave  != -1) &&
              (researcherFKToSave != -1));
    }
        
   // For manipulating multiple records
    
    /**
     * Removes a reference; deletes all authors
     * @param referenceFK PK of the reference to be deleted
     * @return TISFO
     */
    public boolean removeReference(long referenceFK){    return deleteElementsByPrimaryFK(referenceFK);    }
  
    
    /** 
     * Adds a set to an XML document. Overrides the super class because order matters
     * @param xml XMLAssist being used to construct the XML document
     * @param includeComments true if comments are to be included
     * @param referencePK PK of the reference being added
     */
    public void addSetToXML(XMLAssist xml, boolean includeComments, long referencePK){
      addSetToXML(xml, includeComments, DB.getReferenceAuthorsView().getField_AuthorOrder(), referencePK);    
    }
  
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
     */
    protected ABCMdbTable baseTable(){                     return DB.getReferencesAndAuthorsTable();                                    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
     */
    protected ABCMdbField baseTablePKField(){              return DB.getReferencesAndAuthorsTable().getField_PK();                      }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
     */
    protected ABCMdbField baseTablePrimaryFKField(){       return DB.getReferencesAndAuthorsTable().getField_ReferenceFK();             }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
     */
    protected ABCMdbField baseTableSecondaryFKField(){     return DB.getReferencesAndAuthorsTable().getField_ResearcherFK();            }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
     */
    protected ABCMdbAbstractViewable recordsView(){        return DB.getReferenceAuthorsView();                                         }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
     */
    protected ABCMdbField recordsViewPKField(){            return DB.getReferenceAuthorsView().getField_PK();                           }
        
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
     */
    protected ABCMdbField recordsViewPrimaryFKField(){     return DB.getReferenceAuthorsView().getField_Reference_PK();                 }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
     */
    protected ABCMdbField recordsViewPrimaryNameField(){   return null;                                                                 }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
     */
    protected ABCMdbField recordsViewSecondaryNameField(){ return DB.getReferenceAuthorsView().getField_Researcher_LastNameFirstName(); }
  
    
  } // End inner class
  
  /* Constructor */
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public ReferenceManager(AbstractABCMDBGateway db){     
    super(db);
    RAM = new ReferenceAuthorManager(db, this);
  }

  
  /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  public ABCMdbTable baseTable(){                  return DB.getReferencesTable();                                          }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  public ABCMdbField baseTablePKField(){           return DB.getReferencesTable().getField_PK();                            }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){         return DB.getReferencesTable().getField_Title();                         }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){     return DB.getReferencesTypesAndAddedByView();                            }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){         return DB.getReferencesTypesAndAddedByView().getField_Reference_PK();    }
  
  

  
  /* XML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){                return "references"; }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    
    ABCMReferencesTypesAndAddedByView view = DB.getReferencesTypesAndAddedByView();
    
    int IDFieldCol              = rs.findColumn(view.getField_Reference_PK().getName());
    int titleFieldCol           = rs.findColumn(view.getField_Reference_Title().getName());
    int dateFieldCol            = rs.findColumn(view.getField_Reference_Date().getName());
    int locFieldCol             = rs.findColumn(view.getField_Reference_Location().getName());
    int summaryFieldCol         = rs.findColumn(view.getField_Reference_Summary().getName());
    int abstractFieldCol        = rs.findColumn(view.getField_Reference_Abstract().getName());
    int workTypeFieldCol        = rs.findColumn(view.getField_ReferenceType().getName());
    int researcherNameFieldCol  = rs.findColumn(view.getField_Researcher_Name().getName());
    

    while(rs.next()){
      xml.open("reference", "NUMBER=\"" + rs.getString(IDFieldCol) + "\"");
        xml.addLine("title",                    rs.getString(titleFieldCol));
        xml.addLine("workdate",                     rs.getString(dateFieldCol));
  
        xml.addLine("typeOfWork",               rs.getString(workTypeFieldCol));
        xml.addLine("addedby",                  rs.getString(researcherNameFieldCol));
          
        if(singleInstance){
          String sum = rs.getString(summaryFieldCol);
          String abs = rs.getString(abstractFieldCol);
          if((sum!=null) && (sum.length() > 0))
            xml.addLine("summary", sum);
          if((abs!=null) && (abs.length() > 0))
            xml.addLine("abstract", abs);
        }
        
        if(rs.getString(locFieldCol)!= null)    
            xml.addLine("foundin", rs.getString(locFieldCol));
        
        // Add authors
        xml.open("authors");
          RAM.addSetToXML(xml, includeComments, rs.getLong(IDFieldCol));
        xml.closeTag();
        
        if(includeComments){
          CommentManager cm = new CommentManager(DB);
          xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.REFERENCE, rs.getLong(IDFieldCol)), XMLAssist.INDENT_BLOCK);
        }
      xml.closeTag();
    }
  }
  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName(){     return "xmlRefListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlRefToHTML.xsl";     }

  
  /**
   * Set the values to be saved as a new Reference in the HWM database.
   * @param title Title of referenct
   * @param dte date of reference (as a String; may be 'n.d.' or 'in press', etc.)
   * @param summary A summary of the work
   * @param abst The abstract for the work
   * @param workType A number identifying the kind of work, from the tblReferenceWorkTypes table
   * @param addedBy User ID of the person adding this reference
   * @param loc text indicating where this work is found, i.e. "American Antiquity 12(3):1-20"
   * @param auts collection of long integers identifying the researchers who are listed as authors;
   * order is preserved, so auts[0] is the first author.
   */
  public void setInstanceToSave(String title, String dte, String summary, String abst, 
      long workType, long addedBy, String loc, long[] auts){

    titleToSave       = AbstractABCMDBGateway.makeSQLSafe(title);
    dateToSave        = AbstractABCMDBGateway.makeSQLSafe(dte);
    summaryToSave     = AbstractABCMDBGateway.makeSQLSafe(summary);
    abstractToSave    = AbstractABCMDBGateway.makeSQLSafe(abst);
    typeOfWorkToSave  = workType;
    researcherAdding  = addedBy;
    locationToSave    = AbstractABCMDBGateway.makeSQLSafe(loc);
    authors           = new long[auts.length];
    for(int x = 0; x < auts.length; x++){      authors[x] = auts[x];     }    
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    titleToSave      = null;
    dateToSave       = null;
    summaryToSave    = null;
    abstractToSave   = null;
    typeOfWorkToSave = -1;
    researcherAdding = -1;
    locationToSave   = null;
    authors          = null;
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){ 
    return ((titleToSave      != null ) &&
            (dateToSave       != null ) &&
            (summaryToSave    != null ) &&
            (abstractToSave   != null ) &&
            (typeOfWorkToSave != -1   ) &&
            (researcherAdding != -1   ) &&
            (locationToSave   != null ) &&
            (authors          != null));
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){ 
    if(!instanceToSaveIsValid()) return null;
    return DB.getReferencesTable().getDatum(titleToSave, 
                                            dateToSave, 
                                            summaryToSave, 
                                            abstractToSave, 
                                            "" + typeOfWorkToSave, 
                                            "" + researcherAdding, 
                                            locationToSave);
  }
  
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#saveInstanceToDB() 
   */
  public boolean saveInstanceToDB(){
    long[] tempAuthors = authors;
    boolean successFlag = super.saveInstanceToDB();
    if(successFlag){
      long newPK = getLastSavedID();
      for(int i = 0; i < tempAuthors.length; i++){
        RAM.setInstanceToSave(newPK, i, tempAuthors[i]);
        successFlag = successFlag && RAM.saveInstanceToDB(); // Note short-circuit
      }
    }
    return successFlag;
  }
  
  /** 
   * This overrides the parent class because deleting a reference also
   * deletes all the authors associated with it
   * @param PK the primary key of the reference to be deleted 
   * @return true if successful, false otherwise
   */
  public boolean deleteItem(long PK){
    return RAM.deleteElementsByPrimaryFK(PK) && super.deleteItem(PK);
  }  
  
  /**
   * Gets a reference in a TeX format
   * @param refID the ID of the reference to retriev
   * @return the TeX format text
   */
  public String getReferenceAsTEX(long refID){
    StringWriter out = new StringWriter();
    try{
      Connection con = DB.getAdminConnection();
      Statement st1 = con.createStatement();
      
      ResultSet ref = st1.executeQuery(DB.getReferencesTable().getSelectSQL(DB.getOwner(), 
          AbstractABCMDBGateway.where(DB.getReferencesTable().getField_PK().getFullFieldName(DB.getOwner()), refID)));
      
      if(ref.next()){
        String title    = ref.getString(DB.getReferencesTable().getField_Title().getName());
        String location = ref.getString(DB.getReferencesTable().getField_Location().getName());
        String refDate  = ref.getString(DB.getReferencesTable().getField_ReferenceDate().getName());
        
        
        // Get the authors
        Statement st2 = con.createStatement();
        ResultSet authors = st2.executeQuery(DB.getReferencesAndAuthorsTable().getSelectSQL(DB.getOwner(), 
            AbstractABCMDBGateway.where(DB.getReferencesAndAuthorsTable().getField_ReferenceFK().getFullFieldName(DB.getOwner()), refID) + " " + 
            AbstractABCMDBGateway.orderBy(DB.getReferencesAndAuthorsTable().getField_AuthorOrder().getFullFieldName(DB.getOwner()))));
        
        boolean first = true;
        while(authors.next()){
          if(!first){
            if(authors.isLast()) out.write(" and ");
            else                 out.write(", ");    
          }
          first = false;
          String author = DB.getStringItem(DB.getResearchersTable(), DB.getResearchersTable().getField_PK(), 
              authors.getLong(DB.getReferencesAndAuthorsTable().getField_ResearcherFK().getName()), DB.getResearchersTable().getField_LastNameFirstName());
          out.write(author);
        }
        out.write(" (" + refDate + ") {\\em " + title + "} " + location + ".\n");
        
        authors.close();
        st2.close();
        
      }
      
      ref.close();
      st1.close();
      con.close();
    }
    catch(Exception E){
      return "";
    }
    return out.toString(); 
  }  
}
