package abcm.db.managers;

import java.sql.ResultSet;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.views.ABCMResearchersAndSessionsView;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.utilities.XMLAssist;

/**
 * Manages the entry and retrieval of Researcher Logins in the
 * HWM database.
 * 
 * Researcher Logins are never deleted.
 * 
 * Saving a researcher login automatically records a time as a session
 * start time; special methods are provided to set a session end time.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ResearcherLoginsManager extends AbstractDBManager {

  /** Researcher FK to save  */
  long researcherFKToSave;
  
  /* Constructor */
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public ResearcherLoginsManager(AbstractABCMDBGateway db){
    super(db);
  }

  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){  return DB.getResearchersAndSessionsView();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){      return DB.getResearchersAndSessionsView().getField_LoginRecord_PK();     }
    
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){             return "sessions";  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMResearchersAndSessionsView view = DB.getResearchersAndSessionsView();
    
    int researcherNameFieldCol = rs.findColumn(view.getField_Researcher_Name().getName());
    int sessionBeginFieldCol   = rs.findColumn(view.getField_SessionBegin().getName());

    while(rs.next()){
      xml.open("session");
      xml.addLine("researcher", rs.getString(researcherNameFieldCol));
      xml.addLine("date",       rs.getString(sessionBeginFieldCol));
      xml.closeTag();
    }      
  }

  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlResearcherLoginListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlResearcherLoginToHTML.xsl"; }
  
  
  /* Saving */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    researcherFKToSave = -1;
  } 
  
  /**
   * Sets the instance to be saved
   * @param researcherID
   */
  public void setInstanceToSave(long researcherID){
    researcherFKToSave = researcherID;  
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return (researcherFKToSave != -1);
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave() {
    String loginTime = DB.currentTimeString();  // retrieve once so the inserted values are identical
    return (instanceToSaveIsValid() ? 
        DB.getResearcherSessionsTable().getDatum("" + researcherFKToSave, loginTime, loginTime, null) : 
        null);  
  }
  
  /* Special */
  
  /**
   * Sets a session end time to be saved
   * @param sessionID the PK of the session to be recorded as having ended.
   * @return true if successful, false otherwise
   */
  public boolean setSessionEnd(long sessionID){
    return setSessionEnd(sessionID, DB.currentTimeString());
  }
  
  /**
   * Marks a session record in the database as having ended and records the time
   * @param sessionID PK of the session ending
   * @param endTime time at which the session ends
   * @return TISFO
   */
  public boolean setSessionEnd(long sessionID, String endTime){
    return DB.updateItem(DB.getResearcherSessionsTable(), 
                         DB.getResearcherSessionsTable().getField_PK(), sessionID, 
                         DB.getResearcherSessionsTable().getField_SessionEndTime(), endTime);
  }
  
}
