package abcm.db.managers;

import java.sql.ResultSet;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMResearchersTable;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.utilities.XMLAssist;

/**
 * Manages the entry and retrieval of Researchers in the
 * HWM database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ResearcherManager extends AbstractNonAuditableDBManager {

  /** Name of researcher to save  */
  private String nameToSave;
  /** Name of researcher to save in Last, First format  */
  private String lastFirstToSave;
  /** Description of research to save  */
  private String descriptionToSave;

  /* Constructor */

  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public ResearcherManager(AbstractABCMDBGateway db){   
    super(db);    
  }    
  
  /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  public ABCMdbTable baseTable(){                      return DB.getResearchersTable();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  public ABCMdbField baseTablePKField(){               return DB.getResearchersTable().getField_PK();                 }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){             return DB.getResearchersTable().getField_LastNameFirstName();  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField() 
   */
  public ABCMdbField baseTableDescriptionField(){      return DB.getResearchersTable().getField_Description();        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){         return DB.getResearchersTable();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){             return DB.getResearchersTable().getField_PK();                 }
  
  
  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){                return "researchers"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMResearchersTable view = (ABCMResearchersTable)recordsView();

    int PKFieldCol                 = rs.findColumn(view.getField_PK().getName());
    int researcherNameFieldCol     = rs.findColumn(view.getField_ResearcherName().getName());
    int lastNameFirstNameFieldCol  = rs.findColumn(view.getField_LastNameFirstName().getName());
    int descriptionFieldCol        = rs.findColumn(view.getField_Description().getName());
    
    while(rs.next()){
      xml.open("researcher", "NUMBER=\"" + rs.getString(PKFieldCol) + "\"");
        xml.addLine("name",                                rs.getString(researcherNameFieldCol));
        xml.addLine("officialName",                        rs.getString(lastNameFirstNameFieldCol));
        if(rs.getString(descriptionFieldCol)!=null)       
          xml.addLine("description",                       rs.getString(descriptionFieldCol));
      xml.closeTag();
    }
  }
  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlResearcherListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlResearcherToHTML.xsl"; }

  
  /* Saving */
  
  /**
   * Set the values to be saved as a new researcher
   * @param researcherName name to be saved
   * @param researcherLastNameFirstName name to be saved in last, first order
   * @param description description of the researcher
   */
  public void setInstanceToSave(String researcherName, String researcherLastNameFirstName, String description){
    nameToSave        = AbstractABCMDBGateway.makeSQLSafe(researcherName);
    lastFirstToSave   = AbstractABCMDBGateway.makeSQLSafe(researcherLastNameFirstName);
    descriptionToSave = AbstractABCMDBGateway.makeSQLSafe(description);
  }  
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){    
    return ((nameToSave != null)  && (lastFirstToSave != null));   
  }
    
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){   
    return (instanceToSaveIsValid() ? DB.getResearchersTable().getDatum(nameToSave, lastFirstToSave, descriptionToSave) : null);   
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    nameToSave        = null;
    lastFirstToSave   = null;
    descriptionToSave = null;
  }
    
}
