package abcm.db.managers;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Vector;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMSimulationDataTable;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.math.SimDate;
import abcm.sim.SimulationDataSeries;
import abcm.sim.SimulationDatum;
import abcm.utilities.XMLAssist;

/**
 * Database Manager for table in which Simulation Data (Data Series) are stored. 
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SimulationDataManager extends AbstractNonAuditableDBManager {

  /** Series FK to be saved  */        private long   seriesFKToSave;
  /** Simulation Date to be saved  */  private double simulationDateToSave;
  /** Numeric value to be saved  */    private double numericValueToSave;
  /** Text value to be saved  */       private String stringValueToSave;
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public SimulationDataManager(AbstractABCMDBGateway db){    super(db);     }
  
  /* General */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){                return DB.getSimulationDataTable();               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){         return DB.getSimulationDataTable().getField_PK(); }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField(){       return DB.getSimulationDataTable().getField_PK(); }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView(){   return DB.getSimulationDataTable();               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField(){       return DB.getSimulationDataTable().getField_PK(); }  
  
  
  /* XML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName() {          return "data";                                    }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
    ABCMSimulationDataTable instanceView = DB.getSimulationDataTable();    
    while(rs.next()){
      xml.open("datum");
        xml.addLine("date", SimDate.convertToString((rs.getDouble(instanceView.getField_SimulationDate().getName()))));
        xml.addLine("value", ((rs.getString(instanceView.getField_StringValue().getName()) == null) ||
                              (rs.getString(instanceView.getField_StringValue().getName()).length() == 0))  ? 
            Double.toString((rs.getDouble(instanceView.getField_NumericValue().getName()))) : 
            rs.getString(instanceView.getField_StringValue().getName()));
      xml.closeTag();
    }
  }
  
  /**
   * Adds a set of Simulation Data to an XML document
   * @param xml XMLAssist object constructing the XML document
   * @param seriesFK FK to the series being writted
   * @throws Exception
   */
  public void addSetToXML(XMLAssist xml, long seriesFK) throws Exception{
    ABCMSimulationDataTable instanceView = DB.getSimulationDataTable();
          
    String sql = instanceView.getSelectSQL(DB.getOwner(), 
        AbstractDBGateway.where(instanceView.getField_DataSeriesFK().getName(), seriesFK) + " " +  
        AbstractDBGateway.orderBy(instanceView.getField_SimulationDate().getName()));
    
    Connection con = DB.getAdminConnection();
    Statement st = con.createStatement();
    ResultSet rs2 = st.executeQuery(sql);
    
    addInstancesToXML(xml, rs2, false);
    
    rs2.close();
    st.close();
    con.close();
  }

  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return ""; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return ""; }
  
  /* Saving */
  
  /**
   * Sets the instance to be saved
   * @param seriesFK FK to the series to be saved
   * @param simulationDate Simulation Date to be saved
   * @param numericValue numeric value to be saved
   * @param stringValue text value to be saved
   */
  public void setDataToSave(long seriesFK, double simulationDate, double numericValue, String stringValue){
    seriesFKToSave       = seriesFK; 
    simulationDateToSave = simulationDate;
    numericValueToSave   = numericValue;
    stringValueToSave    = stringValue;      
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave() {
    if(seriesFKToSave == -1) return null;
    return DB.getSimulationDataTable().getDatum("" + seriesFKToSave, "" + simulationDateToSave, "" + numericValueToSave, stringValueToSave);
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    seriesFKToSave = -1;
    simulationDateToSave = -1;
    numericValueToSave = -1;
    stringValueToSave = "";
  }
    
  /**
   * Deletes all Simulation Data related to the specified series
   * @param seriesPK PK of the series to be deleted
   * @return TISFO
   */
  public boolean deleteAllSimulationDataRelatedToDataSeries(long seriesPK){
    return DB.deleteItem(DB.getSimulationDataTable(), DB.getSimulationDataTable().getField_DataSeriesFK(), seriesPK);
  }
  
  /**
   * Saves all data related to the given series
   * @param series Simulation Data Series to be saved- that is, the Series object that contains all the data
   * @param PK FK to the Simulation Data Series- that is, the Data Series record in the ABCM db
   * @return TISFO
   */
  public boolean saveDataSeriesData(SimulationDataSeries series, long PK){
    System.out.println("Saving series: " + series.getName());
    if(series == null) {
      errorReport = "Data series to be saved was null!";
      return false;
    }
    Vector V = series.getSeries();
    for(int i = 0; i < V.size(); i++){
      SimulationDatum sd = (SimulationDatum)V.get(i);
      System.out.println("SAVING SIM DATUM on date " + sd.getSimulationDate().getString() + " val = " + sd.getNumericValue());
      setDataToSave(PK, sd.getSimulationDate().value(), (series.isNumeric() ? sd.getNumericValue() : 0), (series.isNumeric() ? "" : sd.getTextValue()));
      if(!saveInstanceToDB()) return false; // Abandon ship
    }
    return true;
  }

  // TODO Deleted 'getOwner' and 'canBeEdited' that did not implement an interface and returned
  // dummy values; I believe these were not needed... check this.
  
}
