package abcm.db.managers;

import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMSimulationDataSeriesTable;
import abcm.db.components.tables.ABCMSimulationDataTable;
import abcm.db.managers.abstractparents.AbstractNonAuditableDBManager;
import abcm.math.SimDate;
import abcm.sim.SimulationDataSeries;
import abcm.utilities.XMLAssist;

/**
 * Provides an interface for storing and retrieving Simulation Data Series
 * from the HWM Database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SimulationDataSeriesManager extends AbstractNonAuditableDBManager{

  /** Value to indicate that a data series has not been initialized  */  
  private static final long DEFAULT_NON_INITIALIZED = -1;
    
  /** Series to be saved  */  
  private SimulationDataSeries seriesToSave;
  /** Name of series to be saved  */  
  private String               seriesNameToSave;
  /** Description of series to be saved  */  
  private String               seriesDescriptionToSave;
  /** ID of Run associated with this series to be saved  */  
  private long                 runIDToSave;
  

  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public SimulationDataSeriesManager(AbstractABCMDBGateway db){
    super(db);
    runIDToSave = DEFAULT_NON_INITIALIZED;
  }


  /* General */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){                 return DB.getSimulationDataSeriesTable();                            }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){          return DB.getSimulationDataSeriesTable().getField_PK();              }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField(){        return DB.getSimulationDataSeriesTable().getField_Name();            }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField() 
   */
  protected ABCMdbField baseTableDescriptionField(){ return DB.getSimulationDataSeriesTable().getField_SimulationRunFK(); }

  /* XML creation */
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){  return DB.getSimulationDataSeriesTable();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){      return DB.getSimulationDataSeriesTable().getField_PK();                 }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){             return "simulationDataSeries"; }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMSimulationDataSeriesTable listView = (ABCMSimulationDataSeriesTable)recordsView();
    
    int IDFieldCol           = rs.findColumn(listView.getField_PK().getName());
    int runNumberFieldCol    = rs.findColumn(listView.getField_SimulationRunFK().getName());
    int seriesNameFieldCol   = rs.findColumn(listView.getField_Name().getName());
    int descriptionFieldCol  = rs.findColumn(listView.getField_Description().getName());
    
    while(rs.next()){
      xml.open("series");
        xml.addLine("ID", rs.getLong(IDFieldCol));
        xml.addLine("RunNumber", rs.getLong(runNumberFieldCol));
        xml.addLine("seriesName", rs.getString(seriesNameFieldCol));
        if(!rs.getString(descriptionFieldCol).startsWith("null")) xml.addLine("desc", rs.getString(descriptionFieldCol));
        
        if(singleInstance){
          xml.open("data");
            SimulationDataManager sdm = new SimulationDataManager(DB);
            sdm.addSetToXML(xml, rs.getLong(listView.getField_PK().getName()));
          xml.closeTag();
        }
      xml.closeTag();    
    }    
  }
  
  /**
   * Adds a set of data series to an XML document
   * @param xml XMLAssist object that is constructing the document
   * @param simulationRunID ID of the simulation run
   * @param includeComments true if comments are to be included
   * @throws Exception
   */
  public void addSetToXML(XMLAssist xml, long simulationRunID, boolean includeComments) throws Exception{
    Connection con = DB.getAdminConnection();
    Statement st   = con.createStatement();
    
    ABCMSimulationDataSeriesTable simulationDataSeries = DB.getSimulationDataSeriesTable();
    
    String sqlDataSeries = simulationDataSeries.getSelectSQL(DB.getOwner(), 
        AbstractDBGateway.where(simulationDataSeries.getField_SimulationRunFK().getName(), simulationRunID)); 
    
    ResultSet rs = st.executeQuery(sqlDataSeries);
    
    addInstancesToXML(xml, rs, includeComments);
    
    rs.close();
    st.close();
    con.close();
  }
  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlSimulationDataSeriesListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlSimulationDataSeriesToHTML.xsl"; }


  /**
   * Sets the values to be saved as an instance of a Simulation Data Series.
   * @param series the Series that will be saved
   * @param name the name to be assigned this Series
   * @param runID the ID of the run from which this Series was created
   */
  public void setInstanceToSave(SimulationDataSeries series, String name, long runID){
    seriesToSave     = series;
    seriesNameToSave = name;
    runIDToSave      = runID;
  }

  /**
   * Set the description of the Series to be saved
   * @param seriesDescriptionToSave the description to be saved 
   */
  public void setSeriesDescriptionToSave(String seriesDescriptionToSave) {
    this.seriesDescriptionToSave = seriesDescriptionToSave;
  }

  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){
    return (instanceToSaveIsValid() ? 
        DB.getSimulationDataSeriesTable().getDatum("" + runIDToSave, seriesNameToSave,
            (seriesDescriptionToSave != null ? seriesDescriptionToSave : ""), (seriesToSave.isNumeric() ? "1" : "0")) :
        null);    
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    seriesToSave            = null;
    seriesNameToSave        = null;
    runIDToSave             = DEFAULT_NON_INITIALIZED;
    seriesDescriptionToSave = null;
  }
  
  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return ((seriesToSave     != null) &&
            (seriesNameToSave != null) &&
            (runIDToSave      != DEFAULT_NON_INITIALIZED));
  }
  
  /**
   * This is overridden from the parent class because it saves not only the entry into the
   * database for this series, but all the series data, too.
   * @return true if successful, false otherwise
   */
  public boolean saveInstanceToDB() {
    if(datumToSave() == null) return false; // Validates that there is a series to be saved
    
    // Sometimes an instance is re-saved to save the data again without reconstructing the
    // whole thing
    long PK = -1;
    SimulationDataSeries toBeSaved = seriesToSave; // The usual 'save' routine clears the 'seriesToBeSaved' variable; need a temp var
    if(super.saveInstanceToDB()) PK = super.getLastSavedID();
    else{
      try{
        String sql = DB.getSimulationDataSeriesTable().getSelectSQL(DB.getOwner(), 
            "WHERE " + DB.getSimulationDataSeriesTable().getField_SimulationRunFK().getName() + " = " + runIDToSave + " AND "
                     + DB.getSimulationDataSeriesTable().getField_Name().getName() + " = '" + seriesToSave.getName() + "'");

        Connection con = DB.getAdminConnection(); 
        Statement st = con.createStatement();
        ResultSet rs = st.executeQuery(sql);
        
        if(rs.next())     PK = rs.getLong(DB.getSimulationDataSeriesTable().getField_PK().getName());
        
        rs.close();
        st.close();
        con.close();
      }
      catch(Exception e){
        addToErrorReport("Error occurred retrieving PK for previously saved instance of SimulationDataSeries: " + e.getMessage()); 
      }       
    }
    if(PK != -1){ // That is, if a PK was recovered or created- either way...
      SimulationDataManager sdm = new SimulationDataManager(DB);
      
      // First, remove all previous data that may have been stored
      sdm.deleteAllSimulationDataRelatedToDataSeries(PK);
            
      // Now save the data series's elements
      return sdm.saveDataSeriesData(toBeSaved, PK);      
    }  
    else return false;
  }
  
 
  /**
   * Retrieve a data series from the HWM Database
   * @param PK the primary key of the series to be retrieved
   * @return the Simulation Data Series requested, if available (null if no
   * series with the specified PK exists)
   */
  public SimulationDataSeries getDataSeriesFromDB(long PK){
    String name = getName(PK);
    long runNumber = getRunNumber(PK);
    return getDataSeriesFromDB(runNumber, name);
  }
  
  
  /**
   * Gets the run number associated with a particular data series stored in the HWM
   * Database
   * @param PK the primary key of the data series
   * @return the run number of the simulation run that generated this series
   */
  public long getRunNumber(long PK){
    return DB.getNumericItem(DB.getSimulationDataSeriesTable(), DB.getSimulationDataSeriesTable().getField_PK(), PK, 
                               DB.getSimulationDataSeriesTable().getField_SimulationRunFK());
  }
  
  
  /**
   * Gets a simulation data series created from data already stored
   * in the HWM database.
   * @param runNumber
   * @param Name
   * @return a Simulation Data Series created from one already stored
   * in the HWM database
   */
  public SimulationDataSeries getDataSeriesFromDB(long runNumber, String Name){
    SimulationDataSeries sds = null;
    
    try{
      ABCMSimulationDataSeriesTable table      = DB.getSimulationDataSeriesTable();
      ABCMSimulationDataTable       dataTable  = DB.getSimulationDataTable();
      
      String sql = table.getSelectSQL(DB.getOwner(), 
          "WHERE " + table.getField_SimulationRunFK().getName() + "= " + runNumber + " AND "
                   + table.getField_Name().getName() + "= '" + Name + "'");
      
      Connection con = DB.getGuestConnection(); 
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      // If there is at least one record
      if(rs.next()){
        boolean isNumeric = rs.getBoolean(table.getField_IsNumeric().getName());
        long seriesPK     = rs.getLong(table.getField_PK().getName());
        
        sds = new SimulationDataSeries(runNumber + ":" + Name, isNumeric);    
        
        // Now fill in the data
        rs.close();
                
        sql = dataTable.getSelectSQL(DB.getOwner(), 
            "WHERE " + dataTable.getField_DataSeriesFK().getName() + " = " + seriesPK);
        
        rs = st.executeQuery(sql);
        while(rs.next()){
          if(isNumeric){
            sds.addElement(new SimDate(rs.getDouble(dataTable.getField_SimulationDate().getName())), 
                           rs.getDouble(dataTable.getField_NumericValue().getName()));
          }
          else {
            sds.addElement(new SimDate(rs.getDouble(dataTable.getField_SimulationDate().getName())), 
                           rs.getDouble(dataTable.getField_StringValue().getName()));            
          }
        }        
      }
      
      st.close();
      con.close();
      return sds;
    }
    catch(Exception e){
      addToErrorReport(e.getMessage());
    }
    return null;
  }  
  
  /** 
   * Overridden from the parent class because all simulation data is deleted
   * as well.
   * @param PK PK of the data series to be removed
   * @return TISFO 
   */
  public boolean deleteItem(long PK){
    SimulationDataManager sdm = new SimulationDataManager(DB);
    if(!sdm.deleteAllSimulationDataRelatedToDataSeries(PK)){
      addToErrorReport(sdm.getErrorReport());
      return false;
    }
    return super.deleteItem(PK);
  }
  
  /** 
   * Creates an HTML option list of Simulation Data Series.
   * Overridden from the parent class because the order must be by two fields, 
   * simulation run and data series name
   * @return the HTML option list 
   */
  public String getOptionList(){
    String optionList = "";
    try{
      Connection con = DB.getAdminConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(DB.getSimulationDataSeriesTable().getSelectSQL(DB.getOwner(), 
          "ORDER BY " + DB.getSimulationDataSeriesTable().getField_SimulationRunFK().getName() + ", " + 
                       DB.getSimulationDataSeriesTable().getField_Name().getName()));
      
      optionList = createOptionList(rs, DB.getSimulationDataSeriesTable().getField_PK(), DB.getSimulationDataSeriesTable().getField_Name(), 
          DB.getSimulationDataSeriesTable().getField_SimulationRunFK());
      
      rs.close();
      st.close();
      con.close();
      
    }
    catch(Exception E){
      addToErrorReport("Could not get option list of simulation data series: " + E.getMessage());
    }
    return optionList;
  }
  
  /**
   * Creates an HTML option list of Simulation Data Series
   * @param rs result set containing the records to be listed
   * @param IDField field in which the ID will be recovered
   * @param nameField field in which the name will be recovered
   * @param runNumberField field in which the run number will be recovered
   * @return HTML option list
   * @throws Exception
   */
  public String createOptionList(ResultSet rs, ABCMdbField IDField, ABCMdbField nameField, ABCMdbField runNumberField) throws Exception{
    StringWriter ret = new StringWriter();
    while(rs.next()){
      ret.write("<OPTION Value=\"" + rs.getLong(IDField.getName()) + "\">Run " + rs.getString(runNumberField.getName()) + ": " + rs.getString(nameField.getName()) + "</OPTION>");
    }
    return ret.toString();    
  }

  public SimulationDataSeries getDataSeriesFromNarrative(String narrativeName, String command, int commandColumn){
    NarrativeManager NM = new NarrativeManager(this.DB);
    
    return null;
  }
  
}
