package abcm.db.managers;


import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Date;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMSimulationDetailsView;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.sim.Simulation;
import abcm.utilities.XMLAssist;


/**
 * Manages the storage and retrieval of stored Simulations in the
 * HWM Database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SimulationManager extends AbstractAuditableDBManager {

/** Name of Simulation to be Saved  */ private String nameToSave;
/** Use ID to be saved  */             private long   userIDToSave;
/** Code Model ID to be saved  */      private long   codeModelIDToSave;
/** Data Model ID to be saved  */      private long   dataModelIDToSave;
/** Description to be saved  */        private String descToSave;

  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public SimulationManager(AbstractABCMDBGateway db){
    super(db);
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){          return DB.getSimulationsTable();               }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){   return DB.getSimulationsTable().getField_PK(); }
  
  
  /* Auditable */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField baseTableOwnerField(){         return DB.getSimulationsTable().getField_ResearcherFK();              }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView(){  return DB.getCanBeEdited_SimulationsView();                           }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField canBeEditedViewPKField(){      return DB.getCanBeEdited_SimulationsView().getField_Simulation_PK();  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){      return DB.getSimulationDetailsView();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){          return DB.getSimulationDetailsView().getField_Simulation_PK();      }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){          return DB.getSimulationDetailsView().getField_Simulation_Title();   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField() 
   */
  public ABCMdbField baseTableDescriptionField(){   return DB.getSimulationDetailsView().getField_Simulation_Desc();    }
  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){                return "simulations"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    
    ABCMSimulationDetailsView view = DB.getSimulationDetailsView();
    
    int simulationPKFieldCol      = rs.findColumn(view.getField_Simulation_PK().getName());
    int simulationTitleFieldCol   = rs.findColumn(view.getField_Simulation_Title().getName());
    int canBeEditedFieldCol       = rs.findColumn(view.getField_CanBeEdited().getName());
    int dateCreatedFieldCol       = rs.findColumn(view.getField_Simulation_DateCreated().getName());
    int researcherNameFieldCol    = rs.findColumn(view.getField_Researcher_Name().getName());
    int codeModelPKFieldCol       = rs.findColumn(view.getField_CodeModel_PK().getName());
    int codeModelTitleFieldCol    = rs.findColumn(view.getField_CodeModel_Title().getName());
    int dataModelPKFieldCol       = rs.findColumn(view.getField_DataModel_PK().getName());
    int dataModelTitleFieldCol    = rs.findColumn(view.getField_DataModel_Title().getName());
    int simulationDescFieldCol    = rs.findColumn(view.getField_Simulation_Desc().getName());
    int truthValueFieldCol        = rs.findColumn(view.getField_TruthValue().getName());
    int truthValuePKFieldCol      = rs.findColumn(view.getField_TruthValue_PK().getName());
    
    while(rs.next()){
      xml.open("simulation");
        xml.addLine("ID",             rs.getLong(simulationPKFieldCol));
        xml.addLine("title",          rs.getString(simulationTitleFieldCol));
        xml.addLine("canBeEdited",    ((rs.getString(canBeEditedFieldCol) != null) ? "Yes" : "No"));
        xml.addLine("dateCreated",    rs.getString(dateCreatedFieldCol));
        xml.addLine("createdBy",      rs.getString(researcherNameFieldCol));
        xml.addLine("codeModelID",    rs.getLong(codeModelPKFieldCol));
        xml.addLine("codeModelTitle", rs.getString(codeModelTitleFieldCol));
        xml.addLine("dataModelID",    rs.getLong(dataModelPKFieldCol));
        xml.addLine("dataModelName",  rs.getString(dataModelTitleFieldCol));
        xml.addLine("description",    rs.getString(simulationDescFieldCol));
        xml.addLine("truthValue",     rs.getString(truthValueFieldCol));
      xml.closeTag();
    }
  }

  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlSimulationListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlSimulationToHTML.xsl"; }
  
  
  /**
   * Sets the values to be saved as an instance of the Simulation
   * @param name the name to be assigned to this Simulation
   * @param userID the user ID of the user who is saving this Simulation
   * @param codeModelID the primary key of the Code Model used in this Simulation
   * @param dataModelID the primary key of the Data Model used in this Simulation
   * @param description a description of the Simulation
   */
  public void setInstanceToSave(String name, long userID, long codeModelID, long dataModelID, String description){
    nameToSave = name;
    userIDToSave = userID;
    codeModelIDToSave = codeModelID;
    dataModelIDToSave = dataModelID;
    descToSave = description;
  }
  

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){
    return DB.getSimulationsTable().getDatum(nameToSave, "" + codeModelIDToSave, "" + dataModelIDToSave, 
        DB.currentTimeString(), "" + userIDToSave, descToSave);    
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    nameToSave = null;
    userIDToSave = -1;
    codeModelIDToSave = -1;
    dataModelIDToSave = -1;
    descToSave = null;
  }

  
  /**
   * Gets a specific Simulation from the HWM Database.
   * @param SimulationPK primary key of the requested Simulation.
   * @return the requested Simulation object
   */
  public Simulation getSimulation(long SimulationPK){
    String sql = "";
    Simulation s = null;
    try{
      sql = DB.getSimulationsAndResearchersView().getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(DB.getSimulationsAndResearchersView().getField_Simulation_PK().getName(), SimulationPK));
      
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      
      ResultSet rs = st.executeQuery(sql);
      
      if(rs.next()){
        String title = rs.getString(DB.getSimulationsAndResearchersView().getField_Simulation_Title().getName());
        long codeMod = rs.getLong(DB.getSimulationsAndResearchersView().getField_CodeModel_PK().getName());
        long dataMod = rs.getLong(DB.getSimulationsAndResearchersView().getField_DataModel_PK().getName());
        Date d       = rs.getDate(DB.getSimulationsAndResearchersView().getField_Simulation_DateCreated().getName());
        String by    = rs.getString(DB.getSimulationsAndResearchersView().getField_Researcher_Name().getName());
        
        s = new Simulation(title, d, by, codeMod, dataMod);
        
        rs.close();
        
        st.close();
        con.close();
        
      }
      
    }
    catch(Exception e){
      errorReport +="Error while loading simulation " + SimulationPK + ": " + e.getMessage();
    }
    
    return s;
  }

}  
  
