package abcm.db.managers;

import java.sql.Connection;
import java.sql.Date;
import java.sql.ResultSet;
import java.sql.Statement;
import java.text.SimpleDateFormat;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.views.ABCMSimulationRunsAndResearchersView;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.sim.SimulationRun;
import abcm.utilities.XMLAssist;

/**
 * Manages the creation and saving of simulation run numbers in the database.
 * 
 * Special functions: the creation of new simulation Run IDs with or without
 * random number seeds specified.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SimulationRunManager extends AbstractDBManager{

  /**   */  private long   simulationFKToSave; 
  /**   */  private long   parameterSetFKToSave; 
  /**   */  private long   probeSetFKToSave; 
  /**   */  private long   runModeFKToSave;
  /**   */  private long   researcherFKToSave;
  /**   */  private String notesToSave;
  /**   */  private long   randomNumberSeedToSave;
  /**   */  private Date   dateCreatedToSave;
  
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public SimulationRunManager(AbstractABCMDBGateway db){   super(db);     }

  
  /* XML creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){    return DB.getSimulationRunsAndResearchersView();                               }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){        return DB.getSimulationRunsAndResearchersView().getField_SimulationRun_PK();   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  public String xmlOuterNodeName(){               return "simulationRuns"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMSimulationRunsAndResearchersView view = DB.getSimulationRunsAndResearchersView();
    
    int simulationRunPKFieldCol    = rs.findColumn(view.getField_SimulationRun_PK().getName());
    int dateTimeBegunFieldCol      = rs.findColumn(view.getField_SimulationRun_DateTimeBegun().getName());
    int simulationPKFieldCol       = rs.findColumn(view.getField_Simulation_PK().getName());
    int simulationTitleFieldCol    = rs.findColumn(view.getField_Simulation_Title().getName());
    int paramSetPKFieldCol         = rs.findColumn(view.getField_ParameterSet_PK().getName());
    int paramSetNameFieldCol       = rs.findColumn(view.getField_ParameterSet_Name().getName());
    int probeSetPKFieldCol         = rs.findColumn(view.getField_ProbeSet_PK().getName());
    int probeSetNameFieldCol       = rs.findColumn(view.getField_ProbeSet_Name().getName());
    int runModePKFieldCol          = rs.findColumn(view.getField_SimulationRunMode_PK().getName());
    int runModeTypeFieldCol        = rs.findColumn(view.getField_SimulationRunMode_Type().getName());
    int researcherNameFieldCol     = rs.findColumn(view.getField_Researcher_Name().getName());
    int simulationRunNotesFieldCol = rs.findColumn(view.getField_SimulationRun_Notes().getName());
    int randomNumberSeedFieldCol   = rs.findColumn(view.getField_SimulationRun_RandomNumberSeed().getName());
    int simulationRunReRunFieldCol = rs.findColumn(view.getField_SimulationRun_ReRun_FK().getName());
    int truthValueFieldCol         = rs.findColumn(view.getField_TruthValue().getName());
    int combinedTruthValueFieldCol = rs.findColumn(view.getField_CombinedTruthValue().getName());
    
    while(rs.next()){
      xml.open("simulationRun");
        xml.addLine("ID",             rs.getLong(simulationRunPKFieldCol));
        xml.addLine("timeRun",        rs.getString(dateTimeBegunFieldCol));
        xml.addLine("simulationID",   rs.getLong(simulationPKFieldCol));
        xml.addLine("simulation",     rs.getString(simulationTitleFieldCol));
        xml.addLine("parameterSetID", rs.getLong(paramSetPKFieldCol));
        xml.addLine("parameterSet",   rs.getString(paramSetNameFieldCol));
        xml.addLine("probeSetID",     rs.getLong(probeSetPKFieldCol));
        xml.addLine("probeSet",       rs.getString(probeSetNameFieldCol));
        xml.addLine("mode",           rs.getString(runModeTypeFieldCol));
        xml.addLine("runBy",          rs.getString(researcherNameFieldCol));
        xml.addLine("notes",          rs.getString(simulationRunNotesFieldCol));
        xml.addLine("RNS",            rs.getLong(randomNumberSeedFieldCol));
        if(rs.getString(simulationRunReRunFieldCol)!= null) xml.addLine("ReRun", rs.getString(simulationRunReRunFieldCol));
        if(rs.getString(truthValueFieldCol)!= null) xml.addLine("truthValue", rs.getString(truthValueFieldCol));
        if(singleInstance){
          SimulationDataSeriesManager sdsm = new SimulationDataSeriesManager(DB);
          sdsm.addSetToXML(xml, instancePK, includeComments);
        }
      xml.closeTag();
    }
    
    
  }
  
  
  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlSimulationRunListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlSimulationRunToHTML.xsl"; }
  
  
  /* Saving */
  
  /**
   * Sets the instance of the Simulation Run to be saved
   * @param simFK FK to the Simulation used
   * @param paramSetFK FK to the Parameter Set used
   * @param probeSetFK FK to the Probe set used
   * @param runModeFK FK to the Run Mode in which this Simulation Run was executed
   * @param researcherFK FK to the Researcher who executed this Simulation
   * @param notes notes about this Simulation Run
   * @param randomNumberSeed random number seed
   * @param dateCreated date and time the run was performed
   */
  public void setInstanceToSave(long simFK, 
                                long paramSetFK, 
                                long probeSetFK, 
                                long runModeFK, 
                                long researcherFK, 
                                String notes, 
                                long randomNumberSeed,
                                Date dateCreated){
    simulationFKToSave     = simFK; 
    parameterSetFKToSave   = paramSetFK; 
    probeSetFKToSave       = probeSetFK;  
    runModeFKToSave        = runModeFK;
    researcherFKToSave     = researcherFK;
    notesToSave            = AbstractDBGateway.makeSQLSafe(notes);
    randomNumberSeedToSave = randomNumberSeed;
    dateCreatedToSave      = dateCreated;
    
  }
    
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave() {
    simulationFKToSave     = -1; 
    parameterSetFKToSave   = -1; 
    probeSetFKToSave       = -1;  
    runModeFKToSave        = -1;
    researcherFKToSave     = -1;
    notesToSave            = null;
    randomNumberSeedToSave = -1;
    dateCreatedToSave      = null;
  }

  /**
   * Determines if the instance to be saved is valid
   * @return true if the instance to be saved has been set, false otherwise
   */
  public boolean instanceToSaveIsValid(){
    return ((simulationFKToSave     != -1  ) && 
            (parameterSetFKToSave   != -1  ) &&
            (probeSetFKToSave       != -1  ) &&
            (runModeFKToSave        != -1  ) &&
            (researcherFKToSave     != -1  ) &&
            (notesToSave            != null) &&
            (randomNumberSeedToSave != -1  ) &&
            (dateCreatedToSave      != null));
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave() {                             
    return (instanceToSaveIsValid() ? 
//             DB.getSimulationRunsTable().getDatum((new SimpleDateFormat("MM/dd/yyyy HH:mm:ss")).format(dateCreatedToSave),
    		DB.getSimulationRunsTable().getDatum((new SimpleDateFormat(DB.getSQLSyntaxEngine().getDateTimeInsertFormat())).format(dateCreatedToSave),
                                                "" + simulationFKToSave, 
                                                "" + parameterSetFKToSave, 
                                                "" + probeSetFKToSave, 
                                                "" + runModeFKToSave, 
                                                "" + researcherFKToSave, 
                                                notesToSave, 
                                                "" + randomNumberSeedToSave) :
             null);
  }

 
  /* Special */
  
  /**
   * Creates a new Simulation Run with no Random number specified
   * @param userID the ID of the person creating this SimulationRun
   * @param simID the primary key of the Simulation being used in this Simulation Run
   * @param ParamSetID the primary key of the Parameter Set being used in the Simulation Run
   * @param ProbeSetID the primary key of the Probe Set being used in the Simulation Run
   * @param mode identifies the mode in which this Simulation Run is created
   * @param notes any notes associated with this Simulation Run
   * @return a new Simulation Run
   */
  public SimulationRun createNewSimulationRun(long userID, 
                                              long simID, 
                                              long ParamSetID, 
                                              long ProbeSetID, 
                                              int mode, 
                                              String notes ){
    return createNewSimulationRun(userID, simID, ParamSetID, ProbeSetID, mode, notes, (long)Math.random());
  }
  
  /**
   * Creates a new Simulation Run
   * @param userID the ID of the person creating this SimulationRun
   * @param simID the primary key of the Simulation being used in this Simulation Run
   * @param ParamSetID the primary key of the Parameter Set being used in the Simulation Run
   * @param ProbeSetID the primary key of the Probe Set being used in the Simulation Run
   * @param mode identifies the mode in which this Simulation Run is created
   * @param notes any notes associated with this Simulation Run
   * @param RNS a random number seed
   * @return a new Simulation Run
   */
  public SimulationRun createNewSimulationRun(long userID, 
      long simID, long ParamSetID, long ProbeSetID, int mode, String notes, long RNS){
    
    SimulationRun sr = null;
    Date d = new Date(System.currentTimeMillis());
    setInstanceToSave(simID, ParamSetID, ProbeSetID, mode, userID, notes, RNS, d);
    if(saveInstanceToDB()){
      sr = new SimulationRun(lastSavedID, userID, simID, ParamSetID, ProbeSetID, notes, mode, d, RNS);
    }
    return sr;
  }
  
  /**
   * Marks in the HWM Database whether a specified Simulation Run is
   * reproducible or not.
   * @param runIsReproducible the status of the run
   * @param simRunID the primary key of the run
   */
  public void setRunIsReproducible(boolean runIsReproducible, long simRunID){
    String setVal = "0";
    if(runIsReproducible) setVal = "1";
    try{
      Connection con = DB.getGuestConnection(); 
      Statement st = con.createStatement();
      
      String sql = "" +
          "UPDATE " + DB.getSimulationRunsTable().getFullName(DB.getOwner()) + " " + 
          "SET "    + DB.getSimulationRunsTable().getField_IsReproducible().getName() + " = " + setVal + " " +
          "WHERE "  + DB.getSimulationRunsTable().getField_SimulationFK().getName() + " = " + simRunID;
      
      st.executeUpdate(sql);
      
      st.close();
      con.close();
    }
    catch(Exception e){
      addToErrorReport("An error occurred while setting run is reproducible. " + e.getMessage());
    }    
  }    

}
