package abcm.db.managers.abstractparents;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.Auditable;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbField;

/**
 * The <code>AbstractAuditableDBManager</code> class provides an extension to the
 * AbstractDBManager class in such a way that a manager can easily manage objects
 * subject to the audit rules of the ABCM environment. Child classes must provide
 * a 'canBeEdited(long PK)' method, which returns true if the record can be edited
 * and false if not. In general, this will be achieved by calling this class's
 * 'canBeEdited(Viewable, Field, PK) method, which is provided for convenience.
 * Similarly, child classes must provide a 'getOwner(long PK)' method, which is
 * usually crafted from the getOwner(Viewable, Field, PK) stub provided here.
 * 
 * Using both of these, this class enforces the rules about auditability:
 * deletions can be made only made if the object can be edited, and by the object's
 * owner. This class provides a simple 'isOwner(long PK, userID)' method that compares the
 * ID to the results of the getOwner(PK) operation.
 * 
 * Most importantly, the 'deleteItem(Table, Field, PK)' method incorporates the audit rules; when
 * deleteItem(PK) invokes this the audit rule is automatically enforced.
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractAuditableDBManager extends AbstractDBStandardizedElementManager implements Auditable {

  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public AbstractAuditableDBManager(AbstractABCMDBGateway db){
    super(db);
  }
  
  /**
   * Gets the field in the base table that indicates the owner of the item
   * @return the base table's 'owner' field
   */
  public abstract ABCMdbField             baseTableOwnerField();
  
  /**
   * Gets the view that indicates whether a record can be edited
   * @return the view that indicates whether a record can be edited
   */
  public abstract ABCMdbAbstractViewable  canBeEditedView();
  
  /**
   * Gets the field of the view that indicates whether a record can be modified 
   * @return the field that indicates whether an item can be edited
   */
  public abstract ABCMdbField             canBeEditedViewPKField();
  
  /** 
   * @see abcm.db.Auditable#canBeEdited(long) 
   */
  public final boolean canBeEdited(long PK){
    return DB.entryExists(canBeEditedView(), canBeEditedViewPKField(), PK);
  }  
  
  /** 
   * @see abcm.db.Auditable#getOwner(long) 
   */
  public final long getOwner(long PK){
    return DB.getNumericItem(baseTable(), baseTablePKField(), PK, baseTableOwnerField());
  }
  
  /** 
   * @see abcm.db.Auditable#isOwner(long, long) 
   */
  public boolean isOwner(long PK, long userID) {    return (userID == getOwner(PK));     }
  
  /**
   * Deletes the specified item
   * @param PK PK of record to be deleted
   * @param user user attempting the delete
   * @return TISFO
   */
  public boolean deleteItem(long PK, long user){
    return permitModification(PK, user) && 
           DB.deleteItem(baseTable(), baseTablePKField(), PK);
  }  
  
  /**
   * Determine if this object can be modified
   * @param PK PK of object
   * @param user PK of user attempting modification
   * @return true if the modification is permissible, false otherwise
   */
  public boolean permitModification(long PK, long user){
    if(!canBeEdited(PK)) {
      addToErrorReport("Item cannot be modified: it cannot be edited due to auditing rules.");
      return false;
    }
    if(!isOwner(PK, user)){
      addToErrorReport("Item cannot be modified: specified user is not owner.");
      return false;
    }
    return true;
  }

}
