package abcm.db.managers.abstractparents;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.utilities.XMLAssist;

/**
 * A 'Link Table' links two related tables. One of these is 'dominant',
 * however. For example, if a Narrative may belong to a History, and this
 * will be represented by an entry in the 'HistoryNarratives' table.
 * If (to make an example from something that is actually not permitted by
 * the other audit rules in the db) the History is deleted, the entry in 
 * the 'HistoryNarratives' table would also be deleted- in fact, all such entries for that History 
 * would be deleted. This would be perfectly reasonable and would not
 * be any loss to the database: the Narrative would be untouched, only a little
 * less used. But deleting the Narrative and removing it from the database-
 * by, for example, deleting all the entries in the 'HistoryNarratives' table
 * that refer to it- would be a profound loss: all of the histories that
 * relied on it would now be changed. Hence the 'History' is dominant,
 * in that what matters is protecting the History. The History
 * 'owns' the Narratives that it links. This situation is common enough
 * that some simple structures are provided to make managing this kind of table easier. 
 * 
 * @author John T. Murphy
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractDBLinkTableManager extends AbstractDBManager {
  
  /* Constructor */
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public AbstractDBLinkTableManager(AbstractABCMDBGateway db){ super(db); }
  
  /* Base Definitions */
  
  // Base table
  /**
   * Gets the Base Table. The Base Table is the table being managed- the linking table
   * @return the Base Table for this manager
   */
  protected abstract ABCMdbTable             baseTable();
  
  /**
   * Gets the primary key for the Base Table
   * @return the base table's primary key field
   */
  protected abstract ABCMdbField             baseTablePKField();
  
  /**
   * Gets the field in the Base Table that contains the foreign key
   * to the primary table.
   * @return the foreign key field linking this table to the primary table
   */
  protected abstract ABCMdbField             baseTablePrimaryFKField();
  
  /**
   * Gets the field in the Base Table that contains the foreign key to the
   * secondary table 
   * @return the foreign key field linking this table to the secondary table
   */
  protected abstract ABCMdbField             baseTableSecondaryFKField();
      
  // Records view  
  /**
   * Gets the primary FK in the records view
   * @return the primary FK in the records view
   */
  protected abstract ABCMdbField             recordsViewPrimaryFKField();
  
  /**
   * Gets the primary name field in the records view
   * @return the primary name field in the records view
   */
  protected abstract ABCMdbField             recordsViewPrimaryNameField();
  
  /**
   * Gets the secondary name field in the records view
   * @return the secondary name field in the records view
   */
  protected abstract ABCMdbField             recordsViewSecondaryNameField();

  
  /* Ordered links */
  // Some links have an order associated with them; that is,
  // the parent record 'owns' several child records, but
  // wants them to be thought of as existing in a particular
  // order. If this is the case, the 'order' field will be specified
  // here. All transactions that add or remove records will
  // also update the order of the child records.
  // Implementing classes need only override this method so that
  // it does not return null and all necessary updates to
  // order fields will be undertaken automatically.
  // Note that if 'ordered' is assumed, the assumption that
  // the links will be unique is put aside- that is,
  // for non-ordered links, it is assumed that you may not
  // link the same two records twice, but for ordered links
  // it is assumed that the same link can occur twice,
  // differentiated by the order. This difference is seen in
  // both getters and the option lists returned.
  
  /**
   * For ordered linked tables (that is, tables in which there is
   * not only a link between table A and table B but an ordered
   * link for which it is assumed that if A has N links to B
   * then they are to be maintained as numbered 1, 2, 3... N),
   * a field will be specified here that serves as the numbering
   * field for the links. If this is null (which it is by default)
   * the link is assumed to be unordered.
   * @return the name of the order field for the links between the
   * parent and child tables
   */
  protected ABCMdbField orderField(){            return null; }
  
  
  /**
   * The order field (see orderField() method) in the table will
   * have a corresponding field in the records View.
   * @return the order field in the records view, if any (null otherwise)
   */
  protected ABCMdbField recordsViewOrderField(){ return null; }
  
  // Implementing classes should also override these two methods to provide instruction
  // when saving new records; the dominant PK is the PK of the main
  // value, the new order to save is the new order.

  /**
   * Implementing classes should override this method so that it returns
   * the PK to be saved as the FK to the dominant table. This allows
   * order management to be performed automatically. The value to be
   * returned should be set before a 'save' is executed.
   * @return the PK of the record to be linked in the dominant table
   */
  protected long getOrderedDominantPKToSave(){ return -1;             }
  
  /**
   * Implementing classes should override this method so that it returns
   * the order value to be saved when a new link is established. 
   * @return the new order value to be saved
   */
  protected long getOrderedNewOrderToSave(){   return Long.MAX_VALUE; }


  /* Getters */
  
  /**
   * Gets the primary key in the Base Table based on the two FKs from the
   * linked tables. Note: Assumes this combination will be unique; otherwise
   * returns an arbitrarily selected matching instance.
   * @param primaryFK FK value from the primary table
   * @param secondaryFK FK value from the secondary table
   * @return the PK of the matching entry in the Base Table. Returns -1 if not found.
   */
  public long getPK(long primaryFK, long secondaryFK){
    long PK = -1;
    try{
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      
      String sql = baseTable().getSelectSQL(DB.getOwner(), "WHERE " + baseTablePrimaryFKField().getName()   + " = " + primaryFK + " AND " + 
                                                                      baseTableSecondaryFKField().getName() + " = " + secondaryFK);
      
      ResultSet rs = st.executeQuery(sql);
      
      if(rs.next()) PK = rs.getLong(recordsViewPKField().getName());
    
      rs.close();
      st.close();
      con.close();
      
    }
    catch(Exception e){
      addToErrorReport("Could not get primary key: " + e.getMessage());
    }
    return PK;
  }
  
  /**
   * Gets the record PK given the dominant PK and the order number
   * of the entry requested. Returns -1 if 'order' is not used by the
   * implementing class.
   * @param dominantPK FK of parent record
   * @param order order number of record sought
   * @return record PK of entry specified
   */
  public long getPKFromDominantPKAndOrder(long dominantPK, long order){
    return DB.getNumericItem(
        baseTable().getSelectSQL(DB.getOwner(), 
            " WHERE " + this.baseTablePrimaryFKField().getName() + " =" + dominantPK + " AND " +
                        this.orderField().getName()              + " =" + order), this.baseTablePKField().getName());
  }
  
  /* Record Manipulation */

  /**
   * Deletes an item from the Base Table.
   * Note that if ordered items are being used, the order
   * numbers are collapsed to ensure that they are contiguous
   * after the record is removed.
   * @param recordPK primary key of the entry to be deleted
   * @return true if successful, false otherwise
   */
  public boolean deleteItem(long recordPK){
    if(orderField() == null) return DB.deleteItem(baseTable(), baseTablePKField(), recordPK);
    
    // Look up primary key of dominant table
    long dominantPK = DB.getNumericItem(baseTable(), baseTablePKField(), recordPK, baseTablePrimaryFKField());
    if(dominantPK == -1){
      addToErrorReport("Could  not detect PK of dominant field to delete.");
      return false;
    }
    
    // Delete the requested record
    boolean successfulDelete = DB.deleteItem(baseTable(), baseTablePKField(), recordPK);
    
    if(!successfulDelete){
      addToErrorReport("Delete in link table manager was unsuccessful: " + DB.getErrorReport());
      return false;
    }
    
    // Collapse and renumber
    return collapseOrderNumbers(dominantPK);
  }
   
  
  
  
  /**
   * Saves an instance to the database.
   * If order is being used, the order numbers are expanded to make
   * room for the new entry, then re-collapsed after the entry is
   * made, to ensure that they are contiguous.
   * @see abcm.db.managers.abstractparents.AbstractDBManager#saveInstanceToDB()
   */
  public boolean saveInstanceToDB(){
    // If no order, just use the superclass's method
    if(orderField() == null) return super.saveInstanceToDB();
    
    // Otherwise:
    long dominantPK = getOrderedDominantPKToSave();
    if(!this.incrementOrders(dominantPK, getOrderedNewOrderToSave())){
      addToErrorReport("Error when attempting to save ordered value; could not increment orders. ");
      return false;
    }
    if(!super.saveInstanceToDB()){
      addToErrorReport("Error when attempting to save ordered value: could not be saved.");
      return false;
    }
    return collapseOrderNumbers(dominantPK);
  }
    
  /**
   * Deletes all elements deriving from a single FK to the primary
   * table. It is equivalent to deleting a history and all of its narratives
   * at once. Order is not an issue because all entries are removed.
   * @param primaryFK FK to the primary table to entries to be deleted
   * @return true if successful, false otherwise
   */
  public boolean deleteElementsByPrimaryFK(long primaryFK){
    return DB.deleteItem(baseTable(), baseTablePrimaryFKField(), primaryFK);
  }
  
  /**
   * Deletes the entries that match both FKs. Often this will be exactly one
   * element (but it needn't be; if order is being used, ALL entries will
   * be deleted.).
   * @param primaryFK FK to the primary table
   * @param secondaryFK FK to the secondary table
   * @return TISFO
   */
  public boolean deleteElementsByBothPrimaryKeys(long primaryFK, long secondaryFK){    
    boolean successFlag = false;
    try {
      Connection con = DB.getGuestConnection();
      Statement st   = con.createStatement();
      
      String sql = "DELETE " + 
                   "FROM "  + baseTable().getFullName(DB.getOwner()) + " " +
                   "WHERE " + baseTablePrimaryFKField().getName() + " = " + primaryFK + " AND " +
                              baseTableSecondaryFKField().getName() + " = " + secondaryFK;
      st.execute(sql);
      successFlag = true;
      
      if(orderField() != null) collapseOrderNumbers(primaryFK);
      
      st.close();
      con.close();
      
    } 
    catch (Exception e) {
      addToErrorReport("Error when removing elements from linked table: " + e.getMessage());
    }
    return successFlag;      
  }
  
  
  /**
   * Gets the option list of all elements in records() that have
   * the specified value for the 'primary' FK
   * @param primaryFK the required primary FK
   * @return the HTML option list
   */
  public String getOptionList(long primaryFK){   
    return getOptionList(primaryFK, true);
  }

  /**
   * Gets the option list of all elements in records() that have the specified value
   * for the 'primary' FK; optionally uses the name field from the primary table.
   * If this is an unordered linked table, the value used as the index will be the
   * FK to the secondary table- commonly this will be a unique value. If it is an ordered
   * linked table (recordsViewOrderField() is specified), the value will be the order index.
   * @param primaryFK FK to be extracted
   * @param showPrimaryName true if the name field from the primary table should be used in the option list
   * @return the HTML option list
   */
  public String getOptionList(long primaryFK, boolean showPrimaryName){
    return DB.getOptionList(
        // Table or view from which view is drawn
        recordsView(),                   
        // Field from which 'value' is to be drawn
        (recordsViewOrderField() == null ? recordsViewPKField() : recordsViewOrderField()),
        // Collection of fields used to create displayed portion of list
        (recordsViewOrderField() != null ?
            (showPrimaryName ? 
                new ABCMdbField[]{ recordsViewOrderField(), recordsViewPrimaryNameField(), recordsViewSecondaryNameField()}   :
                new ABCMdbField[]{ recordsViewOrderField(), recordsViewSecondaryNameField()})
            :
        (showPrimaryName ? 
            new ABCMdbField[]{ recordsViewPrimaryNameField(), recordsViewSecondaryNameField()}   :
            new ABCMdbField[]{ recordsViewSecondaryNameField()})),
        // Field to use to limit this list- the dominant FK field
        recordsViewPrimaryFKField(), 
        // FK of instance of dominant FK being listed
        primaryFK);
  }
  
  /**
   * Adds a set of records to an XML document. The set is defined by the value relating
   * records to the FK to the primary table.
   * @param xml the XMLAssist object constructing the xml document
   * @param includeComments true if comments should be included 
   * @param FK1 foreign key value to link to the primary table
   */
  public void    addSetToXML(XMLAssist xml, boolean includeComments, long FK1){
    addSetToXML(xml, includeComments, null, FK1);
  }

  /**
   * Adds a set of records to an XML document. The set is defined by the value relating
   * records to the FK to the primary table.
   * @param xml the XMLAssist object constructing the xml document
   * @param includeComments true if comments should be included 
   * @param orderByField field that can be sorted
   * @param FK1 foreign key value to link to the primary table
   */
  protected void addSetToXML(XMLAssist xml, boolean includeComments, ABCMdbField orderByField, long FK1){
    addSetToXML(xml, includeComments, getFilteredOrderedSQL(orderByField, FK1));   
  }
  
  /**
   * Adds a set of records to an XML document. The set is defined by the SQL statement passed
   * @param xml the XMLAssist object constructing the xml document
   * @param includeComments true if comments should be included 
   * @param sql the SQL statement to retrieve this set
   */
  protected void addSetToXML(XMLAssist xml, boolean includeComments, String sql){
    try {
          
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      addInstancesToXML(xml, rs, includeComments);
      
      rs.close();
      st.close();
      con.close();
    } 
    catch (Exception e) {
      addToErrorReport("Error adding set to XML: " + e.getMessage());
    }
  }
  
  /**
   * Gets an SQL statement that selects records from the Base Table based on
   * the value of the foreign key to the primary table, ordered by some field
   * in the Base Table
   * @param orderByField field to use to order. If null, ignored.
   * @param FK1 value of foreign key
   * @return SQL SELECT statement
   */
  protected String getFilteredOrderedSQL(ABCMdbField orderByField, long FK1){
    return recordsView().getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(recordsViewPrimaryFKField().getName(), FK1) + 
          ((orderByField != null) ? (" " + AbstractDBGateway.orderBy(orderByField.getName())) : 
            (orderField() != null ? " " + AbstractDBGateway.orderBy(orderField().getName()) : "")));
  }
  
  /**
   * Calls getXML (defined in superclass) restricted to set of records defined by the value of
   * the foreign key to the primary table. Order is unspecified; comments and XML header are omitted.
   * @param FK1 value of foreign key
   * @return the XML for the specified set of records
   */
  public String getSetXML(long FK1){
    return getSetXML(null, FK1, EXCLUDE_INSTANCE_HEADER, EXCLUDE_COMMENTS);
  }
  
  /**
   * Calls getXML (defined in superclass) restricted to set of records defined by the value of
   * the foreign key to the primary table. Order is unspecified; comments are excluded.
   * @param FK1 value of foreign key
   * @param includeHeader flag to specify if XML header is included
   * @return the XML for the specified set of records
   */
  public String getSetXML(long FK1, boolean includeHeader){
    return getSetXML(null, FK1, includeHeader, EXCLUDE_COMMENTS);
  }
  
  /**
   * Calls getXML (defined in superclass) restricted to set of records defined by the value of
   * the foreign key to the primary table. Order of records is unspecified
   * @param FK1 value of foreign key
   * @param includeHeader flag to specify if XML header is included
   * @param includeComments flag to specify if comments are included
   * @return the XML for the specified set of records
   */
  public String getSetXML(long FK1, boolean includeHeader, boolean includeComments){
    return getSetXML(null, FK1, includeHeader, includeComments);
  }
  
  /**
   * Calls getXML (defined in superclass) restricted to set of records defined by the value of
   * the foreign key to the primary table. Comments and the XML header are omitted.
   * @param orderByField field by which to order records
   * @param FK1 value of foreign key
   * @return the XML for the specified set of records
   */
  public String getSetXML(ABCMdbField orderByField, long FK1){
    return getSetXML(orderByField, FK1, EXCLUDE_INSTANCE_HEADER, EXCLUDE_COMMENTS);
  }
  
  /**
   * Calls getXML (defined in superclass) restricted to set of records defined by the value of
   * the foreign key to the primary table. Comments are excluded by default.
   * @param orderByField field by which to order records
   * @param FK1 value of foreign key
   * @param includeHeader flag to specify if XML header is included
   * @return the XML for the specified set of records
   */
  public String getSetXML(ABCMdbField orderByField, long FK1, boolean includeHeader){
    return getSetXML(orderByField, FK1, includeHeader, EXCLUDE_COMMENTS);
  }  
  
  /**
   * Calls getXML (defined in superclass) restricted to set of records defined by the value of
   * the foreign key to the primary table.
   * @param orderByField field by which to order records
   * @param FK1 value of foreign key
   * @param includeHeader flag to specify if XML header is included
   * @param includeComments flag to specify if comments are included
   * @return the XML for the specified set of records
   */
  public String getSetXML(ABCMdbField orderByField, long FK1, boolean includeHeader, boolean includeComments){
    return getXML(getFilteredOrderedSQL(orderByField, FK1), includeHeader, includeComments);
  }
  
  /**
   * Gets an HTML representation of a subset of records defined by shared value to the
   * foreign key to the primary table.
   * @param FK1 value of foreign key
   * @param includeHeaders true if HTML headers are to be included
   * @param xslRootDir directory in which XSL stylesheets are stored
   * @return the HTML for the requested set of records.
   */
  public String getSetHTML(long FK1, boolean includeHeaders, String xslRootDir){
    if(getListStylesheetName() == null) return "";
    return getHTML(getSetXML(FK1), includeHeaders, xslRootDir, getListStylesheetName());
  }

  
  // Manage ordered elements (optional- depends on child class implementation of orderField();)
  
  
  
  protected boolean collapseOrderNumbers(long dominantPK){
    boolean successFlag = false;
    try {
      String dbOwner = DB.getOwner();
      ABCMdbTable table = baseTable();
      
      String tableName = table.getFullName(dbOwner);
      
      String IDFieldName          = baseTablePKField().getFullFieldName(dbOwner);
      String shortIDFieldName     = baseTablePKField().getName();
      String secondaryIDFieldName = baseTableSecondaryFKField().getFullFieldName(dbOwner);
      String primaryIDFieldName   = baseTablePrimaryFKField().getFullFieldName(dbOwner);
      String orderFieldName       = orderField().getFullFieldName(dbOwner);
      String shortOrderFieldName  = orderField().getName();
      
      String sqlGetList =
          "SELECT "   + IDFieldName + ", " + secondaryIDFieldName + ", " + orderFieldName + " " +
          "FROM "     + tableName + " " +
          "WHERE "    + primaryIDFieldName + " = " + dominantPK + " " +
          "ORDER BY " + orderFieldName +
          ";"; 
      
      Connection con = DB.getGuestConnection();
      Statement stGetList = con.createStatement();
      Statement stUpdate = con.createStatement();
      
      ResultSet rs = stGetList.executeQuery(sqlGetList);

      long instrCount = 1;
      while(rs.next()){
        if(rs.getInt(shortOrderFieldName)!= instrCount){
          String sqlUpdate = 
              "UPDATE " + tableName + " " + 
              "SET " + shortOrderFieldName + " =" + instrCount + " " +
              "WHERE " + IDFieldName + " = " + rs.getInt(shortIDFieldName) +
              ";";
          stUpdate.executeUpdate(sqlUpdate);
        }
        instrCount++;
      }
      
      successFlag = true;
    }
    catch(Exception e){
      addToErrorReport("Could not collapse Ordered values in Link Table Manager: " + e.getMessage());
    }
    return successFlag;
  }

  
  public boolean incrementOrders(long dominantPK, long newOrder){
    try{
      String dbOwner = DB.getOwner();
      ABCMdbTable table = baseTable();
      String tableName = table.getFullName(dbOwner);
      String orderFieldName = orderField().getFullFieldName(dbOwner);
      String dominantIDFieldName = this.baseTablePrimaryFKField().getFullFieldName(dbOwner);
      
      String sqlIncrement = 
        "UPDATE " + tableName + " " +
        "SET " + orderFieldName + " = " + orderFieldName + " + 1 " + 
        "WHERE " + dominantIDFieldName + " = " + dominantPK + " " + 
        "AND " + orderFieldName + " >= " + newOrder + ";";
      
      Connection con = DB.getAdminConnection();
      Statement st = con.createStatement();
      
      st.executeUpdate(sqlIncrement);
      
      st.close();
      con.close();
    }
    catch(Exception e){
      addToErrorReport("Could not increment narrative orders: + " + e.getMessage());
      return false;
    }
    return true;
  }

  public boolean changeOrder(long recordPK, long newOrder){
    long dominantPK = DB.getNumericItem(baseTable(), this.baseTablePKField(), recordPK, this.baseTablePrimaryFKField());
    long oldOrder   = DB.getNumericItem(baseTable(), baseTablePKField(), recordPK, this.orderField());
    return changeOrder(dominantPK, oldOrder, newOrder);    
  }
  
  public boolean changeOrder(long dominantPK, long oldOrder, long newOrder){
    if(oldOrder == newOrder) return true; // Check and skip work if not needed
    if(newOrder <= 0){
      addToErrorReport("Order number must be greater than zero.");
      return false;
    }
    
    
    // Get the PK of the record to be changed
    long PK = getPKFromDominantPKAndOrder(dominantPK, oldOrder);
    System.out.println(dominantPK + " " + oldOrder + " Retrieved PK = " + PK);

    if(PK == -1) return false;
    
    if(!incrementOrders(dominantPK, newOrder)) return false;

    
    
    // Change the order number of the specified record
    try{
      String dbOwner = DB.getOwner();
      ABCMdbTable table = baseTable();
      
      String tableName = table.getFullName(dbOwner);
      
      String IDFieldName          = baseTablePKField().getFullFieldName(dbOwner);
      String shortOrderFieldName  = orderField().getName();
      
      Connection con = DB.getGuestConnection();
      Statement stUpdate = con.createStatement();
      
      String sqlUpdate = 
        "UPDATE " + tableName + " " + 
        "SET " + shortOrderFieldName + " =" + newOrder + " " +
        "WHERE " + IDFieldName + " = " + PK +
        ";";
      stUpdate.executeUpdate(sqlUpdate);
      stUpdate.close();
      con.close();

    }
    catch(Exception E){
      addToErrorReport("An error occurred while updating the order number in changeOrder: " + E.getMessage());
      collapseOrderNumbers(dominantPK); // Should try this even if other elements have not been successful
      return false;                     // Reflecting the error
    }
    
    // Collapse order numbers 
    if(!collapseOrderNumbers(dominantPK)) return false;
    
    // If everything completed successfully
    return true;    
  }
  
}
