package abcm.db.managers.abstractparents;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import abcm.db.ABCMXMLTransformer;
import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.utilities.XMLAssist;

/**
 * Provides the most basic functionality required for managing a single
 * object's collection of records in the database. This is one level higher
 * than providing database operations; that is, this class makes use of the
 * functionality provided by AbstractDBGateway and implementations thereof.
 * 
 * Implementations of this class must provide the following:
 * 
 * <ul>
 * <li>A ResultSet returned by the records() method; this is, generally, a human-readable
 * version of the records being managed- that is, with foreign keys replaced by
 * human-readable values when appropriate.</li>
 * <li>An ABCMdbField returned by the PKField() method; this corresponds to the Primary
 * Key field in the table being managed (not necessarily the human-readable view returned by
 * records()). The Primary Key is assumed to be arbitrary and numeric (usually an Identity field)</li>
 * <li>xmlOuterNodeName a name assigned to the collection of objects listed when an XML version of the
 * list of records is requested. Usually the XML result looks like: "Objects""Object"-data-"/Object""/Objects", 
 * in which case "Objects" is provided by this method.</li>
 * <li>a method to add instances to an XML document. This must translate the fields into XML names. This is assumed
 * to work from a ResultSet with the same structure as that returned by 'records()', even if the ResultSet is 
 * actually created through other means.</li>
 * <li>getListStylesheetName() and getInstanceStylesheetName(): these are filenames; both permit the translation from XML into HTML
 * if an appropriate style sheet is passed.</li>
 * <li>An ABCMDBDatum object that can be used to save an instance in to the table the implementing manager governs.</li>
 * <li>A method to clear the object to be saved; generally speaking, a calling object must set values to be saved, then
 * invoke the 'save' method. Between these the values to be saved are in private variables, which are automatically cleared
 * when the save occurs. The implementing object must provide the method to clear these values.</li> 
 * </ul>
 * 
 * When given these values, the abstract class provides the following functionality to its child classes:
 * 
 * <ul>
 * <li>XMLCreation: XML representations of specific instances or lists of the objects being managed can be
 * created, with options for including XML header information and object comments. Note that this relies on the
 * <code>addInstancesToXML</code> method which must be provided by the implementing class. Of special importance are two
 * protected variables available to the implementation: singleInstance is a boolean flag that indicates that only one
 * instance (record) is being represented; instance PK is the primary key of that instance. This allows the implementation
 * to select between alternative representations- generally, single instances will include detailed information, including
 * information about child records, that lists will omit.</li>
 * <li>HTMLCreation: the XML representation can be translated into HTML (via the specified stylesheet)</li>
 * <li>Saving: data can be saved to the database, including automated retrieval of the PK of the last object saved</li>
 * <li>Error Reporting: A standard method of recording and reporting errors, including those from the DBGateway object.</li>
 * </ul>
 * 
 * 
 * @author John T. Murphy
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractDBManager{

  /** Indicates that the instance header should be included in the XML representation        */
  public static final boolean INCLUDE_INSTANCE_HEADER = true;
  
  /** Indicates that the instance header should not be included in the XML representation    */
  public static final boolean EXCLUDE_INSTANCE_HEADER = false;
  
  /** Indicates that comments should be included in the XML and HTML representations         */
  public static final boolean INCLUDE_COMMENTS = true;
  
  /** Indicates that comments should not be included in the XML and HTML representations     */
  public static final boolean EXCLUDE_COMMENTS = false;

  /** DBGateway object that will be used by this manager                                     */
  protected AbstractABCMDBGateway DB;
  
  /** A String that contains the most recent error messages from db or other operations      */
  protected String                errorReport = "";
  
  /** The numeric PK of the last item saved by this manager                                  */
  protected long                  lastSavedID = -1;

  /** 
   * A flag indicating that the XML being created is creating a representation of a single 
   * instance, not a list, of these objects
   */
  protected boolean               singleInstance;
  
  /** 
   * If a single instance is being created, this variable will store the numeric primary 
   * key of that instance 
   */
  protected long                  instancePK;

  
  /* Constructor */

  /**
   * Sets the DBGateway for this manager and calls the clearInstanceToSave method to
   * initialize the object.
   * @param db DBGateway to be used by this manager
   */
  public AbstractDBManager(AbstractABCMDBGateway db){
    if(db == null) System.out.println("DB IS NULL!!!!!");
    DB                          = db;
    clearInstanceToSave();
  }

  
  /* Error report */
  /**
   * Returns the errorReport String as well as the error report from the DBGateway object
   * in use.
   * @return String error report message
   */
  public String getErrorReport(){
    return (errorReport != null ? "DBManager error Report: " + errorReport : "") + "\n" +
     (((DB != null) && (DB.getErrorReport() != null)) ?  "DB Gateway reports the following errors:" + "\n" + DB.getErrorReport() : ""); }
  
  /**
   * Clears the current errorReport String and sets it to a default (empty string)
   */
  protected void clearErrorReport(){ errorReport = ""; }
  
  /**
   * Adds an error message to this object's errorReport string.
   * @param addendum the message to be added
   */
  protected void addToErrorReport(String addendum){
    System.err.println(addendum);
    errorReport += addendum + System.getProperty("line.separator");
  }
  
  
  /* General */
  /**
   * Returns a view or table that contains a representation of the records this manager works with.
   * Often this is a human-readable view that replaces numeric FK fields with text.
   * @return the required view or table
   */
  protected abstract ABCMdbAbstractViewable recordsView();

  /**
   * The Field that is the primary key in the table that contains the records being managed
   * by this dbManager. Note that this is not a field in the same viewable that is returned
   * by 'records()'
   * @return the primary key field of the underlying table managed by this manager
   */
  protected abstract ABCMdbField recordsViewPKField();

  
  /* XML creation */
  
  /**
   * Gets the name of the outermost node when creating an xml list of the elements managed by this manager. 
   * @return an xml node name
   */
  protected abstract String xmlOuterNodeName();

  /**
   * Adds representations of instances of the elements managed by this class to an xml document. 
   * @param xml an XMLAssist object building the document
   * @param rs a recordset listing instances of this object
   * @param includeComments true if comments associated with this object are to be included, false if not
   * @throws Exception
   */
  protected abstract void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception;
  
  // XML Lists
  /**
   * Gets an xml document representing a list of the elements managed by this class. The document omits xml headers and
   * comments about the objects
   * @return the text of the xml document
   */
  public String getListXML(){ return getListXML(EXCLUDE_INSTANCE_HEADER, EXCLUDE_COMMENTS); }
  
  /**
   * Gets an xml document representing a list of the elements managed by this class. Omits comments.
   * @param includeHeader true if the xml document should include all standard xml headers; false if only an xml snippet is required
   * @return the text of the xml document
   */
  public String getListXML(boolean includeHeader){ return getListXML(includeHeader, EXCLUDE_COMMENTS); }
  
  /**
   * Gets an xml document representing a list of the elements managed by this class.
   * @param includeHeader true if the xml document should include all standard xml headers; false if only an xml snippet is required
   * @param includeComments true if the document should include comments in the db related to this object
   * @return the text of the xml document
   */
  public String getListXML(boolean includeHeader, boolean includeComments){
    if(recordsView() == null) return "";
    singleInstance = false;
    instancePK = -1;    
    String sql = recordsView().getSelectAllSQL(DB.getOwner());
    return getXML(sql, includeHeader, includeComments);
  }
  
  // XML Instances
  /**
   * Gets an xml document representing a single instance of an element managed by this class. By default it
   * omits the xml header but includes comments about the instance.
   * @param PK primary key of the element in the associated database
   * @return the text of the xml document
   */
  public String getInstanceXML(long PK){ return getInstanceXML(PK, EXCLUDE_INSTANCE_HEADER, INCLUDE_COMMENTS); }
 
  /**
   * Gets an xml document representing a single instance of an element managed by this class. Includes comments by default.
   * @param PK primary key of the element in the associated database
   * @param includeHeader true if xml headers should be included, false if not
   * @return the text of the xml document
   */
  public String getInstanceXML(long PK, boolean includeHeader) {  return getInstanceXML(PK, includeHeader, INCLUDE_COMMENTS);           }
  
  /**
   * Gets an xml document representing a single instance of an element managed by this class. 
   * @param PK primary key of the element in the associated database
   * @param includeHeader true if xml headers should be included, false if not
   * @param includeComments true if comments about this object should be included, false if not
   * @return the text of the xml document
   */
  public String getInstanceXML(long PK, boolean includeHeader, boolean includeComments){
    if(recordsView() == null) return "";
    singleInstance = true;
    instancePK     = PK;
    String sql = recordsView().getSelectSQL(DB.getOwner(), AbstractDBGateway.where(recordsViewPKField().getName(), PK));
    return getXML(sql, includeHeader, includeComments);
  }

  /**
   * Gets an xml document representing objects represented by this class. This may be one or more than one
   * instance, dependent on the sql statement passed. This method makes use of an instance variable 
   * called 'singleInstance', which is used to inform the method as to whether it is being used to transform
   * one or more instances; it uses the xml outer node if it is being used to build a list and omits this
   * if it is being used for a single instance. 
   * @param sql SQL statement that is used to extract instances of the objects from the associated db 
   * @param includeHeader true if xml headers should be included, false if not
   * @param includeComments true if comments about this object should be included, false if not
   * @return the text of the xml document
   */
  protected String getXML(String sql, boolean includeHeader, boolean includeComments){
    XMLAssist xml = new XMLAssist();
                
    try{
      // Get result sets
      Connection con  = DB.getGuestConnection();
      Statement  st   = con.createStatement();
      ResultSet  rs   = st.executeQuery(sql);

      // Build the xml output
      if(!singleInstance) xml.open(xmlOuterNodeName());
      
      addInstancesToXML(xml, rs, (includeComments ? INCLUDE_COMMENTS : EXCLUDE_COMMENTS));
        
      if(!singleInstance) xml.closeTag();
      
      rs.close();
      st.close();
      con.close();
      
    } 
    catch(Exception e){
      addToErrorReport("Error in getXML(): " + xmlOuterNodeName() + " - " + e.getMessage() + " ORIGINAL SQL: " + sql);
    }    
    if (includeHeader)      return xml.getXML(XMLAssist.INCLUDE_HEADER);
    else                    return xml.getXML(XMLAssist.EXCLUDE_HEADER);    
  }
  
  /* HTML Creation */
  
  /**
   * Gets the name of the xsl stylesheet associated with the elements managed by this dbmanager and
   * used to create html representations of lists of these elements.
   * @return an xsl stylesheet name
   */
  protected abstract String getListStylesheetName();
  
  /**
   * Gets the name of the xsl stylesheet associated with the elements managed by this dbmanager and
   * used to create an html representations of a single instance of these elements.
   * @return an xsl stylesheet name
   */
  protected abstract String getInstanceStylesheetName();
  
    
  /**
   * Returns an HTML representation of a list of the items managed by this class.
   * @param includeHeaders true if full HTML headers are to be created, or false if only the
   * HTML elements describing this item are included 
   * @param xslRootDir the directory in which the xsl stylesheet required for this transformation is found
   * @return the HTML representation
   */
  public String getListHTML(boolean includeHeaders, String xslRootDir) {
    return getHTML(getListXML(), includeHeaders, xslRootDir, getListStylesheetName());
  }

  /**
   * Returns an HTML representation of a single instance of the items managed by this class
   * @param PK primary key of the instance in the associated db
   * @param includeHeaders true if headers should be included in the HTML document, false otherwise
   * @param xslRootDir directory in which the xsl stylesheet needed for this transformation is found
   * @return text of the HTML document
   */
  public String getInstanceHTML(long PK, boolean includeHeaders, String xslRootDir) {
    return getHTML(getInstanceXML(PK), includeHeaders, xslRootDir, getInstanceStylesheetName());
  }
  
  /**
   * Returns an HTML representation of the xml document passed.
   * @param xml the xml document from which the html is to be made
   * @param includeHeaders true if HTML headers should be included
   * @param xslRootDir directory in which the xsl stylesheet will be found
   * @param stylesheet name of the stylesheet to be used
   * @return text of the HTML document
   */
  protected String getHTML(String xml, boolean includeHeaders, String xslRootDir, String stylesheet){
    if(stylesheet == null) return "";
    if(xslRootDir == null) return "";
    if(xml == null || xml.length() == 0) return "";
    return ABCMXMLTransformer.getHTML(xml, xslRootDir + stylesheet, includeHeaders);
  }
  
  
  /* Saving */

  /**
   * Gets an instance of an ABCMdbDatum that represents the next element
   * to be saved to the database.
   * @return the requested datum element
   */
  protected abstract ABCMdbDatum datumToSave();

  /**
   * Clears any previously set instance to be saved in the database. This is called automatically
   * after an instance is saved.
   */
  protected abstract void clearInstanceToSave();
  
  /**
   * Gets the PK that resulted from the last save operation
   * @return the ID of the last saved element
   */
  public long getLastSavedID(){ return lastSavedID; }
  
  /**
   * Saves an instance to the database. The instance saved has generally been set by
   * some previous method call and is returned by the datumToSave() method.
   * @return true if successful, false otherwise
   */
  public boolean saveInstanceToDB(){
    boolean successFlag = false;
    ABCMdbDatum datum = datumToSave();
    if(datum != null){
      try {
        lastSavedID = DB.insertAndReturnPK(datum.getInsertSQL(DB.getOwner(), DB.getSQLSyntaxEngine()));
        successFlag = true;
        clearInstanceToSave();
      }
      catch (Exception e) {  
        addToErrorReport("Error in saveInstanceToDB(): " + e.getMessage());   
      }
    }
    return successFlag;
  }
    
}