package abcm.db.managers.abstractparents;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;

/**
 * This class extends the functionality of its parent, AbstractDBManager, to
 * include functions common and convenient when the data being managed conform to
 * a simple structure in which there is, in addition to the arbitrary numeric PK field
 * a field that provides the human-PK, i.e. a unique name. This is common (but not universal;
 * names of people, for example, are often not contained in single fields but in several, 
 * and this class could not be applied in this case) and quite useful.
 * 
 * Additionally, a second field containing a 'description' may optionally be specified.
 * 
 * With this additional information, it becomes possible to search for PK values based on
 * name (and vice versa), and create option lists.
 * 
 * @author John T. Murphy
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractDBStandardizedElementManager extends  AbstractDBManager {

  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public AbstractDBStandardizedElementManager(AbstractABCMDBGateway db){ super(db); }
  
  /* General */

  /**
   * Gets the table containing the objects managed by this class
   * @return the table containing the objects managed by this class
   */
  protected abstract ABCMdbTable baseTable();
  
  /**
   * Gets the primary key field of the table containing the objects managed by this class
   * @return the primary key field for these objects
   */
  protected abstract ABCMdbField baseTablePKField();
  
  /**
   * Gets the field in the base table that represents the name of these objects.
   * @return the base table name field
   */
  protected abstract ABCMdbField baseTableNameField();
  
  /**
   * Gets the field in the base table containing the 'description' of these objects. Note that
   * this is defined concretely here to return 'null'; implementing classes may chose to override
   * this in order to supply a description field, but this is optional
   * @return the description field
   */
  protected ABCMdbField baseTableDescriptionField(){ return null; } // This is not required, but used by some
                                                                    // child classes for option lists

  /* Basic information retrieval */

  /**
   * Returns the numeric PK value associated with the unique value in the specified
   * 'name' field
   * @param name name to be looked up
   * @return PK of the requested element; returns -1 if name is not found
   */
  public long getPK(String name){
    if((baseTableNameField() == null) || (baseTablePKField() == null)) return -1;             // Validate
    return DB.getNumericItem(baseTable(), baseTableNameField(), name, baseTablePKField());
  }
  
  /**
   * Returns the unique 'name' value associated with the specified numeric PK.
   * @param PK the PK of the requested element
   * @return the name of the requested element 
   */
  public String getName(long PK){
    if((baseTableNameField() == null) || (recordsViewPKField() == null)) return "";             // Validate    
    return DB.getStringItem(baseTable(), baseTablePKField(), PK, baseTableNameField());
  }
  
  /**
   * Returns the description associated with the specified numeric PK. Returns an empty
   * string if the description field is null.
   * @param PK primary key of the element requested
   * @return a description of the requested element
   */
  public String getDescription(long PK){
    if((baseTableDescriptionField() == null) || (baseTablePKField() == null)) return "";      // Validate
    return DB.getStringItem(baseTable(), baseTablePKField(), PK, baseTableDescriptionField());
  }
  
  /**
   * Returns the description associated with the specified unique name.
   * @param name name of the element requested
   * @return a description of the requested element
   */
  public String getDescription(String name){
    return getDescription(getPK(name));
  }

  
  /* Standard option list */
  
  /**
   * A convenience method that invokes getOptionList(boolean, boolean) using default
   * values of true (use numeric values) and false (do not include a description)
   * @return the HTML option list
   */
  public String getOptionList(){   
    return getOptionList(true, false);    
  }
  
  /**
   * A convenience method that invokes getOptionList(boolean, boolean) with a caller-specified
   * value for whether numeric values are to be used; uses the default (do not include a description)
   * for the second option.
   * @param useNumericValues true if numeric values are to be used in the 'Value' portion of the option
   * list, false otherwise.
   * @return the HTML option list
   */
  public String getOptionList(boolean useNumericValues){
    return getOptionList(useNumericValues, false);
    
  }

  /**
   * Returns the collection of records managed by this object as an HTML option list.
   * @param useNumericValues if true, the 'Values' portion of the option list will
   * be drawn from the numeric PK field; if false, it will be drawn from the String
   * in the 'name' field
   * @param includeDescription if true, the option list will display not only the name
   * of each item but the description as well (with a default placeholder in between).
   * @return the HTML option list
   */
  public String getOptionList(boolean useNumericValues, boolean includeDescription){
    if(baseTableNameField() == null) return "<OPTION></OPTION>";
    if((includeDescription == false) || (baseTableDescriptionField() == null))
      return DB.getOptionList(baseTable(), (useNumericValues ? baseTablePKField() :  baseTableNameField()), baseTableNameField());
    else 
      return DB.getOptionList(baseTable(), 
                            (useNumericValues ? baseTablePKField() : baseTableNameField()),
                            new ABCMdbField[]{ baseTableNameField(), baseTableDescriptionField()});
  }
  
  
}
