package abcm.db.managers.abstractparents;

import java.sql.ResultSet;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.components.ABCMObjectCommentTable;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.utilities.XMLAssist;

/**
 * One special kind of link table is an 'Object Comment' table. An ABCM database stores
 * comments that can be associated with multiple objects; in fact, a comment can be
 * associated with any or (almost) all of the other kinds of objects, like references,
 * Simulations, Code Models, etc. These associations are managed by a simple kind of
 * linking table.
 * 
 * @author John T. Murphy
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractObjectCommentManager extends AbstractDBLinkTableInnerManager{
 
  /** Primary Key of the object, to be saved as FK  */
  protected long objectPKToSave;
  /** Primary Key of the comment, to be saved as FK  */
  protected long commentPKToSave;
    
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   * @param containedBy Containing class (also a Manager)
   */
  public AbstractObjectCommentManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
    super(db, containedBy);
  }
      
  
  /**
   * Implementing classes must specify the object comment table they are managing
   * @return gets the base table, which will be an Object Comment Table
   */
  public abstract ABCMObjectCommentTable getBaseTable();
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
   */
  protected final ABCMdbTable baseTable(){                 return getBaseTable();       }
  
  /** 
   * Overridden to return null; not used in this child class
   * @return null 
   */
  protected final ABCMdbField baseTablePKField(){          return null;                 }
  
  /** 
   * Returns the results of getCommentIDField()
   * @return the comment ID field 
   */
  protected final ABCMdbField baseTablePrimaryFKField(){   return getCommentIDField();  }
  
  /** 
   * Returns the results of getObjectIDField()
   * @return the object ID field 
   */
  protected final ABCMdbField baseTableSecondaryFKField(){ return getObjectIDField();   }
  
  /** 
   * Overridden to take no action; not used in this child class
   * @return null 
   */
  protected final ABCMdbAbstractViewable recordsView(){         return null;  }
  /** 
   * Overridden to take no action; not used in this child class
   * @return null 
   */
  protected final ABCMdbField recordsViewPKField(){             return null;  }  
  /** 
   * Overridden to take no action; not used in this child class
   * @return null 
   */
  protected final ABCMdbField recordsViewPrimaryFKField(){      return null;  }
  /** 
   * Overridden to take no action; not used in this child class
   * @return null 
   */
  protected final ABCMdbField recordsViewPrimaryNameField(){    return null;  }
  /** 
   * Overridden to take no action; not used in this child class
   * @return null 
   */
  protected final ABCMdbField recordsViewSecondaryNameField(){  return null;  }
  
  /** 
   * Overridden to take no action; not used in this child class
   * @throws Exception generic exception
   */
  protected final void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {    }

  /**
   * Sets the instance to be saved
   * @param objectID primary key of the object to which the comment is to be associated
   * @param commentID primary key of the comment
   */
  public final void setInstanceToSave(long objectID, long commentID){
    objectPKToSave  = objectID;
    commentPKToSave = commentID;
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  protected final void clearInstanceToSave(){
    objectPKToSave  = -1;
    commentPKToSave = -1;
  }
  
  /**
   * Gets the status of the instance to be saved
   * @return true if the instance to be saved is set, false otherwise
   */
  public final boolean instanceToSaveIsValid(){
    return ((objectPKToSave  != -1) &&
            (commentPKToSave != -1));
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){   return getBaseTable().getDatum(objectPKToSave, commentPKToSave);   }
  
  
  /* Abstract Methods */
  
  /**
   * Implementing classes must indicate the field that serves as the FK to the Comments table.
   * @return the Field that is the FK to the Comments Table
   */
  public abstract ABCMdbField getCommentIDField();

  /**
   * Implementing classes must indicate the field that serves as the FK to the appropriate object table.
   * @return the Field that is the FK to the object's table
   */
  public abstract ABCMdbField getObjectIDField();
    
}
