package abcm.db.managers.stubs;

import java.sql.ResultSet;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMSimulationConfigurationsAndResearchersView;
import abcm.db.managers.CommentManager;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.sim.Configuration;
import abcm.utilities.XMLAssist;

/**
 *  Provides an interface for managing Configurations as stored in the 
 *  HWM Database, including routines to transform them to displayable
 *  HTML, XML, etc.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractConfigurationManager extends AbstractAuditableDBManager {
  
  /** Stylesheet name for list  */
  private static final String listStylesheetName = "xmlConfigurationListToHTML.xsl";
  /** Stylesheet name for instances */
  private static final String instanceStylesheetName = "xmlConfigurationToHTML.xsl";
  
  /** Flag indicating that default Configuration is required  */
  protected static final int DEFAULT_PK = -2;

  /** Name of Configuration to be saved  */
  private String nameToSave;
  /** Description of Configuration to be saved  */
  private String descToSave;
  /** ID of user who is owner of Configuration to be saved  */
  private long userIDToSave;

  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public AbstractConfigurationManager(AbstractABCMDBGateway db){
    super(db);
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){              return DB.getSimulationConfigurationsTable();                                                     }
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public    ABCMdbField baseTableOwnerField(){    return DB.getSimulationConfigurationsTable().getField_ResearcherFK();                             }
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){       return DB.getSimulationConfigurationsTable().getField_PK();                                       }
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField(){     return DB.getSimulationConfigurationsTable().getField_Name();                                     }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView(){ return DB.getSimulationConfigurationsAndResearchersView();                                        }
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField(){     return DB.getSimulationConfigurationsAndResearchersView().getField_SimulationConfiguration_PK();  }
  
  
  /* Saving */  
  /**
   * Sets an instance of a Configuration to be saved
   * @param name name to be assigned
   * @param description description of Configuration
   * @param userID PK of user who will own this Configuration
   */
  public void setInstanceToSave(String name, String description, long userID){
    nameToSave   = name;
    descToSave   = description;
    userIDToSave = userID;
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  protected ABCMdbDatum datumToSave(){    return DB.getSimulationConfigurationsTable().getDatum(nameToSave, descToSave, 
      DB.currentTimeString(), userIDToSave);  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  protected void clearInstanceToSave(){  
    nameToSave   = null;
    descToSave   = null;
    userIDToSave = -1;
  }
  
  
  /* XML Creation */
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName(){    return "simconfiglist";  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMSimulationConfigurationsAndResearchersView view = (ABCMSimulationConfigurationsAndResearchersView)recordsView();
    
    int PKFieldCol        = rs.findColumn(view.getField_SimulationConfiguration_PK().getName());
    int simConfigNameCol  = rs.findColumn(view.getField_SimulationConfiguration_Name().getName());
    int researcherNameCol = rs.findColumn(view.getField_Researcher_Name().getName());
    int descCol           = rs.findColumn(view.getField_SimulationConfiguration_Desc().getName());
    int canBeEditedCol    = rs.findColumn(view.getField_CanBeEdited().getName());
    int truthValueCol     = rs.findColumn(view.getField_TruthValue().getName());
    
    while(rs.next()){
      xml.open("simconfig");
        xml.addLine("ID",  rs.getString(PKFieldCol));
        xml.addLine("name", rs.getString(simConfigNameCol));
        xml.addLine("createdBy", rs.getString(researcherNameCol));
        xml.addLine("desc", rs.getString(descCol));
        xml.addLine("canBeEdited", ((rs.getString(canBeEditedCol) != null) ? "Yes" : "No"));
        xml.addLine("truthValue", rs.getString(truthValueCol));
        
        if(singleInstance){
          addConfigurationDetailsToXML(xml, instancePK);
        }
        
        if(includeComments){
          CommentManager cm = new CommentManager(DB);
          xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.CONFIGURATION, instancePK), XMLAssist.INDENT_BLOCK);
        }    
      xml.closeTag();
    }
    
    
    
  }

  /* Auditing */
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView(){ return DB.getCanBeEdited_SimulationConfigurationsView();  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField canBeEditedViewPKField(){     return DB.getCanBeEdited_SimulationConfigurationsView().getField_SimulationConfiguration_PK();  }
  
  /* Abstract methods to be imposed lower in the class hierarchy */
  /**
   * Implementing classes must provide for returning a default Configuration
   * @return the default Configuration
   */
  public abstract Configuration getDefaultConfiguration();
  
  /**
   * Implementing classes must retrieve a Configuration given its primary key
   * @param PK primary key of Configuration
   * @return the requested Configuration
   */
  public abstract Configuration getConfiguration(long PK);
  
  /**
   * Implementing classes must add Configuration details to an XML document 
   * @param xml XMLAssist object creating the XML document
   * @param PK primary key of Configuration to be added
   */
  public abstract void addConfigurationDetailsToXML(XMLAssist xml, long PK);

  /* Convenience stubs */
  /**
   * Gets a Configuration from the ABCM Database given its name.
   * @param configurationName the name of the requested Configuration
   * @return the requested Configuration
   */
  public final Configuration getConfiguration(String configurationName){
    return getConfiguration(getPK(configurationName));
  }

  
  
  
}
