package abcm.db.managers.stubs;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Vector;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.views.ABCMDataModelsWithNamesOfElementsView;
import abcm.db.components.views.ABCMSummariesAndResearchersView;
import abcm.db.components.views.ABCMSummaryAnalysesAndDetailsView;
import abcm.db.components.views.ABCMSummaryCodeModelsAndDetailsView;
import abcm.db.components.views.ABCMSummaryConfigurationsAndDetailsView;
import abcm.db.components.views.ABCMSummaryDataModelsAndDetailsView;
import abcm.db.components.views.ABCMSummaryHistoriesAndDetailsView;
import abcm.db.components.views.ABCMSummaryNarrativesAndDetailsView;
import abcm.db.components.views.ABCMSummaryParameterSetsView;
import abcm.db.components.views.ABCMSummaryProbeSetsView;
import abcm.db.components.views.ABCMSummaryRunsAndDetailsView;
import abcm.db.components.views.ABCMSummarySimulationsAndDetailsView;
import abcm.db.managers.CodeModelManager;
import abcm.db.managers.CommentManager;
import abcm.db.managers.HistoryManager;
import abcm.db.managers.ParameterSetManager;
import abcm.db.managers.ProbeSetManager;
import abcm.db.managers.SimulationManager;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.math.SimDate;
import abcm.utilities.XMLAssist;

/**
 * Provides methods for the storage and retrieval of Summaries from the HWM Database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractSummaryManager extends AbstractAuditableDBManager {

  /** Name of summary to be saved */
  private String summaryNameToSave;
  
  /** ID of user to be saved with summary */
  private long userIDToSave;

  /** Summary Analyses Manager to be used internally */
  private SummaryAnalysesManager SAM;
  
  /**
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class SummaryAnalysesManager extends AbstractDBLinkTableInnerManager{

    /** FK of summary to be saved  */
    private long   summaryFKToSave;
    /** FK of analysis to save  */
    private long   analysisFKToSave;
    /** Description to be saved  */
    private String descriptionToSave;
    /** Analysis order to save */
    private int    analysisOrderToSave;

    /**
     * Constructor
     * @param db Instance of an ABCM Gateway to provide database access.
     * @param containedBy enclosing class
     */
    public SummaryAnalysesManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
     */
    protected ABCMdbTable baseTable(){                     return DB.getSummaryAnalysesTable();                                    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
     */
    protected ABCMdbField baseTablePKField(){              return DB.getSummaryAnalysesTable().getField_PK();                      }


    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
     */
    protected ABCMdbField baseTablePrimaryFKField(){       return DB.getSummaryAnalysesTable().getField_SummaryFK();               }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
     */
    protected ABCMdbField baseTableSecondaryFKField(){     return DB.getSummaryAnalysesTable().getField_AnalysisFK();              }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
     */
    protected ABCMdbAbstractViewable recordsView(){        return DB.getSummaryAnalysesAndDetailsView();                           }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
     */
    protected ABCMdbField recordsViewPKField(){            return DB.getSummaryAnalysesTable().getField_PK();                      }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
     */
    protected ABCMdbField recordsViewPrimaryFKField(){     return DB.getSummaryAnalysesAndDetailsView().getField_Summary_PK();     }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
     */
    protected ABCMdbField recordsViewPrimaryNameField(){   return DB.getSummaryAnalysesAndDetailsView().getField_Summary_Title();  }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
     */
    protected ABCMdbField recordsViewSecondaryNameField(){ return DB.getSummaryAnalysesAndDetailsView().getField_Analysis_Title(); }

    // Summaries include analyses in specific orders
    /** (non-Javadoc)
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#orderField()
     */
    protected ABCMdbField orderField(){                    return DB.getSummaryAnalysesTable().getField_Order();                   }
    
    protected long getOrderedDominantPKToSave(){           return summaryFKToSave;                                                 }
    
    protected long getOrderedNewOrderToSave(){             return analysisOrderToSave;                                             }
    
    protected ABCMdbField recordsViewOrderField(){         return DB.getSummaryAnalysesAndDetailsView().getField_AnalysisOrder();  }
        
    /* XML */
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
      
      ABCMSummaryAnalysesAndDetailsView view = DB.getSummaryAnalysesAndDetailsView();
      
      int summaryPKFieldCol           = rs.findColumn(view.getField_Summary_PK().getName());
      int summaryTitleFieldCol        = rs.findColumn(view.getField_Summary_Title().getName());
      int analysisPKFieldCol          = rs.findColumn(view.getField_Analysis_PK().getName());
      int analysisTitleFieldCol       = rs.findColumn(view.getField_Analysis_Title().getName());
      int analysisDateCreatedFieldCol = rs.findColumn(view.getField_Analysis_DateCreated().getName());
      int analysisDescFieldCol        = rs.findColumn(view.getField_Analysis_Desc().getName());
      int researcherNameFieldCol      = rs.findColumn(view.getField_Researcher_LastNameFirstName().getName());
            
      while(rs.next()){
        xml.open("analysisListed");
          xml.addLine("analysisTitle", rs.getString(analysisTitleFieldCol));
          
          // For each analysis get the runs to which it applies
          String sql = DB.getAnalysesAndRunsView().getSelectSQL(DB.getOwner(), 
              AbstractDBGateway.where(DB.getAnalysesAndRunsView().getField_Analysis_PK().getName(), rs.getLong(analysisPKFieldCol)) + " " + 
              AbstractDBGateway.orderBy(DB.getAnalysesAndRunsView().getField_SimulationRun_PK().getName()));

          Connection con = DB.getAdminConnection();
          Statement st = con.createStatement();
          ResultSet rs2 = st.executeQuery(sql);

          int simulationRunFieldCol = rs2.findColumn(DB.getAnalysesAndRunsView().getField_SimulationRun_PK().getName());
          
          xml.open("analysisRuns");
            while(rs2.next()){
              xml.open("analysisRun");
                xml.addLine("analysisRunID", rs2.getLong(simulationRunFieldCol));
              xml.closeTag();
            }
          xml.closeTag();
          
        xml.closeTag();  // Analysis Listed
        rs2.close();
        st.close();
      }      
    }  
    
    
    /* Saving */
    
    /**
     * Sets the instance to be saved
     * @param summaryFK FK of summary to be saved
     * @param analysisFK FK of analysis to be saved
     * @param description description to be saved
     * @param order order of the analysis within this summary
     */
    public void setInstanceToSave(long summaryFK, long analysisFK, String description, int order){
      summaryFKToSave     = summaryFK;
      analysisFKToSave    = analysisFK;
      descriptionToSave   = (description != null ? description : "");
      analysisOrderToSave = order; 
    }
    
    /**
     * Determines if the instance to be saved is valid
     * @return true if the instance to be saved has been set and is valid, false otherwise
     */
    public boolean instanceToSaveIsValid(){
      return ((summaryFKToSave   != -1) && 
              (analysisFKToSave  != -1) && 
              (descriptionToSave != null));
    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
     */
    protected void clearInstanceToSave() {
      summaryFKToSave   = -1;
      analysisFKToSave  = -1;
      descriptionToSave = null;
      
    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
     */
    protected ABCMdbDatum datumToSave() {
      return (instanceToSaveIsValid() ? 
          DB.getSummaryAnalysesTable().getDatum("" + summaryFKToSave, "" + analysisFKToSave, descriptionToSave, analysisOrderToSave) : null);
    }
    
    /**
     * Gets the list of the primary keys of the analyses, in the correct order. Someday this might be better in the super class
     * @param summaryPK
     * @return the list of analysis PKs in order
     */
    protected Vector<Long> getListOfAnalysisPKsInOrder(long summaryPK){
      try{
        Vector<Long> analysisFKs = new Vector<Long>();
        
        Connection con = DB.getAdminConnection();
        
        String sql = baseTable().getSelectSQL(DB.getOwner(),
            AbstractDBGateway.where(DB.getSummaryAnalysesTable().getField_SummaryFK().getName(), summaryPK) + " " + 
            AbstractDBGateway.orderBy(orderField().getName()));
        Statement st = con.createStatement();
        ResultSet rs = st.executeQuery(sql);
        
        
        while(rs.next())  analysisFKs.add(new Long(rs.getLong(DB.getSummaryAnalysesTable().getField_AnalysisFK().getName())));    
        
        rs.close();
        st.close();
        
        return analysisFKs; 
      }
      catch(Exception E){
        addToErrorReport("Could not retrieve list of Analysis FKs " + E.getMessage());
      }
      return null;
    }
    
    /**
     * Gets the descriptions of the analyses in the listed order
     * @param summaryPK
     * @return the list of analyses' descriptions
     */
    protected Vector<String> getListOfAnalysisDescriptionsInOrder(long summaryPK){
      try{
        Vector<String> analysisDescriptions = new Vector<String>();
        
        Connection con = DB.getAdminConnection();
        
        String sql = baseTable().getSelectSQL(DB.getOwner(),
            AbstractDBGateway.where(DB.getSummaryAnalysesTable().getField_SummaryFK().getName(), summaryPK) + " " + 
            AbstractDBGateway.orderBy(orderField().getName()));
        Statement st = con.createStatement();
        ResultSet rs = st.executeQuery(sql);
        
        
        while(rs.next())  analysisDescriptions.add(rs.getString(DB.getSummaryAnalysesTable().getField_Description().getName()));    
        
        rs.close();
        st.close();
        
        return analysisDescriptions; 
      }
      catch(Exception E){
        addToErrorReport("Could not retrieve list of Analysis FKs " + E.getMessage());
      }
      return null;
    }
    
    /**
     * Updates the description of an analysis within the summary
     * @param summaryPK
     * @param analysisOrder
     * @param newDescription
     * @return true if successful
     */
    protected boolean updateAnalysisDescription(long summaryPK, long analysisOrder, String newDescription){
      long PK = this.getPKFromDominantPKAndOrder(summaryPK, analysisOrder);
      if(PK == -1){
        addToErrorReport("Could not update description; no analysis with order number " + analysisOrder + " was found."); 
        return false;        
      }
      return DB.updateItem(baseTable(), baseTablePKField(), PK, 
          DB.getSummaryAnalysesTable().getField_Description(), AbstractDBGateway.makeSQLSafe(newDescription));
    }
    
  } // end inner class
  

  // public class
  
  /**
   * Wraps the details of a summary.
   * 
   *  Variables are not in JavaDoc; simply memoize instances of a Summary
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class SummaryDetails{
    private boolean        valid         = false;
    private String         name          = null;
    private String         desc          = null;
    private String         summ          = null;
    private String         concl         = null;
    private String         createdBy     = null;
    private String         createdOn     = null;
    private Vector<Long>   analysisFKs   = null;
    private Vector<String> descriptions  = null;
    
    public SummaryDetails(long PK){
      try{
        Connection con = DB.getAdminConnection();
        
        // First hit the main recordset to get the description, summary, and conclusion
        String sql = DB.getSummariesAndResearchersView().getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(DB.getSummariesAndResearchersView().getField_Summary_PK().getName(), PK));

        Statement st = con.createStatement();
        ResultSet rs = st.executeQuery(sql);
        
        if(rs.next()){
          name      = rs.getString(DB.getSummariesAndResearchersView().getField_Summary_Title().getName());
          desc      = rs.getString(DB.getSummariesAndResearchersView().getField_Summary_Desc().getName());
          summ      = rs.getString(DB.getSummariesAndResearchersView().getField_Summary_Summary().getName());
          concl     = rs.getString(DB.getSummariesAndResearchersView().getField_Summary_Conclusion().getName());
          createdOn = rs.getString(DB.getSummariesAndResearchersView().getField_Summary_DateCreated().getName());
          createdBy = rs.getString(DB.getSummariesAndResearchersView().getField_Researcher_LastNameFirstName().getName());
        }
        
        rs.close();
        st.close();
        
        analysisFKs  = SAM.getListOfAnalysisPKsInOrder(PK);
        descriptions = SAM.getListOfAnalysisDescriptionsInOrder(PK); 
        
        con.close();
        valid = true;
      }
      
      catch(Exception E){
        // Do nothing; result will be null entries and 'invalid' flag
      }
      // Invalid objects should contain no information- better none than incomplete
      if(!valid){
        name          = null;
        desc          = null;
        summ          = null;
        concl         = null;
        createdBy     = null;
        createdOn     = null;
        analysisFKs   = null;
        descriptions  = null;
      }
      
    }
    
    public boolean        isValid(){                 return valid;         }
    public String         getName(){                 return name;          }
    public String         getDescription(){          return desc;          }
    public String         getSummary(){              return summ;          }
    public String         getConclusion(){           return concl;         }
    public String         getCreatedBy(){            return createdBy;     }
    public String         getCreatedOn(){            return createdOn;     }
    public Vector<Long>   getAnalysisFKsInOrder(){   return analysisFKs;   }
    public Vector<String> getAnalysisDescsInOrder(){ return descriptions;  }
    
  }
  
  /* Constructor */
  
  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public AbstractSummaryManager(AbstractABCMDBGateway db){    
    super(db);
    SAM = new SummaryAnalysesManager(DB, this);
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  public ABCMdbTable baseTable(){                   return DB.getSummariesTable();                                        }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){         return DB.getSummariesTable().getField_PK();                          }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  public ABCMdbField baseTableNameField(){          return DB.getSummariesAndResearchersView().getField_Summary_Title();  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableDescriptionField() 
   */
  public ABCMdbField baseTableDescriptionField(){   return DB.getSummariesAndResearchersView().getField_Summary_Desc();   }
  
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  public ABCMdbAbstractViewable recordsView(){      return DB.getSummariesAndResearchersView();                           }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  public ABCMdbField recordsViewPKField(){          return DB.getSummariesAndResearchersView().getField_Summary_PK();     }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName(){  return "summaries"; }
  
  
  /* XML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  public void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    ABCMSummariesAndResearchersView view = DB.getSummariesAndResearchersView();
    
    int summaryPKFieldCol                   = rs.findColumn(view.getField_Summary_PK().getName());
    int summaryTitleFieldCol                = rs.findColumn(view.getField_Summary_Title().getName());
    int summaryDescFieldCol                 = rs.findColumn(view.getField_Summary_Desc().getName());
    int summarySummaryFieldCol              = rs.findColumn(view.getField_Summary_Summary().getName());
    int summaryConclusionFieldCol           = rs.findColumn(view.getField_Summary_Conclusion().getName());
    int dateCreatedFieldCol                 = rs.findColumn(view.getField_Summary_DateCreated().getName());
    int researcherNameFieldCol              = rs.findColumn(view.getField_Researcher_LastNameFirstName().getName());
        
    while(rs.next()){
      xml.open("summary");
        xml.addLine("ID",           rs.getLong(summaryPKFieldCol));
        xml.addLine("title",        rs.getString(summaryTitleFieldCol));
        xml.addLine("desc",         rs.getString(summaryDescFieldCol));
        xml.addLine("summaryText",  rs.getString(summarySummaryFieldCol));
        xml.addLine("conclusion",   rs.getString(summaryConclusionFieldCol));
        xml.addLine("date",         rs.getString(dateCreatedFieldCol));
        xml.addLine("createdBy",    rs.getString(researcherNameFieldCol));
        
        if(singleInstance){
          Connection con = DB.getAdminConnection();
          
          // Add list of analyses
          xml.open("analysesList");
            SAM.addSetToXML(xml, includeComments, instancePK);
          xml.closeTag();  // AnalysesList
          
          // List of runs
          xml.open("runList");
            addRunsXMLToSummary(xml, instancePK, con);
          xml.closeTag();
          
          // List of simulations
          xml.open("simulationList");
            addSummarySimulationsAndDetails(xml, instancePK, con);
          xml.closeTag();
        
          xml.open("dataModelList");
            addSummaryDataModelsAndDetails(xml, instancePK, con);
          xml.closeTag();
          
          xml.open("historyList");
            addSummaryHistoryList(xml, instancePK, con);
          xml.closeTag();
       
          xml.open("narrativeList");
            addSummaryNarrativesAndDetails(xml, instancePK, con);
          xml.closeTag();
        
          xml.open("codeModelList");
            addSummaryCodeModelsAndDetails(xml, instancePK, con);
          xml.closeTag();

          
          xml.open("configurationInfo");
            xml.open("configurationsList");
              addSummaryConfigurationsAndDetails(xml, instancePK, con);
            xml.closeTag();
            xml.open("configurationDetails");
              xml.addBlock(getSpecificConfigurationsDetails(instancePK).getXML(XMLAssist.EXCLUDE_HEADER), true);
            xml.closeTag();
          xml.closeTag();
          
          xml.open("parameterSetsList");
            addParameterSetsAndDetails(xml, instancePK, con);
          xml.closeTag();
    
          xml.open("probeSetList");
            addProbeSetsAndDetails(xml, instancePK, con);
          xml.closeTag();
          
          xml.open("analysesFull");
          xml.closeTag();

          if(includeComments){
            CommentManager cm = new CommentManager(DB);
            xml.addBlock(cm.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.SUMMARY, instancePK), XMLAssist.INDENT_BLOCK);
          }          

          
          con.close();
        }
          
        
      xml.closeTag();
    }    
  }
  
  /**
   * Adds XML describing simulation runs to an XML document
   * @param xml XMLAssist object constructing the XML document
   * @param summaryPK primary key of the summary being listed
   * @param con database connection
   * @throws Exception
   */
  private void addRunsXMLToSummary(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    ABCMSummaryRunsAndDetailsView view = DB.getSummaryRunsAndDetailsView();
    
    String sql = view.getSelectSQL(DB.getOwner(), 
        AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK) + " " + 
        AbstractDBGateway.orderBy(view.getField_SimulationRun_PK().getName()));
    
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);

    
    int summaryPKFieldCol            = rs.findColumn(view.getField_Summary_PK().getName());
    int simulationRunPKFieldCol      = rs.findColumn(view.getField_SimulationRun_PK().getName());
    int simulationRunAtFieldCol      = rs.findColumn(view.getField_SimulationRun_DateTimeBegun().getName());
    int simultionRunNotesFieldCol    = rs.findColumn(view.getField_SimulationRun_Notes().getName());
    int simulationRunRNSFieldCol     = rs.findColumn(view.getField_SimulationRun_RandomNumberSeed().getName());
    int simulationRunReRunFKFieldCol = rs.findColumn(view.getField_SimulationRun_ReRun_FK().getName());
    int simulationTitleFieldCol      = rs.findColumn(view.getField_Simulation_Title().getName());
    int paramSetNameFieldCol         = rs.findColumn(view.getField_ParameterSet_Name().getName());
    int probeSetNameFieldCol         = rs.findColumn(view.getField_ProbeSet_Name().getName());
    int researcherNameFieldCol       = rs.findColumn(view.getField_Researcher_Name().getName());    
    int truthValueFieldCol           = rs.findColumn(view.getField_TruthValue().getName());
    
    while(rs.next()){
      xml.open("run");
        xml.addLine("RunID",        rs.getLong(simulationRunPKFieldCol));
        xml.addLine("RunAt",        rs.getString(simulationRunAtFieldCol));
        xml.addLine("RunBy",        rs.getString(researcherNameFieldCol));
        xml.addLine("Simulation",   rs.getString(simulationTitleFieldCol));
        xml.addLine("ParameterSet", rs.getString(paramSetNameFieldCol));
        xml.addLine("ProbeSet",     rs.getString(probeSetNameFieldCol));
        xml.addLine("truthValue",   rs.getString(truthValueFieldCol));
      xml.closeTag();      
    }
    
    rs.close();
    st.close();
  }
  
  /**
   * Adds the Simulations and their Details to the Summary
   * @param xml XMLAssist object constructing the XML document
   * @param summaryPK primary key of the summary being listed
   * @param con database connection
   * @throws Exception
   */
  private void addSummarySimulationsAndDetails(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    ABCMSummarySimulationsAndDetailsView view = DB.getSummarySimulationsAndDetailsView(); 
    // Get the list of included simulations
    String sql = DB.getSummarySimulationsAndDetailsView().getSelectSQL(DB.getOwner(), 
        AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK) + " " + 
        AbstractDBGateway.orderBy(view.getField_Simulation_Title().getName()));
    
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);
    
    int simulationIDFieldCol = rs.findColumn(view.getField_Simulation_PK().getName());
    
    SimulationManager sm = new SimulationManager(DB);
    
    while(rs.next()){
      xml.addBlock(sm.getInstanceXML(rs.getLong(simulationIDFieldCol), 
                                     AbstractDBManager.EXCLUDE_INSTANCE_HEADER, 
                                     AbstractDBManager.INCLUDE_COMMENTS), XMLAssist.INDENT_BLOCK);
    }
    
    rs.close();
    st.close();
  }

  /**
   * Adds Data Models and their details to the summary
   * @param xml XMLAssist object constructing the XML document
   * @param summaryPK primary key of the summary being listed
   * @param con database connection
   * @throws Exception
   */
  private void addSummaryDataModelsAndDetails(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    ABCMSummaryDataModelsAndDetailsView view = DB.getSummaryDataModelsAndDetailsView();
    
    String sql = view.getSelectSQL(DB.getOwner(), 
        AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK) + " " + 
        AbstractDBGateway.orderBy(view.getField_DataModel_Name().getName()));
    
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);
    
    int summaryPKFieldCol                   = rs.findColumn(view.getField_Summary_PK().getName());
    int dataModelPKFieldCol                 = rs.findColumn(view.getField_DataModel_PK().getName());
    int dataModelNameFieldCol               = rs.findColumn(view.getField_DataModel_Name().getName());
    int dataModelDescFieldCol               = rs.findColumn(view.getField_DataModel_Description().getName());
    int dataModelDateCreatedFieldCol        = rs.findColumn(view.getField_DataModel_DateCreated().getName());    
    int historyTitleFieldCol                = rs.findColumn(view.getField_History_Title().getName());
    int simulationConfigurationNameFieldCol = rs.findColumn(view.getField_SimulationConfiguration_Name().getName());
    int researcherNameFieldCol              = rs.findColumn(view.getField_Researcher_Name().getName());
    
    while(rs.next()){
      ABCMDataModelsWithNamesOfElementsView dataModelsListView = DB.getDataModelsWithNamesOfElementsView();
      sql = dataModelsListView.getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(dataModelsListView.getField_DataModel_PK().getName(), rs.getLong(dataModelPKFieldCol)));
      Statement st2 = con.createStatement();
      ResultSet rs2 = st2.executeQuery(sql);
      
      int canBeEditedFieldCol                  = rs2.findColumn(dataModelsListView.getField_CanBeEdited().getName());
      int combinedTruthValueFieldCol           = rs2.findColumn(dataModelsListView.getField_CombinedTruthValue().getName());
      int dateModelDateCreatedFieldCol         = rs2.findColumn(dataModelsListView.getField_DataModel_DateCreated().getName());
      int dataModelListDescFieldCol            = rs2.findColumn(dataModelsListView.getField_DataModel_Desc().getName());
      int dataModelListNameFieldCol            = rs2.findColumn(dataModelsListView.getField_DataModel_Name().getName());
      int dataModelListPKFieldCol              = rs2.findColumn(dataModelsListView.getField_DataModel_PK().getName());
      int dataModelListHistoryPKFieldCol       = rs2.findColumn(dataModelsListView.getField_History_PK().getName());
      int dataModelListHistoryTitleFieldCol    = rs2.findColumn(dataModelsListView.getField_History_Title().getName());
      int dataModelListResearcherNameFieldCol  = rs2.findColumn(dataModelsListView.getField_Researcher_Name().getName());
      int dataModelListSimConfigNameFieldCol   = rs2.findColumn(dataModelsListView.getField_SimulationConfiguration_Name().getName());
      int dataModelListSimConfigPKFieldCol     = rs2.findColumn(dataModelsListView.getField_SimulationConfiguration_PK().getName());
      int dataModelListTruthValueFieldCol      = rs2.findColumn(dataModelsListView.getField_TruthValue().getName());
      int dataModelListTruthValuePKFieldCol    = rs2.findColumn(dataModelsListView.getField_TruthValue_PK().getName());
      
      xml.open("dataModel");
        if(rs2.next()){
          xml.addLine("dataModelTitle",       rs2.getString(dataModelListNameFieldCol));
          xml.addLine("dataModelDescription", rs2.getString(dataModelListDescFieldCol));
          xml.addLine("historyTitle",         rs2.getString(dataModelListHistoryTitleFieldCol));
          xml.addLine("configurationTitle",   rs2.getString(dataModelListSimConfigNameFieldCol));
          xml.addLine("dataModelCreatedBy",   rs2.getString(dataModelListResearcherNameFieldCol));
          xml.addLine("truthValue",           rs2.getString(dataModelListTruthValueFieldCol));
        }
      xml.closeTag(); // dataModel        
      rs2.close();
      st2.close();        
    }
    rs.close();
    st.close();
  }
  
  /**
   * Adds Histories and their details to the Summary
   * @param xml XMLAssist object constructing the XML document
   * @param summaryPK primary key of the summary being listed
   * @param con database connection
   * @throws Exception
   */
  private void addSummaryHistoryList(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    ABCMSummaryHistoriesAndDetailsView view = DB.getSummaryHistoriesAndDetailsView();
    
    String sql = view.getSelectSQL(DB.getOwner(), AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK));
    
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);
    
    int historyPKFieldCol      = rs.findColumn(view.getField_History_PK().getName());
    
    HistoryManager HM = new HistoryManager(DB);
    
    while(rs.next()){
      xml.addBlock(HM.getInstanceXML(rs.getInt(historyPKFieldCol), AbstractDBManager.EXCLUDE_INSTANCE_HEADER, AbstractDBManager.INCLUDE_COMMENTS), XMLAssist.INDENT_BLOCK );
    }
    rs.close();
    st.close();
         
  }
  
  private void addSummaryHistoriesDetails(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    ABCMSummaryHistoriesAndDetailsView view = DB.getSummaryHistoriesAndDetailsView();
    
    String sql = view.getSelectSQL(DB.getOwner(), AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK));
    
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);
    
    int historyPKFieldCol      = rs.findColumn(view.getField_History_PK().getName());
    
    HistoryManager HM = new HistoryManager(DB);
    
    while(rs.next()){
      int histPK = rs.getInt(historyPKFieldCol);
      xml.addBlock(HM.getInstanceXML(histPK, AbstractDBManager.EXCLUDE_INSTANCE_HEADER, AbstractDBManager.INCLUDE_COMMENTS), XMLAssist.INDENT_BLOCK );
    }
    rs.close();
    st.close();
    
  }
  /**
   * Adds Narratives and their Details to the Summary
   * @param xml XMLAssist object constructing the XML document
   * @param summaryPK primary key of the summary being listed
   * @param con database connection
   * @throws Exception
   */
  private void addSummaryNarrativesAndDetails(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    
    ABCMSummaryNarrativesAndDetailsView view = DB.getSummaryNarrativesAndDetailsView();
    
    String sql = view.getSelectSQL(DB.getOwner(), AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK));
    
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);

    int summaryPKFieldCol             = rs.findColumn(view.getField_Summary_PK().getName());
    int narrativePKFieldCol           = rs.findColumn(view.getField_Narrative_PK().getName());
    int narrativeNameFieldCol         = rs.findColumn(view.getField_Narrative_Name().getName());
    int narrativeBaseDateFieldCol     = rs.findColumn(view.getField_Narrative_BaseDate().getName());
    int narrativeDescFieldCol         = rs.findColumn(view.getField_Narrative_Description().getName());
    int narrativeIsSeasonalFieldCol   = rs.findColumn(view.getField_Narrative_IsSeasonal().getName());
    int researcherNameFieldCol        = rs.findColumn(view.getField_Researcher_Name().getName());
    int narrativeDateCreatedFieldCol  = rs.findColumn(view.getField_Narrative_DateCreated().getName());
    
    Vector narrList = new Vector();
    
    while(rs.next()){
      narrList.add(0, new Integer(rs.getInt(narrativePKFieldCol)));
      xml.open("narrativeInList");
        xml.addLine("ID", rs.getString(narrativePKFieldCol));
        xml.addLine("name", rs.getString(narrativeNameFieldCol));
        xml.addLine("baseDate",  SimDate.convertToString(rs.getDouble(narrativeBaseDateFieldCol)));
        xml.addLine("researcher", rs.getString(researcherNameFieldCol));
        xml.addLine("dateCreated", rs.getString(narrativeDateCreatedFieldCol));
        xml.addLine("narrativeDescription", rs.getString(narrativeDescFieldCol));
        xml.addLine("isSeasonal", (rs.getBoolean(narrativeIsSeasonalFieldCol) ? "Yes" : "No"));
        CommentManager CM = new CommentManager(DB);
        xml.open("narrativeComments");
          xml.addBlock(CM.getListOfCommentsForObjectAsXML(AbstractABCMDBGateway.NARRATIVE, rs.getInt(narrativePKFieldCol)), true);
        xml.closeTag();
      xml.closeTag();            
    }
    rs.close();
    st.close();
  }

  private void addSummaryCodeModelsAndDetails(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    
    ABCMSummaryCodeModelsAndDetailsView view = DB.getSummaryCodeModelsAndDetailsView();
    
    String sql = view.getSelectSQL(DB.getOwner(), AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK));
    
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);
    
    int IDCol          = rs.findColumn(view.getField_CodeModel_PK().getName());
    int titleCol       = rs.findColumn(view.getField_CodeModel_Title().getName());
    int descCol        = rs.findColumn(view.getField_CodeModel_Desc().getName());
    int researcherCol  = rs.findColumn(view.getField_Researcher_Name().getName());
    int dateCreatedCol = rs.findColumn(view.getField_CodeModel_DateCreated().getName());
    
    
    CodeModelManager CM = new CodeModelManager(DB);
  
    while(rs.next()){
        xml.addBlock(CM.getInstanceXML(rs.getInt(IDCol)), true);    
    }
    return;
  }
  
  /**
   * Adds Configurations and their details to the Summary
   * @param xml XMLAssist object constructing the XML document
   * @param summaryPK primary key of the summary being listed
   * @param con database connection
   * @throws Exception
   */
  private void addSummaryConfigurationsAndDetails(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    // Get the list of Configurations
    ABCMSummaryConfigurationsAndDetailsView view = DB.getSummaryConfigurationsAndDetailsView();
   
    String sql = view.getSelectSQL(DB.getOwner(), AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK));
    
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);
    
    int summaryPKFieldCol                   = rs.findColumn(view.getField_Summary_PK().getName());
    int summarySimConfigPKFieldCol          = rs.findColumn(view.getField_SimulationConfiguration_PK().getName());
    int summarySimConfigNameFieldCol        = rs.findColumn(view.getField_SimulationConfiguration_name().getName());
    int summarySimConfigDescFieldCol        = rs.findColumn(view.getField_SimulationConfiguration_Desc().getName());
    int summaryResearcherNameFieldCol       = rs.findColumn(view.getField_Researcher_Name().getName());
    int summaryResearcherPKFieldCol         = rs.findColumn(view.getField_Researcher_PK().getName());
    int summarySimConfigDateCreatedFieldCol = rs.findColumn(view.getField_SimulationConfiguration_DateCreated().getName());
    int summarySimConfigTruthValueFieldCol  = rs.findColumn(view.getField_TruthValue().getName());
    
    while(rs.next()){
      xml.open("configurationInList");
        xml.addLine("ID",          rs.getString(summarySimConfigPKFieldCol));
        xml.addLine("name",        rs.getString(summarySimConfigNameFieldCol));
        xml.addLine("createdBy",   rs.getString(summaryResearcherNameFieldCol));
        xml.addLine("desc",        rs.getString(summarySimConfigDescFieldCol));
        xml.addLine("truthValue",  rs.getString(summarySimConfigTruthValueFieldCol));
    
        // And a detail block that must be supplied by the extensions to this abstract class
        xml.addBlock(getSpecificConfigurationsList(rs.getLong(summarySimConfigPKFieldCol)).getXML(XMLAssist.EXCLUDE_HEADER), true);
      xml.closeTag();
    }

    rs.close();
    st.close();

  }
  
  /**
   * Adds Parameter Sets and their details to the Summary
   * @param xml XMLAssist object constructing the XML document
   * @param summaryPK primary key of the summary being listed
   * @param con database connection
   * @throws Exception
   */
  private void addParameterSetsAndDetails(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    // Get the list of parameter Sets
    ABCMSummaryParameterSetsView view = DB.getSummaryParameterSetsView();
    
    String sql = view.getSelectSQL(DB.getOwner(), AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK));
    
    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);
    
    int summaryPKFieldCol  = rs.findColumn(view.getField_Summary_PK().getName());
    int paramSetIDFieldCol = rs.findColumn(view.getField_ParameterSet_PK().getName());
    
    ParameterSetManager psm = new ParameterSetManager(DB);
    while(rs.next()){
      xml.addBlock(psm.getInstanceXML(rs.getLong("intParameterSetID"), 
          AbstractDBManager.EXCLUDE_INSTANCE_HEADER, AbstractDBManager.INCLUDE_COMMENTS), XMLAssist.INDENT_BLOCK);
    }
    
    rs.close();
    st.close();
  }
    
  /**
   * Adds Probe Sets and their Details to the Summary
   * @param xml XMLAssist object constructing the XML document
   * @param summaryPK primary key of the summary being listed
   * @param con database connection
   * @throws Exception
   */
  private void addProbeSetsAndDetails(XMLAssist xml, long summaryPK, Connection con) throws Exception{
    ABCMSummaryProbeSetsView view = DB.getSummaryProbeSetsView();
    
    String sql = view.getSelectSQL(DB.getOwner(), AbstractDBGateway.where(view.getField_Summary_PK().getName(), summaryPK)); 

    Statement st = con.createStatement();
    ResultSet rs = st.executeQuery(sql);
    
    int summaryPKFieldCol = rs.findColumn(view.getField_Summary_PK().getName());
    int probeSetPKFieldCol = rs.findColumn(view.getField_ProbeSet_PK().getName());
    
    ProbeSetManager prsm = new ProbeSetManager(DB);
    
    while(rs.next()){
      xml.addBlock(prsm.getInstanceXML(rs.getLong(probeSetPKFieldCol), 
          AbstractDBManager.EXCLUDE_INSTANCE_HEADER, AbstractDBManager.INCLUDE_COMMENTS), XMLAssist.INDENT_BLOCK);
    }

    rs.close();
    st.close();          
  }
  
//  /**
//   * Adds Analyses to the Summary
//   * @param xml XMLAssist object constructing the XML document
//   * @param summaryPK primary key of the summary being listed
//   * @param con database connection
//   * @throws Exception
//   */
//  private void addAnalysesToSummary(XMLAssist xml, long summaryPK, Connection con) throws Exception{
//    // Display each analysis
//    ABCMSummaryAnalysesTable table = DB.getSummaryAnalysesTable();
//    
//    String sql = table.getSelectSQL(DB.getOwner(), AbstractDBGateway.where(table.getField_SummaryFK().getName(), summaryPK));
//    
//    Statement st = con.createStatement();
//    ResultSet rs = st.executeQuery(sql);
//    
//    int summaryFKFieldCol  = rs.findColumn(table.getField_SummaryFK().getName());
//    int analysisFKFieldCol = rs.findColumn(table.getField_AnalysisFK().getName());
//    
//    AnalysisManager am = new AnalysisManager(DB);
//    while(rs.next()){
//      xml.addBlock(am.getInstanceXML(rs.getLong(analysisFKFieldCol), 
//          AbstractDBManager.EXCLUDE_INSTANCE_HEADER, AbstractDBManager.INCLUDE_COMMENTS), XMLAssist.INDENT_BLOCK);
//    }
//
//    rs.close();
//    st.close();
//  }
  
  // Abstract methods that must be provided by implementing classes
  
  /**
   * Returns the list of specific configurations included in this summary;
   *  Returns the list in an xml format that can be inserted into the summary.
   * @param configID PK of Configuration
   * @return the list of specific configurations included in this summary, in XML format
   */
  public abstract XMLAssist getSpecificConfigurationsList(long configID);
  
  /**
   * Returns the specific details of the specific configurations in this
   * summary. Returns this in an xml format that can be inserted into the summary.
   * @param PK PK of Configuration
   * @return details of the configuration, in XML format
   */
  public abstract XMLAssist getSpecificConfigurationsDetails(long PK);
  
  

  /* Saving */
  
  // Note that only names and user IDs are created when this is invoked;
  // descriptions, summaries, conclusions are added later.
  
  /**
   * Sets an instance of a Summary to be saved. Instance includes only
   * summary name and user ID; other elements are added later
   * @param newSummaryName name of summary
   * @param userID ID of user who will 'own' summary
   */
  public void setInstanceToSave(String newSummaryName, long userID){
    summaryNameToSave = newSummaryName;
    userIDToSave         = userID;
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  public void clearInstanceToSave(){
    summaryNameToSave = null;
    userIDToSave         = -1;
  }
  
  /**
   * Determines if the instance to save has been set and is valid
   * @return true if the instance to save is valid
   */
  public boolean instanceToSaveIsValid(){
    return ((summaryNameToSave  != null) && 
            (userIDToSave       != -1));
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  public ABCMdbDatum datumToSave(){
    return DB.getSummariesTable().getDatum(summaryNameToSave, "", "", "", "" + userIDToSave, DB.currentTimeString());    
  }
  
  
  /* Deleting */
  
  /**
   * Deletes a Summary
   * @param PK Primary Key of Summary to be deleted
   * @param user ID (PK) of user to be deleted
   * @return TISFO
   */
  public boolean deleteItem(long PK, long user){
    return permitModification(PK, user)      &&                                                 // Validate
           SAM.deleteElementsByPrimaryFK(PK) &&                                                 // remove child records
           super.deleteItem(PK, user);  // and delete from main table
  }  

  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName() {    return "xmlSummaryListToHTML.xsl"; }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlSummaryToHTML.xsl"; }
  
  
  /* Auditing */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField baseTableOwnerField(){        return DB.getSummariesTable().getField_ResearcherFK();           }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView(){ return DB.getCanBeEdited_SummariesView();                        }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField canBeEditedViewPKField(){     return DB.getCanBeEdited_SummariesView().getField_Summary_PK();  } 
  
  
  /* Special */
  
  /**
   * Modify the description, summary, and conclusion of a Summary stored in the HWM Database
   * @param summaryPK primary key of the Summary whose description is to be modified
   * @param desc new description
   * @param summary new summary
   * @param conclusion new conclusion
   * @param user user ID to be used to confirm that the user has edit permission
   * @return true if successful, false if not
   */
  public boolean modifyDescriptionAndConclusions(long summaryPK, String desc, String summary, String conclusion, long user){
    return modifyDescription(summaryPK, desc, user)        && 
           modifySummary(summaryPK, summary, user)         && 
           modifyConclusions(summaryPK, conclusion, user);
  }
  
  /**
   * Modify the description of a Summary stored in the HWM Database
   * @param summaryPK primary key of the Summary whose description is to be modified
   * @param desc new description
   * @param user user ID to be used to confirm that the user has edit permission
   *  @return true if successful, false if not
   */
  public boolean modifyDescription(long summaryPK, String desc, long user){
    return permitModification(summaryPK, user) &&
           DB.updateItem(DB.getSummariesTable(), DB.getSummariesTable().getField_PK(), summaryPK, 
                         DB.getSummariesTable().getField_Description(), desc);
  }
  
  /**
   * Modify the summary of a Summary stored in the HWM Database
   * @param summaryPK primary key of the Summary whose description is to be modified
   * @param summary new summary
   * @param user user ID to be used to confirm that the user has edit permission
   * @return true if successful, false if not
   */
  public boolean modifySummary(long summaryPK, String summary, long user){
    return permitModification(summaryPK, user) &&
           DB.updateItem(DB.getSummariesTable(), DB.getSummariesTable().getField_PK(), summaryPK, 
                         DB.getSummariesTable().getField_Summary(), summary);    
  }
  
  /**
   * Modify the conclusion of a Summary stored in the HWM Database
   * @param summaryPK primary key of the Summary whose description is to be modified
   * @param conclusion new conclusion
   * @param user user ID to be used to confirm that the user has edit permission
   * @return true if successful, false if not
   */
  public boolean modifyConclusions(long summaryPK, String conclusion, long user){
    return permitModification(summaryPK, user) &&
           DB.updateItem(DB.getSummariesTable(), DB.getSummariesTable().getField_PK(), summaryPK, 
                         DB.getSummariesTable().getField_Conclusion(), conclusion);
  }
  
  /**
   * Add an Analysis to the specified Summary.
   * @param summaryPK primary key of the Summary to which the Analysis is to be added
   * @param analysisPK primary key of the Analysis to be added
   * @param desc description of this Analysis's contribution to the Summary
   * @param order ordinal position at which this analysis should be entered
   * @param user user ID to be used to confirm that the user has edit permission
   * @return true if successful, false if not
   */
  public boolean addAnalysis(long summaryPK, long analysisPK, String desc, int order, long user){
    SAM.setInstanceToSave(summaryPK, analysisPK, desc, order);
    return permitModification(summaryPK, user) && 
           SAM.saveInstanceToDB();   
  }
  
  /**
   * Removes an Analysis from the specified Summary. Note that this will return
   * true if no error is encountered, even if the Analysis is not present in the
   * specified Summary
   * @param summaryPK      primary key of summary
   * @param analysisOrder  order number of analysis to be removed
   * @param user           user ID to be used to confirm that the user has edit permission
   * @return true if successful, false otherwise
   */
  public boolean removeAnalysis(long summaryPK, long analysisOrder, long user){
    return permitModification(summaryPK, user) && 
           SAM.deleteElementsByBothPrimaryKeys(summaryPK, analysisOrder);
  }
    
  /**
   * Change the order in which an analysis appears in this summary. Does not change
   * the description of the analysis (and hence is better than having to remove and replace
   * it)
   * @param summaryPK PK of the summary affectedd
   * @param oldOrder  old ordinal position of the analysis
   * @param newOrder  new ordinal position of the analysis
   * @param user      ID of the user making this changed; used to confirm that the user has edit permission
   * @return true if successful, false otherwise
   */
  public boolean changeAnalysisOrder(long summaryPK, long oldOrder, long newOrder, long user){
    return permitModification(summaryPK, user) && SAM.changeOrder(summaryPK, oldOrder, newOrder);
  }
  
  public boolean modifyDescriptionOfAnAnalysis(long summaryPK, long analysisOrder, String newDescription, long user){
    return permitModification(summaryPK, user) && SAM.updateAnalysisDescription(summaryPK, analysisOrder, newDescription);
  }
  
  /**
   * Get the list of Analyses in the specified Summary as an HTML option list
   * @param summaryPK primary key of the specified Summary
   * @return a String containing the HTML Option list of the Analysis in the specified Summary
   */
  public String getAnalysesAsOptionList(long summaryPK){
    return SAM.getOptionList(summaryPK, false);    
  } 
  
  
  
  /**
   * Gets a summary details object for this summary; this allows quick (immutable) access 
   * to the description and other components of the summary without repeated database hits
   * @param summaryPK PK of the summary for which the summary details object is requested
   * @return the requested SummaryDetails object
   */
  public AbstractSummaryManager.SummaryDetails getSummaryDetailsObject(long summaryPK){ 
    return new SummaryDetails(summaryPK); 
  }
  
}
