package abcm.graphics;

import java.awt.Color;
import java.io.StringWriter;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import abcm.math.BasicMath;
import abcm.math.Point;
import abcm.utilities.TextFormatter;
import abcm.utilities.XMLAssist;

/**
 * Class that permits easy construction of well-formatted SVG documents.
 * 
 * The main advantage of this class is that it allows documents to be added
 * conveniently as subdocuments, while maintaining the information needed to make
 * each one visible to a human. For example, suppose you have created a document that
 * depicts a map legend. You would like to add it to any map, but the scale and size
 * of the map page (and even the orientation) may vary. Using this class you can
 * add a group and establish a new frame of reference inside the old one, but still
 * know how to scale text so that the font is legible and oriented up-down and left-to-right.
 * SVG does not natively provide this.
 * 
 * This class works in conjunction with the SVGRepresentable class. SVGRepresentables
 * include sequences of basic commands (addLine, addRectangle, etc.) that construct an
 * SVG drawing. In addition to these basic commands, groups can be added that contain
 * other SVGRepresentable objects.
 * 
 * Note that this documentation has not been double-checked.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SVGAssist {

  /* Inner classes */
  
  /**
   * Class svgFrameOfRef stores four values: the units per inch in horizontal
   * and vertical dimensions, and the width and height in inches of the current
   * viewport. A class like this is always available for an svg document, so that
   * objects that are attempting to figure out how to display themselves can get
   * the dimensions of the virtual screen on which they are going to be displayed.
   */
  class svgFrameOfRef{
    
    /**
     * X position of the top left corner
     */
    double topLeftX; 
    
    /**
     * Y position of the top left corner
     */
    double topLeftY; 
    
    /**
     * X position of the bottom right corner
     */
    double bottomRightX;
    
    /**
     * Y position of the bottom right corner
     */
    double bottomRightY;
    
    /**
     * Units in this frame of reference that equal one visible inch in the X dimension
     */
    double unitsPerInchX;
    
    /**
     * Units in this frame of reference that equal one visible inch in the Y dimension
     */
    double unitsPerInchY;
    
    /**
     * An ID for this frame of reference
     */
    String ID;
        
    /**
     * Basic constructor.
     * Note that the values given can result in 'flipping' the frame of reference- that
     * is, a value for xRight can be less than xLeft. The new frame of reference is
     * appropriately position and 'flipped'. This applies for both x and y dimensions.
     * @param xLeft x coordinate of the top left corner of this frame of reference, 
     * in the coordinate system in which it is to be placed 
     * @param yTop y coordinate of the top left corner of this frame of reference,
     * in the coordinate system in which it is to be placed
     * @param xRight x coordinate of the bottom right corner of this frame of reference,
     * in the coordinate system in which it is to be placed
     * @param yBottom y coordinate system of the bottom right corner of this frame of reference,
     * in the coordinate system in which it is to be placed
     * @param units_Per_Inch_X units in the new coordinate system that will equal one visible
     * inch, along the X dimension 
     * @param units_Per_Inch_Y units in the new coordinate system that will equal one visible
     * inch, along the Y dimension
     * @param Identifier A simple identifier for this frame of reference
     */
    public svgFrameOfRef(
        double xLeft, double yTop,
        double xRight, double yBottom,
        double units_Per_Inch_X, 
        double units_Per_Inch_Y, 
        String Identifier){
      
      topLeftX = xLeft; 
      topLeftY = yTop; 
      bottomRightX = xRight; 
      bottomRightY = yBottom;
      unitsPerInchX = units_Per_Inch_X;
      unitsPerInchY = units_Per_Inch_Y;
      ID = Identifier;
    }
    
    /**
     * Returns the point that marks the horizontal center of this frame of reference,
     * in the external coordinate system
     * @return the horizontal center
     */
    public double getXCenter(){ return (topLeftX + bottomRightX)/2d;}
    
    /**
     * Returns the point that marks the vertical center of this frame of reference,
     * in the external coordinate system
     * @return the vertical center
     */    
    public double getYCenter(){ return (topLeftY + bottomRightY)/2d;}
    
    
    /**
     * Gets the minimum X (which may be either top left or bottom right x coordinate),
     * in the external frame of reference
     * @return the minumum x value
     */
    public double getMinX()   { return Math.min(topLeftX, bottomRightX);   }
    
    /**
     * Gets the minimum Y (which may be either top left or bottom right x coordinate)
     * in the external frame of reference
     * @return the minumum y value
     */
    public double getMinY()   { return Math.min(topLeftY, bottomRightY);   }
    
    /**
     * Gets the absolute range in the external frame of reference in the x dimension
     * @return abs(right x - left x)
     */
    public double getXRange(){  return Math.abs(bottomRightX - topLeftX); }
    
    /**
     * Gets the absolute range in the external frame of reference in the y dimension
     * @return abs(right y - left y)
     */
    public double getYRange(){  return Math.abs(bottomRightY - topLeftY); }
    
    
    /**
     * Prints a simple string reporting the values of the instance variables of this frame of
     * reference. Useful mainly for debugging.
     */
    public void report(){
      System.out.print(" ID= " + ID);
      System.out.print(" x1 = " + BasicMath.roundDouble(topLeftX,3));
      System.out.print(" y1 = " + BasicMath.roundDouble(topLeftY,3));
      System.out.print(" x2 = " + BasicMath.roundDouble(bottomRightX,3));
      System.out.print(" y2 = " + BasicMath.roundDouble(bottomRightY,3));
      System.out.print(" upiX = " + BasicMath.roundDouble(unitsPerInchX,3));
      System.out.print(" upiY = " + BasicMath.roundDouble(unitsPerInchY,3));
      System.out.print(" vpwi = " + BasicMath.roundDouble(viewportWidthInInches(),3));
      System.out.print(" vphi= " + BasicMath.roundDouble(viewportHeightInInches(),3));
      System.out.println();
    }
    
    /**
     * Returns the width of the viewport for this frame of reference in inches
     * @return the width in inches
     */
    public double viewportWidthInInches() { return (Math.abs(bottomRightX - topLeftX) / unitsPerInchX);  }
    
    /**
     * Returns the height of the viewport for this frame of reference in inches
     * @return the height in inches
     */
    public double viewportHeightInInches(){ return (Math.abs(bottomRightY - topLeftY) / unitsPerInchY);  }

    /**
     * Determines whether the visible direction 'right' is positive within this frame of
     * reference. This is non-trivial, as it depends not only on whether this frame is 'flipped'
     * but whether the frame of reference in which it is nested (and all higher-level frames,
     * up to the top level) is also 'flipped'
     * @return true if increasing x moves right on the visible page, false otherwise
     */
    public boolean rightPositive(){ return bottomRightX > topLeftX; }
    
    /**
     * Determines whether the visible direction 'down' is positive within this frame of
     * reference. This is non-trivial, as it depends not only on whether this frame is 'flipped'
     * but whether the frame of reference in which it is nested (and all higher-level frames,
     * up to the top level) is also 'flipped'
     * @return true if increasing y moves downward on the visible page, false otherwise
     */
    public boolean downPositive() { return bottomRightY > topLeftY; }
    
  }
  
  /* Static Values */

  /**
   * Default width of an SVG page, in pixels (value is 1100)
   */
  public static final int    DEFAULT_SVG_PAGE_WIDTH_PIXELS       = 1100;
  
  /**
   * Default height of an SVG page, in pixels (value is 850)
   */
  public static final int    DEFAULT_SVG_PAGE_HEIGHT_PIXELS      =  850;
  
  /**
   * Default width of an SVG page, in inches (value is 11)
   */
  public static final double DEFAULT_SVG_PAGE_WIDTH_INCHES       =   11;
  
  /**
   * Default height of an SVG page, in inches (value is 8.5)
   */
  public static final double DEFAULT_SVG_PAGE_HEIGHT_INCHES      =    8.5;

  /**
   * Minimum value for the height and/or width of an SVG page, in pixels (value is 100) 
   */
  public static final int    MIN_PIXELS = 100;
  
  /**
   * Minimum value for the height and/or width of an SVG page, in inches (value is .5)
   */
  public static final double MIN_INCHES =    .5;
  
  /**
   * Default color for popup menus' borders (value is 'blue')
   */
  public static final String DEFAULT_POPUP_BORDER_STROKE         = "blue";
  
  /**
   * Default width for popup menus' borders (value is 3px)
   */
  public static final double DEFAULT_POPUP_BORDER_STROKE_WIDTH   = 3;      // Units are undefined
  
  /**
   * Default value for X offset from a popup menu to the point to which it is anchored;
   * value is .25 inches 
   */
  public static final double DEFAULT_POPUP_X_OFFSET_INCHES       = 0.25;
  
  /**
   * Default value for Y offset from a popup menu to the point to which it is anchored;
   * value is .25 inches 
   */  
  public static final double DEFAULT_POPUP_Y_OFFSET_INCHES       = 0.25;

  
  /* Static Methods */
  
  // SVG Document Preamble
  /**
   * Gets a well-formatted SVG preamble, with namespace and viewport declaration.
   * Note that 'inches' is presumed to be the 'real' unit of measurement, but in
   * fact this is arbitrary, too.
   * @param pageXSizeInPixels base width of the SVG viewport
   * @param pageYSizeInPixels base height of the SVG viewport
   * @param pageWidthInInches width of the SVG output, in inches (or other unit)
   * @param pageHeightInInches height of the SVG output, in inches (or other unit)
   * @return a well-formatted SVG preamble
   * 
   */
  public static String getSVGPreamble(int pageXSizeInPixels, int pageYSizeInPixels, 
      double pageWidthInInches, double pageHeightInInches){
    // NOTE REVISED 3/11/09- xmlns:svg -> xmlns
    return "xmlns=\"http://www.w3.org/2000/svg\" " +
           "xmlns:xlink=\"http://www.w3.org/1999/xlink\" " +
           "width=\"" + pageXSizeInPixels + "px\" height=\"" + pageYSizeInPixels + "px\" " +
           "viewBox=\"0 0 " + pageWidthInInches + " " + pageHeightInInches + "\"";
  }  
  
  /**
   * Makes a block of text within an SVG frame. Each line in the original text
   * becomes a line in the SVG.
   * @param x the x position at which to locate the text
   * @param y the y position at which to locate the text 
   * @param ySize the height of a line of the text (in points)
   * @param t the text to be placed
   * @param unitsPerInchY units per inch in the x direction in the framework in which this text is being placed
   * @param unitsPerInchX units per inch in the x direction in the framework in which this text is being placed
   * @return the SVG code required to make the block of text as needed
   */
  public static String makeText(double x, double y, 
      double ySize, String t, double unitsPerInchY, double unitsPerInchX){
    XMLAssist temp = new XMLAssist();
    temp.open("g", "transform=\"scale(" + unitsPerInchX/72 + "," + unitsPerInchY/72 + ")\"");
      temp.open("text","x=\"" + x + "\" y=\"" + y + "\" style=\"font-size:" + ySize + "pt;\"");
        temp.addBlock(SVGAssist.makeTextSpans(x, ySize+2, t), XMLAssist.INDENT_BLOCK); 
      temp.closeTag();
    temp.closeTag();
    return temp.getXML(XMLAssist.EXCLUDE_HEADER);
  }
  
  
  /**
   * Converts a string to a series of SVG text spans. Each
   * text span has an X starting point and a vertical size.
   * Spans are parsed into separate spans by carriage returns.
   * @param x the starting point for this text span
   * @param ySize the y size of this text span
   * @param t the string to be converted to spans
   * @return a string that consists of a series of text spans
   */
  public static String makeTextSpans(double x, double ySize, String t){
    String parse = t;
    parse = TextFormatter.removeDoubleLinefeeds(parse);
    parse += "\n";
    StringWriter out = new StringWriter();
    StringWriter temp = new StringWriter();
    while(parse.length() > 0){      
      String S = parse.substring(0, 1);
      if(S.equalsIgnoreCase("\n")){
        if(temp.toString().length() > 1) out.write(maketspan(x, ySize, temp.toString()) + "\n");
        temp = new StringWriter();
      }
      else{  temp.write(S);  }
      parse = parse.substring(1);
    }
    return out.toString();
  }
  
  /**
   * Simple tspan tag creator.
   * @param x x Position at which the text will be located
   * @param dy line height
   * @param t text to be included in the tag
   * @return the formatted tspan SVG tag, with x and dy attributes, and text included
   */
  
  public static String maketspan(double x, double dy, String t){
    return "<tspan x=\"" + x + "\" dy=\"" + dy +"\">" + t + "</tspan>";   
  }
   
  
  /**
   * Takes a java Color and returns a string representing it as an svg rgb attribute
   * i.e. rgb(100,100,100)
   * @param C java Color to be represented
   * @return rgb SVG tag
   */
  public static String getRGB(Color C){
    int r = C.getRed();
    int g = C.getGreen();
    int b = C.getBlue();
    return "rgb(" + r + "," + g + "," + b + ")";
  }
 
 
  // Assistance
  /**
   * Takes a collection of points and converts it into the string 
   * of points and coordinates in the required format.
   * @param points the points to be represented
   * @return a String representing these points in the format x1,y1 x2,y2 x3,y3 etc.
   */
  public static String getPointsList(Collection points){
    String pointsList = "";
    Iterator i = points.iterator();
    while(i.hasNext()){
      Point p = (Point)i.next();
      pointsList += p.getX() + ", " + p.getY() + " "; 
    }
    return pointsList;
  }
  
  
  /**
   * Returns an array of SVGRepresentable objects from a collection of those
   * objects. This is a convenience transformation only.
   * @param Coll collection of SVGRepresentable objects
   * @return array of SVGRepresentable objects
   */
  public static SVGRepresentable[] getArrayFromCollection(Collection Coll){
    SVGRepresentable[] a = new SVGRepresentable[Coll.size()];
    int c = 0;
    Iterator I = Coll.iterator();
    while(I.hasNext()){
      a[c] = (SVGRepresentable)I.next();
      c++;
    }
    return a;
  }
  
  
  /**
   * Returns an array of SVGRepresentable objects from a Hashtable of those
   * objects. This is a convenience transformation only.
   * @param H Hashtable of SVGRepresentable objects
   * @return array of SVGRepresentable objects
   */
  public static SVGRepresentable[] getArrayFromHashtable(Hashtable H){
    SVGRepresentable[] a = new SVGRepresentable[H.size()];
    int c = 0;
    Enumeration E = H.elements();
    while(E.hasMoreElements()){
      a[c] = (SVGRepresentable)E.nextElement();
      c++;
    }
    return a;
  }
  
  /**
   * Given a set of parameters, creates a new Frame Of References and pushes it onto
   * the stack of Frames of Reference
   * @param xLeft x position of top left corner, in original frame of reference
   * @param yTop y position of top left corner, in original frame of reference
   * @param xRight x position of bottom right corner, in original frame of reference
   * @param yBottom y position of bottom right corner, in original frame of reference
   * @param unitsPerInchX units per inch that will be used in the new frame of reference in the x direction
   * @param unitsPerInchY units per inch that will be used in the new frame of reference in the y direction
   * @param ID a unique ID for this frame of reference
   */
  /* Frame of Reference Management*/
  private void pushFOR(
      double xLeft, double yTop, double xRight, double yBottom, 
      double unitsPerInchX, double unitsPerInchY, String ID){

    pushFOR(new svgFrameOfRef(
        xLeft, yTop, xRight, yBottom, unitsPerInchX, unitsPerInchY, ID));  
  }
  
  /**
   * Pushes the specified frame of reference onto the 'stack' of Frames of Reference
   * @param svgFOR Frame of Reference to be pushed
   */
  private void pushFOR(svgFrameOfRef svgFOR){   FramesOfRef.add(0, svgFOR); }
  
  /**
   * Gets (but does not POP) the topmost frame of reference off the stack. Returns null if the stack is empty.
   * @return the topmost Frame of Reference
   */
  private svgFrameOfRef getFOR(){ return ((FramesOfRef.size() > 0 ) ? (svgFrameOfRef)FramesOfRef.get(0) : null);    }
  
  /**
   * Pops the topmost Frame of Reference off the stack. Returns null if the stack is empty
   * @return the topmost Frame of Reference
   */
  private svgFrameOfRef popFOR(){ return ((FramesOfRef.size() > 0 ) ? (svgFrameOfRef)FramesOfRef.remove(0) : null); }
  
  /* Instance Variables */
  
  /**
   * The XMLAssist object that constructs the bulk of the SVG document
   */
  private XMLAssist xml      = new XMLAssist();
  
  /**
   * The XMLAssist object that constructs the definitions portion of the SVG document
   */
  private XMLAssist xmlDefs  = new XMLAssist();
  
  /**
   * The collection (stack) of nested Frames of Reference
   */
  private Vector FramesOfRef = new Vector();
  
  /**
   * The collection of SVGRepresentable objects that will be drawn in this document
   */
  private Vector svg_objects = new Vector();
  
  /**
   * Width of the SVG page, in pixels
   */
  private int pageXSizeInPixels      = DEFAULT_SVG_PAGE_WIDTH_PIXELS;
  
  /**
   * Height of the SVG page, in pixels
   */
  private int pageYSizeInPixels      = DEFAULT_SVG_PAGE_HEIGHT_PIXELS;
  
  /**
   * Width of the SVG page, in inches
   */
  private double pageWidthInInches   = DEFAULT_SVG_PAGE_WIDTH_INCHES;
  
  /**
   * Height of the SVG page, in inches
   */
  private double pageHeightInInches  = DEFAULT_SVG_PAGE_HEIGHT_INCHES;
  
  
  /* Constructors */
  /**
   * Default constructor; sets all values to defaults
   */
  public SVGAssist(){   init();  }
  
  /**
   * Constructor with parameters
   * @param windowXSizeInPixels width of the SVG window, in pixels
   * @param windowYSizeInPixels height of the SVG window, in pixels
   * @param windowWidthInInches width of the SVG page, in inches
   * @param windowHeightInInches height of the SVG page, in inches
   */
  
  public SVGAssist(int windowXSizeInPixels, int windowYSizeInPixels, 
                   double windowWidthInInches, double windowHeightInInches){
    // Set instance variables
    pageXSizeInPixels      = windowXSizeInPixels;
    pageYSizeInPixels      = windowYSizeInPixels;
    pageWidthInInches      = windowWidthInInches;
    pageHeightInInches     = windowHeightInInches;
    init();
  }
  
  
  /**
   * Creates the definitions and the initial frame of reference
   *
   */
  private void init(){
    xmlDefs.open("defs");
    // Create the initial frame of reference; one inch is one unit
    svgFrameOfRef F = new svgFrameOfRef(0,0, pageWidthInInches, pageHeightInInches, 1,1, "BASE");
    pushFOR(F);    
  }

  /* Add an object to be depicted */
  /**
   * Adds an SVGRepresentable object to the collection of objects to be depicted
   * @param svgObj the SVGRepresentable object to be added
   */
  public void addSVGObject(SVGRepresentable svgObj){ svg_objects.add(svgObj); }
  
  /* Create the SVG document and return it in bytes */
  /**
   * Gets the SVG being created
   * @return the SVG document requested
   */
  public byte[] getSVG(){
    // Add the preamble
    xml.open("svg", SVGAssist.getSVGPreamble(pageXSizeInPixels, pageYSizeInPixels, 
                   pageWidthInInches, pageHeightInInches));
      // Add objects
      for(int i = 0; i < svg_objects.size(); i++){    
        ((SVGRepresentable)svg_objects.get(i)).addToSVG(this);     
      }
    xml.closeTag();
    return xml.getXML().getBytes();    
  }
  
  /* Scale information */
  // This is needed by some elements as the elements are being added
  /**
   * Gets the units per inch of the current Frame of Reference, in the X direction
   * @return units per inch X
   */
  public double getUnitsPerInchX()           { return (getFOR().unitsPerInchX); }
  
  /**
   * Gets the units per inch of the current Frame of Reference, in the Y direction
   * @return units per inch Y
   */
  public double getUnitsPerInchY()           { return (getFOR().unitsPerInchY); }
  
  /**
   * Gets the width of the current viewport in inches
   * @return width of the current viewport in inches
   */
  public double getViewportWidthInInches()   { return (getFOR().viewportWidthInInches()); }
  
  /**
   * Gets the height of the current viewport in inches
   * @return height of the current viewport in inches
   */
  public double getViewportHeightInInches()  { return (getFOR().viewportHeightInInches()); }
  
  /**
   * Gets the width of the current viewport in the units of that viewport
   * @return width of the current viewport in the units of that viewport
   */
  public double getViewportWidthInUnits()    { return getFOR().unitsPerInchX * getFOR().viewportWidthInInches(); }
  
  /**
   * Gets the height of the current viewport in the units of that viewport
   * @return height of the current viewport in the units of that viewport
   */  
  public double getViewportHeightInUnits()   { return getFOR().unitsPerInchY * getFOR().viewportHeightInInches(); }
  
  /* Add SVG Initial Elements (Scripts and Defs) */
  // These should appear before any of the other elements
 
  /**
   * Adds the `definitions' section to the SVG document, with the specified string as its
   * content
   * @param defString String containing the definitions required in this SVG document
   */
  // Definitions and Scripts
  public void addDef(String defString){
    xml.open("defs");
      xml.addBlock(defString,XMLAssist.INDENT_BLOCK);
    xml.closeTag();
  }
  
  /**
   * Adds a script to this SVG document
   * @param scriptName name of the script to be added
   */

  public void addScript(String scriptName){
    xml.addLine("script", "type=\"text/javascript\" xlink:href=\"" + scriptName + "\"", "");
  }
  
  /* Add basic SVG elements */
  
  // Basic SVG Items
  /**
   * Adds an SVG line in the current frame of reference.
   * @param x1 x1 coordinate
   * @param y1 y1 coordinate
   * @param x2 x2 coordinate
   * @param y2 y2 coordinate
   * @param stroke color
   * @param stroke_width width of line
   */
  public void addLine(double x1, double y1, double x2, double y2, String stroke, double stroke_width){
    addLine(x1, y1, x2, y2, stroke, stroke_width,"");
  }
  
  
  /**
   * Adds an SVG line in the current frame of reference.
   * @param x1 x1 coordinate
   * @param y1 y1 coordinate
   * @param x2 x2 coordinate
   * @param y2 y2 coordinate
   * @param stroke color
   * @param stroke_width width of line
   * @param otherTags any other attributes to be included (formatted as part of the svg line tag)
   */
  public void addLine(double x1, double y1, double x2, double y2, String stroke, double stroke_width, String otherTags){
    xml.addLine("line", "x1=\"" + x1 + "\" y1=\"" + y1 + "\" " +
                        "x2=\"" + x2 + "\" y2=\"" + y2 + "\" " +
                        "stroke=\"" + stroke + "\" " + 
                        "stroke-width=\"" + stroke_width + "\" " + otherTags, "");
  }
  
  
  /**
   * Adds an SVG circle
   * @param x x coordinate of the center of the circle
   * @param y y coordinate of the center of the circle
   * @param r radius
   * @param fill color of fill
   */
  public void addCircle(double x, double y, double r, String fill){
    addCircle(x,y,r, fill, "");
  }
  
  
  /**
   * Adds an SVG circle
   * @param x x coordinate of the center of the circle
   * @param y y coordinate of the center of the circle
   * @param r radius
   * @param fill color of fill
   * @param otherTags any other attributes to be included
   */
  public void addCircle(double x, double y, double r, String fill, String otherTags){    
    xml.addLine("<circle cx=\"" + x + "\" cy=\"" + y + "\" r=\"" + r + "\" fill=\"" + fill + "\" " + otherTags + "/>\n", true);  
  }
 
  
  /**
   * Adds a rectangle
   * @param x top left x coordinate
   * @param y top left right coordinate
   * @param width width
   * @param height height
   * @param stroke color
   * @param stroke_width width of border
   * @param fill fill color
   */
  public void addRectangle(double x, double y, double width, double height, 
      String stroke, double stroke_width, String fill){
    addRectangle(x, y, width, height, stroke, stroke_width, fill, "");
  }
  
  
  /**
   * Adds a rectangle
   * @param x top left x coordinate
   * @param y top left right coordinate
   * @param width width
   * @param height height
   * @param stroke color
   * @param stroke_width width of border
   * @param fill fill color
   * @param otherTags any other attributes to be included
   */
  public void addRectangle(double x, double y, double width, double height, 
      String stroke, double stroke_width, String fill, String otherTags){
      xml.addLine("rect", "x=\"" + x + "\" " + 
                        "y=\"" + y + "\" " +
                        "width=\"" + width + "\" " +
                        "height=\"" + height + "\" " +
                        "stroke=\"" + stroke + "\" " +
                        "stroke-width=\"" + stroke_width + "\" " +
                        "fill=\"" + fill + "\" " + otherTags,"");
  }
    
  
  /**
   * Adds a polygon. The pointsList can be created using the getPointsList method
   * @param pointsList a String representing the list of points
   * @param fill fill color for the polygon
   * @param stroke stroke color for the polygon
   * @param stroke_width width of the polygon stroke
   */
  public void addPolygon(String pointsList, String fill, String stroke, double stroke_width){
    addPolygon(pointsList, fill, stroke, stroke_width, "");  
  }
  
  
  /**
   * Adds a polygon. The pointsList can be created using the getPointsList method
   * @param pointsList a String representing the list of points
   * @param fill fill color for the polygon
   * @param stroke stroke color for the polygon
   * @param stroke_width width of the polygon stroke
   * @param otherTags any other attributes
   */
  public void addPolygon(String pointsList, String fill, String stroke, double stroke_width, String otherTags){
    xml.addLine("polygon",  "points=\"" + pointsList + "\" " +
                            "fill=\"" + fill + "\" " +
                            "stroke=\"" + stroke +"\" " +
                            "stroke-width=\"" + stroke_width + "\" " + otherTags,"");
  }
  
  
  /**
   * Adds a background color to any group
   * @param backgroundColor
   */
  public void addBackground(String backgroundColor){
    svgFrameOfRef f = getFOR();
    addRectangle(f.getMinX(), f.getMinY(), f.getXRange(), f.getYRange(), "none",0,backgroundColor, "opacity=\"1\"");
  }
  
  
  /**
   * Adds an image to any group
   * @param xPos top left X
   * @param yPos top left Y
   * @param width
   * @param height
   * @param location the location of the image file
   * @param aspectRatio SVG specification of how aspect ratio is to be handled (i.e. "xMidYMid")
   */
  public void addImage(double xPos, double yPos, double width, double height, String location, String aspectRatio){
    xml.addLine("image", "x=\"0\" y=\"0\" width=\"" + width + "\" height=\"" + height + "\" xlink:href=\"" + location + "\" preserveAspectRatio=\"" + aspectRatio + "\"","");  
  }  

  
  /**
   * Adds a 'use' tag in the current ref. A 'use' tag makes use of a previously
   * defined SVG object
   * @param xPos
   * @param yPos
   * @param href the ID by which the defined object is to be referred
   */
  public void addUse(double xPos, double yPos, String href){
    xml.addLine("use", "x=\"" + xPos + "\" y=\"" + yPos + "\" xlink:href=\"" + href + "\"","");
  }
 
  
  /**
   * Adds a comment in the SVG code
   * @param comment comment to be added
   */
  private void writeComment(String comment){ 
    xml.addBlock("\n<!-- " + comment + " -->", XMLAssist.INDENT_BLOCK); 
  }

  
  /**
   * Adds text that will be presented in the SVG document
   * @param text text to be added
   * @param margin margin to be left
   * @param maxFontSize largest font size to be used
   * @param linespacing spacing of the lines (dy)
   */
  public void addText(String text, double margin, int maxFontSize, int linespacing){
    double viewportHeight = getFOR().viewportHeightInInches() - margin * 2;
    double viewportWidth  = getFOR().viewportWidthInInches() - margin * 2;
    double ratio = viewportHeight / viewportWidth;
    int initialLineLength = TextFormatter.getLengthOfLongestLine(text);
    int len = initialLineLength;
    double modRatio;
    double textHeight;
    double textWidth;
    String mod;
    do{
      mod = TextFormatter.wordWrap(text, len);
      textHeight = TextFormatter.getTextHeight(mod, maxFontSize, linespacing);
      textWidth  = TextFormatter.getTextWidth(mod, maxFontSize);
      modRatio   = textHeight / textWidth;
      len--;
    }while((modRatio < ratio) && 
           ((textHeight > viewportHeight) || (textWidth > viewportWidth)) &&
           (len > 20));
    xml.addBlock(makeText(0,0,10,mod,getFOR().unitsPerInchY, getFOR().unitsPerInchX), XMLAssist.INDENT_BLOCK);
  }
  
  /* Add Group Elements*/
  // Framed Viewports
  
  /**
   * Adds a framed viewport. This essentially takes a subsection of the current window and
   * defines a new coordinate system. The new viewport is pushed onto the stack, and the units
   * per inch in the new coordinate system are calculated. A frame around it can be added, 
   * of a specified thickness
   * @param xPos X position (in the external coordinate system)
   * @param yPos Y position (in the external coordinate system)
   * @param windowXSize width, in the external coordinate system
   * @param windowYSize height, in the external coordinate system
   * @param windowNewLowerXBound lower X boundary of new coordinate system
   * @param windowNewLowerYBound lower Y boundary of new coordinate system
   * @param windowNewUpperXBound upper X boundary of new coordinate system
   * @param windowNewUpperYBound upper Y boundary of new coordinate system
   * @param frameThicknessInInches thickness of the frame, in inches 
   * @param borderColor color to be used for the border
   * @param svg_objects collection of SVG objects to be drawn inside this viewport
   * @param ID an identifier for this viewport
   */
  public void addFramedViewport(double xPos, double yPos, 
                     double windowXSize, double windowYSize, 
                     double windowNewLowerXBound, double windowNewLowerYBound, 
                     double windowNewUpperXBound, double windowNewUpperYBound,
                     double frameThicknessInInches, String borderColor,
                     SVGRepresentable[] svg_objects, String ID){
    
    addFramedViewport(xPos, yPos, windowXSize, windowYSize, windowNewLowerXBound, windowNewLowerYBound, windowNewUpperXBound, windowNewUpperYBound,
        frameThicknessInInches, borderColor, svg_objects, ID, "");    
  }
    

    /**
   * Adds a framed viewport. This essentially takes a subsection of the current window and
   * defines a new coordinate system. The new viewport is pushed onto the stack, and the units
   * per inch in the new coordinate system are calculated. A frame around it can be added, 
   * of a specified thickness
   * @param xPos X position (in the external coordinate system)
   * @param yPos Y position (in the external coordinate system)
   * @param windowXSize width, in the external coordinate system
   * @param windowYSize height, in the external coordinate system
   * @param windowNewLowerXBound lower X boundary of new coordinate system
   * @param windowNewLowerYBound lower Y boundary of new coordinate system
   * @param windowNewUpperXBound upper X boundary of new coordinate system
   * @param windowNewUpperYBound upper Y boundary of new coordinate system
   * @param frameThicknessInInches thickness of the frame, in inches 
   * @param borderColor color to be used for the border
   * @param svg_objects collection of SVG objects to be drawn inside this viewport
   * @param ID an identifier for this viewport
   * @param otherTags other parameters to be included
   */
  public void addFramedViewport(double xPos, double yPos, 
                     double windowXSize, double windowYSize, 
                     double windowNewLowerXBound, double windowNewLowerYBound, 
                     double windowNewUpperXBound, double windowNewUpperYBound,
                     double frameThicknessInInches, String borderColor,
                     SVGRepresentable[] svg_objects, String ID, String otherTags){

    addFramedViewport(xPos, yPos, windowXSize, windowYSize, windowNewLowerXBound, windowNewLowerYBound,
        windowNewUpperXBound, windowNewUpperYBound, frameThicknessInInches, borderColor, svg_objects,
        ID, otherTags,"");
  }
  
  
    /**
   * Adds a framed viewport. This essentially takes a subsection of the current window and
   * defines a new coordinate system. The new viewport is pushed onto the stack, and the units
   * per inch in the new coordinate system are calculated. A frame around it can be added, 
   * of a specified thickness
   * @param xPos X position (in the external coordinate system)
   * @param yPos Y position (in the external coordinate system)
   * @param windowXSize width, in the external coordinate system
   * @param windowYSize height, in the external coordinate system
   * @param windowNewLowerXBound lower X boundary of new coordinate system
   * @param windowNewLowerYBound lower Y boundary of new coordinate system
   * @param windowNewUpperXBound upper X boundary of new coordinate system
   * @param windowNewUpperYBound upper Y boundary of new coordinate system
   * @param frameThicknessInInches thickness of the frame, in inches 
   * @param borderColor color to be used for the border
   * @param svg_objects collection of SVG objects to be drawn inside this viewport
   * @param ID an identifier for this viewport
   * @param otherTags other parameters to be included as part of the group tag created
   * @param innerTags attributes to be assigned to the inner svg window
   */
  public void addFramedViewport(double xPos, double yPos, 
                 double windowXSize, double windowYSize, 
                 double windowNewLowerXBound, double windowNewLowerYBound, 
                 double windowNewUpperXBound, double windowNewUpperYBound,
                 double frameThicknessInInches, String borderColor,
                 SVGRepresentable[] svg_objects, String ID, String otherTags, String innerTags){

    
    xml.open("g", otherTags);
      double frameHorizontalThickness  = frameThicknessInInches * this.getUnitsPerInchX();
      double frameVerticalThickness    = frameThicknessInInches * this.getUnitsPerInchY();
      // Calculate the corners of the middle of the frame
      // Note that the actual external points are used in some places to make
      // sure that the corners are drawn like a rectangle; a rectangle is
      // not used, however, because the thicknesses of the horizontal and vertical lines
      // may vary
      double LeftX = xPos + frameVerticalThickness / 2d;  
      double RightX = xPos + windowXSize - frameVerticalThickness / 2d;
      double TopY = yPos + frameHorizontalThickness / 2d;
      double BottomY = yPos + windowYSize - frameHorizontalThickness / 2d;
      writeComment("Frame for " + ID);
      
      // Top Line
      addLine(xPos, TopY, xPos + windowXSize, TopY, borderColor, frameHorizontalThickness);
      // Bottom Line
      addLine(xPos, BottomY, xPos + windowXSize, BottomY, borderColor, frameHorizontalThickness);
      // Left Line
      addLine(LeftX, yPos, LeftX, yPos + windowYSize, borderColor, frameHorizontalThickness);
      // Right Line
      addLine(RightX, yPos, RightX, yPos + windowYSize,  borderColor, frameHorizontalThickness);
      // Now recalculate the corners for the inside of the frame
      LeftX   += frameVerticalThickness / 2d;
      TopY    += frameHorizontalThickness / 2d;
      RightX  -= frameVerticalThickness / 2d;
      BottomY -= frameHorizontalThickness / 2d;
      // Now create a new SVG Group
      this.addSVG(LeftX, TopY, (RightX - LeftX), (BottomY - TopY), 
                  windowNewLowerXBound, windowNewLowerYBound, 
                  windowNewUpperXBound, windowNewUpperYBound, 
                  svg_objects, "", innerTags);
    xml.closeTag();
  }

  /**
   * Adds an SVG viewport to the current frame of reference
   * @param xPos x position, in the current coordinate system
   * @param yPos y position, in the current coordinate system
   * @param windowXSize width, in the current coordinate system
   * @param windowYSize height, in the current coordinate system
   * @param windowNewLowerXBound new lower boundary of x coordinates in the new window
   * @param windowNewLowerYBound new lower boundary of y coordinates in the new window
   * @param windowNewUpperXBound new upper boundary of x coordinates in the new window
   * @param windowNewUpperYBound new upper boundary of y coordinates in the new window
   * @param svg_objects collection of SVGRepresentable objects to be depicted in this SVG window
   * @param ID a unique ID for this SVG window
   */  // SVG Elements
  
  public void addSVG(double xPos, double yPos, 
                     double windowXSize, double windowYSize, 
                     double windowNewLowerXBound, double windowNewLowerYBound, 
                     double windowNewUpperXBound, double windowNewUpperYBound,
                     SVGRepresentable[] svg_objects, String ID){
    addSVG(xPos, yPos, windowXSize, windowYSize, windowNewLowerXBound, windowNewLowerYBound,
        windowNewUpperXBound, windowNewUpperYBound, svg_objects, ID, "");
  }   
  
  
  
  /**
   * Adds an SVG viewport to the current frame of reference
   * @param xPos x position, in the current coordinate system
   * @param yPos y position, in the current coordinate system
   * @param windowXSize width, in the current coordinate system
   * @param windowYSize height, in the current coordinate system
   * @param windowNewLowerXBound new lower boundary of x coordinates in the new window
   * @param windowNewLowerYBound new lower boundary of y coordinates in the new window
   * @param windowNewUpperXBound new upper boundary of x coordinates in the new window
   * @param windowNewUpperYBound new upper boundary of y coordinates in the new window
   * @param svg_objects collection of SVGRepresentable objects to be depicted in this SVG window
   * @param ID a unique ID for this SVG window
   * @param otherTags other attributes to be assigned to this window
   */
  public void addSVG(double xPos, double yPos, 
                     double windowXSize, double windowYSize, 
                     double windowNewLowerXBound, double windowNewLowerYBound, 
                     double windowNewUpperXBound, double windowNewUpperYBound,
                     SVGRepresentable[] svg_objects, String ID, String otherTags){
  
    // Get the current frame of reference
    svgFrameOfRef svgFOR = getFOR();
    
    double xRange = Math.abs(windowNewUpperXBound - windowNewLowerXBound);
    double yRange = Math.abs(windowNewUpperYBound - windowNewLowerYBound);
    
    boolean reverseX = (windowNewUpperXBound < windowNewLowerXBound);
    boolean reverseY = (windowNewUpperYBound < windowNewLowerYBound);
        
    double newUnitsPerInchX = Math.abs(xRange / (windowXSize / svgFOR.unitsPerInchX));
    double newUnitsPerInchY = Math.abs(yRange / (windowYSize / svgFOR.unitsPerInchY));

    
    pushFOR(windowNewLowerXBound, windowNewLowerYBound, windowNewUpperXBound, windowNewUpperYBound,
         newUnitsPerInchX, newUnitsPerInchY, ID);
    
    writeComment(ID);
    // open an SVG tag
    xml.open("svg", "x=\""        + xPos + "\" " + 
                    "y=\""        + yPos + "\" " +
                    "width=\""    + windowXSize + "\" " +
                    "height=\""   + windowYSize + "\" " +
                    "viewBox=\" " + windowNewLowerXBound + " " + windowNewLowerYBound + " " + xRange  + " " + yRange + "\" " +
                    "preserveAspectRatio=\"xMidYMid\"");
      // A transform may be required here; this is so that ranges can go negative and it will be handled
      // automatically
      boolean doTransform = (reverseX || reverseY); 
      if(doTransform){
        double translateToX    = (reverseX ? 2 * windowNewLowerXBound : 0);
        double translateXScale = (reverseX ? -1                       : 1);

        double translateToY    = (reverseY ? 2 * windowNewLowerYBound : 0);
        double translateYScale = (reverseY ? -1                       : 1);

        xml.open("g", "transform=\"translate(" + translateToX + ", " + translateToY + ") scale(" + translateXScale + ", " + translateYScale + ")\" " + otherTags); 
      }
        // Depict all objects, passing this object (giving them access to the current frame of ref)
        for(int i = 0; i < svg_objects.length; i++){   svg_objects[i].addToSVG(this); }
      // close the g tag if there was a transform
      if(doTransform){ xml.closeTag(); }
    // close the svg tag
    xml.closeTag();
    // pop the frame of reference off the stack
    popFOR();
  }
  
  
  /* Groups */
  
  /**
   * Adds a group containing SVGRepresentable objects
   * @param xPos x position
   * @param yPos y position
   * @param ID unique identifier, that can be used by scripts
   * @param otherTags other attributes to this group
   * @param svg_objects collection of SVGRepresentable objects to be depicted
   */
  public void addG(double xPos, double yPos, String ID, String otherTags, SVGRepresentable[] svg_objects){
    addG(xPos, yPos, 0, 0, ID, otherTags, svg_objects);
  }
    
  
  /**
   * Adds a group at 0,0 (that is a group with no translate) containing SVGRepresentable objects
   * @param ID unique identifier, that can be used by scripts
   * @param otherTags other attributes to this group
   * @param svg_objects collection of SVGRepresentable objects to be depicted
   */
  public void addG(String ID, String otherTags, SVGRepresentable[] svg_objects){
    addG(0,0,0,0, ID, otherTags, svg_objects);
  }  
  
  
  
  /**
   * Adds a group containing an SVG object to be depicted. Group does
   * not translate (that is, it is at 0,0) nor rescale
   * @param ID unique identifier, that can be used by scripts
   * @param otherTags other attributes to this group
   * @param svg_object SVGRepresentable object to be depicted
   */
  public void addG(String ID, String otherTags, SVGRepresentable svg_object){
    addG(0,0,0,0, ID, otherTags, new SVGRepresentable[] { svg_object });
  }
    
  
  /**
   * Adds a group
   * @param xPos x position
   * @param yPos y position
   * @param newXScale new scale to be applied in the x dimension
   * @param newYScale new scale to be applied in the y dimension
   * @param ID unique identifier, that can be used in scripts
   * @param otherTags other attributes to this group
   * @param svg_object SVGRepresentable object to be depicted
   */
  public void addG(double xPos, double yPos, double newXScale, double newYScale, 
        String ID, String otherTags, SVGRepresentable svg_object){
    addG(xPos, yPos, newXScale, newYScale, ID, otherTags, new SVGRepresentable[] {svg_object} );
  }
    
 
  /**
   * Adds a group
   * @param xPos x position
   * @param yPos y position
   * @param newXScale new scale to be applied in the x dimension
   * @param newYScale new scale to be applied in the y dimension
   * @param ID unique identifier, that can be used in scripts
   * @param otherTags other attributes to this group
   * @param svg_objects SVGRepresentable objects to be depicted
   */  
  public void addG(double xPos, double yPos, double newXScale, double newYScale, 
        String ID, String otherTags, SVGRepresentable[] svg_objects){
    // Create the new frame of reference & push it onto the stack
    svgFrameOfRef svgFOR = getFOR();
    newXScale = (newXScale == 0 ? 1 : newXScale);
    newYScale = (newYScale == 0 ? 1 : newYScale);
    
    pushFOR(
          svgFOR.topLeftX - xPos, svgFOR.topLeftY - yPos,
          (svgFOR.bottomRightX - xPos) / newXScale, (svgFOR.bottomRightY - yPos) / newYScale,
          svgFOR.unitsPerInchX / newXScale, svgFOR.unitsPerInchY / newYScale, ID);
    
    // Open the g tag
    String translate = ((xPos != 0 && yPos != 0) ? "translate(" + xPos + "," + yPos + ")"       : "");
    String scale = ((newXScale * newYScale != 0) ? "scale(" + newXScale + "," + newYScale + ")" : "");
    String transform;
    if((translate == "") && (scale=="")) transform="";
    else{
      String spacer = "";
      if((translate != "") && (scale!="")) spacer = " ";
      transform="transform=\"" + translate + spacer + scale + "\"";
    }
    String spacer = "";
    if(otherTags != "") spacer=" ";
    xml.addBlock("<!-- " + ID + " -->", XMLAssist.INDENT_BLOCK);
    xml.open("g", transform + spacer + otherTags);
      // Depict all objects
      for(int i = 0; i < svg_objects.length; i++){   
        svg_objects[i].addToSVG(this);     
      }
    // close the g tag
    xml.closeTag();  
    // pop the frame of reference off the stack, if there was a new one at all
    popFOR();
  }

  /**
   * Automates the creation of a 'popup' window in a given viewbox. This is mostly
   * helpful when the viewbox is a map. The issue that this method solves is that the
   * map's coordinate space is scaled one way, while a displayable window must be scaled
   * appropriately for text.
   * 
   * This method does the following:
   * 
   * Accepts an x,y position as the anchor point of the popup window.
   * Accepts a recommended width and height of the popup window.
   * Determines the actual displayed width of the popup window. This will
   * take the recommendations as a maximum, but may rescale to smaller dimensions
   * if the viewport is not large enough to accommodate it.
   * Determines where the popup window should be positioned in relation to the
   * anchor point; if the point is left of the center of the viewport, the popup
   * window will appear to the right of the anchor point, and if it is above center,
   * the popup window will appear below, etc.
   * Accepts information about whether the calling viewport is oriented left-right or right-left, up-down
   * or down-up, and reverses the scale appropriately if needed. Also accepts an ID that can be used
   * by scripts to make the popup appear and be hidden
   * @param x x coordinate of the anchor point for this popup
   * @param y y coordinate of the anchor point for this popup
   * @param recommendedWidth 
   * @param recommendedHeight 
   * @param xOffset x offset from anchor point, in inches
   * @param yOffset y offset from anchor point, in inches 
   * @param content text to be placed in this popup
   * @param borderStroke color of border
   * @param borderWidth width of border
   * @param ID an ID for this popup window
   * @param otherTags used for popups that require other attributes (a popup is an SVG G tag)
   * 
   */
  public void addPopup(
      double x, double y, 
      double recommendedWidth, double recommendedHeight, 
      double xOffset, double yOffset, 
      SVGRepresentable[] content, 
      String borderStroke, double borderWidth, String ID, String otherTags){
    
    double windowWidth  = recommendedWidth; 
    double windowHeight = recommendedHeight;
    
    double mp = .40; // MaximumProportion of display area allowed to be covered by window
    
    double ratio = 1;
    
    if((windowWidth  > (mp * getViewportWidthInInches())) ||
       (windowHeight > (mp * getViewportHeightInUnits()))){
      ratio = Math.min(((mp * getViewportWidthInInches()) / recommendedWidth), ((mp * getViewportHeightInInches()) / recommendedWidth));
      windowWidth  *= ratio;
      windowHeight *= ratio;
    }
        
    boolean isLeft  = (x < getFOR().getXCenter());
    boolean isAbove = (y < getFOR().getYCenter());
    
    double xPos = x + (isLeft  ? xOffset : -1 * (xOffset + windowWidth)) * getUnitsPerInchX();
    double yPos = y + (isAbove ? yOffset : -1 * (yOffset + windowHeight))* getUnitsPerInchY();
    
    // If the FOR indicates that it is reflected, the routine will re-orient the frame as needed
    double topLeftX     = ((getFOR().rightPositive()) ? 0                      :      recommendedWidth  );
    double topLeftY     = (!(getFOR().downPositive()) ?      recommendedHeight : 0                      );
    double bottomRightX = ((getFOR().rightPositive()) ?      recommendedWidth  : 0                      );
    double bottomRightY = (!(getFOR().downPositive()) ? 0                      :      recommendedHeight );
    
    double upiX = getUnitsPerInchX();
    double upiY = getUnitsPerInchY();
    
    addFramedViewport(xPos, yPos, windowWidth * upiX, windowHeight * upiY,
           topLeftX, topLeftY, bottomRightX, bottomRightY, borderWidth, borderStroke,
           content, ID, otherTags);
            
  }  

}
