package abcm.graphics;

import java.util.ArrayList;
import java.util.HashSet;

import abcm.math.Point;
import abcm.utilities.XMLAssist;

/**
 * This class wraps basic X3D functions so that
 * classes that implement X3DRepresentable can
 * easily add themselves to an X3D document.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class X3DAssist {

  /** Navigation 'FLY'  */
  public static final String FLY      = "\"FLY\"";
  /** Navigation 'WALK'  */
  public static final String WALK     = "\"WALK\"";
  /** Navigation 'PAN'  */
  public static final String EXAMINE  = "\"PAN\"";
  /** Navigation 'SLIDE'  */
  public static final String SLIDE    = "\"SLIDE\"";  
  /** Navigation 'ANY'  */
  public static final String ANY      = "\"ANY\"";
  
  /** String describing the background for this scene  */
  protected String backgroundString = null;
  
  /** Array of Strings describing viewpoints  */
  protected ArrayList viewpointlist = new ArrayList();
  
  /** String describing the navigation options for this scene  */
  protected String navigationString = null;
  
  /** String describing the directional lighting options for this scene  */
  protected String directionalLightString = null; 
  
  
  /** Collection of Strings describing directional lighting by time of day  */
  protected String[] directionalLightingStringsByTimeOfDay = new String[]{
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='0.000 -1.000 -0.2'/>",   // 12:00 AM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='-0.133 -0.867 -0.2'/>",  // 1:00  AM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='-0.267 -0.733 -0.2'/>",  // 2:00  AM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='-0.400 -0.600 -0.2'/>",  // 3:00  AM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='-0.533 -0.467 -0.2'/>",  // 4:00  AM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='-0.667 -0.333 -0.2'/>",  // 5:00  AM
    "<DirectionalLight ambientIntensity='0.30' color='1 0.60 0.00' direction='0.800 -0.200 -0.2'/>",   // 6:00  AM
    "<DirectionalLight ambientIntensity='0.40' color='1 0.65 0.10' direction='0.667 -0.333 -0.2'/>",   // 7:00  AM
    "<DirectionalLight ambientIntensity='0.50' color='1 0.70 0.20' direction='0.533 -0.467 -0.2'/>",   // 8:00  AM
    "<DirectionalLight ambientIntensity='0.60' color='1 0.75 0.30' direction='0.400 -0.600 -0.2'/>",   // 9:00  AM
    "<DirectionalLight ambientIntensity='0.70' color='1 0.80 0.40' direction='0.267 -0.733 -0.2'/>",   // 10:00 AM
    "<DirectionalLight ambientIntensity='0.80' color='1 0.85 0.55' direction='0.133 -0.867 -0.2'/>",   // 11:00 AM
    "<DirectionalLight ambientIntensity='0.90' color='1 0.90 0.70' direction='0.000 -1.000 -0.2'/>",   // 12:00 PM
    "<DirectionalLight ambientIntensity='0.80' color='1 0.80 0.55' direction='-0.133 -0.867 -0.2'/>",  // 1:00  PM
    "<DirectionalLight ambientIntensity='0.70' color='1 0.70 0.40' direction='-0.267 -0.733 -0.2'/>",  // 2:00  PM
    "<DirectionalLight ambientIntensity='0.60' color='1 0.60 0.30' direction='-0.400 -0.600 -0.2'/>",  // 3:00  PM
    "<DirectionalLight ambientIntensity='0.50' color='1 0.50 0.20' direction='-0.533 -0.467 -0.2'/>",  // 4:00  PM
    "<DirectionalLight ambientIntensity='0.35' color='1 0.40 0.10' direction='-0.667 -0.333 -0.2'/>",  // 5:00  PM
    "<DirectionalLight ambientIntensity='0.20' color='1 0.30 0.00' direction='-0.800 -0.200 -0.2'/>",  // 6:00  PM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='0.667 -0.333 -0.2'/>",   // 7:00  PM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='0.533 -0.467 -0.2'/>",   // 8:00  PM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='0.400 -0.600 -0.2'/>",   // 9:00  PM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='0.267 -0.733 -0.2'/>",   // 10:00 PM
    "<DirectionalLight ambientIntensity='0.25' color='1 0.9 0.8' direction='0.133 -0.867 -0.2'/>"};  // 11:00 PM
  
  /** The X3D document being created  */
  private XMLAssist x3d            = new XMLAssist();
  
  /** A collection of items and their definitions to be added in the document's opening  */
  private HashSet   definitions    = new HashSet();
  
  /** A separate xml document for the creation of the definitions section  */
  private XMLAssist definitionsXML = new XMLAssist();
  
  /** The collection of objects to be added  */
  private ArrayList x3d_objects    = new ArrayList();
  
  /** A flag indicating whether any of the objects attempts to use the 'billboard' function  */
  private boolean usesBillboard    = false;

  /**
   * Gets the standard X3D preamble
   * @return the standard X3D preamble
   */
  public static String getX3DPreamble(){
    return "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema-instance\" " +
           "xsd:noNamespaceSchemaLocation=\"http://www.web3d.org/specifications/x3d-3.0.xsd\" " +
           "profile=\"Immersive\"";
  }
  
  /**
   * Returns the complete document as a byte[] array
   * @return the completed X3D document
   */
  public byte[] getX3D(){
    x3d.open("Group");  // Actually should be a transform to put everything centered on 0,0 and oriented n+/-, e+/-
      for(int i = 0; i < x3d_objects.size(); i++){    
        ((X3DRepresentable)x3d_objects.get(i)).addToX3D(this);     
      }
    x3d.closeTag();
    XMLAssist finalX3D = new XMLAssist();
    finalX3D.open("X3D", getX3DPreamble());
      finalX3D.open("Scene");
        finalX3D.open("Group");
          if(directionalLightString != null) x3d.addBlock(directionalLightString);
          finalX3D.addBlock(definitionsXML.getXML(true));
          ArrayList v = (ArrayList)viewpointlist.clone();
          while(v.size() > 0) finalX3D.addBlock((String)v.remove(0));
          if(backgroundString != null) finalX3D.addBlock(backgroundString);
          if(navigationString != null) finalX3D.addBlock(navigationString);
          finalX3D.addBlock(x3d.getXML(true));
        finalX3D.closeTag();
      finalX3D.closeTag();  // Scene
    finalX3D.closeTag(); // X3D
    return finalX3D.getXML().getBytes();
  }
  
  /**
   * Adds an x3d-representable object to this document
   * @param x3d the object to be added
   */
  public void addX3DObject(X3DRepresentable x3d){
    x3d_objects.add(x3d);  
  }
 
  /**
   * Adds a 'billboard' to the x3d environment.
   * @param p location at which billboard will be based
   * @param billboardText text to be displayed on the billboard
   */
  public void addBillboard(Point p, String billboardText){
    if(usesBillboard == false){
      System.out.println("Should be adding definitions here...");
      addBillboardDefinition();
      usesBillboard = true;
    }
    // First, convert the point to the VRML Coordinate system
    // Seven elements are needed:
    // Location
    String loc = p.getX() + " " + p.getZ() + " " + p.getY();
    
    // Text
    // The text passed will be delimited with \n,
    // but must be converted to separate strings.
    String[] outText = breakText(billboardText);
    String outputText = "";
    for(int i = 0; i< outText.length; i++){
      outputText += "\"" + outText[i] + "\"";
      if(i < outText.length - 1) outputText+=",";
    }
    
    int numberOfLines = outText.length;
    
    // How large will the plaque appear?
    // Standard height is 10 lines per meter, but max size is 3 m
    // and min size is 1 m
    double plaqueHeight = Math.max(1, Math.min(3, numberOfLines / 10));
    
    // Standard width is .5 m
    double plaqueWidth = .5;
    double plaqueThickness = .1;
    String plaqueSize = "" + plaqueWidth + " " + plaqueHeight + " " + plaqueThickness;
    
    // What is the position of the plaque anchor?
    // Will be 1/2 the height of the plaque plus one meter for the post
    // Expressed as an offset
    String plaqueAnchor = "0 " + (1 + plaqueHeight /2) + " 0";
    
    // What will be the scale of the text?
    // Usually this will be 1/10, but may be smaller;
    double scaleValue = (numberOfLines > 10 ? plaqueHeight / numberOfLines : .1);
    String textScale = "" + scaleValue + " " + scaleValue + " " + scaleValue;
    
    // How wide is the text allowed to be?
    // This is a single number, but must be passed as a string
    // to be applied to each line
    // Note that the text width is given scaled with the text
    double textWidthValue = plaqueWidth - .02;
    String textWidth = "" + textWidthValue / scaleValue;
    
    
    // Where will the text be positioned?
    // This is just a slight offset from the top left corner
    // of the plaque, based on the difference between plaque width and text width
    // Also, the text position must be the bottom of the first line;
    String textPosition = "" + (-1 * (plaqueWidth / 2) + ((plaqueWidth - textWidthValue)/2)) + " " +
                               ((plaqueHeight / 2) - 1 * scaleValue) + " " + 
                               " .1";  // The small z value puts the text in front of the plaque
    
    
    x3d.open("ProtoInstance", "name='Marker'");
      x3d.addLine("fieldValue", "name='markerPosition' value='" + loc + "'","");
      x3d.addLine("fieldValue", "name='displayText' value='" + outputText + "'","");
      x3d.addLine("fieldValue", "name='plaqueSize' value='" + plaqueSize + "'","");
      x3d.addLine("fieldValue", "name='plaqueAnchor' value='" + plaqueAnchor + "'","");
      x3d.addLine("fieldValue", "name='textWidth' value='" + textWidth + "'","");
      x3d.addLine("fieldValue", "name='textPosition' value='" + textPosition + "'","");
      x3d.addLine("fieldValue", "name='textScale' value='" + textScale + "'","");
    x3d.closeTag(); // Close the Proto instance    
  }
  
  /**
   * Breaks a text into smaller subtexts for easier display
   * @param text text to be parsed
   * @return text in smaller pieces
   */
  public static String[] breakText(String text){
    int lineCount = 0;
    text = text.replaceAll("\n\n","\n");
    String t = text + "\n";
    while(t.length() > 1 && lineCount < Integer.MAX_VALUE){
      if(t.contains("\n")) {
        lineCount++;
        if(t.length() > 1) t = t.substring(t.indexOf("\n") + 1);
        else t="";
      }
      else{
        t = "";
      }      
    }
    String[] ret = new String[lineCount];
    lineCount = 0;
    t = text + "\n";
    while(t.length() > 1 && lineCount < Integer.MAX_VALUE){
      if(t.contains("\n")) {
        ret[lineCount] = t.substring(0,t.indexOf("\n"));
        if(t.length() > 1) t = t.substring(t.indexOf("\n") + 1);
        else t="";
        lineCount++;
      }
      else{
        t = "";
      }      
    }
    return ret;
  }
  
  /**
   * Adds the definition of the 'billboard' object to the x3d document
   */
  private void addBillboardDefinition(){
    System.out.println("Adding definitions to definitions XMLAssist...");
      definitionsXML.open("ProtoDeclare","name='Marker'");
      definitionsXML.open("ProtoInterface");
        String AT = "accessType='initializeOnly' ";
        definitionsXML.addLine("field", "name='markerPosition' " + "type='SFVec3f' "  +  AT + " value='0.0  0.0 0.1'","");
        definitionsXML.addLine("field", "name='displayText' "    + "type='MFString' " +  AT + " value='*'"           ,"");
        definitionsXML.addLine("field", "name='plaqueSize' "     + "type='SFVec3f' "  +  AT + " value='0.5  1.0 0.1'","");
        definitionsXML.addLine("field", "name='plaqueAnchor' "   + "type='SFVec3f' "  +  AT + " value='0    1.5 0.0'","");
        definitionsXML.addLine("field", "name='textWidth' "      + "type='SFFloat' "  +  AT + " value='4.5'"         ,"");
        definitionsXML.addLine("field", "name='textPosition' "   + "type='SFVec3f' "  +  AT + " value='-.24 0.4 0.1'","");
        definitionsXML.addLine("field", "name='textScale' "      + "type='SFVec3f' "  +  AT + " value='0.1  0.1 0.1'","");
      definitionsXML.closeTag(); // Interface
      definitionsXML.open("ProtoBody");
        definitionsXML.open("Transform");
          definitionsXML.open("IS");
            definitionsXML.addLine("connect", "nodeField='translation' protoField='markerPosition'","");
          definitionsXML.closeTag();  // IS
          definitionsXML.open("Shape");
            definitionsXML.open("Appearance"); // Sphere Appearance
              definitionsXML.addLine("Material", "diffuseColor='0 1 0' emissiveColor='1 1 0'","");
            definitionsXML.closeTag(); // Appearance
            definitionsXML.addLine("Sphere", "radius='0.1'","");
          definitionsXML.closeTag(); // Shape
          definitionsXML.open("Transform", "translation='0 0.55 0'"); // Transform for post
            definitionsXML.open("Shape"); // Post shape
              definitionsXML.open("Appearance"); // Post appearance
                definitionsXML.addLine("Material", "diffuseColor='0 1 0' emissiveColor='0 1 1'","");
              definitionsXML.closeTag(); // Post Appearance
              definitionsXML.addLine("Cylinder", "radius='0.05' height='0.9'","");
            definitionsXML.closeTag(); // Post
          definitionsXML.closeTag(); // Close transform for post
          definitionsXML.open("Transform"); // Transform to plaque position
            definitionsXML.open("IS");
              definitionsXML.addLine("connect", "nodeField='translation' protoField='plaqueAnchor'","");
            definitionsXML.closeTag();
            definitionsXML.open("Billboard","axisOfRotation='0 0 0'"); // Open billboard group
              definitionsXML.open("Shape");
                definitionsXML.open("Appearance");
                  definitionsXML.addLine("Material", "diffuseColor='0 0 0' emissiveColor='0 0 1'","");
                definitionsXML.closeTag();
                definitionsXML.open("Box");
                  definitionsXML.open("IS");
                    definitionsXML.addLine("connect", "nodeField='size' protoField='plaqueSize'","");
                  definitionsXML.closeTag();
                definitionsXML.closeTag(); // Close the plaque box
              definitionsXML.closeTag(); // Close the plaque shape
              definitionsXML.open("Transform"); // Move to position Text
                definitionsXML.open("IS");
                  definitionsXML.addLine("connect", "nodeField='translation' protoField='textPosition'","");
                  definitionsXML.addLine("connect", "nodeField='scale' protoField='textScale'","");
                definitionsXML.closeTag();
                definitionsXML.open("Shape"); // Text Shape
                  definitionsXML.open("Appearance");
                    definitionsXML.addLine("Material", "diffuseColor='0 1 0' emissiveColor='1 1 1'","");
                  definitionsXML.closeTag();
                  definitionsXML.open("Text");
                    definitionsXML.open("IS");
                      definitionsXML.addLine("connect", "nodeField='string' protoField='displayText'","");
                      definitionsXML.addLine("connect", "nodeField='maxExtent' protoField='textWidth'","");
                    definitionsXML.closeTag();
                  definitionsXML.closeTag(); // Text
                definitionsXML.closeTag(); // Text Shape
              definitionsXML.closeTag(); // End transform to text position
            definitionsXML.closeTag(); // Close billboard group
          definitionsXML.closeTag(); // Transfrom to plaque position
        definitionsXML.closeTag(); // Transform to Marker Position
      definitionsXML.closeTag(); // ProtoBody
    definitionsXML.closeTag(); // Marker  
  }
  
  /**
   * Adds a definition; a definition is placed at the head of the x3d document.
   * @param definitionName name of the definition
   * @param definition text of the definition
   */
  public void addDefinition(String definitionName, String definition){
    if(!definitions.contains(definitionName)){
      System.out.println("Adding definition: " + definitionName);
      definitionsXML.addBlock(definition);
      definitions.add(definitionName);
    }
  }
  
  /**
   * Adds a NavigationInfo tag to the X3D text. Only one can exist, so re-setting
   * this overrides the original.
   * @param primaryNavigation one of the available navigation strings to be the primary one. Note: as a hack, may pass multiple
   * ones in a properly formatted String (i.e:"FLY" "EXAMINE" "WALK"- that is, include spaces between each one.)
   * @param any true if any kind of navigation is allowed, false otherwise. Adds '"ANY"' to the list of navigation possibilities.
   * @param speed speed to be used for initial navigation
   * @param headlight true if the headlight should be initially 'on', false if 'off'
   */
  public void addNavigation(String primaryNavigation, boolean any, long speed, boolean headlight){
    navigationString = "<NavigationInfo type='" + primaryNavigation + (any ? " " + ANY : "") + "' " + "speed='" + speed + "' headlight='" + headlight + "'/>"; 
  }
  
  /**
   * Adds directional lighting for this scene
   * @param intensity intensity of ambient light
   * @param colorRed 'red' portion of color
   * @param colorGreen 'green' portion of color
   * @param colorBlue 'blue' portion of color
   * @param directionX 'X' component of direction
   * @param directionY 'Y' component of direction
   * @param directionZ 'Z' component of direction
   */
  public void addDirectionalLight(double intensity, 
      double colorRed, double colorGreen, double colorBlue, 
      double directionX, double directionY, double directionZ){
    directionalLightString = "<DirectionalLight ambientIntensity='" + intensity + "' " +
                                                "color='" + colorRed + " " + colorGreen + " " + colorBlue + "' " +
                                                "direction='" + directionX + " " + directionY + " " + directionZ + "'/>";
                                                
  }
  
  /**
   * Adds 'sunlight' to the scene. 
   * @param timeOfDayAsFractionOf24Hours Time of day, expressed as a fraction of the 24-hour
   * day (i.e. noon = .5)
   */
  public void addSunlight(double timeOfDayAsFractionOf24Hours){
    System.out.print("Adding sunlight! ");
    directionalLightString = directionalLightingStringsByTimeOfDay[(int) ((timeOfDayAsFractionOf24Hours % 1) * 24)];
    System.out.println(directionalLightString);
  }
  
  /**
   * Adds a viewpoint. Note that the viewpoint is given in x,y,z coordinates
   * and placed in X3D (that is, x, z, y) coordinates
   * @param name Name to be assigned to this viewpoint
   * @param xPos x coordinate
   * @param yPos y coordinate
   * @param zPos z coordinate
   */
  public void addViewpoint(String name, double xPos, double yPos, double zPos){
    viewpointlist.add("<Viewpoint description='" + name + "' position='" + xPos + " " + zPos + " " + yPos + "'/>");
  }
  
  /**
   * Adds a sky and background to the X3D scene being created.
   * @param skyAngles angles beginning at 'straight up' from which, in concentric circles, the sky colors will be
   * pinned and factored as linear gradients
   * @param skyColors list of sky colors, beginning at 'straight up' and one for each skyAngle listed
   * @param groundAngles angles beginning at 'straight down' from which, in concentric circles, the ground colors
   * will be pinned and factored as linear gradients
   * @param groundColors list of ground colors, beginning at 'straight down, and one for each groundAngle listed
   */
  public void addSkyAndGround(float[] skyAngles, float[][] skyColors, float[] groundAngles, float[][] groundColors){
    String s = "<Background groundAngle='";
    for(int i = 0; i < groundAngles.length; i++) s += groundAngles[i] + ((i + 1) < groundAngles.length ? " " : "");
    s += "' groundColor='";
    for(int i = 0; i < groundColors.length; i++) s+= groundColors[i][0] + " " + groundColors[i][1] + " " + groundColors[i][2] + ((i + 1) < groundColors.length ? " " : "");
    s += "' skyAngle='";
    for(int i = 0; i < skyAngles.length; i++) s += skyAngles[i] + ((i + 1) < skyAngles.length ? " " : "");
    s += "' skyColor='";
    for(int i = 0; i < skyColors.length; i++) s+= skyColors[i][0] + " " + skyColors[i][1] + " " + skyColors[i][2] + ((i + 1) < skyColors.length ? " " : "");
    s += "'/>";
    backgroundString = s;
  }
  
  /**
   * Provides direct access to the underlying xml; eventually this will
   * be eliminated and all objects adding themselves to this document
   * will do so via this class's methods.
   * @return direct access to the XML document being created
   */
  public XMLAssist getXML(){ return x3d; } // TODO Eliminate direct access to X3DAssist xml
  
  
}
