package abcm.math;

/**
 * Provides simple mathematical operations like rounding, etc.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public final class BasicMath {

  /**
   * Private constructor to prevent instantiation; all methods of this 
   * class are static.
   */
  private BasicMath(){}
  
  /**
   * For rounding operations, the default precision is 5 decimal places.
   */
  private static final int DEFAULT_PRECISION = 5;

  /**
   * Rounds a float according to standard rules (less than .5 is rounded down,
   * .5 or higher is rounded up).
   * @param x value to be rounded
   * @return the rounded value
   */
  public static float roundFloat(float x){
    return roundFloat(x, DEFAULT_PRECISION);
  }

  /**
   * Rounds a double according to standard rules (less than .5 is rounded down,
   * .5 or higher is rounded up).
   * @param x value to be rounded
   * @return rounded value
   */
  public static double roundDouble(double x){
    return roundDouble(x, DEFAULT_PRECISION);
  }

  /**
   * Rounds a float according to standard rules (less than .5 is rounded down,
   * .5 or higher is rounded up).
   * @param x value to be rounded
   * @param precision number of decimal places of precision to which to round the value
   * @return rounded value
   */
  public static float roundFloat(float x, int precision){
    return (float)(Math.floor(Math.pow(10, precision) * x + .5)/Math.pow(10, precision));
  }

  /**
   * Rounds a double according to standard rules (less than .5 is rounded down,
   * .5 or higher is rounded up).
   * @param x value to be rounded
   * @param precision number of decimal places of precision to which to round the value
   * @return rounded value
   */
  public static double roundDouble(double x, int precision){
    double result = ((Math.floor(Math.pow(10, precision) * x + .5))/Math.pow(10, precision));
    // This line corrects a problem that occurs when rounding to a negative
    // precision. Because the division operation is inexact, Java gives an
    // erroneous result, i.e. 123456.654321 to precision -5 will return 100000.00000000001
    if(precision < 0) result = Math.round(result);
    return result;
  }

  /**
   * Returns the cube root of the value specified.
   * @param x the value for which the cube root is to taken
   * @return the cube root of the value specified
   */
  public static double cbrt(double x){
    // return Math.pow(x,1d/3d);
    // WHEN ABLE TO USE JAVA 1.5, use this:
    // return Math.cbrt(x);
    // TODO Check use of BasicMath.cbrt- may be obsolete
    double mult = (x < 0 ? -1 : 1);
    return mult * Math.pow(Math.E, Math.log(Math.abs(x))/3d);
  }

  /**
   * Returns the 3- dimensional determinant given the matrix
   * of values passed. Matrix is: [[ a  b  c ] [ d  e  f] [g h i]]
   * @param a (see method description)
   * @param b (see method description)
   * @param c (see method description)
   * @param d (see method description)
   * @param e (see method description)
   * @param f (see method description)
   * @param g (see method description)
   * @param h (see method description)
   * @param i (see method description)
   * @return determinant of 3x3 matrix
   */
  public static double determinant3D(
      double a, double b, double c, 
      double d, double e, double f, 
      double g, double h, double i){
    
    return a * (e * i - h * f) - b * (d * i - g * f) + c * ( d * h - g * e);
  }

}
