package abcm.math;

import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import visad.Delaunay;


/**
 * Contains basic geometric functions that manipulate points and calculate
 * distances, headings, angles, etc. All methods are provided as static methods.
 * Also provides methods for Delaunay triangulation.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Geometry {
  
  /** Value of precision for calculations known to be limited by the
   * precision of Java double calculations (see 'pointLiesOnSegment' for an example)
   */
  public static final double epsilon = .00000003d;
  
  
  /* Points */
  
  /**
   * Gets the distance between two points
   * @param p1 The first point
   * @param p2 The second point
   * @return the distance between p1 and p2
   */
  public static double distance(Point p1, Point p2){
    return Math.sqrt(
        Math.pow((p1.getX() - p2.getX()),2) +
        Math.pow((p1.getY() - p2.getY()),2) +
        Math.pow((p1.getZ() - p2.getZ()),2));
  }
  
  /**
   * Gets the planar distance between two points;
   * ignores the z dimension
   * @param p1 The first point
   * @param p2 The second point
   * @return the distance in the X,Y plane between p1 and p2
   */
  public static double planarDistance(Point p1, Point p2){
    return Math.sqrt(
        Math.pow((p1.getX() - p2.getX()),2) +
        Math.pow((p1.getY() - p2.getY()),2));
  }
  
  /**
   * Gets the slope from one point to another.
   * @param p1 The first point
   * @param p2 The second point
   * @return The slope (vertical) of the vector from the first point to the second point
   */
  public static double slope(Point p1, Point p2){
    double rise = p2.getZ() - p1.getZ(); 
    double run = planarDistance(p1, p2);
    double slope =(run != 0 ? rise/run : 0);
    slope = BasicMath.roundDouble(slope,6);
    return slope;
  }
  
  /**
   * Determines if the points passed as arguments are listed 
   * so that they would proceed from point 1 to point 2
   * to point 3 in a clockwise direction.
   * Result is given by solving for the determinant of the
   * matrix formed by [[p1.x, p1.y, p2.x] [p2.x, p2.y, p3.x] [p3.x, p3.y, 1]],
   * which will be less than zero only if the points are clockwise.
   * @param p1 first point
   * @param p2 second point
   * @param p3 third point
   * @return true if the points are clockwise, false if not
   */
  public static boolean pointsAreListedClockwise(Point p1, Point p2, Point p3){
    return(BasicMath.determinant3D(p1.getX(), p1.getY(), 1, p2.getX(), p2.getY(), 1, p3.getX(), p3.getY(), 1) < 0);    
  }
  
  /**
   * Gives the planar heading from one point to another. 
   * Uses pilot's compass, 0 = north, clockwise, in radians. 
   * Returns zero if points are equal in 2-D space (x1 = x2, y1 = y2).
   * Heading is given in Radians. 
   * @param origin Point from which the heading is to be calculated
   * @param destination Point to which the heading is to be calculated
   * @return planar heading, in radians (0 = north, clockwise increasing),
   *     from origin to destination
   */
  public static double heading(Point origin, Point destination){
    double heading;
    double xDiff = destination.getX()-origin.getX();
    double yDiff = destination.getY()-origin.getY();
    
    heading = 0;
    // If the points' x and y coordinates are equal, return zero;
    // Only proceed if at least one of them is nonzero
    if((xDiff != 0) || (yDiff != 0)){
      // If the points lie along a single N-S line
      if(xDiff == 0){
        // If the second point is due north of the first, return zero
        if(yDiff > 0 ){
          heading = 0;        
        }
        // The second point is due south of the first; return pi
        else{
          heading = Math.PI;
        }
      }
      // The points do not lie along a single N-S line
      else{
        // if the point lies to the east
        if(xDiff > 0){
          // If the points lie along an E-W line
          if(yDiff == 0){
            heading = Math.PI/2;
          }
          else{
            // if the point is in the NE quadrant
            if(yDiff > 0){
              heading = Math.atan(xDiff/yDiff);
            }
            // Otherwise, SE quadrant
            else{
              heading = Math.PI + Math.atan(xDiff/yDiff);              
            }
          }
        }
        // The point lies to the west
        else{
          if(yDiff == 0){
            heading = Math.PI * (-0.5);
          }
          else{
            // if the point is in the NW quadrant
            if(yDiff > 0){
              heading = (-1) * Math.atan(xDiff * (-1)/(yDiff));
            }
            // The point is in the SW quadrant
            else{
              heading = Math.PI * -0.5 - Math.atan(yDiff / xDiff);
            }
          }
        }
      }
    }
    return heading;
  }
  
  /**
   * Returns the difference in heading from one vector to another.
   * Arguments are three points defining two connected line segments
   * first point is the joint point; second and third are starting
   * and ending points, so that V1 = p1 to joint, V2 =
   * calling instance to p3.
   * @param joint middle vertex, where change of heading occurs
   * @param p1 Start of V1
   * @param p3 End of V2
   * @return Angle change from V1 to V2, in radians 
   */
  public static double headingChange(Point joint, Point p1, Point p3){
    double heading1 = heading(p1, joint);
    double heading2 = heading(joint, p3);

    double HeadingChange = heading2 - heading1;
    if(HeadingChange > Math.PI){
      HeadingChange = -1 * (Math.PI - (HeadingChange - Math.PI)); 
    }
    if(HeadingChange <= (-1 * Math.PI)){
      HeadingChange += (2 * Math.PI);
    }
    return HeadingChange;
  }

  /**
   * Returns a new point at the same planar (x,y) coordinates but
   * with a new vertical value
   * @param base reference point
   * @param offset vertical offset
   * @return the new point
   */
  public static Point getVerticalOffsetPoint(Point base, double offset){
    return new Point(base.getX(), base.getY(), base.getZ() + offset);
  }
  
  /**
   * Gets the point directly above the specified point, by the offset amount.
   * @param base the base point from which the new point is reckoned
   * @param offset amount above the base point
   * @return the new point
   */
  public static Point getPointAbove(Point base, double offset){
    return getVerticalOffsetPoint(base, offset);
  }
  
  /**
   * Gets the point directly below the specified point, by the offset amount.
   * @param base the base point from which the new point is reckoned
   * @param offset amount above the base point
   * @return the new point
   */
  public static Point getPointBelow(Point base, double offset){
    return getVerticalOffsetPoint(base, -1 * offset);
  }
  
  /**
   * Gets the coordinates found a specified distance and heading
   * from a given point. Note: always returns the same 'z' as
   * the base point; does not take into account vertical slope
   * @param base Point from which new point should be reckoned
   * @param headingInRadians heading, in radians, along which the new point
   * should be calculated. Note: can be over 2Pi or below zero
   * @param distance Distance from the base point to the desired new point
   * @return point with required coordinates
   */
  public static Point getOffsetPoint(Point base, double headingInRadians, double distance){
    return(new Point(base.getX() + distance * Math.sin(headingInRadians), base.getY() + distance * Math.cos(headingInRadians), base.getZ()));
  }
  
  /**
   * Returns a point between the two specified points, at some specified
   * fraction along the vector that connects them. Can accept fractions > 1
   * @param base the first endpoint
   * @param end the second endpoint
   * @param frac the fraction of distance along the vector defined by these endpoints
   * @return the new point at the projected fraction
   */
  public static Point pointBetween(Point base, Point end, double frac){
    return new Point( (base.getX() + (end.getX() - base.getX()) * frac), 
                      (base.getY() + (end.getY() - base.getY()) * frac),
                      (base.getZ() + (end.getZ() - base.getZ()) * frac));
  }
  
  /**
   * Returns the point that marks the intersection of two lines defined
   * by a point and a vector. The intersection may occur at any point along the line
   * so defined (even if the intersection is some negative number 'behind' one
   * or both of the vectors).
   * 
   * Returns null if the two headings are parallel, unless the two points
   * both lie along the vector so defined, in which case the function returns
   * P1.
   * 
   * If the two points have the same coordinates, returns P1 regardless of heading
   * NOTE: 2D only! Returns a point with z = 0
   * @param Point1 base point
   * @param heading1 heading away from base point 1
   * @param Point2 second base point
   * @param heading2 heading away from base point 2
   * @return point that marks the intersection of the two heading vectors based at the specified points
   */
  public static Point intersection(Point Point1, double heading1, Point Point2, double heading2){
    Point P1 = new Point(Point1.getX(), Point1.getY());
    Point P2 = new Point(Point2.getX(), Point2.getY());
    
    double x;
    double y;

    Point calc1 = getOffsetPoint(P1, heading1, 1);
    Point calc2 = getOffsetPoint(P2, heading2, 1);
    
    double x1 = P1.getX();
    double x2 = calc1.getX();
    double x3 = P2.getX();
    double x4 = calc2.getX();
    
    double y1 = P1.getY();
    double y2 = calc1.getY();
    double y3 = P2.getY();
    double y4 = calc2.getY();
    
    double den = (y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1);
    if(den == 0) return null;
    
    double v1num = (x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3);
//    There are two ways to do this; they both arrive at the 
//    same answer, of course, but one parameterizes one vector
//    and the second chooses the other. The second method would be:
//    double v2num = (x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3);
    
    x = x1 + (v1num / den) * (x2 - x1);
    y = y1 + (v1num / den) * (y2 - y1);
      
    return new Point(x, y);    
  }

  /**
   * Gives the magnitude of the z-vector that represents the cross
   * product of the two vectors beginning from the specified
   * vertex point and moving to the specified end points
   * @param vertex point of intersection
   * @param end1 end point of the first vector (extending from vertex)
   * @param end2 end point of the second vector (extending from vertex)
   * @return the magnitude of the cross-product
   */
  public static double crossProduct2D(Point vertex, Point end1, Point end2){
    return  (((end1.getX() - vertex.getX()) * (end2.getY() - vertex.getY())) - 
             ((end1.getY() - vertex.getY()) * (end2.getX() - vertex.getX())));    
  }

  
  
  /* Points in relation to Line Segments */
  
  /**
   * Determines if a given test point lies on the line segment
   * defined by two other points (in 2-Dimensions)
   * @param end1 end point 1
   * @param end2 end point 2
   * @param test the point to be tested
   * @return true if the test point lies on the segment, false otherwise
   */
  public static boolean pointLiesOnSegment(Point end1, Point end2, Point test){
    if(end1.isPlanarEqual(end2)) return test.isPlanarEqual(end1);
    if(Math.abs(end1.X - end2.X) < epsilon){
      return ((Math.abs(test.X - end1.X) < epsilon) &&  // Test point has same X coordinate  - compare with epsilon instead?
              ((test.Y >= end1.Y && test.Y <= end2.Y) || (test.Y >= end2.Y && test.Y <= end1.Y))); // Test point falls between the two points
    }
    double a = (end2.Y - end1.Y) / (end2.X - end1.X);
    double b = end1.Y - a * end1.X;
    return ((Math.abs(test.Y - (a * test.X + b)) < epsilon) &&
            ((test.X >= (end1.X - epsilon) && test.X <= (end2.X + epsilon)) || (test.X >= (end2.X - epsilon) && test.X <= (end1.X + epsilon))) && 
            ((test.Y >= (end1.Y - epsilon) && test.Y <= (end2.Y + epsilon)) || (test.Y >= (end2.Y - epsilon) && test.Y <= (end1.Y + epsilon))));
  }
  
  /**
   * Determines if a given test point lies on the given segment (in 2 dimensions)
   * @param lineSeg the line segment to be tested
   * @param test the test point
   * @return true if the point lies on the line segment, false otherwise
   */
  public static boolean pointLiesOnSegment(LineSegment lineSeg, Point test){ 
    return pointLiesOnSegment(lineSeg.P1, lineSeg.P2, test);    
  }
  
  /**
   * Returns the closest point on the line segment defined by P1 and P2 to P3.
   * May return one of the endpoints. If the two points defining the line segment
   * are equal, return P1.
   * @param P1 One end of the line segment
   * @param P2 The other end of the line segment
   * @param P3 The third, reference point
   * @return a point object representing the coordinates of the closest point
   * on the specified line segment to the other point provided 
   */
  public static Point closestPointOnSegment(Point P1, Point P2, Point P3){
    if(P1.isEqual(P2)) return P1;
    
    // Some useful intermediate variables
    double gX = P2.getX() - P1.getX();
    double gY = P2.getY() - P1.getY();
    double gZ = P2.getZ() - P1.getZ();
    double kX = P1.getX() - P3.getX();
    double kY = P1.getY() - P3.getY();
    double kZ = P1.getZ() - P3.getZ();

    // This minimizes the distance function along the vector of the segment
    // C is a parameter indicating distance along the vector, with unit
    // distance 1 = length of segment      
    double A = (gX * gX) + (gY * gY) + (gZ * gZ);
    double B = 2 * ((gX * kX) + (gY * kY) + (gZ * kZ));
    double C = ((-1) * B) / ( 2 * A);
    
    if(C <= 0) return P1;
    if(C >= 1) return P2;
    return new Point(P1.getX() + C * gX, P1.getY() + C * gY, P1.getZ() * C * gZ );
  }
  
  /**
   * Returns the closest point on the line segment defined by P1 and P2 to P3.
   * May return one of the endpoints. If the two points defining the line segment
   * are equal, return P1.
   * @param lineSeg
   * @param point
   * @return a point object representing the coordinates of the closest point
   * on the specified line segment to the other point provided 
   */
  public static Point closestPointOnSegment(LineSegment lineSeg, Point point){
    return closestPointOnSegment(lineSeg.P1, lineSeg.P2, point);
  }

  
  /* Line Segments */
  /**
   * Returns true if the two line segments intersect in 2D space, false otherwise
   * see http://www.vb-helper.com/howto_segments_intersect.html for VB implementation
   * @param lineSeg1 line segment 1
   * @param lineSeg2 the line to be tested for intersection
   * @return true if there is an intersection, false if not, and false if the two
   * segments are parallel. Note that this should return 'true' if the two
   * segments share an endpoint (and are not parallel), but this sometimes
   * fails due to the limitations of double-precision math.
   */
  public static boolean lineSegmentsIntersectIn2D(LineSegment lineSeg1, LineSegment lineSeg2){
    double x1 = lineSeg1.P1.X;
    double y1 = lineSeg1.P1.Y;
    double x2 = lineSeg1.P2.X;
    double y2 = lineSeg1.P2.Y;
    double a1 = lineSeg2.P1.X;
    double b1 = lineSeg2.P1.Y;
    double a2 = lineSeg2.P2.X;
    double b2 = lineSeg2.P2.Y;
    
    double xdiff = x2 - x1;
    double ydiff = y2 - y1;
    double adiff = a2 - a1;
    double bdiff = b2 - b1;
    
    if((adiff * ydiff - bdiff * xdiff) == 0 )  return false;  // Segments are parallel
    
    // This parameterizes the point of intersection along each segment 
    double s = (xdiff * (b1 - y1) + ydiff * (x1 - a1)) / (adiff * ydiff - bdiff * xdiff);
    double t = (adiff * (y1 - b1) + bdiff * (a1 - x1)) / (bdiff * xdiff - adiff * ydiff);
    
    // The two segments intersect if the parameters for each are between 0 and 1
    return (s >= 0 && s <= 1 &&  t >= 0 && t <= 1);
  }
  
  
  /**
   * Returns the point of intersection with another line
   * segment, assuming both extended indefinitely; point returned is the point on segment 1
   * segment (including appropriate z value) based on 
   * where the line segments intersect in 2D dimensions only
   * @param L1 line segment one
   * @param L2 line segment two
   * @return the point at which the two segments intersect. Returns null if the two line segments are parallel
   */
  public static Point pointOfIntersection2D(LineSegment L1, LineSegment L2){
    double x1 = L1.P1.X;
    double y1 = L1.P1.Y;
    double x2 = L1.P2.X;
    double y2 = L1.P2.Y;
    double z1 = L1.P1.Z;
    double z2 = L1.P2.Z;
    
    double a1 = L2.P1.X;
    double b1 = L2.P1.Y;
    double a2 = L2.P2.X;
    double b2 = L2.P2.Y;
    
    double xdiff = x2 - x1;
    double ydiff = y2 - y1;
    double zdiff = z2 - z1;
    double adiff = a2 - a1;
    double bdiff = b2 - b1;
    
    if((adiff * ydiff - bdiff * xdiff) == 0 ) return null;  // Segments are parallel
    
    // This parameterizes the point of intersection along each segment 
//    double s = (xdiff * (b1 - y1) + ydiff * (x1 - a1)) / (adiff * ydiff - bdiff * xdiff);
    double t = (adiff * (y1 - b1) + bdiff * (a1 - x1)) / (bdiff * xdiff - adiff * ydiff);
    
    // The two segments intersect if the parameters for each are between 0 and 1; use 't' to
    // calculate the offset from the first endpoint of LineSeg 1 to the new point
    return new Point(x1 + xdiff * t, y1 + ydiff * t, z1 + zdiff * t);
  }

  
  /**
   * Returns the point of intersection of two line segments provided that the
   * intersection falls on one of the segments (they are not extended)
   * @param L1 line segment 1
   * @param L2 line segment 2
   * @return point of intersection, if any (returns null if the segments are parallel or if they
   * do not intersect)
   */
  public static Point restrictedPointOfIntersection2D(LineSegment L1, LineSegment L2){
    if(L1.P1.isPlanarEqual(L2.P1) || L1.P1.isPlanarEqual(L2.P2)) return new Point(L1.P1.getX(), L1.P1.getY(), L1.P1.getZ());
    if(L1.P2.isPlanarEqual(L2.P1) || L1.P2.isPlanarEqual(L2.P2)) return new Point(L1.P2.getX(), L1.P2.getY(), L1.P2.getZ());
    
    double x1 = L1.P1.X;
    double y1 = L1.P1.Y;
    double x2 = L1.P2.X;
    double y2 = L1.P2.Y;
    double z1 = L1.P1.Z;
    double z2 = L1.P2.Z;
    
    double a1 = L2.P1.X;
    double b1 = L2.P1.Y;
    double a2 = L2.P2.X;
    double b2 = L2.P2.Y;
    
    double xdiff = x2 - x1;
    double ydiff = y2 - y1;
    double zdiff = z2 - z1;
    double adiff = a2 - a1;
    double bdiff = b2 - b1;
    
    if((adiff * ydiff - bdiff * xdiff) == 0 ) return null;  // Segments are parallel
    
    // This parameterizes the point of intersection along each segment 
    double s = (xdiff * (b1 - y1) + ydiff * (x1 - a1)) / (adiff * ydiff - bdiff * xdiff);
    double t = (adiff * (y1 - b1) + bdiff * (a1 - x1)) / (bdiff * xdiff - adiff * ydiff);
    
    if(Double.isNaN(z1 + zdiff * t))
       System.out.println("NaN ERROR in restrictedPointOfIntersection.");
    
    if(Math.abs(1-t) < Geometry.epsilon) t = 1;
    if(Math.abs(1-s) < Geometry.epsilon) s = 1;
    if(Math.abs(0-t) < Geometry.epsilon) t = 0;
    if(Math.abs(0-s) < Geometry.epsilon) s = 0;
    
    // The two segments intersect if the parameters for each are between 0 and 1; use 't' to
    // calculate the offset from the first endpoint of LineSeg 1 to the new point
    return ( (( (t >= 0) && (t <= 1) ) &&
              ( (s >= 0) && (s <= 1) )) ? new Point(x1 + xdiff * t, y1 + ydiff * t, z1 + zdiff * t) : null);
  }

  
  /**
   * Returns true if the given line segment crosses this one in two dimensions.
   * The segments' intersection is based on their actual lengths, not any extension of them.
   * @param L1 Line Segment 1
   * @param L2 Line Segment 2
   * @return true if the two segments intersect, false if not (including if the two segments are parallel)
   */
  public static boolean lineSegmentsCrossIn2D(LineSegment L1, LineSegment L2){
    double x1 = L1.P1.X;
    double y1 = L1.P1.Y;
    double x2 = L1.P2.X;
    double y2 = L1.P2.Y;
    double a1 = L2.P1.X;
    double b1 = L2.P1.Y;
    double a2 = L2.P2.X;
    double b2 = L2.P2.Y;
    
    double xdiff = x2 - x1;
    double ydiff = y2 - y1;
    double adiff = a2 - a1;
    double bdiff = b2 - b1;
    
    if(Math.abs(adiff * ydiff - bdiff * xdiff) <= epsilon ) return false;  // Segments are parallel
    
    // This parameterizes the point of intersection along each segment 
    double s = (xdiff * (b1 - y1) + ydiff * (x1 - a1)) / (adiff * ydiff - bdiff * xdiff);
    double t = (adiff * (y1 - b1) + bdiff * (a1 - x1)) / (bdiff * xdiff - adiff * ydiff);
    
    // The two segments intersect if the parameters for each are between 0 and 1
    return (s > 0 && s < 1 &&  t > 0 && t < 1);
  }
  
  /* Triangles */
    
  /**
   * Returns the area of the triangle defined by the three points given.
   * Note: Area is calculated for the 2-D positions only.
   * @param p1 First vertex of the triangle
   * @param p2 Second vertex of the triangle
   * @param p3 Third vertex of the triangle
   * @return the area of the triangle
   */
  public static double areaOfTriangle(Point p1, Point p2, Point p3){
    return Math.abs((0.5) * (p1.X * p2.Y + p1.Y * p3.X + p2.X * p3.Y - p3.X * p2.Y - p3.Y * p1.X - p2.X * p1.Y));    
  }
  
  /**
   * Calculates the 3-D area of a triangle in 3 space
   * @param p1 first vertex
   * @param p2 second vertex
   * @param p3 third vertex
   * @return area
   */
  public static double areaOfTriangle3D(Point p1, Point p2, Point p3){
    double a = p1.distanceToPoint(p2);
    double b = p2.distanceToPoint(p3);
    double c = p3.distanceToPoint(p1);
    double p = (a + b + c) / 2d;
    return Math.sqrt(p * (p - a) * (p - b) * (p - c)); // Heron's formula    
  }
  
  /**
   * Calculates the area of a quadrilateral in 3 space.
   * Note: does not enforce if the points are coplanar;
   * assumes transect from p2 to p3
   * @param p1 first vertex
   * @param p2 second vertex
   * @param p3 third vertex
   * @param p4 fourth vertex
   * @return area
   */
  public static double areaOfQuadrilateral3D(Point p1, Point p2, Point p3, Point p4){
    return areaOfTriangle3D(p1, p2, p3) + areaOfTriangle3D(p2, p4, p3);
  } 

  /**
   * FROM: http://www.mathpages.com/home/kmath393.htm
   * 
   *       V  =  (z1+z2+z3)(x1y2-x2y1+x2y3-x3y2+x3y1-x1y3)/6
   *       
   * @param p1 vertex 1
   * @param p2 vertex 2
   * @param p3 vertex 3
   * @return the volume under the triangle defined by the three points
   */
  public static double volumeUnderTriangle(Point p1, Point p2, Point p3){
    return (p1.Z + p2.Z + p3.Z) * (Math.abs((p1.X * p2.Y - p2.X * p1.Y) + (p2.X * p3.Y - p3.X * p2.Y) + (p3.X * p1.Y - p1.X * p3.Y))) / 6d;
  }
  
  
  /**
   * 
   * See: http://mathforum.org/library/drmath/view/54505.html
   * @param v1 Vertex one
   * @param v2 Vertex two
   * @param v3 Vertex three
   * @param P  Point to be tested
   * @return true if the test point is inside the triangle formed by the other three,
   * false otherwise 
   */
  public static boolean pointIsInside(Point v1, Point v2, Point v3, Point P){
    double cp1 = crossProduct2D(v1, v2, P);
    double cp2 = crossProduct2D(v2, v3, P);
    double cp3 = crossProduct2D(v3, v1, P);
//    System.out.println("pointIsInside is Returning (true if ALL positive or ALL negative):\n       " + cp1 + " " + cp2 + " " + cp3);
    return(((cp1 >= 0) && (cp2 >= 0) && (cp3 >= 0)) ||
           ((cp1 <= 0) && (cp2 <= 0) && (cp3 <= 0)));
  }
  
  
  /**
   * Determines if a specified point lies inside the circle described by the
   * three points given.
   * The determinant of the matrix:
   * 
   * Ax  Ay  (Ax^2 + Ay^2)  1
   * Bx  By  (Bx^2 + By^2)  1
   * Cx  Cy  (Cx^2 + Cy^2)  1
   * Dx  Dy  (Dx^2 + Dy^2)  1
   * 
   * is positive if if the point is inside the triangle, assuming that 
   * A, B, and C are counter-clockwise. See Wikipedia, "Delaunay Triangulation"
   * @param v1 first vertex of the triangle to be tested
   * @param v2 second vertex of the trianngle to be tested
   * @param v3 third vertex of the triangle to be tested
   * @param P point to be tested
   * @return true if P is circumscribed by the triangle described by v1, v2, v3, false
   * otherwise
   */
  public static boolean pointIsCircumscribed(Point v1, Point v2, Point v3, Point P){
    return((      v1.X * (v2.Y * ((v3.X * v3.X + v3.Y * v3.Y) - (P.X * P.X + P.Y * P.Y)) - (v2.X * v2.X + v2.Y * v2.Y) * (v3.Y - P.Y) + (v3.Y * (P.X * P.X + P.Y * P.Y) - P.Y * (v3.X * v3.X + v3.Y * v3.Y))) 
                - v1.Y * (v2.X * ((v3.X * v3.X + v3.Y * v3.Y) - (P.X * P.X + P.Y * P.Y)) - (v2.X * v2.X + v2.Y * v2.Y) * (v3.X - P.X) + (v3.X * (P.X * P.X + P.Y * P.Y) - P.X * (v3.X * v3.X + v3.Y * v3.Y))) 
                + (v1.X * v1.X + v1.Y * v1.Y) * (v2.X * (v3.Y - P.Y) - v2.Y * (v3.X - P.X) + (v3.X * P.Y - P.X * v3.Y))  
                -     (v2.X * (v3.Y * (P.X * P.X + P.Y * P.Y) - P.Y * (v3.X * v3.X + v3.Y * v3.Y)) - v2.Y * (v3.X * (P.X * P.X + P.Y * P.Y) - P.X * (v3.X * v3.X + v3.Y * v3.Y)) + (v2.X * v2.X + v2.Y * v2.Y) * (v3.X * P.Y - P.X * v3.Y))) > 0);
  }

  /**
   * Calculates the elevation of a Point if it
   * is projected onto the plane defined by the three vertices given.
   * @param v1 Vertex one
   * @param v2 Vertex two
   * @param v3 Vertex three
   * @param P  Point to be tested
   * @return the elevation of the projection of the test point onto the triangle
   */
  public static double getProjectedElevation(Point v1, Point v2, Point v3, Point P){
      // Using determinants, the formulas are:
  //    double M1_det = (((v2.Y - v1.Y) * (v3.Z - v1.Z)) - ((v3.Y - v1.Y) * (v2.Z - v1.Z)));
  //    double M2_det = (((v2.X - v1.X) * (v3.Z - v1.Z)) - ((v3.X - v1.X) * (v2.Z - v1.Z)));
  //    double M3_det = (((v2.X - v1.X) * (v3.Y - v1.Y)) - ((v3.X - v1.X) * (v2.Y - v1.Y)));
      
  //    return v1.Z + ((((P.Y - v1.Y) * M2_det) - ((P.X - v1.X) * M1_det))/(M3_det));
      // Faster implementation skips allocation of variables
    
    if(Double.isInfinite(v1.Z + 
            ((((P.Y - v1.Y) * (((v2.X - v1.X) * (v3.Z - v1.Z)) - ((v3.X - v1.X) * (v2.Z - v1.Z)))) - ((P.X - v1.X) * (((v2.Y - v1.Y) * (v3.Z - v1.Z)) - ((v3.Y - v1.Y) * (v2.Z - v1.Z))))) 
           /
            (((v2.X - v1.X) * (v3.Y - v1.Y)) - ((v3.X - v1.X) * (v2.Y - v1.Y))))))
        System.out.println("Infinite Projection!");
      return 
         v1.Z + 
            ((((P.Y - v1.Y) * (((v2.X - v1.X) * (v3.Z - v1.Z)) - ((v3.X - v1.X) * (v2.Z - v1.Z)))) - ((P.X - v1.X) * (((v2.Y - v1.Y) * (v3.Z - v1.Z)) - ((v3.Y - v1.Y) * (v2.Z - v1.Z))))) 
           /
            (((v2.X - v1.X) * (v3.Y - v1.Y)) - ((v3.X - v1.X) * (v2.Y - v1.Y))));
    }
  
  /**
   * Calculates the elevation of a Point if it
   * is projected onto the plane defined by the three vertices given.
   * @param v1 Vertex one
   * @param v2 Vertex two
   * @param v3 Vertex three
   * @param x  x-coordinate of the point to be tested
   * @param y  y-coordinate of the point to be tested
   * @return the elevation of the projection of the test point onto the triangle
   */
  public static double getProjectedElevation(Point v1, Point v2, Point v3, double x, double y){
    return getProjectedElevation(v1, v2, v3, new Point(x,y));
  }
  
  /**
   * Volume under triangle
   * 
   */
  
  /* Quadrilaterals */
  /**
   * Gets the area of the quadrilateral defined by the given four points.
   * Note that this is not perfect- it assumes that the quadrilateral
   * would be convex as drawn- but it is more general than just
   * using squares. (Actually it can give the right answer for some
   * concave quadrilaterals, but not all).
   * Note: Area is calculated for the 3-D faces of two triangular halves
   * based on vertices [1, 2, 4] and [2, 3, 4].
   * @param p1 First vertex
   * @param p2 Second vertex
   * @param p3 Third vertex
   * @param p4 Fourth vertes
   * @return The area of the quadrilateral defined by the four points.
   */
  public static double areaOfQuadrilateral(Point p1, Point p2, Point p3, Point p4){
    return areaOfTriangle(p1, p2, p4) + areaOfTriangle(p2, p3, p4);
  }

  
  
  /* Circles */
    
  /**
   * Returns the center point of the circle inscribed by the three given points. Returns
   * the center point if the three points are colinear
   * @param P1 Point 1 
   * @param P2 Point 2
   * @param P3 Point 3
   * @return The calculated center point of the circle incsribed by the three given points.
   */
  public static Point centerOfCircle(Point P1, Point P2, Point P3){
    double x1 = P1.getX();
    double x2 = P2.getX();
    double x3 = P3.getX();
    
    double h1 = P1.headingTo(P2);
    double h2 = P1.headingTo(P3);
    double h3 = P2.headingTo(P3);
    
    // If the heading from P1 to P2 is the same as the heading from P2
    // to P3, then the three points are colinear, and
    // the one that is closer to P1 is the 'middle' one
    if(h1 == h2){
      return (P1.distanceToPoint(P2) < P1.distanceToPoint(P3) ? P2 : P3);
    }
    // If the heading from P1 to P2 is the exact negative of the heading
    // from P1 to P3, then the three points are colinear, and P1 is
    // the 'middle' one
    if(h1 == (-1 * h2)){
      return P1;
    }
    
    // The points are not colinear- continue
    // Take the two short chords- avoid the one that crosses the other
    // two and thereby avoid the diameter
    double c1 = P1.distanceToPoint(P2);
    double c2 = P1.distanceToPoint(P3);
    double c3 = P2.distanceToPoint(P3);
    
    Point centerChord1 = pointBetween(P1, P2, .5);
    Point centerChord2 = pointBetween(P1, P3, .5);
    
    double heading1 = h1 + (.5 * Math.PI);
    double heading2 = h2 + (.5 * Math.PI);

    // If chord from P1 to P2 is longest, replace
    // with the other chord
    if((c1 > c2) && (c1 > c3)){
      centerChord1 = pointBetween(P2, P3, .5);
      heading1 = h3 + (.5 * Math.PI);
    }
    if((c2 > c1) && (c2 > c3)){
      centerChord2 = pointBetween(P2, P3, .5);
      heading2 = h3 + (.5 * Math.PI);
    }    
    Point C = intersection(centerChord1, heading1, centerChord2, heading2);
    return C;
  }
  
    
  
  
  
  /* Delaunay Triangulation */
  
  /**
   * Given a list of points returns the DeLaunay triangulation, as a vector
   * populated by polygons. Each polygon represents a DeLaunay cell. There will
   * be N-2 cells, where N is the number of Points. Note that there should be
   * no instance of 4 or more cocircular points.
   * @param pointsCollection the list of points for which the DeLaunay Triangulation is
   * to be calculated
   * @return the collection of Triangles that comprise the DeLaunay Triangulation.
   */
  public static Collection getDelaunayTriangulation(Collection pointsCollection){
    if(pointsCollection.size() < 3) return null;
    
    List points = new Vector();
    Hashtable pointsCorrespondence = new Hashtable();
    // 'Squiggle' the points to avoid regular intervals, which can
    // lead to cocircular points.
    // This is done deterministically to avoid issues with random
    // number generation. 
    // Note that the triangulation is calculated based on the shifted points,
    // but the actual triangles use the original points
    Iterator I = pointsCollection.iterator();
    double c = 0;
    double denominator = pointsCollection.size();
    while(I.hasNext()){
      Point original = (Point)I.next();
      Point P = getOffsetPoint(original, (c/denominator) * 2 * Math.PI, .0001d);
      points.add(P);
      pointsCorrespondence.put(P, original);
      c++;
    }
    
    Collection polyList = new Vector();
    HashSet existingTriangles = new HashSet();
    long pointCount = points.size();
    System.err.println("POINT COUNT = " + pointCount);
    // Transform the list of points into a list of polygons that represent
    // the DeLaunay triangulation
    // Note: This is the slow way!
    // Loop through each point
    for(int i = 0; i < (pointCount-2); i++){
      System.err.println("INNER LOOP ONE: (" + pointCount + ")  " + i);
      // For each point, loop through all the other points except
      // the ones that have already been done in the main loop
      Point P1 = (Point)points.get(i);
      for(int j = i+1; j < (pointCount-1); j++){
        
        // For each of these points, loop through all the remaining points
        // except the ones from the secondary loop
        Point P2 = (Point)points.get(j);
        if(!(P2.isPlanarEqual(P1))){
          for(int k = j+1; k < pointCount; k++){
            Point P3 = (Point)points.get(k);    
            if((!P3.isPlanarEqual(P1)) && (!P3.isPlanarEqual(P2))){
              
              Triangle t     = new Triangle(P1, P2, P3);
              int l = 0;
              boolean hasPointInside = false;
              while((l < pointCount) && (!hasPointInside)){
                if((l != i) &&
                   (l != j) &&
                   (l != k)){
                  Point T = (Point)points.get(l);
                  if(T.isPlanarEqual(P1) || T.isPlanarEqual(P2) || T.isPlanarEqual(P3)) {
                    hasPointInside = true;
                  }
                  else {
                    if(t.pointIsCircumscribed(T)) {
//                      System.out.println("Triangle: " + P1.coordStringShort(1) + " " + P2.coordStringShort(1)+ " " + P3.coordStringShort(1) + " circumscribes " + T.coordStringShort(1));
                      hasPointInside = true;
                    }
                  }
                }
                l++;
              }
              if(!hasPointInside){
//                System.out.println("Triangle: " + P1.coordStringShort(1) + " " + P2.coordStringShort(1)+ " " + P3.coordStringShort(1) + " circumscribes NO POINTS");
                // Add this triangle only if it hasn't already been added
                // Note that the overhead is lower for adding a key than
                // checking all three keys... so add them all but check only
                // one
                t = new Triangle((Point)pointsCorrespondence.get(P1), (Point)pointsCorrespondence.get(P2), (Point)pointsCorrespondence.get(P3));
//                t.makeCounterClockwise();
                String key1 = ((Point)t.points.get(0)).coordString() + ":" + 
                              ((Point)t.points.get(1)).coordString() + ":" +
                              ((Point)t.points.get(2)).coordString();
                String key2 = ((Point)t.points.get(1)).coordString() + ":" + 
                              ((Point)t.points.get(2)).coordString() + ":" +
                              ((Point)t.points.get(0)).coordString();
                String key3 = ((Point)t.points.get(2)).coordString() + ":" + 
                              ((Point)t.points.get(0)).coordString() + ":" +
                              ((Point)t.points.get(1)).coordString();
                if(!existingTriangles.contains(key1)){ 
                   polyList.add(t);
                   existingTriangles.add(key1);
                   existingTriangles.add(key2);
                   existingTriangles.add(key3);
                }
                else{ System.err.println("TRYING TO ADD ONE THAT ALREADY EXISTS!"); }
              }
            }
            else{System.err.println("FOUND TWO EQUAL: " + P1.coordString() + "\n" + P2.coordString() + "\n " + P3.coordString());}
          }
        }else{System.err.println("FOUND TWO EQUAL2: " + P1.coordString() + "\n" + P2.coordString());}
      }      
    }
    return polyList;  
  }

  
  /**
   * Gets a Delaunay Triangulation as produced using the VisAd toolkit. Note that the
   * VisAd algorithm requires the coordinates of the  points to be integer values (this is true
   * even though the array in which the values are passed is an array of floats); this adaptation of it
   * begins by multiplying the x,y coordinates of all points by ten thousand. If greater
   * precision is required, the points must be transformed (or the method re-implemented).
   * When the collection is returned, however, it is comprised of triangles
   * that have been constructed using the original, unaltered points. 
   * @param pointsCollection the collection of points to be used to generate
   * the Delaunay triangulation
   * @return a collection of Triangles forming the Delaunay triangulation
   */
  public static Collection getVisadDelaunayTriangulation(List<Point> pointsCollection){
    float[][] samples = new float[2][pointsCollection.size()];
    int i = 0;
    for(Point p: pointsCollection){
      samples[0][i] = (float)(p.getX() * 10000); // Visad algorithm converts to ints; this keeps some precision
      samples[1][i] = (float)(p.getY() * 10000);
      i++;
    }
    Vector triangulation = new Vector();
    try{
      Delaunay d = Delaunay.factory(samples, false);
      int[][] tris = d.Tri;
      for(int[] triangle: tris){
        triangulation.add(new Triangle(pointsCollection.get(triangle[0]), pointsCollection.get(triangle[1]), pointsCollection.get(triangle[2])));
      }
      return triangulation;      
    }
    catch(Exception E){
      System.err.println("Fast VISAD Delaunay failed- using slower Geometry.Delaunay");
      return getDelaunayTriangulation(pointsCollection);
    }
  }
  
  
  
  
  
  
  
  
    
}

