package abcm.math;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;


/**
 *  A basic line segment, using points in 3D as endpoints.
 *  Includes simple text display methods and methods to calculate whether a given instance
 *  intersects another line.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class LineSegment implements ABCMDeepCloneable{

  /* Static Values */
  
  /**
   * A static counter used to create unique identifiers for instances
   */
  private static long ID;
  
  
  /* Instance Variables */
  
  /**
   * A given endpoint
   */
  public Point P1;
  
  /**
   * A second endpoint
   */
  public Point P2;
  
  /**
   * A unique identifier (assigned by the constructor)
   */
  public String IDString;
  
  
  /* Constructor */
  
  /**
   * Constructor that takes two endpoints. A unique identifier is automatically assigned.
   * @param p1 First endpoint
   * @param p2 Second endpoint
   */
  public LineSegment(Point p1, Point p2){
    P1 = p1;
    P2 = p2;
    IDString = "LINE_" + (ID++);
  }

  
  /* Statistics */
  
  /**
   * Returns the length of the Line (the distance between the endpoints)
   * @return the length of the Line
   */
  public double length(){    return Geometry.distance(P1, P2);    }

  
  /* Geometry */
  
  /**
   * Returns true if the two line segments intersect in 2D space, false otherwise
   * see http://www.vb-helper.com/howto_segments_intersect.html for VB implementation
   * @param L2 the line to be tested for intersection
   * @return true if there is an intersection, false if not, and false if the two
   * segments are parallel. Note that this should return 'true' if the two
   * segments share an endpoint (and are not parallel), but this sometimes
   * fails due to the limitations of double-precision math.
   */
  public boolean intersectsIn2D(LineSegment L2){
    return Geometry.lineSegmentsIntersectIn2D(this, L2);
  }

  
  /**
   * Returns the point of intersection with another line
   * segment, assuming both extended indefinitely; point returned is the point on THIS line
   * segment (including appropriate z value) based on 
   * where the line segments intersect in 2D dimensions only
   * @param L2 the other line segment
   * @return the point at which the two segments intersect. Returns null if the two line segments are parallel
   */
  public Point pointOfIntersection2D(LineSegment L2){
    return Geometry.pointOfIntersection2D(this, L2);
  }

  /**
   * Returns true if the given line segment crosses this one in two dimensions.
   * The segments' intersection is based on their actual lengths, not any extension of them.
   * @param L2 the other line segment
   * @return true if the two segments intersect, false if not (including if the two segments are parallel)
   */
  public boolean crossesIn2D(LineSegment L2){
    return Geometry.lineSegmentsCrossIn2D(this, L2);    
  }

  
  /* Reporting */
  
  /**
   * Returns a prettified display string for this Line Segment.
   * @return a String in the form Pt1 - Pt2.
   */
  public String display(){
    return P1.coordString(10) + " - " + P2.coordString(10);
  }
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    return (cloneStore.containsKey(this) ? cloneStore.get(this) : new LineSegment((Point)this.P1.getDeepClone(cloneStore), (Point)this.P2.getDeepClone(cloneStore)));
  }
  
}
