package abcm.math;

import Jama.Matrix;

/**
 * The LinearSystemSolver class wraps the JAMA matrix functions into a single, easy
 * to use class that can be applied to the creation of systems of linear equations,
 * and to the solution using matrix algebra. Specifically, the class allows the creation
 * of a system with any arbitrary number of rows and columns, and a corresponding 
 * one-dimensional array of non-homogenous terms. Together these represent the left
 * and right sides of a system of linear equations- the main matrix stores the coefficients
 * of the unknowns on the left sides of the equations, and the non-homogenous terms store 
 * terms on the right sides. The methods allows for the retrieval and setting of all these values.
 * Additionally, the class invokes matrix operations to derive the solution of the system,
 * when possible, and access to the solved values.
 * <br/><br/>
 * Additional functions allows for the display of the system, and for the easy specification
 * of known terms. Note that the matrix is not limited to being square, and so one way
 * known terms could be added would be to place a 1 in the appropriate column as a coefficient
 * and the term's value on the right, however, this would add another row to the matrix, which
 * could slow processing. This class instead performs a reduction by subtracting out the term
 * from the remaining equations. However, the matrix will still behave as if it had N unknowns-
 * both the display and the programmatic interface will mask the fact that the underlying matrix 
 * is now smaller.
 * <br/><br/>
 * This class also provides an easy means to add additional rows to the main matrix, which will
 * correspond to adding a new equation to the system.
 * <br/><br/>
 * For additional information on how the matrix solutions are performed, see the JAMA documentation.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class LinearSystemSolver {

  /**
   * Basic matrix in which coefficients are stored
   */
  Matrix M;
  
  /**
   * Matrix in which non-harmonic values are stored
   */
  Matrix NHTs;
  
  /**
   * Matrix in which results (originally unknown values) are stored when calculated
   */
  Matrix Unknowns;
  
  /**
   * A message that indicates any problem encountered during the calculations.
   */
  private String errorMessage;
  
  /**
   * Constructor that takes a single number of values and creates a 
   * square matrix, randomly populated
   * @param values the number of rows and columns to create
   */
  public LinearSystemSolver(int values){
    init(values, values);
    M = Matrix.random(values, values);
  }
 
  /**
   * Constructor that allows a matrix to be created with Y rows
   * and X columns.
   * @param rows the number of rows
   * @param cols the number of columns (and, hence, unknowns)
   */
  public LinearSystemSolver(int rows, int cols){
   init(rows,cols); 
  }
  
  /**
   * Initializes the matrices
   * @param rows number of matrix rows
   * @param cols number of matrix columns
   */
  private void init(int rows, int cols){
    M = new Matrix(rows, cols);
    NHTs = new Matrix(rows,1);
    Unknowns = new Matrix(rows,1);
  }
   
  /**
   * Sets the value of an element of the main matrix to a
   * specific, given value.
   * @param row row of the element to be set
   * @param col column of the element to be set
   * @param val new value for the specified element
   */
  public void setElement(int row, int col, double val){  M.set(row, col, val);  }
  
  /**
   * Gets the value of the element at a specified row and column
   * @param row the row of the value to be retrieved
   * @param col the column of the value to be retrieved
   * @return the value at the specified row and column of the main
   * matrix
   */
  public double getElement(int row, int col){   return M.get(row, col);    }
  
  /**
   * Sets the value of one of the Non-Homogenous terms
   * @param row row of the term to be set
   * @param val new value
   */
  public void setNHT(int row, double val){     NHTs.set(row,0,val);    }
  
  /**
   * Returns a value corresponding to the column number
   * of the unknown requested. 
   * @param col the column of the unknown requested 
   * @return column number corresponding to the unknown requested
   */
  public double getSolvedValue(int col){
    double retVal = Unknowns.get(col,0);
    if(retVal == -0.0) retVal = 0;
    return retVal;
  }
  
  /**
   * Define a known value.
   * 
   * When initially created, the matrix may have more columns than rows.
   * This is corrected by indicating that the unknown represented by a given
   * column should actually have a specific value. This column is then
   * removed by multiplying the coefficients as stored in that column
   * by the specified value and subtracting that from the NHTs column in
   * the corresponding rows. The column is then deleted.
   * @param col column of the known value
   * @param val new value to be defined
   */
  public void defineKnownValue(int col, double val){
    int newRow = addEquation();
    setElement(newRow, col, 1);
    setNHT(newRow, val);
  }
  
  /**
   * Perform the matrix solution. Under certain conditions the matrix cannot
   * be solved.
   * @return true if the matrix could actually be solved, false otherwise
   */
  public boolean solve(){
    errorMessage = null;
    if(M.getColumnDimension() > M.getRowDimension()){
      Unknowns = new Matrix(M.getRowDimension(),1);
      System.out.println("!!!!!!!  MATRIX HAS INCORRECT DIMENSIONS: cols = " + M.getColumnDimension() + "   rows = " + M.getRowDimension());
      return false;
    }
    try{
      Unknowns = M.solve(NHTs);
    }
    catch (RuntimeException rte){
      errorMessage = "Attempt to solve linear system failed: " + rte.getMessage();
      print();
      System.out.println("!!!!!!!   Attempt to solve linear system failed: " + rte.getMessage());
      return false;
    }
    return true;
  }

  /**
   * Prints a prettified version of the matrices (including the solution) to the standard output.
   */
  public void print(){
    System.out.println("\nPRINTING Linear System Solver: \n");
    System.out.println();
    System.out.print("        ");
    for(int j = 0; j < Unknowns.getRowDimension(); j++){
      System.out.print(fPrint(j,10) + "  ");
    }
    System.out.println();
    System.out.print("        ");
    for(int j = 0; j < Unknowns.getRowDimension(); j++){
      System.out.print(fPrint(Unknowns.get(j,0),10) + "  ");
    }
    System.out.println();
    for(int i = 0; i < M.getRowDimension(); i++){
      System.out.print("Row: " + i + "  ");
      for(int j = 0; j < M.getColumnDimension(); j++){
        System.out.print(fPrint(getElement(i,j),12) + "\t");
      }
      System.out.print("| " + NHTs.get(i,0));
      System.out.println(); 
    }
    System.out.println();
  }
   
  /**
   * Expands the matrix to include one more equation. Returns the row number
   * for that equation (not the size of the new matrix). Also expands the 
   * NHTs matrix to match. Note that this does not change the number of 
   * columns, and hence the number of defined values, the column index 
   * and other supporting arrays are unchanged.
   * @return the row number for the newly added equation
   */
  public int addEquation(){
    int originalRowSize = M.getRowDimension();
    int originalColSize = M.getColumnDimension();
    // The value to be returned will be the row number of the new
    // equation. This will be the original matrix size - 1 (which gives
    // the original number of the last row) + 1 (the new row's number).
    // int retVal = (M.getRowDimension() + 1) - 1; // BUT:
    int retVal = M.getRowDimension();              // Is simpler!
    // Expand the base matrix
    Matrix newM = new Matrix(originalRowSize + 1, M.getColumnDimension());
    newM.setMatrix(0, originalRowSize - 1, 0, originalColSize - 1, M);
    M = newM;
    // Expand the NHTs matrix
    Matrix newNHTs = new Matrix(originalRowSize + 1, 1);
    newNHTs.setMatrix(0, originalRowSize - 1, 0, 0, NHTs);
    NHTs = newNHTs;
    return retVal;
  }   

  /**
   * Formats a single matrix value to ensure that all values are aligned
   * @param val the value to be printed
   * @param width the maximum width of a value in the matrix (including the
   * decimal point- in other words, the number of characters.
   * @return the formatted value, as a String
   */
  private String fPrint(double val, int width){
    if(val == 0) return ("o                ").substring(0,width);
    double newVal = Math.round(val * Math.pow(10,(width-2)))/Math.pow(10,(width-2));
    return ("" + newVal+ "           ").substring(0,width);
  }

  /**
   * Allows easy determination if all the values resulting from the calculations are zero.
   * This is particularly relevant when the Lineas System is being used for certain kinds
   * of calculations, especially water flow.
   * @return true if all calculated values are zero, false otherwise.
   */
  public boolean isNotZero(){
    for(int i = 0; i < Unknowns.getRowDimension(); i++){
      if(Unknowns.get(i, 0) != 0) return true; 
    }
    return false;
  }

  /**
   * Returns the String indicating if an error was encountered during the calculations.
   * @return A String error message.
   */
  public String getErrorMessage(){
    if(errorMessage == null) return "No error.";
    return errorMessage;
  }
  
}
