package abcm.math;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;

/**
 * A Point is fundamentally a set of three-dimensional coordinates. 
 * The Point class provides basic geometric and trigonometric functions for
 * manipulating 3-D points as a unit and making measurements between
 * points.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Point implements ABCMDeepCloneable{

  /**
   * A static value used to assign each point a unique ID
   */
  private static long pointNumbers = 0;
  
  /**
   * The X position of this Point
   */
  protected double X;
  
  /**
   * The Y position of this Point
   */
  protected double Y;
  
  /**
   * The Z position of this Point
   */
  protected double Z;
  
  /**
   * A name for this point.
   */
  protected String name;
  
  /**
   * An Identifier for this point.
   */
  protected String ID;
  
  /**
   * Constructor that takes two arguments: X and Y. Z is assumed to be zero
   * @param x New X coordinate for this point
   * @param y New Y coordinate for this point
   */
  public Point(double x, double y){
    this(x, y, 0, null);        
  }
  
  /**
   * Constructor that takes two positional arguments, X and Y, assuming Z to be zero, and
   * a third that specifies the name of the Point.
   * @param x X position
   * @param y Y position
   * @param n new Point name
   */
  public Point(double x, double y, String n){
    this(x, y, 0, n);
  }
  
  
  /**
   * Constructor that takes three arguments: X, Y, and Z coordinates.
   * @param x New X coordinate for this point
   * @param y New Y coordinate for this point
   * @param z New Z coordinate for this point
   */
  public Point(double x, double y, double z){
    this(x, y, z, null);    
  }
  
  /**
   * Constructor that takes three positional arguments, X, Y, and Z, and
   * a fourth that specifies the name of the Point.
   * @param x X position
   * @param y Y position
   * @param z Z position
   * @param n new Point name
   */
  public Point(double x, double y, double z, String n){
    X = x;
    Y = y;
    Z = z;
    if(Double.isNaN(z))
        System.out.println("CREATING A POINT WITH Z NaN");
    if(Double.isInfinite(z))
        System.out.println("CREATING A POINT WITH Z INFINITY");
    ID = "P" + pointNumbers;
    pointNumbers++;
    name = (n == null ? ID : n);    
  }
   
  
  
  /* Geometry */

  /**
   * Returns the 3-D Euclidean distance from the calling instance to 
   * the other specified point. 
   * @param p Point to which distance is to be measured.
   * @return The distance from this point to p
   */
  public double distanceToPoint(Point p){
    return Geometry.distance(this, p);
  }
  
  /**
   * Returns the distance from the calling instance to the other specified
   * point ignoring elevation difference.
   * This is useful for examining 2-D maps.
   * @param p Point to which distance is to be measured.
   * @return The distance from this point to p, ignoring the Z dimension
   */
  public double planarDistanceToPoint(Point p){
    return Geometry.planarDistance(this, p);
  }  
  
  /**
   * Returns the slope from the calling instance to the other specified
   * point. 
   * Slope is returned as a fraction (rise over run).
   * @param p Point to which slope is to be calculated
   * @return The slope from this point to p
   */
  public double slopeToPoint(Point p){
    return Geometry.slope(this, p);
  }

  /**
   * Compare two points.
   * @param p Point to which this one is to be compared
   * @return True if this point's X, Y, and Z coordinates are all equal to p's, false otherwise.
   */
  public boolean isEqual(Point p){
    return ((X == p.getX()) && (Y == p.getY()) && (Z == p.getZ())); 
  }

  /**
   * Compare two points, only on X and Y dimensions.
   * @param p the point to which this one is to be compared
   * @return true if the point has the same X and Y coordinates
   * as this one, false otherwise.
   */
  public boolean isPlanarEqual(Point p){
//    if(
//        (!(  (X == p.getX()) && (Y == p.getY())  )) 
//      && ((Math.abs(X - p.getX()) < .0000001) && (Math.abs(Y - p.getY()) < .0000001))) 
//          System.out.println("CLOSE TO PLANAR EQUAL!!!!\n   " + coordString(20) + "\n   " + p.coordString(20) + " " + (X - p.getX()) + " " + (Y - p.getY()));
    return isPlanarEqual(p, Geometry.epsilon);
  }
  
  /**
   * Returns true if the second point is within the specified epsilon of this one.
   * Note that distance to epsilon is calculated along x and y axis independently,
   * not pythagorean
   * @param p the specified other point
   * @param epsilon the error value to be used
   * @return true if the two points are within epsilon of each other on both
   * x and y axes, false otherwise
   */
  public boolean isPlanarEqual(Point p, double epsilon){
    return ((Math.abs(X - p.getX()) < epsilon) && (Math.abs(Y - p.getY()) < epsilon));
  }
  
  /**
   * Gives the planar heading (pilot's compass, 0 = north, clockwise, in radians) from
   * this point to another. Returns zero if points are equal (x1 = x2, y1 = y2).
   * Heading is given in Radians. 
   * @param p Point to which the heading is to be calculated
   * @return planar heading, in radians (0 = north, clockwise increasing),
   *     from calling instance to p
   */
  public double headingTo(Point p){
    return Geometry.heading(this, p);
  }
  
  /**
   * Returns the difference in heading from one vector to another.
   * This point is the joint point; takes as arguments the starting
   * and ending points, so that V1 = p1 to calling instance, V2 =
   * calling instance to p3.
   * @param p1 Start of V1
   * @param p3 End of V2
   * @return Degree of angle change from V1 to V2. 
   */
  public double headingChange(Point p1, Point p3){
    return Geometry.headingChange(this, p1, p3);
  }
  
  
  /* Reporting */
  
  /**
   * Returns a String to print out the coordinates of this Point.
   * Assumes 1/100 accuracy is desired.
   * Uses X: x  Y: y  Z: z format.
   * @return String indicating the coordinates of this point
   */
  public String coordString(){  return coordString(2);   }
  
  /**
   * Returns a String to print out the coordinates of this Point.
   * Uses X: x  Y: y  Z: z format.
   * @param precision the precision to which the coordinates are to be shown 
   * @return String indicating the coordinates of this point to the specified accuracy
   */
  public String coordString(int precision){
    return "X: " + BasicMath.roundDouble(X, precision) + "   " +
           "Y: " + BasicMath.roundDouble(Y, precision) + "   " +
           "Z: " + BasicMath.roundDouble(Z, precision);
  }

  /**
   * Returns a String to print out the coordinates of this point, in abbreviated form.
   * Format is (x,y) (Z is omitted).
   * Assumes 1/100 accuracy is desired.
   * @return String indicating the (X,Y) coordinates of this point
   */
  public String coordStringShort(){  return coordStringShort(2);   }

  /**
   * Returns a String to print out the coordinates of this point, in abbreviated form.
   * Format is (x,y) (Z is omitted). Precision is user-specified.
   * @param precision the precision to which the coordinates are to be shown
   * @return String indicating the (X,Y) coordinates of this point
   */  
  public String coordStringShort(int precision){
      return "(" + BasicMath.roundDouble(X, precision) + "," +
                   BasicMath.roundDouble(Y, precision) + ")";
  }
  
  /* Get Methods */
  
  /**
   * Gets the name of the Point object
   * @return the Point's name
   */
  public String getID(){   return ID;   }

  /**
   * Gets the name of this point.
   * @return the point name
   */
  public String getName(){ return name; }

  
  /**
   * Gets the X coordinate of this point.
   * @return The X coordinate of this point
   */
  public double getX(){    return X;    }

  /**
   * Gets the Y coordinate of this point.
   * @return The Y coordinate of this point
   */
  public double getY(){    return Y;    }

  /**
   * Gets the Z coordinate of this point.
   * @return The Z coordinate of this point
   */
  public double getZ(){    return Z;    }
  

  /** 
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore) 
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Point ret = new Point(this.X, this.Y, this.Z, this.name);
    cloneStore.put(this, ret);
    return ret;
  }

  
}
