package abcm.math;

import java.util.Collection;
import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableVector;

/**
 * A Polygon object consists of a Vector of Point objects that
 * are considered to be the vertices of the Polygon.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Polygon implements ABCMDeepCloneable{

  /** Debugging constant for debugging the 'buffer' routine */
  private static boolean DEBUG_BUFFER = false;
  
  /** The collection of points that constitute the vertices of this Polygon  */
  protected ABCMDeepCloneableVector points;
  
  /** Collection of sides for this Polygon */
  private ABCMDeepCloneableVector sides = null; 

  
  /** Minimum x value of all points in this Polygon */
  private double minX = Double.NaN;
  
  /** Maximum x value of all points in this Polygon */
  private double maxX = Double.NaN;
  
  /** Minimum y value of all points in this Polygon */
  private double minY = Double.NaN;
  
  /** Maximum y value of all points in this Polygon */
  private double maxY = Double.NaN;
  
  /** Minimum z value of all points in this Polygon */
  private double minZ = Double.NaN;
  
  /** Maximum z value of all points in this Polygon */
  private double maxZ = Double.NaN;
  
  /** Center point */
  private Point centerPoint = null;
  
  /**
   * Null constructor for cloning only
   * Constructor
   */
  private Polygon(){ }
  
  /**
   * A constructor that takes a Vector of Points.
   * @param pointsList the vertices of the Polygon
   */
  public Polygon(Collection pointsList){   
    points = new ABCMDeepCloneableVector();
    points.addAll(pointsList);
  }
  
  /**
   * Gets the Vector of vertices.
   * @return the Vector containing the vertices of the Polygon
   */
  public Vector getPoints(){   return new Vector(points);  }
  
  /**
   * Gets the point that is the simple center of this polygon
   * (The average of all the vertices' coordinates)
   * @return the simple center of this polygon, as a new Point
   */
  public Point getCenterPoint(){
    if(centerPoint == null){
      if(points.size() == 0) centerPoint = new Point(0,0,0);
      else{
        double xSum = 0;
        double ySum = 0;
        double zSum = 0;
        for(int i = 0; i < points.size(); i++){
          Point p = (Point)points.elementAt(i);
          xSum += p.getX();
          ySum += p.getY();
          zSum += p.getZ();
        }
        centerPoint =  new Point(xSum / points.size(), 
                                 ySum / points.size(), 
                                 zSum / points.size());
      }
    }
    return centerPoint;
  }
  
  /**
   * Determines if the first three points in the polygon are 'clockwise' or
   * 'counterclockwise'. If the former, reverses all polygon's points. 
   */
  public void makeCounterClockwise(){
    if(points.size() <= 2) return;
    Point p1 = (Point)points.get(0);
    Point p2 = (Point)points.get(1);
    Point p3 = (Point)points.get(2);

    double det = BasicMath.determinant3D(p1.getX()*1000, p1.getY()* 1000, 1, p2.getX()*1000, p2.getY()*1000, 1, p3.getX()*1000, p3.getY()*1000, 1);
    if(det < 0){ // Must reverse the direction
      Vector np = new Vector();
      for(int i = points.size() - 1; i >= 0; i--){
        np.addElement(points.elementAt((i)));        
      }
      points.clear();
      points.addAll(np);
    }   
  }
  
  /**
   * Removes any points that are actually equal
   */
  public void removeDuplicatePoints(){
    for(int i = points.size() - 1; i > 0; i--){
      if(((Point)points.get(i)).isEqual((Point)points.get(i-1))) points.remove(i);
    }
  }
  

  /**
   * Sets the min and max x, y, and z values for this polygon
   */
  private void setMinsAndMaxes(){
    minX = Double.POSITIVE_INFINITY;
    maxX = Double.NEGATIVE_INFINITY;
    
    minY = Double.POSITIVE_INFINITY;
    maxY = Double.NEGATIVE_INFINITY;
    
    minZ = Double.POSITIVE_INFINITY;
    maxZ = Double.NEGATIVE_INFINITY;
        
    for(int i = 0; i < points.size(); i++){
      Point P = (Point)points.get(i);
      if(P.X > maxX) maxX = P.X;
      if(P.X < minX) minX = P.X;
      
      if(P.Y > maxY) maxY = P.Y;
      if(P.Y < minY) minY = P.Y;
      
      if(P.Z > maxZ) maxZ = P.Z;
      if(P.Z < minZ) minZ = P.Z;      
    }
  }
  
  
  /**
   * @return highest x value of all points in this polygon
   */
  public double getMaxX(){
    if(Double.isNaN(maxX)) setMinsAndMaxes();
    return maxX;
  }

  /**
   * @return lowest x value of all points in this polygon
   */
  public double getMinX(){
    if(Double.isNaN(minX)) setMinsAndMaxes();
    return minX;
  }
  
  /**
   * @return highest y value of all points in this polygon
   */
  public double getMaxY(){
    if(Double.isNaN(maxY)) setMinsAndMaxes();
    return maxY;
  }

  /**
   * @return lowest y value of all points in this polygon
   */
  public double getMinY(){
    if(Double.isNaN(minY)) setMinsAndMaxes();
    return minY;
  }

  /**
   * @return highest z value of all points in this polygon
   */
  public double getMaxZ(){
    if(Double.isNaN(maxZ)) setMinsAndMaxes();
    return maxZ;
  }

  /**
   * @return lowest z value of all points in this polygon
   */
  public double getMinZ(){
    if(Double.isNaN(minZ)) setMinsAndMaxes();
    return minZ;
  }

  /**
   * Returns true if the polygon's highest and lowest Z values encompass
   * the specified elevation. Inclusive comparison (less-than-or-equal-to,
   * greater-than-or-equal-to) is used.
   * @param z the elevation to be tested
   * @return true if minZ is less than or equal to Z and Z is less than or equal to maxZ
   */
  public boolean includesElevation(double z){
    if(Double.isNaN(minZ)){
          
    }
    return ((z <= maxZ) && (z >= minZ));
  }
  
  /**
   * Returns true if a 2-D projection of this polygon includes this point,
   * false otherwise.
   * @param P point to be tested
   * @return true if a 2-D projection of this polygon includes this point,
   * false otherwise.
   */
  public boolean pointIsInside(Point P){
    Point C = getCenterPoint();
    Point v1, v2;
    for(int i = 0; i < points.size() - 1; i++){
      v1 = (Point)points.get(i);
      v2 = (Point)points.get(i+1);
//      System.out.println("Point Is Inside Testing (A):");
//      System.out.println("   P1 = " + v1.coordString());
//      System.out.println("   P2 = " + v2.coordString());
//      System.out.println("   C  = " + C.coordString());
//      System.out.println("   P  = " + P.coordString());
//      System.out.println("   pointIsInside(C, P1, P2, P) = " + Geometry.pointIsInside(C, v1, v2, P));
//      System.out.println("   pointLiesOnSegment(P1, P2, P) = " + Geometry.pointLiesOnSegment(v1, v2, P));
      if((Geometry.pointIsInside(C, v1, v2, P)) || Geometry.pointLiesOnSegment(v1, v2, P)) return true;
    }
    // Close the loop
    v1 = (Point)points.get(points.size()-1);
    v2 = (Point)points.get(0);
//    System.out.println("Point Is Inside Testing (B):");
//    System.out.println("   P1 = " + v1.coordString());
//    System.out.println("   P2 = " + v2.coordString());
//    System.out.println("   C  = " + C.coordString());
//    System.out.println("   P  = " + P.coordString());
//    System.out.println("   pointIsInside(C, P1, P2, P) = " + Geometry.pointIsInside(C, v1, v2, P));
//    System.out.println("   pointLiesOnSegment(P1, P2, P) = " + Geometry.pointLiesOnSegment(v1, v2, P));
    if((Geometry.pointIsInside(C, v1, v2, P)) || Geometry.pointLiesOnSegment(v1, v2, P)) return true;
    return false;    
  }
  
  /**
   * Determines if a given point is inside this polygon or not.
   * See pointIsInside(Point, Point, Point, Point) for details.
   * @param x the x coordinate of the point to be tested
   * @param y the y coordinate of the point to be tested
   * @return true if the point is inside this triangle, false otherwise
   */  
  public boolean pointIsInside(double x, double y){
    return pointIsInside(new Point(x,y));
  }
  
  /**
   * Returns true if a 2-D projection of this polygon includes this
   * point (but the point does not lie on one of the sides of the projection),
   * false otherwise
   * @param P point to be tested
   * @return true if the point is inside but not on the projection of the polygon
   */
  public boolean pointIsInsideExclusive(Point P){
    
    boolean pointIsInside = false;
    
    Point C = getCenterPoint();
    Point v1, v2;
    int i = 0;
    while(( i < points.size() && pointIsInside == false)){
      v1 = (Point)points.get(i);
      v2 = (Point)points.get(((i+1)< points.size() ? i + 1: 0));
//      System.out.println("Point Is Inside Exclusive Testing (A):");
//      System.out.println("   P1 = " + v1.coordString(10));
//      System.out.println("   P2 = " + v2.coordString(10));
//      System.out.println("   C  = " + C.coordString(10));
//      System.out.println("   P  = " + P.coordString(10));
//      System.out.println("   pointIsInside(C, P1, P2, P) = " + Geometry.pointIsInside(C, v1, v2, P));
//      System.out.println("   pointLiesOnSegment(P1, P2, P) = " + Geometry.pointLiesOnSegment(v1, v2, P));
      pointIsInside = Geometry.pointIsInside(C, v1, v2, P);
      i++;
    }
    if(!pointIsInside) return false; // Point is not in polygon at all- return false
    
    boolean pointIsOnEdge = false;
    i = 0;
    getSides();
    while(i < sides.size() && pointIsOnEdge == false){
      LineSegment LS = (LineSegment)sides.get(i);
//      System.out.println("SEGMENT: "  + LS.display() + " result: " + Geometry.pointLiesOnSegment(LS, P));
      pointIsOnEdge = Geometry.pointLiesOnSegment(LS, P);
      i++;
    }
    return !pointIsOnEdge;  // If the point lies on one of the edges, it is 'excluded'
    
//    if(Geometry.pointIsInside(C, v1, v2, P) && !Geometry.pointLiesOnSegment(v1, v2, P)){
//        return true;
//      }
      
    
//    return false;
  }
  
  
  /**
   * Returns a new version of this polygon expanded outward
   * from a central point
   * @param fraction by which this polygon should be expanded (or reduced; fraction will generally be greater than 1)
   * @return new, larger polygon
   */
  public Polygon expand(double fraction){
    Vector V = new Vector();
    Point center = this.getCenterPoint();
    for(int i = 0; i < points.size(); i++){
      V.add(Geometry.getOffsetPoint(center, Geometry.heading(center, (Point)points.get(i)), center.distanceToPoint((Point)points.get(i)) * fraction));
    }
    return new Polygon(V);
  }
  
  /**
   * returns a new polygon with vertices some set additional distance further away from the center point
   * @param extra distance to be added
   * @return new polygon
   */
  public Polygon enlarge(double extra){
    Vector V = new Vector();
    Point center = this.getCenterPoint();
    System.out.println("\n\nEnlarging...");
    for(int i = 0; i < points.size(); i++){
      V.add(Geometry.getOffsetPoint(center, Geometry.heading(center, (Point)points.get(i)), center.distanceToPoint((Point)points.get(i)) + extra));
      System.out.println("Original: " + ((Point)points.get(i)).coordString());
      System.out.println("New: " + (Geometry.getOffsetPoint(center, Geometry.heading(center, (Point)points.get(i)), center.distanceToPoint((Point)points.get(i)) + extra)).coordString());
    }
    return new Polygon(V);
  }
  
  
  /**
   * Returns a new polygon large enough to include the given buffer distance outward from
   * the existing vertices
   * @param bufferDistance distance to use
   * @return new polygon
   */
  public Polygon buffer(double bufferDistance){
    Vector V = new Vector();
    if(DEBUG_BUFFER) System.out.println("Buffering...");
    for(int i = 0; i < points.size(); i ++){
      Point previousVertex = (i == 0 ?(Point)points.get(points.size() -1 ) : (Point)points.get(i-1));
      Point currentVertex = (Point)points.get(i);
      Point nextVertex =  ((i + 1) == points.size() ? (Point)points.get(0) : (Point)points.get(i+1));
      double newHeadingForOffset = Geometry.heading(previousVertex, currentVertex) + 
                                   -1 * ((Math.PI - Geometry.headingChange(currentVertex, previousVertex, nextVertex)) / 2d);
      Point newPoint = Geometry.getOffsetPoint(currentVertex, newHeadingForOffset, bufferDistance);
      
      V.add(newPoint);
      if(DEBUG_BUFFER) System.out.println("Original: " + currentVertex.coordString() + " " + currentVertex.ID);
      if(DEBUG_BUFFER) System.out.println("New: " + newPoint.coordString());
      
      previousVertex = currentVertex; // Set up for next
    }
    return new Polygon(V);
  }
  
  /**
   * Gets a list of the sides of this Polygon
   * @return a Vector containing a line segments that form the sides of this Polygon
   */
  public Vector getSides(){
    if(sides == null){
      sides = new ABCMDeepCloneableVector();
      for(int j = 0; j < (points.size() - 1); j++){
        sides.add(new LineSegment((Point)points.get(j), (Point)points.get(j+1)));
      }
      sides.add(new LineSegment((Point)points.get(points.size() - 1), (Point)points.get(0)));
    }
    return new Vector(sides);        
  }
  
  
  /* Reporting */
  
  /**
   * Lists the Points to the standard output.
   */
  public void listPoints(){
    for(int i = 0; i < points.size(); i++){
      System.out.print(((Point)points.elementAt(i)).getName() + " ");
    }
  }
  
  /** 
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore) 
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Polygon clone = new Polygon();
    clone.points = (ABCMDeepCloneableVector)points.getDeepClone(cloneStore);
    if(sides != null) clone.sides = (ABCMDeepCloneableVector)sides.getDeepClone(cloneStore);
    clone.minX = minX;
    clone.maxX = maxX;
    clone.minY = minY;
    clone.maxY = maxY;
    clone.minZ = minZ;
    clone.maxZ = maxZ;
    if(centerPoint != null) clone.centerPoint = (Point)cloneStore.getClone(centerPoint);
    return clone;
  }
  
}

