package abcm.math;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;

/**
 *
 * A Date in the simulation. The simulation will be conceptually divided into years,
 * months, and days, but the Java 'Date' object only works after 1970. This object
 * will provide needed support for the simulation's concept of 'Date'
 * 
 * This date object is a double that represents the number of days since some 
 * zero date. The fractional portion represents the time of day (i.e. .5 means 'noon').
 * 
 * Note that there are no leap years- all years have 365 days, February always 
 * has 28 days.
 * 
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SimDate implements ABCMDeepCloneable, Comparable{
  
  /**
   * A date is fundamentally a double, measuring days since January 1, 9999 BC.
   */
  protected double dayCount;
    
  /**
   * Constructor; value is initialized to zero
   */
  public SimDate(){
    dayCount = 0.0;
  }
  
  /**
   * Constructor; value is initialized to the value given,
   * which is passed as a double
   * @param d new daycount value.
   */
  public SimDate(double d){
    dayCount = d;
  }

  /**
   * Constructor in which the date object is created using
   * as its initial value the value equivalent to the
   * date indicated by the String passed as an argument.
   * @param s new value for the date object, in String form.
   */  
  public SimDate(String s){
    dayCount = parseString(s);
  }
  
  /**
   * Returns the value of this date object. Value is a count of days since
   * Jan 1, 9999 BC
   * @return value of this date object, in days since 1 Jan 9999 BC.
   */
  public double value(){
    return dayCount;
  }

  /**
   * Increment the day count value of this date object.
   */  
  public void incDay(){
    dayCount++;
  }

  /**
   * Increase the value of this date object by moving it
   * one month forward. Adds the number of days in the current month
   * to the current daycount.
   *
   */  
  public void incMonth(){
    dayCount+= Year.daysInMonth(Year.getMonthInt((long)(dayCount % 365)));
  }
  
  /**
   * Move forward one year (365 days) of simulation time.
   */
  public void incYear(){
    dayCount+= 365;
  }

  /**
   * Return a number indicating the year of this instance of
   * the SimDate object; will be a nonzero number
   * between -9999 and 9999.
   * @return the year, positive for AD and negative for BC.
   */  
  public long getYear(){
    return findYear(dayCount);
  }
  
  /**
   * Returns the year (9999BC - 9999AD) indicated by the daycount
   * passed as double
   * 
   * A year of zero indicates a fault in the routine;
   * there is no year zero.
   * @param d the date to be tested, as a straight count of days since 1 Jan  9999 BC 
   * (that is, corresponding to the 'daycount' variable of another SimDate instance).
   * @return the year; BC years are negative.
   */
  public static long findYear(double d){
    long retVal =0;
    long dayC = (long)Math.floor(d);

    // Determine if this is BC or AD
    if(dayC < 9999 * Year.DAYS_IN_YEAR){
      retVal = -9999 + (long)Math.floor((dayC / Year.DAYS_IN_YEAR));
    }
    else{
      dayC = (dayC - (9999 * Year.DAYS_IN_YEAR));
      retVal = (long)Math.floor((dayC / Year.DAYS_IN_YEAR)) + 1;
    }
  
    return retVal;
  
  }
  
  /**
   * Gets the current value of this SimDate object displayed as
   * a string in proper SimDate format
   * @return A string indicating the value of this SimDate object,
   * i.e. "01-JAN-0100 AD"
   */
  public String getString(){
    return convertToString(dayCount);
  }
  
  /**
   * Gets the current value of this SimDate object displayed as
   * a string in proper SimDate format
   * @param includeZeroFractional if true, includes a fractional part for 
   * a decimal even if the decimal is zero (omitted by default)
   * @return A string indicating the value of this SimDate object,
   * i.e. "01-JAN-0100 AD"
   */
  public String getString(boolean includeZeroFractional){
    return convertToString(dayCount, includeZeroFractional);
  }
  
  /**
   * Get the fractional part of this SimDate (represents a fraction
   * of a 24-hour day)
   * @return the fraction part of the day count; zero if non has been set
   */
  public double getFraction(){ return dayCount - Math.floor(dayCount); }
  
  /**
   * Sets the date of this SimDate object to a value equal to the
   * date passed as a string
   * @param s the new date, in the form "01-JAN-0100 AD"
   * @return true if successful, false otherwise
   */
  public boolean setDate(String s){
    boolean retVal = false;
    double newVal = parseString(s);
    if (newVal > -1) {
      dayCount = newVal;
      retVal = true;
    }
    return retVal; 
  }
  
  /**
   * Sets the date of this SimDate object to a value equal to the
   * value passed as a double
   * @param d the new date, in count of days since 01-JAN-9999 BC
   * @return true if successful, false otherwise
   */  
  public boolean setDate(double d){
    boolean retVal = false;
    if (d >= 0){
      dayCount = d;
      retVal = true;
    }
    return retVal;    
  }
  
  /**
   * Converts a double to a String showing the date represented
   * by the double.
   * @param d the date to be converted, as a straight count of days since 1 Jan 9999 BC 
   * (that is, corresponding to the 'daycount' variable of another SimDate instance).
   * @return A String representing this date in SimDate format (i.e. 01-JAN-1000-AD)
   */
  public static String convertToString(double d){
    return convertToString(d, false);
  }
  
  /**
   * Converts a double to a String showing the date represented by
   * the double.
   * @param d the date to be converted, as a count of days since 1 Jan 9999 BC
   * @param includeZeroFractional if true, includes a fractional part for decimal
   * even if the decimal is zero (omitted by default).
   * @return A String representing this date in SimDate format (i.e. 01-JAN-1000-AD)
   */
  public static String convertToString(double d, boolean includeZeroFractional){
    String retVal = "";
    long dayOfYear = ((long)Math.floor(d) % Year.DAYS_IN_YEAR);

    retVal = "00" + Year.dayOfMonth(dayOfYear);
    retVal = retVal.substring(retVal.length()-2);
    retVal += "-";
    // Now determine the month
    retVal += Year.getMonthABB(dayOfYear);
    retVal += "-";

    // And the year
    if(d < 365){
      retVal += "SEASONS";
    }
    else {
      long y = findYear(d);
      String Y = "0000" + Math.abs(y);
      Y = Y.substring(Y.length() - 4);
      if (y < 0)
        retVal += Y + "-BC";
      else
        retVal += Y + "-AD";
    }
     
    // If there is no fractional part, done...
    // Otherwise, add it
    if(includeZeroFractional || d != Math.floor(d)){
      retVal += " +" + BasicMath.roundDouble((d - Math.floor(d)),3);
    }    
    return retVal;    
  }
  
  /**
   * Converts a string in proper SimDate format to the double
   * in which it is stored.
   * Proper SimDate format is like "15-MAR-0044-BC"
   * Seasonal Dates ("15-MAR-SEASONS") are supported; these return
   * a double equal to the day of the year only.
   * @param s String to be converted to double
   * @return the value of the date represented by the String passed.
   */
  public static double parseString(String s){
    if(s == null)       return -1;
    if(s.length() < 14) return -1;
    double retVal = -1;
    try{
      int dayOfYear = Year.dayOfYear(s.substring(0,6));
      // Determine if this is a seasonal date or not
      if(s.substring(7,14).compareTo("SEASONS") == 0){
        retVal++;  // First day of year is day 1, not zero
      } else {
        // Deal with BC years
        if(s.substring(12,14).compareTo("BC") == 0){
          retVal = (9999 - Integer.parseInt(s.substring(7,11))) * Year.DAYS_IN_YEAR;  
        }
        else{
          retVal = 9999 * Year.DAYS_IN_YEAR +
                  (Integer.parseInt(s.substring(7,11)) - 1) * Year.DAYS_IN_YEAR;
        }
      }
      retVal += dayOfYear;
      if(s.indexOf('+') > -1){
        retVal += Double.parseDouble(s.substring(s.indexOf("+")+ 1)); 
      }
      
    }
    catch(NumberFormatException nfe){
      return -1;
    }   
    return retVal;
  }
  
  /**
   * Get the day of the year on which this SimDate falls
   * @return an integer from 1 to 365 indicating the day of the year.
   */
  public int dayOfYear(){
    return Year.dayOfYear(this.getString().substring(0,6));
  }

  /**
   * Sets the fractional part of a date; this corresponds to 'time of day', so that
   * a fraction of .5 means noon.
   * @param fraction fractional value to be set.
   */
  public void setFraction(double fraction){
    dayCount = Math.floor(dayCount) + fraction;
  }
  
  /** 
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore) 
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    SimDate ret = new SimDate(dayCount);
    cloneStore.put(this, ret);
    return ret;
  }

  /**
   * @see java.lang.Comparable#compareTo(java.lang.Object)
   */
  public int compareTo(Object other) {
    if(!(other instanceof SimDate)) return 1;
    double otherDateValue = ((SimDate)other).value();
    return (this.value() > otherDateValue ? 1 : 
                 (this.value() < otherDateValue ? -1 : 0));
  }
  
}
