package abcm.math;

import java.util.Vector;

/**
 *
 *  A Triangle is a special class of polygon. Simple functions are provide to
 *  perform standard calculations that are often required of Triangles (i.e. determine 
 *  whether a given point lies inside or outside the specified triangle).
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
/**
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Triangle extends Polygon{
  
  /* Proxies for the points of this triangle */
  /** Point 1 */
  protected Point P1;  
  /** Point 2 */
  protected Point P2;  
  /** Point 3 */
  protected Point P3;
  
  /**
   * Basic constructor.
   * Re-orders the points to be listed in clockwise order.
   * @param p1 Vertex 1
   * @param p2 Vertex 2
   * @param p3 Vertex 3
   */
  public Triangle(Point p1, Point p2, Point p3){
    super((clockwise(p1, p2, p3) ? 
        makeVector(p1, p3, p2) : 
        makeVector(p1, p2, p3)));
    P1 = (Point)super.getPoints().get(0);
    P2 = (Point)super.getPoints().get(1);
    P3 = (Point)super.getPoints().get(2);    
//    ID = "T" + count++;
  }
  
  /**
   * Creates a vector that lists the three specified points.
   * This method allows the main constructor to take three points and call the
   * superclass's constructor in the first line.
   * @param p1 Vertex 1
   * @param p2 Vertex 2
   * @param p3 Vertex 3
   * @return A vector listing the three points specified
   */
  private static Vector makeVector(Point p1, Point p2, Point p3){
    Vector V = new Vector(3);
    V.add(p1);
    V.add(p2);
    V.add(p3);
    return V;
  }
  
  /**
   * Returns true if the vertices of the triangle are colinear.
   * NOTE: For the moment THIS ALWAYS RETURNS FALSE
   * @return true if the vertices are colinear, false otherwise (always returns false)
   */
  public boolean verticesAreColinear(){    return false;   }
  
  /**
   * Determines if a specified point lies inside the circle that
   * includes all points of this triangle.
   * See pointIsCircumscribed(Point, Point, Point, Point) for details.
   * @param P point to be tested
   * @return true if the point is circumscribed by this triangle, false
   * otherwise. 
   */
  public boolean pointIsCircumscribed(Point P){
    return Geometry.pointIsCircumscribed(P1, P2, P3, P);
  }
  
  /**
   * Calculates the elevation of a Point projected onto the plane defined by
   * this triangle
   * @param P Point to be projected
   * @return projected elevation
   */
  public double getProjectedElevation(Point P){
    return Geometry.getProjectedElevation(P1, P2, P3, P);
  }
  
  /**
   * Calculates the elevation of a Point projected onto the plane defined by
   * this triangle
   * @param x x coordinate of the Point to be projected
   * @param y y coordinate of the Point to be projected
   * @return projected elevation
   */
  public double getProjectedElevation(double x, double y){
    return Geometry.getProjectedElevation(P1, P2, P3, new Point(x,y));
  }
  
  /**
   * Gets a line that crosses this triangle at the specified
   * elevation
   * @param z the requested elevation
   * @return a LineSegment object that crosses this triangle at the specified elevation
   */
  public LineSegment getCrossLine(double z){
    if(!includesElevation(z)) return null;
    if((P1.getZ() == P2.getZ()) && (P1.getZ() == P3.getZ())) return null;  // If the triangle is horizontal, return null
    // The triangle includes the elevation but is not horizontal;
    // This means one of the vertices is above the specified
    // elevation and the other two are below or equal to it, or
    // one of the vertices is below and the other two above or equal to it.
    Point Anchor = P1;
    Point Leg1   = P2;
    Point Leg2   = P3;
    Point temp;
    while(!(((Anchor.getZ() < z) && (Leg1.getZ() >= z) && (Leg2.getZ() >= z))  ||
            ((Anchor.getZ() > z) && (Leg1.getZ() <= z) && (Leg2.getZ() <= z)))){
      if(Anchor.getZ() == z && (((Leg1.getZ() < z) && (Leg2.getZ() < z)) ||
                                ((Leg1.getZ() > z) && (Leg2.getZ() > z)))) {
        return new LineSegment(Anchor,Anchor);
      }
      temp   = Anchor; // Rotate to test...
      Anchor = Leg1;
      Leg1   = Leg2;
      Leg2   = temp;      
    }
    return new LineSegment(Geometry.pointBetween(Anchor, Leg1, Math.abs(Anchor.getZ() - z) / Math.abs(Anchor.getZ() - Leg1.getZ())),
                           Geometry.pointBetween(Anchor, Leg2, Math.abs(Anchor.getZ() - z) / Math.abs(Anchor.getZ() - Leg2.getZ())));    
  }
  
  /**
   * Gets the area of this triangle projected onto the horizontal plane
   * @return the area of this triangle projected onto the horizontal plane 
   */
  public double getArea(){
    return Geometry.areaOfTriangle(P1, P2, P3);
  }
  
  /**
   * Gets the volume of the projection of this triangle
   * onto a plane at z = the specified value.
   * @param z the plane on which this triangle is to be projected,
   * sweeping out the requested volume
   * @return the volume of the solid described
   */
  public double getVolumeAbove(double z){
    double A = getArea();
    double avgHeight = ((P1.getZ() - z) +  
                        (P2.getZ() - z) + 
                        (P3.getZ() - z)) / 3d;
    return (A * avgHeight);
  }
  
  /**
   * Returns true if any of this triangle's sides is intersected by
   * the specified Line Segment in 2D.
   * @param L the segment to be tested
   * @return true if intersection occurs, false if not
   */
  public boolean anySideIsIntersected(LineSegment L){
    LineSegment S1 = new LineSegment(P1, P2);
    LineSegment S2 = new LineSegment(P2, P3);
    LineSegment S3 = new LineSegment(P3, P1);
    return L.intersectsIn2D(S1) ||
           L.intersectsIn2D(S2) ||
           L.intersectsIn2D(S3);
  }
  
  /**
   * CHANGED FROM 'counterClockwise' 9/15/08 because it actually
   * returns TRUE if the points are clockwise and FALSE if not
   * @param P1
   * @param P2
   * @param P3
   * @return true if the points are in clockwise order, false otherwise
   */
  public static boolean clockwise(Point P1, Point P2, Point P3){
    Point P1t = P1;
    Point P2t = new Point(P1.X + 1000 * (P2.X - P1.X), P1.Y + 1000 * (P2.Y - P1.Y));
    Point P3t = new Point(P1.X + 1000 * (P3.X - P1.X), P1.Y + 1000 * (P3.Y - P1.Y));
    double det = BasicMath.determinant3D(P1t.getX(), P1t.getY(), 1, 
                                         P2t.getX(), P2t.getY(), 1, 
                                         P3t.getX(), P3t.getY(), 1);
    return (det < 0);
  }
  
  
}
