package abcm.math;

/**
 * A Year object represents a collection of months of varying lengths.
 * We use this to navigate and ameliorate the oddities of our calendar system.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Year {

  /**
   * Number of days in a year. Note that this is always 365; no leap years are allowed.
   */
  public static final int DAYS_IN_YEAR = 365;
  
  /**
   * Array of months in the standard year.
   */
  public static final Month[] months = 
  { 
    new Month("January",   31,   0),
    new Month("February",  28,  31),
    new Month("March",     31,  59),
    new Month("April",     30,  90),
    new Month("May",       31, 120),
    new Month("June",      30, 151),
    new Month("July",      31, 181),
    new Month("August",    31, 212),
    new Month("September", 30, 243),
    new Month("October",   31, 273),
    new Month("November",  30, 304),
    new Month("December",  31, 334) 
  };
    
  /**
   * Returns the day of the month (i.e. 'the 15th') given a day
   * of the year (i.e. day 299).
   * 
   * This value is not zero-based; the first day is day 1. The
   * number passed, however, should be zero-based, i.e. day
   * 10 of the year = Jan 11th (Jan 1 is day zero).
   * @param dayOfYear
   * @return day of the month
   */
  public static int dayOfMonth(long dayOfYear){
    int retVal = 0;
    dayOfYear++;
    int i = 0;
    while(dayOfYear > months[i].getDays()){
      dayOfYear-=months[i].getDays();
      i++;
    }
    retVal = (int)dayOfYear;
    
    return retVal;
  }
  
  /**
   * Returns the abbreviated name of the month given the day of the
   * year. Days are passed as zero-based (Jan 1 = 0).
   * @param dayOfYear the number of the day of the year 
   * @return Abbreviated month name, i.e. JAN.
   */
  public static String getMonthABB(long dayOfYear){
    return months[getMonthInt(dayOfYear)].getABB();
  }
    
  /**
   * Returns an integer indicating the month into which a given day of
   * the year falls. Days of years are zero-based (Jan 1 = day 0);
   * Months are zero-based (Jan = 0).
   * 
   * @param dayOfYear
   * @return integer indicating month (Jan = 0; Feb = 1...)
   */
  public static int getMonthInt(long dayOfYear){
    dayOfYear++;
    int i = 0;
    while(dayOfYear > months[i].getDays()){
      dayOfYear-=months[i].getDays();
      i++;
    }
    return i;
  }
  
  /**
   * Returns the number of days in the specified month.
   * Months passed are zero-based (Jan = 0)
   * @param month integer representing which month of the year
   * @return number of days in the requested month.
   */
  public static int daysInMonth(int month){
    return months[month].getDays();
  }
  
  /**
   * Returns an integer indicating which month based on the 
   * string passed; strings can be either full month names
   * or ABB format (i.e. January or JAN).
   * @param m
   * @return integer (zero-based) indicating which month
   */
  public static int monthFromString(String m){
    int retVal = 0;
    for(int i = 11; i>0; i--){
      if((months[i].getName().compareTo(m) == 0) || 
         (months[i].getName().substring(0,3).toUpperCase()).compareTo(m) == 0){
        retVal = i;
        break; 
      }
    }
    return retVal;
  }
  
  /**
   * Returns the day of the year from a string in the form
   * '01-JAN' or '01-January'
   * @param s String whose date is to be considered
   * @return day of the year represented by that date
   */
  public static int dayOfYear(String s){
    int m = monthFromString(s.substring(3));
    int d = Integer.parseInt(s.substring(0,2));
    return (months[m].getMonthFirstDay() + d - 1);
  }
    
}