package abcm.model;

import abcm.db.managers.CodeModelOptionValuesManager;
import abcm.db.managers.CommandsManager;
import abcm.db.managers.GlossaryManager;
import abcm.math.SimDate;
import abcm.model.tools.ScriptLibrarian;
import abcm.sim.AbstractModelComponent;
import abcm.sim.HistoryInstruction;
import abcm.utilities.QuotedLineParser;

/**
 * Root class for all ABCM parser objects. The parser converts text input into
 * operational instructions. This class defines the commands inherent in
 * an ABCM modeling environment; extensions of this class can add additional
 * commands as needed, following the pattern given in the constructor here.  
 * 
 * Note that the inner classes here are not documented in JavaDoc; see the online
 * definitions of commands for details on each one.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractABCMParser extends AbstractModelComponent{
  
  /** Maximum number of command indexes  */
  public static final int MAX_COMMANDS = 2000;
  
  /** Environment in which commands are executed  */
  protected AbstractABCModelEnvironment E;
  
  /** String containing a report of the failure to add commands  */
  protected String onFailureReport;
  
  /** List of commands, to maximum index given by MAX_COMMANDS  */
  private ParseAction[] commands = new ParseAction[MAX_COMMANDS];  
  
  /** Maximum number of arguments allowed                       */
  public static final int MAX_ARGS = 10; // maximum entries on a line in a command
  
  /**
   *
   * Takes a single command entry as an entire string
   * and parses it into discrete chunks. Uses quotes
   * to make chunks that contain spaces. Last quote is
   * optional, so that:
   * 
   * This is a "test of the quoting functions
   * 
   * parses to:
   * 
   * [0] = This
   * [1] = is
   * [2] = a
   * [3] = test of the quoting functions
   * 
   * Note: returns a String of 10 entries regardless of the length of
   * the input.
   * @param cText
   * @return a String array with the pieces of this String as entries
   */
  public static String[] parseQuoted(String cText){
    return QuotedLineParser.parseQuoted(cText, MAX_ARGS);
  }  

  // 
  
  /** Used internally to allow text entries to be converted to numeric option values  */
  private CodeModelOptionValuesManager cmovm;
  
  /* Commands  implemented as objects that extend ParseAction */

    /* Environment Commands */
  
  // Logging in
  class Login                          extends ParseAction{
    private String name = "Login";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult u = parseArgString(name, command, 1);  if(!u.success) return u;
      ABCMResult p = parseArgString(name, command, 2);  if(!p.success) return p;
      return E.login(u.text, p.text);
    }
  }
  
  // Help
  class Define                         extends ParseAction{
    private String name = "Define";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult term = parseArgString(name, command, 1); if(!term.success) return term;
      GlossaryManager gm = new GlossaryManager(E.getDBGateway());
      return new ABCMResult(gm.getDefinition(term.text), true);                            
    }
  }
  class Help                           extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      return new ABCMResult(getCommandsManager().getListOfCommandsAsString(), true);   
    }
  }
  class HelpOn                         extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult term = parseArgString("HelpOn", command, 1); if(!term.success) return term;
      return new ABCMResult(getCommandsManager().getHelpString(term.text), true);   
    }
  }  
  
  // Script Processing
  class ProcessScript                  extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult fileToProcess = parseArgString("ProcessScript", command, 1); if(!fileToProcess.success) return fileToProcess;
      return new ABCMResult(env.processScript(fileToProcess.text), true);                                 
    }
  }
  class ProcessScriptSilent            extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult fileToProcess = parseArgString("ProcessScript", command, 1); if(!fileToProcess.success) return fileToProcess;
      env.processScript(fileToProcess.text);
      return new ABCMResult("Executing script '" + fileToProcess.text + "' in quiet mode...", true);
    }
  }
  class ProcessNarrative               extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult narr = parseArgString("ProcessNarrative", command, 1); if(!narr.success) return narr;
      return new ABCMResult(env.processScript(narr.text, AbstractABCModelEnvironment.ScriptProcessor.SCRIPT_IS_NARRATIVE), true);
    }
  }
  class Test                           extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      return new ABCMResult(env.processScript("Test.abcm"), true);   
    }
  }

  // Script manipulation
  class ConvertScriptToNarrative       extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult inputFileName  = parseArgString("ConvertScriptToNarrative", command, 1); if(!inputFileName.success) return inputFileName;
      ABCMResult outputFileName = parseArgString("ConvertScriptToNarrative", command, 2); if(!outputFileName.success) return outputFileName;
      ABCMResult baseDate       = parseArgString("ConvertScriptToNarrative", command, 3); if(!baseDate.success) return baseDate;
      ScriptLibrarian sl = new ScriptLibrarian(env);
      if(!sl.convertScriptToNarrative(inputFileName.text, outputFileName.text, baseDate.text)) return new ABCMResult("Could not convert script: " + sl.getErrorReport(), false);
      return new ABCMResult("Converted script " + inputFileName.text + " to narrative " + outputFileName.text, true);
    }
  }
  
  class CompileNarrative               extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult inputFileName  = parseArgString("CompileNarrative", command, 1); if(!inputFileName.success) return inputFileName;
      ABCMResult outputFileName = parseArgString("CompileNarrative", command, 2); if(!outputFileName.success) return outputFileName;
      ScriptLibrarian sl = new ScriptLibrarian(env);
      if(!sl.compileNarrative(inputFileName.text, outputFileName.text)) return new ABCMResult("Could not compile narrative: " + sl.getErrorReport(), false);
      else return new ABCMResult("Compiled Narrative " + inputFileName.text + " to " + outputFileName.text, true);
    }
  }
  
  class UploadNarrative                extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult narrativeName = parseArgString("UploadNarrative", command, 1); if(!narrativeName.success) return narrativeName;
      ABCMResult newName       = parseArgString("UploadNarrative", command, 2); if(!newName.success) return newName;
      ABCMResult desc          = parseArgString("UploadNarrative", command, 3); if(!desc.success) return desc;
      ABCMResult baseDate      = parseArgString("UploadNarrative", command, 4); if(!baseDate.success) return baseDate;
      if(SimDate.parseString(baseDate.text) < 0) return new ABCMResult("Could not upload narrative; base date '" + baseDate.text + " is invalid", false);
      ABCMResult isSeasonal    = parseArgString("UploadNarrative", command, 5); if(!isSeasonal.success) return isSeasonal;
      ScriptLibrarian sl = new ScriptLibrarian(env);
      boolean seasonalFlag = isSeasonal.text.equalsIgnoreCase("True");
      if(sl.uploadCompiledNarrative(narrativeName.text, newName.text, desc.text, baseDate.text, seasonalFlag).success == false)
          return new ABCMResult("Could not upload narrative: " + sl.getErrorReport(), false);
      else
          return new ABCMResult("Narrative " + narrativeName.text + " uploaded.", true);
    }
  }
  
  class StoreHistory                   extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult historyName = parseArgString("StoreHistory", command, 1); if(!historyName.success) return historyName;
      ABCMResult historyDesc = parseArgString("StoreHistory", command, 2); if(!historyDesc.success) return historyDesc;
      ABCMResult truthValue  = parseArgString("StoreHistory", command, 3); if(!truthValue.success) return truthValue;
      return env.storeHistory(historyName.text, historyDesc.text, truthValue.text);
    }
  }
  
  class AddNarrativeToHistory          extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult historyName   = parseArgString("AddNarrativeToHistory", command, 1); if(!historyName.success) return historyName;
      ABCMResult narrativeName = parseArgString("AddNarrativeToHistory", command, 2); if(!narrativeName.success) return narrativeName;
      ABCMResult baseDate      = parseArgString("AddNarrativeToHistory", command, 3); if(!baseDate.success) return baseDate;
      ABCMResult order         = parseArgDouble("AddNarrativeToHistory", command, 4); if(!order.success) return order;
      return env.addNarrativeToHistory(historyName.text, narrativeName.text, baseDate.text, (int)order.value);
    }
  }
  
  // Conditional execution
  class IfUnsuccessful                 extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.getLastResult().success){
        String[] newCommand = new String[command.length - 1];
        for(int i = 0; i < command.length - 1; i++) newCommand[i] = command[i + 1];
        return processCommand(newCommand[0], newCommand);
      }
      else return new ABCMResult("- skipped - ", true);
    }
  }
  class IfSuccessful                   extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(env.getLastResult().success){
        String[] newCommand = new String[command.length - 1];
        for(int i = 0; i < command.length - 1; i++) newCommand[i] = command[i + 1];
        return processCommand(newCommand[0], newCommand);
      }
      else return new ABCMResult("- skipped - ", true);
    }
  }
  
  
  // Model Creation  
  class CreateModel                    extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult simName  = parseArgString("CreateModel", command, 1); if(!simName.success) return simName;
      ABCMResult paramSet = parseArgString("CreateModel", command, 2); if(!paramSet.success) return paramSet;
      ABCMResult probeSet = parseArgString("CreateModel", command, 3); if(!probeSet.success) return probeSet;
      ABCMResult notes    = parseArgString("CreateModel", command, 4); if(!notes.success) notes = new ABCMResult("", true);
      return env.createModel(simName.text, paramSet.text, probeSet.text, notes.text);
    }
  }  
  class CreateTestModel                extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      return env.createTestModel();   
    }
  }

  class SetRandomNumberSeed            extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult rns = parseArgDouble("SetRandomNumberSeed", command, 1); if(!rns.success) return rns;
      if(!env.hasModel()) return noModelMsg("set random number seed");
      return env.getModel().setRandomNumberSeed((long)rns.value);
    }
  }
  
  // Model tear down
  class DisposeAndReset                extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      return env.disposeAndReset();
    }
  }  
  
  // Manage demos
  class StartDemo                      extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult demo = parseArgString("StartDemo", command, 1); if(!demo.success) return demo;
      return env.startDemo(demo.text);
    }
  }
  class ContinueDemo                   extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      return env.continueDemo();   
    }
  }
  class SkipDemoStep                   extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      return env.skipDemoStep();   
    }
  }  
  class SetDemoStep                    extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult step = parseArgDouble("SetDemoStep", command, 1); if(!step.success) return step;
      return env.setDemoStep((int)step.value);
    }
  }
  class StepDemoBackwards              extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      return env.stepDemoBackwards(); 
    }       
  }
  class StopDemo                       extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult demo = parseArgString("StopDemo", command, 1); if(!demo.success) return demo;
      return new ABCMResult(env.stopDemo(demo.text), true);   
    }
  }  

  // Graphs
  class CreateDataGraph                extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult graph = parseArgString("CreateDataGraph", command, 1); if(!graph.success) return graph;
      return env.createDataSeriesGraph(graph.text);
    }
  }
  class ListGraphs                     extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      return new ABCMResult(env.listGraphs(), true);
    }
  }  
  class ShowDataGraph                  extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult graph = parseArgString("ShowDataGraph", command, 1); if(!graph.success) return graph;
      return new ABCMResult(env.showDataGraph(graph.text), true);
    }
  }
  class LoadDataSeries                 extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult num   = parseArgDouble("LoadDataSeries", command, 1); if(!num.success)   return num;
      ABCMResult name  = parseArgString("LoadDataSeries", command, 2); if(!name.success)  return name;
      ABCMResult graph = parseArgString("LoadDataSeries", command, 3); if(!graph.success) return graph;
      return env.loadDataSeriesFromDB((long)num.value, name.text, graph.text); 
    }
  }    
  class DropSeriesFromGraph            extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult series = parseArgString("DropSeriesFromGraph", command, 1); if(!series.success) return series;
      ABCMResult graph  = parseArgString("DropSeriesFromGraph", command, 2); if(!graph.success) return graph;
      return env.dropDataSeriesFromAnalysis(series.text, graph.text);
    }
  }
  class DropGraph                      extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult graph = parseArgString("DropGraph", command, 1); if(!graph.success) return graph;
      return env.dropAnalysis(graph.text);
    }
  }

  // Transcript
  class BookmarkTranscript             extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult name = parseArgString("BookmarkTranscript", command, 1); if(!name.success) return name;      
      return env.bookmarkTranscript(name.text);
    }
  }
    

  // Branching
  class MarkAndClone                   extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult name = parseArgString("MarkAndClone", command, 1); if(!name.success) return name;
      return env.markAndCloneWithNewRNG(name.text);
    }
  }
  class RestoreToMark                  extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult name = parseArgString("RestoreToMark", command, 1); if(!name.success) return name;
      return env.restoreToMark(name.text);
    }
  }
  class AdoptCurrentBranch             extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      return env.clearAllMarksAndAcceptCurrentBranch();
    }
  }
  
  
  /* Model Commands */
  
  
  /**
   * Simple method to return a standard message when no model is detected
   * @param action name of the action attempted
   * @return message indicating that the action could not be performed
   */
  protected final ABCMResult noModelMsg(String action){  return new ABCMResult("Could not " + action + "; no model exists.", false);    }
  
  // Model Status
  class ModelStatus                    extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("retrieve model status");
      return new ABCMResult(env.getModel().modelStatus(), true);  
    }
  }  
  class ViewConfigurationDetails       extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("view configuration details");  
      return env.getModel().getConfigurationReport();   
    }
  }  
  class SetCodeModelOption             extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set code model option");
      ABCMResult option = parseArgString("SetCodeModelOption", command, 1); if(!option.success) return option;
      ABCMResult val    = parseArgDouble("SetCodeModelOption", command, 2); 
      return env.getModel().setCodeModelOption(option.text, (int)val.value);
    }
  }
  class SetParameter                   extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set parameter");
      ABCMResult param = parseArgString("SetParameter", command, 1); if(!param.success) return param;
      ABCMResult val   = parseArgDouble("SetParameter", command, 2); if(!val.success) return val;
      return env.getModel().setParameter(param.text, val.value);
    }
  }
  class ShowParameters                 extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("show parameters");
      return env.getModel().getParameterList();
    }
  }

  // Model time management
  class SetPeriod                      extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set period");
      ABCMResult newPeriod = parseArgString("SetPeriod", command, 1); if(!newPeriod.success) return newPeriod;
      return E.getModel().setCurrentPeriod(newPeriod.text);
    }
  }
  class SetDate                        extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set simulation date");
      ABCMResult newDate = parseArgString("SetDate", command, 1); if(!newDate.success) return newDate;
      return E.getModel().setModelTime(newDate.text); 
    }
  }
  class ShowDate                       extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("show model date");
      return E.getModel().showDate();   
    }
  }

  // Step
  class Step                           extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("step model");  
      return E.getModel().step();   
    }
  }
  class StepDays                       extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("step model");
      ABCMResult days = parseArgDouble("StepDays", command, 1); if(!days.success) return days;  
      return env.getModel().StepForwardNDays((long)days.value);
    }
  }
  class Finish                         extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("finish");  
      return env.getModel().stepToEnd();   
    }
  }
  class StepToDate                     extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("step model");
      ABCMResult newDate = parseArgString("StepToDate", command, 1); if(!newDate.success) return newDate;
      return env.getModel().stepToDate(newDate.text);
    }
  }

  // Stop
  class Stop                           extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("mark end of narrative");
      return new ABCMResult("Model Stop at " + env.getModel().getModelTime().getString(), true);
    }
  }
  
  // Load model elements
  class LoadSimulation                 extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("load simulation");
      ABCMResult sim = parseArgString("LoadSimulation", command, 1); if(!sim.success) return sim;
      return env.getModel().loadSimulation(sim.text);
    }
  }  
  class LoadCodeModel                  extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("load code model");
      ABCMResult cm = parseArgString("LoadCodeModel", command, 1); if(!cm.success) return cm;
      return env.getModel().loadCodeModel(cm.text);
    }
  }
  class LoadDataModel                  extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("load data model");
      ABCMResult dm = parseArgString("LoadDataModel", command, 1); if(!dm.success) return dm;
      return env.getModel().loadDataModel(dm.text);
    }
  }
  class LoadHistory                    extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("load history");
      ABCMResult hist = parseArgString("LoadHistory", command, 1); if(!hist.success) return hist;
      return env.getModel().loadHistory(hist.text);
    }
  }
  class LoadConfiguration              extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("load configuration");
      ABCMResult config = parseArgString("LoadConfiguration", command, 1); if(!config.success) return config;
      return env.getModel().loadConfiguration(config.text);
    }
  }
  class LoadParameterSet               extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("load parameter set");
      ABCMResult paramset = parseArgString("LoadParameterSet", command, 1); if(!paramset.success) return paramset;
      return env.getModel().loadParameterSet(paramset.text);
    }
  }
  
  // Space
  class AddPoint                       extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("add point");
        ABCMResult name = parseArgString("AddPoint", command, 1); if(!name.success) return name;
        ABCMResult X = parseArgDouble("AddPoint", command, 2); if(!X.success) return X;
        ABCMResult Y = parseArgDouble("AddPoint", command, 3); if(!Y.success) return Y;
        ABCMResult Z = parseArgDouble("AddPoint", command, 4); if(!Z.success) return Z;
      return env.getModel().addPoint(name.text, X.value, Y.value, Z.value);
    }
  }  
  class AddSurfacePoint                extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("add surface point");
        ABCMResult name = parseArgString("AddPoint", command, 1); if(!name.success) return name;
        ABCMResult X = parseArgDouble("AddPoint", command, 2); if(!X.success) return X;
        ABCMResult Y = parseArgDouble("AddPoint", command, 3); if(!Y.success) return Y;
      return env.getModel().addSurfacePoint(name.text, X.value, Y.value);
    }
  }
  class ViewListOfPoints               extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get list of points");  
      return env.getModel().getListOfPoints();   
    }
  }
  class SetNAndEOrientation            extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set N and E orientation");
      ABCMResult no = parseArgString("SetNAndEOrientation", command, 1); if(!no.success) return no;
      ABCMResult eo = parseArgString("SetNAndEOrientation", command, 2); if(!eo.success) return eo;
      boolean northIsPositive = env.getModel().getSpace().northPositive();
            if((no.text.compareToIgnoreCase("POSITIVE") == 0) ||
               (no.text.compareToIgnoreCase("+")        == 0)) 
                  { northIsPositive = true; }
            if((no.text.compareToIgnoreCase("NEGATIVE") == 0) ||
               (no.text.compareToIgnoreCase("-")        == 0)) 
                  { northIsPositive = false; }
      
      boolean eastIsPositive = env.getModel().getSpace().eastPositive();
            if(( eo.text.compareToIgnoreCase("POSITIVE") == 0) ||
               ( eo.text.compareToIgnoreCase("+")        == 0)) 
                  {  eastIsPositive = true; }
            if(( eo.text.compareToIgnoreCase("NEGATIVE") == 0) ||
               ( eo.text.compareToIgnoreCase("-")        == 0)) 
                  {  eastIsPositive = false; }
      return env.getModel().setNorthAndEastOrientation(northIsPositive, eastIsPositive);   
    }
  } 
  
  // Populations and agents
  class CreatePopulation               extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("create population");
      return env.getModel().createPopulation();
    }
  }
  
  class PopulationDo extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("send do message to population");
      ABCMResult instruction = parseArgString("PopulationDo", command, 1); if(!instruction.success) return instruction;
      String newCommand = "";
      for(int i = 1; i < command.length; i++) newCommand += command[i] + (i < (command.length - 1) ? " " : "");
      return env.getModel().populationDo(newCommand);
    }
  }
  
  class AgentDo extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("send do message to agent");
      ABCMResult agentID     = parseArgString("AgentDo", command, 1); if(!agentID.success)     return agentID;
      ABCMResult instruction = parseArgString("AgentDo", command, 2); if(!instruction.success) return instruction;
      String newCommand = "";
      for(int i = 2; i < command.length; i++) newCommand += command[i] + (i < (command.length - 1) ? " " : "");
      return env.getModel().agentDo(agentID.text, newCommand);
    }
  }
  // Data Series
  class CreateNumericDataSeries        extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("create data series");
      ABCMResult seriesName = parseArgString("CreateNumericDataSeries", command, 1); if(!seriesName.success) return seriesName;
      ABCMResult valueName  = parseArgString("CreateNumericDataSeries", command, 2); if(!valueName.success) return valueName;
      return env.getModel().createNumericDataSeries(seriesName.text, valueName.text);
    }
  }
  class ViewDataSeriesPrintout         extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){  
      if(!env.hasModel()) return noModelMsg("view data series printout");
      ABCMResult series = parseArgString("ViewDataSeriesPrintout", command, 1); if(!series.success) return series;
      return env.getModel().getDataSeriesPrintout(series.text);
    }
  }
  class ListDataSeries                 extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("list data series");
      return env.getModel().getListOfDataSeries();
    }
  }
  class SaveDataSeries                 extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
        if(!env.hasModel()) return noModelMsg("save data series");
        ABCMResult name = parseArgString("SaveDataSeries", command, 1); if(!name.success) return name;
      return env.getModel().saveDataSeries(name.text);
    }
  }
  class SaveAllDataSeries              extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("save the data series");
      return env.getModel().saveAllDataSeries();
    }
  }
  class AddSeriesToGraph               extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("add the series");
      ABCMResult series = parseArgString("AddSeriesToGraph", command, 1); if(!series.success) return series;
      ABCMResult graph = parseArgString("AddSeriesToGraph", command, 2); if(!graph.success) return graph;
      return env.getModel().addDataSeriesToGraph(series.text, graph.text);
    }
  }
    
  // Probes
  class CreateProbe                    extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("create probe");

      ABCMResult name     = parseArgString("CreateProbe", command, 1); if(!name.success) return name;
      ABCMResult baseD    = parseArgString("CreateProbe", command, 2); if(!baseD.success) return baseD;
      if(SimDate.parseString(baseD.text) < 0) return new ABCMResult("Specified base date ('" + baseD.text + ") is invalid.", false);
      ABCMResult freq     = parseArgDouble("CreateProbe", command, 3); if(!freq.success) return freq;
      ABCMResult series   = parseArgString("CreateProbe", command, 4); if(!series.success) return series;
      ABCMResult retrieve = parseArgString("CreateProbe", command, 5); if(!retrieve.success) return retrieve;
      return env.getModel().createProbe(name.text, new SimDate(baseD.text), (long)freq.value, 
          series.text, series.text, true, retrieve.text);
    }
  }
  class ListProbes                     extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("list probes");  
      return env.getModel().getListOfProbes();
    }
  }  
  class DropProbe                      extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("drop probe");
      ABCMResult probe = parseArgString("DropProbe", command, 1); if(!probe.success) return probe;
      return env.getModel().dropProbe(probe.text);
    }
  }
  class TurnProbeOn                    extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("turn probe on");
      ABCMResult probe = parseArgString("TurnProbeOn", command, 1); if(!probe.success) return probe;
      return env.getModel().turnProbeOn(probe.text);
    }
  }  
  class TurnProbeOff                   extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("turn probe off");
      ABCMResult probe = parseArgString("TurnProbeOff", command, 1); if(!probe.success) return probe;
      return env.getModel().turnProbeOff(probe.text);   
    }
  }
  class CreateMonthlyProbe             extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("create monthly probe");
      ABCMResult name     = parseArgString("CreateMonthlyProbe", command, 1); if(!name.success) return name;
      ABCMResult series   = parseArgString("CreateMonthlyProbe", command, 2); if(!series.success) return series;
      ABCMResult retrieve = parseArgString("CreateMonthlyProbe", command, 3); if(!retrieve.success) return retrieve;
      return env.getModel().createMonthlyProbe(name.text, series.text, series.text, 
          true, retrieve.text);
    }
  }
  class SetProbeOffset                 extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set probe offset");
      ABCMResult probe  = parseArgString("SetProbeOffset", command, 1); if(!probe.success) return probe;
      ABCMResult offset = parseArgDouble("SetProbeOffset", command, 2); if(!offset.success) return offset;
      return env.getModel().setProbeOffset(probe.text, offset.value);
    }
  }
  class LoadProbeSet                   extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("load probe set");
      ABCMResult ps = parseArgString("LoadProbeSet", command, 1); if(!ps.success) return ps;
      return env.getModel().loadProbeSet(ps.text);
    }
  }

  
  class StoreProbe                     extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult probeName = parseArgString("StoreProbe", command, 1); if(!probeName.success) return probeName;
      ABCMResult probeDesc = parseArgString("StoreProbe", command, 2); if(!probeDesc.success) return probeDesc;
      ABCMResult series    = parseArgString("StoreProbe", command, 3); if(!series.success)    return series;
      ABCMResult domain    = parseArgString("StoreProbe", command, 4); if(!domain.success)    return domain;
      ABCMResult numeric   = parseArgString("StoreProbe", command, 5); if(!numeric.success)   return numeric;
      ABCMResult baseDate  = parseArgString("StoreProbe", command, 6); if(!baseDate.success)  return baseDate;
      if(SimDate.parseString(baseDate.text) < 0) return new ABCMResult("Could not store probe; base date '" + baseDate.text + "' is invalid", false);
      ABCMResult frequency = parseArgDouble("StoreProbe", command, 7); if(!frequency.success) return frequency;
      ABCMResult retrieve  = parseArgString("StoreProbe", command, 8); if(!retrieve.success)  return retrieve;
      ABCMResult offset    = parseArgDouble("StoreProbe", command, 9); if(!offset.success)    return offset;
      return (env.storeProbe(probeName.text, probeDesc.text, series.text, domain.text, (numeric.text.equalsIgnoreCase("TRUE")), 
          new SimDate(baseDate.text), (int)frequency.value, retrieve.text, offset.value));
    }
  }
  
  class StoreMonthlyProbe              extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult probeName = parseArgString("StoreProbe", command, 1); if(!probeName.success) return probeName;
      ABCMResult probeDesc = parseArgString("StoreProbe", command, 2); if(!probeDesc.success) return probeDesc;
      ABCMResult series    = parseArgString("StoreProbe", command, 3); if(!series.success)    return series;
      ABCMResult domain    = parseArgString("StoreProbe", command, 4); if(!domain.success)    return domain;
      ABCMResult numeric   = parseArgString("StoreProbe", command, 5); if(!numeric.success)   return numeric;
      ABCMResult retrieve  = parseArgString("StoreProbe", command, 6); if(!retrieve.success)  return retrieve;
      ABCMResult offset    = parseArgString("StoreProbe", command, 7); if(!offset.success)    return offset;
      return (env.storeMonthlyProbe(probeName.text, probeDesc.text, series.text, domain.text, (numeric.text.equalsIgnoreCase("TRUE")), 
          retrieve.text, offset.value));
    }
  }
  
  class StoreProbeSet                  extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult probeSetName = parseArgString("StoreProbe", command, 1); if(!probeSetName.success) return probeSetName;
      ABCMResult probeSetDesc = parseArgString("StoreProbe", command, 2); if(!probeSetDesc.success) return probeSetDesc;
      return(env.storeProbeSet(probeSetName.text, probeSetDesc.text)); 
    }
  }
  
  class AddProbeToProbeSet             extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult probeName    = parseArgString("StoreProbe", command, 1); if(!probeName.success) return probeName;
      ABCMResult probeSetName = parseArgString("StoreProbe", command, 2); if(!probeSetName.success) return probeSetName;
      return(env.addProbeToProbeSet(probeName.text, probeSetName.text));
    }
  }
  
  
  // Seasonal Narratives
  class TurnSeasonalNarrativeOff       extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("turn seasonal narrative on");
      ABCMResult narr = parseArgString("TurnSeasonalNarrativeOff", command, 1); if(!narr.success) return narr;
      return env.getModel().turnSeasonalNarrativeOff(narr.text);   
    }
  }  
  class TurnSeasonalNarrativeOn        extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("turn seasonal narrative off");
      ABCMResult narr = parseArgString("TurnSeasonalNarrativeOn", command, 1); if(!narr.success) return narr;
      return env.getModel().turnSeasonalNarrativeOn(narr.text);
    }
  }
  class SwitchSeasonalNarrative        extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("switch seasonal narrative");
      ABCMResult narr1 = parseArgString("SwitchSeasonalNarrative", command, 1); if(!narr1.success) return narr1;
      String narrList = narr1.text + ":";
      for(int i = 2; i < command.length; i++){
        ABCMResult narr = parseArgString("SwitchSeasonalNarrative", command, i);
        if(narr.success) narrList += narr.text + ":";
      }
      String[] narratives = narrList.split(":");
      return env.getModel().chooseSeasonalNarrative(narratives);
    }
  }
  
//  class SaveInstructions extends ParseAction{
//    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
//      return new ABCMResult(   
//    }
//  }
  
  
  
  
  
  /* Parsing Engine */
  
  /**
   * Constructor. When invoked adds the environment and abstract model-specific
   * actions to the array of actions to be executed.
   */
  public AbstractABCMParser (AbstractABCModelEnvironment env){

    E = env;
    cmovm = new CodeModelOptionValuesManager(E.getDBGateway());
    
    /* Add commands */

      /* Environment Commands*/  
    
    // Logging In
    addCommand("Login",                    new Login());
    
    // Help
    addCommand("Define",                   new Define());
    addCommand("Help",                     new Help()); 
    addCommand("HelpOn",                   new HelpOn());
    
    // Script Processing
    addCommand("ProcessScript",            new ProcessScript());
    addCommand("ProcessScriptSilent",      new ProcessScriptSilent());
    addCommand("ProcessNarrative",         new ProcessNarrative());
    addCommand("Test",                     new Test());
    addCommand("ConvertScriptToNarrative", new ConvertScriptToNarrative());
    addCommand("CompileNarrative",         new CompileNarrative());
    addCommand("UploadNarrative",          new UploadNarrative());
    addCommand("StoreHistory",             new StoreHistory());
    addCommand("AddNarrativeToHistory",    new AddNarrativeToHistory());
   
    // Conditional Execution
    addCommand("IfUnsuccessful",           new IfUnsuccessful());
    addCommand("IfSuccessful",             new IfSuccessful());
      
    // Model Creation
    addCommand("CreateModel",              new CreateModel());
    addCommand("CreateTestModel",          new CreateTestModel());
    addCommand("SetRandomNumberSeed",      new SetRandomNumberSeed());
    
    // Model tear down
    addCommand("DisposeAndReset",          new DisposeAndReset());
    
    // Model Status
    addCommand("ModelStatus",              new ModelStatus());
    addCommand("ViewConfigurationDetails", new ViewConfigurationDetails());
    addCommand("SetCodeModelOption",       new SetCodeModelOption());
    addCommand("SetParameter",             new SetParameter());
    addCommand("ShowParameters",           new ShowParameters());
    
    // Manage demos
    addCommand("StartDemo",                new StartDemo());
    addCommand("ContinueDemo",             new ContinueDemo());
    addCommand("SkipDemoStep",             new SkipDemoStep());
    addCommand("SetDemoStep",              new SetDemoStep());
    addCommand("StepDemoBackwards",        new StepDemoBackwards());
    addCommand("StopDemo",                 new StopDemo());

    // Graphs
    addCommand("CreateDataGraph",          new CreateDataGraph());
    addCommand("ListGraphs",               new ListGraphs());
    addCommand("ShowDataGraph",            new ShowDataGraph());
    addCommand("LoadDataSeries",           new LoadDataSeries());
    addCommand("DropSeriesFromGraph",      new DropSeriesFromGraph());
    addCommand("DropGraph",                new DropGraph());
    
    
    /* Model Commands*/
    
    // Model time management
    addCommand("SetPeriod",                new SetPeriod());
    addCommand("SetDate",                  new SetDate());
    addCommand("ShowDate",                 new ShowDate());
    
    // Step
    addCommand("Step",                     new Step());
    addCommand("StepDays",                 new StepDays());
    addCommand("Finish",                   new Finish());
    addCommand("StepToDate",               new StepToDate());
    
    // Stop
    addCommand("Stop",                     new Stop());
    // Load model elements
    addCommand("LoadSimulation",           new LoadSimulation());
    addCommand("LoadCodeModel",            new LoadCodeModel());
    addCommand("LoadDataModel",            new LoadDataModel());
    addCommand("LoadHistory",              new LoadHistory());
    addCommand("LoadConfiguration",        new LoadConfiguration());
    addCommand("LoadParameterSet",         new LoadParameterSet());
     
    // Space
    addCommand("AddPoint",                 new AddPoint());
    addCommand("AddSurfacePoint",          new AddSurfacePoint());
    addCommand("ViewListOfPoints",         new ViewListOfPoints());
    addCommand("SetNAndEOrientation",      new SetNAndEOrientation());
    
    // Agents
    addCommand("CreatePopulation",         new CreatePopulation());
    addCommand("PopulationDo",             new PopulationDo());
    addCommand("AgentDo",                  new AgentDo());
    
    // Data Series
    addCommand("CreateNumericDataSeries",  new CreateNumericDataSeries());
    addCommand("ViewDataSeriesPrintout",   new ViewDataSeriesPrintout());
    addCommand("ListDataSeries",           new ListDataSeries());
    addCommand("SaveDataSeries",           new SaveDataSeries());
    addCommand("SaveAllDataSeries",        new SaveAllDataSeries());
    addCommand("AddSeriesToGraph",         new AddSeriesToGraph());
    
    // Probes
    addCommand("CreateProbe",              new CreateProbe());
    addCommand("ListProbes",               new ListProbes());
    addCommand("DropProbe",                new DropProbe());
    addCommand("TurnProbeOn",              new TurnProbeOn());
    addCommand("TurnProbeOff",             new TurnProbeOff());
    addCommand("CreateMonthlyProbe",       new CreateMonthlyProbe());
    addCommand("SetProbeOffset",           new SetProbeOffset());
    addCommand("LoadProbeSet",             new LoadProbeSet());
    addCommand("StoreProbe",               new StoreProbe());
    addCommand("StoreMonthlyProbe",        new StoreMonthlyProbe());
    addCommand("StoreProbeSet",            new StoreProbeSet());
    addCommand("AddProbeToProbeSet",       new AddProbeToProbeSet());    
    
    
    // Seasonal Narratives
    addCommand("TurnSeasonalNarrativeOff", new TurnSeasonalNarrativeOff());
    addCommand("TurnSeasonalNarrativeOn",  new TurnSeasonalNarrativeOn());
    addCommand("SwitchSeasonalNarrative",  new SwitchSeasonalNarrative());
 
    // Transcript management
    addCommand("BookmarkTranscript",       new BookmarkTranscript());
    
    // Branching
    addCommand("MarkAndClone",             new MarkAndClone());
    addCommand("RestoreToMark",            new RestoreToMark());
    addCommand("AdoptCurrentBranch",       new AdoptCurrentBranch());
    
//    addCommand(SaveInstructions, new SaveInstructions());
//    addCommand(GetSVG, new GetSVG());
//    addCommand(GetX3D, new GetX3D());
//    addCommand(View3D, new View3D());
    
  }
  
  
  
  
  
  /**
   * Retrieves a result by executing the command.
   * @param commandPK Primary key of the requested command
   * @param command array with element[0] equal to the name of the command and the remaining elements equal to its arguments.
   * @return the result of executing the command
   */
  private final ABCMResult pc(int commandPK, String[] command){
    if(command != null && command[0] != null && command[0].startsWith("**")) return new ABCMResult("", true);
    if(commandPK < 0) {
      String c = "";
      for(int i = 0; i < command.length; i++) c += (command[i] != null ? command[i] : "");
      return new ABCMResult("Unknown command: " + c, false);
    }
    if(commandPK >= MAX_COMMANDS) return new ABCMResult("Unknown command with ID number greater than maximum", false);
    if(commands[commandPK] == null) return new ABCMResult("Unknown command with null value " + commandPK + " (" + command[0] + ") in commands array", false);
    return commands[commandPK].executeCommand(command, E);    
  }
    
  /**
   * Processes command sent as a primary key of a command
   * and an array of arguments. 
   * @param commandPK
   * @param command
   * @return a response that can be written to an output window
   */
  public final ABCMResult processCommand(int commandPK, String[] command){
    return pc(commandPK, command); 
  }
  
  /**
   * Processes command sent as a single String, space-delimited,
   * the first element being the command and the subsequent ones
   * being arguments.
   * @param cText
   * @return a response that can be written to an output window
   */
  public final ABCMResult processCommand(String cText){
    String[] command = new String[MAX_ARGS];
    command = parseQuoted(cText);
    // Get the primary key of the initial command
    int commandPK = (int)getCommandsManager().getCommandPK(command[0]); 
    return processCommand(commandPK, command);
  }
  
  /**
   * Process a command sent as a String and an array of
   * Strings as arguments.
   * @param command
   * @param argList
   * @return a response that can be written to an output window
   */
  public final ABCMResult processCommand(String command, String[] argList){
    int commandPK = (int)getCommandsManager().getPK(command);
    return processCommand(commandPK, argList);
  }
  
  /**
   * Processes a command sent as a History Instruction
   * If this is passed a null it returns a String indicating
   * that no operation took place. This is because some
   * methods of parsing a History Instruction List return
   * null when a Seasonal Narratives is turned 'off.'
   * @param HI The History Instruction to be executed
   * @return a response that can be written to an output window
   */
  public final ABCMResult processCommand(HistoryInstruction HI){
    if(HI == null) return new ABCMResult ("No operation/Skipped Command (Seasonal Narrative may be turned off...)", false);
    return processCommand(HI.CommandID, HI.argList);     
  }

  
  
  
  /**
   * Confirms that the string being processed has enough arguments
   * to process the argument in question
   * @param actionName name of the action being performed
   * @param commands list of commands to be processed
   * @param argNum number of arguments expected
   */
  protected final ABCMResult parseArgString(String actionName, String[] commands, int argNum){
    try{
      if(commands[argNum] == "") return new ABCMResult("Could not perform " + actionName + "; Not enough arguments.", false);
      return new ABCMResult(commands[argNum], true);
    }
    catch(ArrayIndexOutOfBoundsException A){
      return new ABCMResult("Could not perform " + actionName + "; Not enough arguments (required: " + argNum + ").", false);
    }
    catch(Exception E){
      return new ABCMResult("Could not perform " + actionName + "; Argument " + argNum + " is invalid: " + E.getMessage(), false);
    }
  }
  
  /**
   * Confirms that the string being processed has enough arguments and returns a
   * double if the argument should return a numeric argument
   * If the argument is in the form RAND:a,b
   * where a and b are correctly formatted double values,
   * the value returned will be a random number selected from
   * a uniform distribution between a (inclusive) and b (not inclusive).
   * @param actionName name of the action being performed
   * @param commands list of commands to be processed
   * @param argNum number of arguments expected
   */ 
  protected final ABCMResult parseArgDouble(String actionName, String[] commands, int argNum ){
    try{
      return new ABCMResult(Double.parseDouble(commands[argNum]), true);
    }
    catch(NumberFormatException NFE){
      if(commands[argNum].startsWith("RAND:")){
        try{
          String vals[] = commands[argNum].substring(5).split(",");
          double val0 = Double.parseDouble(vals[0]);
          double val1 = Double.parseDouble(vals[1]);
          double newResult = val0 + (val1 - val0) * this.random();
          return new ABCMResult("New random value from '" + commands[argNum] + "' is " + newResult, newResult, true);
        }
        catch(Exception E){
          return new ABCMResult("Could not create random double from '" + commands[argNum] + "'", false);
        }        
      }
      double ret = cmovm.getOptionValue(commands[argNum]);
      if(ret >=0) return new ABCMResult((double)ret, true);
      
      if(this.getParameterSet().hasParameter(commands[argNum])){
        ret = paramVal(commands[argNum]);
        return new ABCMResult("Parameter value '" + commands[argNum] + "' is " + ret, ret, true);
      }
      
      return new ABCMResult("Could not parse '" + commands[argNum] + "' into double.", false);
    }
    catch(ArrayIndexOutOfBoundsException A){
      return new ABCMResult("Could not perform " + actionName + "; Not enough arguments.", false);
    }
    catch(Exception E){
      return new ABCMResult("Could not perform " + actionName + "; Argument " + argNum + " is invalid: " + E.getMessage(), false);
    }
  }
  
  
  /**
   * Adds a command to this parser. 
   * @param commandIndex the index to which this command is to be assigned
   * @param command the command being added
   * @return a message indicating the success or failure of this operation
   */
  protected final ABCMResult addCommand(int commandIndex, ParseAction command){
    if(command==null)                  return new ABCMResult("You cannot set a null command", false);
    if(commandIndex < 0)               return new ABCMResult("You cannot set a command index less than zero", false);
    if(commandIndex >= MAX_COMMANDS)   return new ABCMResult("Command index requested exceeds maximum of " + (MAX_COMMANDS -1), false);
    if(commands[commandIndex] != null) return new ABCMResult("A command with the specified index (" + commandIndex + ") already exists", false);
    commands[commandIndex] = command;
    return new ABCMResult("Successfully added command", true);
  }
  
  /**
   * Adds a command to this parser. Uses the name for the command; the index is derived by
   * looking up the value in the database.
   * @param commandName name of the command to be added
   * @param command command being added
   * @return a message indicating the success or failure of this operation
   */
  protected final ABCMResult addCommand(String commandName, ParseAction command){
    int commandIndex = (int)getCommandsManager().getCommandPK(commandName);
    if(commandIndex <= 0) return new ABCMResult("Could not add command '" + commandName + "'; the command is not present in the specified database.", false);
    return addCommand(commandIndex, command);
  }  
  
  /**
   * Gets the environment's current commands manager
   * @return the current commands manager
   */
  protected CommandsManager getCommandsManager(){
    return E.getCommandsManager();
  }
  
}
