package abcm.model;

import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.db.managers.CodeModelManager;
import abcm.db.managers.DataModelManager;
import abcm.db.managers.HistoryManager;
import abcm.db.managers.ParameterSetManager;
import abcm.db.managers.ProbeManager;
import abcm.db.managers.ProbeSetManager;
import abcm.db.managers.SimulationDataSeriesManager;
import abcm.db.managers.SimulationManager;
import abcm.db.managers.SimulationRunManager;
import abcm.db.managers.stubs.AbstractConfigurationManager;
import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.graphics.X3DAssist;
import abcm.graphics.X3DRepresentable;
import abcm.math.BasicRNG;
import abcm.math.SimDate;
import abcm.model.agents.AbstractABCModelPopulation;
import abcm.sim.AbstractModelComponent;
import abcm.sim.Configuration;
import abcm.sim.DataModel;
import abcm.sim.History;
import abcm.sim.HistoryInstruction;
import abcm.sim.ParameterSet;
import abcm.sim.Probe;
import abcm.sim.Simulation;
import abcm.sim.SimulationDataSeries;
import abcm.sim.SimulationRun;
import abcm.space.DatumPoint;
import abcm.space.Space;
import abcm.utilities.XMLAssist;

/**
 * Fundamental class from which all ABCM Models should be created.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractABCModel extends AbstractModelComponent implements SVGRepresentable, X3DRepresentable {

  /** Minimum width of the legend in SVG output  */
  private static final double MIN_LEGEND_WIDTH                    =   3.0;
  /** Maximum width of the legend  in SVG output  */
  private static final double MAX_LEGEND_WIDTH                    =   6.0;
  /** Fraction of the SVG map output that should be accorded to the legend  */
  private static final double LEGEND_FRACTION                     =   6d / 21d;  // Six of 21 inches is ideal
  /** Margin to be used on SVG representation  */
  private static final double MARGIN                              =    .5;
  
  /** Current simulation date  */
  protected SimDate      modelTime      = new SimDate("01-JAN-9999 BC");
  /** Nominal category of time in which the model is currently positioned  */
  protected String       currentPeriod  = "- no period set -";
  
  /** The simulation in use in this model  */
  protected Simulation    simulation;
  /** The data model in use in this model  */
  protected DataModel     dataModel;
  /** The history in use in this model  */
  protected History       history;
  /** The model's current configuration  */
  protected Configuration configuration;

  // Data Collection
  /** The collection of data series in this model  */
  protected ABCMDeepCloneableHashtable dataSeries = new ABCMDeepCloneableHashtable();
  /** The collection of probes in this model  */
  protected ABCMDeepCloneableHashtable probes     = new ABCMDeepCloneableHashtable();
  
  // Environment
  /** The environment in which this model exists  */
  protected AbstractABCModelEnvironment env;

  // Agents
  /** The population of agents in this model **/
  protected AbstractABCModelPopulation population;
  
  // Status
  /** True if the model is currently executing a step, false otherwise  */
  protected boolean inStep          = false;
  /** True if no change has been made to the model that will modify its input so that the result is not securely auditable, false otherwise  */
  protected boolean modelIntegrity  = true;
  /** ID Number assigned to the current run. Will be -1 if run is test run  */
  protected long    runID           = -1;
  
  /* Constructors */
  
  /**
   * Constructor
   * @param modelEnvironment The programmatic environment in which this model will exist
   */
  public AbstractABCModel(AbstractABCModelEnvironment modelEnvironment){
    env       = modelEnvironment;
    inStep    = false;
  }
  
  /**
   * Constructor
   * @param modelEnvironment The programmatic environment in which this model will exist
   * @param sr The Simulation Run associated with this model
   */
  public AbstractABCModel(AbstractABCModelEnvironment modelEnvironment, SimulationRun sr){
    env       = modelEnvironment;
    inStep    = true;
      loadSimulation(sr.getSimulationPK());
      loadParameterSet(sr.getParameterSetPK(), sr);
      loadProbeSet(sr.getProbeSetPK());
    inStep = false;
    runID = sr.getSimulationRunID();    
  }

  
  
  /* Data integrity */
  // A model is said to have complete integrity only if the information
  // that was used to create the model is employed, unchanged, until the
  // end of the simulation run and the moment when all collected
  // data are stored. In an interactive session this is not guaranteed.
  // The method of determining if integrity has been compromised
  // is to trap all changes to the model. If they are performed during
  // a simple 'step' procedure, they will not affect model integrity,
  // since the step procedure invokes the elements prescribed by the
  // loaded History. However, invoking any non-interrogative command
  // outside a step procedure will revoke data integrity (even if it 
  // is to set a new value that is equal to the old one).
  //
  // Model Integrity implies that the run would be completely reproducible
  // provided the same input information and RNS were used.
  
  
  /**
   * Checks whether this model has 'integrity'- that is, whether the results
   * can be securely linked to the auditing information about the model input 
   */
  protected void checkIntegrity(){
    if(!inStep) modelIntegrity = false;
  }
  
  /**
   * Returns true if the model can be securely linked to the auditing information
   * about the model input, false if a change has been made that violates integrity
   * @return true if model integrity is intact, false otherwise
   */
  public boolean modelIntegrityIsIntact(){
    return modelIntegrity;
  }
  

  /* Load tree */
  
  /*
   * A Model runs by creating a new 'Simulation Run', which is defined
   * by three elements:
   * 
   * A 'Simulation'
   * A 'Parameter Set'
   * A 'Probe Set'
   * 
   * A 'simulation' includes two elements: a Code Model and a Data Model.
   * 
   * A 'Code Model' is a list of code options that dictate which versions
   * or options of code are run.
   * 
   * A 'Data Model' includes two elements: a 'History' and a 'Configuration'
   * 
   * All of these are loaded by a load tree.
   */
  
  /**
   * Loads a simulation
   * @param simPK The primary key of the simulation as stored in the HWM database.
   * @return A String indicating the success of the operation.
   */
  public ABCMResult loadSimulation(long simPK){
    checkIntegrity();
    SimulationManager sm = new SimulationManager(env.getDBGateway());
    simulation = sm.getSimulation(simPK);
    if(simulation == null) {
      // This return value is sometimes unprinted; use err.println to report
      System.err.println(sm.getErrorReport());
      return new ABCMResult("Could not load simulation. " + sm.getErrorReport(), false);
    }
    loadCodeModel(simulation.getCodeModelPK());
    loadDataModel(simulation.getDataModelPK());
    return new ABCMResult("Loaded simulation " + simulation.getSimName(), true);
  } 
    
  /**
   * Loads a simulation
   * @param simulationName the name of the simulation as stored in the HWM database
   * @return A String indicating the success of the operation.
   */
  public ABCMResult loadSimulation(String simulationName){    
    SimulationManager sm = new SimulationManager(env.getDBGateway());
    long simPK = sm.getPK(simulationName);
    return loadSimulation(simPK);
  }
  
    /**
     * Loads a code model
     * @param CodeModelPK The primary key of the code model as stored in the HWM database
     * @return A String indicating the success of the operation
     */
    public ABCMResult loadCodeModel(long CodeModelPK){
      checkIntegrity();
      CodeModelManager cmm = new CodeModelManager(env.getDBGateway());
      AbstractModelComponent.setModelComponent(cmm.getCodeModelFromDB(CodeModelPK), null, null);
      return new ABCMResult("Loaded code model.", true);
    }
    
    /**
     * Loads a code model with the code model name passed as a string.
     * @param codeModelName
     * @return A message indicating the success of the operation.
     */
    public ABCMResult loadCodeModel(String codeModelName){
      CodeModelManager cmm = new CodeModelManager(env.getDBGateway());
      long PK = cmm.getPK(codeModelName);
      return loadCodeModel(PK);
    }
    
    /**
     * Loads a data model
     * @param DataModelPK The primary key of the Data Model as stored in the HWM database
     * @return A String indicating the success of the operation
     */
    public ABCMResult loadDataModel(long DataModelPK){
      checkIntegrity();
      DataModelManager dmm = new DataModelManager(env.getDBGateway());
      dataModel = dmm.getDataModel(DataModelPK);
      if(dataModel == null) {
        // This return value is sometimes unprinted; use err.println to report
        System.err.println("Could not load data model: " + dmm.getErrorReport());
        return new ABCMResult("Could not load data model", false);
      }
      loadHistory(dataModel.getHistoryPK());
      loadConfiguration(dataModel.getConfigurationPK());
      return new ABCMResult("Successfully loaded data model", true);
    }
    
    /**
     * Loads a data model by name
     * @param dataModelName Name of the data model in the HWM Database to be loaded
     * @return an ABCMResult indicating the success or failure of this operation
     */
    public ABCMResult loadDataModel(String dataModelName){
      DataModelManager dm = new DataModelManager(env.getDBGateway());
      long PK = dm.getPK(dataModelName);
      ABCMResult result = loadDataModel(PK);
      if(!result.success) return result;
      return new ABCMResult(result.text + " '" + dataModelName + "'", true);
    }
   
      /**
       * Loads a History
       * @param HistoryPK the primary key of the History as stored in the HWM database
       * @return an ABCMResult indicating the success or failure of this operation
       */
      public ABCMResult loadHistory(long HistoryPK){
        checkIntegrity();
        HistoryManager hm = new HistoryManager(env.getDBGateway());
        history = hm.getHistoryFromDB(HistoryPK);
        modelTime.setDate(history.getFirstDate()-1);
        return new ABCMResult("Successfully loaded history", true);
      }
      
      /**
       * Loads a History
       * @param historyName the name of the History as stored in the HWM database
       * @return A string indicating the success of the operation.
       */
      public ABCMResult loadHistory(String historyName){
        HistoryManager HM = new HistoryManager(env.getDBGateway());
        long HPK = HM.getPK(historyName);
        if(HPK == -1) return new ABCMResult("History '" + historyName + "' could not be loaded.", false);
        else{
          ABCMResult result = loadHistory(HPK);
          if(!result.success) return result;
          return new ABCMResult(loadHistory(HPK).text + " '" + historyName + "'.", true);
        }
      }      
      
      /**
       * Loads a configuration
       * @param ConfigurationPK The primary key of the Configuration as stored in the HWM database
       * @return A String indicating the success of the operation
       */
      public ABCMResult loadConfiguration(long ConfigurationPK){
        checkIntegrity();
        AbstractConfigurationManager cm = env.getConfigurationManager();
        configuration = cm.getConfiguration(ConfigurationPK);
        if(configuration == null) System.out.println("ERROR loading config");
        if(configuration == null) return new ABCMResult("Could not load configuration (" + ConfigurationPK + ") :" + cm.getErrorReport(), false);
        else return new ABCMResult("Successfully loaded configuration " + configuration.getName(), true);
      }
      
      /**
       * Loads a configuration
       * @param configName the name of the Configuration as stored in the HWM database.
       * @return A String indicating the success of the operation
       */
      public ABCMResult loadConfiguration(String configName){
        AbstractConfigurationManager cm = env.getConfigurationManager();
        configuration = cm.getConfiguration(cm.getPK(configName));
        
        if( configuration == null ) return new ABCMResult("Could not load configuration '" + configName + "' :" + cm.getErrorReport(), false);
        else                        return new ABCMResult("Successfully loaded configuration '" + configuration.getName() + "'", true);
      }
      
  /**
   * Loads a parameter set.
   * Note that if the new parameter set specifies a different random number generator
   * type, a new RNG is created and substituted for the existing one. This is true
   * even when the new parameter set has no value for RNG (that is, if the previous parameter
   * set had a nonzero value for RNG, and the new one has no value, a new instance of the
   * default RNG is created). 
   * @param parameterSetPK the primary key of the parameter set as stored in the HWM database.
   * @return A Sring indicating the success of the operation
   */
  public ABCMResult loadParameterSet(long parameterSetPK){
    return loadParameterSet(parameterSetPK, null);
  }
  
  /**
   * Loads a parameter set using the specified simulation run.
   * The simulation run can be used to pass a specified random
   * number seed
   * @param parameterSetPK PK of the parameter set being loaded
   * @param sr simulation run to use
   * @return an ABCMResult indicating the success or failure of this operation
   */
  public ABCMResult loadParameterSet(long parameterSetPK, SimulationRun sr){
    checkIntegrity();
    int rngType = (AbstractModelComponent.getCountOfParameters() > 0 ? (int)paramVal("RANDOMNUMBERGENERATORTYPE") : 0);
    ParameterSetManager psm = new ParameterSetManager(env.getDBGateway());
    ParameterSet P = psm.getParameterSet(parameterSetPK);
    BasicRNG rng = null;
    int newRNGType = (int)P.getParameter("RANDOMNUMBERGENERATORTYPE"); 
    if(rngType != newRNGType) rng = (sr == null ? new BasicRNG(newRNGType) : new BasicRNG(newRNGType, sr.getRandomNumberSeed())); 
    AbstractModelComponent.setModelComponent(null, P, rng);
    return new ABCMResult("Successfully loaded parameter set", true);
  }
  
  /**
   * Loads a parameter set by name
   * @param paramSetName Name of the parameter set in the HWM Databse to be loaded
   * @return an ABCMResult indicating the success or failure of this operation
   */
  public ABCMResult loadParameterSet(String paramSetName){
    ParameterSetManager psm = new ParameterSetManager(env.getDBGateway());
    long PK = psm.getPK(paramSetName);
    return new ABCMResult(loadParameterSet(PK).text + " '" + paramSetName + "'", true);
  }  
  
  
  /**
   * Loads a probe set
   * 
   * Note that loading a probe set does not affect model integrity
   * 
   * @param probeSetPK the primary key of the probe set as stored in the HWM database
   * @return A String indicating the success of the operation
   */  
  public ABCMResult loadProbeSet(long probeSetPK){
    int count = 0;
    ProbeSetManager psm = new ProbeSetManager(env.getDBGateway());
    Vector v = psm.getProbesInProbeSet(probeSetPK);
    if((v == null) || (v.size() == 0)) return new ABCMResult("No probes loaded; probe set may be empty or may not exist ", 0, false);
    while(v.size() > 0){
      count++;
      Probe p = (Probe)v.remove(0);
      loadProbe(p);
    }
    return new ABCMResult("Loaded " + count + " probes ", count, true);
  }
  
  /**
   * Loads a probe set
   * @param probeSetName The name of the probe set as stored in the HWM database
   * @return A String indicating the success of the operation
   */
  public ABCMResult loadProbeSet(String probeSetName){
    ProbeSetManager psm = new ProbeSetManager(env.getDBGateway());
    long pk = psm.getPK(probeSetName);
    if(pk == -1) return new ABCMResult("No probe set named '" + probeSetName + "' exists", 0, false);
    ABCMResult result = loadProbeSet(pk);
    return new ABCMResult(result.text + "from Probe Set ' " + probeSetName + "'", result.value, result.success);
  }
  
    /**
     * Loads a probe
     * @param probePK the primary key of the probe as stored in the HWM database.
     * @return A String indicating the success of the operation
     */
    public ABCMResult loadProbe(long probePK){
      ProbeManager pm = new ProbeManager(env.getDBGateway());
      Probe P = pm.getProbe(probePK);
      loadProbe(P);
      return new ABCMResult("Successfully loaded probe", true);
    }

    /**
     * Loads the specified probe
     * @param P probe to be loaded
     */
    private void loadProbe(Probe P){
      if(P.isMonthly()){
        this.createMonthlyProbe(P.getProbeName(), P.getSeriesName(), P.getDomain(), P.isNumeric(), P.getRetrieveCommand(), P.getOffset());
      }
      else{
        this.createProbe(P.getProbeName(), new SimDate(P.getBaseDate()), P.getFrequency(), P.getSeriesName(), P.getDomain(), P.isNumeric(), P.getRetrieveCommand(), P.getOffset());
      }
    }
  
    
  /* Setting options and parameters*/  
    
  /**
   * Sets a code model option, identified by a String, to a particular
   * value
   * @param option The name of the option to be set
   * @param val The value to which the option will be set
   * @return an ABCMResult indicating the success or failure of this operation
   */
  public ABCMResult setCodeModelOption(String option, int val){
    checkIntegrity();
    getCodeModel().setCodeModelOption(option, val);
    return new ABCMResult("Set code model option '" + option + "' to value " + val, true);
  }
  
  /**
   * Sets a parameter by name
   * @param paramName The name of the parameter to be set
   * @param value The value to which the parameter is to be set
   * @return an ABCMResult indicating the success or failure of this operation
   */
  public ABCMResult setParameter(String paramName, double value){
    checkIntegrity();
    getParameterSet().setParameter(paramName, value);
    return new ABCMResult("Successfully set parameter '" + paramName + "' to " + value + ".", true);
  }
 
  /**
   * Sets the random number seed by creating a new RNG of the
   * type indicated by the parameter set.
   * @param randomNumberSeed
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult setRandomNumberSeed(long randomNumberSeed){
    BasicRNG rng = null;
    int newRNGType = (int)this.getParameterSet().getParameter("RANDOMNUMBERGENERATORTYPE"); 
    rng = new BasicRNG(newRNGType, randomNumberSeed); 
    AbstractModelComponent.setModelComponent(null, null, rng);
    return new ABCMResult("Set random number seed to " + randomNumberSeed, true);
  }

  /**
   * Gets the run number of the current simulation run being run by this model
   * @return the model's run number; will be -1 if a test model.
   */
  public long getRunNumber(){    return runID;  }  
  
  /**
   * Gets the Space object of the model
   * @return the Space object of the model
   */
  public abstract Space getSpace();
  
    /**
   * Returns the simulation's current History object
   * @return The History object of the Simulation
   */
  public History getHistory(){   return history;  }

  // General Step method
  /**
   * The model's overall 'Step' method; consists of moving forward
   * one simulation day, processing any instructions for that day,
   * and outputting to the GUI if appropriate.
   * @return A String that can be written to an output window
   */
  public ABCMResult step(){
    System.out.println("\n\nSTEPPING " + modelTime.getString());
    inStep = true;
    StringWriter out = new StringWriter();
    out.write("STEPPING " + modelTime.getString() + "\n");
    if(modelTime.dayOfYear() == 0) System.out.println("Beginning new year for Run Number " + runID + " Date: " + modelTime.getString());
    // If there is a history. There almost always will be, but not
    // necessarily for testing
    if(history != null){
      history.setDate(modelTime);
      // Execute all historical instructions
      while(history.hasMoreCommandsOnCurrentDate()){
        out.write(modelTime.getString() + ": " + processCommand(history.getNextCommand()).text + "\n");
      }
    }

    // Step the model
    modelStep();
    
    // Now step any agents in the model; agents go after the rest of the model's daytime step
    stepPopulation();
    
    // Capture Data
    captureData();
    
    // Update the date to the next day
    modelTime.incDay();
       
    //env.setSimDate(modelTime.getString());
    env.updateGUI();
    
    inStep = false;
    
    return new ABCMResult(out.toString(), true);
  }  

  /**
   * Fundamental 'step' method of a model. Extensions to this
   * class should implement this method to perform their central
   * routines.
   */
  public abstract void modelStep();

  /**
   * Allow a population, if any, to change or make decisions
   *
   */
  public void stepPopulation(){
    if(population != null) population.step();
  }

  
/* Step Commands */
  
  /**
   * Step the model forward to the specified date.
   * @param toDate Date to which the model is to be run
   * @return a response that can be written to an output window
   */
  public ABCMResult stepToDate(double toDate){
    StringWriter out = new StringWriter();
    while(modelTime.value() < toDate){
      out.write(step().text);
    }
    return new ABCMResult(out.toString(), true);
  }
  
  /**
   * Step the model forward a specified number of days.
   * @param days Number of days through which to step
   * @return A response that can be written to the output window.
   */
  public ABCMResult StepForwardNDays(long days){
    StringWriter out = new StringWriter();
    double toDate = modelTime.value() + days;
    while(modelTime.value() < toDate){
      out.write(step().text);
    }
    return new ABCMResult(out.toString(), true);
  }
  
  /**
   * Steps the model forward to the specified date.
   * @param toDate Date to which the model is to be run
   * @return a response that can be written to an output window
   */
  public ABCMResult stepToDate(String toDate){
    // Some cleverness will be helpful...
    // If the date is only the day and month, figure
    // out when the next instance of this is...
    double toDateVal = 0;
    if(toDate.length() == "01-JAN".length()){
      toDateVal = SimDate.parseString(toDate + "-SEASONS");
      while(modelTime.value() > toDateVal) toDateVal += 365;
    }
    else{
      toDateVal = SimDate.parseString(toDate);
    }
    return new ABCMResult(stepToDate(toDateVal).text + "\n" +
           "Stepped to date: " + modelTime.getString(), true);
  }
  
  /**
   * Steps the model forward all the way to the end date.
   * @return A string indicating success of the operation.
   */
  public ABCMResult stepToEnd(){
    if(history == null) return new ABCMResult ("Could not step to end of model; no History has been loaded.", false);
    return stepToDate(history.getLastDate() + 1);
  }
  
  /**
   * Capture data according to whatever probes are set
   */
  public void captureData(){
    Enumeration e = probes.elements();
    while(e.hasMoreElements()){
      Probe p = (Probe)e.nextElement();
//      System.out.println("Probe: " + p.getProbeName() + " executing...");
      if(p.isTimeToUpdate(modelTime)){
        SimulationDataSeries series = p.getSeries();
        String  retrieveCommand     = p.getRetrieveCommand();
        SimDate datumTimeStamp      = new SimDate(modelTime.value() + p.getOffset());
        ABCMResult result = getValue(retrieveCommand);
        if(result.success){
          if(series.isNumeric()) series.addElement(datumTimeStamp, result.value);
          else                   series.addElement(datumTimeStamp, result.text);
          System.out.println("Series " + series.getName() + " domain " + series.getDomain() + " has " + series.getSeries().size() + " elements");
        }
        else{
          System.out.println("Attempting to update probe " + p.getProbeName() + " and encountered problem: " + result.text);
        }
      }
    }
  }

  /* Parse Commands */
  /**
   * Call the parser to process a command given as a String.
   * The String should be in a format where the first element
   * matches a valid command, and the remainder serve as arguments;
   * elements are space-delimited.
   * @param s String of command to be processed
   * @return a response that can be written to an output window.
   */
  public ABCMResult processCommand(String s){
    return env.processCommand(s);
  }
  
  /**
   * Call the parser to process a command given in int, String[]
   * format- that is, the first element is an integer that is
   * the primary key of a specific valid command, and the
   * array of Strings contains the list of arguments.
   * @param c primary key of command to be executed
   * @param argL Array of String arguments to the command
   * @return a response that can be written to an output window
   */
  public ABCMResult processCommand(int c, String[] argL){
    return env.processCommand(c, argL);
  }
  
  /**
   * Process a specified History Instruction
   * @param HI History Instruction to be processed
   * @return a response that can be written to an output window
   */
  public ABCMResult processCommand(HistoryInstruction HI){
    return env.processCommand(HI); 
  }  
  
  /**
   * Sets the current simulation date/time
   * @param d new simulation date
   */
  public void setModelTime(double d){
    checkIntegrity();
    modelTime.setDate(d);
  }
  
  /**
   * Gets the current model time
   * @return the current model time
   */
  public SimDate getModelTime(){
    return modelTime;
  }
  
  /**
   * Set the model time to the specified Simulation Date
   * @param sd New Simulation Date, in HWM Date format (01-JAN-1234-AD)
   * @return an ABCMResult indicating the success or failure of this operation
   */
  public ABCMResult setModelTime(String sd){
    checkIntegrity();
    double newTime = (new SimDate(sd)).value();
    if(newTime > 0){
      setModelTime(newTime);
      return new ABCMResult("Set Model Date to " + modelTime.getString(), true);
    }
    else return new ABCMResult("Invalid date or other issue prevented date " +
                "from being reset; current simulation date is " + 
                modelTime.getString(), false);
  }
 
  /* Time */
  /**
   * Set the value of the 'period' variable for this model 
   * @param newPeriod new value for the 'period' variable
   * for this model.
   * @return a response that can be displayed to an output window
   */
  public ABCMResult setCurrentPeriod(String newPeriod){
    checkIntegrity();
    currentPeriod = newPeriod;
    return new ABCMResult("Period now set to: " + currentPeriod, true);
  }

  /* Space */
  /**
   * Add a point to the Model's collection of known points
   * @param pName name for the new point
   * @param x X coordinate of the point
   * @param y Y coordinate of the point
   * @param z Z coordinate of the point
   * @return a response that can be written to an output window
   */
  public ABCMResult addPoint(String pName, double x, double y, double z){
    checkIntegrity();
    getSpace().addPoint(pName, new DatumPoint(x, y, z, pName));
    return new ABCMResult("Point " +  pName + " added successfully.", true);
  }
  
  /**
   * Adds a point whose x,y coordinates are given and whose z coordinate
   * is determined from the existing datum points. Default is zero if
   * coordinates given are not within existing surface.
   * @param pName name to be assigned to the new point
   * @param x new point's x coordinate
   * @param y new point's y coordinate
   * @return an ABCMResult object indicating success or failure of the operation
   * (note: always succeeds; even if a point with the same name has been
   * added previously, new point will overwrite).
   */
  public ABCMResult addSurfacePoint(String pName, double x, double y){
//    checkIntegrity();
//    double newZ = getSpace().getZ(new Point(x,y));
//    if(Double.isNaN(newZ)) newZ = 0;
//    getSpace().addPoint(pName, new DatumPoint(x, y, newZ));
    getSpace().addSurfacePoint(pName, new DatumPoint(x, y, 0, pName));
    return new ABCMResult("Point " + pName + " added successfully; z coordinate will be calculated.", true);
  }
  
  /**
   * Gets a list of the points known to this model
   * @return a String containing a list of points known to the model
   */
  public ABCMResult getListOfPoints(){
    return new ABCMResult(getSpace().getListOfPoints(), true);
  }

  /**
   * Sets the orientation of the space in this model
   * @param northIsPositive true if increasing numbers represent north, false if they represent south
   * @param eastIsPositive true if increasing numbers represent east, false if they represent west
   * @return a message indicating the new values set by this operation
   */
  public ABCMResult setNorthAndEastOrientation(boolean northIsPositive, boolean eastIsPositive){
    getSpace().setNorthIsPositive(northIsPositive);
    getSpace().setEastIsPositive(eastIsPositive);
    return new ABCMResult("Set: North is " + (northIsPositive ? "Positive" : "Negative") + "; " +
                "East is "  + (eastIsPositive  ? "Positive" : "Negative") + "; ", true);
  }  
  
  /* Populations and Agents */
  /**
   * Sets the current population to null; allows a new population to be initialized.
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public final ABCMResult clearPopulation(){ 
    population = null;
    return new ABCMResult("Population is cleared; must re-create and re-populate.", true);
  }
  
  /**
   * Creates a new empty population. Will fail if this model already contains a population
   * @return an ABCMResult indicating success or failure of this operation
   */
  public final ABCMResult createPopulation(){
    if(population != null) return new ABCMResult("Population is not yet cleared; must clear existing population before creating a new one", false);
    population = getPopulation();
    return new ABCMResult("New population created. Contains " + population.countOfAgents() + " agent" + (population.countOfAgents() > 1 ? "s" : "") + ".", true);
  } 
  
  /**
   * Implementing classes should provide a means to create a population
   * of agents. 
   * @return the new population (usually empty but optionally initialized with agents)
   */
  public abstract AbstractABCModelPopulation getPopulation();

  
  /* Pass commands to population and agents */
  /**
   * Directs the population to take some action by passing a command to it. Fails if
   * no population has been created.
   * @param command the command to be executed
   * @return an ABCMResult indicating the outcome of this action, if successful, or
   * cause of failure.
   */
  public ABCMResult populationDo(String command){
    if(population == null) return new ABCMResult("No population of agents exists", false);
    return population.processCommand(command);
  }
  
  /**
   * Directs a specific agent in the population to take some action by passing a command
   * to it. Fails if there is no population or if there is no agent by the specified
   * ID
   * @param agentID ID String of the agent to act
   * @param command command that the agent should undertake
   * @return an ABCMResult object indicating the outcome of this operation, if successful, or
   * cause of failure.
   */
  public ABCMResult agentDo(String agentID, String command){
    if(population == null) return new ABCMResult("No population of agents exists", false);
    return population.sendCommandToAgent(agentID, command);
  }
    
  /* ************************************************************************** */
  /* GUI                                                                        */
  /* ************************************************************************** */
  /**
   * Adds a Data Series to a graph
   * @param seriesName The name of the Data Series to be added. Must be a Data
   * Series that already exists in the model.
   * @param graphName The name of the graph to which this Data Series will be
   * added. Must be an extant graph.
   * @return an ABCMResult indicating the success or failure of this operation
   */
  public ABCMResult addDataSeriesToGraph(String seriesName, String graphName){
    if(!dataSeries.containsKey(seriesName)) return new ABCMResult("Data series '" + seriesName + "' does not exist.", false);
    return env.addSeriesToGraph(graphName, (SimulationDataSeries)dataSeries.get(seriesName), seriesName);     
    
  }
  
  /**
   * Gets the specified simulation data series if it exists
   * @param seriesName The series to be retrieved
   * @return the specified series; if no series with the given name exists, returns null
   */
  public SimulationDataSeries getDataSeries(String seriesName){
    if(!dataSeries.containsKey(seriesName)) return null;
    return (SimulationDataSeries)dataSeries.get(seriesName);
  }
  

  /* ************************************************************************** */
  /* Generic Information Portals                                                */
  /* ************************************************************************** */
    
  /**
   * Gets a value from the model using the specified request command
   * @param requestCommand command to be parsed specifying the value to be returned
   * @return the result of the request command
   */
  public ABCMResult getValue(String requestCommand){ return env.processCommand(requestCommand); }

  /**
   * Gets a report indicating the model's status
   * @return a model status report
   */
  protected String modelStatus(){
    String retVal = "";
    retVal = "Model status report:\n\n";
    // Run number
    if(runID > -1) {
      retVal += "  Run Number: " + runID + "\n";
      retVal += "  Model Is Reproducible: " + modelIntegrity + "\n";
    }
    else           retVal += "  Non-saveable run\n";
    
    retVal += "  User:          " + env.getUser() + "\n";               
    retVal += "  Simulation:    " + (simulation != null ? simulation.getSimName() : "No simulation loaded.") + "\n"; 
    retVal += "  " +  getCodeModel().getReport();                   
    retVal += "  Data Model:    " + (dataModel!= null ? "  " + dataModel.getName()      : "No Data model loaded.") + "\n";
    retVal += "  History:       " + (history != null ? "  " + showDate().text   : "No History loaded.") + "\n";
    retVal += "  Configuration: " + (configuration != null ? "  " + configuration.getName()    : "No configuration loaded.") + "\n";
    if(configuration != null) retVal += configuration.getReport();
    retVal += "  " + getParameterSet().getReport();
    retVal += "  Probes: " + probes.size() + "\n";
    return retVal;
  }
  
  
  /**
   * Gets a report describing the current configuration- listing all plants,
   * plant stages, and values, etc.
   * @return a String containing a report of the details of the current 
   * configuration object.
   */
  public ABCMResult getConfigurationReport(){
    if(configuration == null) return new ABCMResult("No configuration has been created for this model.", false);
    return new ABCMResult(configuration.getReport(), true);
  }
  
  /**
   * Returns a report of the model's current time and period.
   * Output is formatted: "History: 01-JAN-0123-AD (Period)"
   * If no history is set, returns "<No history set> - 01-JAN-0123-AD (Period)"
   * @return the report of model's current time and period. 
   */
  public ABCMResult showDate(){
    String retVal = "";
    if(history != null) retVal += "History: " + history.getName() + " ";
    else          retVal += "No history set - ";
    retVal += modelTime.getString() + " (" + currentPeriod + ")";
    return new ABCMResult(retVal, true);
  }  

  
  /**
   * Returns a list of the parameters and values in use by the model
   * @return The list of parameters and values in the Parameter Set for this model
   */
  public ABCMResult getParameterList(){
    return new ABCMResult(getParameterSet().getReport(), true);
  }  


  /* ************************************************************************** */
  /* Data Series and Probes                                                     */
  /* ************************************************************************** */
  /**
   * Creates a numeric data series
   * 
   * The name provided for the data series must not already be in use as a Data
   * series name, nor an empty String. 
   * @param seriesName The name of the Data Series to be created.
   * @param variableName The name to be assigned to the variable of the Data Series
   * created.
   * @return A String indicating the success or failure of the operation
   */
  public ABCMResult createNumericDataSeries(String seriesName, String variableName){
    // Validation
    if((seriesName == null ) || (variableName) == null) return new ABCMResult("Cannot use null values for either series or variable name.", false);
    if((seriesName.length() == 0 ) || (variableName.length() == 0)) return new ABCMResult("Cannot use empty strings for either series or variable name.", false);
    if(dataSeries.containsKey(seriesName)) {
      return new ABCMResult("Data series " + seriesName + " already exists", false);
    }
    SimulationDataSeries sds = new SimulationDataSeries(variableName, SimulationDataSeries.IS_NUMERIC, seriesName);
    dataSeries.put(seriesName, sds);
    return new ABCMResult("Created Data Series '" + seriesName + "' with variable '" + variableName + "'", true);
  }
  
  /**
   * Get a printout of a Data Series and its values
   * @param seriesName The name of the Data Series to be reported
   * @return a report of the Data Series and its data
   */
  public ABCMResult getDataSeriesPrintout(String seriesName){
    if(!dataSeries.containsKey(seriesName)) return new ABCMResult("Model does not contain data series '" + seriesName + "'", false);
    return new ABCMResult(((SimulationDataSeries)dataSeries.get(seriesName)).getSeriesPrintout(), true);
  }
  
  /**
   * Get a list of the Data Series contained in this model
   * @return a String containing a list of Data Series in this model
   */
  public ABCMResult getListOfDataSeries(){
    String list = "Data Series available: " + dataSeries.size() + "\n\n";
    if(dataSeries.size() > 0){
      Enumeration e = dataSeries.keys();
      while(e.hasMoreElements()) list += "Series Name: " + (String)e.nextElement() + "\n";
    }
    return new ABCMResult(list, true);
  }
  
  /**
   * Creates a probe
   * @param probeName the name of the probe to be created
   * @param baseDate base date for the probe
   * @param frequency frequency at which the probe is to be executed
   * @param dataSeriesName data series to which the probe sends its results
   * @param dataSeriesDomain name of the domain in the data series to which the probe sends its results
   * @param numeric true if the probe returns numeric values, false otherwise
   * @param retrieveCommand command executed by this probe
   * @return - A String indicating the successful creation of the probe
   */
  public ABCMResult createProbe(String probeName, SimDate baseDate, long frequency, 
      String dataSeriesName, String dataSeriesDomain, boolean numeric, String retrieveCommand){
    return createProbe(probeName, baseDate, frequency, dataSeriesName, dataSeriesDomain, numeric, retrieveCommand, -1);      
  }
  
  /**
   * Creates a probe
   * @param probeName the name of the probe to be created
   * @param baseDate base date for the probe
   * @param frequency frequency at which the probe is to be executed
   * @param dataSeriesName data series to which the probe sends its results
   * @param dataSeriesDomain name of the domain in the data series to which the probe sends its results
   * @param numeric true if the probe returns numeric values, false otherwise
   * @param retrieveCommand command executed by this probe
   * @param offset double value specifying offset from 12:00 midnight to be used when
   * issuing the command. Useful for recovering time-dependent data
   * @return a message indicating whether the probe has been successfully created
   */
  public ABCMResult createProbe(String probeName, SimDate baseDate, long frequency, 
      String dataSeriesName, String dataSeriesDomain, boolean numeric, String retrieveCommand, double offset){
    // If there is already a probe with this name, can't add the new one
    if(probes.containsKey(probeName))           return new ABCMResult("Probe '" + probeName + "' already exists; " +
                                                       "could not create new probe.", false);
    // If there is no data series with the specified name, create one
    if(!dataSeries.containsKey(dataSeriesName)) {
      dataSeries.put(dataSeriesName, new SimulationDataSeries(dataSeriesDomain, numeric, dataSeriesName));
    }
    // Note there is some slippage; the method call passes a domain name, and this is used
    // if the simulation data series is created anew. However, if there is already an SDS,
    // the existing domain name is used- there is no check to make sure they match.
    SimulationDataSeries sds = (SimulationDataSeries)dataSeries.get(dataSeriesName); 
    Probe p = new Probe(probeName, baseDate, frequency, 
       dataSeriesName, sds.getDomain(), numeric, retrieveCommand);
    if(offset >= 0) p.setOffset(offset);
    p.setSeries(sds);
    probes.put(probeName, p);
    return new ABCMResult("Created probe '" + probeName + " new size of Probes = " + probes.size(), true);      
  }
  
  
  /**
   * Creates a monthly probe.
   * 
   * This method will fail to create a new probe if the data series
   * does not exist, or if the name specified for the probe
   * already refers to a probe in the model.
   * 
   * @param probeName New name by which probe will be referenced
   * @param dataSeriesName Data series which this probe will store to
   * @param dataSeriesDomain name of the domain in teh data series to which the probe sends its
   * results
   * @param numeric true if the probe returns numeric values, false otherwise
   * @param retrieveCommand Command used by the probe to get data from the model
   * @return A message indicating the success or failure of this method
   */
  public ABCMResult createMonthlyProbe(String probeName, String dataSeriesName, String dataSeriesDomain, 
      boolean numeric, String retrieveCommand){
    return createMonthlyProbe(probeName, dataSeriesName, dataSeriesDomain, numeric, retrieveCommand, -1);        
  }
  
  /**
   * Creates a monthly probe.
   * 
   * This method will fail to create a new probe if the data series
   * does not exist, or if the name specified for the probe
   * already refers to a probe in the model.
   * 
   * @param probeName the name of the probe to be created
   * @param dataSeriesName data series to which the probe sends its results
   * @param dataSeriesDomain name of the domain in the data series to which the probe sends its results
   * @param numeric true if the probe returns numeric values, false otherwise
   * @param retrieveCommand command executed by this probe
   * issuing the command. Useful for recovering time-dependent data
   * @param offset double value specifying offset from 12:00 midnight to be used when
   * issuing the command. Useful for recovering time-dependent data
   * @return A message indicating the success or failure of this method
   */
  public ABCMResult createMonthlyProbe(String probeName, String dataSeriesName, String dataSeriesDomain, 
      boolean numeric, String retrieveCommand, double offset){
    if(probes.containsKey(probeName))           return new ABCMResult("Probe '" + probeName + "' already exists; could not create new probe.", false);
    if(!dataSeries.containsKey(dataSeriesName)) {
      dataSeries.put(dataSeriesName, new SimulationDataSeries(dataSeriesDomain, numeric, dataSeriesName));
    }
    SimulationDataSeries sds = (SimulationDataSeries)dataSeries.get(dataSeriesName); 
    Probe p = new Probe(probeName, dataSeriesName, sds.getDomain(), numeric,  retrieveCommand);
    p.setSeries(sds);
    if(offset >= 0) p.setOffset(offset);
    probes.put(probeName, p);
    return new ABCMResult("Created monthly probe '" + probeName, true);        
  }
  
  
  /**
   * Drops a Probe from this model
   * @param probe the name of the Probe to be dropped
   * @return A String indicating the success or failure of this method
   */
  public ABCMResult dropProbe(String probe){
    if(probes.containsKey(probe)) {
      probes.remove(probe);
      return new ABCMResult("Probe '" + probe + "' dropped.", true);
    }
    return new ABCMResult("Probe '" + probe + "' does not exist.", false);
  }
  
  /**
   * Turns a Probe on. This will cause it to begin or resume collecting
   * data.
   * @param probe the name of the Probe to be activated
   * @return a String indicating the success or failure of the action
   */
  public ABCMResult turnProbeOn(String probe){
    if(probes.containsKey(probe)) {
      ((Probe)probes.get(probe)).turnProbeOn();
      return new ABCMResult("Probe '" + probe + "' is turned on.", true);
    }
    return new ABCMResult("Probe '" + probe + "' does not exist.", false);
  }
  
  /**
   * Turns a Probe off. This will cause it to suspend the collection
   * of data.
   * @param probe the name of the Probe to be deactivated
   * @return a String indicating the success or failure of the operation.
   */
  public ABCMResult turnProbeOff(String probe){
    if(probes.containsKey(probe)) {
      ((Probe)probes.get(probe)).turnProbeOff();
      return new ABCMResult("Probe '" + probe + "' is turned off.", true);
    }
    return new ABCMResult("Probe '" + probe + "' does not exist.", false);
  }
  
  /**
   * Gets a list of the Probes in this model.
   * @return a list of the Probes in t he model.
   */
  public ABCMResult getListOfProbes(){
    String list = "Probes available: \n\n";
    Enumeration e = probes.keys();
    while(e.hasMoreElements()){
      String probeName = (String)e.nextElement();
      Probe p = (Probe)probes.get(probeName);
      list += "    " + probeName + " " + p.getReport() + "\n";
    }
    return new ABCMResult(list, true);
  }

  /**
   * Sets the offset value for an extant probe
   * @param probe name of the probe to be set
   * @param offset new value for the probe's offset
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult setProbeOffset(String probe, double offset){
    if(probes.containsKey(probe)) {
      double o = ((Probe)probes.get(probe)).setOffset(offset);
      return new ABCMResult("Probe '" + probe + "' offset set to " + o + ".", true);
    }
    return new ABCMResult("Probe '" + probe + "' does not exist.", false);
  }
  
  /* ************************************************************************** */
  /* Cloning this model object.                                                 */
  /* ************************************************************************** */
  
  /**
   * Implementing classes must be able to return a complete deep clone of the model
   * object. Most implementations will need to call the cloneModelCore() method.
   * @param cloneStore the Clone Store for the copy process
   * @return a 'Deep Clone' of the model object
   */
  public abstract AbstractABCModel getModelClone(ABCMCloneStore cloneStore);
  
  /**
   * Clones the instance variables defined in the abstract model class;
   * implementing classes creating Deep Clones of the model object should
   * call this
   * @param modelToClone the model object that will be the clone returned
   * @param cloneStore the Clone Store in which copies of objects can be
   * placed and retrieved
   */
  protected void cloneModelCore(AbstractABCModel modelToClone, ABCMCloneStore cloneStore){
    
    if(cloneStore == null) cloneStore = new ABCMCloneStore();
    
    // Instance variables:
    modelToClone.currentPeriod   = currentPeriod;
    modelToClone.inStep          = inStep;
    modelToClone.runID           = -1; 
    modelToClone.modelIntegrity  = modelIntegrity;
    
    modelToClone.modelTime       = (modelTime     != null ? (SimDate)modelTime.getDeepClone(cloneStore) : null);
    modelToClone.simulation      = (simulation    != null ? (Simulation)simulation.getDeepClone(cloneStore) : null);
    modelToClone.dataModel       = (dataModel     != null ? (DataModel)dataModel.getDeepClone(cloneStore) : null);
    modelToClone.history         = (history       != null ? (History)history.getDeepClone(cloneStore) : null);
    modelToClone.configuration   = (configuration != null ? (Configuration)configuration.getDeepClone(cloneStore) : null);
    modelToClone.dataSeries      = (dataSeries    != null ? (ABCMDeepCloneableHashtable)dataSeries.getDeepClone(cloneStore) : null);
    modelToClone.probes          = (probes        != null ? (ABCMDeepCloneableHashtable)probes.getDeepClone(cloneStore) : null);
    
  }
  
  /* ************************************************************************** */
  /* Saving data to the database                                                */
  /* ************************************************************************** */
  /**
   * Saves the values in a Data Series to the HWM Database.
   * @param seriesName the name of the Data Series to be saved
   * @return a String indicating the success or failure of the action
   */
  public ABCMResult saveDataSeries(String seriesName){
    if(runID == -1) return new ABCMResult("Cannot save data series; run was not allocated an ID Number.", false);
    if(!dataSeries.containsKey(seriesName)) return new ABCMResult("No series '" + seriesName + "' exists.", false);
    SimulationDataSeriesManager sdsm = new SimulationDataSeriesManager(env.getDBGateway());
    SimulationDataSeries sds = (SimulationDataSeries)dataSeries.get(seriesName);
    sdsm.setInstanceToSave(sds, seriesName, runID);
    sdsm.setSeriesDescriptionToSave(sds.getDescription());
    sdsm.saveInstanceToDB();
    SimulationRunManager srm = new SimulationRunManager(env.getDBGateway());
    srm.setRunIsReproducible(modelIntegrity, runID);
    System.out.println(sdsm.getErrorReport());
    return new ABCMResult("Saved series to database.", true);
  }
  
  /**
   * Saves all data from all Data Series in the model
   * @return a String indicating the success or failure of the operation
   */
  public ABCMResult saveAllDataSeries(){
    System.out.println("Saving all data series: run number: " + runID);
    if(runID ==  -1) return new ABCMResult("Cannot save data series; run was not allocated an ID Number.", false);
    String seriesList = "";
    Enumeration e = dataSeries.keys();
    while(e.hasMoreElements()){
      String seriesName = (String)e.nextElement();
      saveDataSeries(seriesName);
      seriesList += seriesName + "\n";
    }
    return new ABCMResult("Saved data series:\n\n" + seriesList, true);
  }

  
  
  /**
   * Get a programmatically accessible list (Enumeration) of the Data series
   * contained in this model.
   * @return the collection of Data Series in the model
   */
  public Enumeration getDataSeriesNames(){   return dataSeries.keys();  }
  
  
  /**
   * Turns the specified seasonal narrative 'on'
   * @param narrativeName narrative to be activated
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult turnSeasonalNarrativeOn(String narrativeName){
    if(history == null) return new ABCMResult("Could not set seasonal narrative; no history exists", false);
    history.setSeasonalNarrativeStatus(narrativeName, true);
    return new ABCMResult("Set seasonal narrative '" + narrativeName + "' to 'On'", true);
  }
  
  
  /**
   * Turns the specified seasonal narrative 'off'
   * @param narrativeName narrative to be deactivated
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult turnSeasonalNarrativeOff(String narrativeName){
    if(history == null) return new ABCMResult("Could not set seasonal narrative; no history exists", false);
    history.setSeasonalNarrativeStatus(narrativeName, false);
    return new ABCMResult("Set seasonal narrative '" + narrativeName + "' to 'Off'", true);
  }
  
  /**
   * Select randomly from among a given list of narratives. The list may include the same
   * narrative name more than once to make it more likely to be the one selected
   * @param narratives names of the narratives to be chosen
   * @return a message indicating the success or failure of this operation, including the
   * name of the narrative selected
   */
  public ABCMResult chooseSeasonalNarrative(String[] narratives){
    int narrativeCount = narratives.length;
    if(narrativeCount == 0) return new ABCMResult("No narratives listed.", false);
    int selectedNarrative = (int)(Math.floor(getRNG().nextDouble() * narrativeCount));
    String narrativeName = narratives[selectedNarrative];
    for(int x = 0; x < narrativeCount; x++){    turnSeasonalNarrativeOff(narratives[x]);    } 
    turnSeasonalNarrativeOn(narrativeName);
    return new ABCMResult("Set seasonal narrative '" + narrativeName + "' to 'On'", true); 
  }

  /* ************************************************************************** */
  /* SVG                                                                        */
  /* ************************************************************************** */
  
  /** 
   * @see abcm.graphics.SVGRepresentable#addToSVG(abcm.graphics.SVGAssist) 
   */
  public void addToSVG(SVGAssist svg){
    
    Space space = getSpace();
    
    // The Model is added to the SVG object by adding two
    // SVGRepresentable objects: the Legend and the space object
    // The Legend is an inner class of the model;

    /**
     * Class representing the legend to be added to the map
     */
    class Legend implements SVGRepresentable{
      private double viewportTopLeftX;
      private double viewportTopLeftY;
      private double viewportWidth;
      private double viewportHeight;
    
      class Logo implements SVGRepresentable{
        private String imageLoc;
        
        public Logo(String imageLocation){
          imageLoc = imageLocation;
        }
        
        public void addToSVG(SVGAssist svg){
          String imageOutputLocation = env.getImageOutputLocation();
          double width = svg.getViewportWidthInUnits();
          double height = svg.getViewportHeightInUnits();
          if(imageLoc != null) svg.addImage(0, 0, width, height, imageLoc, "xMinYMin meet");
        }
      }
     
      
      /**
       * A class representing the image of a compass rose
       * @author John T. Murphy
       */
      class CompassRose implements SVGRepresentable{
        
        public void addToSVG(SVGAssist svg){
          // Need defs for compass rose
          XMLAssist xml = new XMLAssist();    
          xml.addLine("polygon", "fill=\"black\" stroke=\"black\" stroke-width=\"1\" points=\"-5,-10 0,-20 5,-10\" id=\"rosePoint\" ","");
          xml.addLine("polyline", "fill=\"none\" stroke-linejoin=\"round\" stroke-width=\"2\" stroke=\"black\" points=\"-4,1 -4,-10 4,0 4,-11\" id=\"compassN\"","");
          xml.open("symbol", "id=\"compassRose\" viewBox=\"-40 -40 80 80\"");
            xml.addLine("use", "xlink:href=\"#rosePoint\" transform=\"rotate(0)\"","");
            xml.addLine("use", "xlink:href=\"#rosePoint\" transform=\"rotate(90)\"","");
            xml.addLine("use", "xlink:href=\"#rosePoint\" transform=\"rotate(180)\"","");
            xml.addLine("use", "xlink:href=\"#rosePoint\" transform=\"rotate(270)\"","");
            xml.addLine("use", "xlink:href=\"#compassN\" transform=\"translate(0,-25)\"","");
            xml.addLine("<rect x=\"-10\" y=\"-10\" width=\"20.0\" height=\"20\" stroke=\"black\" stroke-width=\"1\" fill=\"none\"/>\n", XMLAssist.INDENT_BLOCK);
          xml.closeTag(); // Close Symbol    
          svg.addDef(xml.getXML(XMLAssist.EXCLUDE_HEADER));
          svg.addUse(0,0,"#compassRose");
        }
      }
    
      /**
       * A class in which 'buttons' can be placed on an SVG representation
       * @author John T. Murphy
       */
      class ControlBox implements SVGRepresentable{
        double height;
        double width = 6;
        Vector svgControls = new Vector();
    
        class svgControl implements SVGRepresentable{
          double height = .5;
          
          public String bColor;
          public String lbl;
          public String click;
          
          public svgControl(String label, String buttonColor, String onClick){
            bColor = buttonColor;
            lbl    = label;
            click  = onClick;
          }
              
          public void addToSVG(SVGAssist svg){
            
            class label implements SVGRepresentable{
              String t = "";
              label(String text){
                t = text;
              }
              public void addToSVG(SVGAssist svg){
                svg.addText(t, .1, 15, 1);
              }
            }
            
            // Add the button
            svg.addRectangle(.25 * height, .25 * height,
                             .75 * height, .75 * height,
                             "black", .01, bColor,
                             click);
            svg.addSVG(.75, .25 * height, svg.getViewportWidthInUnits(), height, 0,0,svg.getViewportWidthInUnits(), height, new SVGRepresentable[] {new label(lbl)}, "Label");
          }
        }
        
        
        public ControlBox(){
          height = .25;
        }
        
        public void addControl(String label, String buttonColor, String onClick){
          svgControl c = new svgControl(label, buttonColor, onClick);
          svgControls.add(c);
          height += c.height;
        }
        
        public void addToSVG(SVGAssist svg){
          double yPos = .125;
          double width = svg.getViewportWidthInUnits();
          for(int i = 0; i < svgControls.size(); i++){
            // Create a new SVG box
            svgControl c = (svgControl)svgControls.get(i);
            svg.addSVG(0, yPos,
                       width, c.height, 
                       0, 0, 
                       width, c.height, 
                       new SVGRepresentable[] {c}, "Control");
            yPos += c.height;
          }
        }
      }     
      
      class LegendTextArea implements SVGRepresentable{
        String text = "";
        
        public void addToText(String newText){     text += newText;      }
        public void addToSVG(SVGAssist svg){
          svg.addText(text, .2, 14, 2);
        }
      }
      
      public Legend(double widthInInches, double heightInInches){
         viewportTopLeftX = 0;
         viewportTopLeftY = 0;
         viewportWidth = widthInInches;
         viewportHeight =heightInInches; 
      }
      
      // These are available so Model sets them correctly
      public double getViewportTopLeftX(){ return viewportTopLeftX; }
      public double getViewportTopLeftY(){ return viewportTopLeftY; }
      public double getViewportWidth(){    return viewportWidth; }
      public double getViewportHeight(){   return viewportHeight; }
      
      public void addToSVG(SVGAssist svg){
        
        // Logo
        Logo iisLogo       = new Logo(env.getImageOutputLocation() + "\\" + env.getLogoImageName());
        
        // Compass rose
        CompassRose cr     = new CompassRose(); // Compass rose is infinitely scaleable- doesn't matter
        
        // Control Box
        ControlBox  cb     = new ControlBox();
        // ... and add the controls as appropriate
        cb.addControl("Hide Gridlines","gray","onclick=\"hideGridlines(evt);\"");
        if(getSpace().includeContourLines){
          cb.addControl("Hide Topo Lines","green","onclick=\"hideContourLines(evt);\""); }
        if(getSpace().includeDelaunayTriangulation){
          cb.addControl("Hide Delaunay","blue","onclick=\"hideDelaunayTriangulation(evt);\""); }

        // Legend Text area
        LegendTextArea LTA = new LegendTextArea();
        
        // Now add each of these as an SVG
        
        // First the Logo
        double logoWidth  = svg.getViewportWidthInUnits();
        double logoHeight = 2 * svg.getUnitsPerInchY(); 

        svg.addSVG(0,0,
                   logoWidth, logoHeight, 
                   0, 0, 
                   logoWidth, logoHeight,
                   new SVGRepresentable[]{iisLogo}, "Logo");
        
        // Next the compass rose
        double compassRoseHeightInches  = 2;
        double compassRoseWidth         = svg.getViewportWidthInUnits();
        double compassRoseHeight        = compassRoseHeightInches * svg.getUnitsPerInchY();
        double compassRoseXPosition     = 0;
        double compassRoseYPosition     = svg.getViewportHeightInUnits() - compassRoseHeight;
        
        svg.addSVG(compassRoseXPosition, compassRoseYPosition, 
                   compassRoseWidth, compassRoseHeight, 
                   0, 0, 
                   compassRoseWidth, compassRoseHeight, 
                   new SVGRepresentable[] { cr }, "Compass Rose");
        
        // Then the Control Box
        double controlBoxHeightInches  = cb.height;
        double controlBoxHeightUnits   = controlBoxHeightInches * svg.getUnitsPerInchY();
        double controlBoxXPosition     = 0;
        double controlBoxYPosition     = compassRoseYPosition - controlBoxHeightUnits;
        
        svg.addFramedViewport(controlBoxXPosition, controlBoxYPosition,
                  svg.getViewportWidthInUnits(), controlBoxHeightUnits,
                  0, 0,
                  cb.width, cb.height,  .01, "black",
                  new SVGRepresentable[] { cb }, "Control Box");
        
        // Finally the Text Area
        LTA.addToText(modelStatus());
        double textAreaXPosition = .25 * svg.getUnitsPerInchX();
        double textAreaYPosition = logoHeight + .25 * svg.getUnitsPerInchY();
        double textAreaWidth = (svg.getViewportWidthInInches() - .25) * svg.getViewportWidthInUnits();
        double textAreaHeight = svg.getViewportHeightInInches() - logoHeight - compassRoseHeightInches - controlBoxHeightInches;
        
        
        svg.addSVG(textAreaXPosition, textAreaYPosition, 
                   textAreaWidth, textAreaHeight, 
                   0, 0, 
                   textAreaWidth, textAreaHeight, 
                   new SVGRepresentable[]{ LTA },"Legend Text"); 
         
      }
    }
    
    /* End Inner Classes; begin addToSVG for Model*/
    // Need the map script
    svg.addScript("map.js");    
    
    // Place the legend
    // The left of the legend should be offset from the left of the page
    double legendLeft = MARGIN * svg.getUnitsPerInchX();
    // The top of the legend should be offset in from the top of the page
    double legendTop = MARGIN * svg.getUnitsPerInchY();
    // The legend width will be some fraction of the page- between a minimum and a max
    // First calculate it in inches, because this is how the max and min are specified
    double legendWidth = svg.getViewportWidthInInches() * LEGEND_FRACTION;
    if(legendWidth < MIN_LEGEND_WIDTH)      legendWidth = MIN_LEGEND_WIDTH;
    if(legendWidth > MAX_LEGEND_WIDTH)      legendWidth = MAX_LEGEND_WIDTH;
    // ... then convert to current viewport units (usually unnecessary; = 1)
    legendWidth *= svg.getUnitsPerInchY();
    // The legend height will be as much of the page as possible
    double legendHeight = (svg.getViewportHeightInInches() - MARGIN * 2) * svg.getUnitsPerInchY();
    
    // A Legend is an SVG Representable object
    Legend L = new Legend(legendWidth * svg.getUnitsPerInchX(), legendHeight * svg.getUnitsPerInchY());
    svg.addFramedViewport(legendLeft, legendTop,               // Positioned at the left margin, top margin
                   legendWidth, legendHeight,                  // is X page inches wide, Y page inches high
                   L.getViewportTopLeftX(), L.getViewportTopLeftY(),   // Internal coordinates
                   L.getViewportWidth(),    L.getViewportHeight(),              
                   .1, "green",                                // Border width and color
                   new SVGRepresentable[] {L}, "Legend" );     // Internal comment in SVG
    

    // Space
    
    double mapWidthInInches  = svg.getViewportWidthInInches() - MARGIN * 2d - legendWidth;
    double mapHeightInInches = legendHeight;
    
    // Space is also an SVG representable object; it already exists
    // but needs the correct window. The window must be scaled to show
    // only the ranges specified for display.
    double wBound = space.getDisplayWesternBoundary();
    double eBound = space.getDisplayEasternBoundary();
    double nBound = space.getDisplayNorthernBoundary();
    double sBound = space.getDisplaySouthernBoundary();
    
    // Must rescale these to match the map area aspect ratio; an unfun set of calculations, but one
    // that guarantees that the requested portion of the map can appear and will not be distorted
    double mapXRange = Math.abs(eBound - wBound);
    double mapYRange = Math.abs(sBound - nBound);
    
    double xRatio = mapXRange / mapWidthInInches;
    double yRatio = mapYRange / mapHeightInInches;
    
    double ratio = (xRatio > yRatio ? xRatio : yRatio);
          
    double mapXCenter = (wBound + eBound)/2d;
    double mapYCenter = (nBound + sBound)/2d;
              
    wBound = mapXCenter + ((wBound - mapXCenter)/(Math.abs(wBound - mapXCenter))) * (ratio * mapWidthInInches  / 2d);
    nBound = mapYCenter + ((nBound - mapYCenter)/(Math.abs(nBound - mapYCenter))) * (ratio * mapHeightInInches / 2d);
    eBound = mapXCenter + ((eBound - mapXCenter)/(Math.abs(eBound - mapXCenter))) * (ratio * mapWidthInInches  / 2d);
    sBound = mapYCenter + ((sBound - mapYCenter)/(Math.abs(sBound - mapYCenter))) * (ratio * mapHeightInInches / 2d);
            
    
    // Tell the space object these are the boundaries (primarily for gridlines)
    double origXMin = space.getDisplayWesternBoundary();
    double origYMin = space.getDisplayNorthernBoundary();
    double origXMax = space.getDisplayEasternBoundary();
    double origYMax = space.getDisplaySouthernBoundary();
    
    space.setDisplayRange(wBound, nBound, eBound, sBound);
    
    // Keep in mind that this method will perform the re-orientation needed if min > max
    svg.addFramedViewport(MARGIN + legendWidth, MARGIN,   // Positioned adjacent to legend, at top margin
          svg.getViewportWidthInInches() - MARGIN * 2d - legendWidth,  // width is what's left after legend 
          legendHeight,                                   // height is same as legend
          wBound, nBound,                                      // Top Left
          eBound, sBound,                                      // Bottom Right
          .1, "green",
          new SVGRepresentable[] {space}, "Space", "", "id=\"map\"");    
    
    // Set the space's display range back to the user-specified values
    space.setDisplayRange(origXMin, origYMin, origXMax, origYMax);    
    
    return;    
  }
  
  /* ************************************************************************** */
  /* X3D                                                                        */
  /* ************************************************************************** */
  /** 
   * @see abcm.graphics.X3DRepresentable#addToX3D(abcm.graphics.X3DAssist) 
   */
  public void addToX3D(X3DAssist x3dAssist){
    getSpace().addToX3D(x3dAssist);
  }
  
}
