package abcm.model;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.views.ABCMParameterAppliesToCodeModelDomainsView;
import abcm.db.managers.AnalysisManager;
import abcm.db.managers.CommandsManager;
import abcm.db.managers.HistoryManager;
import abcm.db.managers.NarrativeManager;
import abcm.db.managers.ParameterCategoriesManager;
import abcm.db.managers.ParameterManager;
import abcm.db.managers.ParameterSetManager;
import abcm.db.managers.ProbeManager;
import abcm.db.managers.ProbeSetManager;
import abcm.db.managers.SimulationDataSeriesManager;
import abcm.db.managers.SimulationManager;
import abcm.db.managers.SimulationRunManager;
import abcm.db.managers.TruthValueManager;
import abcm.db.managers.stubs.AbstractConfigurationManager;
import abcm.graphics.SVGRepresentable;
import abcm.graphics.X3DRepresentable;
import abcm.math.BasicRNG;
import abcm.math.SimDate;
import abcm.sim.ABCMGenericGUI;
import abcm.sim.AbstractModelComponent;
import abcm.sim.Analysis;
import abcm.sim.CodeModel;
import abcm.sim.HistoryInstruction;
import abcm.sim.ParameterSet;
import abcm.sim.SimulationDataSeries;
import abcm.sim.SimulationRun;
import abcm.space.Space;
import abcm.utilities.ScriptReader;
import abcm.web.ABCMSession;
import abcm.web.SVGViewer;
import abcm.web.X3DViewer;


/**
 * Provides the container context in which ABCM Models run.
 *
 * The most basic function of the model environment is to provide
 * a portal to the subclasses from which the model itself is made.
 * For example, when creating an ABCM model one of the requirements
 * is that a new subclass of the <code>AbstractDBGateway</code> object
 * is created and endowed with the ability to contact the specified
 * database. The model environment must provide programmatic access
 * to that new subcless; hence, when extending the AbstractABCModelEnvironment
 * object, the 'getDBGateway()' method should return an instance of the
 * new DBGateway subclass. The same applies for models and the extended
 * ConfigurationManager.
 * 
 * Additionally, the ModelEnvironment provides a portal to file locations-
 * where scripts are found, images and transcripts written, etc.
 * 
 * Also, the ModelEnvironment provides some functionality for running a model:
 * specifically it provides a way to store analyses that include data from runs
 * retrieved from the database and not related to any given model.
 * 
 * Finally, the ModelEnvironment must provide hooks for the GUI to know what
 * to display.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractABCModelEnvironment {

  /* Static variables */
  
  /** Extension used to identify script files (and demos) */
  public static final String SCRIPT_EXTENSION             = "_abcm";
  /** Extension used to identify narrative files */
  public static final String NARRATIVE_EXTENSION          = "_narr";
  /** Extension used to identify compiled narrative files */
  public static final String COMPILED_NARRATIVE_EXTENSION = "_naru";
  /** Extension used to identify ABCM transcripts */
  private static String TRANSCRIPT_SUFFIX                 = ".abcm.txt";

  /** Maximum length of a transcript segment */
  protected static int SEGMENT_LENGTH = 16384;
  
  /** Instance variables */
  
  /** Represents the mode in which the environment is being run (i.e. interactive, batch)  */
  protected int                   mode;
 
  /** Model contained within this environment  */
  protected AbstractABCModel      m       = null;
  /** Session within which this environment exists  */
  protected ABCMSession           session = null;
  
  /** Collection of analyses active in this environment  */
  protected Hashtable             analyses = new Hashtable();
  
  /** Name of the demo currently being run in this environment  */
  protected String                demoName;
  /** Current step of the demo being run in this environment  */
  protected int                   demoStep;


  /** The last (previous) command executed  */
  private String     lastCommand               = null;
  
  /** The response to the previous command  */
  private StringWriter     lastInterchange           = new StringWriter();
  /** Transcript of commands and responses in this session  */
  private StringWriter     transcript                = new StringWriter();
  /** The last result obtained  */
  private ABCMResult lastResult                = null;

  /**  Used to manage active parameters in this environment */
  private ParameterManager pm;

  /** Collection of models at 'marked' paused positions */
  protected Vector    marks = new Vector();
  
  /** Collection of models at 'marked' paused positions, indexed by assigned name */
  protected Hashtable marksByName = new Hashtable();
  
  /** The original model (regardless of any marks set) */
  protected AbstractABCModel original = null;
  
  
  /* Inner class: StoredModel */
  
  /**
   * Utility class to stoore a copy of a model, capturing its instance variables.
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class StoredModel{
    AbstractABCModel  model;
    ParameterSet      P;
    CodeModel         C;
    BasicRNG          RNG;
    
    /** Constructor
     * @param m           The model itself
     * @param paramSet    The current parameter set
     * @param codeModel   The current code model
     * @param rng         The current Basic Random Number Generator
     */
    public StoredModel(AbstractABCModel m, ParameterSet paramSet, CodeModel codeModel, BasicRNG rng){
      model = m;
      P     = paramSet;
      C     = codeModel;
      RNG   = rng;
    }
  }
  
  /* Inner class - Script Processor */
  
  
  /**
   * Processes a script file by passing the commands in it to the parser
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class ScriptProcessor{
  
    /** Flag indicating whether the script being processed is a narrative  */
    private boolean isNarr = false;
    
    /** Utility flag */
    public static final boolean SCRIPT_IS_NARRATIVE     = true;
    
    /** Utility flag */
    public static final boolean SCRIPT_IS_NOT_NARRATIVE = false;
    
    /**
     * Constructor that is informed that it exists in a given environment
     * and will be used to process a narrative
     * @param scriptIsNarrative flag indicating that the script is a narrative
     */
    public ScriptProcessor(boolean scriptIsNarrative){
      isNarr = scriptIsNarrative;
    }
    
    
    /**
     * Processes a script
     * @param fileName the file name of the script to be processed
     * @return a String containing the output of the execuution
     */
    public String process(String fileName){
      // Get the script via a script reader
      ScriptReader sr  = new ScriptReader(fileName);    
      if(!sr.getSuccessFlag()) return sr.getErrorReport();
  
      StringWriter out = new StringWriter();
      while(!sr.EOF()){
        String nextLine = sr.getNextLine();
        if(nextLine!= null && nextLine.length() > 0){
          if(isNarr){
            if(nextLine.length() > 15) nextLine = nextLine.substring(15);  // Skip past dates in .nar files
          }
          out.write(">" + nextLine + "\n");
          String output = processCommand(nextLine).text; 
          out.write(output);
          if(output.length() > 0 ) out.write("\n");
        }
      }
      sr.closeFile();
      return out.toString();
    }
  }
    
  
  /* Constructor */
  
  /**
   * Constructor
   * @param modeID numeric representation of the mode in which this environment exists
   */
  protected AbstractABCModelEnvironment(int modeID){    
    mode = modeID;
  }
  

  /**
   * Constructor
   * @param S Session within which this environment exists
   * @param modeID numeric representation of the mode in which this environment exists
   */
  protected AbstractABCModelEnvironment(ABCMSession S, int modeID){
    session = S;
    mode    = modeID;
  }
  
  /* Fundamentals */
  
  // Any basic object that extends this should implement these to provide gateways
  // to the new classes that will be used in the new modeling context.
  
  
  /**
   * Gets the DB Gateway being used by this environment
   * @return the environment's DB gateway
   */
  public    abstract AbstractABCMDBGateway                   getDBGateway();

  /**
   * Gets the parser being used by this environment
   * @return the environment's model
   */
  public    abstract AbstractABCMParser                         getParser();

  /**
   * Gets the configuration manager being used by this environment
   * @return the environment's configuration manager
   */
  public    abstract AbstractConfigurationManager getConfigurationManager();  
  
  /**
   * Gets the commands manager being used by this environment
   * @return the environment's commands manager
   */
  public    abstract CommandsManager                   getCommandsManager();
  
  /**
   * Gets a new model in this environment
   * @return a new model object created within this environment
   */
  protected abstract AbstractABCModel                            newModel();

  /**
   * Gets a new model in this environment to be associated with the specified simulation run
   * @param sr simulation run with which the new model is to be linked
   * @return the newly created model
   */
  protected abstract AbstractABCModel                            newModel(SimulationRun sr);
  
  
  /* File Management */
  
  // Usually a model environment needs to know where to put and find files
  // This will vary based on the environment
  // Usually these will be defined not by the immediate subclass but by the classes below that,
  // which will represent different implementations of the same ABC Modeling environment
  
  
  /**
   * Gets the directory in which XSL stylesheets are found
   * @return the XSL stylesheet directory
   */
  public abstract String getXSLRootDir();
  
  /**
   * Gets the directory to which images should be written
   * @return the image directory
   */
  public abstract String getImageOutputLocation();
  
  /**
   * Gets the prefix to be added to URLs of images written out
   * @return the URL prefix for images
   */
  public abstract String getImageURLPrefix(); 
  
  /**
   * Gets the directory in which scripts are found
   * @return the script directory
   */
  public abstract String getScriptDir(); 
  
  /**
   * Gets the directory to which transcripts are to be written
   * @return the transcript directory
   */
  public abstract String getTranscriptLocation();
  
  /**
   * Gets the prefix to be added to URLS of transcripts
   * @return the URL prefix for transcripts
   */
  public abstract String getTranscriptURLPrefix();
  
  /**
   * Returns true if the specified script file exists, false otherwise. Looks in the script directory
   * obtained by getScriptDir();
   * @param fileName script file sought
   * @return true if the file is present, false otherwise
   */
  public final boolean scriptFileExists(String fileName){
    File F = new File(getScriptDir() + fileName);
    return F.exists();
  }
  
  /**
   * Calls getClosestScriptFile(filename, preferNarrative) using
   * 'false' as the argument to preferNarrative.
   * @param fileName file sought
   * @return result of search
   */
  public final ABCMResult getClosestScriptFileName(String fileName){
    return getClosestScriptFileName(fileName, false);
  }
  
  /**
   * Gets the name of a script file that is 'close to' the given name. The given
   * name may lack an extension, for example; this method will find the script file
   * that has the same base name; the preferNarrative flag indicates which file
   * to return if both a narrative and a script file exist.
   * @param fileName File sought
   * @param preferNarrative true if the narrative version of the file is to be returned
   * in preference to the script version, false otherwise
   * @return the name of the script file, if one is found; an unsuccessful result otherwise
   */
  public final ABCMResult getClosestScriptFileName(String fileName, boolean preferNarrative){
    String s = null;
    if(scriptFileExists(fileName))                             s = fileName;
    else{
      if(!preferNarrative){
        if(scriptFileExists(fileName + getScriptExtension()))          s = fileName + getScriptExtension();
        else if(scriptFileExists(fileName + getNarrativeExtension()))  s = fileName + getNarrativeExtension();
      }
      else{
        if(scriptFileExists(fileName + getNarrativeExtension()))       s = fileName + getNarrativeExtension();
        else if(scriptFileExists(fileName + getScriptExtension()))     s = fileName + getScriptExtension();
      }
    }
    if(s == null) return new ABCMResult("No file matching name '" + fileName + "' found: ", false);
    else          return new ABCMResult(s, true);
  }
  
  // Any model environment needs to know how to get user IDs
  // Usually these will be defined not by the immediate subclass but by the classes below that,
  // making it implementation specific rather than model-type specific
  
  /**
   * Gets the ID used by the Admin login to the database
   * @return the Admin ID
   */
  public abstract String getAdminDBID();
  
  /**
   * Gets the ID used by the Guest login to the database
   * @return the Guest ID
   */
  public abstract String getGuestDBID();
  
  /**
   * Gets the password used by the Admin login to the database
   * @return the Admin password
   */
  public abstract String getAdminDBPWD();
  
  /**
   * Gets the password used by the Guest login to the database
   * @return the Guest password
   */  
  public abstract String getGuestDBPWD();
  
  
  /**
   * Implementation can define a specific script file extension infix; for example,
   * if the implementation is 'Hohokam Water Management', and the infix is 'hwm',
   * scripts, narratives, and compiled narratives will be named:
   *
   * MyScript.hwm.abcm
   * MyScript.hwm.nar
   * MyScript.hwm.nru
   * @return infix to use within file names  
   */ 
  protected abstract String getScriptFileExtensionInfix();
  
  /**
   * Gets the extension to be added to script files. Includes the infix specified by getScriptFileExtensionInfix()
   * @return the script file extension
   */
  public String getScriptExtension(){            return "." + getScriptFileExtensionInfix() + SCRIPT_EXTENSION;             }
  
  /**
   * Gets the extension to be added to narrative files. Includes the infix specified by getScriptFileExtensionInfix()
   * @return the narrative file extension
   */
  public String getNarrativeExtension(){         return "." + getScriptFileExtensionInfix() + NARRATIVE_EXTENSION;          }
  
  /**
   * Gets the extension to be added to compiled narrative files. Includes the infix specified by getScriptFileExtensionInfix()
   * @return the compiled narrative file extension
   */
  public String getCompiledNarrativeExtension(){ return "." + getScriptFileExtensionInfix() + COMPILED_NARRATIVE_EXTENSION; }
  
  /* Model Creation */
  
  /**
   * Creates a model in this environment. Uses the current mode. Will fail if the specified
   * simulation, parameter set, and probe set cannot be loaded.
   * @param simName name of the ABCM Simulation to be loaded
   * @param paramSet name of the ABCM Parameter Set to be loaded
   * @param probeSet name of the ABCM Probe Set to be loaded
   * @param notes any notes about this model
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult createModel(String simName, String paramSet, String probeSet, String notes){
    return createModel(simName, paramSet, probeSet, notes, this.getMode());
  }
  
  /**
   * Creates a model in this environment. Will fail if the specified
   * simulation, parameter set, and probe set cannot be loaded.
   * @param simName name of the ABCM Simulation to be loaded
   * @param paramSet name of the ABCM Parameter Set to be loaded
   * @param probeSet name of the ABCM Probe Set to be loaded
   * @param notes any notes about this model
   * @param currentMode mode in which this environment exists
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult createModel(String simName, String paramSet, String probeSet, String notes, int currentMode){

    if(!isAuthorized()) return new ABCMResult("You must log in before you create a model.", false);
    
    SimulationManager sm = new SimulationManager(getDBGateway());
    long simID = sm.getPK(simName);
    if(simID == -1){
      return new ABCMResult("Could not locate Simulation " + simName, false);
    }
    
    ParameterSetManager psm = new ParameterSetManager(getDBGateway());
    long paramSetID = psm.getPK(paramSet);
    if(paramSetID == -1) {
      return new ABCMResult("Could not locate Parameter Set " + paramSet, false);
    }
    
    ProbeSetManager pbsm = new ProbeSetManager(getDBGateway());
    long probeSetID = pbsm.getPK(probeSet);
    if(probeSetID == -1) {
      return new ABCMResult("Could not locate Probe Set " + probeSet, false);
    }
    ABCMResult retVal = createModel(simID, paramSetID, probeSetID, notes, currentMode);
    if(getModel() != null) extendTranscript("Create Model: " + getRunNumber());
    return retVal; 
  }
  
  /**
   * Creates a model in this environment. Uses the current mode. Will fail if the specified
   * simulation, parameter set, and probe set cannot be loaded.
   * @param simID PK in the database of the ABCM Simulation to be loaded
   * @param paramSetID PK in the database of the ABCM Parameter Set to be loaded
   * @param probeSetID PK in the database of the ABCM Probe Set to be loaded
   * @param notes any notes about this model
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult createModel(long simID, long paramSetID, long probeSetID, String notes){
    return createModel(simID, paramSetID, probeSetID, notes, this.getMode());
  }
  
  /** 
   * Creates a model in this environment. Will fail if the specified
   * simulation, parameter set, and probe set cannot be loaded.
   * @param simID PK in the database of the ABCM Simulation to be loaded
   * @param paramSetID PK in the database of the ABCM Parameter Set to be loaded
   * @param probeSetID PK in the database of the ABCM Probe Set to be loaded
   * @param notes any notes about this model
   * @param currentMode mode in which this environment exists
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult createModel(long simID, long paramSetID, long probeSetID, String notes, int currentMode){
    if(!isAuthorized()) return new ABCMResult("You must log in before you create a model.", false);
    
    long RNS = (long)Math.floor(Math.random()*100000000);
        
    SimulationRunManager srm = new SimulationRunManager(getDBGateway());
    SimulationRun sr = srm.createNewSimulationRun(getUserID(), 
          simID, paramSetID, probeSetID, currentMode, 
          notes, RNS);  
    m = newModel(sr);
    System.out.println(m.modelStatus());
    if(m == null) return new ABCMResult("Failed to create model.", false);
    marksByName.clear();
    marks.clear();
    marks.add(m);
    ABCMCloneStore cloneStore = new ABCMCloneStore();
    marksByName.put("ORIGINAL", new StoredModel(m, 
                                                AbstractModelComponent.getPClone(cloneStore), 
                                                AbstractModelComponent.getCMClone(cloneStore),
                                                AbstractModelComponent.getCurrentRNG()));
    return new ABCMResult("Successfully created a new model. Model run number is " + m.getRunNumber(), true);
  }
  
  /**
   * Creates a test model. Test models cannot be saved as they are assumed to have no associated
   * initial simulation data.
   * @return a message indicating the success or failure of this operation. No test model can be created
   * if a model or test model already exist.
   */
  public final ABCMResult createTestModel(){
    if(m != null){
      if(this.modelIsTestModel()) return new ABCMResult("Already in test model; use 'DisposeAndReset' to clear model", false);
      else                        return new ABCMResult("A model is being used; use 'DisposeAndReset' to clear model", false); 
    }
    clearModel();
    m = newModel();
    if(m == null) return new ABCMResult("Failed to create test model.", false);
    m.loadParameterSet("DEFAULT");
    return new ABCMResult("Test model created. Results from a test model may not be saved.\n", true);
  }


  /* Branching Models */
  /**
   * Marks the current model and creates a clone; stores the original
   * under the assigned name for later retrieval. 
   * @param markName name to be used for retrieval; if the name is
   * already in use to store a model, will fail
   * @return an ABCM Result indicating the success or failure of this operation
   */
  public final ABCMResult markAndCloneWithNewRNG(String markName){
    if(marksByName.containsKey(markName)) return new ABCMResult("Could not mark model by the name '" + markName+ "'; a mark of that name already exists.", false);
    m.checkIntegrity();
    ABCMCloneStore cloneStore = new ABCMCloneStore();
    AbstractABCModel newModel = m.getModelClone(cloneStore);
    marks.add(m);
    StoredModel SM = new StoredModel(m, AbstractModelComponent.getPClone(cloneStore), 
                                        AbstractModelComponent.getCMClone(cloneStore), 
                                        AbstractModelComponent.getCurrentRNG());
    marksByName.put(markName, SM);
    AbstractModelComponent.setModelComponent(null, null, new BasicRNG((int)SM.P.getParameter("RandomNumberGeneratorType"), (long)SM.RNG.nextDouble() * Long.MAX_VALUE));
    if(original == null) original = m;
    m = newModel;
    return new ABCMResult("Now using new model; mark at '" + markName + "' is saved.", true);
  }
  
  
  /**
   * Deletes any branched models subsequent to the one specified by name, and
   * returns to the marked model specified.
   * @param markName name of branch to be restored. If no matching name is found,
   * aborts
   * @return an ABCMResult indicating the success or failure of this operation
   */
  public final ABCMResult restoreToMark(String markName){
    if(!marksByName.containsKey(markName)) return new ABCMResult("Could not restore to mark; no mark '" + markName + "' is stored.", false);
    StoredModel SM = (StoredModel)marksByName.get(markName);
    m = SM.model;
    AbstractModelComponent.setModelComponent(SM.C, SM.P, SM.RNG); // Put these back to stored versions
    // Clear back to marked in 'marks' vector and all related marks in 'marksByName'
    for(int i = marks.size() - 1; i > marks.indexOf(m); i--){
      AbstractABCModel modelToRemove = (AbstractABCModel)marks.remove(i);
      Enumeration E = marksByName.keys();
      while(E.hasMoreElements()){
        String key = (String)E.nextElement();
        if(((StoredModel)marksByName.get(key)).model.equals(modelToRemove)){
          marksByName.remove(key);
          break;
        }
      }      
    }
    return new ABCMResult("Model is restored to mark '" + markName +"'.", true);
  }
  
  
  /**
   * Clears all previously saved marks and considers the current branch to be
   * the originalmodel.
   * @return an ABCMResult indicating success or failure of this operation
   */
  public final ABCMResult clearAllMarksAndAcceptCurrentBranch(){
    if(marks.size() == 1) return new ABCMResult("Current branch is only branch.", false);
    m.runID = original.runID;
    original = m;
    marksByName.clear();
    marks.clear();
    return new ABCMResult("Current branch is now considered main branch.", true);
  }
  

  
  /* Model Methods */
  
  /**
   * Returns true if there is a model present in this environment.
   * @return true if a model is present
   */
  public final boolean hasModel(){ return m != null;  }
  
  
  /**
   * Returns true if the model is a test model. Test models have -1 as their run
   * number. Data from test models cannot be saved. It is important to note, however,
   * that if this method is executed in an HWMEnvironment that has no model (model
   * is null), the result will be true rather than false; this is because it is
   * better to fail in a way that stops inappropriate data transactions than in
   * the way that permits them.
   * @return true if the model is a test model or no model exists, false otherwise
   */
  public final boolean modelIsTestModel(){
    if(m == null)              return true; // neither is correct, but better to err by saying it's a test model
    if(m.getRunNumber() == -1) return true;
    else                       return false;
  }

  
  /**
   * Gets a value indicating whether the resident model is reproducible
   * or not. Returns false if there is no model.
   * @return true if the model is reproducible from data in the
   * database, false if not (i.e. if the user has entered additional,
   * unrecorded commands
   */
  public final boolean modelIsReproducible(){
    if(m == null) return false;
    return m.modelIntegrityIsIntact();
  }  

  
  /* Model Discard */
  
  /**
   * This can be overloaded to allow more complex operations when a model is to be disposed
   * @return a message indicating the success or failure of this operation.
   */
  public ABCMResult disposeAndReset(){
    if(getGUI() != null){    getGUI().disposeAndReset(); }
    clearModel();
    analyses = new Hashtable();
    lastInterchange.getBuffer().setLength(0);
    stopDemo(demoName);
    return new ABCMResult("Model environment is reset.", true);
  }
  
  
  /**
   * Clears the model out of this Environment object
   * Also clears marks out of branches
   */
  public final void clearModel(){
    this.marks.clear();
    this.marksByName.clear();
    m = null;    
 }
  
  
  /* Get methods */

  
  /**
   * Gets the model contained by this environment
   * @return the model contained by this environment
   */
  public AbstractABCModel getModel() { return m; }
  
  /**
   * Returns true if this environment contains a marked model with
   * the specified mark name
   * @param markName name sought
   * @return true if a matching name is found, false otherwise
   */
  public boolean hasModel(String markName){
    return marksByName.containsKey(markName);
  }
  
  /**
   * Gets the model with the specified branch point name
   * @param markName name sought
   * @return model with that name
   */
  public AbstractABCModel getModelFromBranchPoint(String markName){
    return ((StoredModel)marksByName.get(markName)).model;
  }
  
  /**
   * Gets the run number of the current Model's Simulation Run
   * Returns -1 if there is no current Model.
   * @return the run number of the current Model's Simulation Run
   */
  public final long getRunNumber(){ return (m != null ? m.getRunNumber() : -1);   }
  
  
  /**
   * Gets the mode in which this Environment exists.
   * @return the mode in which this Environment exists.
   */
  public final int getMode(){             return mode;                }
 
  
  /**
   * Gets the Space object associated with the Model in this HWM Environment
   * @return the model object's Space object
   */
  public final Space getSpace(){          return m.getSpace();        }
  
  
  /* Execute parse methods */
  
  /**
   * Causes the parser to execute the command provided
   * @param command the command to be executed
   * @return the results of executing the command
   */
  public final ABCMResult processCommand(String command){
    return processCommand(command, false);
  }
  
  
  /**
   * Causes the parser to execute the command provided
   * @param c the primary key of the command to be executed
   * @return the results of executing the command
   */
  public final ABCMResult processCommand(int c){
    String[] argList = null;
    return processCommand(c, argList);
  }
  
  
  /**
   * Causes the parser to execute the command provided
   * @param c the command to be executed
   * @param argList list of arguments to the command
   * @return the results of executing the command
   */
  public final ABCMResult processCommand(int c, String[] argList){
    return getParser().processCommand(c, argList);
  }
  
  
  /**
   * Causes the parser to execute the History Instruction provided
   * @param HI the History Instruction to be executed
   * @return the results of executing the command
   */
  public final ABCMResult processCommand(HistoryInstruction HI){
    return getParser().processCommand(HI);
  }
 
  
  /**
   * Processes a command and, if requested, stores that command in the transcript.
   * @param command the command to be processed
   * @param includeInTranscript true if the command should be stored in the transcript
   * @return a String representing the results of the command
   */
  public ABCMResult processCommand(String command, boolean includeInTranscript){
    if(command == null || command.length() == 0) 
      if(lastCommand == null) return new ABCMResult("No command.", false);
      else command = lastCommand;
    lastCommand = command;
    lastResult = getParser().processCommand(command);
    lastInterchange.getBuffer().setLength(0); // Do this after 'processCommand' because processing the command may involve running this method!
    if(command.startsWith("**")){
      command = formatComment(command);
      transcript.write("> " + command + "\n");
      lastInterchange.write("> " + command + "\n");
    }
    else {
      if(includeInTranscript){
        transcript.write("> ");
        transcript.write(command);
        transcript.write("\n");
        transcript.write(lastResult.text);
        transcript.write("\n\n");
      }
      lastInterchange.write("> <b>");
      lastInterchange.write(command);
      lastInterchange.write("</b>\n");
      lastInterchange.write(lastResult.text);  
    }
    extendTranscript(transcript.toString());
    transcript.getBuffer().setLength(0);
    return lastResult;
  }  

  
  
  
  /**
   * Comments can be formatted in an implementation-specific way by a subclass
   * @param comment comment to be formatted
   * @return re-formatted comment
   */
  protected abstract String formatComment(String comment);

  /**
   * Invokes a script processor.
   * @param fileName script to be processed
   * @return output of processing operation
   */
  public String processScript(String fileName){
    return processScript(fileName, ScriptProcessor.SCRIPT_IS_NOT_NARRATIVE);
  }
    
  /**
   * Invokes a script processor
   * @param fileName script to be procesed
   * @param scriptIsNarrative true if the script is actually a narrative, false otherwise
   * @return output of processing operation
   */
  public String processScript(String fileName, boolean scriptIsNarrative){
    ABCMResult file = this.getClosestScriptFileName(fileName);
    if(!file.success) return file.text;
    ScriptProcessor sp = new ScriptProcessor(scriptIsNarrative);
    return sp.process(this.getScriptDir() + file.text);
  }
  
  /**
   * Gets the result of the preceding operation
   * @return Gets the result of the preceding operation
   */
  public ABCMResult getLastResult(){ return lastResult; }
  
  
  /* Transcript Management */
 
  /**
   * Gets the file name of the current transcript file.
   * File name does not include location.
   * @return the current transcript file name
   */
  private String getTranscriptFileName(){
    return getTranscriptFileName(getUserLogin());
  }
  
  /**
   * Gets the transcript file name for the specified user
   * @param user the user. File name does not include location
   * @return the file name
   */
  private String getTranscriptFileName(String user){
    return user + "_" + new SimpleDateFormat("yyyy.MM.dd").format(new Date()) + TRANSCRIPT_SUFFIX;
  }
  
  /**
   * Gets the file name of the file used to store transcript
   * bookmarks for the current user. File name does not
   * include location.
   * @return the name of the bookmark file
   */
  private String getBookmarkFileName(){
    return getBookmarkFileName(getUserLogin());
  }
  
  /**
   * Gets the file name of the file used to store
   * transcript bookmarks for the specified user.
   * File name does not include location.
   * @param user the user
   * @return the file name
   */
  private String getBookmarkFileName(String user){
    return user + TRANSCRIPT_SUFFIX;
  }
  
  
  /**
   * Extends the transcript by saving the specified text
   * @param text
   */
  public boolean extendTranscript(String text){
    return extendTranscriptFile(getTranscriptFileName(), text);
  }
  
  
  /**
   * Extends the specified file by appending the specified text
   * @param fileName
   * @param text
   */
  private boolean extendTranscriptFile(String fileName, String text){
    if(text.length() > 0 ){
      try{
        FileWriter fw = new FileWriter(
            new File(getTranscriptLocation() + fileName), 
            true);
        fw.write(text);
        fw.close();
        return true;
      }
      catch(IOException ioe){
        System.err.println("An error occurred while extending the transcript: " + ioe.getMessage());
        return false;
      } 
    }
    return false;
  }
  
  /**
   * Writes a target html tag to the transcript and
   * a link to it in the bookmarks file for the current user
   * @param bookmark text to be added
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult bookmarkTranscript(String bookmark){
    if(bookmark.length() > 0){
      if(!extendTranscriptFile(getBookmarkFileName(), 
            getTranscriptFileName() + ": " + bookmark + " (" + new SimpleDateFormat("yyyy.MM.DD hh:mm a").format(new Date()) + ")\n"))
          return new ABCMResult("Some error occurred while writing to the bookmark file; bookmark not saved", false);
      if(!extendTranscript("\n\nBOOKMARK: " + bookmark +"\n\n")) 
          return new ABCMResult("Some error occurred while writing to transcript; bookmark was saved in bookmark file but may not appear in transcript.", false);
      return new ABCMResult("Bookmarked Transcript.", true);
    }
    return new ABCMResult("No bookmark provided.", false);
  }
  
  /**
   * Gets a list of the existing transcript files for the current user
   * @return the list of files
   */
  public Vector<String> getTranscriptList(){
    return getTranscriptList(getUserLogin());
  }
  
  /**
   * Gets a list of existing transcript files for the specified user
   * @param username the name of the user
   * @return the list of files
   */
  public Vector<String> getTranscriptList(String username){
    Vector<String> ret = new Vector<String>();
    try{
      File F = new File(getTranscriptLocation());
      if(!F.isDirectory()) return ret; // Should never happen
      File[] flist = F.listFiles();
      for(File f: flist){
        if(f.getName().startsWith(username) && f.getName().compareTo(getBookmarkFileName()) != 0) ret.add(f.getName());
      }
    }
    catch(Exception E){
       System.out.println("An error occurred getting list of transcripts for user '" + username + "'");
    }
    return ret;
  }

  /**
   * Gets the contents of the bookmark file for the current user
   * @return the contents of the bookmark file
   */
  public String getBookmarkList(){
    return getBookmarkList(getUserLogin());
  }
  
  /**
   * Gets the contents of the bookmark file for the specified user
   * @param user the user
   * @return the contents of the bookmark file
   */
  public String getBookmarkList(String user){
    StringWriter ret = new StringWriter();
    try{
      FileReader fr = new FileReader(new File(getTranscriptLocation() + getBookmarkFileName(user)));
      
      char[] i = new char[1];
      while(fr.ready()){        
        fr.read(i);
        ret.write(i[0]); 
      }
      fr.close();
    }
    catch(Exception E){
      System.out.println("An error occurred getting bookmark list for user '" + user + ".");
      return "";
    }
    return ret.toString();
  }
  
  /**
   * Gets the list of bookmarks in an HTML format for the current user
   * @return the bookmarks
   */
  public String getHTMLFormatBookmarkList(){
    return getHTMLFormatBookmarkList(getUserLogin());
  }
  
  /**
   * Gets the list of bookmarks in an HTML format
   * @param user the specified user
   * @return the bookmarks
   */
  public String getHTMLFormatBookmarkList(String user){
    String source = getBookmarkList(user);
    if(source.length() == 0) return "";
    StringWriter ret = new StringWriter();
    String[] lines = source.split("\n");
    for(String l: lines){
      if(l.length() > 0){
        String file = l.substring(0, l.indexOf(":"));
        ret.write("<a target=\"_blank\" href=\"" + getTranscriptURLPrefix() + file + "\">" + file + "</a>" + l.substring(l.indexOf(":")) + "\n");
      }
    }
    return ret.toString();
  }
  
  /**
   * Gets a String that contains the last command the last response to that
   * command.
   * @return the last command and the corresponding response.
   */
  public String getLastInterchange(){    return (lastInterchange.toString() != null ? lastInterchange.toString() : "");  }
  

  /* GUI Methods */  
  
  /**
   * Gets the GUI contained in this Environment
   * @return the Environment's GUI
   */
  public abstract ABCMGenericGUI getGUI();
 
  
  /**
   * Returns whether this Environment has a GUI
   * @return true if the Environment has a GUI, false if not
   */
  public final boolean hasGUI(){    return (getGUI() != null);  }
    
  
  /**
   * Updates the GUI
   */
  public final void updateGUI(){    if (getGUI() != null) getGUI().update();    }
  

  /**
   * Creates a data series graph with the specified name. Will fail if a graph with that
   * name already exists.
   * @param name name to be assigned to the new series graph
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult createDataSeriesGraph(String name){ 
    if(analyses.containsKey(name)) return new ABCMResult("Could not create analysis; a graph with that name already exists.", false);
    else{
      Analysis analysis = new Analysis();
      analyses.put(name, analysis);
    }
    return new ABCMResult("New analysis graph created", true); 
  }   
  
  
  /**
   * Loads a data series from the ABCM database and inserts it into a data series graph
   * @param runNumber run number of the data series to be extracted
   * @param seriesName name of the data series to be extracted
   * @param graphName graph to which the data series is to be added
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult loadDataSeriesFromDB(long runNumber, String seriesName, String graphName){
    if(!analyses.containsKey(graphName)) return new ABCMResult("No graph by that name has been created.", false);
    Analysis analysis = (Analysis)analyses.get(graphName);
    SimulationDataSeriesManager sdsm = new SimulationDataSeriesManager(getDBGateway());
    SimulationDataSeries sds = sdsm.getDataSeriesFromDB(runNumber, seriesName);
    if(sds == null) return new ABCMResult("Could not retrieve data series with given run number and name.", false);
    analysis.addSeries(sds, runNumber + "-" + seriesName);
    return new ABCMResult("Loaded data series " + runNumber + "-" + seriesName + " into " + graphName, true);
  }
  
  
  /**
   * Adds the specified data series to the given graph
   * @param graphName name of the graph to which the series is to be added
   * @param sds the simulation data series to be added
   * @param seriesName name to be assigned to that series in the graph
   * @return a message indicating the success or failure of the operation
   */
  public final ABCMResult addSeriesToGraph(String graphName, SimulationDataSeries sds, String seriesName){
    if(!analyses.containsKey(graphName)) return new ABCMResult("No graph by that name has been created.", false);
    Analysis analysis = (Analysis)analyses.get(graphName);
    if(analysis.addSeries(sds, seriesName)) return new ABCMResult("Successfully added '" + seriesName + "' to '" + graphName + "'", true);
    else return new ABCMResult("Could not add series to graph; possibly series is non-numeric, or series with same name may already exist", false);
  }
  
  /**
   * Implementing classes should determine how to display a data graph
   * @param name name of the graph to be displayed
   * @return some message indicating success or failure
   */
  public abstract String showDataGraph(String name);
  
  /**
   * Gets a list of the graphs available in this environment; implementation-specific.
   * @return a list of graphs
   */
  public abstract String listGraphs();
  
  
  /**
   * Returns, for a given graph, the number of Data Series in the graph.
   * @param graphName the name of the graph
   * @return the number of Data Series in that graph
   */
  public final long numberOfSeriesInGraph(String graphName){
    if(analyses.containsKey(graphName)) return ((Analysis)analyses.get(graphName)).getSeries().size();
    else                                return 0;
  }
  
  
  /* Parameters */
  /**
   * Initializes the Parameter Manager for this Environment object
   * @return the Parameter Manager for this object 
   */
  private ParameterManager getParameterManager(){
    if(pm == null) pm = new ParameterManager(getDBGateway());
    return pm;
  }
  
  
  /**
   * Gets a Vector that lists the categories into which Parameters fall
   * @return a Vectore of Parameter Categories
   */
  public Vector getParameterCategories(){                  return (new ParameterCategoriesManager(getDBGateway())).getParameterCategories();     }
  
  
  /**
   * Gets the list of all the Parameters that fall under a specified Parameter Category
   * @param cat the name of the Parameter Category
   * @return a list of all the Parameters in the specified category
   */
  public Vector getParametersInCategory(String cat){       return getParameterManager().getParametersInCategory(cat); }
  
  
  /**
   * Gets a programmatic list (Enumeration) of all Parameters
   * @return all Parameters, Enumerated
   */
  public Enumeration getListOfParameters(){                return getParameterManager().getParameters();              }
    
  
  /**
   * Gets the description of a specified parameter.
   * @param paramName the name of the parameter
   * @return the description of the requested parameter
   */
  public String getParameterDescription(String paramName){ return getParameterManager().getDescription(paramName);    }
  
  
  /**
   * Gets the default value of the specified parameter.
   * @param paramName the name of the parameter whose default value is requested
   * @return the default value of the parameter, as a String
   */
  public String getParameterDefaultValue(String paramName){return getParameterManager().getDefaultValue(paramName);   }
  
  
  /**
   * Gets the units in which a parameter is specified.
   * @param paramName the requested parameter
   * @return the units in which that parameter should be stored
   */
  public String getParameterUnits(String paramName){       return getParameterManager().getUnits(paramName);          }
  
  
  /**
   * Gets the minimum value for a specified parameter
   * @param paramName the name of the parameter
   * @return the parameter's minimum value
   */
  public String getParameterMinValue(String paramName){    return getParameterManager().getMin(paramName);            }
  
  
  /**
   * Gets the maximum value for a specified parameter
   * @param paramName the name of the parameter
   * @return the parameter's maximum value
   */
  public String getParameterMaxValue(String paramName){    return getParameterManager().getMax(paramName);            }
  
  
  /**
   * Gets the ID of a specified parameter
   * @param paramName the name of the parameter
   * @return the parameter's ID
   */
  public String getParameterID(String paramName){          return getParameterManager().getID(paramName);             }
  
  /**
   * Gets the current value of the specified parameter in the
   * extant model. Returns -1 if the model is null
   * @param paramName the name of the parameter
   * @return the parameter value
   */
  public double getParameterCurrentValue(String paramName){
    return m.paramVal(paramName);
  }
  
   
  /**
   * Drops a data series from the specified analysis
   * @param dataSeriesName name of the series to be dropped
   * @param analysisName name of the analysis from which the series is to be dropped
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult dropDataSeriesFromAnalysis(String dataSeriesName, String analysisName){
    if(!analyses.containsKey(analysisName)) return new ABCMResult("No analysis named '" + analysisName + "' exists.", false);
    Analysis A = (Analysis)analyses.get(analysisName);
    if(!A.getSeries().containsKey(dataSeriesName)) return new ABCMResult("No series named '" + dataSeriesName + "' exists in analysis '" + analysisName + "'", false);
    A.dropSeries(dataSeriesName);
    return new ABCMResult("Dropped series '" + dataSeriesName + "' from analysis '" + analysisName + "'", true);
  }
  
  /**
   * Drops an analysis from this environment
   * @param analysisName name of the analysis to be dropped
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult dropAnalysis(String analysisName){
    if(analyses.containsKey(analysisName)){
      analyses.remove(analysisName);
      return new ABCMResult("Removed analysis '" + analysisName + ".'", true);
    }
    else return new ABCMResult("No analysis '" + analysisName + ".'", false);
  }
 
  
  /* Code Model and Parameter Options */
  
  /**
   * Gets the current value of the Code Model witin the
   * specified domain. Returns -1 if there is no current model.
   * @param domain the domain to be retrieved, i.e. "Random
   * Number Generator Type"
   * @return the value of the current Code Model on the
   * specified domain
   */
  public final long getCodeModelOptionValue(String domain){
    return m.codeModelOption(domain); 
  }
  

 /**
   * Returns true if the specified parameter applies within the
   * Code Model of the current model. Returns false if no
   * model exists; returns true if the parameter applies
   * to the fundamental HWM Engine and not to any of the
   * specific Code Models
   * @param paramName the name of the parameter to be checked
   * @return true if the parameter applies to the engine or the
   * current Code Model, false otherwise
   */
  public final boolean parameterApplies(String paramName){
    // If there is no model, return false
    if(m == null) return false;
    
    // Some parameters apply to the engine, not any specific code
    // model; these should always be considered to apply
    if(getParameterManager().isEngineApplicable(paramName)) return true;
    
    // TODO Determine if this can be done in Parameter Manager
    
    String sql = "";
    try{
      Connection con = getDBGateway().getGuestConnection(); 
      Statement st = con.createStatement();
      
      String dbOwner = getDBGateway().getOwner();
      ABCMParameterAppliesToCodeModelDomainsView view = getDBGateway().getParameterAppliesToCodeModelDomainsView();
      sql = view.getSelectSQL(dbOwner, AbstractDBGateway.where(view.getField_Parameter_Name().getFullFieldName(dbOwner),paramName));
      ResultSet rs = st.executeQuery(sql);
      String domainName = "";
      long currentValue = 0;
      while(rs.next()){
        // Note that this is inefficient, retrieving the current
        // value several times instead of storing it. However, in
        // practice most parameters will apply only to one
        // code model domain, so there will be little redundancy
        String newDomainName = rs.getString("vcharInCodeDomainIDString");
        if(domainName.compareTo(newDomainName) != 0){
          domainName = newDomainName;
          currentValue = getCodeModelOptionValue(domainName);
        }
        if(currentValue == rs.getLong("intCodeModelOptionValue")){
          rs.close();
          st.close();
          con.close();
          return true;
        }
      }
      rs.close();
      st.close();
      con.close();
    }
    catch(Exception E){
      // Better to assume that a parameter DOES apply
      System.out.println("ERROR: " + E.getMessage());
      return true;
    }    
    return false;
  }

  
  /* Session Management */
  
  /**
   * Performs a login; note that this does not guarantee
   * that the login is or is not successful.
   * @param userName the user name to be used to log in
   * @param password the password to be used to log in
   * @return a String indicating that the login was executed
   */
  public final ABCMResult login(String userName, String password){
    if(session == null) return new ABCMResult("No session.", false);
    session.setLogin(userName, password, System.currentTimeMillis());
    if(session.IsAuthorized()){
      disposeAndReset();
      lastInterchange.getBuffer().setLength(0);
      extendTranscript("Login at: " + new SimpleDateFormat("yyyy.MM.dd hh:mm:ss a").format(new Date()) + "\n");
      return new ABCMResult("Session logged in.", true);      
    }
    log(session.getErrorString());
    return new ABCMResult("Session login not completed: " + session.getErrorString(), false);
  }

  /**
   * Writes a log entry
   * @param logEntry entry to be written
   * @return true if the write was successful, false otherwise
   */
  private final boolean log(String logEntry){
    try{
      File log = new File(getLogfile());
      FileWriter writer = new FileWriter(log, true);
      writer.write(DateFormat.getInstance().format(new Date(System.currentTimeMillis())) + "\n" + logEntry);
      writer.close();
      return true;
    }
    catch(Exception E){

    }
    return false;
  }
  
  /**
   * Gets the name of the log file
   * @return name of the environment's log file
   */
  private final String getLogfile(){
    return this.getTranscriptLocation() + "LOG.txt";
  }
  
  /* Session Interface */
  
  /**
   * Gets the user who is logged in to the current session
   * @return the name of the user logged in
   */
  public final String  getUser(){         return (session!= null ? session.getUsername() : "(anonymous)");  }
  
  /**
   * Gets the user ID of the user logged in to the current session
   * @return the user ID
   */
  public final long    getUserID(){       return (session!= null ? session.getUserID()   : -1);             }
  
  /**
   * Gets the user name of the user logged in to the current session
   * @return the current user name; returns '(anonymous)' if the session is null
   */
  public final String  getUsername(){     return (session!= null ? session.getUsername() : "(anonymous)");  }
  
  /**
   * Gets the login name of the user logged in to the current session
   * @return the current user's login name; returns '(anonymous)' if the session is null
   */
  public final String  getUserLogin(){    return (session!= null ? session.getLoginName(): "(anonymous)");  }
  
  /**
   * Determines if the session has been logged in by an authorized user
   * @return true if the session is logged in, false if not
   */
  public final boolean isAuthorized(){    return (session!= null ? session.IsAuthorized(): false);          }
  
  /**
   * Determines if the current user is an administrator
   * @return true if the current user is an administrator, false otherwise
   */
  public final boolean isAdministrator(){ return (session!= null ? session.IsAdministrator() : false);      } 
  
  /* Demos */  
  
  /**
   * Initiates the processing of a demo
   * @param DemoName name of the demo to statr
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult startDemo(String DemoName){
    demoName = DemoName;
    demoStep = 0;
    return continueDemo();
  }
  
  /**
   * Continues a demo, if one has been started, by moving on to the next demo step
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult continueDemo(){
    if(demoName == null) return new ABCMResult("No demo is in progress.", false);
    demoStep++;
    return new ABCMResult(this.processScript(demoName + "." + demoStep + getScriptExtension()), true);
  }
  
  /**
   * Skips the next demo step and moves on to the next one
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult skipDemoStep(){
    demoStep++;
    return new ABCMResult("Skipping to next demo step", true);
  }
    
  /**
   * Sets the current demo step to some arbitrary number
   * @param demoStep step to be set
   * @return a message indicating the success or failure of this operation
   */
  public final ABCMResult setDemoStep(int demoStep){
    this.demoStep = demoStep;
    return new ABCMResult("Set demo step to " + demoStep, true);
  }
    
  /**
   * Decrements the demo step counter, so that the preceding demo step
   * will be repeated. Will fail if current demo step is already zero
   * or lower
   * @return an ABCMResult object indicating the success or failure of this operation
   */
  public final ABCMResult stepDemoBackwards(){
    if(demoStep > 1){
      this.demoStep--;
      return new ABCMResult("Preparing to repeat step " + demoStep, true);
    }
    else{
      return new ABCMResult("Demo step already at zero", false);      
    }    
  }
  
  /**
   * Gets the name of the current demo
   * @return the name of the current demo
   */
  public final String getDemoName(){ return demoName; }
    
  /**
   * Gets the number of the current demo step
   * @return the current demo step
   */
  public final int    getDemoStep(){ return demoStep; }
  
  /**
   * Stops processing of a demo. 
   * @param demoToStop name of demo to be stopped. Must match current demo
   * @return a string indicating the success or failure of this operation. // TODO Probably wrong!
   */
  public final String stopDemo(String demoToStop){
    if(demoName == null) return "No current demo.";
    if((demoToStop.length() > 0 ) &&
       (demoName.compareToIgnoreCase(demoToStop) > 0)){
          return "Not in demo '" + demoToStop + "'; ignoring stop command.";
    }
    String oldDemoName = demoName;
    demoName = null;
    demoStep = 0;
    return "Demo '" + oldDemoName + "' ended.";
  }

    
  /* Images */  

  /**
   * Gets the name of the image to be used as a logo for this environment
   * @return the logo image name
   */
  public abstract String getLogoImageName();
  
  /**
   * Wraps the getImage() function of the Analysis Manager, which
   * allows the retrieval of the data series graph. Uses a default
   * image height and width, and displays all data from the graph's
   * data series.
   * @param graphName the name of the graph to be drawn
   * @return the name of the file where the SVG for this graph is stored
   */
  public String getGraphImage(String graphName){
    if(!analyses.containsKey(graphName)) return "No such graph exists.";
    Analysis analysis = (Analysis)analyses.get(graphName);
    AnalysisManager AM = new AnalysisManager(getDBGateway());
    return AM.getImage(analysis, graphName, getImageOutputLocation());
  }
 
  /**
   * Wraps the getImage() function of the Analysis Manager, which allows the retrieval
   * of the data series graph. Uses the specified height and width, and displays
   * all data from the graph's data series.
   * @param graphName the name of the graph to be drawn
   * @param ImageHeight the height of the graph image, in pixels
   * @param ImageWidth the width of the graph image, in pixels
   * @return the name of the file where the SVG for this graph is stored
   */
  public String getGraphImage(String graphName, long ImageHeight, long ImageWidth){
    if(!analyses.containsKey(graphName)) return "No such graph exists.";
    Analysis analysis = (Analysis)analyses.get(graphName);
    AnalysisManager AM = new AnalysisManager(getDBGateway());
    return AM.getImage(analysis, graphName, ImageHeight, ImageWidth, getImageOutputLocation());
  }
  
  /**
   * Wraps the getImage() function of the Analysis Manager, which allows the retrieval
   * of the data series graph. Uses the specified height and width. The user may also
   * specify a limit to the range of dates that are displayed.
   * @param graphName the name of the graph to be drawn
   * @param ImageHeight the height of the graph image, in pixels
   * @param ImageWidth the width of the graph image, in pixels
   * @param startDate the date of the earliest data to be included on the graph
   * @param endDate the date of the latest data to be included on the graph
   * @return the name of the file where the SVG for this graph is stored
   */
  public String getGraphImage(String graphName, long ImageHeight, long ImageWidth, String startDate, String endDate){
    if(!analyses.containsKey(graphName)) return "No such graph exists.";
    Analysis analysis = (Analysis)analyses.get(graphName);
    AnalysisManager AM = new AnalysisManager(getDBGateway());
    return AM.getImage(analysis, graphName, ImageHeight, ImageWidth, startDate, endDate, getImageOutputLocation());
  }
  
  /**
   * Gets the image of a specific analysis. This completely wraps the functionality
   * of the AnalysisManager.getImage(PK) method.
   * @param analysisID the primary key of the Analysis in the HWM Database
   * @return the name of the file where the SVG image is stored
   */
  public String getGraphImage(long analysisID){
    AnalysisManager AM = new AnalysisManager(getDBGateway());
    return AM.getImage(analysisID, getImageOutputLocation());
  }

  // Specifically for running in an HTML environment
  
  /**
   * Wraps the function of the HWMSVGMapViewer, and allows the Web Application
   * to open the SVG maps.
   * @return the file name where the SVG map is stored
   */
  public String getMapImage(){
    if(m == null) return "unsuccessful";

    SVGViewer svg = new SVGViewer();
    
    return this.getImageURLPrefix() + svg.getSVG(new SVGRepresentable[]{ m }, this.getImageOutputLocation(), 
        this.getUsername() + "_" + (new SimpleDateFormat("yyyy_MM_dd__HH_mm_ss.SSS")).format(new Date(System.currentTimeMillis())));
  }
  
  /**
   * Gets a map image of the model in this environment
   * @param mapMinX minimum X coordinate to be included in the map
   * @param mapMinY minimum Y coordinate to be included in the map
   * @param mapMaxX maximum X coordinate to be included in the map
   * @param mapMaxY maximum Y coordinate to be included in the map
   * @param width actual SVG width of the image
   * @param height actual SVG height of the image
   * @param pageWidth page width of the image
   * @param pageHeight page height of the image
   * @param showContours true if the map should include contours
   * @param showDelaunay true if the map should include a Delaunay triangulation of the landscape
   * @return the name of the image as saved on disk
   */
  public String getMapImage(double mapMinX, double mapMinY, double mapMaxX, double mapMaxY, double width, 
      double height, double pageWidth, double pageHeight, boolean showContours, boolean showDelaunay){
    SVGViewer svg = new SVGViewer();
    svg.setPageWidthAndHeightInInches(pageWidth, pageHeight);
    svg.setPageWidthAndHeightInPixels((int)width, (int)(height));
    m.getSpace().setDisplayRange(mapMinX, mapMinY, mapMaxX, mapMaxY);
    m.getSpace().includeContourLines = showContours;
    m.getSpace().includeDelaunayTriangulation = showDelaunay;
    return this.getImageURLPrefix() + svg.getSVG(new SVGRepresentable[]{ m }, this.getImageOutputLocation());
  }
  
  /**
   * Gets a map image of the model in this environment
   * @param width actual SVG width of the image
   * @param height actual SVG height of the image
   * @param pageWidth page width of the image
   * @param pageHeight page height of the image
   * @param showContours true if the map should include contours
   * @param showDelaunay true if the map should include a Delaunay triangulation of the landscape
   * @return the name of the image as saved on disk
   */
  public String getMapImage(double width, double height, double pageWidth, double pageHeight, 
      boolean showContours, boolean showDelaunay){
    SVGViewer svg = new SVGViewer();
    svg.setPageWidthAndHeightInInches(pageWidth,pageHeight);
    svg.setPageWidthAndHeightInPixels((int)width, (int)(height));
    getSpace().includeContourLines = showContours;
    getSpace().includeDelaunayTriangulation = showDelaunay;
    System.out.println("SET TO " + showContours + " and " + showDelaunay);
    return this.getImageURLPrefix() + svg.getSVG(new SVGRepresentable[]{ m }, this.getImageOutputLocation(),
        this.getUsername() + "_" + (new SimpleDateFormat("yyyy_MM_dd__HH_mm_ss.SSS")).format(new Date(System.currentTimeMillis())));
  }
 
  /**
   * Gets an X3D image of the model space
   * @return name of the X3D image file as saved on disk
   */
  public String getX3DImage(){
    if(m == null) return "unsuccessfulAttempt";
    X3DViewer x3d = new X3DViewer();
    return x3d.getX3D(new X3DRepresentable[]{ m }, this.getImageOutputLocation(), 
        this.getUsername() + "_" + (new SimpleDateFormat("yyyy_MM_dd__HH_mm_ss.SSS")).format(new Date(System.currentTimeMillis())));
  }
  
  /**
   * Returns an HTML table listing the Analysis data for the specified analysis,
   * if the analysis is a graph in the current environment.
   * @param analysisName the name of the Analysis to be retrieved
   * @return the HTML table of the Analysis data
   */
  public String getAnalysisData(String analysisName){
    return getAnalysisData(analysisName, false);
  }
  
  /**
   * Returns a table listing the Analysis data for the specified analysis, if the
   * analysis is a graph in the current environment. If ASCII is selected, the
   * result is human-readable (dates are given in human-readable form).
   * @param analysisName the name of the Analysis to be retrieved
   * @param ASCII true if the data should be returned as ASCII; returns HTML otherwise
   * @return the table of analysis data
   */
  public String getAnalysisData(String analysisName, boolean ASCII){
    if(!analyses.containsKey(analysisName)) return "";
    Analysis a = (Analysis)analyses.get(analysisName);
    return (ASCII ? a.getAnalysisAsASCIITable() : a.getAnalysisAsHTMLTable());
  }
  
  /**
   * Gets an HTML Option List that lists the names of all the series contained in the
   * various analyses and graphs in the current GUI.
   * @return the HTML Option List of all the series in all the GUI's graphs
   */
  public String getGraphedSeriesAsOptionList(){
    String retVal = "";
    Hashtable analysesRecovered = new Hashtable();
    
    // Loop through all the available analyses
    Enumeration e = analyses.elements();
    while(e.hasMoreElements()){
      Analysis a = (Analysis)e.nextElement();
      Hashtable h = a.getSeries();
      // Loop through the series in this analysis
      Enumeration seriesNames = h.keys();
      while (seriesNames.hasMoreElements()){
        // Add names to a hashtable- will eliminate duplicates
        String seriesName = (String)seriesNames.nextElement();
        analysesRecovered.put(seriesName, seriesName);
      }
    }
    
    Enumeration names = analysesRecovered.keys();
    while(names.hasMoreElements()){
      String n = (String)names.nextElement();
      retVal+="<OPTION VALUE=\"" + n + "\">" + n + "</OPTION>";
    }
    return retVal;
  }

  /**
   * Gets the list of graphs being managed in this GUI as an HTML Option List.
   * @return a String containing the HTML Option List of graphs
   */
  public String getGraphsAsOptionList(){
    String retVal = "";
    Enumeration e = analyses.keys();
    while(e.hasMoreElements()){
      String g = (String)e.nextElement();
      retVal+="<OPTION VALUE=\"" + g + "\">" + g + "</OPTION>";
    }
    return retVal;
  }
  
  /**
   * Gets the list of Data Series being managed in this GUI as an HTML Option List.
   * @return a String containing the HTML Option List of Data Series
   */
  public String getDataSeriesAsOptionList(){
    if(!hasModel()) return "";
    String retVal = "";
    Enumeration e = getModel().getDataSeriesNames();
    while(e.hasMoreElements()){
      String sn = (String)e.nextElement();
      retVal+="<OPTION VALUE=\"" + sn + "\">" + sn + "</OPTION>";
    }
    return retVal;
  }

  /* DB */
  
  /**
   * Create the database in this model environment. This should only be run when the database
   * is to be initially set up.
   * @return true if successful, false if not
   */
  public boolean createDB(){
    if(getDBGateway().createDB()){
      System.out.println("Database created successfully.");
      return true;
    }
    else{
      System.out.println("An error occurred; database not completed.");
      return false;
    }
  }
  
  /**
   * Stores a probe in the database.
   * @param name name of the probe to be saved
   * @param description description of the action of the probe
   * @param dataSeriesName name of the data series to be used by the probe
   * @param dataSeriesDomain name of the domain in the data series used by the probe
   * @param isNumeric true if the probe returns a numeric value, false otherwise
   * @param isMonthly true if the probe is monthly, false otherwise
   * @param baseDate base date from which execution of the probe is to be offset
   * @param frequency frequency, in days, at which the probe is to be executed
   * @param retrieve commands used by the probe to retrieve data from the model
   * @param offset double indicating some temporal offset from 12:00 am; useful in retrieving time-dependent data
   * @return a message indicating the success or failure of this operation
   */
  protected ABCMResult storeProbe(String name, String description, String dataSeriesName, String dataSeriesDomain, boolean isNumeric,
      boolean isMonthly, SimDate baseDate, int frequency, String retrieve, double offset){
    if(getUserID() == -1) return new ABCMResult("Could not store probe; no current user", false);
    ProbeManager pm = new ProbeManager(this.getDBGateway());
    pm.setInstanceToSave(getUserID(), name, description, dataSeriesName, dataSeriesDomain, isNumeric, isMonthly, 
        baseDate.value(), frequency, retrieve, offset);
    if(!pm.saveInstanceToDB()) return new ABCMResult("Could not save probe: " + pm.getErrorReport(), false);
    else                       return new ABCMResult("Probe " + name + " saved to db.", true);
  }
  
  /**
   * Stores a probe in the database.
   * @param name name of the probe to be saved
   * @param description description of the action of the probe
   * @param dataSeriesName name of the data series to be used by the probe
   * @param dataSeriesDomain name of the domain in the data series used by the probe
   * @param isNumeric true if the probe returns a numeric value, false otherwise
   * @param baseDate base date from which execution of the probe is to be offset
   * @param frequency frequency, in days, at which the probe is to be executed
   * @param retrieve commands used by the probe to retrieve data from the model
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult storeProbe(String name, String description, String dataSeriesName, String dataSeriesDomain, boolean isNumeric,
      SimDate baseDate, int frequency, String retrieve){
    return storeProbe(name, description, dataSeriesName, dataSeriesDomain, isNumeric, baseDate, frequency, retrieve, 0);
  }
  
  /**
   * Stores a probe in the database.
   * @param name name of the probe to be saved
   * @param description description of the action of the probe
   * @param dataSeriesName name of the data series to be used by the probe
   * @param dataSeriesDomain name of the domain in the data series used by the probe
   * @param isNumeric true if the probe returns a numeric value, false otherwise
   * @param baseDate base date from which execution of the probe is to be offset
   * @param frequency frequency, in days, at which the probe is to be executed
   * @param retrieve commands used by the probe to retrieve data from the model
   * @param offset double indicating some temporal offset from 12:00 am; useful in retrieving time-dependent data
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult storeProbe(String name, String description, String dataSeriesName, String dataSeriesDomain, boolean isNumeric,
      SimDate baseDate, int frequency, String retrieve, double offset){
    return storeProbe(name, description, dataSeriesName, dataSeriesDomain, isNumeric, false, baseDate, frequency, retrieve, offset);    
  }
  
  /**
   * Stores a monthly probe in the database.
   * @param name name of the probe to be saved
   * @param description description of the action of the probe
   * @param dataSeriesName name of the data series to be used by the probe
   * @param dataSeriesDomain name of the domain in the data series used by the probe
   * @param isNumeric true if the probe returns a numeric value, false otherwise
   * @param retrieve commands used by the probe to retrieve data from the model
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult storeMonthlyProbe(String name, String description, String dataSeriesName, String dataSeriesDomain, boolean isNumeric,
      String retrieve){
    return storeMonthlyProbe(name, description, dataSeriesName, dataSeriesDomain, isNumeric, retrieve, 0);
  }
  
  /**
   * Stores a monthly probe in the database.
   * @param name name of the probe to be saved
   * @param description description of the action of the probe
   * @param dataSeriesName name of the data series to be used by the probe
   * @param dataSeriesDomain name of the domain in the data series used by the probe
   * @param isNumeric true if the probe returns a numeric value, false otherwise
   * @param retrieve commands used by the probe to retrieve data from the model
   * @param offset double indicating some temporal offset from 12:00 am; useful in retrieving time-dependent data
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult storeMonthlyProbe(String name, String description, String dataSeriesName, String dataSeriesDomain, boolean isNumeric,
      String retrieve, double offset){
    return storeProbe(name, description, dataSeriesName, dataSeriesDomain, isNumeric, true, new SimDate(), 0, retrieve, offset);
  }
  
  /**
   * Stores a probe set in the database
   * @param name name to be assigned to the probe set; must be unique
   * @param desc description of the probe set
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult storeProbeSet(String name, String desc){
    if(getUserID() == -1) return new ABCMResult("Could not store probe set; no current user", false);
    ProbeSetManager psm = new ProbeSetManager(this.getDBGateway());
    psm.setInstanceToSave(name, desc, getUserID());
    if(!psm.saveInstanceToDB()) return new ABCMResult("Could not save probe set: " + psm.getErrorReport(), false);
    else                        return new ABCMResult("Probe set " + name + " saved.", true);
  }
  
  
  /**
   * Adds a probe to the specified probe set 
   * @param probeName name of the probe to add
   * @param probeSetName name of the probe set to which it is to be added
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult addProbeToProbeSet(String probeName, String probeSetName){
    if(getUserID() == -1) return new ABCMResult("Could not add probe set; no current user", false);
    ProbeSetManager psm = new ProbeSetManager(this.getDBGateway());
    ProbeManager pm = new ProbeManager(this.getDBGateway());
    long probePK = pm.getPK(probeName);
    if(probePK == -1) return new ABCMResult("Probe '" + probeName + "' does not exist", false);
    long probeSetPK = psm.getPK(probeSetName);
    if(probeSetPK == -1) return new ABCMResult("Probe Set '" + probeSetName + "' does not exist", false);
    
    if(!psm.addAProbe(probeSetPK, probePK)) return new ABCMResult("Could not add probe to probe set: " + psm.getErrorReport(), false);
    else return new ABCMResult("Probe '" + probeName + "' added to Probe Set '" + probeSetName + "'", true);
  }
  
  /**
   * Stores a history in the database
   * @param historyName name of the history to be saved
   * @param historyDesc description
   * @param truthValue value indicating the truth value of this history
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult storeHistory(String historyName, String historyDesc, String truthValue){
    if(getUserID() < 0) return new ABCMResult("Could not store history; user ID is not valid", false);
    TruthValueManager TM = new TruthValueManager(getDBGateway());
    long tv = TM.getPK(truthValue);
    if(tv < 0) return new ABCMResult("Could not store history; truth value '" + truthValue + " is not valid", false);
    HistoryManager HM = new HistoryManager(getDBGateway());
    HM.setInstanceToSave(historyName, historyDesc, getUserID(), tv);
    if(!HM.saveInstanceToDB()) return new ABCMResult("Could not store history: " + HM.getErrorReport(), false);
    return new ABCMResult("History '" + historyName + " stored.", true);
  }
  
  /**
   * Adds a narrative to the specified history
   * @param historyName name of the history to which this narrative is to be added
   * @param narrativeName name of the narrative to be added
   * @param baseDate base date for this narrative in this history
   * @param order order in which this narrative is to be executed relative to other narratives. If the order
   * is higher than any pre-existing narratives' orders, other narratives will be shuffled downward to give this
   * one the recommended priority. Resulting values will be collapsed to the smallest numbers in sequence beginning
   * with one.
   * @return a message indicating the success or failure of this operation
   */
  public ABCMResult addNarrativeToHistory(String historyName, String narrativeName, String baseDate, int order){
    HistoryManager HM = new HistoryManager(getDBGateway());
    NarrativeManager NM = new NarrativeManager(getDBGateway());
    long historyPK = HM.getPK(historyName);
    if(historyPK < 0) return new ABCMResult("Could not add narrative to history; history '" + historyName + " is unknown: " + HM.getErrorReport(), false);
    
    long narrativePK = NM.getPK(narrativeName);
    if(narrativePK < 0) return new ABCMResult("Could not add narrative to history; narrative '" + narrativeName + " is unknown: " + NM.getErrorReport(), false);
    
    if(SimDate.parseString(baseDate) < 0) return new ABCMResult("Could not add narrative to history; base date '" + baseDate + "' is invalid", false);
    
    if(!HM.insertNarrative(historyPK, narrativePK, SimDate.parseString(baseDate), order))
        return new ABCMResult("Could not add narrative to history: " + HM.getErrorReport(), false);
    return new ABCMResult("Added narrative '" + narrativeName + " to history " + historyName + " at base date " + baseDate, true);
    
  }
  
}
