package abcm.model.agents;

import java.util.Hashtable;

import abcm.model.ABCMResult;
import abcm.sim.AbstractModelComponent;
import abcm.utilities.QuotedLineParser;

/**
 * An Actor is any component that can accept direct instructions,
 * parse them, and return the results of those actions. This structure
 * allows interactivity with both agents and populations
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractABCMActor extends AbstractModelComponent{

  /* Inner Classes */
  /**
   * An Action is something that an actor can perform; actions are
   * added to ActorCommandParser
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  protected abstract class Action{
    
    /**
     * Causes this Action object to do what it is intended to do
     * @param actor the actor undertaking the action
     * @param command the action to be undertaken
     * @return an ABCMResult indicating the results of the action
     */
    public abstract ABCMResult execute(AbstractABCMActor actor, String[] command);
  
  }
  
  /**
   * Causes the actor to return a String indicating its status
   */
  protected class Action_Report extends Action{
    
    /**
     * @see abcm.model.agents.AbstractABCMActor.Action#execute(abcm.model.agents.AbstractABCMActor, java.lang.String[])
     */
    public ABCMResult execute(AbstractABCMActor actor, String[] command){ 
      return (actor != null ? new ABCMResult(actor.report(), true) : new ABCMResult("Could not get agent report; agent specified is null.", false));
    }
  }
  
  /**
   * A Parser allows actors to link commands to actions
   */
  protected abstract class ActorCommandParser{
    private Hashtable actions;
    
    /**
     * Constructor
     */
    public ActorCommandParser(){
      actions = new Hashtable();
      addCommand("REPORT", new Action_Report());
    }
    
    /**
     * Adds a command to the parser
     * @param commandName the name of the command to be executed
     * @param action the action to be executed
     */
    protected void addCommand(String commandName, Action action){
      String c = commandName.toUpperCase();
      if(actions.containsKey(c)) actions.remove(c);
      actions.put(c, action);
    }
    
    /**
     * Causes this parser object to process the command
     * @param actor the actor undertaking the action
     * @param command the command to be executed
     * @param commandList arguments to command
     * @return an ABCMResult indicating the outcome of the action
     */
    public final ABCMResult processCommand(AbstractABCMActor actor, String command, String[] commandList){
      String c = command.toUpperCase();
      if(actions.containsKey(c)) return ((Action)actions.get(c)).execute(actor, commandList);
      else                       return new ABCMResult("Could not process agent command; command '" + command + "' is unknown", false);
    }    
  }

  /* Processing Commands */
  /**
   * Processes a command
   * @param command the command to be processed, as a single String
   * @return an ABCMResult indicating the outcome of the action
   */
  public final ABCMResult processCommand(String command){
    if(command.length() == 0) command="NO_COMMAND";
    String[] commandList = QuotedLineParser.parseQuoted(command, 1000);
    //    System.out.println("I'm actor " + getID() + " and I'm executing command: " + command);
    return getActorCommandParser().processCommand(this, commandList[0], commandList);
  }

  /**
   * Implementing classes must provide access to the parser that
   * they are using; child classes to the AbstractABCMActor may (almost
   * certainly will) wish to extend the ActorCommandParser to include 
   * additional functionality.
   * @return the parser this Actor object uses
   */
  protected abstract ActorCommandParser getActorCommandParser();

  /* Reporting */
  /**
   * Implementing classes should provide a simple String report
   * @return a report describing the status of this object
   */
  public abstract String report();

  /**
   * Implementing classes should provide a simple String ID
   * @return the agent's String ID
   */
  public abstract String getID();
}
