package abcm.model.agents;

import abcm.model.AbstractABCModel;

/**
 * Agents may be of any kind needed by implementing classes.
 * They will inherit a basic parser functionality and some other
 * fundamental features (a common method for generating ID Strings,
 * for example).
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractABCMAgent extends AbstractABCMActor{
  
  /* Static Variables */
  /** Counter that tracks the number of agents and can be used to assign unique agent IDs  */
  private static int agentCounter = 0;
  
  /* Instance Variables*/
  /** String ID for this agent  */
  private   String            agentID;
  /** Agent's pointer to the Model in which it exists   */
  protected AbstractABCModel  M;
  
  /* Constructor */
  /**
   * Constructor 
   * @param model Model in which this agent exists
   * @param pop Population in which this agent exists
   */
  public AbstractABCMAgent(AbstractABCModel model, AbstractABCModelPopulation pop){
    M = model;
    agentID = getAgentIDPrefix() + agentCounter + getAgentIDSuffix();  
    agentCounter++;
    pop.addAgent(this);
  }
  
  /* Initialization */
  /**
   * Implementing classes should provide a prefix for agent IDs
   * @return a String to be prefixed to agent IDs
   */
  protected abstract String getAgentIDPrefix();
  
  
  /**
   * Implementing classes should provide a suffix for agent IDs
   * @return a String to be appended to agent IDs
   */
  protected abstract String getAgentIDSuffix();
  
  
  
  /* Getters */
  /**
   * Gets the agent ID
   * @return the String ID for this agent
   */
  public final String getID(){ return agentID; }
  
  
  /* Actions */
  /**
   * Implementing classes should provide some basic step mechanism
   */
  public abstract void step();
  
  /**
   * Utility output that prepends the agent ID to send messages to the output screen (for debugging)
   * @param msg the message to be sent
   */
  protected void out(String msg){
    System.out.println("AGENT " + getID() + ": " + msg);
  }
  
}
