package abcm.model.agents;

import java.awt.Color;
import java.util.Hashtable;

import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.model.AbstractABCModel;

/**
 * Instances of agents that have X,Y locations- that is, are positioned spatially
 * in an ABCM model.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractABCMLocatedAgent extends AbstractABCMAgent implements SVGRepresentable{

  public static final String TRIANGLE = "Triangle";
  public static final String SQUARE   = "Square";
  public static final String CIRCLE   = "Circle";
  
  public static final String RED      = "Red";
  public static final String BLUE     = "Blue";
  public static final String GREEN    = "Green";
  public static final String ORANGE   = "Orange";
  public static final String YELLOW   = "Yellow";
  
  /** Defines a list of colors identified by String name */
  private static Hashtable colors = new Hashtable();
  
  // Initialize the colors hashtable
  {
    colors.put(RED,    Color.RED   );
    colors.put(BLUE,   Color.BLUE  );
    colors.put(GREEN,  Color.GREEN );
    colors.put(ORANGE, Color.ORANGE);
    colors.put(YELLOW, Color.YELLOW);
  }
  
  /** E-W position of this agent */
  protected double x;
  
  /** N-S position of this agent */
  protected double y;
  
  /** String identifying the shape of the icon to be used for depictions of this agent */
  protected String iconType     = TRIANGLE;
  
  /** Size of this agent, in inches, in 2-D depictions */
  protected double sizeInInches = .1d;
  
  /** Color to be used when depicting this agent */
  protected String color        = RED;
  
  /**
   * Creates a new instance of this agent at the specified location
   * @param model the ABCM model in which this agent will operate
   * @param pop the population in which this agent will exist
   * @param initialXLocation initial E-W location of the agent
   * @param initialYLocation initial N-S location of the agent
   */
  public AbstractABCMLocatedAgent(AbstractABCModel model, AbstractABCModelPopulation pop, 
      double initialXLocation, double initialYLocation) {
    super(model, pop);
    x = initialXLocation;
    y = initialYLocation;
  }

  
  /**
   * Sets the icon that will be used to identify this agent in 2-D maps. One of the
   * names of a valid icon type must be used or the icon type will not be set.
   * @param newIconType one of the list of permitted icon types (i.e. 'Triangle', 'Square', 'Circle').
   * @return true if successful, false otherwise. 
   */
  public boolean setIcon(String newIconType){
    if(newIconType.compareTo(SQUARE)   == 0){ iconType = SQUARE;   return true;  }
    if(newIconType.compareTo(CIRCLE)   == 0){ iconType = CIRCLE;   return true;  }
    if(newIconType.compareTo(TRIANGLE) == 0){ iconType = TRIANGLE; return true;  }
    return false;
  }
  
  /**
   * Moves this agent from one location to another
   * @param newX new X position
   * @param newY new Y position
   */
  public void move(double newX, double newY){
    x = newX;
    y = newY;
  }
  
  /**
   * @see abcm.graphics.SVGRepresentable#addToSVG(abcm.graphics.SVGAssist)
   */
  public void addToSVG(SVGAssist svg){
    if(iconType.compareTo(CIRCLE) == 0) svg.addCircle(x, y, svg.getUnitsPerInchX() * sizeInInches, SVGAssist.getRGB((Color)colors.get(color)));
    if(iconType.compareTo(SQUARE) == 0) svg.addRectangle(x, y, svg.getUnitsPerInchX() * sizeInInches, svg.getUnitsPerInchY() * sizeInInches, 
        SVGAssist.getRGB((Color)colors.get(color)), svg.getUnitsPerInchX() * sizeInInches, SVGAssist.getRGB((Color)colors.get(color)));    
  }
  
  
}
