package abcm.model.agents;

import java.util.Hashtable;
import java.util.Vector;

import abcm.model.ABCMResult;
import abcm.model.AbstractABCModel;

/**
 * An AbstractABCModelPopulation is a container class for agents.
 * 
 * Populations can also execute instructions on command.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractABCModelPopulation extends AbstractABCMActor{

  /** List of agents in this population  */
  protected Hashtable<String, AbstractABCMAgent> agentList = new Hashtable<String, AbstractABCMAgent>();

  /** Pointer to the model in which this population exists  */
  protected AbstractABCModel M;
  
  /**
   * Constructor
   * @param model model in which this population exists
   */
  public AbstractABCModelPopulation(AbstractABCModel model){
    M = model;
    createAgents();
  }
  
  /**
   * Resets the agent list by destroying all agents and
   * creating a new collection in the manner specified by the
   * Code Model object
   */
  public void resetAgents(){
    destroyAgents();
    createAgents();
  }
  
  /**
   * Removes all agents.
   */
  public void destroyAgents(){
    agentList = new Hashtable<String, AbstractABCMAgent>();
  }
  
  /**
   * Returns the count of agents in the agent list.
   * @return count of the agents in the agent list.
   */
  public long countOfAgents(){
    return agentList.size();
  }
  
  /**
   * Calls the step method for all agents in the
   * agent list, in random order 
   */
  public void step(){
    takePopulationActions();
    Vector v = getVectorOfAgents();
    while(v.size() > 0){
      AbstractABCMAgent a = (AbstractABCMAgent)(v.remove((int)Math.floor(random() * v.size())));
      a.step();
    }    
  }
  
  /**
   * This method is called prior to invoking agent actions during
   * the step() method; it can be overridden as a simple means
   * of providing population-level actions without re-writing
   * the agent loop.
   */
  public void takePopulationActions(){}
  
  /**
   * Transform the hashtable of agents into a vector,
   * generally so that they can be removed in order or in
   * random order.
   * @return a Vector containing all the Agents in this Population
   */
  private Vector<AbstractABCMAgent> getVectorOfAgents(){    return new Vector<AbstractABCMAgent>(agentList.values());   }
  
  /**
   * Adds an agent to this population
   * @param agent the agent to be added
   */
  protected final void addAgent(AbstractABCMAgent agent){ agentList.put(agent.getID(), agent); }
  
  /**
   * Gets an agent when called with the specified agent's ID String
   * @param agentID the ID String of the agent to be retrieved
   * @return the requested agent; if no agent with the specified ID
   * is found, returns null
   */
  public AbstractABCMAgent getAgent(String agentID){
    return agentList.get(agentID);
  }
  
  /**
   * Implementing classes should be able to create populations
   * of agents.
   */
  public abstract void createAgents();
  
  /**
   * Sends the command to the agent
   * @param agentID ID of agent to be sent command
   * @param command command to be sent to agent
   * @return the ABCMResult indicating the agent's outcome, or an error
   */
  public final ABCMResult sendCommandToAgent(String agentID, String command){
    AbstractABCMAgent agent = getAgent(agentID);
    if(agent == null) return new ABCMResult("No agent with ID '" + agentID + "' exists.", false);
    return agent.processCommand(command);    
  }
  
  /**
   * Chooses an agent at random from the list and returns it. Uses current
   * random number generator
   * @return a randomly chosen agent
   */
  public final AbstractABCMAgent getAgentAtRandom(){
    return (agentList.size() == 0 ?  null : agentList.get(Math.floor(agentList.size() * this.random())));    
  }
  
}
