package abcm.model.tools;

import java.io.StringWriter;

import abcm.db.managers.CommandsManager;
import abcm.math.SimDate;
import abcm.model.AbstractABCMParser;
import abcm.model.AbstractABCModelEnvironment;
import abcm.utilities.QuotedLineParser;
import abcm.utilities.ScriptReader;

/**
 * Converts a narrative file in which commands are stored in their English
 * versions to one in which all commands are numeric; the resulting format
 * is one that can be easily uploaded to an ABCM database.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class NarrativeCompiler {

  
  /**
   * Constructor; private scope prevents this class from being instantiated or extended.
   */
  private NarrativeCompiler(){}
  
  /**
   * Given an input file containing a narrative, returns a StringWriter
   * containing the 'compiled' version, ready to be uploaded to a db.
   * @param inputFileName name of the input file
   * @param env model environment, providing access to the database.
   * @return StringWriter containing text of compiled version
   */
  public static StringWriter compileNarrative(String inputFileName, AbstractABCModelEnvironment env){
    CommandsManager cm = env.getCommandsManager();
    ScriptReader sr = new ScriptReader(inputFileName);
    StringWriter ret = new StringWriter();
    
    int warningCount = 0;
    long line = 1;
    long order = 0;
    double oldDate = 0;
    double newDate = 1;
    while(!sr.EOF()){
      String nextLine = sr.getNextLine();
      if(nextLine != null){
        if(!nextLine.startsWith("**") && nextLine.length() > 0){
          nextLine = nextLine.replaceAll("\t"," ");
          String[] inputLine = QuotedLineParser.parseQuoted(nextLine, AbstractABCMParser.MAX_ARGS + 1);
          String onDate = inputLine[0];
          newDate = SimDate.parseString(onDate);
          if(newDate == -1) {
            System.err.println("WARNING: Invalid date '" + onDate + "' in line: " + line);
            warningCount++;
          }
          else{
            if(newDate != oldDate) order = 1;
            else                   order++;
            oldDate = newDate;
            
            ret.write(newDate + "\t" + order + "\t");
            long PK = cm.getCommandPK(inputLine[1]);
            if(PK==0) {
              System.err.println("WARNING: Unknown command in line: " + line + " '" + inputLine[1] + "'");
              warningCount++;
            }
            ret.write(PK + "");
            int i = 2;
            while((i < inputLine.length) && (inputLine[i] != null) && (inputLine[i].length() > 0)){
              ret.write("\t" + inputLine[i]);
              i++;
            }
          }
          ret.write("\n");
        }
      }
      line++;
    }
    if(warningCount > 0) System.err.println(warningCount + " warnings.");
    return ret;
  } 
  
}
