package abcm.model.tools;

import java.io.StringWriter;

import abcm.math.SimDate;
import abcm.model.AbstractABCMParser;
import abcm.utilities.QuotedLineParser;
import abcm.utilities.ScriptReader;

/**
 * Converts a narrative file into a script.
 * 
 * Conversion includes adding 'steps' between dates on which
 * the narrative has instructions. Result is a StringWriter
 * object with the new script's text, not a new file.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class NarrativeToScript {

  /**
   * Constructor; private scope prevents this class from being instantiated or extended.
   */
  private NarrativeToScript(){}
  
  /**
   * Converts a narrative file to a script file.
   * @param inputFileName name of the file to be converted
   * @return StringWriter containing the text of the new script.
   */
  public static StringWriter narrativeToScript(String inputFileName){
    StringWriter ret = new StringWriter();
    
    boolean isSeasonal = false;
    double currentDate = -1;

    ScriptReader sp = new ScriptReader(inputFileName);
    
    int line = 1;
    while(!sp.EOF()){
      String nextLine = sp.getNextLine();
      if(nextLine != null){
        nextLine = nextLine.replaceAll("\t"," ");
        if(nextLine.length() > 0 && !nextLine.startsWith("**")){
          String[] inputLine = QuotedLineParser.parseQuoted(nextLine, AbstractABCMParser.MAX_ARGS + 2);
          double newDate = SimDate.parseString(inputLine[0]);
          if(currentDate > -1){          
            if(currentDate != newDate){
              if(newDate == (currentDate + 1)){
                ret.write("Step\n");
              }
              else if(newDate < (currentDate + 20)){
                ret.write("StepNDays " + (int)(newDate - currentDate) + "\n");
              }
              else{
                ret.write("StepToDate " + inputLine[0] + "\n");
              }
              currentDate = newDate;
            }
          }
          else {
            currentDate = SimDate.parseString(inputLine[0]);
          }
          String retVal = "";
          for(int x = 1; x < inputLine.length; x++){
            retVal += inputLine[x] + " ";
          }
          ret.write(retVal + "\n");
        }
        line++;
      }
    }
    return ret;
  }  
}
