package abcm.model.tools;

import java.sql.Connection;
import java.sql.Statement;

import abcm.db.components.ABCMdbDatum;
import abcm.db.managers.NarrativeManager;
import abcm.model.ABCMResult;
import abcm.model.AbstractABCMParser;
import abcm.model.AbstractABCModelEnvironment;
import abcm.utilities.QuotedLineParser;
import abcm.utilities.ScriptReader;

/**
 * Uploads a Narrative (in compiled narrative format) to an ABCM database
 * as a stored narrative.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class NarrativeUploader {

  
  /**
   * Constructor; private scope prevents this class from being instantiated or extended.
   */
  private NarrativeUploader(){}

  /**
   * Upload the specified narrative into the database
   * @param inputFileName name of file in which compiled narrative is stored
   * @param newName name to be assigned to this narrative in the db. Must not duplicate an extant narrative name
   * @param desc description
   * @param baseDate base date for this narrative as stored
   * @param isSeasonal true if the narrative is a seasonal narrative
   * @param env environment in which this upload is taking place (will provide access to database gateway)
   * @return true if successful, false otherwise
   */
  public static ABCMResult uploadNarrative(String inputFileName, String newName, String desc, String baseDate, 
      boolean isSeasonal, AbstractABCModelEnvironment env){
    System.out.println("Uploading " + inputFileName);
    
    if(!env.isAuthorized()) return new ABCMResult("Could not upload narrative; user is not authorized", false);
    String userName = env.getUsername();
    ScriptReader sp = new ScriptReader(inputFileName);
    if(!sp.getSuccessFlag()){
      return new ABCMResult("Could not find script: " + sp.getErrorReport(), false);
    }
    long user = env.getUserID();
    
    NarrativeManager nm = new NarrativeManager(env.getDBGateway());
    if((nm.setInstanceToSave(newName, desc, user, baseDate, isSeasonal) == false) || (nm.saveInstanceToDB() == false)){
      System.err.println(nm.getErrorReport());
      return new ABCMResult("Failed to upload narrative: " + nm.getErrorReport(), false);
    }
    
    long newNarrID = nm.getLastSavedID();
    Connection con = env.getDBGateway().getAdminConnection();
    
    try{
      con.setAutoCommit(false);
      Statement st = con.createStatement();  
      int line = 1;
      System.err.println("Processing line: " + line);
      while(!sp.EOF()){
        String nextLine = sp.getNextLine();
        if(nextLine != null && nextLine.length() > 0){
          nextLine = nextLine.replaceAll("\t"," ");
          String[] inputLine = QuotedLineParser.parseQuoted(nextLine, AbstractABCMParser.MAX_ARGS + 2);
          
          int x = inputLine.length-1;
          while( (inputLine[x].length() == 0) && (x >= 0))  x--;
          int argLength = x - 2;
          String[] newArgs = new String[argLength];
          for(int y = 0; y < argLength; y++){
            newArgs[y] = inputLine[y + 3];
          }
          // This is the correct way to do this, but it's slow...
//          if(!nm.insertInstructionOnDate(newNarrID,Double.parseDouble(inputLine[0]), 
//                                     Integer.parseInt(inputLine[1]), 
//                                     Integer.parseInt(inputLine[2]),
//                                     newArgs, false)){
//            System.err.println("Error reading line " + line + ": " + nextLine);
//            System.err.println(nm.getErrorReport());
//            errorReport = "Error reading line " + line + ": " + nextLine + " \n" + nm.getErrorReport();
//            sp.closeFile();
//            return false;
//          }
          ABCMdbDatum d = env.getDBGateway().getNarrativeInstructionsTable().getDatum(
              "" + newNarrID,
              "" + Double.parseDouble(inputLine[0]), 
              "" + Integer.parseInt(inputLine[1]), 
              "" + Integer.parseInt(inputLine[2]), 
              (newArgs.length > 0 ? newArgs[0] : ""), 
              (newArgs.length > 1 ? newArgs[1] : ""), 
              (newArgs.length > 2 ? newArgs[2] : ""),
              (newArgs.length > 3 ? newArgs[3] : ""),
              (newArgs.length > 4 ? newArgs[4] : ""),
              (newArgs.length > 5 ? newArgs[5] : ""),
              (newArgs.length > 6 ? newArgs[6] : ""),
              (newArgs.length > 7 ? newArgs[7] : ""),
              (newArgs.length > 8 ? newArgs[8] : ""),
              (newArgs.length > 9 ? newArgs[9] : ""),
              "");
          st.execute(d.getInsertSQL(env.getDBGateway().getOwner(),env.getDBGateway().getSQLSyntaxEngine()));
        }
        line++;
        if(line % 50 == 0) System.out.println("   Processing line: " + line);
      }
      sp.closeFile();
      con.commit();
      st.close();
      con.close();      
      System.out.println("Done uploading " + inputFileName);
      return new ABCMResult("Successfully uploaded narrative", newNarrID, true);
    }    
    catch(Exception e){
      try{
        con.rollback();
      }
      catch(Exception e2){}
      sp.closeFile();
      System.err.println("Could not upload narrative :" + e.getMessage());
      return new ABCMResult("Could not upload narrative: " + e.getMessage(), false);
    }  
  }
 
}
