package abcm.model.tools;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.OutputStream;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.Vector;

import abcm.math.SimDate;
import abcm.model.ABCMResult;
import abcm.model.AbstractABCModelEnvironment;
import abcm.utilities.ScriptReader;
import abcm.utilities.TextFormatter;

/**
 * Provides methods that manage the library of scripts that reside in the directory
 * specified by the Model Environment. These methods include delete, copy, and save,
 * as well as some interaction with the database.
 * <p>
 * This class deals with four distinct file types: scripts, demos, scripted narratives,
 * and compiled narratives. See the ABCM documentation for details.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ScriptLibrarian {

  /** File extension used to identify script files  */
  private String script_extension;             
  /** File extension used to identify narrative files  */
  private String narrative_extension;          
  /** File extension used to identify compiled narrative files */
  private String compiled_narrative_extension; 
  
  /** Line used in scripts to indicate demo steps */
  public static final String DEMO_BREAK                   = "** <Demo_Break>"; 
  
  /** Programmatic environment in which this librarian works  */
  private AbstractABCModelEnvironment env;
  /** Most recent error encountered by this object  */
  private String errorReport = "";
  
  
  /**
   * No-argument constructor.
   * @param E Model Environment in which this librarian is to operate
   */
  public ScriptLibrarian(AbstractABCModelEnvironment E){
    env = E;
    script_extension             = E.getScriptExtension();
    narrative_extension          = E.getNarrativeExtension();
    compiled_narrative_extension = E.getCompiledNarrativeExtension();    
  }

  /* Internal Retrieval */
  
  /**
   * Gets a Hashtable that lists all the demos.
   * A demo is a file like "MyDemo.abcm" that is
   * accompanied by files like "MyDemo.1.abcm",
   * "MyDemo.2.abcm" etc.
   * @return the Hashtable of demo files
   */
  public Hashtable<String, String> getDemos(){
    Hashtable<String, String> demos = new Hashtable<String, String>();
    String dir = env.getScriptDir();
    File   directory = new File(dir);
    File[] files = directory.listFiles();
    if(files != null){
      for(int i = 0; i <files.length; i++){
        String fileName = files[i].getName();
        if(fileName.matches("\\w*\\.\\d*\\" + script_extension + "\\z")){
          String shortName = fileName.split("\\.\\d*\\" + script_extension + "\\z")[0];
          demos.put(shortName, shortName);
        }
      }
    }
    return demos;
  }
  
  /* Make catalog available externally*/
  
    /* Scripts */
  
  /**
   * Returns a list of script files in the library as an HTML option list.
   * Scripts are all files with .abcm extension except
   * those that are demos (A demo is a file like 
   * "MyDemo.abcm" that is accompanied by files like 
   * "MyDemo.1.abcm", "MyDemo.2.abcm" etc.)
   * @return the list of script files as an option list
   */
  public String getListOfScriptsAsOptionList(){
    String dir = env.getScriptDir();
    File   directory = new File(dir);
    File[] files = directory.listFiles();
    TreeSet ts = new TreeSet();
    
    if(files != null){
      Hashtable demos = getDemos();
      for(int i = 0; i <files.length; i++){
        String fileName = files[i].getName();
        if(fileName.matches(".*\\" + script_extension + "\\z")){
          String shortName = fileName.split("\\.\\d*\\" + script_extension + "\\z")[0];
          shortName = shortName.split("\\" + script_extension + "\\z")[0];
          if(!demos.containsKey(shortName)){
            
            ts.add(shortName);
          }
        }
      }
    }
    // Sort these
    String options = "";
    Iterator I = ts.iterator();
    while(I.hasNext()){
      options += "<option>" + (String)I.next() + "</option>";
    }
    return options;
  }

    /* Demos */
 
  /**
   * Returns a list of the demo files in the library as an HTML option list.
   * A demo is a file like "MyDemo.abcm" that is accompanied by files like 
   * "MyDemo.1.abcm", "MyDemo.2.abcm" etc. The list includes just the short
   * names, without the numerals or extensions.
   * @return the list of demo files as an HTML option list
   */
  public String getListOfDemosAsOptionList(){
    String options = "";
    Hashtable demos = getDemos();
    Enumeration e = demos.elements();
    // Sort them (Hashtable is unsorted)
    TreeSet ts = new TreeSet();
    while(e.hasMoreElements()){
      ts.add(e.nextElement());
    }
    Iterator I = ts.iterator();
    while(I.hasNext()){
      options += "<option>" + (String)I.next() + "</option>";
    }
    return options;
  }
  
    /* Scripted Narratives */

  /**
   * Gets a list of scripted narratives in the library as an HTML option list.
   * A scripted narrative is a file with an '.nar' extension.
   * @return the list of scripted narratives as an HTML option list
   */
  public String getListOfScriptedNarrativesAsOptionList(){
    String dir = env.getScriptDir();
    File   directory = new File(dir);
    File[] files = directory.listFiles();
    String options = "";
    if(files != null){
      for(int i = 0; i <files.length; i++){
        String fileName = files[i].getName();
        if(fileName.matches(".*\\" + narrative_extension + "\\z")){
          String shortName = TextFormatter.getFileWithoutExtension(fileName);
          options +="<option>" + shortName + "</option>";
        }
      }
    }
    return options;
  }
  
    /* Compiled Narratives */
 
  /**
   * Gets a list of compiled narratives in the library as an HTML option list.
   * A compiled narrative is a file with an '.nar' extension.
   * @return the list of compiled narratives as an HTML option list
   */
  public String getListOfCompiledNarrativesAsOptionList(){
    String dir = env.getScriptDir();
    File   directory = new File(dir);
    File[] files = directory.listFiles();
    String options = "";
    if(files != null){
      for(int i = 0; i <files.length; i++){
        String fileName = files[i].getName();
        if(fileName.matches(".*\\" + compiled_narrative_extension + "\\z")){
          String shortName = TextFormatter.getFileWithoutExtension(fileName);
          options +="<option>" + shortName + "</option>";
        }
      }
    }
    return options;
  }
 

  /* Actions */
  
    /* Save */
  
  /**
   * Saves a script or a demo to the library. The script is specified by the text passed; the file name
   * is passed, and the file created will overwrite any previous file with that name.
   * If the 'saveAsDemo' parameter is true, the text will be parsed to find 'demo break' tags; these
   * will be used to split the text into different demo steps. A single file will be created that
   * stores the whole script, but a collection of files, each representing a separate demo step,
   * will also be created (hence 'Test.abcm' plus 'Test.1.abcm', 'Test.2.abcm', etc.)
   * Note that if the saveAsDemo parameter is true the routine assumes that the filename passed has no
   * extension, and adds it to both the main demo script and all the steps; if the parameter is
   * false, the file is saved with the name exactly as specified- no extension is added.
   * @param scriptText text of the script to be saved
   * @param fileToSaveName name of the file to be created; should not include extension
   * @param saveAsDemo true if the text should be parsed and separate demo steps created, false if not
   * @return true if all files are saved successfully, false otherwise
   */
  public boolean saveToScriptLibrary(String scriptText, String fileToSaveName, boolean saveAsDemo){
    
    class ScriptToSave{
      String fileName;
      StringWriter text = new StringWriter();
      
      /**
       * Constructor that sets name of script to save.
       * @param name name of script to be saved
       */
      public ScriptToSave(String name){
        fileName = name;
      }
      
      public void addText(String textToAdd){
        text.write(textToAdd);
      }
      
      public boolean save(){
         File F = new File(env.getScriptDir() + fileName);
         try{
           FileWriter FW = new FileWriter(F);
           FW.write(text.toString());
           FW.close();
         }
         catch(Exception E){
           errorReport = "Could not save script due to exception " + E.getMessage();
           return false;
         }
         return true;
      }
    }

    if(scriptText == null){
      errorReport = "Script text passed was null; cannot save.";
      return false;
    }
    if(scriptText.length() == 0){
      errorReport = "Script text passed was empty String; cannot save.";
      return false;
    }        
    
    if(saveAsDemo) deleteDemo(fileToSaveName);  // This will overwrite anything previously saved
    else           deleteFile(fileToSaveName);
    
    Vector filesToSave = new Vector();
    
    ScriptToSave mainScript = new ScriptToSave(fileToSaveName + (saveAsDemo ? script_extension : ""));
    filesToSave.add(mainScript);
    
    String[] lines = scriptText.split("\n");
//    String[] lines = scriptText.split("^."); // Should split on line separators...
    int demoStep = 1;
    ScriptToSave demoPart = new ScriptToSave(fileToSaveName + "." + demoStep + script_extension);
    if(saveAsDemo){      
      filesToSave.add(demoPart);
    }
    for(int i = 0; i < lines.length; i++){
      String in = lines[i];
      
      if(in.contains(DEMO_BREAK)){
        if(saveAsDemo){
          demoStep++;
          demoPart = new ScriptToSave(fileToSaveName + "." + demoStep + script_extension);
          filesToSave.add(demoPart);
        }
      }
      else{
        if(saveAsDemo) demoPart.addText(in + "\n");  // Note that you don't want the 'Demo Break' lines in the demos
      }
      mainScript.addText(in + "\n");                 // ... but you do want them here
    }   
    boolean retVal = true;
    
    // Save all the files created
    for(int i = 0; i < filesToSave.size(); i++){
      retVal = retVal && ((ScriptToSave)filesToSave.get(i)).save();
    }
    
    return retVal;
  }

  /**
   * Saves the specified text to a script file with the given name. Will add script suffix to filename if it is not present.
   * @param text text to be written
   * @param file name of file to be created.
   * @return true if successful, false if not
   */
  public boolean saveScript(String text, String file){
    String fileName = file;
    if(!fileName.matches(".*\\" + script_extension + "\\z")) fileName += script_extension;
    return saveToScriptLibrary(text, fileName, false);
  }
  
  /**
   * Saves the specified text to a demo file with the given name. Will add demo suffix to filename if it is not present.
   * @param text text to be written
   * @param file name of file to be created.
   * @return true if successful, false if not
   */
  public boolean saveDemo(String text, String file){
    String fileName = file;
    if(fileName.matches(".*\\" + script_extension + "\\z")) fileName = TextFormatter.getFileWithoutExtension(fileName);
    return saveToScriptLibrary(text, fileName, true);
  }
  
  /**
   * Saves the specified text to a narrative file with the given name. Will add narrative suffix to filename if it is not present.
   * @param text text to be written
   * @param file name of file to be created.
   * @return true if successful, false if not
   */
  public boolean saveNarrative(String text, String file){
    String fileName = file;
    if(!fileName.matches(".*\\" + narrative_extension + "\\z")) fileName += narrative_extension;
    return saveToScriptLibrary(text, fileName, false);
  }
  
  /**
   * Saves the specified text to a compiled narrative file with the given name. Will add compiled narrative suffix to filename if it is not present.
   * @param text text to be written
   * @param file name of file to be created.
   * @return true if successful, false if not
   */
  public boolean saveCompiledNarrative(String text, String file){
    String fileName = file;
    if(!fileName.matches(".*\\" + compiled_narrative_extension + "\\z")) fileName += compiled_narrative_extension;
    return saveToScriptLibrary(text, fileName, false);
  }
  
    /* Retrieve */
  
  
  /**
   * Retrieves a StringWriter containing the text of the specified file.
   * @param fileName name of the file to be read
   * @return text of the file
   */
  public StringWriter getFile(String fileName){
    StringWriter ret = new StringWriter();
    
    try{
      if(!env.scriptFileExists(fileName)){
        errorReport = "No file with specified script name exists; cannot retrieve.";
        return null;
      }
//      ScriptReader sr = new ScriptReader(env.getScriptDir() + fileName);
//      while(!sr.EOF()){
//        ret.write(sr.getNextLine() + "\n");
////        String nextLine = sr.getNextLine();
////        if(nextLine != null)  ret.write(nextLine + "\n");
//      }
      FileReader fr = new FileReader(env.getScriptDir() + fileName);
      while(fr.ready()) ret.write(fr.read());
      fr.close();
    }
    catch(Exception E){
      errorReport = "Error retrieving file: " + E.getMessage();
      return null;
    }
    return ret;
  }
  
  /**
   * Retrieves the text of the specified script.
   * @param file name of the script file to be read. Will add the script file name suffix if it is not present.
   * @return the text of the file
   */
  public StringWriter getScript(String file){
    String fileName = file;
    if(!fileName.matches(".*\\" + script_extension + "\\z")) fileName += script_extension;
    return getFile(fileName);             
  }
  
  /**
   * Retrieves the text of the specified demo.
   * @param file name of the demo file to be read. Will add the demo file name suffix if it is not present.
   * @return the text of the file
   */
  public StringWriter getDemo(String file){
    String fileName = file;
    if(!fileName.matches(".*\\" + script_extension + "\\z")) fileName += script_extension;
    return getFile(fileName);             
  }
  
  /**
   * Retrieves the text of the specified narrative.
   * @param file name of the narrative file to be read. Will add the narrative file name suffix if it is not present.
   * @return the text of the file
   */
  public StringWriter getNarrative(String file){         
    String fileName = file;
    if(!fileName.matches(".*\\" + narrative_extension + "\\z")) fileName += narrative_extension;
    return getFile(fileName);          
  }
  
  /**
   * Retrieves the text of the specified compiled narrative.
   * @param file name of the compiled narrative file to be read. Will add the compiled narrative file name suffix if it is not present.
   * @return the text of the file
   */
  public StringWriter getCompiledNarrative(String file){ 
    String fileName = file;
    if(!fileName.matches(".*\\" + compiled_narrative_extension + "\\z")) fileName += compiled_narrative_extension;
    return getFile(fileName); 
  }
  
    /* Delete */
  
  /**
   * Deletes a file from the script library. The file to be deleted is passed
   * explicitly, including the full filename (but not pathname, of course); other
   * methods of this object provide the file extensions as needed, but this one
   * takes the full name including extension.
   * @param fileName the file to be deleted
   * @return true if the delete is successful, false otherwise
   */
  public boolean deleteFile(String fileName){
    if(!env.scriptFileExists(fileName)){
      errorReport = "Could not delete file; no file named '" + fileName + "' exists.";
      return false;
    }
    String dir  = env.getScriptDir();
    File   file = new File(dir + fileName);
    return file.delete();
  }
  
  /**
   * Deletes the specified file from the library directory.
   * @param scriptFileName the name of the script to be deleted; should not include extension, 
   * but if it is included it will be stripped
   * @return true if successful, false otherwise
   */
  public boolean deleteScript(String scriptFileName){
    String fileName = scriptFileName;
    if(!fileName.matches(".*\\" + script_extension + "\\z")) fileName += script_extension;
    return deleteFile(fileName);
  }
  
  /**
   *  Deletes the demo file and all its member files (i.e.
   *  MyDemo.abcm and MyDemo.1.abcm, MyDemo.2.abcm, etc.)
   *  @param scriptDemoName the base name of the demo to be deleted; should not include any extension or demo step
   *  (hence 'MyDemo' not 'MyDemo.abcm' or 'MyDemo.1' or 'MyDemo.1.abcm'
   *  @return true if the delete of the base file was successful, false otherwise
   */
  public boolean deleteDemo(String scriptDemoName){
    String dir = env.getScriptDir();
    File   directory = new File(dir);
    File[] files = directory.listFiles();
    if(files != null){
      for(int i = 0; i <files.length; i++){
        String fileName = files[i].getName();
        if(fileName.matches("\\A" + scriptDemoName + "\\.\\d*\\" + script_extension + "\\z")){
          File tbd = new File(dir + fileName);
          tbd.delete();
        }
      }
    }
    return deleteFile(scriptDemoName + script_extension);
  }
  
  /**
   *  Deletes a scripted narrative file from the library.
   *  @param scriptedNarrativeFileName the name of the scripted narrative to be deleted; should not include
   *  file extension, but if it is included it will be stripped
   *  @return true if successful, false otherwise  
   */
  public boolean deleteNarrative(String scriptedNarrativeFileName){
    String fileName = scriptedNarrativeFileName;
    if(!fileName.matches(".*\\" + narrative_extension + "\\z")) fileName += narrative_extension;
    return deleteFile(fileName);
  }

  /**
   *  Deletes a compiled narrative file from the library.
   *  @param compiledNarrativeFileName the name of the compiled narrative to be deleted; should not include
   *  file extension, but if it is included it will be stripped
   *  @return true if successful, false otherwise  
   */  
  public boolean deleteCompiledNarrative(String compiledNarrativeFileName){
    String fileName = compiledNarrativeFileName;
    if(!fileName.matches(".*\\" + compiled_narrative_extension + "\\z")) fileName += compiled_narrative_extension;
    return deleteFile(fileName);
  }

    /* Copy */
  
  /**
   * Creates a new copy of the specified file.
   * @param fileName name of the file to be copied
   * @param copyName name to be assigned to the new copy
   * @return true if successful, false otherwise
   */
  public boolean copyFile(String fileName, String copyName){
    if(!env.scriptFileExists(fileName)){
      errorReport = "No file with specified name exists; cannot copy.";
      return false;
    }
    if(env.scriptFileExists(copyName)){
      errorReport = "File with specified destination name already exists; cannot overwrite. Delete file first.";
      return false;
    }
    
    try{
      OutputStream out = new FileOutputStream(new File(env.getScriptDir() + copyName));
      ScriptReader sr = new ScriptReader(env.getScriptDir() + fileName);
      while(!sr.EOF()){
        String nextLine = sr.getNextLine();
        if(nextLine != null && nextLine.length() > 0){
          out.write(nextLine.getBytes());
          out.write('\n');
        }
      }      
      out.close();
    }
    catch(Exception E){
      errorReport = "Error copying file: " + E.getMessage();
      return false;
    }
    return true;
  }
  
  /**
   * Makes a copy of a script in the library. Note that this is the only way to rename
   * files or change their types (by changing the file extension). In the unlikely event that
   * a user has a .nar file and needs to save it, she can save it as XYZ.abcm (the extension
   * will be added as a default), then copy it as XYZ.nar, and finally delete the XYZ.abcm file.
   * @param scriptName the name of the script to be copied. Should include any file extension.
   * @param copyName the new name to be assigned to the copy. Should include any file extension.
   * @return true if successful, false otherwise
   */
  public boolean copyScript(String scriptName, String copyName){
    String fileName = scriptName;
    String newName  = copyName; 
    if(!fileName.matches(".*\\" + script_extension + "\\z")) fileName += script_extension;
    if(!newName.matches(".*\\" + script_extension + "\\z")) newName += script_extension;
    return copyFile(fileName, newName);
  }
  
  /**
   * Makes a copy of a demo in the library. Includes the base file (for example, 'MyDemo.abcm')
   * and all demo step files (like 'MyDemo.1.abcm', 'MyDemo.2.abcm').
   * @param demoName the name of the demo to be copied; should not include any demo step indicator nor
   * the file extension (thus 'MyDemo' and not 'MyDemo.abcm' or 'MyDemo.1.abcm')
   * @param newDemoName the new name to be assigned to this copy
   * @return true if the copy is successful, false otherwise
   */
  public boolean copyDemo(String demoName, String newDemoName){
    copyFile(demoName + script_extension, newDemoName + script_extension);
    int i = 1;
    while(env.scriptFileExists(demoName + "." + i + script_extension)){
      copyScript(demoName + "." + i + script_extension, newDemoName + "." + i + script_extension);
      i++;
    }
    return true;
  }
  
  /**
   * Makes a copy of a scripted narrative in the library.
   * @param scriptedNarrativeName the name of the scripted narrative to be copied. Should include any file extension.
   * @param copyName the new name to be assigned to the copy. Should include any file extension.
   * @return true if successful, false otherwise
   */
  public boolean copyNarrative(String scriptedNarrativeName, String copyName){ 
    String fileName = scriptedNarrativeName;
    String newName  = copyName; 
    if(!fileName.matches(".*\\" + narrative_extension + "\\z")) fileName += narrative_extension;
    if(!newName.matches(".*\\" + narrative_extension + "\\z")) newName += narrative_extension;
    return copyFile(fileName, newName); 
  }
  
  /**
   * Makes a copy of a compiled narrative in the library.
   * @param compiledNarrativeName the name of the compiled narrative to be copied. Should include any file extension.
   * @param copyName the new name to be assigned to the copy. Should include any file extension.
   * @return true if successful, false otherwise
   */
  public boolean copyCompiledNarrative(String compiledNarrativeName, String copyName){
    String fileName = compiledNarrativeName;
    String newName  = copyName; 
    if(!fileName.matches(".*\\" + compiled_narrative_extension + "\\z")) fileName += compiled_narrative_extension;
    if(!newName.matches(".*\\" + compiled_narrative_extension + "\\z")) newName += compiled_narrative_extension;
    return copyFile(fileName, newName); 
  }
  
    /* Convert */
  
  /**
   * Converts a script file into a scripted narrative. See {@link ScriptToNarrative} for details.
   * Will fail if the input file does not exist or if the output file
   * does exist (you should delete any existing file first; it will not be
   * overwritten).
   * @param inputFileName name of the file to be converted (without extension)
   * @param outputFileName name of the file to be created (without extension)
   * @param baseDate base date to be used for this narrative
   * @return true if successful; false otherwise
   */
  public boolean convertScriptToNarrative(String inputFileName, String outputFileName, String baseDate){
    String inFile  = inputFileName;
    String outFile = outputFileName;
    
    if(!inFile.matches(".*\\" + script_extension + "\\z")) inFile += script_extension;
    if(!outFile.matches(".*\\" + narrative_extension + "\\z")) outFile += narrative_extension;
    
    if(!env.scriptFileExists(inFile)){
      errorReport = "Script file specified does not exist; cannot make narrative.";
      return false;
    }
    if(env.scriptFileExists(outFile)){
      errorReport = "Narrative with destination name already exists; cannot overwrite. Delete file first.";
      return false;
    }    
    
    StringWriter narr = ScriptToNarrative.scriptToNarrative(env.getScriptDir() + inFile, baseDate);
    if(narr==null) {
      errorReport= "No narrative could be created";
      return false;
    }
    File F = new File(env.getScriptDir() + outFile);
    try{
      FileWriter FW = new FileWriter(F);
      FW.write(narr.toString());
      FW.close();
    }
    catch(Exception E){
      errorReport = "Exception occurred while translating script to narrative: " + E.getMessage();
      return false;
    }
    return true;
  }
   
  /**
   * Converts a narrative file into a script file. See {@link NarrativeToScript} for details.
   * Will fail if the input file does not exist or if the output file
   * does exist (you should delete any existing file first; it will not be
   * overwritten).
   * @param inputFileName the name of the file to be converted (without extension)
   * @param outputFileName the name of the file to be created (without extension)
   * @return true if successful, false otherwise
   */
  public boolean convertNarrativeToScript(String inputFileName, String outputFileName){
    String inFile  = inputFileName;
    String outFile = outputFileName;
    
    if(!inFile.matches(".*\\" + narrative_extension + "\\z")) inFile += narrative_extension;
    if(!outFile.matches(".*\\" + script_extension + "\\z")) outFile += script_extension;
  
    if(!env.scriptFileExists(inFile)){
      errorReport = "File '" + inFile + "' does not exist.";
      return false;
    }
    if(env.scriptFileExists(outFile)){
      errorReport = "File '" + outFile + "' already exists; cannot overwrite. Delete first.";
      return false;
    }
    
    
    System.out.println("CONVERTING: " + inFile + " to " + outFile);
    
    StringWriter script = NarrativeToScript.narrativeToScript(env.getScriptDir() + inFile);
    if(script==null) {
      errorReport = "No script could be created";
      return false;
    }
    File F = new File(env.getScriptDir() + outFile);
    try{
      FileWriter FW = new FileWriter(F);
      FW.write(script.toString());
      FW.close();
     }
    catch(Exception E){
      errorReport = "An exception occurred: " + E.getMessage();
      return false;
    }
    return true;
  }

  /**
   * Compiles a narrative from a version in which English commands are used and comments
   * may be included to one that uses numerical commands and omits all comments. The resulting file
   * is in a format ready to be uploaded and stored in an ABCM database.
   * @param narrativeName name of the original narrative file
   * @param outputFileName name of the destination file
   * @return true if successful, false otherwise
   */
  public boolean compileNarrative(String narrativeName, String outputFileName){
    String narrName = narrativeName;
    String outFile = outputFileName;
    
    if(!narrName.matches(".*\\" + narrative_extension + "\\z")) narrName += narrative_extension;
    if(!outFile.matches(".*\\" + compiled_narrative_extension + "\\z")) outFile += compiled_narrative_extension;
    
    
    if(!env.scriptFileExists(narrName)){
      errorReport = "No narrative '" + narrName + "' exists.";
      return false;
    }
    if(env.scriptFileExists(outFile)){
      errorReport = "Specified destination file '" + outFile + "' already exists. Delete first.";
      return false;
    }
    
    
    StringWriter nru = NarrativeCompiler.compileNarrative(env.getScriptDir() + narrName, env);
    if(nru==null) {
      errorReport="No nru could be created";
      return false;
    }
    File F = new File(env.getScriptDir() + outFile);
    try{
      FileWriter FW = new FileWriter(F);
      FW.write(nru.toString());
      FW.close();
     }
    catch(Exception E){
      errorReport="Error compiling narrative: " + E.getMessage();
      return false;
    }
    return true;
  }
  
  /**
   * Converts a compiled narrative, in which commands are stored numerically, into
   * a decompiled one, in which English commands are used.
   * @param narrativeName name of the compiled narrative file
   * @param outputFileName file to which the decompiled narrative is to be stored
   * @return true if successful, false otherwise
   */
  public boolean decompileNarrative(String narrativeName, String outputFileName){
    String narrName = narrativeName;
    String outFile = outputFileName;
    
    if(!narrName.matches(".*\\" + compiled_narrative_extension + "\\z")) narrName += compiled_narrative_extension;
    if(!outFile.matches(".*\\" + narrative_extension + "\\z")) outFile += narrative_extension;

    if(!env.scriptFileExists(narrName)){
      errorReport = "No narrative '" + narrName + "' exists.";
      return false;
    }
    if(env.scriptFileExists(outFile)){
      errorReport = "Specified destination file '" + outFile + "' already exists. Delete first.";
      return false;
    }    
    
    StringWriter nar = NarrativeDecompiler.decompileNarrative(env.getScriptDir() + narrName, env);
    
    if(nar==null) {
      errorReport="No nru could be created";
      return false;
    }
    File F = new File(env.getScriptDir() + outFile);
    try{
      FileWriter FW = new FileWriter(F);
      FW.write(nar.toString());
      FW.close();
     }
    catch(Exception E){
      errorReport="Error decompiling narrative: " + E.getMessage();
      return false;
    }
    return true;
  }

    /* Upload */  
  
  
  /**
   * Uploads a compiled narrative to an ABCM database.
   * @param narrativeName name of the compiled narrative file. If the compiled narrative file extension is not included, it will be added
   * @param newName name of narrative to be st ored in the database; must not duplicate any extant narrative.
   * @param desc description of this narrative
   * @param baseDate based date to be stored for this narrative
   * @param isSeasonal true if the narrative is seasonal, false otherwise
   * @return true if successful, false otherwise
   */
  public ABCMResult uploadCompiledNarrative(String narrativeName, String newName, String desc, 
      String baseDate, boolean isSeasonal){
    String narr = narrativeName;
    if(!narr.matches(".*\\" + narrative_extension + "\\z")) narr += compiled_narrative_extension;
    
    if(!env.scriptFileExists(narr)){
      errorReport = "File '" + narr + "' does not exist.";
      return new ABCMResult(errorReport, false);
    }
    if(!env.isAuthorized()){
      errorReport = "User is not authorized; log in first.";
      return new ABCMResult(errorReport, false);
    }
    if(SimDate.parseString(baseDate) < 0 ){
      errorReport = "Specified base date '" + baseDate + "' is invalid.";
      return new ABCMResult(errorReport, false);
    }
    ABCMResult result = NarrativeUploader.uploadNarrative(env.getScriptDir() + narr, newName, desc, baseDate, isSeasonal, env);
    if(!result.success) errorReport = result.text;
    return result;
  }
  
  
  
  
  /* Error reporting */
  
  /**
   * Gets the error report string for this script librarian. This will have a description
   * of the last error that occurred (i.e. a file that could not be created).
   * @return the current error report string
   */
  public String getErrorReport(){ return errorReport; }
  
}
