package abcm.model.tools;

import java.io.StringWriter;

import abcm.math.SimDate;
import abcm.model.AbstractABCMParser;
import abcm.utilities.QuotedLineParser;
import abcm.utilities.ScriptReader;

/**
 *  Converts a script file to a narrative file. To do this
 *  requires an input file name and the Base Date from which the
 *  script is said to commence. 
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class ScriptToNarrative {
  
  /**
   * A string explaining the usage of this class 
   */
  private static String helpString = "Usage: Transform a script to a narrative." +
      "\n\n" +
      "Arguments:\n\n" +
      "1: Input file name.\n" +
      "2: Base Date, in ABCM Date format.";
  
  /**
   * Given an input file name and a base date, converts the script file stored in the input
   * file to a narrative, beginning on the specified base date.
   * @param inputFileName name of input file
   * @param baseDate base date to be used
   * @return StringWriter containing the text of the narrative that results
   */
  public static StringWriter scriptToNarrative(String inputFileName, String baseDate){
    System.out.println("Processing: " + inputFileName);
    double currentDate = SimDate.parseString(baseDate);
    
    if(currentDate == -1) {
      System.err.println("Could not complete: unrecognizable base date '" + baseDate + "'");
      return null;  
    }

    StringWriter ret = new StringWriter();
    
    ScriptReader sp = new ScriptReader(inputFileName);

    while(!sp.EOF()){
      String nextLine = sp.getNextLine();
      if(nextLine!=null){
        nextLine = nextLine.replaceAll("\t"," ");
        if(nextLine.length() > 0 && !nextLine.startsWith("**")){
          String[] inputLine = QuotedLineParser.parseQuoted(nextLine, AbstractABCMParser.MAX_ARGS + 2);
          if(inputLine[0].toUpperCase().compareTo("STEP") == 0) currentDate++;
          else if(inputLine[0].toUpperCase().compareTo("STEPDAYS") == 0) currentDate += Long.parseLong(inputLine[1]);
          else if(inputLine[0].toUpperCase().compareTo("STEPTODATE") == 0) currentDate = SimDate.parseString(inputLine[1]);
          else ret.write(SimDate.convertToString(currentDate) + "\t" + nextLine.replaceAll(" +","\t") + "\n");
        }
      }
    }
    return ret;
  }

  /**
   * Constructor; private scope prevents this class from being instantiated or extended.
   */
  private ScriptToNarrative(){}
  
  /**
   * Runs this application.
   * @param args a list of arguments for the application. If there
   * are fewer than 2 arguments, or if the first argument is '-h',
   * prints a help string. 
   */
  public static void main(String[] args){
    boolean isSeasonal = false;
    
    if(args.length < 2 || args[0].startsWith("-h")){
      System.out.println(helpString);
      return;
    }
    String inputFileName = args[0];
    String baseDate = args[1];
    
    System.err.println("Transforming: '" + inputFileName + "' into narrative with base date: " + baseDate);
    
    System.out.println(scriptToNarrative(inputFileName, baseDate).toString());
  }
}

