package abcm.sim;

/**
 * An object implementing the HWMGUI interface provides a standard
 * set of methods that any HWMGUI must offer. These include creating
 * and manipulating data series graphs and disposing/resetting the model.
 * <br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public interface ABCMGenericGUI {

  /**
   * Creates a data series graph with the specified name.
   * The implementation is responsible for how it handles
   * name conflicts.
   * @param graphName name of the new data series graph
   * @return a String indicating the success or failure of this operation
   */
  public String createDataSeriesGraph(String graphName);
  
  /**
   * Loads a data series from the HWM Database into the specified
   * graph.  
   * @param RunID the run number for the required series
   * @param SeriesName the name of the required data series
   * @param GraphName the name of the graph to which the series should
   * be added
   * @return a String indicating the success or failure of the operation
   */
  public String loadDataSeriesFromDB(long RunID, String SeriesName, String GraphName);
  
  /**
   * Lists the graphs currently resident in the GUI.
   * @return a String containing the list of graphs
   */
  public String listGraphs();
  
  /**
   * Disposes and resets the current model, and discards all series and
   * data series graphs in residence. The implementation may wish to
   * provide an opportunity to confirm the operation.
   * @return a String indicating the success of the operation
   */
  public String disposeAndReset();

  /**
   * Updates the GUI. Implementation-specific.
   */
  public void update();
  
  /**
   * Display the graph specified.
   * @param graphName the name of the graph to be displayed
   * @return a String providing information that may be required for display
   */
  public String showDataSeriesGraph(String graphName);

  /**
   * Add the specified series to the specified graph
   * @param graphName the name of the graph to which the series should be added
   * @param sds the Simulation Data Series to be added
   * @param seriesName name to be assigned to this series in the graph
   * @return a String providing information that may be required for display
   */
  public String addSeriesToGraph(String graphName, SimulationDataSeries sds, String seriesName);
  
}
