package abcm.sim;

import abcm.cloning.ABCMCloneStore;
import abcm.math.BasicRNG;

/**
 * The <code>ModelComponent</code> object contains three fundamental
 * container objects that should be used in common across all model objects
 * within a given model environment: a <code>CodeModel</code> Object, a <code>ParameterSet</code>,
 * and a Basic Random Number Generator (<code>BasicRNG</code>).
 * 
 * The <code>ModelComponent</code> object contains these as static variables, not instance
 * variables, so there is exactly one instance of each object that is shared in common
 * with all subclasses.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractModelComponent implements Selectable{
  
  /** Code Model that all instances will use  */
  private static CodeModel    C     = new CodeModel();
  /** Parameter Set that all instances will use  */
  private static ParameterSet P     = new ParameterSet();  
  /** Random Number Generator that all instances will use  */
  private static BasicRNG     RNG   = new BasicRNG(BasicRNG.DEFAULT_JAVA_RNG);
  
  /** Flag indicating status as selected **/
  private boolean isSelected = false;
  
  /**
   * Sets the components of this ModelComponent
   * @param c the Code Model
   * @param p the Parameter Set
   * @param rng the Random Number Generator
   */
  public static final void setModelComponent(CodeModel c, ParameterSet p, BasicRNG rng){
    if(c   != null) C   =   c;
    if(p   != null) P   =   p;
    if(rng != null) RNG = rng;
  }
  
  /**
   * Gets the value of the given Code Model Option Domain for the current Code Model
   * @param domain domain to be checked
   * @return value returned by the Code Model
   */
  public int     codeModelOption(String domain) { return C.getCodeModelOption(domain); }
  
  /** 
   * Gets the value of the specified parameter from the current Parameter Set.
   * @param paramName name of the parameter requested
   * @return the value of that parameter
   */
  public double  paramVal(String paramName)     { return P.getParameter(paramName);    }
  
  /**
   * Gets a random number from the current Random Number Generator. Number is uniform
   * distribution between zero and one.
   * @return a randomly generated number
   */
  public double  random()                       { return RNG.nextDouble();             }

  /**
   * Gets the current Code Model
   * @return the current Code Model
   */
  protected CodeModel       getCodeModel() { return C;   }
  
  /**
   * Gets the current Parameter Set
   * @return the current Parameter Set
   */
  protected ParameterSet getParameterSet() { return P;   }
  
  /**
   * Gets the current Random Number Generator
   * @return the current Random Number Generator
   */
  protected BasicRNG              getRNG() { return RNG; }
  
  /**
   * Gets a deep clone of the parameter set
   * @param cloneStore Clone Store to be used
   * @return deep clone of parameter set
   */
  public static ParameterSet getPClone(ABCMCloneStore cloneStore){ 
    return (ParameterSet)P.getDeepClone(cloneStore);
  }
  
  /**
   * Gets a deep clone of the code model
   * @param cloneStore Clone Store to be used
   * @return deep clone of the Code Model
   */
  public static CodeModel getCMClone(ABCMCloneStore cloneStore){
    return (CodeModel)C.getDeepClone(cloneStore);
  }
  
  /**
   * Gets the current RNG. Public access is required because most RNGs are not
   * Cloneable 
   * @return the current RNG
   */
  public static BasicRNG  getCurrentRNG(){ return RNG; } // Public access is required because RNG is not cloneable
  
  /**
   * Gets a count of the parameters in the Parameter Set.
   * This is an indication that this object has not been
   * initialized.
   * @return the count of parameters in the parameter set.
   */
  public static int getCountOfParameters(){
    return P.getCountOfParameters();
  }
  
  
  /* Object selection */
  
  /* (non-Javadoc)
   * @see abcm.sim.Selectable#select()
   */
  public void select(){   isSelected = true; }
  
  /* (non-Javadoc)
   * @see abcm.sim.Selectable#unSelect()
   */
  public void unSelect(){ isSelected = false; }  

  /* (non-Javadoc)
   * @see abcm.sim.Selectable#isSelected()
   */
  public boolean isSelected(){ return isSelected;  }

}
