package abcm.sim;

import java.awt.Color;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Vector;

import abcm.graphics.ColorSequence;
import abcm.math.BasicMath;
import abcm.math.SimDate;
import abcm.math.Year;
import abcm.utilities.XMLAssist;

/**
 * An Analysis consists of a collection of data series that have been grouped
 * together for some purpose. A simple example would be data from a particular
 * probe used for several simulation runs; the analysis would allow side-by-side
 * comparison of the separate runs' results. The Analysis object provides easy
 * means to create, manage, and output such analyses.
 * 
 * Note that many of the instance variables are considered self-explanatory and are not
 * included in JavaDoc.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Analysis {
  

  /** Flag indicating whether output is intended to be read by humans  */
  public static final boolean HUMAN_READABLE = true;
  /** Flag indicating whether output is intended to be read by the R statistical program  */
  public static final boolean R_READABLE     = false;
  
  /** The collection of Series in this Analysis */
  private Hashtable Series;
  
  // Constants for generating SVG Graphs
  
  // In the left margin
  private static final long   SVG_LEFT_MARGIN               = 45;
  private static final long   SVG_Y_AXIS_TIC_WIDTH          = 5;
  private static final long   SVG_Y_AXIS_LABEL_WIDTH        = 20;
  private static final long   SVG_Y_AXIS_LABEL_SPACE        = 2;
  
  // In the right margin
  private static final long   SVG_RIGHT_MARGIN              = 15;
  private static final long   SVG_LEGEND_WIDTH              = 100;
  private static final long   SVG_LEGEND_SPACE              = 25;
  
  // In the top margin
  private static final long   SVG_TOP_MARGIN                = 5;
  private static final long   SVG_HEADER_HEIGHT             = 15;
  private static final long   SVG_HEADER_SPACE              = 2;
  
  // In the bottom margin
  private static final long   SVG_BOTTOM_MARGIN             = 5;
  
  private static final long   SVG_X_AXIS_TIC_HEIGHT         = 5;
  private static final long   SVG_X_AXIS_LABEL_HEIGHT       = 10;
  private static final long   SVG_X_AXIS_LABEL_SPACE        = 2;
  
  // Strokes and other sizes
  private static final long   SVG_AXIS_STROKE_WIDTH         = 2;
  private static final long   SVG_X_AXIS_TIC_STROKE_WIDTH   = 2;
  private static final long   SVG_Y_AXIS_TIC_STROKE_WIDTH   = 2;
  private static final long   SVG_DATA_SERIES_STROKE_WIDTH  = 1;
  private static final long   SVG_DATA_POINT_RADIUS         = 2;
  private static final long   SVG_VERT_HASH_STROKE_WIDTH    = 1;
  private static final long   SVG_HORZ_HASH_STROKE_WIDTH    = 1;
  private static final long   SVG_Y_AXIS_LABEL_HEIGHT       = 10;
  private static final long   SVG_X_AXIS_LABEL_INTERVAL     = 100;
  private static final long   SVG_Y_AXIS_LABEL_INTERVAL     = 100;
  
  // Colors and fonts
  private static final String SVG_AXIS_STROKE_COLOR       = "#000000";
  private static final String SVG_BACKGROUND_COLOR        = "#C0C0C0";
  private static final String SVG_GRAPH_AREA_COLOR        = "#FFFFFF";
  private static final String SVG_HEADER_COLOR            = "#000000";
  private static final String SVG_HEADER_FONT_FAMILY      = "Helvetica";
  private static final String SVG_VERT_HASH_STROKE_COLOR  = "#888888";
  private static final String SVG_HORZ_HASH_STROKE_COLOR  = "#888888";
  private static final String SVG_X_AXIS_FONT_FAMILY      = "Helvetica";
  private static final String SVG_X_AXIS_LABEL_COLOR      = "#000000";
  private static final String SVG_Y_AXIS_FONT_FAMILY      = "Helvetica";
  private static final String SVG_Y_AXIS_LABEL_COLOR      = "#000000";
  private static final String SVG_LEGEND_FONT_FAMILY      = "Helvetica";
  private static final String SVG_LEGEND_LABEL_COLOR      = "#000000";
  
  // Other Constants
  private static final long   SVG_ROUGH_X_TIC_INTERVAL_PX = 150;
  private static final long   SVG_ROUGH_Y_TIC_INTERVAL_PX = 150;
  private static final long   SVG_LEGEND_ENTRY_HEIGHT     = 30;
  private static final long   SVG_LEGEND_ENTRY_SPACE      = 2;
  private static final long   SVG_LEGEND_FONT_HEIGHT      = 10;
  private static final long   SVG_LEGEND_LEFT_MARGIN      = 3;
  private static final long   SVG_LEGEND_BAR_SPACE        = 2;
  private static final long   SVG_LEGEND_BAR_STROKE_WIDTH = 5;

  
  /**
   * Generic constructor
   */
  public Analysis(){
    Series = new Hashtable();
  }
  
  /**
   * Gets the Hashtable containing the list of
   * data series kept by this graph.
   * @return a Hashtable containing the data
   * series stored by the graph
   */
  public Hashtable getSeries(){
    return Series;
  }
  
  
  /**
   * Get a specific data series that is stored by
   * this graph.
   * @param seriesName
   * @return The requested data series
   */
  public SimulationDataSeries getDataSeries(String seriesName){
    if(Series.containsKey(seriesName))
        return (SimulationDataSeries)Series.get(seriesName);
    else
        return null;
  }
  
  /**
   * Adds a series to this graph by passing the series and
   * a name by which it will be referred. Will fail if
   * a series is already contained by this graph with
   * the specified name, or if the user has attempted to
   * add a non-numeric series.
   * @param series the series to be added
   * @param seriesName name to be assigned to this series
   * @return true if successful, false if not
   */
  public boolean addSeries(SimulationDataSeries series, String seriesName){
    if(!series.isNumeric())            return false;
    if(Series.containsKey(seriesName)) return false;
    Series.put(seriesName, series);
    return true;
  }
  
  /**
   * Removes a particular series from this analysis.
   * @param seriesName the series to be removed.
   */
  public void dropSeries(String seriesName){
    if(Series.containsKey(seriesName)) Series.remove(seriesName);
  }
  
  /**
   * Gets the highest value of all the data series in this graph.
   * Returns zero if there are no series in this graph. If there
   * are series but none includes any data, returns the value
   * of the series.getHighestValue() function.
   * @return The highest value of all data series in this graph.
   */
  public double highestValue(){
    if(Series.size() == 0 )   return 0;
    Enumeration e = Series.elements();
    double retVal = ((SimulationDataSeries)e.nextElement()).getHighestValue();
    while(e.hasMoreElements()){
      double val = ((SimulationDataSeries)e.nextElement()).getHighestValue();
      if(val > retVal) retVal = val;
    }
    return retVal;
  }
  
  /**
   * Gets the lowest value of all the data series in this graph.
   * Returns zero if there are no series in this graph. If there
   * are series but none includes any data, returns the value of the
   * series.getLowestValue() function.
   * @return The lowest value of all data series in this graph.
   */
  public double lowestValue(){
    if(Series.size() == 0 )   return 0;
    Enumeration e = Series.elements();
    double retVal = ((SimulationDataSeries)e.nextElement()).getLowestValue();
    while(e.hasMoreElements()){
      double val = ((SimulationDataSeries)e.nextElement()).getLowestValue();
      if(val < retVal) retVal = val;
    }
    return retVal;
  }
  
  /**
   * Returns the lowest simulation date of all series in this graph,
   * as a double. If there are no series in this graph, returns zero. 
   * If there are series but none has any data, returns zero.
   * @return the lowest simulation date of all series in this graph,
   * as a double.
   */
  public double earliestTime(){
    if(Series.size() == 0 ){
      System.out.println("Earliest time SERIES IS EMPTY!!");
      return 0;
    }
    Enumeration e = Series.elements();
    if(!e.hasMoreElements()) return 0;
    Vector v = new Vector();
    while(e.hasMoreElements()){
      SimulationDataSeries sds = (SimulationDataSeries)e.nextElement();
      SimDate sd = sds.getEarliestDate();
      if(sd != null) v.add(sd);
    }
    if(v.size() == 0) return 0;
    SimDate sd = (SimDate)v.remove(0);
    double retVal = sd.value();
    while(v.size() > 0){
      sd = (SimDate)v.remove(0);
      if(sd.value() < retVal) retVal = sd.value();
    }
    System.err.println("Calculated earliest time as: " + retVal);
    return retVal;
  }
  
  /**
   * Returns the latest simulation date of all series in this graph,
   * as a double. If there are no series in this graph, returns zero. 
   * If there are series but none has any data, returns zero.
   * @return the latest simulation date of all series in this graph,
   * as a double.
   */
  public double latestTime(){
    if(Series.size() == 0 ){
      System.out.println("Latest time SERIES IS EMPTY!!");
      return 0;
    }
    Enumeration e = Series.elements();
    if(!e.hasMoreElements()) return 0;
    Vector v = new Vector();
    while(e.hasMoreElements()){
      SimulationDataSeries sds = (SimulationDataSeries)e.nextElement();
      SimDate sd = sds.getLastDate();
      if(sd != null) v.add(sd);
    }
    if(v.size() == 0) return 0;
    SimDate sd = (SimDate)v.remove(0);
    double retVal = sd.value();
    while(v.size() > 0){
      sd = (SimDate)v.remove(0);
      if(sd.value() > retVal) retVal = sd.value();
    }
    System.err.println("Calculated latest time as: " + retVal);
    return retVal;
  }
  
  /**
   * Returns the contents of this Analysis's data series
   * as an HTML Table.
   * @return the HTML table
   */
  public String getAnalysisAsHTMLTable(){
    // Utility String
    String noValue = " - ";
    
    // Create the output writer
    StringWriter out = new StringWriter();
    
    // Get the source data
    Vector  tableHeaders = getTableHeaders();
    TreeMap rows         = getRowsInSeriesTable();
    
    // Write the html table header
    out.write("<table class=\"analysisData\">");
    
    // Write the header row; begin with the left column, Date
    out.write("<tr><th>Date</th>");
    // Loop through the headers
    for(int i = 0; i < tableHeaders.size(); i++) out.write("<th>" + (String)tableHeaders.get(i) + "</th>");
    // Close the header row
    out.write("</tr>");
    
    // Now the data
    // Loop through the rows
    Iterator I = rows.keySet().iterator();
    while(I.hasNext()){
      SimDate rowDate = (SimDate)I.next();
      Double row[] = (Double[]) rows.get(rowDate);
      
      out.write("<tr><td class=\"analysisDate\">" + rowDate.getString() + "</td>");
      
      for(int i = 0; i < row.length; i++)
          out.write("<td class=\"analysisData\">" + (row[i] != null ? ("" + row[i]) : noValue) + "</td>");

      out.write("</tr>");
    }
    out.write("</table>");
    return out.toString();
  }
  
  /**
   * Gets the analysis as an ASCII table.
   * @return the ASCII table
   */
  public String getAnalysisAsASCIITable(){
    return getAnalysisAsASCIITable(HUMAN_READABLE);
  }
  
  /**
   * Gets the analysis as an ASCII table, in one of two forms: a human-readable
   * form in which Simulation Dates are given as Strings, and a second form
   * in which dates are given as double values.
   * @param humanReadable true if the human-readable form is requested,
   * false otherwise
   * @return the ASCII string that presents this table.
   */
  public String getAnalysisAsASCIITable(boolean humanReadable){
    // Utility String
    String noValue = (humanReadable ? " - ": "0");
    
    // Create the output writer
    StringWriter out = new StringWriter();
    
    // Get the source data
    Vector  tableHeaders = getTableHeaders();
    TreeMap rows         = getRowsInSeriesTable();
        
    // Write the header line
    out.write("Date");
    // Loop through the headers for each series
    for(int i = 0; i < tableHeaders.size(); i++) out.write("\t" + (String)tableHeaders.get(i));
    // And finish the line with a return
    out.write("\r\n");
    
    // Now the data
    // Loop through the rows
    Iterator I = rows.keySet().iterator();
    while(I.hasNext()){
      SimDate rowDate = (SimDate)I.next();
      Double row[] = (Double[]) rows.get(rowDate);
      String dateString = (humanReadable ? rowDate.getString(true) : "" + rowDate.value());
      out.write(dateString);
      for(int i = 0; i < row.length; i++) out.write("\t" + (row[i] != null ? ("" + row[i]) : noValue));
      out.write("\r\n");      
    }
    return out.toString();
  }
  
  /**
   * Gets the table headers for this Analysis as an
   * Enumeration.
   * @return the table headers for this Analysis
   */
  public Vector<String> getTableHeaders(){
    return seriesSorted();
  }
  
  /**
   * Gets a TreeMap that includes all of the rows of this Analysis.
   * This is a map of Simulation Dates and Double[] arrays; the arrays
   * have the data in the order returned by getTableHeaders().
   * @return the rows hash table
   */
  public TreeMap getRowsInSeriesTable(){
    Vector headers = getTableHeaders();
    TreeMap rows = new TreeMap();
    int elementCounter = 0;
    for(int j = 0; j < headers.size(); j++){
      String seriesName = (String)headers.get(j);
      // We begin to create the return value, with the headings
      SimulationDataSeries series = (SimulationDataSeries)Series.get(seriesName);
      
      Vector v = series.getSeries();

      for(int i = 0; i < v.size(); i++){
        SimulationDatum sd = (SimulationDatum)v.get(i);
        SimDate date = sd.getSimulationDate();
        Double cols[] = new Double[Series.size()];
        if(!rows.containsKey(date)){
          rows.put(date, cols);
        }
        else {
          cols = (Double[])rows.get(date);
        }
        cols[elementCounter] = new Double(sd.getNumericValue());
      }
      elementCounter++;
    }
    return rows;
  }
  
  /**
   * Gets the analysis as an XML document.
   * @param heading a heading to be prepended to the XML document
   * @return the XML document, as a String
   */
  public String getAsXML(String heading){
    XMLAssist xml = new XMLAssist();
    xml.open("analysis");
      xml.addLine("analysisName", heading);
      double eTime = earliestTime();
      double lTime = latestTime();
      xml.addLine("earliestAnalysisDate", eTime);
      xml.addLine("earliestAnalysisSimDate", SimDate.convertToString(eTime));
      xml.addLine("latestAnalysisDate", lTime);
      xml.addLine("latestAnalysisSimDate",SimDate.convertToString(lTime));
      xml.addLine("lowestAnalysisValue", lowestValue());
      xml.addLine("highestAnalysisValue", highestValue());
      Enumeration e = Series.elements();
      while(e.hasMoreElements()){
        ((SimulationDataSeries)e.nextElement()).getAsXML(xml);
      }      
    xml.closeTag();
    return xml.getXML();
  }
     
  /**
   * Gets the data in this Analysis in an SVG graph using the specified 
   * image height and width and including all the data.
   * @param heading a heading to appear at the top of the graphs
   * @param imageWidth the width of the SVG image, in pixels
   * @param imageHeight the height of the SVG image, in pixels
   * @return the SVG file depicting this image, as a String
   */
  public String getAsSVG(String heading, long imageWidth, long imageHeight){
    return getAsSVG(heading, imageWidth, imageHeight, 0, 0);
    
  }
  
  /**
   * Gets the data in this Analysis in an SVG graph using the specified 
   * image height and width and limiting the data to those that fall
   * between the specified start and end dates.
   * @param heading a heading to appear at the top of the graphs
   * @param imageWidth the width of the SVG image, in pixels
   * @param imageHeight the height of the SVG image, in pixels
   * @param firstDate the date of the earliest data to be placed on the graph
   * @param lastDate the date of the latest data to be placed on the graph
   * @return the SVG file depicting this image, as a String
   */  
  public String getAsSVG(String heading, long imageWidth, long imageHeight, String firstDate, String lastDate){
    return getAsSVG(heading, imageWidth, imageHeight, SimDate.parseString(firstDate), SimDate.parseString(lastDate));
  }
  
  /**
   * Gets the data in this Analysis in an SVG graph using the specified 
   * image height and width and limiting the data to those that fall
   * between the specified start and end dates.
   * @param heading a heading to appear at the top of the graphs
   * @param imageWidth the width of the SVG image, in pixels
   * @param imageHeight the height of the SVG image, in pixels
   * @param firstDate the date of the earliest data to be placed on the graph
   * @param lastDate the date of the latest data to be placed on the graph
   * @return the SVG file depicting this image, as a String
   */  
  public String getAsSVG(String heading, long imageWidth, long imageHeight, double firstDate, double lastDate){
  
    long leftMargin   = SVG_LEFT_MARGIN + 
                        SVG_Y_AXIS_TIC_WIDTH + 
                        SVG_Y_AXIS_LABEL_WIDTH + 
                        SVG_Y_AXIS_LABEL_SPACE;
    
    long rightMargin  = SVG_RIGHT_MARGIN +
                        SVG_LEGEND_WIDTH +
                        SVG_LEGEND_SPACE;
    
    long topMargin    = SVG_TOP_MARGIN + 
                        SVG_HEADER_HEIGHT + 
                        SVG_HEADER_SPACE;
    
    long bottomMargin = SVG_BOTTOM_MARGIN + 
                        SVG_X_AXIS_TIC_HEIGHT +
                        SVG_X_AXIS_LABEL_HEIGHT +
                        SVG_X_AXIS_LABEL_SPACE;
    
    // Adjust to make sure the image is at least the minimum size
    imageWidth  = Math.max(imageWidth, leftMargin + rightMargin + 100);
    imageHeight = Math.max(imageHeight, topMargin + bottomMargin + 100);
    
    ColorSequence cs = new ColorSequence();
    XMLAssist xml = new XMLAssist();
    xml.open("svg", "xmlns:xlink=\"http://www.w3.org/1999/xlink\" " + 
                    "xmlns:svg=\"http://www.w3.org/2000/svg\" " + 
                    "height=\""+ imageHeight + "\" width=\"" + imageWidth + "\" " +
                    "viewBox=\"0 0 " + imageWidth + " " +  imageHeight + "\" preserveAspectRatio=\"none\"");
    
    // Write the Java Script link
    xml.addLine("  <script type=\"text/javascript\" xlink:href=\"graph.js\"/>\n", XMLAssist.INDENT_BLOCK);
    
    // First, provide the gray background
    xml.addLine("<rect x=\"0\" y=\"0\" width=\"" + imageWidth + "\" height=\"" + imageHeight + "\" fill=\"" + SVG_BACKGROUND_COLOR + "\"/>\n", XMLAssist.INDENT_BLOCK);
    
    // Position the header with the main text label
    long topOfHeader     = SVG_TOP_MARGIN;
    long VcenterOfHeader = topOfHeader + SVG_HEADER_HEIGHT/2;
    xml.open("text", "x=\"" + (imageWidth / 2) + "\" " + 
                     "y=\"" + VcenterOfHeader + "\" " +
                     "dominant-baseline=\"mathematical\" " +
                     "font-size=\"" + SVG_HEADER_HEIGHT + "\" " +
                     "font-family=\"" + SVG_HEADER_FONT_FAMILY + "\" " +
                     "stroke=\"" + SVG_HEADER_COLOR + "\" " +
                     "text-anchor=\"middle\"");
    xml.addLine(heading + "\n", XMLAssist.INDENT_BLOCK);
    xml.closeTag();
    
    // 
    long OriginXPosition     = leftMargin;
    long OriginYPosition     = imageHeight - bottomMargin;
      
    long axisHeight          = imageHeight - bottomMargin - topMargin;
    long axisWidth           = imageWidth - leftMargin - rightMargin;
       
    xml.open("g", "transform=\"translate(" + OriginXPosition + "," + OriginYPosition + "), scale(1,-1)\"");
    xml.addLine("<rect x=\"0\" y=\"0\" width=\"" + axisWidth + "\" height=\"" + axisHeight + "\" fill=\"" + SVG_GRAPH_AREA_COLOR + "\"/>\n", XMLAssist.INDENT_BLOCK);
    xml.addLine("<polyline points=\"" + axisWidth + ",0, 0,0, 0," + axisHeight + "\" fill=\"none\" stroke-width=\"" + SVG_AXIS_STROKE_WIDTH + "\" stroke=\"" + SVG_AXIS_STROKE_COLOR + "\"/>\n", XMLAssist.INDENT_BLOCK);

    //  Adjust date ranges; the oddness prevents crossed dates.
    // Understand the range of the data; graph must be within this
    double earliestDataInSeries = earliestTime();
    double latestDataInSeries   = latestTime();
    
    // If the user entered dates that are backwards, switch them.
    if(lastDate < firstDate){
      double newFirstDate = lastDate;
      lastDate = firstDate;
      firstDate = newFirstDate;
    }
    
    
    // If the range selected by the user falls wholly outside the range of the data,
    // just show the whole data set
    if(firstDate > latestDataInSeries || lastDate < earliestDataInSeries){
      firstDate = earliestDataInSeries;
      lastDate  = latestDataInSeries;
    }
    
    // If the user entered equal dates, 
    
    // If the user selected a range that exceeds the range of the data, trim it
    double graphStartDateAndTime = Math.max(earliestDataInSeries, firstDate);
    double graphEndDateAndTime   = Math.min(latestDataInSeries, lastDate);
    
    // If the start and end times are now equal, just go to the end of the data set
    if(graphStartDateAndTime == graphEndDateAndTime) graphEndDateAndTime = latestDataInSeries;
        
    double highestValue = highestValue();
    double lowestValue  = lowestValue();
    
    double dateScale    = graphEndDateAndTime - graphStartDateAndTime;
    double valueScale   = highestValue - lowestValue;
    
    // Add the axes tics at the extremes
    makeXAxisTic(xml, axisWidth);
    makeXAxisTic(xml, 0);
    
    makeYAxisTic(xml, axisHeight);
    makeYAxisTic(xml, 0);
    
    // With labels
    makeXLabel(xml, "" + SimDate.convertToString(graphEndDateAndTime), axisWidth);
    makeXLabel(xml, "" + SimDate.convertToString(graphStartDateAndTime), 0);

    makeYLabel(xml, "" + BasicMath.roundDouble(lowestValue,6), 0);
    makeYLabel(xml, "" + BasicMath.roundDouble(highestValue,6), axisHeight);
    
    // Add some vertical hashes
    // There should be hashes about every 200px in the original image (not zoomed)
    // However, the scale will vary, so some artifice will be required to determine
    // where they should be placed.
    long roughNoHashes = axisWidth/SVG_ROUGH_X_TIC_INTERVAL_PX;
    double duration = graphEndDateAndTime - graphStartDateAndTime;
    double roughHashInterval = duration/roughNoHashes;
    boolean monthly = false;
    if(roughHashInterval < 120) monthly = true;
    boolean trimonthly = false;
    if((roughHashInterval >= 120) && (roughHashInterval <= 400)) trimonthly = true;
    double hashInterval = 365;
    if(roughHashInterval > (3 * 365)   ) hashInterval = 3   * 365;
    if(roughHashInterval > (10 * 365)  ) hashInterval = 10  * 365;
    if(roughHashInterval > (25 * 365)  ) hashInterval = 25  * 365;
    if(roughHashInterval > (50 * 365)  ) hashInterval = 50  * 365;
    if(roughHashInterval > (100 * 365) ) hashInterval = 100 * 365;
    if(roughHashInterval > (200 * 365) ) hashInterval = 200 * 365;
    if(roughHashInterval > (300 * 365) ) hashInterval = 300 * 365;
    if(roughHashInterval > (500 * 365) ) hashInterval = 500 * 365;
    
    double lastHash = 0;
    
    for(double d = graphStartDateAndTime+1; d < graphEndDateAndTime; d++){
      boolean makeHash = false;
      SimDate D = new SimDate(d);
      if(monthly){
        if(Year.dayOfMonth(D.dayOfYear()) == 1) makeHash = true;
      }
      if(trimonthly){
        if(((Year.getMonthInt(D.dayOfYear()) % 3) == 0) && 
             (Year.dayOfMonth(D.dayOfYear()) == 1)) makeHash = true;
      }
      if(!(monthly || trimonthly)){
        if((d - SimDate.parseString("01-JAN-0001 AD")) % hashInterval == 0) makeHash = true;    
      }
      if(makeHash){
        double xPos = ((d - graphStartDateAndTime)/ dateScale) * axisWidth;   
        // Draw the hashmark
        xml.addLine("<polyline points=\"" + xPos + ",0, " + xPos + "," + axisHeight + "\" fill=\"none\" stroke-width=\"" + SVG_VERT_HASH_STROKE_WIDTH + "\" stroke=\"" + SVG_VERT_HASH_STROKE_COLOR + "\"/>\n", XMLAssist.INDENT_BLOCK);
        // Draw the label
        if((xPos - lastHash > SVG_X_AXIS_LABEL_INTERVAL) && ((axisWidth - xPos) > SVG_X_AXIS_LABEL_INTERVAL)){
          makeXLabel(xml, SimDate.convertToString(d), (long)xPos);
          makeXAxisTic(xml, xPos);
          lastHash = xPos;
        }
        
      }    
    }
    
    // Add some horizontal hashes
    // There should be a hash mark every so often, as with the X hashes
    double range = highestValue - lowestValue;
    // This will give the order of magnitude (base 10) of the range
    int magnitude = (int)(Math.floor((Math.log(range)/Math.log(10))));
    int ticCount = 1;
    for(int c = 0; c < 10; c++){
      double val = c * Math.pow(10, magnitude);
        if((val >= lowestValue) && (val <= highestValue)){
          ticCount++;
       }
    }
    roughHashInterval =axisHeight / ticCount;
    double inc = 1;
    while(roughHashInterval > SVG_Y_AXIS_LABEL_INTERVAL){
      roughHashInterval /= 2;
      inc /= 2;
    }
    for(double c = 0; c < 10; c+= inc){
      double val = c * Math.pow(10, magnitude);
        if((val >= lowestValue) && (val <= highestValue)){
          double Y = ((val - lowestValue)/valueScale) * axisHeight;
          xml.addLine("<polyline points=\"0," + Y + ", " + axisWidth + "," + Y + "\" fill=\"none\" stroke-width=\"" + SVG_HORZ_HASH_STROKE_WIDTH + "\" stroke=\"" + SVG_HORZ_HASH_STROKE_COLOR + "\"/>\n", XMLAssist.INDENT_BLOCK);
          makeYAxisTic(xml, Y);
          makeYLabel(xml, "" + BasicMath.roundDouble(val,3), (long)Y);
       }
    }
    
    // Loop through the keys
    double legendEntryNumber = 1;
    String legendEntries = "";
    Vector seriesInOrder = seriesSorted();
    int colorCounter = 0;
    for(int seriesCounter = 0; seriesCounter < seriesInOrder.size(); seriesCounter++){
      String seriesName = (String)seriesInOrder.get(seriesCounter);
      xml.open("g", "id=\"" + seriesName + "\"");
        Color c = cs.getColor(colorCounter);
        String colorString = "rgb(" + c.getRed() + "," + c.getGreen() + "," + c.getBlue() + ")";
        legendEntries += makeLegendEntry(seriesName, legendEntryNumber,colorString);
        legendEntryNumber++;
        SimulationDataSeries sds = (SimulationDataSeries)(Series.get(seriesName));
        if(sds.isNumeric()){
          StringWriter PolyPts = new StringWriter();
          PolyPts.write("<polyline points=\"");
          xml.open("g"); // Series Points
            // Plot the data points; use ColorSequence for colors
            Vector v = new Vector(sds.getSeries());
            // An inelegance: to minimize performance time for the SVG generation
            // remove elements that are identical to the preceding and following
            // data points.
            int n = 0;
            if(v.size() > 3){
              for(int i = v.size() - 2; i >0 ; i--){
                double prec            = ((SimulationDatum)v.get(i-1)).getNumericValue();
                double curr            = ((SimulationDatum)v.get(i  )).getNumericValue();
                double next            = ((SimulationDatum)v.get(i+1)).getNumericValue();
                if((curr == prec) && (curr == next)) {
                  v.remove(i);
                  n++;
                }
              }
            }
            for(int i = 0; i < v.size(); i++){
              SimulationDatum sd = (SimulationDatum)v.get(i);
              double date = sd.timeValue();
              if((date >= graphStartDateAndTime) && (date <= graphEndDateAndTime) ){
                double value = sd.getNumericValue();
                double X = ((date - graphStartDateAndTime)/ dateScale) * axisWidth;
                double Y = ((value - lowestValue)/valueScale) * axisHeight;
                PolyPts.write(X + "," + Y + " ");
                xml.addLine("<circle r=\"" + SVG_DATA_POINT_RADIUS + "\" cx=\"" + X + "\" cy=\"" + Y + "\" fill=\"" + colorString + "\" tooltip=\"enable\"><hint>A DATA POINT</hint></circle>\n",XMLAssist.INDENT_BLOCK);
              }
            }
          xml.closeTag();
          PolyPts.write("\" fill=\"none\" stroke-width=\"" + SVG_DATA_SERIES_STROKE_WIDTH + "\" stroke=\"" + colorString + "\" onClick=\"showInfo(evt)\"/>\n");
          xml.open("g");
          xml.addLine(PolyPts.toString(), XMLAssist.INDENT_BLOCK);
          xml.closeTag();
        }
      xml.closeTag();  // Series Group
      colorCounter++;
    }
   
    
    // 
    xml.closeTag();  // g
    // Add the legend
    xml.open("g","transform=\"translate(" + (imageWidth - SVG_RIGHT_MARGIN - SVG_LEGEND_WIDTH) + "," + topMargin + ")\"");
    xml.addBlock(legendEntries, XMLAssist.INDENT_BLOCK);
    xml.closeTag(); 
    xml.closeTag();  // SVG
    return xml.getXML();
  }
  
  
  /**
   * Returns a vector of the series names,
   * sorted- this ensures that internally it is possible
   * manipulate the series in several places but in the same order.
   * @return Vector of series names in order
   */
  private Vector<String> seriesSorted(){
    Vector v = new Vector();
    Enumeration E = Series.keys();
    while(E.hasMoreElements()){
      String s = (String)E.nextElement();
      int i = 0;
      while((i < v.size()) && ((String)v.get(i)).compareTo(s) < 0){
        i++;
      }
      if(i == v.size()) v.add(s);
      else              v.add(i, s);
    }
    return v;
  }
  
  /**
   * Generates the SVG for an X-Axis label at the specified X position and
   * adds it to an XML document.
   * @param xml the XMLAssist object that is being used to create the SVG
   * @param value the value to be used for the label
   * @param X the position at which the label will appear
   */
  private void makeXLabel(XMLAssist xml, String value, long X){
    long Y = 0 + SVG_X_AXIS_TIC_HEIGHT  + SVG_X_AXIS_LABEL_SPACE + SVG_X_AXIS_LABEL_HEIGHT/2;
    String labelSVG = "x=\"" + X + "\" " + 
                      "y=\"" + Y + "\" " +
                      "dominant-baseline=\"mathematical\" " +
                      "font-size=\"" + SVG_X_AXIS_LABEL_HEIGHT + "\" " +
                      "font-family=\"" + SVG_X_AXIS_FONT_FAMILY + "\" " +
                      "stroke=\"" + SVG_X_AXIS_LABEL_COLOR + "\" " +
                      "text-anchor=\"middle\"";


    xml.open("g","transform=\"scale(1,-1)\"");
      xml.open("text", labelSVG);
        xml.addLine(value + "\n", XMLAssist.INDENT_BLOCK);
      xml.closeTag();
    xml.closeTag();
  }
  
  /**
   * Generates the SVG for a Y-Axis label at the specified Y position and
   * adds it to an XML document.
   * @param xml the XMLAssist object that is being used to create the SVG
   * @param value the value to be used for the label
   * @param Y the position at which the label will appear
   */
  private void makeYLabel(XMLAssist xml, String value, long Y){
    long X = 0 - SVG_Y_AXIS_TIC_WIDTH  - SVG_Y_AXIS_LABEL_SPACE;
    String labelSVG = "x=\"" + X + "\" " + 
                      "y=\"-" + Y + "\" " +
                      "dominant-baseline=\"mathematical\" " +
                      "font-size=\"" + SVG_Y_AXIS_LABEL_HEIGHT + "\" " +
                      "font-family=\"" + SVG_Y_AXIS_FONT_FAMILY + "\" " +
                      "stroke=\"" + SVG_Y_AXIS_LABEL_COLOR + "\" " +
                      "text-anchor=\"end\"";

    xml.open("g","transform=\"scale(1,-1)\"");
      xml.open("text", labelSVG);
        xml.addLine(value + "\n", XMLAssist.INDENT_BLOCK);
      xml.closeTag();
    xml.closeTag();
  }
  
  /**
   * Generates the SVG for an X-axis tic mark, at the specified position,
   * and adds it to an XML document.
   * @param xml the XMLAssist object being used to create the document
   * @param X the position at which the tic mark will appear
   */
  private void makeXAxisTic(XMLAssist xml, double X){
    xml.addLine("<polyline points=\"" + X + ",0, " + X + ",-" + SVG_X_AXIS_TIC_HEIGHT + "\" fill=\"none\" stroke-width=\"" + SVG_X_AXIS_TIC_STROKE_WIDTH + "\" stroke=\"" + SVG_AXIS_STROKE_COLOR + "\"/>\n", XMLAssist.INDENT_BLOCK);
  }

  /**
   * Generates the SVG for a Y-axis tic mark, at the specified position,
   * and adds it to an XML document.
   * @param xml the XMLAssist object being used to create the document
   * @param Y the position at which the tic mark will appear
   */
  private void makeYAxisTic(XMLAssist xml, double Y){
    xml.addLine("<polyline points=\" 0," + Y + ", -" + SVG_Y_AXIS_TIC_WIDTH + "," + Y + "\" fill=\"none\" stroke-width=\"" + SVG_Y_AXIS_TIC_STROKE_WIDTH + "\" stroke=\"" + SVG_AXIS_STROKE_COLOR + "\"/>\n", XMLAssist.INDENT_BLOCK);
  }
  
  /**
   * Generates the SVG for a legend entry and adds it to an XML document.
   * @param seriesName the name of the series to be added to the legend
   * @param legendEntryNo the number of this entry
   * @param colorString the name of the color to be used for the line related to
   * this series
   * @return a String containing the SVG for the legend
   */
  private String makeLegendEntry(String seriesName, double legendEntryNo, String colorString ){
    String retVal = "";
    double xPos = SVG_LEGEND_LEFT_MARGIN;
    double yPos = legendEntryNo * SVG_LEGEND_ENTRY_HEIGHT;
    retVal += "<text x=\"" + xPos + "\" " +
                    "y=\"" + yPos + "\" " +
                    "dominant-baseline=\"mathematical\" " +
                    "font-size=\"" + SVG_LEGEND_FONT_HEIGHT + "\" " +
                    "font-family=\"" + SVG_LEGEND_FONT_FAMILY + "\" " +
                    "stroke=\"" + SVG_LEGEND_LABEL_COLOR + "\" " +
                    "text-anchor=\"left\"" +
                    ">" +
                    seriesName + 
                    "</text>\n";
    retVal += "<polyline points=\" " + xPos + "," + (yPos + SVG_LEGEND_FONT_HEIGHT + SVG_LEGEND_BAR_SPACE) + ", " + (xPos + SVG_LEGEND_WIDTH) + "," + (yPos + SVG_LEGEND_FONT_HEIGHT + SVG_LEGEND_BAR_SPACE) + "\" fill=\"none\" stroke-width=\"" + SVG_LEGEND_BAR_STROKE_WIDTH + "\" stroke=\"" + colorString + "\" onclick=\"hideSeries(evt, '" + seriesName + "');\"/>\n";
    return retVal;
  }

  /**
   * Gets an extended analysis. This includes standard analysis plus an average and,
   * optionally, a standard deviation.
   * @param StDev true if Standard Deviation should be calculated
   * @return an extended analysis.
   */
  public Analysis getExtendedAnalysis(boolean StDev){
    
    class pt{
      Vector values = new Vector();
      
      public pt(double value){
        addVal(value);
      }
      
      public void addVal(double value){
       values.add(new Double(value)); 
      }
      public double getSum(){
        double sum = 0;
        for(int i = 0; i < values.size(); i++){
          sum += ((Double)values.get(i)).doubleValue();
        }
        return sum;
      }
      
      public double getAvg(){ return getSum() / values.size(); }
    }
  
    class hash{
      Hashtable Hash = new Hashtable();
      public void addPt(SimDate sd, double val){
        pt p;
        if(!Hash.containsKey(sd)){  p = new pt(val);        }
        else                     {  p = (pt)Hash.get(sd);   }
      }
      
    }
    
    
    
    
    Analysis extended = new Analysis();
    
    SimulationDataSeries AVG = new SimulationDataSeries("Average", true);
    
    hash H = new hash();
    
    Enumeration e = Series.keys();
    while(e.hasMoreElements()){
      String name = (String)e.nextElement();
      SimulationDataSeries S = (SimulationDataSeries)Series.get(name); 
      extended.addSeries(S, name);
      Vector s = S.getSeries();
      for(int i = 0; i < s.size(); i++){
        SimulationDatum SD = (SimulationDatum)s.get(i);
        H.addPt(SD.getSimulationDate(), SD.getNumericValue());
      }
    }
    Enumeration dates = H.Hash.keys();
    while(dates.hasMoreElements()){
      SimDate SD = (SimDate)dates.nextElement();
      double d = ((pt)(H.Hash.get(SD))).getAvg();
      AVG.addElement(SD, d);
    }
    
    extended.addSeries(AVG, "Average");
    return extended;
  }
  
}
