package abcm.sim;

import java.util.Enumeration;
import java.util.Hashtable;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;

/**
 * Stores a collection of keys and values that
 * are used to select from optional processing paths
 * within the code. This is in lieu of having separately
 * written, stored, and compiled versions of the code.
 * Intuitively, a code option chooses options from the
 * behavior of the model. For example, one option may
 * be for plants to grow using one production model,
 * while an alternative option uses a different model.
 * To implement this, a code model key is created,
 * and the code executed is dependent on the value of that
 * key in the code model.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class CodeModel implements ABCMDeepCloneable{

  /** Table of options and values  */
  private Hashtable optionList = new Hashtable();

  /**
   * Sets a code model option. The domain is specified by a 
   * String, and this option is set to a specific value.
   * 
   * @param domain  Key representing the code model option
   * @param value   Value to be associated with this option.
   */
  public void setCodeModelOption(String domain, int value){
    optionList.put(domain, new Integer(value));
  }
    
  /**
   * Gets the value for a code model option.
   * 
   * @param domain  Key representing the requested code model option
   * @return an integer representing the value for this code model option.
   */
  public int getCodeModelOption(String domain){
    if(optionList.containsKey(domain)) 
        return ((Integer)optionList.get(domain)).intValue();
    else
        return 0;
  }

  /**
   * Gets a list of the keys contained by this Code Model object.
   * @return an Enumeration listing the keys
   */
  public Enumeration getKeys(){
    return optionList.keys();
  }
  
  /**
   * Returns the number of values stored in this Code Model
   * Object.
   * @return number of values stored
   */
  public int countOfValues(){
    return optionList.size();
  }
  
  /**
   * Generates a string listing the options and values
   * stored in this Code Model object.
   * @return the Code Model report
   */
  public String getReport(){
    String retVal = "Code Model with " + optionList.size() + " keys\n\n";
    Enumeration e = optionList.keys();
    while(e.hasMoreElements()){
      String opt = (String)e.nextElement();
      int val    = getCodeModelOption(opt);
      retVal += opt + " value: " + val + "\n";
    }
    return retVal + "\n";
  }
  
  /**
   * Returns a copy of the hashtable that stores this option list; this should
   * only be used for reporting, not modification.
   * @return the copy of the option list Hashtable
   */
  public Hashtable getOptionList(){ return (Hashtable)optionList.clone(); }
  
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    CodeModel ret = new CodeModel();
    cloneStore.put(this, ret);
    ret.optionList = (Hashtable)optionList.clone();
    return ret;
  }
  
}
