package abcm.sim;

import java.io.StringReader;
import java.util.Hashtable;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.helpers.DefaultHandler;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.math.SimDate;
import abcm.math.Year;

/**
 * A History object is used to manage Histories while
 * the simulation runs.
 * 
 * The History object manages both the seasonal and overall
 * narratives. It includes methods that can begin the history
 * execution on a given date, proceed through both seasonal and
 * overall narratives until there are no more commands to execute
 * on that date.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class History implements ABCMDeepCloneable{

  /** Maximum number of instructions in any History  */
  private static final int MAXINSTRUCTIONS = 100000;
  
  /** List of all instructions in this History  */
  private HistoryInstruction instructionList[] = new HistoryInstruction[MAXINSTRUCTIONS];
  /** Pointer to current Seasonal Narrative instruction  */
  private int seasonalIndex;
  /** Pointer to current History instruction  */
  private int overallIndex;
  
  /** Flag indicating whether all Seasonal Narratives (collectively) will continue to execute on this date  */
  private boolean hasMoreSeasonalCommandsOnCurrentDate;
  /** Flag indicating whether the History should continue to execute on this date  */
  private boolean hasMoreOverallCommandsOnCurrentDate;
  
  /** Name of this History  */
  private String name;
  /** Current Simulation Date  */
  private SimDate currentDate = new SimDate();
  
  /** Hashtable indicating status of Seasonal Narratives (on or off)  */
  private Hashtable SeasonalNarrativesOnOff =  new Hashtable();
  
  
  /**
   * Constructor
   * @param historyName Name of this History object
   */
  public History(String historyName){
    name = historyName;
  }
  
  /**
   * Gets the name of this History
   * @return the History name
   */  
  public String getName(){
    return name;
  }
  
  /**
   * Writes the list of History instructions to the standard output 
   */
  public void writeInstructionList(){
    int x = 0;
    while(!(instructionList[x] == null)){
      System.out.print("x: " + x + "  Date: " + SimDate.convertToString(instructionList[x].date) + "(" + instructionList[x].date + ") command: " + instructionList[x].CommandID + " ");
      int i = 1;
      while(instructionList[x].argList[i] != null){
        System.out.print("  " + instructionList[x].argList[i]);
        i++;
      }
      System.out.print("\n");
      x++;
    }
  }
  
  // Manipulation and execution
  /**
   * Sets the execution date for this history object.
   * 
   * This finds the first instruction in the seasonal
   * and overall narratives and sets the pointer to return
   * 
   * If no instruction falls on this day of the year (for
   * seasonal instructions) or on this date (for overall 
   * instructions, the respective index will be set to -1.
   * @param s the Simulation Date to be set
   */
  public void setDate(SimDate s){
    currentDate = s;
    double dateVal = s.value();
    double seasonalDateVal = dateVal % 365;
    int index = 0;
    seasonalIndex = -1;
    overallIndex = -1;
    while(instructionList[index] != null && instructionList[index].date < dateVal){
      if(seasonalIndex < 0){
        if(instructionList[index].date == seasonalDateVal){
          seasonalIndex = index;
        }          
      }
      index++;
    }
    if(instructionList[index] != null){
      if(instructionList[index].date == dateVal){
        overallIndex = index;
      }
    }
      
  }

  /**
   * Whether a given seasonal narrative is turned on or off.
   * Will return 'true' if the seasonal narrative has
   * never been turned on or off, thus making 'on' the default.
   * @param narrativeName The narrative to be queried
   * @return true if the narrative is on, false if not
   */
  public boolean getSeasonalNarrativeStatus(String narrativeName){
    Boolean B = (Boolean)SeasonalNarrativesOnOff.get(narrativeName);
    return (SeasonalNarrativesOnOff.get(narrativeName) == null ? true : B.booleanValue());
  }
  
  /**
   * Turns a seasonal narrative 'on' or 'off
   * @param narrativeName Name of the narrative whose satus is to be set
   * @param status the value of the status
   * 
   */
  public void setSeasonalNarrativeStatus(String narrativeName, boolean status){
    SeasonalNarrativesOnOff.put(narrativeName, new Boolean(status));
  }
  
  /**
   * Gets the next command that should be executed
   * on the specified date.
   * 
   * This will return null if there is no next command on the
   * current specified date.
   * @return The History Instruction object that is the next command
   * in this history on the specified date.
   */
  public HistoryInstruction getNextCommand(){
    HistoryInstruction retVal = null;
    boolean useSeasonal = false;
    // If there is at least one command
    if(hasMoreCommandsOnCurrentDate()){
      // If there is both a seasonal and an overall command queued up
      if((seasonalIndex > -1 && (instructionList[seasonalIndex].date == currentDate.dayOfYear())) &&
         ((overallIndex > -1) && (instructionList[overallIndex] != null) && (instructionList[overallIndex].date == currentDate.value()))){
        // Choose the one with the lower narrative order
        useSeasonal = 
          ((instructionList[seasonalIndex].narrativeOrder) <
           (instructionList[overallIndex].narrativeOrder));
      }
      else {
        useSeasonal = ((seasonalIndex > -1) && (instructionList[seasonalIndex].date == currentDate.dayOfYear()));
      }
      if(useSeasonal){
        if(getSeasonalNarrativeStatus(instructionList[seasonalIndex].narrativeName)){
           retVal = instructionList[seasonalIndex];
        }
        seasonalIndex++;
      }
      else {
        retVal = instructionList[overallIndex];
        overallIndex++;
      }
    }
    return retVal;
  }
  
  /**
   * Determines whether there are any more commands to be
   * executed on the specified date.
   * @return true if there are more commands on the current date, false if not.
   */
  public boolean hasMoreCommandsOnCurrentDate(){
    boolean retVal = false;
    double cdv = currentDate.value();
    if(cdv > 0){
      if(seasonalIndex > -1){
        if(instructionList[seasonalIndex] != null){
          if(instructionList[seasonalIndex].date == (currentDate.dayOfYear()))
            retVal = true;
        }
      }
      if(retVal == false){
        if(overallIndex > -1){
          if(instructionList[overallIndex] != null){
            if(instructionList[overallIndex].date == cdv){
              retVal = true; 
            }
          }
        }
      }      
    }
    return retVal;
  }

  /**
   * Gets the date of the earliest (non-seasonal) instruction in this
   * History.
   * @return the date of the earliest instruction, excluding seasonal
   * narratives, in this History
   */
  public double getFirstDate(){
    int index = 0;
    // Count forward until you're out of instructions or until the first
    // instruction that is not a seasonal narrative
    while(instructionList[index] != null && 
          instructionList[index].date < (Year.DAYS_IN_YEAR + 1)) 
      index++;
    if(instructionList[index] == null) return 0;
    else return instructionList[index].date;
  }
  
  /**
   * Gets the date of the last instruction executed in this History, as
   * a double
   * @return double representing the date of the last instruction in this History 
   */
  public double getLastDate(){
    if(instructionList[0] == null) return 0; // Empty history
    int i = MAXINSTRUCTIONS - 1;
    while(instructionList[i] == null) i--;
    return instructionList[i].date;
  }
  
  /**
   * Translates the set of instructions from the XML format
   * in which they are received from the HistoryManager object
   * into an InstructionList array that can be stepped through
   * by the simulation.
   * @param XML XML formatted instruction list
   */
  public void setInstructionsFromXML(String XML){
    SAXParserFactory SPF = SAXParserFactory.newInstance();
    SPF.setValidating(false);
    try{
      SAXParser p = SPF.newSAXParser();
      parseHistoryXML pHxml = new parseHistoryXML();
      p.parse(new InputSource(new StringReader(XML)), pHxml);
    }
    catch (Exception e){
      System.out.println("Some error occurred while parsing instructions from XML: " + e.getMessage());
    }   
    
  }
  
  /**
   * This class operationalizes the parsing of an XML format
   * History, with all of its instructions, into the instruction
   * list array for this instance of the History object.
   * 
   * For details on the class operation, see the documentation
   * related to the SAX parser.
   * 
   */
  private class parseHistoryXML extends DefaultHandler {
  
  /** Index of current argument */
  private int argIndex = 0;
  
  /** Index of the current instruction */
  private int instructionIndex;
  
  /** Parser status tag; see code for meaning */
  private static final int OUTSIDEALLTAGS = 0;
  /** Parser status tag; see code for meaning */
  private static final int inINSTRUCTIONSET = 1;
  /** Parser status tag; see code for meaning */
  private static final int inNARRATIVENAME = 2;
  /** Parser status tag; see code for meaning */
  private static final int inNARRATIVEORDER = 3;
  /** Parser status tag; see code for meaning */
  private static final int inINSTRUCTIONORDER = 4;
  /** Parser status tag; see code for meaning */
  private static final int inDATE = 5;
  /** Parser status tag; see code for meaning */
  private static final int inCOMMAND = 6;
  /** Parser status tag; see code for meaning */
  private static final int inCOMMANDPK = 7;
  /** Parser status tag; see code for meaning */
  private static final int inARGUMENTS = 8;
  /** Parser status tag; see code for meaning */
  private static final int inARG = 9;
  
  /** Used to identify current mode while parsing */
  private int mode=OUTSIDEALLTAGS;
  
  /** Used while parsing to represent current text */
  private String text;
  
    /**
     * @see org.xml.sax.helpers.DefaultHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
     */
    public void startElement(String URI, String localName, 
        String qualifiedName, Attributes atts){

      if(qualifiedName.equals("instructionset")){
        mode = inINSTRUCTIONSET;
        instructionIndex = Integer.parseInt(atts.getValue(0));
        instructionList[instructionIndex] = new HistoryInstruction();
        instructionList[instructionIndex].instructionOrder = instructionIndex;

      }else if(qualifiedName.equals("narrativeName")){
        mode = inNARRATIVENAME;
      }else if(qualifiedName.equals("narrativeOrder")){
        mode = inNARRATIVEORDER;
      }else if(qualifiedName.equals("instructionOrder")){
        mode = inINSTRUCTIONORDER;
      }else if(qualifiedName.equals("date")){
        instructionList[instructionIndex].date = Double.parseDouble(atts.getValue(0));
        mode = inDATE;
      }else if(qualifiedName.equals("command")){
        mode = inCOMMAND;
      }else if(qualifiedName.equals("commandPK")){
        mode = inCOMMANDPK;
      }else if(qualifiedName.equals("arguments")){
        mode = inARGUMENTS;
      }else if(qualifiedName.equals("arg")){
        argIndex = Integer.parseInt(atts.getValue(0));
        mode = inARG;
      }
      // Set the characters returned to ""; SAX may call 'characters' several times.
      text ="";
    }
    
    /**
     * @see org.xml.sax.helpers.DefaultHandler#characters(char[], int, int)
     */
    public void characters(char[] chars, int firstChar, int dataLength){
      text += new String(chars, firstChar, dataLength);
      switch (mode){
        case inINSTRUCTIONSET:
          break;
        case inNARRATIVENAME:
          instructionList[instructionIndex].narrativeName = text;
          break;
        case inNARRATIVEORDER:
          instructionList[instructionIndex].narrativeOrder = Integer.parseInt(text);
          break;
        case inINSTRUCTIONORDER:
          instructionList[instructionIndex].instructionOrder = Integer.parseInt(text);
          break;
        case inDATE:
          break;
        case inCOMMAND:
          break;
        case inCOMMANDPK:
          instructionList[instructionIndex].CommandID = Integer.parseInt(text);
          break;
        case inARGUMENTS:
          break;
        case inARG:
          instructionList[instructionIndex].argList[argIndex] = text;
          break; 
        default:
          break;
      }
    }
    
    /**
     * @see org.xml.sax.helpers.DefaultHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
     */
    public void endElement(String URI, String localName, String qualifiedName){ mode = OUTSIDEALLTAGS;  }
    
  }
    
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    History H = new History(name);
    cloneStore.put(this, H);
    H.currentDate = new SimDate(this.currentDate.value());
    H.hasMoreOverallCommandsOnCurrentDate = this.hasMoreOverallCommandsOnCurrentDate;
    H.hasMoreSeasonalCommandsOnCurrentDate = this.hasMoreSeasonalCommandsOnCurrentDate;
    H.instructionList = new HistoryInstruction[instructionList.length];
    for(int i = 0; i < instructionList.length; i++){
      H.instructionList[i] = (HistoryInstruction)cloneStore.getClone(instructionList[i]);
    }
    H.overallIndex = this.overallIndex;
    H.seasonalIndex = this.seasonalIndex;
    H.SeasonalNarrativesOnOff = (Hashtable)this.SeasonalNarrativesOnOff.clone(); // This is a map of strings and booleans, so no deep clone is needed
    return H;
  }
  
}
