package abcm.sim;

import java.util.Enumeration;
import java.util.Hashtable;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;

/**
 * A Parameter Set represents a collection of parameters and their
 * current values. The object can be passed from one object to any objects it creates
 * and manages, allowing them all to use the same set of parameters.
 * <br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ParameterSet implements ABCMDeepCloneable{

  /** Collection of parameters  */
  private Hashtable params;

  /**
   * Generic constructor; initializes the parameters table.
   */
  public ParameterSet(){
    params = new Hashtable();
  }
  
  /**
   * Sets the value of a parameter
   * @param paramName name of the parameter to be set
   * @param paramValue value to which this parameter is to be set
   */
  public void setParameter(String paramName, double paramValue){
    paramName = paramName.toUpperCase();
    params.remove(paramName);  // Take out any pre-existing value
    params.put(paramName, new Double(paramValue));
  }

  /**
   * Gets the value of a parameter. Returns zero if the
   * specified parameter does not exist in the Parameter Set
   * @param paramName name of the parameter to be returned
   * @return the value of the specified parameter
   */
  public double getParameter(String paramName){
    paramName = paramName.toUpperCase();
    if(params.containsKey(paramName)){
      return ((Double)params.get(paramName)).doubleValue();
    }
    else{
      System.err.println("Requested Parameter '" + paramName + "' has no value in current Parameter Set; returning zero.");
      return 0;
    }
  }
  
  /**
   * Determine if the specified parameter exists in this parameter set
   * @param paramName name of the parameter sought
   * @return true if it is found in the parameter set, false otherwise
   */
  public boolean hasParameter(String paramName){
    return params.containsKey(paramName);
  }

  /**
   * Gets a report listing all parameters and values in this Parameter Set
   * @return the Parameter Set report
   */
  public String getReport(){
    int width = 30;
    String retVal = "";
    Enumeration e = params.keys();
    int maxDecimal = 0;
    while(e.hasMoreElements()){
      String key = (String)e.nextElement();
      if(key.length() > width) width = key.length();
      maxDecimal = Math.max(maxDecimal, params.get(key).toString().indexOf("."));
    }
    
    width+=3; // Three more spaces for good measure
    e = params.keys();
    String buffer = "";
    for(int i = 0; i < width; i++) buffer += " "; // Create a string to tack on and then trim
    while(e.hasMoreElements()){
      String key = (String)e.nextElement();
      String val = params.get(key).toString();
      while(val.indexOf(".") < maxDecimal) val = " " + val;
      retVal+= (key + buffer).substring(0, width) + val + "\n";
    }
    return "Parameters: " + params.size() + "\n\n" + retVal;    
  }

  
  /**
   * Gets a count of the parameters in this parameter set
   * @return the count of parameters
   */
  public int getCountOfParameters(){    return params.size();   }
  
  /** 
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore) 
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    ParameterSet PS = new ParameterSet();
    PS.params = (Hashtable)params.clone(); // Only a shallow clone is required
    return PS;
  }
  
}
