
package abcm.sim;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.math.SimDate;

/**
 * A Probe is an instruction that tells the model to
 * record certain data to a certain data series at
 * a specified interval, beginning with a base date. Methods
 * exist to turn probes on or off, which activates or interrupts
 * the data collection they perform.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Probe implements ABCMDeepCloneable{

  /** Name of this probe  */
  private String               name;
  /** True if the probe is executed monthly, false otherwise  */
  private boolean              monthly            = false;  
  /** Frequency, in Simulation Days, at which the probe should be updated  */
  private long                 frequency          = 1;
  /** Base Date from which updates are calculated with the specified frequency  */
  private long                 baseDate;
  /** Name of the Data Series to which this probe sends its output */
  private String               seriesName;
  /** Name of the domain in the Data Series to which this probe sends its output  */
  private String               seriesDomain;
  /** Flag indicating whether the probe collects numeric or textual data  */
  private boolean              numeric            = true;
  /** Data Series to which the probe will direct its output  */
  private SimulationDataSeries series;
  /** Command the probe uses to retrieve data from the simulation  */
  private String               retrieveCommand;
  /** Switch that can be set to suspend collection of data  */
  private boolean              on                 = true;
  /** Numeric 'offset' that can be passed as part of the command to retrieve data (useful for data referenced by a finer time scale than days)  */
  private double               offset             = 0;  // When saving a datum, this number (less than one) 
                                                        // will be added to the SimDate. Allows storage of 
                                                        // information (usually flow hydrographs) at a finer 
                                                        // level than a full day, provided the model element 
                                                        // interrogated is equipped to provide this

  /**
   * Constructor
   * @param probeName Name of this probe
   * @param BaseDate zero date from which updates are calculated with the specified frequency
   * @param freq frequency, in days, at which this probe should be updated
   * @param dataSeriesName Data Series for this Probe
   * @param dataSeriesDomain Domain in Data Series to which this probe directs data
   * @param isNumeric True if the probe is numeric, false if textual
   * @param RetrieveCommand Command that the probe should execute to collect data.
   */
  public Probe(String probeName, SimDate BaseDate, long freq, 
      String dataSeriesName, String dataSeriesDomain, boolean isNumeric, String RetrieveCommand ){
    name            = probeName;
    monthly         = false;
    baseDate        = (long)BaseDate.value();
    frequency       = freq;
    seriesName      = dataSeriesName;
    seriesDomain    = dataSeriesDomain;
    numeric         = isNumeric;
    retrieveCommand = RetrieveCommand;
  }
  
  /**
   * Constructor
   * @param probeName Name of this probe
   * @param dataSeriesName Data Series for this Probe
   * @param dataSeriesDomain Domain in Data Series to which this probe directs data
   * @param isNumeric True if the probe is numeric, false if textual
   * @param RetrieveCommand Command that the probe should execute to collect data.
   */
  public Probe(String probeName, String dataSeriesName, String dataSeriesDomain, 
      boolean isNumeric, String RetrieveCommand){
    name            = probeName;
    monthly         = true;
    seriesName      = dataSeriesName;
    seriesDomain    = dataSeriesDomain;
    numeric         = isNumeric;
    retrieveCommand = RetrieveCommand;
  }
  
  /* Additional Setters */
  
  
  /**
   * Sets this probes offset value
   * @param O the new offset value
   * @return the newly set offset value
   */
  public double setOffset(double O){
    offset = Math.abs(O) % 1d;
    return offset;
  }
  
  /**
   * Sets the data series for this probe
   * @param sds data series to be used
   */
  public void setSeries(SimulationDataSeries sds){    series = sds;  }

  
     
  /* Determine if this probe should be updated on the given date*/ 
  
  
  /**
   * Returns true if the specified simulation date is one on which
   * the probe is supposed to collect data
   * @param simDate the Date to be tested
   * @return true if the probe should update, false otherwise
   */
  public boolean isTimeToUpdate(SimDate simDate){
    boolean update = false;
    if(on){
      if(monthly){
        if(simDate.getString().startsWith("01"))                     update = true;
      }
      else { 
        if((((long)simDate.value() - baseDate) % frequency) == 0)  update = true; 
//        if(update) System.out.println("I'm probe " + this.name + " and I'm sure it's time to update! (freq = " + this.frequency + ") simdatevalue " + simDate.value() + " baseDate " + baseDate);
      }
    }
    return update;
  }
  
  
  /* Basic probe information/ getters*/
  
  /**
   * Returns the base date for this probe
   * @return Returns the baseDate.
   */
  public long getBaseDate() {    return baseDate;  }

  
  /**
   * Returns the frequency for this probe
   * @return Returns the frequency.
   */
  public long getFrequency() {    return frequency;  }

  
  /**
   * Returns the retrieve command for this probe
   * @return Returns the retrieveCommand.
   */
  public String getRetrieveCommand() {    return retrieveCommand;  }
  
  
  /**
   * Returns the series into which this probe places data
   * @return Returns the series.
   */
  public SimulationDataSeries getSeries() {   return series; }
  
  
  /**
   * Gets the name of the series into which this probe places data
   * @return the name of the probe's data series
   */
  public String getSeriesName(){
    if(series     != null) return series.getName();
    if(seriesName != null) return seriesName;
    else                   return "Unknown";
  }
  
  /**
   * Gets the domain of the data series to which this probe directs its data 
   * @return the data series domain name
   */
  public String getDomain(){ return seriesDomain; }
  
  /**
   * Returns the state of this probe (true = on, false = off)
   * @return true if the probe is turned on, false otherwise.
   */
  public boolean isOn(){        return on; }
  
  /**
   * Gets whether this probe is monthly or not.
   * @return true if the probe is monthly, false if not.
   */
  public boolean isMonthly(){    return monthly;  }
 
  /**
   * Gets the probe's status as numeric or textual
   * @return true if the probe is numeric, false otherwise
   */
  public boolean isNumeric(){     return numeric; }
  
  /**
   * Gets the offset value for this probe
   * @return the probe's offset value
   */
  public double getOffset(){ return offset; }
  
  /**
   * Gets the name of this probe
   * @return the probe's name
   */
  public String getProbeName(){ return name; }
  
  /* Probe actions */
  
  /**
   * Turns this probe on.
   */
  public void turnProbeOn() {   on = true;   }
  
  /**
   * Turns this probe off
   */
  public void turnProbeOff(){   on = false;  }
  
  /* Reporting */
  
  /**
   * Gets a report describing this probe 
   * @return a brief description of this probe
   */
  public String getReport(){
    String report = "";
    if(monthly){
      report += "*Monthly* ";
    }
    else{
      report += SimDate.convertToString(baseDate) + " FREQ: " + frequency + " ";
    }
    report += seriesName + " " + offset + "\n";
    report += " RETRIEVE: " + retrieveCommand;
    return report;
  }

  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Probe newClone = new Probe(this.name, new SimDate(this.baseDate), this.frequency, this.seriesName, this.seriesDomain, this.numeric, this.retrieveCommand);
    cloneStore.put(this, newClone);
    newClone.monthly = this.monthly;
    newClone.offset  = this.offset;
    newClone.on      = this.on;
    newClone.series  = (SimulationDataSeries)(cloneStore.getClone(series));
    return newClone;
  }
}
