package abcm.sim;

import java.io.StringWriter;
import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableVector;
import abcm.math.SimDate;
import abcm.utilities.XMLAssist;

/**
 * A Simulation Data Series is a collection of data taken at specific
 * dates during a simulation run, and including either text or numeric
 * data from the run.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SimulationDataSeries implements ABCMDeepCloneable {
  
  /** Flag value indicating numeric data are stored  */
  public static final boolean IS_NUMERIC = true;
  /** Flag value indicating text data are stored  */
  public static final boolean IS_TEXT    = false;
  
  /** Collection of data points  */
  private ABCMDeepCloneableVector    series;
  /** Flag indicating collection is numeric (vs. textual)  */
  private boolean                    numeric;
  /** Highest (numeric) value  */
  private double                     highestValue;
  /** Lowest (numeric) value  */
  private double                     lowestValue;
  /** Run Number associated with this data series  */
  private long                       runNumber;
  /** Description of this data series  */
  private String                     seriesDescription;
  /** Name for this data series  */
  private String                     seriesName;
  /** Domain described by this series  */
  private String                     domain;
  
  /** Flag indicating whether these data have been saved  */
  private boolean                    saved;
  
  /**
   * Basic constructor.
   * @param variable the name of the domain represented by this series.
   * @param isNumeric true if the Data Series is numeric, false otherwise 
   */
  public SimulationDataSeries(String variable, boolean isNumeric){
    domain = variable;
    numeric = isNumeric;
    init();
  }

  /**
   * Constructor that includes a Data Series name
   * @param variable the name of the domain represented by this series.
   * @param isNumeric true if the Data Series is numeric, false otherwise
   * @param name a new name for the Data Series 
   */
  public SimulationDataSeries(String variable, boolean isNumeric, String name){
    domain = variable;
    numeric = isNumeric;
    seriesName = name;
    init();
  }
  
  /**
   * Initializes the series vector
   */
  private void init(){
    series = new ABCMDeepCloneableVector();
  }
  
  /**
   * Adds a numeric element, in the proper series order per SimDate.
   * Note that if this method is called on a series that is not
   * numeric, it will fail and return false.
   * @param simulationDate Simulation Date associated with the element to be added
   * @param value value to be added
   * @return returns true if successful, false if not.
   */
  public boolean addElement(SimDate simulationDate, double value){
    if(!numeric) return false;
    else{
      saved = false;
      SimulationDatum datum = new SimulationDatum(simulationDate, domain, value);
      if(series.size() == 0){
        highestValue = value;
        lowestValue = value;
        series.add(datum);
        return true;
      }
      // Series already has at least one element in it
      if(value > highestValue) highestValue = value;
      if(value < lowestValue)  lowestValue  = value;
      double time = simulationDate.value();
      int i = series.size() - 1;
      
      // If the new element falls after the last one
      if(time >= ((SimulationDatum)series.get(i)).getSimulationDate().value()){
        series.add(datum);
        return true;
      }
      // If not, trace backwards until it is fit in in chronological
      // order
      while(i >= 0 && (time < ((SimulationDatum)series.get(i)).getSimulationDate().value())){
        i--;
      }
      i++; // Jog back upward to the insertion point
      series.insertElementAt(datum, i);
      return true;
    }    
  }
  
  /**
   * Adds a String element, in the proper series order per SimDate
   * Returns false if unsuccessful, as, for example, when a String
   * element is added to a series that is numeric
   * @param simulationDate Simulation Date associated with this value
   * @param value value to be added
   * @return true if successful, false if not
   */
  public boolean addElement(SimDate simulationDate, String value){
    if(numeric) return false;
    else {
      saved = false;
      int i = series.size() - 1;
      while((i > 0) && ( ((SimulationDatum)series.get(i)).timeValue() >= simulationDate.value()))    i--;
      i++;
      series.insertElementAt(new SimulationDatum(simulationDate, domain, value), i);
      return true;
    }
  }
    
  /**
   * Gets the simulation date of the earliest datum
   * point in this series.
   * @return The simulation date on which this series begins
   */
  public SimDate getEarliestDate(){
    if(series.size() == 0 ) return null;
    else return ((SimulationDatum)series.get(0)).getSimulationDate();
  }
  
  /**
   * Gets the simulation date of the last datum point in
   * this series
   * @return The simulation date on which the series ends
   */
  public SimDate getLastDate(){
    if(series.size() == 0 ) return null;
    return ((SimulationDatum)series.get(series.size() - 1)).getSimulationDate();
  }
  
  /**
   * Gets the series
   * @return a vector representing the series itself
   */
  public Vector getSeries(){
    return series;
  }
  
  /**
   * Gets a printout of this data series
   * @return A string representing this data series
   */
  public String getSeriesPrintout(){
    StringWriter so = new StringWriter();
    for(int i = 0; i < series.size(); i++){
      so.write(((SimulationDatum)series.get(i)).toString() + "\n");
    }
    return so.toString();
  }

  /**
   * Indicates whether the Data Series is numeric
   * @return true if the Data Series is numeric, false otherwise
   */
  public boolean isNumeric(){
    return numeric;
  }
  
  /**
   * Gets the highest value among all the numeric data stored
   * @return the highest value among all the numeric data stored
   */
  public double getHighestValue(){
    return highestValue;
  }

  /**
   * Gets the lowest value among all the numeric data stored
   * @return the lowest value among all the numeric data stored
   */
  public double getLowestValue(){
    return lowestValue;
  }
  
  /**
   * Sets the description for this Data Series
   * @param newDescription new description for this series
   */
  public void setDescription(String newDescription){
    seriesDescription = newDescription;
  }

  /**
   * Gets the description of this Data Series
   * @return the description of this Data Series
   */
  public String getDescription(){
    return seriesDescription;
  }

  /**
   * Determine if this Data Series has been saved or not
   * @return true if the Data Series has been saved, false otherwise
   */
  public boolean hasBeenSaved(){
    return saved;
  }

  /**
   * Marks this Data Series as having been saved
   */
  public void markAsSaved(){
    saved = true;
  }

  /**
   * Set the name of this Data Series
   * @param newName new name for the Data Series
   */
  public void setName(String newName){
    seriesName = newName;
  }
  
  /**
   * Gets the name of this Data Series
   * @return the Data Series name
   */
  public String getName(){
    if(seriesName != null) return seriesName;
    return "unnamed";
  }

  /**
   * Gets the SimulationDataSeries as an XML document.
   * @return the XML representation of this data series, as a String
   */
  public String getAsXML(){
    XMLAssist xml = new XMLAssist();
    return (getAsXML(xml)).getXML();
  }
  
  /**
   * Adds a SimulationDataSeries' XML information to an XMLAssist object creating
   * an xml document.
   * @param xml the XMLAssist object to which this information is to be added
   * @return the XMLAssist object with the information added
   */
  public XMLAssist getAsXML(XMLAssist xml){
    xml.open("simulationDataSeries");
      xml.addLine("name",seriesName);
      xml.addLine("isNumeric", "" + numeric);
      if(series.size() > 0){
        if(numeric){
          xml.addLine("highestValue", highestValue);
          xml.addLine("lowestValue", lowestValue);
        }
        xml.addLine("firstDate", ((SimulationDatum)series.get(0)).getSimulationDate().value());
        xml.addLine("firstSimDate", ((SimulationDatum)series.get(0)).getSimulationDate().getString());
        xml.addLine("lastDate", ((SimulationDatum)series.get(series.size()-1)).getSimulationDate().value());
        xml.addLine("lastSimDate", ((SimulationDatum)series.get(series.size()-1)).getSimulationDate().getString());
        xml.open("data");
          for(int i = 0; i < series.size(); i++){
            SimulationDatum sd = (SimulationDatum)series.get(i);
            xml.open("datum");
              SimDate s = sd.getSimulationDate();
              xml.addLine("simDate", s.getString());
              xml.addLine("date", s.value());
              xml.addLine("value", (numeric ? "" + sd.getNumericValue() : sd.getTextValue()));
            xml.closeTag();
          }
        xml.closeTag();
      }
    xml.closeTag();
    return xml;
  }

  /**
   * Returns the last value before the time specified.
   * This will make it possible to save values at arbitrary times but
   * return them at standard intervals. For example, suppose a value
   * is stored every 3 days, but we wish to graph that value every 5 days.
   * Thus:
   * 
   * Day 1  value = 100
   * Day 4  value = 200
   * Day 7  value = 300
   * Day 10 value = 400
   * Day 13 value = 500
   * Day 16 value = 600
   * Day 19 value = 700
   * Day 22 value = 800
   * 
   * Graph by 5 days:
   * Day 1  value = 100
   * Day 6  value = 200
   * Day 11 value = 400
   * Day 16 value = 600
   * Day 20 value = 700
   * 
   * This will be useful only for some kinds of values; stage hydrographs
   * is one of them.
   * @param valueAt the Simulation Date stamp before which the value should be returned
   * @return numeric value of last datum prior to the given Simulation Date
   */
  public double getLastValueBefore(SimDate valueAt){
    double retVal = 0;
    double time = valueAt.value();
    if(!isNumeric())       return 0;
    if(series.size() == 0) return 0;
    if((((SimulationDatum)series.get(0)).getSimulationDate()).value() > time) return 0;
    retVal = ((SimulationDatum)series.get(0)).getNumericValue();
    for(int i = 1; i < series.size(); i++){
      if((((SimulationDatum)series.get(i)).getSimulationDate()).value() > time) return retVal;
      retVal = ((SimulationDatum)series.get(i)).getNumericValue();
    }
    return retVal;
  }
  
  /**
   * Gets a numeric value from the data series that is a linearly interpolated value
   * between the value taken most nearly after the specified time and the value taken most nearly
   * prior to that time. If this series is not numeric, or if there is no data in it,
   * zero is returned. If a datum was taken at precisely the time
   * specified, that value is returned without adjustment.
   * @param valueAt Simulation Date for later value
   * @return interpolated value between 
   */
  public double getScaledValueAt(SimDate valueAt){
    double lastVal = 0;
    double nextVal = 0;
    double lastTime = 0;
    double nextTime = 0;
    double val = 0;
    double time = valueAt.value();
    if(!isNumeric())       return 0;
    if(series.size() == 0) return 0;
    SimDate SD = (((SimulationDatum)series.get(0)).getSimulationDate());
    if(SD == null)         return 0;
    if(SD.value() > time)  return 0;
    
    lastVal = ((SimulationDatum)series.get(0)).getNumericValue();
    lastTime = ((SimulationDatum)series.get(0)).timeValue();
    
    for(int i = 1; i < series.size(); i++){
      nextVal = ((SimulationDatum)series.get(i)).getNumericValue();
      nextTime = ((SimulationDatum)series.get(i)).timeValue();
      val = nextVal;
      if(time == nextTime) return val;
      if((((SimulationDatum)series.get(i)).getSimulationDate()).value() > time) {
        return lastVal + (nextVal - lastVal) * ((time - lastTime)/ (nextTime - lastTime));
      }
      lastVal = nextVal;
      lastTime = nextTime;
    }
    return val;
  }
  
  /**
   * Gets the domain of this data series
   * @return the series domain
   */
  public String getDomain(){ return domain; }
  
  /** 
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore) 
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    SimulationDataSeries newSDS = new SimulationDataSeries(domain, numeric, seriesName);
    cloneStore.put(this, newSDS);
    newSDS.highestValue      = this.highestValue;
    newSDS.lowestValue       = this.lowestValue;
    newSDS.runNumber         = this.runNumber;
    newSDS.seriesDescription = this.seriesDescription;
    newSDS.series            = (ABCMDeepCloneableVector)series.getDeepClone(cloneStore); // SimulationDatum objects are immutable and don't need deep cloning
    return newSDS;
  
  }
  
}

