package abcm.sim;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.math.SimDate;

/**
 * Represents one data point drawn from the simulation.
 * 
 * The datum point's value may be either numeric or string.
 * 
 * The datum point is specified with a given Simulation Date.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SimulationDatum implements ABCMDeepCloneable{

  /** Limit on the width of the name of a domain  */
  private static final int DOMAIN_TEXT_WIDTH     = 10; 
  
  /** Simulation Date stamp for this datum  */
  private SimDate   sd;
  /** Domain ID for this datum  */
  private String    domain;
  /** Numeric value for this datum  */
  private double    numericValue;
  /** String value for this datum  */
  private String    textValue;
  
  /** Flag indicating whether datum is numeric or text  */
  private boolean isNumericValue;
  
  /**
   * Constructor for numeric value
   * @param simDate Simulation Date stamp
   * @param variable Domain ID
   * @param numberValue Numeric value
   */
  SimulationDatum(SimDate simDate, String variable, double numberValue){
    sd = simDate;
    domain = variable;
    numericValue = numberValue;
    isNumericValue = true;
  }
  
  /**
   * Constructor for text value
   * @param simDate Simulation Date stamp
   * @param variable Domain ID
   * @param stringValue Text value
   */
  SimulationDatum(SimDate simDate, String variable, String stringValue){
    sd = simDate;
    domain = variable;
    textValue = stringValue;
    isNumericValue = false;
  }
  
  /** 
   * @see java.lang.Object#toString() 
   */
  public String toString(){
    String d = domain;
    while(d.length() < DOMAIN_TEXT_WIDTH) d+= " ";
    if(isNumericValue)
      return d + " at " + sd.getString() + " = " + numericValue;
    else
      return d + " at " + sd.getString() + " = " + textValue;
  }
  
  /**
   * Returns true if this datum point is numeric, false if not.
   * @return true if this datum point is numeric, false if not.
   */
  public boolean isNumeric(){
    return isNumericValue;
  }
  
  /**
   * Returns true if this datum point is textual, false if not.
   * @return true if this datum point is textual, false if not.
   */
  public boolean isText(){
    return !isNumericValue;
  }
  
  /**
   * Returns the domain of this Simulation Datum
   * @return the domain.
   */
  public String getDomain() {
    return domain;
  }

  /**
   * Returns the numeric value contained in this Simulation Datum
   * @return the numeric value contained in this Simulation Datum
   */
  public double getNumericValue() {
    return numericValue;
  }

  /**
   * Returns the simulation date for this Datum
   * @return the simulation date for this Datum
   */
  public SimDate getSimulationDate() {
    return sd;
  }

  /**
   * Returns the text value conatined in this Simulation Datum
   * @return the text value conatined in this Simulation Datum
   */
  public String getTextValue() {
    return textValue;
  }
  
  /**
   * Gets the Simulation Date of this datum point as a pure
   * numeric value (double).
   * @return a double indicating when during the simulation 
   * the datum point was taken.
   */
  public double timeValue(){
    return sd.value();
  } 
  
  
  /** 
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore) 
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    SimulationDatum newDatum = 
        (isNumericValue ? new SimulationDatum((SimDate)sd.getDeepClone(null), domain, numericValue) : 
                          new SimulationDatum((SimDate)sd.getDeepClone(null), domain, textValue));
    cloneStore.put(this, newDatum);
    return newDatum;
  }
}
