package abcm.sim;

import java.util.Date;

/**
 * A Simulation Run is a programmatic representation of the
 * record of a simulation being run. It includes elements
 * like a Run ID, the ID of the person who initiated the run,
 * the primary keys of the simulation, parameter set, and probe
 * set of the run, notes, a time run, and the mode in which
 * the simulation was run.
 * 
 * This object also allows for the creation of Simulation Runs
 * in conjunction with the HWM Database, from which the Run ID
 * is retrieved.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SimulationRun {

  
  /** Indicates that this run was executed in GUI mode  */
  public static final int GUI_MODE                = 0;
  /** Indicates that this run was executed in batch mode  */
  public static final int BATCH_MODE              = 1;
  /** Indicates that this run was executed in server mode  */
  public static final int SERVER_MODE             = 2;
  /** Indicates that this run was executed in command line mode  */
  public static final int COMMAND_LINE_MODE       = 3;
  /** Indicates that this run was executed in script mode  */
  public static final int SCRIPT_MODE             = 4;
  /** Indicates that this run was executed in the online interactive mode  */
  public static final int ONLINE_INTERACTIVE_MODE = 5;
  /** Indicates that this was run in a standalone application */
  public static final int STANDALONE_MODE         = 6;
  
  /** Unique ID for this run  */
  private long      simulationRunID;
  /** PK of user in database who initiated this run */
  private long      createdByID;
  /** PK of the Simulation used for this run  */
  private long      simulationPK;
  /** PK of the Parameter set used for this run  */
  private long      parameterSetPK;
  /** PK of the Probe set used for this run  */
  private long      probeSetPK;
  /** Notes associated with this run  */
  private String    notes;
  /** Mode in which this run was executed  */
  private int       mode;
  /** Date and time at which this run was created  */
  private Date      timeRun;
  /** Random number seed used for this run */
  private long      RNS;
  
   
  /**
   * Constructor that takes all information needed to constitute the run
   * @param RunID ID to be assigned to this run
   * @param UserID PK of user who initiated this run
   * @param SimPK PK of Simulation being used for this run
   * @param ParamSetPK PK of Parameter Set being used for this run
   * @param ProbeSetPK PK of Probe Set being used for this run
   * @param Notes Notes associated with this run
   * @param Mode Mode in which this run was executed
   * @param timeBegun Time at which this run was begun
   * @param randomNumberSeed random number seed to be used for this run
   */
  public SimulationRun(long RunID, 
                       long UserID, 
                       long SimPK, 
                       long ParamSetPK, 
                       long ProbeSetPK, 
                       String Notes, 
                       int Mode,
                       Date timeBegun,
                       long randomNumberSeed){
    
    simulationRunID = RunID;
    createdByID     = UserID;
    simulationPK    = SimPK;
    parameterSetPK  = ParamSetPK;
    probeSetPK      = ProbeSetPK;
    notes           = Notes;
    mode            = Mode;
    timeRun         = timeBegun;
    RNS             = randomNumberSeed;
        
  }

  /**
   * Returns the ID of the person who created the run.
   * @return the ID of the person who created the run
   */
  public long getCreatedByID() {        return createdByID;        }

  /**
   * Returns an integer indicating the mode in which the simulation was run
   * @return the mode.
   */
  public int getMode() {                return mode;               }

  /**
   * Returns any notes associated with this Simulation Run
   * @return any notes associated with this Simulation Run
   */
  public String getNotes() {            return notes;              }

  /**
   * Returns the primary key of the Parameter Set used for this Simulation Run
   * @return the primary key of the Parameter Set used for this Simulation Run
   */
  public long getParameterSetPK() {     return parameterSetPK;     }

  /**
   * Return the primary key of the Probe Set used for this Simulation Run
   * @return the primary key of the Probe Set used for this Simulation Run
   */
  public long getProbeSetPK() {         return probeSetPK;         }

  /**
   * Returns the primary key of the Simulation used for this Simulation Run
   * @return the primary key of the Simulation used for this Simulation Run
   */
  public long getSimulationPK() {       return simulationPK;       }

  /**
   * Returns the Run ID for this Simulation Run
   * @return the Run ID for this Simulation Run
   */
  public long getSimulationRunID() {    return simulationRunID;    }

  /**
   * Returns the time that this Simulation Run was created. Note that this
   * reflects the time the run was begun, not when it ended
   * @return the time that this Simulation Run was created
   */
  public Date getTimeRun() {            return timeRun;            }
  
  /**
   * @return the random number seed used for this simulation run
   */
  public long getRandomNumberSeed() {   return RNS;                }

}
