package abcm.space;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.graphics.X3DAssist;
import abcm.graphics.X3DRepresentable;
import abcm.math.Point;
import abcm.utilities.XMLAssist;

/**
 * DatumPoint
 * 
 * A DatumPoint is essentially a Point as defined in the hwmMath package,
 * but it can be drawn using the SVGRepresentable interface.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class DatumPoint extends Point implements SVGRepresentable, X3DRepresentable, ABCMDeepCloneable {

  /** Definition as appears in X3D documents  */
  private static String x3dDefinition = 
    "<Appearance DEF='Point'>\n" +
    "  <Material diffuseColor='0 1 0' emissiveColor='1 1 0'/>\n" +
    "</Appearance>\n";
  
  /** Flag to include labels on SVG representation  */
  private boolean includePointLabels = false;
  
  /* Constructors */
  /**
   * Creates a datum point with the specified 3-D coordinates. A name for the point is automatically assigned.
   * @param X x coordinate
   * @param Y y coordinate
   * @param Z z coordinate
   */
  public DatumPoint(double X, double Y, double Z){           super(X, Y, Z);      }
  
  /**
   * Creates a datum point with the specified 2-D coordinates; z is assumed to be zero.
   * @param X x coordinate
   * @param Y y coordinate
   */
  public DatumPoint(double X, double Y){                     super(X, Y);         }
  
  /**
   * Creates a datum point with the specified 3-D coordinates and name
   * @param X x coordinate
   * @param Y y coordinate
   * @param Z z coordinate
   * @param n name
   */
  public DatumPoint(double X, double Y, double Z, String n){ super(X, Y, Z, n);   }
  
  /**
   * Creates a datum point with the specified 2-d coordinates and name; z is assumed to be zero.
   * @param X x coordinate
   * @param Y y coordinate
   * @param n
   */
  public DatumPoint(double X, double Y, String n){           super(X, Y, n);      }
  
  
  /**
   * Sets the flag for whether point lables are to be included or not
   * @param include new value for flag
   */
  public void setIncludePointLabels(boolean include){ 
    includePointLabels = include; 
  }
  
  /**
   * Adds an SVG representation of this Point to an XML
   * document being constructed by the specified SVGAssist
   * object.
   * @param svg the SVGAssist object that is depicting this object
   */
  public void addToSVG(SVGAssist svg){
    class DatumPointDrawer implements SVGRepresentable{
      Point Parent;
      
      public DatumPointDrawer(Point parent){
        Parent = parent;
      }
      public void addToSVG(SVGAssist svg){
        // Points are always drawn with 1/20th inch radii
        svg.addCircle(X, Y, svg.getUnitsPerInchX()/50, "blue", "onclick=\"showPointLabel(evt, '" + ID + "','" + name + " (" + coordString() + ")');\"");
        if(includePointLabels){
          svg.addPopup(X, Y, 2d, .5, .5, .5, new SVGRepresentable[]{                // Note anonymous inner class to create point label
              new SVGRepresentable(){ 
                public void addToSVG(SVGAssist svg){
                  svg.addBackground("white");
                  svg.addText("Point: " + Parent.getName() + " " + Parent.coordString(3), 
                    .1, 15, 1);
                }
              }
            }, 
              "grey", 1d/100d, "Point: " + Parent.getID(), "visibility=\"hidden\" id=\"pointLabel_" + ID + "\"  onclick=\"hideSelf(evt,'pointLabel_" + ID + "');\"");
        }
      }
    }
    
    svg.addG(super.ID, "", new DatumPointDrawer(this));
    
  }
  
  

  /**
   * Adds an X3D representation of this Point to an xml document
   * being constructed by the specified XMLAssist object.
   * for this Point
   * @param x3d the X3DAssist object creating the output document
   */
  public void addToX3D(X3DAssist x3d){
    XMLAssist xml = x3d.getXML();
    x3d.addDefinition("Point", x3dDefinition);
    x3d.addBillboard(this, coordString());
    xml.open("Transform", "translation='" + getX() + " " + getZ() + " " + getY() + "'"); 
      xml.open("Shape");
        xml.addLine("Appearance", "USE='Point'","");
        xml.addLine("Sphere", "radius='0.1'","");
      xml.closeTag(); // Close the ball
      xml.open("Transform", "translation='0 0.5 0'");
        xml.open("Shape");
          xml.addLine("Appearance", "USE='Point'","");
          xml.addLine("Cylinder", "radius='0.05' height='0.8'","");
        xml.closeTag();
      xml.closeTag();
    xml.closeTag();
  }
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(ABCMCloneStore cloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    DatumPoint ret = new DatumPoint(this.X, this.Y, this.Z, this.name);
    cloneStore.put(this, ret);
    return ret;
  }
  
}
