package abcm.space;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.cloning.ABCMDeepCloneableVector;
import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.graphics.X3DAssist;
import abcm.graphics.X3DRepresentable;
import abcm.math.Point;
import abcm.math.Polygon;

/**
 * A Space object is the container object for the Canal Systems and Fields in
 * the HWM Simulation. The Space object not only provides programmatic access to
 * these members, but provides means of representing the entire space (in SVG or
 * X3D). <br/><br/> Copyright: 2005 Arizona State University<br/><br/>
 * 
 * Note: Spatial Elements and Spatial Element Group is under development.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Space implements ABCMDeepCloneable, SVGRepresentable, X3DRepresentable, HasElevation {

  // Public values
  /** Useful utility flag: defines E = + to be 'true' */
  public static final boolean EAST_POSITIVE = true;
  /** Useful utility flag: defines W = + to be 'false' */
  public static final boolean WEST_POSITIVE = false;
  /** Useful utility flag: defines N = + to be 'true' */
  public static final boolean NORTH_POSITIVE = true;
  /** Useful utility flag: defines S = + to be 'false' */
  public static final boolean SOUTH_POSITIVE = false;

  // Containers
  /** Collection of datum points that define the landscape */
  private ABCMDeepCloneableHashtable datumPoints          = new ABCMDeepCloneableHashtable();
  /** Collection of groups of spatial elements that are located within the landscape */
  private ABCMDeepCloneableHashtable spatialElementGroups = new ABCMDeepCloneableHashtable();
  /** Collection of spatial elements in the landscape */
  private ABCMDeepCloneableVector    spatialElements      = new ABCMDeepCloneableVector();
  
  /** Time of day when drawing 3-D representation  */
  protected double timeOfDay = .5; // Default is noon
  
  /** X3D representation of the daytime sky color  */
  protected float[][] daySkyColor   = new float[][] {{0f, 0.2f, 0.7f}, {0f, 0.5f, 1f}, {1f, 1f, 1f}}; 
    
  /** X3D representation of the nighttime sky color  */
  protected float[][] nightSkyColor = new float[][] {{0.0f, 0.0f, 0f}, {0.0f, 0.00f, 0.0f}, {0.3f, 0.3f, 0.3f}};

  
  
  
//  // Points added onto the surface are processed in bulk only
//  // when needed, to avoid overhead of regenerating topography for each one
//  // However, this is not cloned: points are added before cloning
  
  /**
   * Utility class that stores the information about a datum point.
   *  
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class SurfaceDatumPoint{ 
    /** Name to be assigned to this datum */
    public String name; 
    /** Point to be added */
    public DatumPoint point; 
    
    /**
     * Constructor
     * @param n name
     * @param P datum point
     */
    public SurfaceDatumPoint(String n, DatumPoint P){ name = n; point = P; } }
  
  /**
   * List of surface points to be added in bulk
   */
  private List<SurfaceDatumPoint> surfacePoints = new ArrayList<SurfaceDatumPoint>();

  // This should be called before every get operation that needs the full
  // list of points, but will return very quickly if there are none to be added
  /**
   * Internally called method that adds all of the surface points in the surfacePoints list
   * to the actual collection of datum points.</br>
   * 
   * The issue is processing speed. If a point is added that defines a new z coordinate
   * the terrain must be regenerated. Alternatively, if a point is added that is supposed
   * to be placed on the existing surface, it will not change the original terrain.
   * Therefore, it is possible to add multiple points to the existing surface in bulk.
   */
  private void processSurfacePoints(){
    if(surfacePoints.size() == 0) return;
    TriangulatedTopography tt =  getFundamentalTriangulatedTopographyWithoutSurfacePoints();
    Vector V = (tt == null ? new Vector() : getFundamentalTriangulatedTopographyWithoutSurfacePoints().getTriangles());
    while(surfacePoints.size() > 0){
      SurfaceDatumPoint sdp = surfacePoints.remove(0);
      boolean inside = false;
      int i = 0;
      while((inside == false) && (i < V.size())){
        double z = ((TopoTriangle)V.get(i)).getZ(sdp.point);
        if(!Double.isNaN(z)){
          addPoint(sdp.name, new DatumPoint(sdp.point.getX(), sdp.point.getY(), z)); // Note that if this is called the topography gets reset to null
          inside = true;
        }
        i++;
      }  
      if(!inside){
        addPoint(sdp.name, sdp.point); // If it's not within the boundaries, just add it, but note that it doesn't effect the current topography!
      }
    }
  }
  
  
  /**
   * Adds a point whose x,y coordinate is given but whose coordinate will be determined by the existing
   * terrain.
   * @param name Name by which the point is to be indexed
   * @param point datum point to be stored
   * @return true (does not fail)
   */
  public boolean addSurfacePoint(String name, DatumPoint point){
    surfacePoints.add(new SurfaceDatumPoint(name, point));
    return true;
  }

  
  // Switches
  private boolean eastIsPositive  = EAST_POSITIVE;
  private boolean northIsPositive = NORTH_POSITIVE;
  public  boolean drawPoints = true;
  public  boolean includeContourLines = false;
  public  boolean includeDelaunayTriangulation = false;
  
  // Ranges
  private double maxX;
  private double minX;
  private double maxY;
  private double minY;
  private double maxZ;
  private double minZ;

  // Internal ranges for display calculations
  private double displayWesternBoundary;
  private double displayNorthernBoundary;
  private double displayEasternBoundary;
  private double displaySouthernBoundary;  
  private double contourLineMin;
  private double contourLineMax;
  private double contourLineInterval;

  /** Delaunay Triangulation (memoized)  */
  private TriangulatedTopography fundamentalTriangulatedTopography;
  private TriangulatedTopography fullTriangulatedTopography;
  
  /**
   * Constructor that sets orientation according to defaults
   */
  public Space() {  }

  /**
   * Constructor that allows orientation to be specified
   * 
   * @param eastIsPos
   *          true if x coordinates increase in eastward direction, false
   *          otherwise
   * @param northIsPos
   *          true if y coordinates increase in northern direction, false
   *          otherwise
   */
  public Space(boolean eastIsPos, boolean northIsPos) {
    eastIsPositive  = eastIsPos;
    northIsPositive = northIsPos;
  }

  /* Add elements */

  /**
   * Add a point to this Space.
   * 
   * Tracks the maximum and minimum X and Y coordinates entered.
   * 
   * Note that no error will occur if a Point with the specified name has
   * already been entered; the new point will overwrite the old one.
   * 
   * @param name name of the point being entered
   * @param dp   DatumPoint being entered
   */
  public void addPoint(String name, DatumPoint dp) {
    // If this is the first point added
    setMinAndMax(dp.getX(), dp.getX(), dp.getY(), dp.getY(), dp.getZ(), dp.getZ());

    // If there is already a point with this name, remove it
    if (datumPoints.containsKey(name))   datumPoints.remove(name);
    
    // Add this point
    datumPoints.put(name, dp);

    // Set the Delaunay Triangulation to null; this must be rebuilt when a point is added
    fundamentalTriangulatedTopography = null;
    
  }

  /**
   * Adds a spatial element to this Space object, to the specified group.
   * @param group group to which it is to be added
   * @param name name that it is to be assigned
   * @param element the element to be added
   */
  public boolean addSpatialElement(String group, String name, SpatialElement element){
    processSurfacePoints();
    ABCMDeepCloneableHashtable spatialElementGroup = getSpatialElementGroupHashtable(group);
    if(spatialElementGroup.containsKey(name)) return false; // Does NOT overwrite extant object with the same name!
    
    // Adjust min and max values of the space based on the new values for this object
    setMinAndMax(element.getMaxX(), element.getMinX(), 
                 element.getMaxY(), element.getMinY(), 
                 element.getMaxZ(), element.getMinZ());
    
    // And store object
    spatialElementGroup.put(name, element);
    spatialElements.add(element);
    return true;
  }
  
  /**
   * Removes a spatial element
   * @param group group to which the element belongs
   * @param name name assigned to the element
   * @return true (in all cases, even if the element did not originally appear in the group)
   */
  public boolean removeSpatialElement(String group, String name){
    processSurfacePoints();
    ABCMDeepCloneableHashtable spatialElementGroup = getSpatialElementGroupHashtable(group);
    if(spatialElementGroup.containsKey(name)){
      SpatialElement element = (SpatialElement)spatialElementGroup.remove(name);
      if(element != null) spatialElements.remove(element);
    }
    // Technically the space's min and max values should be re-set, but this shouldn't
    // matter often, and it may even be nice to leave it as if the object were still there...
    return true;
  }
  
  /**
   * Gets the table of spatial element groups
   * @param groupName the name of the group requested
   * @return the Hashtable representing that group
   */
  private ABCMDeepCloneableHashtable getSpatialElementGroupHashtable(String groupName){
    if(!spatialElementGroups.containsKey(groupName)) spatialElementGroups.put(groupName, new ABCMDeepCloneableHashtable());
    return (ABCMDeepCloneableHashtable)spatialElementGroups.get(groupName);
  }
  
  /**
   * Sets the upper and lower coordinate boundaries for the space
   * @param xHigh x maximum
   * @param xLow  x minimum
   * @param yHigh y maximum
   * @param yLow  y minimum
   * @param zHigh z maximum
   * @param zLow  z minimum
   */
  private void setMinAndMax(double xHigh, double xLow, double yHigh, double yLow, double zHigh, double zLow){
    if ((datumPoints.size() == 0) && (spatialElementGroups.size() == 0)) {
      maxX = xHigh;
      minX = xLow;
      maxY = yHigh;
      minY = yLow;
      maxZ = zHigh;
      minZ = zLow;
    } else {
      if (xHigh > maxX)
        maxX = xHigh;
      if (xLow < minX)
        minX = xLow;
      if (yHigh > maxY)
        maxY = yHigh;
      if (yLow < minY)
        minY = yLow;
      if (zHigh > maxZ)
        maxZ = zHigh;
      if (zLow < minZ)
        minZ = zLow;
    }
    
    resetDisplayRanges();
  }
  

  /* Maximum and Minimum extents */

  /**
   * Returns the maximum X coordinate of all Points entered
   * 
   * @return Returns the maxX.
   */
  public double getMaxX() {
    processSurfacePoints();
    if (maxX == minX)
      return maxX + 1;
    else
      return maxX;
  }

  /**
   * Returns the maximum Y coordinate of all Points entered
   * 
   * @return Returns the maxY.
   */
  public double getMaxY() {
    processSurfacePoints();
    if (maxY == minY)
      return maxY + 1;
    else
      return maxY;
  }

  /**
   * Gets the minimum X value used by any point in this space
   * 
   * @return Returns the minX.
   */
  public double getMinX() {
    processSurfacePoints();
    if (minX == maxX)
      return minX - 1;
    else
      return minX;
  }

  /**
   * Gets the minimum Y value used by any point in this space
   * 
   * @return Returns the minY.
   */
  public double getMinY() {
    processSurfacePoints();
    if (minY == maxY)
      return minY - 1;
    else
      return minY;
  }

  /**
   * Gets the minimum Z value used by any point in this space 
   * @return the minZ
   */
  public double getMinZ() {
    processSurfacePoints();
    return minZ;
  }

  /**
   * Gets the maximum Z value used by any point in this space
   * @return the maxZ
   */
  public double getMaxZ() {
    processSurfacePoints();
    return maxZ;
  }

  /* Get elements */

  /**
   * Retrieves a point from this Space object by name
   * @param datumPointName the name of the requested point
   * @return the requested Point; will return null if the Point has not been
   *         entered.
   */
  public DatumPoint getPoint(String datumPointName) {
    processSurfacePoints();
    return (DatumPoint) datumPoints.get(datumPointName);
  }

  /**
   * Resets the display ranges to match the extent of the points contained
   * by the space object
   */
  public void resetDisplayRanges(){
    processSurfacePoints();
    if(eastIsPositive){
      displayWesternBoundary = minX;
      displayEasternBoundary = maxX;
    }
    else{
      displayWesternBoundary = maxX;
      displayEasternBoundary = minX;
    }
    if(northIsPositive){
      displayNorthernBoundary = maxY;
      displaySouthernBoundary = minY;
    }
    else{
      displayNorthernBoundary = minY;
      displaySouthernBoundary = maxY;
    }
  }
  
  /**
   * Sets the boundaries of the area that will be displayed.
   * Returns false if the ranges are invalid (a minimum specified
   * is greater than or equal to its corresponding maximum),
   * true otherwise.
   * @param w western boundary
   * @param n northern boundary
   * @param e eastern boundary
   * @param s southern boundary
   * @return true if the result was successful; false if the boundaries given are invalid
   */
  public boolean setDisplayRange(double w, double n, double e, double s){
    System.out.println("Setting display ranges " + w + " " + n + " " + e + " " + s);
    if(eastIsPositive   && w > e) return false;
    if(!eastIsPositive  && e > w) return false;
    if(northIsPositive  && s > n) return false;
    if(!northIsPositive && n > s) return false;
    displayWesternBoundary  = w;
    displayNorthernBoundary = n;
    displayEasternBoundary  = e;
    displaySouthernBoundary = s;
    return true;
  }
    
  /* Display */
  
  /** 
   * @see abcm.graphics.SVGRepresentable#addToSVG(abcm.graphics.SVGAssist) 
   */
  public void addToSVG(SVGAssist svg){
    
    
    /**
     * A ContourLineGroup is the collection of contour lines
     */
    class ContourLineGroup implements SVGRepresentable{
      private Vector clines_vector = new Vector();
      
      /**
       * A Cline is a collection of contour lines at a given
       * elevation (though not necessarily sequential or continuous) 
       */
      class Cline implements SVGRepresentable{
        private Vector contourLines;
        private double value;
        
        /**
         * Simple constructor that sets the instance values
         * @param linesVector the collection of contour lines
         * @param val the elevation of this set of contour lines
         */
        public Cline(Vector linesVector, double val){       
          contourLines = linesVector;
          value = val;
        }

        /**
         * Adds this collection of contour lines to the svg document
         */
        public void addToSVG(SVGAssist svg){
          if(contourLines.size() > 0){
            String otherTags = "";
            svg.addG(0,0,"Contour Lines At " + value, otherTags, SVGAssist.getArrayFromCollection(contourLines)); 
          }
        }
      }
        
      /**
       * Adds a collection of contour lines, as a Cline, to the ContourLineGroup
       * @param V vector of contour lines
       * @param val elevation
       */
      public void addCline(Vector V, double val){
        clines_vector.add(new Cline(V, val));
      }
        
      /**
       * Adds this ContourLineGroup to the svg document
       * @param svg SVG Assist object being used to create the svg document
       */
      public void addToSVG(SVGAssist svg){
        svg.addG("Contour Lines","id=\"contourLines\" visibility=\"display\"", SVGAssist.getArrayFromCollection(clines_vector));
      }
    }

    /**
     * A GridlineGroup is a collecion of gridlines. As a convenience, it includes
     * the extreme values for the extent of the lines in both directions
     */
    class GridlineGroup implements SVGRepresentable{
      Vector glines_vector = new Vector();
      double minX, minY, maxX, maxY;
      
      /**
       * A helper class representing one gridline. A gridline can be
       * either 'latitude' (running W-E and having a N/S value) or
       * 'longitude' (running N-S and having an E/W value). 
       */
      class Gridline implements SVGRepresentable{
        boolean lat;
        double val;
        double end1, end2;
        
        /**
         * Creates a Gridline; a gridline is either a latitude or a longitude, and
         * has a value (either N/S or E/W) and either an W-E extent or a N-S extent.
         * @param value the value at which this Gridline appears
         * @param end_1 the N or W extent of this Gridline
         * @param end_2 the S or E extent of this Gridline
         * @param latitude whether this line is a latitude or longitude
         */
        public Gridline(double value, double end_1, double end_2, boolean latitude){
          lat  = latitude;
          val  = value;
          end1 = end_1;
          end2 = end_2;
        }
        
        /**
         * Adds this Gridline to an SVG document
         * @param svg the SVGAssist object being used to create this SVG document
         */
        public void addToSVG(SVGAssist svg){
          double x1 = (lat ? end1 : val);
          double y1 = (lat ? val  : end1);
          double x2 = (lat ? end2 : val);
          double y2 = (lat ? val  : end2);
          double w  = (lat ? svg.getUnitsPerInchY() : svg.getUnitsPerInchY())/ 100d;
          svg.addLine(x1, y1, x2, y2, (val == 0 ? "red" : "gray"), w);
          // And add the label
          // ?
        }
      }

      /**
       * A simple constructor that creates a GridlineGroup
       * @param westLimit the lowest X value
       * @param northLimit the lowest Y value
       * @param eastLimit the highest X value
       * @param southLimit the highest Y value
       */
      public GridlineGroup(double westLimit, double northLimit, double eastLimit, double southLimit){
        minX = westLimit;
        minY = northLimit;
        maxX = eastLimit;
        maxY = southLimit;
      }
      
      /**
       * This method adds a grid line to this GridLineGroup.
       * @param value the value (either N/S or E/W)
       * @param latitude whether this line is a latitude or longitude
       */
      public void addGLine(double value, boolean latitude){
        double end_1 = (!latitude ? minY : minX);
        double end_2 = (!latitude ? maxY : maxX);
        glines_vector.add(new Gridline(value, end_1, end_2, latitude));
      }
      
      /**
       * Adds this GridlineGroup to the svg document
       * @param svg the SVGAssist object being used to create this svg document
       */
      public void addToSVG(SVGAssist svg){
        svg.addG("Grid Lines","id=\"gridlines\"", SVGAssist.getArrayFromCollection(glines_vector));
      }
    }
    
    // Begin display creation
    processSurfacePoints();
    
    // This will only run if there has been no initialization
    if(displayWesternBoundary == displayEasternBoundary || displayNorthernBoundary == displaySouthernBoundary) setDisplayRange(minX, minY, maxX, maxY);
    
    // Determine the actual ranges to display. This will depend on the ranges
    // of the SVG viewport and the actual page dimensions
    double displayAreaWidth  = svg.getViewportWidthInInches();
    double displayAreaHeight = svg.getViewportHeightInInches(); 
    
    if(displayAreaWidth  == 0) displayAreaWidth  = 1;
    if(displayAreaHeight == 0) displayAreaHeight = 1;
    
    double xRange = Math.abs(displayEasternBoundary - displayWesternBoundary);
    double yRange = Math.abs(displaySouthernBoundary - displayNorthernBoundary);
    
    double xCompression = xRange / displayAreaWidth;
    double yCompression = yRange / displayAreaHeight;
    
    System.out.println("xComp: " + xCompression + " yComp: " + yCompression);
    
    // If the x axis has to be compressed more to fit, then
    // rescale the y ranges, and vice versa. Keep the current center.
    if(xCompression > yCompression){
      yRange = xRange * ( displayAreaHeight / displayAreaWidth);
    }
    else{
      xRange      = yRange * ( displayAreaWidth / displayAreaHeight);
    } 
    
    
    if(includeDelaunayTriangulation)  getFundamentalTriangulatedTopography().addToSVG(svg);
    
    // Contour lines
    if(includeContourLines){
      setContourLineIntervals();
           
      ContourLineGroup clg = new ContourLineGroup();
      for(double d = contourLineMin; d < contourLineMax; d+= contourLineInterval){
        Vector V = getFundamentalTriangulatedTopography().getContourLines(d);
        if(V != null) clg.addCline(V, d);
      }
      clg.addToSVG(svg);
    }
    
    // Gridlines    
    GridlineGroup glg = new GridlineGroup(displayWesternBoundary, displayNorthernBoundary, displayEasternBoundary, displaySouthernBoundary);
    System.out.println("   displayWesternBoundary  : " + displayWesternBoundary);
    System.out.println("   displayEasternBoundary  : " + displayEasternBoundary);
    System.out.println("   displayNorthernBoundary : " + displayNorthernBoundary);
    System.out.println("   displaySouthernBoundary : " + displaySouthernBoundary);
    
    // Calculate the intervals and ranges for N/S gridlines (showing E-W position)
    int magnitude = (int)(Math.floor((Math.log(xRange)/Math.log(10))));
    int ticCount = 0;
    for(int c = 0; c < 10; c++){
      double val = c * Math.pow(10, magnitude);
      // These can't both be true; accommodates east/westIsPositive
      if((val >= displayWesternBoundary) && (val <= displayEasternBoundary)){     ticCount++;    }
      if((val >= displayEasternBoundary) && (val <= displayWesternBoundary)){     ticCount++;    }
    }
    
    double roughHashInterval = svg.getViewportWidthInInches() / ticCount;
    double inc = 1;
    double GRIDLINE_INTERVAL = 1;
    while(roughHashInterval > GRIDLINE_INTERVAL){
      roughHashInterval /= 2;
      inc /= 2;
    }
      
    for(double c = -10; c < 10; c+= inc){
      double val = c * Math.pow(10, magnitude);
      // See comment above
      if((val >= displayWesternBoundary) && (val <= displayEasternBoundary)){  glg.addGLine(val, false); }
      if((val >= displayEasternBoundary) && (val <= displayWesternBoundary)){  glg.addGLine(val, false); }
    }
    
    for(double c = -10; c < 10; c+= inc){
      double val = c * Math.pow(10, magnitude);
      // See comment above
      if((val >= displaySouthernBoundary) && (val <= displayNorthernBoundary)){  glg.addGLine(val, true);  }
      if((val >= displayNorthernBoundary) && (val <= displaySouthernBoundary)){  glg.addGLine(val, true);  }
    }
    
    glg.addToSVG(svg);

    // Add each collection
    svg.addG(0,0,"Points","",SVGAssist.getArrayFromHashtable(datumPoints));
    
    Enumeration E = spatialElementGroups.keys();
    while(E.hasMoreElements()){
      String group = (String)E.nextElement();
      svg.addG(0,0, group,"",SVGAssist.getArrayFromHashtable(getSpatialElementGroupHashtable(group)));
    }
    
  } 

  /**
   * @see abcm.graphics.X3DRepresentable#addToX3D(abcm.graphics.X3DAssist)
   */
  public void addToX3D(X3DAssist x3dAssist){
    processSurfacePoints();
    // Create ground and sky...
    x3dAssist.addSkyAndGround(new float[]{1.309f, 1.571f},
                              ((timeOfDay >= .25 && timeOfDay <= .75) ? daySkyColor : nightSkyColor),
                              new float[]{1.309f, 1.571f},
                              new float[][] {{0.1f, 0.1f, 0f}, {0.4f, 0.25f, 0.2f}, {0.6f, 0.6f, 0.6f}});
    
    x3dAssist.addViewpoint("DEFAULT", (this.getMaxX() + this.getMinX()) / 2d, (this.getMaxY() + this.getMinY()), getMeanElevation() + 2);
    x3dAssist.addNavigation(X3DAssist.FLY, true, 256, false);
    x3dAssist.addSunlight(timeOfDay);
    
    Enumeration e = datumPoints.elements();
    while(e.hasMoreElements()){
      DatumPoint dp = (DatumPoint)e.nextElement();
      dp.addToX3D(x3dAssist);
    }
    getTriangulatedTopography().addToX3D(x3dAssist);
    
    // Add all spatial elements
    Enumeration elements = spatialElements.elements();
    while(elements.hasMoreElements()){
      SpatialElement se = (SpatialElement)elements.nextElement();
      se.addToX3D(x3dAssist);      
    }
    
  }
  
  /**
   * Sets the interval at which contour lines will be drawn
   */
  private void setContourLineIntervals() {
    processSurfacePoints();
    double zRange = maxZ - minZ;
    int magnitude = (int) (Math.floor((Math.log(zRange) / Math.log(10))));

    double roughInterval = Math.pow(10, magnitude);
    double inc = 1;
    while ((zRange / roughInterval) < 8) {
      roughInterval /= 2;
      inc /= 2;
    }

    double increment = inc * Math.pow(10, magnitude);
    double lowRange = Math.floor(minZ / increment) * increment
        + (minZ < 0 ? -1 : 0);
    double highRange = Math.floor(maxZ / increment) * increment
        + (maxZ < 0 ? -1 : 0);

    contourLineInterval = increment;
    contourLineMin = lowRange;
    contourLineMax = highRange;
  }

  /**
   * Gets the interval at which contour lines are to be drawn
   * @return the contour line interval
   */
  public double getContourLineInterval() {
    processSurfacePoints();
    setContourLineIntervals();
    return contourLineInterval;
  }

  
  /* Basic interrogatives */

  /**
   * Determines if the Space object contains a Point object with the specified
   * name
   * 
   * @param pointName
   *          the name of the Point being sought
   * @return true if the Point is contained, false otherwise
   */
  public boolean containsPoint(String pointName) {
    processSurfacePoints();
    return datumPoints.containsKey(pointName);
  }

  /**
   * Gets whether this Space object is oriented so that Y coordinates increase
   * north or south
   * 
   * @return true if north is positive, false if south is positive
   */
  public boolean northPositive() {
    return northIsPositive;
  }

  /**
   * Gets whether this Space object is oriented so that X coordinates increase
   * east or west
   * 
   * @return true if east is positive, false if west is positive
   */
  public boolean eastPositive() {
    return eastIsPositive;
  }

  /**
   * Gets the fundamental triangulated topography for this space; this
   * is the topography without any spatial elements
   * @return the basic topography as a Triangulated Topography
   */
  public TriangulatedTopography getFundamentalTriangulatedTopography() {
    processSurfacePoints();
    return getFundamentalTriangulatedTopographyWithoutSurfacePoints();
  }
  
  /**
   * This returns the triangulated topography without processing surface points
   * @return the fundamental triangulated topography for this space.
   */
  protected TriangulatedTopography getFundamentalTriangulatedTopographyWithoutSurfacePoints(){
    if (fundamentalTriangulatedTopography != null) return fundamentalTriangulatedTopography;
    if(datumPoints.size() < 3) fundamentalTriangulatedTopography = null;
    else                       fundamentalTriangulatedTopography = new TriangulatedTopography(datumPoints.values());
    return fundamentalTriangulatedTopography;
  }
  

  /**
   * Gets the triangulated topography for this Space object,
   * including spatial objects (SpatialElement)it contains
   * @return the triangulated topography for this Space
   */
  public TriangulatedTopography getTriangulatedTopography(){
    processSurfacePoints();
    System.out.println("Getting triangulated topography for space!!!!!");
    if(fullTriangulatedTopography != null){
      boolean newTT = false;
      for(int i = 0; i < spatialElements.size(); i++){
        SpatialElement SE = (SpatialElement)spatialElements.get(i);
        if(SE.altered() == true){
          newTT = true;
          break;
        }
      }
      if(newTT == false) return fullTriangulatedTopography;
    }
    // At least one of the spatial elements has not been polled
    // in its current state; TriangulatedTopography must be rebuilt
    Vector boundaryPolygons = new Vector();
    Vector bermPolygons = new Vector();
    for(int i = 0; i < spatialElements.size(); i++){
      SpatialElement SE = (SpatialElement)spatialElements.get(i);
      boundaryPolygons.addAll(SE.getBoundaryPolygons());
      bermPolygons.addAll(SE.getBermPolygons());
    }   
    fullTriangulatedTopography = getFundamentalTriangulatedTopography();
    fullTriangulatedTopography.punchHoles(boundaryPolygons, bermPolygons);
    for(int i = 0; i < spatialElements.size(); i++){
      fullTriangulatedTopography.addTopoTriangles(((SpatialElement)spatialElements.get(i)).getTriangulatedTopographyWithinBoundaryPolygons().getTriangles());
    }
    return fullTriangulatedTopography;
  }
  
  
  
  /**
   * Gets the mean elevation points in the space- a simple (max + min) / 2
   * 
   * @return the mean elevation for the space, given the points entered so far
   */
  public double getMeanElevation() {
    processSurfacePoints();
    return (getMinZ() + getMaxZ()) / 2;
  }

  /* Other interrogatives */

  /**
   * Gets a subset of the Delaunay triangulation comprised only of those
   * triangles that have any point that falls within the boundaries of the four
   * points given.
   * 
   * @param p1 corner of the boundary box
   * @param p2 corner of the boundary box
   * @param p3 corner of the boundary box
   * @param p4 corner of the boundary box
   * @return Vector of triangles touching or contained in the boundary box
   */
  public Vector getDelaunaySubset(Point p1, Point p2, Point p3, Point p4) {
    processSurfacePoints();
    return getFundamentalTriangulatedTopography().getSubset(p1, p2, p3, p4);
  }


  
  /* Reporting */
  /**
   * Gets a list of the points known to this model
   * 
   * @return a String containing a list of points known to the model
   */
  public String getListOfPoints() {
    processSurfacePoints();
    StringWriter out = new StringWriter();
    out.write("\nPoints List:\n");
    Enumeration e = datumPoints.keys();
    if (e.hasMoreElements()) {
      while (e.hasMoreElements()) {
        String pointKey = (String) e.nextElement();
        out.write("  " + pointKey + ": ");
        Point p = (Point) datumPoints.get(pointKey);
        out.write(p.coordString() + "\n");
      }
    } else {
      out.write(" No points\n");
    }
    return out.toString();
  }

  /**
   * Gets the eastern boundary of the display
   * @return the eastern boundary of the display
   */
  public double getDisplayEasternBoundary() {   
    return displayEasternBoundary;  
  }

  /**
   * Gets the western boundary of the display
   * @return the eastern boundary of the display
   */
  public double getDisplayWesternBoundary() {    return displayWesternBoundary;  }

  /**
   * Gets the southern boundary of the display
   * @return the eastern boundary of the display
   */
  public double getDisplaySouthernBoundary() {   return displaySouthernBoundary;  }

  /**
   * Gets the northern boundary of the display
   * @return the eastern boundary of the display
   */
  public double getDisplayNorthernBoundary() {   return displayNorthernBoundary;  }

  /**
   * Gets the flag indicating if east is represented by positive coordinates
   * @return true if east is positive
   */
  public boolean getEastIsPositive() {
    return eastIsPositive;
  }

  /**
   * Sets the 'east is positive' flag
   * @param eastIsPositive new value
   */
  public void setEastIsPositive(boolean eastIsPositive) {
    this.eastIsPositive = eastIsPositive;
    resetDisplayRanges();
  }

  /**
   * Gets the flag indicating if north is represented by positive coordinates
   * @return true if north is positive
   */  
  public boolean getNorthIsPositive() {
    return northIsPositive;
  }

  /**
   * Sets the 'north is positive' flag
   * @param northIsPositive new value
   */
  public void setNorthIsPositive(boolean northIsPositive) {
    this.northIsPositive = northIsPositive;
    resetDisplayRanges();
  }

  
  /**
   * @see abcm.space.HasElevation#getZ(abcm.math.Point)
   */
  public double getZ(Point P){
    processSurfacePoints();
    if(P.getX() < this.minX || P.getX() > this.maxX || P.getY() < this.minY || P.getY() > this.maxY) return Double.NaN;
    // If P falls within the boundaries of any spatial element, use that
    // spatial element's getZ
    Enumeration E = spatialElements.elements();
    while(E.hasMoreElements()){
      SpatialElement se = (SpatialElement)E.nextElement();
      Vector V = se.getBoundaryPolygons();
      for(int i = 0; i < V.size(); i++){
        Polygon poly = (Polygon)V.get(i);
        if(poly.pointIsInside(P)) {
          return se.getZ(P);
        }
      }
    }
    TriangulatedTopography tt = getFundamentalTriangulatedTopography();
    Vector V = tt.getTriangles();
    for(int i = 0; i < V.size(); i++){
      TopoTriangle T = (TopoTriangle)V.get(i);
      double z = T.getZ(P);
      if(!Double.isNaN(z)){
        return z;  // NaN is the return value if the point is not inside the TopoTriangle; no need to do the test twice
      }
    }
    return Double.NaN;
  }

  /**
   * Child classes will want to override this...
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    processSurfacePoints();
    Space space = new Space();
    cloneStore.put(this, space);
    cloneCoreSpaceObjects(space, cloneStore);
    return space;
  }

  /**
   * Separate method for cloning the core objects of a space object.
   * @param newSpace The new space object being created
   * @param cloneStore  The clone store associated with this cloning event
   */
  protected void cloneCoreSpaceObjects(Space newSpace, ABCMCloneStore cloneStore){
    newSpace.timeOfDay                         = timeOfDay;
    newSpace.daySkyColor                       = daySkyColor;
    newSpace.nightSkyColor                     = nightSkyColor;
    newSpace.eastIsPositive                    = eastIsPositive;
    newSpace.northIsPositive                   = northIsPositive;
    newSpace.drawPoints                        = drawPoints;
    newSpace.includeContourLines               = includeContourLines;
    newSpace.includeDelaunayTriangulation      = includeDelaunayTriangulation;
    newSpace.maxX                              = maxX;
    newSpace.minX                              = minX;
    newSpace.maxY                              = maxY;
    newSpace.minY                              = minY;
    newSpace.maxZ                              = maxZ;
    newSpace.minZ                              = minZ;
    newSpace.displayWesternBoundary            = displayWesternBoundary;
    newSpace.displayNorthernBoundary           = displayNorthernBoundary;
    newSpace.displayEasternBoundary            = displayEasternBoundary;
    newSpace.displaySouthernBoundary           = displaySouthernBoundary;  
    newSpace.contourLineMin                    = contourLineMin;
    newSpace.contourLineMax                    = contourLineMax;
    newSpace.contourLineInterval               = contourLineInterval;
    processSurfacePoints();
    newSpace.datumPoints                       = (datumPoints          != null ? (ABCMDeepCloneableHashtable)datumPoints.getDeepClone(cloneStore) : null);
    newSpace.spatialElementGroups              = (spatialElementGroups != null ? (ABCMDeepCloneableHashtable)spatialElementGroups.getDeepClone(cloneStore) : null);
    newSpace.spatialElements                   = (spatialElements      != null ? (ABCMDeepCloneableVector)spatialElements.getDeepClone(cloneStore) : null);
    
    newSpace.fundamentalTriangulatedTopography = (fundamentalTriangulatedTopography != null ? (TriangulatedTopography)fundamentalTriangulatedTopography.getDeepClone(cloneStore) : null);
    newSpace.fullTriangulatedTopography        = (fullTriangulatedTopography        != null ? (TriangulatedTopography)fullTriangulatedTopography.getDeepClone(cloneStore) : null);
  
  }
}

