package abcm.space;

import java.util.Vector;

import abcm.cloning.ABCMDeepCloneable;
import abcm.graphics.SVGRepresentable;
import abcm.graphics.X3DRepresentable;

/**
 * A SpatialElement is an element that can be added into a landscape and the
 * landscape, as defined by datum points, is not changed. 
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public interface SpatialElement extends SVGRepresentable, X3DRepresentable, HasElevation, ABCMDeepCloneable {

  /**
   * Gets the list of polygons that define the boundaries of this
   * SpatialElement. For example, an item being placed into a
   * landscape may have multiple polygons for its own representation,
   * but the boundary polygons are the ones that mark the outside
   * edges of the object.
   * @return a collection of boundary polygons
   */
  public Vector getBoundaryPolygons();

  /**
   * Gets the list of polygons that define a 'buffer zone' around this spatial
   * element.
   * @return collection of polygons that define the berm around this spatial elements
   */
  public Vector getBermPolygons();

  /**
   * Should return 'true' if the berm or boundary polygon
   * set would be different since this was last polled
   * (and when first polled). This allows memoization.
   * @return true if the spatial element has been changed since last rendered,
   * false otherwise
   */
  public boolean altered();
  
  /**
   * The maximum X coordinate for this spatial element
   * @return maximum X coordinate
   */
  public double getMaxX();
  
  /**
   * The maximum Y coordinate for this spatial element
   * @return maximum Y coordinate
   */
  public double getMaxY();
  
  /**
   * The maximum Z coordinate for this spatial element
   * @return maximum Z coordinate
   */
  public double getMaxZ();
  
  /**
   * The minimum X coordinate for this spatial element
   * @return minimum X coordinate
   */
  public double getMinX();
  
  /**
   * The minimum Y coordinate for this spatial element
   * @return minimum Y coordinate
   */
  public double getMinY();
  
  /**
   * The minimum Z coordinate for this spatial element
   * @return minimum Z coordinate
   */
  public double getMinZ();

  /**
   * Get the triangulated topography within the boundary polygons
   * @return a triangulated topography representing this spatial element
   * within the boundary polygons
   */
  public TriangulatedTopography getTriangulatedTopographyWithinBoundaryPolygons();
  
}
