package abcm.space;

import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.math.Geometry;
import abcm.math.LineSegment;
import abcm.math.Point;
import abcm.math.Triangle;

/** 
 *  A Triangle object extended to provide basic functions related to
 *  using this as a Delaunay cell representing a surface or landscape.
 *   
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class TopoTriangle extends Triangle implements SVGRepresentable, HasElevation, ABCMDeepCloneable {
  
  /** Debug flag for debugging the volume differential calculations */
  private static boolean DEBUG_VOL_DIFF = false;
  
  /** SVG fill to be used for a standard topo triangle  */
  private static final String TOPO_TRIANGLE_FILL   = "none";
  /** SVG stroke (color) to be used for a standard topo triangle  */
  private static final String TOPO_TRIANGLE_STROKE = "blue";
  
  /** ID counter */
  static int count = 0;

  
  /** ID Assigned to this triangle */
  public String ID = "T" + (count++); 
  
  /**
   * Constructor
   * @param P1 end point number 1
   * @param P2 end point number 2
   * @param P3 end point number 3
   */
  public TopoTriangle(Point P1, Point P2, Point P3){ 
    super(P1, P2, P3);
  }
  
  /**
   * Constructor
   * @param T A Triangle object from which this TopoTriangle is to be made
   */
  public TopoTriangle(Triangle T){
    super((Point)T.getPoints().get(0), (Point)T.getPoints().get(1), (Point)T.getPoints().get(2));
    if(getMinZ() < -50){
      System.out.println("TOO LOW!!! Attemption to construct a topo triangle with a z value of " + getMinZ());
    }
  } 
  
  /**
   * Get a Contour Line that crosses this TopoTriangle at the specified
   * elevation
   * @param z the requested elevation
   * @return a Line object that crosses this triangle at the specified elevation
   */
  public ContourLine getContourLine(double z){
    if(!includesElevation(z)) return null;
    LineSegment L = super.getCrossLine(z);
    if(L == null) return null;
    return new ContourLine(L.P1, L.P2, z);    
  }
  
  /** 
   * @see abcm.graphics.SVGRepresentable#addToSVG(abcm.graphics.SVGAssist) 
   */
  public void addToSVG(SVGAssist svg){
    String pointsList = SVGAssist.getPointsList(points);    
    double stroke_width = (svg.getUnitsPerInchX() / 100d) * 2d;
    svg.addPolygon(pointsList, TOPO_TRIANGLE_FILL, TOPO_TRIANGLE_STROKE, stroke_width);
  }  

  /** 
   * @see abcm.space.HasElevation#getZ(abcm.math.Point) 
   */
  public double getZ(Point P){
    if(!pointIsInside(P)) return Double.NaN;
    return getProjectedElevation(P);
  }
  
  /**
   * @return a String representing the coordinates of the vertices of this topo triangle
   */
  public String ptString(){
    return "[" + P1.coordString() + " *** " + P2.coordString() + " *** " + P3.coordString() + "]"; 
  }
  
  /**
   * @return vertex 1
   */
  public Point getP1(){ return P1; }
  
  /**
   * @return vertex 2
   */
  public Point getP2(){ return P2; }
  
  /**
   * @return vertex 3
   */
  public Point getP3(){ return P3; }


  /**
   * Assume that the projection of the
   * given line segment will slice this topo
   * triangle into separate pieces (as many
   * as three). This method returns a vector
   * that includes the new triangles. The vector
   * will be empty if the line segment
   * does not intersect the topo triangle
   * in a way that would divide it. 
   * @param seg the line segment to be projected
   * @return the resultant triangles
   */
  public Vector<TopoTriangle> slice(LineSegment seg){
    Vector<TopoTriangle> results = new Vector<TopoTriangle>();
    
    Vector  sides = this.getSides();
    LineSegment side1 = (LineSegment)sides.get(0);
    LineSegment side2 = (LineSegment)sides.get(1);
    LineSegment side3 = (LineSegment)sides.get(2);
    
    Point side1Int = Geometry.restrictedPointOfIntersection2D(side1, seg);
//    if(side1Int != null) System.out.println(side1Int.coordString(9));
    Point side2Int = Geometry.restrictedPointOfIntersection2D(side2, seg);
//    if(side2Int != null) System.out.println(side2Int.coordString(9));
    Point side3Int = Geometry.restrictedPointOfIntersection2D(side3, seg);
//    if(side3Int != null) System.out.println(side3Int.coordString(9));
    
    if(side1Int != null)
        if(side1Int.isPlanarEqual(P1) || side1Int.isPlanarEqual(P2)) side1Int = null;
    
    if(side2Int != null)
        if(side2Int.isPlanarEqual(P2) || side2Int.isPlanarEqual(P3)) side2Int = null;
    
    if(side3Int != null)
        if(side3Int.isPlanarEqual(P3) || side3Int.isPlanarEqual(P1)) side3Int = null;

    
    // If this genuinely crosses side 1
    if(side1Int != null){
      // Then, if it genuinely crosses side 2
      if(side2Int != null){
        results.add(new TopoTriangle(side1Int, P2, side2Int));
        results.add(new TopoTriangle(side1Int, side2Int, P3));
        results.add(new TopoTriangle(side1Int, P3, P1));
        return results;
      }
      // Or if it genuinely crosses side 3
      else if(side3Int != null){
        results.add(new TopoTriangle(side1Int, side3Int, P1));
        results.add(new TopoTriangle(side1Int, P2, P3));
        results.add(new TopoTriangle(side1Int, P3, side3Int));
        return results;
      }
      // Or it could go through the vertex between 2 and 3 (P3)
      else{
        results.add(new TopoTriangle(side1Int, P2, P3));
        results.add(new TopoTriangle(side1Int, P3, P1));
        return results;
      }
    }
    else if(side2Int != null){
      if(side3Int != null){
        results.add(new TopoTriangle(side2Int, P3, side3Int));
        results.add(new TopoTriangle(side2Int, side3Int, P1));
        results.add(new TopoTriangle(side2Int, P1, P2));
        return results;
      }
      else{
        results.add(new TopoTriangle(side2Int, P3, P1));
        results.add(new TopoTriangle(side2Int, P1, P2));
        return results;
      }
    }
    else if(side3Int != null){
      // The only possibility remaining is:
      results.add(new TopoTriangle(side3Int, P2, P3));
      results.add(new TopoTriangle(side3Int, P1, P2));
      return results;
    }
    else{ // All possible intersect points are null- empty!
      return results;
    }
  }

  
  /**
   * @return true if all coordinates of all three vertices are planar equal,
   * false otherwise
   */
  public boolean isSinglePoint(){
    return P1.isPlanarEqual(P2, .0001) || P2.isPlanarEqual(P3, .0001);
  }
  
  /**
   * @return the ID of this topo triangle
   */
  public String getID(){ return ID; }
  
  /**
   * Returns true if a 2-Dimensional projection
   * of this triangle along the z-axis touches
   * any part of the other triangle
   * @param other
   * @return true if any part of the other triangle is directly above or below any part of this one
   */
  public boolean isVerticallyAssociated(TopoTriangle other){
    // First check: if any vertex of A is inside B
    if ((other.pointIsInside(this.P1)) ||
        (other.pointIsInside(this.P2)) ||
        (other.pointIsInside(this.P3)) ||
    
    // Second check: if any vertex of B is inside A
        (this.pointIsInside(other.P1)) ||
        (this.pointIsInside(other.P2)) ||
        (this.pointIsInside(other.P3))) return true;
    // Third check: if any side of A intersects any side of B
    LineSegment S1 = new LineSegment(P1, P2);
    LineSegment O1 = new LineSegment(other.P1, other.P2);
    if(S1.crossesIn2D(O1)) return true;
    LineSegment S2 = new LineSegment(P2, P3);
    if(S2.crossesIn2D(O1)) return true;
    LineSegment O2 = new LineSegment(other.P2, other.P3);
    if(S1.crossesIn2D(O2) || S2.crossesIn2D(O2)) return true;
    LineSegment S3 = new LineSegment(P3, P1);
    if(S3.crossesIn2D(O1) || S3.crossesIn2D(O2)) return true;
    LineSegment O3 = new LineSegment(other.P2, other.P1);
    if(S1.crossesIn2D(O3) || S2.crossesIn2D(O3) || S3.crossesIn2D(O3)) return true;
    return false;            
  }
  
  /**
   * @param other
   * @return returns true if the other triangle and this one have three equal vertices (note: checks all
   * possible correspondences, this.vertex 1 = other.vertex 1, this.vertex 1 = other.vertex 2, etc.
   */
  public boolean isCoterminous(TopoTriangle other){
    return ((P1.isPlanarEqual(other.P1) && P2.isPlanarEqual(other.P2) && P3.isPlanarEqual(other.P3)) ||
            (P2.isPlanarEqual(other.P1) && P3.isPlanarEqual(other.P2) && P1.isPlanarEqual(other.P3)) ||
            (P3.isPlanarEqual(other.P1) && P1.isPlanarEqual(other.P2) && P2.isPlanarEqual(other.P3)));
  }
  
  /**
   * @param P vertex sought
   * @return gets the line segment representing the side beginning with the specified vertex.  Null if the point doesn't match any of the vertices.
   */
  public LineSegment getSideBeginning(Point P){
    if     (P.equals(P1)) return new LineSegment(P1, P2);
    else if(P.equals(P2)) return new LineSegment(P2, P3);
    else if(P.equals(P3)) return new LineSegment(P3, P1);
    return null;
  }
  
  /**
   * @param P vertex sought
   * @return gets the line segment representing the side ending with the specified vertex. Null if the point doesn't match any of the vertices.
   */
  public LineSegment getSideEnding(Point P){
    if     (P.equals(P1)) return new LineSegment(P3, P1);
    else if(P.equals(P2)) return new LineSegment(P1, P2);
    else if(P.equals(P3)) return new LineSegment(P2, P3);
    return null;
  }
  
  /**
   * Rotates the vertices of this triangle. Creates a new triangle with vertex 1 = this.vertex 2, 
   * vertex 2 = this.vertex 3, and vertex 3 = this.vertex 1.
   * @return the rotated, new triangle.
   */
  public TopoTriangle rotate(){
    return new TopoTriangle(P2, P3, P1);
  }
  
  /**
   * Gets the volume differential between the two 3-D topo triangles.
   * Assumes that the other triangle is the 'lower' triangle. The result is
   * a volume differential that splits out volume above from volume below,
   * returning both as positive numbers (hence the 'lower' triangle could
   * actually be, all or in part, above this one, and the volume differential
   * will be calculated appropriately).
   * @param lowerTriangle the other triangle, considered to be 'below' this one
   * @return the volume difference between the two triangles
   */
  public TopographicVolumeDifferential getVolumeDifference(TopoTriangle lowerTriangle){
    if(DEBUG_VOL_DIFF)  System.out.println("Getting volume difference:");
    if(!isCoterminous(lowerTriangle)){
      System.err.println("Getting Volume Difference: NON COTERMINOUS TRIANGLE:\n" + this.ptString() + "\n" + lowerTriangle.ptString());
      return new TopographicVolumeDifferential();
    }
    TopoTriangle other = lowerTriangle;
    while(!P1.isPlanarEqual(other.P1) || !P2.isPlanarEqual(other.P2) || !P3.isPlanarEqual(other.P3)) 
      other = other.rotate(); 
    
    if(DEBUG_VOL_DIFF){
      System.out.println(P1.getX() + " " + P1.getY() + "              " + P2.getX() + " " + P2.getY() + "              " + P3.getX() + " " + P3.getY() + "              ");
      System.out.println("P1 Z: " + P1.getZ() + " " + other.P1.getZ());    
      System.out.println("P2 Z: " + P2.getZ() + " " + other.P2.getZ());
      System.out.println("P3 Z: " + P3.getZ() + " " + other.P3.getZ());
    }
    
    // If the triangles are exactly the same (all z coordinates match) there is no difference- return zeroes
    if(((Math.abs(P1.getZ() - other.P1.getZ())) + 
        (Math.abs(P2.getZ() - other.P2.getZ())) +
        (Math.abs(P3.getZ() - other.P3.getZ())))  < Geometry.epsilon){
      if(DEBUG_VOL_DIFF) System.out.println("Triangles are approximately equal- returning zeros");
      return new TopographicVolumeDifferential();
    }

    if(DEBUG_VOL_DIFF) System.out.println("     Z1 above: " + (P1.getZ() - other.P1.getZ()) + "\n     Z2 above: " + (P2.getZ() - other.P2.getZ()) + "\n     Z3 above: " + (P3.getZ() - other.P3.getZ()));
    
    
    // If the triangles are so narrow as to be lines, there is effectively no difference- return zero
    if(P1.isPlanarEqual(P2) || P2.isPlanarEqual(P3) || P3.isPlanarEqual(P1)){
      if(DEBUG_VOL_DIFF) System.out.println("Triangle is basically linear; returning zeros");
      return new TopographicVolumeDifferential();
    }
    
    // Don't allow tiny differences to interfere
    other = new TopoTriangle(new Point(other.getP1().getX(), other.getP1().getY(), (Math.abs(P1.getZ() - other.getP1().getZ()) > Geometry.epsilon ? other.getP1().getZ() : P1.getZ())),
                             new Point(other.getP2().getX(), other.getP2().getY(), (Math.abs(P2.getZ() - other.getP2().getZ()) > Geometry.epsilon ? other.getP2().getZ() : P2.getZ())),
                             new Point(other.getP3().getX(), other.getP3().getY(), (Math.abs(P3.getZ() - other.getP3().getZ()) > Geometry.epsilon ? other.getP3().getZ() : P3.getZ())));
    
    // Two cases: if the z coordinates of corresponding points are all either above or below each other, it's easy:
    if((P1.getZ() >= other.P1.getZ() && P2.getZ() >= other.P2.getZ() && P3.getZ() >= other.P3.getZ()) ||
       (P1.getZ() <= other.P1.getZ() && P2.getZ() <= other.P2.getZ() && P3.getZ() <= other.P3.getZ())){
         
      // Need to decide a floor:
      double floor = (getMinZ() < other.getMinZ() ? getMinZ() : other.getMinZ());
      // Transpose...
      TopoTriangle U = new TopoTriangle(new Point(P1.getX(), P1.getY(), (P1.getZ() - floor)),
                                        new Point(P2.getX(), P2.getY(), (P2.getZ() - floor)),
                                        new Point(P3.getX(), P3.getY(), (P3.getZ() - floor)));
      
      TopoTriangle L = new TopoTriangle(new Point(other.P1.getX(), other.P1.getY(), (other.P1.getZ() - floor)),
                                        new Point(other.P2.getX(), other.P2.getY(), (other.P2.getZ() - floor)),
                                        new Point(other.P3.getX(), other.P3.getY(), (other.P3.getZ() - floor)));
      
      double volumeUnderU = Geometry.volumeUnderTriangle(U.P1, U.P2, U.P3);
      if(DEBUG_VOL_DIFF) System.out.println("UPPER: " + U.ptString() + "\n     volumeUnder: " + volumeUnderU + " floor: " + floor);
      double volumeUnderL = Geometry.volumeUnderTriangle(L.P1, L.P2, L.P3);
      if(DEBUG_VOL_DIFF) System.out.println("LOWER: " + L.ptString() + "\n     volumeUnder: " + volumeUnderL + " floor: " + floor);
      double totalVolume = volumeUnderU - volumeUnderL;
      if(DEBUG_VOL_DIFF) System.out.println("DIFF " + totalVolume);
      if(totalVolume > 0) return new TopographicVolumeDifferential(totalVolume, 0);
      else                return new TopographicVolumeDifferential(0, -1 * totalVolume);      
    }
    else{
      // Find the point that is 'different'- only one point will be above
      // Assume it's point 1;
      Point dUpper = P1;
      Point dLower = other.P1;
      Point c1Upper = P2;
      Point c1Lower = other.P2;
      Point c2Upper = P3;
      Point c2Lower = other.P3;
      // Check- if it's point 2, change:
      if((P1.getZ() >= other.P1.getZ() && P2.getZ() < other.P2.getZ()  && P3.getZ() >= other.P3.getZ()) ||
         (P1.getZ() <= other.P1.getZ() && P2.getZ() > other.P2.getZ() && P3.getZ() <= other.P3.getZ())){
        dUpper = P2;
        dLower = other.P2;
        c1Upper = P3;
        c1Lower = other.P3;
        c2Upper = P1;
        c2Lower = other.P1;     
      }
    
      if((P1.getZ() >= other.P1.getZ() && P2.getZ() >= other.P2.getZ() && P3.getZ() < other.P3.getZ()) ||
         (P1.getZ() <= other.P1.getZ() && P2.getZ() <= other.P2.getZ() && P3.getZ() > other.P3.getZ())){
        dUpper = P3;
        dLower = other.P3;
        c1Upper = P1;
        c1Lower = other.P1;
        c2Upper = P2;
        c2Lower = other.P2;     
      }
      
      // Now we know the points that both lie above (or below) each other, and the one point that is different
      // We need to create three coterminous triangles from these points
      double f1 = (c1Upper.getZ() - c1Lower.getZ())/((dLower.getZ() - c1Lower.getZ()) - (dUpper.getZ() - c1Upper.getZ()));
      Point intersect1 = new Point(c1Upper.getX() + f1 * (dUpper.getX() - c1Upper.getX()),
                                   c1Upper.getY() + f1 * (dUpper.getY() - c1Upper.getY()),
                                   c1Upper.getZ() + f1 * (dUpper.getZ() - c1Upper.getZ()));
      
      double f2 = (c2Upper.getZ() - c2Lower.getZ())/((dLower.getZ() - c2Lower.getZ()) - (dUpper.getZ() - c2Upper.getZ()));
      Point intersect2 = new Point(c2Upper.getX() + f2 * (dUpper.getX() - c2Upper.getX()),
                                   c2Upper.getY() + f2 * (dUpper.getY() - c2Upper.getY()),
                                   c2Upper.getZ() + f2 * (dUpper.getZ() - c2Upper.getZ()));
     
      // These two points plus the other know points define three triangles for each of the upper and lower
      // triangles
      TopographicVolumeDifferential ret = new TopographicVolumeDifferential();
      if(DEBUG_VOL_DIFF) {
        System.out.println("SUBDIVIDING TRIANGLES...");
        System.out.println("DUPPER " + dUpper.coordString());
        System.out.println("DLOWER " + dLower.coordString());
        System.out.println("c1Upper " + c1Upper.coordString());
        System.out.println("c1Lower " + c1Lower.coordString());
        System.out.println("c2Upper " + c2Upper.coordString());
        System.out.println("c2Lower " + c2Lower.coordString());
      }
      ret.add((new TopoTriangle(dUpper, intersect1, intersect2)).getVolumeDifference(new TopoTriangle(dLower, intersect1, intersect2)));
      ret.add((new TopoTriangle(intersect1, intersect2, c1Upper)).getVolumeDifference(new TopoTriangle(intersect1, intersect2, c1Lower)));
      ret.add((new TopoTriangle(c1Upper, c2Upper, intersect2)).getVolumeDifference(new TopoTriangle(c1Lower, c2Lower, intersect2)));
      return ret;
      
    }
    
    
  }
  
  /** 
   * @see abcm.math.Polygon#getDeepClone(abcm.cloning.ABCMCloneStore) 
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    TopoTriangle clone = new TopoTriangle((Point)(cloneStore.getClone(P1)), 
                                          (Point)(cloneStore.getClone(P2)), 
                                          (Point)(cloneStore.getClone(P3)));
    cloneStore.put(this, clone);
    return clone;
  }
  
}













