package abcm.space;

import java.util.Vector;

import abcm.math.Geometry;
import abcm.math.LineSegment;
import abcm.math.Point;

/**
 *
 * A class that takes two triangles and returns a collection of triangles
 * that collectively form the polygon that marks the interior
 * of the intersection of the two triangles.
 * Assume triangle A and triangle B exist such that when they are
 * both projected onto the x,y plane at least one point can be found such
 * that the point is inside both projections. Given this, find the
 * triangle or set of triangles that includes all such points, such that
 * each such point exists in only one triangle and the z coordinates of
 * the returned triangles are those that would be found if these triangles
 * were projected onto the second triangle.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class TopoTriangleProjector {

  /** Debug flag */
  static boolean debug = false;

  /**
   * Constructor; private scope prevents the class from being instantiated or extended.
   */
  private TopoTriangleProjector(){} 
  
  /**
   * Gets the set of triangles that reflect the interior of the intersection of
   * the two triangles, such that the z coordinates are derived from the second
   * triangle.
   * @param triangleBeingProjected the triangle being projected
   * @param baseProjectedOnto the triangle onto which the first triangle is projected
   * @return the set of triangles that represents the projection of the first triangle onto the second
   */
  public static Vector project(TopoTriangle triangleBeingProjected, TopoTriangle baseProjectedOnto){
    if(debug){
      System.out.println("PROJECTING...");
      System.out.println("  T: " + triangleBeingProjected.ID + "  "  + triangleBeingProjected.ptString());
      System.out.println("  A: " + baseProjectedOnto.ID + "  " + baseProjectedOnto.ptString());
    }
  
    
    boolean vertexT1IsInsideA;
    boolean vertexT2IsInsideA;
    boolean vertexT3IsInsideA;
    int countOfTVerticesInsideA;
    boolean TIsProperlyRotated;
    
    boolean vertexA1IsInsideT;
    boolean vertexA2IsInsideT;
    boolean vertexA3IsInsideT;
    int countOfAVerticesInsideT;
    boolean AIsProperlyRotated;
    
    Vector newTriangles = new Vector(); 
    
    
    
    TopoTriangle A = baseProjectedOnto;
    TopoTriangle T = triangleBeingProjected;
    
    int rotations = 0;
    do{
      vertexT1IsInsideA = A.pointIsInside(T.getP1());
      vertexT2IsInsideA = A.pointIsInside(T.getP2());
      vertexT3IsInsideA = A.pointIsInside(T.getP3());
      
      countOfTVerticesInsideA = (vertexT1IsInsideA ? 1 : 0) + (vertexT2IsInsideA ? 1 : 0) + (vertexT3IsInsideA ? 1 : 0);
    
      TIsProperlyRotated = (countOfTVerticesInsideA == 0 ||                           
                            countOfTVerticesInsideA == 1 && vertexT1IsInsideA ||     
                            countOfTVerticesInsideA == 2 && vertexT1IsInsideA && vertexT2IsInsideA ||
                            countOfTVerticesInsideA == 3);
      if(TIsProperlyRotated ==  false) T = T.rotate();
      rotations++;
      if(rotations >= 8)
        System.out.println("Breaking Rotations of T");
    }while(TIsProperlyRotated == false);
    
    /* CASE A */
    // Simplest case possible: all of the vertices of T are in A;
    // all that need be returned is a triangle with T's points in
    // the X/Y plane projected onto A's elevations
    if(countOfTVerticesInsideA == 3){
      if(debug) System.out.println("ALL OF T VERTICES ARE INSIDE A; returning T projected onto A");
      TopoTriangle newT = new TopoTriangle(new Point(T.getP1().getX(), T.getP1().getY(), A.getZ(T.getP1())), 
                                        new Point(T.getP2().getX(), T.getP2().getY(), A.getZ(T.getP2())), 
                                        new Point(T.getP3().getX(), T.getP3().getY(), A.getZ(T.getP3())));
      newTriangles.add(newT);
      if(debug) System.out.println("T = " + newT.ptString());
      return newTriangles;
    }
  
    rotations = 0;
    do{
      vertexA1IsInsideT = T.pointIsInside(A.getP1());
      vertexA2IsInsideT = T.pointIsInside(A.getP2());
      vertexA3IsInsideT = T.pointIsInside(A.getP3());
    
      countOfAVerticesInsideT = (vertexA1IsInsideT ? 1 : 0) + (vertexA2IsInsideT ? 1 : 0) + (vertexA3IsInsideT ? 1 : 0);
    
      AIsProperlyRotated = (countOfAVerticesInsideT == 0 ||                           
                            countOfAVerticesInsideT == 1 && vertexA1IsInsideT ||     
                            countOfAVerticesInsideT == 2 && vertexA1IsInsideT && vertexA2IsInsideT ||
                            countOfAVerticesInsideT == 3);
      if(AIsProperlyRotated == false) A = A.rotate();
      rotations++;
      if(rotations >= 8)
        System.out.println("Breaking rotations of A");
    }while(AIsProperlyRotated == false);
      /* CASE ?? */
    // Second simple case: A is entirely within T
    if(countOfAVerticesInsideT == 3){
      TopoTriangle newT = new TopoTriangle(new Point(A.getP1().getX(), A.getP1().getY(), A.getP1().getZ()), 
                                        new Point(A.getP2().getX(), A.getP2().getY(), A.getP2().getZ()),
                                        new Point(A.getP3().getX(), A.getP3().getY(), A.getP3().getZ()));
      newTriangles.add(newT);
      if(debug) System.out.println("ALL OF A VERTICES ARE INSIDE T; returning a new triangle of A");
      if(debug) System.out.println("T = " + newT.ptString());
      return newTriangles;
    }
    
    // Map all the points
    Point ASide1IntersectionWithTSide1 = Geometry.restrictedPointOfIntersection2D((LineSegment)A.getSides().get(0), (LineSegment)T.getSides().get(0));
    Point ASide1IntersectionWithTSide2 = Geometry.restrictedPointOfIntersection2D((LineSegment)A.getSides().get(0), (LineSegment)T.getSides().get(1));
    Point ASide1IntersectionWithTSide3 = Geometry.restrictedPointOfIntersection2D((LineSegment)A.getSides().get(0), (LineSegment)T.getSides().get(2));
    
    Point ASide2IntersectionWithTSide1 = Geometry.restrictedPointOfIntersection2D((LineSegment)A.getSides().get(1), (LineSegment)T.getSides().get(0));
    Point ASide2IntersectionWithTSide2 = Geometry.restrictedPointOfIntersection2D((LineSegment)A.getSides().get(1), (LineSegment)T.getSides().get(1));
    Point ASide2IntersectionWithTSide3 = Geometry.restrictedPointOfIntersection2D((LineSegment)A.getSides().get(1), (LineSegment)T.getSides().get(2));
    
    Point ASide3IntersectionWithTSide1 = Geometry.restrictedPointOfIntersection2D((LineSegment)A.getSides().get(2), (LineSegment)T.getSides().get(0));
    Point ASide3IntersectionWithTSide2 = Geometry.restrictedPointOfIntersection2D((LineSegment)A.getSides().get(2), (LineSegment)T.getSides().get(1));
    Point ASide3IntersectionWithTSide3 = Geometry.restrictedPointOfIntersection2D((LineSegment)A.getSides().get(2), (LineSegment)T.getSides().get(2));
    
    int countOfTimesASide1IsCrossed = (ASide1IntersectionWithTSide1 != null ? 1 : 0) + 
                                      (ASide1IntersectionWithTSide2 != null ? 1 : 0) +
                                      (ASide1IntersectionWithTSide3 != null ? 1 : 0);
    
    int countOfTimesASide2IsCrossed = (ASide2IntersectionWithTSide1 != null ? 1 : 0) + 
                                      (ASide2IntersectionWithTSide2 != null ? 1 : 0) +
                                      (ASide2IntersectionWithTSide3 != null ? 1 : 0);
    
    int countOfTimesASide3IsCrossed = (ASide3IntersectionWithTSide1 != null ? 1 : 0) + 
                                      (ASide3IntersectionWithTSide2 != null ? 1 : 0) +
                                      (ASide3IntersectionWithTSide3 != null ? 1 : 0);
    
    int numberOfASidesCrossed = (countOfTimesASide1IsCrossed > 0 ? 1 : 0) + 
                                (countOfTimesASide2IsCrossed > 0 ? 1 : 0) + 
                                (countOfTimesASide3IsCrossed > 0 ? 1 : 0); 

    /* CASE ?? */
    // Third simple case: Triangles do not overlap at all
    if(numberOfASidesCrossed    == 0 && 
       countOfAVerticesInsideT  == 0 && 
       countOfTVerticesInsideA  == 0) {
      if(debug) System.out.println("The Two Triangles do not cross; no projection");
      return newTriangles; // If none of these is > 0, triangles do not overlap at all
    }
    
//    int countOfTimesTSide1IsCrossed = (ASide1IntersectionWithTSide1 != null ? 1 : 0) + 
//                                      (ASide2IntersectionWithTSide1 != null ? 1 : 0) +
//                                      (ASide3IntersectionWithTSide1 != null ? 1 : 0);
//    
//    int countOfTimesTSide2IsCrossed = (ASide1IntersectionWithTSide2 != null ? 1 : 0) + 
//                                      (ASide2IntersectionWithTSide2 != null ? 1 : 0) +
//                                      (ASide3IntersectionWithTSide2 != null ? 1 : 0);
//    
//    int countOfTimesTSide3IsCrossed = (ASide1IntersectionWithTSide3 != null ? 1 : 0) + 
//                                      (ASide2IntersectionWithTSide3 != null ? 1 : 0) +
//                                      (ASide3IntersectionWithTSide3 != null ? 1 : 0);

    Point projectedT1 = new Point(T.getP1().getX(), T.getP1().getY(), Geometry.getProjectedElevation(A.getP1(), A.getP2(), A.getP3(), T.getP1()));
    Point projectedT2 = new Point(T.getP2().getX(), T.getP2().getY(), Geometry.getProjectedElevation(A.getP1(), A.getP2(), A.getP3(), T.getP2()));
    Point projectedT3 = new Point(T.getP3().getX(), T.getP3().getY(), Geometry.getProjectedElevation(A.getP1(), A.getP2(), A.getP3(), T.getP3()));
    
    Vector ASide1Points = getPointsInOrder(A.getP1(), ASide1IntersectionWithTSide1, ASide1IntersectionWithTSide2, ASide1IntersectionWithTSide3, A.getP2());
    Vector ASide2Points = getPointsInOrder(A.getP2(), ASide2IntersectionWithTSide1, ASide2IntersectionWithTSide2, ASide2IntersectionWithTSide3, A.getP3());
    Vector ASide3Points = getPointsInOrder(A.getP3(), ASide3IntersectionWithTSide1, ASide3IntersectionWithTSide2, ASide3IntersectionWithTSide3, A.getP1());
    Vector TSide1Points = getPointsInOrder(projectedT1, ASide1IntersectionWithTSide1, ASide2IntersectionWithTSide1, ASide3IntersectionWithTSide1, projectedT2);
    Vector TSide2Points = getPointsInOrder(projectedT2, ASide1IntersectionWithTSide2, ASide2IntersectionWithTSide2, ASide3IntersectionWithTSide2, projectedT3);
    Vector TSide3Points = getPointsInOrder(projectedT3, ASide1IntersectionWithTSide3, ASide2IntersectionWithTSide3, ASide3IntersectionWithTSide3, projectedT1);
    
    // With the two triangles properly rotated, it's a snap:
    Vector points = new Vector();
    if(countOfAVerticesInsideT == 0 && countOfTVerticesInsideA == 0){
      if(debug){
        System.out.println("Case A0 T0");
        System.out.println("A: " + A.ptString());
        System.out.println("T: " + T.ptString());
      }
      // In this case, the polygon will be entirely internal to
      // A. One need merely get the intersections of the T sides with
      // the A sides in clockwise order.
      for(int i = 1; i < TSide1Points.size() - 1; i++) points.add(TSide1Points.get(i));
      for(int i = 1; i < TSide2Points.size() - 1; i++) points.add(TSide2Points.get(i));
      for(int i = 1; i < TSide3Points.size() - 1; i++) points.add(TSide3Points.get(i));
    }
    
    if(countOfAVerticesInsideT == 0 && countOfTVerticesInsideA == 1){
      if(debug){
        System.out.println("Case A0 T1");
        System.out.println("A: " + A.ptString());
        System.out.println("T: " + T.ptString());
      }
      // In this case, the polygon will start with T1 (the only vertex inside
      // A) and go through all other T intersects:
      points.add(TSide1Points.get(0));
      for(int i = 1; i < TSide1Points.size() - 1; i++) points.add(TSide1Points.get(i));
      for(int i = 1; i < TSide2Points.size() - 1; i++) points.add(TSide2Points.get(i));
      for(int i = 1; i < TSide3Points.size() - 1; i++) points.add(TSide3Points.get(i));
    }
    
    if(countOfAVerticesInsideT == 0 && countOfTVerticesInsideA == 2){
      if(debug){
        System.out.println("Case A0 T2");
        System.out.println("A: " + A.ptString());
        System.out.println("T: " + T.ptString());
      }
      // In this case the polygon will start with T1 and T2 (which are inside
      // A) and go through the remaining T intersects. T Side 1 will not intersect
      // meaningfully and can be ignored
      points.add(TSide1Points.get(0));
      points.add(TSide2Points.get(0));
      for(int i = 1; i < TSide2Points.size() - 1; i++) points.add(TSide2Points.get(i));
      for(int i = 1; i < TSide3Points.size() - 1; i++) points.add(TSide3Points.get(i));
    }
    
    if(countOfAVerticesInsideT == 1 && countOfTVerticesInsideA == 0){
      if(debug){
        System.out.println("Case A1 T0");
        System.out.println("A: " + A.ptString());
        System.out.println("T: " + T.ptString());
      }
      // In this case the first vertex will be A's vertex1.
      points.add(A.getP1());
      // After that it will be the T-intersections on A's sides:
      for(int i = 1; i < ASide1Points.size() - 1; i++) points.add(ASide1Points.get(i));
      for(int i = 1; i < ASide2Points.size() - 1; i++) points.add(ASide2Points.get(i));
      for(int i = 1; i < ASide3Points.size() - 1; i++) points.add(ASide3Points.get(i));
    }
    
    if(countOfAVerticesInsideT == 1 && countOfTVerticesInsideA == 1){
      if(debug){
        System.out.println("Case A1 T1 HERE");
        System.out.println("A: " + A.ptString());
        System.out.println("T: " + T.ptString());
      }
      // If this is the case, there can only be four points:
      points.add(A.getP1());
      points.add(ASide1Points.get(1));
      for(int i = 0; i < TSide1Points.size() - 1; i++) points.add(TSide1Points.get(i));
      for(int i = 1; i < TSide2Points.size() - 1; i++) points.add(TSide2Points.get(i));
      for(int i = 1; i < TSide3Points.size() - 1; i++) points.add(TSide3Points.get(i));
    }
    
    if(countOfAVerticesInsideT == 1 && countOfTVerticesInsideA == 2){
      if(debug){
        System.out.println("Case A1 T2");
        System.out.println("A: " + A.ptString());
        System.out.println("T: " + T.ptString());
      }
      // If this is the case, there can only be five points:
      points.add(A.getP1());
      points.add(ASide1Points.get(1));
      points.add(TSide1Points.get(0));
      points.add(TSide2Points.get(0));
      points.add(TSide2Points.get(1));
    }
    
    if(countOfAVerticesInsideT == 2 && countOfTVerticesInsideA == 0){
      if(debug){
        System.out.println("Case A2 T0");
        System.out.println("A: " + A.ptString());
        System.out.println("T: " + T.ptString());
      }
      
      points.add(A.getP1());
      points.add(A.getP2());
      for(int i = 1; i < ASide2Points.size() - 1; i++) points.add(ASide2Points.get(i));
      for(int i = 1; i < ASide3Points.size() - 1; i++) points.add(ASide3Points.get(i));
    }
    
    if(countOfAVerticesInsideT == 2 && countOfTVerticesInsideA == 1){
      if(debug){
        System.out.println("Case A2 T1 HERE");
        System.out.println("A: " + A.ptString());
        System.out.println("T: " + T.ptString());
      }
      
      points.add(A.getP1());
      points.add(A.getP2());
      points.add(ASide2Points.get(1));
      for(int i = 1; i < TSide1Points.size() - 1; i++) points.add(TSide1Points.get(i));
      for(int i = 1; i < TSide2Points.size() - 1; i++) points.add(TSide2Points.get(i));
      for(int i = 1; i < TSide3Points.size() - 1; i++) points.add(TSide3Points.get(i));
    }
    
    if(countOfAVerticesInsideT == 2 && countOfTVerticesInsideA == 2){
      if(debug){
        System.out.println("Case A2 T2");
        System.out.println("A: " + A.ptString());
        System.out.println("T: " + T.ptString());
      }
      // If this is the case, there can only be six points:
      points.add(A.getP1());
      points.add(A.getP2());
      points.add(ASide2Points.get(1));
      points.add(TSide1Points.get(0));
      points.add(TSide2Points.get(0));
      points.add(TSide2Points.get(1));
    }
    
    if(debug){
      System.out.println("Points Size: " + points.size());
      for(int i = 0; i < points.size(); i++){
        System.out.println("POINTS: " + i + " " + ((Point)points.get(i)).coordStringShort());
      }
    } 
    for(int i = points.size() - 1; i > 0; i--){
      Point P1 = (Point)points.get(i);
      for(int j = (i - 1); j >=0; j--){
        if(P1.isPlanarEqual((Point)points.get(j))){
          points.remove(i);
          break;
        }        
      }
    }
    if(debug)  System.out.println("Points Size With Planar Removed: " + points.size());
    Point baseVertex = (Point)points.get(0);
    for(int i = 1; i < points.size() - 1; i++){
      if(debug) System.out.println("Creating: " + baseVertex.coordString() + " " + ((Point)points.get(i)).coordString() + " " + ((Point)points.get(i+1)).coordString());
      newTriangles.add(new TopoTriangle(baseVertex, (Point)points.get(i), (Point)points.get(i+1)));
    }
    
    return newTriangles; 
  }
  
  /**
   * Given a set of points possibly representing one end of a line segment, a collection of points along that line
   * segment, and the end point of the line segment, removes co-equal points and ensures that the points are in
   * order from nearest to furthest.
   * @param vertex1 near vertex
   * @param intersectA point 1
   * @param intersectB point 2
   * @param intersectC point 3
   * @param vertex2 far vertex
   * @return the list of points in order from near to far, with co-equal points removed.
   */
  private static Vector getPointsInOrder(Point vertex1, Point intersectA, Point intersectB, Point intersectC, Point vertex2){
    Vector pointSort = new Vector();
    if(intersectA != null) pointSort.add(intersectA);
    if(intersectB != null) pointSort.add(intersectB);
    if(intersectC != null) pointSort.add(intersectC);
    Vector points = new Vector();
    points.add(vertex1);
    while(pointSort.size() > 0){
      Point toRemove = (Point)pointSort.get(0);
      Point compare = toRemove;
      double min = vertex1.planarDistanceToPoint(toRemove);      
      for(int i = 1; i < pointSort.size(); i++){
        compare = (Point)pointSort.get(i); 
        double minCompare = vertex1.planarDistanceToPoint(compare);
        if(minCompare < min){
          toRemove = compare;
          min      = minCompare;
        }
      }
      pointSort.remove(toRemove);
      points.add(toRemove);
    }
    points.add(vertex2);
    
    return points;   
  }
  
}
