package abcm.space;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JPanel;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableVector;
import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.graphics.X3DAssist;
import abcm.graphics.X3DIndexedFaceSetCreator;
import abcm.graphics.X3DRepresentable;
import abcm.math.Geometry;
import abcm.math.LineSegment;
import abcm.math.Point;
import abcm.math.Polygon;
import abcm.math.Triangle;
import abcm.utilities.XMLAssist;

/**
 * Represents a Triangulation that is a surface or landscape.
 * 
 * Note that this class includes a number of static final debugging flags;
 * these are not described in the JavaDoc; each turns on or off debugging
 * for specific methods or routines- see the code for details.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class TriangulatedTopography implements SVGRepresentable, X3DRepresentable, ABCMDeepCloneable, HasElevation {
  
  /* Debugging flags */
  private static final boolean DEBUG_ADD     = false;
  private static final boolean DEBUG_X3D     = false;
  private static final boolean DEBUG_PUNCH   = false;
  private static final boolean DEBUG_OVERLAY = false;
  private static final boolean DEBUG_REMOVE  = false;
  private static final boolean DEBUG_VOLUME  = false;
  private static final boolean DEBUG_PROJECT = false;
  
  /* Inner class */
  /**
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  class TTDraw extends JFrame{
      
      /** Serial ID */
      private static final long serialVersionUID = -2792569775519026813L;
      /** Default width of window on screen, in pixels */
      protected static final int defaultScreenWidthInPixels = 700;
      /** Default height of window on screen, in pixels */
      protected static final int defaultScreenHeightInPixels = 300;
      
      /** Minimum x coordinate to be drawn */
      public float minX;
      /** Maximum x coordinate to be drawn */
      public float maxX;
      /** Minimum y coordinate to be drawn */
      public float maxY;
      /** Maximum y coordinate to be drawn */
      public float minY;
      
      /** 
       * Collection of triangles to be drawn (if none other are specified as argument to 
       * repaint; this may not be possible in latest rev.) 
       */
      public Vector triangles;
      /** Collection of polygons to be drawn overlain on the triangles */
      public Vector polygons;
      /** 
       * Collection of triangles to be emphasized- actually, triangles found in the
       * collection being drawn but NOT in this collection will be highlighted. The
       * use is to take a vector drawn at one time and use it as 'compare' with the
       * vector drawn later to see changes.
       */
      public Vector compare;
      
      /** A single triangle that can be specifically highlighted   */
      public TopoTriangle active;
      
      /** A single LineSegment that can be specifically highlighted */
      public LineSegment C1;
      /** A single LineSegment that can be specifically highlighted */
      public LineSegment C2;
      /** A single Point that can be specifically highlighted */
      public Point activePoint;
      
      /**
       * Constructor
       * @param xMin minimum X value to be shown
       * @param xMax maximum X value to be shown
       * @param yMin minimum Y value to be shown
       * @param yMax maximum Y value to be shown
       * @param name Name to be displayed on this window
       * @param trianglesToDraw collection of triangles to draw
       * @param polygonsToDraw collection of polygons to draw
       * @param compareWith collection of triangles to be compared with the drawn collection
       */
      public TTDraw(float xMin, float xMax, float yMin, float yMax, String name, 
          Vector trianglesToDraw, Vector polygonsToDraw, Vector compareWith){
        minX = xMin;
        maxX = xMax;
        minY = yMin;
        maxY = yMax;
        setTitle(name);
        triangles = trianglesToDraw;
        polygons = polygonsToDraw;
        compare = compareWith;
        setBounds(100, -900, defaultScreenWidthInPixels, defaultScreenHeightInPixels);
      }

      /**
       * Inner class that allows the panel to drawn the image required
       *  @author John T. Murphy<br/>
       *  Global Institute of Sustainability,<br/>
       *  Arizona State University<br/>
       *  For license and disclaimers, see the Overview.
       */
      private class ViewPane extends JPanel{
    
        /** Required ID */
        private static final long serialVersionUID = 123037986L;
        
        /**
         * Paints this, with the following specifics (note: colors and thicknesses may be adjusted
         * in code for effect and may not match what is given here):
         * 
         * The basic layer of triangles (triangles) are drawn in black, line thickness = 1</br>
         * 
         * If there is a set of triangles in 'compare', then if the triangle in 'triangles' is NOT
         * found in the 'compare' set, the triangle is drawn in green, line thickness = 3</br>
         * 
         * If there is an 'active' triangle (active), it is drawn in blue, line thickness = 2</br>
         * 
         * If there is a set of polygons (polygons) to be drawn, it is drawn (in its entirety) in yellow, line thickness = 1</br>
         * 
         * If 'C1' (a line segment) is not null, it is drawn in red, line thickness = 1</br>
         * 
         * If 'C2' (another line segment) is not null, it is drawn in yellow, line thickness = 1</br>
         * 
         * If there is an active Point (activePoint), it is marked with a green X, line thickness = 1</br>
         * 
         *  @param G graphics window to be used when repainting this element
         */
        public void paintComponent(Graphics G){
          Color trianglesColor = new Color(Color.BLACK.getRed()/255f, Color.BLACK.getGreen()/255f, Color.BLACK.getBlue()/255f, .1f);
          
          
          Graphics2D g2D = (Graphics2D)G;
          float cw = getSize().width/(maxX - minX);
          float ch = getSize().height/(maxY - minY);
          
          for(int i = 0; i < triangles.size(); i++){
            TopoTriangle t = (TopoTriangle)triangles.get(i);
            Point P1 = (Point)(t.getPoints().get(0));
            Point P2 = (Point)(t.getPoints().get(1));
            Point P3 = (Point)(t.getPoints().get(2));
           
            
            g2D.setColor(trianglesColor);
            g2D.setStroke(new BasicStroke(1));
            
            if(compare != null && !compare.contains(t)){
              g2D.setColor(Color.green);
              g2D.setStroke(new BasicStroke(3));
            }
            if(active != null && active.equals(t)){
              System.out.println("DRAWING ACTIVE!!! " + active.ID);
              System.out.println(((LineSegment)active.getSides().get(0)).display());
              System.out.println(((LineSegment)active.getSides().get(1)).display());
              System.out.println(((LineSegment)active.getSides().get(2)).display());
              g2D.setColor(Color.blue);
              g2D.setStroke(new BasicStroke(3));
            }
            g2D.drawLine((int)((P1.getX() - minX) * cw), (int)((P1.getY() - minY) * ch), (int)((P2.getX() - minX) * cw), (int)((P2.getY() - minY) * ch));
            g2D.drawLine((int)((P2.getX() - minX) * cw), (int)((P2.getY() - minY) * ch), (int)((P3.getX() - minX) * cw), (int)((P3.getY() - minY) * ch));
            g2D.drawLine((int)((P3.getX() - minX) * cw), (int)((P3.getY() - minY) * ch), (int)((P1.getX() - minX) * cw), (int)((P1.getY() - minY) * ch));
          }
          if(polygons != null){
            for(int i = 0; i < polygons.size(); i++){
              Polygon poly = (Polygon)polygons.get(i);
              Vector points = poly.getPoints();
              for(int p = 0; p < points.size() - 1; p++){                
                Point P1 = (Point)points.get(p);
                Point P2 = (Point)points.get(p + 1);
                g2D.setColor(Color.yellow);
                g2D.setStroke(new BasicStroke(1));
                g2D.drawLine((int)((P1.getX() - minX) * cw), (int)((P1.getY() - minY) * ch), (int)((P2.getX() - minX) * cw), (int)((P2.getY() - minY) * ch));  
              }
              Point P1 = (Point)points.get(points.size() - 1);
              Point P2 = (Point)points.get(0);
              g2D.setColor(Color.yellow);
              g2D.setStroke(new BasicStroke(1));
              g2D.drawLine((int)((P1.getX() - minX) * cw), (int)((P1.getY() - minY) * ch), (int)((P2.getX() - minX) * cw), (int)((P2.getY() - minY) * ch));
            }
          }
          if(C1 != null){
            g2D.setColor(Color.red);
            g2D.setStroke(new BasicStroke(1));
            g2D.drawLine((int)((C1.P1.getX() - minX) * cw), (int)((C1.P1.getY() - minY) * ch), (int)((C1.P2.getX() - minX) * cw), (int)((C1.P2.getY() - minY) * ch));
          }
          
          if(C2 != null){
            g2D.setColor(Color.green);
            g2D.setStroke(new BasicStroke(1));
            g2D.drawLine((int)((C2.P1.getX() - minX) * cw), (int)((C2.P1.getY() - minY) * ch), (int)((C2.P2.getX() - minX) * cw), (int)((C2.P2.getY() - minY) * ch));
          }
          if(activePoint != null){
            g2D.setColor(Color.green);
            g2D.setStroke(new BasicStroke(1));
            g2D.drawLine((int)(((activePoint.getX() - 10) - minX) * cw), (int)(((activePoint.getY() - 10) - minY) * ch), (int)(((activePoint.getX() + 10) - minX) * cw), (int)(((activePoint.getY() + 10) - minY) * ch));
            g2D.drawLine((int)(((activePoint.getX() + 10) - minX) * cw), (int)(((activePoint.getY() - 10) - minY) * ch), (int)(((activePoint.getX() - 10) - minX) * cw), (int)(((activePoint.getY() + 10) - minY) * ch));
          }          
        }     
      }
      
      /**
       * Makes this view pane visible
       * @param xPos x position
       * @param yPos y position
       * @param width x dimension (pixels)
       * @param height y dimension (pixels)
       */
      public void makeVisible(int xPos, int yPos, int width, int height){
        setBounds(xPos, yPos, width, height);
        setSize(height, width);
//        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        getContentPane().add(new ViewPane());
        setVisible(true);
      }
      
      /**
       * Repaints this pane
       * @param topoTrianglesToDraw collection of triangles to be drawn (overwrites those
       * in the instance variable/constructor)
       */
      public void repaint(Vector topoTrianglesToDraw){
        triangles = (Vector)topoTrianglesToDraw.clone();
        super.repaint();
      }
    }

  
  /* Static Constants */
  
  /** String giving x3d definition of terrain object  */
  private static String x3dDefinition = 
    "<Appearance DEF='Terrain'>\n" +
    "  <Material diffuseColor='0.9 0.5 0.1' emissiveColor='0 0 0'/>\n" +
    "</Appearance>";
  

  /* Instance Variables */
  
  /** Vector of topographic triangles  */
  protected ABCMDeepCloneableVector topoTriangles = new ABCMDeepCloneableVector();
  
  /* Constructor */
  
  /**
   * Constructor that includes all points in the points list.
   * @param points List of points to be included in this topography. 
   */
  public TriangulatedTopography(Collection points){
    if(points.size() == 0) return;
    //Collection dT = Geometry.getDelaunayTriangulation(points);
    Collection dT = Geometry.getVisadDelaunayTriangulation(new Vector(points));
    Iterator iterator = dT.iterator();
    while(iterator.hasNext()){
      topoTriangles.add(new TopoTriangle((Triangle)iterator.next()));
    }
  }

  /* External Modification of Triangle List */
  
  public void addTopoTriangles(Collection triangles){
    if(topoTriangles == null) topoTriangles = new ABCMDeepCloneableVector();
    Iterator iterator = triangles.iterator();
    while(iterator.hasNext()) topoTriangles.add((ABCMDeepCloneable)iterator.next());
  }
  
  /**
   * Gets the list of topo triangles that make up this object
   * @return the list of triangles in this topography
   */
  public Vector getTriangles(){
    Vector V = new Vector();
    V.addAll(topoTriangles);
    return V;
  }
  

  /* Subset Manipulation */
  
  /**
   * Gets a subset of this topography bounded by a quadrilateral defined by
   * the founr points specified.
   * @param p1 corner point 1
   * @param p2 corner point 2
   * @param p3 corner point 3
   * @param p4 corner point 4
   * @return the extracted subset
   */
  public Vector getSubset(Point p1, Point p2, Point p3, Point p4){
    Vector V = topoTriangles;
    Vector Vsub = new Vector();
    LineSegment l1 = new LineSegment(p1, p2);
    LineSegment l2 = new LineSegment(p2, p3);
    LineSegment l3 = new LineSegment(p3, p4);
    LineSegment l4 = new LineSegment(p4, p1);
    Triangle B1 = new Triangle(p1, p2, p4);
    Triangle B2 = new Triangle(p2, p3, p4);
    for(int i = 0; i < V.size(); i++){
      Triangle T = (Triangle)V.get(i);
      // Return true if the triangle falls entirely within one
      // of the two triangles B1 and B2, or
      // if any of the triangles's sides intersect any of the line
      // segments, or
      // if either of the two large triangles falls within this
      // triangle
      boolean add = B1.pointIsInside((Point)T.getPoints().get(0)) ||
                    B2.pointIsInside((Point)T.getPoints().get(0)) ||
                    T.anySideIsIntersected(l1) ||
                    T.anySideIsIntersected(l2) ||
                    T.anySideIsIntersected(l3) ||
                    T.anySideIsIntersected(l4) ||
                    T.pointIsInside(p3);

      if(add) Vsub.add(T);
    }
    return Vsub;
  }
  
  /**
   * Gets a list of triangles in increasing X-coordinate order,
   * useful for shortening searches
   * @return the list of triangles from low to high min X coordinate
   */
  private Vector getListOfTrianglesInIncreasingXCoordinateOrder(){
    Vector V = new Vector();
    for(int i = 0; i < topoTriangles.size(); i++){
      TopoTriangle toPlace = (TopoTriangle)topoTriangles.get(i);
      for(int j = 0; j < V.size(); j++){
        if(((TopoTriangle)V.get(j)).getMinX() > toPlace.getMinX()){
          V.add(j, toPlace);
          break;
        }
      }
      if(!V.contains(toPlace)) V.add(toPlace);
    }
    return V;
  }

  
  /* Actions */
  
  /**
   * Adds a new point. Point is added within an existing triangle,
   * which is then discarded
   * @param P
   * @return the new point added
   */  
  public Point addPointAtProjectedElevation(Point P){
    if(DEBUG_ADD) System.out.println("Adding " + P.coordString() + " to " + topoTriangles.size());
    Point firstOnSideVertex = null;
    Point newCenter = null;
    boolean firstTriangle = true;
    for(int i = (topoTriangles.size() -1); i >=0; i--){ // Backwards because new triangles are added at the end
      TopoTriangle T = (TopoTriangle)topoTriangles.get(i); //
      if(DEBUG_ADD) System.out.println("Checking " + idTriangle(T));
      double newZ = T.getZ(P);
      if(!Double.isNaN(newZ)){
        newCenter = new Point(P.getX(), P.getY(), newZ);
        Point P1 = T.getP1();
        Point P2 = T.getP2();
        Point P3 = T.getP3();
        // If this point matches an existing vertex, return it
        // Do this before any other checks
        if(P1.isPlanarEqual(P, Geometry.epsilon)){
          if(DEBUG_ADD) System.out.println("Point is planar equal to " + P1.coordString());
          return P1;
        }
        if(P2.isPlanarEqual(P, Geometry.epsilon)){
          if(DEBUG_ADD) System.out.println("Point is planar equal to " + P2.coordString());
          return P2;
        }
        if(P3.isPlanarEqual(P, Geometry.epsilon)){
          if(DEBUG_ADD) System.out.println("Point is planar equal to " + P3.coordString());
          return P3;
        }
        
        if(T.pointIsInsideExclusive(P)){
          if(DEBUG_ADD) System.out.println("Removing " + idTriangle(T) + " because point " + P.coordString() + " is inside...");
          topoTriangles.remove(T); // Take this triangle out of the collection
          topoTriangles.add(i, new TopoTriangle(newCenter, P1, P2));
          topoTriangles.add(i, new TopoTriangle(newCenter, P2, P3));
          topoTriangles.add(i, new TopoTriangle(newCenter, P3, P1));
          if(i == 1414){
            if(DEBUG_ADD)
                System.out.println("Adding:\n  " + idTriangle((TopoTriangle)topoTriangles.get(topoTriangles.size() - 3)) +
                                   "       \n  " + idTriangle((TopoTriangle)topoTriangles.get(topoTriangles.size() - 2)) +
                                   "       \n  " + idTriangle((TopoTriangle)topoTriangles.get(topoTriangles.size() - 1)));
          }
              
              
          return newCenter;
        }
        else{
          Point oppositeVertex = null;
          Point V1 = null;
          Point V2 = null;
          if(Geometry.pointLiesOnSegment(P1, P2, P)){  oppositeVertex = P3;   V1 = P1;  V2 = P2;  }
          if(Geometry.pointLiesOnSegment(P2, P3, P)){  oppositeVertex = P1;   V1 = P2;  V2 = P3;  }
          if(Geometry.pointLiesOnSegment(P3, P1, P)){  oppositeVertex = P2;   V1 = P3;  V2 = P1;  }
          if(oppositeVertex == null) System.err.println("ERROR: Problem in addPointAtProjectedElevation...");
          else{
            // The point lies on a side of this triangle. A second triangle SHOULD be
            // created (unless the point was on the very edge of the topography)
            // corresponding to the adjacent triangle
            if(DEBUG_ADD) System.out.println("Removing " + idTriangle(T) + " because point is on side");
            topoTriangles.remove(T);
            if(firstOnSideVertex != null) newCenter         = firstOnSideVertex;
            else                          firstOnSideVertex = newCenter;
          
            TopoTriangle T1 = new TopoTriangle(newCenter, V1, oppositeVertex);
            TopoTriangle T2 = new TopoTriangle(newCenter, V2, oppositeVertex);
            topoTriangles.add(i, T1);
            topoTriangles.add(i, T2);
          }
          if(firstTriangle) firstTriangle = false;
          else              return newCenter; // if this is the second triangle, return- have to be done...
        }
      }      
    }
    if(newCenter == null) System.err.println("DID NOT PLACE POINT " + P.coordString());
    return newCenter;
  }
  
  /**
   * 'Cuts' a hole in this TriangulatedTopography. The result is
   * the original TriangulatedTopography, but with all triangles now
   * stopping at the borders of the collections of polygons
   * passed. The elevations of the remaining portions of the Triangulated
   * Topography are not changed. The result is a set of points
   * where the triangles intersect the polygons; this is returned,
   * with a map from each of these points to the corresponding
   * point on the original polygons.
   * @param boundaryPolygons collection of polygons that mark the outside boundaries
   * of this object
   * @param bermPolygons a collection of polygons that mark some short distance
   * away from the outside boundaries of this object, so that a 'berm' separating
   * it from other elements of the space can be created
   */
  public void punchHoles(Vector boundaryPolygons, Vector bermPolygons){

    if(DEBUG_PUNCH) draw(0, 0, 900, 900, -500, 600, -400, 500, this.toString() + "BEFORE");
    
    // Overlay the bermPolygons
    if(DEBUG_PUNCH) System.out.println("Overlaying berm polygons");
    if(bermPolygons != null) overlayPolygons(bermPolygons);   
    if(DEBUG_PUNCH) draw(0, 0, 900, 900, 100, 600, -400, 100, this.toString() + "BERM");

    // Put all the boundary polygons in
    if(DEBUG_PUNCH) System.out.println("Overlaying boundary polygons");
    HashMap map = overlayPolygons(boundaryPolygons);
    if(DEBUG_PUNCH) draw(0, 0, 900, 900, 100, 600, -400, 100, this.toString() + "BOUNDARY");
    
    // Remove all the interior triangles
    if(DEBUG_PUNCH) System.out.println("Removing interior points");
    removeAllPointsInteriorToPolygons(boundaryPolygons);
    if(DEBUG_PUNCH) draw(0, 0, 900, 900, 100, 600, -400, 100, this.toString() + "REMOVED");
    
    // Now take the map of points and replace triangles that touch the boundary
    // polygons with ones that go to the original elevation
    if(DEBUG_PUNCH) System.out.println("Replacing with proper elevations");
    replacePoints(map);
    if(DEBUG_PUNCH) draw(0, 0, 900, 900, 100, 600, -400, 100, this.toString() + "REPLACE");
    
  }
  
  /**
   * Takes one collection of polygons and projects all of their
   * vertices onto the existing topography, forming new triangles
   * but leaving the elevations of the landscape unchanged
   * @param polygons
   * @return a HashMap of points added to points from the vertices of the
   * original polygons
   */
  public HashMap overlayPolygons(Vector polygons){
    HashMap map = new HashMap();
    
    if(polygons == null || polygons.size()== 0) return map;
    
    if(DEBUG_OVERLAY) System.out.println("\n\n\nOverlaying polygons: " + polygons.size() + " of 'em into " + topoTriangles.size() + " triangles...");
    
    TTDraw frame = null;
    
    if(DEBUG_OVERLAY) frame = draw(0, 0, 800, 800, -500f, 4000f, -2500f, 2500f, this.toString());
    
    if(DEBUG_OVERLAY){
      float zoom = .75f;
      float centerx = 450f;
      float centery = 1250.25f;
      frame = draw(200, -1000, 1200, 1000, centerx - zoom, centerx + zoom, centery - zoom, centery + zoom, this.toString());
    }
    
    if(DEBUG_OVERLAY) frame.polygons = polygons;
    
    // These drive certain tests given below
//    String[] polygonSidesToPauseAt = new String[]{"LINE_765", "LINE_767"};
    String[] polygonSidesToPauseAt = new String[]{};
    Vector compare = new Vector();
    Vector compareFinal = (Vector)topoTriangles.clone();
    
    // Make a list of all the polygons' sides
    Vector sides = new Vector();
    for(int i = 0; i < polygons.size(); i++) sides.addAll(((Polygon)polygons.get(i)).getSides());
    
  
    // Eliminate duplicate sides:
    if(DEBUG_OVERLAY) System.out.println("Removing sides: initial = " + sides.size());
    for(int i = 0; i < sides.size() - 1; i++){
      LineSegment side1 = (LineSegment)sides.get(i);
      for(int j = i + 1; j < sides.size(); j++){
        LineSegment side2 = (LineSegment)sides.get(j);
        if(DEBUG_OVERLAY) frame.repaint(topoTriangles);
        if((side1.P1.isPlanarEqual(side2.P1, .0000001) && side1.P2.isPlanarEqual(side2.P2, .0000001)) ||
           (side1.P2.isPlanarEqual(side2.P1, .0000001) && side1.P1.isPlanarEqual(side2.P2, .0000001))){
          sides.remove(j); // Take the duplicate out
          j--;             // And decrement the index, so that when it is incremented you're back in the right spot
        }
      }
    }
    if(DEBUG_OVERLAY) System.out.println("Removing sides: final = " + sides.size());
    System.out.print("");

    Point N;
    for(Object sideObject: sides){
      LineSegment side = (LineSegment)sideObject;
      N = addPointAtProjectedElevation(side.P1);
      if(N != null) map.put(N, side.P1);
      N = addPointAtProjectedElevation(side.P2);
      if(N != null) map.put(N, side.P2);
    }
    
    if(DEBUG_OVERLAY) System.out.println("\n\n*****  NOW DOING SIDE INTERSECTIONS...");
    if(DEBUG_OVERLAY){
      System.out.print("");
    }
    
    for(Object sideObject: sides){
      LineSegment side = (LineSegment)sideObject;
    
      // Some debugging code will help here
      boolean exploringASide = false;
      
      if(DEBUG_OVERLAY){
        System.out.println("EXAMINING SIDE: " + side.display());
        if(polygonSidesToPauseAt.length == 0){
          exploringASide = true;
        }
        else{
          for(String ps: polygonSidesToPauseAt){
            if(ps.compareTo(side.IDString) == 0){
              exploringASide = true;
              break;
            }
          }
        }
      }
      // This is for debugging...
      if(DEBUG_OVERLAY && exploringASide) compare = (Vector)topoTriangles.clone();
      
      if(DEBUG_OVERLAY && exploringASide){
        frame.C1 = side;
        frame.C2 = null;
        frame.compare = null;
        frame.active = null;
        System.out.println("Drawing with polygon side being tested highlighted:\n   " + side.IDString + " @ " + side.display());
        frame.repaint(topoTriangles);
        System.out.print("");
      }
      
      Vector newTrianglesToAdd = new Vector();
      Vector newTriangles = new Vector();
      
      float minX = Math.min((float)side.P1.getX(), (float)side.P2.getX());
      float maxX = Math.max((float)side.P1.getX(), (float)side.P2.getX());
      
      float minY = Math.min((float)side.P1.getY(), (float)side.P2.getY());
      float maxY = Math.max((float)side.P1.getY(), (float)side.P2.getY());
      
      
      if(DEBUG_OVERLAY) System.out.println(" LOOPING THROUGH " + topoTriangles.size() + " TRIANGLES");
      for(int i = 0; i < topoTriangles.size(); i++){
        TopoTriangle T = (TopoTriangle)topoTriangles.get(i);
//        System.out.println("Triangle " + i + " (" + topoTriangles.size() + ") = " + T.getID());
        if(!(minX > T.getMaxX() ||
             minY > T.getMaxY() ||
             maxX < T.getMinX() ||
             maxY < T.getMinY())){
          newTriangles = T.slice(side);
          if(newTriangles.size() > 0){
            
  //          System.out.println("Removing " + T.ID + " and replacing with " + newTriangles.size() + " triangles");
            if(DEBUG_OVERLAY && exploringASide){
              frame.active = T;
              frame.compare = compare;
              frame.repaint(topoTriangles);
              System.out.print("");
            }
            topoTriangles.remove(i);
            i--;
            newTrianglesToAdd.addAll(newTriangles);
          }
        }
      }
      if(DEBUG_OVERLAY) System.out.println("topoTriangles originally had " + topoTriangles.size() + " and I'm adding " + newTrianglesToAdd.size());
      topoTriangles.addAll(newTrianglesToAdd);
      if(DEBUG_OVERLAY && exploringASide){
        frame.active = null;
        frame.C1 = null;
        frame.C2 = null;
        frame.compare = compare;
        frame.repaint(topoTriangles);
        System.out.print("");
      }
    }
    if(DEBUG_OVERLAY){
      frame.active = null;
      frame.C1 = null;
      frame.C2 = null;
      frame.compare = compareFinal;
      frame.polygons = polygons;
      frame.repaint(topoTriangles);
      System.out.print("");
    }  
    return map;
  }
  
    
  /**
   * Removes all points that are interior to the collection of polygons given
   * @param boundaryPolygons
   */
  public void removeAllPointsInteriorToPolygons(Vector boundaryPolygons){
    TTDraw frame = null;
    if(DEBUG_REMOVE) frame = draw(0, 0, 900, 900, 60.9f, 61.0f, -96.0f, -95.97f, this.toString() + "REMOVE");
    Vector V = new Vector();
    if(DEBUG_REMOVE) frame.polygons = V;
    for(int t = topoTriangles.size() - 1; t >= 0; t--){
      TopoTriangle T = (TopoTriangle)topoTriangles.get(t);
      if(DEBUG_REMOVE){
        frame.active = T;
        System.out.println("Testing to remove: " + idTriangle(T));
        frame.repaint(topoTriangles);
      }
      for(int p = 0; p < boundaryPolygons.size(); p++){
        Polygon P = (Polygon)boundaryPolygons.get(p);
        V.clear();
        V.add(P);
        if(DEBUG_REMOVE) frame.repaint(topoTriangles);
        System.out.print("");
        if(P.pointIsInside(T.getCenterPoint())){
          topoTriangles.remove(t);
          break;
        }
      }
    }  
  }
  
  /**
   * Replaces triangles that may have been created with one
   * set of points with new triangles made from a corresponding
   * set of points. The correspondence may be of any kind, but
   * typically one set lying on an abstract plane is replaced with
   * a set created within a topographical landscape and having the
   * correct elevations.
   * @param points
   */
  public void replacePoints(Map points){
    TopoTriangle N = null;
    for(int i = 0; i < topoTriangles.size(); i++){
      TopoTriangle T = (TopoTriangle)topoTriangles.get(i);
      Object O1 = points.get(T.getP1());
      Object O2 = points.get(T.getP2());
      Object O3 = points.get(T.getP3());
      
      if(O1 != null || O2 != null || O3 != null){
        Point P1 = (O1 == null ? T.getP1() : (Point)O1);
        Point P2 = (O2 == null ? T.getP2() : (Point)O2);
        Point P3 = (O3 == null ? T.getP3() : (Point)O3);
        N = new TopoTriangle(P1, P2, P3);
        topoTriangles.set(i, N);
      }
    }
  }
  
  /**
   * Returns a map of triangles in this triangulated topography
   * with coterminous triangles (triangles that have vertices that
   * are equivalent in the x,y dimensions, even if they differ
   * in z values). Non-coterminous triangles are not included
   * in the map. The triangles returned may be rotated versions
   * of the triangles in the other topography. Triangles that
   * have equal z values are also eliminated and are not returned.
   * The result is a map of triangles in this one and the other
   * one that differ in the z dimension only.
   * @param other
   * @return the map of triangles that match the other triangulated topography
   */
  public Map getCoterminousNonequalPairs(TriangulatedTopography other){
    Vector V1 = getListOfTrianglesInIncreasingXCoordinateOrder();
    Vector V2 = other.getListOfTrianglesInIncreasingXCoordinateOrder();
    Map ret = new HashMap();
    for(int i = V1.size() - 1; i >=0; i--){
      TopoTriangle toPlace = (TopoTriangle)V1.get(i);
      for(int j = V2.size() - 1; j >=0; j--){
        TopoTriangle compare = (TopoTriangle)V2.get(j);
        if(toPlace.isCoterminous(compare)){
          V1.remove(i);
          V2.remove(j);
          while(!toPlace.getP1().isPlanarEqual(compare.getP1()) || !toPlace.getP2().isPlanarEqual(compare.getP2()) || !toPlace.getP3().isPlanarEqual(compare.getP3())) 
              compare = compare.rotate();
          if((Math.abs(toPlace.getP1().getZ()- compare.getP1().getZ()) > Geometry.epsilon) ||
             (Math.abs(toPlace.getP2().getZ()- compare.getP2().getZ()) > Geometry.epsilon) ||
             (Math.abs(toPlace.getP3().getZ()- compare.getP3().getZ()) > Geometry.epsilon)){
             ret.put(toPlace, compare);
          }
          break;
        }
      }
    }
    return ret;    
  }
  
  /**
   * Returns a new triangulated topography that is the result
   * of projected another topography onto this one. That is, the
   * pattern of vertices and sides from the other topography is
   * replicated within the existing vertices and sides of this
   * one, but with the elevation values of this one. The
   * metaphor is a light shining through the other one, casting
   * shadows onto this one.
   * @param other
   * @return the projection of the other triangulated topography onto this one
   */
  public TriangulatedTopography getProjection(TriangulatedTopography other){
    TriangulatedTopography ret = new TriangulatedTopography(new Vector());
    
    TTDraw projecting          = null;
    TTDraw returning           = null;
    
    if(DEBUG_PROJECT){
      this.draw(0,0, 400, 400, -300, 600, -300, 300, "GETTING PROJECTION", null);
      projecting          = other.draw(0, 400, 400, 400, -300, 600, -300, 300, "PROJECTING OTHER", null);
      returning           = ret.draw(400, 0, 400, 400, -300, 600, -300, 300, "RETURNING", null);
    }
    
    for(int i = 0; i < other.topoTriangles.size(); i++){
      TopoTriangle T = (TopoTriangle)other.topoTriangles.get(i);
      if(DEBUG_PROJECT) projecting.active = T;
      if(DEBUG_PROJECT) projecting.repaint();
      TriangulatedTopography interim = getProjection(T);
      ret.addTopoTriangles(interim.getTriangles());
      if(DEBUG_PROJECT) returning.repaint(ret.topoTriangles);
    }
    return ret;
  }
  
  /**
   * Gets a projection of a single triangle onto this one, as if that triangle's
   * shadow were cast on this one. Returns only elements interior to the specified
   * triangle.
   * @param A the other triangle
   * @return the requested projection
   */
  public TriangulatedTopography getProjection(TopoTriangle A){
    TriangulatedTopography TT = new TriangulatedTopography(new Vector());
    Vector V = getSubset(A.getMinX(), A.getMinY(), A.getMaxX(), A.getMaxY()).topoTriangles;
    for(int j = 0; j < V.size(); j++){
      TT.addTopoTriangles(TopoTriangleProjector.project(A, (TopoTriangle)V.get(j)));
    }
    return TT;
  }
  
  /**
   * Returns the set of all triangles in this Triangulated Topography
   * that touch some point within the given square
   * @param minX
   * @param minY
   * @param maxX
   * @param maxY
   * @return the set of all triangles within the square
   */
  public TriangulatedTopography getSubset(double minX, double minY, double maxX, double maxY){
    Vector V = new Vector();
    for(int i = 0; i < topoTriangles.size(); i++){
      TopoTriangle T = (TopoTriangle)topoTriangles.get(i);
      if( (T.getMaxX() > minX) && (T.getMinX() < maxX) && (T.getMaxY() > minY) && (T.getMinY() < maxY)) V.add(T);
    }
    TriangulatedTopography ret = new TriangulatedTopography(new Vector());
    ret.addTopoTriangles(V);
    return ret;
  }
  
  /**
   * Calculates the topographic volume difference between this triangulated topography
   * and another one. Assumes that volumes above the other one and below this one are
   * 'positive', while volumes in which this one lies underneath the other one are 'negative'.
   * @param lower
   * @return the topographic volume differential between this and another triangulated topography
   */
  public TopographicVolumeDifferential getTopographicVolumeDifferential(TriangulatedTopography lower){
    if(DEBUG_VOLUME) System.out.println("Getting volume differential...");
    TopographicVolumeDifferential ret = new TopographicVolumeDifferential();
    TriangulatedTopography UPPER_WITH_LOWER_PROJECTED = this.getProjection(lower);
    
    TriangulatedTopography LOWER_WITH_UPPER_PROJECTED = lower.getProjection(UPPER_WITH_LOWER_PROJECTED);
    Map coterminous = UPPER_WITH_LOWER_PROJECTED.getCoterminousNonequalPairs(LOWER_WITH_UPPER_PROJECTED);
    
    TriangulatedTopography displayKeys = new TriangulatedTopography(new Vector());
    displayKeys.addTopoTriangles(coterminous.values());
      
    TTDraw keys = null;
    if(DEBUG_VOLUME) keys = displayKeys.draw(0, 0, 800, 800, 150, 510, -50, 150, "KEYS FROM MAP", null);
    
    Set A = coterminous.keySet();
    
    Iterator i = A.iterator();
    int c = 0;
    if(DEBUG_VOLUME) System.out.println("\n\n\nBeginning loop through map -------------------------------------");
    while(i.hasNext()){
      if(DEBUG_VOLUME) System.out.println("Looping through map of triangles: " + (c++) + " of " + A.size());
      TopoTriangle T1 = (TopoTriangle)i.next();
      TopoTriangle T2 = (TopoTriangle)coterminous.get(T1);
      if(DEBUG_VOLUME){
        keys.active = T2;
        keys.repaint();
      }
      ret.add(T1.getVolumeDifference(T2));
    }
    return ret;
  }

  
  /* Reporting */
  
  /**
   * Writes a list of the entire collection of triangles to the standard output
   *
   */
  public void write(){
    for(int i = 0; i < topoTriangles.size(); i++){
      TopoTriangle T = (TopoTriangle)topoTriangles.get(i);
      System.out.println(idTriangle(T));
    }
  } 
  
  /**
   * Identifies the triangle with its position in the triangles list (for debugging)
   * @param T
   * @return a String identifying the triangle
   */
  public String idTriangle(TopoTriangle T){
    return "T:" + T.getID() + " " + topoTriangles.indexOf(T) + " " + T.ptString();
  }

  
  /* Graphics */
  
  /** 
   * @see abcm.graphics.SVGRepresentable#addToSVG(abcm.graphics.SVGAssist) 
   */
  public void addToSVG(SVGAssist svg){
    System.out.println("Adding delaunay to SVG; svgUPIX: " + svg.getUnitsPerInchX());
    SVGRepresentable[] triangles = new SVGRepresentable[topoTriangles.size()];
    for(int i = 0; i< topoTriangles.size(); i++){triangles[i] = (SVGRepresentable)topoTriangles.get(i); }
    svg.addG(0,1,"delaunayTriangulation","id=\"delaunayTriangulation\" visibility=\"display\"",triangles);
  }
  
  /** 
   * @see abcm.graphics.X3DRepresentable#addToX3D(abcm.graphics.X3DAssist)
   */
  public void addToX3D(X3DAssist x3d){
    if(DEBUG_X3D) draw(0, 0, 900, 900, -500, 600, -400, 500, this.toString()+ "FINAL");
    XMLAssist xml = x3d.getXML();

    x3d.addDefinition("Terrain", x3dDefinition);

    X3DIndexedFaceSetCreator ifsCreator = new X3DIndexedFaceSetCreator(topoTriangles);
    
    xml.addLine("<!-- Terrain -->\n", XMLAssist.INDENT_BLOCK);
    xml.open("Group");
      xml.open("Shape");
        xml.addLine("Appearance", "USE='Terrain'","");
        xml.open("IndexedFaceSet", "coordIndex=\"" + ifsCreator.getFaceList() + "\"" + " solid=\"false\"");
          xml.addLine("Coordinate", "point=\"" + ifsCreator.getPointList() + "\"", "");
        xml.closeTag(); // Indexed Face Set
      xml.closeTag(); // Terrain Shape
    xml.closeTag(); // Terrain Group
    
  }
  
  /**
   * Adds contour lines to the triangular topography. 
   * @param svg the SVGAssist object to which these should be added
   * @param contourInterval the interval at which the lines should be drawn
   */
  public void addContourLinesToSVG(SVGAssist svg, double contourInterval){
    for(int i = 0; i < topoTriangles.size(); i++){
      TopoTriangle T = (TopoTriangle)topoTriangles.get(i);
      ContourLine CL = T.getContourLine(contourInterval);
      if(CL!=null) CL.addToSVG(svg);      
    }
  }
  
  /**
   * Gets a collection of ContourLines for this topography 
   * @param contourInterval the contour line interval
   * @return the collection of countour lines
   */
  public Vector getContourLines(double contourInterval){
    Vector V = new Vector();
    for(int i = 0; i < topoTriangles.size(); i++){
      TopoTriangle T = (TopoTriangle)topoTriangles.get(i);
      ContourLine CL = T.getContourLine(contourInterval);
      if(CL!=null) V.add(CL);      
    }
    return V;
  }
  
  /**
   * Draws this triangulated topography to an output window
   * @param xPos
   * @param yPos
   * @param height
   * @param width
   * @param minX
   * @param maxX
   * @param minY
   * @param maxY
   * @param name
   * @return the output window TTDraw frame
   */
  public TTDraw draw(int xPos, int yPos, int height, int width, float minX, float maxX, float minY, float maxY, String name){
    return draw(xPos, yPos, height, width, minX, maxX, minY, maxY, name, null);
  }
  
  /**
   * Draws this triangulated topography to an output window
   * @param xPos
   * @param yPos
   * @param height
   * @param width
   * @param minX
   * @param maxX
   * @param minY
   * @param maxY
   * @param name
   * @param compare
   * @return the output window TTDraw frame
   */
  public TTDraw draw(int xPos, int yPos, int height, int width, float minX, float maxX, float minY, float maxY, String name, Vector compare){
    TTDraw frame = new TTDraw(minX, maxX, minY, maxY, name, (Vector)this.topoTriangles.clone(), null, compare);
    frame.makeVisible(xPos, yPos, width, height);
    return frame;
  }
  
  
  /* Object Cloning */
  
  /** 
   * Note that this is a shallow clone; for a deep clone, use the getDeepClone() method
   * @see java.lang.Object#clone()
   */
  public Object clone(){
    TriangulatedTopography TT =  new TriangulatedTopography(new Vector());
    TT.topoTriangles = (ABCMDeepCloneableVector)this.topoTriangles.clone();
    return TT;
  }
  
  /**
   * 
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    TriangulatedTopography clone = new TriangulatedTopography(new Vector());
    cloneStore.put(this, clone);
    clone.addTopoTriangles((ABCMDeepCloneableVector)topoTriangles.getDeepClone(cloneStore));
    return clone;
  }
  

  /** 
   * @see abcm.space.HasElevation#getZ(abcm.math.Point) 
   */
  public double getZ(Point P){
    for(int i = 0; i < topoTriangles.size(); i++){
      TopoTriangle T = (TopoTriangle)topoTriangles.get(i);
      double z = T.getZ(P);
      if(!Double.isNaN(z)) return z;
    }
    return Double.NaN;
  }
  
}
