package abcm.utilities;

/**
 * Provides a method for parsing a line of text with quotes
 * into chunks, taking quoted sections as one chunk.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public final class QuotedLineParser {

  /**
   * Parses a string into 'words', counting quoted sections as
   * single words. Thus:
   * 
   * This is my "Very Big Example"
   * 
   * becomes:
   * [0] This
   * [1] is
   * [2] my
   * [3] Very Big Example
   * 
   * Trims all words beyond the allowed number of entries
   * 
   * @param cText text to be parsed
   * @param entries maximum number of entries
   * @return an array of parsed words
   */
  public static String[] parseQuoted(String cText, int entries){
    final int OUT_OF_QUOTE = 0;
    final int IN_QUOTE = 1;
    final int IN_WHITESPACE = 2;
    
    String[] commands = new String[entries];
    for(int i = 0; i < entries; i++) commands[i]="";
    
    int mode = OUT_OF_QUOTE;
    int cc = 0;
    while((cText.length()) > 0 && (cc < entries)){
      switch (mode){
        case OUT_OF_QUOTE:
          if(cText.startsWith("\"")){
            mode = IN_QUOTE;
            cText = cText.substring(1);
            break;
          }
          if(cText.startsWith(" ")){
            mode = IN_WHITESPACE;
            cText = cText.substring(1);
            cc++;
            break;
          }
          commands[cc] += cText.substring(0,1);
          cText = cText.substring(1);
          break;
        case IN_QUOTE:
          if(cText.startsWith("\"")){
            mode = IN_WHITESPACE;
            cText = cText.substring(1);
            cc++;
            break;
          }
          commands[cc] += cText.substring(0,1);
          cText = cText.substring(1);
          break;
        case IN_WHITESPACE:
          if(cText.startsWith(" ")) cText = cText.substring(1);
          else mode = OUT_OF_QUOTE;
          break;
      }
    }
    
    return commands;
    
  }

}
