package abcm.utilities;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

/**
 * This is a generic class used by all routines that process script
 * files (including .abcm, .nar, and .nru files). It wraps basic functions
 * like reading the next input line, etc.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ScriptReader {
  
  /**
   * Indicates that the status of the current read operation is inside a line
   */
  private static final int IN_LINE              = 0;
  
  /**
   * Indicates that the status of the current read operation is not inside a line
   */
  private static final int NOT_IN_LINE          = 1;
  
  /**
   * Indicates that the status of the current read operation is EOF
   */
  private static final int END_OF_FILE          = 2;

  /**
   * A FileReader object used to read the files
   */
  private FileReader fr;
  
  /**
   * Indicates that the status of the current read operation
   */
  private int mode;

  /**
   * A String describing the reasons for failure of the recent read operation
   */  
  private String errorReport;
  
  /**
   * A flag indicating whether the recent operation was successful
   */
  private boolean successFlag;
  
  /**
   * Constructor that provides the name of the file to be read
   * @param fileName name of the file to be read
   */
  public ScriptReader(String fileName){
    errorReport = "No Script Reader Error.";
    mode = IN_LINE;
    try{
      fr = new FileReader(fileName);
      successFlag = true;
    }
    catch(FileNotFoundException fnfe){
      errorReport = "Trying to find '" + fileName  + "' but: " + fnfe.getMessage();
      mode = END_OF_FILE;
      successFlag = false;
    }
  }
  
  /**
   * Determines whether the end of the file has been reached
   * @return true if EOF, false if not
   */
  public boolean EOF(){   return (mode == END_OF_FILE);     }
  
  
  /**
   * Returns the current status of the read operation
   * @return mode
   */
  public int getMode(){   return mode;   }
  
  /**
   * Gets the next line in the script file. Parses to avoid
   * empty lines
   * @return the next input line from the file
   */
  public String getNextLine(){
    String nextLine = "";
    try{      
      char[] i = new char[1];
      if(!fr.ready()){
        mode = END_OF_FILE;
        fr.close();
        return null;
      }
      while(fr.ready()){
        fr.read(i);
        // If you're in a line, keep parsing
        if(mode == IN_LINE){
          switch (i[0]){
            case 10: //
            case 13: // End of the line has been reached
              mode = NOT_IN_LINE;
              return nextLine;            
            default: // Just add the character to the string
              nextLine += i[0];
              mode = IN_LINE;
          }
        } // If you're not in a line, look for the first character of the next line
        else{
          switch(i[0]){
            case 10: // Do nothing
            case 13: // Do nothing
            case 32: // Do nothing
              mode = NOT_IN_LINE;
              break;
            default: // This is a new character; add it to the line and return to parsing
              nextLine += i[0];
              mode = IN_LINE;
          }
        }
      }
    }
    catch(IOException ioe){
      System.err.println("EXCEPTION OCCURRED IN SCRIPT READER getNextLine");
      return null;
    }
    mode = END_OF_FILE;
    return nextLine;
  }
  
  /**
   * Gets an error report, indicating the character of the last error encountered.
   * @return the error report.
   */
  public String getErrorReport() {    return errorReport;     }

  /**
   * Gets a flag indicating the success or failure of the last
   * read operation.
   * @return Returns the successFlag.
   */
  public boolean getSuccessFlag() {   return successFlag;     }

  /**
   * Closes the file being read. This will stop all reading and should only be done
   * after file is read or if an error has occured
   *
   */
  public void closeFile(){
    try{
      fr.close();
    }
    catch(Exception E){
      errorReport = E.getMessage();
    }
  }
  
}
