package abcm.utilities;

import java.io.StringWriter;

import abcm.math.BasicMath;

/**
 * Provides support methods for text transformation,
 * etc., for the HWM Application.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class TextFormatter {

  /**
   * Performs a wordWrap without indentation (indent = 0);
   * see wordWrap(String, int, int)
   * @param inText Text to be parsed
   * @param width Width at which to begin looking for 
   * space to break line.
   * @return The parsed version of the text, with line
   * breaks included.
   */
  public static String wordWrap(String inText, int width){
    return wordWrap(inText, width, 0);
  }
  
  /**
   * Takes a single long line of text and includes breaks
   * at spaces given the specified width. Note that
   * this does NOT break at the first space prior to
   * the width, but rather the first space AFTER. It is
   * therefor useful for breaking long lines but not
   * perfect at it.<br/>
   * <br/>
   * <b>Example:</b><br/>
   * <br/>
   * Input text exceeds the width of a line (20 characters)<br/>
   * <br/>
   * <b>becomes:</b><br/>
   * <br/>
   * Input text exceeds the<br/> 
   * width of a line (20 characters)
   * @param inText Text to be parsed
   * @param width Width at which to begin looking for 
   * space to break line.
   * @param indent number of spaces to indent, when a line is indented
   * @return The parsed version of the text, with line
   * breaks included.
   */  
  public static String wordWrap(String inText, int width, int indent){
    String indentString = "";
    for(int i = 0; i < indent; i++){ indentString+=" "; }
    if(inText.length() + indent < width) return indentString + inText;
    
    // Else it's too long
    StringWriter out = new StringWriter();
    out.write(indentString);
    int col = indent;
    String[] words = inText.split(" ");
    for(String word: words){
      if((word.length() + col) > width){
        out.write("\n");
        col = 0;
      }
      out.write(word);
      out.write(" ");
      col += word.length() + 1;
    }
    return out.toString();       
  }
  
  /**
   * Takes a collection of lines and returns them all as a single
   * String collectively wrapped using the wordWrap method.
   * @param inText the input text for this method
   * @param width the declared width at which text is to be wrapped
   * @return the reformatted text
   */
  public static String lineWrap(String inText, int width){
    return lineWrap(inText, width, 0);
  }
  
  /**
   * Takes a collection of lines and returns them all as a single
   * String collectively wrapped using the wordWrap method.
   * @param inText the input text for this method
   * @param width the declared width at which text is to be wrapped
   * @param indent number of spaces to indent
   * @return the reformatted text
   */
  public static String lineWrap(String inText, int width, int indent){
    String[] lines = inText.split("\n");
    StringWriter out = new StringWriter();
    for(String line: lines){
      out.write(wordWrap(line, width, indent));
      out.write("\n");
    }
    return out.toString();
  }
  
  /**
   * Returns a formatted string that takes a given value, rounds it
   * to a specific precision, and adds units to it.
   * @param value the value to be displayed
   * @param precision the precision required; negative numbers round to
   * powers of 10 above zero, while positive numbers specify the number
   * of decimals to display
   * @param unitID the unit to be displayed
   * @return a formatted value with a unit, i.e. '1.25 m'
   */
  public static String roundWithUnits(double value, int precision, String unitID){
    if(value == 0) return "0.0 " + unitID;
    // First, determine the magnitude of the value
    int i = (int)(Math.floor(Math.log(value)/Math.log(10)));
    double newVal;
    // If the value passed is so large it will be displayed in scientific notation
    if( i >  6){
      // If the precision is negative or zero, the result should be the value 
      // rounded to 10 ^ precision, so 1.2345678E12 with precision -9 will give 1.234E12
      if(precision <= 0) return roundWithUnits(value, i + precision, unitID);
      else{
        // The precision is positive
        // This will round the notational result to the number of decimal places, so
        // 1.2345678E12 with precision 4 returns 1.2345E12
        double adj = value / Math.pow(10,i);
        double adjRounded = ((Math.floor(Math.pow(10, precision) * adj + .5))/Math.pow(10, precision));
        if(precision < 0) adjRounded = Math.round(adjRounded);
        newVal = Double.parseDouble(adjRounded + "E" + i);
      }
    }
    // If the value passed is so small it will be displayed in scientific notation,
    else if( i < -6){
      // If the precision is negative or zero, this is meaningless and returns the original number
      if(precision <= 0) newVal = value;
      else{
        // The precision is positive
        // This will round the notational result to the number of decimal places, so
        // 1.2345678E12 with precision 4 returns 1.2345E12
        double adj = value / Math.pow(10,i);
        double adjRounded = BasicMath.roundDouble(adj, precision);
        newVal = Double.parseDouble(adjRounded + "E" + i);
      }
    }
    // The value passed is will not be displayed in scientific notation,
    //   and the precision specified is positive
    // This performs a normal round and returns the result   
    else{
      newVal = BasicMath.roundDouble(value, precision);
    }
        
    String Val = "" + newVal;
    
    // Rounding errors occur and can result in numbers like 1.234000000000002 when
    // 1.234 is the desired result. This will trim these...
    
    // Strip off the .0, if appropriate
    if(precision <=0 && !(Val.indexOf('E')>-1)) Val = Val.substring(0, Val.length()-2);
    
    // Insert a space if appropriate
    Val = Val.replaceAll("E", " x 10^");
    
    return Val + " " + unitID;
  }

  /**
   * Pads a String from a given double value to a specified width
   * by adding trailing spaces. If the value's initial representation
   * is longer than the specified width, the returned String is
   * not trimmed and includes the entire value. This is mainly
   * useful for prettifying output.
   * @param value value to be represented
   * @param width width to which it should be padded
   * @return padded String, i.e. "6.234    "
   */
  public static String toWidth(double value, int width){
    String s = "" + value;
    if(s.length() > width) return s;
    while(s.length() < width) s = "               " + s;
    return s.substring(s.length() - width);
  }  

  /**
   * Removes all instances of double linefeeds from a given String.
   * All multiple instances of linefeeds will be removed (so that
   * "\n\n\n\n\n" becomes "\n")
   * @param x text to be reformatted
   * @return reformatted text
   */
  public static String removeDoubleLinefeeds(String x){
    String parse = x;
    int lineDiff = 1;
    while(lineDiff != 0){
      int startLength = parse.length();
      parse = parse.replaceAll("\n\n", "\n");
      lineDiff = startLength - parse.length();
    }
    return parse;
  }

  /**
   * Calculates the approximate width of a given text, in inches.
   * Assumes that on average a character is 1/2 the font size
   * given- that is, if the font is 15 pt, the average character
   * will be 7.5 pts wide. The number returned is in inches, 72 pts per
   * inch.
   * @param t The text to be assessed
   * @param fontSize the size of the font, in pts
   * @return the width of the longest line, in pts
   */
  public static double getTextWidth(String t, int fontSize){
    return getLengthOfLongestLine(t) * (fontSize / 2d) / 72d;
  }
  
  /**
   * Returns the number of characters in the longest line
   * of the text passed
   * @param t the text to be passed
   * @return the number of characters in the longest line
   */
  public static int getLengthOfLongestLine(String t){
    int charCount = 0;
    int maxCharCount = 0;
    String parse = t + "\n";
    parse = TextFormatter.removeDoubleLinefeeds(parse);
    while(parse.length() > 0){
      String S = parse.substring(0, 1);
      if(S.equalsIgnoreCase("\n")){
        maxCharCount = (charCount > maxCharCount ? charCount : maxCharCount);
        charCount = 0;
      }
      else{  charCount++;  }
      parse = parse.substring(1);
    }
    return maxCharCount;
  }
  
  /**
   * Returns the height of the given text, in inches.
   * Note that it measures from the top of the first line to the
   * bottom of the last line- only the space between lines is
   * included, not the trailing space at the bottom.
   * @param t the text to be passed
   * @param fontSize the size of the font, in pts (72 pts per inch
   * @param linespacing the size of the space between lines (in pts)
   * @return the height of the text, in inches.
   */
  public static double getTextHeight(String t, int fontSize, int linespacing){
    long lineCount = lineCount(t);
    return ((lineCount * fontSize) + ((lineCount - 1) * linespacing))/ 72d;
  }
  
  /**
   * Gets a count of the lines of text in the specified input string.
   * Double linefeeds are removed before the count is made
   * @param t text to be examined
   * @return count of lines in the specified text
   */
  public static long lineCount(String t){
    long lineCount = 0;
    String parse = t + "\n";
    parse = TextFormatter.removeDoubleLinefeeds(parse);    
    while(parse.length() > 0){
      String S = parse.substring(0, 1);
      if(S.equalsIgnoreCase("\n")){
        lineCount++;
      }
      parse = parse.substring(1);
    }
    return lineCount;
  }
  
  /**
   * Trims the trailing .xyz extension off of a file name.
   * Note that the extension is not assumed to be three characters; hence
   * myFile.type becomes myFile.
   * @param fileName file name to be trimmed
   * @return trimmed file name
   */
  public static String getFileWithoutExtension(String fileName){
    if(fileName == null) return null;
    int index = fileName.lastIndexOf(".");
    if(index == -1) return fileName;
    return fileName.substring(0,index);     
  }
  
  /**
   * Returns the file extension from a file name. Note that this only
   * assumes that the String passed will include at least one dot
   * character; the returned String does not include the dot character itself.
   * If there is no dot character, the returned string is null
   * @param fileName the file name to be parsed
   * @return the extension of the file
   */
  public static String getFileExtension(String fileName){
    if(fileName == null) return null;
    int index = fileName.lastIndexOf(".");
    if(index == -1) return null;
    return fileName.substring(index+1);     
  }

  /**
   * Formats a string to eliminate characters that cannot be processed by TeX
   * @param text the text to be corrected
   * @return the corrected version of the text
   */
  public static String makeTEXSafe(String text){
    if(text == null) return null;
    String ret = text.replaceAll("\\Q_\\E", "\\\\_");
    System.out.println("Converted " + text + " into " + ret);
    return ret;
  }
  
}
