package abcm.utilities;

import java.io.StringWriter;
import java.util.Vector;

/**
 * Object that allows the easy creation of a well-formatted XML
 * document. Methods are provided for adding tags and attributes
 * in single-line fashion; opening and closing of tags is
 * handled automatically.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class XMLAssist {  
  
  /**
   * Standard XML header
   */
  public static final String XMLHEADER = "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n";

  /**
   * Static value for use by calling methods to indicate that the header should be excluded
   */
  public static final boolean EXCLUDE_HEADER = true;
  
  /**
   * Static value for use by calling methods to indicate that the header should be included
   */
  public static final boolean INCLUDE_HEADER = false;
  
  /**
   * Static value for use by calling methods to indicate that the block added should be indented
   */
  public static final boolean INDENT_BLOCK = true;
  
  /**
   * Static value for use by calling methods to indicate that the block added should not be indented
   */
  public static final boolean DO_NOT_INDENT_BLOCK = false;
  
  /**
   * StringWriter object that constructs the xml output
   */
  private StringWriter xml;
  
  /**
   * Collection of 'open' tags that have been issued for this xml document; this will
   * operate like a stack that allows nesting levels to be tracked backwards
   */
  private Vector openTags;  

  /**
   * Constructor with no arguments.
   *
   */
  public XMLAssist(){
    xml = new StringWriter(5120000);
    openTags = new Vector();
  }
  
  /**
   * Opens a new tag in the XML documents
   * @param tag the name of the tag to be opened
   */
  public void open(String tag){
    addLine(getIndent(openTags.size()) + "<" + tag + ">\n");
    openTags.add(0,tag);
  }

  /**
   * Opens a new tag in the XML documents
   * @param tag the name of the tag to be opened
   * @param attributes any attributes that will be appended to this tag
   */
  public void open(String tag, String attributes){
    addLine(getIndent(openTags.size()) + "<" + tag + " " + attributes + ">\n");
    openTags.add(0,tag);
  }
  
  /**
   * Closes a tag. The XML Assist object will keep a stack of tags that are
   * open and close them in the reverse order. If this method is invoked
   * more times than there are open tags, no action is taken.
   */
  public void closeTag(){
    if(openTags.size() > 0){
      if(openTags.size() > 1)
        addLine(getIndent(openTags.size() - 1) + "</" + (String)openTags.remove(0) + ">\n");
      else
        addLine(getIndent(openTags.size() - 1) + "</" + (String)openTags.remove(0) + ">");
    }
  }
  
  /**
   * Gets the current XML document, as a String. Using this
   * method includes the standard XML header.
   * @return the current XML document
   */
  public String getXML(){
    return getXML(false);
  }  
  
  /**
   * Gets the current XML document, as a String. The argument
   * specifies whether the XML header is to be included or not
   * @param excludeHeader true if the header is to be excluded, false if not
   * @return the XML document
   */
  public String getXML(boolean excludeHeader){
    if(excludeHeader) return xml.toString();
    else{
//      StringWriter sw = new StringWriter(xml.getBuffer().length()+XMLHEADER.length());
//      sw.write(XMLHEADER);
//      sw.write(xml.toString());
      return xml.toString();
//      return sw.toString();
    }
  }
  
  /**
   * Appends a line to the XML document; no transformation is made
   * @param newLine the text of the line to be added
   */
  public void addLine(String newLine){
    xml.write(newLine);
  }
  
  /**
   * Appends a line to the XML document. If the doIndent argument is true,
   * will indent the line to match the indentation given the current
   * number of tags.
   * @param newLine the text of the line to be added
   * @param doIndent true if the new line is to be indented, false otherwise
   */
  public void addLine(String newLine, boolean doIndent){
    if(doIndent) addLine(getIndent(openTags.size()) + newLine);
    else         addLine(newLine);
  }
  
  /**
   * Appends a block of text to the XML document.
   * @param newBlock the text to be appended
   */
  public void addBlock(String newBlock){
    xml.write(newBlock + "\n");
  }
  
  /**
   * Appends a block of text to the XML document. If the doIndent
   * argument is true, each line of the block is indented to match
   * the rest of the document, given the number of tags that have been
   * opened.
   * @param newBlock the text to be added
   * @param doIndent true if the new line is to be indented, false otherwise
   */
  public void addBlock(String newBlock, boolean doIndent){
    String block = newBlock;
    if(doIndent) block = getIndent(openTags.size()) + newBlock.replaceAll("\n", "\n" + getIndent(openTags.size()));
    addBlock(block);
  }
  
  
  /* Basic tag and value makers */
  /**
   * Makes a new tag, including opening the tag,
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param tag the name of the new tag to be appended
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, String value){
    return "<" + tag + ">" + value + "</" + tag + ">";
  }

  /**
   * Makes a new tag, including opening the tag,
   * including any attributes of the tag
   * inserting the value, and closing the tag, all on a single
   * line of the XML document. If the 'value' string is
   * zero-length, uses the shortcut form to close the tag.
   * @param tag the name of the new tag to be appended
   * @param attributes any attributes of the tag; these are inserted verbatim
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, String attributes, String value){
    return (value.length() == 0 ?
             ("<" + tag + " " + attributes + "/>") :
             ("<" + tag + " " + attributes + ">" + value + "</" + tag + ">"));
  }
  
  /**
   * Makes a new tag, including opening the tag,
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param tag the name of the new tag to be appended
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, int value){
    return "<" + tag + ">" + value + "</" + tag + ">";
  }

  /**
   * Makes a new tag, including opening the tag,
   * including any attributes of the tag
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param tag the name of the new tag to be appended
   * @param attributes any attributes of the tag; these are inserted verbatim
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, String attributes, int value){
    return "<" + tag + " " + attributes + ">" + value + "</" + tag + ">";
  }

  /**
   * Makes a new tag, including opening the tag,
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param tag the name of the new tag to be appended
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, long value){
    return "<" + tag + ">" + value + "</" + tag + ">";
  }

  /**
   * Makes a new tag, including opening the tag,
   * including any attributes of the tag
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param tag the name of the new tag to be appended
   * @param attributes any attributes of the tag; these are inserted verbatim
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, String attributes, long value){
    return "<" + tag + " " + attributes + ">" + value + "</" + tag + ">";
  }

  /**
   * Makes a new tag, including opening the tag,
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param tag the name of the new tag to be appended
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, double value){
    return "<" + tag + ">" + value + "</" + tag + ">";
  }
  
  /**
   * Makes a new tag, including opening the tag,
   * including any attributes of the tag
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param tag the name of the new tag to be appended
   * @param attributes any attributes of the tag; these are inserted verbatim
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, String attributes, double value){
    return "<" + tag + " " + attributes + ">" + value + "</" + tag + ">";
  }
    
  /**
   * Makes a new tag, including opening the tag,
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param tag the name of the new tag to be appended
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, Boolean value){
    return "<" + tag + ">" + value + "</" + tag + ">";
  }

  /**
   * Makes a new tag, including opening the tag,
   * including any attributes of the tag
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param tag the name of the new tag to be appended
   * @param attributes any attributes of the tag; these are inserted verbatim
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String maketag(String tag, String attributes, Boolean value){
    return "<" + tag + " " + attributes + ">" + value + "</" + tag + ">";
  }
  
  /* Basic line making */
  /**
   * Makes the next line of the XML document, including opening 
   * the next tag preceded by a correct indent based on the number of open tags
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param indent the number of spaces this line should be indented
   * @param tag the name of the new tag to be appended
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String makeline(int indent, String tag, String value){
    return getIndent(indent) + "<" + tag + ">" + value + "</" + tag + ">\n";
  }
  
  /**
   * Makes the next line of the XML document, including opening 
   * the next tag (with attributed) preceded by a correct 
   * indent based on the number of open tags
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param indent the number of spaces this line should be indented
   * @param tag the name of the new tag to be appended
   * @param attributes any attributes of the tag; these are inserted verbatim 
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String makeline(int indent, String tag, String attributes, String value){
    return getIndent(indent) + maketag(tag, attributes, value) + "\n";
//    return getIndent(indent) + "<" + tag + " " + attributes + ">" + value + "</" + tag + ">\n";
  }

  /**
   * Makes the next line of the XML document, including opening 
   * the next tag preceded by a correct indent based on the number of open tags
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param indent the number of spaces this line should be indented
   * @param tag the name of the new tag to be appended
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String makeline(int indent, String tag, int value){
    return getIndent(indent) + "<" + tag + ">" + value + "</" + tag + ">\n";
  }
  
  /**
   * Makes the next line of the XML document, including opening 
   * the next tag (with attributed) preceded by a correct 
   * indent based on the number of open tags
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param indent the number of spaces this line should be indented
   * @param tag the name of the new tag to be appended
   * @param attributes any attributes of the tag; these are inserted verbatim 
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String makeline(int indent, String tag, String attributes, int value){
    return getIndent(indent) + "<" + tag + " " + attributes + ">" + value + "</" + tag + ">\n";
  }

  /**
   * Makes the next line of the XML document, including opening 
   * the next tag preceded by a correct indent based on the number of open tags
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param indent the number of spaces this line should be indented
   * @param tag the name of the new tag to be appended
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String makeline(int indent, String tag, long value){
    return getIndent(indent) + "<" + tag + ">" + value + "</" + tag + ">\n";
  }
  
  /**
   * Makes the next line of the XML document, including opening 
   * the next tag (with attributed) preceded by a correct 
   * indent based on the number of open tags
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param indent the number of spaces this line should be indented
   * @param tag the name of the new tag to be appended
   * @param attributes any attributes of the tag; these are inserted verbatim 
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String makeline(int indent, String tag, String attributes, long value){
    return getIndent(indent) + "<" + tag + " " + attributes + ">" + value + "</" + tag + ">\n";
  }

  /**
   * Makes the next line of the XML document, including opening 
   * the next tag preceded by a correct indent based on the number of open tags
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param indent the number of spaces this line should be indented
   * @param tag the name of the new tag to be appended
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String makeline(int indent, String tag, double value){
    return getIndent(indent) + "<" + tag + ">" + value + "</" + tag + ">\n";
  }
  
  /**
   * Makes the next line of the XML document, including opening 
   * the next tag (with attributed) preceded by a correct 
   * indent based on the number of open tags
   * inserting the value, and closing the tag, all on a single
   * line of the XML document.
   * @param indent the number of spaces this line should be indented
   * @param tag the name of the new tag to be appended
   * @param attributes any attributes of the tag; these are inserted verbatim 
   * @param value the value to be inserted within the tag
   * @return the new tag
   */
  public static String makeline(int indent, String tag, String attributes, double value){
    return getIndent(indent) + "<" + tag + " " + attributes + ">" + value + "</" + tag + ">\n";
  }
   
    
  /* Add a line to the document being built */
  /**
   * Adds the next line to the document being built.
   * @param tag the tag to be added
   * @param value the value to be given to this tag
   */
  public void addLine(String tag, String value){
    addLine(makeline(openTags.size(), tag, value));
  }
  
  /**
   * Adds the next line to the document being built.
   * @param tag the tag to be added
   * @param attributes any attributes to be inserted; these are inserted verbatim
   * @param value the value to be given to this tag
   */
  public void addLine(String tag, String attributes, String value){
    addLine(makeline(openTags.size(), tag, attributes, value));
  }

  /**
   * Adds the next line to the document being built.
   * @param tag the tag to be added
   * @param value the value to be given to this tag
   */
  public void addLine(String tag, int value){
    addLine(makeline(openTags.size(), tag, value));
  }

  /**
   * Adds the next line to the document being built.
   * @param tag the tag to be added
   * @param attributes any attributes to be inserted; these are inserted verbatim
   * @param value the value to be given to this tag
   */
  public void addLine(String tag, String attributes, int value){
    addLine(makeline(openTags.size(), tag, attributes, value));
  }
  
  /**
   * Adds the next line to the document being built.
   * @param tag the tag to be added
   * @param value the value to be given to this tag
   */
  public void addLine(String tag, long value){
    addLine(makeline(openTags.size(), tag, value));
  }

  /**
   * Adds the next line to the document being built.
   * @param tag the tag to be added
   * @param attributes any attributes to be inserted; these are inserted verbatim
   * @param value the value to be given to this tag
   */
  public void addLine(String tag, String attributes, long value){
    addLine(makeline(openTags.size(), tag, attributes, value));
  }
  
  /**
   * Adds the next line to the document being built.
   * @param tag the tag to be added
   * @param value the value to be given to this tag
   */
  public void addLine(String tag, double value){
    addLine(makeline(openTags.size(), tag, value));
  }
  
  /**
   * Adds the next line to the document being built.
   * @param tag the tag to be added
   * @param attributes any attributes to be inserted; these are inserted verbatim
   * @param value the value to be given to this tag
   */
  public void addLine(String tag, String attributes, double value){
    addLine(makeline(openTags.size(), tag, attributes, value));
  }
  
  
  /* Indents are spaces, but this is the fastest way to do it */
  /**
   * Returns a String of spaces that serve as an indent,
   * given the number of tages that are open.
   * @param indentLevel the number of open tags, or the 'Indent Level'
   * @return a sequence of n spaces, where n is a multiple of the Indent Level
   */
  private static String getIndent(int indentLevel){
    switch(indentLevel){
      case 0:  return "";
      case 1:  return "  ";
      case 2:  return "    ";
      case 3:  return "      ";
      case 4:  return "        ";
      case 5:  return "          ";
      case 6:  return "            ";
      case 7:  return "              ";
      case 8:  return "                ";
      case 9:  return "                  ";
      case 10: return "                    ";      
    }
    // if it's bigger than 10, do it the long way
    String retVal = "";
    for(int i = 0; i < indentLevel; i++){
      retVal += "  ";
    }
    return retVal;
  }
  
   
}
