package abcm.web;

import abcm.db.AbstractABCMDBGateway;
import abcm.db.managers.ProjectResearcherManager;
import abcm.db.managers.ResearcherLoginsManager;

/**
 * Represents a session in the HWM website
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ABCMSession implements java.io.Serializable {
  
  /** Serial ID  */
  public static final long serialVersionUID =  296110818;
  
  /** Flag indicating whether the session has been authorized  */
  private boolean isAuthorized     = false;
  
  /** Flag indicating whether the current user is an administrator  */
  private boolean isAdmin          = false;
  /** Current user's username */
  private String  userName         = "";
  /** Current user's username */              // TODO Separate username and loginName variables; resolve any issues that fall from this
  private String  loginName        = "";
  /** Current user's ID  */
  private long    userID           = 0;
  /** Most recent error encountered by this session  */
  private String  errorString      = ""; 
  
  /** Access point to database  */
  private AbstractABCMDBGateway DB;

  /**
   * Constructor
   * @param db Instance of an ABCM Gateway to provide database access.
   */
  public ABCMSession(AbstractABCMDBGateway db){
    DB = db;
  }

  /**
   * Logs this session in. If the username and password are valid,
   * sets the blnIsAuthorized variable to true.
   * @param username Username
   * @param password Password
   * @param createdAt time at which this session was created, in millis
   */
  public void setLogin(String username, String password, long createdAt){
    loginName        = username;
    userName         = username;
    isAuthorized     = false;
    isAdmin          = false;
    errorString      = "";
    
    ProjectResearcherManager prm = new ProjectResearcherManager(DB);
    if(prm.confirmLogin(username, password)){
      long ID      = prm.getPK(username);
      userID       = prm.getResearcherID(ID);
      userName     = prm.getName(ID); // This converts the username as entered into the one that is stored (i.e. JOhnMUrphy -> JohnMurphy)
      isAuthorized = true;
      isAdmin      = prm.isAdmin(username);
      
      ResearcherLoginsManager rlm = new ResearcherLoginsManager(DB);
      
      rlm.setInstanceToSave(userID); // Note: Saves the user ID, not the project researcher ID; this is a debatable choice...
      rlm.saveInstanceToDB();
      
    }
    else{
      userID           = 0;
      userName         = "";
      isAuthorized     = false;
      errorString      = prm.getErrorReport() + "\nDB: " + DB.getErrorReport();      
    }

  }
    
  /**
   * Returns true if the session is authorized; false if not
   * @return true if the session is authorized; false if not
   */
  public boolean IsAuthorized (){      return isAuthorized;  }
  
  /**
   * Returns true if the person logged in has administrator status
   * Note: does not re-check status, so if status has changed since
   * login, this will not reflect that change.
   * @return true if the person logged in has administrator status
   */
  public boolean IsAdministrator(){    return isAdmin;       }
    
  /**
   * Returns the database ID by which this user is identified
   * @return the database ID by which this user is identified
   */
  public long getUserID(){             return userID;           }
  
  /**
   * Gets a String indicating the most recent error that has occurred
   * in methods of this object
   * @return a String error message
   */
  public String getErrorString(){      return (((errorString != null) && (errorString.length() > 0)) ? "SESSION ERROR STRING " + errorString : "");        }
  
  /**
   * Gets the username of the person logged in on this session.
   * If noone is logged in, returns "Anonymous".
   * @return current Username
   */
  public String getUsername(){         return (isAuthorized ? userName: "Anonymous"); } 
  
  /**
   * Gets the login name of the person logged in on this session.
   * If no individual is logged in, returns "Anonymous".
   * @return the current login name
   */
  public String getLoginName(){        return (isAuthorized ? loginName: "Anonymous");   }

}