package abcm.web;


import java.io.File;
import java.io.FileOutputStream;

import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;

/**
 * Creates and saves an SVG file.
 * 
 * In keeping with the SVG format the file will have two scales that should be kept in mind: the first
 * is the size of the image in pixels, and the second is the intended display size in physical units
 * like inches (which are used here). One purpose of the viewer is to reconcile these two, so that
 * text that is to be displayed in the image appears at a reasonable size when the image is reproduced
 * at the intended physical size.
 *
 * This class is fairly brittle: it does not include even simple validation, so invoking a setter
 * for, say, page width, where the value passed is invalid (for example, negative), will not be
 * trapped and may cause unpredictable results.
 * 
 * There is a set of variables with a 'map' prefix that are used specifically for creating maps
 * in a standard ABCM format; these can be used to adjust the range, in x,y coordinates, that
 * the map displays. They are settable through a single method that does provide
 * some minor validation (so mins are less than maxes).
 * 
 * Most of the instance variables and their getters and setters are considered self-explanatory
 * and are not JavaDoc documented. Note that variables with the 'map' prefix
 * are used in placing and depicting a spatial map with a restricted range.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SVGViewer {

  /* Page, Margin, and Border Defaults */
  public static final double DEFAULT_PAGE_WIDTH                  =  11.0;
  public static final double DEFAULT_PAGE_HEIGHT                 =   8.5;
  public static final int    DEFAULT_PAGE_X_SIZE_IN_PIXELS       = 900;
  public static final double DEFAULT_TOP_MARGIN                  =   0.25;
  public static final double DEFAULT_BOTTOM_MARGIN               =   0.25;
  public static final double DEFAULT_LEFT_MARGIN                 =   0.25;
  public static final double DEFAULT_RIGHT_MARGIN                =   0.25;
  public static final double DEFAULT_BORDER_WIDTH_IN_INCHES      =   0.1;
  public static final double MIN_PAGE_HEIGHT                     =   7.0;
  public static final double MIN_PAGE_WIDTH                      =   7.0;

  /** The value for page height in pixels is derived proportionally from the defaults for width and height in inches and width in pixels  */
  public static final int    DEFAULT_PAGE_Y_SIZE_IN_PIXELS       = ((int)Math.floor((DEFAULT_PAGE_X_SIZE_IN_PIXELS / DEFAULT_PAGE_WIDTH) * DEFAULT_PAGE_HEIGHT));
  
  
  /* Instance Variables */
  private double topMargin           = DEFAULT_TOP_MARGIN;
  private double bottomMargin        = DEFAULT_BOTTOM_MARGIN;
  private double leftMargin          = DEFAULT_LEFT_MARGIN;
  private double rightMargin         = DEFAULT_RIGHT_MARGIN;
  private double pageWidthInInches   = DEFAULT_PAGE_WIDTH;
  private double pageHeightInInches  = DEFAULT_PAGE_HEIGHT;
  private int    pageXSizeInPixels   = DEFAULT_PAGE_X_SIZE_IN_PIXELS;
  private int    pageYSizeInPixels   = DEFAULT_PAGE_Y_SIZE_IN_PIXELS;
  private double borderWidthInInches = DEFAULT_BORDER_WIDTH_IN_INCHES;
  

 /* MAP ONLY */  
  
  private double mapXRange;
  private double mapYRange;
  
  private double mapXMin;
  private double mapXMax;
  private double mapYMin;
  private double mapYMax;
  
  private double mapXBuffer;
  private double mapXLowerBound;
  private double mapYBuffer;
  private double mapYLowerBound;
   
  
  /* Constructor (none) */
  // public SVGViewer(){ }
  
  
  /* Getters and Setters */
  public double getBottomMargin()       { return bottomMargin;        }
  public double getLeftMargin()         { return leftMargin;          }
  public double getPageHeightInInches() { return pageHeightInInches;  }
  public double getPageWidthInInches()  { return pageWidthInInches;   }
  public double getRightMargin()        { return rightMargin;         }
  public double getTopMargin()          { return topMargin;           }
  public int    getPageYSizeInPixels()  { return pageYSizeInPixels;   }
  public int    getPageXSizeInPixels()  { return pageXSizeInPixels;   }
  public double getBorderWidthInInches(){ return borderWidthInInches; }
    

  public void setPageHeightInInches(double newPageHeightInInches) 
     { pageHeightInInches = Math.max(newPageHeightInInches, MIN_PAGE_HEIGHT); }
  
  public void setPageWidthInInches(double newPageWidthInInches) 
     { pageWidthInInches  = Math.max(newPageWidthInInches, MIN_PAGE_WIDTH  ); }
  
  public void setBottomMargin(double newBottomMargin)              {  bottomMargin        = newBottomMargin;        }
  public void setLeftMargin(double newLeftMargin)                  {  leftMargin          = newLeftMargin;          }
  public void setRightMargin(double newRightMargin)                {  rightMargin         = newRightMargin;         }
  public void setTopMargin(double newTopMargin)                    {  topMargin           = newTopMargin;           }
  public void setPageXSizeInPixels(int newPageXSizeInPixels)       {  pageXSizeInPixels   = newPageXSizeInPixels;   }
  public void setPageYSizeInPixels(int newPageYSizeInPixels)       {  pageYSizeInPixels   = newPageYSizeInPixels;   }
  public void setBorderWidthInInches(double newBorderWidthInInches){  borderWidthInInches = newBorderWidthInInches; }

  /**
   * Sets all four margins (top, bottom, left, right) to the new value
   * @param newMargin
   */
  public void setMargin(double newMargin){
    setBottomMargin(newMargin);
    setTopMargin(newMargin);
    setLeftMargin(newMargin);
    setRightMargin(newMargin);
  }
  
  public void setPageWidthAndHeightInInches(double newPageWidth, double newPageHeight){
    setPageWidthInInches(newPageWidth);
    setPageHeightInInches(newPageHeight);
  }
  
  public void setPageWidthAndHeightInPixels(int newPageWidth, int newPageHeight){
    setPageXSizeInPixels(newPageWidth);
    setPageYSizeInPixels(newPageHeight);
  }
  
  
  /**
   * Creates a representation of the specified object. The SVG image will
   * be sized according to the settings of this object (page height, etc.)
   * Returns the name of the svg file.
   * @param svg_objects list of objects to be represented
   * @param imageOutputLocation path to the output location to which the file will be saved
   * @return the name of the file created
   */
  public String getSVG(SVGRepresentable[] svg_objects, String imageOutputLocation){
    return getSVG(svg_objects, imageOutputLocation, "");
  } 
  
  /**
   * Creates a representation of the specified object. The SVG image will
   * be sized according to the settings of this object (page height, etc.)
   * Returns the name of the svg file.
   * @param svg_objects list of objects to be represented
   * @param imageOutputLocation path to the output location to which the file will be saved
   * @param filePrefix prefix to be prepended to the file name
   * @return the name of the file created
   */
  public String getSVG(SVGRepresentable[] svg_objects, String imageOutputLocation, String filePrefix){
    String retVal = "unsuccessful";
    // Create a file and write the svg to it
    try{
      File svgOut = File.createTempFile(filePrefix + "_map",".svg", new File(imageOutputLocation));
//      svgOut.deleteOnExit();        // TODO Restore 'delete on exit' for map files
      FileOutputStream svgOutput = new FileOutputStream(svgOut.getAbsoluteFile());
      SVGAssist svg = new SVGAssist(pageXSizeInPixels, pageYSizeInPixels, 
          pageWidthInInches, pageHeightInInches);
      for(int i=0;i < svg_objects.length; i++){    svg.addSVGObject(svg_objects[i]);    }
      svgOutput.write(svg.getSVG());
      svgOutput.close();
      
      // Returns the name of the file created
      retVal = svgOut.getName();
    }
    catch(Exception e){
      System.out.println(e.getMessage());
      e.printStackTrace();
      return "Could not create image";
    }
    return retVal;
  }
 
  
  /**
   * Sets the range to be displayed in a map
   * @param min_X minimum x coordinate
   * @param min_Y minimum y coordinate
   * @param max_X maximum x coordinate
   * @param max_Y maximum y coordinate
   * @return true if successful, false otherwise; will be unsuccessful if a minimum is greater than the corresponding maximum
   */
  public boolean setRange(double min_X, double min_Y, double max_X, double max_Y){
    /* Trap issues */
    if((min_X >= max_X) ||
       (min_Y >= max_Y)    ) return false;
    
    // If all is OK, set the variables. This is the only entry
    // portal to re-setting these variables, so it memoizes a bunch of them
    mapXMin = min_X;
    mapYMin = min_Y;
    mapXMax = max_X;
    mapYMax = max_Y;
    
    mapXRange     = max_X - min_X;
    mapYRange     = max_Y - min_Y;
    
    // Add some padding
    mapXBuffer     = (mapXRange < 100d ? 10d : mapXRange/10d);
    mapXLowerBound = min_X - mapXBuffer;
    mapYBuffer     = (mapYRange < 100d ? 10d : mapYRange/10d);
    mapYLowerBound = min_Y - mapYBuffer;
 
    mapXRange += (2d*mapXBuffer);
    mapYRange += (2d*mapYBuffer);
    
    return true;
  }
  
}
