package flowag.basic;

import java.util.Collection;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.sim.AbstractModelComponent;

/**
 * An Environment object represents the natural environment in which
 * plants grow. The Environment object provides basic information 
 * such as temperature and daily sunlight hours that can be used in 
 * the plant productivity routines.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Environment extends AbstractModelComponent implements ABCMDeepCloneable{

  /**
   * Default value for mean percent annual daytime hours (@value)
   */
  public static final double DEFAULT_MEAN_PCT_ANNUAL_DAYTIME_HOURS   =  0.27;  // Value at equator (one day = 12 hours / 12 * 365 = 1/365 = .0027, or .27%)
  
  /**
   * Default value for mean daily temperature (@value C)
   */
  public static final double DEFAULT_MEAN_DAILY_TEMPERATURE          =  20.0;  // Baseline value 

  
  /** Switch indicating no rainfall will occur automatically   */
  public static final int NO_RAINFALL = 0;
  /** Switch turning on basic rainfall engine. */
  public static final int BASIC_RAINFALL = 1;
  
  /** Lower altitude for lapse rate temperature calculations                */ 
  protected double lowerAltitude;
  /** Temperature at lower altitude for lapse rate temperature calculations */ 
  protected double lowerAltitudeTemp;
  /** Upper altitude for lapse rate temperature calculations                */ 
  protected double upperAltitude;
  /** Temperature at upper altitude for lapse rate temperature calculations */ 
  protected double upperAltitudeTemp;
  /** Calculated value for lapse rate                                       */ 
  private double   lapseRate;
  
  /** Probability of rainfall occurring (should be between 0 and 1, inclusive, but this is not validated) */ 
  protected double basicDailyRainfallProbability;
  /** Maximum of range of possible rainfall from basic rainfall algorithm, in mm */ 
  protected double basicDailyRainfallMaxAmount;
  /** Minimum of range of possible rainfall from basic rainfall algorithm, in mm */ 
  protected double basicDailyRainfallMinAmount;
  
  /** Template for water arriving on fields via rainfall; can be used to set chemical concentration.  */
  protected Water rainSource = new Water(1);
  
  /**
   * Gets the current probability of rain occurring
   * @return the current probability of rainfall (range: 0-1)
   */
  public double getBasicDailyRainfallProbability(){     return basicDailyRainfallProbability;   }

  /**
   * Sets the probability of rain occuring. Does not perform validation.
   * @param basicDailyRainfallProbability new value for probability of rainfall
   */
  public void setBasicDailyRainfallProbability(double basicDailyRainfallProbability){
    this.basicDailyRainfallProbability = basicDailyRainfallProbability;
  }

  /**
   * Gets the current daily maximum rainfall amount
   * @return the current daily maximum rainfall amount
   */
  public double getBasicDailyRainfallMaxAmount(){       return basicDailyRainfallMaxAmount;     }

  /**
   * Set the rainfall engine's basic daily maximum amount.
   * @param basicDailyRainfallMaxAmount new value for maximum amount
   */public void setBasicDailyRainfallMaxAmount(double basicDailyRainfallMaxAmount){ 
    this.basicDailyRainfallMaxAmount = basicDailyRainfallMaxAmount;
  }

  /**
   * Gets the current daily minimum rainfall amount
   * @return the current daily minimum rainfall amount
   */
  public double getBasicDailyRainfallMinAmount(){       return basicDailyRainfallMinAmount;     }

  /**
   * Set the rainfall engine's basic daily minimum amount.
   * @param basicDailyRainfallMinAmount new value for minimum amount
   */
  public void setBasicDailyRainfallMinAmount(double basicDailyRainfallMinAmount) {
    this.basicDailyRainfallMinAmount = basicDailyRainfallMinAmount;
  }
  
  /**
   * Sets the rainfall watersource, which will determine chemical content in rainfall (if any)
   * @param newSource new water object to be used as rainfall watersource
   */
  public void setRainSource(Water newSource){ rainSource = newSource; }

  
  
  /** Mean percentage of the Annual Daytime hours that are 
   *  currently occurring on days in this environment. 
   *  This is used in the Blaney-Criddle calculation of
   *  Evapo-transpiration rate                              */  
  private double meanPctAnnualDaytimeHours;   
  
  /** Mean daily temperature. This is used in the 
   * Blaney-Criddle calculation of Evapo-transpiration rate */  
  private double meanDailyTemperature;
  
  /**
   * Generic constructor. Sets the values to their defaults.
   */
  public Environment(){
    meanPctAnnualDaytimeHours = DEFAULT_MEAN_PCT_ANNUAL_DAYTIME_HOURS;
    meanDailyTemperature      = DEFAULT_MEAN_DAILY_TEMPERATURE;
    lowerAltitudeTemp         = DEFAULT_MEAN_DAILY_TEMPERATURE;
  }
  
  /* Temperature */
  
  /**
   * Sets the mean daily temperature.
   * @param temp The meanDailyTemperature to set.
   */
  public void setMeanDailyTemperature(double temp) {    
    meanDailyTemperature = temp;
    setMeanDailyTempByAltitude(0, temp, 1000, temp);
  }
  
  
  /**
   * Sets the mean daily temperature at a pair of altitudes. This allows
   * calculation of a lapse rate and thus calculation of a temperature
   * (linear gradient) at any altitude.
   * @param lowerAlt lower altitude
   * @param lowerAltTemp temperature at lower altitude
   * @param upperAlt upper altitude
   * @param upperAltTemp temperature at upper altitude
   */
  public void setMeanDailyTempByAltitude(double lowerAlt, double lowerAltTemp, double upperAlt, double upperAltTemp){
    lowerAltitude        = lowerAlt;
    lowerAltitudeTemp    = lowerAltTemp;
    upperAltitude        = upperAlt;
    upperAltitudeTemp    = upperAltTemp;
    lapseRate            = (upperAltitude != lowerAltitude ? (lowerAltitudeTemp - upperAltitudeTemp)/ (upperAltitude - lowerAltitude) : 0);
    meanDailyTemperature = lowerAltitudeTemp;
  }
  
  
  /**
   * Gets the mean daily temperature.
   * @return the meanDailyTemperature.
   */
  public double getMeanDailyTemperature(){  
    return meanDailyTemperature;   
  }
  
  /**
   * Gets the mean daily temperature at the specified altitude. Calculated using lapse
   * rate.
   * @param altitude the altitude requested
   * @return the mean daily temperature at that altitude.
   */
  public double getMeanDailyTempByAltitude(double altitude){
    return lowerAltitudeTemp - (altitude - lowerAltitude) * lapseRate;  
  }
  
  
  /* Daylight Hours*/
  
  /**
   * Gets the mean percentage of annual daytime hours.
   * @return the meanPctAnnualDaytimeHours.
   */  
  public double getMeanPctAnnualDaytimeHours() {   
    return meanPctAnnualDaytimeHours;  
  }
  
  /**
   * Sets the mean percentage of annual daytime hours.
   * @param hours The meanPctAnnualDaytimeHours to set.
   */
  public void setMeanPctAnnualDaytimeHours(double hours) {    
    meanPctAnnualDaytimeHours = hours;   
  }

  
  /**
   * Causes rain to fall on all rainfall recipient objects passed in the collection.
   * Amount of rainfall is determined by the rainfall engine.
   * @param rainRecipients objects to receive rainfall
   */
  public void rain(Collection rainRecipients){
    switch (codeModelOption("RAINFALL_PROCEDURE")){
      case BASIC_RAINFALL:
        if(random() < this.basicDailyRainfallProbability){
          double amount = Math.max(this.basicDailyRainfallMinAmount, 0) + random() * (Math.min((this.basicDailyRainfallMaxAmount - this.basicDailyRainfallMinAmount), 0));
          for(Object rainRecipient: rainRecipients) // Much of this can be eliminated with generics...
              if(rainRecipient instanceof RainRecipient) ((RainRecipient)rainRecipient).rain(amount, rainSource);
        }
        break;      
      case NO_RAINFALL:
      default:
        break;
    }
    
  }

  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Environment newEnvironment = new Environment();
    cloneStore.put(this, newEnvironment);
    newEnvironment.meanDailyTemperature = this.meanDailyTemperature;
    newEnvironment.lowerAltitude        = this.lowerAltitude;
    newEnvironment.upperAltitude        = this.upperAltitude;
    newEnvironment.upperAltitudeTemp    = this.upperAltitudeTemp;
    newEnvironment.lapseRate            = this.lapseRate;
    newEnvironment.meanPctAnnualDaytimeHours = this.meanPctAnnualDaytimeHours;
    return newEnvironment;
  }
}
