package flowag.basic;

/**
 * Provides simple unit conversion methods; for example, million acre-feet to cubic meters.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class UnitConversion {

  /**
   * Private constructor to prevent instantiation
   */
  private UnitConversion(){}
  
  /**
   * Converts a value in million acre-feet to cubic meters
   * @param maf amount to be converted
   * @return equivalent in cubic meters
   */
  public static double millionAcreFeetToCubicMeters(double maf){
    return maf * 1000000d * 1233.487661d; 
  }
  
  /**
   * Converts a value from cubic meters to million acre-feet
   * @param m cubic meter value
   * @return equivalent value in million acre-feet
   */
  public static double cubicMeterToMillionAcreFeet(double m){
    return m * (.000000000810713192178);
  }
  
  /**
   * Converts a value of million acre-feet per day to cubic meters per second.
   * @param mafPerDay amount to be converted
   * @return equivalent amount in cubic meters per second
   */
  public static double mafPerDayToCubicMetersPerSecond(double mafPerDay){
    return millionAcreFeetToCubicMeters(mafPerDay) / (24d * 60d * 60d);
  }
  
  /**
   * Converts a value of million acre-feet per year to cubic meters per second.
   * @param mafPerYear amount to be converted
   * @return equivalent amount in cubic meters per second
   */
  public static double mafPerYearToCubicMetersPerSecond(double mafPerYear){
    return mafPerDayToCubicMetersPerSecond(mafPerYear / 365d);
  }
  
  /**
   * Converts a flow value in cubic meters per second to million
   * acre-feet per day
   * @param Q the initial flow value
   * @return the converted value
   */
  public static double cubicMetersPerSecondToMillionAcreFeetPerDay(double Q){
    return cubicMeterToMillionAcreFeet(Q) * (24 * 60 * 60);
  }
  
  /**
   * Converts a time expressed in two parts into a single time. The original time is expressed
   * as a base time that is a fraction of a day (assumed to be between zero and one) and 
   * a number of seconds that are to be added to the base time. The procedure converts the
   * seconds to a fraction of a day (by dividing by 24 * 60 * 60) and adds this to the
   * base time. The original values are not validated, but if the result is below zero, zero is
   * returned, and if it is above one, one is returned. Otherwise the value will be a figure
   * between zero and one that represents the point in time as a fraction of the day.
   * @param baseTimeAsFractionOfDay the base of the time to be converted
   * @param intervalAfterBaseTimeInSeconds the additional interval to be added to the base time
   * @return the converted time
   */
  public static double convertTime(double baseTimeAsFractionOfDay, double intervalAfterBaseTimeInSeconds){
    return Math.max(0, Math.min(1, baseTimeAsFractionOfDay + ((intervalAfterBaseTimeInSeconds)/(60 * 60 * 24))));
  }
  
  /**
   * Converts a value in cubic meters into one in cubic feet
   * @param valueInCubicMeters original value
   * @return new value
   */
  public static double cubicMetersToCubicFeet(double valueInCubicMeters){
    return valueInCubicMeters * 35.3146667d;
  }
  
}
