package flowag.basic;

import java.util.Enumeration;
import java.util.Hashtable;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.sim.AbstractModelComponent;

/**
 * A Water Object represents a volume of water. It may contain various
 * chemicals in varying concentrations.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Water extends AbstractModelComponent implements ABCMDeepCloneable{

  /** Chemicals contained in this Water object */ private ABCMDeepCloneableHashtable chemicals = new ABCMDeepCloneableHashtable();
  /** Volume, in cubic meters                  */ private double volume;

  /**
   * Constructs a water object with the specified volume
   * @param newVolume volume of the new water object
   */
  public Water(double newVolume){
    volume = newVolume;
  }
  
  /**
   * Creates a new volume of water with the given volume but with
   * a collection of chemicals in the same proportions as the water
   * sample provided
   * @param newVolume The volume of the new instance of the water object
   * that is being created
   * @param sample A sample water object with chemicals in the same
   * proportions.
   */
  public Water(double newVolume, Water sample){
    volume = newVolume;
    chemicals = (ABCMDeepCloneableHashtable)sample.chemicals.getDeepClone(new ABCMCloneStore());    
  }
  
  /**
   * Evaporates a given fraction of the water. This will
   * increase the concentration of the nutrients
   * @param fraction The fraction of the water to be evaporated off
   */
  public void evaporate(double fraction){
    Chemical c;
    if ((fraction > 0) && (fraction <= 1)){
      volume *= (1d - fraction);
      // The concentration of each chemical will increase
      // as the volume of solvent diminishes
      Enumeration E = chemicals.elements();
      while(E.hasMoreElements()){
        c = (Chemical)E.nextElement();
        double oldCon = c.getConcentration();
        double solventLoss = (1d - fraction);
        double newCon = oldCon / solventLoss;
        c.setConcentration(newCon);
      }
    }
  }
  
  /**
   * Gets the concentration of a specific chemical.
   * @param chemName Name of the chemical whose concentration is requested
   * @return Returns the concentration of the chemical with the name
   * passed as argument.
   */
  public double getConcentration(String chemName){
    double conctr = 0;
    if (chemicals.containsKey(chemName)){
      conctr = ((Chemical)chemicals.get(chemName)).getConcentration();
    }
    return conctr;
  }
  
  /**
   * Allows the user to add a chemical to this instance of the water
   * object. If the chemical already exists, its concentration is increased.
   * If not, it is added anew.
   * @param c Chemical object to be added
   */
  public void addChemical(Chemical c){
    String chemName = c.getName();
    if(chemicals.containsKey(chemName)){
      Chemical C = (Chemical)chemicals.get(chemName);
      C.addNew(c);
    }
    else{
      chemicals.put(chemName, c);
    }
  }
  
  /**
   * Gets the volume of this water object
   * @return Returns the volume of this instance of the water object
   */
  public double getVolume(){
    return volume;
  }
  
  // TODO Water.getChemicals returns a clone of the hashtable, but is it a deep clone?
  
  /**
   * Gets the hashtable listing the chemicals found in this
   * water object.
   * @return The hashtable list of chemicals found in this
   * water object
   */
  public Hashtable getChemicals(){
    return (Hashtable)chemicals.clone();
//    Hashtable H = new Hashtable();
//    Enumeration E = chemicals.keys();
//    while(E.hasMoreElements()){
//      String k = (String)E.nextElement();
//      double c = ((Chemical)chemicals.get(k)).getConcentration();
//      H.put(k, new BasicChemical(k, c));
//    }
//    return H;
  }
  
  /**
   * Adds a different instance of a water object to this one. 
   * @param w the water object to be added to this one
   * @return the new volume of the water object
   */
  public double mixIn(Water w){
    double newVolume = this.volume + w.volume;
    // And adjust the concentrations of any chemical already in this instance
    // but not in the other- dilution
    Enumeration myChems = chemicals.keys();
    while(myChems.hasMoreElements()){
      String key = (String)myChems.nextElement();
      if(!w.getChemicals().containsKey(key)){
        Chemical C = ((Chemical)chemicals.get(key));
        C.setConcentration((C.getConcentration() * volume) / newVolume);
      }
    }
    
    
    Enumeration E = w.getChemicals().elements();
    while(E.hasMoreElements()){
      Chemical C = (Chemical)E.nextElement();
      if(chemicals.containsKey(C.getName())){
        // Must adjust concentration based on comparative volume
        Chemical extantChemical = (Chemical)chemicals.get(C.getName());
        extantChemical.setConcentration(((this.volume * extantChemical.getConcentration()) + (w.volume * C.getConcentration())) / newVolume);
      }
      else{
        chemicals.put(C.getName(), C);  
      }
    }
    
    // Finally, actually set the volume for this instance
    volume += w.getVolume();
    
    // And clear the instance added
    w.clear();
    
    // Return the new volume
    return volume;
  }
  
  /**
   * Remove a portion of the water from this water object.
   * Returns a new water object with the specified volume.
   * This instance is reduced in volume by the volume
   * of the new water object returned. If the volume to remove
   * is larger than the current volume, the entire volume
   * is returned and this instance is set to volume zero.
   * The concentrations of all chemicals remain unchanged.
   * @param volumeToRemove The volume to be removed from this instance
   * @return A new water object with the specified volume.
   */
  public Water removePortion(double volumeToRemove){
    Water w = new Water(Math.min(volume, volumeToRemove), this);
    volume = Math.max(volume - volumeToRemove,0);
    return w;
  }

  /**
   * Allows 'destroying' this object without requiring that a new one be created.
   * @return true
   */
  public boolean clear(){
    if(chemicals != null) chemicals.clear();
    volume = 0;
    return true;
  }

  // TODO Water.removeChemical returns false if the chemical is not present- but should it? 
  
  /**
   * Remove a particular chemical from this sample of water.
   * @param chemicalName
   * @return true if successful, false if not (i.e. if the chemical is not
   * present in this water to begin with).
   */
  public boolean removeChemical(String chemicalName){
    if(!chemicals.containsKey(chemicalName)) return false;
    chemicals.remove(chemicalName);
    return true;
  }
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Water newWater = new Water(volume);
    newWater.chemicals = (ABCMDeepCloneableHashtable)chemicals.getDeepClone(cloneStore);
    return newWater;
  }
  
  /**
   * Gets a report describing this water object
   * @return the requested report
   */
  public String report(){
    String rep = "";
    rep += "Volume: " + this.getVolume() + "\n";
    Enumeration E = chemicals.elements();
    while(E.hasMoreElements()){
      Chemical c = (Chemical)E.nextElement();
      rep += "  " + c.getName() + ": " + c.getConcentration() + " ppm\n";
    }
    return rep;
  }
}
