package flowag.bio;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Enumeration;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.math.Geometry;
import abcm.math.Point;
import abcm.math.SimDate;
import abcm.sim.AbstractModelComponent;
import flowag.basic.Water;


/**
 * A specific instance of a single kind of plant being grown in a field.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Crop extends AbstractModelComponent implements ABCMDeepCloneable{

  /** Field in which this crop grows                          */ protected Field                             field;
  /** Plant object represented by this crop                   */ protected Plant                             plant;            
  /** Date on which plot was planted                          */ protected SimDate                           datePlanted;
  /** Age of this crop                                        */ protected int                               cropAge;
  /** Current stage of plant growth for this crop             */ protected int                               currentStage; 
  /** Days the crop has been in its current growth stage      */ protected int                               timeInCurrentStage;
  /** Density at which this crop was planted in the field     */ protected double                            density;
  /** Life span of this crop (from Plant characteristics      */ protected int                               lifeSpan;
  /** Is this crop alive?                                     */ protected boolean                           alive;
  /** Number of consecutive days water need has not been met  */ protected long                              consecutiveDaysWaterNeedNotMet;

  /** Drought sensitivity of this plant during the current stage */ protected double droughtSensitivityOfCurrentStage;
  /** Duration of the current stage                              */ protected int    durationOfCurrentStage;
  /** Crop Factor during the current stage                       */ protected double cropFactorOfCurrentStage;
  
  /** Record of the maximum number of consecutive days water 
   * available has been short of water needed 
   */ 
  protected long   maxConsecutiveDaysWaterNeedNotMet;
  /** Total water received during this crop's growth */ 
  protected double totalActualWater;
  /** Total of water need for all days (so far) during this plant's growth */ 
  protected double totalWaterNeed;
  /** Total of minimum water needed by this plant for all days (so far) during growth */ 
  protected double totalMinimumWater;

  /**
   * Value by which harvest should be adjusted based on chemical needs 
   */
  protected double chemicalNeedAdjustment;
  
  /**
   * Flag indicating that this crop should be removed from the field (i.e. because it has died) 
   */
  protected boolean remove; // Flag that indicates that this crop should be automatically removed

  /** Code Model Switch: Production Model */            
      private static final int PRODUCTION_MODEL_1 = 1;
  
  /** Limit on the number of days a plant can live without water */ 
      private static final int DAYS_WITHOUT_WATER_LIMIT = 7;

  // Code Model Switches
  /** Code Model Switch: No Plant Growth           */ public static final int NO_PLANT_GROWTH            = 0;
  /** Code Model Switch: Basic Plant Growth        */ public static final int PLANT_GROWTH_MODEL_1       = 1;
  /** Code Model Switch: Second Basic Plant Growth */ public static final int PLANT_GROWTH_MODEL_2       = 2;


  /**
   * Constructor; creates a new crop object of a certain
   * kind of plant, planted on a certain date.
   * @param f Field in which this crop is placed
   * @param p Plant of which this crop consists
   * @param Density the density of this plant in this field
   * @param sd simulation date on which this crop was planted
   */
  public Crop(Field f, Plant p, double Density, SimDate sd){
    field                            = f;
    plant                            = p;
    density                          = Density;
    datePlanted                      = new SimDate(sd.value());
    lifeSpan                         = p.getLifeSpan();
    alive                            = true;
    currentStage                     = 1;
    cropAge                          = 0;
    timeInCurrentStage               = 0;
    droughtSensitivityOfCurrentStage = p.getDroughtSensitivityOfStage(currentStage);
    durationOfCurrentStage           = p.getDurationOfStage(currentStage);
    cropFactorOfCurrentStage         = p.getCropFactorOfStage(currentStage);
    chemicalNeedAdjustment           = 1;
  }

  /* Getters */  
  
  /**
   * Gets the density at which this crop was planted
   * @return the density, as a fraction between 0 and 1.
   */
  public double getDensity(){    return density;     }

  
  /**
   * Gets a boolean result indicating whether this plant
   * is alive or dead.
   * @return Returns the status of this plant as alive or dead.
   */
  public boolean isAlive() {     return alive;       }

  /**
   * Get the 'maturity' of this plant- the ratio of how
   * long it has lived to its lifespan. Note that if
   * lifespan is zero (which should never happen),
   * value returned is zero.
   * @return the current maturity level of this plant.
   */
  public double getMaturity(){   return (lifeSpan != 0 ? ((double)cropAge) / ((double)lifeSpan) : 0);   }

  
  /**
   * Gets the yield potential for this crop.
   * 
   * The yield potential is the yield that this crop
   * can provide if it is nurtured under ideal conditions
   * for the remainder of its life, given whatever damage
   * it has already suffered.
   * @return the yield potential for this crop
   */
  public double getPotentialYield(){
    
    double overallDroughtSensitivity = plant.getDroughtSensitivity();
    long daysWithoutWaterPermitted = (long)Math.floor(DAYS_WITHOUT_WATER_LIMIT * (1 - overallDroughtSensitivity));
    if(maxConsecutiveDaysWaterNeedNotMet > (daysWithoutWaterPermitted + Crop.DAYS_WITHOUT_WATER_LIMIT)) return 0;
    if(cropAge > (lifeSpan * 1.2d))                                         return 0;
    
    double baseYield = 1; // Assume plant is perfectly healthy
    if(totalMinimumWater == 0) return 1; // Should only be on first day...
    // First adjust based on total water received.
    // If the water received is above the minimum needed
    if(totalActualWater > totalMinimumWater){
      double maxAdjust = Math.sqrt(overallDroughtSensitivity * 10d)/10d; // So, .9 becomes .3, .4 becomes .2
      double fraction = (totalWaterNeed - totalActualWater) / (totalWaterNeed - totalMinimumWater);
      baseYield -= maxAdjust * fraction;
    }
    // Otherwise, the water received is below the minimum needed; this hits harder
    else{
      double minAdjust = Math.sqrt(overallDroughtSensitivity * 10d)/10d; // So, .9 becomes .3, .4 becomes .2
      baseYield -= minAdjust;
      double fraction = totalActualWater/ totalMinimumWater;
      baseYield *= fraction;
    }
    
    
    // Now adjust for consecutive days missing need
    double adjust = 1;
    if(maxConsecutiveDaysWaterNeedNotMet > daysWithoutWaterPermitted){
      long daysOverLimit = maxConsecutiveDaysWaterNeedNotMet - daysWithoutWaterPermitted;
      adjust = 1 - daysOverLimit / DAYS_WITHOUT_WATER_LIMIT; // Note: will not be > 1 due to initial check (above)
    }
    baseYield *= adjust;
    baseYield *= chemicalNeedAdjustment;
    if(Double.isNaN(baseYield)) System.err.println("BASE YIELD IS NaN!!");
    // Finally, adjust for time after crop life span but before harvest
    if(cropAge <= lifeSpan * 1.1d) return baseYield; // Ten percent of growing season is allowed for harvest
    if(cropAge >= lifeSpan * 1.2d) return 0;         // But twenty percent means all crop is lost- must harvest within that time!
    double daysIntoDecline = cropAge - (lifeSpan * 1.1d);
    double daysOfDecline = (lifeSpan * 1.2d) - (lifeSpan * 1.1d);
    double fraction = daysIntoDecline / daysOfDecline;
    
    return baseYield * (1 - fraction); // Drop linearly from 100% to 0& between 10% and 20% past lifespan    
  }
  
  /**
   * Gets the yield if this crop were harvested today. May be zero if the crop is not yet
   * mature.
   * @return yield if crop is available immediately.
   */
  public double getHarvestableYield(){
    return (cropAge < lifeSpan ? 0 : getPotentialYield());
  }
  
  
  /* Actions (Growth) */
  
  /**
   * Grows this crop one day's worth, given the current
   * soil conditions.
   * 
   * Note that this may alter the soil conditions as
   * water/nutrients are extracted or emitted
   */
  public void step(){
//    System.out.println("CROP STEP; " + codeModelOption("PLANT_PRODUCTION"));
    switch(codeModelOption("PLANT_PRODUCTION")){
      case PLANT_GROWTH_MODEL_1:
//        System.out.println("\n\nRUNNING PLANT GROWTH MODEL 1 FOR CROP " + plant.getName() + " current age = " + cropAge + " lifespan = "  + lifeSpan);
        cropAge++;                                                               // Increase the crop's age
        if(cropAge < lifeSpan){                                                  // Once lifespan reached, plant processes end
          timeInCurrentStage++;                                                    // And the time in the current stage
          // Determine if it's time to move to the next stage
          if(timeInCurrentStage > durationOfCurrentStage){
            currentStage++;
            timeInCurrentStage               = 0;
            droughtSensitivityOfCurrentStage = plant.getDroughtSensitivityOfStage(currentStage);
            durationOfCurrentStage           = plant.getDurationOfStage(currentStage);
            cropFactorOfCurrentStage         = plant.getCropFactorOfStage(currentStage);
          }
          
          double waterNeed         = field.getETo() * cropFactorOfCurrentStage * density;
          double minimumWaterNeed  = waterNeed * droughtSensitivityOfCurrentStage;
          double waterAvailable    = field.getWaterAvailableMM();
          double waterToRemove     = Math.min(waterNeed, waterAvailable);
          
          
          if(waterAvailable < minimumWaterNeed){
//            System.out.println("CROP IN FIELD " + field.name + "                     HAS WATER SHORTAGE: " + waterAvailable + " vs. " + minimumWaterNeed);
            consecutiveDaysWaterNeedNotMet++;
            maxConsecutiveDaysWaterNeedNotMet = Math.max(consecutiveDaysWaterNeedNotMet, maxConsecutiveDaysWaterNeedNotMet);
            // Die?
          }
          else{
//            System.out.println("CROP IN FIELD                     " + field.name + " HAS ENOUGH WATER: " + waterAvailable + " vs. " + minimumWaterNeed);
            consecutiveDaysWaterNeedNotMet = 0;          
          }
          totalActualWater       += waterToRemove;
          totalWaterNeed         += waterNeed;
          totalMinimumWater      += minimumWaterNeed;
          Water waterRemoved = field.removeWaterMM(waterToRemove);
//          System.out.println("REMOVING " + waterToRemove);
          // Now adjust the chemical need adjustment value based on soil
          // chemistry
          Enumeration E = plant.chemicalEffects.keys();
          while(E.hasMoreElements()){
            Plant.PlantChemicalEffect effect = (Plant.PlantChemicalEffect)plant.chemicalEffects.get((String)E.nextElement());
            String chemical = effect.chemical;
            double need = effect.need;
            double present = (double)this.field.getChemicalConcentration(chemical);
            System.out.print("Chemical is: " + chemical + " need is " + need + " present is " + present);
            if(((need > 0) && (present < need)) || ((need < 0) && (present > (Math.abs(need))))){
              chemicalNeedAdjustment -= (.5d) / lifeSpan;
              System.out.println(" chemical need adjustment is " + chemicalNeedAdjustment);
              break; // Once it is short or over on any of them, the adjustment is made- don't do any more!
            }
            else{System.out.println(" need is met.");}
                        
          }
          processChemicals();
        }
        else{
          if(cropAge > lifeSpan * 1.25){
            remove = true;  // Once crop is past harvestable, automatically remove self
          }
        }
        break;
      case NO_PLANT_GROWTH: // This will be the default case
      default:
        cropAge++;
        if(cropAge >= lifeSpan) alive = false;
        break;
    }
  }
  
  /**
   * Processes soil/field chemicals, if consumption or
   * emission of chemicals is appropriate.
   */
  protected void processChemicals(){
    Enumeration E = plant.chemicalEffects.keys();
    while(E.hasMoreElements()){
      Plant.PlantChemicalEffect effect = (Plant.PlantChemicalEffect)plant.chemicalEffects.get((String)E.nextElement());
      System.out.println("PROCESSING EFFECT " + effect.chemical + " " + effect.consume);
      String chemical = effect.chemical;
      double consume = effect.consume * density;
      if(consume > 0) this.field.removeChemical(chemical, consume);  
      if(consume < 0) this.field.addChemical(chemical, (-1 * consume));          
    }
  }
  
  /**
   * Remove this crop from this field, often done after the
   * crop is harvested. By default this does nothing, but it could
   * deposit chemicals onto the ground (as, for example, being plowed under). 
   */
  public void remove(){
    
  }
  
  

  /* Reporting */

  /**
   * Gets a report of this object's status.
   * @return a report of this object's status
   */
  public String getReport(){
    String retVal = "";
    retVal += "  Name:                 " + plant.getName() + "\n";
    retVal += plant.getReport();
    retVal += "  Density:              " + density + "\n";
    retVal += "  Date Planted:         " + datePlanted.getString() + "\n"; 
    retVal += "  Crop Age:             " + cropAge + "\n";
    retVal += "  Crop Stage:           " + currentStage + "\n";
    retVal += "  Crop potential yield: " + getPotentialYield() + "\n";
    retVal += "  COLOR:                " + getColor().toString();
    return retVal;
  }
  
  
  /* Graphics */
  
  /**
   * Gets a color by which to display this crop.
   * Color indicates two things:
   * Immediate health is indicated by hue; if the plant 
   * has been watered recently, it will be a nice healthy green;
   * if several days have passed without it being watered, it will
   * trend toward a burned-out red.
   * 
   * Long-term health is indicated by translucency; if the plant
   * has never experienced shortages, it will be perfectly
   * opaque, but the more frequently it experiences dry conditions
   * the more translucent it will become.
   * @return a Color object to be used to depict this crop
   */
  public Color getColor(){
    double yp = getPotentialYield();
    return (isAlive() ?
        new Color(0, 255, (int)(255 - yp * 255), 0) :
        new Color(255, 0, 0, 255));
  }
  
  /**
   * Inner class that allows a crop to be drawn within a field.
   * A Crop is distributed uniformly throughout a field at a given
   * density, but it is drawn as if it physically was contained in
   * a certain fraction of the field along one dimension. Along the
   * second dimension, the maturity of the crop is represented.
   * The color represents the crop's health and well-being (esp.
   * whether it is alive or too dry).
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class CropSVGDrawer implements SVGRepresentable{
    
    /** Crop corner points to be drawn */ ArrayList pts = new ArrayList();

    /**
     * Inner class that performs the write action
     * when a crop is drawn to an SVG document.
     * 
     *  @author John T. Murphy<br/>
     *  Global Institute of Sustainability,<br/>
     *  Arizona State University<br/>
     *  For license and disclaimers, see the Overview.
     */
    private class CropWriter implements SVGRepresentable{
      /** List of the points that will be drawn as corners */
      private String pointsList;
     
      /**
       * Basic constructor that initializes this with a set of points.
       * @param pts the points to be drawn.
       */
      public CropWriter(ArrayList pts){
        pointsList = SVGAssist.getPointsList(pts);
      }
      
      /**
       * @see abcm.graphics.SVGRepresentable#addToSVG(abcm.graphics.SVGAssist)
       */
      public void addToSVG(SVGAssist svg) {
        String cropColor = SVGAssist.getRGB(getColor());
        svg.addPolygon(pointsList, cropColor, cropColor, svg.getUnitsPerInchX()/100);
      }      
    }
    
    /**
     * Basic constructor that allows the corner points of the crop
     * to be initialized.
     * @param corner1 one anchor corner of the field
     * @param firstTarget the first 'target' point, representing progress 'across' the field
     * @param secondTarget the second 'target' point, representing progress 'across' the field
     * @param corner4 the second anchor corner of the field
     */
    public CropSVGDrawer(Point corner1, Point firstTarget, Point secondTarget, Point corner4){
      Point corner2 = Geometry.pointBetween(corner1, firstTarget, Math.min(getMaturity(), 1));
      Point corner3 = Geometry.pointBetween(corner4, secondTarget, Math.min(getMaturity(), 1));
                    
      pts.add(corner1);
      pts.add(corner2);
      pts.add(corner3);
      pts.add(corner4);
    }
    
    /**
     * @see abcm.graphics.SVGRepresentable#addToSVG(abcm.graphics.SVGAssist)
     */
    public void addToSVG(SVGAssist svg){
      svg.addG(plant.getName() + datePlanted.getString(),"", new CropWriter(pts));      
    }
    
  }
  

  

  /* Object Cloning */
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Crop clone = new Crop((Field)cloneStore.getClone(this.field), (Plant)cloneStore.getClone(this.plant), 
        density, (SimDate)cloneStore.getClone(this.datePlanted));
    cloneStore.put(this, clone);
    clone.alive                             = this.alive;
    clone.consecutiveDaysWaterNeedNotMet    = this.consecutiveDaysWaterNeedNotMet;
    clone.cropAge                           = this.cropAge;
    clone.currentStage                      = this.currentStage;
    clone.lifeSpan                          = this.lifeSpan;
    clone.timeInCurrentStage                = this.timeInCurrentStage;
    
    clone.chemicalNeedAdjustment            = this.chemicalNeedAdjustment;
    clone.cropFactorOfCurrentStage          = this.cropFactorOfCurrentStage;
    
    clone.droughtSensitivityOfCurrentStage  = this.droughtSensitivityOfCurrentStage;
    clone.durationOfCurrentStage            = this.durationOfCurrentStage;
    
    clone.maxConsecutiveDaysWaterNeedNotMet = this.maxConsecutiveDaysWaterNeedNotMet;
    clone.remove                            = this.remove;
    clone.totalActualWater                  = this.totalActualWater;
    clone.totalMinimumWater                 = this.totalMinimumWater;
    clone.totalWaterNeed                    = this.totalWaterNeed;
    
    return clone;
  }
  
}
