package flowag.bio;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableVector;
import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.graphics.X3DAssist;
import abcm.graphics.X3DIndexedFaceSetCreator;
import abcm.math.Geometry;
import abcm.math.Point;
import abcm.math.Polygon;
import abcm.math.SimDate;
import abcm.sim.AbstractModelComponent;
import abcm.space.SpatialElement;
import abcm.space.TopoTriangle;
import abcm.space.TriangulatedTopography;
import abcm.utilities.TextFormatter;
import abcm.utilities.XMLAssist;
import flowag.basic.BasicChemical;
import flowag.basic.Chemical;
import flowag.basic.Environment;
import flowag.basic.RainRecipient;
import flowag.basic.Water;
import flowag.bio.Crop.CropSVGDrawer;


/**
 * A Field is a plot of land in which plants grow. A Field has a single
 * soil object that determines the characteristics of the soil, and a collection
 * of Crop objects that represent plants growing. The Field is delimited by
 * four corner points that mark its location in space and determine its area.
 * The Field object must therefore act as an interface between the spatially
 * determined Field and the aspatial Soil object, especially with respect
 * to the addition of water, which is done in mm (abstract) but must be converted
 * to and from m^3 based on the surface area of the entire field.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Field extends AbstractModelComponent implements SpatialElement, RainRecipient, ABCMDeepCloneable {

/* Public Constants */

  /** Multiplier found in the Blaney-Criddle equations for calculating ETo.  Value is @value       */
           public  static final double BLANEY_CRIDDLE_MULTIPLIER        =   .46;
  /** An addend value found in the Blaney-Criddle equations for calculating ETo. Value is @value.  */
           public  static final double BLANEY_CRIDDLE_ADDEND            =  8.00;

/* Private Constants */

  /** Default percent of annual daytime hours; value is @value                            */ 
           private static final double DEFAULT_PCT_ANNUAL_DAYTIMEHOURS   =   .270;
  /** Default mean daily temperature; will give Blaney-Criddle ETo of 5  (@value)         */ 
           private static final double DEFAULT_MEAN_DAILY_TEMPERATURE    = 22.86635;
  /** Default standard evapotranspiration rate (@value)                                   */ 
           private static final double DEFAULT_STANDARD_ETO              =  5.000;
  /** Depth of 'soil' when displayed on field; this merely elevates the field above 
   * the landscape to make it visible
   */
           private static final double DISPLAY_FIELD_SOIL_DEPTH          = .2;
           
/* Static methods */
  
  /**
   * Calculate the volume of water (in cubic meters) given a fields area and a depth of
   * water in millimeters
   * @param MM depth in millimeters
   * @param a area
   * @return volume
   */
  public static double calcVolumeFromMM(double MM, double a){         return (a * MM) / 1000d;        }
           
  /**
   * Calculate a depth of water, in millimeters, equivalent to a volume of water
   * distributed across an area
   * @param volume the volume of water
   * @param a the area
   * @return the depth, in millimeters
   */
  public static double calcMMFromVolume(double volume, double a){    return (volume / a) * 1000d;     }
           
           
           
/* Static Variables */
  /** Private static value used for creating IDs */ private static int fieldCount;

/* Instance Variables */
  /** Name of this field */
       public String name;
  
  /** NW Corner of the field. 
   * Note that geography is for convenience only; the actual arrangement of the corners is not validated */  
       private Point  NW;
  /** NE Corner of the field. 
   * Note that geography is for convenience only; the actual arrangement of the corners is not validated */  
       private Point  NE;
  /** SE Corner of the field. 
   * Note that geography is for convenience only; the actual arrangement of the corners is not validated */  
       private Point  SE;
  /** SW Corner of the field. 
   * Note that geography is for convenience only; the actual arrangement of the corners is not validated */  
       private Point  SW;
  /** Soil found in this field (uniform)          */  
       private Soil   soil;
  /** Collection of crops in this field           */  
       private ABCMDeepCloneableVector crops = new ABCMDeepCloneableVector();
  /** Standard ETo rate, used when setting directly and not calculated from environment */  
       private double ETo_Set = Double.NaN;
       
  /** Mean daily temperature for this field. If not set, default is Double.NaN  */
       private double meanDailyTemperature_Set  = Double.NaN;
  /** Mean percent annual daytime hours. If not set, default is Double.NaN */
       private double meanPctAnnualDaytimeHours = Double.NaN;
  /** Accumulated water received since the last time step */
       private double waterReceivedSinceLastStep = 0;
  /** Water received during the last time step */
       private double waterReceivedAtLastStep    = 0;
       
  /** Field onto which this field flow water when water is too deep */
       protected Field outflowField;
  /** Flag indicating if this field has an outflow field and should direct ALL received water onto it */
       protected boolean feeder = false;
  
  /** Record of total harvest last year */
       protected double totalHarvestLastYear = 0;
  /** Accumulation of total harvest this year */
       protected double totalHarvestThisYear = 0;
  /** Counter of days (used for recordkeeping at intervals, usually yearly) */
       protected int dayCounter = 0;
       
  /** A private ID for this field                 */  
       private String fieldID;
  
  /** Land area of the field; calculated once,
    * when the (immutable) field corners are
    * established                                  */  
       private double area;
   
       
  /** The effective area is the area actually used for farming; may be less than the total area, which
   * allows concentration of water from the whole field only a portion of it
   */
       private double effectiveArea;
  /**  Area in 2-D, avoiding slope issues */
       private double rainfallArea; 
  /** Pointer to the (climate) environment        */  
       private Environment E;

  /** Minimum X coordinate */  protected double minX; 
  /** Minimum Y coordinate */  protected double minY; 
  /** Minimum Z coordinate */  protected double minZ; 
  /** Maximum X coordinate */  protected double maxX; 
  /** Maximum Y coordinate */  protected double maxY; 
  /** Maximum Y coordinate */  protected double maxZ;
  /** Mean of Z coordinates */ protected double meanZ;
  
  /** Allows Field as SpatialElement to be flagged if it has been changed */
  protected boolean altered = true;

  /** Collection of polygons indicating the boundary of this field; used for spatial representation  */
  protected ABCMDeepCloneableVector boundaryPolygons = null;
  
  /** Collection of polygons indicating the berm area around this field; used for spatial representation */
  protected ABCMDeepCloneableVector bermPolygons = null;
  
  /** Triangulated Topography that represents this field; used for spatial representation   */
  protected TriangulatedTopography triangulatedTopographyOfField = null;
  
/* Constructor */
  /**
   * Constructor with locations for the plot's corners.
   * @param fieldName       name for this field
   * @param NorthWestPoint  North West corner
   * @param NorthEastPoint  North East corner
   * @param SouthEastPoint  South East corner
   * @param SouthWestPoint  South West corner
   * @param physicalEnvironment an Environment for this field to exist within
   */
  public Field(String fieldName, Point NorthWestPoint, Point NorthEastPoint, Point SouthEastPoint, 
      Point SouthWestPoint, Environment physicalEnvironment){
    name = fieldName;
    
    // Specify dimensions
    NW = NorthWestPoint;
    NE = NorthEastPoint;
    SE = SouthEastPoint;
    SW = SouthWestPoint;
    E  = physicalEnvironment;
    
    
    // Set mins and maxs
    minX = Math.min(NW.getX(), Math.min(NE.getX(), Math.min(SE.getX(), SW.getX())));
    minY = Math.min(NW.getY(), Math.min(NE.getY(), Math.min(SE.getY(), SW.getY())));
    minZ = Math.min(NW.getZ(), Math.min(NE.getZ(), Math.min(SE.getZ(), SW.getZ())));
    
    maxX = Math.max(NW.getX(), Math.max(NE.getX(), Math.max(SE.getX(), SW.getX())));
    maxY = Math.max(NW.getY(), Math.max(NE.getY(), Math.max(SE.getY(), SW.getY())));
    maxZ = Math.max(NW.getZ(), Math.max(NE.getZ(), Math.max(SE.getZ(), SW.getZ())));
    
    meanZ = (minZ + maxZ) / 2d;
    
    // Area is a useful stat to have; it never changes, so
    // calculate it here
    area = Geometry.areaOfQuadrilateral3D(NW, NE, SE, SW); // Oddly, can and maybe should use a code model switch here.
                                                           // The issue is which area matters: 2-D only or 3-D?
    effectiveArea = area;
    rainfallArea = Geometry.areaOfTriangle(NW, NE, SE) + Geometry.areaOfTriangle(SE, SW, NW); // Note that this is 2-D- rainfall is measured as if falling on a plane
    
    // A field contains soil and crops; initialize these containers
    // here
    soil = new Soil();
    crops = new ABCMDeepCloneableVector();
    
    
    // Assign a name
    fieldID = "F" + fieldCount;
    fieldCount++;
    
  }
  
  
/* Set Methods */
  /**
   * Sets the ETo_Set variable. This is the value that
   * is used when ETo is simply user-set, not calculated.
   * @param val the ETo_Set to set.
   */
  public void setETo_Set(double val){  ETo_Set = (val > 0 ? val : Double.NaN);   }
  
  
  /**
   * Sets the outflow field for this field; water in excess of this field's
   * capacity will be disgorged onto the specified outflow field
   * @param f the outflow field
   */
  public void setOutflowField(Field f){ outflowField = f; }
  
  /**
   * Clears any previously set outflow field; water in excess of this field's
   * capacity will become standing water 
   */
  public void clearOutflowField(){      outflowField = null; }
  
  /**
   * Sets a flag indicating if this field is a feeder field. If true, all water received on
   * this field is redirected onto the outflow field (if any).
   * @param isFeederField new value for this flag
   */
  public void setFeeder(boolean isFeederField){ feeder = isFeederField; }
  
  
  /* Get methods */
  /**
   * Gets the area of the plot
   * @return the area of the plot
   */
  public double getArea(){                        return area;                                }

  /**
   * Gets the effective area of this field
   * @return the effective area of this field
   */
  public double getEffectiveArea(){               return effectiveArea;                       }
  
  /**
   * Sets the value for the effective area of this field. Converts (magically!) water
   * that was distributed across the old effective area of the field into a new depth
   * in the new effective area
   * @param newEffectiveArea new value for effective area. Must be less than or equal to
   * physical area, and greater than or equal to zero. 
   * @return true if successful, false otherwise
   */
  public boolean setEffectiveArea(double newEffectiveArea){
    if(newEffectiveArea > area || newEffectiveArea <= 0) return false;
    // Convert soil water to new area
    double currentSoilWaterAvailable = soil.getWaterAvailableMM();
    Water W = soil.extractWaterMM(currentSoilWaterAvailable);
    double newWaterDepth = Field.calcMMFromVolume(Field.calcVolumeFromMM(currentSoilWaterAvailable, effectiveArea), newEffectiveArea);
    soil.addWaterMM(W, newWaterDepth);
    for(int i = 0; i < crops.size(); i++){
      Crop c = (Crop)crops.get(i);
      c.density = c.density / (newEffectiveArea / effectiveArea);
    }
    effectiveArea = newEffectiveArea;
    return true;
  }
  
  /**
   * Sets the effective area based on a fractional value
   * @param newEffectiveFraction fraction, between zero and one
   * @return true if successful, false otherwise
   */
  public boolean setEffectiveFraction(double newEffectiveFraction){
    if(newEffectiveFraction <= 0 || newEffectiveFraction > 1) return false;
    return setEffectiveFraction(newEffectiveFraction * area);
  }
  
  
  /* Environment */
  // Many of these elements use a pattern:
  //
  // If a user-defined value is set, use that
  // If not, defer to the _Environment_ variable, if that is set
  // If environment is not set, use built-in default
  
  /**
   * Sets the mean daily temperature. If this is set, it will override the
   * value supplied by the Environment
   * @param newTemp new temperature value
   */
  public void setMeanDailyTemperature(double newTemp){ this.meanDailyTemperature_Set = newTemp; }
  /**
   * Clears the user-set mean daily temperature value, allowing temperature to again
   * be determined by the Environment
   */
  public void clearMeanDailyTemperature(){             this.meanDailyTemperature_Set = Double.NaN; }
  
  /**
   * Gets the mean daily temperature for this field
   * @return the meanDailyTemperature.
   */
  public double getMeanDailyTemperature(){
//    System.out.println("Getting mean daily temp: " + 
//        (!Double.isNaN(this.meanDailyTemperature_Set) ? this.meanDailyTemperature_Set : (E != null ? E.getMeanDailyTempByAltitude(meanZ):Field.DEFAULT_MEAN_DAILY_TEMPERATURE)));
    return(!Double.isNaN(this.meanDailyTemperature_Set) ? this.meanDailyTemperature_Set :   // If user-set value, use it 
               (E != null ? E.getMeanDailyTempByAltitude(meanZ):                             // Generally get temp from environment
                 Field.DEFAULT_MEAN_DAILY_TEMPERATURE));                                     // Otherwise use built-in default 
  }
  
  
  
  /**
   * Sets the value for mean percentage annual daytime hours. If this is
   * set, the value will override that given by the environment
   * @param newVal new value for mean percentage annual daytime hours
   */
  public void setMeanPctAnnualDaytimeHours(double newVal){ this.meanPctAnnualDaytimeHours = newVal;     }
  
  /**
   * Clears the value for mean percentage annual daytime hours, allowing this
   * value to again be provided by the environment 
   */
  public void clearMeanPctAnnualDaytimeHours(){            this.meanPctAnnualDaytimeHours = Double.NaN; }
  
  
  /**
   * Get the mean pecentage of annual daytime hours
   * in effect on this day.
   * @return the meanPctAnnualDaytimeHours.
   */
  public double getMeanPctAnnualDaytimeHours(){   
    return (!Double.isNaN(this.meanPctAnnualDaytimeHours) ? this.meanPctAnnualDaytimeHours :   // If user-set value, use it
      E != null ? E.getMeanPctAnnualDaytimeHours():                                            // Generally get temp from environment
        DEFAULT_PCT_ANNUAL_DAYTIMEHOURS);                                                      // Otherwise use built-in default
  }  
  
  /**
   * Gets the Evapotranspiration rate of the standard grass
   * crop.
   * @return the evapotranspiration rate of the standard grass
   * crop.
   */
  public double getETo(){
//    System.out.println("Getting ETo : " + (!Double.isNaN(ETo_Set) ? ETo_Set : getCalculatedEtO()));
    return (!Double.isNaN(ETo_Set) ? ETo_Set : getCalculatedEtO());    
  }
  
  /**
   * Get the standard evapotranspiration rate (ETo)
   * as calculated using the Blaney-Criddle method
   * @return the ETo as calculated using the Blaney-Criddle method
   */
  public double getCalculatedEtO(){
    return
      getMeanPctAnnualDaytimeHours() * (
          BLANEY_CRIDDLE_MULTIPLIER * getMeanDailyTemperature() + 
          BLANEY_CRIDDLE_ADDEND);
  }
  
  
  
  
  
  
  
/* Action Methods */
  
  /**
   * Sow by planting a crop in the field.
   * Note that this requires a simulation date which
   * becomes that crop's start date, a plant, and
   * a value that indicates how densely the plant
   * is to be planted.
   * @param p the Plant to be sown
   * @param plantDensity the density at which the plant is to be sown
   * @param sd the Simulation Date on which this planting occurs
   */
  public void sow(Plant p, double plantDensity, SimDate sd){
    // If you are trying to plant this crop and there is already on planted
    // with the same _plant_ on the same _date_ then just merge them
    for(int i = 0; i < crops.size(); i++){
      Crop C = (Crop)crops.get(i);
      if(C.plant == p && C.datePlanted.value() == sd.value()){
        C.density+= plantDensity;
        return;
      }      
    }
   
    Crop c = new Crop(this, p, plantDensity, sd);
    // The crop vector is an ordered one...
    if(crops.size() == 0) crops.add(c);
    else{
      for(int i = 0; i < crops.size(); i++){
        if(((Crop)crops.get(i)).datePlanted.value() >= sd.value()) crops.add(i, c);
        break;
      }
      if(!crops.contains(c)) crops.add(c);
    }
    System.out.println("SOWING " + p.getName() + " at density " + plantDensity + " on date " + sd.getString() + " now crops has " + crops.size());
  }
  
  
  /**
   * Harvests a specific crop from this field.
   * @param plantName name of the plant whose crop is to be harvested
   * @param index index number of the crop (if two crops planted using the same plant but
   * at different times exist)
   * @return the yield of the crop
   */
  public double harvest(String plantName, int index){
    Crop C = getCrop(plantName, index);
    if(C == null) return 0;
    C.remove();
    crops.remove(C);
    double ret = C.getHarvestableYield();
    totalHarvestThisYear+=ret;
    return ret;
  }
  
/* Interrogatives */
  
  /**
   * Gets a Crop by plant name; by default uses the first crop (index = 0)
   * for that plant
   * @param plantName the name of the plant
   * @return the Crop object requested; null if it does not exist
   */
  protected Crop getCrop(String plantName){ return getCrop(plantName, 0); }
  
  /**
   * Gets a Crop by plant name and index. Index discriminates between two crops planted with the same
   * plant but on different dates
   * @param plantName the name of the plant
   * @param index index of the plant's Crop
   * @return the Crop object requested; null if it does not exist
   */
  protected Crop getCrop(String plantName, int index){
    if(crops.size() == 0) return null;
    if(crops.size() == 1) return (Crop)crops.get(0); // If there is only one, return it... allows default probing
    Vector listOfCrops = new Vector();
    for(int i = crops.size() - 1; i >=0; i--){
      if(((Crop)crops.get(i)).plant.getName().compareToIgnoreCase(plantName) == 0) listOfCrops.add(crops.get(i));
    }
    if(listOfCrops.size() == 0) return null;
    if(index > (listOfCrops.size() - 1)) return ((Crop)listOfCrops.lastElement());
    if(index < 0) return (Crop)listOfCrops.get(0);
    return (Crop)listOfCrops.get(index);
  }
  
  /**
   * Gets the potential yield for this crop (if allowed to mature)
   * @param plantName name of the plant to be harvested
   * @param index index of the crop to be harvested
   * @return the amount of the potential yield for this crop
   */
  public double getPotentialYield(String plantName, int index){
    Crop C = getCrop(plantName, index);
    if(C == null) return Double.NaN;
    return C.getPotentialYield();
  }
  
  /**
   * Gets the amount of yield immediately harvestable for a given crop.
   * @param plantName name of the plant to be harvested
   * @param index index of the crop to be harvested
   * @return amount immediately harvestable
   */
  public double getHarvestableYield(String plantName, int index){
    Crop C = getCrop(plantName, index);
    if(C == null) return Double.NaN;
    return C.getHarvestableYield();
  }
  
  

/* Step Methods */  
  
  /**
   * Steps this field through its actions for one day
   */
  public void step(){
//    if(this.name.compareTo("Field_N_A1") == 0 ||
//       this.name.compareTo("Field_N_F5") == 0){ 
//      System.out.println("Stepping field " + name);
//      System.out.println(soil.getWater().report());
//    }
    
    waterReceivedAtLastStep = waterReceivedSinceLastStep;
    waterReceivedSinceLastStep = 0;
    
    dayCounter ++;
    if(dayCounter == 365){
      this.totalHarvestLastYear = this.totalHarvestThisYear;
      totalHarvestThisYear = 0;
      dayCounter = 0;
    }
    
//    if(crops.size() > 0)
//        System.out.println("Processing Field; water received at last step = " + waterReceivedAtLastStep + " now " + soil.getWaterAvailableMM() + " mm");
    soil.processWater();
    grow();
  } 
  
  /**
   * Gets the total amount harvested during the last year
   * @return the total of last year's harvest
   */
  public double getTotalHarvestLastYear(){ return totalHarvestLastYear; }
  
  /**
   * Grows the crops in this field for 1 day
   * This takes the crops in random order and
   * calls each crop's step method. This method will
   * affect the crop's stage but will also alter
   * the soil object in which it is growing
   * by removing or contributing water and/or
   * nutrients.
   */
  public void grow(){
    // Randomize the order of the crops
    // and process each one in sequence
//    if(this.name.compareTo("Field_N_A1") == 0) System.out.println("Field.grow() with " + crops.size() + " crops...");
    Vector grabBag = new Vector();
    grabBag.addAll(crops);
//    if(this.name.compareTo("Field_N_A1") == 0) System.out.println("grabBag.size = " + grabBag.size());
    while(grabBag.size() > 0){
      Crop c = (Crop)grabBag.remove((int)Math.floor(getRNG().nextDouble()*grabBag.size()));
//      System.out.println("Growing " + c.plant.getName());
      c.step();
      if(c.remove){
//        System.out.println("Crop " + c.plant.getName() + " is asking to be removed...");
        c.remove();
        crops.remove(c);
      }
    }
  }

  
  /* Water */
  
  // Conceptually there are two different ways to deal with water:
  //     By absolute volume
  //     By depth, in MM, distributed across the field
  // 
  // Canals generally deal with absolute volume. Rainfall, conversely,
  // is usually specified in MM.
  // Internally, the soil object is abstract, lacking any real dimensions.
  // Water is always dealt with in MM inside the soil object. The Field
  // object does not contain any water directly, but converts any
  // amount added to it to a MM of water to be kept inside the soil
  // object.
  
  /**
   * Calculates an actual volume of water given a number
   * of mm of water that would fall on this field. The calculation
   * is simply area * MM / 1000.
   * @param MM the mm of water
   * @return the volume of water
   */
  public double calcVolumeFromMM(double MM){         return calcVolumeFromMM(MM, effectiveArea); }
  
  /**
   * Calculates a number of mm of water given a volume
   * of water, as if that volume of water were spread evenly
   * across the surface area of the field.
   * @param volume the volume of water from which mm are to be calculated
   * @return the mm of water that volume represents
   */
  private double calcMMFromVolume(double volume){    return calcMMFromVolume(volume, effectiveArea);    }
  
  /**
   * Rains water on field surface (area) but converts it
   * into mm based on rainfall area.
   * @param MM
   * @param rainSource
   */
  public void rain(double MM, Water rainSource){
    addWater(new Water(calcVolumeFromMM(MM, rainfallArea), rainSource)); // NOTE: not 'effective area'- the MM is across the entire surface, ignoring slope!
  }
  
  
  /**
   * Adds a volume of water to this field.
   * 
   * Fields do not directly store water; they mediate
   * between 'real' volumes of water and 'abstract' soils.
   * Thus to 'store' the water that is added requires converting
   * the volume to a depth, in mm, and adding that to
   * the 'soil' object (which can be conceived of as having
   * 1 m sq size).
   * @param w Water object to be added
   */
  public void addWater(Water w){
//    System.out.print("Adding water vol = " + w.getVolume() + " to field " + this.name + " start: " + this.getWaterAvailableMM());
    if(feeder && outflowField != null) outflowField.addWater(w);
    else if(outflowField == null)      addWaterToSoil(w);
    else{
      // There is an outflow field set for this; any water added over saturation is directed there
      double originalSoilWaterDepth = soil.getWaterAvailableMM();
      double soilDepth = soil.getSoilDepth();
      double mmToAdd = Field.calcMMFromVolume(w.getVolume(), effectiveArea);
      double mmOver = originalSoilWaterDepth + mmToAdd - soilDepth;
      if(mmOver > 0){
        double volumeToAddHere = calcVolumeFromMM(mmToAdd - mmOver);
        addWaterToSoil(new Water(volumeToAddHere, w));
        outflowField.addWater(new Water((w.getVolume() - volumeToAddHere), w));
      }
      else{ // It won't go over the limit
        addWaterToSoil(w);
      }
    }
//    System.out.println(" new depth = " + this.getWaterAvailableMM());
  }
  
  /**
   * Adds a water object (which will include a volume) to the
   * Soil object contained by this field. Will transform the
   * water into a different volume that represents just the
   * mm of water that are applied to the Soil object, but
   * will do so in a way that maintains the concentrations of
   * chemicals.
   * @param w the Water object to be added
   */
  private void addWaterToSoil(Water w){
    double mm = calcMMFromVolume(w.getVolume());
    waterReceivedSinceLastStep += mm;
    soil.addWaterMM(w, mm);
  }
  
  
  /**
   * Returns the amount of water available for plant growth
   * as a depth, in mm.
   * @return water available, in mm
   */
  public double getWaterAvailableMM(){    return soil.getWaterAvailableMM();    } 
  
  /**
   * Gets the water that had been received at the time the last step method
   * executed; generally this will be one day's worth.
   * @return the value of the water elememt returned
   */
  public double getWaterReceivedAtLastStep(){ return waterReceivedAtLastStep;   }
  
  /**
   * Removes water from this field, by specifying the
   * depth of water to be removed.
   * @param mmWaterToExtract the number of mm to remove
   * @return Water object extracted
   */
  public Water removeWaterMM(double mmWaterToExtract){
    soil.extractWaterMM(mmWaterToExtract);
    return new Water(calcVolumeFromMM(mmWaterToExtract), soil.getWater());
  }
  
/* Reporting */
  /**
   * Gets a report describing this Field.
   * @return a string that is a printable report of this object's
   * current attributes and state.
   */
  public String getReport(){
    String retVal = "";
    retVal += "Field Coordinates: \n" +
              "  " + NW.coordString()  + "   "  + NE.coordString() + "\n" +
              "  " + SW.coordString()  + "   "  + SE.coordString() + "\n";
    
    boolean debug = (getParameterSet().getParameter("DEBUG") != 0);
    int rdp = (int)(Math.floor(getParameterSet().getParameter("ROUNDING_DECIMAL_PLACES")));
    
    String area = (debug ? TextFormatter.roundWithUnits(getArea(), rdp, "m^2") : "" + getArea()) + "\n";
    String temp = (debug ? TextFormatter.roundWithUnits(getMeanDailyTemperature(), rdp, "C") : "" + getMeanDailyTemperature()) + "\n";
    String Eto  = (debug ? TextFormatter.roundWithUnits(getETo(), rdp, "mm/day") : "" + getETo()) + "\n";
    
    retVal += "Area: " + getArea() + "\n";
    retVal += "T:    " + getMeanDailyTemperature() + "\n";
    retVal += "ETo:  " + getETo() + "\n";
    retVal += "Soil:\n" + soil.getReport() + "\n";
    retVal += "Crops: " + crops.size() + "\n";
    for(int i = 0; i<crops.size(); i++){
      Crop c = (Crop)crops.get(i);
      retVal += c.getReport() + "\n"; 
    }
    
    return retVal;
  }
  
  /**
   * Gets the total density of all Crops planted in the Field.
   * @return the total of the densities of all Crops in this Field.
   */
  public double getTotalDensity(){
    double d = 0;
    for(int i = 0; i < crops.size(); i++){
      d+=((Crop)crops.elementAt(i)).getDensity();
    }
    return d;
  }

  /**
   * Adds SVG describing this Field to an XML document.
   * @param svg the SVGAssist object that contains the
   * document to which this Field is to be added
   */
  public void addToSVG(SVGAssist svg){
    
    class FieldDrawer implements SVGRepresentable{
      
      public void addToSVG(SVGAssist svg){
        System.err.println("Adding A FIELD to the SVG: " + name);
        double wSoil = soil.getWaterAvailableMM() /1000d;
        double dSoil = soil.getSoilDepth();
        double opacity = ((wSoil <= dSoil)  ? wSoil/dSoil : 0.5);
//        System.out.println(NW.getX() + "," + NW.getY() + "SOIL WATER = " + wSoil + " SOIL DEPTH = " + dSoil + " opacity: " + opacity);
//        if(wSoil > dSoil) System.out.println("WATER IS DEEPER THAN SOIL...");        
        ArrayList pointsList = new ArrayList();
        pointsList.add(SW);
        pointsList.add(NW);
        pointsList.add(NE);
        pointsList.add(SE);
        String pts = SVGAssist.getPointsList(pointsList);
        // First the outline
        svg.addPolygon(pts,"none","green",svg.getUnitsPerInchX()/100d,  "id=\"field_" + fieldID + "_Border\" onclick=\"showFieldDetails(evt,'" + fieldID + "');\"");
        // Next the soil
        svg.addPolygon(pts,SVGAssist.getRGB(soil.getColor()),"none",0, "opacity=\""+ opacity +"\" " + " id=\"field_" + fieldID + "_Border\" onclick=\"showFieldDetails(evt,'" + fieldID + "');\"");
        class fieldDesc implements SVGRepresentable{
          public void addToSVG(SVGAssist svg){
            svg.addBackground("white");
            svg.addText(getReport(),.1,15,2);
          }
        }
        String out = getReport();
        double h = TextFormatter.getTextHeight(out,15,1);
        double w = TextFormatter.getTextWidth(out,15);
        svg.addPopup(NE.getX(),NE.getY(),w,h,.5,1.5, new SVGRepresentable[] {new fieldDesc()},"orange",.01, fieldID, "visibility=\"hidden\" id=\"fieldLabel_" + fieldID + "\" onclick=\"hideSelf(evt,'fieldLabel_" + fieldID+"');\"");
        if(crops.size()>0){
          SVGRepresentable[] CROPS = new SVGRepresentable[crops.size()];
          double d=0;
          boolean anyAlive = false;
          for(int i=0;i<crops.size(); i++){
            Crop crop = (Crop)crops.get(i);
            System.err.println("   ADDING A CROP " + crop.plant.getName());
            if(crop.isAlive()) anyAlive = true;
            double dens = crop.getDensity();
            
            // The representation of the crop will occupy some portion
            // of the space defined by the field.
            // Note that this represents by linear distance, not by true area...
            Point corner1 = Geometry.pointBetween(SE, SW, d);
            Point corner2 = Geometry.pointBetween(NE, NW, d);
            d = d + dens;
            Point corner3 = Geometry.pointBetween(NE, NW, d);
            Point corner4 = Geometry.pointBetween(SE, SW, d);
            
            CropSVGDrawer CD = crop.new CropSVGDrawer(corner1, corner2, corner3, corner4);
            CROPS[i]=CD;
          }
          svg.addG("Crops","",CROPS);        
        }
      }
    }
    
    FieldDrawer FD = new FieldDrawer();
    svg.addG("Field: " + fieldID, "id=\"field_" + fieldID + "\"", FD);    
  }


  /**
   * Gets the sum of all the densities of crops currently in the field
   * @return the sum of all the densities of the field's crops
   */
  protected double getTotalFractionOfFieldOccupied(){
    double totalFractionOfFieldOccupied = 0;
    Iterator I = crops.iterator();
    while(I.hasNext()){
      Crop c = (Crop)I.next();
      totalFractionOfFieldOccupied += c.getDensity();
    }
    return totalFractionOfFieldOccupied;
  }
  
  /** 
   * @see abcm.graphics.X3DRepresentable#addToX3D(abcm.graphics.X3DAssist) 
   */
  public void addToX3D(X3DAssist x3d){
    
    // Fields are 1m high when fully grown; a box around the field at 1m will represent this,
    // and the field will 'grow' toward the top of the box.
    
    // First create a representation of the soil
    Point soilWaterTop1 = new Point(SW.getX(), SW.getY(), SW.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
    Point soilWaterTop2 = new Point(NW.getX(), NW.getY(), NW.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
    Point soilWaterTop3 = new Point(NE.getX(), NE.getY(), NE.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
    Point soilWaterTop4 = new Point(SE.getX(), SE.getY(), SE.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
    
    double fractionOfWater = Math.min(soil.getSaturation(), 1);
    fractionOfWater = 0; // TEST
    
    Point soilWaterBottom1 = new Point(SW.getX(), SW.getY(), SW.getZ() + DISPLAY_FIELD_SOIL_DEPTH * ( 1 - fractionOfWater));
    Point soilWaterBottom2 = new Point(NW.getX(), NW.getY(), NW.getZ() + DISPLAY_FIELD_SOIL_DEPTH * ( 1 - fractionOfWater));
    Point soilWaterBottom3 = new Point(NE.getX(), NE.getY(), NE.getZ() + DISPLAY_FIELD_SOIL_DEPTH * ( 1 - fractionOfWater));
    Point soilWaterBottom4 = new Point(SE.getX(), SE.getY(), SE.getZ() + DISPLAY_FIELD_SOIL_DEPTH * ( 1 - fractionOfWater));
    
    // Now create an indexed face set for the water
    Vector<Polygon> polygonList = new Vector<Polygon>();
    Vector<Point> pointList = new Vector<Point>();
    
    if(fractionOfWater > 0){
      
      polygonList.clear();
      
      // The water will be represented by six faces:
      // South end
      pointList.clear();
      pointList.add(soilWaterTop1);
      pointList.add(soilWaterTop4);
      pointList.add(soilWaterBottom4);
      pointList.add(soilWaterBottom1);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // North end
      pointList.clear();
      pointList.add(soilWaterTop3);
      pointList.add(soilWaterTop2);
      pointList.add(soilWaterBottom2);
      pointList.add(soilWaterBottom3);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // East end
      pointList.clear();
      pointList.add(soilWaterTop4);
      pointList.add(soilWaterTop3);
      pointList.add(soilWaterBottom3);
      pointList.add(soilWaterBottom4);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // West end
      pointList.clear();
      pointList.add(soilWaterTop2);
      pointList.add(soilWaterTop1);
      pointList.add(soilWaterBottom1);
      pointList.add(soilWaterBottom2);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // Triangle 1 (top)
      pointList.clear();
      pointList.add(soilWaterTop1);
      pointList.add(soilWaterTop2);
      pointList.add(soilWaterTop3);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // Triangle 2 (top)
      pointList.clear();
      pointList.add(soilWaterTop1);
      pointList.add(soilWaterTop3);
      pointList.add(soilWaterTop4);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      X3DIndexedFaceSetCreator ifsc = new X3DIndexedFaceSetCreator(polygonList);
      
      XMLAssist xml = x3d.getXML();
      xml.open("Group");
        xml.open("Shape");
          xml.open("Appearance");
            xml.addLine("Material", "diffuseColor='0.0 0.0 1.0'", "");
          xml.closeTag();
          xml.open("IndexedFaceSet", "coordIndex=\"" + ifsc.getFaceList() + "\"" + " solid=\"false\"");
            xml.addLine("Coordinate", "point=\"" + ifsc.getPointList() + "\"", "");
          xml.closeTag(); // Indexed Face Set
        xml.closeTag(); // Terrain Shape
      xml.closeTag(); // Terrain Group

      
    }
    // And an indexed face set for the soil
    if(fractionOfWater < 1){
      
      polygonList.clear();
      
      // The water will be represented by six faces:
      // South end
      pointList.clear();
      pointList.add(soilWaterBottom1);
      pointList.add(soilWaterBottom4);
      pointList.add(SW);
      pointList.add(SE);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // North end
      pointList.clear();
      pointList.add(soilWaterBottom3);
      pointList.add(soilWaterBottom2);
      pointList.add(NW);
      pointList.add(NE);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // East end
      pointList.clear();
      pointList.add(soilWaterBottom4);
      pointList.add(soilWaterBottom3);
      pointList.add(NE);
      pointList.add(SE);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // West end
      pointList.clear();
      pointList.add(soilWaterBottom2);
      pointList.add(soilWaterBottom1);
      pointList.add(SW);
      pointList.add(NW);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // Triangle 1 (top)
      pointList.clear();
      pointList.add(soilWaterBottom1);
      pointList.add(soilWaterBottom2);
      pointList.add(soilWaterBottom3);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      // Triangle 2 (top)
      pointList.clear();
      pointList.add(soilWaterBottom1);
      pointList.add(soilWaterBottom3);
      pointList.add(soilWaterBottom4);
      polygonList.add(new Polygon(pointList));
      pointList.clear();

      X3DIndexedFaceSetCreator ifsc = new X3DIndexedFaceSetCreator(polygonList);
      
      XMLAssist xml = x3d.getXML();

      Color c = soil.getColor();
      
      xml.open("Group");
        xml.open("Shape");
          xml.open("Appearance");
            xml.addLine("Material", "diffuseColor='" + (c.getRed() / 256d) + " " + (c.getGreen() / 256d) + " " + (c.getBlue() / 256d) + "'", "");
          xml.closeTag();
          xml.open("IndexedFaceSet", "coordIndex=\"" + ifsc.getFaceList() + "\"" + " solid=\"false\"");
            xml.addLine("Coordinate", "point=\"" + ifsc.getPointList() + "\"", "");
          xml.closeTag(); // Indexed Face Set
        xml.closeTag(); // Terrain Shape
      xml.closeTag(); // Terrain Group

      
    }
    
    
    // Crops
    // Each crop gets a fraction of the field's area, defined as a fraction of the N-S length,
    // but goes from E-W entirely
    double startFraction = 0;
    double endFraction = 0;
    
    Iterator I = crops.iterator();
    while(I.hasNext()){
      Crop c = (Crop)I.next();
      double d = c.getDensity();
      endFraction = endFraction + d;
      double h = c.getMaturity();
      Point corner1 = Geometry.pointBetween(SW, SE, startFraction); // Clockwise?
      Point corner2 = Geometry.pointBetween(NW, NE, startFraction);
      Point corner3 = Geometry.pointBetween(NW, NE, endFraction);
      Point corner4 = Geometry.pointBetween(SW, SE, endFraction);
      
      Point mid1    = (startFraction > 0 ? Geometry.pointBetween(NW, SE, startFraction) : Geometry.pointBetween(corner1, corner2, 0.999));
      Point mid2    = (endFraction   < 1 ? Geometry.pointBetween(NW, SE, endFraction)   : Geometry.pointBetween(corner3, corner4, 0.999));
      
      corner1 = new Point(corner1.getX(), corner1.getY(), corner1.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
      corner2 = new Point(corner2.getX(), corner2.getY(), corner2.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
      corner3 = new Point(corner3.getX(), corner3.getY(), corner3.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
      corner4 = new Point(corner4.getX(), corner4.getY(), corner4.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
      mid1 = new Point(mid1.getX(), mid1.getY(), mid1.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
      mid2 = new Point(mid2.getX(), mid2.getY(), mid2.getZ() + DISPLAY_FIELD_SOIL_DEPTH);
      
      Point upperCorner1 = new Point(corner1.getX(), corner1.getY(), corner1.getZ() + h); 
      Point upperCorner2 = new Point(corner2.getX(), corner2.getY(), corner2.getZ() + h);
      Point upperCorner3 = new Point(corner3.getX(), corner3.getY(), corner3.getZ() + h);
      Point upperCorner4 = new Point(corner4.getX(), corner4.getY(), corner4.getZ() + h);
      Point upperMid1    = new Point(mid1.getX(), mid1.getY(), mid1.getZ() + h);
      Point upperMid2    = new Point(mid2.getX(), mid2.getY(), mid2.getZ() + h);
      
      polygonList.clear();
      
      // Near end
      pointList.clear();
      pointList.add(corner1);
      pointList.add(upperCorner1);
      pointList.add(upperCorner4);
      pointList.add(corner4);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      // Far end
      pointList.clear();
      pointList.add(corner3);
      pointList.add(upperCorner3);
      pointList.add(upperCorner2);
      pointList.add(corner2);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      // Near Side 1
      pointList.clear();
      pointList.add(upperMid1);
      pointList.add(upperCorner1);
      pointList.add(corner1);
      pointList.add(mid1);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      // Near Side 2
      pointList.clear();
      pointList.add(upperCorner2);
      pointList.add(upperMid1);
      pointList.add(mid1);
      pointList.add(corner2);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      // Top 1
      pointList.clear();
      pointList.add(upperMid1);
      pointList.add(upperMid2);
      pointList.add(upperCorner4);
      pointList.add(upperCorner1);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      // Top 2
      pointList.clear();
      pointList.add(upperCorner2);
      pointList.add(upperCorner3);
      pointList.add(upperMid2);
      pointList.add(upperMid1);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      // Far Side 1
      pointList.clear();
      pointList.add(upperCorner4);
      pointList.add(upperMid2);
      pointList.add(mid2);
      pointList.add(corner4);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      // Far Side 2
      pointList.clear();
      pointList.add(upperMid2);
      pointList.add(upperCorner3);
      pointList.add(corner3);
      pointList.add(mid2);
      polygonList.add(new Polygon(pointList));
      pointList.clear();
      
      X3DIndexedFaceSetCreator ifsc = new X3DIndexedFaceSetCreator(polygonList);
      
      // etc...
      
      startFraction = endFraction;
      
      XMLAssist xml = x3d.getXML();
      System.out.println("Adding a crop X3D: " + c.plant.getName() + " planted on " + c.datePlanted.getString());
      xml.open("Group");
        xml.open("Shape");
          xml.open("Appearance");
            xml.addLine("Material", "diffuseColor='0.2 0.5 0.1'", "");
          xml.closeTag();
          xml.open("IndexedFaceSet", "coordIndex=\"" + ifsc.getFaceList() + "\"" + " solid=\"false\"");
            xml.addLine("Coordinate", "point=\"" + ifsc.getPointList() + "\"", "");
          xml.closeTag(); // Indexed Face Set
        xml.closeTag(); // Terrain Shape
      xml.closeTag(); // Terrain Group

//      x3d.addBillboard(p, billboardText)
    }
            
  }
  
  /**
   * @see abcm.space.SpatialElement#getBermPolygons()
   */
  public Vector getBermPolygons(){
    if(bermPolygons != null) return bermPolygons;
    double distance1 = NE.distanceToPoint(SW);
    double distance2 = NW.distanceToPoint(SE);
    double frac1 = (distance1 + 1)/distance1;
    double frac2 = (distance2 + 1)/distance2;
    
    Point newNE = Geometry.pointBetween(SW, NE, frac1);
    Point newSE = Geometry.pointBetween(NW, SE, frac2);
    Point newSW = Geometry.pointBetween(NE, SW, frac1);
    Point newNW = Geometry.pointBetween(SE, NW, frac2);
    
    TopoTriangle T1 = new TopoTriangle(newNW, newNE, newSE);
    TopoTriangle T2 = new TopoTriangle(newSE, newSW, newNW);
    ABCMDeepCloneableVector V = new ABCMDeepCloneableVector();
    V.add(T1);
    V.add(T2);
    bermPolygons = V; 
    return V;
  }
  
  /**
   * @see abcm.space.SpatialElement#getBoundaryPolygons()
   */
  public Vector getBoundaryPolygons(){
    if(boundaryPolygons != null) return boundaryPolygons;
    TopoTriangle T1 = new TopoTriangle(NW, NE, SE);
    TopoTriangle T2 = new TopoTriangle(SE, SW, NW);
    ABCMDeepCloneableVector V = new ABCMDeepCloneableVector();
    V.add(T1);
    V.add(T2);
    boundaryPolygons = V; 
    return V;
  }
  
  /**
   * @see abcm.space.SpatialElement#getTriangulatedTopographyWithinBoundaryPolygons()
   */
  public TriangulatedTopography getTriangulatedTopographyWithinBoundaryPolygons(){
    if(triangulatedTopographyOfField != null) return triangulatedTopographyOfField;
    triangulatedTopographyOfField = new TriangulatedTopography(new Vector());
    triangulatedTopographyOfField.addTopoTriangles(getBoundaryPolygons());
    return triangulatedTopographyOfField;
  }
  
  /**
   * Will return true on first call, but not thereafter;
   * @see abcm.space.SpatialElement#altered()
   */
  public boolean altered(){
    if(altered = true){
      altered = false;
      return true;
    }
    return false;
  }
  
  /**
   * @see abcm.space.SpatialElement#getMinX()
   */
  public double getMinX(){ return minX; }
  
  /**
   * @see abcm.space.SpatialElement#getMinY()
   */
  public double getMinY(){ return minY; }
  
  /**
   * @see abcm.space.SpatialElement#getMinZ()
   */
  public double getMinZ(){ return minZ; }
  
  /**
   * @see abcm.space.SpatialElement#getMaxX()
   */
  public double getMaxX(){ return maxX; }
  
  /**
   * @see abcm.space.SpatialElement#getMaxY()
   */
  public double getMaxY(){ return maxY; }
  
  /**
   * @see abcm.space.SpatialElement#getMaxZ()
   */
  public double getMaxZ(){ return maxZ; }
  
  /**
   * @see abcm.space.HasElevation#getZ(abcm.math.Point)
   */
  public double getZ(Point p){
    return getTriangulatedTopographyWithinBoundaryPolygons().getZ(p);
  }
  
  
  /** 
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore) 
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Field clone = new Field(name, (Point)cloneStore.getClone(NW), (Point)cloneStore.getClone(NE), 
        (Point)cloneStore.getClone(SE), (Point)cloneStore.getClone(SW), (Environment)cloneStore.getClone(E));
    cloneStore.put(this, clone);
    clone.ETo_Set = this.ETo_Set;
    clone.crops = (ABCMDeepCloneableVector)this.crops.getDeepClone(cloneStore);
    clone.area = this.area;
    clone.effectiveArea = this.effectiveArea;
    clone.fieldID = this.fieldID;
    if(this.isSelected()) clone.select(); // Default is not selected; only need to mark selected if it is
    clone.soil = (Soil)cloneStore.getClone(this.soil);
    clone.altered = altered;
    if(boundaryPolygons != null) clone.boundaryPolygons = (ABCMDeepCloneableVector)boundaryPolygons.getDeepClone(cloneStore);
    else clone.boundaryPolygons = null;
    
    if(bermPolygons != null) clone.bermPolygons = (ABCMDeepCloneableVector)bermPolygons.getDeepClone(cloneStore);
    else clone.bermPolygons = null;
    
    return clone;
  }
  
  /**
   * Determines if the specified point lies inside the
   * boundaries of the field
   * @param P the point to be checked
   * @return true if the point is inside the field (including on edges),
   * false otherwise
   */
  public boolean pointIsInside(Point P){
    Vector V = getBoundaryPolygons();
    for(int i = 0; i < V.size(); i++){
      if(((Polygon)V.get(i)).pointIsInside(P)) return true;
    }
    return false;
  }
  
  /**
   * Gets the concentration of a chemical in this field
   * @param chem the chemical to be analyzed
   * @return concentration of that chemical in this field's soil
   */
  public double getChemicalConcentration(Chemical chem){
    return soil.getChemicalConcentration(chem);    
  }
  
  /**
   * Gets the current concentration of a chemical in this field
   * @param chem chemical to be analyzed
   * @return concentration of that chemical in this field's soil
   */
  public double getChemicalConcentration(String chem){
    return soil.getChemicalConcentration(chem);
  }
  
  /**
   * Adds a chemical to the soil in this field, at the specified
   * concentration
   * @param name name of the chemical to be added
   * @param concentration concentration to be added
   */
  public void addChemical(String name, double concentration){
    soil.addChemical(new BasicChemical(name, concentration));
  }
  
  /**
   * Removes all or a portion of a chemical from the soil in this field 
   * @param name name of the chemical to remove
   * @param concentration concentration to be removed
   */
  public void removeChemical(String name, double concentration){
    soil.removeChemical(name, concentration);
  }
}
