package flowag.bio;

import abcm.sim.AbstractModelComponent;

/**
 * Represents what is collected from a crop. These can 'act', in the
 * sense that they can age or be processed.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Harvest extends AbstractModelComponent{

  /** Which plant was harvested                       */   private Plant  plant;
  /** Age of this harvest. Harvests age over time     */   private int    age;
  /** A record of the labor required to gather  this harvest. Can also be used to accumulate
   *  labor as processing takes place.                */   
                                                           public  double labor;
  /** Size of this harvest, in bushels                */   private double bushels;
  /** Lifespan of this harvest before it
   * becomes unusable.                                */   public  double lifespan;
  
  /**
   * Basic constructor in which plant type and size of harvest are specified.
   * Age is set to zero.
   * @param p Plant harvested
   * @param Bushels size of harvest
   */ 
  public Harvest(Plant p, double Bushels){
    plant = p;
    age = 0;
    lifespan = plant.getUnprocessedStorageLifespan(); 
  }
  
  /**
   * Constructor in which plant type, size of harvest, and initial
   * age of harvest are specified.
   * @param p Plant harvested
   * @param Bushels size of harvest
   * @param a initial age of harvest
   */
  private Harvest(Plant p, double Bushels, int a){
    plant = p;
    age = a;
    lifespan = plant.getUnprocessedStorageLifespan();
  }
  
  /**
   * Basic step function; causes age to be incremented.
   * @return true
   */
  public boolean step(){
    age++;
    return true;
  }
  
  /**
   * Allows a portion of the harvest to be removed into
   * another harvest. Allocates labor proportionally.
   * Returns null if bushelsToRemove is zero or less.
   * Returns the entire contents of original Harvest object
   * if bushelsToRemove is larger than Harvest size.
   * The new Harvest object has the same age as the original.
   * @param bushelsToRemove amount to be removed.
   * @return a new Harvest object
   */
  public Harvest getPortion(double bushelsToRemove){
    if(bushelsToRemove <= 0 ) return null;
    double b = Math.max(bushels, bushelsToRemove);
    double laborInNewPortion = labor * (bushelsToRemove/bushels);
    labor = labor - laborInNewPortion;
    bushels -= b;
    Harvest H = new Harvest(plant, b, age); // Note: removed portion is still aged!
    H.labor = laborInNewPortion;
    return H;
  }
  
  /**
   * Gets the labor required to process the harvest. This
   * is a derived value from the Plant type being processed,
   * scaled for the size of the Harvest.
   * @return labor required to process the harvest.
   */
  public double getLaborRequiredToProcess(){
    return plant.getLaborRequiredToProcessStoredHarvest() * bushels;
  }
  
  /**
   * Process this harvest into something that will last longer.
   * Note that the labor used will only be equal to the amount
   * of labor that must be used to process the entire Harvest. If
   * the agent wishes to process only a portion of the Harvest, it must
   * split the Harvest into two objects. 
   * @param labor the amount of labor agent is willing to expend
   * @return the amount of labor expended
   */
  public double process(double labor){
    double laborRequired = getLaborRequiredToProcess();
    if(labor <= laborRequired) return 0;
    labor += laborRequired;                                     // Add this to the amount of labor stored
    lifespan = plant.getProcessedStorageLifespan();                 // Extend the lifespan
    return laborRequired;                                       // Return the labor used. 
  }
  
}
