package flowag.bio;

import java.util.Iterator;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.cloning.ABCMDeepCloneableVector;
import abcm.sim.AbstractModelComponent;

/**
 * A plant is a generic plant object;an instance of a plant object 
 * is no more than a store of the information that a Crop object needs to
 * move through the stages from planting to harvest.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Plant extends AbstractModelComponent implements ABCMDeepCloneable{

   /* Default Values */
  /** Default sensitivity to drought for a plant */ public static double DEFAULT_DROUGHT_SENSITIVITY     =   0.90;

  /* Instance variables */  
  /** A name for this plant instance     */ protected String name;
  /** The life stages of this plant      */ protected ABCMDeepCloneableVector stages             = new ABCMDeepCloneableVector();
  /** The chemical effects on this plant */ protected ABCMDeepCloneableHashtable chemicalEffects = new ABCMDeepCloneableHashtable(); 


/* Inner classes */
  /**
   * PlantStage
   * 
   * A plant stage is a container object for the variables that
   * describe a stage in the plant's life cycle.
   * 
   * A stage is described by:
   * 
   *   A duration    - the number of days that the stage will last
   *   A crop factor - the ratio of this plant's evapotranspiration
   *      rate to standard grass's, during this stage.
   *   A drought sensitivity - The minimum ratio of water available to
   *      water need that a plant must meet before having damage
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class PlantStage implements ABCMDeepCloneable{
    /** Duration of this stage, in days               */ public int    duration;
    /** Crop factor for this plant during this stage. This is the ratio of this plant's  
      *   EvapoTranspiration rate to that of a standard grass  */
                                                         public double cropFactor;                                                       
    /** Sensitivity to drought during this stage      */ public double droughtSensitivity;
    /** Labor required to keep this plant alive       */ public double laborRequired;
    
    /**
     * Basic constructor for a plant stage that allows all values to be set.
     * @param d duration, in days
     * @param cf crop factor
     * @param ds drought sensitivity
     */
    private PlantStage(int d, double cf, double ds){
      duration = d;
      cropFactor = cf;
      droughtSensitivity = ds;
    }
    
    /**
     * Basic constructor for a plant stage that allows duration and crop factor
     * to be set but uses default drought sensitivity.
     * @param d duration, in days
     * @param cf crop factor
     */
    private PlantStage(int d, double cf){
      duration = d;
      cropFactor = cf;
      droughtSensitivity = DEFAULT_DROUGHT_SENSITIVITY;
    }
    
    /**
     * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
     */
    public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
      PlantStage newStage = new PlantStage(duration, cropFactor, droughtSensitivity);
      cloneStore.put(this, newStage);
      return newStage;
    }
  }

  /**
   * Gets a default Plant with one default stage
   * @param plantName name to be assigned to this plant
   * @return the requested default Plant
   */
  public static Plant getDefaultPlant(String plantName){
    Plant p = new Plant(plantName);
    p.addStage(1, 90, 1.0, 1.0);
    return p;
  }
  
  
  /**
   * A ChemicalEffect is a set of values that determine how the presence or absence of
   * chemicals in the soil can affect a plant. 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class PlantChemicalEffect implements ABCMDeepCloneable{
    
    /** The name of the chemical to affect the plant  */
    protected String chemical;
    /** Value for what concentration the plant needs, or can be imacted by, for this chemical */
    protected double need;
    /** Value for how much the plant will reduce (or increase) the concentration of a plant in the soil where it grows */
    protected double consume;
    
    /**
     * Constructor
     * @param chemicalName name of the chemical
     * @param chemicalNeed value of need
     * @param consumption value of consumption
     */
    public PlantChemicalEffect(String chemicalName, double chemicalNeed, double consumption){
      System.out.println("INITIALIZING for " + name + " CHEMICAL = " + chemicalName);
      chemical = chemicalName;
      need     = chemicalNeed;
      consume  = consumption;
    }
    
    /**
     * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
     */
    public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
      PlantChemicalEffect newEffect = new PlantChemicalEffect(chemical, need, consume);
      cloneStore.put(this, newEffect);
      return newEffect;
    }
  }
  
  
  
  /**
   * Construct a new plant object and assign it the
   * given name.
   * @param newName the new name for the plant
   */
  public Plant(String newName){
    name = newName;
  }
  
/* Set Methods */

  /**
   * Add a stage to this plant.
   * @param stageNum The number of the stage to be added. Note that
   * this will push any previously added stages that were added
   * with equal or higher stage numbers one position higher.
   * Note that the argument is cardinal while the stage internally
   * is ordinal (what from the outside is called 'stage 1' is
   * stored internally as 'stage 0').
   * @param duration The number of days the stage will last
   * @param cropFactor The crop factor of the current stage
   * @param droughtSensitivity The drought sensitivity of the current stage
   */
  public void addStage(int stageNum, int duration, 
      double cropFactor, double droughtSensitivity){
    PlantStage p = new PlantStage(duration, cropFactor, droughtSensitivity);
    stages.insertElementAt(p, stageNum - 1);
  }
    
  
  /**
   * Add a chemical effect to this plant
   * @param chemical name of the chemical
   * @param need value of plant need
   * @param consume value of plant consumption
   */
  public void addChemicalEffect(String chemical, double need, double consume){
    if(chemicalEffects.containsKey(chemical)) chemicalEffects.remove(chemical);
    chemicalEffects.put(chemical, new PlantChemicalEffect(chemical, need, consume));
  }
  
  
/* Get Methods */  
  /**
   * Get the name of this plant object
   * @return the name of the plant object
   */
  public String getName() {
    return name;
  }  

  /**
   * Gets the drought sensitivity of this plant
   * Note: returns the weighted average of all stages.
   * If no stages, returns default
   * @return Returns the droughtSensitivity.
   */
  public double getDroughtSensitivity() {
    int ls = getLifeSpan();
    if(stages.size() == 0) return DEFAULT_DROUGHT_SENSITIVITY;
    else{
      double ds = 0;
      for(int i = 0; i < stages.size(); i++){
        PlantStage ps = (PlantStage)stages.get(i);
        ds += ps.droughtSensitivity * (double)ps.duration;
      }
      return ds / (double)ls;
    }
  }
  
  /**
   * Gets the life span of this plant, in days. Will be
   * equal to the sum of all the stages' durations.
   * @return the life span of the plant
   */
  public int getLifeSpan(){
    int ls = 0;
    for(int i = 0; i < stages.size(); i++){
      ls += ((PlantStage)stages.get(i)).duration;
    }
    return ls;
  }
  
  /**
   * Get a count of the number of stages this plant has.
   * @return the number of stages in the growth cycle of this plant.
   */
  public int getCountOfStages(){    return stages.size();     }

  /**
   * Returns true if the given stage number is a valid one (that
   * a stage with that number exists)
   * @param stageNum the requested stage number
   * @return true if it is valid
   */
  private boolean stageNumIsValid(int stageNum){
    return ((stageNum > 0) && (stageNum <= stages.size()));
  }
  
  /**
   * Get the duration of the specified stage of this plant
   * @param stageNum The number of the stage for which the 
   * duration is requested
   * @return the duration, in days, of the specified stage
   * of this plant.
   */
  public int getDurationOfStage(int stageNum){
    return(stageNumIsValid(stageNum) ? ((PlantStage)stages.get(stageNum-1)).duration : 0);
  }

  /**
   * Get the crop factor for the plant at a specified stage
   * @param stageNum The number of stage for which the crop factor
   * is to be returned.
   * @return the crop factor of the plant at that stage
   */
  public double getCropFactorOfStage(int stageNum){
    return(stageNumIsValid(stageNum) ? ((PlantStage)stages.get(stageNum-1)).cropFactor : 0);
  }
  
  /**
   * Get the Drought Sensitivity for the plant at a specified stage
   * @param stageNum The number of the stage for which the
   * drought sensitivity is sought
   * @return the drought sensitivity of the plant at that stage
   */
  public double getDroughtSensitivityOfStage(int stageNum){
    return(stageNumIsValid(stageNum) ? ((PlantStage)stages.get(stageNum-1)).droughtSensitivity : 0);    
  }  

  
  
  /**
   * Gets a report describing this plant
   * @return the Plant status report
   */
  public String getReport(){
    
    String retVal = "";
    retVal += "Name: " + name + "\n";
    for(int i = 0; i < stages.size(); i++){
      PlantStage ps = (PlantStage)stages.get(i);
      retVal += "  Stage: " + (i + 1) + "\n";
      retVal += "     Duration:           " + ps.duration + "\n";
      retVal += "     Crop Factor:        " + ps.cropFactor + "\n";
      retVal += "     Drough Sensitivity: " + ps.droughtSensitivity + "\n";
    }
    Iterator I = chemicalEffects.keySet().iterator();
    while(I.hasNext()){
      String chem = (String)I.next();
      PlantChemicalEffect pce = (PlantChemicalEffect)chemicalEffects.get(chem);
      retVal += "  Chemical: " + chem + " need: " + pce.need + " consume: " + pce.consume + "\n";
    }
    return retVal;
  }
  
  // TODO Labor issues in Plant.java
  
  /**
   * Gets the labor required to tend to this plant in the specified stage
   * Not implemented. 
   * @param stageNum the number of the stage requested
   * @return the labor required; returns zero if the stage given is below zero or
   * above the actual highest stage number
   */
  public double getLaborRequiredInCurrentStage(int stageNum){
    if(stageNum < 0) return 0;
    if(stageNum > (getCountOfStages() - 1)) return 0;
    if(stages.get(stageNum-1)!=null){
      return ((PlantStage)stages.get(stageNum-1)).laborRequired;
    }
    else return 0;
  }
  
//  public double getTotalLaborRequired(){
//    double labor = 0;
//    for(int i=0; i<stages.size(); i++){
//      labor+=((PlantStage)stages.get(i)).laborRequired;
//    }
//    return labor;
//  }
  
//  public double getHarvestLabor(){return 1;}
//  public double getYieldToBushels() { return 1; }
  
  /**
   * Used only by Harvest object; not currently implemented
   * @return duration this plant can be stored without processing after harvest
   */
  public double getUnprocessedStorageLifespan() {return 100;}
  
  
  /**
   * Used only by Harvest object; not currently implemented
   * @return duration this plant can be stored after processing
   */
  public double getProcessedStorageLifespan() {return 200;}
  
  /**
   * Used only by Harvest object; not currently implemented
   * @return labor required to process this plant after harvest
   */
  public double getLaborRequiredToProcessStoredHarvest() {return 1;}

  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Plant newClone = new Plant(name);
    cloneStore.put(this, newClone);
    newClone.stages = (ABCMDeepCloneableVector)stages.getDeepClone(null);
    newClone.chemicalEffects = (ABCMDeepCloneableHashtable)chemicalEffects.getDeepClone(null);
    return newClone;
  }
  
}
