package flowag.bio;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.List;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.sim.AbstractModelComponent;
import abcm.sim.ParameterSet;
import abcm.utilities.TextFormatter;
import flowag.basic.BasicChemical;
import flowag.basic.Chemical;
import flowag.basic.Water;

/**
 * A Soil object allows water and chemicals to impact plants growing in
 * it. Soil objects are always abstract when compared to Field objects;
 * that is, Fields have a definite location in space, and the volume of
 * water placed on them is measurable in real terms, whereas soils
 * are one-dimensional (depth and depth, in mm, of water).
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
/**
 * @author John
 *
 */
public class Soil extends AbstractModelComponent implements ABCMDeepCloneable{

  
  /** Default fineness of soil (@value unitless) */  public static final double DEFAULT_SOIL_FINENESS = .1;
  
  /** Chemicals found in this soil               */   
  private ABCMDeepCloneableHashtable<String, Chemical> chemicals = new ABCMDeepCloneableHashtable<String, Chemical>();
  
  /** Depth of soil (mm)                         */   private double soilDepth_mm  = 1000 * paramVal("DEFAULT_SOIL_DEPTH"); // 
  /** Water contained in this soil               */   private Water water          = new Water(0);
  
  /**
   * Default name assigned to 'null' chemical placeholder
   */
  public static final String NULL_CHEMICAL_NAME = "NULL_CHEMICAL";
  
  /* Constants */
  
  // Code Model Switches
  /** Code Model Switch: No evaporation          */   public static final int NO_EVAPORATION    = 0;
  /** Code Model Switch: Basic evaporation       */   public static final int BASIC_EVAPORATION = 1;
  
  /** Code Model Switch: No seepage              */   public static final int NO_SEEPAGE           = 0;
  /** Code Model Switch: Basic Seepage           */   public static final int BASIC_SEEP_PROCEDURE = 1;
  
  /* Constructor */
  /**
   * Constructor 
   */
  public Soil(){
    chemicals.put(NULL_CHEMICAL_NAME, new BasicChemical(NULL_CHEMICAL_NAME, paramVal("SOIL_CHEMICAL_CONCENTRATION_LIMIT")));
  }
  
  /**
   * Provides public access to the water object; allows
   * assessment of the water quality (which chemicals in
   * which concentrations). Note that the volume of this
   * water object is arbitrary.
   * @return the current water object.
   */
  public Water getWater(){    return water;  }
  
  /**
   * Processes Water through this soil object; currently this does
   * nothing, but water may seep or evaporate away. More importantly,
   * water may transfer some of its chemicals to the soil, so that even
   * when the water is removed the chemicals remain.
   */
  public void processWater(){
    evaporate();
    seep();
  }
  
  /**
   * Adds an instance of a chemical to this soil
   * object. The Chemical object has a concentration
   * built in, and a name. If the chemical is already
   * present in the soil, it is added to it such that
   * its concentration increases.
   * Note that this will eventually hit a ceiling,
   * at which point it begins scaling chemicals
   * proportionally. Until then it is strictly additive
   * @param c the chemical to be added
   */
  public void addChemical(Chemical c){
    List<Chemical> l = new ArrayList<Chemical>();
    l.add(c);
    addChemicals(l);
  }
  
  /**
   * Adds a list of chemicals, then rescales. Can accommodate adding
   * the null chemical
   * @param listOfChemicalsToAdd the collection of chemicals to be added
   */
  private void addChemicals(List<Chemical> listOfChemicalsToAdd){
//    System.out.println("Adding chemicals!");
    for(Chemical c: listOfChemicalsToAdd){
      String name                   = c.getName();
      double concentrationToBeAdded = c.getConcentration();
      
//      System.out.println("Adding chemical to soil name = " + c.getName() + " conc = " + c.getConcentration());
      
      if(!chemicals.containsKey(name)) chemicals.put(name, new BasicChemical(name, 0)); // Placeholder- zero concentration
      
      Chemical nullInTable = ((Chemical)chemicals.get(Soil.NULL_CHEMICAL_NAME));
      
      // If you're adding the null chemical
      // Assume you know what you're doing
      if(c.getName().compareToIgnoreCase(Soil.NULL_CHEMICAL_NAME) == 0){
        nullInTable.setConcentration(nullInTable.getConcentration() + concentrationToBeAdded);
      }
      else{ // Otherwise, add a normal chemical
        double currentNullConcentration = ((Chemical)chemicals.get(Soil.NULL_CHEMICAL_NAME)).getConcentration();
        Chemical inTable = ((Chemical)chemicals.get(name));
//        System.out.println("Found a chemical in the table: " + inTable.getName() + " " + inTable.getConcentration());
        inTable.setConcentration(inTable.getConcentration() + concentrationToBeAdded);
        if(concentrationToBeAdded < currentNullConcentration) // That is, if there's space available
            nullInTable.setConcentration(Math.max(nullInTable.getConcentration() - concentrationToBeAdded, 0));
        else
            nullInTable.setConcentration(0);
      }
    }
    rescaleConcentrations(chemicals.values(), paramVal("SOIL_CHEMICAL_CONCENTRATION_LIMIT"));
  }
  
  /**
   * Removes some portion (or all) of a chemical's concentration
   * @param name the name of the chemical to be diminished
   * @param concentrationToRemove the concentration to be take out
   */
  public void removeChemical(String name, double concentrationToRemove){
    if(!chemicals.containsKey(name)) return;
    Chemical C = (Chemical)chemicals.get(name);
    double actualValueToBeRemoved = Math.min(concentrationToRemove, C.getConcentration()); 
    C.setConcentration(C.getConcentration() - actualValueToBeRemoved);
    Chemical N = (Chemical)chemicals.get(Soil.NULL_CHEMICAL_NAME);
    N.setConcentration(N.getConcentration() + actualValueToBeRemoved); // This pads the chemicals back up to the full amount, including null    
  }
  
  
  /**
   * Evaporates a portion of the water contained in this soil.
   * Chooses evaporation routine based on code model switch.
   */
  public void evaporate(){
    
    if(getWaterAvailableMM() == 0) return;
    switch(codeModelOption("SOIL_EVAPORATION")){
      case BASIC_EVAPORATION:{
        double mmToEvaporate = Math.min(getWaterAvailableMM(), paramVal("SOIL_EVAPORATE_FLAT_RATE"));
        Water w = extractWaterMM(mmToEvaporate);
        List<Chemical> l = new ArrayList<Chemical>();
        
        if(w.getChemicals().size() > 0){
          Enumeration E = w.getChemicals().elements();
          while(E.hasMoreElements()){
            Chemical C = (Chemical)E.nextElement();
            C.setConcentration(C.getConcentration() * (mmToEvaporate / soilDepth_mm) * (paramVal("SOIL_CHEMICAL_EVAP_CONV_RATIO")));
            l.add(C);
          }
           addChemicals(l);
        }
        break;
      }
      case NO_EVAPORATION:
      default:
        break;
    }
  }
  
  /**
   * Seeps some of the water out of this soil
   */
  public void seep(){
    
    switch(codeModelOption("SEEP_PROCEDURE")){
      case BASIC_SEEP_PROCEDURE:{
        if(getWaterAvailableMM() == 0) return; // If no water, done!
        System.out.println("SEEPING...");
        addChemicalsFromSeep(                                                 // Add the chemicals from the seeping water
            extractWaterMM(                                                     // Remove the seeping water from the water object
                Math.min(getWaterAvailableMM(), paramVal("SOIL_SEEP_FLAT_RATE")))); // How much water
        break;
      }
      case NO_SEEPAGE:
      default:{ // NO SEEP
//        System.out.println("No seepage occurring");
      }
    }
  }
  
  /**
   * Seeps some water; adjusts the water content to include the 'null' chemical
   * if the chemical content in the water is below the 'clean' limit.
   * @param w
   */
  private void addChemicalsFromSeep(Water w){
    double seepageConversionRatio         = paramVal("SOIL_CHEMICAL_SEEP_CONV_RATIO");
    double seepageConcentrationCleanLimit = paramVal("SOIL_CHEMICAL_SEEP_CLEAN_LIMIT");
    
    // First, calculate the total concentration of chemicals in the water seeping
    // If this is UNDER the 'CLEAN' limit, add 'NULL' chemical to bump it up
    // This implies that whenever water seeps, it has the capacity, if it is not
    // too laden with chemicals, to remove some chemicals from the soil
    Enumeration E1 = w.getChemicals().elements();
    
    // Get the concentration in the water to be seeped
    double total = 0;
    while(E1.hasMoreElements())  total += ((Chemical)E1.nextElement()).getConcentration();
    
    // Is it less than the 'clean' limit? If so, add the null chemical in, up to the limit
    if(total < seepageConcentrationCleanLimit){
      w.addChemical(new BasicChemical(NULL_CHEMICAL_NAME, seepageConcentrationCleanLimit - total));
      System.out.println("Seepage concentration is below clean limit " + total + " " + seepageConcentrationCleanLimit);
    }
    else
        System.out.println("Seepage concentration is above clean limit " + total + " " + seepageConcentrationCleanLimit);
    
    
    // Now, loop through the list of chemicals in the water
    // and add these to the chemicals in the soil
    // The way this is done uses a 'conversion ratio'
    
    Enumeration E2 = w.getChemicals().elements();
    List<Chemical> l = new ArrayList<Chemical>();
    while(E2.hasMoreElements()){
      Chemical C = (Chemical)E2.nextElement();
      // Calculate the concentration to be added.
      // The formula for this is:
      //
      //  take depth of water to be seeped out in mm    (equivalent to volume of the water object passed * 1000)
      //  divide by depth of soil
      //  then multiply by a conversion factor
      C.setConcentration(C.getConcentration() * (((w.getVolume() * 1000) / this.soilDepth_mm) * seepageConversionRatio));
//      System.out.println("Adding chemical " + C.getName() + " in concentration " + C.getConcentration() + " in seep " + seepageConversionRatio + " " + w.getVolume());
      l.add(C);
    }
    addChemicals(l);
  }
  
  /**
   * Rescales the concentrations of chemicals passed in the collection to the specified limit. Concentrations
   * remain in the same proportion but scaled to the different ceiling. 
   * @param chemicalsToRescale collection of chemicals to adjust
   * @param limit ceiling on total concentration
   */
  private void rescaleConcentrations(Collection chemicalsToRescale, double limit){
    System.out.println("\nRescaling to limit of " + limit);
    double total = 0;
    for(Object chemical: chemicalsToRescale) total+= ((Chemical)chemical).getConcentration();
    System.out.println("Original total " + total);

    double newTotal = 0;
    for(Object chemicalObject: chemicalsToRescale){
      Chemical chem = (Chemical)chemicalObject;
      System.out.print(((chem).getName() + "                          ").substring(0,30) + " @ " + chem.getConcentration());
      chem.setConcentration(chem.getConcentration() * (limit / total));
      System.out.println(" newval = " + chem.getConcentration());
      newTotal += chem.getConcentration();
    }
    System.out.println("New total: " + newTotal + "\n");
    
  }
  
  
  
  /**
   * Adds a specific depth of water to the soil object.
   * @param waterToAdd A model water object to be added; the actual
   * water added will have the same composition as the model
   * @param MM millimeters of water to be added
   */
  public void addWaterMM(Water waterToAdd, double MM){
    double current = this.getWaterAvailableMM();
    if(current < soilDepth_mm) water.mixIn(new Water( (Math.min(soilDepth_mm - current, MM) / 1000d), waterToAdd));  
  }
  
  /**
   * Takes some of the water out of this soil object.
   * @param MM Volume of water to be taken out, expressed as millimeters
   * @return the volume of water actually extracted; may
   * be less than the requested volume
   */  
  public Water extractWaterMM(double MM){                 return water.removePortion(MM / 1000);              }
  
  /**
   * Gets the amount of water available, in mm.
   * @return mm of water available
   */
  public double getWaterAvailableMM(){                    return water.getVolume() * 1000;                   }
  
  /**
   * Gets a printable report of the status of this object.
   * @return a string object that includes a report of the status of this object.
   */  
  public String getReport(){
    ParameterSet P = this.getParameterSet();
    boolean debug = (P.getParameter("DEBUG") != 0);
    int rdp = (int)(Math.floor(P.getParameter("ROUNDING_DECIMAL_PLACES")));
  
    String vol      = (debug ? TextFormatter.roundWithUnits(water.getVolume(), rdp, "m^3") : "" + water.getVolume()) + "\n";
    String depth    = (debug ? TextFormatter.roundWithUnits(soilDepth_mm, rdp, "m") : "" + soilDepth_mm ) + "\n";
    
    String retVal = "";
    retVal += "  Soil Water Content: " + vol + "\n";
    retVal += "  Soil Depth:         " + depth + "\n";
    retVal += "  Chemicals:          " + chemicals.size();
    Enumeration e = chemicals.elements();
    while(e.hasMoreElements()){
      Chemical c = (Chemical)e.nextElement();
      String con = (debug ? TextFormatter.roundWithUnits(c.getConcentration(), rdp, "ppm") : "" + c.getConcentration());
      retVal += "      " + c.getName() + " - " + con + "\n";
    }
    retVal += "\n";
    return retVal;
  }
  
  

  


  
  /* Getters for soil conditions */

  /**
   * Gets the depth of this soil object, in meters
   * @return the depth of this soil object
   */
  public double getSoilDepth(){    return soilDepth_mm / 1000d;       }

  
  /**
   * Returns the percent of water that this soil can hold that it is actually
   * holding
   * @return the water available divided by the soil depth
   */
  public double getSaturation(){
    return getWaterAvailableMM() / this.soilDepth_mm;
  }
  

  /**
   * Gets the concentration in this soil of the specified chemical.
   * @param chem the chemical requested
   * @return the concentration of the requested chemical
   */
  public double getChemicalConcentration(Chemical chem){
    return (this.chemicals.containsKey(chem.getName()) ? ((Chemical)chemicals.get(chem.getName())).getConcentration() : 0);
  }
  
  /**
   * Gets the concentration in this soil of the specified chemical
   * @param chem the name of the chemical to be returned 
   * @return the concentration of that chemical in this soil
   */
  public double getChemicalConcentration(String chem){
    return (this.chemicals.containsKey(chem) ? ((Chemical)chemicals.get(chem)).getConcentration() : 0);
  }

  
  /* Representations */
  /**
   * Gets a color to use when representing a field with this soil.
   * The color will be based on soil type and moisture.
   * @return Color to be used when displaying this soil
   */
  public Color getColor(){
    double waterAvail = getWaterAvailableMM();
    if(waterAvail <= soilDepth_mm) {
      int opacity = (int)(Math.floor((255 * (waterAvail / soilDepth_mm))));
      return new Color((int)165,(int)42,(int)42,opacity);
    }
    else return new Color(0, 0, 128, 128);
    
  }

  
  /* Object Cloning */
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Soil clone = new Soil();
    cloneStore.put(this, clone);
    clone.chemicals = (ABCMDeepCloneableHashtable<String, Chemical>)chemicals.getDeepClone(cloneStore);
    clone.soilDepth_mm = this.soilDepth_mm;
    clone.water = (Water)cloneStore.getClone(this.water);
    return clone;
  }
  

  
}
