package flowag.channels;


import java.util.Vector;

import abcm.math.BasicMath;
import abcm.math.Point;
import abcm.sim.AbstractModelComponent;
import flowag.basic.Global;

/**
 * Any object that implements a ChannelCrossSection interface must provide
 * a set of values that represent the hydraulic characteristics of that
 * cross section. These are, generally, represented as functions f(z), where
 * z is the depth of water as measured from the bottom of the cross section.</br></br>
 * 
 * The programmer can represent the internal channel geometry in any way
 * desired. The interface requires that top width, bottom width, and channel
 * depth be retrievable as generic variables; for some channel geometries
 * the bottom width will be zero by definition. Beyond this, however, the 
 * cross-sectional shape of the channel can be represented in any way provided
 * that the hydraulic values are provided in the specified ways.</br></br>
 * 
 * All units are meters.</br></br>
 * 
 * For creating channels using these cross sections, it may be
 * easiest to add corresponding methods to the ChannelSegmentFactory, but this
 * is only a convenience.</br></br>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractChannelCrossSection extends AbstractModelComponent implements ChannelCrossSection{

  /**
   * Returns the dK/dZ value given the specified water depth.
   * @param waterDepth the depth of water for which the dK/dZ value is to be calculated
   * @param roughness Manning's roughness coefficient to be used in this calculation
   * @return the dK/dZ value at the specified water depth
   */
  public double getDKDz(double waterDepth, double roughness) {
    // TODO Error trap getDKDZ if water depth is too low or high
    double e = .01; // A small epsilon value
    return (getConveyance(waterDepth + e, roughness) - getConveyance(waterDepth - e, roughness))/(2*e);
  }
  
  /**
   * Get the critical depth given the specified discharge.
   * @param discharge the requested discharge
   * @return the critical depth given the specified discharge.
   */
  public double getCriticalDepth(double discharge){
    double minY = .000001;
    double maxY = 100;
    double interval;
    double result = 0;
    double minEnergy = 0;
    do{
      interval = (maxY - minY)/50;
      double lowestResult = 10000;
      double newMinY = 0;
      double newMaxY = 0;
      
      for(double test = minY; test < maxY; test+= interval){
        result = getEnergy(discharge, test);
        if(result < lowestResult){
          newMinY = test - interval;
          newMaxY = test + interval;
          lowestResult = result;
          minEnergy = test;
        }
      }
      minY = newMinY;
      maxY = newMaxY;
    }while(interval > .001);
    return minEnergy;
  }
  
  /**
   * @see flowag.channels.ChannelCrossSection#getEnergy(double, double)
   */
  public double getEnergy(double discharge, double depth){
    double A = getFlowArea(depth);
    double V = discharge/A;
    return depth + (V * V) / (2d * Global.AccelerationDueToGravity);
  }
  
  /**
   * @see flowag.channels.ChannelCrossSection#getDepthPerManningDischarge(double, double, double)
   */
  public double getDepthPerManningDischarge(double Q, double roughness, double slope){
    if(Q == 0) return 0;
    double z = this.getChannelHeight();
    double testQ = this.getManningDischarge(z, roughness, slope);
    if(testQ < Q) return -1;
    double epsilon = .00001;
    double interval = z/2d;
    while(Math.abs(testQ - Q) > epsilon){
      testQ = this.getManningDischarge(z - interval, roughness, slope);
      if(testQ > (Q - epsilon)) z = z - interval;
      interval /= 2d;
    }
    double check = this.getManningDischarge(z, roughness, slope);
//    System.out.println("DPMD: Calc: " + z + " for Q of " + Q + " CHECK: " + check + " DIFF: " + (Q - check));
    return z;
  } 
  
  /**
   * @see flowag.channels.ChannelCrossSection#calculateWaterDepthFromSpecificForce(double, double)
   */
  public double calculateWaterDepthFromSpecificForce(double SF, double Q){
    double tolerance = .01;
  
    double depth = getChannelHeight();
    double interval = depth;
  
    double e = tolerance + 1;
    while(e > tolerance){
      double area            = getFlowArea(depth);
      double centroidDepth   = centroidDepth(depth);
      double depthToCentroid = depth - centroidDepth;
      double calculatedSF = Math.pow(Q, 2) / (Global.AccelerationDueToGravity * area) + area * depthToCentroid;
      e = Math.abs(SF - calculatedSF);
      if(e > tolerance){
        if(calculatedSF > SF){
          if(interval == depth){
            depth   *= 2;
            interval = depth;
          }
          else{
            interval /= 2;
            depth += interval;
          }
        }
        else{
          interval /= 2;
          depth -= interval;
        }
      }
    }    
    return depth;
  }
  
  
  /* HEC Output */
  
  /**
   * @see flowag.channels.ChannelCrossSection#report()
   */
  public String report(){
    String s = "";
    s += "Top Width:      " + getTopWidth() + "\n";
    s += "Bottom Width:   " + getBottomWidth() + "\n";
    s += "Channel Height: " + getChannelHeight() + "\n";
    s += "Points:\n";
    
    Vector CSPoints = getCutAndSurfaceLine(new Point(0,0,0), 0);
    
    s+= "Cut-and-Surface line for origin = (0,0,0) and heading = 0 (All y's are zero)\n";
    
    for(int i = 0; i < CSPoints.size(); i++){
      s+= "   [" + BasicMath.roundDouble(((Point)CSPoints.get(i)).getX(),2) + ", " + BasicMath.roundDouble(((Point)CSPoints.get(i)).getZ(),2) + "]\n"; 
    }
    
//    for(double x = 0; x < getChannelHeight(); x+= getChannelHeight()/20){
//      s+= "x = " + x + "\n";
//      s+= "   WaterSurfaceWidth     = " + getWaterSurfaceWidth(x) + "\n";
//      s+= "   FlowArea              = " + getFlowArea(x) + "\n";
//      s+= "   WettedPerimeter       = " + getWettedPerimeter(x) + "\n";
//      s+= "   AverageDepth          = " + getAvgDepth(x) + "\n";
//      s+= "   HydraulicRadius       = " + getHydraulicRadius(x) + "\n";
//      
//      s+= "   Conveyance  (.01)     = " + getConveyance(x,.01) + "\n";
//      s+= "   dA/dz                 = " + getDADz(x) + "\n";
////      s+= "   dK/Dz                 = " + getDKDz(x, .01) + "\n";
////      s+= "   Frict. Slope (10,.01) = " + getFrictionSlope(10,x,.01) + "\n";
////      s+= "   Manning Vel.          = " + getManningVelocity(x, .01, .01) + "\n";
////      s+= "   Manning Disch.        = " + getManningDischarge(x, .01, .01) + "\n";
////      s+= "   Critical Depth (10)   = " + getCriticalDepth(10) + "\n";
//      
//      
//    }
    
    return s;
  }
  
}
