package flowag.channels;

import abcm.math.Geometry;
import abcm.math.Point;
import abcm.math.SimDate;
import abcm.sim.AbstractModelComponent;
import flowag.basic.Global;

/**
 * Parent class for a family of Channel Segment objects. Provides operations that
 * all channel segments are expected to perform. In particular, maintains head and end coordinates,
 * upstream and downstream discharge, roughness, basic hydrograph functions, and functions related to channel
 * geometry such as slope.
 *   
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractChannelSegment extends AbstractModelComponent implements ChannelSegment {

  /** Coordinates of the upstream end of this segment (at cross section bottom) */ 
  private Point headCoords;
  /** Coordinates of the downstream end of this segment (at cross section bottom)*/ 
  private Point endCoords;
  /** Calculated slope between head and end coordinates (at cross section bottom)*/ 
  private double bedSlope;
  
  
  /** Heading from head coordinates to end coordinates */ protected double h;


  /** Channel Cross Section at upstream  end  */ protected ChannelCrossSection upstreamCrossSection;
  /** Channel Cross Section at downstream end */ protected ChannelCrossSection downstreamCrossSection;
  
  /** Manning's roughness coefficient*/ protected double roughness;
  
  /** Current upstream discharge                                     */ 
  protected double Q1;
  /** Current downstream discharge                                   */ 
  protected double Q2;
  /** Current upstream stage (absolute elevation, not water depth)   */ 
  protected double z1;
  /** Current downstream stage (absolute elevation, not water depth) */ 
  protected double z2;

  /** Useful constant equivalent to turning heading left  */ protected static double LEFT  = -1 *  Math.PI / 2d;
  /** Useful constant equivalent to turning heading right */ protected static double RIGHT =       Math.PI / 2d;
  
  // Inner class to manage hydrographs
  /** */ protected SingleDayHydrographEngine hydrographEngine = new SingleDayHydrographEngine();
  
  /**
   * Basic constructor requiring instance information
   * @param HeadCoordinates upstream end
   * @param EndCoordinates downstream end coordinates
   * @param n roughness
   * @param upstreamCross cross section at the upstream end of this segment
   * @param downstreamCross cross section at the downstream end of this segment
   */
  public AbstractChannelSegment(Point HeadCoordinates, Point EndCoordinates, double n, 
      ChannelCrossSection upstreamCross, ChannelCrossSection downstreamCross){
    headCoords             = HeadCoordinates;
    endCoords              = EndCoordinates;
    bedSlope               = Geometry.slope(headCoords, endCoords);
    roughness              = n;
    upstreamCrossSection   = upstreamCross;
    downstreamCrossSection = downstreamCross;
    h                      = headCoords.headingTo(endCoords);
    clearHydrographs();
    setUpstreamDepth(0);
    setDownstreamDepth(0);
  }
  
  
  /**
   * Resets the head and end coordinates for this segment
   * @param h new head coordinates
   * @param e new end coordinates
   */
  protected void setHeadAndEndCoordinates(Point h, Point e){
    headCoords = h;
    endCoords  = e;
    bedSlope = Geometry.slope(h, e);
  }
 
  
  /**
   * Gets the Point object that represents the center of the bottom
   * of the channel segment at the upstream end.
   * @return current Head Coordinates
   */
  public Point getHeadCoords()                   { return headCoords; }
  
  /**
   * Gets the Point object that represents the center of the bottom
   * of the channel segment at the downstream end.
   * @return current End Coordinates
   */
  public Point getEndCoords()                    { return endCoords;  }

  
  /**
   * Gets the Point object that represents the upstream corner of the left bank
   * @return the upstream left bank coordinate
   */
  public Point getUpstreamLeftBankCoordinate(){
    Point P = Geometry.getOffsetPoint(headCoords, h + LEFT,  upstreamCrossSection.getTopWidth()   / 2d);
    return new Point(P.getX(), P.getY(), headCoords.getZ() + upstreamCrossSection.getChannelHeight());
  }
  
  /**
   * Gets the Point object that represents the upstream corner of the right bank
   * @return the upstream right bank coordinate
   */
  public Point getUpstreamRightBankCoordinate(){
    Point P = Geometry.getOffsetPoint(headCoords, h + RIGHT, upstreamCrossSection.getTopWidth()   / 2d);
    return new Point(P.getX(), P.getY(), headCoords.getZ() + upstreamCrossSection.getChannelHeight());
  }
  
  /**
   * Gets the Point object that represents the downstream corner of the left bank
   * @return the downstream left bank coordinate
   */
  public Point getDownstreamLeftBankCoordinate(){
    Point P = Geometry.getOffsetPoint(endCoords, h + LEFT,  downstreamCrossSection.getTopWidth()   / 2d);
    return new Point(P.getX(), P.getY(), endCoords.getZ() + downstreamCrossSection.getChannelHeight());
  }
  
  /**
   * Gets the Point object that represents the downstream corner of the right bank
   * @return the downstream right bank coordinate
   */
  public Point getDownstreamRightBankCoordinate(){
    Point P = Geometry.getOffsetPoint(endCoords, h + RIGHT, downstreamCrossSection.getTopWidth()   / 2d);
    return new Point(P.getX(), P.getY(), endCoords.getZ() + downstreamCrossSection.getChannelHeight());
  }
  
  
  
/**
   * Sets the current upstream discharge.
   * @param Q new value for discharge, in m^3/s
   */
  public void setUpstreamDischarge(double Q)     { Q1 = Q; }
  
  /**
   * Sets the current downstream discharge.
   * @param Q new value for discharge, in m^3/s
   */
  public void setDownstreamDischarge(double Q)   { Q2 = Q; }
  
  /**
   * Sets the current upstream stage.
   * @param z new value for stage, in m.
   */
  public void setUpstreamStage(double z)         { z1 = z; }
  
  /**
   * Sets the current downstream stage.
   * @param z new value for stage, in m.
   */
  public void setDownstreamStage(double z)       { z2 = z; }
  
  /**
   * Sets the Manning roughness coefficient for the entire length of the channel.
   * @param n new value for roughness coefficient
   */
  public void setRoughness(double n)             { roughness = n; }

  /**
   * Gets the value of the upstream discharge rate.
   * @return current value of upstream discharge rate, in m^3/s
   */
  public double getUpstreamDischarge()           { return Q1; }
  
  /**
   * Gets the value of the downstream discharge rate.
   * @return current value of downstream discharge rate, in m^3/s
   */
  public double getDownstreamDischarge()         { return Q2; }
  
  /**
   * Gets the value of the current upstream water stage.
   * @return current value of upstream stage, in m above a datum
   */
  public double getUpstreamStage()               { return z1; }
  
  /**
   * Gets the value of the current downstream water stage.
   * @return current value of downstream stage, in m above a datum
   */
  public double getDownstreamStage()             { return z2; }
  
  /**
   * Gets the current value of the roughness coefficient along the length of the channel.
   * @return Manning's n for this channel
   */
  public double getRoughness()                   { return roughness; }

  
  
  /* Allow manipulation by depth instead of stage  */
  
  /**
   * Sets the current upstream stage, but does so using a depth
   * above the Head Coordinates' z-value. For example, if the HeadCoordinate
   * is at an elevation of z = 4m, and this method is called with
   * a depth of 1.2m, the stage will be set to 5.2m.
   * @param newDepth depth of the water by which the stage will be set for the upstream
   * cross section
   */
  public void setUpstreamDepth(double newDepth)  { setUpstreamStage(getHeadCoords().getZ() + newDepth); }
  
  /**
   * Sets the current downstream stage, but does so using a depth
   * above the End Coordinates' z-value. For example, if the EndCoordinate
   * is at an elevation of z = 4m, and this method is called with
   * a depth of 1.2m, the stage will be set to 5.2m.
   * @param newDepth depth of the water by which the stage will be set for the upstream
   * cross section
   */
  public void setDownstreamDepth(double newDepth){ setDownstreamStage(getEndCoords().getZ() + newDepth); }
  
  /**
   * Gets the depth of the water presuming the water is at a given
   * stage at the upstream cross-section.
   * @param stage the stage for which a depth value should be calculated 
   * @return the depth of the water at the given stage, equal to stage argument
   * passed minus the z-value of the Head Coordinates
   */
  public double getUpstreamWaterDepth(double stage)      { return stage - getHeadCoords().getZ();                      }
  
  /**
   * Gets the depth of the water presuming the water is at a given
   * stage at the downstream cross-section.
   * @param stage the stage for which a depth value should be calculated 
   * @return the depth of the water at the given stage, equal to stage argument
   * passed minus the z-value of the End Coordinates
   */
  public double getDownstreamWaterDepth(double stage)    { return stage - getEndCoords().getZ();                       }
  
  /**
   * Gets the depth of the water at the upstream cross-section given
   * the current upstream water stage. 
   * @return the depth of the water, equal to the current stage minus
   * the z-value of the Head Coordinates
   */
  public double getUpstreamWaterDepth()                  { return getUpstreamWaterDepth(getUpstreamStage());                   }
  
  /**
   * Gets the depth of the water at the downstream cross-section given
   * the current upstream water stage. 
   * @return the depth of the water, equal to the current stage minus
   * the z-value of the End Coordinates
   */
  public double getDownstreamWaterDepth()                { return getDownstreamWaterDepth(getDownstreamStage());                 }
  
  /* Get values determine by location in space  */
  
  /**
   * Gets the current slope of the bed of this channel segment. The slope
   * is calculated as the change in elevation from the HeadCoordinates to the
   * EndCoordinates divided by the planar distance between them.
   * @return bed slope of this channel segment.
   */
  public double getBedSlope()                            { return bedSlope; }
  
  /**
   * Gets the length of the channel segment. The length is the planar
   * distance between the Head and End Coordinates; elevation change
   * is ignored. Example: if the
   * HeadCoordinates = (0,0,0) and the EndCoordinates = (100,0,-.5),
   * the length returned is 100.00. 
   * @return the length of the channel in the X,Y plane
   */
  public double getLength()                              { return getHeadCoords().planarDistanceToPoint(getEndCoords());    }
  
  /* Construction Volume and Area values */
  
  /**
   * Gets the total construction volume of this channel segment. Note
   * that this is not the volume of water, but is the volume of the
   * segment as constructed.
   * @return the total construction volume of the ChannelSegment
   */
  public double getTotalVolume(){
    return(getLength() * ((getUpstreamArea() + getDownstreamArea())/ 2d));    
  }
  
  /**
   * Gets the bottom width of the upstream cross section.
   * @return bottom width of the upstream cross section
   */
  public double getUpstreamBottomWidth()         { return   upstreamCrossSection.getBottomWidth();               }
  
  /**
   * Gets the top width of the upstream cross section.
   * @return top width of the upstream cross section
   */
  public double getUpstreamTopWidth()            { return   upstreamCrossSection.getTopWidth();                  }
  
  /**
   * Gets the channel height of the upstream cross section.
   * @return channel height of the upstream cross section
   */
  public double getUpstreamChannelHeight()       { return   upstreamCrossSection.getChannelHeight();             }
  
  /**
   * Gets the area of the upstream cross section. This is the construction
   * area, not the flow area.
   * @return construction area of the upstream cross section
   */
  public double getUpstreamArea()                { return   upstreamCrossSection.getArea();                      }

  /**
   * Gets the bottom width of the downstream cross section.
   * @return bottom width of the downstream cross section
   */
  public double getDownstreamBottomWidth()       { return downstreamCrossSection.getBottomWidth();               }
  
  /**
   * Gets the top width of the downstream cross section.
   * @return top width of the downstream cross section
   */
  public double getDownstreamTopWidth()          { return downstreamCrossSection.getTopWidth();                  }
  
  /**
   * Gets the channel height of the downstream cross section.
   * @return channel height of the downstream cross section
   */
  public double getDownstreamChannelHeight()     { return downstreamCrossSection.getChannelHeight();             }
  
  /**
   * Gets the area of the upstream cross section. This is the construction
   * area, not the flow area.
   * @return construction area of the upstream cross section
   */
  public double getDownstreamArea()              { return downstreamCrossSection.getArea();                      }
  
  
  
  /**
   * Gets the width of the water surface, given current stage, at the upstream
   * cross section.
   * @return water surface width at the upstream cross section
   */
  public double getUpstreamWaterSurfaceWidth()   { return   upstreamCrossSection.getWaterSurfaceWidth(getUpstreamWaterDepth());              }
  
  /**
   * Gets the average depth, given current stage, at the upstream
   * cross section.
   * @return average depth at the upstream cross section
   */
  public double getUpstreamAvgDepth()            { return   upstreamCrossSection.getAvgDepth(getUpstreamWaterDepth());                       }
  
  /**
   * Gets the flow area, given current stage, at the upstream
   * cross section.
   * @return flow area at the upstream cross section
   */
  public double getUpstreamFlowArea()            {
    return   upstreamCrossSection.getFlowArea(getUpstreamWaterDepth());                       
  }
  
  /**
   * Gets the upstream cross sectional area of water flow assuming the given depth
   * @param waterDepth depth of the water to be assumed
   * @return cross sectional area of flow
   */
  public double getUpstreamFlowArea(double waterDepth){
    return upstreamCrossSection.getFlowArea(waterDepth);
  }
  
  /**
   * Gets the wetted perimeter, given current stage, at the upstream
   * cross section.
   * @return wetted perimeter at the upstream cross section
   */
  public double getUpstreamWettedPerimeter()     { return   upstreamCrossSection.getWettedPerimeter(getUpstreamWaterDepth());                }
  
  /**
   * Gets the hydraulic radius, given current stage, at the upstream
   * cross section.
   * @return hydraulic radius at the upstream cross section
   */
  public double getUpstreamHydraulicRadius()     { return   upstreamCrossSection.getHydraulicRadius(getUpstreamWaterDepth());                }
  
  /**
   * Gets the dA/dZ value, given current stage, at the upstream
   * cross section.
   * @return dA/dZ value at the upstream cross section
   */
  public double getUpstreamDADz()                { return   upstreamCrossSection.getDADz(getUpstreamWaterDepth());                           }
  
  /**
   * Gets the conveyance, given current stage, at the upstream
   * cross section.
   * @return conveyance value at the upstream cross section
   */
  public double getUpstreamConveyance()          { return   upstreamCrossSection.getConveyance(getUpstreamWaterDepth(), roughness);          }
  
  /**
   * Gets the dK/dZ value, given current stage, at the upstream
   * cross section.
   * @return dK/dZ value at the upstream cross section
   */
  public double getUpstreamDKDz()                { return   upstreamCrossSection.getDKDz(getUpstreamWaterDepth(), roughness);                }


  /**
   * Gets the width of the water surface, given current stage, at the downstream
   * cross section.
   * @return water surface width at the downstream cross section
   */
  public double getDownstreamWaterSurfaceWidth() { return downstreamCrossSection.getWaterSurfaceWidth(getDownstreamWaterDepth());            }
  
  /**
   * Gets the flow area, given current stage, at the downstream
   * cross section.
   * @return flow area at the downstream cross section
   */
  public double getDownstreamAvgDepth()          { return downstreamCrossSection.getAvgDepth(getDownstreamWaterDepth());                     }
  
  /**
   * Gets the flow area, given current stage, at the downstream
   * cross section.
   * @return flow area at the downstream cross section
   */
  public double getDownstreamFlowArea()          { return downstreamCrossSection.getFlowArea(getDownstreamWaterDepth());                     }
  
  /**
   * Gets the cross-sectional area of flow at the downstream end, assuming the
   * specified depth of water
   * @param waterDepth depth of water
   * @return cross-sectional area of flow at downstream end of segment
   */
  public double getDownstreamFlowArea(double waterDepth){ return downstreamCrossSection.getFlowArea(waterDepth); }
  
  /**
   * Gets the wetted perimeter, given current stage, at the downstream
   * cross section.
   * @return wetted perimeter at the downstream cross section
   */
  public double getDownstreamWettedPerimeter()   { return downstreamCrossSection.getWettedPerimeter(getDownstreamWaterDepth());              }
  
  /**
   * Gets the hydraulic radius, given current stage, at the downstream
   * cross section.
   * @return hydraulic radius at the downstream cross section
   */
  public double getDownstreamHydraulicRadius()   { return downstreamCrossSection.getHydraulicRadius(getDownstreamWaterDepth());              }
  
  /**
   * Gets the dA/dZ value, given current stage, at the downstream
   * cross section.
   * @return dA/dZ value at the downstream cross section
   */
  public double getDownstreamDADz()              { return downstreamCrossSection.getDADz(getDownstreamWaterDepth());                         }
  
  /**
   * Gets the conveyance, given current stage, at the downstream
   * cross section.
   * @return conveyance at the downstream cross section
   */
  public double getDownstreamConveyance()        { return downstreamCrossSection.getConveyance(getDownstreamWaterDepth(), roughness);        }
  
  /**
   * Gets the dK/dZ value, given current stage, at the downstream
   * cross section.
   * @return dK/dZ value at the downstream cross section
   */
  public double getDownstreamDKDz()              { return downstreamCrossSection.getDKDz(getDownstreamWaterDepth(), roughness);              }
  
  /* Empirical hydraulic values (based on Manning's equation) */
  
  /**
   * Gets the velocity as calculated by the Manning equation
   * given the specified stage, passed as an argument, at the upstream cross section.
   * @param stage water stage to be used
   * @return the velocity as calculated by Manning's equation, given
   * the current stage at the upstream cross section.
   */
  public double getUpstreamManningVelocity(double stage){
    return upstreamCrossSection.getManningVelocity(getUpstreamWaterDepth(stage), roughness, getBedSlope());
  }
  
  /**
   * Gets the discharge as calculated by the Manning equation
   * given the specified stage, passed as an argument, at the downstream cross section.
   * 
   * @param stage water stage to be used
   * @return the discharge as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getUpstreamManningDischarge(double stage){
    return upstreamCrossSection.getManningDischarge(getUpstreamWaterDepth(stage), roughness, getBedSlope());
  }
  
  /**
   * Gets the velocity as calculated by the Manning equation
   * given the specified stage, passed as an argument, at the downstream cross section.
   * @param stage water stage to be used
   * @return the velocity as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getDownstreamManningVelocity(double stage){
    return downstreamCrossSection.getManningVelocity(getDownstreamWaterDepth(stage), roughness, getBedSlope());
  }

  /**
   * Gets the discharge as calculated by the Manning equation
   * given the specified stage, passed as an argument, at the downstream cross section.
   * @param stage water stage to be used
   * @return the discharge as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getDownstreamManningDischarge(double stage){
//    System.out.println("Getting downstream manning discharge for stage " + stage + " which is depth " + getDownstreamWaterDepth(stage) + " n= " + this.roughness + " slope = " + this.getBedSlope());
    return downstreamCrossSection.getManningDischarge(getDownstreamWaterDepth(stage), roughness, getBedSlope());
  }
  
  /**
   * Gets the downstream Manning discharge assuming the water depth
   * is a fraction of the channel construction depth
   * @param fraction fraction to be used
   * @return value of discharge at the specified depth assuming Manning's equation
   */
  public double getDownstreamManningDischargePerFractionOfChannelHeight(double fraction){
//    System.out.println("Getting manning discharge for fraction " + fraction + " using endcoords");
    return getDownstreamManningDischarge(this.endCoords.getZ() + fraction * this.getDownstreamChannelHeight());
  }
  
  /**
   * @see flowag.channels.ChannelSegment#getUpstreamDepthPerManningDischarge(double)
   */
  public double getUpstreamDepthPerManningDischarge(double Q){
    return upstreamCrossSection.getDepthPerManningDischarge(Q, roughness, getBedSlope());
  }

  
  /**
   * @see flowag.channels.ChannelSegment#getDownstreamDepthPerManningDischarge(double)
   */
  public double getDownstreamDepthPerManningDischarge(double Q){
    return downstreamCrossSection.getDepthPerManningDischarge(Q, roughness, getBedSlope());
  }
  
  
  /**
   * @see flowag.channels.ChannelSegment#getFlowAreaAtPosition(double)
   */
  public double getFlowAreaAtPosition(double x){
    if(x < 0 || x > getLength()) return 0;
    double d = getDepthAtPosition(x);
    double fraction = x / getLength();
    return fraction      *   upstreamCrossSection.getFlowArea(d) + 
          (1 - fraction) * downstreamCrossSection.getFlowArea(d);
  }
  
  
  
  /**
   * @see flowag.channels.ChannelSegment#getTotalWater(double, double)
   */
  public double getTotalWater(double upstreamWaterDepth, double downstreamWaterDepth){  
    double A1 = upstreamCrossSection.getFlowArea(upstreamWaterDepth);
    double A2 = downstreamCrossSection.getFlowArea(downstreamWaterDepth);
    double A = (A1 + A2)/2;
    double length = getHeadCoords().planarDistanceToPoint(getEndCoords());
    return length * A;
  }
  
  
  
  
  
  
  /* Flow Volume and Area values */
  
  /**
   * Gets the average of the flow area at the upstream and downstream cross
   * sections given the current stages.
   * @return the average flow area of the two cross sections given their
   * current stages 
   */
  public double getAverageFlowArea()             { return (getUpstreamFlowArea() + getDownstreamFlowArea())/2.0d; }
  
  
  
  /* Calculations */ 
  
  /**
   * Gets the velocity head at the upstream cross section. The velocity head
   * is calculated as the square of the velocity divided by twice the acceleration
   * due to gravity (V^2 / 2g).
   * @return the velocity head at the upstream cross section
   */
  public double getUpstreamVelocityHead(){ 
    return Math.pow(getUpstreamFlowVelocity(),2)/(2 * Global.AccelerationDueToGravity);   
  }

    /**
   * Gets the Specific Energy at the upstream cross section. The Specific Energy
   * is the sum of the stage and the velocity head.
   * @return the Specific Energy at the upstream cross section.
   */
  public double getUpstreamSpecificEnergy(){
    return getUpstreamStage() + getUpstreamVelocityHead(); 
  }

  /**
   * Gets the velocity head at the downstream cross section. The velocity head
   * is calculated as the square of the velocity divided by twice the acceleration
   * due to gravity (V^2 / 2g).
   * @return the velocity head at the downstream cross section
   */
  public double getDownstreamVelocityHead(){ 
    return Math.pow(getDownstreamFlowVelocity(),2)/(2 * Global.AccelerationDueToGravity); 
  }
  
  /**
   * Gets the Specific Energy at the downstream cross section. The Specific Energy
   * is the sum of the stage and the velocity head.
   * @return the Specific Energy at the downstream cross section.
   */
  public double getDownstreamSpecificEnergy(){ 
    return getDownstreamStage() + getDownstreamVelocityHead(); 
  }
  
  /**
   * Gets the friction slope, given current discharge, at the upstream
   * cross section. Note that this is NOT the same as the instantaneous
   * friction slope calculated for the cross section, which is derived from Manning's
   * equation. This is the change in Specific Energy over the length of the
   * segment, (SE1 - SE2) / x. Note that this will be the same for both
   * the upstream and downstream cross sections, but 
   * @return the friction slope of the upstream cross section
   */
  public double getFrictionSlope(){
    return (getUpstreamSpecificEnergy() - getDownstreamSpecificEnergy()) / getLength();
  }
  
  /**
   * Stub to return the friction slope for the 'upstream' cross section.
   * @return the friction slope of this segment
   */
  public double getUpstreamFrictionSlope(){ 
    return getFrictionSlope();
  }
  
  /**
   * Stub to return the friction slope for the 'downstream' cross section.
   * @return the friction slope of this segment
   */
  public double getDownstreamFrictionSlope(){
    return getFrictionSlope();
  }
  
  /**
   * Gets the flow velocity at the upstream cross section, given
   * the current stage and discharge. This is the quotient of
   * the discharge divided by the flow area. Under conditions of 
   * normal flow this will equal the velocity as calculated by Manning's
   * equation. Returns zero if depth is zero.
   * @return current upstream flow velocity
   */
  public double getUpstreamFlowVelocity(){ 
    double d = getUpstreamFlowArea();
    return ((d != 0) ? getUpstreamDischarge() / d : 0);
  }
  
  /* Derived hydraulic values */
  
  // This is legacy code: confirm and delete...:
  
  // TODO Catch the bug in this section!!! note Q1 instead of Q2; did this affect results?
  // This was moved upward to the abstract class... check it there (it was 'corrected' there)
  
//  /**
//   * Gets the flow velocity at the downstream cross section, given
//   * the current stage and discharge. This is the quotient of
//   * the discharge divided by the flow area. Under conditions of 
//   * normal flow this will equal the velocity as calculated by Manning's
//   * equation. Returns zero if depth is zero.
//   * @return current downstream flow velocity
//   */
//  public double getDownstreamFlowVelocity(){ 
//    double d = downstreamCrossSection.getFlowArea(getDownstreamWaterDepth());
//    return ((d != 0) ? Q1 / d : 0);
//  }

  
  /**
   * Gets the flow velocity at the downstream cross section, given
   * the current stage and discharge. This is the quotient of
   * the discharge divided by the flow area. Under conditions of 
   * normal flow this will equal the velocity as calculated by Manning's
   * equation. Returns zero if depth is zero.
   * @return current downstream flow velocity
   */
  public double getDownstreamFlowVelocity(){ 
    double d = getDownstreamFlowArea();
    return ((d != 0) ? getDownstreamDischarge() / d : 0);
  }

/**
   * Gets the velocity as calculated by the Manning equation
   * given the current stage at the upstream cross section.
   * @return the velocity as calculated by Manning's equation, given
   * the current stage at the upstream cross section.
   */
  public double getUpstreamManningVelocity(){
    return getUpstreamManningVelocity(getUpstreamStage());
  }
  
  /**
   * Gets the discharge as calculated by the Manning equation
   * given the current stage at the downstream cross section.
   * @return the discharge as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getUpstreamManningDischarge(){
    return getUpstreamManningDischarge(getUpstreamStage());
  }

  /**
   * Gets the velocity as calculated by the Manning equation
   * given the current stage at the downstream cross section.
   * @return the velocity as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getDownstreamManningVelocity(){
    return getDownstreamManningVelocity(getDownstreamStage());
  }
  
  /**
   * Gets the discharge as calculated by the Manning equation
   * given the current stage at the downstream cross section.
   * @return the discharge as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getDownstreamManningDischarge(){
    return getDownstreamManningDischarge(getDownstreamStage());
  }
  
  /**
   * Sets the stage and discharge values for the upstream cross
   * section. The argument passed is the stage value; the discharge
   * is set to the value given by Manning's equation for normal
   * flow at that stage.
   * @param stage the stage value to be set, and from which discharge
   * is also calculated and set
   */
  public void setUpstreamStagePlusManningDischarge(double stage){
    setUpstreamStage(stage);
    setUpstreamDischarge(getUpstreamManningDischarge(stage));    
  }

  /**
   * Sets the stage and discharge values for the downstream cross
   * section. The argument passed is the stage value; the discharge
   * is set to the value given by Manning's equation for normal
   * flow at that stage.
   * @param stage the stage value to be set, and from which discharge
   * is also calculated and set
   */
  public void setDownstreamStagePlusManningDischarge(double stage){
    setDownstreamStage(stage);
    setDownstreamDischarge(getDownstreamManningDischarge(stage));
  }
  
  /**
   * Sets the stage and discharge values for the upstream cross
   * section. The argument passed is a depth value, from which
   * stage is calculated by adding the depth to the z value of the
   * Head Coordinates; the discharge is set to the value given by Manning's 
   * equation for normal flow at that stage.
   * @param depth the depth value from which the stage value to be set is calculated, 
   * after which discharge is also calculated and set
   */
  public void setUpstreamDepthPlusManningDischarge(double depth){
    setUpstreamDepth(depth);
    setUpstreamStagePlusManningDischarge(getUpstreamStage());
  }
  
  /**
   * Sets the stage and discharge values for the upstream cross
   * section. The argument passed is a depth value, from which
   * stage is calculated by adding the depth to the z value of the
   * End Coordinates; the discharge is set to the value given by Manning's 
   * equation for normal flow at that stage.
   * @param depth the depth value from which the stage value to be set is calculated, 
   * after which discharge is also calculated and set
   */
  public void setDownstreamDepthPlusManningDischarge(double depth){
    setDownstreamDepth(depth);
    setDownstreamStagePlusManningDischarge(getDownstreamStage());
  }  
  
  /* Intermediate values along segment axis */
  
  /**
   * Gets the stage at the position along the channel segment,
   * calculated from stage at head and end positions.
   * @param x distance from head coordinats at which stage is to be calculated
   * @return stage at the specified point
   */
  public double getStageAtPosition(double x){
    double z1 = getUpstreamStage();
    double z2 = getDownstreamStage();
    if(x < 0 || x > getLength()) return 0;
    return z1 + (x / getLength()) * (z2 - z1);
  }
  
  /**
   * @see flowag.channels.ChannelSegment#getBottomZAtPosition(double)
   */
  public double getBottomZAtPosition(double x){
    if(x < 0 || x > getLength()) return 0;
    return getHeadCoords().getZ() + (x/getLength()) * (getEndCoords().getZ() - getHeadCoords().getZ());
  }
  
  /**
   * 
   * @see flowag.channels.ChannelSegment#getDepthAtPosition(double)
   */
  public double getDepthAtPosition(double x){
    return getStageAtPosition(x) - getBottomZAtPosition(x); 
  }
  
  /**
   * 
   * @see flowag.channels.ChannelSegment#getWaterVolumeBetweenPositions(double, double)
   */
  public double getWaterVolumeBetweenPositions(double x1, double x2){
    if(x2 < x1){
      double swap = x1;
      x1 = x2;
      x2 = swap;
    }
    if(((x1 < 0) || (x1 > getLength()) ||
        (x2 < 0) || (x2 > getLength()))) return 0;
    return (getFlowAreaAtPosition(x1) + getFlowAreaAtPosition(x2)) / 2 * (x2 - x1);
  }

  /* Flow through the segment */
  
  /**
   * 
   * @see flowag.channels.ChannelSegment#averageFlowVelocity()
   */
  public double averageFlowVelocity(){
    return (getUpstreamFlowVelocity() + getDownstreamFlowVelocity()) / 2d;
  }

  /**
   * 
   * @see flowag.channels.ChannelSegment#timeToFlowThroughSegment()
   */
  public double timeToFlowThroughSegment(){
    double Vavg = averageFlowVelocity();
    // It is possible for the average flow velocity to be zero. and
    // if that's true, water does NOT flow through the
    // segment, hence this variable is irrelevant
    return ( Vavg != 0 ? getLength()/Vavg : 0);
  }
  
  /**
   * 
   * @see flowag.channels.ChannelSegment#flowAcceleration()
   */
  public double flowAcceleration(){
    // If the time to flow through the segment is zero, it means that
    // water is either not moving, or the canal is emptying in both
    // directions, or filling from both sides.
    double X = timeToFlowThroughSegment();
    return (timeToFlowThroughSegment() != 0 ? (getDownstreamFlowVelocity() - getUpstreamFlowVelocity()) / X : 0d); 
  }
  
  /**
   * 
   * @see flowag.channels.ChannelSegment#getTotalWater()
   */
  public double getTotalWater(){
    return getTotalWater(getUpstreamStage(), getDownstreamStage());
  }
  
  /* Hydrograph */
  
  /**
   * 
   * @see flowag.channels.Hydrographable#clearHydrographs()
   */
  public void clearHydrographs(){    
    hydrographEngine.clearHydrographs();  
  }
  
  /**
   * 
   * @see flowag.channels.Hydrographable#resetHydrographs()
   */
  public void resetHydrographs(){
    hydrographEngine.resetHydrographs();
  }
  
  /**
   * 
   * @see flowag.channels.Hydrographable#setUpstreamDischarge(double, abcm.math.SimDate)
   */
  public void setUpstreamDischarge(double Q, SimDate SD){
    setUpstreamDischarge(Q);
    hydrographEngine.setUpstreamDischarge(getUpstreamDischarge(), SD);
  }

  /**
   * 
   * @see flowag.channels.Hydrographable#setUpstreamStage(double, abcm.math.SimDate)
   */
  public void setUpstreamStage(double z, SimDate SD){
    setUpstreamStage(z);
    hydrographEngine.setUpstreamStage(getUpstreamStage(), SD);
  }
  
  /**
   * 
   * @see flowag.channels.Hydrographable#setDownstreamDischarge(double, abcm.math.SimDate)
   */
  public void setDownstreamDischarge(double Q, SimDate SD){
    setDownstreamDischarge(Q);
    hydrographEngine.setDownstreamDischarge(getDownstreamDischarge(), SD);
  }

  /**
   * 
   * @see flowag.channels.Hydrographable#setDownstreamStage(double, abcm.math.SimDate)
   */
  public void setDownstreamStage(double z, SimDate SD){
    setDownstreamStage(z);
    hydrographEngine.setDownstreamStage(getDownstreamStage(), SD);
  }
  
  /**
   * 
   * @see flowag.channels.Hydrographable#getUpstreamFlowAt(abcm.math.SimDate)
   */
  public double getUpstreamFlowAt(SimDate SD){
    return hydrographEngine.getUpstreamFlowAt(SD);
  }

  /**
   * 
   * @see flowag.channels.Hydrographable#getUpstreamStageAt(abcm.math.SimDate)
   */
  public double getUpstreamStageAt(SimDate SD){
    return hydrographEngine.getUpstreamStageAt(SD);
  }

  /**
   * 
   * @see flowag.channels.Hydrographable#getDownstreamFlowAt(abcm.math.SimDate)
   */
  public double getDownstreamFlowAt(SimDate SD){
    return hydrographEngine.getDownstreamFlowAt(SD);
  }

  /**
   * 
   * @see flowag.channels.Hydrographable#getDownstreamStageAt(abcm.math.SimDate)
   */
  public double getDownstreamStageAt(SimDate SD){
    return hydrographEngine.getDownstreamStageAt(SD);
  }
  
}
