package flowag.channels;

import java.util.Hashtable;
import java.util.Vector;

import abcm.sim.AbstractModelComponent;

/**
 * This abstract class implements a number of features that can be used by a family of descendant
 * classes to calculate flow through a connected set of canal segments. Among the implementation
 * details are standard ways to aggregate and use the results of these calculations, and to
 * apply them to the original canal system so that the calculated dz and dQ values are added
 * to the original values, thus updating the system.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public abstract class AbstractFlowCalculator  extends AbstractModelComponent implements FlowCalculator{

  
  /* Static methods */
  /**
   * Set the entire flow system to 'empty'
   * @param initialSegment the initial segment, from which to track downstream
   * @param results a result set that is configured as the system is tracked so that 
   * its dz and dQ values are those that are required to set the depth in the
   * actual system to zero 
   */
  public static void setToEmpty(CanalSegment initialSegment, FlowZAndQAdjustments results){
    results.setInitialUpstreamValues(initialSegment, -1 * initialSegment.getUpstreamWaterDepth(), -1 * initialSegment.getUpstreamDischarge());
    setDownstreamToEmpty(initialSegment, results);
  }
  
  /**
   * Sets all segments downstream of the given initial segment to empty, and adjusts
   * results to match required dz and dQ values. Recursively calls itself for all
   * downstream segments until end is reached
   * @param initialSegment initial segment from which to start
   * @param results results object to set to negative values for all encountered depths and flows
   */
  private static void setDownstreamToEmpty(CanalSegment initialSegment, FlowZAndQAdjustments results){
    if(!initialSegment.isEmpty())
        results.addDownstreamValues(initialSegment, -1 * initialSegment.getDownstreamWaterDepth(), -1 * initialSegment.getDownstreamDischarge());
    Vector V = initialSegment.getOutflowCanals();
    for(int i = 0; i < V.size(); i++) setDownstreamToEmpty((CanalSegment)V.get(i), results);
  }
  
  
  /** Initial canal segment */ 
  protected CanalSegment                         initialCanalSegment;
  /** Set of boundary conditions for downstream segments */ 
  protected Hashtable                            downstreamBoundaryConditions;
  /** Single downstream boundary condition; if others are not supplied, this will be applied to all */ 
  protected DownstreamBoundaryCondition          generalDownstreamBoundaryCondition;
  /** Upstream boundary condition to be applied to initial segment */ 
  protected UpstreamBoundaryCondition            ubc;

  /** Last calculated duration; default is Double.NaN */ 
  protected double                 lastCalculatedDuration   = Double.NaN;
  /** Set of results- values used to adjust z and Q values in actual system, and thus effect the changes required by these calculations */ 
  protected FlowZAndQAdjustments   results;
  
  // Boundary conditions
  /**
   * @see flowag.channels.FlowCalculator#setUpstreamBoundaryCondition(flowag.channels.FlowCalculator.UpstreamBoundaryCondition)
   */
  public void setUpstreamBoundaryCondition(UpstreamBoundaryCondition condition){ ubc = condition; }
  /**
   * @see flowag.channels.FlowCalculator#setDownstreamBoundaryCondition(flowag.channels.FlowCalculator.DownstreamBoundaryCondition)
   */
  public void setDownstreamBoundaryCondition(DownstreamBoundaryCondition condition){ generalDownstreamBoundaryCondition = condition; }
  /**
   * @see flowag.channels.FlowCalculator#setDownstreamBoundaryCondition(flowag.channels.CanalSegment, flowag.channels.FlowCalculator.DownstreamBoundaryCondition)
   */
  public void setDownstreamBoundaryCondition(CanalSegment cs, DownstreamBoundaryCondition condition){
    downstreamBoundaryConditions.remove(cs);
    if(condition != null) downstreamBoundaryConditions.put(cs, condition);
  }
  
  
  
  /* Constructor */
  /**
   * Constructor
   * @param initialSegment Initial canal segment
   */
  public AbstractFlowCalculator(CanalSegment initialSegment){
    initialCanalSegment = initialSegment;
  }
  
  // Calculations
  // The structures allow a two-step process: calculation can be performed
  // once to verify that it is performed without failure, then
  // results can be returned if the calculation was completed.
  // Objects using this class may prefer to invoke getResults(duration) directly
  // without testing calculate(duration) first, but in cases where performance
  // of calculations and performance of setting values for the result set
  // are issues, being able to do these separately may be an advantage.
  
  /**
   * @see flowag.channels.FlowCalculator#calculate(double)
   */
  public final boolean calculate(double duration){
    lastCalculatedDuration = duration;
    results = null;
    if(Double.isNaN(duration)) return false;
    return performCalculations(duration);
  }
  
  
  /**
   * This method automatically returns the results of the last calculated
   * duration. If the last calculated duration is NaN, null is returned.
   * @see flowag.channels.FlowCalculator#getResults()
   */
  public FlowZAndQAdjustments getResults(){
    return (Double.isNaN(lastCalculatedDuration) ? null : getResults(lastCalculatedDuration));
  }
  
  /**
   * @see flowag.channels.FlowCalculator#getResults(double)
   */
  public FlowZAndQAdjustments getResults(double duration){
    if(duration != lastCalculatedDuration){
      if(!calculate(duration)) return null;
    }
    if(results == null) setResults();
    return results; 
  }
  
  /**
   * This method should perform all calculations necessary to derive the
   * new flow values to be returned. However, it need not create a return
   * set; it need merely indicate whether the calculations were completed
   * successfully. 
   * @param duration duration of time through which flow occurs
   * @return true if the calculations perform successfully, false otherwise
   */
  protected abstract boolean performCalculations(double duration);
  
  
  /**
   * This method should take the current state of this object
   * and set the FlowCalculator.FlowZAndQAdjustments 'results'
   * object from the superclass to a set that contains
   * all of the information needed to update the actual canal segments.
   * It is called automatically when getResults(duration) is invoked 
   */
  protected abstract void setResults();
  
  
}
