package flowag.channels;

import abcm.math.Point;

/**
* As an additional convenience it includes hydrographs for upstream and
 * downstream flow and stage. These can be used to store information about
 * water flow throughout a day.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class BasicChannelSegment extends AbstractChannelSegment{
    
  /* Constructors */
  /**
   * Creates a channel segment with the given starting and end coordinates,
   * roughness, and upstream and downstream channel cross sections.
   * @param HeadCoordinates the center point of the channel segment bottom at the
   * upstream end
   * @param EndCoordinates the center point of the channel segment bottom at the
   * downstream end
   * @param n Manning's roughness coefficient
   * @param upstreamCross Cross Section at the upstream end
   * @param downstreamCross Cross Section at the downstream end
   */
  protected BasicChannelSegment(Point HeadCoordinates, Point EndCoordinates, double n, 
      ChannelCrossSection upstreamCross, ChannelCrossSection downstreamCross){
    super(HeadCoordinates, EndCoordinates, n, upstreamCross, downstreamCross);
  }

  /**
   * Creates a channel segment with the given starting and end coordinates, roughness,
   * and upstream and downstream channel cross sections as given by a String specification
   * (see ChannelCrossSectionFactory for details).
   * @param HeadCoordinates
   * @param EndCoordinates
   * @param n
   * @param upstreamChannelCrossSectionSpecification
   * @param downstreamChannelCrossSectionSpecification
   */
  protected BasicChannelSegment(Point HeadCoordinates, Point EndCoordinates, double n, 
      String upstreamChannelCrossSectionSpecification, 
      String downstreamChannelCrossSectionSpecification){
    super(HeadCoordinates, EndCoordinates, n, 
          ChannelCrossSectionFactory.getChannelCrossSection(upstreamChannelCrossSectionSpecification),
          ChannelCrossSectionFactory.getChannelCrossSection(downstreamChannelCrossSectionSpecification));    
  }
  
  
  
  /* Incomplete */
  
  /**
   * NONFUNCTIONAL Changes the shape of this Channel Segment by changing the upper and lower
   * cross-section. Must return a value indicating the difference between the volumes
   * of the original and new forms.
   * @deprecated
   * @param UpstreamChannelCrossSection not used
   * @param DownstreamChannelCrossSection not used
   * @return intended to return the difference between the volumes of the old and new channel geometries; actually returns zero
   */
  public double changeShape(ChannelCrossSection UpstreamChannelCrossSection, 
      ChannelCrossSection DownstreamChannelCrossSection){
    return 0;
  }
    
  /**
   * @see flowag.channels.ChannelSegment#infill(double)
   */
  public void infill(double fillDepth){
    // fill in the upstream and downstream cross sections
    // And move the headcoordinates upwards
    upstreamCrossSection.fill(fillDepth);
    Point headCoords = getHeadCoords();
    Point endCoords  = getEndCoords();
    Point newHeadCoords = new Point(headCoords.getX(), headCoords.getY(), headCoords.getZ() + fillDepth, headCoords.getName());
    downstreamCrossSection.fill(fillDepth);
    Point newEndCoords = new Point(endCoords.getX(), endCoords.getY(), endCoords.getZ() + fillDepth, endCoords.getName());
    setHeadAndEndCoordinates(newHeadCoords, newEndCoords);
  }
  
  /**
   * @see flowag.channels.ChannelSegment#excavate(double)
   */
  public void excavate(double depth){
    // Dig this channel segment out a bit further
    upstreamCrossSection.dig(depth);
    Point headCoords = getHeadCoords();
    Point endCoords  = getEndCoords();
    
    Point newHeadCoords = new Point(headCoords.getX(), headCoords.getY(), headCoords.getZ() - depth, headCoords.getName());
    downstreamCrossSection.dig(depth);
    Point newEndCoords = new Point(endCoords.getX(), endCoords.getY(), endCoords.getZ() + depth, endCoords.getName());
    setHeadAndEndCoordinates(newHeadCoords, newEndCoords);
  }

  /**
   * Calculates the total surface area of the construction of the canal.
   * @return the total surface area, in m^2, of the construction of the
   * canal.
   */
  public double getTotalConstructionSurfaceArea(){
    double s1 = upstreamCrossSection.getWettedPerimeter(upstreamCrossSection.getChannelHeight());
    double s2 = downstreamCrossSection.getWettedPerimeter(downstreamCrossSection.getChannelHeight());
    return .5 * (s1 + s2) * getLength();
  }
  
  /**
   * @see flowag.channels.ChannelSegment#getLaborToChangeDepth(double)
   */
  public double getLaborToChangeDepth(double depthChange){
    if(depthChange == 0) return 0;
    else if(depthChange < 0){
      // This briefly reconfigures the channel segment to calculate the volume,
      // then sets it back to what it was.
      double currentVolume = this.getTotalVolume();
      this.infill(Math.abs(depthChange));
      double newVolume = this.getTotalVolume();
      this.excavate(depthChange);
      return Math.abs(newVolume - currentVolume);
    }
    else{
      double currentVolume = this.getTotalVolume();
      this.excavate(depthChange);
      double newVolume = this.getTotalVolume();
      this.infill(depthChange);
      return Math.abs(newVolume - currentVolume);
    }
  }

  
}
