package flowag.channels;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.cloning.ABCMDeepCloneableVector;
import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.graphics.X3DAssist;
import abcm.graphics.X3DIndexedFaceSetCreator;
import abcm.math.BasicMath;
import abcm.math.Geometry;
import abcm.math.Point;
import abcm.math.Polygon;
import abcm.math.Triangle;
import abcm.space.SpatialElement;
import abcm.space.TopoTriangle;
import abcm.space.TriangulatedTopography;
import abcm.utilities.TextFormatter;
import abcm.utilities.XMLAssist;
import flowag.basic.Water;
import flowag.bio.Field;

/**
 * Represents a straight-line channel and provides
 * functions for adjusting the size and shape of the channel and
 * calculating water flow within it.
 * 
 * Curved or tapering canals can be created by assembling
 * CanalSegments in sequence. The CanalSegment class includes
 * methods for extending or branching the canal to other 
 * segments. Additional methods are provided for dealing with
 * fields and plots, and for text reporting of the canals.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class CanalSegment extends BasicChannelSegment implements SpatialElement{
  
  /* Static class */
  /**
   * This class extends the basic triangulated topography class by providing a simpler
   * constructor and a standard method for adding to X3D.
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public static class CanalBedTriangulatedTopography extends TriangulatedTopography{
    
    /** Standard X3D material definition */
    public static String x3dCanalBedDefinition = 
      "<Appearance DEF='CanalBed'>\n" +
      "  <Material diffuseColor='0.9 0.1 0.1' emissiveColor='0 0 0'/>\n" +
      "</Appearance>";

    
    /**
     * Constructor
     */
    public CanalBedTriangulatedTopography(){
      super(new Vector());
    }
    
    /**
     * @see abcm.space.TriangulatedTopography#addToX3D(abcm.graphics.X3DAssist)
     */
    public void addToX3D(X3DAssist x3d){
      XMLAssist xml = x3d.getXML();
  
      x3d.addDefinition("CanalBed", x3dCanalBedDefinition);
  
      X3DIndexedFaceSetCreator ifsCreator = new X3DIndexedFaceSetCreator(topoTriangles);
      
      xml.addLine("<!-- CanalBed -->\n", XMLAssist.INDENT_BLOCK);
      xml.open("Group");
        xml.open("Shape");
          xml.addLine("Appearance", "USE='CanalBed'","");
          xml.open("IndexedFaceSet", "coordIndex=\"" + ifsCreator.getFaceList() + "\"" + " solid=\"false\"");
            xml.addLine("Coordinate", "point=\"" + ifsCreator.getPointList() + "\"", "");
          xml.closeTag(); // Indexed Face Set
        xml.closeTag(); // Terrain Shape
      xml.closeTag(); // Terrain Group
      
    }

  }
  
  /** X3D standard definition for appearance of water */
  public static final String x3dWaterDefinition = 
        "<Appearance DEF='Water'>\n" +
        "  <Material diffuseColor='0.2 0.5 0.1' emissiveColor='0 0 0'/>\n" +
        "</Appearance>";
  
  /* Public Constants */
  /** Default value for the bottom width of a canal. */
  public static final double DEFAULT_BOTTOM_WIDTH = 17;
  
  /** Default value for the top width of a canal. */
  public static final double DEFAULT_TOP_WIDTH    = 20;
  
  /** 
   * Default value for the construction depth of a canal. 
   * Note that this is the depth of construction, not the depth 
   * of the water. 
   */
  public static final double DEFAULT_CHANNEL_HEIGHT = 5;
  
  /** Default value for the roughness (Manning's) of a canal. **/
  public static final double DEFAULT_ROUGHNESS = .01;
  
  /** Default value for depth considered 'empty' **/
  public static final double EMPTY_DEPTH = .005;
  
  
  /* Code Model Switches */
  /** Code Model Switch for Wear */ public static final int NO_WEAR                     = 0;
  /** Code Model Switch for Wear */ public static final int STANDARD_WEAR               = 1;
  
  /** Code Model Switch for Evaporation */ public static final int NO_CHANNEL_EVAPORATION      = 0;
  /** Code Model Switch for Evaporation */ public static final int BASIC_CHANNEL_EVAPORATION   = 1;
  
  /** Canal Type Name */ public static final String DISTRIBUTION = "Distribution";
  /** Canal Type Name */ public static final String MAIN         = "Main";
  /** Canal Type Name */ public static final String FIELD_CANAL  = "Field";
  
  /* Static Variables */
  /** Counter for creating unique ID*/ private static int canalSegmentCount = 1;
  
  
  /* Static Methods */

  /**
   * Returns the next canal ID and increments internal ID counter
   * @return a String representing a CanalSegment ID of the form "C00001".
   */
  private static String getNextID(){
    String CanalSegmentID = "00000" +  canalSegmentCount;
    CanalSegmentID = "C" + CanalSegmentID.substring((CanalSegmentID.length()-5));
    canalSegmentCount++;
    return CanalSegmentID;
  }
  
  
  /**
   * Given two cut and surface lines, maps all of the points
   * from line 1 onto corresponding points on line 2. Typically
   * the process is then executed using the same lines but switching
   * positions; the result is the creation of two lists of points
   * that both have the same number of points at corresponding locations
   * (fractionally across the cut line) than can be matched up
   * for representation or interpolation
   * @param line1 the first list of points
   * @param line2 the second list of points
   * @return a new list of points 
   */
  public static Vector mergeCutAndSurfaceLines(Vector line1, Vector line2){
    Vector merged = new Vector();
    merged.addAll(line1);
    
    // Some numerical values needed
    Point base1 = (Point)merged.get(0);
    Point end1  = (Point)merged.get(merged.size() - 1);
    double width1 = base1.planarDistanceToPoint(end1);
        
    Point base2 = (Point)line2.get(0);
    Point end2  = (Point)line2.get(line2.size() - 1);
    double width2 = base2.planarDistanceToPoint(end2);
   
    // Loop through all points on line2 except the ends (because we already
    // know those and have matching points in line1 already)
    for(int i = 1; i < (line2.size() - 1); i++){
      Point toPlace = (Point)line2.get(i);
      double fraction = base2.planarDistanceToPoint(toPlace) / width2;
      // Now loop through the points in the new line to see where this one fits
      for(int t = 1; t < merged.size(); t++){
        Point farPoint = (Point)merged.get(t);
        double endFrac = base1.planarDistanceToPoint(farPoint) / width1;
        if(Math.abs(endFrac - fraction) < .00001) break;
        if(endFrac > fraction){
          Point nearPoint = (Point)merged.get(t - 1);
          // Must calculate how far along the line segment it is
          double beginFrac = base1.planarDistanceToPoint(nearPoint) / width1;
          double smallFrac = fraction - beginFrac;
          double segFrac = smallFrac / (endFrac - beginFrac); // Note: no points on top of one another!
          Point toAdd = Geometry.pointBetween(nearPoint, farPoint, segFrac);
          merged.add(t, toAdd);
          break;
        }
      }
    }    
    return merged;
  }
  
  
  /* Instance variables */
  /** Inflow gate */
  private Gate                        headGate; 
  /** Outflow gate. Note that this is only applicable if there are no connected canals */
  private Gate                        endGate;                       
  /** Collection of outflow gates */
  private ABCMDeepCloneableVector     outflowCanalGates = new ABCMDeepCloneableVector();
  /** Collection of fields (plots) irrigated by this canal segment */
  private ABCMDeepCloneableVector     plotsIrrigated    = new ABCMDeepCloneableVector();
  /** ID for this canal segment */
  private String                      canalSegmentID;
  /** Flag indicating whether this canal segment is considered attested in the real world or only interpolated */
  private boolean                     isInterpolated = false;        
  /** Flag indicating whether this canal is still in active use */                                                                    
  private boolean                     isAbandoned = false;
  /** String indicating canal type (default is 'Unknown') */
  private String                      canalType = "Unknown";
  /** Water object that represents the water in the canal segment. Will include water quality information. */
  private Water                       water; 
  
  /** Head coordinates at surface level; these remain unchanged even if the depth of the canal segment is adjusted. */
  private Point                       anchorHeadCoords;
  /** End coordinates at surface level; these remain unchanged even if the depth of the canal segment is adjusted. */
  private Point                       anchorEndCoords;
  
  /** Record of the original roughness value; used when the canal is repaired */
  private double                      originalRoughness; 
  /** User-settable value for maximum outflow; default is Double.NaN */
  private double                      userSetMaximumOutflow = Double.NaN;
  
  /** Vector of points for the upstream left corners of outflow canals; used for graphical representation */
  private ABCMDeepCloneableHashtable  upstreamLeftCornersOfOutflows;
  /** Vector of points for the upstream right corners of outflow canals; used for graphical representation */
  private ABCMDeepCloneableHashtable  upstreamRightCornersOfOutflows;
  /** Vector of points for the downstream left corners of outflow canals; used for graphical representation */
  private Point                       downstreamLeftDrawCorner;
  /** Vector of points for the downstream right corners of outflow canals; used for graphical representation */
  private Point                       downstreamRightDrawCorner;
  /** Polygon that represents the shape of the channel at the junction it flows into, if applicable */
  private Polygon                     outflowJunctionPolygon;


  
  
  /* Constructors */
  
  /**
   * Constructor that takes start point, end point, and name
   * to be assigned to this CanalSegment. The CanalSegment is 
   * given a default Parameter Set and a default Code Model.
   * Also by default, the channel is trapezoidal in cross-section
   * and has a size determined by the default values for top width,
   * bottom width, and channel height. A default value for Manning's coefficient will be applied.
   * @param hCoords Point object indicating head coordinates
   * @param eCoords Point object indicating end coordinates
   * @param ID Name of Canal Segment
   */
  public CanalSegment(Point hCoords, Point eCoords, String ID){
    this(hCoords, eCoords, ID, DEFAULT_ROUGHNESS, 
        new TrapezoidalChannelCrossSection(DEFAULT_TOP_WIDTH, DEFAULT_BOTTOM_WIDTH, DEFAULT_CHANNEL_HEIGHT), 
        new TrapezoidalChannelCrossSection(DEFAULT_TOP_WIDTH, DEFAULT_BOTTOM_WIDTH, DEFAULT_CHANNEL_HEIGHT));
  }
  
  
  /**
   * Constructor that takes a Head Gate, end point, a Code Model, and a
   * Parameter Set. The CanalSegment created begins at the coordinates of the
   * HeadGate object, and is assigned the name specified. Also by default, 
   * the channel is trapezoidal in cross-section
   * and has a size determined by the default values for top width,
   * bottom width, and channel height. A default value for Manning's coefficient will be applied.
   * @param HG HeadGate that will form the starting point for this segment
   * @param eCoords Point object indicating end coordinates
   * @param newName name to be assigned
   */
  public CanalSegment(HeadGate HG, Point eCoords, String newName){
    this(HG.getLocus(), eCoords,
        newName, DEFAULT_ROUGHNESS, 
        new TrapezoidalChannelCrossSection(DEFAULT_TOP_WIDTH, DEFAULT_BOTTOM_WIDTH, DEFAULT_CHANNEL_HEIGHT), 
        new TrapezoidalChannelCrossSection(DEFAULT_TOP_WIDTH, DEFAULT_BOTTOM_WIDTH, DEFAULT_CHANNEL_HEIGHT));
    headGate = HG.getOutflow(); // This sets the inflow gate for this segment
                                // to the outflow gate of the specified head gate
    headGate.setHeadGateAndOutflowCanal(HG, this);
  }

  /**
   * Constructor that takes start and end points, name to be assigned to the new canal
   * segment, roughness coefficient, and dimensions, and a cross section, code model,
   * and parameter set. The CanalSegment will have the assigned name, position, and 
   * cross-section, and the assigned Code Model and Parameter Set.
   * @param hCoords Coordinates for the upstream end of the canal
   * @param eCoords Coordinates for the downstream end of the canal
   * @param ID Name of Canal Segment
   * @param nRoughness Manning's roughness coefficient
   * @param xs Cross Section to be applied to the Canal Segment
   */
  public CanalSegment(Point hCoords, Point eCoords, String ID, double nRoughness, 
      ChannelCrossSection xs){
    this(hCoords, eCoords, ID, nRoughness, xs, xs);
  }

  /**
   * Constructor
   * @param hCoords
   * @param eCoords
   * @param ID
   * @param nRoughness
   * @param uxs
   * @param dxs
   */
  public CanalSegment(Point hCoords, Point eCoords, String ID, double nRoughness, 
      ChannelCrossSection uxs, ChannelCrossSection dxs){
    super(Geometry.getPointBelow(hCoords, uxs.getChannelHeight()), Geometry.getPointBelow(eCoords, dxs.getChannelHeight()),
        nRoughness, 
        uxs.getCopy(), 
        dxs.getCopy());
    anchorHeadCoords = hCoords;
    anchorEndCoords  = eCoords;
    canalSegmentID = ID;
    headGate           = new Gate(this, this);
    water              = new Water(0);
    endGate            = new Gate(this, this);
    originalRoughness  = super.getRoughness();
  }


  /* Basic Setters */
  /**
   * Sets the type of this canal
   * @param newType new type for this canal
   */
  public void setCanalType(String newType){ canalType = newType; }
  
  /* Basic Getters */
  
  /**
   * Returns the canal ID of this canal.
   * @return the ID of this canal
   */
  public String getCanalSegmentID(){
    return canalSegmentID;
  }  
  
  /**
   * Gets a String that indicates this canal's 'type'. Type can be flexibly assigned,
   * but will generally use one of the defined static String type variables.
   * @return the canal's type
   */
  public String getType(){ return canalType; }  
  
  
  
  
  
  /* Get Methods related to canal system structure */

  // Inflow
  
  /**
   * Get Inflow Canal Gate.
   * 
   * Each canal segment has an inflow gate
   * @return The inflow gate for this canal segment
   */
  public Gate getInflowGate(){
    return headGate;
  }

  
  /**
   * Returns the Inflow CanalSegment for this canal.
   * @return the Inflow CanalSegment for this canal
   */
  public CanalSegment getInflowCanal(){
    return (headGate.getSource() == null ? this : headGate.getSource());
  }
  
  
  // Outflow
  
  /**
   * Returns the Vector of outflow canals for this canal.
   * @return a Vector listing the outflow canals for this canal
   */
  public Vector getOutflowCanalGates(){    return outflowCanalGates;   }
    
  /**
   * Returns a Vector containing the list of outflow canals
   * for this CanalSegment.
   * @return the Vector listing outflow canals
   */
  public Vector getOutflowCanals(){
    Vector v = new Vector();
    Vector cg = getOutflowCanalGates();
    for(int i = 0; i < cg.size(); i++){
      v.add(((Gate)cg.get(i)).getOutflow());          
    }
    return v;
  }

  /**
   * Returns a Vector of just the outflow canals that are open. 
   * @return a Vector of open outflow canals
   */
  public Vector getOpenOutflowCanals(){
    Vector v = new Vector();
    Vector cg = getOutflowCanalGates();
    if(cg == null) return v;
    for(int i = 0; i < cg.size(); i++){
      if(((Gate)cg.get(i)).isOpen())  v.add(((Gate)cg.get(i)).getOutflow());          
    }
    return v;
  }  
  
  // Specialty: downstream 'major' vs. 'minor' outflows
  
  /**
   * Gets the next downstream segment. If more than one,
   * chooses the one whose heading is closest to this one.
   * If none, returns null
   * @return the next downstream CanalSegment object 
   */
  public CanalSegment getNextDownstreamSegment(){
    if(outflowCanalGates.size() == 0 ) return null;
    else if (outflowCanalGates.size() == 1){
      return ((Gate)outflowCanalGates.get(0)).getOutflow();
    }
    else{
      double hc = 180;
      Point headCoords = super.getHeadCoords();
      Point endCoords = super.getEndCoords();
      CanalSegment retVal = ((Gate)outflowCanalGates.get(0)).getOutflow();
      hc = endCoords.headingChange(headCoords, retVal.getEndCoords());
      for(int i = 1; i < outflowCanalGates.size(); i++){
        CanalSegment cs = ((Gate)outflowCanalGates.get(i)).getOutflow();
        double hcCompare = endCoords.headingChange(headCoords, cs.getEndCoords());
        if(Math.abs(hcCompare) < Math.abs(hc)){
          retVal = cs;
          hc = hcCompare;
        }       
      }
      return retVal;
    }
  }  
  
  /**
   * Returns a Vector listing all the outflow canals except the
   * major one.
   * @return a vector listing all the outflow canals except the
   * major one.
   */
  public Vector getBranches(){
    if(outflowCanalGates.size()<=1) return new Vector();
    else{
      Vector v = new Vector();
      CanalSegment ds = getNextDownstreamSegment();
      for(int i = 0; i < outflowCanalGates.size(); i++){
        CanalSegment c = ((Gate)outflowCanalGates.get(i)).getOutflow(); 
        if(!ds.equals(c)) v.add(c); 
      }
      return v;
    }
  }
  
  

  
  
  
  /* Canal System Creation methods */    
  
  
  /**
   * Extend a canal by adding a new canal segment onto the
   * end of this canal. The segment may be disallowed
   * if the 'allowUphillSegments' flag is set to false
   * and the coordinates of the end point are at a higher
   * elevation than the end point of this segment.
   * @param newEndPoint point to which the new segment is to extend 
   * @param cs system in which this segment and the new one exist
   * @param newCanalName new name for the new canal segment
   * @return true if successful, false otherwise.
   */
  public boolean extendCanal(Point newEndPoint, CanalSystem cs, String newCanalName){
    boolean retVal = false;
    Point endCoords  = getEndCoords();
    
    if(cs.allowUphillSegments || ((newEndPoint.getZ() - downstreamCrossSection.getChannelHeight()) <= endCoords.getZ())){
      endGate = null;
      CanalSegment newCanal1 = new CanalSegment(this.anchorEndCoords, newEndPoint, newCanalName, this.getRoughness(), super.downstreamCrossSection);
      newCanal1.setInflowCanal(this);
      addOutflowCanal(newCanal1);
      newCanal1.canalType = this.canalType;
      cs.addCanal(newCanal1);
      retVal = true;
    }
    else{
      System.err.println("Could not extend canal: new Z coordinate for downstream segment (" + newEndPoint.coordString() + " )is higher than that of upstream segment (" + endCoords.coordString() + ")");
      retVal = false;
    }
    return retVal;
  }
  
  /**
   * Extends this canal segment as the inflow to another. Note that the method
   * may fail if the inflow gate has an inflow canal already or a Headgate, or 
   * if the new canal would be of invalid slope.
   * @param joinTo the canal segment into which the segment to be extended will flow
   * @param cs the canal segment to be extended
   * @param newCanalName the new name for the newly created canal segment
   * @return true if successful, false otherwise.
   */
  public boolean extendCanalIntoSegment(CanalSegment joinTo, CanalSystem cs, String newCanalName){
    if(joinTo.getInflowCanal() != null) return false;               // Canal to flow into must have no inflow already
    if(joinTo.getInflowGate().getHeadGate() != null) return false;  // ditto, essentially
    boolean retVal = false;
    Point endCoords  = getEndCoords();
    Point newEndPoint = joinTo.getHeadCoords();
    
    if(cs.allowUphillSegments || ((newEndPoint.getZ() - downstreamCrossSection.getChannelHeight()) <= endCoords.getZ())){
      endGate = null;
      CanalSegment newCanal1 = new CanalSegment(this.anchorEndCoords, newEndPoint, newCanalName, this.getRoughness(), super.downstreamCrossSection);
      newCanal1.setInflowCanal(this);
      addOutflowCanal(newCanal1);
      cs.addCanal(newCanal1);
      joinTo.setInflowCanal(newCanal1);
      newCanal1.addOutflowCanal(joinTo);      
      newCanal1.canalType = this.canalType;
      retVal = true;
    }
    else{
      System.err.println("Could not extend canal: new Z coordinate for downstream segment (" + newEndPoint.coordString() + " )is higher than that of upstream segment (" + endCoords.coordString() + ")");
      retVal = false;
    }
    return retVal;    
  }
  
  /**
   * Disconnects this canal segment from all inflow and outflow canals, isolating it.
   */
  public void disconnect(){
    disconnectFromInflow();
    disconnectAllOutflows();
  }
  
  /**
   * Disconnects this canal segment from any inflow canal; manages the recordkeeping for the inflow canal also.
   */
  private void disconnectFromInflow(){
    if((this.headGate != null) && (this.headGate.getSource() != null)) this.headGate.getSource().removeOutflowCanalSegment(this);
    this.headGate.setSource(this);
  }
  
  /**
   * Removes a specific outflow canal segment from the collection of outflow segments for this canal segment.
   * @param outflowCS the segment to be removed.
   */
  private void removeOutflowCanalSegment(CanalSegment outflowCS){
    Vector V = this.getOutflowCanalGates();
    for(int i = 0; i < V.size(); i++){
      Gate G = (Gate)V.get(i);
      if(G.getSource() != null && G.getSource().equals(outflowCS)){
        this.outflowCanalGates.remove(G);
        return;
      }
    }
  }
  
  
  
  /**
   * Removes all outflow canal segments; manages the recordkeeping in all segments affected. 
   */
  private void disconnectAllOutflows(){
    Vector V = this.getOutflowCanalGates();
    for(int i = 0; i < V.size(); i++){
      Gate G = (Gate)V.get(i);
      if(G.getSource() != null) G.getSource().disconnectFromInflow();
    }
    endGate = new Gate(this,this);
  }
  
  
    
  
  /* Methods to modify the shape of this canal segment */
  
  /**
   * Sets the profile of this Canal Segment.
   * The profile is passed as a specification that is used by the ChannelCrossSectionFactory
   * object to create a channel cross section.
   * @param spec a String representing the specifical of a ChannelCrossSection
   * @return true if successful, false otherwise
   */
  public boolean setProfile(String spec){
    return setUpstreamProfile(spec) & setDownstreamProfile(spec); // Note: not short-circuited; both will execute, though if the first one fails the second should fail as well    
  }
  
  /**
   * Sets the upstream profile of this segment, and changes the cross-sections as
   * appropriate
   * @param spec String specification of a cross-section
   * @return true if successful, false otherwise
   */
  public boolean setUpstreamProfile(String spec){
    if(this.getUpstreamWaterDepth() > 0 || this.getDownstreamWaterDepth() > 0) return false;
    ChannelCrossSection CCS = ChannelCrossSectionFactory.getChannelCrossSection(spec);
    if(CCS == null) return false;
    upstreamCrossSection = CCS;
    Point newHeadCoords = Geometry.getPointBelow(anchorHeadCoords, CCS.getChannelHeight());
    this.z1 = newHeadCoords.getZ();
    setHeadAndEndCoordinates(newHeadCoords, getEndCoords());
    return true;    
  }
  
  /**
   * Sets the downstream cross sectional profile
   * @param spec a profile specification. @see flowag.channels.ChannelCrossSectionFactory.getChannelCrossSection(String) for details.
   * @return true if successful, false otherwise
   */
  public boolean setDownstreamProfile(String spec){
    if(this.getUpstreamAvgDepth() > 0 || this.getDownstreamAvgDepth() > 0) return false;
    ChannelCrossSection CCS = ChannelCrossSectionFactory.getChannelCrossSection(spec);
    if(CCS == null) return false;
    downstreamCrossSection = CCS;
    Point newEndCoords = Geometry.getPointBelow(anchorEndCoords, CCS.getChannelHeight());
    this.z2 = newEndCoords.getZ();
    setHeadAndEndCoordinates(getHeadCoords(), newEndCoords);
    return true;    
  }
    
  /**
   * Sets new values for the upstream and downstream (matching) cross sections of this segment
   * @param topWidth new value for the construction top width of the segment
   * @param bottomWidth new value for the construction bottom width of the segment
   * @param channelHeight new value for the construction height of the segment
   * @return true if successful, false otherwise
   */
  public boolean setProfile(double topWidth, double bottomWidth, double channelHeight){
    if(this.getUpstreamWaterDepth() > 0 || this.getDownstreamWaterDepth() > 0) return false;
    upstreamCrossSection.setTopWidth(topWidth);
    upstreamCrossSection.setBottomWidth(bottomWidth);
    upstreamCrossSection.setChannelHeight(channelHeight);
    downstreamCrossSection.setTopWidth(topWidth);
    downstreamCrossSection.setBottomWidth(bottomWidth);
    downstreamCrossSection.setChannelHeight(channelHeight);
    Point newHeadCoords = Geometry.getPointBelow(anchorHeadCoords, upstreamCrossSection.getChannelHeight());
    Point newEndCoords  = Geometry.getPointBelow(anchorEndCoords, downstreamCrossSection.getChannelHeight());
    setHeadAndEndCoordinates(newHeadCoords, newEndCoords);
    return true;    
  }
  
  /**
   * Sets new values for the upstream cross section of this segment
   * @param topWidth new value for the construction top width of the segment
   * @param bottomWidth new value for the construction bottom width of the segment
   * @param channelHeight new value for the construction height of the segment
   * @return true if successful, false otherwise
   */
  public boolean setUpstreamProfile(double topWidth, double bottomWidth, double channelHeight){
    if(this.getUpstreamWaterDepth() > 0 || this.getDownstreamWaterDepth() > 0) return false;
    upstreamCrossSection.setTopWidth(topWidth);
    upstreamCrossSection.setBottomWidth(bottomWidth);
    upstreamCrossSection.setChannelHeight(channelHeight);
    Point newHeadCoords = Geometry.getPointBelow(anchorHeadCoords, upstreamCrossSection.getChannelHeight());
    setHeadAndEndCoordinates(newHeadCoords, getEndCoords());
    return true;    
  }
  
  /**
   * Sets new values for the downstream cross section of this segment
   * @param topWidth new value for the construction top width of the segment
   * @param bottomWidth new value for the construction bottom width of the segment
   * @param channelHeight new value for the construction height of the segment
   * @return true if successful, false otherwise
   */
  public boolean setDownstreamProfile(double topWidth, double bottomWidth, double channelHeight){
    if(this.getUpstreamAvgDepth() > 0 || this.getDownstreamAvgDepth() > 0) return false;
    downstreamCrossSection.setTopWidth(topWidth);
    downstreamCrossSection.setBottomWidth(bottomWidth);
    downstreamCrossSection.setChannelHeight(channelHeight);
    Point newEndCoords  = Geometry.getPointBelow(anchorEndCoords, downstreamCrossSection.getChannelHeight());
    setHeadAndEndCoordinates(getHeadCoords(), newEndCoords);
    return true;    
  }
  


  

  
  
  
  
  

  /* Set Methods related to canal system structure */
  
  /**
   * Sets the Inflow CanalSegment for this canal.
   * @param canalSegment the new Inflow CanalSegment
   * @return true if successful, false otherwise 
   */
  private boolean setInflowCanal(CanalSegment canalSegment){
    // Confirm that inflow canal's end coordinates equal this canal's head coordinates
    if(this.getHeadCoords().isEqual(canalSegment.getEndCoords())){
      headGate.setSource(canalSegment);
      return true;
    }
    System.err.println("CANNOT SET INFLOW CANAL for Seg " + this.canalSegmentID + " with new inflow of " + canalSegment.canalSegmentID + ": DOWNSTREAM SEGMENT'S UPSTREAM COORDINATES ARE " + this.getHeadCoords().coordString(10) + " WHILE UPSTREAM ENDS AT " + canalSegment.getEndCoords().coordString(10));
    return false;
  }

  /**
   * Adds the specified canal to the collection of outflow canals.
   * Outflow canals are always listed in order such that the endpoints of the outflowing
   * canals are in clockwise arrangement.
   * @param C
   */
  public void addOutflowCanal(CanalSegment C){
    if(outflowCanalGates.size() == 0){
      outflowCanalGates.add(C.getInflowGate());
      downstreamLeftDrawCorner = C.getUpstreamLeftBankCoordinate();
      downstreamRightDrawCorner = C.getUpstreamRightBankCoordinate();
      return;
    }
    double headingChangeToNewCanal = Geometry.headingChange(getEndCoords(), getHeadCoords(), C.getEndCoords());
    for(int i = 0; i < outflowCanalGates.size(); i++){
      if(headingChangeToNewCanal < Geometry.headingChange(getEndCoords(), getHeadCoords(), ((Gate)outflowCanalGates.get(i)).getOutflow().getEndCoords())){
        outflowCanalGates.add(i, C.getInflowGate());
        break;
      }
    }
    // If not added yet, add to end
    if(!outflowCanalGates.contains(C.getInflowGate())) outflowCanalGates.add(C.getInflowGate());
    
    // Add to map of corner points...
    double previousHeadingChange = -1 * Math.PI;
    double newHeadingChange = 0;
    Vector pointsList = new Vector();
    CanalSegment previousCanalSegment = null;
    double r = this.downstreamCrossSection.getTopWidth() / 2d;
    upstreamLeftCornersOfOutflows = new ABCMDeepCloneableHashtable();
    upstreamRightCornersOfOutflows = new ABCMDeepCloneableHashtable();
    double newZ = this.getDownstreamLeftBankCoordinate().getZ();
    Point headCoords = getHeadCoords();
    Point endCoords = getEndCoords();
    double originalHeading = headCoords.headingTo(endCoords);
    for(int i = 0; i < outflowCanalGates.size(); i++){
      CanalSegment out = ((Gate)outflowCanalGates.get(i)).getOutflow();
      newHeadingChange = Geometry.headingChange(endCoords, headCoords, out.getEndCoords());
      Point newPoint = Geometry.getOffsetPoint(endCoords, ((newHeadingChange + previousHeadingChange) / 2d) + originalHeading, r);
      newPoint = new Point(newPoint.getX(), newPoint.getY(), newZ);
      previousHeadingChange = newHeadingChange;
      if(i == 0) downstreamLeftDrawCorner = newPoint;
      upstreamLeftCornersOfOutflows.put(out, newPoint);
      pointsList.add(newPoint);
      if(previousCanalSegment != null) upstreamRightCornersOfOutflows.put(previousCanalSegment, newPoint);
      previousCanalSegment = out;
    }
    // Clean up the last one
    double newSplit = originalHeading + (Math.PI + previousHeadingChange) / 2d;
    Point newPoint = Geometry.getOffsetPoint(endCoords, newSplit, r);
    newPoint = new Point(newPoint.getX(), newPoint.getY(), newZ);
    upstreamRightCornersOfOutflows.put(previousCanalSegment, newPoint);
    downstreamRightDrawCorner = newPoint;
    pointsList.add(newPoint);
    outflowJunctionPolygon = new Polygon(pointsList);
  }
  
  /**
   * Clears the collection of outflow canal Gates.
   */
  private void clearOutflowCanalGates(){   
    outflowCanalGates.clear();
    downstreamLeftDrawCorner       = null;
    downstreamRightDrawCorner      = null;
    outflowJunctionPolygon         = null;
    upstreamRightCornersOfOutflows = null;
    upstreamLeftCornersOfOutflows  = null;    
  }
  

  
  
  /* Gate status getters */
    
  /**
   * Returns true if the inflow gate to this Canal Segment is open.
   * @return true if the inflow gate to this Canal Segment is open.
   */
  public boolean inflowGateIsOpen(){  return (headGate != null ? headGate.isOpen() : false); }
  
  /**
   * Determine if the outflow gate of this canal segment is open.
   * @return true if there is an endgate and it is open, or 
   * if there is no end gate but there are more than zero open
   * outflow canals (per getOpenOutflowCanals())
   */
  public boolean outflowGateIsOpen(){
    if(endGate == null) return (getOpenOutflowCanals().size() > 0);
    return endGate.isOpen(); 
  }
  

  
  
  /* FOR FLOW SYSTEM CREATION */
  /* 
   * These are public methods that give information about the canal segment
   * that can be used to determine its status, its relationship to other
   * segments adjacent to it, and its role in the flowing system
   */

  /**
   * Returns true if there is any water in this canal segment. Determined by
   * whether upstream or downstream depth is greater than 0; if _either_ of
   * them is, returns true (c.f. 'isEmpty')
   * 
   * @return true if there is any water in this canal segment
   */
  public boolean hasWater(){  return ((getUpstreamWaterDepth() > 0)  || (getDownstreamWaterDepth() > 0));  }
  
  /**
   * Returns true if there is no water in this canal segment; false otherwise.
   * Returns true only if both upstream and downstream depths are below the
   * threshold for emptiness. (c.f. 'hasWater')
   * 
   * @return true or false
   */
  public boolean isEmpty(){
    return ((getUpstreamWaterDepth() < EMPTY_DEPTH) && (getDownstreamWaterDepth() < EMPTY_DEPTH));
  }
  
  /**
   * Returns true if the upstream canal segment of this one exists and has
   * water in it (using 'hasWater()'), false otherwise.
   * @return true or false
   */
  public boolean upstreamCanalHasWater(){
    return (headGate.getSource() != null ? headGate.getSource().hasWater() : false);
  }  
  
  /**
   * Returns true if all downstream outflow branches of this canal segment
   * are empty; false otherwise.
   * @return true or false
   */
  public boolean allDownstreamOutflowSegsAreEmpty(){
    Vector out = getOpenOutflowCanals();
    for(int i = 0; i < out.size(); i++){
      CanalSegment CS = (CanalSegment)out.get(i);
      if(!CS.isEmpty()) return false; 
    }
    return true;
  } 
 
  /**
   * Returns the total volume of water in this and all downstream canal
   * segments. If there are multiple downstream segments, only the
   * ones with open inflow gates are included in the total.
   * @return the total volume of water in this and all downstream canal
   * segments
   */
  public double totalWaterInThisAndOutflowSegments(){
    double tw = 0;
    tw += getTotalWater();
    Vector V = getOpenOutflowCanals();
    for(int i = 0; i < V.size(); i++){
      tw+=((CanalSegment)V.get(i)).getTotalWater();
    }
    return tw;
  }
  
  /**
   * Returns true if there is a no end gate for this segment,
   * which indicates that it is the end of the known system.
   * @return true if this is the end of the defined system
   */
  public boolean isEndOfDefinedSystem(){ return !(endGate==null);}
  

  
  
  
  /* Opening and Closing Gates */
  
  /**
   * Closes the inflow gate of this Canal Segment
   * Note that this should have the effect of setting
   * the discharge rates upstream and downstream
   * to zero; however, this can be recalculated by
   * the flow routine.
   * @return true if the close method is executed
   * successfully, false otherwise
   */
  public boolean closeInflowGate(){
    if(headGate.isBroken()) return false;
    headGate.setOpen(false);
    return true;
  }
  
  /**
   * Opens the inflow gate to this canal segment. Note: the original theory
   * was not implemented; this merely changes the status of the gate.
   * Ideally it would balance the water on either side in terms of
   * amount, stage, and quality (discharge remaining zero on both sides)
   * @return true
   */
  public boolean openInflowGate(){
    headGate.setOpen(true);
    return true;
    
//    // If this is already open, do nothing
//    if(headGate.isOpen()) return true;
//    
//    // Must adjust water on both sides of gate
//    
//    // Get the relevant values before beginning
//    double originalUpstreamSideStage = this.getInflowCanal().getDownstreamStage();
//    double originalDownstreamSideStage = this.getUpstreamStage();
//    
//    // The new value will be between these two
//    double originalTotalVolume = this.getInflowCanal().totalWaterInThisAndOutflowSegments();
//    
//    // Start with the lower stage and work upward
//    // Note that even if this is a junction, there should still be only ONE
//    // stage value on the other side of the gate
//    double lowStage  = Math.min(originalUpstreamSideStage, originalDownstreamSideStage);
//    double highStage = Math.max(originalUpstreamSideStage, originalDownstreamSideStage);
//    if(lowStage != highStage){
//      double stageInterval = highStage - lowStage;
//      double adjust = stageInterval / 2d;
//      double newStage = lowStage + adjust;
//      headGate.setOpen(true);
//         
//      double MINIMUM_TOLERABLE_DIFF = .01;
//      do{
//        this.setUpstreamStage(newStage);
//        
//        adjust = adjust / 2d;
//        if(this.getInflowCanal().totalWaterInThisAndOutflowSegments() > originalTotalVolume) newStage -= adjust;
//        else newStage += adjust;
//      }while((this.getInflowCanal().totalWaterInThisAndOutflowSegments() - originalTotalVolume) > MINIMUM_TOLERABLE_DIFF);
//    }
//    // This should already be zero, but just to make sure;
//    this.setUpstreamDischarge(0);
//    
//    
//    
//    return true;
  }
  
  /**
   * Closes the end gate of this canal segment.
   * Note that canal segments will only have end gates if they
   * are termini- that is, if they have no outflow segments.
   * When the end gate is closed, discharge will be set to zero,
   * however, this can be done by the flow calculation routine.
   * @return true if successful, false otherwise
   */
  public boolean closeEndGate(){
    if(endGate == null) return false;
    if(!endGate.isBroken()) {
      endGate.setOpen(false);
      return true;
    }
    return false;
  }
  
  /**
   * Opens the end gate of this canal segment.
   * Canal segments only have end gates if they are termini- that is, if
   * they have no outflow segments. When the gate is opened,
   * discharge is permitted, but the outflow will be considered normal flow.
   * @return true if successful, false otherwise.
   */
  public boolean openEndGate(){
    if(endGate == null) return false;
    endGate.setOpen(true);
    return true;
  } 


  /* Overload up and downstream get and set */
  /*
   * These methods are overloaded to ensure that the values 
   * are consistent within the canal system- i.e., that the
   * stage at a point where canal segment A joins B
   * is the same whether you are looking at the
   * downstream of A or the upstream of B.
  
   * Note that this will be somewhat forced in the
   * case of junctions where the canal branches. It 
   * will be possible to set a common stage for such
   * junctions, but not a common discharge. However, this
   * is not a concern because the flow routines use these
   * values differently. It is also not possible to set a 
   * single depth; only a stage can be set.
   *
   * The structure employed is as follows:
   *
   * Upstream:
   *   When setting upstream values, one of the following conditions will apply:
   *    - The canal segment may be an initial segment, and thus have no inflow canals
   *     - The canal segment may be an extension of another, in which case it is
   *           + the only extension
   *           + one of several extensions (a junction)
   * Also, the inflow gate may be open or closed. In the case of a junction,
   * the only extensions that matter are the ones with OPEN inflow gates.
   *  Downstream:
   * 
   * When setting downstream values, one of the following conditions will apply:
   *     - The canal segment may be a terminus (the end of the canal)
   *     - The canal segment may be continued by a single canal segment
   *     - The canal segment may be continued by several canal segments
   * The inflow gates of any continuing canal segments, or the `end gate' of a terminus,
   * may be open or closed.
   * 
   * Each of the methods below deals with the appropriate contingency
   */
  
  //  First, a collection of methods that overload the basic operators
  
  /**
   * Sets the upstream stage of this CanalSegment. Will also set
   * the downstream stage of any inflow canal (which may, in turn,
   * set the upstream stages of any other outflow canals).
   * @param z the new stage to be set
   */
  public void setUpstreamStage(double z){         setUpstreamStage(z, this);        }
  
  /**
   * Sets the downstream stage value for this CanalSegment
   * object, and the corresponding upstream stage
   * values for any outflow CanalSegments. The downstream stage 
   * value will be the z value at the downstream cross section. However,
   * if the CanalSegment has one or more outflow canals,
   * this method will automatically set the upstream stage
   * values for all of them to the given stage.
   * @param z the stage value to be set
   */
  public void setDownstreamStage(double z){       setDownstreamStage(z, this);      }

  /**
   * Sets the upstream discharge for this CanalSegment.
   * This will also set the downstream discharge for any
   * inflow canal segment, provided that this CanalSegment
   * is the only outflow segment for that canal.
   * @param Q the new discharge value
   */
  public void setUpstreamDischarge(double Q){     setUpstreamDischarge(Q, this);    }

  /**
   * Sets the downstream discharge for this CanalSegment object.
   * Will also set this as the upstream discharge for any
   * outflow CanalSegment, but only if there is just one. 
   * If there are multiple outflow canals, the only action
   * taken will be to set this CanalSegment's downstream
   * discharge.<br/><br/>
   * This is unfortunate- ideally, the downstream and
   * upstream discharges of connected CanalSegments would be kept
   * consistent. However, in the case of multiple outflow canals,
   * there is no way to apportion the discharge among them. Solving
   * this is part of the job of the flow algorithm.
   * @param Q the new discharge value
   */
  public void setDownstreamDischarge(double Q){   setDownstreamDischarge(Q, this);  }
 
  // Now the actual methods, that incorporate appropriate recursion to ensure
  // that the strategy is enforced
  
  /**
   * Sets the upstream stage of a canal segment, and simultaneously sets
   * the downstream stage of any upstream segment to which it is connected,
   * if the inflow gate for this canal is open.
   * This method acts like the corresponding setUpstreamStage(double)
   * method, but passes an additional parameter that represents
   * which CanalSegment object called it. This prevents recursion because
   * this CanalSegment will not call the second method on the same
   * CanalSegment that called the first.
   * Sets the upstream stage for this segment, AND the downstream stage
   * for any upstream segment.
   * @param z the new stage to be set
   * @param calledBy the CanalSegment object that has called this method
   */
  private void setUpstreamStage(double z, CanalSegment calledBy){
    // First, set the upstream z value for this segment
    super.setUpstreamStage(z);
    // Next, decide if recursion is appropriate; this will only happen
    // when this method is called by THIS object.
    if(calledBy.equals(this)){
      // If the inflow gate is open, the upstream segment's downstream
      // z should be set to match
      if(inflowGateIsOpen()){
        CanalSegment C = getInflowCanal();
        if(C == null) return; // It is possible that this is an initial segment
        if(C.getCanalSegmentID() != canalSegmentID) C.setDownstreamStage(z, this);
      }
    }
  }
  
  /**
   * Sets the downstream stage value for this CanalSegment.
   * This method acts in accord with the description provided
   * for setDownstreamStage(double), but passes as an additional
   * argument the CanalSegment that calls the argument. This is
   * part of a scheme to prevent infinite recursion- the
   * method will not set the upstream stage of the CanalSegment
   * that called it. 
   * @param z the new stage to be set
   * @param calledBy CanalSegment calling this method
   */
  private void setDownstreamStage(double z, CanalSegment calledBy){
    // First, set the downstream z value for this segment
    super.setDownstreamStage(z);
    // Now get the list of all other canals into which this one flows.
    // Note that this will omit any if the inflow gate is closed
    Vector V = getOpenOutflowCanals();
    // Loop through this and get all the ones EXCEPT the one that
    // called this one.
    for(int i = 0; i < V.size(); i++){
      CanalSegment cs = (CanalSegment)V.get(i);
      if(!cs.equals(calledBy)) {
        // Set its upstream value
        cs.setUpstreamStage(z, this);      
      }
    }
  }
  
  /**
   * Sets the upstream discharge for this Canal Segment.
   * This method acts like the description provided for the
   * public setUpstreamDischarge(double) method, but passes
   * an additional parameter that indicates what CanalSegment
   * object has called the method. This prevents recursion.
   * @param Q new discharge value
   * @param calledBy CanalSegment that has called this method
   */
  private void setUpstreamDischarge(double Q, CanalSegment calledBy){
    super.setUpstreamDischarge(Q);
    CanalSegment C = getInflowCanal();
    if(calledBy.equals(this)){
      if(C == null) return;
      if(this.inflowGateIsOpen()){
        if(C.getCanalSegmentID() != canalSegmentID){
          // Only set this if there is only ONE open outflow canal;
          // If there are two, discharge can't be set
          if(C.getOpenOutflowCanals().size() == 1) {
            C.setDownstreamDischarge(Q);
          }
        }
      }
    }
  }
  
  /**
   * Sets the downstream discharge as described in the public
   * method setDownstreamDischarge(double). However, this
   * also tags the method with an additional parameter that
   * indicates which CanalSegment has called this method. This
   * prevents recursion.
   * @param Q new discharge value
   * @param calledBy the CanalSegment that has called this method
   */
  private void setDownstreamDischarge(double Q, CanalSegment calledBy){
    super.setDownstreamDischarge(Q);
    // If this routine is called by this canal segment,
    // then you might need to set an adjacent canal segment's values
    if(calledBy.equals(this)){
      Vector V = getOpenOutflowCanals();
      if(V.size() == 1){
        CanalSegment cs = (CanalSegment)V.get(0);
        cs.setUpstreamDischarge(Q, this);
      }
    }
  }
  
  /**
   * Sets the stage and discharge values for the upstream cross
   * section. The argument passed is the stage value; the discharge
   * is set to the value given by Manning's equation for normal
   * flow at that stage. Overridden from the super class because
   * in some cases it will be advantageous to set this not to exactly
   * the calculated discharge value, but to the headgate's flow
   * value- mainly to eliminate rounding errors and make sure the
   * two values are the same.
   * @param stage the stage value to be set, and from which discharge
   * is also calculated and set
   */
  public void setUpstreamStagePlusManningDischarge(double stage){
    setUpstreamStage(stage);
    double calculatedValue = getUpstreamManningDischarge(stage);
    double hgVal = Double.NaN;
    if((this.headGate != null) && (this.headGate.isOpen()) && (this.headGate.getHeadGate() != null)) hgVal = this.headGate.getHeadGate().getFlow();
    if(Math.abs(hgVal - calculatedValue) < .15) calculatedValue = hgVal;
    setUpstreamDischarge(calculatedValue);
  }
  
  /* Get hydraulic values */

  /*
   * This set of methods returns hydraulic values, but does so
   * in a way that accommodates the canal system network structure.
   * 
   * In some cases this means that the value returned will be based
   * on the geometry of the upstream or downstream segment rather than 
   * the current segment.
   */
  
  /**
   * Returns the water surface width at the downstream end of this
   * CanalSegment. Note that if this CanalSegment has one (and
   * only one) open outflow canal, the value returned will be
   * the upstream water surface width of the outflow canal.
   * @return water surface width at the downstream end of this
   * CanalSegment
   */
  public double getDownstreamWaterSurfaceWidth() {
    return ((getOpenOutflowCanals().size() == 1) ?
                 ((CanalSegment)getOutflowCanals().get(0)).getUpstreamWaterSurfaceWidth() :
                 super.getDownstreamWaterSurfaceWidth());
  }
  
  /**
   * Returns the Average Depth at the downstream end of this
   * CanalSegment. Note that if this CanalSegment has one (and
   * only one) outflow canal, the value returned will be the
   * upstream Average Depth of the outflow canal.
   * @return the Average Depth at the downstream end of this CanalSegment
   */
  public double getDownstreamAvgDepth(){ 
    return ((getOpenOutflowCanals().size() == 1) ?
                 ((CanalSegment)getOutflowCanals().get(0)).getUpstreamAvgDepth() :
                 super.getDownstreamAvgDepth());
  }
  
  /**
   * Returns the Flow Area at the downstream end of this CanalSegment.
   * Note that if this CanalSegment has one (and only one) outflow
   * canal, the value returned will be the Flow Area of the upstream
   * end of the outflow canal.
   * @return the Flow Area at the downstream end of this CanalSegment.
   */
  public double getDownstreamFlowArea(){
    return ((getOpenOutflowCanals().size() == 1) ?
                 ((CanalSegment)getOutflowCanals().get(0)).getUpstreamFlowArea() :
                 super.getDownstreamFlowArea());                     
  }
  
  /**
   * @see flowag.channels.AbstractChannelSegment#getDownstreamFlowArea(double)
   */
  public double getDownstreamFlowArea(double waterDepth){
    return ((getOpenOutflowCanals().size() == 1) ?
        ((CanalSegment)getOutflowCanals().get(0)).getUpstreamFlowArea(waterDepth) :
        super.getDownstreamFlowArea(waterDepth));
  }
  
  /**
   * Returns the Wetted Perimeter at the downstream end of this CanalSegment.
   * Note that if this CanalSegment has one (and only one) outflow
   * canal, the value returned will be the Wetted Perimeter of the upstream
   * end of the outflow canal.
   * @return the Wetted Perimeter at the downstream end of this CanalSegment.
   */
  public double getDownstreamWettedPerimeter(){
    return ((getOpenOutflowCanals().size() == 1) ?
                 ((CanalSegment)getOutflowCanals().get(0)).getUpstreamWettedPerimeter() :
                 super.getDownstreamWettedPerimeter());              
  }
  
  /**
   * Returns the Hydraulic Radius at the downstream end of this CanalSegment.
   * Note that if this CanalSegment has one (and only one) outflow
   * canal, the value returned will be the Hydraulic Radius of the upstream
   * end of the outflow canal.
   * @return the Hydraulic Radius at the downstream end of this CanalSegment.
   */
  public double getDownstreamHydraulicRadius(){
    return ((getOpenOutflowCanals().size() == 1) ?
                 ((CanalSegment)getOutflowCanals().get(0)).getUpstreamHydraulicRadius() :
                 super.getDownstreamHydraulicRadius());              
  }
  
  /**
   * Returns the dA/dZ value at the downstream end of this CanalSegment.
   * Note that if this CanalSegment has one (and only one) outflow
   * canal, the value returned will be the dA/dZ value of the upstream
   * end of the outflow canal.
   * @return the dA/dZ value at the downstream end of this CanalSegment.
   */
  public double getDownstreamDADz(){
    return ((getOpenOutflowCanals().size() == 1) ?
        ((CanalSegment)getOutflowCanals().get(0)).getUpstreamDADz() :
        super.getDownstreamDADz());              
  }
  
  /**
   * Returns the Conveyance value at the downstream end of this CanalSegment.
   * Note that if this CanalSegment has one (and only one) outflow
   * canal, the value returned will be the Conveyance of the upstream
   * end of the outflow canal.
   * @return the Conveyance value at the downstream end of this CanalSegment.
   */
  public double getDownstreamConveyance(){ 
    return ((getOpenOutflowCanals().size() == 1) ?
        ((CanalSegment)getOutflowCanals().get(0)).getUpstreamConveyance() :
        super.getDownstreamConveyance());        
  }
  
  /**
   * Returns the dK/dZ value at the downstream end of this CanalSegment.
   * Note that if this CanalSegment has one (and only one) outflow
   * canal, the value returned will be the dK/dZ value of the upstream
   * end of the outflow canal.
   * @return the dK/dZ value at the downstream end of this CanalSegment.
   */
  public double getDownstreamDKDz(){
    return ((getOpenOutflowCanals().size() == 1) ?
        ((CanalSegment)getOutflowCanals().get(0)).getUpstreamDKDz() :
        super.getDownstreamDKDz());
  }
  
  /**
   * Returns the upstream friction slope for this canal segment.
   * If this canal segment is an outflow canal, returns the downstream
   * friction slope for its upstream segment.
   * @return upstream friction slope
   */
  public double getUpstreamFrictionSlope(){ 
    return (((getInflowCanal() != null) && (this.inflowGateIsOpen())) ? 
                      getInflowCanal().getDownstreamFrictionSlope() : 
                      super.getUpstreamFrictionSlope());
  }

  /**
   * Returns the average of the upstream and downstream discharges
   * @return the average of the upstream and downstream discharges
   */
  public double getAverageDischarge(){ return ((getUpstreamDischarge() + getDownstreamDischarge()) / 2d);  }

  
  /* User settable maximum outflow */
  
  /**
   * Sets the user-settable maximum outflow
   * @param userSetMax
   */
  public void setMaximumOutflow(double userSetMax){
    userSetMaximumOutflow = (userSetMax >= 0 ? userSetMax : Double.NaN);
  }
  
  /**
   * Gets the maximum outflow; if the user has set a value,
   * that value will be used, but if not, it will be calculated
   * from slope, cross-section (downstream), and roughness
   * assuming 2/3 full and using Manning's equation.
   * Note that this is always a fiction: in truth, the flow of
   * a canal segment can't be throttled at the downstream end.
   * @return the maximum outflow allowed for this segment
   */
  public double getMaximumOutflow(){
    return (!Double.isNaN(this.userSetMaximumOutflow) ? 
        this.userSetMaximumOutflow : 
        getDownstreamManningDischargePerFractionOfChannelHeight(2d/3d)); 
  }
  
  
  
  /* Status */
  
  /**
   * Returns true if the canal segment is considered to be interpolated
   * (that is, not archaeologically attested); false otherwise.
   * @return true or false
   */
  public boolean getIsInterpolated(){   return isInterpolated;  }
  
  /**
   * Sets whether this canal segment is considered to be interpolated
   * (archaeologically attested) or not.
   * @param interpolated
   */
  public void setIsInterpolated(boolean interpolated){  isInterpolated = interpolated;   }

  /**
   * Indicates whether this segment is abandoned
   * @return true if the segment is abandoned, false otherwise
   */
  public boolean getIsAbandoned(){ return isAbandoned; }
  
  /**
   * Sets the status of this segment as abandoned or not
   * @param abandoned true if abandoned, false otherwise
   */
  public void setIsAbandoned(boolean abandoned){ isAbandoned = abandoned; } 

  /**
   * Sets the status of this segment and all downstream branches (recursively to termini)
   * to 'abandoned' 
   */
  public void abandonReach(){
    setIsAbandoned(true);
    this.headGate.setOpen(false);
    Vector V = getOutflowCanals();
    for(int i = 0; i < V.size(); i++)
        ((CanalSegment)V.get(i)).abandonReach();    
  }
  

  /* Wear */
   
  /**
   * Causes erosion and degradation to this Channel Segment.
   * @param time the amount of time for which wear should be 
   * calculated
   */
  public void wear(double time){
    switch (codeModelOption("CANAL_WEAR")){
      case STANDARD_WEAR:{
        // Roughness gets adjusted
        setRoughness(roughness + getParameterSet().getParameter("N_INC_PER_SECOND_PER_Q") * time * getAverageDischarge());
        
        // Canal Fills in slightly or erodes out
        double flowAverageVelocity = getUpstreamFlowVelocity() + getDownstreamFlowVelocity()/2d; 
        if(flowAverageVelocity > getParameterSet().getParameter("EROSION_THRESHOLD")){
          excavate(getParameterSet().getParameter("CANAL_ERODE_PER_SECOND_PER_Q") * time * getAverageDischarge());
        }
        if(flowAverageVelocity < getParameterSet().getParameter("INFILL_THRESHOLD")){
          infill(getParameterSet().getParameter("CANAL_FILL_PER_SECOND_PER_Q") * time * getAverageDischarge());
        }
                
        break;
      }
//      case Schumm(1969)- see Graybill, Gregory, Funkhousr, and Nials(2006) for reference to equations about channel 
//      morphology change through time...
      case NO_WEAR:
      default: {
        // No action
        break;
      }
    }
  }
 

    
  
  /* Water */

  /**
   * Gets a sample of water that represents the water quality flowing into this segment.
   * @return a water sample from the upstream end of the canal segment
   */
  public Water getInflowWaterSample(){
    if(this.headGate.getHeadGate() != null) return new Water(1, this.headGate.getHeadGate().getWaterSample());
    if(this.getInflowCanal().equals(this)) return new Water(1, this.water);
    return new Water(1, this.getInflowCanal().getOutflowWaterSample());
  }
  
  /**
   * @see flowag.channels.AbstractChannelSegment#getTotalWater(double, double)
   */
  public double getTotalWater(double upstreamWaterDepth, double downstreamWaterDepth){  
    double A1 = getUpstreamFlowArea(upstreamWaterDepth);
    double A2 = getDownstreamFlowArea(downstreamWaterDepth);
    double A = (A1 + A2)/2;
    double length = getHeadCoords().planarDistanceToPoint(getEndCoords());
    return length * A;
  }
  
  /**
   * Gets a water sample from the outflow end, representing the quality of water leaving this canal segment
   * @return a water sample from the downstream end of the canal
   */
  public Water getOutflowWaterSample(){
    Water W = new Water(1, getInflowWaterSample());
    switch(codeModelOption("EVAPORATION_ALGORITHM")){
      case NO_CHANNEL_EVAPORATION:{
        return W;
      }
      case BASIC_CHANNEL_EVAPORATION:{
        double mmLossPerSecond = getParameterSet().getParameter("EVAPORATION_MM_PER_DAY") / (24d * 60d * 60d);
        double v = super.averageFlowVelocity();
        double l = super.getLength();
        double flowTime = (v != 0 ? (l / v) : 0);
        double usd = getUpstreamWaterDepth();
        double dsd = getDownstreamWaterDepth();
        double afterFlowUpstreamDepth   = (v > 0 ? usd : usd - (flowTime * mmLossPerSecond)); // Note flow is directional!
        double afterFlowDownstreamDepth = (v < 0 ? dsd : dsd - (flowTime * mmLossPerSecond));
        
        double volumeBeforeFlow = getTotalWater(usd, dsd);
        double volumeAfterFlow  = getTotalWater(afterFlowUpstreamDepth, afterFlowDownstreamDepth);
        W.evaporate((volumeBeforeFlow - volumeAfterFlow)/volumeBeforeFlow);
        return W;        
      }
    }
    return null;
  }
  
  
  /* Fields */
  
  /**
   * Adds a Field to the collection of plots irrigated by this canal.
   * If the plot is already among the plots irrigated by this canal, 
   * it is not added again.
   * @param p Plot to be added
   */
  public void addPlot(Field p){
    if(!(plotsIrrigated.contains(p))){
      plotsIrrigated.add(p);
    }
  }

  /**
   * Removes the specified plot from the collection of plots 
   * irrigated by this canal. If the plot is not among the plots 
   * irrigated by this canal, it is not removed, but no error occurs.
   * @param p Plot to be removed
   */
  public void removePlot(Field p){
    plotsIrrigated.remove(p);    
  }
  
  /**
   * Indicates whether this canal segment irrigates the specified field
   * @param f the field requested
   * @return true if the field is irrigated by this canal segment, false otherwise
   */
  public boolean irrigatesField(Field f){
    return plotsIrrigated.contains(f);
  }
  
  /**
   * Number of plots irrigated by this canal
   * @return the number of plots irrigated by this canal
   */
  public long plotsIrrigated(){
    return plotsIrrigated.size();
  }
  
  /**
   * Total area irrigated by this canal.
   * Does not include area irrigated by canals fed by this canal.
   * @return The sum of the areas of all the plots fed by this canal
   */
  public double areaIrrigated(){
    double A = 0;
    for(int i = 0; i<plotsIrrigated.size(); i++){
      Field P = (Field)plotsIrrigated.get(i);
      A += P.getArea();
    }
    return A;
  }
  
  /**
   * Total area irrigated by this canal, including that irrigated by
   * all canals fed by this canal.
   * @param OnlyOpenGates True if you trace through the canal tree
   * only using gates that are 'open'; false if you include all
   * canals regardless of gate status.
   * @return The sum of the areas of all the plots fed by this canal
   * including plots fed by canals that are fed from this canal.
   */
  public double areaIrrigatedAll(boolean OnlyOpenGates){
    double A = 0;
    for(int i = 0; i<plotsIrrigated.size(); i++){
      Field P = (Field)plotsIrrigated.get(i);
      A += P.getArea();
    }
    for(int i = 0; i<outflowCanalGates.size(); i++){
      Gate G = (Gate)outflowCanalGates.get(i);
      // Only calculate if gate is
      if(!OnlyOpenGates || G.isOpen()){
        CanalSegment c = G.getOutflow();
        // Recursive- calls self for all downstream canals
        A += c.areaIrrigatedAll(OnlyOpenGates); 
      }      
    }
    return A;
  }  
  
  /**
   * Deposits water onto fields. Water quality is determined by getOutflowWaterSample().
   * Water volume is calculated from the outflow rate over the specified time.
   * @param t time interval over which water is discharged
   */
  public void waterFields(double t){
    if(plotsIrrigated == null || plotsIrrigated.size() == 0) return;
    if(this.getDownstreamDischarge() == 0) return;
    Water W = this.getOutflowWaterSample();
    double totalV = this.getDownstreamDischarge() * t;
    double totalA = areaIrrigated();
    for(int i = 0; i < plotsIrrigated.size(); i++){
      Field F = (Field)plotsIrrigated.get(i);
      double newWaterVol = totalV * (F.getArea() / totalA);
      F.addWater(new Water(newWaterVol, W)); // Apportion water proportionally to all irrigated fields
    }    
  }
  
  
  /* Selecting Segments of the Canal System */
  
  /**
   * Marks this canal segment and all downstream canal
   * segments as selected
   */
  public void selectWithDownstream(){
    this.select();
    Vector v = getOutflowCanals();
    for(int i = 0; i < v.size(); i++){
      ((CanalSegment)v.get(i)).selectWithDownstream();
    }
  }
  
  /**
   * Marks this canal segment and all downstream canal
   * segments as unselected
   */
  public void unSelectWithDownstream(){
    this.unSelect();
    Vector v = getOutflowCanals();
    for(int i = 0; i < v.size(); i++){
      ((CanalSegment)v.get(i)).unSelectWithDownstream();
    }
  }

  
  /* Text Reporting */
 
  /**
   * Returns the type of the canal (equivalent to the type of downstream cross section).
   * @return a String representing the type of the canal
   */
  public String getCanalType(){  return downstreamCrossSection.type();   }
   
  /**
   * Returns a String representing the status report for this CanalSegment
   * @return String representing a status report for this CanalSegment
   */
  public String report(){
    String Report = "";
    Point headCoords = getHeadCoords();
    Point endCoords = getEndCoords();
    ChannelCrossSection UpstreamCrossSection = super.upstreamCrossSection;
    ChannelCrossSection DownstreamCrossSection = super.downstreamCrossSection;
    
    boolean debug = (getParameterSet().getParameter("DEBUG") != 0);
    int rdp = (int)(Math.floor(getParameterSet().getParameter("ROUNDING_DECIMAL_PLACES")));
    rdp=9;
    
    // These are the variables we'll be retrieving
    String bedSlope = (!debug ? TextFormatter.roundWithUnits(getBedSlope(), rdp, "m/m") : "" + getBedSlope());
    String length   = (!debug ? TextFormatter.roundWithUnits(getLength(), rdp, "m") : "" + getLength());
    String upsTopW  = (!debug ? TextFormatter.roundWithUnits(UpstreamCrossSection.getTopWidth(), rdp, "m") : "" + UpstreamCrossSection.getTopWidth());
    String upsBotW  = (!debug ? TextFormatter.roundWithUnits(UpstreamCrossSection.getBottomWidth(), rdp, "m") : "" + UpstreamCrossSection.getBottomWidth());
    String upsChH   = (!debug ? TextFormatter.roundWithUnits(UpstreamCrossSection.getChannelHeight(), rdp, "m") : "" + UpstreamCrossSection.getChannelHeight());
    String dwnTopW  = (!debug ? TextFormatter.roundWithUnits(DownstreamCrossSection.getTopWidth(), rdp, "m") : "" + DownstreamCrossSection.getTopWidth());
    String dwnBotW  = (!debug ? TextFormatter.roundWithUnits(DownstreamCrossSection.getBottomWidth(), rdp, "m") : "" + DownstreamCrossSection.getBottomWidth());
    String dwnChH   = (!debug ? TextFormatter.roundWithUnits(DownstreamCrossSection.getChannelHeight(), rdp, "m") : "" + DownstreamCrossSection.getChannelHeight());
    String totVol   = (!debug ? TextFormatter.roundWithUnits(getTotalVolume(), rdp, "m^3") : "" + getTotalVolume());
    String watVol   = (!debug ? TextFormatter.roundWithUnits(water.getVolume(), rdp, "m^3") : "" + water.getVolume());
    String n        = (!debug ? TextFormatter.roundWithUnits(super.getRoughness(), 4, "(dimensionless)") : "" + super.getRoughness());
    
    // Upstream values
    String upsWatD  = (!debug ? TextFormatter.roundWithUnits(getUpstreamWaterDepth(), rdp, "m") : "" + getUpstreamWaterDepth());
    String upsWatW  = (!debug ? TextFormatter.roundWithUnits(getUpstreamWaterSurfaceWidth(), rdp, "m") : "" + getUpstreamWaterSurfaceWidth());
    String upsWetP  = (!debug ? TextFormatter.roundWithUnits(getUpstreamWettedPerimeter(), rdp, "m") : "" + getUpstreamWettedPerimeter());
    String upsFlowA = (!debug ? TextFormatter.roundWithUnits(getUpstreamFlowArea(), rdp, "m") : "" + getUpstreamFlowArea());
    String upsHydrR = (!debug ? TextFormatter.roundWithUnits(getUpstreamHydraulicRadius(), rdp, "m") : "" + getUpstreamHydraulicRadius());
    String upsQ     = (!debug ? TextFormatter.roundWithUnits(getUpstreamDischarge(), rdp, "m^3/s") : "" + getUpstreamDischarge());
    String upsV     = (!debug ? TextFormatter.roundWithUnits(getUpstreamFlowVelocity(), rdp, "m") : "" + getUpstreamFlowVelocity());
    String upsK     = (!debug ? TextFormatter.roundWithUnits(getUpstreamConveyance(), rdp, "m") : "" + getUpstreamConveyance());
    String upsSf    = (!debug ? TextFormatter.roundWithUnits(getUpstreamFrictionSlope(), rdp, "m") : "" + getUpstreamFrictionSlope());
    String upsdAdz  = (!debug ? TextFormatter.roundWithUnits(getUpstreamDADz(), rdp, "m") : "" + getUpstreamDADz());
    String upsdKdz  = (!debug ? TextFormatter.roundWithUnits(getUpstreamDKDz(), rdp, "m") : "" + getUpstreamDKDz());
    String upsAvgD  = (!debug ? TextFormatter.roundWithUnits(getUpstreamAvgDepth(), rdp, "m") : "" + getUpstreamAvgDepth());
    String upsMannV = (!debug ? TextFormatter.roundWithUnits(getUpstreamManningVelocity(), rdp, "m") : "" + getUpstreamManningVelocity());
    String upsMannQ = (!debug ? TextFormatter.roundWithUnits(getUpstreamManningDischarge(), rdp, "m") : "" + getUpstreamManningDischarge());
    
    // Downstream values
    String dwnWatD  = (!debug ? TextFormatter.roundWithUnits(getDownstreamWaterDepth(), rdp, "m") : "" + getDownstreamWaterDepth());
    String dwnWatW  = (!debug ? TextFormatter.roundWithUnits(getDownstreamWaterSurfaceWidth(), rdp, "m") : "" + getDownstreamWaterSurfaceWidth());
    String dwnWetP  = (!debug ? TextFormatter.roundWithUnits(getDownstreamWettedPerimeter(), rdp, "m") : "" + getDownstreamWettedPerimeter());
    String dwnFlowA = (!debug ? TextFormatter.roundWithUnits(getDownstreamFlowArea(), rdp, "m") : "" + getDownstreamFlowArea());
    String dwnHydrR = (!debug ? TextFormatter.roundWithUnits(getDownstreamHydraulicRadius(), rdp, "m") : "" + getDownstreamHydraulicRadius());
    String dwnQ     = (!debug ? TextFormatter.roundWithUnits(getDownstreamDischarge(), rdp, "m^3/s") : "" + getDownstreamDischarge());
    String dwnV     = (!debug ? TextFormatter.roundWithUnits(getDownstreamFlowVelocity(), rdp, "m") : "" + getDownstreamFlowVelocity());
    String dwnK     = (!debug ? TextFormatter.roundWithUnits(getDownstreamConveyance(), rdp, "m") : "" + getDownstreamConveyance());
    String dwnSf    = (!debug ? TextFormatter.roundWithUnits(getDownstreamFrictionSlope(), rdp, "m") : "" + getDownstreamFrictionSlope());
    String dwndAdz  = (!debug ? TextFormatter.roundWithUnits(getDownstreamDADz(), rdp, "m") : "" + getDownstreamDADz());
    String dwndKdz  = (!debug ? TextFormatter.roundWithUnits(getDownstreamDKDz(), rdp, "m") : "" + getDownstreamDKDz());
    String dwnAvgD  = (!debug ? TextFormatter.roundWithUnits(getDownstreamAvgDepth(), rdp, "m") : "" + getDownstreamAvgDepth());
    String dwnMannV = (!debug ? TextFormatter.roundWithUnits(getDownstreamManningVelocity(), rdp, "m") : "" + getDownstreamManningVelocity());
    String dwnMannQ = (!debug ? TextFormatter.roundWithUnits(getDownstreamManningDischarge(), rdp, "m") : "" + getDownstreamManningDischarge());
    
    
    
    Report += "   Canal: " + canalSegmentID + "" + " (Type: " + getCanalType() + ")\n";
    Report += "     Head:                  " + headCoords.coordString() + "\n";
    Report += "     End:                   " + endCoords.coordString() + "\n";
    Report += "     Slope:                 " + bedSlope + "\n";
    Report += "     Length:                " + length + "\n";
    Report += "     Up. Top Width:         " + upsTopW + "\n";
    Report += "     Up. Bottom Width:      " + upsBotW + "\n";
    Report += "     Up. Ch. Height:        " + upsChH + "\n";
    Report += "     Dwn. Top Width:        " + dwnTopW + "\n";
    Report += "     Dwn. Bottom Width:     " + dwnBotW + "\n";
    Report += "     Dwn. Ch. Height:       " + dwnChH + "\n";
    
    Report += "     Total Volume:          " + totVol + "\n"; 
    Report += "     Water Volume:          " + watVol + "\n";
    Report += "     Inflow Canal:          " + getInflowCanal().getCanalSegmentID() + "\n";
    Report += "     Outflow Canals:        " + outflowCanalGates.size() + "\n";

    for(int i=0; i<outflowCanalGates.size(); i++){
      CanalSegment c = ((Gate)outflowCanalGates.get(i)).getOutflow();
      Report += "         " + (i +1) + ": " + c.getCanalSegmentID();
      double hc = Math.toDegrees(endCoords.headingChange(headCoords, c.getEndCoords()));
      String HC = (!debug ? TextFormatter.roundWithUnits(hc, 2, "deg") : "" + hc);
      Report += " heading change: " + HC + "\n";
    }
    Report += "     Roughness:             " + n + "\n";

    // Upstream
//    Report += "     Ups. Water Depth:      " + upsWatD + "\n";
//    Report += "     Ups. Water Width:      " + upsWatW + "\n";
//    Report += "     Ups. Wetted Perimeter: " + upsWetP + "\n";
//    Report += "     Ups. Flow Area:        " + upsFlowA + "\n";    
//    Report += "     Ups. Hydraulic Radius: " + upsHydrR + "\n";
    Report += "     Ups. Discharge:        " + upsQ + " (" + getUpstreamDischarge() + ") \n";
//    Report += "     Ups. Flow Velocity:    " + upsV + "\n";
//    Report += "     Ups. Conveyance:       " + upsK + "\n";
//    Report += "     Ups. Friction Slope:   " + upsSf + "\n";
//    Report += "     Ups. dA/dZ:            " + upsdAdz + "\n";
//    Report += "     Ups. dK/dZ:            " + upsdKdz + "\n";
//    Report += "     Ups. Average Depth:    " + upsAvgD + "\n";
//    Report += "     Ups. Manning Vel.:     " + upsMannV + "\n";
//    Report += "     Ups. Manning Disch.:   " + upsMannQ + "\n";
//
//    // Downstream
//    Report += "     Dwn. Water Depth:      " + dwnWatD + "\n";
//    Report += "     Dwn. Water Width:      " + dwnWatW + "\n";
//    Report += "     Dwn. Wetted Perimeter: " + dwnWetP + "\n";
//    Report += "     Dwn. Flow Area:        " + dwnFlowA + "\n";    
//    Report += "     Dwn. Hydraulic Radius: " + dwnHydrR + "\n";
    Report += "     Dwn. Discharge:        " + dwnQ + "\n";
//    Report += "     Dwn. Flow Velocity:    " + dwnV + "\n";
//    Report += "     Dwn. Conveyance:       " + dwnK + "\n";
//    Report += "     Dwn. Friction Slope:   " + dwnSf + "\n";
//    Report += "     Dwn. dA/dZ:            " + dwndAdz + "\n";
//    Report += "     Dwn. dK/dZ:            " + dwndKdz + "\n";
//    Report += "     Dwn. Average Depth:    " + dwnAvgD + "\n";
//    Report += "     Dwn. Manning Vel.:     " + dwnMannV + "\n";
//    Report += "     Dwn. Manning Disch.:   " + dwnMannQ + "\n";
//
//
//    for(int i = 0; i < plotsIrrigated.size(); i++){
//      Report += ((Field)plotsIrrigated.get(i)).getReport();
//    }

    Report += "\n\n";
    return Report; 
  }  

  /**
   * Gets a summary of the flow through this canal segment
   * @return a String that contains a summary of the flow through this canal segment
   */
  public String flowReport(){
    String r = "";
    r += getCanalSegmentID() + "U Z: " + TextFormatter.toWidth(BasicMath.roundDouble(getUpstreamStage(),8),10) + 
                             "    D: " + TextFormatter.toWidth(BasicMath.roundDouble(getUpstreamWaterDepth(), 8),10) + 
                             "    Q: " + TextFormatter.toWidth(BasicMath.roundDouble(getUpstreamDischarge(),8),10) + "  @  " + super.getHeadCoords().coordString() + "\n";
    Vector C = getOutflowCanals();
    if(C.size() > 0 ){
      for(int i = 0; i < C.size(); i++){
        if(C.size() > 1) r += "Branch: " + i + "\n";
        CanalSegment c = (CanalSegment)C.get(i);
        r += c.flowReport();
      }
    }
    else{
      r += getCanalSegmentID() + "D Z: " + TextFormatter.toWidth(BasicMath.roundDouble(getDownstreamStage(),8),10) + 
                               "    D: " + TextFormatter.toWidth(BasicMath.roundDouble(getDownstreamWaterDepth(), 8),10) + 
                               "    Q: " + TextFormatter.toWidth(BasicMath.roundDouble(getDownstreamDischarge(),8),10) + "  @  " + super.getEndCoords().coordString() + "\n";
    }
    return r;
  }

  /**
   * Gets a HEC-RAS import cross-section as a string
   * @param head head of this canal, or its end
   * @param dist distance from some point to the cross-section point
   * @param streamID ID of the stream to which this segment belongs
   * @param reachID ID of the reach to which this segment belongs
   * @return A String object depicting the geometry of this 
   * cross-section in HEC-RAS input format.
   */
  public String getCrossSection(boolean head, double dist, String streamID, String reachID){
    StringWriter out = new StringWriter();
    
    // Note: this will use the length of this canal segment
    // if it is called to return the end coordinates
    Point headCoords = getHeadCoords();
    Point endCoords = getEndCoords();
    double length = headCoords.planarDistanceToPoint(endCoords);
    out.write("  CROSS-SECTION:\n");
    out.write("    STREAM ID: " + streamID + "\n");
    out.write("    REACH ID: " + reachID + "\n");
    out.write("    STATION: " + dist + "\n");
    out.write("    BANK POSITIONS: 0.0 1.0\n");
    out.write("    REACH LENGTHS: " + length + " , " + length + ", " + length + "\n");
    out.write("    NVALUES:\n" +
              "         0.0, " + getRoughness() + " \n" +
              "         1.0, " + getRoughness() + " \n");
    Vector vLine;
    if(head){
      vLine = super.upstreamCrossSection.getCutAndSurfaceLine(headCoords,headCoords.headingTo(endCoords));
    }
    else{
      vLine = super.downstreamCrossSection.getCutAndSurfaceLine(endCoords, headCoords.headingTo(endCoords));
    }    
    out.write("    CUT LINE:\n");
    Point temp = (Point)vLine.firstElement();
    out.write("       " + temp.getX() + ", " + temp.getY() + " \n");
    temp = (Point)vLine.lastElement();
    out.write("       " + temp.getX() + ", " + temp.getY() + " \n");
    out.write("    SURFACE LINE:\n");
    for(int i = 0; i < vLine.size(); i++){
      temp = (Point)vLine.get(i);
      out.write("       " + temp.getX() + ", " + temp.getY() + ", " + temp.getZ() + " \n");  
    }
    out.write("  END:\n\n");
    
    return out.toString();
  }
  
  /* Spatial Element */
  
  /**
   * @see abcm.space.SpatialElement#getBoundaryPolygons()
   */
  public Vector getBoundaryPolygons(){
    Vector V = new Vector();
    
    Vector poly1Vector = new Vector();
    poly1Vector.add(this.getUpstreamLeftDrawCorner());
    poly1Vector.add(this.getDownstreamLeftDrawCorner());
    poly1Vector.add(this.getDownstreamRightDrawCorner());
    poly1Vector.add(this.getUpstreamRightDrawCorner());
    
    V.add(new Polygon(poly1Vector));
    if(this.outflowCanalGates.size() > 1) V.add(this.outflowJunctionPolygon);    
//    if(true) System.out.println("Canal Segment: " + this.canalSegmentID + " Getting boundary polygons: " + V.size());
    return V;
  }
  
  /**
   * @see abcm.space.SpatialElement#getBermPolygons()
   */
  public Vector getBermPolygons(){
    Vector V = getBoundaryPolygons();
    Vector ret = new Vector();
    for(int i = 0; i < V.size(); i++){
      Polygon P = (Polygon)V.get(i);
      ret.add(P.buffer(10));
    }
    return ret;
  }
  
  
  /**
   * @see abcm.space.SpatialElement#getMaxX()
   */
  public double getMaxX(){
    Vector V = getBoundaryPolygons();
    double max = Double.NEGATIVE_INFINITY;
    for(int i = 0; i < V.size(); i++){
      Polygon P = (Polygon)V.get(i);
      Vector v = P.getPoints();
      for(int j = 0; j < v.size(); j++){
        Point p = (Point)v.get(j);
        if(p.getX() > max) max = p.getX();
      }
    }
    return max;  
  }

  /**
   * @see abcm.space.SpatialElement#getMaxY()
   */
  public double getMaxY(){
    Vector V = getBoundaryPolygons();
    double max = Double.NEGATIVE_INFINITY;
    for(int i = 0; i < V.size(); i++){
      Polygon P = (Polygon)V.get(i);
      Vector v = P.getPoints();
      for(int j = 0; j < v.size(); j++){
        Point p = (Point)v.get(j);
        if(p.getY() > max) max = p.getY();
      }
    }
    return max;  

  }

  /**
   * @see abcm.space.SpatialElement#getMaxZ()
   */
  public double getMaxZ(){
    Vector V = getBoundaryPolygons();
    double max = Double.NEGATIVE_INFINITY;
    for(int i = 0; i < V.size(); i++){
      Polygon P = (Polygon)V.get(i);
      Vector v = P.getPoints();
      for(int j = 0; j < v.size(); j++){
        Point p = (Point)v.get(j);
        if(p.getZ() > max) max = p.getZ();
      }
    }
    return max;  
  }

  /**
   * @see abcm.space.SpatialElement#getMinX()
   */
  public double getMinX(){
    Vector V = getBoundaryPolygons();
    double min = Double.POSITIVE_INFINITY;
    for(int i = 0; i < V.size(); i++){
      Polygon P = (Polygon)V.get(i);
      Vector v = P.getPoints();
      for(int j = 0; j < v.size(); j++){
        Point p = (Point)v.get(j);
        if(p.getX() < min) min = p.getX();
      }
    }
    return min;  
  }

  /**
   * @see abcm.space.SpatialElement#getMinY()
   */
  public double getMinY(){
    Vector V = getBoundaryPolygons();
    double min = Double.POSITIVE_INFINITY;
    for(int i = 0; i < V.size(); i++){
      Polygon P = (Polygon)V.get(i);
      Vector v = P.getPoints();
      for(int j = 0; j < v.size(); j++){
        Point p = (Point)v.get(j);
        if(p.getY() < min) min = p.getY();
      }
    }
    return min;  

  }

  /**
   * @see abcm.space.SpatialElement#getMinZ()
   */
  public double getMinZ(){
    return getEndCoords().getZ();
  }
  
  
  /**
   * Not yet implemented. Returns zero.
   * @see abcm.space.HasElevation#getZ(abcm.math.Point)
   */
  public double getZ(Point P){
    return 0;  
  }

  
  
  /**
   * @see abcm.space.SpatialElement#altered()
   */
  public boolean altered(){    return true;  }
  
  /* Drawing */ 
  
  /**
   * Gets the upstream left corner of this canal segment for drawing.
   * If the canal segment has no inflow or is the only outflow
   * of its upstream neighbor, the cross section corners are used.
   * If it is coming out of a junction, the junction polygon must be used,
   * which is the responsibility of the upstream segment.
   * @return a Point marking the upstream left corner
   */
  public Point getUpstreamLeftDrawCorner(){
    if((this.headGate.getSource() != null)  && (this.headGate.getSource().outflowCanalGates.size() > 1)){
      return (Point)this.headGate.getSource().upstreamLeftCornersOfOutflows.get(this);
    }
    return this.getUpstreamLeftBankCoordinate();
  }
  
  /**
   * Gets the upstream right corner of this canal segment for drawing.
   * If the canal segment has no inflow or is the only outflow
   * of its upstream neighbor, the cross section corners are used.
   * If it is coming out of a junction, the junction polygon must be used,
   * which is the responsibility of the upstream segment.
   * @return a Point marking the upstream left corner
   */
  public Point getUpstreamRightDrawCorner(){
    if((this.headGate.getSource() != null)  && (this.headGate.getSource().outflowCanalGates.size() > 1)){
      return (Point)this.headGate.getSource().upstreamRightCornersOfOutflows.get(this);
    }
    return this.getUpstreamRightBankCoordinate();
  }
  
  // For outflows, there are three possibilities:
  // Either this canal has no downstream segment,
  // or it has one, or it has more than one.
  // If 0, return the simple coordinates
  // If 1, return the _upstream_ coordinates of the
  // downstream outflow segment
  // If >1, return the coordinates in the junction polygon
  
  /**
   * Gets the downstream left corner of this canal segment.
   * Uses one of three possibilities: the canal may have no
   * downstream segment, in which case the downstream cross
   * section is used. If it has exactly one outflow segment,
   * the upstream coordinates of the downstream segment are used.
   * If it has multiple outflows (a junction), the coordinates
   * of the junction polygon are used.
   * @return a point representing the left downstream draw corner
   * of this segment
   */
  public Point getDownstreamLeftDrawCorner(){
    if(this.outflowCanalGates.size() == 0){
      return this.getDownstreamLeftBankCoordinate();
    }
    else if(this.outflowCanalGates.size() == 1){
      return ((CanalSegment)((Gate)this.outflowCanalGates.get(0)).getOutflow()).getUpstreamLeftDrawCorner();
    }
    else{
      return this.downstreamLeftDrawCorner;
    }
  }
  
  /**   
   * Gets the downstream right corner of this canal segment.
   * Uses one of three possibilities: the canal may have no
   * downstream segment, in which case the downstream cross
   * section is used. If it has exactly one outflow segment,
   * the upstream coordinates of the downstream segment are used.
   * If it has multiple outflows (a junction), the coordinates
   * of the junction polygon are used.
   * @return a point representing the right downstream draw corner
   * of this segment
   */
  public Point getDownstreamRightDrawCorner(){
    if(this.outflowCanalGates.size() == 0){
      return this.getDownstreamRightBankCoordinate();
    }
    else if(this.outflowCanalGates.size() == 1){
      return ((CanalSegment)((Gate)this.outflowCanalGates.get(0)).getOutflow()).getUpstreamRightDrawCorner();
    }
    else{
      return this.downstreamRightDrawCorner;
    }
  }
  
  /**
   * Gets an array of all four corners of this segment, for graphical
   * representation. The list is ordered: upstream left, downstream left,
   * downstream right, upstream right.
   * @return the four draw corners.
   */
  public ArrayList getDrawCorners(){
    ArrayList drawCorners = new ArrayList();
    drawCorners.add(this.getUpstreamLeftDrawCorner());
    drawCorners.add(this.getDownstreamLeftDrawCorner());
    drawCorners.add(this.getDownstreamRightDrawCorner());
    drawCorners.add(this.getUpstreamRightDrawCorner());
    return drawCorners;
  }
  
  /**
   * Gets the four corners representing the water in this canal segment.
   * Order is upstream left, downstream left, downstream right, upstream right.
   * @return the list of all four water corners for this segment
   */
  public ArrayList getWaterCorners(){
    ArrayList waterCorners = new ArrayList();
    ArrayList upstream = getUpstreamWaterCorners();
    ArrayList downstream = getDownstreamWaterCorners();
    waterCorners.add(upstream.get(0));
    waterCorners.addAll(downstream);
    waterCorners.add(upstream.get(1));
    return waterCorners;
  }
  
  /**
   * Gets the upstream water corners for this segment. Order is left, right.
   * @return the two corners representing the usptream water corners for this segment.
   */
  public ArrayList getUpstreamWaterCorners(){
    ArrayList waterCorners = new ArrayList();
    // If this is the downstream part of a junction with multiple outflows,
    // must get this from the inflow source
    if((this.headGate.getSource() != null)  && (this.headGate.getSource().outflowCanalGates.size() > 1)){
      waterCorners = this.headGate.getSource().getOutflowWaterCorners(this);
    }
    else{
      Point headCoords = getHeadCoords();
      Point endCoords  = getEndCoords();
      double h = headCoords.headingTo(endCoords);
      double z1 = getUpstreamStage();
      double uwsw = upstreamCrossSection.getWaterSurfaceWidth(z1);
      Point P1 = Geometry.getOffsetPoint(headCoords, h + AbstractChannelSegment.LEFT, uwsw / 5d);
      waterCorners.add(new Point(P1.getX(),P1.getY(), z1));
      Point P2 = Geometry.getOffsetPoint(headCoords, h + AbstractChannelSegment.RIGHT, uwsw / 5d); 
      waterCorners.add(new Point(P2.getX(), P2.getY(), z1));
    }
    return waterCorners;
  }
  
  /**
   * Gets the downstream water corners for this segment. Order is left, right.
   * @return the two corners representing the downstream water corners for this segment.
   */
  public ArrayList getDownstreamWaterCorners(){
    ArrayList waterCorners = new ArrayList();
    Point headCoords = getHeadCoords();
    Point endCoords = getEndCoords();
    // If this is the terminus, just use the normal:
    if(this.outflowCanalGates.size() == 0){
      double h = headCoords.headingTo(endCoords);
      double z2 = getDownstreamStage();
      double dwsw = downstreamCrossSection.getWaterSurfaceWidth(z2);
      waterCorners.add(Geometry.getOffsetPoint(endCoords, h + AbstractChannelSegment.LEFT, dwsw / 5d));
      waterCorners.add(Geometry.getOffsetPoint(endCoords, h + AbstractChannelSegment.RIGHT, dwsw / 5d));
    }
    // If this has only one outflow, use that outflow's upstream coordinates
    else if(this.outflowCanalGates.size() == 1){
      waterCorners = ((Gate)this.outflowCanalGates.get(0)).getOutflow().getUpstreamWaterCorners();
    }
    // Otherwise this is a junction and is treated in a special way
    else{
      Point c1 = this.getDownstreamLeftDrawCorner();
      Point c2 = this.getDownstreamRightDrawCorner();
      double fraction = (this.z2) / (c1.getZ() - endCoords.getZ());
      if(fraction > 1)  fraction = 1;
      if(fraction <= 0) fraction = 0;
      Point wc1 = Geometry.pointBetween(c1, c2, .2 - (fraction * .2d));
      Point wc2 = Geometry.pointBetween(c1, c2, .8 + (fraction * .2d));
      waterCorners.add(new Point(wc1.getX(), wc1.getY(), z2));
      waterCorners.add(new Point(wc2.getX(), wc2.getY(), z2));
    }
    return waterCorners;
  }
  
  /**
   * Gets the list of corners representing the water flowing out
   * of this segment into a specified segment
   * @param outflow the outflow segment requested 
   * @return the corners of the water flowing into the requested segment
   */
  public ArrayList getOutflowWaterCorners(CanalSegment outflow){
    Point c1 = (Point)this.upstreamLeftCornersOfOutflows.get(outflow);
    Point c2 = (Point)this.upstreamRightCornersOfOutflows.get(outflow);
    Point endCoords = getEndCoords();
    double fraction = (this.z2) / (c1.getZ() - endCoords.getZ());
    if(fraction > 1)  fraction = 1;
    if(fraction <= 0) fraction = 0;
    Point wc1 = Geometry.pointBetween(c1, c2, .2 - fraction * .2d);
    Point wc2 = Geometry.pointBetween(c1, c2, .8 + fraction * .2d);
    ArrayList waterCorners = new ArrayList();
    waterCorners.add(new Point(wc1.getX(), wc1.getY(), z2));
    waterCorners.add(new Point(wc2.getX(), wc2.getY(), z2));
    return waterCorners;
  }
  
  /**
   * Gets the list of outflow water corners if this segment
   * flows into a junction. The order is clockwise.
   * @return the list of outflow water corners.
   */
  public ArrayList getJunctionWaterCorners(){
    ArrayList waterCorners = new ArrayList();
    ArrayList selfWaterCorners = this.getDownstreamWaterCorners();
    waterCorners.add(selfWaterCorners.get(0));
    for(int i = 0; i < outflowCanalGates.size(); i++){
      ArrayList outWaterCorners = this.getOutflowWaterCorners(((Gate)outflowCanalGates.get(i)).getOutflow());
      waterCorners.add(outWaterCorners.get(0));
      waterCorners.add(outWaterCorners.get(1));
    }
    waterCorners.add(selfWaterCorners.get(1));
    return waterCorners;
  }
  
  
  /**
   * Adds nodes describing an SVG representation of this CanalSegment, in XML, 
   * to an XML document.
   * @param svg SVGAssist object that is constructing the SVG document to
   * which this is being added
   */
  public void addToSVG(SVGAssist svg){
    
    class CanalSegmentDrawer implements SVGRepresentable{
      
      class CanalSegmentLabelDrawer implements SVGRepresentable{
        
        public void addToSVG(SVGAssist svg){
          svg.addBackground("white");
          svg.addText(report(), .5, 15, 1);   
        }
      }
      
      
      public void addToSVG(SVGAssist svg){
        if(!isAbandoned){
          svg.addPolygon(SVGAssist.getPointsList(getDrawCorners()), "none","red",svg.getUnitsPerInchX()/100d, "onclick=\"showCanalSegDetails(evt, '" + canalSegmentID + "');\"");
          svg.addPolygon(SVGAssist.getPointsList(getWaterCorners()), "blue","blue",svg.getUnitsPerInchX()/100d, "onclick=\"showCanalSegDetails(evt, '" + canalSegmentID + "');\"");
          svg.addPopup(((Point)(getDrawCorners().get(0))).getX(), ((Point)(getDrawCorners().get(0))).getY(), 
              3, 5, 
              .5, .5, 
//              new SVGRepresentable[] {new CanalSegmentLabelDrawer()}, 
//                  "grey", 1/100, getCanalSegmentID(), "visibility=\"hidden\" id=\"canalSegLabel_" + getCanalSegmentID() + "\"  onclick=\"showCanalSegDetails(evt,'canalSegLabel_" + getCanalSegmentID() + "');\"");
          new SVGRepresentable[] {new CanalSegmentLabelDrawer()}, 
            "grey", 1/100, getCanalSegmentID(), "visibility=\"hidden\" id=\"canalSegLabel_" + getCanalSegmentID() + "\"  onclick=\"showCanalSegDetails(evt,'" + getCanalSegmentID() + "');\"");
          }
        else{
          System.out.println("DRAWING ABANDONED!!!");
          svg.addPolygon(SVGAssist.getPointsList(getDrawCorners()), "yellow", "yellow", svg.getUnitsPerInchX()/100d, "onclick=\"showCanalSegDetails(evt, '" + canalSegmentID + "');\"");
        }
        if(outflowCanalGates.size() > 1){
          svg.addPolygon(SVGAssist.getPointsList(outflowJunctionPolygon.getPoints()), "none", "red", svg.getUnitsPerInchX()/100d);
          svg.addPolygon(SVGAssist.getPointsList(getJunctionWaterCorners()), "blue", "blue", svg.getUnitsPerInchX()/100d);
        }
      }
    
    }
    svg.addG("Canal Segment: " + canalSegmentID, "", new SVGRepresentable[]{new CanalSegmentDrawer()});
    
  }

  
  /**
   * Gets a cut-and-surface line at the upstream end of this segment.
   * If it is at a junction, uses getJunctionCutAndSurfaceLine, but
   * otherwise uses this segment's upstream cross section.
   * @return cut-and-surface line as a Vector of Points
   */
  public Vector getUpstreamDrawCutAndSurfaceLine(){
    Point headCoords = getHeadCoords();
    // As before, upstream has only two conditions:
    // If this is an outflow from a junction
    if((this.headGate.getSource() != null)  && (this.headGate.getSource().outflowCanalGates.size() > 1)){
      return getJunctionCutAndSurfaceLine(getUpstreamLeftDrawCorner(), getUpstreamRightDrawCorner(), headCoords.getZ());  
    }
    return upstreamCrossSection.getCutAndSurfaceLine(headCoords, h);
  }
  
  
  /**
   * Gets a cut-and-surface line at the downstream terminus of this canal segment.
   * If this has a single outflow, the upstream end of the downstream
   * segment is used; if it is a terminus the downstream cross section
   * of this segment is used. If a junction, calls getJunctionCutAndSurfaceLine
   * using the left and right corners as arguments.
   * @return cut-and-surface line as a Vector of Point objects. 
   */
  public Vector getDownstreamDrawCutAndSurfaceLine(){
    Point headCoords = getHeadCoords();
    Point endCoords = getEndCoords();

    // If this is a terminus
    if(this.outflowCanalGates.size() == 0){
      return downstreamCrossSection.getCutAndSurfaceLine(endCoords, h);
    }
    // If this has a single outflow
    else if(this.outflowCanalGates.size() == 1){
      return ((Gate)this.outflowCanalGates.get(0)).getOutflow().getUpstreamDrawCutAndSurfaceLine();
    }    
    // If this is a junction
    else{
      // Note that this seems backwards, but it's because the getJunctionCutAndSurfaceLine method expects to be doing upstream ends
      // not downstream
      return getJunctionCutAndSurfaceLine(this.getDownstreamLeftDrawCorner(), this.getDownstreamRightDrawCorner(), endCoords.getZ()); 
    } 
  }
  
  /**
   * Gets a cut and surface line across a junction. The line is faked- it takes the
   * corners passed and creates a trapezoidal section with points at the specified
   * elevation at .2 and .8 across the gap.
   * @param leftCorner left corner of the section
   * @param rightCorner right corner of the section
   * @param zElevation elevation to be used as section bottom
   * @return the Vector of points
   */
  private Vector getJunctionCutAndSurfaceLine(Point leftCorner, Point rightCorner, double zElevation){
    Vector V = new Vector();
    V.add(leftCorner);
    
    Point p1 = Geometry.pointBetween(leftCorner, rightCorner, .2);
    V.add(new Point(p1.getX(), p1.getY(), zElevation));
    
    Point p2 = Geometry.pointBetween(leftCorner, rightCorner, .8);
    V.add(new Point(p2.getX(), p2.getY(), zElevation));
    
    V.add(rightCorner);
    return V;
  }
  
  /**
   * @see abcm.space.SpatialElement#getTriangulatedTopographyWithinBoundaryPolygons()
   */
  public TriangulatedTopography getTriangulatedTopographyWithinBoundaryPolygons(){
    Point headCoords = getHeadCoords();
    Point endCoords = getEndCoords();

    Vector pointsList1 = getUpstreamDrawCutAndSurfaceLine();
    Vector pointsList2 = getDownstreamDrawCutAndSurfaceLine();
    Vector finalPointsList1 = mergeCutAndSurfaceLines(pointsList1, pointsList2);
    Vector finalPointsList2 = mergeCutAndSurfaceLines(pointsList2, pointsList1);
    
    // These will have the same number of points
    Vector topoTriangles = new Vector();
    for(int i = 0; i < (finalPointsList1.size() - 1); i++){
      topoTriangles.add(new TopoTriangle((Point)finalPointsList1.get(i), (Point)finalPointsList1.get(i + 1), (Point)finalPointsList2.get(i)));
      topoTriangles.add(new TopoTriangle((Point)finalPointsList1.get(i + 1), (Point)finalPointsList2.get(i + 1), (Point)finalPointsList2.get(i)));
    }
    // If this is a junction!:
    if(outflowCanalGates.size() > 1){
      
      // First, do the triangles coming off the main trunk
      Point P1 = (Point)pointsList2.get(3);
      Point P2 = (Point)pointsList2.get(2);
      Point P3 = (Point)pointsList2.get(1);
      Point P4 = (Point)pointsList2.get(0);
      
      Point interim1 = Geometry.pointBetween(P1, endCoords, .7);
      Point P6 = new Point(interim1.getX(), interim1.getY(), endCoords.getZ());
      
      Point interim2 = Geometry.pointBetween(P4, endCoords, .7);
      Point P8 = new Point(interim2.getX(), interim2.getY(), endCoords.getZ());
      
      topoTriangles.add(new TopoTriangle(P1, P2, P6));
      topoTriangles.add(new TopoTriangle(P6, P2, endCoords));
      topoTriangles.add(new TopoTriangle(P2, P3, endCoords));
      topoTriangles.add(new TopoTriangle(P3, P8, endCoords));
      topoTriangles.add(new TopoTriangle(P3, P4, P8));
      
      // Next do each of the junction sections in turn
      for(int i = 0; i < outflowCanalGates.size(); i++){
        Vector cutAndSurface = ((Gate)outflowCanalGates.get(i)).getOutflow().getUpstreamDrawCutAndSurfaceLine();
        P1 = (Point)cutAndSurface.get(3);
        P2 = (Point)cutAndSurface.get(2);
        P3 = (Point)cutAndSurface.get(1);
        P4 = (Point)cutAndSurface.get(0);
        
        interim1 = Geometry.pointBetween(P1, endCoords, .7);
        P6 = new Point(interim1.getX(), interim1.getY(), endCoords.getZ());
        
        interim2 = Geometry.pointBetween(P4, endCoords, .7);
        P8 = new Point(interim2.getX(), interim2.getY(), endCoords.getZ());
        
        topoTriangles.add(new TopoTriangle(P1, P2, P6));
        topoTriangles.add(new TopoTriangle(P6, P2, endCoords));
        topoTriangles.add(new TopoTriangle(P2, P3, endCoords));
        topoTriangles.add(new TopoTriangle(P3, P8, endCoords));
        topoTriangles.add(new TopoTriangle(P3, P4, P8));        
      }     
    }
    CanalBedTriangulatedTopography tt = new CanalBedTriangulatedTopography();
    tt.addTopoTriangles(topoTriangles);
    return tt;
  }
  
  /**
   * @see abcm.graphics.X3DRepresentable#addToX3D(abcm.graphics.X3DAssist)
   */
  public void addToX3D(X3DAssist x3d){      
    Point headCoords = getHeadCoords();
    Point endCoords = getEndCoords();

    // First do the interior
    CanalBedTriangulatedTopography tt = (CanalBedTriangulatedTopography)getTriangulatedTopographyWithinBoundaryPolygons();
    tt.addToX3D(x3d);
//    
//    
//    // Then, if there is water, do an indexed face set for that
    ArrayList water = new ArrayList();
    ArrayList waterCorners = getWaterCorners();
    Polygon P1 = new Triangle((Point)waterCorners.get(0), (Point)waterCorners.get(1), (Point)waterCorners.get(3));
    water.add(P1);
    Polygon P2 = new Triangle((Point)waterCorners.get(1), (Point)waterCorners.get(2), (Point)waterCorners.get(3));
    water.add(P2);
    if(outflowCanalGates.size() > 1){
      Point c = new Point(endCoords.getX(), endCoords.getY(), z2);
      ArrayList junctionWaterCorners = getJunctionWaterCorners();
      for(int i = 0; i < junctionWaterCorners.size() - 1; i++){
        water.add(new Triangle((Point)junctionWaterCorners.get(i), (Point)junctionWaterCorners.get(i+1), c));
      }
    }
//     And water
    X3DIndexedFaceSetCreator creator = new X3DIndexedFaceSetCreator(water);
    x3d.addDefinition("Water", x3dWaterDefinition);
    XMLAssist xml = x3d.getXML();
    xml.open("Group");
      xml.open("Shape");
        xml.addLine("Appearance", "USE='Water'","");
        xml.open("IndexedFaceSet", "coordIndex=\"" + creator.getFaceList() + "\"" + " solid=\"false\"");
          xml.addLine("Coordinate", "point=\"" + creator.getPointList() + "\"", "");
        xml.closeTag(); // Indexed Face Set
      xml.closeTag(); // Water Shape
    xml.closeTag(); // Water Group
  }
  
  /* Cloning */
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Point headCoords = getHeadCoords();
    Point endCoords = getEndCoords();

    CanalSegment ret = new CanalSegment((Point)(cloneStore.containsKey(headCoords) ? cloneStore.get(headCoords) : headCoords.getDeepClone(cloneStore)),
                                         (Point)(cloneStore.containsKey(endCoords) ? cloneStore.get(endCoords) : endCoords.getDeepClone(cloneStore)), this.canalSegmentID);
    cloneStore.put(this,ret);
    ret.canalType               = canalType;
    ret.h                       = h;
    ret.Q1                      = Q1;
    ret.Q2                      = Q2;
    ret.z1                      = z1;
    ret.z2                      = z2;
    ret.isInterpolated          = isInterpolated;
    ret.originalRoughness       = originalRoughness;
    ret.roughness               = roughness;
    ret.anchorHeadCoords        = anchorHeadCoords;
    ret.anchorEndCoords         = anchorEndCoords;
    ret.downstreamCrossSection  = (ChannelCrossSection)cloneStore.getClone(downstreamCrossSection);
    if(this.isSelected())       ret.select();
    if(downstreamLeftDrawCorner != null)  ret.downstreamLeftDrawCorner    = (Point)cloneStore.getClone(downstreamLeftDrawCorner);
    if(downstreamRightDrawCorner != null) ret.downstreamRightDrawCorner   = (Point)cloneStore.getClone(downstreamRightDrawCorner);
    if(endGate != null)                   ret.endGate                     = (Gate)cloneStore.getClone(endGate);
    if(headGate != null)                  ret.headGate                    = (Gate)cloneStore.getClone(headGate);
    if(hydrographEngine != null)          ret.hydrographEngine            = (SingleDayHydrographEngine)cloneStore.getClone(hydrographEngine);
    
    ret.outflowCanalGates           = (ABCMDeepCloneableVector)outflowCanalGates.getDeepClone(cloneStore);
    if(outflowJunctionPolygon != null) ret.outflowJunctionPolygon      = (Polygon)cloneStore.getClone(outflowJunctionPolygon);
    ret.plotsIrrigated              = (ABCMDeepCloneableVector)plotsIrrigated.getDeepClone(cloneStore);
    
    ret.upstreamCrossSection = (ChannelCrossSection)cloneStore.getClone(upstreamCrossSection);
    if(upstreamLeftCornersOfOutflows != null) ret.upstreamLeftCornersOfOutflows = (ABCMDeepCloneableHashtable)upstreamLeftCornersOfOutflows.getDeepClone(cloneStore);
    if(upstreamRightCornersOfOutflows != null) ret.upstreamRightCornersOfOutflows = (ABCMDeepCloneableHashtable)upstreamRightCornersOfOutflows.getDeepClone(cloneStore);
    ret.water = (Water)water.getDeepClone(cloneStore); 

    return ret;
  }
  
  
  
  
  /**
   * Indicates whether this segment flows onto the specified field.
   * @param fieldName the field requested
   * @return true if this field is irrigated by this canal segment, false otherwise
   */
  public boolean flowsIntoFieldNamed(String fieldName){
    Iterator I = plotsIrrigated.iterator();
    while(I.hasNext()){
      if((((Field)I.next()).name).compareToIgnoreCase(fieldName) == 0) return true; 
    }
    return false;
  }
  
  /* Obsolete */
//  public boolean isEmptyButHasWaterFlowingInOrOut(){
//    // If this is at a headgate with a nonzero flow
//    if(isEmpty() && ((getInflowGate().getHeadGate() != null && 
//       ((getInflowGate().getHeadGate().getFlow() != 0) ||
//        (getInflowGate().getHeadGate().getFlowChange(0, 24 * 60 * 60) != 0))))) return true;
//    // If the canal segment has an upstream segment with a positive discharge
//    if(getUpstreamDischarge() > 0 && getDownstreamDischarge() == 0) return true;
//    if(getDownstreamDischarge() > 0 && getUpstreamDischarge() == 0) return true;
//    return false;
//  }
  
  
}