package flowag.channels;

import java.awt.Color;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.graphics.SVGAssist;
import abcm.graphics.X3DAssist;
import abcm.math.Point;
import abcm.math.SimDate;
import abcm.sim.AbstractModelComponent;
import abcm.sim.Analysis;
import abcm.sim.SimulationDataSeries;
import abcm.space.SpatialElement;
import abcm.space.TriangulatedTopography;
import flowag.basic.Water;
import flowag.bio.Field;
import flowag.bio.Plant;



/**
 * A CanalSystem object is a container class for a collection of linked
 * CanalSegment objects that together form a 'system'. This object
 * provides basic mechanisms for building a canal system, and
 * summary reports that aggregate information from all canals
 * in that system. Also includes a number of methods that are mainly
 * of use in arranging graphic output.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class CanalSystem extends AbstractModelComponent implements SpatialElement, ABCMDeepCloneable {

  /** Maximum time step allowed for flow calculations. Default is one hour */
  private static double MAXIMUM_TIMESTEP         = 60 * 60;
  
  
  /** Name for this canal system */                  private String CanalSystemName;
  /** Collection of canal segments in this system */ private ABCMDeepCloneableHashtable canals = new ABCMDeepCloneableHashtable();
  /** Collection of fields in this system */         private ABCMDeepCloneableHashtable fields = new ABCMDeepCloneableHashtable();
  /** Collection of canal segments' proxy names */   private ABCMDeepCloneableHashtable canalSegmentProxyNames = new ABCMDeepCloneableHashtable();

  /** Pointer to a selected canal segment */ private CanalSegment selectedCanalSegment;
  /** Minimum X value of coordinates of all canals in this system */ private double minX;
  /** Maximum X value of coordinates of all canals in this system */ private double maxX;
  /** Minimum Y value of coordinates of all canals in this system */ private double minY;
  /** Maximum Y value of coordinates of all canals in this system */ private double maxY;
  /** Minimum Z value of coordinates of all canals in this system */ private double minZ;
  /** Maximum Z value of coordinates of all canals in this system */ private double maxZ;
  /** Collection of canal types */ private Hashtable canalTypesLegend = new Hashtable();

  /**
   * Flag to indicate whether to permit the addition of segments
   * that technically flow 'uphill' (downstream z coordinate higher than upstream)
   */
  protected boolean allowUphillSegments = true; // Default is to allow segments to have positive slope
  
  /** Accumulates the volume of construction for all canal segments to date */
  private double constructionVolumeAccumulator;
  
  // Code Model Switches
  /** Flow algorithm code model switch */ public static final int NO_FLOW                     = 0;
  /** Flow algorithm code model switch */ public static final int HEC_RAS_IMP_FINITE_DIFF     = 1;
  /** Flow algorithm code model switch */ public static final int LPG                         = 2;

  
  /* Constructors and Initialization Methods */
  
  /**
   * Constructor that creates a CanalSystem object with no name.
   */
  public CanalSystem(){
    initCanalSystem("<unnamed>");
  }

  /**
   * Constructor that creates a canal system object with the specified name.
   * @param csName name of the new canal system
   */
  public CanalSystem(String csName){
    initCanalSystem(csName);
  }

  /**
   * Initialize the canal system.
   * This is basically what needs to be done in a constructor,
   * but allows polymorphic constructors to invoke the same code.
   * @param canalSystemName Name for the canal system
   */  
  private void initCanalSystem(String canalSystemName){
    CanalSystemName = canalSystemName;
    initCanalTypesLegend();
  }  
  
  /**
   * Initializes the CanalTypes hashtable
   * This can be blank, or it can have 'pre-loaded'
   * types.
   */
  private void initCanalTypesLegend(){
    Color c = new Color(255,0,0);
    canalTypesLegend.put("Unknown", c);    
  }

  
  /* Setters */

  /**
   * Sets the canal system name.
   * @param string new name for this canal system
   */
  public void setCanalSystemName(String string) {     CanalSystemName = string;     }  

  
  /* Basic Getters */

  /**
   * Gets the name of this Canal System.
   * @return the canal system name.
   */
  
  
  /**
   * Gets the canal system name
   * @return the name of this canal system
   */
  public String getCanalSystemName() {                return CanalSystemName;       }
  
  
  /* Actions */
  
  /* Statistics */

  /**
   * Gets the total area irrigated by all the CanalSegments in this
   * canal system.
   * @return The total area irrigated by the entire Canal System
   */
  public double TotalAreaIrrigated(){
    double TotalIrr = 0;
    Enumeration e;
    CanalSegment c;
    
    e = canals.elements();
    while(e.hasMoreElements()){
      c = (CanalSegment)e.nextElement();
      TotalIrr += c.areaIrrigated();
    }
    
    return TotalIrr;
  }
  
  /**
   * Gets the total construction volume of all
   * canals in this canal system
   * @return total construction volume of all canals
   */
  public double totalVolumeOfCanals(){
    double TotalVolume= 0;
    Enumeration e;
    CanalSegment c;
    
    e = canals.elements();
    while(e.hasMoreElements()){
      c = (CanalSegment)e.nextElement();
      TotalVolume += c.getTotalVolume();
    }
    return TotalVolume;
  }
  
  /**
   * Gets the total volume of water stored in this entire
   * canal system
   * @return total of all water in all segments in this canal system
   */
  public double totalVolumeOfWater(){
    double TotalWater = 0;
    Enumeration e;
    CanalSegment c;
    
    e = canals.elements();
    while(e.hasMoreElements()){
      c = (CanalSegment)e.nextElement();
//      TotalWater += c.getWaterVolume();
    }
    return TotalWater;
  }

  /**
   * Gets the total length of all Canal Segments in this
   * Canal System
   * @return the total length of the Canal System
   */
  public double totalLength(){
    double length = 0;
    Enumeration e = canals.elements();
    while(e.hasMoreElements()){
      length += ((CanalSegment)e.nextElement()).getLength();
    }
    return length;
  }
  
  /* Modifications to Canal Segments or Systems */
  
  /**
   * Extends the last-added canal in this canal system to the specified point.
   * @param p point to which canal system is to be extended
   * @param newCanalName name for new canal segment
   * @return true if successful, false otherwise
   */
  public boolean extendSelectedCanalSegment(Point p, String newCanalName){
    if(selectedCanalSegment == null) return false;
    CanalSegment c = selectedCanalSegment;
    boolean retVal = c.extendCanal(p, this, newCanalName);
    return retVal;
  }

  /**
   * Sets the profile of the last canal segment added.
   * @param topWidth the new to width of the Canal Segment
   * @param bottomWidth the new bottom width of the Canal Segment
   * @param depth the new depth for the Canal Segmemt
   * @return true if successful, false if not
   */
  public boolean setSelectedCanalSegmentProfile(double topWidth, double bottomWidth, double depth){
    return (selectedCanalSegment != null ? selectedCanalSegment.setProfile(topWidth, bottomWidth, depth) : false);
  }

  /**
   * Sets a flag indicating whether the selected canal segment is considered 'interpolated'
   * @param isInterpolated
   * @return false if no segment was selected; true otherwise
   */
  public boolean setSelectedCanalSegmentInterpolated(boolean isInterpolated){
    if(selectedCanalSegment != null){
      selectedCanalSegment.setIsInterpolated(isInterpolated);
      return true;
    }
    return false;
  }

  /**
   * Adds a new canal to this canal system
   * @param CanalObject CanalSegment to be added
   */
  public void addCanal(CanalSegment CanalObject){
    canals.put(CanalObject.getCanalSegmentID(), CanalObject);
    double tempMinX = Math.min(CanalObject.getHeadCoords().getX(),CanalObject.getEndCoords().getX() );
    double tempMaxX = Math.max(CanalObject.getHeadCoords().getX(),CanalObject.getEndCoords().getX() );
    double tempMinY = Math.min(CanalObject.getHeadCoords().getY(),CanalObject.getEndCoords().getY() );
    double tempMaxY = Math.max(CanalObject.getHeadCoords().getY(),CanalObject.getEndCoords().getY() );
    double tempMinZ = Math.min(CanalObject.getHeadCoords().getZ(),CanalObject.getEndCoords().getZ() );
    double tempMaxZ = Math.max(CanalObject.getHeadCoords().getZ(),CanalObject.getEndCoords().getZ() );
    if(canals.size() == 1){
      minX = tempMinX;
      maxX = tempMaxX;
      minY = tempMinY;
      maxY = tempMaxY;
      minZ = tempMinZ;
    }
    else{
      minX = Math.min(minX, tempMinX);
      maxX = Math.max(maxX, tempMaxX);
      minY = Math.min(minY, tempMinY);
      maxY = Math.max(maxY, tempMaxY);
      minZ = Math.min(minZ, tempMinZ);
      maxZ = Math.max(maxZ, tempMaxZ);
    }
    
    // If this canal has a 'type' string set,
    // add it to the list of types found in this canal
    // system and assign it a color value.
    if(CanalObject.getCanalType() != null){
      if(!canalTypesLegend.containsKey(CanalObject.getCanalType()))
        canalTypesLegend.put(CanalObject.getCanalType(), 
           new Color(
             (int)Math.floor(Math.random()*256),
             (int)Math.floor(Math.random()*256), 
             (int)Math.floor(Math.random()*256)));
    }
    
    selectedCanalSegment = CanalObject;
  }
  
  
  /**
   * Removes a canal segment from this canal system; in fact, only disconnects it
   * from inflow and outflow.
   * @param toRemove the segment to be removed
   * @return true (does not fail)
   */
  public boolean removeCanalSegment(CanalSegment toRemove){
    toRemove.disconnect();
    return true;
  }
  
  
  
  
  
  
  
  /* Fields */
  
  /**
   * Links the specified canal segment to the specified field
   * @param canalSegID ID of canal segment
   * @param field Field object
   * @return true if successful, false otherwise (if canal segment ID is not found)
   */
  public boolean linkCanalSegmentToField(String canalSegID, Field field){
    CanalSegment cs = getCanalSegment(canalSegID);
    if(cs == null) return false;
    fields.put(field.name, field); // Each field will appear only once; note: should max and min x and y ranges be adjusted here, too?
    cs.addPlot(field);
    return true;
  }
  
  /**
   * Drops a field from this canal system and any segments that
   * feed into it
   * @param fieldName the name of the field being dropped
   * @param F the field itself
   */
  public void dropField(String fieldName, Field F){
    Enumeration E = canals.elements();
    while(E.hasMoreElements()) ((CanalSegment)E.nextElement()).removePlot(F); // Will remove from all canal segs, even though didn't exist in some
    fields.remove(fieldName);
  }
  
  /**
   * Breaks the link between a canal segment and the specified field.
   * If the field is no longer associated with any canal segment in this
   * system, the field is dropped entirely
   * @param canalSegID ID of canal segment
   * @param field Field object
   * @return true if successful. false otherwise (if canal segment ID is not found)
   */
  public boolean unlinkCanalSegmentFromField(String canalSegID, Field field){
    CanalSegment cs = getCanalSegment(canalSegID);
    if(cs == null) return false;
    cs.removePlot(field); // This will not cause an error even if plot is not linked to field
    // However, now must check all remaining canal segments to see if plot is
    // still irrigated by some other field...
    Enumeration E = canals.elements();
    while(E.hasMoreElements())  if(((CanalSegment)E.nextElement()).irrigatesField(field)) return true;
    // If not:
    fields.remove(field.name);
    return true;
  }
  
  /**
   * Gets a Vector listing all the fields managed by this
   * Canal System. Includes all fields to which Canal Segments
   * in this Canal System are linked
   * @return a vector of fields
   */
  public Vector getAllFields(){
    return new Vector(fields.values());
  }
  
  
  /* Dimensions */  
  
  
  /**
   * @see abcm.space.SpatialElement#getMaxX()
   */
  public double getMaxX() {
    return maxX;
  }
  
  /**
   * @see abcm.space.SpatialElement#getMaxY()
   */
  public double getMaxY() {
    return maxY;
  }
  
  /**
   * @see abcm.space.SpatialElement#getMinX()
   */
  public double getMinX() {
    return minX;
  }
  
  /**
   * @see abcm.space.SpatialElement#getMinY()
   */
  public double getMinY() {
    return minY;
  }

  
  /* Actions */
    
  /**
   * Resets all hydrographs in this canal system
   */
  private void resetHydrographs(){
    Enumeration canalsList = canals.elements();
    while(canalsList.hasMoreElements()){
      ((CanalSegment)canalsList.nextElement()).resetHydrographs();          
    }
  }
  
  /**
   * Causes water to flow through the system for the specified
   * duration
   * @param durationInDays duration through which flow should occur, in days
   * @return true if successful, false if some error occurred
   */
  public boolean flow(double durationInDays){
    resetHydrographs();
    double durationInSeconds = durationInDays * 24 * 60 * 60;   
//    System.out.println("FLOWING: " + codeModelOption("FLOW_PROCEDURE"));
    /* Flow procedures */
    switch(codeModelOption("FLOW_PROCEDURE")){
    /* HEC-RAS Flow - Implicit Finite Difference Approximation */
      case HEC_RAS_IMP_FINITE_DIFF:{
        System.out.println("Running HEC-RAS in Flow System");
        double totalTimeCalculated = 0;
        while(totalTimeCalculated < durationInSeconds){
          double dayFrac = totalTimeCalculated / (24d * 60d * 60d);
//          System.out.println("\nFLOWING AT TIME " + totalTimeCalculated + "/" + dayFrac + " current flow is " + ((CanalSegment)this.canals.get("ProbeSeg1")).getInflowGate().getHeadGate().getFlowAt(dayFrac));
          
          Vector flowSystems = getHECFlowSystems(dayFrac);

          // For each flow system, calculate the maximum time of flow possible, then find the lowest of these among all flow systems
          double t = Math.min(MAXIMUM_TIMESTEP, (durationInSeconds - totalTimeCalculated));
          for(int i = 0; i < flowSystems.size(); i++)
            t = Math.min(t, ((HEC_RAS_FlowCalculator)flowSystems.get(i)).solveForMaxTimestep(t));          
          
          // Flow all flow systems for the lowest maximum time possible (ASSUMES STABILITY AT MIN!!!)
          boolean success = true;
          for(int i = 0; i < flowSystems.size(); i++) 
              success = success & ((HEC_RAS_FlowCalculator)flowSystems.get(i)).calculate(t); // Will do all but will show success of all
          if(!success) System.err.println("AT LEAST ONE FLOW SYSTEM FAILED FOR TIME " + t);
          
          // Now we play a trick: we set the canal segments' discharge and flow values to 
          // their current values plus _half_ of the delta values. Then we execute the
          // move water and wear routines. Then we will add the second half of the delta values
          // back in. The effect of this is that the wear and flow routines will use the
          // median value across the change, effectively using average flow values to
          // calculate the wear during the interval of time t.
          dayFrac = (totalTimeCalculated + t/2d) / (24d * 60d * 60d);
          for(int i = 0; i < flowSystems.size(); i++) {
            ((FlowCalculator)flowSystems.get(i)).getResults().applyFraction(canals.values(), .5d, dayFrac);
          }
          
          // Move the water- this affects chemical composition and puts water on fields
          moveWater(t);
          
          // This allows degradation of canal structure
          degradeCanals(t);
          
          // Increment the overall clock
          totalTimeCalculated += t;
          dayFrac = totalTimeCalculated / (24d * 60d * 60d);
          
          // Now increase the discharges to the final setting, at the end of time t
          for(int i = 0; i < flowSystems.size(); i++) 
              ((FlowCalculator)flowSystems.get(i)).getResults().applyFraction(canals.values(), .5d, dayFrac);
          }
        }
          // Break out of case switch!
        break;
      case LPG:{
        double totalTimeCalculated = 0;
        while(totalTimeCalculated < durationInSeconds){
          double dayFrac = totalTimeCalculated / (24d * 60d * 60d);
//          System.out.println("\nFLOWING AT TIME " + totalTimeCalculated + "/" + dayFrac);
          
          Vector flowSystems = getLPGFlowSystems(dayFrac);

          // For each flow system, calculate the maximum time of flow possible, then find the lowest of these among all flow systems
          double t = Math.min(MAXIMUM_TIMESTEP, (durationInSeconds - totalTimeCalculated));
          
          // Flow all flow systems for the lowest maximum time possible (ASSUMES STABILITY AT MIN!!!)
          boolean success = true;
          for(int i = 0; i < flowSystems.size(); i++) 
              success = success & ((FlowCalculator)flowSystems.get(i)).calculate(t); // Will do all but will show success of all
          if(!success) System.err.println("AT LEAST ONE FLOW SYSTEM FAILED FOR TIME " + t);
          
          // Now we play a trick: we set the canal segments' discharge and flow values to 
          // their current values plus _half_ of the delta values. Then we execute the
          // move water and wear routines. Then we will add the second half of the delta values
          // back in. The effect of this is that the wear and flow routines will use the
          // median value across the change, effectively using average flow values to
          // calculate the wear during the interval of time t.
          dayFrac = (totalTimeCalculated + t/2d) / (24d * 60d * 60d);
          for(int i = 0; i < flowSystems.size(); i++) {
            ((FlowCalculator)flowSystems.get(i)).getResults().applyFraction(canals.values(), .5d, dayFrac);
          }
//          System.out.println("Moving water...");
          // Move the water- this affects chemical composition and puts water on fields
          moveWater(t);
          
          // This allows degradation of canal structure
          degradeCanals(t);
          
          // Increment the overall clock
          totalTimeCalculated += t;
          dayFrac = totalTimeCalculated / (24d * 60d * 60d);
          
          // Now increase the discharges to the final setting, at the end of time t
          for(int i = 0; i < flowSystems.size(); i++) 
              ((FlowCalculator)flowSystems.get(i)).getResults().applyFraction(canals.values(), .5d, dayFrac);        
          }
        }
        break;
      case 1000:{
          System.out.println("RUNNING TEST FLOW PROC");
          resetHydrographs();
          FlowCalculator flowCalc = new TESTFlowCalculator(this.getCanalSegment("ProbeSeg1"), canals.values());
//          flowCalc.getResults(0).applyFraction(canals.values(),1, 0);
        }
        break;
      case 2000:
          
        break;
      case NO_FLOW:{
          
        }
        break;
    }
    return true;
  }
  
  /**
   * Gets a list of all HEC-RAS flow systems in this canal system.
   * Start time is required because boundary conditions, including
   * initial headgate flow levels, must also be calculated.
   * A canal system can contain multiple flow systems due to closed gates
   * or empty sections.
   * @param startTime  start time requested, as a fraction of a day (i.e. 0.5 = noon)
   * @return a list of the LPG flow systems within this canal system.
   */
  private Vector getHECFlowSystems(double startTime){
    Vector flowSystems = HEC_RAS_FlowCalculator.getHECFlowSystems(canals.values());
    for(int i = 0; i < flowSystems.size(); i++){
      HEC_RAS_FlowCalculator flowCalc = (HEC_RAS_FlowCalculator)flowSystems.get(i);
      if(flowCalc.initialCanalSegment.getInflowGate().getHeadGate() != null)
           flowCalc.setUpstreamBoundaryCondition(
               new FlowCalculator.HeadgateBasedUpstreamBoundaryCondition(
                   flowCalc.initialCanalSegment.getInflowGate().getHeadGate(), startTime));
    }
    return flowSystems;
  }
  
  
  /**
   * Gets a list of all LPG flow systems in this canal system.
   * Start time is required because boundary conditions, including
   * initial headgate flow levels, must also be calculated.
   * A canal system can contain multiple flow systems due to closed gates
   * or empty sections.
   * @param startTime  start time requested, as a fraction of a day (i.e. 0.5 = noon)
   * @return a list of the LPG flow systems within this canal system.
   */
  private Vector getLPGFlowSystems(double startTime){
    Vector<LPGFlowCalculator> flowSystems = LPGFlowCalculator.getFlowSystems(canals.values());
    for(int i = 0; i < flowSystems.size(); i++){
      LPGFlowCalculator flowCalc = flowSystems.get(i);
      if(flowCalc.initialCanalSegment.getInflowGate().getHeadGate() != null)        
           flowCalc.setUpstreamBoundaryCondition(
               new FlowCalculator.HeadgateBasedUpstreamBoundaryCondition(
                   flowCalc.initialCanalSegment.getInflowGate().getHeadGate(), startTime));      
    }
    return flowSystems;
  }
  
  
  
  /**
   * When the z and q values are set, move the water through the system and onto
   * the fields.
   * 
   * This method refers only to actual movement of water: the process
   * of delivering water, including chemical load, from one vessel to another
   * in the system. It does not affect the flow and stage values in the segments-
   * it only uses them
   * 
   * @param t the time through which water should be moved, in seconds
   */
  private void moveWater(double t){
    Enumeration E = canals.elements();
    while(E.hasMoreElements()){
      CanalSegment cs = (CanalSegment)E.nextElement();
      cs.waterFields(t);
    }
  }
  
  
  /**
   * Call the degradation function for each canal segment
   * @param t
   */
  private void degradeCanals(double t){
    Enumeration CSList = canals.elements();
    while(CSList.hasMoreElements()){
      ((CanalSegment)CSList.nextElement()).wear(t);
    }  
  }
  
  /**
   * Sows a plant in all of the fields of all of the canal
   * segments in this canal system
   * @param p Plant to be sown
   * @param density density at which plant is to be sown
   * @param s date on which sowing event occurs, as a SimDate object
   */
  public void sow(Plant p, double density, SimDate s){
    for(Object field: fields.values()) ((Field)field).sow(p, density, s);    
  }

  /**
   * Rains the specified number of mm of rain, drawn from the sample
   * water source object, on all of the fields associated with all of the
   * CanalSegment objects in this CanalSystem.
   * @param MM millimeters of rainfall
   * @param rainSource a sample WaterSource object, containing some specified
   * combination of Chemicals and Concentrations
   */
  public void rain(double MM, Water rainSource){
    for(Object field: fields.values()) ((Field)field).rain(MM, rainSource);
  }

 
  /**
   * Resets the Construction Volume Accumulator to zero.
   */
  public void resetConstructionVolumeAccumulator(){    constructionVolumeAccumulator = 0;   }

  
  /* Selecting or Getting Specific Canal Elements */
  
  /** 
   * Methods that result in the creation of new canals
   * set this variable so that the new canal can be returned.
   * @return - The last newly added canal
   */
  public CanalSegment getSelectedCanalSegment(){    return selectedCanalSegment;   }

  /**
   * Sets the specified canal segment to be the
   * selected canal segment; this allows a number of
   * operations to be performed on it directly.
   * @param canalID
   * @return true if there is a canal segment with
   * the specified ID, false otherwise.
   */
  public boolean selectCanalSegment(String canalID){
    CanalSegment CS = getCanalSegment(canalID);
    if(CS == null) return false;
    unselectAllCanalSegments();
    selectedCanalSegment = CS;
    selectedCanalSegment.select();
    return true;
  }
  
  
  /**
   * Unselects all canal segments in this system
   */
  public void unselectAllCanalSegments(){
    selectedCanalSegment = null;
    Enumeration e = canals.elements();
    while(e.hasMoreElements()){
      ((CanalSegment)e.nextElement()).unSelect();
    }
  }
  
  /**
   * Selects all the canal segments in this system
   */
  public void selectAllCanalSegments(){
    selectedCanalSegment = null;
    Enumeration e = canals.elements();
    while(e.hasMoreElements()){
      ((CanalSegment)e.nextElement()).select();
    }
  }
  
  /**
   * Selects the specified canal segment and all segments downstream
   * of it, in all branches
   * @param name the name of the initial segment to select
   */
  public void selectSegmentAndDownstreamSegments(String name){
    CanalSegment CS = getCanalSegment(name);
    if(CS == null) return;
    CS.selectWithDownstream();
  }
  
  /**
   * Adds a proxy name for a specified canal segment. The proxy can
   * be used to refer to the canal segment just as the original name
   * in the future.
   * @param originalName the original name of the canal segment
   * @param newName the new, proxy name for the segment. Overrides any previous proxy
   * definition with this same name 
   * @return a String indicating the success of this addition (does not fail)
   */
  public String addCanalSegmentProxyName(String originalName, String newName){
    canalSegmentProxyNames.put(newName, originalName);
    return "Added '" + newName + "' as proxy for '" + originalName + "'.";
  }
  
  /**
   * Gets a requested canal segment. The segment can be requested by a name
   * or by a proxy name.
   * @param canalSegmentNameOrProxy the name requested
   * @return the segment, if found; null otherwise
   */
  public CanalSegment getCanalSegment(String canalSegmentNameOrProxy){
    if(canals.containsKey(canalSegmentNameOrProxy)) return (CanalSegment)canals.get(canalSegmentNameOrProxy);
    if(canalSegmentProxyNames.containsKey(canalSegmentNameOrProxy)) return getCanalSegment((String)canalSegmentProxyNames.get(canalSegmentNameOrProxy));
    return null;
  }
  
  /**
   * Gets the stage of the specified canal segment's downstream end at 
   * the given time of day (within the last day, per the way hydrographs
   * store and refresh data)
   * @param canalSegName name of the canal segment
   * @param time time of day, as a fraction of a day (i.e. 0.5 = noon)
   * @return the stage value
   */
  public double getDownstreamStageInSegAt(String canalSegName, double time){
    CanalSegment CS = getCanalSegment(canalSegName);
    if(CS == null) return 0;
    SimDate SD = new SimDate("01-JAN-SEASONS");
    SD.setFraction(time);
    return CS.getDownstreamStageAt(SD);
  }

  /**
   * Creates an SVG representation of the specified canal segment's
   * cross section.
   * @param endSegName the segment to be profiled
   * @return The file name of the SVG output.
   */
  public String getProfile(String endSegName){
    CanalSegment CS = getCanalSegment(endSegName);
    if(CS == null) return "Could not create; no seg '" + endSegName + "' exists.";
    Analysis A = new Analysis();
    SimulationDataSeries SDS = new SimulationDataSeries("ChannelBottom", true);
    A.addSeries(SDS, "Bottom");
    SimulationDataSeries[] SDs = new SimulationDataSeries[24];
    for(int i = 0; i < 24; i++){
      SDs[i] = new SimulationDataSeries("At time " + i, true);
      A.addSeries(SDs[i], "Time_" + i);
    }
    
    Vector v = new Vector();
    v.add(0,CS);
    while(CS.getInflowCanal().getCanalSegmentID() != CS.getCanalSegmentID()){
      CS = CS.getInflowCanal();
      v.add(0,CS);
    }
    // First add the upstream portion for the first canal segment
    CanalSegment c;
    c = (CanalSegment)v.get(0);
    double l = 0;
    SDS.addElement(new SimDate(l), c.getHeadCoords().getZ() );
    for(int x = 0; x < 24; x++){
      SimDate S = new SimDate(l);
      System.out.println(S.getString());
      double usStage = c.getUpstreamStageAt(new SimDate(x/24d));
      System.out.println(usStage);
      SDs[x].addElement(S, usStage);
    }
    // Next add each of the downstream portions
    for(int i = 0; i < v.size(); i++){
      c = ((CanalSegment)v.get(i));
      l += c.getLength();
      SimDate S = new SimDate(l);
      System.out.println(S.getString());
      SDS.addElement(S, c.getEndCoords().getZ() );
      for(int x = 0; x < 24; x++){
        SDs[x].addElement(new SimDate(l), c.getDownstreamStageAt(new SimDate(x/24d)));
      }  
    }
    
    return A.getAsSVG("Profile for canal system " + CanalSystemName, 400,600);
  }

    
  /* Graphic output */
  /**
   * @see abcm.graphics.SVGRepresentable#addToSVG(abcm.graphics.SVGAssist)
   */  
  public void addToSVG(SVGAssist svg){
    svg.addG("Canal System: " + this.CanalSystemName, "", SVGAssist.getArrayFromHashtable(canals));
  }
  
  /* Text Reporting */
  
  /**
   * Gets a report describing this canal system.
   * @return a string containing a text report describing this canal system.
   */
  public String report(){
    String Report = "";
    String canalsReport = "";
    
    Report += "Canal System: " + CanalSystemName + "\n\n";
    Report += "  Canals: " + canals.size() + "\n";
    Report += "  Total Area Irrigated: "  + TotalAreaIrrigated() + "\n";
    Report += "  Total Volume of Water in System: " + totalVolumeOfWater() + "\n";
    Report += "\n";
       
    if(canals.size()>0){
      Iterator I = (Iterator)canals.elements();
      while (I.hasNext()){
        CanalSegment c = (CanalSegment)I.next();
        canalsReport = c.report() + canalsReport;
      }      
      Report+= canalsReport;
      Report+="\n\n";
    }
    Report+=" MinX: = " + minX + " MaxX = " + maxX + "\n";
    Report+=" MinY: = " + minY + " MaxY = " + maxY + "\n";
    Report+="\n**** End Canal System '" + CanalSystemName + "' Report\n\n";
    
    return Report;    
  }

  /**
   * Gets a string that includes the flow reports from all Canal Segments.
   * @return a complete flow report
   */
  public String flowReport(){
    String report = "";
    String canalsReport = "";
    report += "Canal System " + CanalSystemName + "\n";
    Enumeration e = canals.elements();
    while(e.hasMoreElements()){
      CanalSegment c = (CanalSegment)e.nextElement();
      if(c.getInflowCanal() == c){
        report += c.flowReport();        
      }
    }
    return report;
  }
  
  /**
   * Gets this canal system as a HEC-RAS import file.
   * This will allow us to compare the results of a HEC-RAS
   * run on the same system.
   * @return a String that is formatted as a HEC-RAS import file
   * describing this canal system
   */
  public String writeHECImportFile(){
    StringWriter out = new StringWriter();
    
    // Header
    out.write("# HEC-RAS Output file\n");
    out.write("BEGIN HEADER:\n");
    out.write("\n");
    out.write("  UNITS: METRIC\n");
    out.write("\n");
    out.write("END HEADER:\n");
    out.write("\n");
    out.write("\n");
    
    // Stream Network
    out.write("BEGIN STREAM NETWORK:\n");
    out.write("\n");
    
    // Create a list of all the Endpoints that mark
    // Junctions
    // Remember, HWM does not allow inflow junctions,
    // so the routine to find these is simple:
    
    Enumeration c = canals.elements();
    Hashtable pts = new Hashtable();
    int ptct = 1;
    while(c.hasMoreElements()){
      CanalSegment cs = (CanalSegment)c.nextElement();
      if(cs.getInflowCanal().equals(cs)){
        Point p = cs.getHeadCoords();
        pts.put(p.coordString(),new Integer(ptct));
        out.write("  ENDPOINT " + p.getX() + ", " + p.getY() + ", " + p.getZ() + ", " + ptct + "\n");
        ptct++;
      }
      if(cs.getOutflowCanalGates().size() != 1) {
         Point p = cs.getEndCoords();
         pts.put(p.coordString(),new Integer(ptct));
         out.write("  ENDPOINT " + p.getX() + ", " + p.getY() + ", " + p.getZ() + ", " + ptct + "\n");
         ptct++;
      }
    }
    out.write("\n");
    
    // Now list the reaches
    // First create a vector of all canal segments that have
    // no inflow
    Vector v = new Vector();
    c = canals.elements();
    while(c.hasMoreElements()){
      CanalSegment cs = (CanalSegment)c.nextElement();
      if(cs.getInflowCanal().equals(cs)){
        v.add(cs);
      }
    }
    
    int streamCount = 1;
    int reachCount = 1;
    double dist = 0;
    Hashtable streamsAndLengths = new Hashtable();
    while(v.size() > 0){
      CanalSegment cs = (CanalSegment)v.remove(0);
      //System.out.println("Doing Canal Segment: " + cs.getCanalSegmentID());
      String StreamID = "STREAM" + streamCount;      
      out.write("  REACH:\n");
      out.write("    STREAM ID: " + StreamID + "\n");
      out.write("    REACH ID: REACH" + reachCount + "\n");
      Point p = cs.getHeadCoords();
      out.write("    FROM POINT: " + ((Integer)pts.get(p.coordString())).intValue() + "\n");
      
      // There's an interruption here... must print out
      // the last point here, before listing the points
      // individually; build this as a string
      String o = "    CENTERLINE:\n";
      // The first point is the headcoords point
      o += "       " + p.getX() + ", " + p.getY() + ", " + p.getZ() + ", " + dist / 1000 + "\n";
      
      Vector vb; // We need this for the criteria of the while loop
      do {
        p = cs.getEndCoords();
        dist += cs.getLength();
        o += "       " + p.getX() + ", " + p.getY() + ", " + p.getZ() + ", " + dist / 1000 + "\n";
        vb = cs.getBranches();
        cs = cs.getNextDownstreamSegment();
        // If there are no branches
        if(vb.size() == 0){          
          if(cs == null){
            streamsAndLengths.put("Stream" + streamCount, new Double(dist));
            streamCount++;
            dist = 0;              
            reachCount++;  
          }
        }
        else{
          v.addAll(vb);
          reachCount++;
          v.add(0,cs);
        }
      }while((cs!=null) && (vb.size() == 0));
      out.write("    TO POINT: " + pts.get(p.coordString()) + "\n");
      out.write(o);
      out.write("  END:\n\n");
    }
    
    out.write("\n");
    out.write("END STREAM NETWORK:\n");
    out.write("\n");
        
    // Cross-Sections
    out.write("BEGIN CROSS-SECTIONS:\n");
    out.write("\n");

    // Much of this algorithm repeats what has come above
    
    v = new Vector();
    c = canals.elements();
    while(c.hasMoreElements()){
      CanalSegment cs = (CanalSegment)c.nextElement();
      if(cs.getInflowCanal().equals(cs)){
        v.add(cs);
      }
    }
    
    // List all the cross-sections
    streamCount = 1;
    int rc = 1;
    dist = 0;
    while(v.size() > 0){
      CanalSegment cs = (CanalSegment)v.remove(0);
      double streamLength = ((Double)streamsAndLengths.get("Stream" + streamCount)).doubleValue();
      //dist = cs.getLength();
      do{
        out.write(cs.getCrossSection(true, (streamLength - dist) / 1000, "STREAM" + streamCount, "REACH" + rc));
        dist+=cs.getLength();
        if(cs.getNextDownstreamSegment()==null){
          out.write(cs.getCrossSection(false, (streamLength - dist) / 1000, "STREAM" + streamCount, "REACH" + rc));    
        }
        Vector vb = cs.getBranches();
        if(vb.size() > 0){ 
          rc++;
          v.addAll(vb);
        }       
        cs = cs.getNextDownstreamSegment();
        if(cs == null) {
          streamCount++;
          rc++;
          dist = 0;
        } 
      }while(cs!=null);
    }
        
    out.write("END CROSS-SECTIONS:\n");
    out.write("\n");
    
    return out.toString().replaceAll("\n","\r\n");
  }

  /**
   * @see abcm.space.SpatialElement#altered()
   */
  public boolean altered(){
    Iterator i = canals.keySet().iterator();
    while(i.hasNext()) if(((CanalSegment)i.next()).altered()) return true;
    return false;
  }

  /**
   * @see abcm.space.SpatialElement#getBermPolygons()
   */
  public Vector getBermPolygons(){
    Vector V = new Vector();
    Enumeration e = canals.elements();
    while(e.hasMoreElements()){
      CanalSegment c = (CanalSegment)e.nextElement();
      if(c.getCanalSegmentID() != "SegR3") V.addAll(c.getBermPolygons());
    }
    return V;
  }

  /**
   * @see abcm.space.SpatialElement#getBoundaryPolygons()
   */
  public Vector getBoundaryPolygons(){
    Vector V = new Vector();
    Enumeration e = canals.elements();
    while(e.hasMoreElements()){
      CanalSegment c = (CanalSegment)e.nextElement();
      if(c.getCanalSegmentID() != "SegR3") V.addAll(c.getBoundaryPolygons());
    }
    return V;
  }

  /**
   * Not implemented; returns zero
   * @see abcm.space.SpatialElement#getMaxZ()
   */
  public double getMaxZ(){
    return 0;
  }

  /**
   * Not implemented; returns zero.
   * @see abcm.space.SpatialElement#getMinZ()
   */
  public double getMinZ(){
    return 0;
  }

  /**
   * @see abcm.graphics.X3DRepresentable#addToX3D(abcm.graphics.X3DAssist)
   */
  public void addToX3D(X3DAssist x3d){
    Enumeration e = canals.elements();
    while(e.hasMoreElements())  ((CanalSegment)e.nextElement()).addToX3D(x3d);
  }

  /**
   * @see abcm.space.HasElevation#getZ(abcm.math.Point)
   */
  public double getZ(Point P){
    return 0;   
  }
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    CanalSystem clone = new CanalSystem();
    cloneStore.put(this, clone);
    clone.CanalSystemName = CanalSystemName;
    clone.minX = minX;
    clone.maxX = maxX;
    clone.minY = minY;
    clone.maxY = maxY;
    clone.minZ = minZ;
    clone.maxZ = maxZ;
  
    clone.selectedCanalSegment = selectedCanalSegment;
    clone.constructionVolumeAccumulator = constructionVolumeAccumulator;
    clone.canals                 = (canals != null ? (ABCMDeepCloneableHashtable)canals.getDeepClone(cloneStore) : null);
    clone.fields                 = (fields != null ? (ABCMDeepCloneableHashtable)fields.getDeepClone(cloneStore) : null);
    clone.canalSegmentProxyNames = (canalSegmentProxyNames != null ? (ABCMDeepCloneableHashtable)canalSegmentProxyNames.getDeepClone(cloneStore) : null);
    clone.canalTypesLegend       = (Hashtable)canalTypesLegend.clone(); // This requires only a shallow clone
    
    return clone;
  }
  
  /**
   * @see abcm.space.SpatialElement#getTriangulatedTopographyWithinBoundaryPolygons()
   */
  public TriangulatedTopography getTriangulatedTopographyWithinBoundaryPolygons(){
    TriangulatedTopography ret = new TriangulatedTopography(new Vector());
    Enumeration e = canals.elements();
    while(e.hasMoreElements())  ret.addTopoTriangles(((CanalSegment)e.nextElement()).getTriangulatedTopographyWithinBoundaryPolygons().getTriangles());
    return ret;
  }


  /**
   * Gets the names of all canal segments that flow into a particular field
   * @param fieldName the field sought
   * @return the name of all segments flowing onto that field
   */
  public Vector<String> getNamesOfInflowCanalSegments(String fieldName){
    Vector<String> ret = new Vector<String>();
    Iterator I = this.canals.values().iterator();
    while(I.hasNext()){
      CanalSegment cs = (CanalSegment)I.next(); 
      if(cs.flowsIntoFieldNamed(fieldName)) ret.add(cs.getCanalSegmentID());
    }
    return ret;
  }
  
  /**
   * Gets the total harvest of all fields in this canal system for the previous calendar year
   * @return the harvest total for the previous year
   */
  public double getTotalHarvestLastYear(){
    double total = 0;
    for(Object field: fields.values()) total += ((Field)field).getTotalHarvestLastYear();
    return total;
  }
  
  /**
   * Indicates whether this canal system includes the specified field
   * @param fieldName name of the field sought
   * @return true if the system contains this field
   */
  public boolean containsField(String fieldName){
    return this.fields.keySet().contains(fieldName);
  }


  
  
  /**
   * Gets a list of the initial segments of all branches of this system
   * that are assigned a type matching the static variable 'DISTRIBUTION'.
   * @return a list of initial distribution segments
   */
  public List<String> getDistributionCanalInitialSegmentNames(){
    Vector<String> orderedList = new Vector<String>();
    for(Object o: canals.keySet()){
      CanalSegment cs = (CanalSegment)(canals.get((String)o));
      if((cs.getType().compareToIgnoreCase(CanalSegment.DISTRIBUTION)== 0) &&
         (cs.getInflowCanal() != null && !cs.getInflowCanal().equals(cs))  &&
         (cs.getInflowCanal().getType().compareToIgnoreCase(CanalSegment.MAIN) == 0)){
        double distance = getDistanceFromInitialSegment((String)o);
        if(orderedList.size() == 0) orderedList.add((String)o);
        else{
          for(int i = 0; i < orderedList.size(); i++){
            double otherDist = getDistanceFromInitialSegment(orderedList.get(i));
            if(distance < otherDist) orderedList.add(i, (String)o);
            break;
          }
        }       
      }
    }
    return new ArrayList<String>(orderedList);  
  }
  
  /**
   * Gets the distance of the upstream end of the specified segment from the initial segment
   * of the canal system.
   * @param csName the name of the canal segment requested
   * @return the distance, in meters, from the head of the canal flowing into the requested segment
   */
  public double getDistanceFromInitialSegment(String csName){
    CanalSegment cs = this.getCanalSegment(csName);
    if(cs == null) return -1;
    
    double dist = 0;
    
    CanalSegment current = cs;
    CanalSegment inflow = cs.getInflowCanal();
    while(inflow != null && !inflow.equals(current)){
      dist += inflow.getLength();
      current = inflow;
      inflow = current.getInflowCanal();
    }
    return dist;  
    
  }
}





