package flowag.channels;

import java.util.Vector;

import abcm.cloning.ABCMDeepCloneable;
import abcm.math.Point;

/**
 * Defines a basic interface for a ChannelCrossSection. A Channel's cross section
 * can have instance variables representing water stage and rate of flow; from these
 * a host of values, such as water surface width and cross-sectional area, can be
 * calculated and used in flow routines. 
 * @author John T. Murphy
 */
public interface ChannelCrossSection extends ABCMDeepCloneable {

  /**
   * Gets the top width of this cross section.
   * @return the top width of this cross section
   */
  public double getTopWidth();
  
  /**
   * Gets the bottom width of this cross section. For some cross
   * sections the bottom width may be meaningless (i.e. a round-bottom
   * channel), but in these cases this method should return zero.
   * @return the bottom width of this cross section
   */
  public double getBottomWidth();
  
  /**
   * Gets the height from the top of the cross section to the bottom of the channel.
   * This is, of course, the construction height and not the water depth
   * @return the channel height for this cross section
   */
  public double getChannelHeight();
  
  /**
   * Sets the top width for this cross section. Most implementations should
   * ensure that the top width is equal to or greater than the bottom width.
   * @param topWidth the new top width
   */
  public void setTopWidth(double topWidth);
  
  /**
   * Sets the bottom width for this cross section. Most implementations should
   * ensure that the bottom width is equal to or less than the top width. Some
   * implementations will not employ a bottom width in the normal sense, because
   * the bottom width will be zero by definition (as, for example, in a round-bottom
   * cross section). 
   * @param bottomWidth the new bottom width
   */
  public void setBottomWidth(double bottomWidth);
  
  /**
   * Sets the new channel height for this cross section. Implementations should
   * ensure that this is greater than zero.
   * @param channelHeight the new channel height for this cross section
   */
  public void setChannelHeight(double channelHeight);
  
  /**
   * Gets the cross-sectional area for this channel cross section.
   * @return the cross-sectional area for this channel cross section
   */
  public double getArea();         // This is the area of the cross-section, regardless of depth of flow
    
  /* Hydraulic Values */
  /**
   * Returns the width of the water surface, given the depth
   * of the water from the bottom of the channel.
   * @param waterDepth the depth of the water
   * @return the width of the surface of the water.
   */
  public double getWaterSurfaceWidth(double waterDepth);
  
  /**
   * Returns the wetted perimeter given the depth
   * of the water from the bottom of the channel.
   * @param waterDepth the depth of the water
   * @return the wetted perimeter
   */
  public double getWettedPerimeter(double waterDepth);
  
  /**
   * Returns the cross-sectional area of the water, given
   * the depth of the water from the bottom of the channel.
   * @param waterDepth the depth of the water
   * @return the cross-sectional area of the water
   */
  public double getFlowArea(double waterDepth);
  
  /**
   * Returns the average depth of the water, given
   * the depth of the water from the bottom of the channel.
   * 
   * The Average Depth is not, as it might seem, the simple average depth of the
   * channel across its cross-section. It is instead defined as the flow area divided
   * by the surface width.
   * @param waterDepth the depth of the water
   * @return the Average Depth of the water cross-section
   */
  public double getAvgDepth(double waterDepth);
  
  /**
   * Returns the Hydraulic Radius, given a specified water depth.
   * @param waterDepth the water depth for which Hydraulic Radius is to be calculated
   * @return the Hydraulic Radius given the specified water depth
   */
  public double getHydraulicRadius(double waterDepth);
  
  /**
   * Returns the Conveyance given the specified water depth and a channel
   * roughness.
   * @param waterDepth the depth of water for which the conveyance is to be calculated
   * @param roughness the roughness coefficient to be applied to this calculation
   * @return the Conveyance of a channel with this cross section and the specified roughness
   */
  public double getConveyance(double waterDepth, double roughness);
  
  /**
   * Returns the dA/dZ value given the specified water depth.
   * @param waterDepth the depth of water for which the dA/dZ value is to be calculated
   * @return the dA/dZ value at the specified water depth
   */
  public double getDADz(double waterDepth);
  
  /**
   * Returns the dK/dZ value given the specified water depth.
   * @param waterDepth the depth of water for which the dK/dZ value is to be calculated
   * @param roughness Manning coefficient to be used in this calculation
   * @return the dK/dZ value at the specified water depth
   */
  public double getDKDz(double waterDepth, double roughness);
  
  
  /**
   * Returns the Friction Slope given the specified water depth,
   * roughness, and discharge.
   * @param discharge flow rate to be used for this calculation
   * @param waterDepth the depth of water for which the Friction Slope is to be calculated
   * @param roughness Manning coefficient to be used in this calculation
   * @return the Friction Slope at the specified water depth
   */
  public double getFrictionSlope(double discharge, double waterDepth, double roughness);

  /**
   * Returns the velocity of normal flow as calculated from the Manning equation,
   * given this cross section at a specific water depth, and specifying a particular
   * Manning roughness and slope. 
   * @param waterDepth the depth of the water to be used in calculating the velocity
   * @param roughness the roughness coefficient to be used
   * @param slope the slope to be assumed
   * @return the velocity of normal flow given the parameters passed
   */
  public double getManningVelocity(double waterDepth, double roughness, double slope);
  
  /**
   * Returns the discharge of normal flow as calculated from the Manning equation,
   * given this cross section at a specific water depth, and specifying a particular
   * Manning roughness and slope. 
   * @param waterDepth the depth of the water to be used in calculating the discharge
   * @param roughness the roughness coefficient to be used
   * @param slope the slope to be assumed
   * @return the discharge of normal flow given the parameters passed
   */
  public double getManningDischarge(double waterDepth, double roughness, double slope);
  
  /**
   * Get the critical depth given the specified discharge.
   * @param discharge the requested discharge
   * @return the critical depth given the specified discharge.
   */
  public double getCriticalDepth(double discharge);
  
  /**
   * Gets the energy associated with the specified depth and discharge
   * @param discharge value of Q
   * @param depth depth (not stage) of water
   * @return energy value
   */
  public double getEnergy(double discharge, double depth);
  
  /**
   * Estimates the depth required to produce the specified discharge,
   * using Manning's equation
   * @param Q discharge being calculated
   * @param roughness roughness of the channel
   * @param slope slope of the channeg
   * @return the estimated depth
   */
  public double getDepthPerManningDischarge(double Q, double roughness, double slope);
  
  
  /**
   * Calculates the water depth given the specific force and flow rate.
   * @param SF specific force
   * @param Q flow rate
   * @return water depth
   */
  public double calculateWaterDepthFromSpecificForce(double SF, double Q);
  
  /**
   * Returns the depth of the centroid. By this is meant the distance from the channel
   * bottom level to the centroid; note that a common value in equations
   * is the depth from the water surface to the centroid, which would be
   * waterDepth - centroidDepth(waterDepth).
   * @param waterDepth
   * @return the depth of the centroid of this cross-section
   */
  public double centroidDepth(double waterDepth);
  
  
  
  
  /* HEC Output */
  
  /**
   * Returns a vector of values representing a cut-and-surface line that was
   * taken across this cross section. The cut-and-surface line is in absolute
   * coordinates assuming that the cross section is situated at an origin and
   * the measurements are taken along a line passing through this origin perpendicular
   * to the heading of the channel, which must be passed as an argument.
   * @param origin Point object representing the bottom, center of this cross
   * section
   * @param channelHeading heading of the channel; the cut-and-surface measurements
   * will be at a heading perpendicular to this
   * @return list of points in cut-and-surface line
   */
  public Vector getCutAndSurfaceLine(Point origin, double channelHeading);

  /**
   * Fills in a portion of the bottom of this cross section, changing its
   * shape. The details of how this occurs will be left to implementing
   * classes (i.e. a trapezoidal section may simply fill flatly up from
   * the bottom, while a parabolic one may result in a new curve).
   * @param fillDepth depth of fill to be added
   */
  public void fill(double fillDepth);  

  /**
   * Excavates a portion of this cross section, changing its shape
   * @param depth depth (below current bottom) to which to dig
   */
  public void dig(double depth);

  /**
   * Gets a String identifying the type of this cross section
   * @return the type of this cross section
   */
  public String type();
  
  
  /**
   * Gets a string describing the character and status of this cross section 
   * @return a description of the cross section
   */
  public String report();

  /**
   * Gets a copy of this channel cross section
   * @return the new channel cross section
   */
  public ChannelCrossSection getCopy();
  
}
