package flowag.channels;

import java.util.Vector;

import abcm.math.Point;


/**
 * A ChannelCrossSectionFactory can create instances of objects implementing the ChannelCrossSection interface
 * based on String descriptions of the channel cross section.
 * 
 * A simple trapezoidal cross-section can be represented by the string "TRAP:a;b;c" in which a, b, and c are numbers
 * representing the cross-section's top-width, bottom-width, and channel-height respectively.
 * 
 * A Parabolic cross-section can be represented by the string "PARA:a;b", in which a and b are numbers
 * representing the cross-section's top-width and channel-height.
 * 
 * An Irregular cross-section can be represented by the String "IRR:x1,z1;x2,z2;x3,z3;x4,z4;x5,z5..." in which each pair
 * of x,z numbers represent a distance from the left channel bank and a depth of the channel at that point.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ChannelCrossSectionFactory {

  /**
   * Gets the requested channel cross section given the specification
   * @param channelCrossSectionSpecification a specification describing the cross-section
   * @return the requested cross-section; null if unsuccessful in parsing the cross section specification
   */
  public static ChannelCrossSection getChannelCrossSection(String channelCrossSectionSpecification){
    String[] Spec = channelCrossSectionSpecification.split(":");
    if(Spec.length < 1) return null; // Uh Oh
    // Trapezoidal
    if(Spec[0].startsWith("TRAP")){
      String[] data = Spec[1].split(";");
      if(data.length < 3) return null; // Uh Oh
      double topWidth = Double.parseDouble(data[0]);
      double bottomWidth = Double.parseDouble(data[1]);
      double channelHeight = Double.parseDouble(data[2]);
      return new TrapezoidalChannelCrossSection(topWidth, bottomWidth, channelHeight);
    }
    else if(Spec[0].startsWith("PARA")){
      String[] data = Spec[1].split(";");
      if(data.length < 2) return null; // Uh Oh
      double topWidth = Double.parseDouble(data[0]);
      double channelHeight = Double.parseDouble(data[1]);
      return new ParabolicChannelCrossSection(topWidth, channelHeight);      
    }
    else if(Spec[0].startsWith("IRR")){
      Vector Points = new Vector();
      String[] data = Spec[1].split(";");
      for(int i = 0; i < data.length; i++){
        String[] datum = data[i].split(",");
        if(datum.length != 2) return null; // Uh oh
        double X = Double.parseDouble(datum[0]);
        double Z = Double.parseDouble(datum[1]);
        Point point = new Point(X, 0, Z);
        Points.add(point);
      }
      return new IrregularChannelCrossSection(Points);
    }
    return null;
  }
  
  
  
}
