package flowag.channels;

import abcm.math.Point;

/**
 * A ChannelSegment is an object that can represent a single, straight-line
 * channel through which water can flow.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public interface ChannelSegment extends Hydrographable{
    
  /**
   * Gets the Point object that represents the center of the bottom
   * of the channel segment at the upstream end.
   * @return current Head Coordinates
   */
  public Point getHeadCoords();
  
  /**
   * Gets the Point object that represents the center of the bottom
   * of the channel segment at the downstream end.
   * @return current End Coordinates
   */
  public Point getEndCoords();

  // TODO Error Traps
  
  /**
   * Sets the current upstream discharge.
   * @param Q new value for discharge, in m^3/s
   */
  public void setUpstreamDischarge(double Q);
  
  /**
   * Sets the current downstream discharge.
   * @param Q new value for discharge, in m^3/s
   */
  public void setDownstreamDischarge(double Q);
  
  /**
   * Sets the current upstream stage.
   * @param z new value for stage, in m.
   */
  public void setUpstreamStage(double z);
  
  /**
   * Sets the current downstream stage.
   * @param z new value for stage, in m.
   */
  public void setDownstreamStage(double z);
  
  /**
   * Sets the Manning roughness coefficient for the entire length of the channel.
   * @param n new value for roughness coefficient
   */
  public void setRoughness(double n);

  /**
   * Sets the current upstream stage, but does so using a depth
   * above the Head Coordinates' z-value. For example, if the HeadCoordinate
   * is at an elevation of z = 4m, and this method is called with
   * a depth of 1.2m, the stage will be set to 5.2m.
   * @param newDepth depth of the water by which the stage will be set for the upstream
   * cross section
   */
  public void setUpstreamDepth(double newDepth);
  
  /**
   * Sets the current downstream stage, but does so using a depth
   * above the End Coordinates' z-value. For example, if the EndCoordinate
   * is at an elevation of z = 4m, and this method is called with
   * a depth of 1.2m, the stage will be set to 5.2m.
   * @param newDepth depth of the water by which the stage will be set for the upstream
   * cross section
   */
  public void setDownstreamDepth(double newDepth);
  
  /**
   * Gets the value of the upstream discharge rate.
   * @return current value of upstream discharge rate, in m^3/s
   */
  public double getUpstreamDischarge();
  
  /**
   * Gets the value of the downstream discharge rate.
   * @return current value of downstream discharge rate, in m^3/s
   */
  public double getDownstreamDischarge();
  
  /**
   * Gets the value of the current upstream water stage.
   * @return current value of upstream stage, in m above a datum
   */
  public double getUpstreamStage();
  
  /**
   * Gets the value of the current downstream water stage.
   * @return current value of downstream stage, in m above a datum
   */
  public double getDownstreamStage();
  
  /**
   * Gets the current value of the roughness coefficient along the length of the channel.
   * @return Manning's n for this channel
   */
  public double getRoughness();

  /* Derived Geometry */
  /**
   * Gets the current slope of the bed of this channel segment. The slope
   * is calculated as the change in elevation from the HeadCoordinates to the
   * EndCoordinates divided by the planar distance between them.
   * @return bed slope of this channel segment.
   */
  public double getBedSlope();
  
  /**
   * Gets the length of the channel segment. The length is the planar
   * distance between the Head and End Coordinates; elevation change
   * is ignored. Example: if the
   * HeadCoordinates = (0,0,0) and the EndCoordinates = (100,0,-.5),
   * the length returned is 100.00. 
   * @return the length of the channel in the X,Y plane
   */
  public double getLength();
  
  /**
   * Gets the depth of the water presuming the water is at a given
   * stage at the upstream cross-section.
   * @param stage the stage for which a depth value should be calculated 
   * @return the depth of the water at the given stage, equal to stage argument
   * passed minus the z-value of the Head Coordinates
   */
  public double getUpstreamWaterDepth(double stage);
  
  /**
   * Gets the depth of the water presuming the water is at a given
   * stage at the downstream cross-section.
   * @param stage the stage for which a depth value should be calculated 
   * @return the depth of the water at the given stage, equal to stage argument
   * passed minus the z-value of the End Coordinates
   */
  public double getDownstreamWaterDepth(double stage);
  
  /**
   * Gets the depth of the water at the upstream cross-section given
   * the current upstream water stage. 
   * @return the depth of the water, equal to the current stage minus
   * the z-value of the Head Coordinates
   */
  public double getUpstreamWaterDepth();
  
  /**
   * Gets the depth of the water at the downstream cross-section given
   * the current upstream water stage. 
   * @return the depth of the water, equal to the current stage minus
   * the z-value of the End Coordinates
   */
  public double getDownstreamWaterDepth();
  
  /**
   * Gets the average of the flow area at the upstream and downstream cross
   * sections given the current stages.
   * @return the average flow area of the two cross sections given their
   * current stages 
   */
  public double getAverageFlowArea();
  
  /**
   * Gets the total construction volume of this channel segment. Note
   * that this is not the volume of water, but is the volume of the
   * segment as constructed.
   * @return the total construction volume of the ChannelSegment
   */
  public double getTotalVolume();

  /* Water Geometry */
  
  /**
   * Gets the total volume of water in this canal given the current
   * values for upstream and downstream stage (and thus depth)
   * @return the total volume of water
   */
  public double getTotalWater();
  
  /**
   * Gets the total volume of water in this segment given the specified upstream and downstream depths
   * @param upstreamWaterDepth upstream depth
   * @param downstreamWaterDepth downstream depth
   * @return volume of water
   */
  public double getTotalWater(double upstreamWaterDepth, double downstreamWaterDepth);
  
  
  /**
   * Changes the shape of this Channel Segment by changing the upper and lower
   * cross-section. Must return a value indicating the difference between the volumes
   * of the original and new forms.
   * @param UpstreamChannelCrossSection new upstream channel cross section
   * @param DownstreamChannelCrossSection new downstream channel cross section
   * @return the difference between the volumes of the old and new channel geometries
   */
  public double changeShape(ChannelCrossSection UpstreamChannelCrossSection, 
      ChannelCrossSection DownstreamChannelCrossSection);

  /* Wrap the dry geometric values */
  /**
   * Gets the bottom width of the upstream cross section.
   * @return bottom width of the upstream cross section
   */
  public double getUpstreamBottomWidth();
  
  /**
   * Gets the top width of the upstream cross section.
   * @return top width of the upstream cross section
   */
  public double getUpstreamTopWidth();
  
  /**
   * Gets the channel height of the upstream cross section.
   * @return channel height of the upstream cross section
   */
  public double getUpstreamChannelHeight();
  
  /**
   * Gets the area of the upstream cross section. This is the construction
   * area, not the flow area.
   * @return construction area of the upstream cross section
   */
  public double getUpstreamArea();

  /**
   * Gets the bottom width of the downstream cross section.
   * @return bottom width of the downstream cross section
   */
  public double getDownstreamBottomWidth();
  
  /**
   * Gets the top width of the downstream cross section.
   * @return top width of the downstream cross section
   */
  public double getDownstreamTopWidth();
  
  /**
   * Gets the channel height of the downstream cross section.
   * @return channel height of the downstream cross section
   */
  public double getDownstreamChannelHeight();
  
  /**
   * Gets the area of the upstream cross section. This is the construction
   * area, not the flow area.
   * @return construction area of the upstream cross section
   */
  public double getDownstreamArea();

  /* Wrap the hydraulic values */
  /**
   * Gets the width of the water surface, given current stage, at the upstream
   * cross section.
   * @return water surface width at the upstream cross section
   */
  public double getUpstreamWaterSurfaceWidth();
  
  /**
   * Gets the average depth, given current stage, at the upstream
   * cross section.
   * @return average depth at the upstream cross section
   */
  public double getUpstreamAvgDepth();
  
  /**
   * Gets the flow area, given current stage, at the upstream
   * cross section.
   * @return flow area at the upstream cross section
   */
  public double getUpstreamFlowArea();
  
  /**
   * Gets the wetted perimeter, given current stage, at the upstream
   * cross section.
   * @return wetted perimeter at the upstream cross section
   */
  public double getUpstreamWettedPerimeter();
  
  /**
   * Gets the hydraulic radius, given current stage, at the upstream
   * cross section.
   * @return hydraulic radius at the upstream cross section
   */
  public double getUpstreamHydraulicRadius();
  
  /**
   * Gets the dA/dZ value, given current stage, at the upstream
   * cross section.
   * @return dA/dZ value at the upstream cross section
   */
  public double getUpstreamDADz();
  
  /**
   * Gets the conveyance, given current stage, at the upstream
   * cross section.
   * @return conveyance value at the upstream cross section
   */
  public double getUpstreamConveyance();
  
  /**
   * Gets the dK/dZ value, given current stage, at the upstream
   * cross section.
   * @return dK/dZ value at the upstream cross section
   */
  public double getUpstreamDKDz();
  
  /**
   * Gets the velocity head at the upstream cross section. The velocity head
   * is calculated as the square of the velocity divided by twice the acceleration
   * due to gravity (V^2 / 2g).
   * @return the velocity head at the upstream cross section
   */
  public double getUpstreamVelocityHead();
  
  /**
   * Gets the Specific Energy at the upstream cross section. The Specific Energy
   * is the sum of the stage and the velocity head.
   * @return the Specific Energy at the upstream cross section.
   */
  public double getUpstreamSpecificEnergy();
  
  /**
   * Gets the velocity head at the downstream cross section. The velocity head
   * is calculated as the square of the velocity divided by twice the acceleration
   * due to gravity (V^2 / 2g).
   * @return the velocity head at the downstream cross section
   */
  public double getDownstreamVelocityHead();
  
  /**
   * Gets the Specific Energy at the downstream cross section. The Specific Energy
   * is the sum of the stage and the velocity head.
   * @return the Specific Energy at the downstream cross section.
   */
  public double getDownstreamSpecificEnergy();
  
  /**
   * Gets the friction slope, given current discharge, at the upstream
   * cross section. Note that this is NOT the same as the instantaneous
   * friction slope calculated for the cross section, which is derived from Manning's
   * equation. This is the change in Specific Energy over the length of the
   * segment, (SE1 - SE2) / x. Note that this will be the same for both
   * the upstream and downstream cross sections, but 
   * @return the friction slope of the upstream cross section
   */
  public double getFrictionSlope();
  
  /**
   * Stub to return the friction slope for the 'upstream' cross section.
   * @return the friction slope of this segment
   */
  public double getUpstreamFrictionSlope();
  
  /**
   * Stub to return the friction slope for the 'downstream' cross section.
   * @return the friction slope of this segment
   */
  public double getDownstreamFrictionSlope();
  
  /**
   * Gets the width of the water surface, given current stage, at the downstream
   * cross section.
   * @return water surface width at the downstream cross section
   */
  public double getDownstreamWaterSurfaceWidth();
  
  /**
   * Gets the flow area, given current stage, at the downstream
   * cross section.
   * @return flow area at the downstream cross section
   */
  public double getDownstreamAvgDepth();
  
  /**
   * Gets the flow area, given current stage, at the downstream
   * cross section.
   * @return flow area at the downstream cross section
   */
  public double getDownstreamFlowArea();
  
  /**
   * Gets the wetted perimeter, given current stage, at the downstream
   * cross section.
   * @return wetted perimeter at the downstream cross section
   */
  public double getDownstreamWettedPerimeter();
  
  /**
   * Gets the hydraulic radius, given current stage, at the downstream
   * cross section.
   * @return hydraulic radius at the downstream cross section
   */
  public double getDownstreamHydraulicRadius();
  
  /**
   * Gets the dA/dZ value, given current stage, at the downstream
   * cross section.
   * @return dA/dZ value at the downstream cross section
   */
  public double getDownstreamDADz();
  
  /**
   * Gets the conveyance, given current stage, at the downstream
   * cross section.
   * @return conveyance at the downstream cross section
   */
  public double getDownstreamConveyance();
  
  /**
   * Gets the dK/dZ value, given current stage, at the downstream
   * cross section.
   * @return dK/dZ value at the downstream cross section
   */
  public double getDownstreamDKDz();
  
  /* Derived hydraulic values */
  /**
   * Gets the flow velocity at the upstream cross section, given
   * the current stage and discharge. This is the quotient of
   * the discharge divided by the flow area. Under conditions of 
   * normal flow this will equal the velocity as calculated by Manning's
   * equation. Returns zero if depth is zero.
   * @return current upstream flow velocity
   */
  public double getUpstreamFlowVelocity();
  
  /**
   * Gets the flow velocity at the downstream cross section, given
   * the current stage and discharge. This is the quotient of
   * the discharge divided by the flow area. Under conditions of 
   * normal flow this will equal the velocity as calculated by Manning's
   * equation. Returns zero if depth is zero.
   * @return current downstream flow velocity
   */
  public double getDownstreamFlowVelocity();

  /* Empirical hydraulic values (based on Manning's equation) */
  /**
   * Gets the velocity as calculated by the Manning equation
   * given the current stage at the upstream cross section.
   * @return the velocity as calculated by Manning's equation, given
   * the current stage at the upstream cross section.
   */
  public double getUpstreamManningVelocity();
  
  /**
   * Gets the discharge as calculated by the Manning equation
   * given the current stage at the downstream cross section.
   * @return the discharge as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getUpstreamManningDischarge();

  /**
   * Gets the velocity as calculated by the Manning equation
   * given the current stage at the downstream cross section.
   * @return the velocity as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getDownstreamManningVelocity();
  
  /**
   * Gets the discharge as calculated by the Manning equation
   * given the current stage at the downstream cross section.
   * @return the discharge as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getDownstreamManningDischarge();

  /**
   * Gets the velocity as calculated by the Manning equation
   * given the specified stage, passed as an argument, at the upstream cross section.
   * @param stage water stage on which calculations are based
   * @return the velocity as calculated by Manning's equation, given
   * the current stage at the upstream cross section.
   */
  public double getUpstreamManningVelocity(double stage);
  
  /**
   * Gets the discharge as calculated by the Manning equation
   * given the specified stage, passed as an argument, at the downstream cross section.
   * @param stage water stage on which calculations are based
   * @return the discharge as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getUpstreamManningDischarge(double stage);
  
  /**
   * Gets the velocity as calculated by the Manning equation
   * given the specified stage, passed as an argument, at the downstream cross section.
   * @param stage water stage on which calculations are based
   * @return the velocity as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getDownstreamManningVelocity(double stage);

  /**
   * Gets the discharge as calculated by the Manning equation
   * given the specified stage, passed as an argument, at the downstream cross section.
   * @param stage water stage on which calculations are based
   * @return the discharge as calculated by Manning's equation, given
   * the current stage at the downstream cross section.
   */
  public double getDownstreamManningDischarge(double stage);
  
  
  /**
   * Gets the upstream depth given the specified discharge and assuming
   * Manning's equation.
   * @param Q discharge rate
   * @return estimated depth
   */
  public double getUpstreamDepthPerManningDischarge(double Q);
  
  /**
   * Gets the downstream depth given the specified discharge and assuming
   * Manning's equation.
   * @param Q discharge rate
   * @return estimated depth
   */
  public double getDownstreamDepthPerManningDischarge(double Q);
  
  /**
   * Sets the stage and discharge values for the upstream cross
   * section. The argument passed is the stage value; the discharge
   * is set to the value given by Manning's equation for normal
   * flow at that stage.
   * @param stage the stage value to be set, and from which discharge
   * is also calculated and set
   */
  public void setUpstreamStagePlusManningDischarge(double stage);
  
  /**
   * Sets the stage and discharge values for the downstream cross
   * section. The argument passed is the stage value; the discharge
   * is set to the value given by Manning's equation for normal
   * flow at that stage.
   * @param stage the stage value to be set, and from which discharge
   * is also calculated and set
   */
  public void setDownstreamStagePlusManningDischarge(double stage);
  
  /**
   * Sets the stage and discharge values for the upstream cross
   * section. The argument passed is a depth value, from which
   * stage is calculated by adding the depth to the z value of the
   * Head Coordinates; the discharge is set to the value given by Manning's 
   * equation for normal flow at that stage.
   * @param depth the depth value from which the stage value to be set is calculated, 
   * after which discharge is also calculated and set
   */
  public void setUpstreamDepthPlusManningDischarge(double depth);
  
  /**
   * Sets the stage and discharge values for the upstream cross
   * section. The argument passed is a depth value, from which
   * stage is calculated by adding the depth to the z value of the
   * End Coordinates; the discharge is set to the value given by Manning's 
   * equation for normal flow at that stage.
   * @param depth the depth value from which the stage value to be set is calculated, 
   * after which discharge is also calculated and set
   */
  public void setDownstreamDepthPlusManningDischarge(double depth);

  /**
   * Gets the volume of water between the two specified points
   * along the channel segment
   * @param x1 distance from head coordinates of the upstream point
   * @param x2 distance from head coordinats of the downstream point
   * @return the volume of water in the channel segment between the
   * two points x1 and x2
   */
  public double getWaterVolumeBetweenPositions(double x1, double x2);
  
  /**
   * Gets the cross-sectional area of water flowing at position x
   * @param x distance from upstream end of segment at which flow
   * area is to be calculated
   * @return the cross-sectional area of the water flowing at the
   * specified position
   */
  public double getFlowAreaAtPosition(double x);
  
  /**
   * Gets the depth of the water at the position specified
   * @param x distance from head coordinates at which depth of
   * water is to be reported
   * @return depth of water at specified point x
   */
  public double getDepthAtPosition(double x);
  
  /**
   * Gets the stage of the water at the position specified
   * @param x distance from head coordinates at which stage of
   * water is to be reported
   * @return depth of water at specified point x
   */
  public double getStageAtPosition(double x);
  
  /**
   * Gets the z coordinate of the bottom of the channel at the
   * specified position
   * @param x distance from the head coordinates at which z coordinate
   * is requested
   * @return z coordinate at specified point x
   */
  public double getBottomZAtPosition(double x);
  
  /**
   * Gets the average flow velocity for the channel segment. This
   * is the simple average of the flow velocity at the upstream
   * and downstream ends
   * @return the calculated average flow velocity
   */
  public double averageFlowVelocity();
  
  /**
   * Gets the calculated time it would take for a given point
   * in the water flow stream to move from the upstream end of the
   * segment to the downstream end. In general this will be length of the
   * segment divided by the average flow velocity. Note that if velocity
   * is zero (water is not moving), zero is returned.
   * @return The calculated time
   */
  public double timeToFlowThroughSegment();
  
  /**
   * Calculates the change in velocity of water from the upstream end
   * to the downstream end of the segment divided by the amount of time
   * required to traverse that distance.
   * @return calculated flow acceleration
   */
  public double flowAcceleration();
  
  
  
  /* Incomplete */
  
  /**
   * Fills this segment with the specified depth, deforming its shape and changing its volume
   * @param fillDepth depth to be added
   */
  public void infill(double fillDepth);
  
  /**
   * Digs this segment out by a further amount, changing its shape and volume
   * @param depth depth by which it is to be increased
   */
  public void excavate(double depth);

  /**
   * Calculates the total surface area of the construction of the canal.
   * @return the total surface area, in m^2, of the construction of the
   * canal.
   */
  public double getTotalConstructionSurfaceArea();
  
  /**
   * Gets the labor required to change the depth by the specified amount
   * @param depthChange depth increase
   * @return labor required
   */
  public double getLaborToChangeDepth(double depthChange);
      
}