package flowag.channels;

import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;

import abcm.math.SimDate;


/**
 * Objects implementing the FlowCalculator interface should 
 * employ a constructor that allows them to be directed to
 * a collection of connected CanalSegment objects forming
 * a flowing system. A common pattern will be two constructors:
 * one that takes an initial canal segment and is expected to
 * track all downstream segments until an end to the flow is reached
 * (either a closed gate or the end of the canal), a second
 * that takes an initial segment and a collection of CanalSegments
 * that are valid; only 'valid' segments are considered to be
 * part of the system, and hence they define the boundaries
 * of the FlowSystem even if the actual canals go further. A variant
 * constructor should allow overriding the 'closed gate' provision
 * to move beyond a gate that is actually closed. Using this
 * approach any subset of a canal system can be specified.
 * 
 * The FlowCalculator allows the specification of upstream
 * boundary conditions for the initial segment and downstream
 * boundary conditions for terminal segments. Closed gates
 * are recognized as boundary conditions and need not
 * be specified explicitly, but if a boundary condition
 * is specified for a node that would otherwise be represented
 * by a closed gate, the explicit boundary condition should
 * overrule the one inferred by the gate status.
 * 
 * 
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public interface FlowCalculator {

  /* Inner classes that implementors can use */
  /**
   * Upstream boundary conditions describe what is known to be happening
   * in a stream or canal system upstream of the initial segment. They can
   * be used to calculate flow and stage changes throughout the system.
   *  
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public abstract class UpstreamBoundaryCondition {

    /** 
     * Should return true if the segment upstream is empty and false if not
     * @return true if the segment upstream of this boundary condition is empty
     */
    public abstract boolean isEmptySegmentUpstream();
    
    /**
     * Some boundary conditions will specify a dQ value; others will not. This
     * should identify whether this implementation does or not
     * @return true if this boundary condition specifies a dQ value, false if not
     */
    public abstract boolean specifiesDQ();
    
    /**
     * Gets the change in flow (dQ) given an interval of time
     * @param interval the duration of the interval
     * @return the calculated dQ over that interval
     */
    public abstract double getDQ(double interval);
    
    /**
     * Some boundary conditions will specify a dz value; others will not. This
     * should identify whether this implementation does or not
     * @return true if this boundary condition specifies a dz value, false if not
     */
    public abstract boolean specifiesDZ();
    
    /**
     * Gets the change in stage (dz) given an interval of time
     * @param interval the duration of the interval
     * @return the calculated dz over the interval
     */
    public abstract double getDZ(double interval);
    
  }
  
  /**
   * Assumes that the upstream boundary is a closed gate; change in
   * Q will be -Q, so that flow goes to zero.
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class ClosedGateUpstreamBoundaryCondition extends UpstreamBoundaryCondition{

    /** Value for dQ (change in flow)  */
    private double dq;
    
    /**
     * Constructor
     * @param canalSegment the canal segment at the boundary
     */
    public ClosedGateUpstreamBoundaryCondition(CanalSegment canalSegment){
      dq = -1 * canalSegment.getUpstreamDischarge();
    }
    
    /**
     * Returns true
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#specifiesDQ()
     */
    public boolean specifiesDQ(){             return true;   }
    /**
     * Returns the negative of the original canal's Q value
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#getDQ(double)
     */
    public double  getDQ(double interval){    return dq;     }
    /**
     * Returns false
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#specifiesDZ()
     */
    public boolean specifiesDZ(){             return false;  }
    /**
     * Returns zero
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#getDZ(double)
     */
    public double  getDZ(double interval){    return 0;      }
    /**
     * Returns false
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#isEmptySegmentUpstream()
     */
    public boolean isEmptySegmentUpstream(){  return false;  }
    

  }
  
  /**
   * Sets a boundary condition based on the output of a headgate.
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class HeadgateBasedUpstreamBoundaryCondition extends UpstreamBoundaryCondition{

    /** The Headgate represented */         protected HeadGate gate;
    /** Time of day (as fraction of day) */ protected double   time;
    
    /**
     * Constructor
     * @param H the Headgate represented
     * @param startTimeAsFractionOfDay the time of day for which the boundary conditions are to be calculated
     */
    public HeadgateBasedUpstreamBoundaryCondition(HeadGate H, double startTimeAsFractionOfDay){
      gate = H;    
      time = startTimeAsFractionOfDay;
    }
    
    /**
     * Allows the start time to be re-set
     * @param startTime new value for start time
     */
    public void resetStartTime(double startTime){
      time = startTime;
    }
    
    /**
     * Returns true
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#specifiesDQ()
     */
    public boolean specifiesDQ(){                 return true;                                 }
    /**
     * Gets the dQ value calculated from the headgate
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#getDQ(double)
     */
    public double  getDQ(double interval){        return gate.getFlowChange(time, interval);   }
    /**
     * Returns false
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#specifiesDZ()
     */
    public boolean specifiesDZ(){                 return false;                                }
    /**
     * Returns zero
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#getDZ(double)
     */
    public double  getDZ(double interval){        return 0;                                    }

    /**
     * Returns false
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#isEmptySegmentUpstream()
     */
    public boolean isEmptySegmentUpstream() {     return false;                                }
  
    /**
     * Gets the value of Q at the end of the specified interval from the given start time. This
     * is the actual Q value, not dQ
     * @param interval time interval
     * @return Q at end of specified interval
     */
    public double  getFinalQ(double interval){    
      return gate.getFlowAt(time) + gate.getFlowChange(time, interval); 
    }
    
  }
  
  /**
   * This boundary condition represents a situation in which
   * nothing is known about the upstream conditions: neither dQ nor
   * dz is specified. The upstream segment is assumed to be empty.
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class OpenGateUpstreamBoundaryCondition extends UpstreamBoundaryCondition{

    
    /**
     * Returns false
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#specifiesDQ()
     */
    public boolean specifiesDQ(){            return false;   }
    /**
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#getDQ(double)
     */
    public double  getDQ(double interval){   return 0;       }
    /**
     * Returns false
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#specifiesDZ()
     */
    public boolean specifiesDZ(){            return false;   }
    /**
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#getDZ(double)
     */
    public double  getDZ(double interval){   return 0;       }
    
    /**
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#isEmptySegmentUpstream()
     */
    public boolean isEmptySegmentUpstream(){ return true;    }
    
  }
  
  /**
   * This upstream boundary condition specifies a simple rate of change for Q.
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class SimpleRateOfChangeUpstreamBoundaryCondition extends UpstreamBoundaryCondition{

    /**
     * Rate at which Q is changing
     */
    private double rate;
    
    /**
     * Constructor; rate of change is initialized and remains constant.
     * @param rateOfChangePerSecond
     */
    public SimpleRateOfChangeUpstreamBoundaryCondition(double rateOfChangePerSecond){
      rate = rateOfChangePerSecond;
    }

    // Getters for delta-Q
    /**
     * Returns true
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#specifiesDQ()
     */
    public boolean specifiesDQ(){               return true;              }
    /**
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#getDQ(double)
     */
    public double getDQ(double interval){       return rate * interval;   }
    
    // Inoperative
    /**
     * Returns false
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#specifiesDZ()
     */
    public boolean specifiesDZ(){               return false;   }
    /**
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#getDZ(double)
     */
    public double  getDZ(double interval){      return 0;       }

    /**
     * Returns false; upstream segment is not assumed to be empty
     * @see flowag.channels.FlowCalculator.UpstreamBoundaryCondition#isEmptySegmentUpstream()
     */
    public boolean isEmptySegmentUpstream(){    return false;   }

  }
  
  /**
   * A Downstream Boundary Condition is an empty class designed for utility and symmetry
   * with Upstream boundary conditions.
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public abstract class DownstreamBoundaryCondition {

  }
  
  /**
   * This object represents a collection of dz and dQ values for
   * a specified network of canal segments over a specific
   * duration. The network may consist of one initial canal
   * segment and any number of downstream ones; intuitively this
   * corresponds to a dendritic arrangment branching from a single
   * node, but this is not enforced (any segments may be specified
   * as 'downstream' segments, but whether they are actually connected
   * to the upstream segments, or whether all segments that could be
   * connected to the initial segment actually are, is not enforced).
   * 
   * Additionally, instances of this object can be invoked to apply
   * the delta values to a collection of canal segments, performing
   * the adjustments to their flow values.
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.   
   */
  public class FlowZAndQAdjustments{
   
    /** Initial canal segment                             */
    protected CanalSegment initialCanalSegment;
    /** Initial z (stage) for initial canal segment       */
    protected double       initialCanalSegmentUpstreamZ;
    /** Initial Q (flow) for initial canal segment        */
    protected double       initialCanalSegmentUpstreamQ;
    /**
     * Hashtable of delta-z (dz) values, indicating the
     * change in stage that results from the flow calculations
     */
    protected Hashtable    downstreamZValues = new Hashtable();
    /**
     * Hashtable of delta-Q (dQ) values, indicating the
     * change in flow rate that results from the flow calculations
     */
    protected Hashtable    downstreamQValues = new Hashtable();
    
    /** Debugging variable; when compiled with 'true' allows verbose output.  */
    private boolean display = false;
    
    /* Setters */
    /**
     * Sets the initial collection of values for the initial canal segment
     * @param initialSegment the initial canal segment
     * @param upstreamZ the initial z value for this canal segment
     * @param upstreamQ the initial Q value for this canal segment
     */
    public void setInitialUpstreamValues(CanalSegment initialSegment, double upstreamZ, double upstreamQ){
      initialCanalSegment = initialSegment;
      initialCanalSegmentUpstreamZ = upstreamZ;
      initialCanalSegmentUpstreamQ = upstreamQ;
    }
    
    /**
     * Adds the specified results to the collection of adjustments to downstream
     * values. 
     * @param cs the canal segment
     * @param downstreamZValue the calculated change in z
     * @param downstreamQValue the calculated change in Q
     */
    public void addDownstreamValues(CanalSegment cs, double downstreamZValue, double downstreamQValue){
      downstreamZValues.put(cs, new Double(downstreamZValue));
      downstreamQValues.put(cs, new Double(downstreamQValue));      
    }
    
    
    /* Actions */
    
    /**
     * Applies the results to a collection of canal segments- that is, takes the calculated values
     * and applies them as changes to the existing values.
     * @param canalSegments the collection of canal segments to which the results should be applied.
     */
    public void apply(Collection canalSegments){
      applyFraction(canalSegments, 1, 0); 
    }
    
    /**
     * Applies the results to a collection of canal segments- that is, takes the calculated values
     * and applies them as changes to the existing values. This variant also specifies the time
     * of day at which the flow change took place, to be recorded in any stage hydrographs the
     * canal segments might have.
     * @param canalSegments the collection of canal segments to which the results should be applied.
     * @param timeOfDay the time of day the changes occurred
     */
    public void apply(Collection canalSegments, double timeOfDay){
      applyFraction(canalSegments, 1, timeOfDay);
    }
    
    /**
     * Applies the results to a collection of canal segments- that is, takes the calculated values
     * and applies them as changes to the existing values. This variant allows a fractional
     * result to be applied, as if only some fraction of the calculated change were to take place
     * (as, for example, if the change is calculated for a long time period and applied only during
     * a portion of it- which may be appropriate for some kinds of calculations and not for others,
     * but this is for the implementation to decide. 
     * @param canalSegments the collection of canal segments to which the results should be applied.
     * @param fraction the fraction of the changes to apply
     */
    public void applyFraction(Collection canalSegments, double fraction){
      applyFraction(canalSegments, fraction, 0);
    }
    
    /**
     * Applies the results to a collection of canal segments- that is, takes the calculated values
     * and applies them as changes to the existing values. This variant allows a fractional
     * result to be applied, as if only some fraction of the calculated change were to take place
     * (as, for example, if the change is calculated for a long time period and applied only during
     * a portion of it- which may be appropriate for some kinds of calculations and not for others,
     * but this is for the implementation to decide. This variant also specifies the time
     * of day at which the flow change took place, to be recorded in any stage hydrographs the
     * canal segments might have.
     * @param canalSegments the collection of canal segments to which the results should be applied.
     * @param fraction the fraction of the changes to apply
     * @param dateValue the time of day the changes occurred
     */
    public void applyFraction(Collection canalSegments, double fraction, double dateValue){
      SimDate SD = new SimDate(dateValue);
          
      if(display){
        CanalSegment displaycs;
        displaycs = initialCanalSegment;
        System.out.print(displaycs.getCanalSegmentID() + " USQ: " + 
            displaycs.getUpstreamDischarge() + " + " +
            initialCanalSegmentUpstreamQ + " = " + 
            (displaycs.getUpstreamDischarge() + initialCanalSegmentUpstreamQ));
        System.out.println(" USZ: " +  
            displaycs.getUpstreamStage() + " + " +
            initialCanalSegmentUpstreamZ + " = " + 
            displaycs.getUpstreamStage() + initialCanalSegmentUpstreamZ);
        Iterator displayCanals = canalSegments.iterator();
        String DS = "";
        while(displayCanals.hasNext()){
          displaycs = (CanalSegment)displayCanals.next();
          if(downstreamZValues.containsKey(displaycs)){
            DS = displaycs.getCanalSegmentID() + " DSQ: " +  
                displaycs.getDownstreamDischarge() + " + " +
                ((Double)downstreamQValues.get(displaycs)).doubleValue() + " = " +
                (displaycs.getDownstreamDischarge() + ((Double)downstreamQValues.get(displaycs)).doubleValue())
            + " DSZ: " + 
                displaycs.getDownstreamStage() + " + " +
                ((Double)downstreamZValues.get(displaycs)).doubleValue() + " = " + 
                displaycs.getDownstreamStage() + ((Double)downstreamZValues.get(displaycs)).doubleValue() * fraction + "\n" + DS;    
          }
        }
        System.out.println(DS);
        display = false; // Run only first time applying... 
      }
      
      // This shouldn't happen, but if it does, fail gracefully
      if(initialCanalSegment != null && canalSegments.contains(initialCanalSegment)) {
        CanalSegment cs = initialCanalSegment;
        initialCanalSegment.setUpstreamStage(initialCanalSegment.getUpstreamStage() + this.initialCanalSegmentUpstreamZ * fraction,         SD);
        initialCanalSegment.setUpstreamDischarge(initialCanalSegment.getUpstreamDischarge() + this.initialCanalSegmentUpstreamQ * fraction, SD);
      }
      Iterator canals = canalSegments.iterator();
      while(canals.hasNext()){
        CanalSegment cs = (CanalSegment)canals.next();
        if(downstreamZValues.containsKey(cs)){
          cs.setDownstreamStage(cs.getDownstreamStage() + ((Double)downstreamZValues.get(cs)).doubleValue() * fraction,           SD);
          cs.setDownstreamDischarge(cs.getDownstreamDischarge() + ((Double)downstreamQValues.get(cs)).doubleValue() * fraction,   SD);          
        }
      }
    }

  }
  
  
  // All of these return null if calculate is unsuccessful...
  
  /**
   * Sets the upstream boundary condition
   * @param ubc the condition to be installed
   */
  public void setUpstreamBoundaryCondition(UpstreamBoundaryCondition ubc);
  /**
   * Sets the (general) downstream boundary condition. This will be applied
   * to all outflowing termini.
   * @param dbc the condition to be installed
   */
  public void setDownstreamBoundaryCondition(DownstreamBoundaryCondition dbc);
  /**
   * Sets the downstream boundary condition for a specific canal segment
   * @param cs the segment for which the boundary condition is to be set
   * @param dbc the boundary condition to install
   */
  public void setDownstreamBoundaryCondition(CanalSegment cs, DownstreamBoundaryCondition dbc);
  
  /**
   * This method is invoked to perform the flow calculations. Boundary conditions
   * and initial z and Q values in the canal segments are used to determine dz and dQ
   * values that can then be applied to simulate flow. Note that the calculation
   * can be unsuccessful (as in the case of complex matrix calculations); the calling
   * program should manage this and assess alternatives to correct for the problem.
   *  
   * @param duration interval over which calculations should be done
   * @return true if successful. 
   */
  public boolean calculate(double duration);
  /**
   * Gets the collection of results; can shortcut and wrap the 'calculate' method
   * @return the results collection
   */
  public FlowZAndQAdjustments getResults();
  /**
   * Gets the collection of results; can shortcut and wrap the 'calculate' method
   * @param duration interval for calculation 
   * @return the results collection over the given interval
   */
  public FlowZAndQAdjustments getResults(double duration);

}
