package flowag.channels;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.sim.AbstractModelComponent;
import flowag.basic.Water;


/**
 * A Gate object lies at the intersection of
 * any two Canal Segments and can be used to interrupt
 * flow between them.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class Gate extends AbstractModelComponent implements ABCMDeepCloneable{

  /** Status of this gate, open or closed  */
  private boolean open;
  /** Gates may be damaged; this flag indicates this status */
  private boolean broken;
  
  /** Represents the canal segment flowing into this gate  */
  private CanalSegment source;
  /** The canal segment into which water flows upon leaving this gate */
  private CanalSegment outflow;
  /** HeadGate associated with this gate */
  private HeadGate headGate;

  /**  
   * Default value for the amount of labor required to repair a broken gate. 
   * Value is 8 man-hours or one day
   */
  private static final double DEFAULT_LABOR_TO_REPAIR = 8; 

  /**
   * Constructor with source and outflow as arguments
   * @param in upstream Canal Segment
   * @param out downstream Canal Segment
   */
  public Gate(CanalSegment in, CanalSegment out){
    source  = in;
    outflow = out;
    headGate = null;
    open = true;
  }

  /**
   * Constructor that takes a HeadGate as a source and 
   * a CanalSegment as outflow.
   * @param in
   * @param out
   */
  public Gate(HeadGate in, CanalSegment out){
    headGate = in;
    source = out;
    outflow = out;
    open = true;
  }
  
  /**
   * No argument constructor- for creating gates that do not yet have sources or outflow
   *
   */
  public Gate(){
    source = null;
    outflow = null;
    headGate = null;
    open = true;
  }

  /**
   * Indicates whether the Gate is open or not
   * @return true if the gate is open, false if not
   */
  public boolean isOpen() {
    if(outflow == null) return true;
    return open;
  }
  
  /**
   * Open or close this Gate. The gate can only be set open
   * or closed if it is not broken- broken gates are always open.
   * @param b true if the gate should be open, false if not
   */
  public void setOpen(boolean b) {
    if(!broken) open = b;
  }

  /**
   * Get the CanalSegment into which this gate distributes water
   * @return the outflow canal segment for this gate.
   */
  public CanalSegment getOutflow() {
    return outflow;
  }
  
  /**
   * Get the CanalSegment from which water flows through this gate
   * @return the outflow canal segment for this gate.
   */
  public CanalSegment getSource() {
    return source;
  }
  
  /**
   * Set the source Canal Segment for this gate
   * @param CS The Canal Segment that will be the source for this Gate
   */
  public void setSource(CanalSegment CS){
    source = CS;
  }
  
  /**
   * Get the HeadGate to which this gate is attached
   * @return the HeadGate to which this gate is attached
   */
  public HeadGate getHeadGate(){
    return headGate;
  }
  
  /**
   * Sets the HeadGate and outflow canal for this gate.
   * @param HG the HeadGate that should become the HeadGate for this Gate.
   * @param cs outflow Canal Segment
   */
  public void setHeadGateAndOutflowCanal(HeadGate HG, CanalSegment cs){
    headGate = HG;
    outflow = cs;
  }

  /**
   * Gets a water sample from upstream of this gate
   * @return the requested water sample
   */
  public Water getUpstreamWaterSample(){
    if (source != outflow) return source.getOutflowWaterSample();
    else return headGate.getWaterSample();
  }
  
  /**
   * Gets a water sample from downstream of this gate
   * @return the requested water sample
   */
  public Water getDownstreamWaterSample(){
    return ((outflow != source)? outflow.getInflowWaterSample(): source.getOutflowWaterSample());
  }

  /**
   * Gets the repair status of this gate
   * @return true if the gate is broken
   */
  public boolean isBroken() { return broken;}
  
  /**
   * Damages the gate by setting the 'broken' variable to 'true'
   */
  public void damage(){ broken = true; open = true; }
  
  
  /**
   * Gets the amount of labor required to repair this gate
   * @return an amount of labor, in man-hours
   */
  public double getLaborToRepair(){ return DEFAULT_LABOR_TO_REPAIR; }
  
  /**
   * Repair this gate
   * @param labor the amount of labor applied to the repair
   * @return the amount of labor actually used in the repair. If the amount
   * applied was not enough, returns zero and the repair is not completed (status remains broken)
   */
  public double repair(double labor){
    if(broken == false) return 0;
    double laborNeeded = getLaborToRepair();
    if(labor >= laborNeeded){
      broken = false;
      return laborNeeded;
    }
    else return 0; // Not enough labor was allocated; this is all-or-nothing,
                   // so the result is zero
  }

  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    Gate ret = new Gate();
    cloneStore.put(this, ret);
    ret.broken = broken;
    ret.open = open;
    
    ret.headGate = (HeadGate)(cloneStore.getClone(headGate));
    ret.outflow = (CanalSegment)(cloneStore.getClone(outflow));
    ret.source =  (CanalSegment)(cloneStore.getClone(source));
    return ret;
  }
  
}

