package flowag.channels;

import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import abcm.math.BasicMath;
import abcm.math.LinearSystemSolver;
import flowag.basic.Global;

/**
 * A FlowSystem represents a collection of channel segments
 * that can model flow using the HEC-RAS implicit finite difference
 * approximation scheme.
 * 
 * For details about the calculations involved, see the FlowAg
 * manual.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class HEC_RAS_FlowCalculator extends AbstractFlowCalculator{

  
  /** Map of cells for flow calculations  */
  private Hashtable cellList           = new Hashtable();
  /** Map of nodes for flow calculations  */
  private Hashtable nodeList           = new Hashtable();
  /** Map of terminal nodes  */
  private Hashtable terminalNodeList   = new Hashtable();
  /** Map of junctions  */
  private Vector junctionList          = new Vector();
  /** List of nodes, ordered  */
  private Vector nodeOrderList         = new Vector();
  /** List of cells, ordered  */
  private Vector cellOrderList         = new Vector();
  /** Linear System Solver used to calculate flow  */
  private LinearSystemSolver LSS;


  /** Duration last used for calculation of Linear System */
  private double  lastDurationForLSS;
  /** Theta value last used for calculation of Linear System */
  private double  lastThetaForLSS;
  /** Flag indicating whether last LSS calculation was successful */
  private boolean lastSuccessOfLSS;
  
  /** Minimum time step value (seconds)  */
  private static final double MINIMUM_TIMESTEP = 10;
  /** Minimum calculated depth (below which any value is set to zero)  */
  private static double MINIMUM_CALCULATED_DEPTH = .001;    // Any depth smaller than this is set to zero
  /** Maximum ratio of stage change per time step (or error results)  */
  private static double MAXIMUM_CHANGE_RATIO     = 20;      // Water Depth cannot change by more than 20x
  /** Maximum absolute stage change per time step (or error results)  */
  private static double MAXIMUM_CHANGE           = .5;      // Water Depth cannot change by more than 50 cm
  
  /** Initial node in the sequence  */
  private Node initialNode;
  
/* Inner Classes */

  /** 
   * A Node is a point between segments.
   * All Nodes have at least one input canal, except
   * headgates
   * Nodes can have more than one output canal
   * Nodes are associated with 2 columns in the
   * continuity and momentum matrix: one for
   * dQ and one for dZ
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   * 
   */
  class Node {
    /** Canal segment flowing into this node */
    public CanalSegment in;
    /** Collection of canal segments flowing out from this node */
    public Vector       outCanals;
    /** Cell object that flows into this node */
    public Cell         inCell;
    /** Collection of nodes that are immediately downstream of this node */
    public Vector       outNodes;
    /** Column in the linear system matrix that represents the dQ value for this node */
    public int          dQColumn;
    /** Column in the linear system matrix that represents the dz value for this node */
    public int          dZColumn;
    /** Name of this node (primarily for debugging) */
    public String       name;
    
    /**
     * Constructor
     * @param oc Vector of outflow canals
     * @param cs inflow canal segment
     * @param newName name for this node
     */
    public Node (Vector oc, CanalSegment cs, String newName){
      outCanals = oc;
      in        = cs;
      name      = newName;
      outNodes  = new Vector();
    }
    
    /**
     * Adds a node to the collection of outflowing nodes
     * @param n node to be added
     */
    public void addOutNode(Node n){     outNodes.add(n);    }
    
    /**
     * Gets the cell that flows into this node
     * @return the node's inflow cell
     */
    public Cell    getInCell()   { return inCell;                                        }
    
    /**
     * Indicates whether this node is at a terminus
     * @return true if the node is at a terminus (has no outflow), false otherwise
     */
    public boolean isAtTerminus(){ return (outCanals == null);                           }

    /**
     * Indicates whether this node is at a junction
     * @return true if the node has more than one outflow, false otherwise
     */
    public boolean isJunction()  { return ((outCanals != null) && (outCanals.size()>1)); }
    
    /**
     * Gets the total of all outflows from this node
     * @return the sum of flow in all outflow canal segments from this node  
     */
    public double getTotalOutflow(){
      double retVal = 0;
      for(int i = 0; i > outCanals.size(); i++){
        CanalSegment cs = (CanalSegment)outCanals.get(i);
        retVal += cs.getUpstreamDischarge();
      }
      return retVal;
    }
  }
  
  /**
   * A cell' is roughly equivalent to a segment
   * but corresponds to 2 specific rows in the
   * continuity and momentum matrix, one for continuity
   * and one for momentum
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  class Cell{
    /** Name for this cell (primarily for debugging) */
    public String       name;
    /** Canal segment that corresponds to this cell */
    public CanalSegment cs;
    /** Node that lies at the upstream end of this cell */
    public Node         startNode;
    /** Node that lies at the downstream end of this cell */
    public Node         endNode;
    /** 
     * Row in the linear system matrix that corresponds to the values derived from this cell's characteristics
     * and the continuity equation
     */
    public int          conRow;
    /** 
     * Row in the linear system matrix that corresponds to the values derived from this cell's characteristics
     * and the momentum equation
     */
    public int          momRow;
    
    /**
     * Constructor
     * @param canSeg Canal Segment to which this cell corresponds
     * @param sn start node
     * @param en end node
     * @param cellName name for this cell
     */
    public Cell(CanalSegment canSeg, Node sn, Node en, String cellName){
      cs        = canSeg;
      startNode = sn;
      endNode   = en;
      name      = cellName;
    }
    
  }

  /**
   * A Junction represents a Node with more than
   * one outflow. (In HWM Nodes with multiple inflows
   * are prohibited.)
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  class Junction{
    /** Node that is immediately upstream of this junction */
    public Node inflowNode;
    /** Collection of nodes immediately downstream of this junction */
    public Vector outNodes;
    
    /**
     * Constructor
     * @param inNode node immediately upstream of this node
     * @param outNode collection of nodes immediately downstream of this one
     */
    public Junction(Node inNode, Vector outNode){
      inflowNode = inNode;
      outNodes = outNode;
    }
    
    /**
     * Gets the name of this junction (primarily for debugging)
     * @return a String representing this junction
     */
    public String getName(){
      return "Junction at: " + inflowNode.name;
    }
  }
  
  
  /* Static Methods */
  /**
   * Returnd true if the specified segment is an initial segment, false otherwise
   * @param cs the segment to be tested
   * @return true or false
   */
  public static boolean isInitialSegment(CanalSegment cs){
    if(cs.getInflowGate().getHeadGate() != null)  return true;
    if(cs.getInflowGate().isOpen()      == false) return true;
    return false;
  }
  
  /**
   * Given a collection of canal segments, returns a collection of HEC-RAS flow systems.
   * A single canal system can have several separate flow systems, depending on 
   * the locations of closed gates or empty segments.
   * @param canals the collection of canal segments to be analyzed
   * @return a collection of separate HEC-RAS flow systems
   */
  public static Vector getHECFlowSystems(Collection canals){
    Vector flowSystems = new Vector();
    Iterator c = canals.iterator();
    CanalSegment cs;
    while(c.hasNext()){
      cs = (CanalSegment)c.next();      
      if(isInitialSegment(cs)){
        HEC_RAS_FlowCalculator flowCalc = new HEC_RAS_FlowCalculator(cs);
        flowSystems.add(flowCalc);
        if(!cs.getInflowGate().isOpen()) 
            flowCalc.setUpstreamBoundaryCondition(new FlowCalculator.ClosedGateUpstreamBoundaryCondition(cs));
      }
      
    }
    return flowSystems;
  }
  
  /* Constructors And Initialization Methods */
  
  /**
   * Constructs a new FlowSystem object beginning with the
   * initial segment passed.
   * @param cs canal segment marking the beginning of the FlowSystem
   */
  public HEC_RAS_FlowCalculator(CanalSegment cs){
    super(cs);
    initialNode = createNetwork(cs);      
  }
  
  /**
   * Creates the network from the initial segment
   * passed.
   * @param cs Initial segment of the network
   * @return the initial node in the network
   */
  private Node createNetwork(CanalSegment cs){
    Node ret = null;
    ret = addNode(null, cs);
    
    // Adjust for junctions
    Enumeration E = nodeList.keys();
    while(E.hasMoreElements()){
      String K = (String)E.nextElement();
      Node N = (Node)nodeList.get(K);
      if(N.isJunction()){
        Cell C = (Cell)N.getInCell();                                  // Get the inflow cell
        Node inNode = C.startNode;                                     // Get the inflow node
        Vector outNodeVector = (Vector)N.outNodes;                     // Get the outflow cells
        Vector nodesVector = new Vector();                             // Get the outflow nodes
        for(int i = 0; i < outNodeVector.size(); i++){
          nodesVector.add(((Node)outNodeVector.elementAt(i)));
        }
        cellList.remove(C.name);                                  // Remove the inflow cell
        cellOrderList.remove(C.name);
        nodeList.remove(N.name);                                  // Remove the central node
        nodeOrderList.remove(N.name);
        // Create a new junction object
        Junction J = new Junction(inNode, nodesVector);
        // Add it to the list of junction objects
        junctionList.add(J);
      }          
    }
    
    
    // Update Node Numbers
    int counter;
    counter = 0;
    for(int i = 0; i < nodeOrderList.size(); i++){
      String K = (String)nodeOrderList.get(i);
      Node N = (Node)nodeList.get(K);
      N.dQColumn = counter++;
      N.dZColumn = counter++;
    }
    
    
    // Update Cell Row Numbers
    counter = 0;
    for(int i = 0; i < cellOrderList.size(); i++){
      String K = (String)cellOrderList.get(i);
      Cell C = (Cell)cellList.get(K);
      C.conRow = counter++;
      C.momRow = counter++;
    }
  
    return ret;
  }
      
  /**
   * Adds a node to the network.
   * This routine is passed a StartNode and a canalSegment. The StartNode will be null if this is the
   * first node being passed. In this case the routine will add the
   * node that is the START of the canal Seg. It will also add the canal segment
   * as a cell. If, conversely, the StartNode has a 
   * node but the canal Segment is null, it will simply add that node, assuming that
   * it is the end of the canal reach.
   * @param startNode initial node for this new node
   * @param canalSeg canal segment for this new node
   * @return the start node, which may or may not have been passed as an argument
   */
  private Node addNode(Node startNode, CanalSegment canalSeg){
    // If the Start Node is null, this is an initial segment. 
    // The action taken will be to add the start node, then
    // add the end of the canal as an end node, then add the segment as a cell.
    Node SNode;
    if(startNode == null){ // This is the initial node
      Vector outcanals = new Vector();
      outcanals.add(canalSeg);
      SNode = new Node(outcanals,                                     // Vector of outflow canals 
                       null,                                          // Inflow canal segment
                       "STARTOF: " + canalSeg.getCanalSegmentID());   // Name
      nodeList.put(SNode.name, SNode);
      nodeOrderList.add(SNode.name);
    }
    // If there is a Start Node, it is simply recorded, as it must have been added
    // earlier. The exception to this is at a junction; the outflowing canals
    // must have their own start nodes.
    else{
      SNode = startNode; 
      // If this is a junction, treat it differently
      if(canalSeg.getInflowCanal().getOutflowCanals().size() > 1){
        Vector outcanals = canalSeg.getOutflowCanals();
        SNode = new Node(outcanals,                                     // Vector of outflow canals 
                         null,                                          // Inflow canal segment
                         "STARTOF: " + canalSeg.getCanalSegmentID());   // Name
        startNode.addOutNode(SNode);                  
        nodeList.put(SNode.name, SNode);
        nodeOrderList.add(SNode.name);
      }
      else{
        // Node is not a junction- do nothing, for now
      }
    }
    // If there is a canal segment:
    // Add the node corresponding to the end of the canal segment as a new node
    // Add the canal segment as a cell
    // Call the routine for all the canal's outflow canals with the end node as
    // the new start node
    if(canalSeg != null){
      // Create the node for the end node
      // Note: if there are no outflow canals, this node is a terminus
      // If there are more than one outflow canals, this node is a junction
      Vector outflowCanalVector = canalSeg.getOutflowCanals();
      if(outflowCanalVector.size() == 0) outflowCanalVector = null;
      
      // Create the node that will be the end node
      Node ENode = new Node(outflowCanalVector,                         // Vector of outflow canals 
                            canalSeg,                                   // Inflow canal segment
                            "ENDOF: " + canalSeg.getCanalSegmentID());  // Name
      
      // Add this segment as a cell
      Cell newCell = new Cell(canalSeg,                                 // Canal Segment Associated 
                              SNode,                                    // Start Node
                              ENode,                                    // End Node
                              "CELL:" + canalSeg.getCanalSegmentID());  // Name
      cellList.put(newCell.name, newCell);
      cellOrderList.add(newCell.name);
      ENode.inCell = newCell;
      nodeList.put(ENode.name, ENode);
      nodeOrderList.add(ENode.name);
      
      Vector cOut = canalSeg.getOpenOutflowCanals();
      if(cOut.size() > 0){
        for(int i = 0; i < cOut.size(); i++){
          addNode(ENode, (CanalSegment)cOut.get(i));
        }
      }
      // Else- this is a terminus; put in the terminal node list
      else{
        terminalNodeList.put(ENode.name, ENode);
      }
    }
    else{
      terminalNodeList.put(startNode.name, startNode);
    }
    return SNode;
  }
 
  
  /* Calculations */
  
  /**
   * Solves for the maximum timestep through which this flow system
   * should be put. There are several factors that influence this:
   * 
   * - If water flow level grows too small, the matrix may be unstable;
   *      a solution to this is to reduce the time step
   * - Water may need to fill either an upstream or downstream canal
   *      segment that does not already have water in it.
   * @param initialDurationInSeconds initial value to use for duration of timestep, in seconds
   * @return maximum time step through which this flow system should be put
   */
  public double solveForMaxTimestep(double initialDurationInSeconds){
    if(canalSystemIsEmpty()) return initialDurationInSeconds;
    double t = initialDurationInSeconds;
    while(t > MINIMUM_TIMESTEP){
      if(calculate(t)) break;
      else             t = t / 2d;
    }
    if(t < MINIMUM_TIMESTEP){
      t = MINIMUM_TIMESTEP;
      System.out.println("     TIME FELL BELOW MINIMUM in solveForMaxTimeStep()");
    }
    return t;
  } 
  

  /**
   * @see flowag.channels.AbstractFlowCalculator#performCalculations(double)
   */
  public boolean performCalculations(double duration){
    if(canalSystemIsEmpty()) return true;
    System.out.println("Performing calculations: duration = " + duration);
    double theta = 1d;
    while(theta > 0.5){
      System.out.println("  THETA = " + theta);
      if(initAndSolveLSS(duration, theta) && solutionIsStable()) return true;
      theta -= .1;
    }
    System.out.println("Theta Grounded with no stable solution");
    return false;
  }  
  
  
  /**
   * Indicates if the solution to the linear system is a set of stable changes
   * to the canal system. Stable changes will keep the dz and dQ values within
   * ranges or within acceptable percent changes.
   * @return true if the solution is stable, false if not
   */
  public boolean solutionIsStable(){
    // Loop through the nodes and compare with the dz values;
    // Determine if they are within specs
    // Confirm that the values will not push stages out of bounds...
    Enumeration c = cellList.elements();
    boolean successFlag = true;
    while(successFlag == true && c.hasMoreElements()){
      Cell C = (Cell)c.nextElement();
      double usz  = C.cs.getUpstreamStage();                   // Previous Upstream Stage
      double dsz  = C.cs.getDownstreamStage();                 // Previous Downstream Stage
      double usd  = C.cs.getUpstreamWaterDepth();              // Previous Upstream Depth
      double dsd  = C.cs.getDownstreamWaterDepth();            // Previous Downstream Depth
      double usDz = LSS.getSolvedValue(C.startNode.dZColumn);  // Calculated Upstream Dz
      double dsDz = LSS.getSolvedValue(C.endNode.dZColumn);    // Calculated Downstream Dz
                
      // Can ignore this if the calculated change is zero
      // Upstream
      if(usDz != 0){
        // There's a rounding issue; in some cases the depth should go to exactly zero.
        if(Math.abs(usd + usDz) < MINIMUM_CALCULATED_DEPTH) usDz = -1 * usd;
        // If setting any of these would make the water depth (not stage) go below zero,
        // this is an instability and must be flagged.
        if((usd + usDz) < 0) {
          System.out.println("UPSTREAM DEPTH GOING BELOW ZERO! orig depth = " + usd + " delta " + usDz + " " + C.cs.getCanalSegmentID());
          //System.out.println("  @CANAL SEG: " + C.cs.getCanalSegmentID());
          //System.out.println("Old Val: " + usd + " dZ " + usDz);
          successFlag = false;
        }
                  
        // If the ratio of the change (derivative) to the current value
        // is too large, flag this. Note that this should be
        // relative except in the case of a zero initial depth, in which
        // case it should be absolute
        if(usd != 0){  // Technically if(usd > 0) should suffice... 
          if((usDz > 1.5) && (Math.abs(usDz / usd) > MAXIMUM_CHANGE_RATIO) ){ // Only check this if the change is more than 1.5 meters...?
            System.out.print("UPSTREAM MAXIMUM CHANGE RATIO BEING EXCEEDED! ");
            System.out.print("usd = " + usd + " usDz " + usDz);
            System.out.println("  @CANAL SEG: " + C.cs.getCanalSegmentID());
            //System.out.println("Old Val: " + usd + " dZ " + usDz + " ratio1: " + (usDz / usd) + " ratio2: " + (usd/usDz));
            successFlag = false;
          }
        }
        else{
          if(Math.abs(usDz) > MAXIMUM_CHANGE){
            //System.out.println("Upstream depth change more than maximum allowed");
            successFlag = false;
          }
        }                
      }
               
      // Downstream
      if(dsDz != 0){
        // See comments in parallel structure above
        if(Math.abs(dsd + dsDz) < MINIMUM_CALCULATED_DEPTH) dsDz = -1 * dsd;
        if((dsd + dsDz) < 0) {
          System.out.println("DOWNSTREAM DEPTH GOING BELOW ZERO!");
          //System.out.println("  @CANAL SEG: " + C.cs.getCanalSegmentID());
          //System.out.println("Old Val: " + dsd + " dZ " + dsDz);
          successFlag = false;
        }
        if(dsd != 0){  // Technically if(dsd > 0) should suffice...
          // System.out.println("CANAL SEG: " + C.cs.getCanalSegmentID());
          // System.out.println("Old Val: " + dsd + " dZ " + dsDz + " ratio1: " + (dsDz / dsd) + " ratio2: " + (dsd/dsDz));
          if(Math.abs(dsDz / dsd) > MAXIMUM_CHANGE_RATIO){
            System.out.print("DOWNSTREAM MAXIMUM CHANGE RATIO BEING EXCEEDED! ");
            System.out.print("dsd = " + dsd + " dsDz " + dsDz);
            System.out.println("  @ CANAL SEG: " + C.cs.getCanalSegmentID());
            //System.out.println("Old Val: " + dsd + " dZ " + dsDz + " ratio1: " + (dsDz / dsd) + " ratio2: " + (dsd/dsDz));
            successFlag = false;
          }
        }
        else{
          if(Math.abs(dsDz) > MAXIMUM_CHANGE){
            // System.out.println("Downstream depth change more than maximum allowed");
            successFlag = false;
          }
        }                
      }             
    }
    // Otherwise, all is well- return 'true'
    return successFlag;

  }
  
  /**
   * Determines if the canal system is empty or not
   * @return true if the canal system is empty (consists entirely of segments below minimum depth),
   * false otherwise
   */
  private boolean canalSystemIsEmpty(){
    Enumeration E = cellList.elements();
    while(E.hasMoreElements()){
      CanalSegment cs = ((Cell)E.nextElement()).cs;
      double MINIMUM_DEPTH = .005d;
      if(cs.getUpstreamWaterDepth() >= MINIMUM_DEPTH || cs.getDownstreamWaterDepth() >= MINIMUM_DEPTH) return false;
    }
    return true;
  }
  
  
  
  /* Matrix Construction and Testing*/
 
  /**
   * Initializes and solves the linear system that forms the basis of the HEC-RAS algorithms
   * @param duration interval over which the calculations are to take place
   * @param theta theta value to be used for these calculations; theta is a parameter
   * that can be adjusted to take more or less conservative estimates of changes
   * @return true if the linear system is solved, false if it is not
   */
  private boolean initAndSolveLSS(double duration, double theta){
    if((duration == lastDurationForLSS) && (theta == lastThetaForLSS)) return lastSuccessOfLSS;
    lastDurationForLSS = duration;
    lastThetaForLSS    = theta;
    buildLSS(duration, theta);
    lastSuccessOfLSS = LSS.solve();
    return lastSuccessOfLSS;
  }
  
  /**
   * Build the linear system that represents the flow system.
   * @param duration the proposed time step
   * @param theta the proposed theta value
   */
  private void buildLSS(double duration, double theta){
    LSS = new LinearSystemSolver(2 * cellList.size(), 2 * nodeList.size());

    double g       = Global.AccelerationDueToGravity;
    
    /* Basic matrix- excludes junctions and termini */
    // Create a row for each of the cells in the cellList
    for(int cc = 0; cc < cellOrderList.size(); cc++){
      Cell c = (Cell)cellList.get((String)cellOrderList.get(cc));
      // First get values that are needed for the 
      // calculations. Get these only once because it's faster
      CanalSegment cs = c.cs;
      double x       = cs.getLength();       
      double z1      = cs.getUpstreamStage();
      double Q1      = cs.getUpstreamDischarge();
      double V1      = cs.getUpstreamFlowVelocity();      
      double K1      = cs.getUpstreamConveyance();
      double dKdz1   = cs.getUpstreamDKDz();
      double A1      = cs.getUpstreamFlowArea();
      double dAdz1   = cs.getUpstreamDADz();            
      double Sf1     = cs.getUpstreamFrictionSlope();
            
      if(!cs.getInflowCanal().equals(cs) &&
          cs.getInflowCanal().getOutflowCanals().size() > 1){
        Sf1 = cs.getDownstreamFrictionSlope();
      }
            
      double z2      = cs.getDownstreamStage();
      double Q2      = cs.getDownstreamDischarge();
      double V2      = cs.getDownstreamFlowVelocity();
      double K2      = cs.getDownstreamConveyance();
      double dKdz2   = cs.getDownstreamDKDz();
      double A2      = cs.getDownstreamFlowArea();
      double dAdz2   = cs.getDownstreamDADz();
      double Sf2     = cs.getDownstreamFrictionSlope();
      double Es      = cs.getUpstreamSpecificEnergy();
      double Es2     = cs.getDownstreamSpecificEnergy();
      double barA    = cs.getAverageFlowArea();
      double barSf   = (Sf1 + Sf2)/2.0d;
 
      double t = duration;

      // Continuity Row:
      
      // CQ1
      double CQ1 = (-1d) * theta / x;
      LSS.setElement(c.conRow, c.startNode.dQColumn, CQ1);

      // CZ1
      double CZ1 = ((.5d) / t) * dAdz1;
      LSS.setElement(c.conRow, c.startNode.dZColumn, CZ1);
        
      // CQ2
      double CQ2 = theta / x;
      LSS.setElement(c.conRow, c.endNode.dQColumn, CQ2);
        
      // CZ2
      double CZ2 = ((.5d) / t) * dAdz2;
      LSS.setElement(c.conRow, c.endNode.dZColumn, CZ2);
        
      // CB
      double CB = -1 * ((Q2 - Q1) / x);
      LSS.setNHT(c.conRow, CB);                // Lateral outflow?
        
        
      // Momentum Row:
      
      //MQ1     
      double MQ1_1 = (0.5d / t);
      double MQ1_2 = V1 * theta / x;
      double MQ1_3 = (Q1 != 0 ? theta * g * barA * Sf1 / Q1 : 0);
      double MQ1 = MQ1_1 - MQ1_2 + MQ1_3;
      LSS.setElement(c.momRow, c.startNode.dQColumn, MQ1);
      
      //MZ1
      double MZ1_1 = ((-1) *  g * barA * theta ) / x;
      double MZ1_2 = 0.5d * g * (z2 - z1) * dAdz1 * ( theta / x);
      double MZ1_3 = (K1 != 0 ? g * theta * barA * dKdz1 * (Sf1 / K1) : 0);
      double MZ1_4 = 0.5d * theta * g * dAdz1 * barSf;
      double MZ1 = MZ1_1 + MZ1_2 - MZ1_3 + MZ1_4;
      LSS.setElement(c.momRow, c.startNode.dZColumn, MZ1);
       
      //MQ2
      double MQ2_1 = (0.5d / t);
      double MQ2_2 = (V2 / x);
      double MQ2_3 = (Q2 != 0 ? theta * g * barA * ( Sf2 / Q2 ) : 0);
      double MQ2 = MQ2_1 + MQ2_2 + MQ2_3;
      LSS.setElement(c.momRow, c.endNode.dQColumn, MQ2);
      
      //MZ2
      double MZ2_1 = (g * barA * theta ) / x;
      double MZ2_2 = 0.5d * g * (z2 - z1) * dAdz2 * ( theta / x);
      double MZ2_3 = (K2 != 0 ? g * theta * barA * dKdz2 * (Sf2 / K2) : 0);
      double MZ2_4 = 0.5d * theta * g * dAdz2 * barSf;
      double MZ2 = MZ2_1 + MZ2_2 - MZ2_3 + MZ2_4;
      LSS.setElement(c.momRow, c.endNode.dZColumn, MZ2);
        
      //MB
      double MB_1 = ((V2 * Q2 - V1 * Q1) / x);
      double MB_2 = ((g * barA * (z2 - z1)) / x);
      double MB_3 = (g * barA * barSf); 
      double MB = (-1) * (MB_1 + MB_2 + MB_3);
      MB = BasicMath.roundDouble(MB, 6);
      LSS.setNHT(c.momRow, MB);
      
    }
    
    /* Junctions */
    // Now create special rows for each of the junctions
    Vector JunctionList = junctionList;
    for(int c = 0; c < JunctionList.size(); c++){
      System.out.println("REVISING FOR A JUNCTION...");
      Junction J = (Junction)JunctionList.get(c);
      // Add a Flow Continuity equation relating the inflow nodes
      // to the outflow nodes:
      int newEq = LSS.addEquation();
      LSS.setElement(newEq, J.inflowNode.dQColumn, (-1d * theta));
      Vector V = J.outNodes;
      for(int q = 0; q < V.size(); q++){
        LSS.setElement(newEq, ((Node)V.get(q)).dQColumn, theta);
      }
      LSS.setNHT(newEq, J.inflowNode.getTotalOutflow());
      // Now add a new Stage Continuity equation for each
      // of the outflow nodes
      for(int q = 0; q < V.size(); q++){
        newEq = LSS.addEquation();
        LSS.setElement(newEq, J.inflowNode.dZColumn,1);
        LSS.setElement(newEq, ((Node)V.get(q)).dZColumn, -1);
      }
    }

    /* Upstream boundary condition */
//    // Upstream boundary conditions correspond to the kind of upstream
//    // segment chose; some are at true headgates, others are
//    // interruptions in flow.
//    
//    // Find all the headgates and determine flow change rate
//    // from river
//    Enumeration nodeEnum;
//    nodeEnum = nodeList.elements();
//    while(nodeEnum.hasMoreElements()){
//      Node n = (Node)nodeEnum.nextElement();
//      if(n.isAtHeadgate()) LSS.defineKnownValue(n.dQColumn, initialDeltaQ);
//    }
    
    // Set the upstream boundary condition
    // In reality there can be four- a newly opened gate with an empty
    // canal behind it, 'manning' flow, a closed gate, or a headgate
    if(ubc != null && ubc.specifiesDQ()) LSS.defineKnownValue(initialNode.dQColumn, ubc.getDQ(duration));
    if(ubc != null && ubc.specifiesDZ()) LSS.defineKnownValue(initialNode.dZColumn, ubc.getDZ(duration));
    
    // if it's an open gate... I don't know what to do!
    // if(ubc.isEmptySegmentUpstream())
        
    
    /* Downstream boundary conditions */

    // There are two kinds of terminal nodes, with 2 associated boundary conditions:
    //    1- End of defined canal system ( = Normal Depth)
    //    2- Closed gates                ( = Q is always zero; dQ = -Q)
    
    // When using Normal Depth as a boundary condition,
    // we draw from HEC-RAS p. 2-47, 48.
    // A Single-valued rating curve is implemented.
    // This is implemented by adding another row
    // to the continuity/momentum matrix and providing
    // values for the dQ and dZ columns.
    // In the code below, the equation used follows the
    // single-value rating curve method, but uses the
    // Manning equation to provide the values of the
    // curve, by making use of the built-in methods of the
    // Canal Segment object to calculate discharge based on
    // stage given the Manning equation.
    Enumeration nodeEnum = terminalNodeList.elements();
    while(nodeEnum.hasMoreElements()){
      Node n = (Node)nodeEnum.nextElement();
      int dQCol = n.dQColumn;
      int dZCol = n.dZColumn;
      CanalSegment cs = n.in;
         
      if(!cs.isEndOfDefinedSystem() || cs.outflowGateIsOpen()){
        double Zn = cs.getDownstreamStage();
        double Qn = cs.getDownstreamDischarge();
        double INTERVAL = 0.01;
        double Sk = Zn;
        double Sk_1 = Sk + INTERVAL; // Assume a test depth change of 1 cm
        // Calculate discharge based on stage using the Manning equation
        double Dk = cs.getDownstreamManningDischarge(Sk);
        double Dk_1 = cs.getDownstreamManningDischarge(Sk_1);
        double CDQm = theta;
        double CDZm = - (Dk - Dk_1)/(Sk - Sk_1);              
        double CDBm = Dk_1 + ((Dk - Dk_1)/(Sk - Sk_1)) * (Zn - Sk_1) - Qn;
        int rowNumber = LSS.addEquation();
        LSS.setElement(rowNumber, dQCol, CDQm);
        LSS.setElement(rowNumber, dZCol, CDZm);
        LSS.setNHT(rowNumber, CDBm);  
      }
      // Otherwise, it is the end of the defined system
      else {
        // Set the dQ to -1 * Q; this will force the
        // Q to zero in the next timestep
        System.out.println("Setting dQ for end of system...");
        double Qn = cs.getDownstreamDischarge();
        int rowNumber = LSS.addEquation();
        LSS.setElement(rowNumber, dQCol, 1);
        LSS.setNHT(rowNumber, -1d * Qn);
      }
    }
    
  }
  
  
  
  /**
   * @see flowag.channels.AbstractFlowCalculator#setResults()
   */
  protected void setResults(){
    double duration = 0;
    results = new FlowCalculator.FlowZAndQAdjustments();
    
    if(canalSystemIsEmpty()){
      // Pre-treatment
      // If the canal system is empty, get the upstream boundary
      // condition, calculate a depth, and set everything to that
      // depth and appropriate manning discharge. This is conjuring
      // water from nowhere, but it might work...
      if(this.ubc.specifiesDQ()){
        if(ubc.getDQ(duration) == 0){ 
          // Do nothing if the canal is empty and there is no flow... it should just sit there...
        }
        else{
          double newQ = ubc.getDQ(duration);
          double newDepth = 
            initialCanalSegment.upstreamCrossSection.getDepthPerManningDischarge(newQ, initialCanalSegment.roughness, initialCanalSegment.getBedSlope());
//          System.out.println("SETTING TO NEW DEPTH: " + newDepth + " and FLOW " + newQ);
          results.setInitialUpstreamValues(initialCanalSegment, newDepth, newQ);
          Enumeration c = cellList.elements();
          while(c.hasMoreElements()){
            Cell C = (Cell)c.nextElement();
            results.addDownstreamValues(C.cs, newDepth - C.cs.getDownstreamWaterDepth(), 
                C.cs.getDownstreamManningDischarge(C.cs.getEndCoords().getZ() + newDepth));
          }
        }
      }      
    }
    else{
      Enumeration c = cellList.elements();
      while(c.hasMoreElements()){
        Cell C = (Cell)c.nextElement();
        if(C.cs.equals(initialCanalSegment)) results.setInitialUpstreamValues(C.cs, LSS.getSolvedValue(C.startNode.dZColumn), LSS.getSolvedValue(C.startNode.dQColumn));
        results.addDownstreamValues(C.cs, LSS.getSolvedValue(C.endNode.dZColumn), LSS.getSolvedValue(C.endNode.dQColumn));
      }
    }
  }
  

  
  /* Reporting */
  
  /**
   * List the elements of the system.
   */
  public void listSystem(){
    System.out.println("Node List:");
    Enumeration ns = nodeList.elements();
    for(int i = 0; i < nodeOrderList.size(); i++){
      String K = (String)nodeOrderList.get(i);
      Node N = (Node)nodeList.get(K);
      System.out.println("  " + N.name);      
    }
    System.out.println("Cell List:");
    Enumeration cs = cellList.elements();
    for(int i = 0; i < cellOrderList.size(); i++){
      String K = (String)cellOrderList.get(i);
      Cell c = (Cell)cellList.get(K);
      System.out.println("  " + c.name);      
    }
    System.out.println("Junction List:");
    Enumeration js = junctionList.elements();
    while(js.hasMoreElements()){
      Junction J = (Junction)js.nextElement();
      System.out.println("  " + J.getName());      
    }
  }
 


  
  
  
}
