package flowag.channels;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.math.Point;
import abcm.sim.AbstractModelComponent;
import flowag.basic.UnitConversion;
import flowag.basic.Water;

/**
 * A HeadGate is an interface between a river and a canal system.
 * 
 * The river has a flowRate which is calculated in million acre-feet per
 * year, then scaled to a rate per day; a seasonal narrative is used for this
 * scaling so that the annual flow can be parceled out over the course
 * of a year.
 * 
 * The per day flow rate is a straight number of acre-feet. This can
 * be recalculated into m^3/s, the units in which the canal system
 * works.
 * 
 * This approach is still limited; it does not permit the canal system to 'fill
 * up' (at which point the head of the river is insufficient to move
 * water into the system), but we will deal with this through other means.<br/><br/>
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class HeadGate extends AbstractModelComponent implements ABCMDeepCloneable{

  
  /** Switch for code model to calculate efficiency */
  public static final int STANDARD_HEADGATE_EFFICIENCY     = 0;
  /** Switch for code model to calculate efficiency */
  public static final int INTERVAL_HEADGATE_EFFICIENCY     = 1;
  /** Switch for code model to calculate efficiency */
  public static final int INTERPOLATED_HEADGATE_EFFICIENCY = 2;

  /**
   * A utility class to store a range of efficiencies (lower and upper bound).
   * Ranges are used in a collection to calculate efficiences; in general, the range object
   * with the lowest value for upper bound greater than the flow through the headgate
   * provides the efficiency value, but other varianted (interpolation) are possible.
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
  */
  private class EfficiencyRange{
    /** The high value for flow boundary */
    public double upperBound;
    /** The efficiency value for the range defined by this boundary */
    public double efficiency;
    /** 
     * Constructor that sets the boundary and efficiency
     * @param bound the upper boundary
     * @param eff the efficiency value 
     */
    public EfficiencyRange(double bound, double eff){
      upperBound = bound;
      efficiency = eff;
    }
    /** 
     * Constructor that sets the boundary and efficiency from String values
     * @param bound the upper boundary
     * @param eff the efficiency value   
     */
    public EfficiencyRange(String bound, String eff){
      try{
        upperBound = Double.parseDouble(bound);
        efficiency = Double.parseDouble(eff);
      }
      catch(NumberFormatException nfe){
        upperBound = -1;
        efficiency = -1;
      }
    }
  }
  
  /** River to which this HeadGate connects */ 
  private River river;                  
  /** Gate into which this HeadGate flows */ 
  private Gate outflow;                 
  /** Location of this HeadGate */ 
  private Point headCoords;            
  /** Measure of the ability of this HeadGate to take water from river */ 
  private double efficiency;
  
  /**
   * Collection of efficiency ranges; only the upper bound of each range is given,
   * so the collection is ordered from lowest to highest.
   */
  private EfficiencyRange[] efficiencyRanges;
  
  /** True if this gate is damaged */ 
  private boolean broken;              
  
  /** Flag indicating the status of the headgate as opened or closed  */
  private boolean closed = false;
  
  
  /** Previous day's discharge value */ 
  private double previousDischarge;     
  
  /** Default efficiency value */ 
  public static final double DEFAULT_EFFICIENCY = 0.3;
  /** Default labor (man-days) required to repair headgate */ 
  public static final double DEFAULT_LABOR_TO_REPAIR_HEADGATE = 100;

  
  /* Constructors */

  /**
   * Constructor
   * @param headCoordinates coordinates at which HeadGate is to be located
   */
  public HeadGate(Point headCoordinates){
    this(null, headCoordinates, DEFAULT_EFFICIENCY);
  }

  
  /**
   * Constructor
   * @param gate  The gate into which the headgate directs its flow
   * @param headCoordinates point marking the origins of this headgate
   * @param eff   efficiency- a variable that allows the simulation
   * to calculate the amount of water that should flow into this headgate
   */
  public HeadGate(Gate gate, Point headCoordinates, double eff){
    outflow = gate;
    efficiency = eff;
    headCoords = headCoordinates;
//    System.out.println("Constructed headgate at " + headCoords.coordString() + " with assigned efficiency " + eff);
  }
  
  
  /* Basic Setters */
  /**
   * Sets the River for this HeadGate
   * @param r the River object to which this HeadGate is attached
   */
  public void setRiver(River r){             river = r;           }

  /**
   * Sets a new outflow gate for this head gate.
   * @param gate the new outflow Gate
   */
  public void setOutflow(Gate gate) {        outflow = gate;      }
   
  /**
   * Sets the efficiency of this head gate.
   * @param eff New value for the efficiency of this head gate.
   */
  public void setEfficiency(double eff) {    efficiency = eff;    }

  
  /**
   * Sets the efficiency ranges from a single String
   * @param rangeString a String representing ranges of efficiency;
   * must be in the form: "upperBound1:efficiency1;upperBound2:efficiency2;"
   */
  public void setEfficiencyRanges(String rangeString){
    if(rangeString == null || rangeString.length() == 0) return;
    String[] vals = rangeString.split(";");
    efficiencyRanges = new EfficiencyRange[vals.length];
    for(int i = 0; i < efficiencyRanges.length; i++){
      String[] eff = vals[i].split(":");
      efficiencyRanges[i] = new EfficiencyRange(eff[0], eff[1]);
    }
    
  }
  
  /* Basic Getters */
  /**
   * Gets the location of this HeadGate
   * @return a Point object representing the location of this HeadGate
   */
  public Point getLocus(){    return headCoords;  }
  
  /**
   * Gets the River object for this HeadGate
   * @return the River object to which this HeadGate is attached
   */
  public River getRiver(){    return river;       }

  /**
   * Gets the outflow gate to which this HeadGate directs its flow
   * @return The gate into which water flows from this head gate
   */
  public Gate getOutflow(){   return outflow;     }
  
  /**
   * Gets the current efficiency of this head gate
   * @return the current efficiency of this head gate
   */
  public double getEfficiency() {   
    return (closed ? 0 : efficiency);    
  }

  /**
   * Gets the current efficiency of this headgate as a function of current flow
   * @param flowLevelInMAFPerDay flow level in million-acre-feet per day
   * @return the current efficiency of this head gate
   */
  public double getEfficiency(double flowLevelInMAFPerDay){
    if(closed) return 0;
	  switch(codeModelOption("HEADGATE_EFFICIENCY")){
  	  case INTERVAL_HEADGATE_EFFICIENCY:
  	    if(efficiencyRanges == null || efficiencyRanges.length == 0)  return getEfficiency();
  		  for(int i = 0; i < efficiencyRanges.length; i++){
  		    if(flowLevelInMAFPerDay < efficiencyRanges[i].upperBound){
  		      return efficiencyRanges[i].efficiency; 
  		    }
  		  }
  		  return efficiencyRanges[efficiencyRanges.length - 1].efficiency; // If it is above the last range, return the last efficiency
  	  case INTERPOLATED_HEADGATE_EFFICIENCY:
        if(efficiencyRanges == null || efficiencyRanges.length == 0) return getEfficiency(); // This allows gates of both kinds...
        double lastEff = 0;
        double lastBound = 0;
        for(int i = 0; i < efficiencyRanges.length; i++){
          if(flowLevelInMAFPerDay < efficiencyRanges[i].upperBound){
            return lastEff + 
               (flowLevelInMAFPerDay - lastBound) * 
                   (efficiencyRanges[i].efficiency - lastEff) / (efficiencyRanges[i].upperBound - lastBound);
          }
          lastEff = efficiencyRanges[i].efficiency;
          lastBound = efficiencyRanges[i].upperBound;
        }
        return efficiencyRanges[efficiencyRanges.length - 1].efficiency; // If it is above the range, return the last value
  	  case STANDARD_HEADGATE_EFFICIENCY:
  	  default:
          return getEfficiency();
	  }
  }
  
  
  /**
   * Closes this headgate
   */
  public void close(){ closed = true;  }
  /**
   * Opens this headgate 
   */
  public void open(){  closed = false; }
  
  /**
   * Gets the repair status of this gate
   * @return true if the gate is broken
   */
  public boolean isBroken() {       return broken;  }
  
    

  /* Flow */
  /**
   * Gets the current flow
   * @return the current flow
   */
  public double getFlow(){
    return ((broken || closed) ? 0 : 
      UnitConversion.mafPerDayToCubicMetersPerSecond(
          river.getDailyMAF(this)) * getEfficiency());
  }

  /**
   * Gets the flow through this headgate at a specified time of day.
   * Time is given as a fraction of the day (i.e. 0.5 = noon). May be used
   * to interpolate between old and new flow values
   * @param fractionalTimeOfDay time of day for which flow is to be calculated
   * @return the flow value
   */
  public double getFlowAt(double fractionalTimeOfDay){
    double flow = river.getFlowAtHeadgateAtTime(this, fractionalTimeOfDay);
    return UnitConversion.mafPerDayToCubicMetersPerSecond(flow * this.getEfficiency(flow));
  }
  
  /**
   * Gets the change in flow level for this headgate, beginning
   * from the specified time (as a fraction of a day) and continuing
   * for the specified interval (in seconds).
   * @param fractionalTimeOfDay start time
   * @param durationInSeconds duration
   * @return change in flow over the interval specified, in cubic meters per second
   */
  public double getFlowChange(double fractionalTimeOfDay, double durationInSeconds){
    return getFlowAt(UnitConversion.convertTime(fractionalTimeOfDay, durationInSeconds)) - getFlowAt(fractionalTimeOfDay);
  }
  
  
  /* Water */
  
  /**
   * Gets a sample of water at this HeadGate (from River)
   * @return a water sample
   */
  public Water getWaterSample(){    return river.getWaterSample();    }
  
 
  
  
  
  /**
   * Gets the labor required to repair this HeadGate
   * @return the labor required to repair this HeadGate
   */
  public double getLaborToRepair(){ return DEFAULT_LABOR_TO_REPAIR_HEADGATE; }
  
  /**
   * Repairs this headgate. Requires an amount of labor to be applied
   * to the repair; if this is insufficient, the repair is not done, the value
   * returned is zero.
   * @param labor the amount of labor to be applied to the repair
   * @return the actual amount of labor expended to effect the repair
   */
  public double repair(double labor){
    if(broken == false) return 0;
    double laborNeeded = getLaborToRepair();
    if(labor >= laborNeeded){
      broken = false;
      return laborNeeded;
    }
    else return 0; // Not enough labor was allocated; this is all-or-nothing,
                   // so the result is zero
  }

  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    HeadGate ret = new HeadGate((Point)cloneStore.getClone(headCoords)); // Points are immutable
    cloneStore.put(this, ret);
    ret.broken = this.broken;
    ret.efficiency = this.efficiency;
    ret.outflow = (Gate)cloneStore.getClone(outflow);
    return ret;
  }
  
  /* Obsolete */
//  public void damage(){ broken = true; }
  
  // Instance variables:
  //private double flowRate;             // Rate of water flow in m^3/s

  // Methods
//  
//  /**
//   * Gets a water object that represents the water that flows through this
//   * headgate given the flow rate for the specified time interval.
//   * @param timeInterval The amount of time being tracked
//   * @param sample A template of water- the water object returned will
//   * have the same chemical concentrations as this sample.
//   * @return A new water object representing the water that has flowed through
//   * this gate during the time interval specified.
//   */
//  public Water getFlowGivenTime(double timeInterval, Water sample){
//    return new Water(flowRate * timeInterval, sample);
//  }
//  
//  public double getFlowRateChangeOverInterval(double interval){
//    return river.getFlowChangeOverInterval(interval);
//  }
//  
//
//  /**
//   * Sets the flow rate for this gate
//   * @param frate new value for the flow rate for this head gate.
//   */
//  public void setFlowRate(double frate) {    flowRate = frate;    }
//   
//  /**
//   * Gets the current flow rate of this head gate. This will be in m^3/s.
//   * @return The current flow rate of this head gate, in m^3/s
//   */
//  public double getFlowRate() {     return flowRate;      }
//
  
  
}
