package flowag.channels;

import abcm.math.SimDate;

/**
 * Hydrographable objects provide methods that can be used
 * to keep a record of stage and flow at times even within a
 * single simulation date.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public interface Hydrographable {

  /**
   * Clears the entire hydrograph record for this object
   */
  public void clearHydrographs();
  
  /**
   * Get the upstream stage as recorded at the specified
   * simulation date. If the time falls between two recorded
   * times, should interpolate between the two recorded points.
   * @param SD the specified simulation date (and time)
   * @return the stage
   */
  public double getUpstreamStageAt(SimDate SD);
  
  /**
   * Get the downstream stage as recorded at the specified
   * simulation date. If the time falls between two recorded
   * times, should interpolate between the two recorded points.
   * @param SD the specified simulation date (and time)
   * @return the stage
   */
  public double getDownstreamStageAt(SimDate SD);
  
  /**
   * Get the upstream flow as recorded at the specified
   * simulation date. If the time falls between two recorded
   * times, should interpolate between the two recorded points.
   * @param SD the specified simulation date (and time)
   * @return the flow
   */
  public double getUpstreamFlowAt(SimDate SD);
  
  /**
   * Get the downstream flow as recorded at the specified
   * simulation date. If the time falls between two recorded
   * times, should interpolate between the two recorded points.
   * @param SD the specified simulation date (and time)
   * @return the flow
   */
  public double getDownstreamFlowAt(SimDate SD);

  /**
   * Reset all hydrographs in this object.
   */
  public void resetHydrographs();
  
  /**
   * Records the given stage as the upstream stage at
   * the specified time.
   * @param z the stage to be recorded
   * @param SD the time stamp to be assigned
   */
  public void setUpstreamStage(double z, SimDate SD);
  
  /**
   * Records the downstream stage at the specified
   * time
   * @param z stage to be recorded
   * @param SD the time stamp to be assigned
   */
  public void setDownstreamStage(double z, SimDate SD);
  
  /**
   * Records the given discharge as the upstream discharge
   * at the specified time
   * @param Q the discharge to be recorded
   * @param SD the time stamp to be assigned
   */
  public void setUpstreamDischarge(double Q, SimDate SD);
  
  /**
   * Records the given discharge as the downstream discharge
   * at the specified time
   * @param Q the discharge to be recorded
   * @param SD the time stamp to be assigned
   */
  public void setDownstreamDischarge(double Q, SimDate SD);
  
}
