package flowag.channels;

import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableVector;
import abcm.math.BasicMath;
import abcm.math.Geometry;
import abcm.math.LineSegment;
import abcm.math.Point;

/**
 * A channel cross section with an irregular profile. The profile
 * is specified as a list of points from one back to the other, 
 * such that each point represents a distance from the first bank
 * and a z coordinate representing the bottom of the canal at that point.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class IrregularChannelCrossSection extends AbstractChannelCrossSection {
  /** Collection of points that define this cross-section  */
  private ABCMDeepCloneableVector points;
  /** Point that represents the midpoint of the cross-section  */
  private Point midpoint;
  /** Channel bottom width  */
  private double bottomWidth;
  /** Channel height  */
  private double channelHeight;
  

  /**
   * Constructor. Private scope because empty constructor is appropriate
   * only for self-cloning
   */
  private IrregularChannelCrossSection(){}
  
  /**
   * Constructor
   * @param pointsList list of points that will make up this cross-section
   */
  public IrregularChannelCrossSection(Vector pointsList){
    // Takes a vector of points, but must validate
    // and sort them
    //
    // Structure:
    //
    // Points proceed from left bank (facing downstream)
    // to right; coordinates reflect distance from arbitrary
    // station on left bank, plus vertical distance from any
    // arbitrary station (all units meters, of course).
    // 
    // Criteria:
    //
    // No two points can have the same x-coordinate
    // No two consecutive points can have the same z-coordinate EXCEPT the
    //   points embracing the center point
    // The points with the lowest and highest X-coordinate points must
    // have the same Z coordinate
    
    // First, sort by X coordinate
    points = new ABCMDeepCloneableVector();
    while(pointsList.size()>0){
      Point toPlace = (Point)pointsList.remove(0);
      for(int j = 0; j < points.size(); j++){
        Point compare = (Point)points.get(j);
        if(compare.getX() > toPlace.getX()) {
          points.add(j,toPlace);
          toPlace = null;
          break;
        }
        // No valid set of points has two with the same X;
        // ignore the second one
        if(compare.getX() == toPlace.getX()){
          toPlace = null;
          break;
        }
      }
      // If it's gone the entire length without finding
      // a place to put it, toPlace will still have a point,
      // and should be added to the end of the vector
      if(toPlace != null) points.add(toPlace);
    }
    
//    for(int i = 0; i < points.size(); i++){
//      Point p = (Point)points.get(i);
//      System.out.println(p.coordString());
//    }
    
    // Now scan through the points in order and determine:
    // - Is this valid?
    //     Z values should decrease up to the midpoint,
    //          then increase
    // - What is the midpoint
    // - what is the channel depth
    // - what is the bottom width
    
    // First, the top should be the same on both sides
    double tld = ((Point)points.get(0)).getZ();
    double trd = ((Point)points.get(points.size()-1)).getZ();  
    if( tld != trd ){System.out.println("Top IS NOT SYMMETRICAL!!!");} 
    
    // Now, make sure it goes down, then up, with no reversals. While
    // doing this, grab the lowest point.
    boolean descending = true;
    midpoint = (Point)points.get(1);
    bottomWidth = 0;
    double deepestDepth;
    Point LastPoint = (Point)points.get(0);
    // Note: the _second_ point must be lower; check this
    // before even doing anything else...
    if(((Point)points.get(1)).getZ() > LastPoint.getZ()) return;  // BOOM- should default.
    Point NextPoint;
    for(int i = 1; i < points.size(); i++){
      NextPoint = (Point)points.get(i);
      if(descending){
        // If the two points are equal in depth
        if(NextPoint.getZ() == LastPoint.getZ()){
          midpoint = new Point((NextPoint.getX() + LastPoint.getX())/2, 0 ,NextPoint.getZ()); // Y is always 0
          bottomWidth = NextPoint.getX() - LastPoint.getX();
          descending=false;
        }
        if(NextPoint.getZ() > LastPoint.getZ()){
          midpoint = LastPoint;
          descending=false;
        }
        LastPoint = NextPoint;
      }
      // We're on the way back up...
      else{
        if(NextPoint.getZ() <= LastPoint.getZ()){System.out.println("Found a problem! Channel is not down-up");} // BOOM- should default.
      }
    }
    
    channelHeight = ((Point)points.get(0)).getZ() - midpoint.getZ();
    
    // Now we're going back through, but this time correcting the
    // coordinates to meet the standards...
    for(int i = 0; i < points.size(); i++){
      Point point = (Point)points.remove(i);
      Point NewP = new Point(point.getX() - midpoint.getX(),0,point.getZ() - midpoint.getZ());
      points.add(i,NewP);
    }
  }
    
  /** 
   * @see flowag.channels.ChannelCrossSection#getTopWidth() 
   */
  public double getTopWidth() {
    return ((Point)points.get(points.size()-1)).getX() - ((Point)points.get(0)).getX();    
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getBottomWidth() 
   */
  public double getBottomWidth() {
    return bottomWidth;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getChannelHeight() 
   */
  public double getChannelHeight() {  return channelHeight;  }

  /** 
   * @see flowag.channels.ChannelCrossSection#setTopWidth(double) 
   */
  public void setTopWidth(double topWidth) {
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#setBottomWidth(double) 
   */
  public void setBottomWidth(double bottomWidth) {
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#setChannelHeight(double) 
   */
  public void setChannelHeight(double channelHeight) {
  }

  /**
   * Gets the point at which water at a given depth
   * would touch the left bank
   * @param depth depth of the water
   * @return calculated point where water surface intersects bank
   */
  public Point getLeftPointAtDepth(double depth){
    if(depth < 0) return new Point(0,0,0);
    if(depth > channelHeight) return (Point)points.get(0);
    Point retPoint = null;
    Point LastPoint = (Point)points.get(0);
    Point NextPoint;
    for(int i = 1; i < points.size() - 1; i++){
      NextPoint = (Point)points.get(i);
      if(NextPoint.getZ() <= depth){
        double Nx = NextPoint.getX();
        double Lx = LastPoint.getX();
        double Nz = NextPoint.getZ();
        double Lz = LastPoint.getZ();
        retPoint = new Point(Lx + ((Nx - Lx) / (Nz - Lz)) * (depth - Lz), 0, depth);
        return retPoint;
      }
    }
    return (retPoint ==  null ? new Point(0,0,0) : retPoint);
  }
  
  /**
   * Gets the point at which water at a given depth
   * would touch the right bank
   * @param depth depth of the water
   * @return calculated point where water surface intersects bank
   */
  public Point getRightPointAtDepth(double depth){
    if(depth < 0) return new Point(0,0,0);
    if(depth > channelHeight) return (Point)points.get(points.size()-1);
    Point retPoint = null;
    Point LastPoint = (Point)points.get(points.size() - 1);
    Point NextPoint;
    for(int i = points.size() - 1 - 1; i > 0; i--){
      NextPoint = (Point)points.get(i);
      if(NextPoint.getZ() <= depth){
        double Nx = NextPoint.getX();
        double Lx = LastPoint.getX();
        double Nz = NextPoint.getZ();
        double Lz = LastPoint.getZ();
        retPoint = new Point(Lx + ((Nx - Lx) / (Nz - Lz)) * (depth - Lz), 0, depth);
        return retPoint;
      }
    }
    return (retPoint ==  null ? new Point(0,0,0) : retPoint);
  }
  
  /** 
   * @see flowag.channels.ChannelCrossSection#getArea() 
   */
  public double getArea() {
    return getFlowArea(channelHeight);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getWaterSurfaceWidth(double) 
   */
  public double getWaterSurfaceWidth(double waterDepth) {
    if(waterDepth == 0) return 0;
    Point L = getLeftPointAtDepth(waterDepth);
    Point R = getRightPointAtDepth(waterDepth);
    if(L != null && R != null) return R.getX() - L.getX();
    return 0;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getWettedPerimeter(double) 
   */
  public double getWettedPerimeter(double waterDepth) {
    if(waterDepth == 0) return 0;
    double P = 0;
    Vector p = wettedPerimeter(waterDepth);
    for(int i = 0; i < p.size(); i++){
      LineSegment L = (LineSegment)p.get(i);
      P += L.length(); 
    }
    return P;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getFlowArea(double) 
   */
  public double getFlowArea(double waterDepth) {
    // TODO Faster algorithm? See centroid(waterDepth)
    double A = 0;
    Vector p = wettedPerimeter(waterDepth);
    if(p.size() == 0) return 0;
    for(int i = 0; i < p.size(); i++){
      LineSegment L = (LineSegment)p.get(i);
      Point P1 = L.P1;
      Point P2 = L.P2;
      // Form a trapezoid with the water surface as one side and the bottom
      // as another, with imaginary verticals forming the remaining two, parallel
      // sides. The area should be the width of this times the average of the
      // two depths.
      double leftX = P1.getX();
      double rightX = P2.getX();
      double sectionWidth = rightX - leftX;
      double top = waterDepth;
      double leftHeight = waterDepth - P1.getZ();
      double rightHeight = waterDepth - P2.getZ();
      double avgHeight = (leftHeight + rightHeight)/2d;
      double sectionArea = sectionWidth * avgHeight;
//      System.out.println("Section " + i + " area: " + sectionArea);
      A += sectionArea;
//      A += (P2.getX() - P1.getX()) * ((waterDepth - (P1.getZ()) + (waterDepth - P2.getZ()))/2); 
    }
    return A;
  }
  
  /**
   * Returns a vector of segments that include only those segments
   * or portions of segments covered by water at the specified depth
   * @param waterDepth water depth
   * @return Immersed segments (from which wetted perimeter can be calculated)
   */
  private Vector wettedPerimeter(double waterDepth){
    Vector perimeter = new Vector();
    Point L = getLeftPointAtDepth(waterDepth);
    Point R = getRightPointAtDepth(waterDepth);
    if(L == null || R == null) return null;
    Point LastPoint = (Point)points.get(0);
    Point NextPoint;
    for(int i = 1; i < points.size(); i++){
      NextPoint = (Point)points.get(i);
      if(LastPoint.getX() <= L.getX() && NextPoint.getX() > L.getX()) {
        LineSegment LS = new LineSegment(L, NextPoint);
        perimeter.add(LS);
      }
      if(LastPoint.getX() >= L.getX() && NextPoint.getX() <= R.getX()){
        LineSegment LS = new LineSegment(LastPoint, NextPoint);
        perimeter.add(LS);
      }
      if(LastPoint.getX() <= R.getX() && NextPoint.getX() > R.getX()){
        LineSegment LS = new LineSegment(LastPoint, R);
        perimeter.add(LS);
      }
      LastPoint = NextPoint;
    }
    return perimeter;
  }
  
  

  /** 
   * @see flowag.channels.ChannelCrossSection#getAvgDepth(double) 
   */
  public double getAvgDepth(double waterDepth) {
    double A = getFlowArea(waterDepth);
    double W = getWaterSurfaceWidth(waterDepth);
    if(W == 0) return 0;
    return A / W;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getHydraulicRadius(double) 
   */
  public double getHydraulicRadius(double waterDepth) {
    if(waterDepth <= 0 ) return 0.0;
    double flowArea = getFlowArea(waterDepth);
    double wettedPerimeter = getWettedPerimeter(waterDepth);
    return flowArea/wettedPerimeter;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getConveyance(double, double) 
   */
  public double getConveyance(double waterDepth, double roughness) {
    if(roughness <= 0) return 0;
    double A = getFlowArea(waterDepth);
    double R = getHydraulicRadius(waterDepth);
    return (A * (BasicMath.cbrt(R * R))) / roughness;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getDADz(double) 
   */
  public double getDADz(double waterDepth) {
    return getWaterSurfaceWidth(waterDepth);
    // The value of dA/dZ will be equal to:
    //   1/2 the value of dA/dZ in a trapezoidal channel cross-section with
    //         slope equal to the slope of the left bank
    //   PLUS
    //   1/2 the value of dA/dZ in a trapezoidal channel cross-section with
    //         slope equal to the slope of the right bank
//    Vector V = wettedPerimeter(waterDepth);
//    Line L = (Line)V.get(0);
//    Line R = (Line)V.get(V.size()-1);
//    
//    TrapezoidalChannelCrossSection TCS1 = new TrapezoidalChannelCrossSection(2 * (midpoint.getX() - L.P1.getX()), 
//                                                                             2 * (midpoint.getX() - L.P2.getX()),
//                                                                             L.P1.getZ() - L.P2.getZ());
//    TrapezoidalChannelCrossSection TCS2 = new TrapezoidalChannelCrossSection(2 * (R.P2.getX() - midpoint.getX()), 
//                                                                             2 * (R.P1.getX() - midpoint.getX()),
//                                                                             R.P2.getZ() - R.P1.getZ());
//    // Note: For a trapezoidal cross section, water depth, provided it is
//    // above zero, is irrelevant- all dA/dZ will be the same. Hence the .5d
//    // here is arbitrary.
//    double dAdz = .5d * TCS1.getDADz(.5d) + .5d * TCS2.getDADz(.5d);
//    return dAdz;
  }

  /** 
   * @see flowag.channels.AbstractChannelCrossSection#getDKDz(double, double) 
   */
  public double getDKDz(double waterDepth, double roughness) {
    // TODO Error trap getDKDZ if water depth is too low or high
    double e = .01; // A small epsilon value
    return (getConveyance(waterDepth + e, roughness) - getConveyance(waterDepth - e, roughness))/(2*e);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getFrictionSlope(double, double, double) 
   */
  public double getFrictionSlope(double discharge, double waterDepth,
      double roughness) {
    if(waterDepth == 0) return 0;
    double R = getHydraulicRadius(waterDepth);
    double A = getFlowArea(waterDepth);
    double Q = discharge;
    double n = roughness;
    double fs =Q * Math.abs(Q)* Math.pow(n,2) / ((BasicMath.cbrt(Math.pow(R,4.0d)) * Math.pow(A,2)));
    return fs;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getManningVelocity(double, double, double) 
   */
  public double getManningVelocity(double waterDepth, double roughness,
      double slope) {
    double R = getHydraulicRadius(waterDepth);
    return BasicMath.cbrt(Math.pow(R,2)) * StrictMath.sqrt(Math.abs(slope)) / roughness;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getManningDischarge(double, double, double) 
   */
  public double getManningDischarge(double waterDepth, double roughness,
      double slope) {
    double R = getHydraulicRadius(waterDepth);
    double A = getFlowArea(waterDepth);
    return A * getManningVelocity(waterDepth, roughness, slope);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getCutAndSurfaceLine(abcm.math.Point, double) 
   */
  public Vector getCutAndSurfaceLine(Point origin, double channelHeading) {
    Vector v = new Vector();

    double leftHeading  = channelHeading - (Math.PI/2);
//    double rightHeading = channelHeading + (Math.PI/2);
    
    for(int i = 0; i < points.size(); i++){
      Point P = (Point)points.get(i);
      Point u = new Point(origin.getX(), origin.getY(), origin.getZ() + P.getZ());
      v.add(Geometry.getOffsetPoint(u, leftHeading, P.getX()));
    }
    return v;
  }

  /**
   * @see flowag.channels.ChannelCrossSection#centroidDepth(double)
   */
  public double centroidDepth(double waterDepth){
    return (centroid(waterDepth).getZ());    
  }
  
  /**
   * Calculates the centroid 
   * @param waterDepth
   * @return the centroid for this cross-section
   */
  private Point centroid(double waterDepth){
    Vector pointsList = wettedPerimeter(waterDepth);
    double areaSum = 0;
    double xSum    = 0;  // For performance, eliminate the calculation of the x coordinate...
    double ySum    = 0;
    for(int i = 0; i < pointsList.size(); i++){
      Point P1 = (Point)pointsList.get(i);
      Point P2 = (i < (pointsList.size() - 1) ? (Point)pointsList.get(i+1) : (Point)pointsList.get(0));
      double x1 = P1.getX();
      double y1 = P1.getZ(); // Remember, z is the vertical dimension here
      double x2 = P2.getX();
      double y2 = P2.getZ(); // ditto...
      areaSum += (x1 * y2) - (x2 * y1);
      xSum    += (x1 + x2) * (x1 * y2 - x2 * y1);
      ySum    += (y1 + y2) * (x1 * y2 - x2 * y1);
    }
    double area = areaSum / 2d;
    double cx = xSum / (6d * area);
    double cz = ySum / (6d * area);
    
    return new Point(cx, 0, cz); 
  }
  
  /** 
   * @see flowag.channels.ChannelCrossSection#fill(double) 
   */
  public void fill(double fillDepth) {
    for(int i = 0; i < points.size(); i++){
      Point P = (Point)points.remove(i);
      Point newPoint = new Point(P.getX(), P.getY(), P.getZ() * (1 - fillDepth/this.channelHeight)); 
      points.add(i, newPoint);
      if(P.equals(midpoint)) midpoint = newPoint;
      this.channelHeight *= (1 - fillDepth/this.channelHeight);
    }
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#dig(double) 
   */
  public void dig(double depth) {
    for(int i = 0; i < points.size(); i++){
      Point P = (Point)points.remove(i);
      Point newPoint = new Point(P.getX(), P.getY(), P.getZ() * (1 + depth/this.channelHeight)); 
      points.add(i, newPoint);
      if(P.equals(midpoint)) midpoint = newPoint;
      this.channelHeight *= (1 + depth/this.channelHeight);
    }

  }

  /** 
   * @see flowag.channels.ChannelCrossSection#type() 
   */
  public String type(){
    return "Irregular";
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getCopy() 
   */
  public ChannelCrossSection getCopy(){
    Vector newPointsList = new Vector(points);
    return new IrregularChannelCrossSection(newPointsList);
  }

  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    IrregularChannelCrossSection clone = new IrregularChannelCrossSection();
    cloneStore.put(this, clone);
    clone.points = (ABCMDeepCloneableVector)points.getDeepClone(cloneStore);
    clone.midpoint = (Point)cloneStore.getClone(midpoint);
    clone.bottomWidth = bottomWidth;
    clone.channelHeight = channelHeight;
    return clone;
  }
}
