package flowag.channels;

import java.util.Collection;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.Vector;

/**
 * LPGFlowCalculator = "Looks Pretty Good"
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class LPGFlowCalculator extends AbstractFlowCalculator{

  /* Code Model */
  /** Code Model Switch ID */ public static final String LPG_DEPTH_CALCULATION = "LPG_DEPTH_CALCULATION";
  /** Code Model Switch for calculation method */ public static final int    MANNING_REVERSE_EST   = 0;
  /** Code Model Switch for calculation method */ public static final int    FRACTION_OF_CHANNEL   = 1;
  
  
  /* Static Methods */
  /**
   * Indicates whether a segment is an initial segment or not.
   * @param cs the canal segment to be tested
   * @return true if the inflow to this segment is not a headgate, false otherwise
   */
  public static boolean isInitialSegment(CanalSegment cs){
    if(cs.getInflowGate().getHeadGate() != null)  return true;
    return false;
  }
  
  /**
   * Returns a collection of LPG flow systems.
   * @param canals collection of canals to test
   * @return list of LPG flow systems derived from that collection
   */
  public static Vector<LPGFlowCalculator> getFlowSystems(Collection<CanalSegment> canals){
    Vector<LPGFlowCalculator> flowSystems = new Vector<LPGFlowCalculator>();
    for(CanalSegment cs: canals){
      if(isInitialSegment(cs)){
        LPGFlowCalculator flowCalc = new LPGFlowCalculator(cs);
        flowSystems.add(flowCalc);
        if(!cs.getInflowGate().isOpen()) 
            flowCalc.setUpstreamBoundaryCondition(new FlowCalculator.ClosedGateUpstreamBoundaryCondition(cs));
      }
      
    }
    return flowSystems;
  }
  
  /* Inner Classes */
  
  /**
   * Utility class that records data about a specific canal segment
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class CanalSegData implements Comparable{
    /** The canal segment with which this data object is associated */     public CanalSegment canalSeg;
    /** Distance from the headgate */                                      public double       distance;
    /** Calculated new value for Q */                                      public double       newQ;
    /** Pointer to the data object for the segment upstream of this one */ public CanalSegData upstreamSegmentData;
    
    /**
     * Constructor
     * @param seg segment represented
     * @param dist distance from headgate
     * @param upstreamSegData segment data object for upstream segment
     */
    public CanalSegData(CanalSegment seg, double dist, CanalSegData upstreamSegData){
      canalSeg            = seg;
      distance            = dist;
      newQ                = 0;
      upstreamSegmentData = upstreamSegData;
    }
    
    /**
     * Adds the given dQ value to the current value for Q
     * @param dQ delta-Q value to add
     */
    public void addQ(double dQ){
      newQ += dQ;
      if(upstreamSegmentData != null) upstreamSegmentData.addQ(dQ);
    }

    /**
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(Object other){
      return (!(other instanceof CanalSegData) ? 0 : 
        ((((CanalSegData)other).distance != distance) ? (int)(distance - ((CanalSegData)other).distance) : 1));
    }   
  }
  
  /* Instance Variables */
  /** Collection of all segments in this system */ 
  private Vector<CanalSegData>  allSegs          = new Vector<CanalSegData>();
  /** Collection of all segments that terminate in fields */
  private TreeSet<CanalSegData> fieldTermini     = new TreeSet<CanalSegData>();
  /** Collection of all segments that are terminal but do not terminate in fields */ 
  private Vector<CanalSegData>  nonFieldTermini  = new Vector<CanalSegData>();
  /** Record of the initial upstream Q value for the initial segment */ 
  private double initialSegmentUpstreamQ;
  
  /* Constructor */
  
  /**
   * Constructor that initializes all data sets
   * to match the data structure needed for performing the
   * flow calculations.
   * @param cs the initial canal segment
   */
  public LPGFlowCalculator(CanalSegment cs){
    super(cs);
    recordSegment(cs, 0, cs.getInflowGate().isOpen(), null);
  }
  
  /**
   * Stores data about a canal segment in the appropriate data
   * structure. Calls itself for the segments outflow canal
   * segments, thereby tracking the entire flow system
   * @param cs the current canal segment
   * @param dist the total distance of all segments leading to this one
   * @param isConnectedToInflow true if the path followed so far
   * has never encountered a closed gate, but false otherwise
   * @param upstreamData the segment data for the immediately upstream segment of this one
   */
  private void recordSegment(CanalSegment cs, double dist, boolean isConnectedToInflow, CanalSegData upstreamData){
    double endDist = dist + cs.getLength();
    Vector outflowCanals = cs.getOutflowCanals();
    CanalSegData thisSegData = new CanalSegData(cs, endDist, upstreamData);
    allSegs.add(thisSegData);
    
    // If this is a terminal segment
    if(outflowCanals.size() == 0){
      if(isConnectedToInflow){
       // Determine which bin to store this in:
        if(cs.outflowGateIsOpen()){  // It's an open terminus
          if(cs.plotsIrrigated() > 0){ // It flows into a field
            fieldTermini.add(thisSegData);
          } 
          else{ // It flows out of the system
            nonFieldTermini.add(thisSegData);
          } 
        }
      }
    }    
    else {
      // Now track down the canal system
      for(int i = 0; i < outflowCanals.size(); i++){
        CanalSegment csOut = (CanalSegment)outflowCanals.get(i);
        recordSegment(csOut, endDist, (isConnectedToInflow && csOut.inflowGateIsOpen()), thisSegData);
      }
    }
  }
  
  
  /* Perform Calculations */
  
  /**
   * @see flowag.channels.AbstractFlowCalculator#performCalculations(double)
   */
  protected boolean performCalculations(double duration){
    // Right now all flow is zero; if all outflows are closed, keep it that way
    if((fieldTermini.isEmpty()) && (nonFieldTermini.size() == 0)) return true;
    // Get the amount flowing into the system
    if(ubc == null) System.out.println("UBC IS NULL!!!");
    double Qin = ((HeadgateBasedUpstreamBoundaryCondition)ubc).getFinalQ(duration);
    
    
    initialSegmentUpstreamQ = Qin; // The initial upstream inflow will be equal to this
    if(Qin == 0) return true; // No need to do more- canal system will be emptied
    
    // Determine the maximum amount that can flow to fields
    double maxFieldQ = 0;
    for(CanalSegData csd: fieldTermini) maxFieldQ += csd.canalSeg.getMaximumOutflow();
    
    // Distribute inflow to the field outflows up to their maximum amounts
    // This is done in order from nearest to farthest; if water runs out, the farther
    // outward outlets lose
    Iterator I = fieldTermini.iterator(); // Two criteria- don't use for-each loop
    while(I.hasNext() && Qin > 0){
      CanalSegData csd = (CanalSegData)I.next();
      double Q = Math.min(csd.canalSeg.getMaximumOutflow(), Qin);
//      System.out.println("Setting field terminus " + csd.canalSeg.getCanalSegmentID() + " to " + Q + " from " + Qin);
      csd.addQ(Q);  // Remember, this adds to all segments UPSTREAM of this one, too!
      Qin -= Q; // New 
    }
    
    // If there is still water to be distributed
    if(Qin > 0){
      // If there are non-field termini, the water will go entirely there distributed evenly among them, but
      // if there are only field termini, it gets distributed evenly among them
      if(nonFieldTermini.size() > 0){
        double QDistrib = Qin / (nonFieldTermini.size());
        for(CanalSegData csd: nonFieldTermini) csd.addQ(QDistrib);
      }
      else{
        double QDistrib = Qin / (fieldTermini.size());
        for(CanalSegData csd: fieldTermini) csd.addQ(QDistrib);
      }
    }
    
    // Possible to check for consistency here- final value for first segment
    // CanalSegData flow should match input flow...
    
    
    
    return true; // This algorithm never fails (miraculous!)
  }

  /* Set Results */
  
  /**
   * @see flowag.channels.AbstractFlowCalculator#setResults()
   */
  protected void setResults(){
    results = new FlowCalculator.FlowZAndQAdjustments();
    
    if(initialSegmentUpstreamQ == 0)  AbstractFlowCalculator.setToEmpty(this.initialCanalSegment, results);
    
    
    for(CanalSegData csd: allSegs){
      if(csd.upstreamSegmentData == null){ // It's the initial segment
        double newDepth = 0;
        switch(codeModelOption(LPG_DEPTH_CALCULATION)){
          case FRACTION_OF_CHANNEL:
            newDepth = (csd.newQ > 0 ? csd.canalSeg.getUpstreamChannelHeight() * .667 : 0);
            break;
          case MANNING_REVERSE_EST:
          default:
            newDepth = csd.canalSeg.getUpstreamDepthPerManningDischarge(initialSegmentUpstreamQ); // THIS IS SLOW!
            break;
        }
        results.setInitialUpstreamValues(csd.canalSeg, 
                                         newDepth - csd.canalSeg.getUpstreamWaterDepth(), 
                                         initialSegmentUpstreamQ - csd.canalSeg.getUpstreamDischarge());
      }
      double newDepth = 0;  
      switch(codeModelOption(LPG_DEPTH_CALCULATION)){
        case FRACTION_OF_CHANNEL:
          newDepth = (csd.newQ > 0 ? csd.canalSeg.getDownstreamChannelHeight() * .667 : 0);
          break;
        case MANNING_REVERSE_EST:
        default:
          newDepth = csd.canalSeg.getDownstreamDepthPerManningDischarge(csd.newQ); // THIS IS SLOW!
          break;
      }
      results.addDownstreamValues(csd.canalSeg, 
                                newDepth - csd.canalSeg.getDownstreamWaterDepth(), 
                                csd.newQ - csd.canalSeg.getDownstreamDischarge() );
    }
  }

}
