package flowag.channels;

import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.math.BasicMath;
import abcm.math.Geometry;
import abcm.math.Point;

/**
 * Represents a channel cross section with a parabolic shape,
 * defined by two parameters: the top width and the height
 * from top to bottom.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class ParabolicChannelCrossSection extends AbstractChannelCrossSection {
  
  /** Number of points in a cut and surface line  */
  private static int CUT_AND_SURFACE_PT_COUNT = 20;
  
  /** Parabola: y = ax^2   a is coefficient  */
  double coeff;
  /** Top Width  */
  double tWidth;
  /** Channel Height  */
  double chHeight;
  
  /**
   * Gets the coefficient that describes a parabola given the other dimensions
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @return the calculated coefficient
   */
  private static double getCoefficient(double topWidth, double channelHeight){
    double x = topWidth / 2d;
    return channelHeight / (x * x);
  }

  /**
   * Calculates the area for a parabola with the given dimensions
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @return the calculated flow area
   */
  public static double getArea(double topWidth, double channelHeight){
    return getFlowArea(topWidth, channelHeight, channelHeight);
  }
  
  /**
   * Calculates the water surface width given the channel dimensions and water depth
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @param waterDepth depth of water
   * @return the calculated water surface width
   */
  public static double getWaterSurfaceWidth(double topWidth, double channelHeight, double waterDepth){
    double c = getCoefficient(topWidth, channelHeight);
    return (2 * waterDepth) / c;
  }

  // REF: c8scdm.pdf
  /**
   * Calculates the wetted perimeter given the channel dimensions and water depth
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @param waterDepth depth of water
   * @return the wetted perimeter
   */
  public static double getWettedPerimeter(double topWidth, double channelHeight, double waterDepth){
    double t = getWaterSurfaceWidth(topWidth, channelHeight, waterDepth);
    return t + (8 * waterDepth * waterDepth) / (3 * t);
  }
  
  /**
   * Calculates the flow area given the channel dimensions and water depth
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @param waterDepth depth of water
   * @return the flow area
   */
  public static double getFlowArea(double topWidth, double channelHeight, double waterDepth){
    double t = getWaterSurfaceWidth(topWidth, channelHeight, waterDepth);
    return (2d / 3d) * t * waterDepth;
  }
  
  /**
   * Calculates the average depth given the channel dimensions and water depth
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @param waterDepth depth of water
   * @return the calculated average depth
   */
  public static double getAvgDepth(double topWidth, double channelHeight, double waterDepth){
    double A = getFlowArea(topWidth, channelHeight, waterDepth);
    double W = getWaterSurfaceWidth(topWidth, channelHeight, waterDepth);
    if(W == 0) return 0;
    return A / W;
  }
  
  /**
   * Calculates the hydraulic radius given the channel dimensions and water depth
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @param waterDepth depth of water
   * @return the calculated hydraulic radius
   */
  public static double getHydraulicRadius(double topWidth, double channelHeight, double waterDepth){
    if(waterDepth <= 0 ) return 0.0;
    double flowArea = getFlowArea(topWidth, channelHeight, waterDepth);
    double wettedPerimeter = getWettedPerimeter(topWidth, channelHeight, waterDepth);
    return flowArea/wettedPerimeter;
  }

  /**
   * Calculates the conveyance given the channel dimensions, roughness and water depth
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @param waterDepth depth of water
   * @param roughness channel roughness (n)
   * @return calculated conveyance
   */
  public static double getConveyance(double topWidth, double channelHeight, double waterDepth, double roughness){
    if(roughness <= 0) return 0;
    double A = getFlowArea(topWidth, channelHeight, waterDepth);
    double R = getHydraulicRadius(topWidth, channelHeight, waterDepth);
    return (A * (BasicMath.cbrt(R * R))) / roughness;
  }

  /**
   * Calculates the 'dA/dz' value (change in area with increase in stage) given the channel dimensions and water depth
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @param waterDepth depth of water
   * @return calculated da/dZ
   */
  public static double getDADz(double topWidth, double channelHeight, double waterDepth){
    double c = getCoefficient(topWidth, channelHeight);
    return 2d * Math.sqrt( 1d / c) * Math.sqrt(waterDepth);
  }
  
  /**
   * Calculates the 'dK/dz' value (change in conveyance with increase in stage) given the channel dimensions,
   * roughness, and water depth
   * @param topWidth channel top width
   * @param channelHeight channel height
   * @param waterDepth depth of water
   * @param roughness channel roughness (n)
   * @return calculated dK/dZ
   */
  public static double getDKDz(double topWidth, double channelHeight, double waterDepth, double roughness){
    double e = .01; // A small epsilon value
    return (getConveyance(topWidth, channelHeight, waterDepth + e, roughness) - 
            getConveyance(topWidth, channelHeight, waterDepth - e, roughness))/(2*e);
  }
  
  
  /**
   * Constructor
   * @param topWidth top width of channel
   * @param channelHeight height of channel
   */
  public ParabolicChannelCrossSection(double topWidth, double channelHeight){
    tWidth = topWidth;
    chHeight = channelHeight;
    coeff = getCoefficient(topWidth, channelHeight);
  }
    
  /** 
   * @see flowag.channels.ChannelCrossSection#getTopWidth() 
   */
  public double getTopWidth() {
    return tWidth;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getBottomWidth() 
   */
  public double getBottomWidth() {
    return 0;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getChannelHeight() 
   */
  public double getChannelHeight() {
    return chHeight;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#setTopWidth(double) 
   */
  public void setTopWidth(double topWidth) {
    tWidth = topWidth;
    coeff = chHeight / (tWidth * tWidth / 2d);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#setBottomWidth(double) 
   */
  public void setBottomWidth(double bottomWidth) {
    // Not applicable for this cross-section
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#setChannelHeight(double) 
   */
  public void setChannelHeight(double channelHeight) {
    chHeight = channelHeight;
    coeff = chHeight / (tWidth * tWidth / 2d);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getArea() 
   */
  public double getArea() {
    return getFlowArea(chHeight);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getWaterSurfaceWidth(double) 
   */
  public double getWaterSurfaceWidth(double waterDepth) {
    return getWaterSurfaceWidth(tWidth, chHeight, waterDepth);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getWettedPerimeter(double) 
   */
  public double getWettedPerimeter(double waterDepth) {
    return getWettedPerimeter(tWidth, chHeight, waterDepth);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getFlowArea(double) 
   */
  public double getFlowArea(double waterDepth) {
    return getFlowArea(tWidth, chHeight, waterDepth);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getAvgDepth(double) 
   */
  public double getAvgDepth(double waterDepth) {
    return getAvgDepth(tWidth, chHeight, waterDepth);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getHydraulicRadius(double) 
   */
  public double getHydraulicRadius(double waterDepth) {
    return getHydraulicRadius(tWidth, chHeight, waterDepth);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getConveyance(double, double) 
   */
  public double getConveyance(double waterDepth, double roughness) {
    return getConveyance(tWidth, chHeight, waterDepth, roughness);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getDADz(double) 
   */
  public double getDADz(double waterDepth) {
    return getDADz(tWidth, chHeight, waterDepth);
  }

  /** 
   * @see flowag.channels.AbstractChannelCrossSection#getDKDz(double, double) 
   */
  public double getDKDz(double waterDepth, double roughness) {
    return getDKDz(tWidth, chHeight, waterDepth, roughness);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getFrictionSlope(double, double, double) 
   */
  public double getFrictionSlope(double discharge, double waterDepth,
      double roughness) {
    if(waterDepth == 0) return 0;
    double R = getHydraulicRadius(waterDepth);
    double A = getFlowArea(waterDepth);
    double Q = discharge;
    double n = roughness;
    double fs =Q * Math.abs(Q)* Math.pow(n,2d) / ((BasicMath.cbrt(Math.pow(R,4.0d)) * Math.pow(A,2d)));
    return fs;
  }

  /**
   * @see flowag.channels.ChannelCrossSection#centroidDepth(double)
   */
  public double centroidDepth(double waterDepth){
    return (3d / 5d) * waterDepth;
  }
  
  /** 
   * @see flowag.channels.ChannelCrossSection#getManningVelocity(double, double, double) 
   */
  public double getManningVelocity(double waterDepth, double roughness,
      double slope) {
    double R = getHydraulicRadius(waterDepth);
    return BasicMath.cbrt(Math.pow(R,2d)) * StrictMath.sqrt(Math.abs(slope)) / roughness;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getManningDischarge(double, double, double) 
   */
  public double getManningDischarge(double waterDepth, double roughness,
      double slope) {
    double R = getHydraulicRadius(waterDepth);
    double A = getFlowArea(waterDepth);
    return A * getManningVelocity(waterDepth, roughness, slope);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#getCutAndSurfaceLine(abcm.math.Point, double) 
   */
  public Vector getCutAndSurfaceLine(Point origin, double channelHeading) {
    double leftHeading  = channelHeading - (Math.PI/2);
//    double rightHeading = channelHeading + (Math.PI/2);
    
    Vector points = new Vector();
    double leftBank = -1d * tWidth / 2d;
    double rightBank = tWidth/2d;
    for(int i = 0; i <= CUT_AND_SURFACE_PT_COUNT; i++){
      double x = leftBank + ((double)i / CUT_AND_SURFACE_PT_COUNT) * tWidth;
      double z = coeff * x * x;
      points.add(new Point(x, 0, z));
    }
    Vector v = new Vector();
    for(int i = 0; i < points.size(); i++){
      Point P = (Point)points.get(i);
      Point u = new Point(origin.getX(), origin.getY(), origin.getZ() + P.getZ());
      v.add(Geometry.getOffsetPoint(u, leftHeading, P.getX()));
    }
    return v;
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#fill(double) 
   */
  public void fill(double fillDepth) {
    chHeight -= fillDepth;
    coeff = getCoefficient(tWidth, chHeight);    
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#dig(double) 
   */
  public void dig(double depth) {
    chHeight += depth;
    coeff = getCoefficient(tWidth, chHeight);
  }

  /** 
   * @see flowag.channels.ChannelCrossSection#type() 
   */
  public String type(){
    return "Parabolic";
  }
  
  /** 
   * @see flowag.channels.ChannelCrossSection#getCopy() 
   */
  public ChannelCrossSection getCopy(){
    return new ParabolicChannelCrossSection(tWidth, chHeight);
  }

  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    ParabolicChannelCrossSection clone = new ParabolicChannelCrossSection(tWidth, chHeight);
    cloneStore.put(this, clone);
    return clone;
  }
  
}
