package flowag.channels;

import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.cloning.ABCMDeepCloneableVector;
import abcm.graphics.SVGAssist;
import abcm.graphics.X3DAssist;
import abcm.math.Point;
import abcm.sim.AbstractModelComponent;
import abcm.space.SpatialElement;
import abcm.space.TriangulatedTopography;
import flowag.basic.UnitConversion;
import flowag.basic.Water;



/**
 * A River is a vaguely defined object that acts merely to convey water
 * to headgates. Rivers can have streamflow values, and will eventually
 * have a stage value. HeadGates are sequential, so that downstream
 * conditions can be different from conditions upstream, but this is not
 * yet implemented.
 *
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class River extends AbstractModelComponent implements SpatialElement, ABCMDeepCloneable {
  /** Name of river  */
  private String    name;
  /** Current flow level of river  */
  private double    flowLevel;
  /** Annual Streamflow value  */
  private double    AnnualStreamflow;
  /** Collection of HeadGates  */
  private ABCMDeepCloneableVector    HeadGates;
  /** List of HeadGates by name  */
  private ABCMDeepCloneableHashtable HeadGateList;
  /** Sample of water indicating water quality in the river  */
  private Water     waterSample;
  /** TodaysDailyFlow   */
  private double    todaysDailyFlowInMAF = 0;
  /** Previous daily flow (used for calculating delta Q)  */
  private double    previousDayDailyFlowInMAF = 0;

  /* Static methods */
  
  /* Constructors */

  /**
   * Basic constructor
   * @param RiverName the name to be assigned to this river object
   */
  public River(String RiverName){
    name = RiverName;
    HeadGates = new ABCMDeepCloneableVector();
    HeadGateList = new ABCMDeepCloneableHashtable();
    waterSample = new Water(1);
  }

  /* Basic Setters */

  /**
   * Sets a name for this River
   * @param string new name for this River
   */
  public void setName(String string)        {   name = string;          }

 
  
  /* Headgates */
  
  /**
   * Adds a head gate to the collection of HeadGates
   * @param HG the HeadGate to be added
   * @param index The position in the HeadGates collection
   * at which the head gate should be added.
   * @param name the name of the HeadGate to be added
   * @return a String indicating the success of the operation
   */
  public String addHeadGate(HeadGate HG, int index, String name){
    if(HeadGateList.containsKey(name)) return "Could not add Head Gate '" + name + "'. A Head Gate with that name already exists.";;
    int indx = Math.min(index, HeadGates.size());
    HeadGates.add(indx, HG);
    HeadGateList.put(name, HG);
    HG.setRiver(this);
    return "Successfully added Head Gate '" + name + "' at position " + getHeadGateIndex(name);
  }
  
  /**
   * Gets a specified HeadGate from this River. 
   * Returns null if no HeadGate with the specified name exists.
   * @param headgateName the name of the requested HeadGate
   * @return the HeadGate requested, if available
   */
  public HeadGate getHeadGate(String headgateName){
    if(!HeadGateList.containsKey(headgateName)) return null;
    return (HeadGate)HeadGateList.get(headgateName);   
  }
  
  /**
   * Gets the index of the specified HeadGate.
   * @param name the name of the HeadGate whose index is requested
   * @return the index of the specified HeadGate
   */
  public int getHeadGateIndex(String name){
    if(!HeadGateList.containsKey(name)) return -1;
    HeadGate HG = (HeadGate)HeadGateList.get(name);
    for(int i = 0; i < HeadGates.size(); i++){
      HeadGate HG2 = (HeadGate)HeadGates.get(i);
      if(HG.equals(HG2)) return i;
    }
    return -1;
  }

  /* Basic getters */
  
  /**
   * Gets the name of this River
   * @return the name of this River
   */
  public String getName() {               return name;                         }
  
  
  
  /* Water */
  
  /**
   * Gets the water sample from this river.
   * A water sample will contain specific concentrations
   * of chemicals
   * @return the water sample for this River object.
   */
  public Water getWaterSample() {         return waterSample;                  }
  
  /**
   * Remove a chemical from this river.
   * @param chemicalName
   * @return true if successful, false otherwise
   */
  public boolean removeChemicalFromRiver(String chemicalName){
    return waterSample.removeChemical(chemicalName);
  }
  
  
  /* Flow */
  
  // Flow setters: Two kinds-
  //     Annual Streamflow is Million Acre-Feet per year
  //     Flow Level is the _fraction_ of Annual Streamflow
  //         that reaches/passes the first headgate _per day_
  
  /**
   * Sets the annual streamflow for this River
   * @param d new value for annual streamflow for this River
   */
  public void setAnnualStreamflow(double d) {  AnnualStreamflow = d;   }
  
  /**
   * Sets the current flow level for this River.
   * The flow level is the fraction of the annual streamflow that
   * will pass on a given day. If water flowed evenly throughout
   * the year, flow level would be a constant 1/365. To
   * distribute flow within a year, adjust this value. The sum
   * of all such values for a given calendar year should be 1, but
   * this is nowhere enforced (it should be taken into account
   * when designing the series of instructions that call this
   * method).
   * @param flow The new value for the current flow level for this River
   */
  public void setFlowLevel(double flow)     {  System.out.println ("Resetting flow level to " + flow); flowLevel = flow;       }
  
  
  // Basic Flow Getters
  
  /**
   * Gets the annual streamflow for this River
   * @return the annual streamflow for this River
   */
  public double getAnnualStreamflow() {   return AnnualStreamflow;             }
  
  /**
   * Gets the current flow level for this River
   * @return the current flow level for this River
   */
  public double getFlowLevel()        {   return flowLevel;                    }
  
  
  // Conversions
  //
  // The most fundamental conversion is from annual flow, in Million Acre-Feet
  // to daily flow, also in Million Acre-Feet, by simple multiplication of
  // the FlowLevel
  
  /**
   * Gets the flow of the river in units of million acre-feet per day.
   * @return a number indicating the volume of water, in million
   * acre-feet per day, that will flow given this river's daily
   * flow rate and annual streamflow.
   */
  public double getCurrentDayDailyFlowInMAF() {  return todaysDailyFlowInMAF;   }
  
  
  /**
   * Gets the daily flow, in Million Acre-Feet, at the specified
   * HeagGate
   * @param HG
   * @return the dail flow, in million acre-feet
   */
  public double getDailyMAF(HeadGate HG){
    if(!HeadGates.contains(HG)) return 0;
    double flowAvailable = getCurrentDayDailyFlowInMAF();
    for(int i = 0; i < HeadGates.size(); i++){
      HeadGate H = (HeadGate)HeadGates.get(i);
      if(H.equals(HG)) return flowAvailable;
      flowAvailable *= (1- H.getEfficiency());
    }
    return 0;
  }
  
  
  /**
   * Gets the basic rate of flow of the river (upstream of
   * all headgates) at the specified time of day.
   * Default behavior interprets flow change linearly,
   * so that if previous day's flow is different from current
   * day's flow, the flow is assumed to have taken one day to
   * change from previous to current, and to have done so at a 
   * constant rate.
   * @param fractionalTimeOfDay time of day, expressed as
   * a value between zero and one representing a fraction
   * of a day
   * @return flow value, in million-acre-feet per day, at the given
   * time of day.
   */
  public double getFlowAtTime(double fractionalTimeOfDay){
    double currentFlow = getCurrentDayDailyFlowInMAF(); 
    if(currentFlow == previousDayDailyFlowInMAF) return currentFlow; 
    //if(previousDayDailyFlowInMAF == 0) return currentFlow;
    if(fractionalTimeOfDay < 0) fractionalTimeOfDay = 0;
    if(fractionalTimeOfDay > 1) fractionalTimeOfDay = 1;
    return previousDayDailyFlowInMAF + 
             (getCurrentDayDailyFlowInMAF() - previousDayDailyFlowInMAF) * fractionalTimeOfDay;
  }
  
  /**
   * Gets the flow, in million-acre-feet per day, at the given
   * time of day at the specified headgate. Because headgates
   * are assumed to remove a fraction of the flow, downstream
   * headgates will have flow rates lower than upstream ones.
   * This algorithm begins with the untouched flow and moves
   * sequentially down the list of headgates, removing flow at
   * each one and recalculating the amoung available at the next
   * one. 
   * @param HG The HeadGate for which flow is sought. 
   * @param fractionalTimeOfDay the time of day, expressed as a fraction
   * of a 24-hour day. If a value less than zero is passed, uses zero;
   * if a value greater than one is passed, uses 1.
   * @return the flow value at the requested HeadGate. If the HeadGate passed
   * is not one of the gates in this collection, returns zero.
   */
  public double getFlowAtHeadgateAtTime(HeadGate HG, double fractionalTimeOfDay){
    if(fractionalTimeOfDay < 0) fractionalTimeOfDay = 0;
    if(fractionalTimeOfDay > 1) fractionalTimeOfDay = 1;
    double flowAvailable = getFlowAtTime(fractionalTimeOfDay);
    for(int i = 0; i < HeadGates.size(); i++){
      HeadGate H = (HeadGate)HeadGates.get(i);
      if(H.equals(HG)) return flowAvailable;
      flowAvailable *= (1- H.getEfficiency(flowAvailable));
    }    
    return 0; // Headgate is not one of the list of headgates
  }
  
  /**
   * Gets the change in flow level at a given headgate starting
   * from the time specified and ending after the duration specified.
   * If the resulting end time is past a full day, assumes that the value
   * ending the day is appropriate (that is, will convert a time
   * of 1.1 days to 1.0).
   * @param HG The HeadGate for which flow change is sought
   * @param fractionalTimeOfDay start time
   * @param durationInSeconds duration, in seconds, for which flow change is sought
   * @return change in flow, in million-acre-feet per day
   */
  public double getFlowChangeOverInterval(HeadGate HG, double fractionalTimeOfDay, double durationInSeconds){
    double time1 = fractionalTimeOfDay;
    double time2 = UnitConversion.convertTime(fractionalTimeOfDay, durationInSeconds);
    double flowAtTime2 = getFlowAtHeadgateAtTime(HG, time2);
    double flowAtTime1 = getFlowAtHeadgateAtTime(HG, time1);
    
    double retVal = getFlowAtHeadgateAtTime(HG, UnitConversion.convertTime(fractionalTimeOfDay, durationInSeconds)) - 
        getFlowAtHeadgateAtTime(HG, fractionalTimeOfDay);
    return retVal;
  }
  
//  /**
//   * Returns the flow change over a given fraction of a day. 
//   * Each day of sim time the flow value for the previous day is recorded. The
//   * next day a new value can be set, but to realistically model
//   * the change in the river, this new value cannot be an instantaneous
//   * change, but must be assumed to take place over the course of a day.
//   * This method returns the amount of change that occurs during 
//   * a given interval. The interval should be passed as a fraction of
//   * a day.
//   */
//  public double getFlowChangeOverInterval(double interval){
//    if(interval <= 0 ) return 0;
//    if(interval >  1 ) return 0;
//    return (getCurrentDayDailyFlowInMAF() - previousDayDailyFlowInMAF) * interval;
//  }
//  
//  /**
//   * Returns the flow change over a given number of seconds.
//   * @param seconds
//   * @return the flow change per seconds specified
//   */  
//  public double getFlowChange(double seconds){
//    return getFlowChangeOverInterval(seconds / (60d * 60d * 24d));
//  }
  
  
  /* Step methods */
  
  /**
   * Steps the river forward by recording today's flow value
   * as the `previous' value.
   */
  public void step(){    
//    System.out.println("  River " + this.name + " stepping forward");
//    System.out.println("       Setting previousDayDailyFlow:");
//    System.out.println("           was " + previousDayDailyFlowInMAF);
//    System.out.println("           new " + getCurrentDayDailyFlowInMAF());
    previousDayDailyFlowInMAF = todaysDailyFlowInMAF;
    todaysDailyFlowInMAF = AnnualStreamflow * flowLevel;   
//    System.out.println("           New daily flow is " + this.todaysDailyFlowInMAF);
  }
  
  
  /* Reporting */
  
  /**
   * Gets a report about this river object
   * @return a report describing the river object
   */
  public String getReport(){ return "The value of code model option 'FlowLevel' is " + codeModelOption("FlowLevel");   }
  
  /* Spatial Element */
  /** 
   * @see abcm.graphics.SVGRepresentable#addToSVG(abcm.graphics.SVGAssist) 
   */
  public void addToSVG(SVGAssist svg){
    // If there are no headgates, there's nothing to display
    if(HeadGates.size() == 0){     return;     }

    // Add headgates as points
    for(int i = 0; i < HeadGates.size(); i++){
      HeadGate Gate1 = (HeadGate)HeadGates.get(i);
      Point P1 = Gate1.getLocus();
//      xml.addLine("<circle cx=\"" + P1.getX() + "\" cy=\"" + P1.getY() + "\" r=\"2\" fill=\"blue\"/>\n", true);
      svg.addCircle(P1.getX(), P1.getY(), svg.getUnitsPerInchX()/4, "red");
      svg.addCircle(P1.getX(), P1.getY(), svg.getUnitsPerInchX()/8, "yellow");
    }
    
    // If there are more than one headgates, add them as lines
    if(HeadGates.size() > 1){
//      xml.addLine("\n");
//      xml.addLine("<!-- River: " + name + " -->\n", XMLAssist.INDENT_BLOCK);
    
      for(int i = 0; i < HeadGates.size() - 1; i++){
        HeadGate Gate1 = (HeadGate)HeadGates.get(i);
        HeadGate Gate2 = (HeadGate)HeadGates.get(i + 1);
        Point P1 = Gate1.getLocus();
        Point P2 = Gate2.getLocus();
//        xml.addLine("line","x1=\"" + P1.getX() + "\" y1=\"" + P1.getY() + "\" x2=\"" + P2.getX() + "\" y2=\"" + P2.getY() + "\" style=\"stroke:rgb(0,0,255);stroke-width=:2\"","");
        svg.addLine(P1.getX(), P1.getY(), P2.getX(), P2.getY(), "rgb(0,0,255)", 2);
        
      }          
    }
    
    
  }
  
  /** 
   * @see abcm.space.SpatialElement#getBoundaryPolygons() 
   */
  public Vector getBoundaryPolygons(){    return new Vector();  }

  /** 
   * @see abcm.space.SpatialElement#getBermPolygons() 
   */
  public Vector getBermPolygons(){    return new Vector();  }
  
  /** 
   * @see abcm.space.SpatialElement#altered() 
   */
  public boolean altered(){    return true;  }
  
  /** 
   * @see abcm.space.SpatialElement#getMaxX() 
   */
  public double getMaxX(){    return 0;  }

  /** 
   * @see abcm.space.SpatialElement#getMaxY() 
   */
  public double getMaxY(){    return 0;  }

  /** 
   * @see abcm.space.SpatialElement#getMaxZ() 
   */
  public double getMaxZ(){    return 0;  }

  /** 
   * @see abcm.space.SpatialElement#getMinX() 
   */
  public double getMinX(){    return 0;  }

  /** 
   * @see abcm.space.SpatialElement#getMinY() 
   */
  public double getMinY(){    return 0;   }

  /** 
   * @see abcm.space.SpatialElement#getMinZ() 
   */
  public double getMinZ(){    return 0;   }

  /** 
   * @see abcm.graphics.X3DRepresentable#addToX3D(abcm.graphics.X3DAssist) 
   */
  public void addToX3D(X3DAssist x3d){
    // Not yet implemented... how cozy!
  }

  /** 
   * @see abcm.space.HasElevation#getZ(abcm.math.Point) 
   */
  public double getZ(Point P){
    return 0;
  }
  
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    River ret = new River(this.name);
    cloneStore.put(this, ret); // Must be added before cloning headgates!
    
    ret.flowLevel = flowLevel;
    ret.AnnualStreamflow = AnnualStreamflow;
    ret.previousDayDailyFlowInMAF = previousDayDailyFlowInMAF;
    
    ret.HeadGates    = (ABCMDeepCloneableVector)HeadGates.getDeepClone(cloneStore);
    ret.HeadGateList = (ABCMDeepCloneableHashtable)HeadGateList.getDeepClone(cloneStore);
    ret.waterSample  = (Water)cloneStore.getClone(waterSample);
    
    return ret;
  }
  
  /**
   * @see abcm.space.SpatialElement#getTriangulatedTopographyWithinBoundaryPolygons()
   */
  public TriangulatedTopography getTriangulatedTopographyWithinBoundaryPolygons(){
    return new TriangulatedTopography(new Vector());
  }
  
}
