package flowag.channels;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.math.SimDate;
import abcm.sim.SimulationDataSeries;
import abcm.sim.SimulationDatum;

/**
 * A BasicHydrographEngine that records a graph of water levels (stage and flow) for a day 
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class SingleDayHydrographEngine implements Hydrographable, ABCMDeepCloneable{
  
  /** Upstream stage hydrograph  */
  private SimulationDataSeries upstreamStageHydrograph   = new SimulationDataSeries("z1", true);
  
  /** Upstream flow hydrograph */
  private SimulationDataSeries upstreamFlowHydrograph    = new SimulationDataSeries("Q1", true);
  
  /** Downstream stage hydrograph */
  private SimulationDataSeries downstreamStageHydrograph = new SimulationDataSeries("z2", true);
  
  /** Downstream flow hydrograph */
  private SimulationDataSeries downstreamFlowHydrograph  = new SimulationDataSeries("Q2", true);

  /**
   * @see flowag.channels.Hydrographable#clearHydrographs()
   */
  public void clearHydrographs(){
    upstreamStageHydrograph   = new SimulationDataSeries("z1", true);
    upstreamFlowHydrograph    = new SimulationDataSeries("Q1", true);
    downstreamStageHydrograph = new SimulationDataSeries("z2", true);
    downstreamFlowHydrograph  = new SimulationDataSeries("Q2", true);
  }
    
  /**
   * @see flowag.channels.Hydrographable#resetHydrographs()
   */
  public void resetHydrographs(){
    if(upstreamStageHydrograph.getSeries().size()>0){
      double uz = ((SimulationDatum)upstreamStageHydrograph.getSeries().get(upstreamStageHydrograph.getSeries().size() - 1)).getNumericValue();
      upstreamStageHydrograph   = new SimulationDataSeries("z1", true);
      upstreamStageHydrograph.addElement(new SimDate("01-JAN-SEASONS +0.0"), uz);
    }
    else upstreamStageHydrograph = new SimulationDataSeries("z1", true);
    
    if(upstreamFlowHydrograph.getSeries().size()>0){
      double uQ = ((SimulationDatum)upstreamFlowHydrograph.getSeries().get(upstreamFlowHydrograph.getSeries().size() - 1)).getNumericValue();
      upstreamFlowHydrograph   = new SimulationDataSeries("Q1", true);
      upstreamFlowHydrograph.addElement(new SimDate("01-JAN-SEASONS +0.0"), uQ);
    }
    else upstreamFlowHydrograph = new SimulationDataSeries("Q1", true);
    
    if(downstreamStageHydrograph.getSeries().size()>0){
      double dz = ((SimulationDatum)downstreamStageHydrograph.getSeries().get(downstreamStageHydrograph.getSeries().size() - 1)).getNumericValue();
      downstreamStageHydrograph   = new SimulationDataSeries("z2", true);
      downstreamStageHydrograph.addElement(new SimDate("01-JAN-SEASONS +0.0"), dz);
    }
    else downstreamStageHydrograph = new SimulationDataSeries("z2", true);
    
    if(downstreamFlowHydrograph.getSeries().size()>0){
      double dQ = ((SimulationDatum)downstreamFlowHydrograph.getSeries().get(downstreamFlowHydrograph.getSeries().size() - 1)).getNumericValue();
      downstreamFlowHydrograph   = new SimulationDataSeries("Q2", true);
      downstreamFlowHydrograph.addElement(new SimDate("01-JAN-SEASONS +0.0"), dQ);
    }
    else downstreamFlowHydrograph = new SimulationDataSeries("Q2", true);
    
  }

  /**
   * Converts one SimDate to another; the new SimDate contains only the fractional part of the first
   * @param original the original SimDate
   * @return the new SimDate
   */
  private SimDate convertSimDate(SimDate original){
    return (original.value() > 1 ? new SimDate(original.getFraction()) : original); 
  }
  
  /**
   * @see flowag.channels.Hydrographable#getUpstreamStageAt(abcm.math.SimDate)
   */
  public double getUpstreamStageAt(SimDate SD){   return   upstreamStageHydrograph.getLastValueBefore(convertSimDate(SD)); }
  
  /**
   * @see flowag.channels.Hydrographable#getDownstreamStageAt(abcm.math.SimDate)
   */
  public double getDownstreamStageAt(SimDate SD){ return downstreamStageHydrograph.getLastValueBefore(convertSimDate(SD)); }
  
  /**
   * @see flowag.channels.Hydrographable#getUpstreamFlowAt(abcm.math.SimDate)
   */
  public double getUpstreamFlowAt(SimDate SD){    return    upstreamFlowHydrograph.getLastValueBefore(convertSimDate(SD)); }
  
  /**
   * @see flowag.channels.Hydrographable#getDownstreamFlowAt(abcm.math.SimDate)
   */
  public double getDownstreamFlowAt(SimDate SD){  return  downstreamFlowHydrograph.getLastValueBefore(convertSimDate(SD)); }

  
  /**
   * @see flowag.channels.Hydrographable#setUpstreamStage(double, abcm.math.SimDate)
   */
  public void setUpstreamStage(double z, SimDate SD){
    upstreamStageHydrograph.addElement(convertSimDate(SD), z);
  }
  
  /**
   * @see flowag.channels.Hydrographable#setDownstreamStage(double, abcm.math.SimDate)
   */
  public void setDownstreamStage(double z, SimDate SD){
    downstreamStageHydrograph.addElement(convertSimDate(SD), z);
  }
  
  /**
   * @see flowag.channels.Hydrographable#setUpstreamDischarge(double, abcm.math.SimDate)
   */
  public void setUpstreamDischarge(double Q, SimDate SD){
    upstreamFlowHydrograph.addElement(convertSimDate(SD), Q);
  }
  
  /**
   * @see flowag.channels.Hydrographable#setDownstreamDischarge(double, abcm.math.SimDate)
   */
  public void setDownstreamDischarge(double Q, SimDate SD){
    downstreamFlowHydrograph.addElement(convertSimDate(SD), Q);
  }
  
  /**
   * Gets the upstream stage at the specified time. Interpolates
   * between the two closest times recorded if it does not have a record at the
   * exact time requested.
   * @param time the time for which data are requested
   * @return the value requested, either stored or calculated
   */
  public double getUpstreamStageAt(double time){
    SimDate SD = new SimDate(time % 1);
    return upstreamStageHydrograph.getScaledValueAt(SD);
  }
  
  /**
   * Gets the upstream discharge at the specified time. Interpolates
   * between the two closest times recorded if it does not have a record at the
   * exact time requested.
   * @param time the time for which data are requested
   * @return the value requested, either stored or calculated
   */
  public double getUpstreamDischargeAt(double time){
    SimDate SD = new SimDate(time % 1);
    return upstreamFlowHydrograph.getScaledValueAt(SD);
  }
  
  /**
   * Gets the downstream stage at the specified time. Interpolates
   * between the two closest times recorded if it does not have a record at the
   * exact time requested.
   * @param time the time for which data are requested
   * @return the value requested, either stored or calculated

   */
  public double getDownstreamStageAt(double time){
    SimDate SD = new SimDate(time % 1);
    return downstreamStageHydrograph.getScaledValueAt(SD);
  }
  
  /**
   * Gets the downstream discharge at the specified time. Interpolates
   * between the two closest times recorded if it does not have a record at the
   * exact time requested.
   * @param time the time for which data are requested
   * @return the value requested, either stored or calculated
   */
  public double getDownstreamDischargeAt(double time){
    SimDate SD = new SimDate(time % 1);
    return downstreamFlowHydrograph.getScaledValueAt(SD);
  }
  
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    SingleDayHydrographEngine newEngine = new SingleDayHydrographEngine();
    cloneStore.put(this, newEngine);
    newEngine.upstreamStageHydrograph   = (SimulationDataSeries)cloneStore.getClone(upstreamStageHydrograph);
    newEngine.upstreamFlowHydrograph    = (SimulationDataSeries)cloneStore.getClone(upstreamFlowHydrograph);
    newEngine.downstreamStageHydrograph = (SimulationDataSeries)cloneStore.getClone(downstreamStageHydrograph);
    newEngine.downstreamFlowHydrograph  = (SimulationDataSeries)cloneStore.getClone(downstreamFlowHydrograph);
    return newEngine;
  }
  
}
