package flowag.channels;

import java.util.Vector;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.math.BasicMath;
import abcm.math.Geometry;
import abcm.math.Point;

/**
 * A TrapezoidalChannelCrossSection implements the ChannelCrossSection
 * interface and provides the formula to calculate basic hydraulic values
 * given a trapezoidal channel geometry.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class TrapezoidalChannelCrossSection extends AbstractChannelCrossSection{
  
  /** top width  */
  private double tWidth;
  /** bottom width  */
  private double bWidth;
  /** channel height  */
  private double chHeight;
  
  /* Static Methods */

  /**
   * Returns the width of the water surface, given a specified channel geometry
   * and a water depth.
   * 
   * This is calculated by:
   * 
   * bottomWidth + (waterDepth / channelHeight) * (topWidth - bottomWidth)
   * 
   * Note: Will assume that the channel continues above the specified channel
   * height; that is, if waterDepth is greater than channelHeight, extrapolates
   * the slope upward beyond the actual channel bed. Returns 0 if waterDepth is less
   * than or equal to zero.
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which the surface 
   * width is to be calculated.
   * @return the width of the surface of the water
   * 
   */
  public static double getWaterSurfaceWidth(double topWidth, 
                                            double bottomWidth, 
                                            double channelHeight, 
                                            double waterDepth){
    
    if(waterDepth <= 0) return 0; 
    return bottomWidth + (waterDepth / channelHeight) * (topWidth - bottomWidth);
  }
  
  /**
   * Returns the wetted perimeter of the cross section, given a specified channel
   * geometry and a water depth.
   * 
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which the wetted perimeter
   * is to be calculated
   * @return the wetted perimeter
   */
  public static double getWettedPerimeter(double topWidth,
      double bottomWidth,
      double channelHeight,
      double waterDepth){
      double z = waterDepth;
    if(z <= 0) return 0;
    double bottomLeg =(z * ((topWidth - bottomWidth)/ (2 * channelHeight))); 
    double wetSide = Math.sqrt(bottomLeg * bottomLeg + z * z);
    return 2 * wetSide + bottomWidth;
  }
  
  /**
   * Returns the cross-sectional area of flow given the specified channel
   * geometry and a depth of water.
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which the flow area is
   * to be calculated
   * @return the cross-sectional area of the water flow
   */
  public static double getFlowArea(double topWidth, 
      double bottomWidth, 
      double channelHeight, 
      double waterDepth){
    if(waterDepth <= 0) return 0;
    double surfaceWidth = getWaterSurfaceWidth(topWidth, bottomWidth, channelHeight, waterDepth);
    return waterDepth * ((surfaceWidth + bottomWidth)/2d);
  }
  
  /**
   * Returns the Average Depth for the channel given a specified geometry and a
   * water depth.
   * 
   * The Average Depth is not, as it might seem, the simple average depth of the
   * channel across its cross-section. It is instead defined as the flow area divided
   * by the surface width.
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which the average depth is to
   * be calculated.
   * @return the average depth of the cross section
   */
  public static double getAvgDepth(double topWidth, 
                            double bottomWidth, 
                            double channelHeight, 
                            double waterDepth){
    double A = getFlowArea(topWidth, bottomWidth, channelHeight, waterDepth);
    double W = getWaterSurfaceWidth(topWidth, bottomWidth, channelHeight, waterDepth);
    if(W == 0) return 0;
    return A / W;
  }
  
  /**
   * Returns the hydraulic radius of the channel given a specified geometry
   * and a depth of water.
   * 
   * The hydraulic radius is defined as the flow area divided by the wetted perimeter.
   * 
   * Returns 0 if the water depth is less than or equal to zero.
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which the hydraulic radius is
   * to be calculated
   * @return the hydraulic radius
   */
  public static double getHydraulicRadius(double topWidth,
                                          double bottomWidth,
                                          double channelHeight,
                                          double waterDepth){
    if(waterDepth <= 0 ) return 0.0;
    double flowArea = getFlowArea(topWidth, bottomWidth, channelHeight, waterDepth);
    double wettedPerimeter = getWettedPerimeter(topWidth, bottomWidth, channelHeight, waterDepth);
    return flowArea/wettedPerimeter;
  }
  
  /**
   * Returns the conveyance of the channel given a specified geometry, a 
   * water depth, and a roughness coefficient.
   * 
   * Conveyance is defined to be:
   * 
   *  K = (A * (R ^ (2/3))) / n
   *  
   *  where A is the flow area, R is the hydraulic radius, and n is the
   *  roughness coefficient.
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which the conveyance is
   * to be calculated.
   * @param roughness roughness coefficient from which conveyance is to
   * be calculated.
   * @return the conveyance of a cross-section with 
   * the specified characteristics 
   */
  public static double getConveyance(double topWidth, 
                                     double bottomWidth, 
                                     double channelHeight, 
                                     double waterDepth, 
                                     double roughness){
    if(roughness <= 0) return 0;
    double A = getFlowArea(topWidth, bottomWidth, channelHeight, waterDepth);
    double R = getHydraulicRadius(topWidth, bottomWidth, channelHeight, waterDepth);
    return (A * (BasicMath.cbrt(R * R))) / roughness;
  }
    
  /**
   * Returns the change in area given a change in water depth.
   * 
   * This term is calculated as:
   * 
   * (((topWidth - bottomWidth)/channelHeight) * waterDepth) + bottomWidth
   * 
   * Returns zero if the water depth specified is less than or equal to zero.
   * 
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which the DADz
   * values is to be calculated
   * @return the change in flow area given a change in water depth.
   */
  public static double getDADz(double topWidth,
                               double bottomWidth,
                               double channelHeight,
                               double waterDepth){
    if(waterDepth <= 0) return 0;
    return bottomWidth + 
    ((topWidth - bottomWidth) / channelHeight) * waterDepth;    
  }
  
  /**
   * Returns the change in conveyance given a change in water depth in
   * a specified channel geometry and water depth
   * 
   * See HWM Documentation for derivation
   * 
   * Note that if water depth is zero, dKdz is calculated for water depth = 1 cm.
   * 
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which DKDz 
   * is to be calculated.
   * @param roughness roughness coefficient
   * @return the change in conveyance per change in water depth
   */
  public static double getDKDz(double topWidth, 
                               double bottomWidth, 
                               double channelHeight, 
                               double waterDepth, 
                               double roughness){
    double t = topWidth;
    double b = bottomWidth;
    double z = (waterDepth > 0 ? waterDepth : .01);
    double d = channelHeight;
    double n = roughness;
    double tmb = t - b;
    
    // Constants
    double C1 = Math.sqrt(1.0d + ((tmb * tmb) / (4.0d * (d * d))));
    double C2 = tmb / 2.0d;
    
    // Common terms
    double term1 = b + 2.0d * C1 * z;
    double term2 = b * z + C2 * z * z;
    
    return ( 1.0d / n ) *
           ((
             Math.pow(term1, 2d/3d) *
             (5d/3d) * Math.pow(term2, 2d/3d) * 
             term1
           -
             Math.pow(term2, 5d/3d) *
             (2.0d/3.0d) * Math.pow(term1, (-1d/3d)) *
             2.0d * C1
           )
           /
           Math.pow(term1, (4d/3d)));
  }

  /**
   * Returns the discharge for the given channel geometry and hydraulic values,
   * including channel slope, as calculated using Manning's formula.
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which discharge
   * is to be calculated.
   * @param roughness Manning's n
   * @param slope the channel slope for which this is being calculated
   * @return the Manning Discharge; see the HWM documentation for formulas
   */
  public static double getManningDischarge(double topWidth, double bottomWidth, 
      double channelHeight, double waterDepth, double roughness, double slope){
    return getFlowArea(topWidth, bottomWidth, channelHeight, waterDepth) * 
           getManningVelocity(topWidth, bottomWidth, channelHeight, waterDepth, roughness, slope);
  }
  
  /**
   * Returns the flow velocity for the given channel geometry and hydraulic values,
   * including channel slope, as calculated using Manning's formula.
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param waterDepth depth of the flow of water for which discharge
   * is to be calculated.
   * @param roughness Manning's n
   * @param slope the channel slope for which this is being calculated
   * @return the flow velocity; see the HWM documentation for formulas
   */
  public static double getManningVelocity(double topWidth, double bottomWidth, 
      double channelHeight, double waterDepth, double roughness, double slope){
    double R = getHydraulicRadius(topWidth, bottomWidth, channelHeight, waterDepth);
    return BasicMath.cbrt(Math.pow(R,2)) * StrictMath.sqrt(Math.abs(slope)) / roughness;
  }
    
  /**
   * Returns the friction slope at this cross section given specified channel
   * geometry, discharge, water depth, and roughness.
   * 
   * Note that this friction slope is only valid under normal flow conditions,
   * which do not apply. 
   * 
   * @param topWidth width of the top of the cross-section
   * @param bottomWidth width of the bottom of the cross-section
   * @param channelHeight height of the channel from top to bottom
   * @param discharge current discharge, in m^3/s
   * @param waterDepth depth of the flow of water
   * @param roughness roughness coefficient
   * @return calculated friction slope
   */
  public static double getFrictionSlope(double topWidth, 
      double bottomWidth, double channelHeight, 
      double discharge, double waterDepth, double roughness){
    // If water depth is zero, return zero
    if(waterDepth == 0) return 0;
    double R = getHydraulicRadius(topWidth, bottomWidth, channelHeight, waterDepth);
    double A = getFlowArea(topWidth, bottomWidth, channelHeight, waterDepth);
    double Q = discharge;
    double n = roughness;
//    System.out.println("Getting friction slope:\n  R= " + R + "\n  A=" + A + "\n  Q=" + Q + "\n  n=" + n );
    double fs =Q * Math.abs(Q)* Math.pow(n,2) / ((BasicMath.cbrt(Math.pow(R,4.0d)) * Math.pow(A,2)));
//    System.out.println("Friction Slope = " + fs);
    return fs;
  }
  
  /* Constructor */
  /**
   * Creates an instance of the cross-section object with the specified dimensions.
   * Initializes upstream and downstream depth and discharge to zero (no water present).
   * @param topWidth width at the top of the cross-section
   * @param bottomWidth width at the bottom of the cross-section
   * @param channelHeight height (construction depth) of the channel
   */
  public TrapezoidalChannelCrossSection(double topWidth,double bottomWidth,double channelHeight){
    if(bottomWidth <0)          bottomWidth   = 0;
    if(topWidth < bottomWidth)  topWidth      = bottomWidth;
    if(topWidth == 0)           topWidth      = 1;
    if(channelHeight <= 0)      channelHeight = 1;
    tWidth = topWidth ;
    bWidth = bottomWidth;
    chHeight = channelHeight;
  }
  
  /* Getters and Setters */  
  /**
   * @see flowag.channels.ChannelCrossSection#getTopWidth()
   */
  public double getTopWidth()                        { return tWidth;                      }
  
  /**
   * @see flowag.channels.ChannelCrossSection#getBottomWidth()
   */
  public double getBottomWidth()                     { return bWidth;                      }
  
  /**
   * @see flowag.channels.ChannelCrossSection#getChannelHeight()
   */
  public double getChannelHeight()                   { return chHeight;                    }
 
  /**
   * @see flowag.channels.ChannelCrossSection#setTopWidth(double)
   */
  public void setTopWidth(double topWidth)           { tWidth   = topWidth;                }
  
  /**
   * Sets the bottom width of the cross-section
   * @param  bottomWidth the new width
   */
  public void setBottomWidth(double bottomWidth)     { bWidth   = bottomWidth;             }  
  
  /**
   * Sets the channel height for this cross-section
   * @param channelHeight the new height
   */
  public void setChannelHeight(double channelHeight) { chHeight = channelHeight;           }
  
  /**
   * @see flowag.channels.ChannelCrossSection#getArea()
   */
  public double getArea()                            { return getFlowArea(chHeight);       }
  
  
  /* Hydraulic Values */
  
  
  /** @see flowag.channels.ChannelCrossSection#getWaterSurfaceWidth(double) */
  public double getWaterSurfaceWidth(double waterDepth){ 
    return getWaterSurfaceWidth(tWidth, bWidth, chHeight, waterDepth);  
  }
  
  /** @see flowag.channels.ChannelCrossSection#getAvgDepth(double)          */
  public double getAvgDepth(double waterDepth){          
    return getAvgDepth(tWidth, bWidth, chHeight, waterDepth);           
  }
  
  /** @see flowag.channels.ChannelCrossSection#getFlowArea(double)   */
  public double getFlowArea(double waterDepth){    
    return getFlowArea(tWidth, bWidth, chHeight, waterDepth);  
  }
  
  /** @see flowag.channels.ChannelCrossSection#getWettedPerimeter(double) */
  public double getWettedPerimeter(double waterDepth){    
    return getWettedPerimeter(tWidth, bWidth, chHeight, waterDepth);  
  }

  /** @see flowag.channels.ChannelCrossSection#getHydraulicRadius(double) */
  public double getHydraulicRadius(double waterDepth){    
    return getHydraulicRadius(tWidth, bWidth, chHeight, waterDepth);  
  }
  
  /** @see flowag.channels.ChannelCrossSection#getDADz(double) */
  public double getDADz(double waterDepth){    
    return getDADz(tWidth, bWidth, chHeight, waterDepth);  
  }
  
  /** @see flowag.channels.ChannelCrossSection#getConveyance(double, double)  */
  public double getConveyance(double waterDepth, double roughness){    
    return getConveyance(tWidth, bWidth, chHeight, waterDepth, roughness);  
  }
  
  /** @see flowag.channels.ChannelCrossSection#getDKDz(double, double)    */
  public double getDKDz(double waterDepth, double roughness){    
    return getDKDz(tWidth, bWidth, chHeight, waterDepth, roughness);  
  }
  
  /** @see flowag.channels.ChannelCrossSection#getManningVelocity(double, double, double)  */
  public double getManningVelocity(double waterDepth, double roughness, double slope){
    return getManningVelocity(tWidth, bWidth, chHeight, waterDepth, roughness, slope);
  }
  

  /** @see flowag.channels.ChannelCrossSection#getManningDischarge(double, double, double)   */
  public double getManningDischarge(double waterDepth, double roughness, double slope){
    return getManningDischarge(tWidth, bWidth, chHeight, waterDepth, roughness, slope);
  }
  
    
  /** @see flowag.channels.ChannelCrossSection#getFrictionSlope(double, double, double)   */
  public double getFrictionSlope(double discharge, double waterDepth, double roughness){
    return getFrictionSlope(tWidth, bWidth, chHeight, discharge, waterDepth, roughness);
  }
  
  /**
   * @see flowag.channels.ChannelCrossSection#centroidDepth(double)
   */
  public double centroidDepth(double waterDepth){                             // This is clearly a bug...
    double topWidth = this.getWaterSurfaceWidth(waterDepth);
    return ((bWidth + 2 * tWidth) / (3 * (tWidth + bWidth))) * waterDepth;
  }
  
  /* HEC-RAS Output */

  /** 
   * @see flowag.channels.ChannelCrossSection#getCutAndSurfaceLine(abcm.math.Point, double) 
   */
  public Vector getCutAndSurfaceLine(Point origin, double channelHeading){
    Vector v = new Vector();

    double leftHeading  = channelHeading - (Math.PI/2);
    double rightHeading = channelHeading + (Math.PI/2);
    
    Point upper = new Point(origin.getX(), origin.getY(), origin.getZ() + chHeight);  
    // Top of left side of bank, upstream end
    v.add(Geometry.getOffsetPoint(upper, leftHeading, tWidth * .5));    
    // Bottom left edge of canal trough, upstream end
    v.add(Geometry.getOffsetPoint(origin, leftHeading, bWidth * .5));
    // Bottom right edge of canal trough, upstream end
    v.add(Geometry.getOffsetPoint(origin, rightHeading, bWidth * .5));
    // Top of right side of bank, upstream end
    v.add(Geometry.getOffsetPoint(upper, rightHeading, tWidth * .5));
    
    return v;
  }
  
  /**
   * Fill in the bottom section of the canal.
   * @param fillDepth depth by which the channel should be filled in
   */
  public void fill(double fillDepth){
    double newBottomWidth = getWaterSurfaceWidth(fillDepth);
    double newChannelHeight = chHeight - fillDepth;
    bWidth   = newBottomWidth;
    chHeight = newChannelHeight;
  }

  /**
   * @see flowag.channels.ChannelCrossSection#dig(double)
   */
  public void dig(double depth){
    if(depth < 0) return;
    double newBottomWidth = tWidth - (tWidth - bWidth) * ((chHeight + depth) / chHeight);
    if(newBottomWidth < 0) return;
    bWidth = newBottomWidth;
    chHeight = chHeight + depth;
  }

  /**
   * @see flowag.channels.ChannelCrossSection#type()
   */
  public String type(){
    return "Trapezoidal";
  }

  /**
   * @see flowag.channels.ChannelCrossSection#getCopy()
   */
  public ChannelCrossSection getCopy(){
    return new TrapezoidalChannelCrossSection(tWidth, bWidth, chHeight);
  }

  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    TrapezoidalChannelCrossSection clone = new TrapezoidalChannelCrossSection(tWidth, bWidth, chHeight);
    cloneStore.put(this, clone);
    return clone;
  }
  
}
