package wm.db;


import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import wm.db.tables.WMPlantConfigurationChemicalEffectsTable;
import wm.db.tables.WMPlantConfigurationStagesTable;
import wm.db.views.WMPlantConfigsAndResearchersView;
import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.managers.CommentManager;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.math.BasicMath;
import abcm.utilities.XMLAssist;

/**
 * A PlantConfigurationManager acts as an interface to storing and 
 * retrieving Plant Configurations from an ABCM Database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class PlantConfigurationManager extends AbstractAuditableDBManager {

  /** Name of plant configuration to be saved                              */
  private String                                  nameToSave;
  /** Description of plant configuration to be saved                       */
  private String                                  descriptionToSave;
  /** ID of user attempting to save the configuration                      */
  private long                                    userIDToSave;
  /** PK of plant to which the configuration to be saved applies           */
  private long                                    plantIDToSave;
  /** PK of truth value to be associated with this configuration           */
  private long                                    truthValueToSave;
  
  /** Internal; plant configuration stage manager object    */
  private PlantConfigurationStageManager          PCSM;
  /** Internal; plant configuration chemical effect manager */
  private PlantConfigurationChemicalEffectManager PCCEM;
  
  /**
   * Database Manager for Plant Configuration Stages 
   *
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class PlantConfigurationStageManager extends AbstractDBLinkTableInnerManager{

    /** FK to plant configuration to save  */
    private long   plantConfigFKToSave; 
    /** Stage Order to save  */
    private int    orderToSave; 
    /** Stage Duration to save  */
    private int    durationToSave; 
    /** Crop Factor to save  */
    private double cropFactorToSave; 
    /** Drought Sensitivity to Save  */
    private double droughtSensitivityToSave; 
    /** Labor value to save  */
    private double laborToSave;
    
    /**
     * Constructor
     * @param db Instance of an ABCM Gateway to provide database access.
     * @param containedBy Enclosing class
     */
    public PlantConfigurationStageManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable() 
     */
    protected ABCMdbTable baseTable(){                     return DB().getPlantConfigurationStagesTable();                                     }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField() 
     */
    protected ABCMdbField baseTablePKField(){              return DB().getPlantConfigurationStagesTable().getField_PK();                       }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField() 
     */
    protected ABCMdbField baseTablePrimaryFKField(){       return DB().getPlantConfigurationStagesTable().getField_PlantConfigurationFK();     }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField() 
     */
    protected ABCMdbField baseTableSecondaryFKField(){     return DB().getPlantConfigurationStagesTable().getField_Order();                    }

    
    /* XML Creation */
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
     */
    protected ABCMdbAbstractViewable recordsView(){        return DB().getPlantConfigurationStagesTable();                                    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
     */
    protected ABCMdbField recordsViewPKField(){            return DB().getPlantConfigurationStagesTable().getField_PK();                      }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField() 
     */
    protected ABCMdbField recordsViewPrimaryFKField(){     return DB().getPlantConfigurationStagesTable().getField_PlantConfigurationFK();    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField() 
     */
    protected ABCMdbField recordsViewPrimaryNameField(){   return DB().getPlantConfigurationStagesTable().getField_Order();                   }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField() 
     */
    protected ABCMdbField recordsViewSecondaryNameField(){ return null;    }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
      
      WMPlantConfigurationStagesTable table = DB().getPlantConfigurationStagesTable();
      
      int orderFieldCol                  = rs.findColumn(table.getField_Order().getName());
      int durationFieldCol               = rs.findColumn(table.getField_Duration().getName());
      int cropFactorFieldCol             = rs.findColumn(table.getField_CropFactor().getName());
      int droughtSensitivityFieldCol     = rs.findColumn(table.getField_DroughtSensitivity().getName());
      int laborRequiredFieldCol          = rs.findColumn(table.getField_Labor().getName());
      
      
      while(rs.next()){
        xml.open("stage");
          xml.addLine("order",              rs.getLong(orderFieldCol));
          xml.addLine("duration",           rs.getLong(durationFieldCol));
          xml.addLine("cropFactor",         BasicMath.roundDouble(rs.getDouble(cropFactorFieldCol), 5));
          xml.addLine("droughtSensitivity", BasicMath.roundDouble(rs.getDouble(droughtSensitivityFieldCol), 5));
          xml.addLine("laborRequired",      rs.getDouble(laborRequiredFieldCol));
        xml.closeTag();
      }
      
    }
    
    
    /* Saving */
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
     */
    protected void clearInstanceToSave(){
      plantConfigFKToSave      = -1; 
      orderToSave              = -1; 
      durationToSave           = -1; 
      cropFactorToSave         = -1; 
      droughtSensitivityToSave = -1; 
      laborToSave              = -1;
    }
    
    /**
     * Sets an instance of a Plant Configuration Stage to save
     * @param plantConfig PK to the plant configuration
     * @param order order of stage
     * @param duration duration of stage
     * @param cropFactor crop factor of stage
     * @param droughtSensitivity drought sensitivity to save
     * @param labor labor value to save
     */
    public void setInstanceToSave(long plantConfig, int order, int duration, double cropFactor, double droughtSensitivity, double labor){
      plantConfigFKToSave      = plantConfig; 
      orderToSave              = order; 
      durationToSave           = duration; 
      cropFactorToSave         = cropFactor; 
      droughtSensitivityToSave = droughtSensitivity; 
      laborToSave              = labor;        
    }
    
    /**
     * Determines if the instance to be saved is value
     * @return true if the instance to be saved has been set, false otherwise
     */
    public boolean instanceToSaveIsValid(){
      return ((plantConfigFKToSave      != -1) && 
              (orderToSave              != -1) && 
              (durationToSave           != -1) &&
              (cropFactorToSave         != -1) &&
              (droughtSensitivityToSave != -1) &&
              (laborToSave              != -1));
    }
    
    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
     */
    protected ABCMdbDatum datumToSave(){      
      return 
         (instanceToSaveIsValid() ? 
             DB().getPlantConfigurationStagesTable().getDatum("" + plantConfigFKToSave, 
                                                              "" + orderToSave, 
                                                              "" + durationToSave, 
                                                              "" + cropFactorToSave, 
                                                              "" + droughtSensitivityToSave, 
                                                              "" + laborToSave) :
             null);
      }

    /** 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#saveInstanceToDB() 
     */
    public boolean saveInstanceToDB(){
      long plantConfig = plantConfigFKToSave;
      return instanceToSaveIsValid()                                            && 
             PCSM.incrementStageNumbersOver(plantConfigFKToSave, orderToSave)   &&
             super.saveInstanceToDB()                                           &&
             collapseStageOrders(plantConfig);
    }
    
    /**
     * Increments the stage order numbers over a given lowest number.
     * This allows for another stage to be inserted
     * @param plantConfigID PK of the plant configuration to be adjusted
     * @param newStage new stage number; all stages numbered equal to and above this will
     * be adjusted upward
     * @return TISFO
     */
    public boolean incrementStageNumbersOver(long plantConfigID, long newStage){
      boolean successFlag = false;
      
      try{
        String plantConfigurationStagesTable = DB().getPlantConfigurationStagesTable().getFullName(DB.getOwner());
        String intOrder                      = DB().getPlantConfigurationStagesTable().getField_Order().getFullFieldName(DB.getOwner());
        String intPlantConfigurationID       = DB().getPlantConfigurationStagesTable().getField_PlantConfigurationFK().getFullFieldName(DB.getOwner());
        
        String sql = 
           "UPDATE " + plantConfigurationStagesTable + " " +
             "SET " + intOrder + " = " + intOrder + " + 1 " +
             "WHERE " + intPlantConfigurationID  + " = " + plantConfigID + " " +
             "AND " + intOrder + " >= " + newStage +
             ";"; 
        
        Connection con = DB.getAdminConnection();
        Statement st = con.createStatement();
        
        st.executeUpdate(sql);
        
        successFlag = true;
      }
      catch(Exception e){
        addToErrorReport("Error performing update of stage orders: " + e.getMessage());
      }
      
      return successFlag;
    }

    /**
     * Collapses the stage orders for a given plant configuration, so that any
     * gaps (i.e. 1, 2, 4, 5) are eliminated
     * @param plantConfigurationID plant configuration to be adjusted
     * @return TISFO
     */
    public boolean collapseStageOrders(long plantConfigurationID){
      boolean successFlag = false;
      
      try {
        String plantConfigurationStagesTable = DB().getPlantConfigurationStagesTable().getFullName(DB.getOwner());
        String intOrder                      = DB().getPlantConfigurationStagesTable().getField_Order().getFullFieldName(DB.getOwner());
        String intPlantConfigurationID       = DB().getPlantConfigurationStagesTable().getField_PlantConfigurationFK().getFullFieldName(DB.getOwner());
        
        
        String sqlGetList = DB().getPlantConfigurationStagesTable().getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(intPlantConfigurationID, plantConfigurationID) + " " + AbstractDBGateway.orderBy(intOrder));
        
        Connection con = DB.getGuestConnection();
        Statement stGetList = con.createStatement();
        Statement stUpdate = con.createStatement();
        
        ResultSet rs = stGetList.executeQuery(sqlGetList);
        long instrCount = 1;
        while(rs.next()){
          if(rs.getInt("intOrder")!= instrCount){
            String sqlUpdate = 
                "UPDATE " + plantConfigurationStagesTable + "  " +
                "SET " + intOrder + "=" + instrCount + " " +
                "WHERE " + intPlantConfigurationID + " = " + plantConfigurationID + " " +
                "AND " + intOrder + " = " + rs.getLong("intOrder") + 
                ";";
            stUpdate.executeUpdate(sqlUpdate);
          }
          instrCount++;
        }
        
        successFlag = true;
      }
      catch (Exception e){
        addToErrorReport("Error in Plant Configuration Manager: Collapse Stages(): " + e.getMessage());
      }
      
      return successFlag; 
    }
    
    /* Deleting */
    /**
     * Delets a plant configuration stage. Collapses stage orders afterward, to eliminate
     * gaps
     * @param plantConfigurationID PK of plant configuration to be adjusted 
     * @param order order number to be deleted
     * @return TISFO
     */
    public boolean deleteStage(long plantConfigurationID, long order){
      boolean successFlag = super.deleteElementsByBothPrimaryKeys(plantConfigurationID, order);
      if(successFlag) collapseStageOrders(plantConfigurationID);
      return successFlag;
    }
    
  }

  /**
   * Manages the table that lists chemical effects on plants
   * 
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  private class PlantConfigurationChemicalEffectManager extends AbstractDBLinkTableInnerManager{
    /** PK of the plant configuration associated with the chemical effect to be saved*/
    private long   PKOfConfigToSave;
    /** name of the chemical for which this is the effect*/
    private String chemicalNameToSave;
    /** value for the plant's need (or impact) of the effect to be saved*/
    private double needToSave;
    /** value for the plant's consumption (or emission) of the effect to be saved*/
    private double consumptionToSave;

    /**
     * Constructor
     * @param db DB Gateway
     * @param containedBy pointer to the object within which this is nested
     */
    public PlantConfigurationChemicalEffectManager(AbstractABCMDBGateway db, AbstractDBManager containedBy){
      super(db, containedBy);
    }
    
    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable()
     */
    protected ABCMdbTable baseTable(){                     return DB().getPlantConfigurationChemicalEffectsTable();                                 }
    
    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField()
     */
    protected ABCMdbField baseTablePKField(){              return DB().getPlantConfigurationChemicalEffectsTable().getField_PK();                   }

    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField()
     */
    protected ABCMdbField baseTablePrimaryFKField(){       return DB().getPlantConfigurationChemicalEffectsTable().getField_PlantConfigurationFK(); }
    
    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField()
     */
    protected ABCMdbField baseTableSecondaryFKField(){     return DB().getPlantConfigurationChemicalEffectsTable().getField_ChemicalName(); } 

    /**
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView()
     */
    protected ABCMdbAbstractViewable recordsView(){        return DB().getPlantConfigurationChemicalEffectsTable(); }

    /**
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField()
     */
    protected ABCMdbField recordsViewPKField(){            return DB().getPlantConfigurationChemicalEffectsTable().getField_PK();                   }

    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField()
     */
    protected ABCMdbField recordsViewPrimaryNameField(){   return null; }

    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField()
     */
    protected ABCMdbField recordsViewPrimaryFKField(){     return DB().getPlantConfigurationChemicalEffectsTable().getField_PlantConfigurationFK(); } 

    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField()
     */
    protected ABCMdbField recordsViewSecondaryNameField(){ return DB().getPlantConfigurationChemicalEffectsTable().getField_ChemicalName(); }
    
    /**
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#getOptionList(long, boolean)
     */
    public String getOptionList(long primaryFK, boolean showPrimaryName){
      return super.getOptionList(primaryFK, false); // There is no primary name
    }

    /**
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean)
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception {
      WMPlantConfigurationChemicalEffectsTable table = DB().getPlantConfigurationChemicalEffectsTable();
      
      int chemicalCol = rs.findColumn(table.getField_ChemicalName().getName());
      int needCol     = rs.findColumn(table.getField_NeedOrResist().getName());
      int consumeCol  = rs.findColumn(table.getField_ConsumeOrEmit().getName());
        
        
      while(rs.next()){
        xml.open("chemEffect");
          xml.addLine("chemical", rs.getString(chemicalCol));
          xml.addLine("need", rs.getDouble(needCol));
          xml.addLine("consume", rs.getDouble(consumeCol));
        xml.closeTag();
      }
    }

    /**
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave()
     */
    protected ABCMdbDatum datumToSave() {
      return DB().getPlantConfigurationChemicalEffectsTable().getDatum(PKOfConfigToSave, chemicalNameToSave, needToSave, consumptionToSave);
    }

    /**
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave()
     */
    protected void clearInstanceToSave(){
      PKOfConfigToSave   = 0;
      chemicalNameToSave = "";
      needToSave         = 0;
      consumptionToSave  = 0;      
    }

    /**
     * Sets the instance of this chemical effect to be saved.
     * @param plantConfigFK PK of the associated plant config
     * @param chemicalName name of the chemical
     * @param need value for the plant's need (or impact)
     * @param consumption value for the plant's consumption (or emission)
     */
    public void setInstanceToSave(long plantConfigFK, String chemicalName, double need, double consumption){
      PKOfConfigToSave = plantConfigFK;
      chemicalNameToSave = chemicalName;
      needToSave = need;
      consumptionToSave = consumption;      
    }  
    
    /**
     * Determine if the instance to save is valid
     * @return true if PK given for the plant configuration is greater than zero; false otherwise
     */
    public boolean instanceToSaveIsValid(){       return (PKOfConfigToSave > 0);     }
    
    /**
     * gets the PK of the specified plant configuration chemical effect
     * @param plantConfigurationPK PK of the plant config
     * @param chemical name of the chemical
     * @return PK of the chemical effect record; -1 if no matching record exists
     */
    public long getPK(long plantConfigurationPK, String chemical){
      return DB.getNumericItem(
          "SELECT * FROM " + this.recordsView().getFullName(DB.getOwner()) + " " + 
          "WHERE " + this.recordsViewPrimaryFKField().getName() + "=" + plantConfigurationPK + " AND " +
          this.recordsViewSecondaryNameField().getName() + "='" + chemical + "'",
          this.recordsViewPKField().getName());
    }
    
    /**
     * Delete a specific chemical effect by specifying a plant configuration and a chemical
     * @param plantConfigurationPK PK of the plant configuration
     * @param chemical name of the chemical
     * @return true if successful, false otherwise
     */
    public boolean deleteChemicalEffect(long plantConfigurationPK, String chemical){      
      return this.deleteItem(getPK(plantConfigurationPK, chemical));
    }
  }
  
  
  
  
  
  
  
  /**
   * Constructor
   * @param db DBGateway to be used
   */
  public PlantConfigurationManager(WMDBGateway db){
    super(db);
    PCSM   = new PlantConfigurationStageManager(DB, this);
    PCCEM  = new PlantConfigurationChemicalEffectManager(DB, this); 
  }
  
  /**
   * Gets the DBGateway as a WMDBGateway
   * @return the DBGateway
   */
  private WMDBGateway DB(){ return (WMDBGateway)DB; }
  
  /* General */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){            return DB().getPlantConfigurationsTable();                    }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){     return DB().getPlantConfigurationsTable().getField_PK();      }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField(){   return DB().getPlantConfigurationsTable().getField_Name();    }

  
  /* XML Generation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView(){   return DB().getPlantConfigsAndResearchersView();                                   }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField(){       return DB().getPlantConfigsAndResearchersView().getField_PlantConfiguration_PK();  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName(){              return "plantConfigs";                                                             }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
    
    WMPlantConfigsAndResearchersView view = DB().getPlantConfigsAndResearchersView();
    
    int PKFieldCol             = rs.findColumn(view.getField_PlantConfiguration_PK().getName());
    int nameFieldCol           = rs.findColumn(view.getField_PlantConfiguration_Name().getName());
    int descriptionFieldCol    = rs.findColumn(view.getField_PlantConfiguration_Description().getName());
    int dateCreatedFieldCol    = rs.findColumn(view.getField_PlantConfiguration_DateCreated().getName());
    int researcherNameFieldCol = rs.findColumn(view.getField_Researcher_Name().getName());
    int plantNameFieldCol      = rs.findColumn(view.getField_Plant_Name().getName());
    int canBeEditedFieldCol    = rs.findColumn(view.getField_CanBeEdited().getName());
    int truthValueFieldCol     = rs.findColumn(view.getField_PlantConfiguration_TruthValue().getName());
        
    while(rs.next()){
      xml.open("plantConfig");
        xml.addLine("configID",          rs.getLong(PKFieldCol));
        xml.addLine("configName",        rs.getString(nameFieldCol));
        xml.addLine("configDesc",        rs.getString(descriptionFieldCol));
        xml.addLine("configDateCreated", rs.getString(dateCreatedFieldCol));
        xml.addLine("configCreatedBy",   rs.getString(researcherNameFieldCol));
        xml.addLine("configPlantName",   rs.getString(plantNameFieldCol));
        xml.addLine("configCanBeEdited",   ((rs.getString(canBeEditedFieldCol) != null) ? "Yes" : "No"));
        xml.addLine("configTruthValue",  rs.getString(truthValueFieldCol));
        
        if(singleInstance) {
          xml.open("stages");
            PCSM.addSetToXML(xml, includeComments, rs.getLong(PKFieldCol));
          xml.closeTag();
          
          xml.open("chemicalEffects");
            PCCEM.addSetToXML(xml, includeComments, rs.getLong(PKFieldCol));
          xml.closeTag();
        }
        
        if(includeComments){
          CommentManager cm = new CommentManager(DB);
          xml.addBlock(cm.getListOfCommentsForObjectAsXML(WMDBGateway.PLANT_CONFIGURATION, rs.getLong(PKFieldCol)), XMLAssist.INDENT_BLOCK);
        }
      xml.closeTag();
    }
  }

  /**
   * Adds a set of Plant Configurations to an XML document
   * @param xml XMLAssist object constructing the document
   * @param includeComments true if comments are to be included 
   * @param plantPK ID of plant with which thw Configurations are to be associated
   */
  protected void addSetToXML(XMLAssist xml, boolean includeComments, long plantPK){
    String sql = DB().getPlantConfigsAndResearchersView().getSelectSQL(DB().getOwner(), 
        AbstractDBGateway.where(DB().getPlantConfigsAndResearchersView().getField_Plant_PK().getName(), plantPK));
    try {
          
      Connection con = DB.getGuestConnection();
      Statement st = con.createStatement();
      ResultSet rs = st.executeQuery(sql);
      
      xml.open("plantConfigs");
      addInstancesToXML(xml, rs, includeComments);
      xml.closeTag();
      
      rs.close();
      st.close();
      con.close();
    } 
    catch (Exception e) {
      addToErrorReport("Error adding set to XML: " + e.getMessage());
    }
  }
  
  
  
  /* HTML Generation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){ return "xmlPlantConfigurationToHTML.xsl";      }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName(){     return "xmlPlantConfigurationListToHTML.xsl";  }

  
  /* Saving */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  protected void clearInstanceToSave(){   
    nameToSave        = null;
    descriptionToSave = null;
    userIDToSave      = -1;
    plantIDToSave     = -1;
    truthValueToSave  = -1;    
  }

  /**
   * Sets values to be saved as an instance of a Plant Configuration
   * @param name name of the Plant Configuration. Should not duplicate any previously saved name
   * @param description description
   * @param userID ID of researcher to be 'owner' of this Plant Configuration
   * @param PlantID ID of Plant with which this Configuration is associated. 
   * @param truthValue truth value to be used for thie Plant Configuration
   */
  public void setInstanceToSave(String name, String description, long userID, long PlantID, long truthValue){
    nameToSave        = name;
    descriptionToSave = description;
    userIDToSave      = userID;
    plantIDToSave     = PlantID;
    truthValueToSave  = truthValue;
  }
  
  /**
   * Returns true if the instance to save has been appropriately set
   * @return true if the instance to save has been set and is valid
   */
  public boolean instanceToSaveIsValid(){
    return ((nameToSave        != null) &&
            (descriptionToSave != null) &&
            (userIDToSave      != -1)   &&
            (plantIDToSave     != -1)   &&
            (truthValueToSave  != -1));
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  protected ABCMdbDatum datumToSave(){     
    return (instanceToSaveIsValid() ? 
            DB().getPlantConfigurationsTable().getDatum(nameToSave, descriptionToSave, DB.currentTimeString(), 
                "" + userIDToSave, "" + plantIDToSave, "" + truthValueToSave) 
            : null);   
  }

  
  /* Auditable */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField baseTableOwnerField(){         return DB().getPlantConfigurationsTable().getField_ResearcherFK();                       }
 
  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView(){  return DB().getCanBeEdited_PlantConfigurationsView();                                    }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField canBeEditedViewPKField(){      return DB().getCanBeEdited_PlantConfigurationsView().getField_PlantConfiguration_PK();   }
   
  /**
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#deleteItem(long, long)
   */
  public boolean deleteItem(long plantConfigurationID, long userID){
    return permitModification(plantConfigurationID, userID)     &&
           PCSM.deleteElementsByPrimaryFK(plantConfigurationID) &&
           PCCEM.deleteElementsByPrimaryFK(plantConfigurationID) &&
           super.deleteItem(plantConfigurationID, userID);
  }
  
  
  /* Special */  
  

  /* Stages */
  /**
   * Delets a Plant Configuration Stage from the associated Database. 
   * @param plantConfigurationID primary key of the Plant Configuration to be deleted
   * @param stageNumber number of the stage to be deleted
   * @param user the user ID of the user requesting the delete operation
   * @return true if successful (even if the primary key provided
   * does not match any extant Plant Configuration primary key)
   */  
  public boolean deletePlantConfigurationStage(long plantConfigurationID, long stageNumber, long user){
    return permitModification(plantConfigurationID, user) &&
           PCSM.deleteStage(plantConfigurationID, stageNumber);    
  }
 
  
  /**
   * Saves a new Plant Configuration Stage as part of the specified
   * Plant Configuration. If the new stage number is equal to or lower
   * than any existing stage numbers, these are bumped upward so
   * that this stage is inserted properly. For example, if there
   * were three stages (1, 2, 3) and this method were called to insert
   * a new stage 2, extant stages 2 and 3 would be incremented, leaving
   * (1, 2, 3, 4) where 2 was the newly inserted stage.
   * @param plantConfigurationID primary key of the Plant Configuration to be
   * adjusted
   * @param intOrder order number of the stage to be inserted
   * @param duration duration, in days, this stage lasts
   * @param cropFactor crop factor of this plant during this Plant Configuration
   * Stage
   * @param cropSensitivity crop sensitivity of this plant during this Plant
   * Configuration Stage
   * @param laborRequirement value for the labor required to tend this plant
   * @param userID PK of the researcher saving this stage; used to confirm edit privileges
   * @return true if successful, false if not
   */
  public boolean saveStage(long plantConfigurationID, int intOrder, int duration, 
      double cropFactor, double cropSensitivity, double laborRequirement, long userID){
    PCSM.setInstanceToSave(plantConfigurationID, intOrder, duration, cropFactor, cropSensitivity, laborRequirement);
    return permitModification(plantConfigurationID, userID) &&
           PCSM.saveInstanceToDB();    
  }

  /* Chemical Effects */
  
  /**
   * Saves the specified chemical effect.
   * @param plantConfigurationID PK of the plant configuration to be associated with this chemical effect
   * @param chemical name of the chemical
   * @param need     value to be used for the plant's need for this chemical (or degree to which it is impacted by it)
   * @param consume  value to be used for the plant's consumption (or, if negative, emission) of this chemical
   * @param userID   ID of used saving this effect; used to confirm that the user has edit permissions
   * @return true if successful, false otherwise
   */
  public boolean saveChemicalEffect(long plantConfigurationID, String chemical, double need, double consume, long userID){
    PCCEM.setInstanceToSave(plantConfigurationID, chemical, need, consume);
    return permitModification(plantConfigurationID, userID) && PCCEM.saveInstanceToDB();
  }
  
  /**
   * Gets the list of chemical effects as an option list
   * @param plantConfigPK PK of plant configuration for which effects are to be listed
   * @return a String with an HTML option list of the chemical effeccts
   */
  public String getChemicalsAsOptionList(long plantConfigPK){
    return PCCEM.getOptionList(plantConfigPK);
  }
 
  /**
   * Deletes a specific chemical effect
   * @param plantConfigurationID PK of the plant configuration
   * @param plantConfigurationChemicalEffectPK PK of the chemical effect record in the linking table
   * @param user ID of the researcher making this db change; used to confirm edit privileges
   * @return true if successful, false otherwise
   */
  public boolean deletePlantConfigurationChemicalEffects(long plantConfigurationID, long plantConfigurationChemicalEffectPK, long user){
    return permitModification(plantConfigurationID, user) &&
        PCCEM.deleteItem(plantConfigurationChemicalEffectPK);
  }
  
}
