package wm.db;


import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import wm.db.views.WMPlantConfigsAndResearchersView;
import wm.db.views.WMPlantsAndCommentsView;
import wm.db.views.WMPlantsAndResearchersView;
import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.managers.CommentManager;
import abcm.db.managers.abstractparents.AbstractAuditableDBManager;
import abcm.math.BasicMath;
import abcm.utilities.TextFormatter;
import abcm.utilities.XMLAssist;

/**
 * Manages the entry and retrieval of Plants in the
 * HWM database.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class PlantManager extends AbstractAuditableDBManager {

  /** PK of user attempting to save new plant */
  private long   userIDToSave;
  /** Name of the plant to be saved */
  private String plantNameToSave;

  /**
   * Constructor
   * @param db Gateway to the database
   */
  public PlantManager(WMDBGateway db){    
    super(db);
  }
  
  /**
   * Gets the WMDBGateway object
   * @return the Gateway object to be used
   */
  private WMDBGateway DB(){    return (WMDBGateway)DB;    }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTable() 
   */
  protected ABCMdbTable baseTable(){               return DB().getPlantsTable();                                  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTablePKField() 
   */
  protected ABCMdbField baseTablePKField(){        return DB().getPlantsTable().getField_PK();                    }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBStandardizedElementManager#baseTableNameField() 
   */
  protected ABCMdbField baseTableNameField(){      return DB().getPlantsTable().getField_Name();                  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView() 
   */
  protected ABCMdbAbstractViewable recordsView(){  return DB().getPlantsAndResearchersView();                     }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField() 
   */
  protected ABCMdbField recordsViewPKField(){      return DB().getPlantsAndResearchersView().getField_Plant_PK(); }
  
  /**
   * Sets the values to be used when saving a plant to the database.
   * @param plantName The name of the plant to be saved
   * @param uid The user ID of the person saving the plant
   */
  public void setInstanceToSave(String plantName, long uid){
    plantNameToSave = AbstractABCMDBGateway.makeSQLSafe(plantName);
    userIDToSave = uid;        
  }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean) 
   */
  protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
   
    WMPlantsAndResearchersView view = DB().getPlantsAndResearchersView();
    
    int IDFieldCol          = rs.findColumn(view.getField_Plant_PK().getName());
    int nameFieldCol        = rs.findColumn(view.getField_Plant_Nm().getName());
    int researcherFieldCol  = rs.findColumn(view.getField_Researcher_Name().getName());   
    int canBeEditedFieldCol = rs.findColumn(view.getField_CanBeEdited().getName());
    
    while(rs.next()){
      xml.open("plant");
        xml.addLine("ID",            rs.getLong(IDFieldCol));
        xml.addLine("name",          rs.getString(nameFieldCol));
        xml.addLine("addedBy",       rs.getString(researcherFieldCol));
        xml.addLine("canBeEdited", ((rs.getString(canBeEditedFieldCol) != null) ? "Yes" : "No"));

        PlantConfigurationManager PCM = new PlantConfigurationManager(DB());
        PCM.addSetToXML(xml, false, rs.getLong(IDFieldCol));
        
        /* Comments */
        if(includeComments){
          WMCommentManager cm = new WMCommentManager(DB());
          xml.addBlock(cm.getListOfCommentsForObjectAsXML(WMDBGateway.PLANT, instancePK), XMLAssist.INDENT_BLOCK);
        }
      xml.closeTag(); 
    }    
  }

  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#xmlOuterNodeName() 
   */
  protected String xmlOuterNodeName(){    return "plantlist";    }
  


  /* HTML Creation */
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName() 
   */
  protected String getInstanceStylesheetName(){  return "xmlPlantToHTML.xsl";      }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName() 
   */
  protected String getListStylesheetName(){      return "xmlPlantListToHTML.xsl";  }

  
  /* Saving */
  
  /**
   * Sets an instance of a plant to be saved
   * @param userID ID of researcher who will be the 'owner' of this plant
   * @param name name of the plant to be saved
   */
  public void setInstanceToSave(long userID, String name){
    userIDToSave    = userID;
    plantNameToSave = name;
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave() 
   */
  protected void clearInstanceToSave(){
    userIDToSave    = -1;
    plantNameToSave = null;
  }

  /**
   * Returns true if the instance to be saved is valid, false otherwise
   * @return true if the instance to be saved has been set
   */
  public boolean instanceToSaveIsValid(){
    return ((userIDToSave    != -1)   &&
            (plantNameToSave != null));
  }
  
  /** 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave() 
   */
  protected ABCMdbDatum datumToSave(){
    return(instanceToSaveIsValid() ?
        DB().getPlantsTable().getDatum(plantNameToSave, "" + userIDToSave) 
        : null);
  }  

  
  /* Auditing */

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#baseTableOwnerField() 
   */
  public ABCMdbField baseTableOwnerField(){          return DB().getPlantsTable().getField_ResearcherFK();        }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedView() 
   */
  public ABCMdbAbstractViewable canBeEditedView(){   return DB().getCanBeEdited_PlantsView();                     }

  /** 
   * @see abcm.db.managers.abstractparents.AbstractAuditableDBManager#canBeEditedViewPKField() 
   */
  public ABCMdbField canBeEditedViewPKField(){       return DB().getCanBeEdited_PlantsView().getField_Plant_PK(); }

  /**
   * Gets a TeX formatted report of the collection of Plant Configurations
   * @return the TeX formatted report
   */
  public String getPlantConfigurationsAsTEX(){
    StringWriter retVal = new StringWriter();
    
    try{
      Connection con = DB.getGuestConnection();
      
      WMPlantsAndResearchersView plantsAndResearchersView = DB().getPlantsAndResearchersView();
      
      String plantPKFieldName         = plantsAndResearchersView.getField_Plant_PK().getName();
      String plantNameFieldName       = plantsAndResearchersView.getField_Plant_Nm().getName();
      String plantResearcherFieldName = plantsAndResearchersView.getField_Researcher_Name().getName();

      
      Statement st1 = con.createStatement();
      ResultSet plantsAndResearchers = st1.executeQuery(plantsAndResearchersView.getSelectSQL(DB.getOwner(), AbstractDBGateway.orderBy(plantNameFieldName)));
      
      while(plantsAndResearchers.next()){
        long   plantPK             = plantsAndResearchers.getLong(plantPKFieldName);
        String plantName           = TextFormatter.makeTEXSafe(plantsAndResearchers.getString(plantNameFieldName));
        String plantResearcherName = TextFormatter.makeTEXSafe(plantsAndResearchers.getString(plantResearcherFieldName));
        
        retVal.write("\\section{Plant: " + plantName + "}\n");
        retVal.write("{\\em Added by " + plantResearcherName + "}\n\n");
        
        CommentManager CM = new CommentManager(DB);
          
        if(DB.getNumericItem(DB().getPlantsInCommentsView(), DB().getPlantsInCommentsView().getField_Plant_PK(), 
            plantPK, DB().getPlantsInCommentsView().getField_CountOfComments()) > 0){
        
          retVal.write("\\subsection{Comments on " + plantName + "}\n\n");
          
          // For each plant get all comments
          WMPlantsAndCommentsView plantsAndCommentsView = DB().getPlantsAndCommentsView();
          
          Statement st2 = con.createStatement();
          ResultSet plantsAndComments = st2.executeQuery(plantsAndCommentsView.getSelectSQL(DB.getOwner(), 
              AbstractABCMDBGateway.where(plantsAndCommentsView.getField_Plant_PK().getName(), plantPK)));
          
          
          while(plantsAndComments.next()){
            long commentID = plantsAndComments.getLong(DB().getPlantsAndCommentsView().getField_Comment_PK().getName());
            retVal.write(CM.getCommentTEX(commentID));
          }
          plantsAndComments.close();
          st2.close();
        }
        
        // For each plant get all configurations
        Statement st3 = con.createStatement();
        
        WMPlantConfigsAndResearchersView plantConfigs = DB().getPlantConfigsAndResearchersView();
        ResultSet configs = st3.executeQuery(plantConfigs.getSelectSQL(DB.getOwner(), 
            AbstractABCMDBGateway.where(plantConfigs.getField_Plant_PK().getName(), plantPK)));
        
        while(configs.next()){
          long   configPK          = configs.getLong(plantConfigs.getField_PlantConfiguration_PK().getName());
          String configurationName = TextFormatter.makeTEXSafe(configs.getString(plantConfigs.getField_PlantConfiguration_Name().getName()));
          String addedBy           = TextFormatter.makeTEXSafe(configs.getString(plantConfigs.getField_Researcher_Name().getName()));
          String truthValue        = configs.getString(plantConfigs.getField_PlantConfiguration_TruthValue().getName());
          String desc              = configs.getString(plantConfigs.getField_PlantConfiguration_Description().getName());
          
          retVal.write("\\subsection{" + plantName + " configuration: " + configurationName + "}\n\n");
          retVal.write("{\\em Added by " + addedBy + "}\n\n");
          retVal.write("Truth Value: {\\em " + truthValue + "}\n\n");
          retVal.write(desc + "\n\n");
                    
          // Get the configuration's stages...
          Statement st5 = con.createStatement();
          
          ResultSet configStages = st5.executeQuery(DB().getPlantConfigurationStagesTable().getSelectSQL(DB.getOwner(), 
              AbstractABCMDBGateway.where(DB().getPlantConfigurationStagesTable().getField_PlantConfigurationFK().getName(), configPK) + " " +
              AbstractABCMDBGateway.orderBy(DB().getPlantConfigurationStagesTable().getField_Order().getName())));
          
          String configStagesOut = "";
          
          while(configStages.next()){
            long stageNo       = configStages.getLong(DB().getPlantConfigurationStagesTable().getField_Order().getName());
            long stageDur      = configStages.getLong(DB().getPlantConfigurationStagesTable().getField_Duration().getName());
            double cropFactor  = configStages.getDouble(DB().getPlantConfigurationStagesTable().getField_CropFactor().getName());
            double droughtSens = configStages.getDouble(DB().getPlantConfigurationStagesTable().getField_DroughtSensitivity().getName());
            double labor       = configStages.getDouble(DB().getPlantConfigurationStagesTable().getField_Labor().getName());
            
            cropFactor = BasicMath.roundDouble(cropFactor, 2);
            droughtSens = BasicMath.roundDouble(droughtSens, 2);
            
            
            configStagesOut += stageNo + " & " + stageDur + " & " + cropFactor + " & " + droughtSens + " & " + labor + "\\\\\n";           
            
          }
          
          if(configStagesOut.length() > 0){
            configStagesOut = "\\begin{center}\n" + 
                              "\\begin{tabular}{|c|c|c|c|c|}\n" +
                              "\\hline\n" +
                              "Stage & Duration & Crop Factor & Drought Sensitivity & Labor\\\\\n" +
                              "\\hline\n" +
                              configStagesOut +
                              "\\hline\n" + 
                              "\\end{tabular}\n" + 
                              "\\end{center}\n\n";
            retVal.write(configStagesOut);
          }
          
          
          // Get the configuration comments
          if(DB.getNumericItem(DB().getCommentsCount_PlantConfigurationsView(), 
                               DB().getCommentsCount_PlantConfigurationsView().getField_PlantConfiguration_PK(), 
                               configPK, 
                               DB().getCommentsCount_PlantConfigurationsView().getField_CountOfComments()) > 0){
            
            retVal.write("\\subsubsection{Comments on " + plantName + " configuration " + configurationName + "}\n\n");
          
            Statement st4 = con.createStatement();
            
            ResultSet configComments = st4.executeQuery(DB().getPlantConfigurationCommentsTable().getSelectSQL(DB.getOwner(), 
                AbstractABCMDBGateway.where(DB().getPlantConfigurationCommentsTable().getField_PlantConfigurationFK().getName(), configPK)));
            
            while(configComments.next()){
              long commentID = configComments.getLong(DB().getPlantConfigurationCommentsTable().getField_CommentFK().getName());
              retVal.write(CM.getCommentTEX(commentID));
              
            }
            configComments.close();
            st4.close();
          }
        }
          
        configs.close();
        st3.close();        
      }
      
      st1.close();
      con.close();
      
    }
    catch(Exception E){
      System.err.println("Could not retrieve plant database: "+ E.getMessage());
    }
    
    
//    WMPlantConfigsAndResearchersView plantConfigsAndResearchers = DB().getPlantConfigsAndResearchersView();
//    
//    WMPlantConfigurationStagesTable plantConfigurationStagesTable = DB().getPlantConfigurationStagesTable();
    
    
    
    return retVal.toString();
  }
  
}
