package wm.db;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import wm.db.tables.WMPlantConfigurationChemicalEffectsTable;
import wm.db.tables.WMPlantConfigurationStagesTable;
import wm.db.tables.WMPlantConfigurationsTable;
import wm.db.tables.WMSimulationConfigurationPlantConfigurationsTable;
import wm.db.views.WMSimulationConfigurationPlantConfigurationsView;
import wm.sim.WMConfiguration;
import abcm.db.AbstractABCMDBGateway;
import abcm.db.AbstractDBGateway;
import abcm.db.components.ABCMdbAbstractViewable;
import abcm.db.components.ABCMdbDatum;
import abcm.db.components.ABCMdbField;
import abcm.db.components.ABCMdbTable;
import abcm.db.components.tables.ABCMSimulationConfigurationsTable;
import abcm.db.managers.abstractparents.AbstractDBLinkTableInnerManager;
import abcm.db.managers.abstractparents.AbstractDBManager;
import abcm.db.managers.stubs.AbstractConfigurationManager;
import abcm.sim.Configuration;
import abcm.utilities.XMLAssist;
import flowag.bio.Plant;


/**
 * DB Manager for WatMan specific ABCM Configurations.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class WMConfigurationManager extends AbstractConfigurationManager{

  /** DB Gateway to be used */                                 private AbstractABCMDBGateway DB;
  /** Simulation Configuration Plant Configuration Manager */  private SimulationConfigurationPlantConfigurationManager SCPCM;

  /**
   * Permits easy access to the WatMan version of the DB gateway
   * @return the DB gateway cast WatMan
   */
  private WMDBGateway DB(){ return ((WMDBGateway)DB); }
  
  /**
   * Manages the linking table between SimulationConfigurations and PlantConfigurations
   *
   *  @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  public class SimulationConfigurationPlantConfigurationManager extends AbstractDBLinkTableInnerManager{

    /** PK of the Simulation Configuration to save  */
    private long simConfigPKToSave;
    /** PK of the Plant Configuration to save  */
    private long plantConfigPKToSave;
  
    // Save some text by providing these here (see below)
    /**
     * Gets the Simulation Configuration Plant Configurations Table
     * @return the Simulation Configuration Plant Configurations Table 
     */
    private WMSimulationConfigurationPlantConfigurationsTable table(){
      return DB().getSimulationConfigurationPlantConfigurationsTable();
    }
    
    /**
     * Gets the Simulation Configuration Plant Configurations View
     * @return the Simulation Configuration Plant Configurations View
     */
    private WMSimulationConfigurationPlantConfigurationsView view(){
      return DB().getSimulationConfigurationPlantConfigurationsView();
    }
    
    /**
     * Constructor
     * @param db Instance of an ABCM Gateway to provide database access.
     * @param containedBy enclosing class
     */
    public SimulationConfigurationPlantConfigurationManager(AbstractABCMDBGateway db, AbstractDBManager containedBy) {
      super(db, containedBy);
    }

    /* Basic retrieval */
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTable()
     */
    protected ABCMdbTable baseTable(){                     return table();                                                             }
  
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePKField()
     */
    protected ABCMdbField baseTablePKField(){              return table().getField_PK();                                               }
  
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTablePrimaryFKField()
     */
    protected ABCMdbField baseTablePrimaryFKField(){       return table().getField_SimulationConfigurationFK();                        }
    
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#baseTableSecondaryFKField()
     */
    protected ABCMdbField baseTableSecondaryFKField(){     return table().getField_PlantConfigurationFK();                             }
  
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsView()
     */
    protected ABCMdbAbstractViewable recordsView(){        return view();                                                              }
    
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#recordsViewPKField()
     */
    protected ABCMdbField recordsViewPKField(){            return view().getField_SimulationConfigurationPlantConfiguration_PK();      }
    
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryFKField()
     */
    protected ABCMdbField recordsViewPrimaryFKField(){     return view().getField_SimulationConfiguration_PK();                        }
  
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewPrimaryNameField()
     */
    protected ABCMdbField recordsViewPrimaryNameField(){   return view().getField_PlantConfiguration_PK();                             }
  
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBLinkTableManager#recordsViewSecondaryNameField()
     */
    protected ABCMdbField recordsViewSecondaryNameField(){ return view().getField_PlantConfiguration_Name();                           }

    /* XML Creation */
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#addInstancesToXML(abcm.utilities.XMLAssist, java.sql.ResultSet, boolean)
     */
    protected void addInstancesToXML(XMLAssist xml, ResultSet rs, boolean includeComments) throws Exception{
      int plantConfigIDCol          = rs.findColumn(view().getField_PlantConfiguration_PK().getName());
      int plantConfigNameCol        = rs.findColumn(view().getField_PlantConfiguration_Name().getName());
      int plantConfigDescCol        = rs.findColumn(view().getField_PlantConfiguration_Description().getName());
      int plantConfigDateCreatedCol = rs.findColumn(view().getField_PlantConfiguration_DateCreated().getName());
      int plantNameCol              = rs.findColumn(view().getField_Plant_Name().getName());
      int plantConfigCreatedByCol   = rs.findColumn(view().getField_Researcher_Name().getName());
      int plantConfigCanBeEditedCol = rs.findColumn(view().getField_CanBeEdited().getName());
      int plantConfigTruthValueCol  = rs.findColumn(view().getField_TruthValue().getName());
      
      while(rs.next()){
        xml.open("plant");
          xml.addLine("plantConfigID",    rs.getLong(plantConfigIDCol));
          xml.addLine("plantConfigName",  rs.getString(plantConfigNameCol));
          xml.addLine("plantConfigDesc",  rs.getString(plantConfigDescCol));
          xml.addLine("dateCreated",      rs.getString(plantConfigDateCreatedCol));
          xml.addLine("plantName",        rs.getString(plantNameCol));
          xml.addLine("createdBy",        rs.getString(plantConfigCreatedByCol));
          xml.addLine("canBeEdited",      ((rs.getString(plantConfigCanBeEditedCol) != null) ? "Yes" : "No"));
          xml.addLine("truthValue",       rs.getString(plantConfigTruthValueCol));
        xml.closeTag();
      }      
    }
    
    /* Saving */
    /**
     * Sets the instance to be saved 
     * @param simConfigPK Simulation Configuration PK
     * @param plantConfigPK Plant Configuration PK
     */
    public void setInstanceToSave(long simConfigPK, long plantConfigPK){
      simConfigPKToSave   = simConfigPK;
      plantConfigPKToSave = plantConfigPK;
    }
    
    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#clearInstanceToSave()
     */
    protected void clearInstanceToSave(){
      simConfigPKToSave   = -1;
      plantConfigPKToSave = -1;
    }

    /**
     * 
     * @see abcm.db.managers.abstractparents.AbstractDBManager#datumToSave()
     */
    protected ABCMdbDatum datumToSave(){
      return 
        (instanceToSaveIsValid() ? 
            DB().getSimulationConfigurationPlantConfigurationsTable().getDatum(simConfigPKToSave, plantConfigPKToSave)  :
            null);
    }
    
    /**
     * Determines the status of the instance to be saved
     * @return true if the instance to be saved is valid, false otherwise
     */
    public boolean instanceToSaveIsValid(){
      return ((simConfigPKToSave != -1) && (plantConfigPKToSave != -1));
    } 
    

    /**
     * Returns an option list, but overrides the superclass because there 
     * is no primary name field
     * @param primaryFK primary field FK
     * @return an HTML option list
     */
    public String getOptionList(long primaryFK){
      return getOptionList(primaryFK, false);
    }
    
  }

  /**
   * Constructor 
   * @param db DBGateway
   */
  public WMConfigurationManager(AbstractABCMDBGateway db){
    super(db);
    DB = db;
    SCPCM = new SimulationConfigurationPlantConfigurationManager(db, this);
  }

  /**
   * 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getListStylesheetName()
   */
  protected String getListStylesheetName(){      return "xmlConfigurationListToHTML.xsl";  }

  /**
   * 
   * @see abcm.db.managers.abstractparents.AbstractDBManager#getInstanceStylesheetName()
   */
  protected String getInstanceStylesheetName(){  return "xmlConfigurationToHTML.xsl";      }
  
  /**
   * 
   * @see abcm.db.managers.stubs.AbstractConfigurationManager#getConfiguration(long)
   */
  public Configuration getConfiguration(long PK){
    if(PK == DEFAULT_PK)  return getDefaultConfiguration();
    WMConfiguration newConfig = null;
    
    errorReport = "No error";
    
    
    // Get the configuration row from the database
    try{
      WMDBGateway wmdb = ((WMDBGateway)DB);
      ABCMSimulationConfigurationsTable simConfigsTable = DB.getSimulationConfigurationsTable();
      
      // Confirm that the configuration actually exists
      if(!DB.entryExists(simConfigsTable, simConfigsTable.getField_PK() , PK)) return null;
      
      // Next, retrieve the configuration name 
      newConfig = new WMConfiguration(getName(PK));
      
      // Prepare a database connection
      Connection con = DB.getGuestConnection(); 
      
      
      // Next, find the plant configurations and add these
        
      // First, get a recordset of all plant configurations in this WMConfiguration
      WMSimulationConfigurationPlantConfigurationsTable simConfigPlantConfigsTable = wmdb.getSimulationConfigurationPlantConfigurationsTable(); 
      String simConfigsPlantConfigsTableSQL = simConfigPlantConfigsTable.getSelectSQL(DB.getOwner(), 
          AbstractDBGateway.where(simConfigPlantConfigsTable.getField_SimulationConfigurationFK().getFullFieldName(DB.getOwner()), PK));
      
      Statement st1 = con.createStatement();
      ResultSet plantConfigsList = st1.executeQuery(simConfigsPlantConfigsTableSQL);
        
      // This is a result set that gives a list of the plant configuration numbers in this simulation configuration
      // Loop through these
      while(plantConfigsList.next()){
        long plantConfigPK = plantConfigsList.getLong(simConfigPlantConfigsTable.getField_PlantConfigurationFK().getName());
        
        // Get the details for each one
        WMPlantConfigurationsTable plantConfigurationsTable = wmdb.getPlantConfigurationsTable();
        String plantConfigSQL = plantConfigurationsTable.getSelectSQL(DB.getOwner(), 
            AbstractDBGateway.where(plantConfigurationsTable.getField_PK().getFullFieldName(DB.getOwner()), plantConfigPK));
        
        Statement st2 = con.createStatement();
        ResultSet plantConfigDetails = st2.executeQuery(plantConfigSQL);
        
        if(plantConfigDetails.next()){
          long plantID = plantConfigDetails.getLong(plantConfigurationsTable.getField_PlantFK().getName());
          String varietyName = plantConfigDetails.getString(plantConfigurationsTable.getField_Name().getName());
          
          PlantManager PM = new PlantManager(wmdb);
          
          String plantName = PM.getName(plantID);
          Plant P = new Plant(varietyName);
          
          // Now find the stages for this plant configuration
          WMPlantConfigurationStagesTable plantConfigurationStagesTable = wmdb.getPlantConfigurationStagesTable();
          String plantConfigurationStagesSQL = plantConfigurationStagesTable.getSelectSQL(DB.getOwner(), 
              AbstractDBGateway.where(plantConfigurationStagesTable.getField_PlantConfigurationFK().getName(), plantConfigPK) + " " +
              AbstractDBGateway.orderBy(plantConfigurationStagesTable.getField_Order().getName()));
          
          Statement st3 = con.createStatement();
          ResultSet plantConfigStages = st3.executeQuery(plantConfigurationStagesSQL);
          
          int stageNum = 1;
          while(plantConfigStages.next()){
            P.addStage(stageNum, 
                        plantConfigStages.getInt(plantConfigurationStagesTable.getField_Duration().getName()), 
                        plantConfigStages.getDouble(plantConfigurationStagesTable.getField_CropFactor().getName()),
                        plantConfigStages.getDouble(plantConfigurationStagesTable.getField_DroughtSensitivity().getName()));
            stageNum++;
          }
          plantConfigStages.close(); // Close this
          st3.close();
          
          // Now deal with chemical effects
          WMPlantConfigurationChemicalEffectsTable plantConfigurationChemicalEffectsTable = wmdb.getPlantConfigurationChemicalEffectsTable();
          String plantConfigurationChemicalEffectsSQL = plantConfigurationChemicalEffectsTable.getSelectSQL(DB.getOwner(), 
              AbstractDBGateway.where(plantConfigurationChemicalEffectsTable.getField_PlantConfigurationFK().getName(), plantConfigPK));
          
          Statement st4 = con.createStatement();
          ResultSet plantConfigChems = st4.executeQuery(plantConfigurationChemicalEffectsSQL);
          
          while(plantConfigChems.next()){
            P.addChemicalEffect(plantConfigChems.getString(plantConfigurationChemicalEffectsTable.getField_ChemicalName().getName()), 
                plantConfigChems.getDouble(plantConfigurationChemicalEffectsTable.getField_NeedOrResist().getName()), 
                plantConfigChems.getDouble(plantConfigurationChemicalEffectsTable.getField_ConsumeOrEmit().getName())); 
          }
          plantConfigChems.close();
          st4.close();
          // And add to the configuration
          newConfig.addPlant(P, varietyName, plantName);
        }
        plantConfigDetails.close(); // There will be only one record; no loop, just an if statement
        st2.close();
      }
      plantConfigsList.close();
      st1.close();
      con.close();
    } catch (Exception e) {
      errorReport = "Error in Configuration Manager getConfiguration(): " + e.getMessage();
      System.err.println("ERROR RETRIEVING CONFIGURATION in WMConfiguration Manager: " + e.getMessage());
      System.err.println("Exception is " + e.getLocalizedMessage());
    }
    return newConfig;
  }

  /**
   * @see abcm.db.managers.stubs.AbstractConfigurationManager#getDefaultConfiguration()
   */
  public Configuration getDefaultConfiguration(){
    WMConfiguration config = new WMConfiguration("DEFAULT");
    Plant p = new Plant("default_plant");      
    p.addStage(1, 30, 0.8,  0.8);
    p.addStage(2, 30, 0.95, 0.8);
    p.addStage(3, 30, 1.1,  0.8);
    p.addStage(4, 30, 1.15, 0.8);
    config.addPlant(p, "default_plant", "default_plant_variety");  
    return config;
  }

  
  /* Plant Configurations */
  /**
   * Adds a specified Plant Configuration to this Configuration.
   * @param configID primary key of the Configuration to which the 
   * Plant Configuration is to be added
   * @param plantConfigID primary key of the Plant Configuration to be added
   * @return true if successful, false otherwise
   */
  public boolean addPlantConfigToConfig(long configID, long plantConfigID){
    SCPCM.setInstanceToSave(configID, plantConfigID);
    return SCPCM.saveInstanceToDB();
  }


  /**
   * Removes the specified Plant Configuration from the 
   * specified Configuration
   * @param configPK primary key of the Configuration from which the Plant Configuration
   * should be removed
   * @param plantPK primary key of the Plant Configuration to be removed
   * @return true if successful, false otherwise. Will not return an
   * error if the Plant Configuration specified is not found in
   * the Configuration.
   */
  public boolean removePlantConfig(long configPK, long plantPK){
    return SCPCM.deleteElementsByBothPrimaryKeys(configPK, plantPK);
  }
  
  /**
   * Removes the specific SimulationConfiguration/PlantConfiguration
   * record
   * @param simConfigPlantConfigPK
   * @return true if successful, false otherwise
   */
  public boolean removePlantConfig(long simConfigPlantConfigPK){
    return SCPCM.deleteItem(simConfigPlantConfigPK);
  }
  
  

  /**
   * Gets the list of Plant Configurations in this Configuration
   * as an HTML Option List.
   * @param PK primary key of the Configuration
   * @return the HTML Option List
   */
  public String getPlantConfigsAsOptionList(long PK){
    return SCPCM.getOptionList(PK);
  }

  /**
   * 
   * @see abcm.db.managers.stubs.AbstractConfigurationManager#addConfigurationDetailsToXML(abcm.utilities.XMLAssist, long)
   */
  public void addConfigurationDetailsToXML(XMLAssist xml, long PK){
    addConfigurationDetailsToXML(xml, PK, false);
  }
  
  /**
   * Adds the details of this configuration to an XML document
   * @param xml XMLAssist object constructing the document 
   * @param PK ID of the configuration to be added
   * @param includeComments true if comments are to be included, false otherwise
   */
  public void addConfigurationDetailsToXML(XMLAssist xml, long PK, boolean includeComments){
    // Configurations can have many varieties of details,
    // but this one only has one: plant configurations
    xml.open("configPlants");
      SCPCM.addSetToXML(xml, includeComments, PK);
    xml.closeTag();
  }  
  
}
