package wm.db;

import java.sql.Connection;

import abcm.db.ABCMdbDataInitializer;
import abcm.db.AbstractDBGateway;
import flowag.db.FlowAgDBInitializer;

/**
 * Used to initialize a WatMan ABCM database. Creates the additional,
 * WatMan-specific tables and adds WatMan specific data (such
 * as commands).
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class WMDBDataInitializer extends ABCMdbDataInitializer{

  /** DB Gateway for the WatMan ABCM Database */ 
  protected        WMDBGateway DB;
  /** Primary Key assigned to the 'Rivers' Command Category                         */ 
  protected long newPK_Category_Rivers;
  /** Primary Key assigned to the 'Canal Systems' Command Category                  */ 
  protected long newPK_Category_CanalSystems;
  /** Primary Key assigned to the 'Fields' Command Category                         */ 
  protected long newPK_Category_Fields;
  /** Primary Key assigned to the 'Rain' Command Category                           */ 
  protected long newPK_Category_Rain;
  /** Primary Key assigned to the 'Environment' Command Category                    */ 
  protected long newPK_Category_Environment;
  
  private static final String setStreamflowDefaultCode = 
    "    River R = getRiver(rName);" + "\n" +
    "    if(R == null) return new ABCMResult(\"River `\" + rName + \"` does not exist.\", false);" + "\n" +    
    "    // Calculate the new value for streamflow" + "\n" +
    "    switch(codeModelOption(\"SET_STREAMFLOW_PROC\")){" + "\n" +
    "      case SET_STREAMFLOW_DEFAULT: // (0)" + "\n" +
    "      default:" + "\n" +
    "        n = newFlow;" + "\n" +
    "    }" + "\n" +
    "    R.setAnnualStreamflow(n);";
  
  private static final String setStreamflowStochasticCode = 
    "    River R = getRiver(rName);" + "\n" +
    "    if(R == null) return new ABCMResult(\"River `\" + rName + \"` does not exist.\", false);" + "\n" +
    "    // Calculate the new value for streamflow" + "\n" +
    "    switch(codeModelOption(\"SET_STREAMFLOW_PROC\")){" + "\n" +
    "      case SET_STREAMFLOW_STOCHASTIC: // (1)" + "\n" +
    "        double r = random();" + "\n" +
    "        double p = paramVal(\"SETFLOWSTOCHASTICITY\");" + "\n" +
    "        n = newFlow * ((double)1 - (p * (double)2 * (r - (double).5)));" + "\n" +
    "        break;" + "\n" +
    "    }" + "\n" +
    "    R.setAnnualStreamflow(n);";
  
  private static final String setStreamflowScaledCode = 
    "   River R = getRiver(rName);" + "\n" +
    "   if(R == null) return new ABCMResult(\"River `\" + rName + \"` does not exist.\", false);" + "\n" +
    "   // Calculate the new value for streamflow" + "\n" +
    "   switch(codeModelOption(\"SET_STREAMFLOW_PROC\")){" + "\n" +
    "     case SET_STREAMFLOW_SCALED: // (2)" + "\n" +
    "       n = newFlow * paramVal(\"SCALE_STREAMFLOW_SET_VALUE\");" + "\n" +
    "       break;" + "\n" +
    "   }" + "\n" +
    "   R.setAnnualStreamflow(n);"; 

  
  private static final String noAgentCode = 
    "// To avoid using agents, do not use the CreatePopulation command";
  
  private static final String oneAgentPerFieldCode = 
    "public void createAgents(){" + "\n" +
    "  WMModel model = (WMModel)super.M;" + "\n" +
    "  // There will be various methods here to create agents," + "\n" + 
    "  // depending on what's needed" + "\n" +
    "  switch (codeModelOption(\"AGENT_CREATION\")){" + "\n" +
    "    case(ONE_AGENT_PER_FIELD):{" + "\n" +
    "      // Add one agent for each field" + "\n" +
    "      HashSet fields = model.getAllFields();" + "\n" +
    "      Iterator I = fields.iterator();" + "\n" +
    "      while(I.hasNext()){" + "\n" +
    "        Field F = (Field)I.next();" + "\n" +
    "        Vector<String> namesOfInflowSegs = model.getNamesOfInflowCanalSegments(F.name);" + "\n" +
    "        String inflowSegName = (namesOfInflowSegs.size() == 1 ? namesOfInflowSegs.get(0) : \"UNKNOWN\");" + "\n" +
    "        Vector<String> namesOfCanalSystems = model.getNamesOfCanalSystems(F.name);" + "\n" +
    "        String canalSysName = (namesOfCanalSystems.size() == 1 ? namesOfCanalSystems.get(0) : \"UNKNOWN\");" + "\n" +
    "        addToAgentList(new WMFieldManagerAgent(model, this, F, canalSysName, inflowSegName));" + "\n" +
    "      }" + "\n" +
    "      break;" + "\n" +
    "    }" + "\n" +      
    "  }";
  
  private static final String oneAgentPerFieldWithCSManCode = 
    "public void createAgents(){" + "\n" +
    "  WMModel model = (WMModel)super.M;" + "\n" +
    "  // There will be various methods here to create agents," + "\n" + 
    "  // depending on what's needed" + "\n" +
    "  switch (codeModelOption(\"AGENT_CREATION\")){" + "\n" +
    "    case(ONE_AGENT_PER_FIELD_WITH_CS_MANAGER):{" + "\n" +
    "      // First the canal system manager(s)" + "\n" +
    "      Hashtable canalSystems = model.getAllCanalSystems();" + "\n" +
    "      Iterator CSI = canalSystems.keySet().iterator();" + "\n" +
    "      while(CSI.hasNext())  addToAgentList(new WMCanalSystemManagerAgent(model, this, (String)CSI.next()));" + "\n" +
    "      " + "\n" +
    "       // Then the field manager agents" + "\n" +
    "      HashSet fields = model.getAllFields();" + "\n" +
    "      Iterator I = fields.iterator();" + "\n" +
    "      while(I.hasNext()){" + "\n" +
    "        Field F = (Field)I.next();" + "\n" +
    "         Vector<String> namesOfInflowSegs = model.getNamesOfInflowCanalSegments(F.name);" + "\n" +
    "        String inflowSegName = (namesOfInflowSegs.size() == 1 ? namesOfInflowSegs.get(0) : \"UNKNOWN\");" + "\n" +
    "        Vector<String> namesOfCanalSystems = model.getNamesOfCanalSystems(F.name);" + "\n" +
    "        String canalSysName = (namesOfCanalSystems.size() == 1 ? namesOfCanalSystems.get(0) : \"UNKNOWN\");" + "\n" +
    "        addToAgentList(new WMFieldManagerAgent(model, this, F, canalSysName, inflowSegName));" + "\n" +
    "      }" + "\n" +
    "      break;" + "\n" +
    "    }" + "\n" +
    "  }";
  
  
  /** Data Initializer from FlowAg toolkit; used to import all FlowAg related assertions */ 
  protected FlowAgDBInitializer flowAgInit; 
  
  /**
   * Basic constructor
   * @param db WMDBGateway to be initialized
   */
  public WMDBDataInitializer(WMDBGateway db){
    super(db);
    DB = db;
  }

  
  /**
   * Adds a plant to the ABCM database
   * @param name name of the plant. Must not duplicate any previously entered name.
   * @param researcherFK ID of researcher associated with this plant
   * @return new PK of plant added
   * @throws Exception
   */
  public long addPlant(String name, long researcherFK) throws Exception{
    System.out.println("Adding plant: " + name);
    return add(DB.getPlantsTable().getDatum(name, researcherFK));
  }
  
  /**
   * Adds a plant configuration to the ABCM database
   * @param name name of the plant configuration. Must not duplicate any previously entered name.
   * @param description description of this configuration
   * @param researcherFK ID of researcher associated with this plant configuration
   * @param plantFK ID of the plant associated with this configuration
   * @param truthValueFK truth value
   * @return new PK of plant configuration
   * @throws Exception
   */
  public long addPlantConfiguration(String name, String description, long researcherFK, long plantFK, long truthValueFK) throws Exception{
    return add(DB.getPlantConfigurationsTable().getDatum(name, description, DB.currentTimeString(), 
        researcherFK, plantFK, truthValueFK));
  }
  
  /**
   * Adds a new stage to a plant configuration. Stage is added directly; no checking regarding
   * order is done, so order must not duplicate any pre-existing stage for this plant configuration,
   * and should not leave 'gaps' in order
   * @param plantConfigFK ID of plant configuration of which this is a stage
   * @param order order in which this stage occurs
   * @param duration duration of this stage, in days
   * @param cropFactor crop factor of plant during this stage
   * @param droughtSensitivity drought sensitivity of plant during this stage
   * @param labor labor required to tend this plant during this stage
   * @return new PK of this plant configuration stage
   * @throws Exception
   */
  public long addPlantConfigurationStage(long plantConfigFK, int order, int duration, double cropFactor, 
      double droughtSensitivity, double labor) throws Exception{
    return add(DB.getPlantConfigurationStagesTable().getDatum(plantConfigFK, order, duration, cropFactor, droughtSensitivity, labor));
  }
  
  /**
   * Adds a new chemical sensitivity to a plant configuration
   * @param plantConfigFK PK of the plant configuration
   * @param chemicalName name of the chemical
   * @param needOrResistance value of the plant's need for, or resistance to, this chemical
   * @param consumeOrEmit value of the plant's propensity to consume, or emit, this chemical
   * @return PK of the new record entered
   * @throws Exception generic exception
   */
  public long addPlantConfigurationChemicalSensitivity(long plantConfigFK, String chemicalName, 
      double needOrResistance, double consumeOrEmit) throws Exception{
    return add(DB.getPlantConfigurationChemicalEffectsTable().getDatum(plantConfigFK, chemicalName, needOrResistance, consumeOrEmit));
  }
  
  /**
   * Adds a plant configuration to a simulation configuration
   * @param simConfigFK the simulation configuration foreign key
   * @param plantConfigFK the plant configuration foreign key
   * @return the new PK of the link formed
   * @throws Exception
   */
  public long addPlantConfigurationToConfiguration(long simConfigFK, long plantConfigFK) throws Exception{
    System.out.println("Adding plant configuration " + plantConfigFK + " to configuration " + simConfigFK);
    return add(DB.getSimulationConfigurationPlantConfigurationsTable().getDatum(simConfigFK, plantConfigFK));
  }
  
  /**
   * Adds a comment to a plant configuration
   * @param commentFK FK of comment
   * @param plantConfigFK FK of plant configuration
   * @return new PK of comment-plant config link
   * @throws Exception 
   */
  public long addCommentToPlantConfiguration(long commentFK, long plantConfigFK) throws Exception{
    System.out.println("Adding comment to plant configuration...");
    return add(DB.getPlantConfigurationCommentsTable().getDatum(plantConfigFK, commentFK));
  }
  
  /**
   * Creates the specified database.
   * @param con The DB Connection to be used. 
   * @see abcm.db.ABCMdbDataInitializer#initialize(java.sql.Connection)
   */
  public void initialize(Connection con) throws Exception{
    super.initialize(con);

    // Add all external components
    flowAgInit = new FlowAgDBInitializer(this);
    flowAgInit.init();
    
    
    System.out.println("Adding WM Specific data...");
    
    // Some helpful values (OBSOLETE)
//    String t             = "1";
//    String dateCreated   = "01/04/2007";
//    String researcherFK  = "1";
//    String title         = "TEST";
//    String name          = "TEST";
//    String description   = "TEST";
//    String truthValueFK  = "1";
//    long   commentFK     = 1;
//    long   objectFK      = 1;
//    String TEST          = "TEST";

    
    // Terms
    addTerm("Canal System",  "CanalSys", "By `canal system` we mean any network of interconnected canal segments. This may include more than one head, but does not include canals that are completely disconnected from one another (use multiple canal systems for this).");
    addTerm("Canal Segment", "CanalSeg", "A Canal Segment is a straight-line prismatic channel that represents a section of a canal. Canal Segments can have either a headgate or another canal segment at their upstream end; they can have any number of canal segments at their downstream end.");
    addTerm("River",         "River",    "A `River` object has only two elements: a name and an annual streamflow. We avoid modeling the river in any other way and instead use it only to provide a source for water for our canals.");
    
    
    // New command categories
    newPK_Category_Environment  = addCommandCategory(190000, "Environment",   "Commands to set values in the climate and environment of the simulation");
    newPK_Category_Rivers       = addCommandCategory(150000, "Rivers",        "Create river systems or modify the streamflow levels in them");
    newPK_Category_CanalSystems = addCommandCategory(160000, "Canal Systems", "Create or modify the structure of a canal system, e.g. by adding segments or branches");
    newPK_Category_Fields       = addCommandCategory(170000, "Fields",        "Manage fields");
    newPK_Category_Rain         = addCommandCategory(180000, "Rain",          "Insert rainfall into the simulation or fields");
   
    /* Actions */
      /* Environment */
    addCommand(newPK_Category_Environment, "SetMeanPctAnnualDayHrs", "Set the mean pct of annual daytime hours", false, false,  
        "Sets the mean percentage of annual daytime hours for the Sim environment", new String[]{ "New mean percent annual daytime hours" });
    addCommand(newPK_Category_Environment, "SetMeanDailyTemp", "Set the mean daily temperature", false, false, 
        "Sets the mean daily temperature for the simulation environment", 
        new String[]{ "New mean daily temperature" });
    addCommand(newPK_Category_Environment, "SetMeanDailyTempAtAlt", "Sets the mean daily temperature by altitude gradient", false, false,
        "Sets the mean daily temperature by low and high altitude values",
        new String[]{ "Lower Altitude", "Temperature at Lower Altitude", "Upper Altitude", "Temperature at Upper Altitude" });
    addCommand(newPK_Category_Environment, "SetBasicDailyRainProb", "Set the probability of rainfall", false, false, "Sets the probability of rain in the basic engine",
        new String[]{ "New probability" });
    addCommand(newPK_Category_Environment, "SetBasicDailyRainMin", "Set the minimum amount of rain", false, false, "Sets the minimum amount of rain if it happens",
        new String[]{ "New amount" });
    addCommand(newPK_Category_Environment, "SetBasicDailyRainMax", "Set the maximum amount of rain", false, false, "Sets the maximum amount of rain if it happens",
        new String[]{ "New amount" });
    
    
      /* Rivers */
    addCommand(newPK_Category_Rivers, "CreateRiver", "Creates a river with the given name", false, false, 
        "Create a river system.", 
        new String[]{ "Name of new river system being created" });
    addCommand(newPK_Category_Rivers, "SetAnnualStreamflow", "Sets the annual streamflow of the river named to the value given", false, false, 
        "Set the annual streamflow, in million acre-feet", 
        new String[]{ "Name of the river system", "New annual streamflow value of the river system" });
    addCommand(newPK_Category_Rivers, "SetCurrentStreamFlow", "Sets the stream flow for a given river (arg 1) to a value (arg2) that represents the fraction of the annual streamflow", false, false, 
        "Set the current streamflow; note that this is not the annual streamflow (current streamflow is a daily value).", 
        new String[]{ "Name of River", "New Streamflow Value" });
    addCommand(newPK_Category_Rivers, "AddHeadGateToRiver", "Creates a headgate on a river", false, false, 
        "Creates a headgate at a point in a specified river", 
        new String[]{ "River system to which Gate is to be added", "Point at which Gate is to be added", "Order index of where gate falls in River", "Name of Gate" });
    addCommand(newPK_Category_Rivers, "SetChemicalInRiverWater", "Sets the presence of a chemical in river water", false, false, 
        "Sets the presence of a chemical in river water, at the specified concentration", 
        new String[]{ "River", "Chemical", "Concentration" });
    
      /* Canal Systems */
    addCommand(newPK_Category_CanalSystems, "CreateCanalSystem", "Adds a new canal system to the model", false, false,
        "Create a canal system", 
        new String[]{ "Canal System Name" });
    addCommand(newPK_Category_CanalSystems, "AddCanalToCanalSystem", "Adds a canal to the canal system", false, false,
        "Add a new canal to a canal system. The new canal is not connected to any previously entered canals, even if the end coordinates match another canals.", 
        new String[]{ "Name of canal system", "Starting point for canal segment", "End point for canal segment", "Name for new canal segment" });
    addCommand(newPK_Category_CanalSystems, "SetCanalSegmentProfile", "Sets the profile of the specified canal segment", false, false, 
        "Sets the profile of a canal segment. The profile is specified by a Profile Specification String like TRAP:5;3;1", 
        new String[]{ "Name of the canal system", "Name of the canal segment", "Profile Specification String" });
    addCommand(newPK_Category_CanalSystems, "SetCanalSegmentType", "Sets the canal segment type", false, false, 
        "Sets the canal segment type", new String[] { "Name of the canal system", "Name of the canal segment", "Type"});
    addCommand(newPK_Category_CanalSystems, "ExtendCanal", "Extends a specified canal in the given canal system to the point specified", false, false, 
        "Extend one canal to another point.", 
        new String[]{ "Name of canal system for new canal", "Name of canal segment being extended", "Name of point to which canal is being extended", "Name for new cana segment" });
    addCommand(newPK_Category_CanalSystems, "ExtendLastCanal", "Extends the last canal added in the given canal system to the specifed point", false, false, 
        "Extend the last saved canal to the specifed point.", 
        new String[]{ "Name of canal system", "Name of point to which canal is being extended", "Name of new canal segment" });
    addCommand(newPK_Category_CanalSystems, "SetCanalRank", "Sets the profile of a canal using preset ranks", false, false, 
        "Specify the name of the canal segment and the rank, as an integer", 
        new String[]{ "Name of the canal system", "Name of the canal segment", "Rank, as integer" });
    addCommand(newPK_Category_CanalSystems, "AddCanalAtGate", "Adds a canal beginning at the specified headgate", false, false,
        "Adds a canal beginning at the specified headgate",
        new String[]{ "River Name", "Canal System Name", "Headgate name", "Point to which Canal Segment will extend", "New Canal Segment Name"});
    addCommand(newPK_Category_CanalSystems, "SetUpstreamDischarge", "Sets the upstream discharge of a canal segment", false, false, 
        "Sets the upstream discharge for the specified segment of the specified canal system to the given value", 
        new String[]{ "Canal System name", "Canal Segment name", "New value for upstream discharge" });
    addCommand(newPK_Category_CanalSystems, "SetUpstreamStage", "Sets the upstream stage", false, false, 
        "Sets the upstream stage value for the specified segment of the given canal system to the specified value", 
        new String[]{ "Canal System name", "Canal Segment name", "New value for upstream stage" });
    addCommand(newPK_Category_CanalSystems, "SetDownstreamDischarge", "Sets the downstream discharge of a canal segment", false, false, 
        "Sets the downstream discharge for the specified segment of the specified canal system to the given value", 
        new String[]{ "Canal System name", "Canal Segment name", "New value for downstream discharge" });
    addCommand(newPK_Category_CanalSystems, "SetDownstreamStage", "Sets the downstream stage", false, false, 
        "Sets the downstream stage value for the specified segment of the given canal system to the specified value", 
        new String[]{ "Canal System name", "Canal Segment name", "New value for downstream stage" });
    addCommand(newPK_Category_CanalSystems, "SetUpstreamQperY", "Sets the upstream stage and discharge for a specified canal segment", false, false, 
        "Sets the upstream stage and the discharge, as it would be for that stage given Mannings equation, for the given segment of the specified canal", 
        new String[]{ "Canal System name", "Canal Segment name", "New depth (not stage); Q will be calculated." });
    addCommand(newPK_Category_CanalSystems, "SetDownstreamQperY", "Sets the downstream stage and discharge for a specified canal segment", false, false, 
        "Sets the downstream stage and the discharge, as it would be for that stage given Mannings equation, for the given segment of the specified canal", 
        new String[]{ "Canal System name", "Canal Segment name", "New depth (not stage); Q will be calculated." });
    addCommand(newPK_Category_CanalSystems, "SetMaximumOutflow", "Sets the artificial maximum outflow for a canal segment", false, false,
        "Sets the artificial maximum outflow for a canal segment",
        new String[]{ "Canal System name", "Canal Segment name", "New Value; if negative, clears value"});
    addCommand(newPK_Category_CanalSystems, "SetCanalSegmentSize", "Sets the dimensions of a canal segment", false, false, 
        "Sets the dimensions (top, bottom, height) for the specified canal segment in the given canal systems", 
        new String[]{ "Canal System name", "Canal Segment name", "New top width", "New bottom width", "New channel height" });
    addCommand(newPK_Category_CanalSystems, "SetCanalSegmentRoughness", "Sets the roughness for a specified canal segment", false, false, 
        "Sets the roughness (Mannings n) for the specified canal segment of the given canal system to the specified value", 
        new String[]{ "Canal System name", "Canal Segment name", "New roughness value" });
    addCommand(newPK_Category_CanalSystems, "SetUpstreamZandQ", "Sets the upstream stage and discharge", false, false, 
        "Sets the upstream stage and discharge for the specified Canal Segment", 
        new String[]{ "Canal System name", "Canal Segment name", "New stage", "New discharge" });
    addCommand(newPK_Category_CanalSystems, "SetDownstreamZandQ", "Sets the downstream stage and discharge", false, false, 
        "Sets the downstream stage and discharge for the specified canal segment", 
        new String[]{ "Canal System name", "Canal Segment name", "New stage", "New discharge" });
    addCommand(newPK_Category_CanalSystems, "SetHeadGateEfficiency", "Sets the efficiency (0-1) of a headgate in the specified river", false, false, 
        "Sets the efficiency of a specified HeadGate in a specified river to the given value", 
        new String[]{ "Name of the river in which the HeadGate is found", "Name of the HeadGate", "Efficiency (0.0 - 1.0)" });
    addCommand(newPK_Category_CanalSystems, "OpenHeadGate", "Opens a headgate", false, false, 
        "Opens a headgate", 
        new String[]{ "Name of the river in which the HeadGate is found", "Name of the HeadGate" });
    addCommand(newPK_Category_CanalSystems, "CloseHeadGate", "Closes a headgate", false, false, 
        "Closes a headgate", 
        new String[]{ "Name of the river in which the HeadGate is found", "Name of the HeadGate" });    
    addCommand(newPK_Category_CanalSystems, "SetHGEfficiencyRanges", "Sets the efficiency of a headgate by range of values", false, false, 
        "Sets the efficiency of a headgate by range of values", 
        new String[]{ "Name of the river in which the HeadGate is found", "Name of the HeadGate", "Efficiency ranges, by string of ordered pairs" });
    addCommand(newPK_Category_CanalSystems, "CloseInGate", "Closes a canal seg inflow gate", false, false, 
        "Closes the inflow gate for a canal segment", 
        new String[]{ "Canal System Name", "Canal Segment Name"});
    addCommand(newPK_Category_CanalSystems, "OpenInGate", "Opens a canal seg gate", false, false, 
        "Opens the outflow gate for a canal segment", 
        new String[]{ "Canal System Name" });
    addCommand(newPK_Category_CanalSystems, "CloseOutGate", "Closes a canal seg outflow gate", false, false, 
        "Closes the outflow gate for a canal segment", 
        new String[]{ "Canal System Name", "Canal Segment Name"});
    addCommand(newPK_Category_CanalSystems, "OpenOutGate", "Opens a canal seg gate", false, false, 
         "Opens the outflow gate for a canal segment", 
         new String[]{ "Canal System Name" });
    addCommand(newPK_Category_CanalSystems, "GenerateHECOutput", 
        "Generates a file that can import the existing canal system into the HEC-RAS software package", true, false,
        "Generates a file that can import the existing canal system into the HEC-RAS software package");

      /* Fields */
    addCommand(newPK_Category_Fields, "CreateField", "Create a field.", false, false, 
        "Create a field.", 
        new String[]{ "NW point of new field", "NE point of new field", "SE point of new field", "SW point of new field", "Name to be assigned to this field" });
    addCommand(newPK_Category_Fields, "DropField", "Removes a field from the model completely.", false, false, 
        "Removes a field from the model completely.", 
        new String[]{ "The name of the field to be dropped" });
    addCommand(newPK_Category_Fields, "LinkCanalSegToField", "Links a canal segment to a field", false, false,
        "Links a canal segment to a field",
        new String[]{ "Canal System Name", "Canal Segment Name", "Field Name"});
    addCommand(newPK_Category_Fields, "UnlinkCanalSegFromField", "Unlinks a canal segment from a field", false, false,
        "Unlinks a canal segment from a field",
        new String[]{ "Canal System Name", "Canal Segment Name", "Field Name"});
    addCommand(newPK_Category_Fields, "PlantInAll", "Plant a plant in all fields at the specified density", false, false, 
        "Plants a plant in all fields at the specified density", 
        new String[]{ "The name of the plant to be planted", "The density at which the plant is to be planted" });
    addCommand(newPK_Category_Fields, "SetEToForField", "Sets the ETo for the specified field", false, false, 
        "Sets the Evapotranspiration rate for the specified field", 
        new String[]{ "Field for which ETo is to be set", "New ETo for the specified field" });
    addCommand(newPK_Category_Fields, "PlantInSelectedFields", "Plants the specified plant at the given density in the selected Fields", false, false, 
        "Plants the given plant at the specified density in the selected Fields", 
        new String[]{ "Variety to be planted", "Density at which to plant"} );
    addCommand(newPK_Category_Fields, "SetFieldEffectiveArea", "Sets a field`s effective area", false, false, "Sets a field`s effective area", 
        new String[]{"Name of the field", "New effective area" });
    addCommand(newPK_Category_Fields, "HarvestPlantInField", "Harvests a plant in a field", false, false, "Harvests a plant in a field",
        new String[]{"Name of the field", "Name of the Plant", "Index of the Plant"});
    
    
      /* Rain */
    addCommand(newPK_Category_Rain, "RainOnField", "Rain on a field", false, false, 
        "Causes a set amount of rain to fall on a specific field", 
        new String[]{ "Name of the field on which rain should fall", "Amount of rain that should fall, in mm" });
    addCommand(newPK_Category_Rain, "RainOnAllFields", "Rain on all fields", false, false, 
        "Causes a set amount of water to fall on all fields", 
        new String[]{ "Amount of rain to fall, in mm" });
    addCommand(newPK_Category_Rain, "RainOnSelectedFields", "Rains on selected fields", false, false,
        "Causes the specified amount of water to fall on selected fields", new String[]{"Amount of rain to fall, in mm"});
    addCommand(newPK_Category_Rain, "RainAtElevations", "Rains specified mm on fields between elevations", false, false,
         "Rains specified mm on fields between elevations", new String[] { "amount of rain, in mm", "lower elevation boundary", "upper elevation boundary"});

  /* Interrogatives */

    /* Environment */
    addCommand(newPK_Category_Environment, "GetMeanDailyTemp", "Gets the mean daily temperature", false, true, 
        "Gets the mean daily temperature");
    addCommand(newPK_Category_Environment, "GetMeanDailyTempAtAlt", "Gets the mean daily temperature at altitude", false, true,
        "Gets the mean daily temperature at a specified altitude", new String[]{"Altitude"});

    /* Rivers */
    addCommand(newPK_Category_Rivers, "GetStreamflowForRiver", "Get the streamflow value for a particular river", false, true, 
        "Get the streamflow for a given river.", 
        new String[]{ "Name of the river" });
    addCommand(newPK_Category_Rivers, "GetFlowLevelForRiver", "Get the flow level for a river", false, true, 
        "Get the annual flow rate for a given river", 
        new String[]{ "Name of river" });
    addCommand(newPK_Category_Rivers, "GetDailyWaterForRiver", "Get the amount of water (MAF) flowing per day in the river given its Streamflow and FlowLevel Values", false, true, 
        "Get the daily flow rate for a given river", 
        new String[]{ "Name of river" });
    addCommand(newPK_Category_Rivers, "GetFlowAtHeadGate", "Gets the flow, in Million Acre-Feet per day, at a given HeadGate", false, true, 
        "Gets the flow from the specified river reaching the specified headgate, in million acre-feet per day.", 
        new String[]{ "Name of the river", "Name of the headgate" });

    /* Canal Systems */
    addCommand(newPK_Category_Interrogative, "GetTotalWaterVolume",       "Get the total volume of water in all canals in all canal systems", true,  true, 
        "Get the total volume of water in all canal systems.");
    addCommand(newPK_Category_CanalSystems, "GetTotalWaterVolumeIn", "Get the total volume of water in the specified canal system", false, true, "Get the total volume of water in the specified canal system/", new String[]{ "Name of canal system" });
    addCommand(newPK_Category_Interrogative, "GetTotalVolumeOfCanals",    "Get the total volume of all canals in all canal systems",          true,  true, 
        "Get the total volume of the canals in the simulation (all canal systems)" );
    addCommand(newPK_Category_CanalSystems, "GetCanalSegDSStageAt", "Gets the stage at a give time for a given Canal Segment", false, true, 
        "Gets the stage at a given time for a given canal segment", 
        new String[]{ "Canal System Name", "Canal Segment Name", "Time (as fraction of day, i.e. 0.2)" });
    addCommand(newPK_Category_CanalSystems, "GetCanalSegDSDischarge", "Gets the downstream discharge for a specific canal segment", false, true,
        "Gets the downstream discharge for a specific canal segment", 
        new String[]{"Canal System Name", "Canal Segment Name"});
    addCommand(newPK_Category_CanalSystems, "GetConcentrationInSegment", "Gets the concentration of a chemical in a specific canal segment",       false, false, 
        "Gets the concentration of a specified chemical in a specific canal segment");
    addCommand(newPK_Category_Interrogative, "GetTotalLengthOfAllCanals", "Get the total length of all canals in the model",                  true,  true, 
        "Returns the total length of all canals in the model");
    addCommand(newPK_Category_Interrogative, "GetCanalSysHarvestLastYr", "Gets the total harvest of a canal system last year", false, true,
        "Gets the total harvest of a canal system last year",	new String[]{"The name of the canal system"});

    /* Fields */
    addCommand(newPK_Category_Fields, "GetFieldSoilChem", "Gets the concentration of a chemical in a field", false, true, 
        "Gets the concentration of a chemical in a field", new String[]{"Name of the field", "Name of the chemical"});
    addCommand(newPK_Category_Fields, "GetFieldPotentialYield", "Gets a value for the current potential yield of the specified crop", true, true,
        "Use this to get the potential yield for a specific crop in a field", new String[]{ "Name of the field", "Name of the plant", "Index of Crop"});
    addCommand(newPK_Category_Fields, "GetFieldHarvestableYield", "Gets a value for the current harvestable yield of the specified crop", true, true,
        "Use this to get the harvestable yield for a specific crop in a field", new String[]{ "Name of the field", "Name of the plant", "Index of Crop"});
    addCommand(newPK_Category_Fields, "GetFieldSoilWaterDepth", "Gets the depth of water available in the specified field", false, true,
        "Gets the depth of water available in the specified field", new String[] { "The name of the field"});
    addCommand(newPK_Category_Fields, "GetFieldWaterRecd", "Gets the water received by a field before the last step method", false, true,
        "Gets the water received by a field before the last step method", new String[]{ "The name of the field"});
    addCommand(newPK_Category_Fields, "GetWaterRecdInSelFields", "Gets the total water received in all selected fields", false, true,
        "Gets the total water received in all selected fields");
    addCommand(newPK_Category_Fields, "GetFieldSumHarvestLastYr", "Gets the total harvest from the field last year", false, true,
        "Gets the total water received in all selected fields", new String[]{"The name of the field"});
    
    /* Rain */


  /* Selection and Object Management */

    /* Environment */

    /* Rivers */
    
    /* Canal Systems */
    addCommand(newPK_Category_CanalSystems, "SetCanalSegmentProxy", "Sets a proxy name for a canal segment", false, false, 
        "Allows the simulation to refer to a canal segment by a second name", 
        new String[]{ "Canal System", "Original name of the canal segment", "Proxy name of the canal segments" });
   
    /* Fields */
    addCommand(newPK_Category_Fields, "SelectField",                     "Selects the Field",                                                      false, false, 
        "Selects the specified Field", 
        new String[]{ "Name of field to be selected"  } );
    addCommand(newPK_Category_Fields, "UnselectField",                   "Unselected the specified Field",                                         false, false, 
        "Unselects the specified field", 
        new String[]{ "Name of field to be unselected"   } );
    addCommand(newPK_Category_Fields, "UnselectAllFields",               "Unselects all Fields",                                                   false, false, 
        "Unselects all Fields");
    
    /* Rain */
    
  /* Reports */
    /* Environment */
    /* Rivers */    
    /* Canal Systems*/
    // View Canal System Report
    addCommand(newPK_Category_CanalSystems, "ViewCanalSystemDetails", "Views the report for a single canal system", true, false, 
        "- instructions yet to be written -", 
        new String[]{ "Name of canal system for which report is requested" });

    /* Fields */
    addCommand(newPK_Category_Fields, "GetFieldReport", "Gets a report showing the details of a specified field", true, false, 
        "Use this to review the status of the crops in a specific field", new String[]{ "Name of the field" });
    
    /* Rain */


    addCommand(newPK_Category_View, "View3D", "View a canal system in 3D", true, false, "Opens a VRML representation of a canal system.", 
              new String[]{  "The name of the canal system" } );
        
   // Setting streamflow
    long cat = addCodeModelDomain("Set Streamflow Procedure", "", "SET_STREAMFLOW_PROC");
      addCodeModelDomainOptionValue(cat * 100 + 1, cat,  0, "SET_STREAMFLOW_DEFAULT", setStreamflowDefaultCode, "Streamflow value is set to exactly the value specified");
      addCodeModelDomainOptionValue(cat * 100 + 2, cat,  1, "SET_STREAMFLOW_STOCHASTIC", setStreamflowStochasticCode, "A stochastic adjustment is made to the streamflow value to be set");
      addCodeModelDomainOptionValue(cat * 100 + 3, cat,  2, "SET_STREAMFLOW_SCALED", setStreamflowScaledCode, "Scales the streamflow value by a constant (set by parameter)");
      
    
    cat = addCodeModelDomain("Agent creation", "Which code WMPopulation uses to create the population", "AGENT_CREATION");
      addCodeModelDomainOptionValue(cat * 100 + 1, cat, 0, "NO_AGENTS", "TBD", "No agents are created" );
      addCodeModelDomainOptionValue(cat * 100 + 2, cat, 1, "ONE_AGENT_PER_FIELD", "TBD", "One Field Manager Agent");
      addCodeModelDomainOptionValue(cat * 100 + 3, cat, 2, "ONE_AGENT_PER_FIELD_WITH_CS_MANAGER", "TBD", "One Field Manager Agent with CS Manager");
  }
  
  
}
