package wm.model;

import java.util.HashSet;
import java.util.Set;

import wm.model.agents.WMPopulation;

import abcm.model.ABCMResult;
import abcm.model.agents.AbstractABCMActor;
import abcm.model.agents.AbstractABCMAgent;

/**
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class WMAgent extends AbstractABCMAgent{
  
  /* Actions */
  
  /**
   * An action that returns a list of this agent's fields
   */
  protected class Action_ListFields extends Action{
    /**
     * @see abcm.model.agents.AbstractABCMActor.Action#execute(abcm.model.agents.AbstractABCMActor, java.lang.String[])
     */
    public ABCMResult execute(AbstractABCMActor actor, String[] command){
      return new ABCMResult(((WMAgent)actor).getFieldList(), true);
    }
  }
  
  /* Parser */
  
  /**
   * A parser that includes the additional commands required by a
   * WatMan agent
   */
  protected class WMAgentCommandParser extends ActorCommandParser{
    /**
     * Constructor
     */
    public WMAgentCommandParser(){
      addCommand("LIST_FIELDS", new Action_ListFields());
    }
  }  
  
  /* Static variables */
  /** A static ActorCommandParser; only one will exist per class  */
  protected static WMAgentCommandParser parser;

  /* Instance variables */
  
  /** Pointer to the WatMan model in which this agent will exist; can be used to avoid casts  */
  protected WMModel myModel;
  
  /** Collection of fields under the control of this agent  */
  protected Set<String> fields;
  

  
  /* Constructor */
  
  /**
   * Class Constructor; requires a <code>Model</code> 
   * 
   * In implementation, the <code>Model</code> object is available so that the <code>Agent</code>
   * object can pass instructions to the simulation; the <code>Code Model</code> object
   * is available to determine alternative actions or algorithms for the Agent object.
   */
  public WMAgent(WMModel model, WMPopulation population){
    super(model, population);
    myModel = model;
    fields = new HashSet<String>();
    if(parser == null) parser = new WMAgentCommandParser();
  }

  /* Required Methods */
  
  /**
   * @see abcm.agents.AbstractABCMAgent#getAgentIDPrefix()
   */
  protected String getAgentIDPrefix(){  return "WMA-";  }

  /**
   * @see abcm.agents.AbstractABCMAgent#getAgentIDSuffix()
   */
  protected String getAgentIDSuffix(){  return ""; }

  /**
   * @see abcm.agents.AbstractABCMActor#getActorCommandParser()
   */
  protected ActorCommandParser getActorCommandParser(){ return parser; }
  
  /* Step Method */
  /**
   * General step method for an agent. Actions taken will depend 
   * on the Code Model in place for this agent.
   */
  public void step(){
     // System.out.println("I'm agent " + this.getID() + " and I'm stepping!");
  }
  
  /* Action Methods */

  
  /* Reporting */
  
  /**
   * Gets a report for this agent.
   * @return a String containing the status of this agent
   */
  public String report(){
    String retVal = "";
    retVal += "Agent: " + getID() + "\n";
    retVal += "  Fields: " + (fields != null ? fields.size() : 0) + "\n";
    return retVal;
  }

  /**
   * Gets a list of the fields this agent owns
   * @return the list of fields this agent owns
   */
  public String getFieldList(){
    String report = "Fields: \n";
    for(String s: fields) report += "   " + s + "\n";
    report += "\n";
    return report;
  }
  
  /**
   * Assigns a field to this WatMan agent. The agent can then manage and monitor the field.
   * @param name Name of the field to be added
   */
  public void assignField(String name){    fields.add(name);   }

  /**
   * Selects all of this agent's fields, unselecting all
   * others
   */
  public void selectAllMyFieldsExclusively(){
    myModel.unselectAllFields();
    for(String fieldName: fields) myModel.processCommand("SelectField " + fieldName);
  }
  
  
  
}
