package wm.model;


import java.io.File;
import java.io.FileOutputStream;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import wm.model.agents.WMPopulation;
import wm.sim.WMConfiguration;
import wm.space.WMSpace;
import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.graphics.SVGRepresentable;
import abcm.math.Point;
import abcm.model.ABCMResult;
import abcm.model.AbstractABCModel;
import abcm.model.agents.AbstractABCModelPopulation;
import abcm.sim.SimulationRun;
import abcm.space.Space;
import flowag.basic.BasicChemical;
import flowag.basic.Chemical;
import flowag.basic.Environment;
import flowag.basic.Water;
import flowag.bio.Field;
import flowag.bio.Plant;
import flowag.channels.CanalSegment;
import flowag.channels.CanalSystem;
import flowag.channels.Gate;
import flowag.channels.HeadGate;
import flowag.channels.River;

/**
 * The WatMan model object defines the ABCM model that concerns the creation
 * of canals and canal systems, the flow of water through those systems,
 * the construction and operation of fields based on those systems, and 
 * other miscellaneous elements of modeling an irrigation-based farming system.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class WMModel extends AbstractABCModel implements SVGRepresentable{

  // Private values
  // System of canal segment ranks may not be used pending further revisions...
//  /** Specifies a portion of a canal that is a main line  */
//  private static final int TRUNK     = 1;
//  /** Specifies a secondary line of a canal  */
//  private static final int SECONDARY = 2;
//  /** Specifies a tertiary canal line  */
//  private static final int TERTIARY  = 3;

  // Code Model Switches
  /** Used to specify that the set streamflow method should use the default approach  */
  public static final int SET_STREAMFLOW_DEFAULT     = 0;
  /** Used to specify that the set streamflow method should use a scaled approach */
  public static final int SET_STREAMFLOW_SCALED      = 2;
  /** Used to specify that the set streamflow method should use the stochastic approach  */
  public static final int SET_STREAMFLOW_STOCHASTIC  = 1;
  
  /** Natural environment in which model actions occur  */
  private Environment E;  
  
  // Simulation objects
  /** Collection of Rivers managed by this model  */
  private ABCMDeepCloneableHashtable rivers       = new ABCMDeepCloneableHashtable();
  /** Collection of Canal Systems managed by this model  */
  private ABCMDeepCloneableHashtable canalSystems = new ABCMDeepCloneableHashtable();
  /** Collection of Fields managed by this model  */
  private ABCMDeepCloneableHashtable fields       = new ABCMDeepCloneableHashtable();

  // Model for rainwater; rainwater can have source concentrations
  /** A source for rainwater (contains no concentrations)  */
  private Water rainSource        = new Water (1);
  
  // Space
  /** Pointer to the model's Space object; must be a WMSpace  */
  private WMSpace space          = new WMSpace();

  /* Constructors */
  /**
   * Constructor
   * @param env Environment (model environment; must be a WMModelEnvironment) in which this model will be run
   */
  public WMModel(WMModelEnvironment env){    
    super(env);    
    initializeNaturalEnvironment();
  }
  
  /**
   * Constructor
   * @param env Environment (model environment; must be a WMModelEnvironment) in which this model will be run
   * @param sr Simulation Run associated with this model
   */
  public WMModel(WMModelEnvironment env, SimulationRun sr){    
    super(env, sr);
    initializeNaturalEnvironment();
  }
  
  /* Initialization */
  /**
   * Initialize the environment. Optionally can enter code model
   * options here
   */
  protected void initializeNaturalEnvironment(){
    E = new Environment();
    E.setRainSource(rainSource);
  }

  /* Clone this object */
  /**
   * @see abcm.model.AbstractABCModel#getModelClone(abcm.cloning.ABCMCloneStore)
   */
  public AbstractABCModel getModelClone(ABCMCloneStore cloneStore){
    WMModel newModel = new WMModel((WMModelEnvironment)env);
    if(cloneStore == null) cloneStore = new ABCMCloneStore();
    
    super.cloneModelCore(newModel, cloneStore);

    newModel.E             = (E            != null ? (Environment)E.getDeepClone(cloneStore) : null);
    newModel.rivers        = (rivers       != null ? (ABCMDeepCloneableHashtable)rivers.getDeepClone(cloneStore) : null);
    newModel.rainSource    = (rainSource   != null ? (Water)rainSource.getDeepClone(cloneStore) : null);
    newModel.space         = (space        != null ? (WMSpace)space.getDeepClone(cloneStore) : null);
    newModel.canalSystems  = (canalSystems != null ? (ABCMDeepCloneableHashtable)canalSystems.getDeepClone(cloneStore) : null);
    newModel.fields        = (fields       != null ? (ABCMDeepCloneableHashtable)fields.getDeepClone(cloneStore) : null);
    
    return newModel;
    
  }
  

  /* Step Actions  */
  
  /** 
   * Fundamental step method for this model. Performs the following actions:
   * <ul>
   *   <li>Flows all water in rivers and canals, and allows rivers to update condition</li>
   *   <li>Allows plants to grow</li>
   *   <li>Allows agents to take actions</li>
   * </ul>  
   * @see abcm.model.AbstractABCModel#modelStep() 
   */
  public void modelStep(){
//    System.out.println("FLOWING IN MODEL STEP");
    // FLOW ALL WATER
    flow();
    
//    System.out.println("RAINING IN MODEL STEP");
    // Rain
    rain();
    
//    System.out.println("GROWING IN MODEL STEP");
    // GROW ALL PLANTS
    grow();
    
    // SPECIAL DEBUGGING CODE- TEST ONLY!
//    System.out.println("River Daily Flow: " + this.processCommand("GetDailyWaterForRiver DefaultRiver").value);
//    System.out.println(processCommand("OpenHeadgate DefaultRiver HG1").text);
//    System.out.println(processCommand("OpenHeadgate DefaultRiver HG2").text);
//    System.out.println("Flow into N Canal System: " + this.processCommand("GetCanalSegDSDischarge NorthCanalSys NorthSeg2").text);
//    CanalSystem N = (CanalSystem)getCanalSystem("NorthCanalSys");
//    if(N != null) System.out.println(N.report());
    
    
  }
  
  /**
   * Flow water through the canal systems.
   *
   * Rivers must first be stepped.
   *
   * Order of Canal Systems is not randomized, in
   * part because there will most commonly be only
   * one Canal System, but also because there
   * is presumed to be no interaction between Canal
   * Systems.
   */
  public void flow(){
    Enumeration r = rivers.elements();
    while(r.hasMoreElements()){
      River R = (River)r.nextElement();
      R.step();
    }
    
    Enumeration e = canalSystems.elements();
    while(e.hasMoreElements()){
      CanalSystem c = (CanalSystem)e.nextElement();
      c.flow(1);
    }
  }
  
  /**
   * Causes environment to rain on all fields
   */
  public void rain(){
    E.rain(fields.values());
  }
  
  /**
   * Allow all plants in all fields to grow one day's worth
   */
  public void grow(){
    Enumeration e = fields.elements();
    while(e.hasMoreElements()){
      ((Field)e.nextElement()).step();
    }   
  }
  
  
  /** 
   * @see abcm.model.AbstractABCModel#getSpace() 
   */
  public Space getSpace(){    return space;  }


  /* ************************************************************************** */
  /*  Internal retrieval                                                        */
  /* ************************************************************************** */
  
    /* Rivers */
  
  /**
   * Gets a River object by name
   * @param riverName the name of the River object requested
   * @return the requested River object; returns null if no River object with
   * that name is found
   */
  private River getRiver(String riverName){
    // If the river specified is in the hashtable, return it
    if(rivers.containsKey(riverName)) return ((River)rivers.get(riverName));
    // If not, then if there is only one river, return that one as a default
    if(rivers.size() == 1)            return (River)rivers.elements().nextElement();
    // If there are several rivers, and none of them match this name,
    // this is a genuine ambiguity; return null
    return null;
  }

  /**
   * Returns an ABCMResult object indicating that the requested River was not found
   * @param action the action requested
   * @param riverName the name of the River sought
   * @return an ABCM result object indicating that the requested River was not found
   */
  private ABCMResult riverNotFound(String action, String riverName){
    return new ABCMResult("Could not " + action + "; River '" + riverName + "' does not exist.", false);
  }
  
  
    /* Canal Systems */
  
  /**
   * Gets a hashtable of all canal systems, keyed by name
   * @return the map of canal system names and systems
   */
  public Hashtable<String, CanalSystem> getAllCanalSystems(){
    Hashtable<String, CanalSystem> t = new Hashtable<String, CanalSystem>();
    for(Object n: canalSystems.keySet()) t.put((String)n, (CanalSystem)canalSystems.get(n));    
    return t;
  }
  
  /**
   * Gets a canal system as requested by name. If there is only one 
   * canal system, that system is returned regardless of the name supplied.
   * @param csName the name requested
   * @return the CanalSystem if it exists and is referred to by the specified
   * name or proxy name; null otherwise
   */
  public CanalSystem getCanalSystem(String csName){
    // If the canal system name is in the list, return it
    if(canalSystems.containsKey(csName)) return (CanalSystem)canalSystems.get(csName);
    
    // If there is only one canal system, return that one regardless of the name passed
    if(canalSystems.size() == 1)         return (CanalSystem)(canalSystems.elements().nextElement());

    // Otherwise, there's no way to tell which canal system is meant- return null
    return null;
  } 

  /**
   * Gets the single returnable ID for a specified canal segment (which
   * may have alternative proxy names)
   * @param C the CanalSegment object requested
   * @param canalSegmentID an ID for this canal segment
   * @return the returnable ID fo the specified canal segment; if none exists, returns 
   * the CanalSegments own ID with single quote marks
   */
  private String canalSegmentReturnID(CanalSegment C, String canalSegmentID){
    String proxy = "'" + canalSegmentID + "'";
    if(canalSegmentID.compareTo(C.getCanalSegmentID()) > 0) proxy = "'" + canalSegmentID + "' [" + C.getCanalSegmentID() + "]";
    return proxy;
  }
  
    /* Canal Segments */
  
  /**
   * Sets a canal segment's proxy name
   * @param canalSystemName the name of the canal system that contains this segment
   * @param originalName the original name of the canal segment
   * @param newName the new proxy name for the segment
   * @return an ABCMResult indicating the success of the operation (will fail if specified Canal System does not exist)
   */
  public ABCMResult setCanalSegmentProxyName(String canalSystemName, String originalName, String newName){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return new ABCMResult("Could not set segment proxy name; no canal system '" + canalSystemName + " exists.", false);
    return new ABCMResult(CS.addCanalSegmentProxyName(originalName, newName), true);
  }
  
  /**
   * Returns an ABCMResult object indicating that the requested canal system could not be found
   * @param action a String identifying the action to be performed
   * @param canalSystemName the name of the canal system to be returned
   * @return a result (success=false) indicating the failure to retrieve the requested canal system
   */
  private ABCMResult canalSystemNotFound(String action, String canalSystemName){
    return new ABCMResult("Could not " + action + "; canal system '" + canalSystemName + "' not found", false);
  }
  
    /* Fields */
  
  /**
   * Gets a field requested by name
   * @param fieldName the field requested
   * @return the requested field, if it exists; null otherwise
   */
  private Field getField(String fieldName){
    if(!fields.containsKey(fieldName)) return null;
    return (Field)fields.get(fieldName);
  }

  /**
   * Gets a Vector that contains a list of all fields; includes 'free' fields
   * and fields associated with Canal Systems
   * @return a Vector that lists all fields
   */
  public HashSet getAllFields(){
    HashSet fieldSet = new HashSet(fields.values());    
    Enumeration CSS = canalSystems.elements();
    while(CSS.hasMoreElements()) fieldSet.addAll(((CanalSystem)(CSS.nextElement())).getAllFields());
    return fieldSet;
  }  
  
  /**
   * Returns an ABCMResult object indicating that the Field object requested was not
   * found in the model
   * @param action a String identifying the action to be performed
   * @param fieldName the name of the field to be returned
   * @return a result (success=false) indicating the failure to retrieve the requested field
   */
  private ABCMResult fieldNotFound(String action, String fieldName){ 
    return new ABCMResult("Could not " + action + "; Field '" + fieldName + "' does not exist.", false);
  }
  
  
  /* ************************************************************************** */  
  /*                           Object Selection                                 */
  /*                                                                            */
  /* Allows selection of one or multiple instances of simulation objects.       */
  /*                                                                            */
  /* Actions may act upon selected objects.                                     */
  /*                                                                            */
  /* ************************************************************************** */  
  
  /* Rivers         */
  // Unselect All
  // Select One
  // Unselect One
  // Select All
  /* Canal Systems  */
  // Unselect All
  // Select One
  // Unselect One
  // Select All  
  /* Canal Segments */
  // Unselect All
  // Select One
  // Unselect One
  // Select All
  // Select All by Group
  // Unselect All in Group
  
  
  /**
   * De-Selects all canal segments
   */
  public void unselectAllCanalSegments(){
    Enumeration css = canalSystems.elements();
    while(css.hasMoreElements()){
      ((CanalSystem)css.nextElement()).unselectAllCanalSegments();
    }
  }
  

  /* Fields         */
  // Select One
  // Unselect One
  // Select All
  // Unselect All
  // Select All by Group
  // Unselect All in Group
  
  /**
   * Selects a specific field
   * @param fieldName the field to be selected
   * @return an ABCMResult object indicating success or failure
   */
  public ABCMResult selectSingleField(String fieldName){
    Field f = getField(fieldName); if(f == null) return fieldNotFound("select field", fieldName);
    f.select();
    return new ABCMResult("Selected Field '" + fieldName + "'.", true);
  }
  
  /**
   * Selects a collection of fields by a list of names 
   * @param fieldNames the names of the fields to be selectedd
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult selectField(String fieldNames){
    String[] names = fieldNames.split(",");
    boolean result = true;
    String resultString = "";
    for(String name: names){
//      System.out.println("Attempting to select field: " + name);
      ABCMResult res = selectSingleField(name);
      resultString += res.text;
      result = result & res.success;
    }
    return new ABCMResult(resultString, result);
  }
  
  /**
   * De-selects a specific field
   * @param fieldName the name of the field to be de-selected
   * @return an ABCMResult object indicating success or failure
   */
  public ABCMResult unselectField(String fieldName){
    Field f = getField(fieldName); if(f == null) return fieldNotFound("unselect field", fieldName);
    f.unSelect();
    return new ABCMResult("Unselected Field '" + fieldName + "'.", true);
  }

  /**
   * Selects all fields, including unassociated fields
   * @return an ABCMResult object indicating the success of this operation, including
   * the number of fields.
   */
  public ABCMResult selectAllFields(){
    HashSet H = getAllFields();
    Iterator I = H.iterator();
    while(I.hasNext()) ((Field)I.next()).select();
    return new ABCMResult("All fields are selected (total: " + H.size() + ")", H.size(), true);
  }
  
  /**
   * De-Selects all fields
   * @return an ABCMResult object indicating success or failure
   */
  public ABCMResult unselectAllFields(){
    HashSet H = getAllFields();
    Iterator I = H.iterator();
    while(I.hasNext()) ((Field)I.next()).unSelect();
    return new ABCMResult("All fields are unselected.", true);
  }
  

  /**
   * Selects fields within certain elevation boundaries
   * @param lowerbound the lower bound for elevation
   * @param upperbound the upper bound for elevation
   * @return an ABCMResult object indicating the success of the operation
   * and the number of fields selected
   */
  public ABCMResult selectFieldsBetweenElevations(double lowerbound, double upperbound){
    HashSet H = getAllFields();
    Iterator I = H.iterator();
    int c = 0;
    while(I.hasNext()){
      Field F = (Field)I.next();
      if((F.getMinZ() >= lowerbound) && F.getMaxZ() <= upperbound){
        F.select();
        c++;
      }
    }
    return new ABCMResult("Selected " + c + " fields.", c, true);
  }
  
  
  /* ************************************************************************** */  
  /*                               Model Actions                                */
  /* In general the list of model actions is the list of things that the model  */
  /* can 'do', or ways the user can interact with the model and set its         */
  /* internal parameters.                                                       */
  /* ************************************************************************** */
  
  /* Environment */
  /**
   * Sets the probability of rainfall via the Environment engine
   * @param newProb new probability value
   * @return an ABCMResult indicating the success or failure of this action
   */
  public ABCMResult setBasicDailyRainfallProbability(double newProb){
    E.setBasicDailyRainfallProbability(newProb);
    return new ABCMResult("Set basic daily rainfall probability to " + newProb, true);
  }
  
  /**
   * Sets the basic daily rainfall amount via the Environment engine
   * @param newAmt new amount value
   * @return an ABCMResult indicating the success or failure of this action
   */
  public ABCMResult setBasicDailyRainfallMinAmount(double newAmt){
    E.setBasicDailyRainfallMinAmount(newAmt);
    return new ABCMResult("Set basic daily rainfall minimum amount to " + newAmt, true);
  }
  
  /**
   * Sets the maximum amount of daily rainfall as generated by the rainfall engine
   * @param newAmt new amount value
   * @return an ABCMResult indicating the success or failure of this action
   */
  public ABCMResult setBasicDailyRainfallMaxAmount(double newAmt){
    E.setBasicDailyRainfallMaxAmount(newAmt);
    return new ABCMResult("Set basic daily rainfall maximum amount to " + newAmt, true);
  }
  
  /* Rivers */
    
  /**
   * Adds a river to the Model's collection of rivers
   * @param rName name of new river
   * @return a response that can be written to an output window 
   */
  public ABCMResult addRiver(String rName){
    checkIntegrity();
    if(rivers.containsKey(rName)) return new ABCMResult("River '" + rName + "' already exists.", false);
    River R = new River(rName);
    rivers.put(rName, R);
    space.addRiver(rName, R);
    return new ABCMResult("Successfully added river '" + rName + "'", true);
  }
  
  /**
   * Sets the annual streamflow variable for a given river
   * to the specified amount
   * @param rName Name of the river whose Streamflow is to be set
   * @param newFlow New value for Streamflow for this river
   * @return a response that can be written to an output window
   */
  public ABCMResult setRiverAnnualStreamFlow(String rName, double newFlow){
    checkIntegrity();
    double n;
    
    // Get the river, if possible
    River R = getRiver(rName);
    if(R == null) return new ABCMResult("River '" + rName + "' does not exist.", false);
    
    // Calculate the new value for streamflow
    switch(codeModelOption("SET_STREAMFLOW_PROC")){
      case SET_STREAMFLOW_STOCHASTIC: // (1)
        double r = random();
        double p = paramVal("SETFLOWSTOCHASTICITY");
        n = newFlow * ((double)1 - (p * (double)2 * (r - (double).5)));
        break;
      case SET_STREAMFLOW_SCALED: // (2)
        n = newFlow * paramVal("SCALE_STREAMFLOW_SET_VALUE");
        break;
      case SET_STREAMFLOW_DEFAULT: // (0)
      default:
        n = newFlow;
    }      
    R.setAnnualStreamflow(n);
    return new ABCMResult("Set annual streamflow of " + R.getName() + " river to " + n, true);  
            
  }
  
  /**
   * Sets the current flow level for a given river to the specified amount
   * @param rName Name of the river whose flow value is to be set
   * @param newFlow Value to which the flow is to be set.
   * @return a response that can be written to an output window
   */
  public ABCMResult setRiverFlowLevel(String rName, double newFlow){
    checkIntegrity();
    River R = getRiver(rName);
    if(R == null) return new ABCMResult("River " + rName + " has not yet been created.", false);
    R.setFlowLevel(newFlow);
    return new ABCMResult("River " + rName + " current stream flow set to " + newFlow, true);
  }
  
  /**
   * Sets the chemical concentration found in a specified River. Returns an
   * ABCMResult object indicating the success or failure of the operation
   * @param riverName the name of the river.
   * @param chemical the chemical to be set
   * @param concentration the new concentration
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult setRiverChemicalConcentration(String riverName, String chemical, double concentration){
    River R = getRiver(riverName);
    if(R == null) return new ABCMResult("No river '" + riverName + "' exists.", false);
    Chemical C = new BasicChemical(chemical, concentration);
    R.getWaterSample().addChemical(C);
    return new ABCMResult("Added '" + chemical + "' to river '" + riverName + "'.", true);
  }

  /**
   * Creates a new HeadGate and adds it to the specified River at the specified
   * point. Will fail if the River or the Point do not exist, or if the
   * River already has a Headgate with the specified name. Note also that
   * the index parameter specifies in what ordinal position this HeadGate will fall
   * within the River among the other HeadGates on the same river.
   * @param riverName the name of the River to which this HeadGate is to be added
   * @param headgateLocus the Point at which the HeadGate is to be added
   * @param headgateName the name to be assigned to this HeadGate
   * @param headgateIndex the order in which this HeadGate is to be placed in the River
   * @return a String indicating the success or failure of the operation
   */
  public ABCMResult addHeadgateToRiver(String riverName, String headgateLocus, String headgateName, int headgateIndex){
    if(!rivers.containsKey(riverName)) return new ABCMResult("Cannot add headgate; river '" + riverName + "' does not exist.", false);
    if(!space.containsPoint(headgateLocus)) return new ABCMResult("Cannot add headgate at point '" + headgateLocus + "'; no point with that name exists.", false);
    Point p = space.getPoint(headgateLocus);
    River R = getRiver(riverName);
    if(R == null) return new ABCMResult("Could not add HeadGate to River; no river '" + riverName + "' exists.", false);
    HeadGate HG = new HeadGate(new Gate(), p, paramVal("DEFAULT_HEADGATE_EFFICIENCY"));
    return new ABCMResult(R.addHeadGate(HG, headgateIndex, headgateName), true);
  }

  /**
   * Sets the efficiency of the specified HeadGate on the specified river.
   * @param riverName the name of the River
   * @param headGate the name of the HeadGate
   * @param efficiency the new efficiency value
   * @return an ABCMResult indicating the success or failure of the action
   */
  public ABCMResult setHeadGateEfficiency(String riverName, String headGate, double efficiency){
    River R = getRiver(riverName);
    if(R == null) return new ABCMResult("Could not set efficiency; no river '" + riverName + "' exists.", false);
    HeadGate HG = R.getHeadGate(headGate);
    if(HG == null) return new ABCMResult("Could not set efficiency; no HeadGate '" + headGate + "' exists on river '" + riverName + "'.", false);
    HG.setEfficiency(efficiency);
    return new ABCMResult("Set efficiency of HeadGate '" + headGate + "' on river '" + riverName + "' to " + efficiency + ".", true);
  }
  
  /**
   * Sets ranges for a headgate's efficiency along the river
   * @param riverName name of the river along which the headgate is located
   * @param headGate name of the headgate
   * @param ranges String specifying the efficiencies and the ranges within which they fall. 
   * Format is: "upperBound1:efficiency1;upperBound2:efficiency2;"
   * @return an ABCMResult indicating the success or failure of the operation 
   */
  public ABCMResult setHGEfficiencyRanges(String riverName, String headGate, String ranges){
    River R = getRiver(riverName);
    if(R == null) return new ABCMResult("Could not set efficiency; no river '" + riverName + "' exists.", false);
    HeadGate HG = R.getHeadGate(headGate);
    if(HG == null) return new ABCMResult("Could not set efficiency; no HeadGate '" + headGate + "' exists on river '" + riverName + "'.", false);
    HG.setEfficiencyRanges(ranges);
    return new ABCMResult("Set efficiency of HeadGate '" + headGate + "' on river '" + riverName + "' to ranges: " + ranges, true);
  }
  
  /**
   * Removes a specified chemical from the specified River
   * @param riverName the name of the River
   * @param chemical the name of the Chemical
   * @return an ABCMResult indicating the success or failure of the action
   */
  public ABCMResult removeChemicalFromRiver(String riverName, String chemical){
    River R = getRiver(riverName);
    if(R == null)                            return new ABCMResult("No river '" + riverName + "' exists.", false);
    if(!R.removeChemicalFromRiver(chemical)) return new ABCMResult("Could not remove chemical '" + chemical + "', perhaps because it is not present", false);
    return new ABCMResult("Removed chemical '" + chemical + "' from river '" + riverName + "'", true);
  }

  
  /* Canal Systems */
  
  /**
   * Creates a canal system with the given name and
   * adds it to the collection of CanalSegment Systems in
   * this model. The name provided must not yet have
   * been used for a CanalSegment System in the model; if it
   * has the CanalSegment System is not added.
   * @param csName the name of the canal system to be created
   * @return a response that can be displayed to an output window
   */
  public ABCMResult createCanalSystem(String csName){
    checkIntegrity();
    if(canalSystems.containsKey(csName)) return new ABCMResult("Canal System named '" + csName + "' already exists.", false); // Name must be unique
    CanalSystem cs = new CanalSystem(csName);
    canalSystems.put(csName, cs);
    space.addCanalSystem(csName, cs);
    return new ABCMResult("Canal System " + csName + " created.", true);
  }
  
  /**
   * Adds a Canal Segment to a Canal System
   * @param csName Canal System to which to add this segment
   * @param Point1Name Name of the point that indicates what will be the head
   * coordinates of the new Canal Segment.
   * @param Point2Name Name of the point that indicates what will be the end
   * coordinates of the new canal Segment
   * @param newCanalName a name for the new Canal Segment
   * @return a response that can be displayed to an output window
   */
  public ABCMResult addCanalToCanalSystem(String csName, String Point1Name, String Point2Name, String newCanalName){
    checkIntegrity();
    // TODO in this method- check for errors and nonexistent elements
    CanalSystem cs = getCanalSystem(csName);    
    Point p1 = space.getPoint(Point1Name);
    Point p2 = space.getPoint(Point2Name);
    cs.addCanal(new CanalSegment(p1, p2, newCanalName));
    return new ABCMResult("Added canal " + newCanalName + " to canal system " + csName, true);
  }
  
  /**
   * Adds a CanalSegment to a CanalSystem beginning from the specified
   * HeadGate and extending to the Specified Point
   * @param riverName river on which HeadGate is found
   * @param csName CanalSystem name
   * @param headgateName HeadGate name
   * @param Point2Name name of the reference point to which the Segment will extend
   * @param newCanalName name to be assigned to the new CanalSegment
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult startCanalAtHeadGate(String riverName, String csName, String headgateName, String Point2Name, String newCanalName){
    checkIntegrity();
    CanalSystem cs = getCanalSystem(csName);
    River R = getRiver(riverName);
    if(R == null) return new ABCMResult("Could not start canal; no river '" + riverName + "' exists", false);
    HeadGate HG = R.getHeadGate(headgateName);
    if(HG == null) return new ABCMResult("Could not start canal; no headgate '" + headgateName + "' exists", false);
    Point p2 = space.getPoint(Point2Name);
    if(p2 == null) return new ABCMResult("Could not start canal; no destination point '" + Point2Name + "' exists", false);
    cs.addCanal(new CanalSegment(HG, p2, newCanalName));
    return new ABCMResult("Successfully added new canal at the specified headgate.", true);
  }  
  
  /**
   * Extends a given Canal Segment by appending another Canal Segment onto it.
   * @param csName Name of the Canal System that includes the Canal Segment 
   * to be extended
   * @param CanalName Name of the Canal Segment to be extended
   * @param P2Name Name of the Point object that has the coordinates to which the
   * Canal Segment is to be extended
   * @param newCanalName a name for the new Canal Segment
   * @return a response that can be displayed to an output window
   */
  public ABCMResult extendCanal(String csName, String CanalName, 
      String P2Name, String newCanalName){
    checkIntegrity();
    Point p2 = space.getPoint(P2Name);
    CanalSystem cs = getCanalSystem(csName);
    if(cs==null) return new ABCMResult("Canal system named '" + csName + "' not found.", false);
    CanalSegment c = cs.getCanalSegment(CanalName);
    if(c==null) return new ABCMResult("Canal named '" + CanalName + "' not found.", false);
    else{
      c.extendCanal(p2, cs, newCanalName);
      return new ABCMResult("Canal " + CanalName + " extended to create " + newCanalName, true);
    }
  }
  
  /**
   * Extends the last CanalSegment added. Allows you to avoid
   * specifying the name of the last segment added.
   * @param csName Name of the Canal System to be extended
   * @param P2Name Name of the point whose coordinates are to be
   * used to specify the end coordinates of the new Canal Segment
   * @param newCanalName a name for the new Canal Segment
   * @return a response that can be displayed to an output window
   */
  public ABCMResult extendSelectedCanalSegment(String csName, String P2Name, String newCanalName){
    checkIntegrity();
    CanalSystem cs = getCanalSystem(csName);
    if(cs == null) return new ABCMResult("No canal system with the name '" + csName + "' exists.", false); 
    Point p2 = space.getPoint(P2Name);
    String retVal = (cs.extendSelectedCanalSegment(p2, newCanalName) ? "Successfully extended the last canal." : "Could not extend canal segment.");
    return new ABCMResult(retVal, true);
  }
  
  /**
   * Sets the profile of the selected Canal Segment created in the specified
   * Canal System
   * @param csName name of the Canal System
   * @param topWidth new top width for the Canal Segment
   * @param bottomWidth new bottom width for the Canal Segment
   * @param depth new dept for the Canal Segment
   * @return a String indicating success of the operation
   */
  public ABCMResult setSelectedCanalSegmentProfile(String csName, double topWidth, double bottomWidth, double depth){
    checkIntegrity();
    CanalSystem cs = getCanalSystem(csName);
    if(cs == null) return new ABCMResult("No canal system with the name '" + csName + "' exists.", false);
    cs.setSelectedCanalSegmentProfile(topWidth, bottomWidth, depth);
    return new ABCMResult("Set selected canal profile.", true);
  }
  
  /**
   * Sets the profile of the a Canal Segment in the specified
   * Canal System by indicating the 'rank' of the segment. Rank corresponds
   * to whether the Canal Segment is a trunk, secondary canal, etc.
   * @param csName name of the Canal System
   * @param segName name of the Canal Segment
   * @param rank rank to which the Canal Segment should be set
   * @return a String indicating the success of the operation
   */
  public ABCMResult setCanalSegmentRank(String csName, String segName, int rank){
    return new ABCMResult("Cannot re-set rank for canal system " + csName + " segment " + segName + "; action temporarily disabled", false);
//    checkIntegrity();
//    double topWidth;
//    double bottomWidth;
//    double depth;
//    String output = "";
//    switch(rank){
//      case TRUNK:
//        topWidth     = paramVal("TRUNK_TOP_WIDTH");
//        bottomWidth  = paramVal("TRUNK_BOTTOM_WIDTH");
//        depth        = paramVal("TRUNK_DEPTH");
//        output = "TRUNK";
//        break;
//      case SECONDARY:
//        topWidth     = paramVal("SECONDARY_TOP_WIDTH");
//        bottomWidth  = paramVal("SECONDARY_BOTTOM_WIDTH");
//        depth        = paramVal("SECONDARY_DEPTH");
//        output = "SECONDARY";
//        break;
//      case TERTIARY:
//        topWidth     = paramVal("TERTIARY_TOP_WIDTH");
//        bottomWidth  = paramVal("TERTIARY_BOTTOM_WIDTH");
//        depth        = paramVal("TERTIARY_DEPTH");
//        output = "TERTIARY";
//        break;
//      default:
//        return new ABCMResult("Could not set Canal Segment; invalid rank specified.", false);
//    }
//    CanalSystem cs = (CanalSystem)canalSystems.get(csName);
//    CanalSegment seg = cs.getCanalSegment(segName);
//    if(!seg.expandCanal(topWidth, bottomWidth, depth)) seg.reduceCanal(topWidth, bottomWidth, depth) ;
//    return new ABCMResult("Set segment '" + segName + "' of canal system '" + csName + "' to " + output, true);
  }

  /**
   * Sets the size of a canal segment. Assumes a trapezoidal cross section
   * @param csName CanalSystem name
   * @param segName CanalSegment name
   * @param topWidth top width
   * @param bottomWidth bottom width
   * @param channelHeight channel (construction) height
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentSize(String csName, String segName, double topWidth, double bottomWidth, double channelHeight){
    CanalSystem cs = (CanalSystem)canalSystems.get(csName);
    if(cs == null) return new ABCMResult("No canal system '" + csName + "' found.", false);
    CanalSegment seg = cs.getCanalSegment(segName);
    if(cs == null) return new ABCMResult("No canal segment '" + csName + "' found.", false);
//    if(!seg.expandCanal(topWidth, bottomWidth, channelHeight)) seg.reduceCanal(topWidth, bottomWidth, channelHeight) ;
    return new ABCMResult("THIS COMMAND IS NOT YET IMPLEMENTED", false);
//    return new ABCMResult("Set segment '" + segName + "' of canal system '" + csName + "' to new size: " + topWidth + ", " + bottomWidth + ", " + channelHeight, true);
  }
  
  /**
   * Sets the type for a specified canal segmnt
   * @param csName name of the canal system
   * @param segName name of the specific canal segment
   * @param type new value for canal segment type
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentType(String csName, String segName, String type){
    CanalSystem cs = (CanalSystem)canalSystems.get(csName);
    if(cs == null) return new ABCMResult("No canal system '" + csName + "' found.", false);
    CanalSegment seg = cs.getCanalSegment(segName);
    if(cs == null) return new ABCMResult("No canal segment '" + csName + "' found.", false);
    seg.setCanalType(type);
    return new ABCMResult("Set canal segment " + segName + " in canal system " + csName + " to type " + type, true);
  }
  
  /**
   * Sets the cross sectional profile (upstream and downstream) of a specified CanalSegment
   * @param csName name of the CanalSystem 
   * @param segName name of the CanalSegment
   * @param spec new cross-sectional specification, as a String
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentProfile(String csName, String segName, String spec){
    CanalSystem cs = (CanalSystem)canalSystems.get(csName);
    if(cs == null) return new ABCMResult("No canal system '" + csName + "' found.", false);
    CanalSegment seg = cs.getCanalSegment(segName);
    if(seg == null) return new ABCMResult("No canal segment '" + segName + "' found.", false);
    if(seg.setProfile(spec)) return new ABCMResult("Set segment '" + segName + "' of canal system '" + csName + "' to new profile.", true);
    return new ABCMResult("Could not set segment profile.", false);
  }
  
  /**
   * Sets the roughness value for a canal segment
   * @param csName name of the CanalSystem 
   * @param segName name of the CanalSegment
   * @param n new roughness value
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentRoughness(String csName, String segName, double n){
    CanalSystem cs = (CanalSystem)canalSystems.get(csName);
    if(cs == null) return new ABCMResult("No canal system '" + csName + "' found.", false);
    CanalSegment seg = cs.getCanalSegment(segName);
    if(seg == null) return new ABCMResult("No canal segment '" + csName + "' found.", false);
    seg.setRoughness(n);
    return new ABCMResult("Set segment '" + segName + "' of canal system '" + csName + "' roughness to: " + n, true);
  }
  
  /**
   * Causes a canal segment to be abandoned 
   * @param csName name of the canal system
   * @param segName name of the specific canal segment
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult abandonCanalSegment(String csName, String segName){
    CanalSystem cs = (CanalSystem)canalSystems.get(csName);
    if(cs == null) return new ABCMResult("No canal system '" + csName + "' found.", false);
    CanalSegment seg = cs.getCanalSegment(segName);
    if(seg == null) return new ABCMResult("No canal segment '" + segName + "' found.", false);
    seg.setIsAbandoned(true);
    return new ABCMResult("Set segment '" + segName + "' of canal system '" + csName + "' to abandoned ", true);
  }
  
  /**
   * Sets the upstream discharge value for a particular CanalSegment in a particular
   * CanalSystem. See notes under the CanalSegment object for the cascading impacts of setting stage
   * and discharge values at the ends of linked Canal Segments.
   * @param canalSystemName the Canal System to which the specified Canal Segment will belong
   * @param canalSegmentID the name of the Canal Segment for which the discharge is to be set
   * @param newDischarge the new discharge value
   * @return a String indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentUpstreamDischarge(String canalSystemName, String canalSegmentID, double newDischarge){
    if(!canalSystems.containsKey(canalSystemName)) return new ABCMResult("Could not set value; canal system '" + canalSystemName + "' not found.", false);
    if((((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)) ==  null) return new ABCMResult(
       "Could not set value; canal system '" + canalSystemName + "' does not contain a segment with ID = '" + 
       canalSegmentID + "'.", false);
    (((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)).setUpstreamDischarge(newDischarge);
    return new ABCMResult("Set Canal System '" + canalSystemName + "', segment '" + canalSegmentID + "' upstream discharge to: " + newDischarge, true);
  }
  
  /**
   * Sets the upnstream stage value for a particular CanalSegment in a particular
   * CanalSystem. See notes under the CanalSegment object for the cascading impacts of setting stage
   * and discharge values at the ends of linked Canal Segments.
   * @param canalSystemName the Canal System to which the specified Canal Segment will belong
   * @param canalSegmentID the name of the Canal Segment for which the discharge is to be set
   * @param newStage the new stage value
   * @return a String indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentUpstreamStage(String canalSystemName, String canalSegmentID, double newStage){
    if(!canalSystems.containsKey(canalSystemName)) return new ABCMResult("Could not set value; canal system '" + canalSystemName + "' not found.", false);
    if((((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)) ==  null) return new ABCMResult(
       "Could not set value; canal system '" + canalSystemName + "' does not contain a segment with ID = '" + 
       canalSegmentID + "'.", false);
    (((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)).setUpstreamStage(newStage);
    return new ABCMResult("Set Canal System '" + canalSystemName + "', segment '" + canalSegmentID + "' upstream stage to: " + newStage, true);
  }
  
  /**
   * Sets the downstream discharge value for a particular CanalSegment in a particular
   * CanalSystem. See notes under the CanalSegment object for the cascading impacts of setting stage
   * and discharge values at the ends of linked Canal Segments.
   * @param canalSystemName the Canal System to which the specified Canal Segment will belong
   * @param canalSegmentID the name of the Canal Segment for which the discharge is to be set
   * @param newDischarge the new discharge value
   * @return a String indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentDownstreamDischarge(String canalSystemName, String canalSegmentID, double newDischarge){
    if(!canalSystems.containsKey(canalSystemName)) return new ABCMResult("Could not set value; canal system '" + canalSystemName + "' not found.", false);
    if((((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)) ==  null) return new ABCMResult(
       "Could not set value; canal system '" + canalSystemName + "' does not contain a segment with ID = '" + 
       canalSegmentID + "'.", false);
    (((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)).setDownstreamDischarge(newDischarge);
    return new ABCMResult("Set Canal System '" + canalSystemName + "', segment '" + canalSegmentID + "' downstream discharge to: " + newDischarge, true);
  }

  /**
   * Sets the downstream stage value for a particular CanalSegment in a particular
   * CanalSystem. See notes under the CanalSegment object for the cascading impacts of setting stage
   * and discharge values at the ends of linked Canal Segments.
   * @param canalSystemName the Canal System to which the specified Canal Segment will belong
   * @param canalSegmentID the name of the Canal Segment for which the discharge is to be set
   * @param newStage the new stage value
   * @return a String indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentDownstreamStage(String canalSystemName, String canalSegmentID, double newStage){
    if(!canalSystems.containsKey(canalSystemName)) return new ABCMResult("Could not set value; canal system '" + canalSystemName + "' not found.", false);
    if((((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)) ==  null) return new ABCMResult(
       "Could not set value; canal system '" + canalSystemName + "' does not contain a segment with ID = '" + 
       canalSegmentID + "'.", false);
    (((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)).setUpstreamStage(newStage);
    return new ABCMResult("Set Canal System '" + canalSystemName + "', segment '" + canalSegmentID + "' downstream stage to: " + newStage, true);
  }
  
  /**
   * Sets the upstream stage of a specified CanalSegment in a specified
   * Canal System to a new value. Sets the discharge to the value calculated from
   * the specified depth using Manning's equation. See notes under the CanalSegment object for the cascading impacts of setting stage
   * and discharge values at the ends of linked Canal Segments.
   * @param canalSystemName the Canal System to which the specified Canal Segment will belong
   * @param canalSegmentID the name of the Canal Segment for which the discharge is to be set
   * @param newDepth the new depth value (note: depth of water in the cross section, not stage)
   * @return a String indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentUpstreamQperY(String canalSystemName, String canalSegmentID, double newDepth){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return new ABCMResult("Could not set value; canal system '" + canalSystemName + "' not found.", false);
    CanalSegment C = CS.getCanalSegment(canalSegmentID);
    if(C == null) return new ABCMResult("Could not set value; canal system '" + canalSystemName + "' does not contain a segment with ID = '" + canalSegmentID + "'.", false);
    C.setUpstreamDepthPlusManningDischarge(newDepth);
    return new ABCMResult("Set Canal System '" + canalSystemName + "', segment '" + canalSegmentID + "' upstream stage to: " + newDepth, true);
  }
  
  /**
   * Sets the downstream stage of a specified CanalSegment in a specified
   * Canal System to a new value. Sets the discharge to the value calculated from
   * the specified stage using Manning's equation. See notes under the CanalSegment object for the cascading impacts of setting stage
   * and discharge values at the ends of linked Canal Segments.
   * @param canalSystemName the Canal System to which the specified Canal Segment will belong
   * @param canalSegmentID the name of the Canal Segment for which the discharge is to be set
   * @param newStage the new stage value
   * @return a String indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentDownstreamQperY(String canalSystemName, String canalSegmentID, double newStage){
    if(!canalSystems.containsKey(canalSystemName)) return new ABCMResult("Could not set value; canal system '" + canalSystemName + "' not found.", false);
    if((((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)) ==  null) return new ABCMResult(
       "Could not set value; canal system '" + canalSystemName + "' does not contain a segment with ID = '" + 
       canalSegmentID + "'.", false);
    (((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)).setDownstreamDepthPlusManningDischarge(newStage);
    return new ABCMResult("Set Canal System '" + canalSystemName + "', segment '" + canalSegmentID + "' downstream stage to: " + newStage, true);
  }
 
  /**
   * Sets the upstream stage and flow values for a specified canal segment. 
   * @param canalSystemName name of the CanalSystem
   * @param canalSegmentID name of the CanalSegment
   * @param newStage new z value
   * @param newDischarge new Q value
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentUpstreamZandQ(String canalSystemName, String canalSegmentID, double newStage, double newDischarge){
    if(!canalSystems.containsKey(canalSystemName)) return new ABCMResult("Could not set value; canal system '" + canalSystemName + "' not found.", false);
    if((((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)) ==  null) return new ABCMResult(
       "Could not set value; canal system '" + canalSystemName + "' does not contain a segment with ID = '" + 
       canalSegmentID + "'.", false);
    (((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)).setUpstreamStage(newStage);
    (((CanalSystem)canalSystems.get(canalSystemName)).getCanalSegment(canalSegmentID)).setUpstreamDischarge(newDischarge);
    return new ABCMResult("Set Canal System '" + canalSystemName + "', segment '" + canalSegmentID + "' upstream stage to: " + newStage +
        " and discharge to " + newDischarge, true);
  }
  
  /**
   * Sets the downstream stage and flow values for a specified canal segment. 
   * @param canalSystemName name of the CanalSystem
   * @param canalSegmentID name of the CanalSegment
   * @param newStage new z value
   * @param newDischarge new Q value
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentDownstreamZandQ(String canalSystemName, String canalSegmentID, double newStage, double newDischarge){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null)  return new ABCMResult("Could not set value; canal system '" + canalSystemName + "' not found.", false);
    CanalSegment C = CS.getCanalSegment(canalSegmentID);
    if(C == null) return new ABCMResult("Could not set value; canal system '" + canalSystemName + 
                         "' does not contain a segment with ID = '" + canalSegmentID + "'.", false);
    C.setDownstreamStage(newStage);
    C.setDownstreamDischarge(newDischarge); 
    return new ABCMResult("Set Canal System '"    + canalSystemName + "', " +
           "segment "              + canalSegmentReturnID(C, canalSegmentID) + " " + 
           "downstream stage to: " + newStage + " " +
           "and discharge to "     + newDischarge, true);
  }
  
  /**
   * Drops a canal segment from a canal system
   * @param canalSystemName the name of the canal system
   * @param canalSegmentID the name of the canal segment
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult dropCanalSegment(String canalSystemName, String canalSegmentID){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null)  return new ABCMResult("Could not drop canal segment; canal system '" + canalSystemName + "' not found.", false);
    CanalSegment C = CS.getCanalSegment(canalSegmentID);
    if(C == null) return new ABCMResult("Could not drop canal segment; canal system '" + canalSystemName + 
                         "' does not contain a segment with ID = '" + canalSegmentID + "'.", false);
    CS.removeCanalSegment(C);
    return new ABCMResult("Dropped canal segment " + canalSegmentID + " from canal system " + canalSystemName, true);
  }
  
  /**
   * Closes a headgate
   * @param riverName the name of the river on which the headgate is located
   * @param headGate the name of the headgate
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult closeHeadGate(String riverName, String headGate){
    River R = getRiver(riverName);
    if(R == null) return new ABCMResult("Could not set efficiency; no river '" + riverName + "' exists.", false);
    HeadGate HG = R.getHeadGate(headGate);
    if(HG == null) return new ABCMResult("Could not set efficiency; no HeadGate '" + headGate + "' exists on river '" + riverName + "'.", false);
    HG.close();
    return new ABCMResult("Closed HeadGate '" + headGate + "' on river '" + riverName + ".", true);
  }
  
  /**
   * Opens a headgate
   * @param riverName the name of the river on which the headgate is located
   * @param headGate the name of the headgate
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult openHeadGate(String riverName, String headGate){
    River R = getRiver(riverName);
    if(R == null) return new ABCMResult("Could not set efficiency; no river '" + riverName + "' exists.", false);
    HeadGate HG = R.getHeadGate(headGate);
    if(HG == null) return new ABCMResult("Could not set efficiency; no HeadGate '" + headGate + "' exists on river '" + riverName + "'.", false);
    HG.open();
    return new ABCMResult("Opened HeadGate '" + headGate + "' on river '" + riverName + ".", true);
  }

  
  /**
   * Closes the inflow gate on a canal segment
   * @param canalSystemName name of the CanalSystem in which this segment is located
   * @param canalSegmentName name of the CanalSegment
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult closeInflowGate(String canalSystemName, String canalSegmentName){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return new ABCMResult("Could not close gate; no canal system named '" + canalSystemName + "' exists.", false);
    CanalSegment C = CS.getCanalSegment(canalSegmentName);
    if(C == null) return new ABCMResult("Could not close gate; no canal segment named '" + canalSegmentName + "' exists in canal system " + CS.getCanalSystemName() + ".", false);
    if(C.closeInflowGate()) return new ABCMResult("Closed inflow gate for canal segment '" + canalSegmentName + "' in canal system " + CS.getCanalSystemName(), true);
    return new ABCMResult("Could not close outflow gates for canal segment '" + canalSegmentName + "' in canal system " + CS.getCanalSystemName(), false);
  }
    
  /**
   * Opens the inflow gate on a canal segment
   * @param canalSystemName name of the CanalSystem in which this segment is located
   * @param canalSegmentName name of the CanalSegment
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult openInflowGate(String canalSystemName, String canalSegmentName){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return new ABCMResult("Could not open gate; no canal system named '" + canalSystemName + "' exists.", false);
    CanalSegment C = CS.getCanalSegment(canalSegmentName);
    if(C == null) return new ABCMResult("Could not open gate; no canal segment named '" + canalSegmentName + "' exists in canal system " + CS.getCanalSystemName() + ".", false);
    
    if(C.openInflowGate()) return new ABCMResult("Opened inflow gate for canal segment '" + canalSegmentName + "' in canal system " + CS.getCanalSystemName(), true);
    
    return new ABCMResult("Could not open inflow gates for canal segment '" + canalSegmentName + "' in canal system " + CS.getCanalSystemName(), false);
  }
  
  /**
   * Closes the outflow gate on a canal segment
   * @param canalSystemName name of the CanalSystem in which this segment is located
   * @param canalSegmentName name of the CanalSegment
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult closeOutflowGate(String canalSystemName, String canalSegmentName){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return new ABCMResult("Could not close gate; no canal system named '" + canalSystemName + "' exists.", false);
    CanalSegment C = CS.getCanalSegment(canalSegmentName);
    if(C == null) return new ABCMResult("Could not close gate; no canal segment named '" + canalSegmentName + "' exists in canal system " + CS.getCanalSystemName() + ".", false);
    if(C.closeEndGate()) return new ABCMResult("Closed inflow gates for canal segment '" + canalSegmentName + "' in canal system " + CS.getCanalSystemName(), true);
    return new ABCMResult("This feature is not yet implemented", false);
//    return new ABCMResult("Could not close outflow gates for canal segment '" + canalSegmentName + "' in canal system " + CS.getCanalSystemName(), false);
  }
  
  /**
   * Opens the outflow gate on a canal segment
   * @param canalSystemName name of the CanalSystem in which this segment is located
   * @param canalSegmentName name of the CanalSegment
   * @return an ABCMResult object indicating the success or failure of the operation
*/
  public ABCMResult openOutflowGate(String canalSystemName, String canalSegmentName){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return new ABCMResult("Could not open gate; no canal system named '" + canalSystemName + "' exists.", false);
    CanalSegment C = CS.getCanalSegment(canalSegmentName);
    if(C == null) return new ABCMResult("Could not open gate; no canal segment named '" + canalSegmentName + "' exists in canal system " + CS.getCanalSystemName() + ".", false);
    if(!C.openEndGate()) return new ABCMResult("Could not open gate; canal segment '" + canalSegmentName + " is not properly configured (may be a junction)", false);    
    return new ABCMResult("Outflow gate of canal segment '" + canalSegmentName + " is open.", false);
  }

  /**
   * Returns a String value containing a report of the CanalSystem
   * details and status
   * @param canalSystemName the name of the canal system
   * @return the requested report
   */
  public String viewCanalSystemDetails(String canalSystemName){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return "Canal System '" + canalSystemName + "' does not exist.";
    return CS.report();
  }

  /* Plants and Fields */

  /**
   * Creates a field. Note that while the arguments are given
   * as 'NW Point', 'SE Point', etc., the actual orientation
   * is irrelevant, provided the points are listed in clockwise
   * order. The points are given by name and must already
   * exist in the model's space. The field is not associated
   * with any canal system.
   * @param NWPoint name of the first corner point
   * @param NEPoint name of the second corner point
   * @param SEPoint name of the third corner point
   * @param SWPoint name of the fourth corner point
   * @param FieldName name to be assigned to this field
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult createField(String NWPoint, String NEPoint, 
      String SEPoint, String SWPoint, String FieldName){
    checkIntegrity();
    // Confirm that all the specified points are in 
    // the points list
    Point pNW;
    Point pNE;
    Point pSE;
    Point pSW;
    if(space.containsPoint(NWPoint)) pNW = space.getPoint(NWPoint);
    else return new ABCMResult("Could not add field; model does not contain point '" + NWPoint + "'", false);
    if(space.containsPoint(NEPoint)) pNE = space.getPoint(NEPoint);
    else return new ABCMResult("Could not add field; model does not contain point '" + NEPoint + "'", false);
    if(space.containsPoint(SEPoint)) pSE = space.getPoint(SEPoint);
    else return new ABCMResult("Could not add field; model does not contain point '" + SEPoint + "'", false);
    if(space.containsPoint(SWPoint)) pSW = space.getPoint(SWPoint);
    else return new ABCMResult("Could not add field; model does not contain point '" + SWPoint + "'", false);
    if(fields.containsKey(FieldName)) return new ABCMResult("Could not add new field '" + 
        FieldName + "'. Field with this identifier already exists.", false);
    Field f = new Field(FieldName, pNW, pNE, pSE, pSW, E);
    fields.put(FieldName, f);
    space.addField(FieldName, f);
    return new ABCMResult("Successfully created field " + FieldName, true);
  }
  
  /**
   * Drops a field from the model's collection of fields.
   * Also removes it from the space's collection.
   * @param fieldName Name of the field to be removed
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult dropField(String fieldName){
    checkIntegrity();
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not drop field; no field '" + fieldName + "' exists.", false);
    Field F = (Field)fields.get(fieldName);
    space.dropField(fieldName);
    //Must remove from any canal segments; done by dropping from all canal systems
    Enumeration E = this.canalSystems.elements();
    while(E.hasMoreElements())  ((CanalSystem)E.nextElement()).dropField(fieldName, F);    
    fields.remove(fieldName);
    return new ABCMResult("Dropped field '" + fieldName + "'", true);
  }
  
  /**
   * Links a canal segment to a field, so that water flowing out of that segment
   * is deposited on the field
   * @param canalSystemName name of the canal system in which the segment is located
   * @param canalSegName name of the segment
   * @param fieldName name of the field
   * @return an ABCMResult indicating the success or failure of the operation
   
   */
  public ABCMResult linkCanalSegmentToField(String canalSystemName, String canalSegName, String fieldName){
    checkIntegrity();
    CanalSystem cs = this.getCanalSystem(canalSystemName);
    if(cs == null) return new ABCMResult("Could not link canal segment to field; no canal system '" + canalSystemName + "' exists", false);
    CanalSegment canalSeg = cs.getCanalSegment(canalSegName);
    if(canalSeg == null) return new ABCMResult("Could not link canal segment to field; no canal segment '" + canalSegName + "' found in canal system '" + canalSystemName + "'", false);
    Field f = this.getField(fieldName);
    if(f == null) return new ABCMResult("Could not link canal segment to field; no field '" + fieldName + "' exists.", false);
    if(!cs.linkCanalSegmentToField(canalSegName, f)) return new ABCMResult("Could not link canal segment to field (unknown error)", false);
    return new ABCMResult("Successfully linked canal segment '" + canalSegName + "' in canal system '" + canalSystemName + "' to field '" + fieldName + "'", true);
  }
  
  /**
   * Disconnects a canal segment from a field
   * @param canalSystemName name of the canal system in which the segment is located
   * @param canalSegName name of the segment
   * @param fieldName name of the field
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult unlinkCanalSegmentFromField(String canalSystemName, String canalSegName, String fieldName){
    checkIntegrity();
    CanalSystem cs = this.getCanalSystem(canalSystemName);
    if(cs == null) return new ABCMResult("Could not unlink canal segment from field; no canal system '" + canalSystemName + "' exists", false);
    CanalSegment canalSeg = cs.getCanalSegment(canalSegName);
    if(canalSeg == null) return new ABCMResult("Could not unlink canal segment from field; no canal segment '" + canalSegName + "' found in canal system '" + canalSystemName + "'", false);
    Field f = this.getField(fieldName);
    if(f == null) return new ABCMResult("Could not unlink canal segment from field; no field '" + fieldName + "' exists.", false);
    if(!cs.unlinkCanalSegmentFromField(canalSegName, f)) return new ABCMResult("Could not link canal segment to field (unknown error)", false);
    return new ABCMResult("Successfully linked canal segment '" + canalSegName + "' in canal system '" + canalSystemName + "' to field '" + fieldName + "'", true);
  }

  /**
   * Sets the effective area of a field; this is the area that is considered to be arable and
   * used, so that when water is deposited on the whole field it is effectively moved to
   * cover only the fraction that is considered effective
   * @param fieldName name of the field
   * @param newEffectiveArea new value for effecctive area
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult setFieldEffectiveArea(String fieldName, double newEffectiveArea){
    Field F = getField(fieldName);
    if(F == null) return new ABCMResult("Could not set field effective area; no field named '" + fieldName + "' exists", false);
    if(F.setEffectiveArea(newEffectiveArea)) return new ABCMResult("Set field '" + fieldName + "' effective area to " + newEffectiveArea, true);
    else return new ABCMResult("Could not set field '" + fieldName + "' effective area to " + newEffectiveArea, true);
  }
  
  
  
  /* Water Flow Magic- Canal Segments */
  /**
   * Sets the maximum outflow permitted from a canal segment.
   * @param canalSystemName name of the canal system in which the segment is located
   * @param canalSegName name of the canal segment
   * @param maxQ new maximum outflow value
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult setCanalSegmentMaximumOutflow(String canalSystemName, String canalSegName, double maxQ){
    checkIntegrity();
    CanalSystem cs = this.getCanalSystem(canalSystemName);
    if(cs == null) return new ABCMResult("Could not set max segment outflow; no canal system '" + canalSystemName + "' exists", false);
    CanalSegment canalSeg = cs.getCanalSegment(canalSegName);
    if(canalSeg == null) return new ABCMResult("Could not set max segment outflow; no canal segment '" + canalSegName + "' found in canal system '" + canalSystemName + "'", false);
    canalSeg.setMaximumOutflow(maxQ);
    return new ABCMResult("Successfully set max outflow for canal segment '" + canalSegName + "' in canal system '" + canalSystemName + "' to "+ (maxQ >=0 ? maxQ + "" : "CLEARED"), true);
  }
  
  /**
   * Gets an instance of a plant of the specified name. Plant is
   * drawn from the current configuration
   * @param plantName the requested plant name
   * @return the requested Plant
   */
  public Plant getPlant(String plantName){
    return (configuration != null ? ((WMConfiguration)configuration).getPlant(plantName) : Plant.getDefaultPlant(plantName));
  }
  
  /**
   * Gets the names of all canal segments flowing into a field
   * @param fieldName name of the field
   * @return a Vector of Strings listing the segments' names
   */
  public Vector<String> getNamesOfInflowCanalSegments(String fieldName){
    Vector<String> ret = new Vector<String>();
    Iterator I = this.canalSystems.values().iterator();
    while(I.hasNext()){
      ret.addAll(((CanalSystem)I.next()).getNamesOfInflowCanalSegments(fieldName));
    }
    return ret;
  }
  
  /**
   * Gets the names of all the canal systems in this model that contain a specific field
   * @param fieldName the name of the field
   * @return a Vector of Strings of canal systems that feed into this field
   */
  public Vector<String> getNamesOfCanalSystems(String fieldName){
    Vector<String> ret = new Vector<String>();
    Iterator I = this.canalSystems.keySet().iterator();
    while(I.hasNext()){
      String key = (String)I.next();
      if(((CanalSystem)this.canalSystems.get(key)).containsField(fieldName)) ret.add(key);
    }
    return ret;
  }
  
  /**
   * Gets an array listing the plants in the current configuration
   * @return a String array of available plant names
   */
  public String[] getListOfAvailablePlants(){
    return (configuration != null ? ((WMConfiguration)configuration).getListOfAvailablePlants() : new String[]{});
  }
  
  
  
  
  /**
   * Sows a plant in the specified field
   * @param plantName The name of the plant to be sown
   * @param fieldName The name of the field in which this plant should be sown
   * @param Density   The density at which this plant will be sown
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult sowPlantInField(String plantName, String fieldName, double Density){
    checkIntegrity();
    Plant p = getPlant(plantName);
    if(fields.containsKey(fieldName)){
      Field f = (Field)fields.get(fieldName);
      f.sow(p, Density, modelTime);
      return new ABCMResult("Successfully sowed " + plantName + " at density " + Density + " in " + fieldName, true);
    }
    else{
      return new ABCMResult("No field '" + fieldName + "' exists", false);
    }    
  }
  
  /**
   * Sows a plant in all selected fields
   * @param plantName the name of the plant to be sown
   * @param Density the density at which this plant will be sown
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult sowPlantInSelectedFields(String plantName, double Density){
    checkIntegrity();
    int counter = 0;
    Plant p = getPlant(plantName);
    HashSet H = getAllFields();
    Iterator I = H.iterator();
    while(I.hasNext()){
      Field F = (Field)I.next();
      if(F.isSelected()) {
        F.sow(p, Density, modelTime);
        counter++;
      }
    }
    return new ABCMResult("Sowed plant '" + plantName + "' (" + p.getName() + ") in " + counter + " fields.", true);
  }
  
  /**
   * Causes water (using the established rainsource chemicals) to be added to
   * a particular field.
   * @param fieldName Name of the field to which water is to be added
   * @param MM millimeters of rain
   * @return an ABCMResult object indicating the success or failure of the operation
   */
  public ABCMResult rainMMOnField(String fieldName, double MM){
    checkIntegrity();
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not find field '" + fieldName + "'.", false);
    Field f = (Field)(fields.get(fieldName));
    f.rain(MM, rainSource);
    return new ABCMResult("Successfully added rain (" + MM + ") to field.", true);
  }
  
  /**
   * Rains on all fields falling between the two specified elevations (inclusive)
   * @param MM millimeters of rain
   * @param lowestElevation lowest elevation
   * @param highestElevation highest elevation
   * @return a result indicating the number of fields receiving rain
   */
  public ABCMResult rainMMOnFieldsBetweenElevations(double MM, double lowestElevation, double highestElevation){
    checkIntegrity();
    Enumeration fieldEnum = fields.elements();
    int countOfFieldsReceivingRain = 0;
    while(fieldEnum.hasMoreElements()){
      Field f = (Field)fieldEnum.nextElement();
      double fieldZ = f.getMinZ();
      if(fieldZ >= lowestElevation && fieldZ <=highestElevation) {
        f.rain(MM, rainSource);
        countOfFieldsReceivingRain++;
      }
    }
    return new ABCMResult("Rain fell on " + countOfFieldsReceivingRain + " between " + lowestElevation + " and " + highestElevation, countOfFieldsReceivingRain, true); 
  }
  
  /**
   * Causes rain to occur on the selected fields. Rainwater is drawn from the rainsource object.
   * @param MM amount of rain to deposit, in millimeters
   * @return an ABCMResult indicating the success or failure of the operation
   */
  public ABCMResult rainOnSelectedFields(double MM){
    Iterator I = fields.values().iterator();
    int fieldCount = 0;
    while(I.hasNext()){
      Field F = (Field)I.next();
      if(F.isSelected()){
        F.rain(MM, rainSource);
        fieldCount++;
      }
    }
    return new ABCMResult("Rained on " + fieldCount + " fields", true);
  }
  
  /**
   * Sets the Evapotranspiration rate of a given Field. Will fail if
   * the Field specified does not exist.
   * @param fieldName the field to be set
   * @param newETo the new evapotranspiration rate
   * @return a String indicating the success or failure of the operation.
   */
  public ABCMResult setEToForField(String fieldName, double newETo){
    checkIntegrity();
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not find field '" + fieldName + "'.", false);
    Field f = (Field)(fields.get(fieldName));
    f.setETo_Set(newETo);
    return new ABCMResult("Set ETo for field '" + fieldName + "' to " + newETo, true);
  }
   
  /**
   * Rain a specified amount of water, in mm, on all
   * fields in the model.
   * @param MM number of millimeters of rainfall
   * @return a String indicating success or failure of the opertion.
   */
  public ABCMResult rainMMOnAllFields(double MM){
    checkIntegrity();
    Enumeration e = fields.keys();
    while(e.hasMoreElements()){
      rainMMOnField((String)e.nextElement(), MM);
    }
    e = canalSystems.elements();
    while(e.hasMoreElements()){
      ((CanalSystem)e.nextElement()).rain(MM, rainSource);
    }
    return new ABCMResult("Rained " + MM + "mm on all existing fields", true);
  }
  
  /**
   * Sow a particular plant in all fields, as specified by density
   * @param plantName The name of the plant to be sown
   * @param density The density at which this plant will be sown
   * @return A message indicating success or failure of this action
   */
  public ABCMResult sowPlantInAllFields(String plantName, double density){
    checkIntegrity();
    
    Plant p = getPlant(plantName);
    Enumeration e = canalSystems.elements();
    while(e.hasMoreElements()){
      ((CanalSystem)e.nextElement()).sow(p, density, modelTime);
    }
    
    e = fields.elements();
    while(e.hasMoreElements()){
      ((Field)e.nextElement()).sow(p, density, modelTime);
    }
    
    return new ABCMResult("Sowed " + p.getName() + " in all fields on " + modelTime.getString(), true);
  }  

  /**
   * Harvests a specific crop in a specific field
   * @param fieldName the name of the field
   * @param plantName the name of the plant whose crop is to be harvested
   * @param index index number of the crop (allows discriminating between multiple
   * crops planted at different times using the same plant)
   * @return an ABCMResult indicating the success or failure of the operation, and, if successful, including a value representing
   * the amount harvested
   */
  public ABCMResult harvestPlantInField(String fieldName, String plantName, int index){
    checkIntegrity();
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not find field '" + fieldName + "'.", false);
    Field f = (Field)(fields.get(fieldName));
    double harvest = f.harvest(plantName, index);    
    return new ABCMResult("Harvest of crop '" + plantName + "' (" + index + ") is " + harvest, harvest, true);
  }
  
  /**
   * Gets the total value harvested in a field since the last reset of its annual harvest accumulator
   * @param fieldName the name of the field
   * @return an ABCM result indicating the success or failure of the operation, and, if successful, 
   * including a value representing the amount harvested 
   */
  public ABCMResult getFieldSumHarvestLastYear(String fieldName){
    checkIntegrity();
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not find field '" + fieldName + "'.", false);
    Field f = (Field)(fields.get(fieldName));
    double harvest = f.getTotalHarvestLastYear();    
    return new ABCMResult("Total Harvest Last Year in field " + fieldName + " is " + harvest, harvest, true);
  }
  
  /**
   * Gets the total amount harvested in the last year along an entire canal system
   * @param canalSystemName the name of the canal system
   * @return an ABCM result indicating the success or failure of the operation, and, if successful, 
   * including a value representing the amount harvested
   */
  public ABCMResult getCanalSysHarvestLastYear(String canalSystemName){
    checkIntegrity();
    // TODO in this method- check for errors and nonexistent elements
    CanalSystem cs = getCanalSystem(canalSystemName);
    if(cs == null) return new ABCMResult("Could not get canal system '" + canalSystemName + "'", false);
    double harvest = cs.getTotalHarvestLastYear();
    return new ABCMResult("Total harvest for canal system is " + harvest, harvest, true);    
  }
  
  /* Environment (Natural)*/
   
  /**
   * Sets the mean percentage of annual daytime hours to a specified value
   * @param d the value to be set
   * @return a String indicating the completion of the operation
   */
  public ABCMResult setMeanPctAnnualDaytimeHours(double d){
    E.setMeanPctAnnualDaytimeHours(d);
    return new ABCMResult("Set Mean Pct Annual Daytime Hours to " + d, true);
  }
  
   
  /**
   * Gets the mean percentage of annual daytime hours
   * @return the mean percentage of annual daytime hours
   */
  public ABCMResult getMeanPctAnnualDaytimeHours(){
    return new ABCMResult(E.getMeanPctAnnualDaytimeHours(), true);
  }
 
  /**
   * Sets the mean daily temperature
   * @param d the value to be set
   * @return an ABCMResult indicating the completion of the operation
   */
  public ABCMResult setMeanDailyTemperature(double d){
    checkIntegrity();
    E.setMeanDailyTemperature(d);
    return new ABCMResult("Set Mean Daily Temperature to " + d, true);
  }
 
  /**
   * Sets the temperature for an altitude gradient
   * @param lowerAlt the lower altitude to be used
   * @param lowerAltTemp the temperature at the lower altitude
   * @param upperAlt the upper altitude to be used
   * @param upperAltTemp the temperature at the upper altitude
   * @return an ABCMResult indicating completion of the operation
   */
  public ABCMResult setMeanDailyTemperatureAtAltitude(double lowerAlt, double lowerAltTemp, double upperAlt, double upperAltTemp){
    checkIntegrity();
    E.setMeanDailyTempByAltitude(lowerAlt, lowerAltTemp, upperAlt, upperAltTemp);
    return new ABCMResult("Set Mean Daily Temperature By Altitude: " + lowerAlt + "->" + lowerAltTemp + "/" + upperAlt + "->" + upperAltTemp, true);
  }
  
  /**
   * Gets the mean daily temperature
   * @return the mean daily temperature
   */
  public ABCMResult getMeanDailyTemperature(){
    return new ABCMResult(E.getMeanDailyTemperature(), true);
  }
  
  /**
   * Gets the mean daily temperature at a specific altitude
   * @param altitude the altitude
   * @return an ABCMResult indicating the temperature at that altitude
   */
  public ABCMResult getMeanDailyTemperatureAtAltitude(double altitude){
    return new ABCMResult(E.getMeanDailyTempByAltitude(altitude), true);
  }
 
  
  /* Populations */
  
  /**
   * Initializes the population object. Passes it a pointer
   * to the model object and the current Code Model.
   */
  public AbstractABCModelPopulation getPopulation(){
    return new WMPopulation(this);
  } 
  
  
  /* ************************************************************************** */  
  /* Reporting                                                                  */
  /*                                                                            */
  /* These functions return string values that describe the state of the model. */
  /* ************************************************************************** */

  /**
   * Return a report outlining the entire status of the Model
   * @return The report, as a String
   */
  public String modelStatus(){
    String retVal = super.modelStatus();    

    retVal += "  Rivers: " + rivers.size() + "\n";
    retVal += "  Canal Systems: " + canalSystems.size() + "\n";
    retVal += "  Fields: " + fields.size(); 
    
    retVal += ((population != null) ? "  Population: " + population.countOfAgents() : "  No Population " ) + "\n"; 
    return retVal;
  }
  
  /**
   * Gets a report of all CanalSystem objects' summaries
   * @return a String containing a report of each canal system
   * contained by the model
   */
  public String getCanalSystemReports(){
    StringWriter out = new StringWriter();
    out.write("\nCanalSystems Report:\n");
    Enumeration e = canalSystems.elements();
    while(e.hasMoreElements()){
      CanalSystem cs = (CanalSystem)e.nextElement();
      out.write(cs.report());
    }
    return out.toString();
  }

  /**
   * Gets a report for the specified field.
   * 
   * If no field with the specified name exists, return 
   * an informative message.
   * @param fieldName The name of the field for which a report
   * is requested
   * @return A report describing the status of this field
   */
  public String getFieldReport(String fieldName){
    if(fields.containsKey(fieldName)){
      return ((Field)fields.get(fieldName)).getReport();
    }
    else{
      return "No field '" + fieldName + "' exists";
    }
  }
  
  
  /* ************************************************************************** */  
  /* Exporting and Viewing                                                      */
  /*                                                                            */
  /* ************************************************************************** */
  /**
   * Generate a file describing the model's canal systems in HEC-RAS import format.
   * @return a String indicating the success of the operation and the new file name.
   */
  public String HECOUT(){
    String fileName = "<No File Opened>";
    try{
//      File f = File.createTempFile("HWM",".HEC");
      File f = new File(env.getImageOutputLocation() + "TEMPHECOUT.HEC");
      FileOutputStream fos = new FileOutputStream(f);
      Enumeration e = canalSystems.elements();
      while(e.hasMoreElements()){
        CanalSystem c = (CanalSystem)e.nextElement();
        fos.write((c.writeHECImportFile()).getBytes());
      }
      fileName = f.getName();
    }
    catch(Exception e){
      return "An error occurred while creating HEC output: " + e.getMessage();
    }
    return "HEC Import file created: " + fileName;
  }
  

  /* ************************************************************************** */
  /* Probe Interrogative Functions                                              */
  /* These functions return numeric values and can thus be used for numeric     */
  /* probes pushing data into numeric data series.                              */
  /* ************************************************************************** */
  
  /**
   * Returns the total volume of water currently present in
   * all canals in all canal systems in the model
   * @return sum of all water in all canals in all canal systems
   * in the model.
   */
  public ABCMResult getTotalVolumeOfWater(){
    double totalVolumeOfWater = 0;
    Enumeration e = canalSystems.elements();
    while(e.hasMoreElements()){
      CanalSystem cs = (CanalSystem)e.nextElement();
      totalVolumeOfWater += cs.totalVolumeOfWater();
    }
    return new ABCMResult(totalVolumeOfWater, true);
  }

  /**
   * Returns the total volume of water in a specific canal system
   * @param CanalSystemName the name of the Canal System for which data are requested
   * @return the total volume of water in the Canal System
   */
  public ABCMResult getTotalVolumeOfWaterInSpecificCanalSystem(String CanalSystemName){
    if(canalSystems.containsKey(CanalSystemName))
      return new ABCMResult(((CanalSystem)canalSystems.get(CanalSystemName)).totalVolumeOfWater(), true);
    else
      return new ABCMResult(0, false);
  }
  
  /**
   * Returns the total canal volume for all Canal Systems
   * @return the total volume for all Canals in all Canal Systems
   */
  public ABCMResult getTotalCanalVolume(){
    double totalVolume = 0;
    Enumeration e = canalSystems.elements();
    while(e.hasMoreElements()){
      CanalSystem cs = (CanalSystem)e.nextElement();
      totalVolume += cs.totalVolumeOfCanals();
    }
    return new ABCMResult(totalVolume, true);
  }
  
  /**
   * Returns the streamflow for a specified River
   * @param riverName the name of the River for which the streamflow is requested
   * @return the current setting of the River's streamflow value
   */
  public ABCMResult getStreamflowForRiver(String riverName){
    River R = getRiver(riverName);
    if (R == null) return riverNotFound("get stream flow", riverName);
    return new ABCMResult(R.getAnnualStreamflow(), true);
  }
  
  /**
   * Returns the flow level for a specified River
   * @param riverName the name of the River for which the flow level is requested
   * @return the current flow level for the River
   */
  public ABCMResult getFlowLevelForRiver(String riverName){
    River R = getRiver(riverName);
    if (R == null) return riverNotFound("get stream flow", riverName);
    return new ABCMResult(R.getFlowLevel(), true);
  }
  
  /**
   * Returns the amount of water flowing daily for the specified river
   * @param riverName the name of the River for which the daily flow is requested
   * @return the daily flow for the specified River
   */
  public ABCMResult getDailyWaterForRiver(String riverName){
    River R = getRiver(riverName);
    if (R == null) return riverNotFound("get stream flow", riverName);
    return new ABCMResult(R.getCurrentDayDailyFlowInMAF(), true);
  }  
  
  /**
   * Returns the total length of all Canals in all Canal Systems in the model
   * @return the total length of all Canals in all Canal Systems in the model 
   */
  public ABCMResult getLengthOfAllCanals(){
    double length = 0;
    Enumeration e = canalSystems.elements();
    while(e.hasMoreElements()){
      length += ((CanalSystem)e.nextElement()).totalLength();
    }
    return new ABCMResult(length, true);
  }
  
  /**
   * Gets the downstream stage for a specified segment at a given time;
   * relies on the segment's internal hydrograph for the past day
   * @param CanalSystemName name of the CanalSystem
   * @param canalSegName name of the CanalSegment (or proxy name)
   * @param time time, specified as a fraction of a day
   * @return an ABCMResult indicting the requested stage value
   */
  public ABCMResult getDownstreamStageForSegmentAt(String canalSystemName, String canalSegName, double time){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return canalSystemNotFound("get downstream stage for segment", canalSystemName);
    return new ABCMResult(CS.getDownstreamStageInSegAt(canalSegName, time), true);    
  }

  public ABCMResult getDownstreamDischargeForSegment(String canalSystemName, String canalSegName){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return canalSystemNotFound("get downstream stage for segment", canalSystemName);
    CanalSegment canalSeg = CS.getCanalSegment(canalSegName);
    if(canalSeg == null) return new ABCMResult("Could not get downstream discharge for segment; segment '" + canalSegName + " does not exist in canal system '" + canalSystemName + "'", false);
    double val = canalSeg.getDownstreamDischarge();
    return new ABCMResult("Downstream discharge for segment " + canalSegName + " in system " + canalSystemName + " is " + val, val, true);
  }
  
  /**
   * Gets the flow at a specified headgate
   * @param riverName name of the River on which headgate is located
   * @param headGate name of the headgate requested
   * @return an ABCMResult with the requested value, if available
   */
  public ABCMResult getFlowAtHeadgate(String riverName, String headGate){
    River R = getRiver(riverName);
    if(R == null) return riverNotFound("get flow at headgate", riverName);
    HeadGate HG = R.getHeadGate(headGate);
    if(HG == null) return new ABCMResult("Could not locate headgate '" + headGate + "' on river '" + riverName + "'.", false);
    return new ABCMResult(R.getDailyMAF(HG), true);
  }
      
  /**
   * Gets the concentration of a specified chemical in a specific CanalSegment
   * @param canalSystemName name of the CanalSystem
   * @param canalSegmentName name of the CanalSegment
   * @param chemicalName name of the Chemical
   * @return the requested concentration
   */
  public ABCMResult getConcentrationInSegment(String canalSystemName, String canalSegmentName, String chemicalName){
    CanalSystem CS = getCanalSystem(canalSystemName);
    if(CS == null) return new ABCMResult("Could not get concentration; no canal system named '" + canalSystemName + "' exists", 0, false);
    CanalSegment C = CS.getCanalSegment(canalSegmentName);
    if(C == null) return new ABCMResult("Could not get concentration; no canal system named '" + canalSegmentName + "' exists", 0, false);
    double conToReturn = C.getOutflowWaterSample().getConcentration(chemicalName);
    return new ABCMResult("" + conToReturn, conToReturn, false);
  }
  
  /**
   * Gets a specified field's soil water depth
   * @param fieldName the name of the field
   * @return an ABCM result indicating the success or failure of the operation, and, if successful, 
   * including a value representing the water depth
   */
  public ABCMResult getFieldSoilWaterDepth(String fieldName){
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not get field soil water depth; no field named '" + fieldName + "' exists", false);
    Field F = (Field)fields.get(fieldName);
    double depth = F.getWaterAvailableMM();
    return new ABCMResult("" + depth, depth, true);
  }
  
  /**
   * Gets the concentration in a specified field of a specified chemical
   * @param fieldName the name of the field
   * @param chemical the name of the chemical
   * @return an ABCM result indicating the success or failure of the operation, and, if successful, 
   * including a value representing the concentration
   */
  public ABCMResult getFieldSoilChem(String fieldName, String chemical){
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not get field chemistry; no field named '" + fieldName + "' exists", false);
    Field F = (Field)fields.get(fieldName);
    double conc = F.getChemicalConcentration(new BasicChemical(chemical, 1));
    return new ABCMResult("" + conc, conc, true);
  }
  
  
  /**
   * Gets the potential yield of a specific crop in a specific field
   * @param fieldName the name of the field
   * @param plantName the name of the plant whose crop is sought
   * @param cropIndex the index of the crop (allows multiple crops of the same plant to be differentiated)
   * @return an ABCM result indicating the success or failure of the operation, and, if successful, 
   * including a value representing the potential yield
   */
  public ABCMResult getFieldPotentialYield(String fieldName, String plantName, int cropIndex){
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not get field potential yield; no field named '" + fieldName + "' does not exist", false);
    Field F = (Field)fields.get(fieldName);
    double plantPotentialYield = F.getPotentialYield(plantName, cropIndex);
    if(Double.isNaN(plantPotentialYield)) return new ABCMResult("Could not get potential yield; no living plant in field named .", false);
    return new ABCMResult("Yield potential is " + plantPotentialYield + " for crop " + cropIndex + " of plant " + plantName + " in field " + fieldName, plantPotentialYield, true);
  }
  
  /**
   * Gets the immediately harvestable yield of a specific crop in a specific field
   * @param fieldName the name of the field
   * @param plantName the name of the plant whose crop is sought
   * @param cropIndex the index of the crop (allows multiple crops of the same plant to be differentiated)
   * @return an ABCM result indicating the success or failure of the operation, and, if successful, 
   * including a value representing the harvestable yield
   */
  public ABCMResult getFieldHarvestableYield(String fieldName, String plantName, int cropIndex){
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not get field harvestable yield; no field named '" + fieldName + "' does not exist", false);
    Field F = (Field)fields.get(fieldName);
    double plantPotentialYield = F.getHarvestableYield(plantName, cropIndex);
    if(Double.isNaN(plantPotentialYield)) return new ABCMResult("Could not get harvestable yield; no living plant in field named .", false);
    return new ABCMResult("Yield potential is " + plantPotentialYield + " for crop " + cropIndex + " of plant " + plantName + " in field " + fieldName, plantPotentialYield, true);
  }
  
  /**
   * Gets the amount of water received by a field at the last time step
   * @param fieldName the name of the field
   * @return an ABCM result indicating the success or failure of the operation, and, if successful, 
   * including a value representing the amount of water
   */
  public ABCMResult getFieldWaterReceivedAtLastStep(String fieldName){
    if(!fields.containsKey(fieldName)) return new ABCMResult("Could not get field harvestable yield; no field named '" + fieldName + "' does not exist", false);
    Field F = (Field)fields.get(fieldName);
    return new ABCMResult("Water added is " + F.getWaterReceivedAtLastStep(), F.getWaterReceivedAtLastStep(), true);
  }
  
  /**
   * Gets the total amount of water received by all selected fields during the last time step
   * @return an ABCM result indicating the success or failure of the operation, and, if successful, 
   * including a value representing the amount of water
   */
  public ABCMResult getWaterReceivedAtLastStepInSelectedFields(){
    Iterator I = fields.values().iterator();
    double w = 0;
    while(I.hasNext()){
      Field F = (Field)I.next();
      if(F.isSelected()) w += F.getWaterReceivedAtLastStep();
    }
    return new ABCMResult("Water added to selected fields is " + w, w, true);
  }
  
  /**
   * Gets a space-delimited list of names of 'distribution' canals in the specified
   * canal segment. These can be parsed by agents to determine points at which to manage the canal system 
   * @param csName the name of the canal system
   * @return an ABCMResult object indicating the success or failure of the operation and containing the
   * requested list
   */
  public ABCMResult getCSDistribCanInitSegNames(String csName){
    CanalSystem cs = getCanalSystem(csName);
    if(cs == null) return new ABCMResult("No canal system named '" + csName + "' exists", false);
    
    List<String> l = cs.getDistributionCanalInitialSegmentNames();
    String ret = l.size() + " segments: \n\n";
    for(String s: l) ret += "   " + s + "\n";
    return new ABCMResult(ret, l, true);
    
  }
  
  
  /**
   * Assigns the fields to randomly selected agents. No guarantee
   * is made about even distribution among the agents, etc.
   * @return a result indicating the success or failure of this operation
   */
  public ABCMResult assignFieldsToAgentsRandomly(){
    Vector F = new Vector(fields.keySet());
    for(int i = 0; i < F.size(); i++)((WMAgent)population.getAgentAtRandom()).assignField((String)F.get(i));
    return new ABCMResult("Assigned " + fields.size() + " fields to agents", true);
  }
  
  /**
   * Assigns the fields that are selected to the WMAgent specifiedd
   * @param agentID ID of the agent to be assigned the fields
   * @return an ABCMResult object indicating success or failure of the operation
   */
  public ABCMResult assignSelectedFieldsToAgent(String agentID){
    WMAgent agent = (WMAgent)population.getAgent(agentID);
    if(agent == null) return new ABCMResult("Could not assign fields; no agent '" + agentID + "' exists", false);
    int fieldCount = 0;
    Vector F = new Vector(fields.entrySet());
    for(int i = 0; i < F.size(); i++){
      Field f = (Field)F.get(i);
      if(f.isSelected()){
        agent.assignField((String)F.get(i));
        fieldCount++;
      }
    }
    return new ABCMResult("Assigned " + fieldCount + " fields to agent '" + agentID + "'", true);
  }
  
  /**
   * Assign a field to an agent
   * @param agentID the ID of the agent to receive the field
   * @param fieldName the name of the field to be assigned
   * @return an ABCMResult object indicating success or failure of the operation
   */
  public ABCMResult assignFieldToAgent(String agentID, String fieldName){
    WMAgent agent = (WMAgent)population.getAgent(agentID);
    if(agent == null) return new ABCMResult("Could not assign fields; no agent '" + agentID + "' exists", false);
    Field F = (Field)fields.get(fieldName);
    if(F == null) return new ABCMResult("Could not assign fields; no field '" + fieldName + "' exists", false);
    agent.assignField(fieldName);
    return new ABCMResult("Assigned field '" + fieldName + "' to agent '" + agentID + "'", true);
  }
}
