package wm.model;


import wm.space.WMSpace;
import abcm.model.ABCMResult;
import abcm.model.AbstractABCMParser;
import abcm.model.AbstractABCModelEnvironment;
import abcm.model.ParseAction;
import abcm.space.TopographicVolumeDifferential;
import abcm.space.TriangulatedTopography;

/**
 * The Parser object translates instructions in various
 * formats into execution statements that can be carried
 * out by the Model object. The list of valid commands
 * is stored in the HWM database; the parser must recognize
 * the commands by comparing them to the list in the db.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class WMParser extends AbstractABCMParser {
  
  /* Actions */

    /* Environment */
  
  protected class SetMeanPctAnnualDayHrs extends ParseAction{
    private String action = "set value";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult dayHrs = parseArgDouble(action, command, 1); if(!dayHrs.success) return dayHrs;
      return ((WMModel)env.getModel()).setMeanPctAnnualDaytimeHours(dayHrs.value);
    }
  }
  protected class SetMeanDailyTemp extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set value");  
      ABCMResult temp = parseArgDouble(action, command, 1); if(!temp.success) return temp;
      return ((WMModel)env.getModel()).setMeanDailyTemperature(temp.value);
    }
  }
  protected class SetMeanDailyTempAtAlt extends ParseAction{
    private String action="set daily temp by altitude";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult lowAlt     = parseArgDouble(action, command, 1); if(!lowAlt.success)     return lowAlt;
      ABCMResult lowAltTemp = parseArgDouble(action, command, 2); if(!lowAltTemp.success) return lowAltTemp;
      ABCMResult upAlt      = parseArgDouble(action, command, 3); if(!upAlt.success)      return upAlt;
      ABCMResult upAltTemp  = parseArgDouble(action, command, 4); if(!upAltTemp.success)  return upAltTemp;
      return((WMModel)env.getModel()).setMeanDailyTemperatureAtAltitude(lowAlt.value, lowAltTemp.value, 
          upAlt.value, upAltTemp.value);
    }
  }

  protected class SetBasicDailyRainProb extends ParseAction{
    private String action = "set basic daily rainfall probability";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set value");  
      ABCMResult prob = parseArgDouble(action, command, 1); if(!prob.success) return prob;
      return ((WMModel)env.getModel()).setBasicDailyRainfallProbability(prob.value);
    }
  }
  
  protected class SetBasicDailyRainMin extends ParseAction{
    private String action = "set basic daily rainfall minimum amount";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set value");  
      ABCMResult amt = parseArgDouble(action, command, 1); if(!amt.success) return amt;
      return ((WMModel)env.getModel()).setBasicDailyRainfallMinAmount(amt.value);
    }
  }
  
  protected class SetBasicDailyRainMax extends ParseAction{
    private String action = "set basic daily rainfall maximum amount";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set value");  
      ABCMResult amt = parseArgDouble(action, command, 1); if(!amt.success) return amt;
      return ((WMModel)env.getModel()).setBasicDailyRainfallMaxAmount(amt.value);
    }
  }
  
  
  
    /* Rivers */
  
  protected class CreateRiver extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("create river");
      ABCMResult river = parseArgString("CreateRiver", command, 1); if(!river.success) return river;
      return ((WMModel)env.getModel()).addRiver(river.text);
    }
  }  
  protected class SetAnnualStreamflow extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set annual streamflow");
      ABCMResult river = parseArgString("SetAnnualStreamflow", command, 1); if(!river.success) return river;
      ABCMResult flow = parseArgDouble("SetAnnualStreamflow", command, 2); if(!flow.success) return flow;
      return ((WMModel)env.getModel()).setRiverAnnualStreamFlow(river.text, flow.value);  
    }
  }  
  protected class SetCurrentStreamFlow extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set current streamflow");
      ABCMResult river = parseArgString("SetCurrentStreamflow", command, 1); if(!river.success) return river;
      ABCMResult flow = parseArgDouble("SetCurrentStreamflow", command, 2); if(!flow.success) return flow;
      return ((WMModel)env.getModel()).setRiverFlowLevel(river.text, flow.value);
    }
  }  
  protected class AddHeadGateToRiver extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("add headgate");
      ABCMResult river = parseArgString("AddHeadGateToRiver", command, 1); if(!river.success) return river;
      ABCMResult point = parseArgString("AddHeadGateToRiver", command, 2); if(!point.success) return point;
      ABCMResult index = parseArgDouble("AddHeadGateToRiver", command, 3); if(!index.success) return index;
      ABCMResult name  = parseArgString("AddHeadGateToRiver", command, 4); if(!name.success) return name;
      return ((WMModel)env.getModel()).addHeadgateToRiver(river.text, point.text, name.text, (int)index.value);   
    }
  }  
  protected class SetChemicalInRiverWater extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set chemical in water");  
      ABCMResult river   = parseArgString("SetChemicalInRiverWater", command, 1); if(!river.success) return river;
      ABCMResult chem    = parseArgString("SetChemicalInRiverWater", command, 2); if(!chem.success) return chem;
      ABCMResult concntr = parseArgDouble("SetChemicalInRiverWater", command, 3); if(!concntr.success) return concntr;
      return ((WMModel)env.getModel()).setRiverChemicalConcentration(river.text, chem.text, concntr.value);   
    }
  }
  
    /* Canal Systems */
  
  protected class CreateCanalSystem extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("create canal system");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      return ((WMModel)env.getModel()).createCanalSystem(csName.text);
    }
  }
  protected class AddCanalToCanalSystem extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("add canal");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult p1 = parseArgString(action, command, 2); if(!p1.success) return p1;
      ABCMResult p2 = parseArgString(action, command, 3); if(!p2.success) return p2;
      ABCMResult segName = parseArgString(action, command, 4); if(!segName.success) return segName;
      return ((WMModel)env.getModel()).addCanalToCanalSystem(csName.text, p1.text, p2.text, segName.text);
    }
  }  
  protected class ExtendCanal extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("extend canal");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult p2 = parseArgString(action, command, 3); if(!p2.success) return p2;
      ABCMResult newCanalName = parseArgString(action, command, 4); if(!newCanalName.success) return newCanalName;
      return ((WMModel)env.getModel()).extendCanal(csName.text, segName.text, p2.text, newCanalName.text);
    }
  }  
  protected class ExtendLastCanal extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("extend canal");  
      ABCMResult canalSystem = parseArgString(action, command, 1); if(!canalSystem.success) return canalSystem;
      ABCMResult p2 = parseArgString(action, command, 2); if(!p2.success) return p2;
      ABCMResult newCanalName = parseArgString(action, command, 3); if(!newCanalName.success) return newCanalName;
      return ((WMModel)env.getModel()).extendSelectedCanalSegment(canalSystem.text, p2.text, newCanalName.text);
    }
  }
  protected class SetCanalRank extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set canal rank");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult rank = parseArgDouble(action, command, 3); if(!rank.success) return rank;
      return ((WMModel)env.getModel()).setCanalSegmentRank(csName.text, segName.text, (int)rank.value);
    }
  }
  protected class AddCanalAtGate extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("add canal");
      ABCMResult riverName = parseArgString(action, command, 1); if(!riverName.success) return riverName;
      ABCMResult csName = parseArgString(action, command, 2); if(!csName.success) return csName;
      ABCMResult headgateName = parseArgString(action, command, 3); if(!headgateName.success) return headgateName;
      ABCMResult p2 = parseArgString(action, command, 4); if(!p2.success) return p2;
      ABCMResult newCanalName = parseArgString(action, command, 5); if(!newCanalName.success) return newCanalName;
      return ((WMModel)env.getModel()).startCanalAtHeadGate(riverName.text, csName.text, headgateName.text, p2.text, newCanalName.text);
    }
  }
  protected class SetUpstreamDischarge extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set discharge");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult Q = parseArgDouble(action, command, 3); if(!Q.success) return Q;
      return ((WMModel)env.getModel()).setCanalSegmentUpstreamDischarge(csName.text, segName.text, Q.value);
    }
  }    
  protected class SetUpstreamStage extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set stage");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult Y = parseArgDouble(action, command, 3); if(!Y.success) return Y;
      return ((WMModel)env.getModel()).setCanalSegmentUpstreamStage(csName.text, segName.text, Y.value);
    }
  }
  protected class SetDownstreamDischarge extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set discharge");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult Q = parseArgDouble(action, command, 3); if(!Q.success) return Q;
      return ((WMModel)env.getModel()).setCanalSegmentDownstreamDischarge(csName.text, segName.text, Q.value);
    }
  }    
  protected class SetDownstreamStage extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set stage");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult Q = parseArgDouble(action, command, 3); if(!Q.success) return Q;
      return ((WMModel)env.getModel()).setCanalSegmentDownstreamStage(csName.text, segName.text, Q.value);
    }
  }
  protected class SetUpstreamQperY extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set value");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult Y = parseArgDouble(action, command, 3); if(!Y.success) return Y;
      return ((WMModel)env.getModel()).setCanalSegmentUpstreamQperY(csName.text, segName.text, Y.value);
    }
  }
  protected class SetDownstreamQperY extends ParseAction{
    private String action = "set downstream Q per Y";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult Y = parseArgDouble(action, command, 3); if(!Y.success) return Y;
      return ((WMModel)env.getModel()).setCanalSegmentDownstreamQperY(csName.text, segName.text, Y.value);
    }
  }
  protected class SetMaximumOutflow extends ParseAction{
    private String action = "set maximum outflow";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult maxQ = parseArgDouble(action, command, 3); if(!maxQ.success) return maxQ;
      return ((WMModel)env.getModel()).setCanalSegmentMaximumOutflow(csName.text, segName.text, maxQ.value);
    }
  }
  protected class SetCanalSegmentSize extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set canal segment size");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult topWidth = parseArgDouble(action, command, 3); if(!topWidth.success) return topWidth;
      ABCMResult bottomWidth = parseArgDouble(action, command, 4); if(!bottomWidth.success) return bottomWidth;
      ABCMResult channelHeight = parseArgDouble(action, command, 5); if(!channelHeight.success) return channelHeight;
      return ((WMModel)env.getModel()).setCanalSegmentSize(csName.text, segName.text, topWidth.value, bottomWidth.value, channelHeight.value);
    }
  }    
  protected class SetCanalSegmentRoughness extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set roughness");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult n = parseArgDouble(action, command, 3); if(!n.success) return n;
      return ((WMModel)env.getModel()).setCanalSegmentRoughness(csName.text, segName.text, n.value);
    }
  }
  protected class SetUpstreamZandQ extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("Set values");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult Z = parseArgDouble(action, command, 3); if(!Z.success) return Z;
      ABCMResult Q = parseArgDouble(action, command, 4); if(!Q.success) return Q;      
      return ((WMModel)env.getModel()).setCanalSegmentUpstreamZandQ(csName.text, segName.text, Z.value, Q.value);
    }
  }
  protected class SetDownstreamZandQ extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set values");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult Z = parseArgDouble(action, command, 3); if(!Z.success) return Z;
      ABCMResult Q = parseArgDouble(action, command, 4); if(!Q.success) return Q;      
      return ((WMModel)env.getModel()).setCanalSegmentDownstreamZandQ(csName.text, segName.text, Z.value, Q.value);
    }
  }
  protected class SetCanalSegmentProfile extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){      
      if(!env.hasModel()) return noModelMsg("set profile");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult spec = parseArgString(action, command, 3); if(!spec.success) return spec;
//      return new ABCMResult("NOT CURRENTLY IMPLEMENTED", true);
      return ((WMModel)env.getModel()).setCanalSegmentProfile(csName.text, segName.text, spec.text);
    }
  }
  protected class SetCanalSegmentType extends ParseAction{
    private String action= "set canal type";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set canal segment size");
      ABCMResult csName  = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult type    = parseArgString(action, command, 3); if(!type.success) return type;
      return ((WMModel)env.getModel()).setCanalSegmentType(csName.text, segName.text, type.text);
    }
  }
  protected class SetHeadGateEfficiency extends ParseAction{
    private String action = "set headgate efficiency";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set efficiency");
      ABCMResult river = parseArgString(action, command, 1); if(!river.success) return river;
      ABCMResult headgate = parseArgString(action, command, 2); if(!headgate.success) return headgate;
      ABCMResult efficiency = parseArgDouble(action, command, 3); if(!efficiency.success) return efficiency;
      return ((WMModel)env.getModel()).setHeadGateEfficiency(river.text, headgate.text, efficiency.value);
    }
  }
  protected class SetHGEfficiencyRanges extends ParseAction{
    private String action = "set headgate efficiency ranges";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult river = parseArgString(action, command, 1); if(!river.success) return river;
      ABCMResult headgate = parseArgString(action, command, 2); if(!headgate.success) return headgate;
      ABCMResult ranges = parseArgString(action, command, 3); if(!ranges.success) return ranges;
      return ((WMModel)env.getModel()).setHGEfficiencyRanges(river.text, headgate.text, ranges.text);
    }
  }
  protected class DropCanalSegment extends ParseAction{
    private String action = "drop canal segment";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      return ((WMModel)env.getModel()).dropCanalSegment(csName.text, segName.text);
    }
  }

  
  protected class CloseHeadGate extends ParseAction{
    private String action = "close headgate";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult river = parseArgString(action, command, 1); if(!river.success) return river;
      ABCMResult headgate = parseArgString(action, command, 2); if(!headgate.success) return headgate;
      return ((WMModel)env.getModel()).closeHeadGate(river.text, headgate.text);
    }
  }
  
  protected class OpenHeadGate extends ParseAction{
    private String action = "open headgate";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult river = parseArgString(action, command, 1); if(!river.success) return river;
      ABCMResult headgate = parseArgString(action, command, 2); if(!headgate.success) return headgate;
      return ((WMModel)env.getModel()).openHeadGate(river.text, headgate.text);
    }    
  }
  
  protected class CloseInGate extends ParseAction{
    private String action = "close in gate";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      return ((WMModel)env.getModel()).closeInflowGate(csName.text, segName.text);
    }
  }
  protected class OpenInGate extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("open gate");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      return ((WMModel)env.getModel()).openInflowGate(csName.text, segName.text);
    }
  }
  protected class CloseOutGate extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("close gate");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      return ((WMModel)env.getModel()).closeOutflowGate(csName.text, segName.text);
    }
  }    
  protected class OpenOutGate extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("open gate");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      return ((WMModel)env.getModel()).openOutflowGate(csName.text, segName.text);
    }
  }
  
    /* Fields */  
  protected class CreateField extends ParseAction{
    private String action = "create field";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult NW = parseArgString(action, command, 1); if(!NW.success) return NW;
      ABCMResult NE = parseArgString(action, command, 2); if(!NE.success) return NE;
      ABCMResult SE = parseArgString(action, command, 3); if(!SE.success) return SE;
      ABCMResult SW = parseArgString(action, command, 4); if(!SW.success) return SW;
      ABCMResult fieldName = parseArgString(action, command, 5); if(!fieldName.success) return fieldName;
      return ((WMModel)env.getModel()).createField(NW.text, NE.text, SE.text, SW.text, fieldName.text);
    }
  }
  protected class DropField extends ParseAction{
    private String action = "drop field";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      return ((WMModel)env.getModel()).dropField(field.text);
    }
  }
  protected class LinkCanalSegToField extends ParseAction{
    private String action = "link canal segment to field";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult canalSys = parseArgString(action, command, 1); if(!canalSys.success) return canalSys;
      ABCMResult canalSeg = parseArgString(action, command, 2); if(!canalSeg.success) return canalSeg;
      ABCMResult field    = parseArgString(action, command, 3); if(!field.success)    return field;
      return ((WMModel)env.getModel()).linkCanalSegmentToField(canalSys.text, canalSeg.text, field.text);
    }
  }
  protected class UnlinkCanalSegFromField extends ParseAction{
    private String action = "unlink canal segment from field";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult canalSys = parseArgString(action, command, 1); if(!canalSys.success) return canalSys;
      ABCMResult canalSeg = parseArgString(action, command, 2); if(!canalSeg.success) return canalSeg;
      ABCMResult field    = parseArgString(action, command, 3); if(!field.success)    return field;
      return ((WMModel)env.getModel()).unlinkCanalSegmentFromField(canalSys.text, canalSeg.text, field.text);
    }
  }
  
  
  
  protected class PlantInAll extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("plant");
      ABCMResult plant = parseArgString(action, command, 1); if(!plant.success) return plant;
      ABCMResult density = parseArgDouble(action, command, 2); if(!density.success) return density;
      return ((WMModel)env.getModel()).sowPlantInAllFields(plant.text, density.value);
    }
  }  
  protected class SetEToForField extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set EtO");  
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      ABCMResult eTo = parseArgDouble(action, command, 2); if(!eTo.success) return eTo;
      return ((WMModel)env.getModel()).setEToForField(field.text, eTo.value);
    }
  }
  protected class PlantInSelectedFields extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("plant");  
      ABCMResult plant = parseArgString(action, command, 1); if(!plant.success) return plant;
      ABCMResult density = parseArgDouble(action, command, 2); if(!density.success) return density;
      return ((WMModel)env.getModel()).sowPlantInSelectedFields(plant.text, density.value);
    }
  }    
  protected class SetFieldEffectiveArea extends ParseAction{
    private String action = "set field effective area";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);  
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      ABCMResult effectiveArea = parseArgDouble(action, command, 2); if(!effectiveArea.success) return effectiveArea;
      return ((WMModel)env.getModel()).setFieldEffectiveArea(field.text, effectiveArea.value);
    }
  }
  protected class HarvestPlantInField extends ParseAction{
    private String action = "harvest plant in a field";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);  
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      ABCMResult plant = parseArgString(action, command, 2); if(!plant.success) return plant;
      ABCMResult index = parseArgString(action, command, 3); if(!index.success) return index; 
      return((WMModel)env.getModel()).harvestPlantInField(field.text, plant.text, (int)index.value);
    }
  }
  
  
  
    /* Rain */  
  protected class RainOnField extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("rain");
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      ABCMResult mm = parseArgDouble(action, command, 2); if(!mm.success) return mm;
      return ((WMModel)env.getModel()).rainMMOnField(field.text, mm.value);
    }
  }
  protected class RainOnAllFields extends ParseAction{    
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("rain");  
      ABCMResult mm = parseArgDouble(action, command, 1); if(!mm.success) return mm;
      return ((WMModel)env.getModel()).rainMMOnAllFields(mm.value);
    }
  }  
  protected class RainOnSelectedFields extends ParseAction{
    private String action = "rain on selected fields";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("rain");  
      ABCMResult mm = parseArgDouble(action, command, 1); if(!mm.success) return mm;
      return ((WMModel)env.getModel()).rainOnSelectedFields(mm.value);
    }
  }
  protected class RainAtElevations extends ParseAction{
    private String action = "rain at elevations";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult mm    = parseArgDouble(action, command, 1); if(!mm.success) return mm;
      ABCMResult lower = parseArgDouble(action, command, 2); if(!lower.success) return lower;
      ABCMResult upper = parseArgDouble(action, command, 3); if(!upper.success) return upper;
      return ((WMModel)env.getModel()).rainMMOnFieldsBetweenElevations(mm.value, lower.value, upper.value);      
    }
  }
  

  
  
  
  
  /* Interrogatives */
  
    /* Environment */
  
  protected class GetMeanPctAnnualDayHrs extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get value");  
      return ((WMModel)env.getModel()).getMeanPctAnnualDaytimeHours();
    }
  }
  protected class GetMeanDailyTemp extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get value");  
      return ((WMModel)env.getModel()).getMeanDailyTemperature();
    }
  }
  protected class GetMeanDailyTempAtAlt extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get value");
      ABCMResult alt = parseArgDouble("", command, 1); if(!alt.success) return alt;
      return ((WMModel)env.getModel()).getMeanDailyTemperatureAtAltitude(alt.value);
    }
  }
  
    /* Rivers */
  
  protected class GetStreamflowForRiver extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get streamflow");
      ABCMResult river = parseArgString("GetStreamflowForRiver", command, 1); if(!river.success) return river;
      return ((WMModel)env.getModel()).getStreamflowForRiver(river.text);
    }
  }
  protected class GetFlowLevelForRiver extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get flow level");
      ABCMResult river = parseArgString("GetFlowLevelForRiver", command, 1); if(!river.success) return river;
      return ((WMModel)env.getModel()).getFlowLevelForRiver(river.text);
    }
  }
  protected class GetDailyWaterForRiver extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get daily water");
      ABCMResult river = parseArgString("GetDailyWaterForRiver", command, 1); if(!river.success) return river;
      return ((WMModel)env.getModel()).getDailyWaterForRiver(river.text);
    }
  }  
  protected class GetFlowAtHeadGate extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get flow");
      ABCMResult river = parseArgString(action, command, 1); if(!river.success) return river;
      ABCMResult headgate = parseArgString(action, command, 2); if(!headgate.success) return headgate;
      return ((WMModel)env.getModel()).getFlowAtHeadgate(river.text, headgate.text);
    }
  }
  
    /* Canal Systems */
    
  protected class GetTotalWaterVolume extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get volume");
      return ((WMModel)env.getModel()).getTotalVolumeOfWater();
    }
  }    
  protected class GetTotalWaterVolumeIn extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get volume");
      ABCMResult canalSystem = parseArgString(action, command, 1); if(!canalSystem.success) return canalSystem;  
      return ((WMModel)env.getModel()).getTotalVolumeOfWaterInSpecificCanalSystem(canalSystem.text);
    }
  }
  protected class GetTotalVolumeOfCanals extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get volume");  
      return ((WMModel)env.getModel()).getTotalCanalVolume();
    }
  }
  protected class GetCanalSegDSStageAt extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get stage");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult time = parseArgDouble(action, command, 3); if(!time.success) return time;
      return ((WMModel)env.getModel()).getDownstreamStageForSegmentAt(csName.text, segName.text, time.value);
    }
  }
  protected class GetCanalSegDSDischarge extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get discharge");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      return ((WMModel)env.getModel()).getDownstreamDischargeForSegment(csName.text, segName.text);
    }
  }
  protected class GetConcentrationInSegment extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get concentration");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult chem = parseArgString(action, command, 3); if(!chem.success) return chem;
      return new ABCMResult(((WMModel)env.getModel()).getConcentrationInSegment(csName.text, segName.text, chem.text).value, true);
    }
  }
  protected class GetTotalLengthOfAllCanals extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get length");  
      return ((WMModel)env.getModel()).getLengthOfAllCanals();
    }
  }

//  protected class GetMaxDischargeAtHG extends ParseAction{
//    private String action = "get maximum discharge";
//    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
//      if(!env.hasModel()) return noModelMsg(action);
//      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
//      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
//      return ((WMModel)env.getModel()).getMaxDischarge(csName.text, segName.text);
//    }
//  }
  
  protected class CompareSpaceVolume extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      ABCMResult branchName = parseArgString(action, command, 1); if(!branchName.success) return branchName;
      if(!env.hasModel()) return noModelMsg("compare space volumes.");  
      if(!env.hasModel(branchName.text)) return new ABCMResult("No branch with the name '" + branchName + "' exists", false);
      WMSpace origSpace = (WMSpace)(env.getModelFromBranchPoint(branchName.text)).getSpace();
      WMSpace currSpace = (WMSpace)(env.getModel().getSpace());
      System.out.println("PARSER DEBUG: Getting t1");
      TriangulatedTopography orig = origSpace.getTriangulatedTopography();
      System.out.println("PARSER DEBUG: Getting t2");
      TriangulatedTopography curr = currSpace.getTriangulatedTopography();
//      curr.write();
      System.out.println("PARSER DEBUG: Getting diff");
      TopographicVolumeDifferential diff = curr.getTopographicVolumeDifferential(orig);
      System.out.println("PARSER DEBUG: Returning");
      return new ABCMResult("Volume Differential is " + diff.getVolumeAbove() + " above and " + 
                                                        diff.getVolumeBelow() + " below for a total of " + 
                                                        (diff.getVolumeAbove() + diff.getVolumeBelow()), 
                                                        (diff.getVolumeAbove() + diff.getVolumeBelow()), true);
    }
  }
   
  protected class GetCanalSysHarvestLastYr extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get concentration");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      return ((WMModel)env.getModel()).getCanalSysHarvestLastYear(csName.text); 
    }       
  }
  
    /* Fields */
  
    /* Rain */
  
  /* Selection and Object Management     */
  
    /* Environment */
    /* Rivers */
    /* Canal Systems */
  
  protected class SetCanalSegmentProxy extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("set proxy");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      ABCMResult segName = parseArgString(action, command, 2); if(!segName.success) return segName;
      ABCMResult newName = parseArgString(action, command, 3); if(!newName.success) return newName;
      return ((WMModel)env.getModel()).setCanalSegmentProxyName(csName.text, segName.text, newName.text);
    }
  }
  
    /* Fields */
  protected class SelectField extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("select field");  
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      return ((WMModel)env.getModel()).selectField(field.text);
    }
  }
  protected class UnselectField extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("unselect field"); 
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      return ((WMModel)env.getModel()).unselectField(field.text);
    }
  }
  protected class UnselectAllFields extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("unselect fields");
      return ((WMModel)env.getModel()).unselectAllFields();
    }
  }
  protected class GetFieldSoilWaterDepth extends ParseAction{
    private String action="get field soil water depth";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      return ((WMModel)env.getModel()).getFieldSoilWaterDepth(field.text);
    }
  }
  protected class GetFieldSoilChem extends ParseAction{
    private String action = "get field soil chemical content";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      ABCMResult chem  = parseArgString(action, command, 2); if(!chem.success) return field;
      return ((WMModel)env.getModel()).getFieldSoilChem(field.text, chem.text);
    }
  }
  protected class GetFieldPotentialYield extends ParseAction{
    private String action = "get potential yield";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      ABCMResult plant = parseArgString(action, command, 2); if(!plant.success) return plant;
      ABCMResult index = parseArgString(action, command, 3); if(!index.success) return index;
      return ((WMModel)env.getModel()).getFieldPotentialYield(field.text, plant.text, (int)index.value);
    }     
  }
  protected class GetFieldHarvestableYield extends ParseAction{
    private String action = "get potential yield";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      ABCMResult plant = parseArgString(action, command, 2); if(!plant.success) return plant;
      ABCMResult index = parseArgString(action, command, 3); if(!index.success) return index;
      return ((WMModel)env.getModel()).getFieldHarvestableYield(field.text, plant.text, (int)index.value);
    }     
  }
  protected class GetFieldWaterRecd extends ParseAction{
    private String action = "get field water received at last step";
    public ABCMResult executeCommand(String[] command,  AbstractABCModelEnvironment env) {
      if(!env.hasModel()) return noModelMsg(action);
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      return ((WMModel)env.getModel()).getFieldWaterReceivedAtLastStep(field.text);      
    }    
  }
  protected class GetWaterRecdInSelFields extends ParseAction{
    private String action = "get water received at last step in selected fields";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);
      return ((WMModel)env.getModel()).getWaterReceivedAtLastStepInSelectedFields();
    }
  }
  
  protected class GetFieldSumHarvestLastYr extends ParseAction{
    private String action = "get total harvest";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg(action);  
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      return((WMModel)env.getModel()).getFieldSumHarvestLastYear(field.text);
    }    
  }
  
    /* Rain */
  
  /* Reports */
  
    /* Environment */
    /* Rivers */
    /* Canal Systems */
  protected class ViewCanalSystemsReport extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("view canal report");
      return new ABCMResult(((WMModel)env.getModel()).getCanalSystemReports(), true);
    }
  }
  protected class ViewCanalSystemDetails extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("view canal system details");
      ABCMResult csName = parseArgString(action, command, 1); if(!csName.success) return csName;
      return new ABCMResult(((WMModel)env.getModel()).viewCanalSystemDetails(csName.text), true);
    }
  }
  protected class GenerateHECOutput extends ParseAction{
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("generate HEC output");  
      return new ABCMResult(((WMModel)env.getModel()).HECOUT(), true);
    }
  }
  
    /* Fields */
  protected class GetFieldReport extends ParseAction{
    private String action = "";
    public ABCMResult executeCommand(String[] command, AbstractABCModelEnvironment env){
      if(!env.hasModel()) return noModelMsg("get field report");
      ABCMResult field = parseArgString(action, command, 1); if(!field.success) return field;
      return new ABCMResult(((WMModel)env.getModel()).getFieldReport(field.text), true);
    }
  }
  
    /* Rain */
  
  
/* *************************** */
/* END INNER CLASS DEFINITIONS */   
/* *************************** */  

  /**
   * Constructor that takes the environment in which this parser
   * will exist
   * @param env HWM Environment in which this parser will be used
   */
  public WMParser(WMModelEnvironment env){
    super(env);
    
  /* Add commands to the array of commands */

    /* Actions */
      /* Environment */
    addCommand("SetMeanPctAnnualDayHrs",    new SetMeanPctAnnualDayHrs());
    addCommand("SetMeanDailyTemp",          new SetMeanDailyTemp());
    addCommand("SetMeanDailyTempAtAlt",     new SetMeanDailyTempAtAlt());
    addCommand("SetBasicDailyRainProb",     new SetBasicDailyRainProb());
    addCommand("SetBasicDailyRainMin",      new SetBasicDailyRainMin());
    addCommand("SetBasicDailyRainMax",      new SetBasicDailyRainMax());

      /* Rivers */
    addCommand("CreateRiver",               new CreateRiver());
    addCommand("SetAnnualStreamflow",       new SetAnnualStreamflow());
    addCommand("SetCurrentStreamFlow",      new SetCurrentStreamFlow());
    addCommand("AddHeadGateToRiver",        new AddHeadGateToRiver());
    addCommand("SetChemicalInRiverWater",   new SetChemicalInRiverWater());
    
      /* Canal Systems */
    addCommand("CreateCanalSystem",         new CreateCanalSystem());
    addCommand("AddCanalToCanalSystem",     new AddCanalToCanalSystem());
    addCommand("ExtendCanal",               new ExtendCanal());
    addCommand("ExtendLastCanal",           new ExtendLastCanal());
    addCommand("SetCanalRank",              new SetCanalRank());
    addCommand("AddCanalAtGate",            new AddCanalAtGate());
    addCommand("SetUpstreamDischarge",      new SetUpstreamDischarge());
    addCommand("SetUpstreamStage",          new SetUpstreamStage());
    addCommand("SetDownstreamDischarge",    new SetDownstreamDischarge());
    addCommand("SetDownstreamStage",        new SetDownstreamStage());
    addCommand("SetUpstreamQperY",          new SetUpstreamQperY());
    addCommand("SetDownstreamQperY",        new SetDownstreamQperY());
    addCommand("SetMaximumOutflow",         new SetMaximumOutflow());    
    addCommand("SetCanalSegmentSize",       new SetCanalSegmentSize());
    addCommand("SetCanalSegmentRoughness",  new SetCanalSegmentRoughness());
    addCommand("SetUpstreamZandQ",          new SetUpstreamZandQ());
    addCommand("SetDownstreamZandQ",        new SetDownstreamZandQ());
    addCommand("SetCanalSegmentProfile",    new SetCanalSegmentProfile());
    addCommand("SetCanalSegmentType",       new SetCanalSegmentType());
    addCommand("SetHeadGateEfficiency",     new SetHeadGateEfficiency());
    addCommand("SetHGEfficiencyRanges",     new SetHGEfficiencyRanges());
    addCommand("CloseInGate",               new CloseInGate());
    addCommand("OpenInGate",                new OpenInGate());
    addCommand("CloseOutGate",              new CloseOutGate());
    addCommand("OpenOutGate",               new OpenOutGate());
    addCommand("CloseHeadGate",             new CloseHeadGate());
    addCommand("OpenHeadGate",              new OpenHeadGate());
    
    
      /* Fields */
    addCommand("CreateField",               new CreateField());
    addCommand("DropField",                 new DropField());
    addCommand("LinkCanalSegToField",       new LinkCanalSegToField());
    addCommand("UnlinkCanalSegFromField",   new UnlinkCanalSegFromField());

    
    addCommand("PlantInAll",                new PlantInAll());
    addCommand("SetEToForField",            new SetEToForField());
    addCommand("PlantInSelectedFields",     new PlantInSelectedFields());
    addCommand("SetFieldEffectiveArea",     new SetFieldEffectiveArea());
    addCommand("HarvestPlantInField",       new HarvestPlantInField());
    
    
    
      /* Rain */
    addCommand("RainOnField",               new RainOnField());
    addCommand("RainOnAllFields",           new RainOnAllFields());
    addCommand("RainOnSelectedFields",      new RainOnSelectedFields());
    addCommand("RainAtElevations",          new RainAtElevations());

    
    
    /* Interrogatives */

      /* Environment */
    addCommand("GetMeanPctAnnualDayHrs",    new GetMeanPctAnnualDayHrs());
    addCommand("GetMeanDailyTemp",          new GetMeanDailyTemp());
    addCommand("GetMeanDailyTempAtAlt",     new GetMeanDailyTempAtAlt());
    
      /* Rivers */
    addCommand("GetStreamflowForRiver",     new GetStreamflowForRiver());
    addCommand("GetFlowLevelForRiver",      new GetFlowLevelForRiver());
    addCommand("GetDailyWaterForRiver",     new GetDailyWaterForRiver());
    addCommand("GetFlowAtHeadGate",         new GetFlowAtHeadGate());
    
    
      /* Canal Systems */
    addCommand("GetTotalWaterVolume",       new GetTotalWaterVolume());
    addCommand("GetTotalWaterVolumeIn",     new GetTotalWaterVolumeIn());
    addCommand("GetTotalVolumeOfCanals",    new GetTotalVolumeOfCanals());
    addCommand("GetCanalSegDSStageAt",      new GetCanalSegDSStageAt());
    addCommand("GetCanalSegDSDischarge",    new GetCanalSegDSDischarge());
    addCommand("GetConcentrationInSegment", new GetConcentrationInSegment());
    addCommand("GetTotalLengthOfAllCanals", new GetTotalLengthOfAllCanals());
    
    addCommand("CompareSpaceVolume",        new CompareSpaceVolume());
    addCommand("GetCanalSysHarvestLastYr",  new GetCanalSysHarvestLastYr());
    
      /* Fields */
    addCommand("GetFieldSoilWaterDepth",    new GetFieldSoilWaterDepth());
    addCommand("GetFieldSoilChem",          new GetFieldSoilChem());
    addCommand("GetFieldPotentialYield",    new GetFieldPotentialYield());
    addCommand("GetFieldHarvestableYield",  new GetFieldHarvestableYield());
    addCommand("GetFieldWaterRecd",         new GetFieldWaterRecd());
    addCommand("GetWaterRecdInSelFields",   new GetWaterRecdInSelFields());
    addCommand("GetFieldSumHarvestLastYr"  ,new GetFieldSumHarvestLastYr());
    
    /* Selection and Object Management */
      /* Environment */
      /* Rivers */
      /* Canal Systems */
    addCommand("SetCanalSegmentProxy",      new SetCanalSegmentProxy());
    
      /* Fields */
    addCommand("SelectField",               new SelectField());
    addCommand("UnselectField",             new UnselectField());
    addCommand("UnselectAllFields",         new UnselectAllFields());
    
    
    /* Reports */
      /* Environment */
      /* Rivers */
      /* Canal Systems */
    addCommand("ViewCanalSystemsReport",    new ViewCanalSystemsReport());
    addCommand("ViewCanalSystemDetails",    new ViewCanalSystemDetails());
    addCommand("GenerateHECOutput",         new GenerateHECOutput());
    
      /* Fields */
    addCommand("GetFieldReport",            new GetFieldReport());
    
  }
  
}
