package wm.model.agents;

import java.util.Vector;

import wm.model.WMAgent;
import wm.model.WMModel;
import abcm.math.SimDate;
import abcm.model.ABCMResult;
import abcm.model.agents.AbstractABCMActor;
import flowag.bio.Field;
import flowag.bio.Plant;

/**
 * Manages a single field by planting crops an opening or closing the gate
 * allowing water to be put onto the field.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class WMFieldManagerAgent extends WMAgent{

  /** Internal parser object */
  protected static WMFieldManagerAgentCommandParser parser;
  
  /** Record of plants that can be planted by all instances of this agent class */
  protected static Vector<Plant> plantsAvailable;
  
  /** Earliest date on which planting can occur */
  protected static int           startPlantingDate  = (new SimDate("01-MAR-SEASONS")).dayOfYear();
  /** Record of the shortest crop's duration */
  protected static int           shortestCrop       = Integer.MAX_VALUE;
  /** Record of the longest crop's duration */
  protected static int           longestCrop        = Integer.MIN_VALUE;
  
  
  /** Agent's field */
  protected Field                myField;
  /** Agent's current crop strategy, as a sequence of crops planted through the year (initially empty) */
  protected Vector<Plant>        myCropStrategy           = new Vector<Plant>();
  /** Index; which crop strategy is currently being used (moves along myCropStrategy through year) */
  protected int                  crop                     = 0;
  /** Accumulator tracking total harvest of all crops during the year */
  protected double               totalHarvestForThisYear  = 0;
  /** Record of last year's total harvest  */
  protected double               totalHarvestForLastYear  = 0;
  /** True if a plant is currently in the field, false otherwise */
  protected boolean              plantInField = false;
  /** Date the next harvest will occur (when current plant crop matures; calculated when planted) */
  protected int                  dayOfYearOfNextHarvest   = 0;
  /** Name of the canal system along which this agent manages crops */
  protected String               canalSysName;
  /** Name of the canal segment that feeds into this agent's field */
  protected String               canalSegName;
  /** True if the canal segment that feeds into this agent's field is open, false if closed */
  protected boolean              myGateIsOpen             = false;
  
  /** Threshhold below which agent will open gate if it is not already open */ 
  protected double               lowerWaterBound          = 12;
  /** Threshhold above which agent will close gate if it is open */
  protected double               upperWaterBound          = 50;
  
  
  /* THE FOLLOWING ARE NOT ACTUALLY IMPLEMENTED IN THE AGENT CODE, BUT COULD BE USED TO CONTROL THIS AGENT */
  
  /* Action Classes */
  /**
   * Causes the agent to plant a crop in its field.
   * 
   * Not actually used in current code; demonstration only.
   * Illustrates how an agent could be endowed with actions that
   * can be triggered by user-entered commands during interactive session.
   * 
   * @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  protected class Action_Plant extends Action{
    /**
     * @see abcm.model.agents.AbstractABCMActor.Action#execute(abcm.model.agents.AbstractABCMActor, java.lang.String[])
     */
    public ABCMResult execute(AbstractABCMActor actor, String[] command){
      if(command.length < 2) return new ABCMResult("Could not execute 'plant'; not enough arguments", false);
      ((WMFieldManagerAgent)actor).plant(command[1], 1);
      return new ABCMResult("Planted " + command[1] + " in field " + myField.name, true);
    }
  }
  
  /**
   * Causes the agent to open its gate.
   * 
   * Not actually used in current code; demonstration only.
   * Illustrates how an agent could be endowed with actions that
   * can be triggered by user-entered commands during interactive session.
   * 
   * @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  protected class Action_OpenGate extends Action{
    /**
     * @see abcm.model.agents.AbstractABCMActor.Action#execute(abcm.model.agents.AbstractABCMActor, java.lang.String[])
     */
    public ABCMResult execute(AbstractABCMActor actor, String[] command){
      ((WMFieldManagerAgent)actor).openGate();
      return new ABCMResult("Opened my gate.", true);
    }
  }
  
  /**
   * Causes the agent to close its gate
   * 
   * Not actually used in current code; demonstration only.
   * Illustrates how an agent could be endowed with actions that
   * can be triggered by user-entered commands during interactive session.
   *  
   * @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  protected class Action_CloseGate extends Action{
    /**
     * @see abcm.model.agents.AbstractABCMActor.Action#execute(abcm.model.agents.AbstractABCMActor, java.lang.String[])
     */
    public ABCMResult execute(AbstractABCMActor actor, String[] command){
      ((WMFieldManagerAgent)actor).closeGate();
      return new ABCMResult("Closed my gate.", true);
    }
  }
  
  /**
   * Causes the agent to harvest the crop in its field
   * 
   * Not actually used in current code; demonstration only.
   * Illustrates how an agent could be endowed with actions that
   * can be triggered by user-entered commands during interactive session.
   * 
   * @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  protected class Action_Harvest extends Action{
    /**
     * @see abcm.model.agents.AbstractABCMActor.Action#execute(abcm.model.agents.AbstractABCMActor, java.lang.String[])
     */
    public ABCMResult execute(AbstractABCMActor actor, String[] command){
      ((WMFieldManagerAgent)actor).harvest();
      return new ABCMResult("Harvested.", true);
    }
  }
  
  
  /* Parser class */
  /**
   * Agent's internal parser. Not currently used, but included as demonstration
   * of how agent could be directed by user-entered commands during an interactive
   * session.
   * 
   * @author John T. Murphy<br/>
   *  Global Institute of Sustainability,<br/>
   *  Arizona State University<br/>
   *  For license and disclaimers, see the Overview.
   */
  protected class WMFieldManagerAgentCommandParser extends WMAgentCommandParser{
    public WMFieldManagerAgentCommandParser(){
      addCommand("PLANT",      new Action_Plant());
      addCommand("OPEN_GATE",  new Action_OpenGate());
      addCommand("CLOSE_GATE", new Action_CloseGate());
    }
  }  
  
  /* END PARSER CODE */
  
  
  /* Static methods */
  protected static void clearPlantsAvailableList(){
	  if(plantsAvailable != null) plantsAvailable.clear();
	  plantsAvailable = null;
  }
  
  /* ID */
  /** 
   * Gets the prefix to an agent's automatically assigned ID
   * @see wm.model.WMAgent#getAgentIDPrefix()
   * @return ID prefix 
   */
  protected String getAgentIDPrefix(){  return "WM-FM-";  }
  
  /**
   * Gets the suffix to an agent's automatically assigned ID
   * @see wm.model.WMAgent#getAgentIDSuffix()
   * @return ID suffix
   */
  protected String getAgentIDSuffix(){  return "";        }
  
  /* Constructor */
  /**
   * Constructor
   * @param model model in which this agent exists
   * @param population population in which this agent exists
   * @param F Field this agent will manage
   * @param canalSystemName canal system in which this agent will operate
   * @param inflowCanalSegmentName name of the segment flowing into the field that the agent will manage
   */
  public WMFieldManagerAgent(WMModel model, WMPopulation population, Field F, String canalSystemName, String inflowCanalSegmentName) {
    super(model, population);
    myField = F;
    canalSysName = canalSystemName;
    canalSegName = inflowCanalSegmentName;
    
    // First agent created stores the list that they all have access to
    if(plantsAvailable == null) initializePlantData(myModel);
    
    closeGate();
  }
  
  
  
  /**
   * Initializes the plant data information for all members of this class
   * @param aModel the model to be mined for plant data
   */
  private static void initializePlantData(WMModel aModel){
    plantsAvailable = new Vector<Plant>();
    String[] plantNames = aModel.getListOfAvailablePlants();
    for(int i = 0; i < plantNames.length; i++){
      Plant p = aModel.getPlant(plantNames[i]);
      plantsAvailable.add(p);
      int lifeSpan = p.getLifeSpan();
      if(lifeSpan < shortestCrop) shortestCrop = lifeSpan;
      if(lifeSpan > longestCrop)  longestCrop  = lifeSpan;
    }
  }
  

  /**
   * @see abcm.model.agents.AbstractABCMAgent#out(java.lang.String)
   */
  public void out(String msg){
    System.out.println("AGENT " + getID() + " (" + myField.name + "):" + msg);
  }
  
  /**
   * @see wm.model.WMAgent#step()
   * Proceeds through a year. If the date is later than the earliest daya available for planting,
   * and there is not already a crop in its field, plants the next crop in its sequence (choosing
   * randomly and adding to the sequence if there is not a next crop known). During the crop's growth
   * it manages water by opening or closing a gate allowing water (if available) onto the field;
   * gate is opened when water falls below athreshhold and closed when it exceeds one. At maturation
   * the crop is harvested. No crop will be planted so late in the year that it cannot mature before the
   * end of the year. 
   */
  public void step(){
    super.step();
    if(myModel.getModelTime().dayOfYear() == 0){
      totalHarvestForLastYear = totalHarvestForThisYear;
      totalHarvestForThisYear = 0;
      crop = 0;
      return; // Take New Year's off!
    }
    int currentDayOfYear = myModel.getModelTime().dayOfYear();
    // First, check the field for anything that needs to be harvested
    if(plantInField){
      if(currentDayOfYear == this.dayOfYearOfNextHarvest){
//        double availableHarvest = M.processCommand("GetFieldHarvestableYield " + myField.name + " plant 0").value; 
//        ABCMResult harvest = M.processCommand("HarvestPlantInField " + myField.name + " " + myCropStrategy.get(crop).getName() + " 0");
        double availableHarvest = harvest();
        // Evaluate the harvest; if it is above a certain threshold, choose a shorter growing crop
        // If it is below a threshold, choose a longer one
        // Assumes shorter = higher crop factor/less water demand
        if(availableHarvest > .95){
          if(myCropStrategy.get(crop).getLifeSpan() == WMFieldManagerAgent.shortestCrop) 
            System.out.println("High HARVEST FOR SHORTEST CROP!! " + myField.name  + " HARVEST IS "+ availableHarvest);
          else{
            out("next year my crop " + crop + " is going to be better! switching up from " + myCropStrategy.get(crop).getLifeSpan() + " " + myField.name + " HARVEST IS " + availableHarvest);
            myCropStrategy.add(crop, getPlantWithGrowingSeasonShorterThan(myCropStrategy.get(crop).getLifeSpan()));
          }          
        }
        else if(availableHarvest < .80){
//          double oldlowerbound = lowerWaterBound;
//          double oldupperbound = upperWaterBound;
//          
//          lowerWaterBound = Math.max(lowerWaterBound-1, 0);
//          upperWaterBound++;
//          out("Adjusting water boundaries: " + oldlowerbound + "-" + oldupperbound + " TO: " + lowerWaterBound + "-" + upperWaterBound);
          
          if(myCropStrategy.get(crop).getLifeSpan() == longestCrop) System.out.println("LOW HARVEST FOR LONGEST CROP! " + availableHarvest + " HARVEST IS " + myField.name);
          else{
            out("next year my crop " + crop + " is going to be bad- switching              down from " + myCropStrategy.get(crop).getLifeSpan() + " " + myField.name + " HARVEST IS " + availableHarvest);
            myCropStrategy.add(crop, getPlantWithGrowingSeasonLongerThan(myCropStrategy.get(crop).getLifeSpan()));
          }
        }
        else out("my crop is fine: HARVEST IS " + availableHarvest);
        totalHarvestForThisYear += availableHarvest;
        crop++;
        plantInField = false;
        closeGate();
      }
      else{ // Plant is in field- manage water
        double waterAvailable = M.processCommand("GetFieldSoilWaterDepth  " + myField.name).value;
        if(waterAvailable < lowerWaterBound && (myGateIsOpen == false))  openGate();
        if(waterAvailable > upperWaterBound && (myGateIsOpen == true ))  closeGate();
      }
      return; // Done all that can be done if plant is in field
    }
    // If there aren't any, then check to see if it's OK to plant    
    if(myModel.getModelTime().dayOfYear() >= startPlantingDate){
      int remainingDays = 365 - myModel.getModelTime().dayOfYear();
      if(remainingDays < shortestCrop) return; // Not enough time to grow
      // You may not have a crop strategy for this number yet; try to get one
      if((myCropStrategy.size() <= crop)    ||
         (myCropStrategy.get(crop) == null) || 
         (myCropStrategy.get(crop).getLifeSpan() > remainingDays)){
          if(myCropStrategy.size() > crop) myCropStrategy.remove(crop);  
          myCropStrategy.add(crop, getPlantWithGrowingSeasonShorterThan(remainingDays));
      }
      if(myCropStrategy.get(crop) != null){
        plant(myCropStrategy.get(crop).getName(), 1);
        if(plantInField) dayOfYearOfNextHarvest = currentDayOfYear + myCropStrategy.get(crop).getLifeSpan();
        out("I just planted " + myCropStrategy.get(crop).getName() + " it's " + currentDayOfYear + " and I will harvest on " + dayOfYearOfNextHarvest);
      }
    }    
  }
  
  /* Actions this agent can take */
  
  /**
   * Opens the agent's gate via the model
   */
  private void openGate(){
//    ABCMResult result = 
        M.processCommand("OpenOutGate " + canalSysName + " " + canalSegName);
//    System.out.println("I'm agent " + getID() + ": " + result.text);
    myGateIsOpen = true;
  }
  
  /**
   * Closes the agent's gate via the model
   */
  private void closeGate(){
//    ABCMResult result = 
      M.processCommand("CloseOutGate " + canalSysName + " " + canalSegName);
//    System.out.println("I'm agent " + getID() + ": " + result.text);
    myGateIsOpen = false;
  }
  
  /**
   * Plants a crop via the model
   * @param plantName plant to be planted
   * @param density density at which it will be planted
   */
  private void plant(String plantName, double density){
    M.processCommand("UnselectAllFields");
    M.processCommand("SelectField " + myField.name);
    plantInField = M.processCommand("PlantInSelectedFields " + plantName + " " + density).success; // Always plant at density 1
    openGate();
    M.processCommand("UnselectAllFields"); // Play nice; clean up when done
  }
  
  /**
   * Harvests the crop in the agent's field via the model
   * @return the amount of the harvest
   */
  private double harvest(){
    return M.processCommand("HarvestPlantInField " + myField.name + " " + myCropStrategy.get(crop).getName() + " 0").value;
  }
  
  
  /* Internal */
  /**
   * Gets a plant with a growing season shorter than the specified value from the inventory of plants
   * @param val longest allowable maturation
   * @return a Plant selected from the possibilities meeting the specified criteria
   */
  private Plant getPlantWithGrowingSeasonShorterThan(int val){
    Vector<Plant> validPlants = new Vector<Plant>();
    for(Plant plant: plantsAvailable) if(plant.getLifeSpan() <= val) validPlants.add(plant);
    if(validPlants.size() == 0) return null;
    return validPlants.get((int)(Math.floor(this.random() * validPlants.size())));
  }
  
  /**
   * Gets a plant with a growing season longer than the specified value from the inventory of plants
   * @param val shortest allowable maturation
   * @return a Plant selected from the possibilities meeting the specified criteria
   */
  private Plant getPlantWithGrowingSeasonLongerThan(int val){
    Vector<Plant> validPlants = new Vector<Plant>();
    for(Plant plant: plantsAvailable) if(plant.getLifeSpan() >= val) validPlants.add(plant);
    if(validPlants.size() == 0) return null;
    return validPlants.get((int)(Math.floor(this.random() * validPlants.size())));
  }

  /**
   * Determine if this agent's gate is open. This is used only for internal recordkeepig and will be eliminated. 
   * @deprecated
   * @return true if the agent's gate is open, false if it is closed
   */
  public boolean gateIsOpen(){ return this.myGateIsOpen; }
  
  /**
   * Determine if this agent is managing a 'north' field (identified by having '_N' in its name). This is a special-purpose
   * method and will be eliminated.
   * @deprecated
   * @return true if the field that this agent is managing is 'north', false if not.
   */
  public boolean north(){ return this.myField.name.contains("_N"); }
  
}
