package wm.model.agents;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import wm.model.WMAgent;
import wm.model.WMModel;
import abcm.model.agents.AbstractABCModelPopulation;
import flowag.bio.Field;

/**
 * Container object for Agents in the WM Model.
 * 
 * A Population object should not be confused with the term 'population'
 * as it is used archaeologically. In archaeological parlance it is
 * reasonable to say that one 'population' was displaced by another;
 * within the HWM simulation, there is only ever one 'Population' object,
 * to which all Agents belong. If different groups of Agents are to be
 * managed, another subgrouping will be developed, but the 'Population'
 * will still be the generic container into which all instances of
 * Agent objects are put.
 * 
 * A Population can contain just one Agent that acts as a proxy for
 * all the people living in the HWM simulation; or it can contain an
 * instance for each household or each individual.
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class WMPopulation extends AbstractABCModelPopulation{

  /** Population will consist of one FieldManager agent per field */
  public static final int ONE_AGENT_PER_FIELD                 = 1;
  /** Population will consist of one FieldManager agent per field plus a canal system manager agent */
  public static final int ONE_AGENT_PER_FIELD_WITH_CS_MANAGER = 2;
  
  /* Static Variables */
  /** Internal: parses actor commands */
  protected static ActorCommandParser parser;
  
  /* Instance Variables */
  /** Model in which this object exists; this is a shortcut so that casting won't be necessary */
  protected WMModel m;
  
  
  /**
   * An extended version of the ActorCommandParser- currently does not
   * include any additional commands
   */
  protected class WMPopulationParser extends ActorCommandParser{
    // Add any population commands here
  }
  
  /**
   * Constructor that takes a Model object and a Code
   * Model object.
   * 
   * When called this will automatically create the Agents
   * in the manner specified by the Code Model object
   * @param model Model object to which this Population object belongs
   */
  public WMPopulation(WMModel model){
    super(model);
    m = model;
    if(parser == null) parser = new WMPopulationParser(); 
  }
    
  /**
   * Creates agents.
   * 
   * The number of Agents and other properties will be
   * determined by the Code Model object. For example,
   * some Code Models may specify that there should be
   * exactly one agent, which will represent a proxy
   * for an entire population. Others will specify that
   * there will be an agent for each household, or
   * each individual.
   */
  public void createAgents(){
    WMModel model = (WMModel)super.M;
    // There will be various methods here to create agents, 
    // depending on what's needed
    switch (codeModelOption("AGENT_CREATION")){
      case(ONE_AGENT_PER_FIELD):{
    	WMFieldManagerAgent.clearPlantsAvailableList();
        // Add one agent for each field
        HashSet fields = model.getAllFields();
        Iterator I = fields.iterator();
        while(I.hasNext()){
          Field F = (Field)I.next();
          Vector<String> namesOfInflowSegs = model.getNamesOfInflowCanalSegments(F.name);
          String inflowSegName = (namesOfInflowSegs.size() == 1 ? namesOfInflowSegs.get(0) : "UNKNOWN");
          Vector<String> namesOfCanalSystems = model.getNamesOfCanalSystems(F.name);
          String canalSysName = (namesOfCanalSystems.size() == 1 ? namesOfCanalSystems.get(0) : "UNKNOWN");
          addToAgentList(new WMFieldManagerAgent(model, this, F, canalSysName, inflowSegName));
        }
        break;
      }
      case(ONE_AGENT_PER_FIELD_WITH_CS_MANAGER):{
    	 WMFieldManagerAgent.clearPlantsAvailableList();
    	  
        // First the canal system manager(s)
        Hashtable canalSystems = model.getAllCanalSystems();
        Iterator CSI = canalSystems.keySet().iterator();
        while(CSI.hasNext())  addToAgentList(new WMCanalSystemManagerAgent(model, this, (String)CSI.next()));
        
        // Then the field manager agents
        HashSet fields = model.getAllFields();
        Iterator I = fields.iterator();
        while(I.hasNext()){
          Field F = (Field)I.next();
          Vector<String> namesOfInflowSegs = model.getNamesOfInflowCanalSegments(F.name);
          String inflowSegName = (namesOfInflowSegs.size() == 1 ? namesOfInflowSegs.get(0) : "UNKNOWN");
          Vector<String> namesOfCanalSystems = model.getNamesOfCanalSystems(F.name);
          String canalSysName = (namesOfCanalSystems.size() == 1 ? namesOfCanalSystems.get(0) : "UNKNOWN");
          addToAgentList(new WMFieldManagerAgent(model, this, F, canalSysName, inflowSegName));
        }
        break;
      }
      default:
        addToAgentList(new WMAgent(model, this));
        addToAgentList(new WMAgent(model, this));
    }
  }
  
  /**
   * Adds the agent to the list of agents, including to the hashtable of agent IDs
   * @param agent the agent to be added
   */
  private void addToAgentList(WMAgent agent){
    agentList.put(agent.getID(), agent);
  }

  /**
   * @see abcm.model.agents.AbstractABCMActor#getActorCommandParser()
   */
  protected ActorCommandParser getActorCommandParser(){  return parser;  }
  
  /**
   * @see abcm.model.agents.AbstractABCMActor#report()
   */
  public String report(){
    return "Count of agents: " + this.agentList.size();
  }

  /**
   * @see abcm.model.agents.AbstractABCMActor#getID()
   */
  public String getID(){ return "WatMan Population"; }

}
