package wm.sim;

import java.util.Enumeration;
import java.util.Iterator;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.sim.Configuration;
import flowag.bio.Plant;

/**
 * A WatMan Configuration is an ABCM Configuration with WatMan
 * specific data. 
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class WMConfiguration extends Configuration{

    /* Containers */
  /** Collection of Plants in this configuration         */ 
  private ABCMDeepCloneableHashtable plants       = new ABCMDeepCloneableHashtable();
  /** Collection of Plants in this configuration by name */ 
  private ABCMDeepCloneableHashtable plantsByName = new ABCMDeepCloneableHashtable();
  /** Collection of Chemicals in this configuration      */ 
  private ABCMDeepCloneableHashtable chemicals    = new ABCMDeepCloneableHashtable();
  
  /**
   * Basic constructor
   * @param name Name to be assigned to this configuration
   */
  public WMConfiguration(String name){
    super(name);
  }
  
  /**
   * Returns the plant object with the specified name.
   * 
   * Note that a plant object actually has two names. This
   * is to allow flexibility in creating Plant Configurations,
   * which may correspond to varieties.
   * 
   * Example:
   * 
   * The plant in the HWM database is 'maize'. One Plant
   * Configuration is created called 'maize-short-season', and
   * a second is created called 'maize-long-season.'
   * 
   * A single history is created that plants and harvests
   * 'maize'. We would like the flexibility of running this
   * with either maize-short-season or maize-long-season.
   * 
   * The Simulation Configuration for one version loads 
   * maize-short-season; it stores two references to it,
   * one in 'Plants', using the key 'maize-short-season',
   * and another in 'Plants-By-Name' where it uses the
   * key 'maize'. When the history executes,
   * it attempts to retrieve the plant 'maize', but finds
   * no entry in 'Plants' with that key. It then tries in
   * 'PlantsByName', where it finds the appropriate plant object.
   * 
   * This also allows a different scenario, wherein the history
   * is aware of the existence of versions, and plants both
   * 'maize-short-season' and 'maize-long-season' at different
   * times. 
   * 
   * In this situation, though, there would be a point of conflict
   * in the PlantsByName table, in that when the second Plant object
   * is added using the key 'maize' it will bump the first from the
   * Hashtable. If the user then tries to plant 'maize', the result
   * should be considered random- one of the plants will be returned,
   * and it will be the last one loaded (but this will depend on the
   * database and is therefore unpredictable).
   * 
   * @param name requested name
   * @return the plant object with the specified name
   */
  public Plant getPlant(String name){
    String retrieve = name.toUpperCase();
    if(plants.containsKey(retrieve))            return (Plant)plants.get(retrieve);
    else if(plantsByName.containsKey(retrieve)) return (Plant)plantsByName.get(retrieve);
    return new Plant(retrieve);
  }
  
  /**
   * Adds a plant to this configuration.
   * 
   * Note that both the plant name and the variety name are stored;
   * each is used as a separate key to the same plant object.
   * @param p the Plant to be added
   * @param varietyName the variety name to be applied to this plant
   * in this Configuration
   * @param name a name by which this plant will be referred in this Configuration
   */
  public void addPlant(Plant p, String varietyName, String name){
    plants.put(varietyName.toUpperCase(), p);
    plantsByName.put(name.toUpperCase(), p);
  }
  
  /**
   * Gets a report describing the Configuration
   * @return the Configuration report
   */
  public String getReport(){
    String retVal = "";
    retVal += "Plants: " + plants.size() + "\n";
    Enumeration e = plants.elements();
    while(e.hasMoreElements()){
      Plant p = (Plant)e.nextElement();
      retVal += p.getReport() + "\n";
    }
    return retVal;
  }
 
  /**
   * Gets a list of the plants available in this configuration
   * @return the request list
   */
  public String[] getListOfAvailablePlants(){
    String[] listOfPlants = new String[plants.size()];
    Iterator I = plants.keySet().iterator();
    int i = 0;
    while(I.hasNext()){
      listOfPlants[i] = (String)I.next();
      i++;
    }
    return listOfPlants;
  }
  
  /**
   * @see abcm.cloning.ABCMDeepCloneable#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    WMConfiguration newConfiguration = new WMConfiguration(this.getName());
    cloneStore.put(this, newConfiguration);
    newConfiguration.chemicals    = (ABCMDeepCloneableHashtable)chemicals.getDeepClone(cloneStore);
    newConfiguration.plants       = (ABCMDeepCloneableHashtable)plants.getDeepClone(cloneStore);
    newConfiguration.plantsByName = (ABCMDeepCloneableHashtable)plantsByName.getDeepClone(cloneStore);
    return newConfiguration;
  }
  
}
