package wm.space;

import abcm.cloning.ABCMCloneStore;
import abcm.cloning.ABCMDeepCloneable;
import abcm.cloning.ABCMDeepCloneableHashtable;
import abcm.graphics.SVGAssist;
import abcm.graphics.SVGRepresentable;
import abcm.space.Space;
import flowag.bio.Field;
import flowag.channels.CanalSegment;
import flowag.channels.CanalSystem;
import flowag.channels.River;

/**
 * A WMSpace object extends the ABCM Space object and
 * enables the simulation to contain canal systems, 
 * rivers, and fields. 
 * 
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class WMSpace extends Space implements SVGRepresentable, ABCMDeepCloneable {

  // Containers
  /** Collection of Canal Systems */ 
  private ABCMDeepCloneableHashtable canalSystems = new ABCMDeepCloneableHashtable();
  /** Collection of Fields        */ 
  private ABCMDeepCloneableHashtable fields       = new ABCMDeepCloneableHashtable();
 
  /**
   * Constructor that sets orientation according to defaults
   */
  public WMSpace() {super();  }

  /**
   * Constructor that allows orientation to be specified
   * 
   * @param eastIsPos
   *          true if x coordinates increase in eastward direction, false
   *          otherwise
   * @param northIsPos
   *          true if y coordinates increase in northern direction, false
   *          otherwise
   */
  public WMSpace(boolean eastIsPos, boolean northIsPos) {
    super(eastIsPos, northIsPos);
  }

  /* Adding elements */
  
  /**
   * Adds a river to this space
   * 
   * @param riverName name of the river being entered
   * @param R         the River object to be entered
   */
  public void addRiver(String riverName, River R) {
    addSpatialElement("Rivers", riverName, R);
//    if (rivers.containsKey(riverName))
//      return;
//    rivers.put(riverName, R);
  }

  /**
   * Adds a Canal System to this space object
   * 
   * @param systemName Name of the new Canal System
   * @param cs         Canal System to be added
   * @return A message indicating the success or failure of the operation
   */
  public String addCanalSystem(String systemName, CanalSystem cs) {
    if (canalSystems.containsKey(systemName)) return "Could not add Canal System.";
    canalSystems.put(systemName, cs);
    this.addSpatialElement("CanalSystems", systemName, cs);
    return "Successfully added Canal System.";
  }

  /**
   * Adds a field to this space object.
   * 
   * @param fieldName  Name of the field to be added
   * @param f          Field object being added
   * @return true if successful, false otherwise
   */
  public boolean addField(String fieldName, Field f) {
    return addSpatialElement("Fields", fieldName, f);
  }

  /* Remove elements */
  
  /**
   * Removes a Field from this Space object.
   * @param fieldName name of the Field to be removed.
   * @return true if the Field was in the Space object and
   * was successfully removed, false otherwise
   */ 
  public boolean dropField(String fieldName){
    return removeSpatialElement("Fields", fieldName);
  }
  
  /* Display */

  
  /**
   * Given a canal segment, returns the volume of dirt that would be required to
   * build that canal segment, in terms of the cubic meters of dirt that would
   * have to be moved.
   * 
   * For the moment this simply returns the construction volume of the canal
   * segment, as if it were being built on a landscape that was flat except for
   * a slope that exactly matches the bed of the proposed canal segment. (That
   * is, the volume of dirt that would have to be moved is exactly equal to the
   * 'construction volume' of the canal segment.) Someday this should be
   * revised, so that the amount of work is calculated according to the real
   * landscape, which may be highly obstructive (i.e. there may be a hill
   * between the endpoints) or may already be shaped in a way that is conducive
   * to the construction of the canal.
   * 
   * Assumes that the canal will be built so that all points above the actual
   * construction are removed- thus the walls above the constructed walls go up
   * vertically. Note that this is different from the way the calculations work;
   * in the flow calculations the walls are assumed to go upward and outward
   * infinitely, continuing the same slope as the side walls (for a trapezoidal
   * segment).
   * 
   * @param cs
   *          the canal segment to be built; usually this will be an instance
   *          that is not actually associated with a canal system, but is being
   *          used to test the proposal that a canal segment with its
   *          characteristics be built
   * @return the amount of work required to build the canal, in
   */
  public double getCubicMetersMovedToBuildCanalSegment(CanalSegment cs) {
	return 0;
//    return cs.getConstructedVolume();
  }
  
  /**
   * @see abcm.space.Space#getDeepClone(abcm.cloning.ABCMCloneStore)
   */
  public ABCMDeepCloneable getDeepClone(ABCMCloneStore cloneStore){
    WMSpace newSpace = new WMSpace();
    cloneStore.put(this, newSpace);
    super.cloneCoreSpaceObjects(newSpace, cloneStore);
    newSpace.canalSystems = (ABCMDeepCloneableHashtable)canalSystems.getDeepClone(cloneStore);
    newSpace.fields       = (ABCMDeepCloneableHashtable)fields.getDeepClone(cloneStore);
    return newSpace;
  }

  /* OBSOLETE */
  
///**
// * Adds this space object to an SVG document.
// * @param svg the SVGAssist document being used to create the
// * svg document
// */
//public void addToSVG(SVGAssist svg){
//  super.addToSVG(svg);
//  
//  // Add each collection- OBSOLETE
////  svg.addG("Rivers","",SVGAssist.getArrayFromHashtable(rivers));
////  svg.addG("Canal Systems","",SVGAssist.getArrayFromHashtable(canalSystems));
////  svg.addG("Fields","",SVGAssist.getArrayFromHashtable(fields));
//  
//}
  
}
