package abcm.model.tools;

import java.io.StringWriter;

import abcm.db.managers.CommandsManager;
import abcm.math.SimDate;
import abcm.model.AbstractABCMParser;
import abcm.model.AbstractABCModelEnvironment;
import abcm.utilities.QuotedLineParser;
import abcm.utilities.ScriptReader;

/**
 * 'Decompiles' a narrative from a format in which the commands
 * are specified by instruction number into one in which the English
 * versions of the commands are used.
 *  
 *  @author John T. Murphy<br/>
 *  Global Institute of Sustainability,<br/>
 *  Arizona State University<br/>
 *  For license and disclaimers, see the Overview.
 */
public class NarrativeDecompiler {

  /**
   * Constructor; private scope prevents this class from being instantiated or extended.
   */
  private NarrativeDecompiler(){}
  
  /**
   * Given the name of a file containing a compiled narrative, returns
   * a StringWriter containing the text of the narrative decompiled into English.
   * @param inputFileName name of input file
   * @param env model environment (with access to db gateway)
   * @return StringWriter containing text of decompiled narrative
   */
  public static StringWriter decompileNarrative(String inputFileName, AbstractABCModelEnvironment env){
    StringWriter ret = new StringWriter();
    
    CommandsManager cm = new CommandsManager(env.getDBGateway());
    ScriptReader sr = new ScriptReader(inputFileName);
    
    int warningCount = 0;
    long line = 1;
    double currentDate = 0;
    long orderCount = 1;
    while(!sr.EOF()){
      String nextLine = sr.getNextLine();
      if(nextLine != null){
        if(!nextLine.startsWith("**") && nextLine.length() > 0){
          nextLine = nextLine.replaceAll("\t"," ");
          String[] inputLine = QuotedLineParser.parseQuoted(nextLine, AbstractABCMParser.MAX_ARGS + 1);
          String onDate = inputLine[0];
          double onDateValue = Double.parseDouble(onDate);
          SimDate newDate = new SimDate(onDateValue);
          if(newDate.value() != currentDate) orderCount = 1;
          else                               orderCount++;
          currentDate = newDate.value();
          ret.write(newDate.getString() + "\t");
          String com = cm.getCommand(Integer.parseInt(inputLine[2]));
          if(com.compareTo("UNKNOWN") == 0) {
            System.err.println("WARNING: Unknown command in line: " + line);
            warningCount++;
          }
          ret.write(com);
          int i = 3;
          while((i < inputLine.length) && (inputLine[i] != null) && (inputLine[i].length() > 0)){
            ret.write("\t" + inputLine[i]);
            i++;
          }
          ret.write("\n");
        }
        line++;
      }
    }
    if(warningCount > 0) System.err.println(warningCount + " warnings.");
    
    return ret;
  }
}
